# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
# pylint: disable=simplifiable-if-statement
u"""
Interfaces for the NAG Mark 27.0 `ieee` Chapter.

``ieee`` - IEEE Arithmetic

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module provides functions to handle various aspects of IEEE
floating-point arithmetic behaviour.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07intro.html
"""

# NAG Copyright 2017-2019.

import sys as _sys

from ..base import utils

def is_finite(x):
    u"""
    Determines whether its argument has a finite value.

    ``is_finite`` determines whether a floating-point number is finite.

    For full information please refer to the NAG Library document for
    x07aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07aaf.html

    Parameters
    ----------
    x : float
        The number whose status is to be determined.

    Returns
    -------
    finite : bool
        ifinite = 1 if and only if `x` is finite, and ifinite = 0
        otherwise.
    """

    excs = []
    fun_name = 'naginterfaces.base.ieee.is_finite'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x07aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_INT, # Output, ifinite.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _ifinite = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _ifinite,
    )
    utils._handle_cb_exception(excs)
    _ifinite_py = utils._EngineIntScalarType.to_py(_ifinite)
    if _ifinite_py == 0:
        _finite_py = False
    elif _ifinite_py == 1:
        _finite_py = True
    else:
        raise utils.NagException(
            {'fun_name': fun_name},
            'Unexpected value for internal ifinite.',
        )
    return_data = _finite_py
    return return_data

def is_nan(x):
    u"""
    Determines whether its argument is a NaN (Not A Number).

    ``is_nan`` determines whether a floating-point number is a NaN (Not
    A Number).

    For full information please refer to the NAG Library document for
    x07ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07abf.html

    Parameters
    ----------
    x : float
        The number whose status is to be determined.

    Returns
    -------
    isnan : bool
        iisnan = 1 if and only if `x` is a NaN, and isnan = 0 otherwise.
    """

    excs = []
    fun_name = 'naginterfaces.base.ieee.is_nan'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x07abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_INT, # Output, iisnan.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _iisnan = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _iisnan,
    )
    utils._handle_cb_exception(excs)
    _iisnan_py = utils._EngineIntScalarType.to_py(_iisnan)
    if _iisnan_py == 0:
        _isnan_py = False
    elif _iisnan_py == 1:
        _isnan_py = True
    else:
        raise utils.NagException(
            {'fun_name': fun_name},
            'Unexpected value for internal iisnan.',
        )
    return_data = _isnan_py
    return return_data

def create_infinity(isign):
    u"""
    Creates a signed infinite value.

    ``create_infinity`` creates a signed infinite value.

    For full information please refer to the NAG Library document for
    x07ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07baf.html

    Parameters
    ----------
    isign : int
        Determines the sign of the infinity to be created.

        If `isign` is greater than or equal to 0, a positive infinity is
        returned, otherwise a negative infinity is returned.

    Returns
    -------
    x : float
        The required infinite value.
    """

    excs = []
    fun_name = 'naginterfaces.base.ieee.create_infinity'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x07baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, isign.
        utils._BYREF_FLOAT64, # Output, x.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _isign = utils._EngineIntScalarType(
        isign, {'fun_name': fun_name, 'entity': 'isign'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _isign,
        _x,
    )
    utils._handle_cb_exception(excs)
    _x_py = utils._EngineFloat64ScalarType.to_py(_x)
    return_data = _x_py
    return return_data

def create_nan(quiet):
    u"""
    Creates a NaN (Not A Number).

    ``create_nan`` creates a NaN (Not A Number).

    For full information please refer to the NAG Library document for
    x07bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07bbf.html

    Parameters
    ----------
    quiet : int
        Determines whether a quiet or a signalling NaN is to be created.
        If `quiet` = 1, the returned NaN is quiet, otherwise it is
        signalling. See reference IEEE Standard for Floating-Point
        Arithmetic for the distinction between the two kinds.

    Returns
    -------
    x : float
        The required NaN value.
    """

    excs = []
    fun_name = 'naginterfaces.base.ieee.create_nan'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x07bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, quiet.
        utils._BYREF_FLOAT64, # Output, x.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _quiet = utils._EngineIntScalarType(
        quiet, {'fun_name': fun_name, 'entity': 'quiet'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _quiet,
        _x,
    )
    utils._handle_cb_exception(excs)
    _x_py = utils._EngineFloat64ScalarType.to_py(_x)
    return_data = _x_py
    return return_data

def get_exception_mode(exceptionmode):
    u"""
    Gets current behaviour of floating-point exceptions.

    ``get_exception_mode`` gets the current IEEE exception halting mode.

    For full information please refer to the NAG Library document for
    x07ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07caf.html

    Parameters
    ----------
    exceptionmode : int, shape (3), modified in place
        On exit: each of the three elements of `exceptionmode` is set to
        1 if the corresponding condition will raise an exception, and is
        set to 0 otherwise. `exceptionmode`[0] concerns floating-point
        overflow, `exceptionmode`[1] concerns floating-point division by
        zero, and `exceptionmode`[2] concerns floating-point invalid
        operation.
    """

    excs = []
    fun_name = 'naginterfaces.base.ieee.get_exception_mode'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x07caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Output, exceptionmode[3].
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _exceptionmode_cls = utils._EngineIntArrayType(
        exceptionmode, {'fun_name': fun_name, 'entity': 'exceptionmode'},
        exp_shape=(3,),
        check_contents=False,
    )
    _exceptionmode = _exceptionmode_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _exceptionmode,
    )
    utils._handle_cb_exception(excs)
    _exceptionmode_cls.refresh_py(_exceptionmode)
    return_data = None
    return return_data

def set_exception_mode(exceptionmode):
    u"""
    Sets behaviour of floating-point exceptions.

    ``set_exception_mode`` sets the current IEEE exception halting mode.

    For full information please refer to the NAG Library document for
    x07cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07cbf.html

    Parameters
    ----------
    exceptionmode : int, shape (3)
        Each of the three elements of `exceptionmode` must contain any
        nonzero value if the corresponding condition should raise an
        exception, and contain 0 otherwise. `exceptionmode`[0] concerns
        floating-point overflow, `exceptionmode`[1] concerns
        floating-point division by zero, and `exceptionmode`[2] concerns
        floating-point invalid operation.
    """

    excs = []
    fun_name = 'naginterfaces.base.ieee.set_exception_mode'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x07cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, exceptionmode[3].
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _exceptionmode = utils._EngineIntArrayType(
        exceptionmode, {'fun_name': fun_name, 'entity': 'exceptionmode'},
        exp_shape=(3,),
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _exceptionmode,
    )
    utils._handle_cb_exception(excs)
    return_data = None
    return return_data
