# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `inteq` Chapter.

``inteq`` - Integral Equations

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the numerical solution of integral
equations.
Provision will be made for most of the standard types of equation (see
[Background to the Problems]).
The following are, however, specifically excluded:

- Equations arising in the solution of partial differential equations by
  integral equation methods. In cases where the prime purpose of an
  algorithm is the solution of a partial differential equation it will
  normally be included in submodule ``pde``.

- Calculation of inverse integral transforms. This problem falls within
  the scope of submodule ``sum``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05intro.html
"""

# NAG Copyright 2017-2019.

import sys as _sys

from ..base import utils

def fredholm2_split(lamda, a, b, k1, k2, g, f, c, n, ind, data=None):
    u"""
    Linear nonsingular Fredholm integral equation, second kind, split
    kernel.

    ``fredholm2_split`` solves a linear, nonsingular Fredholm equation
    of the second kind with a split kernel.

    For full information please refer to the NAG Library document for
    d05aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05aaf.html

    Parameters
    ----------
    lamda : float
        The value of the parameter lambda of the integral equation.

    a : float
        a, the lower limit of integration.

    b : float
        b, the upper limit of integration.

    k1 : callable retval = k1(x, s, data=None)
        `k1` must evaluate the kernel k(x,s) = k_1(x,s) of the integral
        equation for a <= s <= x.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x and s at which k_1(x,s) is to be evaluated.

        s : float
            The values of x and s at which k_1(x,s) is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the kernel k(x,s) = k_1(x,s) evaluated at `x`
            and `s`.

    k2 : callable retval = k2(x, s, data=None)
        `k2` must evaluate the kernel k(x,s) = k_2(x,s) of the integral
        equation for x < s <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x and s at which k_2(x,s) is to be evaluated.

        s : float
            The values of x and s at which k_2(x,s) is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the kernel k(x,s) = k_2(x,s) evaluated at `x`
            and `s`.

    g : callable retval = g(x, data=None)
        `g` must evaluate the function g(x) for a <= x <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x at which g(x) is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(x) evaluated at `x`.

    f : float, shape (`n`), modified in place
        On exit: the approximate values f_i, for i = 1,2,...,`n`, of
        f(x) evaluated at the first `n` of m+1 Chebyshev points x_i,
        (see [Description]).

        If `ind` = 0 or 3, m = `n`-1.

        If `ind` = 1, m = 2*`n`.

        If `ind` = 2, m = 2*`n`-1.

    c : float, shape (`n`), modified in place
        On exit: the coefficients c_i, for i = 1,2,...,`n`, of the
        Chebyshev series approximation to f(x).

        If `ind` = 1 this series contains polynomials of odd order only
        and if `ind` = 2 the series contains even order polynomials
        only.

    n : int
        The number of terms in the Chebyshev series required to
        approximate f(x).

    ind : int
        Determines the forms of the kernel, k(x,s), and the function
        g(x).

        `ind` = 0

          k(x,s) is not centro-symmetric (or no account is to be taken
          of centro-symmetry).

        `ind` = 1

          k(x,s) is centro-symmetric and g(x) is odd.

        `ind` = 2

          k(x,s) is centro-symmetric and g(x) is even.

        `ind` = 3

          k(x,s) is centro-symmetric but g(x) is neither odd nor even.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > `a`.

        (`errno` 2)
            A failure has occurred due to proximity of an eigenvalue.
    """

    excs = []
    fun_name = 'naginterfaces.base.inteq.fredholm2_split'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_k1_name = fun_name + ':k1'

    k1h = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_k1h(_k1, _k1sub, en_data, x, s, retval, _iuser, _ruser):
        """Helper for k1."""
        try:
            _x_k1h = utils._EngineFloat64ScalarType.to_py(x.contents)
            _s_k1h = utils._EngineFloat64ScalarType.to_py(s.contents)
            if data is not None:
                _retval_k1h = k1(
                    _x_k1h,
                    _s_k1h,
                    data,
                )
            else:
                _retval_k1h = k1(
                    _x_k1h,
                    _s_k1h,
                )
            utils._EngineFloat64ScalarType(
                _retval_k1h, {'fun_name': cb_k1_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_k2_name = fun_name + ':k2'

    k2h = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_k2h(_k2, _k2sub, en_data, x, s, retval, _iuser, _ruser):
        """Helper for k2."""
        try:
            _x_k2h = utils._EngineFloat64ScalarType.to_py(x.contents)
            _s_k2h = utils._EngineFloat64ScalarType.to_py(s.contents)
            if data is not None:
                _retval_k2h = k2(
                    _x_k2h,
                    _s_k2h,
                    data,
                )
            else:
                _retval_k2h = k2(
                    _x_k2h,
                    _s_k2h,
                )
            utils._EngineFloat64ScalarType(
                _retval_k2h, {'fun_name': cb_k2_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_g_name = fun_name + ':g'

    gh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_gh(_g, _gsub, en_data, x, retval, _iuser, _ruser):
        """Helper for g."""
        try:
            _x_gh = utils._EngineFloat64ScalarType.to_py(x.contents)
            if data is not None:
                _retval_gh = g(
                    _x_gh,
                    data,
                )
            else:
                _retval_gh = g(
                    _x_gh,
                )
            utils._EngineFloat64ScalarType(
                _retval_gh, {'fun_name': cb_g_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('d05aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, lamda.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        k1h,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        k2h,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        gh,
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineFloat64ArrayType.c_type, # Output, c[n].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ind.
        utils._EngineFloat64ArrayType.c_type, # Workspace, w1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, w2.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wd.
        utils._BYREF_INT, # Input, ldw1.
        utils._BYREF_INT, # Input, ldw2.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _lamda = utils._EngineFloat64ScalarType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    utils._check_callable(k1, {'fun_name': fun_name, 'entity': 'k1'})
    _k1 = utils._UNUSED_CALLBACK
    _k1sub = utils._UNUSED_CALLBACK
    _k1h = utils._cast_callback(
        k1h, cb_k1h,
        {'fun_name': fun_name, 'entity': 'k1'},
    )
    utils._check_callable(k2, {'fun_name': fun_name, 'entity': 'k2'})
    _k2 = utils._UNUSED_CALLBACK
    _k2sub = utils._UNUSED_CALLBACK
    _k2h = utils._cast_callback(
        k2h, cb_k2h,
        {'fun_name': fun_name, 'entity': 'k2'},
    )
    utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'})
    _g = utils._UNUSED_CALLBACK
    _gsub = utils._UNUSED_CALLBACK
    _gh = utils._cast_callback(
        gh, cb_gh,
        {'fun_name': fun_name, 'entity': 'g'},
    )
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _ind = utils._EngineIntScalarType(
        ind, {'fun_name': fun_name, 'entity': 'ind'},
    ).to_ctypes()
    _w1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _w2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _wd = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _ldw1 = utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_ldw1'},
    ).to_ctypes()
    _ldw2 = utils._EngineIntScalarType.c_type(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _lamda,
        _a,
        _b,
        _k1,
        _k1sub,
        _k1h,
        _k2,
        _k2sub,
        _k2h,
        _g,
        _gsub,
        _gh,
        _f,
        _c,
        _n,
        _ind,
        _w1,
        _w2,
        _wd,
        _ldw1,
        _ldw2,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s} and b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: b > a.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A failure has occurred due to proximity of an eigenvalue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_cls.refresh_py(_f)
        _c_cls.refresh_py(_c)
        return_data = None
    _err_class.manage()
    return return_data

def fredholm2_smooth(k, g, lamda, a, b, odorev, ev, n, f, c, data=None):
    u"""
    Linear nonsingular Fredholm integral equation, second kind, smooth
    kernel.

    ``fredholm2_smooth`` solves any linear nonsingular Fredholm integral
    equation of the second kind with a smooth kernel.

    For full information please refer to the NAG Library document for
    d05ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05abf.html

    Parameters
    ----------
    k : callable retval = k(x, s, data=None)
        `k` must compute the value of the kernel k(x,s) of the integral
        equation over the square a <= x <= b, a <= s <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x and s at which k(x,s) is to be calculated.

        s : float
            The values of x and s at which k(x,s) is to be calculated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the kernel k(x,s) evaluated at `x` and `s`.

    g : callable retval = g(x, data=None)
        `g` must compute the value of the function g(x) of the integral
        equation in the interval a <= x <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The value of x at which g(x) is to be calculated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(x) evaluated at `x`.

    lamda : float
        The value of the parameter lambda of the integral equation.

    a : float
        a, the lower limit of integration.

    b : float
        b, the upper limit of integration.

    odorev : bool
        Indicates whether it is known that the solution f(x) is odd or
        even about the mid-point of the range of integration. If
        `odorev` is True then an odd or even solution is sought
        depending upon the value of `ev`.

    ev : bool
        Is ignored if `odorev` is False. Otherwise, if `ev` is True, an
        even solution is sought, whilst if `ev` is False, an odd
        solution is sought.

    n : int
        The number of terms in the Chebyshev series which approximates
        the solution f(x).

    f : float, shape (`n`), modified in place
        On exit: the approximate values f_i, for i = 1,2,...,`n`, of the
        function f(x) at the first `n` of m+1 Chebyshev points (see
        [Description]), where

        [table omitted]

    c : float, shape (`n`), modified in place
        On exit: the coefficients c_i, for i = 1,2,...,`n`, of the
        Chebyshev series approximation to f(x). When `odorev` is True,
        this series contains polynomials of even order only or of odd
        order only, according to `ev` being True or False respectively.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > `a`.

        (`errno` 2)
            A failure has occurred due to proximity of an eigenvalue.
    """

    excs = []
    fun_name = 'naginterfaces.base.inteq.fredholm2_smooth'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_k_name = fun_name + ':k'

    kh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_kh(_k, _ksub, en_data, x, s, retval, _iuser, _ruser):
        """Helper for k."""
        try:
            _x_kh = utils._EngineFloat64ScalarType.to_py(x.contents)
            _s_kh = utils._EngineFloat64ScalarType.to_py(s.contents)
            if data is not None:
                _retval_kh = k(
                    _x_kh,
                    _s_kh,
                    data,
                )
            else:
                _retval_kh = k(
                    _x_kh,
                    _s_kh,
                )
            utils._EngineFloat64ScalarType(
                _retval_kh, {'fun_name': cb_k_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_g_name = fun_name + ':g'

    gh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_gh(_g, _gsub, en_data, x, retval, _iuser, _ruser):
        """Helper for g."""
        try:
            _x_gh = utils._EngineFloat64ScalarType.to_py(x.contents)
            if data is not None:
                _retval_gh = g(
                    _x_gh,
                    data,
                )
            else:
                _retval_gh = g(
                    _x_gh,
                )
            utils._EngineFloat64ScalarType(
                _retval_gh, {'fun_name': cb_g_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('d05abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        kh,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        gh,
        utils._BYREF_FLOAT64, # Input, lamda.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_LOGICAL, # Input, odorev.
        utils._BYREF_LOGICAL, # Input, ev.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Workspace, cm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, f1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._BYREF_INT, # Input, ldcm.
        utils._BYREF_INT, # Input, nt2p1.
        utils._EngineFloat64ArrayType.c_type, # Output, f[n].
        utils._EngineFloat64ArrayType.c_type, # Output, c[n].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(k, {'fun_name': fun_name, 'entity': 'k'})
    _k = utils._UNUSED_CALLBACK
    _ksub = utils._UNUSED_CALLBACK
    _kh = utils._cast_callback(
        kh, cb_kh,
        {'fun_name': fun_name, 'entity': 'k'},
    )
    utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'})
    _g = utils._UNUSED_CALLBACK
    _gsub = utils._UNUSED_CALLBACK
    _gh = utils._cast_callback(
        gh, cb_gh,
        {'fun_name': fun_name, 'entity': 'g'},
    )
    _lamda = utils._EngineFloat64ScalarType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _odorev = utils._EngineLogicalScalarType(
        odorev, {'fun_name': fun_name, 'entity': 'odorev'},
    ).to_ctypes()
    _ev = utils._EngineLogicalScalarType(
        ev, {'fun_name': fun_name, 'entity': 'ev'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _cm = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _f1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _ldcm = utils._EngineIntScalarType.c_type(0)
    _nt2p1 = utils._EngineIntScalarType.c_type(0)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _f_cls = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _f = _f_cls.to_ctypes(can_reference_contents=False)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _k,
        _ksub,
        _kh,
        _g,
        _gsub,
        _gh,
        _lamda,
        _a,
        _b,
        _odorev,
        _ev,
        _n,
        _cm,
        _f1,
        _wk,
        _ldcm,
        _nt2p1,
        _f,
        _c,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s} and b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: b > a.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A failure has occurred due to proximity of an eigenvalue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_cls.refresh_py(_f)
        _c_cls.refresh_py(_c)
        return_data = None
    _err_class.manage()
    return return_data

def volterra2(ck, cg, cf, method, iorder, alim, tlim, yn, errest, nmesh, tol, thresh, work, lwk, data=None):
    u"""
    Nonlinear Volterra convolution equation, second kind.

    ``volterra2`` computes the solution of a nonlinear convolution
    Volterra integral equation of the second kind using a reducible
    linear multi-step method.

    For full information please refer to the NAG Library document for
    d05ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05baf.html

    Parameters
    ----------
    ck : callable retval = ck(t, data=None)
        `ck` must evaluate the kernel k(t) of the integral equation (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of k at the specified point.

    cg : callable retval = cg(s, y, data=None)
        `cg` must evaluate the function g(s,y(s)) in (1).

        Parameters
        ~~~~~~~~~~
        s : float
            s, the value of the independent variable.

        y : float
            The value of the solution y at the point `s`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g at the specified points.

    cf : callable retval = cf(t, data=None)
        `cf` must evaluate the function f(t) in (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f at the specified point.

    method : str, length 1
        The type of method which you wish to employ.

        `method` = 'A'

          For Adams' type formulae.

        `method` = 'B'

          For backward differentiation formulae.

    iorder : int
        The order of the method to be used.

    alim : float
        a, the lower limit of the integration interval.

    tlim : float
        The final point of the integration interval, T.

    yn : float, shape (`nmesh`), modified in place
        On exit: `yn`[i-1] contains the most recent approximation of the
        true solution y(t) at the specified point t = `alim`+i*H, for i
        = 1,2,...,`nmesh`, where H = (`tlim`-`alim`)/`nmesh`.

    errest : float, shape (`nmesh`), modified in place
        On exit: `errest`[i-1] contains the most recent approximation of
        the relative error in the computed solution at the point t =
        `alim`+i*H, for i = 1,2,...,`nmesh`, where H =
        (`tlim`-`alim`)/`nmesh`.

    nmesh : int
        The number of equidistant points at which the solution is
        sought.

    tol : float
        The relative accuracy required in the computed values of the
        solution.

    thresh : float
        The threshold value for use in the evaluation of the estimated
        relative errors. For two successive meshes the following
        condition must hold at each point of the coarser mesh

            |Y_1-Y_2|/max(|Y_1|, |Y_2|, |`thresh`|) <= `tol`,

        where Y_1 is the computed solution on the coarser mesh and Y_2
        is the computed solution at the corresponding point in the finer
        mesh. If this condition is not satisfied then the step size is
        halved and the solution is recomputed.

        Note: `thresh` can be used to effect a relative, absolute or
        mixed error test. If `thresh` = 0.0 then pure relative error is
        measured and, if the computed solution is small and `thresh` =
        1.0, absolute error is measured.

    work : float, shape (`lwk`), modified in place
        On exit: if `errno` = 5 or 6, `work`[0] contains the size of
        `lwk` required for the algorithm to proceed further.

    lwk : int
        The dimension of the array `work`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `method` = 'A' and `iorder` = 2.

            Constraint: if `method` = 'A', 3 <= `iorder` <= 6.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: sqrt(machine precision) <= `tol` <= 1.0.

        (`errno` 1)
            On entry, `alim` = *<value>* and `tlim` = *<value>*.

            Constraint: `tlim` > `alim`.

        (`errno` 1)
            On entry, `alim` = *<value>*.

            Constraint: `alim` >= 0.0.

        (`errno` 1)
            On entry, `method` = 'B' and `iorder` = 6.

            Constraint: if `method` = 'B', 2 <= `iorder` <= 5.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 2 <= `iorder` <= 6.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'A' or 'B'.

        (`errno` 2)
            On entry, `method` = 'A', `iorder` = *<value>* and `nmesh` =
            *<value>*.

            Constraint: if `method` = 'A', `nmesh` >= `iorder`-1.

        (`errno` 2)
            On entry, `method` = 'B', `iorder` = *<value>* and `nmesh` =
            *<value>*.

            Constraint: if `method` = 'B', `nmesh` >= `iorder`.

        (`errno` 3)
            On entry, `lwk` = *<value>*.

            Constraint: `lwk` >= 10*`nmesh`+6; that is, *<value>*.

        (`errno` 4)
            The solution is not converging. See [Further Comments].

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The workspace which has been supplied is too small for the
            required accuracy. The number of extrapolations, so far, is
            *<value>*. If you require one more extrapolation extend the
            size of workspace to: `lwk` = *<value>*.

        (`errno` 6)
            The workspace which has been supplied is too small for the
            required accuracy. The number of extrapolations, so far, is
            *<value>*. If you require one more extrapolation extend the
            size of workspace to: `lwk` = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.inteq.volterra2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_ck_name = fun_name + ':ck'

    ckh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_ckh(_ck, _cksub, en_data, t, retval, _iuser, _ruser):
        """Helper for ck."""
        try:
            _t_ckh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_ckh = ck(
                    _t_ckh,
                    data,
                )
            else:
                _retval_ckh = ck(
                    _t_ckh,
                )
            utils._EngineFloat64ScalarType(
                _retval_ckh, {'fun_name': cb_ck_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_cg_name = fun_name + ':cg'

    cgh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cgh(_cg, _cgsub, en_data, s, y, retval, _iuser, _ruser):
        """Helper for cg."""
        try:
            _s_cgh = utils._EngineFloat64ScalarType.to_py(s.contents)
            _y_cgh = utils._EngineFloat64ScalarType.to_py(y.contents)
            if data is not None:
                _retval_cgh = cg(
                    _s_cgh,
                    _y_cgh,
                    data,
                )
            else:
                _retval_cgh = cg(
                    _s_cgh,
                    _y_cgh,
                )
            utils._EngineFloat64ScalarType(
                _retval_cgh, {'fun_name': cb_cg_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_cf_name = fun_name + ':cf'

    cfh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cfh(_cf, _cfsub, en_data, t, retval, _iuser, _ruser):
        """Helper for cf."""
        try:
            _t_cfh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_cfh = cf(
                    _t_cfh,
                    data,
                )
            else:
                _retval_cfh = cf(
                    _t_cfh,
                )
            utils._EngineFloat64ScalarType(
                _retval_cfh, {'fun_name': cb_cf_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('d05baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        ckh,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        cgh,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        cfh,
        utils._BYREF_STRING, # Input, method(len=1).
        utils._BYREF_INT, # Input, iorder.
        utils._BYREF_FLOAT64, # Input, alim.
        utils._BYREF_FLOAT64, # Input, tlim.
        utils._EngineFloat64ArrayType.c_type, # Output, yn[nmesh].
        utils._EngineFloat64ArrayType.c_type, # Output, errest[nmesh].
        utils._BYREF_INT, # Input, nmesh.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Input, thresh.
        utils._EngineFloat64ArrayType.c_type, # Output, work[lwk].
        utils._BYREF_INT, # Input, lwk.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(ck, {'fun_name': fun_name, 'entity': 'ck'})
    _ck = utils._UNUSED_CALLBACK
    _cksub = utils._UNUSED_CALLBACK
    _ckh = utils._cast_callback(
        ckh, cb_ckh,
        {'fun_name': fun_name, 'entity': 'ck'},
    )
    utils._check_callable(cg, {'fun_name': fun_name, 'entity': 'cg'})
    _cg = utils._UNUSED_CALLBACK
    _cgsub = utils._UNUSED_CALLBACK
    _cgh = utils._cast_callback(
        cgh, cb_cgh,
        {'fun_name': fun_name, 'entity': 'cg'},
    )
    utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = utils._UNUSED_CALLBACK
    _cfsub = utils._UNUSED_CALLBACK
    _cfh = utils._cast_callback(
        cfh, cb_cfh,
        {'fun_name': fun_name, 'entity': 'cf'},
    )
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=1,
    ).to_ctypes()
    _iorder = utils._EngineIntScalarType(
        iorder, {'fun_name': fun_name, 'entity': 'iorder'},
    ).to_ctypes()
    _alim = utils._EngineFloat64ScalarType(
        alim, {'fun_name': fun_name, 'entity': 'alim'},
    ).to_ctypes()
    _tlim = utils._EngineFloat64ScalarType(
        tlim, {'fun_name': fun_name, 'entity': 'tlim'},
    ).to_ctypes()
    _nmesh = utils._EngineIntScalarType(
        nmesh, {'fun_name': fun_name, 'entity': 'nmesh'},
    ).to_ctypes()
    _nmesh_py = utils._EngineIntScalarType.to_py(_nmesh)
    _yn_cls = utils._EngineFloat64ArrayType(
        yn, {'fun_name': fun_name, 'entity': 'yn'},
        exp_shape=(_nmesh_py,),
        check_contents=False,
    )
    _yn = _yn_cls.to_ctypes(can_reference_contents=False)
    _errest_cls = utils._EngineFloat64ArrayType(
        errest, {'fun_name': fun_name, 'entity': 'errest'},
        exp_shape=(_nmesh_py,),
        check_contents=False,
    )
    _errest = _errest_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _thresh = utils._EngineFloat64ScalarType(
        thresh, {'fun_name': fun_name, 'entity': 'thresh'},
    ).to_ctypes()
    _lwk = utils._EngineIntScalarType(
        lwk, {'fun_name': fun_name, 'entity': 'lwk'},
    ).to_ctypes()
    _lwk_py = utils._EngineIntScalarType.to_py(_lwk)
    _work_cls = utils._EngineFloat64ArrayType(
        work, {'fun_name': fun_name, 'entity': 'work'},
        exp_shape=(_lwk_py,),
        check_contents=False,
    )
    _work = _work_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ck,
        _cksub,
        _ckh,
        _cg,
        _cgsub,
        _cgh,
        _cf,
        _cfsub,
        _cfh,
        _method,
        _iorder,
        _alim,
        _tlim,
        _yn,
        _errest,
        _nmesh,
        _tol,
        _thresh,
        _work,
        _lwk,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99907:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'A\' and iorder = 2.',
                    'Constraint: if method = \'A\', 3 <= iorder <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sqrt(machine precision) <= tol <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alim = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'tlim = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: tlim > alim.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alim = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: alim >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'B\' and iorder = 6.',
                    'Constraint: if method = \'B\', 2 <= iorder <= 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 2 <= iorder <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'A\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99903:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'A\', iorder = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nmesh = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'A\', nmesh >= iorder-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'B\', iorder = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nmesh = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'B\', nmesh >= iorder.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lwk >= 10*nmesh+6; that is, {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution is not converging.',
                    'See [Further Comments].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The workspace which has been supplied is too small',
                    'for the required accuracy. The number of',
                    'extrapolations, so far, is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'If you require one more extrapolation extend',
                    'the size of workspace to: lwk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The workspace which has been supplied is too small',
                    'for the required accuracy. The number of',
                    'extrapolations, so far, is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'If you require one more extrapolation extend',
                    'the size of workspace to: lwk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _yn_cls.refresh_py(_yn)
        _errest_cls.refresh_py(_errest)
        _work_cls.refresh_py(_work)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def abel2_weak(ck, cf, cg, initwt, iorder, tlim, tolnl, nmesh, yn, comm, data=None):
    u"""
    Nonlinear convolution Volterra--Abel equation, second kind, weakly
    singular.

    ``abel2_weak`` computes the solution of a weakly singular nonlinear
    convolution Volterra--Abel integral equation of the second kind
    using a fractional Backward Differentiation Formulae (BDF) method.

    For full information please refer to the NAG Library document for
    d05bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05bdf.html

    Parameters
    ----------
    ck : callable retval = ck(t, data=None)
        `ck` must evaluate the kernel k(t) of the integral equation (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of k(t) evaluated at `t`.

    cf : callable retval = cf(t, data=None)
        `cf` must evaluate the function f(t) in (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f(t) evaluated at `t`.

    cg : callable retval = cg(s, y, data=None)
        `cg` must evaluate the function g(s,y(s)) in (1).

        Parameters
        ~~~~~~~~~~
        s : float
            s, the value of the independent variable.

        y : float
            The value of the solution y at the point `s`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(s,y(s)) evaluated at `s` and `y`.

    initwt : str, length 1
        If the fractional weights required by the method need to be
        calculated by the function then set `initwt` = 'I' (Initial
        call).

        If `initwt` = 'S' (Subsequent call), the function assumes the
        fractional weights have been computed on a previous call and are
        stored in `comm`['work'].

    iorder : int
        p, the order of the BDF method to be used.

    tlim : float
        The final point of the integration interval, T.

    tolnl : float
        The accuracy required for the computation of the starting value
        and the solution of the nonlinear equation at each step of the
        computation (see [Further Comments]).

    nmesh : int
        N, the number of equispaced points at which the solution is
        sought.

    yn : float, shape (`nmesh`), modified in place
        On exit: `yn`[i-1] contains the approximate value of the true
        solution y(t) at the point t = (i-1)*h, for i = 1,2,...,`nmesh`,
        where h = `tlim`/(`nmesh`-1).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tolnl` = *<value>*.

            Constraint: `tolnl` > 10*machine precision.

        (`errno` 1)
            On entry, `nmesh` = *<value>* and `iorder` = *<value>*.

            Constraint: `nmesh` = 2^m+2*`iorder`-1, for some m.

        (`errno` 1)
            On entry, `nmesh` = *<value>* and `iorder` = *<value>*.

            Constraint: `nmesh` >= 2*`iorder`+1.

        (`errno` 1)
            On entry, lwk = *<value>*.

            Constraint: lwk >=
            (2*`iorder`+6)*`nmesh`+8*`iorder`^2-16*`iorder`+1; that is,
            *<value>*.

        (`errno` 1)
            On entry, `tlim` = *<value>*.

            Constraints: `tlim` > 10*machine precision.

        (`errno` 1)
            On entry, `initwt` = *<value>*.

            Constraints: `initwt` = 'I' or 'S'.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 4 <= `iorder` <= 6.

        (`errno` 2)
            An error occurred when trying to compute the starting
            values.

        (`errno` 3)
            An error occurred when trying to compute the solution at a
            specific step.
    """

    excs = []
    fun_name = 'naginterfaces.base.inteq.abel2_weak'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_ck_name = fun_name + ':ck'

    ckh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_ckh(_ck, _cksub, en_data, t, retval, _iuser, _ruser):
        """Helper for ck."""
        try:
            _t_ckh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_ckh = ck(
                    _t_ckh,
                    data,
                )
            else:
                _retval_ckh = ck(
                    _t_ckh,
                )
            utils._EngineFloat64ScalarType(
                _retval_ckh, {'fun_name': cb_ck_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_cf_name = fun_name + ':cf'

    cfh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cfh(_cf, _cfsub, en_data, t, retval, _iuser, _ruser):
        """Helper for cf."""
        try:
            _t_cfh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_cfh = cf(
                    _t_cfh,
                    data,
                )
            else:
                _retval_cfh = cf(
                    _t_cfh,
                )
            utils._EngineFloat64ScalarType(
                _retval_cfh, {'fun_name': cb_cf_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_cg_name = fun_name + ':cg'

    cgh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cgh(_cg, _cgsub, en_data, s, y, retval, _iuser, _ruser):
        """Helper for cg."""
        try:
            _s_cgh = utils._EngineFloat64ScalarType.to_py(s.contents)
            _y_cgh = utils._EngineFloat64ScalarType.to_py(y.contents)
            if data is not None:
                _retval_cgh = cg(
                    _s_cgh,
                    _y_cgh,
                    data,
                )
            else:
                _retval_cgh = cg(
                    _s_cgh,
                    _y_cgh,
                )
            utils._EngineFloat64ScalarType(
                _retval_cgh, {'fun_name': cb_cg_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('d05bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        ckh,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        cfh,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        cgh,
        utils._BYREF_STRING, # Input, initwt(len=1).
        utils._BYREF_INT, # Input, iorder.
        utils._BYREF_FLOAT64, # Input, tlim.
        utils._BYREF_FLOAT64, # Input, tolnl.
        utils._BYREF_INT, # Input, nmesh.
        utils._EngineFloat64ArrayType.c_type, # Output, yn[nmesh].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[lwk].
        utils._BYREF_INT, # Input, lwk.
        utils._EngineIntArrayType.c_type, # Workspace, nct.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of initwt.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(ck, {'fun_name': fun_name, 'entity': 'ck'})
    _ck = utils._UNUSED_CALLBACK
    _cksub = utils._UNUSED_CALLBACK
    _ckh = utils._cast_callback(
        ckh, cb_ckh,
        {'fun_name': fun_name, 'entity': 'ck'},
    )
    utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = utils._UNUSED_CALLBACK
    _cfsub = utils._UNUSED_CALLBACK
    _cfh = utils._cast_callback(
        cfh, cb_cfh,
        {'fun_name': fun_name, 'entity': 'cf'},
    )
    utils._check_callable(cg, {'fun_name': fun_name, 'entity': 'cg'})
    _cg = utils._UNUSED_CALLBACK
    _cgsub = utils._UNUSED_CALLBACK
    _cgh = utils._cast_callback(
        cgh, cb_cgh,
        {'fun_name': fun_name, 'entity': 'cg'},
    )
    _initwt = utils._EngineCharScalarType(
        initwt, {'fun_name': fun_name, 'entity': 'initwt'}, exp_chrlen=1,
    ).to_ctypes()
    _iorder = utils._EngineIntScalarType(
        iorder, {'fun_name': fun_name, 'entity': 'iorder'},
    ).to_ctypes()
    _tlim = utils._EngineFloat64ScalarType(
        tlim, {'fun_name': fun_name, 'entity': 'tlim'},
    ).to_ctypes()
    _tolnl = utils._EngineFloat64ScalarType(
        tolnl, {'fun_name': fun_name, 'entity': 'tolnl'},
    ).to_ctypes()
    _nmesh = utils._EngineIntScalarType(
        nmesh, {'fun_name': fun_name, 'entity': 'nmesh'},
    ).to_ctypes()
    _nmesh_py = utils._EngineIntScalarType.to_py(_nmesh)
    _yn_cls = utils._EngineFloat64ArrayType(
        yn, {'fun_name': fun_name, 'entity': 'yn'},
        exp_shape=(_nmesh_py,),
        check_contents=False,
    )
    _yn = _yn_cls.to_ctypes(can_reference_contents=False)
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('work',),
    )
    _iorder_py = utils._EngineIntScalarType.to_py(_iorder)
    if _comm_needs_init:
        _work = utils._EngineFloat64ArrayType.empty_ctypes((2*_iorder_py+6)*_nmesh_py+8*_iorder_py**2-16*_iorder_py+1)
        comm['work'] = _work
    else:
        _work = comm['work']
    _lwk_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwk = utils._EngineIntScalarType(
        _lwk_py, {'fun_name': fun_name, 'entity': '_lwk'},
    ).to_ctypes()
    _nct = utils._EngineIntArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ck,
        _cksub,
        _ckh,
        _cf,
        _cfsub,
        _cfh,
        _cg,
        _cgsub,
        _cgh,
        _initwt,
        _iorder,
        _tlim,
        _tolnl,
        _nmesh,
        _yn,
        _work,
        _lwk,
        _nct,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _initwt._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tolnl = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tolnl > 10*machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmesh = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nmesh = 2^m+2*iorder-1, for some m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmesh = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nmesh >= 2*iorder+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lwk >= (2*iorder+6)*nmesh+8*iorder^2-16*iorder+1;',
                    'that is, {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tlim = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraints: tlim > 10*machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, initwt = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraints: initwt = \'I\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 4 <= iorder <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An error occurred when trying to compute the',
                    'starting values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An error occurred when trying to compute the',
                    'solution at a specific step.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _yn_cls.refresh_py(_yn)
        return_data = None
    _err_class.manage()
    return return_data

def abel1_weak(ck, cf, cg, initwt, iorder, tlim, tolnl, nmesh, yn, comm, data=None):
    u"""
    Nonlinear convolution Volterra--Abel equation, first kind, weakly
    singular.

    ``abel1_weak`` computes the solution of a weakly singular nonlinear
    convolution Volterra--Abel integral equation of the first kind using
    a fractional Backward Differentiation Formulae (BDF) method.

    For full information please refer to the NAG Library document for
    d05be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05bef.html

    Parameters
    ----------
    ck : callable retval = ck(t, data=None)
        `ck` must evaluate the kernel k(t) of the integral equation (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of k(t) evaluated at `t`.

    cf : callable retval = cf(t, data=None)
        `cf` must evaluate the function f(t) in (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f(t) evaluated at `t`.

    cg : callable retval = cg(s, y, data=None)
        `cg` must evaluate the function g(s,y(s)) in (1).

        Parameters
        ~~~~~~~~~~
        s : float
            s, the value of the independent variable.

        y : float
            The value of the solution y at the point `s`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(s,y(s)) evaluated at `s` and `y`.

    initwt : str, length 1
        If the fractional weights required by the method need to be
        calculated by the function then set `initwt` = 'I' (Initial
        call).

        If `initwt` = 'S' (Subsequent call), the function assumes the
        fractional weights have been computed by a previous call and are
        stored in `comm`['work'].

    iorder : int
        p, the order of the BDF method to be used.

    tlim : float
        The final point of the integration interval, T.

    tolnl : float
        The accuracy required for the computation of the starting value
        and the solution of the nonlinear equation at each step of the
        computation (see [Further Comments]).

    nmesh : int
        N, the number of equispaced points at which the solution is
        sought.

    yn : float, shape (`nmesh`), modified in place
        On entry: `yn`[0] must contain the value of y(t) at t = 0 (see
        [Further Comments]).

        On exit: `yn`[i-1] contains the approximate value of the true
        solution y(t) at the point t = (i-1)*h, for i = 1,2,...,`nmesh`,
        where h = `tlim`/(`nmesh`-1).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tolnl` = *<value>*.

            Constraint: `tolnl` > 10*machine precision.

        (`errno` 1)
            On entry, `nmesh` = *<value>* and `iorder` = *<value>*.

            Constraint: `nmesh` = 2^m+2*`iorder`-1, for some m.

        (`errno` 1)
            On entry, `nmesh` = *<value>* and `iorder` = *<value>*.

            Constraint: `nmesh` >= 2*`iorder`+1.

        (`errno` 1)
            On entry, lwk = *<value>*.

            Constraint: lwk >=
            (2*`iorder`+6)*`nmesh`+8*`iorder`^2-16*`iorder`+1; that is,
            *<value>*.

        (`errno` 1)
            On entry, `tlim` = *<value>*.

            Constraint: `tlim` > 10*machine precision

        (`errno` 1)
            On entry, `initwt` = *<value>*.

            Constraint: `initwt` = 'I' or 'S'.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 4 <= `iorder` <= 6.

        (`errno` 2)
            An error occurred when trying to compute the starting
            values.

        (`errno` 3)
            An error occurred when trying to compute the solution at a
            specific step.
    """

    excs = []
    fun_name = 'naginterfaces.base.inteq.abel1_weak'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_ck_name = fun_name + ':ck'

    ckh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_ckh(_ck, _cksub, en_data, t, retval, _iuser, _ruser):
        """Helper for ck."""
        try:
            _t_ckh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_ckh = ck(
                    _t_ckh,
                    data,
                )
            else:
                _retval_ckh = ck(
                    _t_ckh,
                )
            utils._EngineFloat64ScalarType(
                _retval_ckh, {'fun_name': cb_ck_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_cf_name = fun_name + ':cf'

    cfh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cfh(_cf, _cfsub, en_data, t, retval, _iuser, _ruser):
        """Helper for cf."""
        try:
            _t_cfh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_cfh = cf(
                    _t_cfh,
                    data,
                )
            else:
                _retval_cfh = cf(
                    _t_cfh,
                )
            utils._EngineFloat64ScalarType(
                _retval_cfh, {'fun_name': cb_cf_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_cg_name = fun_name + ':cg'

    cgh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cgh(_cg, _cgsub, en_data, s, y, retval, _iuser, _ruser):
        """Helper for cg."""
        try:
            _s_cgh = utils._EngineFloat64ScalarType.to_py(s.contents)
            _y_cgh = utils._EngineFloat64ScalarType.to_py(y.contents)
            if data is not None:
                _retval_cgh = cg(
                    _s_cgh,
                    _y_cgh,
                    data,
                )
            else:
                _retval_cgh = cg(
                    _s_cgh,
                    _y_cgh,
                )
            utils._EngineFloat64ScalarType(
                _retval_cgh, {'fun_name': cb_cg_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('d05beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        ckh,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        cfh,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        cgh,
        utils._BYREF_STRING, # Input, initwt(len=1).
        utils._BYREF_INT, # Input, iorder.
        utils._BYREF_FLOAT64, # Input, tlim.
        utils._BYREF_FLOAT64, # Input, tolnl.
        utils._BYREF_INT, # Input, nmesh.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, yn[nmesh].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[lwk].
        utils._BYREF_INT, # Input, lwk.
        utils._EngineIntArrayType.c_type, # Workspace, nct.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of initwt.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(ck, {'fun_name': fun_name, 'entity': 'ck'})
    _ck = utils._UNUSED_CALLBACK
    _cksub = utils._UNUSED_CALLBACK
    _ckh = utils._cast_callback(
        ckh, cb_ckh,
        {'fun_name': fun_name, 'entity': 'ck'},
    )
    utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = utils._UNUSED_CALLBACK
    _cfsub = utils._UNUSED_CALLBACK
    _cfh = utils._cast_callback(
        cfh, cb_cfh,
        {'fun_name': fun_name, 'entity': 'cf'},
    )
    utils._check_callable(cg, {'fun_name': fun_name, 'entity': 'cg'})
    _cg = utils._UNUSED_CALLBACK
    _cgsub = utils._UNUSED_CALLBACK
    _cgh = utils._cast_callback(
        cgh, cb_cgh,
        {'fun_name': fun_name, 'entity': 'cg'},
    )
    _initwt = utils._EngineCharScalarType(
        initwt, {'fun_name': fun_name, 'entity': 'initwt'}, exp_chrlen=1,
    ).to_ctypes()
    _iorder = utils._EngineIntScalarType(
        iorder, {'fun_name': fun_name, 'entity': 'iorder'},
    ).to_ctypes()
    _tlim = utils._EngineFloat64ScalarType(
        tlim, {'fun_name': fun_name, 'entity': 'tlim'},
    ).to_ctypes()
    _tolnl = utils._EngineFloat64ScalarType(
        tolnl, {'fun_name': fun_name, 'entity': 'tolnl'},
    ).to_ctypes()
    _nmesh = utils._EngineIntScalarType(
        nmesh, {'fun_name': fun_name, 'entity': 'nmesh'},
    ).to_ctypes()
    _nmesh_py = utils._EngineIntScalarType.to_py(_nmesh)
    _yn_cls = utils._EngineFloat64ArrayType(
        yn, {'fun_name': fun_name, 'entity': 'yn'},
        exp_shape=(_nmesh_py,),
    )
    _yn = _yn_cls.to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('work',),
    )
    _iorder_py = utils._EngineIntScalarType.to_py(_iorder)
    if _comm_needs_init:
        _work = utils._EngineFloat64ArrayType.empty_ctypes((2*_iorder_py+6)*_nmesh_py+8*_iorder_py**2-16*_iorder_py+1)
        comm['work'] = _work
    else:
        _work = comm['work']
    _lwk_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwk = utils._EngineIntScalarType(
        _lwk_py, {'fun_name': fun_name, 'entity': '_lwk'},
    ).to_ctypes()
    _nct = utils._EngineIntArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ck,
        _cksub,
        _ckh,
        _cf,
        _cfsub,
        _cfh,
        _cg,
        _cgsub,
        _cgh,
        _initwt,
        _iorder,
        _tlim,
        _tolnl,
        _nmesh,
        _yn,
        _work,
        _lwk,
        _nct,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _initwt._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tolnl = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tolnl > 10*machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmesh = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nmesh = 2^m+2*iorder-1, for some m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmesh = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nmesh >= 2*iorder+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lwk >= (2*iorder+6)*nmesh+8*iorder^2-16*iorder+1;',
                    'that is, {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tlim = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tlim > 10*machine precision',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, initwt = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: initwt = \'I\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 4 <= iorder <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An error occurred when trying to compute the',
                    'starting values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An error occurred when trying to compute the',
                    'solution at a specific step.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _yn_cls.refresh_py(_yn)
        return_data = None
    _err_class.manage()
    return return_data

def volterra_weights(method, iorder, omega, nomg, sw, nwt):
    u"""
    Generate weights for use in solving Volterra equations.

    ``volterra_weights`` computes the quadrature weights associated with
    the Adams' methods of orders three to six and the Backward
    Differentiation Formulae (BDF) methods of orders two to five.
    These rules, which are referred to as reducible quadrature rules,
    can then be used in the solution of Volterra integral and
    integro-differential equations.

    For full information please refer to the NAG Library document for
    d05bw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05bwf.html

    Parameters
    ----------
    method : str, length 1
        The type of method to be used.

        `method` = 'A'

          For Adams' type formulae.

        `method` = 'B'

          For Backward Differentiation Formulae.

    iorder : int
        The order of the method to be used. The number of starting
        weights, p is determined by `method` and `iorder`.

        If `method` = 'A', p = `iorder`-1.

        If `method` = 'B', p = `iorder`.

    omega : float, shape (`nomg`), modified in place
        On exit: contains the first `nomg` convolution weights.

    nomg : int
        The number of convolution weights, n_w.

    sw : float, shape (n, `nwt`), modified in place
        On exit: `sw`[i-1,j] contains the weights W_{i,j}, for j =
        0,1,...,`nwt`-1, for i = 1,2,...,`lensw`, where n is as defined
        in [Description].

    nwt : int
        p, the number of columns in the starting weights.

    Returns
    -------
    lensw : int
        The number of rows in the weights W_{i,j}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'A' or 'B'.

        (`errno` 2)
            On entry, `nomg` = *<value>*.

            Constraint: `nomg` >= 1.

        (`errno` 2)
            On entry, `iorder` = *<value>*.

            Constraint: 2 <= `iorder` <= 6.

        (`errno` 3)
            On entry, `method` = 'B' and `iorder` = 6.

            Constraint: if `method` = 'B', 2 <= `iorder` <= 5.

        (`errno` 3)
            On entry, `method` = 'A' and `iorder` = 2.

            Constraint: if `method` = 'A', 3 <= `iorder` <= 6.

        (`errno` 4)
            On entry, `method` = 'B', `iorder` = *<value>* and `nwt` =
            *<value>*.

            Constraint: if `method` = 'B', `nwt` = `iorder`.

        (`errno` 4)
            On entry, `method` = 'A', `iorder` = *<value>* and `nwt` =
            *<value>*.

            Constraint: if `method` = 'A', `nwt` = `iorder`-1.

        (`errno` 5)
            On entry, `method` = 'B', `iorder` = *<value>* and `nomg` =
            *<value>*, ldsw = *<value>*.

            Constraint: if `method` = 'B', ldsw >= `nomg`+`iorder`-1.

        (`errno` 5)
            On entry, `method` = 'A', `iorder` = *<value>* and `nomg` =
            *<value>*, ldsw = *<value>*.

            Constraint: if `method` = 'A', ldsw >= `nomg`+`iorder`-2.
    """

    excs = []
    fun_name = 'naginterfaces.base.inteq.volterra_weights'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('d05bwft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=1).
        utils._BYREF_INT, # Input, iorder.
        utils._EngineFloat64ArrayType.c_type, # Output, omega[nomg].
        utils._BYREF_INT, # Input, nomg.
        utils._BYREF_INT, # Output, lensw.
        utils._EngineFloat64ArrayType.c_type, # Output, sw[ldsw, nwt].
        utils._BYREF_INT, # Input, ldsw.
        utils._BYREF_INT, # Input, nwt.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((sw, 'sw'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=1,
    ).to_ctypes()
    _iorder = utils._EngineIntScalarType(
        iorder, {'fun_name': fun_name, 'entity': 'iorder'},
    ).to_ctypes()
    _nomg = utils._EngineIntScalarType(
        nomg, {'fun_name': fun_name, 'entity': 'nomg'},
    ).to_ctypes()
    _nomg_py = utils._EngineIntScalarType.to_py(_nomg)
    _omega_cls = utils._EngineFloat64ArrayType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
        exp_shape=(_nomg_py,),
        check_contents=False,
    )
    _omega = _omega_cls.to_ctypes(can_reference_contents=False)
    _lensw = utils._EngineIntScalarType.empty_ctypes()
    _ldsw_py = utils._EngineFloat64ArrayType.get_nd(sw, {'fun_name': fun_name, 'entity': 'sw'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsw = utils._EngineIntScalarType(
        _ldsw_py, {'fun_name': fun_name, 'entity': '_ldsw'},
    ).to_ctypes()
    _nwt = utils._EngineIntScalarType(
        nwt, {'fun_name': fun_name, 'entity': 'nwt'},
    ).to_ctypes()
    _method_py = utils._EngineCharScalarType.to_py(_method)
    _iorder_py = utils._EngineIntScalarType.to_py(_iorder)
    if _method_py in ('A', 'a'):
        _vl_p_py = _iorder_py-1
    elif _method_py in ('B', 'b'):
        _vl_p_py = _iorder_py
    else:
        _vl_p_py = 0
    _vl_n_py = _nomg_py+_vl_p_py-1
    _nwt_py = utils._EngineIntScalarType.to_py(_nwt)
    _sw_cls = utils._EngineFloat64ArrayType(
        sw, {'fun_name': fun_name, 'entity': 'sw'},
        exp_shape=(_vl_n_py, _nwt_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sw = _sw_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _iorder,
        _omega,
        _nomg,
        _lensw,
        _sw,
        _ldsw,
        _nwt,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'A\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nomg = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nomg >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 2 <= iorder <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99907:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'B\' and iorder = 6.',
                    'Constraint: if method = \'B\', 2 <= iorder <= 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'A\' and iorder = 2.',
                    'Constraint: if method = \'A\', 3 <= iorder <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'B\', iorder = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nwt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'B\', nwt = iorder.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'A\', iorder = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nwt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'A\', nwt = iorder-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'B\', iorder = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nomg = {:s}, ldsw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if method = \'B\', ldsw >= nomg+iorder-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'A\', iorder = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nomg = {:s}, ldsw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if method = \'A\', ldsw >= nomg+iorder-2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _omega_cls.refresh_py(_omega)
        _sw_cls.refresh_py(_sw)
        _lensw_py = utils._EngineIntScalarType.to_py(_lensw)
        return_data = _lensw_py
    _err_class.manage()
    return return_data

def abel_weak_weights(iorder, iq, lenfw, wt, sw):
    u"""
    Generate weights for use in solving weakly singular Abel-type
    equations.

    ``abel_weak_weights`` computes the fractional quadrature weights
    associated with the Backward Differentiation Formulae (BDF) of
    orders 4, 5 and 6.
    These weights can then be used in the solution of weakly singular
    equations of Abel type.

    For full information please refer to the NAG Library document for
    d05by

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05byf.html

    Parameters
    ----------
    iorder : int
        p, the order of the BDF method to be used.

    iq : int
        Determines the number of weights to be computed. By setting `iq`
        to a value, 2^(`iq`+1) fractional convolution weights are
        computed.

    lenfw : int
        The dimension of the array `wt`.

    wt : float, shape (`lenfw`), modified in place
        On exit: the first 2^(`iq`+1) elements of `wt` contains the
        fractional convolution weights omega_i, for i =
        0,1,...,2^(`iq`+1)-1. The remainder of the array is used as
        workspace.

    sw : float, shape (2^(`iq`+1)+2*`iorder`-1, 2*`iorder`-1), modified
    in place
        On exit: `sw`[i-1,j] contains the fractional starting weights
        W_{i-1,j}, for j = 0,1,...,2*`iorder`-2, for i = 1,2,...,N,
        where N = (2^(`iq`+1)+2*`iorder`-1).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lenfw` = *<value>* and `iq` = *<value>*.

            Constraint: `lenfw` >= 2^(`iq`+2).

        (`errno` 1)
            On entry, ldsw = *<value>*, `iq` = *<value>* and `iorder` =
            *<value>*.

            Constraint: ldsw >= 2^(`iq`+1)+2*`iorder`-1.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 4 <= `iorder` <= 6.
    """

    excs = []
    fun_name = 'naginterfaces.base.inteq.abel_weak_weights'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('d05byft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iorder.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_INT, # Input, lenfw.
        utils._EngineFloat64ArrayType.c_type, # Output, wt[lenfw].
        utils._EngineFloat64ArrayType.c_type, # Output, sw[ldsw, 2*iorder-1].
        utils._BYREF_INT, # Input, ldsw.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((sw, 'sw'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iorder = utils._EngineIntScalarType(
        iorder, {'fun_name': fun_name, 'entity': 'iorder'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _lenfw = utils._EngineIntScalarType(
        lenfw, {'fun_name': fun_name, 'entity': 'lenfw'},
    ).to_ctypes()
    _lenfw_py = utils._EngineIntScalarType.to_py(_lenfw)
    _wt_cls = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_lenfw_py,),
        check_contents=False,
    )
    _wt = _wt_cls.to_ctypes(can_reference_contents=False)
    _ldsw_py = utils._EngineFloat64ArrayType.get_nd(sw, {'fun_name': fun_name, 'entity': 'sw'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsw = utils._EngineIntScalarType(
        _ldsw_py, {'fun_name': fun_name, 'entity': '_ldsw'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _iorder_py = utils._EngineIntScalarType.to_py(_iorder)
    _sw_cls = utils._EngineFloat64ArrayType(
        sw, {'fun_name': fun_name, 'entity': 'sw'},
        exp_shape=(2**_iq_py+1+2*_iorder_py-1, 2*_iorder_py-1),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sw = _sw_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwk = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iorder,
        _iq,
        _lenfw,
        _wt,
        _sw,
        _ldsw,
        _work,
        _lwk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lenfw = {:s} and iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lenfw >= 2^(iq+2).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldsw = {:s}, iq = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ldsw >= 2^(iq+1)+2*iorder-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 4 <= iorder <= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _wt_cls.refresh_py(_wt)
        _sw_cls.refresh_py(_sw)
        return_data = None
    _err_class.manage()
    return return_data
