# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `machine` Chapter.

``machine`` - Machine Constants

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with parameters which characterise certain
aspects of the computing environment in which the NAG Library is
implemented.
They relate primarily to floating-point arithmetic, but also to integer
arithmetic, the elementary functions and exception handling.
The values of the parameters vary from one implementation of the NAG
Library to another, but within the context of a single implementation
they are constants.

The parameters are intended for use primarily by other functions in the
NAG Library, but users of the NAG Library may sometimes need to refer to
them directly.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02intro.html
"""

# NAG Copyright 2017-2019.

import sys as _sys

from ..base import utils

def sinarg_max():
    u"""
    The largest permissible argument for sin and cos.

    ``sinarg_max`` returns the largest positive argument for which the
    Fortran intrinsic functions SIN and COS return a result with some
    meaningful accuracy.

    For full information please refer to the NAG Library document for
    x02ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02ahf.html

    Returns
    -------
    maxsc : float
        The largest positive argument for which the Fortran intrinsic
        functions SIN and COS return a result with some meaningful
        accuracy.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.sinarg_max'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Output, maxsc.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _maxsc = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _maxsc,
    )
    utils._handle_cb_exception(excs)
    _maxsc_py = utils._EngineFloat64ScalarType.to_py(_maxsc)
    return_data = _maxsc_py
    return return_data

def precision():
    u"""
    The machine precision.

    ``precision`` returns epsilon, the value machine precision.

    For full information please refer to the NAG Library document for
    x02aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02ajf.html

    Returns
    -------
    eps : float
        epsilon, the value machine precision.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.precision'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02ajft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Output, eps.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _eps = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _eps,
    )
    utils._handle_cb_exception(excs)
    _eps_py = utils._EngineFloat64ScalarType.to_py(_eps)
    return_data = _eps_py
    return return_data

def real_smallest():
    u"""
    The smallest positive model number.

    ``real_smallest`` returns the smallest positive floating-point
    number.

    For full information please refer to the NAG Library document for
    x02ak

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02akf.html

    Returns
    -------
    tiny_nag : float
        The smallest positive floating-point number.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.real_smallest'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02akft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Output, tiny_nag.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _tiny_nag = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _tiny_nag,
    )
    utils._handle_cb_exception(excs)
    _tiny_nag_py = utils._EngineFloat64ScalarType.to_py(_tiny_nag)
    return_data = _tiny_nag_py
    return return_data

def real_largest():
    u"""
    The largest positive model number.

    ``real_largest`` returns the largest positive floating-point number.

    For full information please refer to the NAG Library document for
    x02al

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02alf.html

    Returns
    -------
    huge_nag : float
        The largest positive floating-point number.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.real_largest'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02alft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Output, huge_nag.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _huge_nag = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _huge_nag,
    )
    utils._handle_cb_exception(excs)
    _huge_nag_py = utils._EngineFloat64ScalarType.to_py(_huge_nag)
    return_data = _huge_nag_py
    return return_data

def real_safe():
    u"""
    The safe range parameter.

    ``real_safe`` returns the safe range of floating-point arithmetic.

    For full information please refer to the NAG Library document for
    x02am

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02amf.html

    Returns
    -------
    saferange : float
        The safe range of floating-point arithmetic.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.real_safe'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02amft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Output, saferange.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _saferange = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _saferange,
    )
    utils._handle_cb_exception(excs)
    _saferange_py = utils._EngineFloat64ScalarType.to_py(_saferange)
    return_data = _saferange_py
    return return_data

def complex_safe():
    u"""
    The safe range parameter for complex floating-point arithmetic.

    ``complex_safe`` returns the safe range of complex floating-point
    arithmetic.

    For full information please refer to the NAG Library document for
    x02an

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02anf.html

    Returns
    -------
    zrange : float
        The safe range of complex floating-point arithmetic.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.complex_safe'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02anft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Output, zrange.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _zrange = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _zrange,
    )
    utils._handle_cb_exception(excs)
    _zrange_py = utils._EngineFloat64ScalarType.to_py(_zrange)
    return_data = _zrange_py
    return return_data

def integer_max():
    u"""
    The largest representable integer.

    ``integer_max`` returns the largest representable positive integer
    value.

    For full information please refer to the NAG Library document for
    x02bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02bbf.html

    Returns
    -------
    hugeint : int
        The largest representable positive integer value.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.integer_max'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, hugeint.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _hugeint = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _hugeint,
    )
    utils._handle_cb_exception(excs)
    _hugeint_py = utils._EngineIntScalarType.to_py(_hugeint)
    return_data = _hugeint_py
    return return_data

def decimal_digits():
    u"""
    The maximum number of decimal digits that can be represented.

    ``decimal_digits`` returns the maximum number of decimal digits
    which can be accurately represented over the whole range of
    floating-point numbers.

    For full information please refer to the NAG Library document for
    x02be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02bef.html

    Returns
    -------
    iprec : int
        The maximum number of decimal digits which can be accurately
        represented over the whole range of floating-point numbers.
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.decimal_digits'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, iprec.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iprec = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iprec,
    )
    utils._handle_cb_exception(excs)
    _iprec_py = utils._EngineIntScalarType.to_py(_iprec)
    return_data = _iprec_py
    return return_data

def model_base():
    u"""
    The floating-point model parameter, b.

    ``model_base`` returns the base argument b of the model of
    floating-point arithmetic as described in [the X02 Introduction].

    For full information please refer to the NAG Library document for
    x02bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02bhf.html

    Returns
    -------
    irad : int
        The base argument b of the model of floating-point arithmetic as
        described in [the X02 Introduction].
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.model_base'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, irad.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irad = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irad,
    )
    utils._handle_cb_exception(excs)
    _irad_py = utils._EngineIntScalarType.to_py(_irad)
    return_data = _irad_py
    return return_data

def model_digits():
    u"""
    The floating-point model parameter, p.

    ``model_digits`` returns the precision argument p (i.e., the number
    of significant base-b digits) of the model of floating-point
    arithmetic described in [the X02 Introduction].

    For full information please refer to the NAG Library document for
    x02bj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02bjf.html

    Returns
    -------
    idigs : int
        The precision argument p (i.e., the number of significant base-b
        digits) of the model of floating-point arithmetic described in
        [the X02 Introduction].
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.model_digits'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02bjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, idigs.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _idigs = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _idigs,
    )
    utils._handle_cb_exception(excs)
    _idigs_py = utils._EngineIntScalarType.to_py(_idigs)
    return_data = _idigs_py
    return return_data

def model_minexp():
    u"""
    The floating-point model parameter e_min.

    ``model_minexp`` returns the minimum exponent argument e_min of the
    model of floating-point arithmetic described in [the X02
    Introduction].

    For full information please refer to the NAG Library document for
    x02bk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02bkf.html

    Returns
    -------
    minexp : int
        The minimum exponent argument e_min of the model of
        floating-point arithmetic described in [the X02 Introduction].
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.model_minexp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02bkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, minexp.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _minexp = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _minexp,
    )
    utils._handle_cb_exception(excs)
    _minexp_py = utils._EngineIntScalarType.to_py(_minexp)
    return_data = _minexp_py
    return return_data

def model_maxexp():
    u"""
    The floating-point model parameter e_max.

    ``model_maxexp`` returns the maximum exponent argument e_max of the
    model of floating-point arithmetic described in [the X02
    Introduction].

    For full information please refer to the NAG Library document for
    x02bl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x02/x02blf.html

    Returns
    -------
    maxexp : int
        The maximum exponent argument e_max of the model of
        floating-point arithmetic described in [the X02 Introduction].
    """

    excs = []
    fun_name = 'naginterfaces.base.machine.model_maxexp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x02blft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, maxexp.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _maxexp = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _maxexp,
    )
    utils._handle_cb_exception(excs)
    _maxexp_py = utils._EngineIntScalarType.to_py(_maxexp)
    return_data = _maxexp_py
    return return_data
