# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `orthog` Chapter.

``orthog`` - Orthogonalization

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the orthogonalization of vectors in a
finite dimensional space.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f05/f05intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections
import sys as _sys

from ..base import utils

def real_gram_schmidt(a, m, n1, n2):
    u"""
    Gram--Schmidt orthogonalization of n vectors of order m.

    ``real_gram_schmidt`` applies the Schmidt orthogonalization process
    to n vectors in m-dimensional space, n <= m.

    For full information please refer to the NAG Library document for
    f05aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f05/f05aaf.html

    Parameters
    ----------
    a : float, shape (`m`, `n2`), modified in place
        On entry: columns `n1` to `n2` contain the vectors to be
        orthogonalized. The vectors are stored by columns in elements 1
        to m.

        On exit: these vectors are overwritten by the orthonormal
        vectors.

    m : int
        m, the number of elements in each vector.

    n1 : int
        The indices of the first and last columns of A to be
        orthogonalized.

    n2 : int
        The indices of the first and last columns of A to be
        orthogonalized.

    Returns
    -------
    cc : float
        Is used to indicate linear dependence of the original vectors.
        The nearer `cc` is to 1.0, the more likely vector `icol` is
        dependent on vectors `n1` to `icol`-1. See [Further Comments].

    icol : int
        The column number corresponding to `cc`. See [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>* and `n2` = *<value>*.

            Constraint: `n1` <= `n2`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.orthog.real_gram_schmidt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f05aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n2].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._EngineFloat64ArrayType.c_type, # Workspace, s.
        utils._BYREF_FLOAT64, # Output, cc.
        utils._BYREF_INT, # Output, icol.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n2_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _s = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _cc = utils._EngineFloat64ScalarType.empty_ctypes()
    _icol = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _lda,
        _m,
        _n1,
        _n2,
        _s,
        _cc,
        _icol,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s} and n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n1 <= n2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _cc_py = utils._EngineFloat64ScalarType.to_py(_cc)
        _icol_py = utils._EngineIntScalarType.to_py(_icol)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['cc', 'icol'],
        )(
            _cc_py,
            _icol_py,
        )
    _err_class.manage()
    return return_data
