# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `univar` Chapter.

``univar`` - Univariate Estimation

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module deals with the estimation of unknown parameters of a
univariate distribution.
It includes both point and interval estimation using maximum likelihood
and robust methods.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections
import sys as _sys

from ..base import utils

def ci_binomial(n, k, clevel):
    u"""
    Computes confidence interval for the parameter of a binomial
    distribution.

    ``ci_binomial`` computes a confidence interval for the parameter p
    (the probability of a success) of a binomial distribution.

    For full information please refer to the NAG Library document for
    g07aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07aaf.html

    Parameters
    ----------
    n : int
        n, the number of trials.

    k : int
        k, the number of successes.

    clevel : float
        The confidence level, (1-alpha), for two-sided interval
        estimate. For example `clevel` = 0.95 will give a 95% confidence
        interval.

    Returns
    -------
    pl : float
        The lower limit, p_l, of the confidence interval.

    pu : float
        The upper limit, p_u, of the confidence interval.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `n` = *<value>* and `k` = *<value>*.

            Constraint: `n` >= `k`.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            When using the relationship with the gamma distribution the
            series to calculate the gamma probabilities has failed to
            converge.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.ci_binomial'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, clevel.
        utils._BYREF_FLOAT64, # Output, pl.
        utils._BYREF_FLOAT64, # Output, pu.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _clevel = utils._EngineFloat64ScalarType(
        clevel, {'fun_name': fun_name, 'entity': 'clevel'},
    ).to_ctypes()
    _pl = utils._EngineFloat64ScalarType.empty_ctypes()
    _pu = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _k,
        _clevel,
        _pl,
        _pu,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, clevel = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < clevel < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= k.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'When using the relationship with the gamma distribution the series to',
                    'calculate the gamma probabilities has failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _pl_py = utils._EngineFloat64ScalarType.to_py(_pl)
        _pu_py = utils._EngineFloat64ScalarType.to_py(_pu)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['pl', 'pu'],
        )(
            _pl_py,
            _pu_py,
        )
    _err_class.manage()
    return return_data

def ci_poisson(n, xmean, clevel):
    u"""
    Computes confidence interval for the parameter of a Poisson
    distribution.

    ``ci_poisson`` computes a confidence interval for the mean parameter
    of the Poisson distribution.

    For full information please refer to the NAG Library document for
    g07ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07abf.html

    Parameters
    ----------
    n : int
        n, the sample size.

    xmean : float
        The sample mean, x_.

    clevel : float
        The confidence level, (1-alpha), for two-sided interval
        estimate. For example `clevel` = 0.95 gives a 95% confidence
        interval.

    Returns
    -------
    tl : float
        The lower limit, theta_l, of the confidence interval.

    tu : float
        The upper limit, theta_u, of the confidence interval.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `xmean` = *<value>*.

            Constraint: `xmean` >= 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            When using the relationship with the gamma distribution the
            series to calculate the gamma probabilities has failed to
            converge.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.ci_poisson'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmean.
        utils._BYREF_FLOAT64, # Input, clevel.
        utils._BYREF_FLOAT64, # Output, tl.
        utils._BYREF_FLOAT64, # Output, tu.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmean = utils._EngineFloat64ScalarType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
    ).to_ctypes()
    _clevel = utils._EngineFloat64ScalarType(
        clevel, {'fun_name': fun_name, 'entity': 'clevel'},
    ).to_ctypes()
    _tl = utils._EngineFloat64ScalarType.empty_ctypes()
    _tu = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmean,
        _clevel,
        _tl,
        _tu,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, clevel = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < clevel < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmean = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: xmean >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'When using the relationship with the gamma distribution the series to',
                    'calculate the gamma probabilities has failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _tl_py = utils._EngineFloat64ScalarType.to_py(_tl)
        _tu_py = utils._EngineFloat64ScalarType.to_py(_tu)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['tl', 'tu'],
        )(
            _tl_py,
            _tu_py,
        )
    _err_class.manage()
    return return_data

def estim_normal(method, n, x, xc, ic, xmu, xsig, tol, maxit, nobs):
    u"""
    Computes maximum likelihood estimates for parameters of the Normal
    distribution from grouped and/or censored data.

    ``estim_normal`` computes maximum likelihood estimates and their
    standard errors for parameters of the Normal distribution from
    grouped and/or censored data.

    For full information please refer to the NAG Library document for
    g07bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07bbf.html

    Parameters
    ----------
    method : str, length 1
        Indicates whether the Newton--Raphson or EM algorithm should be
        used.

        If `method` = 'N', the Newton--Raphson algorithm is used.

        If `method` = 'E', the EM algorithm is used.

    n : int
        n, the number of observations.

    x : float, shape (`n`)
        The observations x_i, L_i or U_i, for i = 1,2,...,n.

        If the observation is exactly specified -- the exact value, x_i.

        If the observation is right-censored -- the lower value, L_i.

        If the observation is left-censored -- the upper value, U_i.

        If the observation is interval-censored -- the lower or upper
        value, L_i or U_i, (see `xc`).

    xc : float, shape (`n`)
        If the jth observation, for j = 1,2,...,n is an
        interval-censored observation then `xc`[j-1] should contain the
        complementary value to `x`[j-1], that is, if `x`[j-1] <
        `xc`[j-1], then `xc`[j-1] contains upper value, U_i, and if
        `x`[j-1] > `xc`[j-1], then `xc`[j-1] contains lower value, L_i.
        Otherwise if the jth observation is exact or right - or
        left-censored `xc`[j-1] need not be set.

    ic : int, shape (`n`)
        `ic`[i-1] contains the censoring codes for the ith observation,
        for i = 1,2,...,n.

        If `ic`[i-1] = 0, the observation is exactly specified.

        If `ic`[i-1] = 1, the observation is right-censored.

        If `ic`[i-1] = 2, the observation is left-censored.

        If `ic`[i-1] = 3, the observation is interval-censored.

    xmu : float
        If `xsig` > 0.0 the initial estimate of the mean, mu; otherwise
        `xmu` need not be set.

    xsig : float
        Specifies whether an initial estimate of mu and sigma are to be
        supplied.

        `xsig` > 0.0

          `xsig` is the initial estimate of sigma and `xmu` must contain
          an initial estimate of mu.

        `xsig` <= 0.0

          Initial estimates of `xmu` and `xsig` are calculated
          internally from:

          - the exact observations, if the number of exactly specified
            observations is >= 2; or

          - the interval-censored observations; if the number of
            interval-censored observations is >= 1; or

          - they are set to 0.0 and 1.0 respectively.

    tol : float
        The relative precision required for the final estimates of mu
        and sigma. Convergence is assumed when the absolute relative
        changes in the estimates of both mu and sigma are less than
        `tol`.

        If `tol` = 0.0, a relative precision of 0.000005 is used.

    maxit : int
        The maximum number of iterations.

        If `maxit` <= 0, a value of 25 is used.

    nobs : int, shape (4), modified in place
        On exit: the number of the different types of each observation;

        `nobs`[0] contains number of right-censored observations.

        `nobs`[1] contains number of left-censored observations.

        `nobs`[2] contains number of interval-censored observations.

        `nobs`[3] contains number of exactly specified observations.

    Returns
    -------
    xmu : float
        The maximum likelihood estimate, mu^, of mu.

    xsig : float
        The maximum likelihood estimate, sigma^, of sigma.

    sexmu : float
        The estimate of the standard error of mu^.

    sexsig : float
        The estimate of the standard error of sigma^.

    corr : float
        The estimate of the correlation between mu^ and sigma^.

    dev : float
        The maximized log-likelihood, L(mu^,sigma^).

    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'N' or 'E'.

        (`errno` 1)
            On entry, effective number of observations < 2.

        (`errno` 1)
            On entry, i = *<value>* and `ic`[i-1] = *<value>*.

            Constraint: `ic`[i-1] = 0, 1, 2 or 3.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: machine precision < `tol` <= 1.0 or `tol` = 0.0.

        (`errno` 2)
            The chosen method has not converged in *<value>* iterations.

        (`errno` 3)
            The process has diverged.

        (`errno` 3)
            The EM process has failed.

        (`errno` 4)
            Standard errors cannot be computed.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.estim_normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, xc[n].
        utils._EngineIntArrayType.c_type, # Input, ic[n].
        utils._BYREF_FLOAT64, # Input/Output, xmu.
        utils._BYREF_FLOAT64, # Input/Output, xsig.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Output, sexmu.
        utils._BYREF_FLOAT64, # Output, sexsig.
        utils._BYREF_FLOAT64, # Output, corr.
        utils._BYREF_FLOAT64, # Output, dev.
        utils._EngineIntArrayType.c_type, # Output, nobs[4].
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _xc = utils._EngineFloat64ArrayType(
        xc, {'fun_name': fun_name, 'entity': 'xc'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ic = utils._EngineIntArrayType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _xmu = utils._EngineFloat64ScalarType(
        xmu, {'fun_name': fun_name, 'entity': 'xmu'},
    ).to_ctypes()
    _xsig = utils._EngineFloat64ScalarType(
        xsig, {'fun_name': fun_name, 'entity': 'xsig'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _sexmu = utils._EngineFloat64ScalarType.empty_ctypes()
    _sexsig = utils._EngineFloat64ScalarType.empty_ctypes()
    _corr = utils._EngineFloat64ScalarType.empty_ctypes()
    _dev = utils._EngineFloat64ScalarType.empty_ctypes()
    _nobs_cls = utils._EngineIntArrayType(
        nobs, {'fun_name': fun_name, 'entity': 'nobs'},
        exp_shape=(4,),
        check_contents=False,
    )
    _nobs = _nobs_cls.to_ctypes(can_reference_contents=False)
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _x,
        _xc,
        _ic,
        _xmu,
        _xsig,
        _tol,
        _maxit,
        _sexmu,
        _sexsig,
        _corr,
        _dev,
        _nobs,
        _nit,
        _wk,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'N\' or \'E\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, effective number of observations < 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and ic[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ic[i-1] = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: machine precision < tol <= 1.0 or tol = 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The chosen method has not converged in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The process has diverged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The EM process has failed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Standard errors cannot be computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xmu_py = utils._EngineFloat64ScalarType.to_py(_xmu)
        _xsig_py = utils._EngineFloat64ScalarType.to_py(_xsig)
        _nobs_cls.refresh_py(_nobs)
        _sexmu_py = utils._EngineFloat64ScalarType.to_py(_sexmu)
        _sexsig_py = utils._EngineFloat64ScalarType.to_py(_sexsig)
        _corr_py = utils._EngineFloat64ScalarType.to_py(_corr)
        _dev_py = utils._EngineFloat64ScalarType.to_py(_dev)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['xmu', 'xsig', 'sexmu', 'sexsig', 'corr', 'dev', 'nit'],
        )(
            _xmu_py,
            _xsig_py,
            _sexmu_py,
            _sexsig_py,
            _corr_py,
            _dev_py,
            _nit_py,
        )
    _err_class.manage()
    return return_data

def estim_weibull(cens, n, x, ic, gamma, tol, maxit):
    u"""
    Computes maximum likelihood estimates for parameters of the Weibull
    distribution.

    ``estim_weibull`` computes maximum likelihood estimates for
    parameters of the Weibull distribution from data which may be
    right-censored.

    For full information please refer to the NAG Library document for
    g07be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07bef.html

    Parameters
    ----------
    cens : str, length 1
        Indicates whether the data is censored or non-censored.

        `cens` = 'N'

          Each observation is assumed to be exactly specified. `ic` is
          not referenced.

        `cens` = 'C'

          Each observation is censored according to the value contained
          in `ic`[i-1], for i = 1,2,...,n.

    n : int
        n, the number of observations.

    x : float, shape (`n`)
        `x`[i-1] contains the ith observation, x_i, for i = 1,2,...,n.

    ic : int, shape (:)
        Note: the required length for this argument is determined as
        follows: if `cens`='C': `n`; otherwise: 1.

        If `cens` = 'C', `ic`[i-1] contains the censoring codes for the
        ith observation, for i = 1,2,...,n.

        If `ic`[i-1] = 0, the ith observation is exactly specified.

        If `ic`[i-1] = 1, the ith observation is right-censored.

        If `cens` = 'N', `ic` is not referenced.

    gamma : float
        Indicates whether an initial estimate of gamma is provided.

        If `gamma` > 0.0, it is taken as the initial estimate of gamma
        and an initial estimate of beta is calculated from this value of
        gamma.

        If `gamma` <= 0.0, initial estimates of gamma and beta are
        calculated, internally, providing the data contains at least two
        distinct exact observations. (If there are only two distinct
        exact observations, the largest observation must not be exactly
        specified.) See [Further Comments] for further details.

    tol : float
        The relative precision required for the final estimates of beta
        and gamma. Convergence is assumed when the absolute relative
        changes in the estimates of both beta and gamma are less than
        `tol`.

        If `tol` = 0.0, a relative precision of 0.000005 is used.

    maxit : int
        The maximum number of iterations allowed.

        If `maxit` <= 0, a value of 25 is used.

    Returns
    -------
    beta : float
        The maximum likelihood estimate, beta^, of beta.

    gamma : float
        Contains the maximum likelihood estimate, gamma^, of gamma.

    sebeta : float
        An estimate of the standard error of beta^.

    segam : float
        An estimate of the standard error of gamma^.

    corr : float
        An estimate of the correlation between beta^ and gamma^.

    dev : float
        The maximized kernel log-likelihood, L(beta^,gamma^).

    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: machine precision < `tol` <= 1.0 or `tol` = 0.0.

        (`errno` 1)
            On entry, `cens` = *<value>*.

            Constraint: `cens` = 'N' or 'C'.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `ic`[i-1] = *<value>*.

            Constraint: `ic`[i-1] = 0 or 1.

        (`errno` 2)
            On entry, i = *<value>* and `x`[i-1] = *<value>*.

            Constraint: `x`[i-1] > 0.0.

        (`errno` 3)
            Unable to calculate initial values.

        (`errno` 3)
            On entry, there are no exactly specified observations.

        (`errno` 4)
            The chosen method has not converged in *<value>* iterations.

        (`errno` 5)
            The process has diverged.

        (`errno` 5)
            Hessian matrix of the Newton--Raphson process is singular.

        (`errno` 6)
            Potential overflow detected.

    See Also
    --------
    :meth:`naginterfaces.library.examples.univar.estim_weibull_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.estim_weibull'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, cens(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineIntArrayType.c_type, # Input, ic[*].
        utils._BYREF_FLOAT64, # Output, beta.
        utils._BYREF_FLOAT64, # Input/Output, gamma.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Output, sebeta.
        utils._BYREF_FLOAT64, # Output, segam.
        utils._BYREF_FLOAT64, # Output, corr.
        utils._BYREF_FLOAT64, # Output, dev.
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of cens.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _cens = utils._EngineCharScalarType(
        cens, {'fun_name': fun_name, 'entity': 'cens'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _cens_py = utils._EngineCharScalarType.to_py(_cens)
    if _cens_py in ('C', 'c'):
        _ic_dim1 = _n_py
    else:
        _ic_dim1 = 1
    exp_shape_ic = (_ic_dim1,)
    _ic = utils._EngineIntArrayType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
        exp_shape=exp_shape_ic,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType.empty_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _sebeta = utils._EngineFloat64ScalarType.empty_ctypes()
    _segam = utils._EngineFloat64ScalarType.empty_ctypes()
    _corr = utils._EngineFloat64ScalarType.empty_ctypes()
    _dev = utils._EngineFloat64ScalarType.empty_ctypes()
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _cens,
        _n,
        _x,
        _ic,
        _beta,
        _gamma,
        _tol,
        _maxit,
        _sebeta,
        _segam,
        _corr,
        _dev,
        _nit,
        _wk,
        _errbuf,
        _ierr,
        _cens._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: machine precision < tol <= 1.0 or tol = 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cens = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: cens = \'N\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and ic[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ic[i-1] = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and x[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: x[i-1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate initial values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, there are no exactly specified observations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The chosen method has not converged in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The process has diverged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Hessian matrix of the Newton--Raphson process is singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Potential overflow detected.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _gamma_py = utils._EngineFloat64ScalarType.to_py(_gamma)
        _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
        _sebeta_py = utils._EngineFloat64ScalarType.to_py(_sebeta)
        _segam_py = utils._EngineFloat64ScalarType.to_py(_segam)
        _corr_py = utils._EngineFloat64ScalarType.to_py(_corr)
        _dev_py = utils._EngineFloat64ScalarType.to_py(_dev)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['beta', 'gamma', 'sebeta', 'segam', 'corr', 'dev', 'nit'],
        )(
            _beta_py,
            _gamma_py,
            _sebeta_py,
            _segam_py,
            _corr_py,
            _dev_py,
            _nit_py,
        )
    _err_class.manage()
    return return_data

def estim_genpareto(n, y, optopt, asvc, obsvc):
    u"""
    Estimates parameter values of the generalized Pareto distribution.

    ``estim_genpareto`` estimates parameter values for the generalized
    Pareto distribution by using either moments or maximum likelihood.

    For full information please refer to the NAG Library document for
    g07bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07bff.html

    Parameters
    ----------
    n : int
        The number of observations.

    y : float, shape (`n`)
        The n observations y_i, for i = 1,2,...,n, assumed to follow a
        generalized Pareto distribution.

    optopt : int
        Determines the method of estimation, set:

        `optopt` = -2

          For the method of probability-weighted moments.

        `optopt` = -1

          For the method of moments.

        `optopt` = 1

          For maximum likelihood with starting values given by the
          method of moments estimates.

        `optopt` = 2

          For maximum likelihood with starting values given by the
          method of probability-weighted moments.

    asvc : float, shape (4), modified in place
        On exit: the variance-covariance of the asymptotic Normal
        distribution of xi^ and beta^. `asvc`[0] contains the variance
        of xi^; `asvc`[3] contains the variance of beta^; `asvc`[1] and
        `asvc`[2] contain the covariance of xi^ and beta^.

    obsvc : float, shape (4), modified in place
        On exit: if maximum likelihood estimates are requested, the
        observed variance-covariance of xi^ and beta^. `obsvc`[0]
        contains the variance of xi^; `obsvc`[3] contains the variance
        of beta^; `obsvc`[1] and `obsvc`[2] contain the covariance of
        xi^ and beta^.

    Returns
    -------
    xi : float
        The parameter estimate xi^.

    beta : float
        The parameter estimate beta^.

    ll : float
        If maximum likelihood estimates are requested, `ll` contains the
        log-likelihood value L at the end of the optimization; otherwise
        `ll` is set to -1.0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 2)
            On entry, `y`[*<value>*] = *<value>*.

            Constraint: `y`[i-1] >= 0.0 for all i.

        (`errno` 3)
            On entry, `optopt` = *<value>*.

            Constraint: `optopt` = -2, -1, 1 or 2.

        (`errno` 9)
            The optimization of log-likelihood failed to converge; no
            maximum likelihood estimates are returned. Try using the
            other maximum likelihood option by resetting `optopt`. If
            this also fails, moments-based estimates can be returned by
            an appropriate setting of `optopt`.

        (`errno` 10)
            Variance of data in `y` is too low for method of moments
            optimization.

        (`errno` 11)
            The sum of `y` is zero within machine precision.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The asymptotic distribution is not available for the
            returned parameter estimates.

        (`errno` 7)
            The distribution of maximum likelihood estimates cannot be
            calculated for the returned parameter estimates because the
            Hessian matrix could not be inverted.

        (`errno` 8)
            The asymptotic distribution of parameter estimates is
            invalid and the distribution of maximum likelihood estimates
            cannot be calculated for the returned parameter estimates
            because the Hessian matrix could not be inverted.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.estim_genpareto'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, optopt.
        utils._BYREF_FLOAT64, # Output, xi.
        utils._BYREF_FLOAT64, # Output, beta.
        utils._EngineFloat64ArrayType.c_type, # Output, asvc[4].
        utils._EngineFloat64ArrayType.c_type, # Output, obsvc[4].
        utils._BYREF_FLOAT64, # Output, ll.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _optopt = utils._EngineIntScalarType(
        optopt, {'fun_name': fun_name, 'entity': 'optopt'},
    ).to_ctypes()
    _xi = utils._EngineFloat64ScalarType.empty_ctypes()
    _beta = utils._EngineFloat64ScalarType.empty_ctypes()
    _asvc_cls = utils._EngineFloat64ArrayType(
        asvc, {'fun_name': fun_name, 'entity': 'asvc'},
        exp_shape=(4,),
        check_contents=False,
    )
    _asvc = _asvc_cls.to_ctypes(can_reference_contents=False)
    _obsvc_cls = utils._EngineFloat64ArrayType(
        obsvc, {'fun_name': fun_name, 'entity': 'obsvc'},
        exp_shape=(4,),
        check_contents=False,
    )
    _obsvc = _obsvc_cls.to_ctypes(can_reference_contents=False)
    _ll = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _y,
        _optopt,
        _xi,
        _beta,
        _asvc,
        _obsvc,
        _ll,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: y[i-1] >= 0.0 for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, optopt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: optopt = -2, -1, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The asymptotic distribution is not available for the',
                    'returned parameter estimates.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The distribution of maximum likelihood estimates cannot be calculated',
                    'for the returned parameter estimates because the Hessian matrix',
                    'could not be inverted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The asymptotic distribution of parameter estimates is invalid and the',
                    'distribution of maximum likelihood estimates cannot be calculated for the',
                    'returned parameter estimates because the Hessian matrix could not be',
                    'inverted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The optimization of log-likelihood failed to converge; no maximum likelihood',
                    'estimates are returned. Try using the other maximum likelihood option by',
                    'resetting optopt. If this also fails, moments-based estimates can be',
                    'returned by an appropriate setting of optopt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Variance of data in y is too low for method of moments optimization.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The sum of y is zero within machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _asvc_cls.refresh_py(_asvc)
        _obsvc_cls.refresh_py(_obsvc)
        _xi_py = utils._EngineFloat64ScalarType.to_py(_xi)
        _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
        _ll_py = utils._EngineFloat64ScalarType.to_py(_ll)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['xi', 'beta', 'll'],
        )(
            _xi_py,
            _beta_py,
            _ll_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def ttest_2normal(tail, equal, nx, ny, xmean, ymean, xstd, ystd, clevel):
    u"""
    Computes t-test statistic for a difference in means between two
    Normal populations, confidence interval.

    ``ttest_2normal`` computes a t-test statistic to test for a
    difference in means between two Normal populations, together with a
    confidence interval for the difference between the means.

    For full information please refer to the NAG Library document for
    g07ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07caf.html

    Parameters
    ----------
    tail : str, length 1
        Indicates which tail probability is to be calculated, and thus
        which alternative hypothesis is to be used.

        `tail` = 'T'

          The two tail probability, i.e., H_1:mu_x != mu_y.

        `tail` = 'U'

          The upper tail probability, i.e., H_1:mu_x > mu_y.

        `tail` = 'L'

          The lower tail probability, i.e., H_1:mu_x < mu_y.

    equal : str, length 1
        Indicates whether the population variances are assumed to be
        equal or not.

        `equal` = 'E'

          The population variances are assumed to be equal, that is
          sigma_x^2 = sigma_y^2.

        `equal` = 'U'

          The population variances are not assumed to be equal.

    nx : int
        n_x, the size of the X sample.

    ny : int
        n_y, the size of the Y sample.

    xmean : float
        x_, the mean of the X sample.

    ymean : float
        y_, the mean of the Y sample.

    xstd : float
        s_x, the standard deviation of the X sample.

    ystd : float
        s_y, the standard deviation of the Y sample.

    clevel : float
        The confidence level, 1-alpha, for the specified tail. For
        example `clevel` = 0.95 will give a 95% confidence interval.

    Returns
    -------
    t : float
        Contains the test statistic, t_obs or t'_obs.

    df : float
        Contains the degrees of freedom for the test statistic.

    prob : float
        Contains the significance level, that is the tail probability,
        p, as defined by `tail`.

    dl : float
        Contains the lower confidence limit for mu_x-mu_y.

    du : float
        Contains the upper confidence limit for mu_x-mu_y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `ystd` = *<value>*.

            Constraint: `ystd` > 0.0.

        (`errno` 1)
            On entry, `xstd` = *<value>*.

            Constraint: `xstd` > 0.0.

        (`errno` 1)
            On entry, `ny` = *<value>*.

            Constraint: `ny` >= 2.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 2.

        (`errno` 1)
            On entry, `equal` = *<value>*.

            Constraint: `equal` = 'E' or 'U'.

        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'T', 'U' or 'L'.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.ttest_2normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, tail(len=1).
        utils._BYREF_STRING, # Input, equal(len=1).
        utils._BYREF_INT, # Input, nx.
        utils._BYREF_INT, # Input, ny.
        utils._BYREF_FLOAT64, # Input, xmean.
        utils._BYREF_FLOAT64, # Input, ymean.
        utils._BYREF_FLOAT64, # Input, xstd.
        utils._BYREF_FLOAT64, # Input, ystd.
        utils._BYREF_FLOAT64, # Input, clevel.
        utils._BYREF_FLOAT64, # Output, t.
        utils._BYREF_FLOAT64, # Output, df.
        utils._BYREF_FLOAT64, # Output, prob.
        utils._BYREF_FLOAT64, # Output, dl.
        utils._BYREF_FLOAT64, # Output, du.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of tail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of equal.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _tail = utils._EngineCharScalarType(
        tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_chrlen=1,
    ).to_ctypes()
    _equal = utils._EngineCharScalarType(
        equal, {'fun_name': fun_name, 'entity': 'equal'}, exp_chrlen=1,
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _ny = utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _xmean = utils._EngineFloat64ScalarType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
    ).to_ctypes()
    _ymean = utils._EngineFloat64ScalarType(
        ymean, {'fun_name': fun_name, 'entity': 'ymean'},
    ).to_ctypes()
    _xstd = utils._EngineFloat64ScalarType(
        xstd, {'fun_name': fun_name, 'entity': 'xstd'},
    ).to_ctypes()
    _ystd = utils._EngineFloat64ScalarType(
        ystd, {'fun_name': fun_name, 'entity': 'ystd'},
    ).to_ctypes()
    _clevel = utils._EngineFloat64ScalarType(
        clevel, {'fun_name': fun_name, 'entity': 'clevel'},
    ).to_ctypes()
    _t = utils._EngineFloat64ScalarType.empty_ctypes()
    _df = utils._EngineFloat64ScalarType.empty_ctypes()
    _prob = utils._EngineFloat64ScalarType.empty_ctypes()
    _dl = utils._EngineFloat64ScalarType.empty_ctypes()
    _du = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _tail,
        _equal,
        _nx,
        _ny,
        _xmean,
        _ymean,
        _xstd,
        _ystd,
        _clevel,
        _t,
        _df,
        _prob,
        _dl,
        _du,
        _errbuf,
        _ierr,
        _tail._length_,
        _equal._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, clevel = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < clevel < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ystd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ystd > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xstd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: xstd > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ny = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ny >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nx >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, equal = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: equal = \'E\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tail = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tail = \'T\', \'U\' or \'L\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _t_py = utils._EngineFloat64ScalarType.to_py(_t)
        _df_py = utils._EngineFloat64ScalarType.to_py(_df)
        _prob_py = utils._EngineFloat64ScalarType.to_py(_prob)
        _dl_py = utils._EngineFloat64ScalarType.to_py(_dl)
        _du_py = utils._EngineFloat64ScalarType.to_py(_du)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['t', 'df', 'prob', 'dl', 'du'],
        )(
            _t_py,
            _df_py,
            _prob_py,
            _dl_py,
            _du_py,
        )
    _err_class.manage()
    return return_data

def robust_1var_median(n, x, y):
    u"""
    Robust estimation, median, median absolute deviation, robust
    standard deviation.

    ``robust_1var_median`` finds the median, median absolute deviation,
    and a robust estimate of the standard deviation for a set of
    ungrouped data.

    For full information please refer to the NAG Library document for
    g07da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07daf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    x : float, shape (`n`)
        The vector of observations, x_1,x_2,...,x_n.

    y : float, shape (`n`), modified in place
        On exit: the observations sorted into ascending order.

    Returns
    -------
    xme : float
        The median, theta_med.

    xmd : float
        The median absolute deviation, sigma_med.

    xsd : float
        The robust estimate of the standard deviation, sigma'_med.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.robust_1var_median'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, y[n].
        utils._BYREF_FLOAT64, # Output, xme.
        utils._BYREF_FLOAT64, # Output, xmd.
        utils._BYREF_FLOAT64, # Output, xsd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _xme = utils._EngineFloat64ScalarType.empty_ctypes()
    _xmd = utils._EngineFloat64ScalarType.empty_ctypes()
    _xsd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _xme,
        _xmd,
        _xsd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        _xme_py = utils._EngineFloat64ScalarType.to_py(_xme)
        _xmd_py = utils._EngineFloat64ScalarType.to_py(_xmd)
        _xsd_py = utils._EngineFloat64ScalarType.to_py(_xsd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['xme', 'xmd', 'xsd'],
        )(
            _xme_py,
            _xmd_py,
            _xsd_py,
        )
    _err_class.manage()
    return return_data

def robust_1var_mestim(isigma, n, x, ipsi, c, h1, h2, h3, dchi, theta, sigma, maxit, tol, rs, wrk):
    u"""
    Robust estimation, M-estimates for location and scale parameters,
    standard weight functions.

    ``robust_1var_mestim`` computes an M-estimate of location with
    (optional) simultaneous estimation of the scale using Huber's
    algorithm.

    For full information please refer to the NAG Library document for
    g07db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07dbf.html

    Parameters
    ----------
    isigma : int
        The value assigned to `isigma` determines whether sigma^ is to
        be simultaneously estimated.

        `isigma` = 0

          The estimation of sigma^ is bypassed and `sigma` is set equal
          to sigma_c.

        `isigma` = 1

          sigma^ is estimated simultaneously.

    n : int
        n, the number of observations.

    x : float, shape (`n`)
        The vector of observations, x_1,x_2,...,x_n.

    ipsi : int
        Which psi function is to be used.

        `ipsi` = 0

          psi(t) = t.

        `ipsi` = 1

          Huber's function.

        `ipsi` = 2

          Hampel's piecewise linear function.

        `ipsi` = 3

          Andrew's sine wave,

        `ipsi` = 4

          Tukey's bi-weight.

    c : float
        If `ipsi` = 1, `c` must specify the parameter, c, of Huber's psi
        function. `c` is not referenced if `ipsi` != 1.

    h1 : float
        If `ipsi` = 2, `h1`, `h2` and `h3` must specify the parameters,
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2` and `h3` are not referenced if `ipsi` != 2.

    h2 : float
        If `ipsi` = 2, `h1`, `h2` and `h3` must specify the parameters,
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2` and `h3` are not referenced if `ipsi` != 2.

    h3 : float
        If `ipsi` = 2, `h1`, `h2` and `h3` must specify the parameters,
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2` and `h3` are not referenced if `ipsi` != 2.

    dchi : float
        d, the parameter of the chi function. `dchi` is not referenced
        if `ipsi` = 0.

    theta : float
        If `sigma` > 0 then `theta` must be set to the required starting
        value of the estimation of the location parameter theta^. A
        reasonable initial value for theta^ will often be the sample
        mean or median.

    sigma : float
        The role of `sigma` depends on the value assigned to `isigma`,
        as follows:

        - if `isigma` = 1, `sigma` must be assigned a value which
          determines the values of the starting points for the
          calculations of theta^ and sigma^. If `sigma` <= 0.0 then
          ``robust_1var_mestim`` will determine the starting points of
          theta^ and sigma^. Otherwise the value assigned to `sigma`
          will be taken as the starting point for sigma^, and `theta`
          must be assigned a value before entry, see above;

        - if `isigma` = 0, `sigma` must be assigned a value which
          determines the value of sigma_c, which is held fixed during
          the iterations, and the starting value for the calculation of
          theta^. If `sigma` <= 0, ``robust_1var_mestim`` will determine
          the value of sigma_c as the median absolute deviation adjusted
          to reduce bias (see ``robust_1var_median``) and the starting
          point for theta^. Otherwise, the value assigned to `sigma`
          will be taken as the value of sigma_c and `theta` must be
          assigned a relevant value before entry, see above.

    maxit : int
        The maximum number of iterations that should be used during the
        estimation.

    tol : float
        The relative precision for the final estimates. Convergence is
        assumed when the increments for `theta`, and `sigma` are less
        than `tol`*max(1.0, sigma_{k-1}).

    rs : float, shape (`n`), modified in place
        On exit: the Winsorized residuals.

    wrk : float, shape (`n`), modified in place
        On exit: if `sigma` <= 0.0 on entry, `wrk` will contain the n
        observations in ascending order.

    Returns
    -------
    theta : float
        The M-estimate of the location parameter, theta^.

    sigma : float
        Contains the M-estimate of the scale parameter, sigma^, if
        `isigma` was assigned the value 1 on entry, otherwise `sigma`
        will contain the initial fixed value sigma_c.

    nit : int
        The number of iterations that were used during the estimation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ipsi` = *<value>*.

            Constraint: `ipsi` = 0, 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `isigma` = *<value>*.

            Constraint: `isigma` = 0 or 1.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 2)
            On entry, `dchi` = *<value>*.

            Constraint: `dchi` > 0.0.

        (`errno` 2)
            On entry, `h1` = *<value>*, `h2` = *<value>* and `h3` =
            *<value>*.

            Constraint: 0 <= `h1` <= `h2` <= `h3` and `h3` > 0.0.

        (`errno` 2)
            On entry, `c` = *<value>*.

            Constraint: `c` > 0.0.

        (`errno` 3)
            All elements of `x` are equal.

        (`errno` 4)
            Current estimate of `sigma` is zero or negative: `sigma` =
            *<value>*.

        (`errno` 5)
            Number of iterations required exceeds `maxit`: `maxit` =
            *<value>*.

        (`errno` 6)
            All winsorized residuals are zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.robust_1var_mestim'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07dbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, isigma.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_INT, # Input, ipsi.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, h1.
        utils._BYREF_FLOAT64, # Input, h2.
        utils._BYREF_FLOAT64, # Input, h3.
        utils._BYREF_FLOAT64, # Input, dchi.
        utils._BYREF_FLOAT64, # Input/Output, theta.
        utils._BYREF_FLOAT64, # Input/Output, sigma.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Output, rs[n].
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Output, wrk[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _isigma = utils._EngineIntScalarType(
        isigma, {'fun_name': fun_name, 'entity': 'isigma'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ipsi = utils._EngineIntScalarType(
        ipsi, {'fun_name': fun_name, 'entity': 'ipsi'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _h1 = utils._EngineFloat64ScalarType(
        h1, {'fun_name': fun_name, 'entity': 'h1'},
    ).to_ctypes()
    _h2 = utils._EngineFloat64ScalarType(
        h2, {'fun_name': fun_name, 'entity': 'h2'},
    ).to_ctypes()
    _h3 = utils._EngineFloat64ScalarType(
        h3, {'fun_name': fun_name, 'entity': 'h3'},
    ).to_ctypes()
    _dchi = utils._EngineFloat64ScalarType(
        dchi, {'fun_name': fun_name, 'entity': 'dchi'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _rs_cls = utils._EngineFloat64ArrayType(
        rs, {'fun_name': fun_name, 'entity': 'rs'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _rs = _rs_cls.to_ctypes(can_reference_contents=False)
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wrk_cls = utils._EngineFloat64ArrayType(
        wrk, {'fun_name': fun_name, 'entity': 'wrk'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _wrk = _wrk_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _isigma,
        _n,
        _x,
        _ipsi,
        _c,
        _h1,
        _h2,
        _h3,
        _dchi,
        _theta,
        _sigma,
        _maxit,
        _tol,
        _rs,
        _nit,
        _wrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipsi = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ipsi = 0, 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isigma = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isigma = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dchi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dchi > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, h1 = {:s}, h2 = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'h3 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 <= h1 <= h2 <= h3 and h3 > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, c = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: c > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All elements of x are equal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Current estimate of sigma is zero or negative:',
                    'sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Number of iterations required exceeds maxit: maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All winsorized residuals are zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _theta_py = utils._EngineFloat64ScalarType.to_py(_theta)
        _sigma_py = utils._EngineFloat64ScalarType.to_py(_sigma)
        _rs_cls.refresh_py(_rs)
        _wrk_cls.refresh_py(_wrk)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['theta', 'sigma', 'nit'],
        )(
            _theta_py,
            _sigma_py,
            _nit_py,
        )
    _err_class.manage()
    return return_data

def robust_1var_mestim_wgt(chi, psi, isigma, n, x, beta, theta, sigma, maxit, tol, rs, wrk, data=None):
    u"""
    Robust estimation, M-estimates for location and scale parameters,
    user-defined weight functions.

    ``robust_1var_mestim_wgt`` computes an M-estimate of location with
    (optional) simultaneous estimation of scale, where you provide the
    weight functions.

    For full information please refer to the NAG Library document for
    g07dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07dcf.html

    Parameters
    ----------
    chi : callable retval = chi(t, data=None)
        `chi` must return the value of the weight function chi for a
        given value of its argument.

        The value of chi must be non-negative.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `chi` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the weight function phi evaluated at `t`.

    psi : callable retval = psi(t, data=None)
        `psi` must return the value of the weight function psi for a
        given value of its argument.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `psi` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the weight function psi evaluated at `t`.

    isigma : int
        The value assigned to `isigma` determines whether sigma^ is to
        be simultaneously estimated.

        `isigma` = 0

          The estimation of sigma^ is bypassed and `sigma` is set equal
          to sigma_c.

        `isigma` = 1

          sigma^ is estimated simultaneously.

    n : int
        n, the number of observations.

    x : float, shape (`n`)
        The vector of observations, x_1,x_2,...,x_n.

    beta : float
        The value of the constant beta of the chosen `chi` function.

    theta : float
        If `sigma` > 0, `theta` must be set to the required starting
        value of the estimate of the location parameter theta^. A
        reasonable initial value for theta^ will often be the sample
        mean or median.

    sigma : float
        The role of `sigma` depends on the value assigned to `isigma` as
        follows.

        If `isigma` = 1, `sigma` must be assigned a value which
        determines the values of the starting points for the calculation
        of theta^ and sigma^.

        If `sigma` <= 0.0, ``robust_1var_mestim_wgt`` will determine the
        starting points of theta^ and sigma^.

        Otherwise, the value assigned to `sigma` will be taken as the
        starting point for sigma^, and `theta` must be assigned a
        relevant value before entry, see above.

        If `isigma` = 0, `sigma` must be assigned a value which
        determines the values of sigma_c, which is held fixed during the
        iterations, and the starting value for the calculation of
        theta^.

        If `sigma` <= 0, ``robust_1var_mestim_wgt`` will determine the
        value of sigma_c as the median absolute deviation adjusted to
        reduce bias (see ``robust_1var_median``) and the starting point
        for theta.

        Otherwise, the value assigned to `sigma` will be taken as the
        value of sigma_c and `theta` must be assigned a relevant value
        before entry, see above.

    maxit : int
        The maximum number of iterations that should be used during the
        estimation.

    tol : float
        The relative precision for the final estimates. Convergence is
        assumed when the increments for `theta`, and `sigma` are less
        than `tol`*max(1.0, sigma_{k-1}).

    rs : float, shape (`n`), modified in place
        On exit: the Winsorized residuals.

    wrk : float, shape (`n`), modified in place
        On exit: if `sigma` <= 0.0 on entry, `wrk` will contain the n
        observations in ascending order.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    theta : float
        The M-estimate of the location parameter theta^.

    sigma : float
        The M-estimate of the scale parameter sigma^, if `isigma` was
        assigned the value 1 on entry, otherwise `sigma` will contain
        the initial fixed value sigma_c.

    nit : int
        The number of iterations that were used during the estimation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `isigma` = *<value>*.

            Constraint: `isigma` = 0 or 1.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 2)
            On entry, `beta` = *<value>*.

            Constraint: `beta` > 0.0.

        (`errno` 3)
            All elements of `x` are equal.

        (`errno` 4)
            Current estimate of `sigma` is zero or negative: `sigma` =
            *<value>*.

        (`errno` 5)
            Number of iterations required exceeds `maxit`: `maxit` =
            *<value>*.

        (`errno` 6)
            All winsorized residuals are zero.

        (`errno` 7)
            The `chi` function returned a negative value: `chi` =
            *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.robust_1var_mestim_wgt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_chi_name = fun_name + ':chi'

    chih = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_chih(_chi, _chisub, en_data, t, retval, _iuser, _ruser):
        """Helper for chi."""
        try:
            _t_chih = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_chih = chi(
                    _t_chih,
                    data,
                )
            else:
                _retval_chih = chi(
                    _t_chih,
                )
            utils._EngineFloat64ScalarType(
                _retval_chih, {'fun_name': cb_chi_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_psi_name = fun_name + ':psi'

    psih = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_psih(_psi, _psisub, en_data, t, retval, _iuser, _ruser):
        """Helper for psi."""
        try:
            _t_psih = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_psih = psi(
                    _t_psih,
                    data,
                )
            else:
                _retval_psih = psi(
                    _t_psih,
                )
            utils._EngineFloat64ScalarType(
                _retval_psih, {'fun_name': cb_psi_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g07dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        chih,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        psih,
        utils._BYREF_INT, # Input, isigma.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Input, beta.
        utils._BYREF_FLOAT64, # Input/Output, theta.
        utils._BYREF_FLOAT64, # Input/Output, sigma.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Output, rs[n].
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Output, wrk[n].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(chi, {'fun_name': fun_name, 'entity': 'chi'})
    _chi = utils._UNUSED_CALLBACK
    _chisub = utils._UNUSED_CALLBACK
    _chih = utils._cast_callback(
        chih, cb_chih,
        {'fun_name': fun_name, 'entity': 'chi'},
    )
    utils._check_callable(psi, {'fun_name': fun_name, 'entity': 'psi'})
    _psi = utils._UNUSED_CALLBACK
    _psisub = utils._UNUSED_CALLBACK
    _psih = utils._cast_callback(
        psih, cb_psih,
        {'fun_name': fun_name, 'entity': 'psi'},
    )
    _isigma = utils._EngineIntScalarType(
        isigma, {'fun_name': fun_name, 'entity': 'isigma'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _rs_cls = utils._EngineFloat64ArrayType(
        rs, {'fun_name': fun_name, 'entity': 'rs'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _rs = _rs_cls.to_ctypes(can_reference_contents=False)
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wrk_cls = utils._EngineFloat64ArrayType(
        wrk, {'fun_name': fun_name, 'entity': 'wrk'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _wrk = _wrk_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _chi,
        _chisub,
        _chih,
        _psi,
        _psisub,
        _psih,
        _isigma,
        _n,
        _x,
        _beta,
        _theta,
        _sigma,
        _maxit,
        _tol,
        _rs,
        _nit,
        _wrk,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isigma = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isigma = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, beta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: beta > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All elements of x are equal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Current estimate of sigma is zero or negative:',
                    'sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Number of iterations required exceeds maxit: maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All winsorized residuals are zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The chi function returned a negative value: chi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _theta_py = utils._EngineFloat64ScalarType.to_py(_theta)
        _sigma_py = utils._EngineFloat64ScalarType.to_py(_sigma)
        _rs_cls.refresh_py(_rs)
        _wrk_cls.refresh_py(_wrk)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['theta', 'sigma', 'nit'],
        )(
            _theta_py,
            _sigma_py,
            _nit_py,
        )
    _err_class.manage()
    return return_data

def robust_1var_trimmed(n, x, alpha, sx):
    u"""
    Computes a trimmed and winsorized mean of a single sample with
    estimates of their variance.

    ``robust_1var_trimmed`` calculates the trimmed and Winsorized means
    of a sample and estimates of the variances of the two means.

    For full information please refer to the NAG Library document for
    g07dd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07ddf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    x : float, shape (`n`)
        The sample observations, x_i, for i = 1,2,...,n.

    alpha : float
        alpha, the proportion of observations to be trimmed at each end
        of the sorted sample.

    sx : float, shape (`n`), modified in place
        On exit: contains the sample observations sorted into ascending
        order.

    Returns
    -------
    tmean : float
        The alpha-trimmed mean, x__t.

    wmean : float
        The alpha-Winsorized mean, x__w.

    tvar : float
        Contains an estimate of the variance of the trimmed mean.

    wvar : float
        Contains an estimate of the variance of the Winsorized mean.

    k : int
        Contains the number of observations trimmed at each end, k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` <= 1.

        (`errno` 2)
            On entry, `alpha` = *<value>*.

            Constraint: 0.0 <= `alpha` < 0.5.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.robust_1var_trimmed'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07ddft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._BYREF_FLOAT64, # Output, tmean.
        utils._BYREF_FLOAT64, # Output, wmean.
        utils._BYREF_FLOAT64, # Output, tvar.
        utils._BYREF_FLOAT64, # Output, wvar.
        utils._BYREF_INT, # Output, k.
        utils._EngineFloat64ArrayType.c_type, # Output, sx[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _tmean = utils._EngineFloat64ScalarType.empty_ctypes()
    _wmean = utils._EngineFloat64ScalarType.empty_ctypes()
    _tvar = utils._EngineFloat64ScalarType.empty_ctypes()
    _wvar = utils._EngineFloat64ScalarType.empty_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _sx_cls = utils._EngineFloat64ArrayType(
        sx, {'fun_name': fun_name, 'entity': 'sx'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _sx = _sx_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _alpha,
        _tmean,
        _wmean,
        _tvar,
        _wvar,
        _k,
        _sx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n <= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alpha = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= alpha < 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sx_cls.refresh_py(_sx)
        _tmean_py = utils._EngineFloat64ScalarType.to_py(_tmean)
        _wmean_py = utils._EngineFloat64ScalarType.to_py(_wmean)
        _tvar_py = utils._EngineFloat64ScalarType.to_py(_tvar)
        _wvar_py = utils._EngineFloat64ScalarType.to_py(_wvar)
        _k_py = utils._EngineIntScalarType.to_py(_k)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['tmean', 'wmean', 'tvar', 'wvar', 'k'],
        )(
            _tmean_py,
            _wmean_py,
            _tvar_py,
            _wvar_py,
            _k_py,
        )
    _err_class.manage()
    return return_data

def robust_1var_ci(method, n, x, clevel):
    u"""
    Robust confidence intervals, one-sample.

    ``robust_1var_ci`` computes a rank based (nonparametric) estimate
    and confidence interval for the location parameter of a single
    population.

    For full information please refer to the NAG Library document for
    g07ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07eaf.html

    Parameters
    ----------
    method : str, length 1
        Specifies the method to be used.

        `method` = 'E'

          The exact algorithm is used.

        `method` = 'A'

          The iterative algorithm is used.

    n : int
        n, the sample size.

    x : float, shape (`n`)
        The sample observations, x_i, for i = 1,2,...,n.

    clevel : float
        The confidence interval desired.

        For example, for a 95% confidence interval set `clevel` = 0.95.

    Returns
    -------
    theta : float
        The estimate of the location, theta^.

    thetal : float
        The estimate of the lower limit of the confidence interval,
        theta_l.

    thetau : float
        The estimate of the upper limit of the confidence interval,
        theta_u.

    estcl : float
        An estimate of the actual percentage confidence of the interval
        found, as a proportion between (0.0,1.0).

    wlower : float
        The upper value of the Wilcoxon test statistic, W_u,
        corresponding to the lower limit of the confidence interval.

    wupper : float
        The lower value of the Wilcoxon test statistic, W_l,
        corresponding to the upper limit of the confidence interval.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'E' or 'A'.

        (`errno` 2)
            Not enough information to compute an interval estimate since
            the whole sample is identical. The common value is returned
            in `theta`, `thetal` and `thetau`.

        (`errno` 3)
            The iterative procedure used to estimate theta has not
            converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_u, the upper
            confidence limit has not converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_l, the lower
            confidence limit has not converged.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.robust_1var_ci'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Input, clevel.
        utils._BYREF_FLOAT64, # Output, theta.
        utils._BYREF_FLOAT64, # Output, thetal.
        utils._BYREF_FLOAT64, # Output, thetau.
        utils._BYREF_FLOAT64, # Output, estcl.
        utils._BYREF_FLOAT64, # Output, wlower.
        utils._BYREF_FLOAT64, # Output, wupper.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wrk.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _clevel = utils._EngineFloat64ScalarType(
        clevel, {'fun_name': fun_name, 'entity': 'clevel'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType.empty_ctypes()
    _thetal = utils._EngineFloat64ScalarType.empty_ctypes()
    _thetau = utils._EngineFloat64ScalarType.empty_ctypes()
    _estcl = utils._EngineFloat64ScalarType.empty_ctypes()
    _wlower = utils._EngineFloat64ScalarType.empty_ctypes()
    _wupper = utils._EngineFloat64ScalarType.empty_ctypes()
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _x,
        _clevel,
        _theta,
        _thetal,
        _thetau,
        _estcl,
        _wlower,
        _wupper,
        _wrk,
        _iwrk,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, clevel = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < clevel < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'E\' or \'A\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Not enough information to compute an interval estimate since the',
                    'whole sample is identical. The common value is returned in theta, thetal',
                    'and thetau.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure used to estimate theta has not converged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure used to estimate, theta_u,',
                    'the upper confidence limit has not converged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure used to estimate, theta_l,',
                    'the lower confidence limit has not converged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _theta_py = utils._EngineFloat64ScalarType.to_py(_theta)
        _thetal_py = utils._EngineFloat64ScalarType.to_py(_thetal)
        _thetau_py = utils._EngineFloat64ScalarType.to_py(_thetau)
        _estcl_py = utils._EngineFloat64ScalarType.to_py(_estcl)
        _wlower_py = utils._EngineFloat64ScalarType.to_py(_wlower)
        _wupper_py = utils._EngineFloat64ScalarType.to_py(_wupper)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['theta', 'thetal', 'thetau', 'estcl', 'wlower', 'wupper'],
        )(
            _theta_py,
            _thetal_py,
            _thetau_py,
            _estcl_py,
            _wlower_py,
            _wupper_py,
        )
    _err_class.manage()
    return return_data

def robust_2var_ci(method, n, x, m, y, clevel):
    u"""
    Robust confidence intervals, two-sample.

    ``robust_2var_ci`` calculates a rank based (nonparametric) estimate
    and confidence interval for the difference in location between two
    independent populations.

    For full information please refer to the NAG Library document for
    g07eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07ebf.html

    Parameters
    ----------
    method : str, length 1
        Specifies the method to be used.

        `method` = 'E'

          The exact algorithm is used.

        `method` = 'A'

          The iterative algorithm is used.

    n : int
        n, the size of the first sample.

    x : float, shape (`n`)
        The observations of the first sample, x_i, for i = 1,2,...,n.

    m : int
        m, the size of the second sample.

    y : float, shape (`m`)
        The observations of the second sample, y_j, for j = 1,2,...,m.

    clevel : float
        The confidence interval required, 1-alpha; e.g., for a 95%
        confidence interval set `clevel` = 0.95.

    Returns
    -------
    theta : float
        The estimate of the difference in the location of the two
        populations, theta^.

    thetal : float
        The estimate of the lower limit of the confidence interval,
        theta_l.

    thetau : float
        The estimate of the upper limit of the confidence interval,
        theta_u.

    estcl : float
        An estimate of the actual percentage confidence of the interval
        found, as a proportion between (0.0,1.0).

    ulower : float
        The value of the Mann--Whitney U statistic corresponding to the
        lower confidence limit, U_l.

    uupper : float
        The value of the Mann--Whitney U statistic corresponding to the
        upper confidence limit, U_u.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'E' or 'A'.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Not enough information to compute an interval estimate since
            each sample has identical values. The common difference is
            returned in `theta`, `thetal` and `thetau`.

        (`errno` 3)
            The iterative procedure used to estimate theta has not
            converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_u, the upper
            confidence limit has not converged.

        (`errno` 3)
            The iterative procedure used to estimate, theta_l, the lower
            confidence limit has not converged.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.robust_2var_ci'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07ebft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._BYREF_FLOAT64, # Input, clevel.
        utils._BYREF_FLOAT64, # Output, theta.
        utils._BYREF_FLOAT64, # Output, thetal.
        utils._BYREF_FLOAT64, # Output, thetau.
        utils._BYREF_FLOAT64, # Output, estcl.
        utils._BYREF_FLOAT64, # Output, ulower.
        utils._BYREF_FLOAT64, # Output, uupper.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wrk.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _clevel = utils._EngineFloat64ScalarType(
        clevel, {'fun_name': fun_name, 'entity': 'clevel'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType.empty_ctypes()
    _thetal = utils._EngineFloat64ScalarType.empty_ctypes()
    _thetau = utils._EngineFloat64ScalarType.empty_ctypes()
    _estcl = utils._EngineFloat64ScalarType.empty_ctypes()
    _ulower = utils._EngineFloat64ScalarType.empty_ctypes()
    _uupper = utils._EngineFloat64ScalarType.empty_ctypes()
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _x,
        _m,
        _y,
        _clevel,
        _theta,
        _thetal,
        _thetau,
        _estcl,
        _ulower,
        _uupper,
        _wrk,
        _iwrk,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99908:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT.format(_m_py),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, clevel = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < clevel < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'E\' or \'A\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Not enough information to compute an interval estimate since each',
                    'sample has identical values. The common difference is returned in',
                    'theta, thetal and thetau.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure used to estimate theta has not converged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure used to estimate, theta_u,',
                    'the upper confidence limit has not converged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure used to estimate, theta_l,',
                    'the lower confidence limit has not converged.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _theta_py = utils._EngineFloat64ScalarType.to_py(_theta)
        _thetal_py = utils._EngineFloat64ScalarType.to_py(_thetal)
        _thetau_py = utils._EngineFloat64ScalarType.to_py(_thetau)
        _estcl_py = utils._EngineFloat64ScalarType.to_py(_estcl)
        _ulower_py = utils._EngineFloat64ScalarType.to_py(_ulower)
        _uupper_py = utils._EngineFloat64ScalarType.to_py(_uupper)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['theta', 'thetal', 'thetau', 'estcl', 'ulower', 'uupper'],
        )(
            _theta_py,
            _thetal_py,
            _thetau_py,
            _estcl_py,
            _ulower_py,
            _uupper_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def outlier_peirce_1var(n, p, y, mean, var, iout, ldiff, diff, llamb):
    u"""
    Outlier detection using method of Peirce, raw data or single
    variance supplied.

    ``outlier_peirce_1var`` identifies outlying values using Peirce's
    criterion.

    For full information please refer to the NAG Library document for
    g07ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07gaf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    p : int
        p, the number of parameters in the model used in obtaining the
        y. If y is an observed set of values, as opposed to the
        residuals from fitting a model with p parameters, then p should
        be set to 1, i.e., as if a model just containing the mean had
        been used.

    y : float, shape (`n`)
        y, the data being tested.

    mean : float
        If `var` > 0.0, `mean` must contain mu, the mean of y, otherwise
        `mean` is not referenced and the mean is calculated from the
        data supplied in `y`.

    var : float
        If `var` > 0.0, `var` must contain sigma^2, the variance of y,
        otherwise the variance is calculated from the data supplied in
        `y`.

    iout : int, shape (`n`), modified in place
        On exit: the indices of the values in `y` sorted in descending
        order of the absolute difference from the mean, therefore |`y`[
        `iout`[i-2] -1]-mu| >= |`y`[ `iout`[i-1] -1]-mu|, for i =
        2,3,...,`n`.

    ldiff : int
        The maximum number of values to be returned in arrays `diff` and
        `llamb`.

        If `ldiff` <= 0, arrays `diff` and `llamb` are not referenced.

    diff : None or float, shape (`ldiff`), modified in place
        On exit, if not None on entry: `diff`[i-1] holds |y-mu|-sigma^2z
        for observation `y`[`iout`[i-1]-1], for i = 1,2,...,min(`ldiff`,
        `niout`+1, `n`-`p`-1).

    llamb : None or float, shape (`ldiff`), modified in place
        On exit, if not None on entry: `llamb`[i-1] holds log(lambda^2)
        for observation `y`[`iout`[i-1]-1], for i = 1,2,...,min(`ldiff`,
        `niout`+1, `n`-`p`-1).

    Returns
    -------
    niout : int
        The number of potential outliers. The indices for these
        potential outliers are held in the first `niout` elements of
        `iout`. By construction there can be at most `n`-`p`-1 values
        flagged as outliers.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 2)
            On entry, `p` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `p` <= `n`-2.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.outlier_peirce_1var'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07gaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, p.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_FLOAT64, # Input, mean.
        utils._BYREF_FLOAT64, # Input, var.
        utils._EngineIntArrayType.c_type, # Output, iout[n].
        utils._BYREF_INT, # Output, niout.
        utils._BYREF_INT, # Input, ldiff.
        utils._EngineFloat64ArrayType.c_type, # Output, diff[ldiff].
        utils._EngineFloat64ArrayType.c_type, # Output, llamb[ldiff].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, initr.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _p = utils._EngineIntScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _mean = utils._EngineFloat64ScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'},
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    _iout_cls = utils._EngineIntArrayType(
        iout, {'fun_name': fun_name, 'entity': 'iout'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _iout = _iout_cls.to_ctypes(can_reference_contents=False)
    _niout = utils._EngineIntScalarType.empty_ctypes()
    _ldiff = utils._EngineIntScalarType(
        ldiff, {'fun_name': fun_name, 'entity': 'ldiff'},
    ).to_ctypes()
    _ldiff_py = utils._EngineIntScalarType.to_py(_ldiff)
    if _ldiff_py > 0:
        if diff is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'diff'},
                'must not be None when ldiff > 0',
            )
    _diff_cls = utils._EngineFloat64ArrayType(
        diff, {'fun_name': fun_name, 'entity': 'diff'},
        exp_shape=(_ldiff_py,),
        check_contents=False,
        allow_none=True,
    )
    _diff = _diff_cls.to_ctypes(can_reference_contents=False)
    if _ldiff_py > 0:
        if llamb is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'llamb'},
                'must not be None when ldiff > 0',
            )
    _llamb_cls = utils._EngineFloat64ArrayType(
        llamb, {'fun_name': fun_name, 'entity': 'llamb'},
        exp_shape=(_ldiff_py,),
        check_contents=False,
        allow_none=True,
    )
    _llamb = _llamb_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType.c_type(0.0)
    _maxit = utils._EngineIntScalarType.c_type(0)
    _initr = utils._EngineFloat64ScalarType.c_type(0.0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _p,
        _y,
        _mean,
        _var,
        _iout,
        _niout,
        _ldiff,
        _diff,
        _llamb,
        _tol,
        _maxit,
        _initr,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= p <= n-2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iout_cls.refresh_py(_iout)
        if diff is not None:
            _diff_cls.refresh_py(_diff)
        if llamb is not None:
            _llamb_cls.refresh_py(_llamb)
        _niout_py = utils._EngineIntScalarType.to_py(_niout)
        return_data = _niout_py
    _err_class.manage()
    return return_data

def outlier_peirce_2var(n, e, var1, var2):
    u"""
    Outlier detection using method of Peirce, two variances supplied.

    ``outlier_peirce_2var`` returns a flag indicating whether a single
    data point is an outlier as defined by Peirce's criterion.

    For full information please refer to the NAG Library document for
    g07gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g07/g07gbf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    e : float
        e~, the value being tested.

    var1 : float
        sigma^2, the residual variance on fitting model M to y.

    var2 : float
        sigma~^2, the residual variance on fitting model M to y~.

    Returns
    -------
    outl : bool
        True if the value being tested is an outlier.

    x : float
        An estimated value of x, the cutoff that indicates an outlier.

    lx : float
        l, the lower limit for x.

    ux : float
        u, the upper limit for x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 3)
            On entry, `var1` = *<value>*.

            Constraint: `var1` > 0.0.

        (`errno` 4)
            On entry, `var2` = *<value>*.

            Constraint: `var2` > 0.0.

        (`errno` 4)
            On entry, `var1` = *<value>*, `var2` = *<value>*.

            Constraint: `var2` < `var1`.
    """

    excs = []
    fun_name = 'naginterfaces.base.univar.outlier_peirce_2var'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g07gbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, e.
        utils._BYREF_FLOAT64, # Input, var1.
        utils._BYREF_FLOAT64, # Input, var2.
        utils._BYREF_FLOAT64, # Output, x.
        utils._BYREF_FLOAT64, # Output, lx.
        utils._BYREF_FLOAT64, # Output, ux.
        utils._BYREF_INT, # Input, mncall.
        utils._BYREF_INT, # Input, niter.
        utils._BYREF_INT, # Output, ioutl.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _e = utils._EngineFloat64ScalarType(
        e, {'fun_name': fun_name, 'entity': 'e'},
    ).to_ctypes()
    _var1 = utils._EngineFloat64ScalarType(
        var1, {'fun_name': fun_name, 'entity': 'var1'},
    ).to_ctypes()
    _var2 = utils._EngineFloat64ScalarType(
        var2, {'fun_name': fun_name, 'entity': 'var2'},
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType.empty_ctypes()
    _lx = utils._EngineFloat64ScalarType.empty_ctypes()
    _ux = utils._EngineFloat64ScalarType.empty_ctypes()
    _mncall = utils._EngineIntScalarType.c_type(0)
    _niter = utils._EngineIntScalarType.c_type(0)
    _ioutl = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _e,
        _var1,
        _var2,
        _x,
        _lx,
        _ux,
        _mncall,
        _niter,
        _ioutl,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var1 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var1 > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var2 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var2 > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var1 = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'var2 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: var2 < var1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        # pylint: disable=simplifiable-if-statement
        _ioutl_py = utils._EngineIntScalarType.to_py(_ioutl)
        if _ioutl_py == 0:
            _outl_py = False
        elif _ioutl_py == 1:
            _outl_py = True
        else:
            raise utils.NagException(
                {'fun_name': fun_name},
                'Unexpected value for internal ioutl.',
            )
        _x_py = utils._EngineFloat64ScalarType.to_py(_x)
        _lx_py = utils._EngineFloat64ScalarType.to_py(_lx)
        _ux_py = utils._EngineFloat64ScalarType.to_py(_ux)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['outl', 'x', 'lx', 'ux'],
        )(
            _outl_py,
            _x_py,
            _lx_py,
            _ux_py,
        )
    _err_class.manage()
    return return_data
