# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `zeros` Chapter.

``zeros`` - Zeros of Polynomials

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with computing the zeros of a polynomial with
real or complex coefficients.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02intro.html
"""

# NAG Copyright 2017-2019.

import sys as _sys

from ..base import utils

def poly_complex(a, n, scal, z):
    u"""
    All zeros of complex polynomial, modified Laguerre's method.

    ``poly_complex`` finds all the roots of a complex polynomial
    equation, using a variant of Laguerre's method.

    For full information please refer to the NAG Library document for
    c02af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02aff.html

    Parameters
    ----------
    a : float, shape (2, `n`+1)
        If `a` is declared with bounds (2,0:`n`), `a`[0,i] and `a`[1,i]
        must contain the real and imaginary parts of a_i (i.e., the
        coefficient of z^(n-i)), for i = 0,1,...,n.

    n : int
        n, the degree of the polynomial.

    scal : bool
        Indicates whether or not the polynomial is to be scaled. See
        [Further Comments] for advice on when it may be preferable to
        set `scal` = False and for a description of the scaling
        strategy.

    z : float, shape (2, `n`), modified in place
        On exit: the real and imaginary parts of the roots are stored in
        `z`[0,i-1] and `z`[1,i-1] respectively, for i = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a`[0,0] = 0.0 and `a`[1,0] = 0.0.

            Constraint: `a`[0,0] != 0.0 or `a`[1,0] != 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            The iterative procedure has failed to converge. This error
            is very unlikely to occur. If it does, please contact NAG
            immediately, as some basic assumption for the arithmetic has
            been violated.

        (`errno` 3)
            ``poly_complex`` cannot evaluate P(z) near some of its zeros
            without underflow. If this message occurs please contact
            NAG.

        (`errno` 3)
            ``poly_complex`` cannot evaluate P(z) near some of its zeros
            without overflow. If this message occurs please contact NAG.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.zeros.poly_complex_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.poly_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, a[2, n+1].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_LOGICAL, # Input, scal.
        utils._EngineFloat64ArrayType.c_type, # Output, z[2, n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, w.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (z, 'z')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(2, _n_py+1),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _scal = utils._EngineLogicalScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'},
    ).to_ctypes()
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(2, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _z = _z_cls.to_ctypes(can_reference_contents=False)
    _w = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _n,
        _scal,
        _z,
        _w,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[0,0] = 0.0 and a[1,0] = 0.0.',
                    'Constraint: a[0,0] != 0.0 or a[1,0] != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure has failed to converge.',
                    'This error is very unlikely to occur. If it does, please contact NAG',
                    'immediately, as some basic assumption for the arithmetic has been violated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'poly_complex cannot evaluate P(z) near some of its zeros without underflow.',
                    'If this message occurs please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'poly_complex cannot evaluate P(z) near some of its zeros without overflow.',
                    'If this message occurs please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        return_data = None
    _err_class.manage()
    return return_data

def poly_real(a, n, scal, z):
    u"""
    All zeros of real polynomial, modified Laguerre's method.

    ``poly_real`` finds all the roots of a real polynomial equation,
    using a variant of Laguerre's method.

    For full information please refer to the NAG Library document for
    c02ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02agf.html

    Parameters
    ----------
    a : float, shape (`n`+1)
        `a`[i] must contain a_i (i.e., the coefficient of z^(n-i)), for
        i = 0,1,...,n.

    n : int
        n, the degree of the polynomial.

    scal : bool
        Indicates whether or not the polynomial is to be scaled. See
        [Further Comments] for advice on when it may be preferable to
        set `scal` = False and for a description of the scaling
        strategy.

    z : float, shape (2, `n`), modified in place
        On exit: the real and imaginary parts of the roots are stored in
        `z`[0,i-1] and `z`[1,i-1] respectively, for i = 1,2,...,n.
        Complex conjugate pairs of roots are stored in consecutive pairs
        of elements of `z`; that is, `z`[0,i] = `z`[0,i-1] and `z`[1,i]
        = -`z`[1,i-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a`[0] = 0.0.

            Constraint: `a`[0] != 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            The iterative procedure has failed to converge. This error
            is very unlikely to occur. If it does, please contact NAG
            immediately, as some basic assumption for the arithmetic has
            been violated.

        (`errno` 3)
            ``poly_real`` cannot evaluate P(z) near some of its zeros
            without underflow. If this message occurs please contact
            NAG.

        (`errno` 3)
            ``poly_real`` cannot evaluate P(z) near some of its zeros
            without overflow. If this message occurs please contact NAG.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.poly_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, a[n+1].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_LOGICAL, # Input, scal.
        utils._EngineFloat64ArrayType.c_type, # Output, z[2, n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, w.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _scal = utils._EngineLogicalScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'},
    ).to_ctypes()
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(2, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _z = _z_cls.to_ctypes(can_reference_contents=False)
    _w = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _n,
        _scal,
        _z,
        _w,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[0] = 0.0.',
                    'Constraint: a[0] != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure has failed to converge.',
                    'This error is very unlikely to occur. If it does, please contact NAG',
                    'immediately, as some basic assumption for the arithmetic has been violated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'poly_real cannot evaluate P(z) near some of its zeros without underflow.',
                    'If this message occurs please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'poly_real cannot evaluate P(z) near some of its zeros without overflow.',
                    'If this message occurs please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        return_data = None
    _err_class.manage()
    return return_data

def quadratic_complex(ar, ai, br, bi, cr, ci, zsm, zlg):
    u"""
    All zeros of complex quadratic equation.

    ``quadratic_complex`` determines the roots of a quadratic equation
    with complex coefficients.

    For full information please refer to the NAG Library document for
    c02ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02ahf.html

    Parameters
    ----------
    ar : float
        The real part of a.

    ai : float
        The imaginary part of a.

    br : float
        The real part of b.

    bi : float
        The imaginary part of b.

    cr : float
        The real part of c.

    ci : float
        The imaginary part of c.

    zsm : float, shape (2), modified in place
        On exit: the real and imaginary parts of the smallest root in
        magnitude are stored in `zsm`[0] and `zsm`[1] respectively.

    zlg : float, shape (2), modified in place
        On exit: the real and imaginary parts of the largest root in
        magnitude are stored in `zlg`[0] and `zlg`[1] respectively.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, (`ar`,`ai`) = (0,0).

        (`errno` 2)
            On entry, (`ar`,`ai`) = (0,0) and (`br`,`bi`) = (0,0).

        (`errno` 3)
            On entry, (`ar`,`ai`) = (0,0) and the root
            -(`cr`,`ci`)/(`br`,`bi`) overflows: `ar` = *<value>*, `cr` =
            *<value>*, `br` = *<value>*, `ai` = *<value>*, `ci` =
            *<value>* and `bi` = *<value>*.

        (`errno` 4)
            On entry, (`cr`,`ci`) = (0,0) and the root
            -(`br`,`bi`)/(`ar`,`ai`) overflows: `cr` = *<value>*, `br` =
            *<value>*, `ar` = *<value>*, `ci` = *<value>*, `bi` =
            *<value>* and `ai` = *<value>*.

        (`errno` 5)
            On entry, B is so large that B^2 is indistinguishable from
            (B^2-4*A*C) and the root -(`br`,`bi`)/(`ar`,`ai`) overflows:
            B = max(abs(`br`), abs(`bi`)) = *<value>*, A =
            max(abs(`ar`), abs(`ai`)) = *<value>*, C = max(abs(`cr`),
            abs(`ci`)) = *<value>*, `br` = *<value>*, `bi` = *<value>*,
            `ar` = *<value>* and `ai` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.quadratic_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, ar.
        utils._BYREF_FLOAT64, # Input, ai.
        utils._BYREF_FLOAT64, # Input, br.
        utils._BYREF_FLOAT64, # Input, bi.
        utils._BYREF_FLOAT64, # Input, cr.
        utils._BYREF_FLOAT64, # Input, ci.
        utils._EngineFloat64ArrayType.c_type, # Output, zsm[2].
        utils._EngineFloat64ArrayType.c_type, # Output, zlg[2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ar = utils._EngineFloat64ScalarType(
        ar, {'fun_name': fun_name, 'entity': 'ar'},
    ).to_ctypes()
    _ai = utils._EngineFloat64ScalarType(
        ai, {'fun_name': fun_name, 'entity': 'ai'},
    ).to_ctypes()
    _br = utils._EngineFloat64ScalarType(
        br, {'fun_name': fun_name, 'entity': 'br'},
    ).to_ctypes()
    _bi = utils._EngineFloat64ScalarType(
        bi, {'fun_name': fun_name, 'entity': 'bi'},
    ).to_ctypes()
    _cr = utils._EngineFloat64ScalarType(
        cr, {'fun_name': fun_name, 'entity': 'cr'},
    ).to_ctypes()
    _ci = utils._EngineFloat64ScalarType(
        ci, {'fun_name': fun_name, 'entity': 'ci'},
    ).to_ctypes()
    _zsm_cls = utils._EngineFloat64ArrayType(
        zsm, {'fun_name': fun_name, 'entity': 'zsm'},
        exp_shape=(2,),
        check_contents=False,
    )
    _zsm = _zsm_cls.to_ctypes(can_reference_contents=False)
    _zlg_cls = utils._EngineFloat64ArrayType(
        zlg, {'fun_name': fun_name, 'entity': 'zlg'},
        exp_shape=(2,),
        check_contents=False,
    )
    _zlg = _zlg_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ar,
        _ai,
        _br,
        _bi,
        _cr,
        _ci,
        _zsm,
        _zlg,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, (ar,ai) = (0,0).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, (ar,ai) = (0,0) and (br,bi) = (0,0).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, (ar,ai) = (0,0) and the root -(cr,ci)/(br,bi) overflows:',
                    'ar = {:s}, cr = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'br = {:s}, ai = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'ci = {:s} and bi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[5]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[6]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, (cr,ci) = (0,0) and the root -(br,bi)/(ar,ai) overflows:',
                    'cr = {:s}, br = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'ar = {:s}, ci = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'bi = {:s} and ai = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[5]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[6]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, B is so large that B^2 is indistinguishable',
                    'from (B^2-4*A*C) and the root -(br,bi)/(ar,ai) overflows:',
                    'B = max(abs(br), abs(bi)) = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'A = max(abs(ar), abs(ai)) = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'C = max(abs(cr), abs(ci)) = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'br = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'bi = {:s},'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[5]),
                    ),
                    'ar = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[6]),
                    ),
                    'ai = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[7]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _zsm_cls.refresh_py(_zsm)
        _zlg_cls.refresh_py(_zlg)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def quadratic_real(a, b, c, zsm, zlg):
    u"""
    All zeros of real quadratic equation.

    ``quadratic_real`` determines the roots of a quadratic equation with
    real coefficients.

    For full information please refer to the NAG Library document for
    c02aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02ajf.html

    Parameters
    ----------
    a : float
        Must contain a, the coefficient of z^2.

    b : float
        Must contain b, the coefficient of z.

    c : float
        Must contain c, the constant coefficient.

    zsm : float, shape (2), modified in place
        On exit: the real and imaginary parts of the smallest root in
        magnitude are stored in `zsm`[0] and `zsm`[1] respectively.

    zlg : float, shape (2), modified in place
        On exit: the real and imaginary parts of the largest root in
        magnitude are stored in `zlg`[0] and `zlg`[1] respectively.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `a` = 0.0.

        (`errno` 2)
            On entry, `a` = 0.0 and `b` = 0.0.

        (`errno` 3)
            On entry, `a` = 0.0 and the root -`c`/`b` overflows: `a` =
            *<value>*, `c` = *<value>* and `b` = *<value>*.

        (`errno` 4)
            On entry, `c` = 0.0 and the root -`b`/`a` overflows: `c` =
            *<value>*, `b` = *<value>* and `a` = *<value>*.

        (`errno` 5)
            On entry, `b` is so large that `b`^2 is indistinguishable
            from (`b`^2-4*`a`*`c`) and the root -`b`/`a` overflows: `b`
            = *<value>*, `a` = *<value>* and `c` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.quadratic_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02ajft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Input, c.
        utils._EngineFloat64ArrayType.c_type, # Output, zsm[2].
        utils._EngineFloat64ArrayType.c_type, # Output, zlg[2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _zsm_cls = utils._EngineFloat64ArrayType(
        zsm, {'fun_name': fun_name, 'entity': 'zsm'},
        exp_shape=(2,),
        check_contents=False,
    )
    _zsm = _zsm_cls.to_ctypes(can_reference_contents=False)
    _zlg_cls = utils._EngineFloat64ArrayType(
        zlg, {'fun_name': fun_name, 'entity': 'zlg'},
        exp_shape=(2,),
        check_contents=False,
    )
    _zlg = _zlg_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _c,
        _zsm,
        _zlg,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = 0.0 and b = 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = 0.0 and the root -c/b overflows:',
                    'a = {:s}, c = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, c = 0.0 and the root -b/a overflows:',
                    'c = {:s}, b = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b is so large that b^2 is indistinguishable',
                    'from (b^2-4*a*c) and the root -b/a overflows:',
                    'b = {:s}, a = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'c = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _zsm_cls.refresh_py(_zsm)
        _zlg_cls.refresh_py(_zlg)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def cubic_real(u, r, s, t, zeror, zeroi, errest):
    u"""
    All zeros of real cubic equation.

    ``cubic_real`` determines the roots of a cubic equation with real
    coefficients.

    For full information please refer to the NAG Library document for
    c02ak

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02akf.html

    Parameters
    ----------
    u : float
        u, the coefficient of z^3.

    r : float
        r, the coefficient of z^2.

    s : float
        s, the coefficient of z.

    t : float
        t, the constant coefficient.

    zeror : float, shape (3), modified in place
        `zeror`[i-1] contains the real part of the ith root.

    zeroi : float, shape (3), modified in place
        `zeroi`[i-1] contains the imaginary part of the ith root.

    errest : float, shape (3), modified in place
        On exit: `errest`[i-1] contains an approximate error estimate
        for the ith root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `u` = 0.0.

            Constraint: `u` != 0.0.

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.dhseqr``.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.cubic_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02akft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, u.
        utils._BYREF_FLOAT64, # Input, r.
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, t.
        utils._EngineFloat64ArrayType.c_type, # Output, zeror[3].
        utils._EngineFloat64ArrayType.c_type, # Output, zeroi[3].
        utils._EngineFloat64ArrayType.c_type, # Output, errest[3].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _u = utils._EngineFloat64ScalarType(
        u, {'fun_name': fun_name, 'entity': 'u'},
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _t = utils._EngineFloat64ScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _zeror_cls = utils._EngineFloat64ArrayType(
        zeror, {'fun_name': fun_name, 'entity': 'zeror'},
        exp_shape=(3,),
        check_contents=False,
    )
    _zeror = _zeror_cls.to_ctypes(can_reference_contents=False)
    _zeroi_cls = utils._EngineFloat64ArrayType(
        zeroi, {'fun_name': fun_name, 'entity': 'zeroi'},
        exp_shape=(3,),
        check_contents=False,
    )
    _zeroi = _zeroi_cls.to_ctypes(can_reference_contents=False)
    _errest_cls = utils._EngineFloat64ArrayType(
        errest, {'fun_name': fun_name, 'entity': 'errest'},
        exp_shape=(3,),
        check_contents=False,
    )
    _errest = _errest_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _u,
        _r,
        _s,
        _t,
        _zeror,
        _zeroi,
        _errest,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, u = 0.0.',
                    'Constraint: u != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The companion matrix H cannot be formed without overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to converge in lapackeig.dhseqr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _zeror_cls.refresh_py(_zeror)
        _zeroi_cls.refresh_py(_zeroi)
        _errest_cls.refresh_py(_errest)
        return_data = None
    _err_class.manage()
    return return_data

def quartic_real(e, a, b, c, d, zeror, zeroi, errest):
    u"""
    All zeros of real quartic equation.

    ``quartic_real`` determines the roots of a quartic equation with
    real coefficients.

    For full information please refer to the NAG Library document for
    c02al

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02alf.html

    Parameters
    ----------
    e : float
        e, the coefficient of z^4.

    a : float
        a, the coefficient of z^3.

    b : float
        b, the coefficient of z^2.

    c : float
        c, the coefficient of z.

    d : float
        d, the constant coefficient.

    zeror : float, shape (4), modified in place
        `zeror`[i-1] contains the real part of the ith root.

    zeroi : float, shape (4), modified in place
        `zeroi`[i-1] contains the imaginary part of the ith root.

    errest : float, shape (4), modified in place
        On exit: `errest`[i-1] contains an approximate error estimate
        for the ith root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `e` = 0.0.

            Constraint: `e` != 0.0.

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.dhseqr``.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.quartic_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02alft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, e.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, d.
        utils._EngineFloat64ArrayType.c_type, # Output, zeror[4].
        utils._EngineFloat64ArrayType.c_type, # Output, zeroi[4].
        utils._EngineFloat64ArrayType.c_type, # Output, errest[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _e = utils._EngineFloat64ScalarType(
        e, {'fun_name': fun_name, 'entity': 'e'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _d = utils._EngineFloat64ScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _zeror_cls = utils._EngineFloat64ArrayType(
        zeror, {'fun_name': fun_name, 'entity': 'zeror'},
        exp_shape=(4,),
        check_contents=False,
    )
    _zeror = _zeror_cls.to_ctypes(can_reference_contents=False)
    _zeroi_cls = utils._EngineFloat64ArrayType(
        zeroi, {'fun_name': fun_name, 'entity': 'zeroi'},
        exp_shape=(4,),
        check_contents=False,
    )
    _zeroi = _zeroi_cls.to_ctypes(can_reference_contents=False)
    _errest_cls = utils._EngineFloat64ArrayType(
        errest, {'fun_name': fun_name, 'entity': 'errest'},
        exp_shape=(4,),
        check_contents=False,
    )
    _errest = _errest_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _e,
        _a,
        _b,
        _c,
        _d,
        _zeror,
        _zeroi,
        _errest,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, e = 0.0.',
                    'Constraint: e != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The companion matrix H cannot be formed without overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to converge in lapackeig.dhseqr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _zeror_cls.refresh_py(_zeror)
        _zeroi_cls.refresh_py(_zeroi)
        _errest_cls.refresh_py(_errest)
        return_data = None
    _err_class.manage()
    return return_data

def cubic_complex(u, r, s, t, zeror, zeroi, errest):
    u"""
    All zeros of complex cubic equation.

    ``cubic_complex`` determines the roots of a cubic equation with
    complex coefficients.

    For full information please refer to the NAG Library document for
    c02am

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02amf.html

    Parameters
    ----------
    u : complex
        u, the coefficient of z^3.

    r : complex
        r, the coefficient of z^2.

    s : complex
        s, the coefficient of z.

    t : complex
        t, the constant coefficient.

    zeror : float, shape (3), modified in place
        `zeror`[i-1] contains the real part of the ith root.

    zeroi : float, shape (3), modified in place
        `zeroi`[i-1] contains the imaginary part of the ith root.

    errest : float, shape (3), modified in place
        On exit: `errest`[i-1] contains an approximate error estimate
        for the ith root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `u` = (0.0,0.0).

            Constraint: `u` != (0.0,0.0)

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.zhseqr``.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.cubic_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02amft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, u.
        utils._BYREF_COMPLEX128, # Input, r.
        utils._BYREF_COMPLEX128, # Input, s.
        utils._BYREF_COMPLEX128, # Input, t.
        utils._EngineFloat64ArrayType.c_type, # Output, zeror[3].
        utils._EngineFloat64ArrayType.c_type, # Output, zeroi[3].
        utils._EngineFloat64ArrayType.c_type, # Output, errest[3].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _u = utils._EngineComplex128ScalarType(
        u, {'fun_name': fun_name, 'entity': 'u'},
    ).to_ctypes()
    _r = utils._EngineComplex128ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    _s = utils._EngineComplex128ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _t = utils._EngineComplex128ScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _zeror_cls = utils._EngineFloat64ArrayType(
        zeror, {'fun_name': fun_name, 'entity': 'zeror'},
        exp_shape=(3,),
        check_contents=False,
    )
    _zeror = _zeror_cls.to_ctypes(can_reference_contents=False)
    _zeroi_cls = utils._EngineFloat64ArrayType(
        zeroi, {'fun_name': fun_name, 'entity': 'zeroi'},
        exp_shape=(3,),
        check_contents=False,
    )
    _zeroi = _zeroi_cls.to_ctypes(can_reference_contents=False)
    _errest_cls = utils._EngineFloat64ArrayType(
        errest, {'fun_name': fun_name, 'entity': 'errest'},
        exp_shape=(3,),
        check_contents=False,
    )
    _errest = _errest_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _u,
        _r,
        _s,
        _t,
        _zeror,
        _zeroi,
        _errest,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, u = (0.0,0.0).',
                    'Constraint: u != (0.0,0.0)',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The companion matrix H cannot be formed without overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to converge in lapackeig.zhseqr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _zeror_cls.refresh_py(_zeror)
        _zeroi_cls.refresh_py(_zeroi)
        _errest_cls.refresh_py(_errest)
        return_data = None
    _err_class.manage()
    return return_data

def quartic_complex(e, a, b, c, d, zeror, zeroi, errest):
    u"""
    All zeros of complex quartic equation.

    ``quartic_complex`` determines the roots of a quartic equation with
    complex coefficients.

    For full information please refer to the NAG Library document for
    c02an

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02anf.html

    Parameters
    ----------
    e : complex
        e, the coefficient of z^4.

    a : complex
        a, the coefficient of z^3.

    b : complex
        b, the coefficient of z^2.

    c : complex
        c, the coefficient of z.

    d : complex
        d, the constant coefficient.

    zeror : float, shape (4), modified in place
        `zeror`[i-1] contains the real part of the ith root.

    zeroi : float, shape (4), modified in place
        `zeroi`[i-1] contains the imaginary part of the ith root.

    errest : float, shape (4), modified in place
        On exit: `errest`[i-1] contains an approximate error estimate
        for the ith root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `e` = (0.0,0.0).

            Constraint: `e` != (0.0,0.0)

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.zhseqr``.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.zeros.quartic_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c02anft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, e.
        utils._BYREF_COMPLEX128, # Input, a.
        utils._BYREF_COMPLEX128, # Input, b.
        utils._BYREF_COMPLEX128, # Input, c.
        utils._BYREF_COMPLEX128, # Input, d.
        utils._EngineFloat64ArrayType.c_type, # Output, zeror[4].
        utils._EngineFloat64ArrayType.c_type, # Output, zeroi[4].
        utils._EngineFloat64ArrayType.c_type, # Output, errest[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _e = utils._EngineComplex128ScalarType(
        e, {'fun_name': fun_name, 'entity': 'e'},
    ).to_ctypes()
    _a = utils._EngineComplex128ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineComplex128ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineComplex128ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _d = utils._EngineComplex128ScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _zeror_cls = utils._EngineFloat64ArrayType(
        zeror, {'fun_name': fun_name, 'entity': 'zeror'},
        exp_shape=(4,),
        check_contents=False,
    )
    _zeror = _zeror_cls.to_ctypes(can_reference_contents=False)
    _zeroi_cls = utils._EngineFloat64ArrayType(
        zeroi, {'fun_name': fun_name, 'entity': 'zeroi'},
        exp_shape=(4,),
        check_contents=False,
    )
    _zeroi = _zeroi_cls.to_ctypes(can_reference_contents=False)
    _errest_cls = utils._EngineFloat64ArrayType(
        errest, {'fun_name': fun_name, 'entity': 'errest'},
        exp_shape=(4,),
        check_contents=False,
    )
    _errest = _errest_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _e,
        _a,
        _b,
        _c,
        _d,
        _zeror,
        _zeroi,
        _errest,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, e = (0.0,0.0).',
                    'Constraint: e != (0.0,0.0)',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The companion matrix H cannot be formed without overflow.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to converge in lapackeig.zhseqr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _zeror_cls.refresh_py(_zeror)
        _zeroi_cls.refresh_py(_zeroi)
        _errest_cls.refresh_py(_errest)
        return_data = None
    _err_class.manage()
    return return_data
