"""
NumPy-focused utilities for the `naginterfaces` package.
"""

# NAG Copyright 2017-2019.

# pylint: disable=protected-access

import numpy as _np

from ..base import utils as _b_utils

def _check_callback_sorder(
        locus,
        arrays,
        **kwargs
):
    """
    Raises NagTypeError if the storage order of callback `arrays` tuple is not
    uniform.
    Raises NagTypeError if the mutual callback storage order doesn't match that
    from the wrapper.
    """
    cb_sorder = _b_utils._get_mutual_sorder(
        locus,
        arrays,
        in_callback=True,
        spiked_sorder=kwargs.get(
            'spiked_sorder',
            _b_utils._EngineData.default_spiked_sorder,
        ),
    )
    wrapper_sorder = kwargs.get(
        'wrapper_sorder', _b_utils._EngineData.default_sorder
    )
    if cb_sorder != wrapper_sorder:
        raise _b_utils.NagTypeError(
            locus,
            'arrays returned from this callback must use the same storage '
            'order as was used for arrays passed to the enclosing ' +
            locus['fun_name'].split(':')[0] + ' function --- '
            'in this callback storage order \'' + str(cb_sorder) +
            '\' was used, ' +
            'while arrays passed to the enclosing function used storage ' +
            'order \'' + str(wrapper_sorder) + '\''
        )

def _check_is_ndarray( # pylint: disable=too-many-arguments
        locus, array,
        allow_none=False, exp_rank=None, exp_shape=None, exp_class=None,
):
    """
    Raises NagTypeError if `array` is not an ndarray.
    Raises NagTypeError if `array` has incorrect dtype according to the input
    Engine*Array expected class.
    Raises NagShapeError if `array` does not have rank `exp_rank` or
    shape `exp_shape`.
    """
    if allow_none and array is None:
        return
    if not isinstance(array, _np.ndarray):
        raise _b_utils.NagTypeError(
            locus,
            _b_utils._bad_instance_msg(['numpy.ndarray'], allow_none)
        )
    _check_valid_ndarray(locus, array, exp_rank, exp_shape, exp_class)

def _check_valid_ndarray(
        locus, array, exp_rank=None, exp_shape=None, exp_class=None
):
    """
    Raises NagTypeError if ndarray `array` has incorrect dtype according to
    the input Engine*Array expected class.
    Raises NagShapeError if `array` does not have rank `exp_rank` or
    shape `exp_shape`.
    Raises NagShapeError when shape checking if any element of the expected
    shape was negative.
    """

    if exp_class is not None:
        a_dtypes = []
        a_dtypes.extend(exp_class._allowed_np_dtypes)

        if (
                exp_class is _b_utils._EngineFloat32ArrayType or
                exp_class is _b_utils._EngineFloat64ArrayType
        ):
            a_dtypes.extend(
                _b_utils._INT_TYPE_MAP[
                    _b_utils._EngineScalarType._module_key_numpy
                ]
            )

        exp_class._check_dtype(
            array, locus, allowed_dtypes=a_dtypes,
        )

    if not array.shape:
        raise _b_utils.NagShapeError(
            locus,
            'must have non-empty shape'
        )

    if exp_shape is not None:
        if _b_utils._check_shape_good(array, exp_shape):
            return
        _b_utils._check_shape_nonneg(locus, exp_shape)
        if array.ndim != len(exp_shape):
            raise _b_utils.NagShapeError(
                locus,
                _b_utils._bad_rank_msg(len(exp_shape), array.ndim),
            )
        raise _b_utils.NagShapeError(
            locus,
            _b_utils._bad_shape_msg(exp_shape, array.shape)
        )

    if exp_rank is not None:
        if array.ndim != exp_rank:
            raise _b_utils.NagShapeError(
                locus,
                _b_utils._bad_rank_msg(exp_rank, array.ndim),
            )

    if _b_utils._IS_DISCONTIG(array):
        raise _b_utils.NagTypeError(
            locus,
            _b_utils._CONTIG_ERR_STR,
        )

def _complex_to_float_array( # pylint: disable=invalid-name
        z,
        sorder,
):
    """
    Return a 2-d float array for the complex vector z.
    """
    if z.dtype == _np.complex64:
        float_class = _b_utils._EngineFloat32ArrayType
    elif z.dtype == _np.complex128:
        float_class = _b_utils._EngineFloat64ArrayType
    else:
        raise _b_utils.NagException(
            {'fun_name': '_complex_to_float_array'},
            'Inconceivable data instance',
        )
    _z_r = float_class.empty_ndarray((2, len(z)), sorder=sorder)
    _z_r[0, :] = [z_i.real for z_i in z]
    _z_r[1, :] = [z_i.imag for z_i in z]
    return _z_r

def _float_array_to_complex( # pylint: disable=invalid-name
        x,
):
    """
    Return a complex vector for a 2-d float array x.
    """
    if x.dtype == _np.float32:
        complex_class = _b_utils._EngineComplex64ScalarType
    elif x.dtype == _np.float64:
        complex_class = _b_utils._EngineComplex128ScalarType
    else:
        raise _b_utils.NagException(
            {'fun_name': '_float_array_to_complex'},
            'Inconceivable data instance',
        )
    return _np.array(
        [
            complex_class.numpy_type(
                complex(x[0, i], x[1, i])
            ) for i in range(x.shape[1])
        ]
    )

def _to_ndarray( # pylint: disable=too-many-arguments
        locus, array, copy,
        allow_none=False, exp_rank=None, exp_shape=None, exp_class=None
):
    """
    Cast `array` to ndarray, perhaps with copying.

    Raises NagTypeError if the cast/copy totally fails or
    succeeds but has incorrect dtype according to the input
    Engine*Array expected class.
    Raises NagShapeError if the cast/copy does not have rank `exp_rank` or
    shape `exp_shape`.
    """
    if allow_none and array is None:
        return array

    def to_ndarray_only(the_array):
        try:
            return _np.array(the_array, copy=copy, subok=True)
        except (TypeError, ValueError) as v_e:
            raise _b_utils.NagTypeError(
                locus,
                'numpy could not convert this argument to an ndarray; '
                'the error message was \'' + str(v_e) + '\''
            )

    nd_cast = to_ndarray_only(array)

    if array is not None and not nd_cast.shape:
        nd_cast = to_ndarray_only([array])

    if (
            not copy and
            len(nd_cast.shape) == 1 and
            _b_utils._IS_DISCONTIG(nd_cast)
    ):
        nd_cast = _np.ascontiguousarray(nd_cast)

    _check_valid_ndarray(locus, nd_cast, exp_rank, exp_shape, exp_class)

    if (
            exp_class is not None and
            (
                exp_class is _b_utils._EngineFloat32ArrayType or
                exp_class is _b_utils._EngineFloat64ArrayType
            ) and
            nd_cast.dtype in _b_utils._INT_TYPE_MAP[
                _b_utils._EngineScalarType._module_key_numpy
            ]
    ):
        nd_cast = nd_cast.astype(exp_class._underlying_scalar.numpy_type)

    return nd_cast

@_b_utils._withdrawal(
    'naginterfaces.library._utils._withdrawal_test', '42',
    ['Replacement advice.']
)
def _withdrawal_test():
    """
    Test of withdrawal decorator.
    """
    _b_utils._withdrawal_test()

@_b_utils._withdrawal(
    'naginterfaces.library._utils._withdrawal_warn_test', '42',
)
def _withdrawal_warn_test():
    """
    Test of withdrawal decorator on a function that reraises a warning.
    """
    _b_utils._withdrawal_warn_test()

@_b_utils._withdrawal(
    'naginterfaces.library._utils._withdrawal_test', None,
    ['Replacement advice.']
)
def _deprecation_test():
    """
    Test of withdrawal decorator.
    """
    _b_utils._deprecation_test()
