# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `anova` Chapter.

``anova`` - Analysis of Variance

This module is concerned with methods for analysing the results of
designed experiments.
The range of experiments covered include:

- single factor designs with equal sized blocks such as randomized
  complete block and balanced incomplete block designs,

- row and column designs such as Latin squares, and

- complete factorial designs.

Further designs may be analysed by combining the analyses provided by
multiple calls to functions or by using general linear model functions
provided in submodule ``correg``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04intro.html

See Also
--------
``naginterfaces.library.examples.anova`` :
    This subpackage contains examples for the ``anova`` module.
    See also the :ref:`library_anova_ex` subsection.
"""

# NAG Copyright 2017-2019.

import collections as _collections

from ..base import utils as _b_utils
from ..library import _utils

def hier2(y, lsub, nobs):
    u"""
    Two-way analysis of variance, hierarchical classification, subgroups
    of unequal size.

    ``hier2`` performs an analysis of variance for a two-way
    hierarchical classification with subgroups of possibly unequal size,
    and also computes the treatment group and subgroup means.
    A fixed effects model is assumed.

    For full information please refer to the NAG Library document for
    g04ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04agf.html

    Parameters
    ----------
    y : float, array-like, shape (n)
        The elements of `y` must contain the observations y_mij in the
        following order:

            y_111,y_211,...,y_{n_11 11},y_112,y_212,...,y_{n_12
            12},...,y_{11l_1},...,

y_{n_{1l_1}1l_1},...,y_{1ij},...,y_{n_{ij}ij},...,y_{1kl_k},...,y_{n_{kl_k}kl_k}.

        In words, the ordering is by group, and within each group is by
        subgroup, the members of each subgroup being in consecutive
        locations in `y`.

    lsub : int, array-like, shape (k)
        The number of subgroups within group i, l_i, for i = 1,2,...,k.

    nobs : int, array-like, shape (l)
        The numbers of observations in each subgroup, n_{ij}, in the
        following order:

n_11,n_12,...,n_{1l_1},n_21,...,n_{2l_2},...,n_{k1},...,n_{kl_k}

    Returns
    -------
    ngp : int, ndarray, shape (k)
        The total number of observations in group i, n_{i.}, for i =
        1,2,...,k.

    gbar : float, ndarray, shape (k)
        The mean for group i, y__{.i.}, for i = 1,2,...,k.

    sgbar : float, ndarray, shape (l)
        The subgroup means, y__{.ij}, in the following order:

y__.11,y__.12,...,y__.1l_1,y__.21,y__.22,...,y__.2l_2,...,y__{.k1},y__{.k2},...,y__.kl_k.

    gm : float
        The grand mean, y__....

    ss : float, ndarray, shape (4)
        Contains the sums of squares for the analysis of variance, as
        follows;

        - `ss`[0] = Between group sum of squares, `ss`_g,

        - `ss`[1] = Between subgroup within groups sum of squares,
          `ss`_{sg},

        - `ss`[2] = Residual sum of squares, `ss`_res,

        - `ss`[3] = Corrected total sum of squares, `ss`_tot.

    idf : int, ndarray, shape (4)
        Contains the degrees of freedom attributable to each sum of
        squares in the analysis of variance, as follows:

        - `idf`[0] = Degrees of freedom for between group sum of
          squares,

        - `idf`[1] = Degrees of freedom for between subgroup within
          groups sum of squares,

        - `idf`[2] = Degrees of freedom for residual sum of squares,

        - `idf`[3] = Degrees of freedom for corrected total sum of
          squares.

    f : float, ndarray, shape (2)
        Contains the mean square ratios, F_1 and F_2, for the between
        groups variation, and the between subgroups within groups
        variation, with respect to the residual, respectively.

    fp : float, ndarray, shape (2)
        Contains the significances of the mean square ratios, p_1 and
        p_2 respectively.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k > 1.

        (`errno` 2)
            On entry, i = *<value>* and `lsub`[i-1] = *<value>*.

            Constraint: `lsub`[i-1] > 0.

        (`errno` 3)
            On entry, sum_i`lsub`[i] = *<value>* and l = *<value>*.

            Constraint: sum_i`lsub`[i] = l.

        (`errno` 4)
            On entry, i = *<value>* and `nobs`[i-1] = *<value>*.

            Constraint: `nobs`[i-1] > 0.

        (`errno` 5)
            On entry, sum_i`nobs`[i-1] = *<value>* and n = *<value>*.

            Constraint: sum_i`nobs`[i-1] = n.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The total corrected sum of squares is zero.

        (`errno` 7)
            The residual sum of squares is zero.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.anova import hier2 as b_hier2

    fun_name = 'naginterfaces.library.anova.hier2'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _lsub = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lsub'},
        lsub,
        copy=False,
    )
    _k_py = _b_utils._EngineIntArrayType.get_shape(_lsub, {'fun_name': fun_name, 'entity': 'lsub'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _nobs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nobs'},
        nobs,
        copy=False,
    )
    _l_py = _b_utils._EngineIntArrayType.get_shape(_nobs, {'fun_name': fun_name, 'entity': 'nobs'}, exp_rank=1)[0]
    _l = _b_utils._EngineIntScalarType(
        _l_py, {'fun_name': fun_name, 'entity': '_l'},
    ).to_ctypes()
    _ngp = _b_utils._EngineIntArrayType.empty_ndarray(
        (_k_py,),
    )
    _gbar = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    _sgbar = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_l_py,),
    )
    _ss = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _idf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2,),
    )
    _fp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2,),
    )
    try:
        _return_scalar = b_hier2(
            _y,
            _n,
            _k,
            _lsub,
            _nobs,
            _l,
            _ngp,
            _gbar,
            _sgbar,
            _ss,
            _idf,
            _f,
            _fp,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ngp_py = _ngp
    _gbar_py = _gbar
    _sgbar_py = _sgbar
    _gm_py = _return_scalar
    _ss_py = _ss
    _idf_py = _idf
    _f_py = _f
    _fp_py = _fp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ngp', 'gbar', 'sgbar', 'gm', 'ss', 'idf', 'f', 'fp'],
    )(
        _ngp_py,
        _gbar_py,
        _sgbar_py,
        _gm_py,
        _ss_py,
        _idf_py,
        _f_py,
        _fp_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def random(y, iblock, nt, it, tol, irdf):
    u"""
    Analysis of variance, randomized block or completely randomized
    design, treatment means and standard errors.

    ``random`` computes the analysis of variance and treatment means and
    standard errors for a randomized block or completely randomized
    design.

    For full information please refer to the NAG Library document for
    g04bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04bbf.html

    Parameters
    ----------
    y : float, array-like, shape (n)
        The observations in the order as described by `iblock` and `nt`.

    iblock : int
        Indicates the block structure.

        abs(`iblock`) <= 1

          There are no blocks, i.e., it is a completely randomized
          design.

        `iblock` >= 2

          There are `iblock` blocks and the data should be input by
          blocks, i.e., `y` must contain the observations for block 1
          followed by the observations for block 2, etc.

        `iblock` <= -2

          There are abs(`iblock`) blocks and the data is input in
          parallel with respect to blocks, i.e., `y`[0] must contain the
          first observation for block 1, `y`[1] must contain the first
          observation for block 2...`y`[abs(`iblock`)-1] must contain
          the first observation for block
          abs(`iblock`),`y`[abs(`iblock`+1)-1] must contain the second
          observation for block 1, etc.

    nt : int
        The number of treatments. If only blocks are required in the
        analysis then set `nt` = 1.

    it : int, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `nt`>=2: n; otherwise: 1.

        `it`[i-1] indicates which of the `nt` treatments plot i
        received, for i = 1,2,...,n.

        If `nt` = 1, `it` is not referenced.

    tol : float
        The tolerance value used to check for zero eigenvalues of the
        matrix Omega. If `tol` = 0.0 a default value of 10^(-5) is used.

    irdf : int
        An adjustment to the degrees of freedom for the residual and
        total.

        `irdf` >= 1

          The degrees of freedom for the total is set to n-`irdf` and
          the residual degrees of freedom adjusted accordingly.

        `irdf` = 0

          The total degrees of freedom for the total is set to n-1, as
          usual.

    Returns
    -------
    gmean : float
        The grand mean, mu^.

    bmean : float, ndarray, shape (|`iblock`|)
        If abs(`iblock`) >= 2, `bmean`[j-1] contains the mean for the
        jth block, beta^_j, for j = 1,2,...,b.

    tmean : float, ndarray, shape (`nt`)
        If `nt` >= 2, `tmean`[l-1] contains the (adjusted) mean for the
        lth treatment, mu^*+tau^_l, for l = 1,2,...,t, where mu^* is the
        mean of the treatment adjusted observations, y_{ij(l)}-tau^_l.

    tabl : float, ndarray, shape (4, 5)
        The analysis of variance table. Column 1 contains the degrees of
        freedom, column 2 the sum of squares, and where appropriate,
        column 3 the mean squares, column 4 the F-statistic and column 5
        the significance level of the F-statistic. Row 1 is for Blocks,
        row 2 for Treatments, row 3 for Residual and row 4 for Total.
        Mean squares are computed for all but the Total row;
        F-statistics and significance are computed for Treatments and
        Blocks, if present. Any unfilled cells are set to zero.

    c : float, ndarray, shape (`nt`, `nt`)
        If `nt` >= 2, the upper triangular part of `c` contains the
        variance-covariance matrix of the treatment effects, the
        strictly lower triangular part contains the standard errors of
        the difference between two treatment effects (means), i.e.,
        `c`[i-1,j-1] contains the covariance of treatment i and j if j
        >= i and the standard error of the difference between treatment
        i and j if j < i, for j = 1,2,...,t, for i = 1,2,...,t.

    irep : int, ndarray, shape (`nt`)
        If `nt` >= 2, the treatment replications, R_{ll}, for l =
        1,2,...,`nt`.

    r : float, ndarray, shape (n)
        The residuals, r_i, for i = 1,2,...,n.

    ef : float, ndarray, shape (`nt`)
        If `nt` >= 2, the canonical efficiency factors.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irdf` = *<value>*.

            Constraint: `irdf` >= 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, no blocks or treatments in model.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, n = *<value>* and `iblock` = *<value>*.

            Constraint: n >= 2 and if abs(`iblock`) >= 2, n must be a
            multiple of abs(`iblock`).

        (`errno` 3)
            On entry, at least one treatment is not present. Treatment
            *<value>* does not appear in `it`.

        (`errno` 3)
            On entry, i = *<value>*, `it`[i-1] = *<value>* and `nt` =
            *<value>*.

            Constraint: 1 <= `it`[i-1] <= `nt`.

        (`errno` 4)
            On entry, the values in `y` are constant.

        (`errno` 5)
            The computation of the eigenvalues has failed to converge.

        (`errno` 5)
            A computed standard error is zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The treatments are totally confounded with blocks.

        (`errno` 7)
            The residual degrees of freedom is zero.

        (`errno` 7)
            The residual mean square is zero.

        (`errno` 8)
            The design is disconnected.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.anova import random as b_random

    fun_name = 'naginterfaces.library.anova.random'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _it = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'it'},
        it,
        copy=False,
    )
    _iblock = _b_utils._EngineIntScalarType(
        iblock, {'fun_name': fun_name, 'entity': 'iblock'},
    ).to_ctypes()
    _iblock_py = _b_utils._EngineIntScalarType.to_py(_iblock)
    _bmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (abs(_iblock_py),),
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _tmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _tabl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4, 5),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py, _nt_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _irep = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nt_py,),
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _ef = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    try:
        _return_scalar = b_random(
            _n,
            _y,
            _iblock,
            _nt,
            _it,
            _bmean,
            _tmean,
            _tabl,
            _c,
            _irep,
            _r,
            _ef,
            tol,
            irdf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _gmean_py = _return_scalar
    _bmean_py = _bmean
    _tmean_py = _tmean
    _tabl_py = _tabl
    _c_py = _c
    _irep_py = _irep
    _r_py = _r
    _ef_py = _ef
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['gmean', 'bmean', 'tmean', 'tabl', 'c', 'irep', 'r', 'ef'],
    )(
        _gmean_py,
        _bmean_py,
        _tmean_py,
        _tabl_py,
        _c_py,
        _irep_py,
        _r_py,
        _ef_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def rowcol(nrep, nrow, ncol, y, nt, it, tol, irdf):
    u"""
    Analysis of variance, general row and column design, treatment means
    and standard errors.

    ``rowcol`` computes the analysis of variance for a general row and
    column design together with the treatment means and standard errors.

    For full information please refer to the NAG Library document for
    g04bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04bcf.html

    Parameters
    ----------
    nrep : int
        b, the number of replicates.

    nrow : int
        r, the number of rows per replicate.

    ncol : int
        c, the number of columns per replicate.

    y : float, array-like, shape (`nrep`*`nrow`*`ncol`)
        The n = brc observations ordered by columns within rows within
        replicates. That is `y`[ rc(i-1) +r(j-1) +k -1] contains the
        observation from the kth column of the jth row of the ith
        replicate, for k = 1,2,...,c, for j = 1,2,...,r, for i =
        1,2,...,b.

    nt : int
        The number of treatments. If only replicates, rows and columns
        are required in the analysis then set `nt` = 1.

    it : int, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `nt`>1: `nrep`*`nrow`*`ncol`; otherwise: 1.

        If `nt` > 1, `it`[i-1] indicates which of the `nt` treatments
        unit i received, for i = 1,2,...,n.

        If `nt` = 1, `it` is not referenced.

    tol : float
        The tolerance value used to check for zero eigenvalues of the
        matrix Omega. If `tol` = 0.0 a default value of 0.00001 is used.

    irdf : int
        An adjustment to the degrees of freedom for the residual and
        total.

        `irdf` >= 1

          The degrees of freedom for the total is set to n-`irdf` and
          the residual degrees of freedom adjusted accordingly.

        `irdf` = 0

          The total degrees of freedom for the total is set to n-1, as
          usual.

    Returns
    -------
    gmean : float
        The grand mean, mu^.

    tmean : float, ndarray, shape (`nt`)
        If `nt` >= 2, `tmean`[l-1] contains the (adjusted) mean for the
        lth treatment, mu^*+tau^_l, for l = 1,2,...,t, where mu^* is the
        mean of the treatment adjusted observations y_{ijk(l)}-tau^_l.
        Otherwise `tmean` is not referenced.

    tabl : float, ndarray, shape (6, 5)
        The analysis of variance table. Column 1 contains the degrees of
        freedom, column 2 the sum of squares, and where appropriate,
        column 3 the mean squares, column 4 the F-statistic and column 5
        the significance level of the F-statistic. Row 1 is for
        replicates, row 2 for rows, row 3 for columns, row 4 for
        treatments (if `nt` > 1), row 5 for residual and row 6 for
        total. Mean squares are computed for all but the total row,
        F-statistics and significance are computed for treatments,
        replicates, rows and columns. Any unfilled cells are set to
        zero.

    c : float, ndarray, shape (`nt`, `nt`)
        The upper triangular part of `c` contains the
        variance-covariance matrix of the treatment effects, the
        strictly lower triangular part contains the standard errors of
        the difference between two treatment effects (means), i.e.,
        `c`[i-1,j-1] contains the covariance of treatment i and j if j
        >= i and the standard error of the difference between treatment
        i and j if j < i, for j = 1,2,...,t, for i = 1,2,...,t.

    irep : int, ndarray, shape (`nt`)
        If `nt` > 1, the treatment replications, R_{ll}, for l =
        1,2,...,`nt`. Otherwise `irep` is not referenced.

    rpmean : float, ndarray, shape (`nrep`)
        If `nrep` > 1, `rpmean`[i-1] contains the mean for the ith
        replicate, mu^+beta^_i, for i = 1,2,...,b. Otherwise `rpmean` is
        not referenced.

    rmean : float, ndarray, shape (`nrep`*`nrow`)
        `rmean`[j-1] contains the mean for the jth row, mu^+rho^_i, for
        j = 1,2,...,r.

    cmean : float, ndarray, shape (`nrep`*`ncol`)
        `cmean`[k-1] contains the mean for the kth column, mu^+gamma^_k,
        for k = 1,2,...,c.

    r : float, ndarray, shape (`nrep`*`nrow`*`ncol`)
        The residuals, r_i, for i = 1,2,...,n.

    ef : float, ndarray, shape (`nt`)
        If `nt` >= 2, the canonical efficiency factors. Otherwise `ef`
        is not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nrep` = *<value>*.

            Constraint: `nrep` >= 1.

        (`errno` 1)
            On entry, `irdf` = *<value>*.

            Constraint: `irdf` >= 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `ncol` = *<value>*.

            Constraint: `ncol` >= 2.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` >= 1.

        (`errno` 1)
            On entry, `nrow` = *<value>*.

            Constraint: `nrow` >= 2.

        (`errno` 2)
            On entry, at least one treatment is not present. Treatment
            *<value>* does not appear in `it`.

        (`errno` 2)
            On entry, i = *<value>*, `it`[i-1] = *<value>* and `nt` =
            *<value>*.

            Constraint: 1 <= `it`[i-1] <= `nt`.

        (`errno` 3)
            On entry, the values in `y` are constant.

        (`errno` 4)
            The computation of the eigenvalues has failed to converge.

        (`errno` 4)
            A computed standard error is zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The treatments are totally confounded with blocks.

        (`errno` 6)
            The residual degrees of freedom is zero.

        (`errno` 6)
            The residual mean square is zero.

        (`errno` 7)
            The design is disconnected.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.anova import rowcol as b_rowcol

    fun_name = 'naginterfaces.library.anova.rowcol'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _it = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'it'},
        it,
        copy=False,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _tmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _tabl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (6, 5),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py, _nt_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _irep = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nt_py,),
    )
    _nrep = _b_utils._EngineIntScalarType(
        nrep, {'fun_name': fun_name, 'entity': 'nrep'},
    ).to_ctypes()
    _nrep_py = _b_utils._EngineIntScalarType.to_py(_nrep)
    _rpmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrep_py,),
    )
    _nrow = _b_utils._EngineIntScalarType(
        nrow, {'fun_name': fun_name, 'entity': 'nrow'},
    ).to_ctypes()
    _nrow_py = _b_utils._EngineIntScalarType.to_py(_nrow)
    _rmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrep_py*_nrow_py,),
    )
    _ncol = _b_utils._EngineIntScalarType(
        ncol, {'fun_name': fun_name, 'entity': 'ncol'},
    ).to_ctypes()
    _ncol_py = _b_utils._EngineIntScalarType.to_py(_ncol)
    _cmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrep_py*_ncol_py,),
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrep_py*_nrow_py*_ncol_py,),
    )
    _ef = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    try:
        _return_scalar = b_rowcol(
            _nrep,
            _nrow,
            _ncol,
            _y,
            _nt,
            _it,
            _tmean,
            _tabl,
            _c,
            _irep,
            _rpmean,
            _rmean,
            _cmean,
            _r,
            _ef,
            tol,
            irdf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _gmean_py = _return_scalar
    _tmean_py = _tmean
    _tabl_py = _tabl
    _c_py = _c
    _irep_py = _irep
    _rpmean_py = _rpmean
    _rmean_py = _rmean
    _cmean_py = _cmean
    _r_py = _r
    _ef_py = _ef
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['gmean', 'tmean', 'tabl', 'c', 'irep', 'rpmean', 'rmean', 'cmean', 'r', 'ef'],
    )(
        _gmean_py,
        _tmean_py,
        _tabl_py,
        _c_py,
        _irep_py,
        _rpmean_py,
        _rmean_py,
        _cmean_py,
        _r_py,
        _ef_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def factorial(y, lfac, nblock, inter, irdf, mterm, maxt):
    u"""
    Analysis of variance, complete factorial design, treatment means and
    standard errors.

    ``factorial`` computes an analysis of variance table and treatment
    means for a complete factorial design.

    For full information please refer to the NAG Library document for
    g04ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04caf.html

    Parameters
    ----------
    y : float, array-like, shape (n)
        The observations in standard order, see [Description].

    lfac : int, array-like, shape (nfac)
        `lfac`[i] must contain the number of levels for the ith factor,
        for i = 1,2,...,k.

    nblock : int
        The number of blocks. If there are no blocks, set `nblock` = 0
        or 1.

    inter : int
        The maximum number of factors in an interaction term. If no
        interaction terms are to be computed, set `inter` = 0 or 1.

    irdf : int
        The adjustment to the residual and total degrees of freedom. The
        total degrees of freedom are set to n-`irdf` and the residual
        degrees of freedom adjusted accordingly. For examples of the use
        of `irdf` see [Further Comments].

    mterm : int
        The maximum number of terms in the analysis of variance table,
        see [Further Comments].

    maxt : int
        The maximum number of treatment means to be computed, see
        [Further Comments]. If the value of `maxt` is too small for the
        required analysis then the minimum number is returned in
        `imean`[0].

    Returns
    -------
    table : float, ndarray, shape (`mterm`, 5)
        The first `itotal` rows of `table` contain the analysis of
        variance table. The first column contains the degrees of
        freedom, the second column contains the sum of squares, the
        third column (except for the row corresponding to the total sum
        of squares) contains the mean squares, i.e., the sums of squares
        divided by the degrees of freedom, and the fourth and fifth
        columns contain the F ratio and significance level, respectively
        (except for rows corresponding to the total sum of squares, and
        the residual sum of squares). All other cells of the table are
        set to zero.

        The first row corresponds to the blocks and is set to zero if
        there are no blocks.

        The `itotal`th row corresponds to the total sum of squares for
        `y` and the (`itotal`-1)th row corresponds to the residual sum
        of squares.

        The central rows of the table correspond to the main effects
        followed by the interaction if specified by `inter`.

        The main effects are in the order specified by `lfac` and the
        interactions are in lexical order, see [Description].

    itotal : int
        The row in `table` corresponding to the total sum of squares.
        The number of treatment effects is `itotal`-3.

    tmean : float, ndarray, shape (`maxt`)
        The treatment means. The position of the means for an effect is
        given by the index in `imean`. For a given effect the means are
        in standard order, see [Description].

    e : float, ndarray, shape (`maxt`)
        The estimated effects in the same order as for the means in
        `tmean`.

    imean : int, ndarray, shape (`mterm`)
        Indicates the position of the effect means in `tmean`. The
        effect means corresponding to the first treatment effect in the
        ANOVA table are stored in `tmean`[0] up to
        `tmean`[`imean`[0]-1]. Other effect means corresponding to the
        ith treatment effect, i = 1,2,...,`itotal`-3, are stored in
        `tmean`[`imean`[i-1]] up to `tmean`[`imean`[i]-1].

    semean : float, ndarray, shape (`mterm`)
        The standard error of the difference between means corresponding
        to the ith treatment effect in the ANOVA table.

    bmean : float, ndarray, shape (`nblock`+1)
        `bmean`[0] contains the grand mean, if `nblock` > 1, `bmean`[1]
        up to `bmean`[`nblock`] contain the block means.

    r : float, ndarray, shape (n)
        The residuals.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irdf` = *<value>*.

            Constraint: `irdf` >= 0.

        (`errno` 1)
            On entry, `inter` = *<value>* and nfac = *<value>*.

            Constraint: `inter` <= nfac.

        (`errno` 1)
            On entry, `inter` = *<value>*.

            Constraint: `inter` >= 0.

        (`errno` 1)
            On entry, `nblock` = *<value>*.

            Constraint: `nblock` >= 0.

        (`errno` 1)
            On entry, nfac = *<value>*.

            Constraint: nfac >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 4.

        (`errno` 2)
            On entry, i = *<value>* and `lfac`[i-1] = *<value>*

            Constraint: `lfac`[i-1] >= 2.

        (`errno` 2)
            On entry, the number of plots per block is incompatible with
            the number of plot factors.

        (`errno` 2)
            On entry, n = *<value>* and `nblock` = *<value>*.

            Constraint: when `nblock` > 1, n must be a multiple of
            `nblock`.

        (`errno` 3)
            On entry, the values of `y` are constant.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, `mterm` = *<value>*

            Constraint: `mterm` >= *<value>*.

        (`errno` 2)
            On entry, `maxt` = *<value>*

            Constraint: `maxt` >= *<value>*.

        (`errno` 4)
            The residual sum of squares is zero.

        (`errno` 4)
            There are no degrees of freedom for the residual sum of
            squares.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.anova import factorial as b_factorial

    fun_name = 'naginterfaces.library.anova.factorial'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _lfac = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lfac'},
        lfac,
        copy=False,
    )
    _nfac_py = _b_utils._EngineIntArrayType.get_shape(_lfac, {'fun_name': fun_name, 'entity': 'lfac'}, exp_rank=1)[0]
    _nfac = _b_utils._EngineIntScalarType(
        _nfac_py, {'fun_name': fun_name, 'entity': '_nfac'},
    ).to_ctypes()
    _mterm = _b_utils._EngineIntScalarType(
        mterm, {'fun_name': fun_name, 'entity': 'mterm'},
    ).to_ctypes()
    _mterm_py = _b_utils._EngineIntScalarType.to_py(_mterm)
    _table = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mterm_py, 5),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _maxt = _b_utils._EngineIntScalarType(
        maxt, {'fun_name': fun_name, 'entity': 'maxt'},
    ).to_ctypes()
    _maxt_py = _b_utils._EngineIntScalarType.to_py(_maxt)
    _tmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxt_py,),
    )
    _e = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxt_py,),
    )
    _imean = _b_utils._EngineIntArrayType.empty_ndarray(
        (_mterm_py,),
    )
    _semean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mterm_py,),
    )
    _nblock = _b_utils._EngineIntScalarType(
        nblock, {'fun_name': fun_name, 'entity': 'nblock'},
    ).to_ctypes()
    _nblock_py = _b_utils._EngineIntScalarType.to_py(_nblock)
    _bmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nblock_py+1,),
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = b_factorial(
            _n,
            _y,
            _nfac,
            _lfac,
            _nblock,
            inter,
            irdf,
            _mterm,
            _table,
            _tmean,
            _maxt,
            _e,
            _imean,
            _semean,
            _bmean,
            _r,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _table_py = _table
    _itotal_py = _return_scalar
    _tmean_py = _tmean
    _e_py = _e
    _imean_py = _imean
    _semean_py = _semean
    _bmean_py = _bmean
    _r_py = _r
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['table', 'itotal', 'tmean', 'e', 'imean', 'semean', 'bmean', 'r'],
    )(
        _table_py,
        _itotal_py,
        _tmean_py,
        _e_py,
        _imean_py,
        _semean_py,
        _bmean_py,
        _r_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def contrasts(tmean, irep, rms, rdf, ct, tabl, tol, usetx, tx):
    u"""
    Computes sum of squares for contrast between means.

    ``contrasts`` computes sum of squares for a user-defined contrast
    between means.

    For full information please refer to the NAG Library document for
    g04da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04daf.html

    Parameters
    ----------
    tmean : float, array-like, shape (nt)
        The treatment means, tau^_i, for i = 1,2,...,t.

    irep : int, array-like, shape (nt)
        The replication for each treatment mean, n_i, for i = 1,2,...,t.

    rms : float
        The residual mean square, sigma^^2.

    rdf : float
        The residual degrees of freedom.

    ct : float, array-like, shape (nt, nc)
        The columns of `ct` must contain the nc contrasts, that is
        `ct`[i-1,j-1] must contain lambda_{ij}, for j = 1,2,...,nc, for
        i = 1,2,...,t.

    tabl : float, array-like, shape (nc, 5)
        The elements of `tabl` that are not referenced as described
        below remain unchanged.

    tol : float
        The tolerance, epsilon used to check if the contrasts are
        orthogonal and if they are orthogonal to the mean. If `tol` <=
        0.0 the value machine precision is used.

    usetx : bool
        If `usetx` = True the means tau_i^* are provided in `tx` and the
        formula (2) is used instead of formula (1).

        If `usetx` = False formula (1) is used and `tx` is not
        referenced.

    tx : float, array-like, shape (nt)
        If `usetx` = True `tx` must contain the means tau_i^*, for i =
        1,2,...,t.

    Returns
    -------
    est : float, ndarray, shape (nc)
        The estimates of the contrast, Lambda_j, for j = 1,2,...,nc.

    tabl : float, ndarray, shape (nc, 5)
        The rows of the analysis of variance table for the contrasts.
        For each row column 1 contains the degrees of freedom, column 2
        contains the sum of squares, column 3 contains the mean square,
        column 4 the F-statistic and column 5 the significance level for
        the contrast. Note that the degrees of freedom are always one
        and so the mean square equals the sum of squares.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rdf` = *<value>*.

            Constraint: `rdf` >= 1.0.

        (`errno` 1)
            On entry, `rms` = *<value>*.

            Constraint: `rms` > 0.0.

        (`errno` 1)
            On entry, nt = *<value>*.

            Constraint: nt >= 2.

        (`errno` 1)
            On entry, nc = *<value>*.

            Constraint: nc >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The *<value>**<value>* and *<value>**<value>* contrasts are
            not orthogonal.

        (`errno` 2)
            The *<value>**<value>* contrast is not orthogonal to the
            mean.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.anova import contrasts as b_contrasts

    fun_name = 'naginterfaces.library.anova.contrasts'

    _tmean = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tmean'},
        tmean,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt_py = _b_utils._EngineFloat64ArrayType.get_shape(_tmean, {'fun_name': fun_name, 'entity': 'tmean'}, exp_rank=1)[0]
    _nt = _b_utils._EngineIntScalarType(
        _nt_py, {'fun_name': fun_name, 'entity': '_nt'},
    ).to_ctypes()
    _irep = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irep'},
        irep,
        copy=False,
    )
    _ct = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ct'},
        ct,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nc_py = _b_utils._EngineFloat64ArrayType.get_shape(_ct, {'fun_name': fun_name, 'entity': 'ct'}, exp_rank=2)[1]
    _nc = _b_utils._EngineIntScalarType(
        _nc_py, {'fun_name': fun_name, 'entity': '_nc'},
    ).to_ctypes()
    _est = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nc_py,),
    )
    _tabl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tabl'},
        tabl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _tx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tx'},
        tx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        b_contrasts(
            _nt,
            _tmean,
            _irep,
            rms,
            rdf,
            _nc,
            _ct,
            _est,
            _tabl,
            tol,
            usetx,
            _tx,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _tabl_py = _tabl
    _est_py = _est
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['est', 'tabl'],
    )(
        _est_py,
        _tabl_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def confidence(typ, tmean, rdf, c, clevel):
    u"""
    Computes confidence intervals for differences between means computed
    by ``random`` or ``rowcol``.

    ``confidence`` computes simultaneous confidence intervals for the
    differences between means.
    It is intended for use after ``random`` or ``rowcol``.

    For full information please refer to the NAG Library document for
    g04db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04dbf.html

    Parameters
    ----------
    typ : str, length 1
        Indicates which method is to be used.

        `typ` = 'T'

          The Tukey--Kramer method is used.

        `typ` = 'B'

          The Bonferroni method is used.

        `typ` = 'D'

          The Dunn--Sidak method is used.

        `typ` = 'L'

          The Fisher LSD method is used.

        `typ` = 'S'

          The Scheffe's method is used.

    tmean : float, array-like, shape (nt)
        The treatment means, tau^_i, for i = 1,2,...,t.

    rdf : float
        nu, the residual degrees of freedom.

    c : float, array-like, shape (nt, nt)
        The strictly lower triangular part of `c` must contain the
        standard errors of the differences between the means as returned
        by ``random`` and ``rowcol``. That is `c`[i,j], i > j, contains
        the standard error of the difference between the ith and jth
        mean in `tmean`.

    clevel : float
        The required confidence level for the computed intervals,
        (1-alpha).

    Returns
    -------
    cil : float, ndarray, shape (nt*(nt-1)/2)
        The ((i-1)(i-2)/2+j)th element contains the lower limit to the
        confidence interval for the difference between ith and jth means
        in `tmean`, for j = 1,2,...,i-1, for i = 2,3,...,t.

    ciu : float, ndarray, shape (nt*(nt-1)/2)
        The ((i-1)(i-2)/2+j)th element contains the upper limit to the
        confidence interval for the difference between ith and jth means
        in `tmean`, for j = 1,2,...,i-1, for i = 2,3,...,t.

    isig : int, ndarray, shape (nt*(nt-1)/2)
        The ((i-1)(i-2)/2+j)th element indicates if the difference
        between ith and jth means in `tmean` is significant, for j =
        1,2,...,i-1, for i = 2,3,...,t. If the difference is significant
        then the returned value is 1; otherwise the returned value is 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `typ` = *<value>*.

            Constraint: `typ` = 'T', 'B', 'D', 'L' or 'S'.

        (`errno` 1)
            On entry, `clevel` = *<value>*.

            Constraint: 0.0 < `clevel` < 1.0.

        (`errno` 1)
            On entry, `rdf` = *<value>*.

            Constraint: `rdf` >= 1.0.

        (`errno` 1)
            On entry, nt = *<value>*.

            Constraint: nt >= 2.

        (`errno` 2)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: `c`[i-1,j-1] > 0.0.

        (`errno` 3)
            There has been a failure in the computation of the
            studentized range statistic.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.anova import confidence as b_confidence

    fun_name = 'naginterfaces.library.anova.confidence'

    _tmean = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tmean'},
        tmean,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt_py = _b_utils._EngineFloat64ArrayType.get_shape(_tmean, {'fun_name': fun_name, 'entity': 'tmean'}, exp_rank=1)[0]
    _nt = _b_utils._EngineIntScalarType(
        _nt_py, {'fun_name': fun_name, 'entity': '_nt'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cil = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py*(_nt_py-1)//2,),
    )
    _ciu = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py*(_nt_py-1)//2,),
    )
    _isig = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nt_py*(_nt_py-1)//2,),
    )
    b_confidence(
        typ,
        _nt,
        _tmean,
        rdf,
        _c,
        clevel,
        _cil,
        _ciu,
        _isig,
    )
    _cil_py = _cil
    _ciu_py = _ciu
    _isig_py = _isig
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['cil', 'ciu', 'isig'],
    )(
        _cil_py,
        _ciu_py,
        _isig_py,
    )
    return _return_data

def dummyvars(typ, levels, ifact, v):
    u"""
    Computes orthogonal polynomials or dummy variables for
    factor/classification variable.

    ``dummyvars`` computes orthogonal polynomial or dummy variables for
    a factor or classification variable.

    For full information please refer to the NAG Library document for
    g04ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04eaf.html

    Parameters
    ----------
    typ : str, length 1
        The type of dummy variable to be computed.

        - If `typ` = 'P', an orthogonal Polynomial representation is
          computed.

        - If `typ` = 'H', a Helmert matrix representation is computed.

        - If `typ` = 'F', the contrasts relative to the First level are
          computed.

        - If `typ` = 'L', the contrasts relative to the Last level are
          computed.

        - If `typ` = 'C', a Complete set of dummy variables is computed.

    levels : int
        k, the number of levels of the factor.

    ifact : int, array-like, shape (n)
        The n values of the factor.

    v : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `typ`='P': `levels`; otherwise: 1.

        If `typ` = 'P', the k distinct values of the underlying variable
        for which the orthogonal polynomial is to be computed.

        If `typ` != 'P', `v` is not referenced.

    Returns
    -------
    x : float, ndarray, shape (n, :)
        The n by k* matrix of dummy variables, where k* = k-1 if `typ` =
        'P', 'H', 'F' or 'L' and k* = k if `typ` = 'C'.

    rep : float, ndarray, shape (`levels`)
        The number of replications for each level of the factor, r_i,
        for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `typ` = *<value>*.

            Constraint: `typ` = 'P', 'H', 'F', 'L' or 'C'.

        (`errno` 1)
            On entry, n = *<value>* and `levels` = *<value>*.

            Constraint: n >= `levels`.

        (`errno` 1)
            On entry, `levels` = *<value>*.

            Constraint: `levels` >= 2.

        (`errno` 2)
            On entry, not all values of `v` are distinct.

        (`errno` 2)
            On entry, not all levels are present in `ifact`.

        (`errno` 2)
            On entry, i = *<value>*, `ifact`[i-1] = *<value>* and
            `levels` = *<value>*.

            Constraint: 1 <= `levels`[i-1] <= `levels`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The *<value>**<value>* polynomial has all elements zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.anova import dummyvars as b_dummyvars

    fun_name = 'naginterfaces.library.anova.dummyvars'

    _ifact = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ifact'},
        ifact,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ifact, {'fun_name': fun_name, 'entity': 'ifact'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _typ = _b_utils._EngineCharScalarType(
        typ, {'fun_name': fun_name, 'entity': 'typ'}, exp_chrlen=1,
    ).to_ctypes()
    _typ_py = _b_utils._EngineCharScalarType.to_py(_typ)
    _levels = _b_utils._EngineIntScalarType(
        levels, {'fun_name': fun_name, 'entity': 'levels'},
    ).to_ctypes()
    _levels_py = _b_utils._EngineIntScalarType.to_py(_levels)
    if _typ_py in ('P', 'p', 'H', 'h', 'F', 'f', 'L', 'l'):
        _x_dim2 = _levels_py-1
    elif _typ_py in ('C', 'c'):
        _x_dim2 = _levels_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_b_utils._EngineData.default_sorder,
    )
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rep = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_levels_py,),
    )
    try:
        b_dummyvars(
            typ,
            _n,
            _levels,
            _ifact,
            _x,
            _v,
            _rep,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _rep_py = _rep
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'rep'],
    )(
        _x_py,
        _rep_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def icc(mtype, score, rtype=1, smiss=None, alpha=0.05):
    u"""
    Intraclass correlation (ICC) for assessing rater reliability.

    ``icc`` calculates the intraclass correlation (ICC).

    For full information please refer to the NAG Library document for
    g04ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g04/g04gaf.html

    Parameters
    ----------
    mtype : int
        Indicates which model is to be used.

        `mtype` = 1

          The reliability study is a one-factor design, ICC(1,1).

        `mtype` = 2

          The reliability study is a random factorial design, ICC(2,1).

        `mtype` = 3

          The reliability study is a mixed factorial design, ICC(3,1).

    score : float, array-like, shape (nrep, nsubj, nrater)
        The matrix of scores, with `score`[k-1,i-1,j-1] being the score
        given to the ith subject by the jth rater in the kth replicate.

        If rater j did not rate subject i at replication k, the
        corresponding element of `score`, `score`[k-1,i-1,j-1], should
        be set to `smiss`.

    rtype : int, optional
        Indicates which type of reliability is required.

        `rtype` = 1

          Interrater reliability is required.

        `rtype` = 2

          Intrarater reliability is required.

    smiss : None or float, optional
        The value used to indicate a missing score.

        Care should be taken in the selection of the value used to
        indicate a missing score. ``icc`` will treat any score in the
        inclusive range (1+/-0.1^(machine.decimal_digits-2))*`smiss` as
        missing.

        Alternatively, a NaN (Not A Number) can be used to indicate
        missing values, in which case the value of `smiss` and any
        missing values of `score` can be set through a call to
        ``ieee.create_nan``.

    alpha : float, optional
        alpha, the significance level used in the construction of the
        confidence intervals for `icc`.

    Returns
    -------
    icc : float
        An estimate of the intraclass correlation to measure either the
        interrater reliability, rho, or intrarater reliability, gamma,
        as specified by `mtype` and `rtype`.

    lci : float
        An approximate lower limit for the 100*(1-alpha)% confidence
        interval for the ICC.

    uci : float
        An approximate upper limit for the 100*(1-alpha%) confidence
        interval for the ICC.

    fstat : float
        f, the F-statistic associated with `icc`.

    df1 : float
        nu_1 and nu_2, the degrees of freedom associated with f.

    df2 : float
        nu_1 and nu_2, the degrees of freedom associated with f.

    pvalue : float
        P(F >= f:nu_1,nu_1), the upper tail probability from an F
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `mtype` = *<value>*.

            Constraint: `mtype` = 1, 2 or 3.

        (`errno` 21)
            On entry, `rtype` = *<value>*.

            Constraint: `rtype` = 1 or 2.

        (`errno` 31)
            On entry, nrep = *<value>*.

            Constraint: nrep >= 1.

        (`errno` 32)
            On entry, nrep = *<value>*.

            Constraint: when `mtype` = 2 or 3 and `rtype` = 2, nrep >=
            2.

        (`errno` 33)
            On entry, after adjusting for missing data, nrep =
            *<value>*.

            Constraint: nrep >= 1.

        (`errno` 34)
            On entry, after adjusting for missing data, nrep =
            *<value>*.

            Constraint: when `mtype` = 2 or 3 and `rtype` = 2, nrep >=
            2.

        (`errno` 41)
            On entry, nsubj = *<value>*.

            Constraint: nsubj >= 2.

        (`errno` 42)
            On entry, after adjusting for missing data, nsubj =
            *<value>*.

            Constraint: nsubj >= 2.

        (`errno` 51)
            On entry, nrater = *<value>*.

            Constraint: nrater >= 2.

        (`errno` 52)
            On entry, after adjusting for missing data, nrater =
            *<value>*.

            Constraint: nrater >= 2.

        (`errno` 61)
            Unable to calculate the ICC due to a division by zero.

            This is often due to degenerate data, for example all scores
            being the same.

        (`errno` 71)
            On entry, mscore = *<value>*.

            Constraint: mscore = 1 or 2.

        (`errno` 91)
            On entry, `alpha` = *<value>*.

            Constraint: 0 < `alpha` < 1.

        (`errno` 92)
            On entry, `alpha` = *<value>*.

            `alpha` is too close to either zero or one.

            This error is unlikely to occur.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 62)
            On entry, a replicate, subject or rater contained all
            missing data.

            All output quantities have been calculated using the reduced
            problem size.

        (`errno` 101)
            `icc` does not fall into the interval [`lci`,`uci`].

            All output quantities have been calculated.

        (`errno` 102)
            The estimate of at least one variance component was
            negative.

            Negative estimates were set to zero and all output
            quantities calculated as documented.

    See Also
    --------
    :meth:`naginterfaces.library.examples.anova.icc_ex.main`
    """

    from ..base.anova import icc as b_icc

    fun_name = 'naginterfaces.library.anova.icc'

    _score = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'score'},
        score,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrep_py = _b_utils._EngineFloat64ArrayType.get_shape(_score, {'fun_name': fun_name, 'entity': 'score'}, exp_rank=3)[0]
    _nrep = _b_utils._EngineIntScalarType(
        _nrep_py, {'fun_name': fun_name, 'entity': '_nrep'},
    ).to_ctypes()
    _nsubj_py = _b_utils._EngineFloat64ArrayType.get_shape(_score, {'fun_name': fun_name, 'entity': 'score'}, exp_rank=3)[1]
    _nsubj = _b_utils._EngineIntScalarType(
        _nsubj_py, {'fun_name': fun_name, 'entity': '_nsubj'},
    ).to_ctypes()
    _nrater_py = _b_utils._EngineFloat64ArrayType.get_shape(_score, {'fun_name': fun_name, 'entity': 'score'}, exp_rank=3)[2]
    _nrater = _b_utils._EngineIntScalarType(
        _nrater_py, {'fun_name': fun_name, 'entity': '_nrater'},
    ).to_ctypes()
    try:
        _return_scalars = b_icc(
            mtype,
            rtype,
            _nrep,
            _nsubj,
            _nrater,
            _score,
            smiss,
            alpha,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _icc_py = _return_scalars.icc
    _lci_py = _return_scalars.lci
    _uci_py = _return_scalars.uci
    _fstat_py = _return_scalars.fstat
    _df1_py = _return_scalars.df1
    _df2_py = _return_scalars.df2
    _pvalue_py = _return_scalars.pvalue
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['icc', 'lci', 'uci', 'fstat', 'df1', 'df2', 'pvalue'],
    )(
        _icc_py,
        _lci_py,
        _uci_py,
        _fstat_py,
        _df1_py,
        _df2_py,
        _pvalue_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data
