# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `inteq` Chapter.

``inteq`` - Integral Equations

This module is concerned with the numerical solution of integral
equations.
Provision will be made for most of the standard types of equation (see
[Background to the Problems]).
The following are, however, specifically excluded:

- Equations arising in the solution of partial differential equations by
  integral equation methods. In cases where the prime purpose of an
  algorithm is the solution of a partial differential equation it will
  normally be included in submodule ``pde``.

- Calculation of inverse integral transforms. This problem falls within
  the scope of submodule ``sum``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05intro.html
"""

# NAG Copyright 2017-2019.

from __future__ import absolute_import

import collections as _collections

from math import sqrt

from ..base import utils as _b_utils
from ..library import machine
from ..library import _utils

def fredholm2_split(lamda, a, b, k1, k2, g, n, ind, data=None):
    u"""
    Linear nonsingular Fredholm integral equation, second kind, split
    kernel.

    ``fredholm2_split`` solves a linear, nonsingular Fredholm equation
    of the second kind with a split kernel.

    For full information please refer to the NAG Library document for
    d05aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05aaf.html

    Parameters
    ----------
    lamda : float
        The value of the parameter lambda of the integral equation.

    a : float
        a, the lower limit of integration.

    b : float
        b, the upper limit of integration.

    k1 : callable retval = k1(x, s, data=None)
        `k1` must evaluate the kernel k(x,s) = k_1(x,s) of the integral
        equation for a <= s <= x.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x and s at which k_1(x,s) is to be evaluated.

        s : float
            The values of x and s at which k_1(x,s) is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the kernel k(x,s) = k_1(x,s) evaluated at `x`
            and `s`.

    k2 : callable retval = k2(x, s, data=None)
        `k2` must evaluate the kernel k(x,s) = k_2(x,s) of the integral
        equation for x < s <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x and s at which k_2(x,s) is to be evaluated.

        s : float
            The values of x and s at which k_2(x,s) is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the kernel k(x,s) = k_2(x,s) evaluated at `x`
            and `s`.

    g : callable retval = g(x, data=None)
        `g` must evaluate the function g(x) for a <= x <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x at which g(x) is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(x) evaluated at `x`.

    n : int
        The number of terms in the Chebyshev series required to
        approximate f(x).

    ind : int
        Determines the forms of the kernel, k(x,s), and the function
        g(x).

        `ind` = 0

          k(x,s) is not centro-symmetric (or no account is to be taken
          of centro-symmetry).

        `ind` = 1

          k(x,s) is centro-symmetric and g(x) is odd.

        `ind` = 2

          k(x,s) is centro-symmetric and g(x) is even.

        `ind` = 3

          k(x,s) is centro-symmetric but g(x) is neither odd nor even.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    f : float, ndarray, shape (`n`)
        The approximate values f_i, for i = 1,2,...,`n`, of f(x)
        evaluated at the first `n` of m+1 Chebyshev points x_i, (see
        [Description]).

        If `ind` = 0 or 3, m = `n`-1.

        If `ind` = 1, m = 2*`n`.

        If `ind` = 2, m = 2*`n`-1.

    c : float, ndarray, shape (`n`)
        The coefficients c_i, for i = 1,2,...,`n`, of the Chebyshev
        series approximation to f(x).

        If `ind` = 1 this series contains polynomials of odd order only
        and if `ind` = 2 the series contains even order polynomials
        only.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > `a`.

        (`errno` 2)
            A failure has occurred due to proximity of an eigenvalue.
    """

    from ..base.inteq import fredholm2_split as b_fredholm2_split

    def cb_k1(x, s, data=None):
        """Wrapping user's k1. Called as base k1."""
        if data is not None:
            _retval_k1 = k1(
                x,
                s,
                data,
            )
        else:
            _retval_k1 = k1(
                x,
                s,
            )
        return _retval_k1

    def cb_k2(x, s, data=None):
        """Wrapping user's k2. Called as base k2."""
        if data is not None:
            _retval_k2 = k2(
                x,
                s,
                data,
            )
        else:
            _retval_k2 = k2(
                x,
                s,
            )
        return _retval_k2

    def cb_g(x, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _retval_g = g(
                x,
                data,
            )
        else:
            _retval_g = g(
                x,
            )
        return _retval_g

    fun_name = 'naginterfaces.library.inteq.fredholm2_split'

    _b_utils._check_callable(k1, {'fun_name': fun_name, 'entity': 'k1'})
    _k1 = cb_k1
    _b_utils._check_callable(k2, {'fun_name': fun_name, 'entity': 'k2'})
    _k2 = cb_k2
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'})
    _g = cb_g
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_fredholm2_split(
        lamda,
        a,
        b,
        _k1,
        _k2,
        _g,
        _f,
        _c,
        _n,
        ind,
        data,
    )
    _f_py = _f
    _c_py = _c
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['f', 'c'],
    )(
        _f_py,
        _c_py,
    )
    return _return_data

def fredholm2_smooth(k, g, lamda, a, b, odorev, ev, n, data=None):
    u"""
    Linear nonsingular Fredholm integral equation, second kind, smooth
    kernel.

    ``fredholm2_smooth`` solves any linear nonsingular Fredholm integral
    equation of the second kind with a smooth kernel.

    For full information please refer to the NAG Library document for
    d05ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05abf.html

    Parameters
    ----------
    k : callable retval = k(x, s, data=None)
        `k` must compute the value of the kernel k(x,s) of the integral
        equation over the square a <= x <= b, a <= s <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The values of x and s at which k(x,s) is to be calculated.

        s : float
            The values of x and s at which k(x,s) is to be calculated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the kernel k(x,s) evaluated at `x` and `s`.

    g : callable retval = g(x, data=None)
        `g` must compute the value of the function g(x) of the integral
        equation in the interval a <= x <= b.

        Parameters
        ~~~~~~~~~~
        x : float
            The value of x at which g(x) is to be calculated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(x) evaluated at `x`.

    lamda : float
        The value of the parameter lambda of the integral equation.

    a : float
        a, the lower limit of integration.

    b : float
        b, the upper limit of integration.

    odorev : bool
        Indicates whether it is known that the solution f(x) is odd or
        even about the mid-point of the range of integration. If
        `odorev` is True then an odd or even solution is sought
        depending upon the value of `ev`.

    ev : bool
        Is ignored if `odorev` is False. Otherwise, if `ev` is True, an
        even solution is sought, whilst if `ev` is False, an odd
        solution is sought.

    n : int
        The number of terms in the Chebyshev series which approximates
        the solution f(x).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    f : float, ndarray, shape (`n`)
        The approximate values f_i, for i = 1,2,...,`n`, of the function
        f(x) at the first `n` of m+1 Chebyshev points (see
        [Description]), where

        [table omitted]

    c : float, ndarray, shape (`n`)
        The coefficients c_i, for i = 1,2,...,`n`, of the Chebyshev
        series approximation to f(x). When `odorev` is True, this series
        contains polynomials of even order only or of odd order only,
        according to `ev` being True or False respectively.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > `a`.

        (`errno` 2)
            A failure has occurred due to proximity of an eigenvalue.
    """

    from ..base.inteq import fredholm2_smooth as b_fredholm2_smooth

    def cb_k(x, s, data=None):
        """Wrapping user's k. Called as base k."""
        if data is not None:
            _retval_k = k(
                x,
                s,
                data,
            )
        else:
            _retval_k = k(
                x,
                s,
            )
        return _retval_k

    def cb_g(x, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _retval_g = g(
                x,
                data,
            )
        else:
            _retval_g = g(
                x,
            )
        return _retval_g

    fun_name = 'naginterfaces.library.inteq.fredholm2_smooth'

    _b_utils._check_callable(k, {'fun_name': fun_name, 'entity': 'k'})
    _k = cb_k
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'})
    _g = cb_g
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_fredholm2_smooth(
        _k,
        _g,
        lamda,
        a,
        b,
        odorev,
        ev,
        _n,
        _f,
        _c,
        data,
    )
    _f_py = _f
    _c_py = _c
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['f', 'c'],
    )(
        _f_py,
        _c_py,
    )
    return _return_data

def volterra2(ck, cg, cf, method, iorder, alim, tlim, nmesh, tol, thresh, lwk, data=None):
    u"""
    Nonlinear Volterra convolution equation, second kind.

    ``volterra2`` computes the solution of a nonlinear convolution
    Volterra integral equation of the second kind using a reducible
    linear multi-step method.

    For full information please refer to the NAG Library document for
    d05ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05baf.html

    Parameters
    ----------
    ck : callable retval = ck(t, data=None)
        `ck` must evaluate the kernel k(t) of the integral equation (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of k at the specified point.

    cg : callable retval = cg(s, y, data=None)
        `cg` must evaluate the function g(s,y(s)) in (1).

        Parameters
        ~~~~~~~~~~
        s : float
            s, the value of the independent variable.

        y : float
            The value of the solution y at the point `s`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g at the specified points.

    cf : callable retval = cf(t, data=None)
        `cf` must evaluate the function f(t) in (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f at the specified point.

    method : str, length 1
        The type of method which you wish to employ.

        `method` = 'A'

          For Adams' type formulae.

        `method` = 'B'

          For backward differentiation formulae.

    iorder : int
        The order of the method to be used.

    alim : float
        a, the lower limit of the integration interval.

    tlim : float
        The final point of the integration interval, T.

    nmesh : int
        The number of equidistant points at which the solution is
        sought.

    tol : float
        The relative accuracy required in the computed values of the
        solution.

    thresh : float
        The threshold value for use in the evaluation of the estimated
        relative errors. For two successive meshes the following
        condition must hold at each point of the coarser mesh

            |Y_1-Y_2|/max(|Y_1|, |Y_2|, |`thresh`|) <= `tol`,

        where Y_1 is the computed solution on the coarser mesh and Y_2
        is the computed solution at the corresponding point in the finer
        mesh. If this condition is not satisfied then the step size is
        halved and the solution is recomputed.

        Note: `thresh` can be used to effect a relative, absolute or
        mixed error test. If `thresh` = 0.0 then pure relative error is
        measured and, if the computed solution is small and `thresh` =
        1.0, absolute error is measured.

    lwk : int
        The dimension of the array `work`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    yn : float, ndarray, shape (`nmesh`)
        `yn`[i-1] contains the most recent approximation of the true
        solution y(t) at the specified point t = `alim`+i*H, for i =
        1,2,...,`nmesh`, where H = (`tlim`-`alim`)/`nmesh`.

    errest : float, ndarray, shape (`nmesh`)
        `errest`[i-1] contains the most recent approximation of the
        relative error in the computed solution at the point t =
        `alim`+i*H, for i = 1,2,...,`nmesh`, where H =
        (`tlim`-`alim`)/`nmesh`.

    work : float, ndarray, shape (`lwk`)
        If `errno` = 5 or 6, `work`[0] contains the size of `lwk`
        required for the algorithm to proceed further.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `method` = 'A' and `iorder` = 2.

            Constraint: if `method` = 'A', 3 <= `iorder` <= 6.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: sqrt(machine precision) <= `tol` <= 1.0.

        (`errno` 1)
            On entry, `alim` = *<value>* and `tlim` = *<value>*.

            Constraint: `tlim` > `alim`.

        (`errno` 1)
            On entry, `alim` = *<value>*.

            Constraint: `alim` >= 0.0.

        (`errno` 1)
            On entry, `method` = 'B' and `iorder` = 6.

            Constraint: if `method` = 'B', 2 <= `iorder` <= 5.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 2 <= `iorder` <= 6.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'A' or 'B'.

        (`errno` 2)
            On entry, `method` = 'A', `iorder` = *<value>* and `nmesh` =
            *<value>*.

            Constraint: if `method` = 'A', `nmesh` >= `iorder`-1.

        (`errno` 2)
            On entry, `method` = 'B', `iorder` = *<value>* and `nmesh` =
            *<value>*.

            Constraint: if `method` = 'B', `nmesh` >= `iorder`.

        (`errno` 3)
            On entry, `lwk` = *<value>*.

            Constraint: `lwk` >= 10*`nmesh`+6; that is, *<value>*.

        (`errno` 4)
            The solution is not converging. See [Further Comments].

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The workspace which has been supplied is too small for the
            required accuracy. The number of extrapolations, so far, is
            *<value>*. If you require one more extrapolation extend the
            size of workspace to: `lwk` = *<value>*.

        (`errno` 6)
            The workspace which has been supplied is too small for the
            required accuracy. The number of extrapolations, so far, is
            *<value>*. If you require one more extrapolation extend the
            size of workspace to: `lwk` = *<value>*.
    """

    from ..base.inteq import volterra2 as b_volterra2

    def cb_ck(t, data=None):
        """Wrapping user's ck. Called as base ck."""
        if data is not None:
            _retval_ck = ck(
                t,
                data,
            )
        else:
            _retval_ck = ck(
                t,
            )
        return _retval_ck

    def cb_cg(s, y, data=None):
        """Wrapping user's cg. Called as base cg."""
        if data is not None:
            _retval_cg = cg(
                s,
                y,
                data,
            )
        else:
            _retval_cg = cg(
                s,
                y,
            )
        return _retval_cg

    def cb_cf(t, data=None):
        """Wrapping user's cf. Called as base cf."""
        if data is not None:
            _retval_cf = cf(
                t,
                data,
            )
        else:
            _retval_cf = cf(
                t,
            )
        return _retval_cf

    fun_name = 'naginterfaces.library.inteq.volterra2'

    _b_utils._check_callable(ck, {'fun_name': fun_name, 'entity': 'ck'})
    _ck = cb_ck
    _b_utils._check_callable(cg, {'fun_name': fun_name, 'entity': 'cg'})
    _cg = cb_cg
    _b_utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = cb_cf
    _nmesh = _b_utils._EngineIntScalarType(
        nmesh, {'fun_name': fun_name, 'entity': 'nmesh'},
    ).to_ctypes()
    _nmesh_py = _b_utils._EngineIntScalarType.to_py(_nmesh)
    _yn = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmesh_py,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmesh_py,),
    )
    _lwk = _b_utils._EngineIntScalarType(
        lwk, {'fun_name': fun_name, 'entity': 'lwk'},
    ).to_ctypes()
    _lwk_py = _b_utils._EngineIntScalarType.to_py(_lwk)
    _work = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lwk_py,),
    )
    try:
        b_volterra2(
            _ck,
            _cg,
            _cf,
            method,
            iorder,
            alim,
            tlim,
            _yn,
            _errest,
            _nmesh,
            tol,
            thresh,
            _work,
            _lwk,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _yn_py = _yn
    _errest_py = _errest
    _work_py = _work
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['yn', 'errest', 'work'],
    )(
        _yn_py,
        _errest_py,
        _work_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def abel2_weak(ck, cf, cg, initwt, tlim, nmesh, comm, iorder=4, tolnl=sqrt(machine.precision()), data=None):
    u"""
    Nonlinear convolution Volterra--Abel equation, second kind, weakly
    singular.

    ``abel2_weak`` computes the solution of a weakly singular nonlinear
    convolution Volterra--Abel integral equation of the second kind
    using a fractional Backward Differentiation Formulae (BDF) method.

    For full information please refer to the NAG Library document for
    d05bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05bdf.html

    Parameters
    ----------
    ck : callable retval = ck(t, data=None)
        `ck` must evaluate the kernel k(t) of the integral equation (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of k(t) evaluated at `t`.

    cf : callable retval = cf(t, data=None)
        `cf` must evaluate the function f(t) in (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f(t) evaluated at `t`.

    cg : callable retval = cg(s, y, data=None)
        `cg` must evaluate the function g(s,y(s)) in (1).

        Parameters
        ~~~~~~~~~~
        s : float
            s, the value of the independent variable.

        y : float
            The value of the solution y at the point `s`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(s,y(s)) evaluated at `s` and `y`.

    initwt : str, length 1
        If the fractional weights required by the method need to be
        calculated by the function then set `initwt` = 'I' (Initial
        call).

        If `initwt` = 'S' (Subsequent call), the function assumes the
        fractional weights have been computed on a previous call and are
        stored in `comm`['work'].

    tlim : float
        The final point of the integration interval, T.

    nmesh : int
        N, the number of equispaced points at which the solution is
        sought.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    iorder : int, optional
        p, the order of the BDF method to be used.

    tolnl : float, optional
        The accuracy required for the computation of the starting value
        and the solution of the nonlinear equation at each step of the
        computation (see [Further Comments]).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    yn : float, ndarray, shape (`nmesh`)
        `yn`[i-1] contains the approximate value of the true solution
        y(t) at the point t = (i-1)*h, for i = 1,2,...,`nmesh`, where h
        = `tlim`/(`nmesh`-1).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tolnl` = *<value>*.

            Constraint: `tolnl` > 10*machine precision.

        (`errno` 1)
            On entry, `nmesh` = *<value>* and `iorder` = *<value>*.

            Constraint: `nmesh` = 2^m+2*`iorder`-1, for some m.

        (`errno` 1)
            On entry, `nmesh` = *<value>* and `iorder` = *<value>*.

            Constraint: `nmesh` >= 2*`iorder`+1.

        (`errno` 1)
            On entry, lwk = *<value>*.

            Constraint: lwk >=
            (2*`iorder`+6)*`nmesh`+8*`iorder`^2-16*`iorder`+1; that is,
            *<value>*.

        (`errno` 1)
            On entry, `tlim` = *<value>*.

            Constraints: `tlim` > 10*machine precision.

        (`errno` 1)
            On entry, `initwt` = *<value>*.

            Constraints: `initwt` = 'I' or 'S'.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 4 <= `iorder` <= 6.

        (`errno` 2)
            An error occurred when trying to compute the starting
            values.

        (`errno` 3)
            An error occurred when trying to compute the solution at a
            specific step.
    """

    from ..base.inteq import abel2_weak as b_abel2_weak

    def cb_ck(t, data=None):
        """Wrapping user's ck. Called as base ck."""
        if data is not None:
            _retval_ck = ck(
                t,
                data,
            )
        else:
            _retval_ck = ck(
                t,
            )
        return _retval_ck

    def cb_cf(t, data=None):
        """Wrapping user's cf. Called as base cf."""
        if data is not None:
            _retval_cf = cf(
                t,
                data,
            )
        else:
            _retval_cf = cf(
                t,
            )
        return _retval_cf

    def cb_cg(s, y, data=None):
        """Wrapping user's cg. Called as base cg."""
        if data is not None:
            _retval_cg = cg(
                s,
                y,
                data,
            )
        else:
            _retval_cg = cg(
                s,
                y,
            )
        return _retval_cg

    fun_name = 'naginterfaces.library.inteq.abel2_weak'

    _b_utils._check_callable(ck, {'fun_name': fun_name, 'entity': 'ck'})
    _ck = cb_ck
    _b_utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = cb_cf
    _b_utils._check_callable(cg, {'fun_name': fun_name, 'entity': 'cg'})
    _cg = cb_cg
    _nmesh = _b_utils._EngineIntScalarType(
        nmesh, {'fun_name': fun_name, 'entity': 'nmesh'},
    ).to_ctypes()
    _nmesh_py = _b_utils._EngineIntScalarType.to_py(_nmesh)
    _yn = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmesh_py,),
    )
    b_abel2_weak(
        _ck,
        _cf,
        _cg,
        initwt,
        iorder,
        tlim,
        tolnl,
        _nmesh,
        _yn,
        comm,
        data,
    )
    _yn_py = _yn
    return _yn_py

def abel1_weak(ck, cf, cg, initwt, tlim, yn, comm, iorder=4, tolnl=sqrt(machine.precision()), data=None):
    u"""
    Nonlinear convolution Volterra--Abel equation, first kind, weakly
    singular.

    ``abel1_weak`` computes the solution of a weakly singular nonlinear
    convolution Volterra--Abel integral equation of the first kind using
    a fractional Backward Differentiation Formulae (BDF) method.

    For full information please refer to the NAG Library document for
    d05be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05bef.html

    Parameters
    ----------
    ck : callable retval = ck(t, data=None)
        `ck` must evaluate the kernel k(t) of the integral equation (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of k(t) evaluated at `t`.

    cf : callable retval = cf(t, data=None)
        `cf` must evaluate the function f(t) in (1).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f(t) evaluated at `t`.

    cg : callable retval = cg(s, y, data=None)
        `cg` must evaluate the function g(s,y(s)) in (1).

        Parameters
        ~~~~~~~~~~
        s : float
            s, the value of the independent variable.

        y : float
            The value of the solution y at the point `s`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(s,y(s)) evaluated at `s` and `y`.

    initwt : str, length 1
        If the fractional weights required by the method need to be
        calculated by the function then set `initwt` = 'I' (Initial
        call).

        If `initwt` = 'S' (Subsequent call), the function assumes the
        fractional weights have been computed by a previous call and are
        stored in `comm`['work'].

    tlim : float
        The final point of the integration interval, T.

    yn : float, array-like, shape (nmesh)
        `yn`[0] must contain the value of y(t) at t = 0 (see [Further
        Comments]).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    iorder : int, optional
        p, the order of the BDF method to be used.

    tolnl : float, optional
        The accuracy required for the computation of the starting value
        and the solution of the nonlinear equation at each step of the
        computation (see [Further Comments]).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    yn : float, ndarray, shape (nmesh)
        `yn`[i-1] contains the approximate value of the true solution
        y(t) at the point t = (i-1)*h, for i = 1,2,...,nmesh, where h =
        `tlim`/(nmesh-1).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tolnl` = *<value>*.

            Constraint: `tolnl` > 10*machine precision.

        (`errno` 1)
            On entry, nmesh = *<value>* and `iorder` = *<value>*.

            Constraint: nmesh = 2^m+2*`iorder`-1, for some m.

        (`errno` 1)
            On entry, nmesh = *<value>* and `iorder` = *<value>*.

            Constraint: nmesh >= 2*`iorder`+1.

        (`errno` 1)
            On entry, lwk = *<value>*.

            Constraint: lwk >=
            (2*`iorder`+6)*nmesh+8*`iorder`^2-16*`iorder`+1; that is,
            *<value>*.

        (`errno` 1)
            On entry, `tlim` = *<value>*.

            Constraint: `tlim` > 10*machine precision

        (`errno` 1)
            On entry, `initwt` = *<value>*.

            Constraint: `initwt` = 'I' or 'S'.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 4 <= `iorder` <= 6.

        (`errno` 2)
            An error occurred when trying to compute the starting
            values.

        (`errno` 3)
            An error occurred when trying to compute the solution at a
            specific step.
    """

    from ..base.inteq import abel1_weak as b_abel1_weak

    def cb_ck(t, data=None):
        """Wrapping user's ck. Called as base ck."""
        if data is not None:
            _retval_ck = ck(
                t,
                data,
            )
        else:
            _retval_ck = ck(
                t,
            )
        return _retval_ck

    def cb_cf(t, data=None):
        """Wrapping user's cf. Called as base cf."""
        if data is not None:
            _retval_cf = cf(
                t,
                data,
            )
        else:
            _retval_cf = cf(
                t,
            )
        return _retval_cf

    def cb_cg(s, y, data=None):
        """Wrapping user's cg. Called as base cg."""
        if data is not None:
            _retval_cg = cg(
                s,
                y,
                data,
            )
        else:
            _retval_cg = cg(
                s,
                y,
            )
        return _retval_cg

    fun_name = 'naginterfaces.library.inteq.abel1_weak'

    _b_utils._check_callable(ck, {'fun_name': fun_name, 'entity': 'ck'})
    _ck = cb_ck
    _b_utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = cb_cf
    _b_utils._check_callable(cg, {'fun_name': fun_name, 'entity': 'cg'})
    _cg = cb_cg
    _yn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yn'},
        yn,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nmesh_py = _b_utils._EngineFloat64ArrayType.get_shape(_yn, {'fun_name': fun_name, 'entity': 'yn'}, exp_rank=1)[0]
    _nmesh = _b_utils._EngineIntScalarType(
        _nmesh_py, {'fun_name': fun_name, 'entity': '_nmesh'},
    ).to_ctypes()
    b_abel1_weak(
        _ck,
        _cf,
        _cg,
        initwt,
        iorder,
        tlim,
        tolnl,
        _nmesh,
        _yn,
        comm,
        data,
    )
    _yn_py = _yn
    return _yn_py

def volterra_weights(method, iorder, nomg, nwt):
    u"""
    Generate weights for use in solving Volterra equations.

    ``volterra_weights`` computes the quadrature weights associated with
    the Adams' methods of orders three to six and the Backward
    Differentiation Formulae (BDF) methods of orders two to five.
    These rules, which are referred to as reducible quadrature rules,
    can then be used in the solution of Volterra integral and
    integro-differential equations.

    For full information please refer to the NAG Library document for
    d05bw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05bwf.html

    Parameters
    ----------
    method : str, length 1
        The type of method to be used.

        `method` = 'A'

          For Adams' type formulae.

        `method` = 'B'

          For Backward Differentiation Formulae.

    iorder : int
        The order of the method to be used. The number of starting
        weights, p is determined by `method` and `iorder`.

        If `method` = 'A', p = `iorder`-1.

        If `method` = 'B', p = `iorder`.

    nomg : int
        The number of convolution weights, n_w.

    nwt : int
        p, the number of columns in the starting weights.

    Returns
    -------
    omega : float, ndarray, shape (`nomg`)
        Contains the first `nomg` convolution weights.

    lensw : int
        The number of rows in the weights W_{i,j}.

    sw : float, ndarray, shape (n, `nwt`)
        `sw`[i-1,j] contains the weights W_{i,j}, for j =
        0,1,...,`nwt`-1, for i = 1,2,...,`lensw`, where n is as defined
        in [Description].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'A' or 'B'.

        (`errno` 2)
            On entry, `nomg` = *<value>*.

            Constraint: `nomg` >= 1.

        (`errno` 2)
            On entry, `iorder` = *<value>*.

            Constraint: 2 <= `iorder` <= 6.

        (`errno` 3)
            On entry, `method` = 'B' and `iorder` = 6.

            Constraint: if `method` = 'B', 2 <= `iorder` <= 5.

        (`errno` 3)
            On entry, `method` = 'A' and `iorder` = 2.

            Constraint: if `method` = 'A', 3 <= `iorder` <= 6.

        (`errno` 4)
            On entry, `method` = 'B', `iorder` = *<value>* and `nwt` =
            *<value>*.

            Constraint: if `method` = 'B', `nwt` = `iorder`.

        (`errno` 4)
            On entry, `method` = 'A', `iorder` = *<value>* and `nwt` =
            *<value>*.

            Constraint: if `method` = 'A', `nwt` = `iorder`-1.

        (`errno` 5)
            On entry, `method` = 'B', `iorder` = *<value>* and `nomg` =
            *<value>*, ldsw = *<value>*.

            Constraint: if `method` = 'B', ldsw >= `nomg`+`iorder`-1.

        (`errno` 5)
            On entry, `method` = 'A', `iorder` = *<value>* and `nomg` =
            *<value>*, ldsw = *<value>*.

            Constraint: if `method` = 'A', ldsw >= `nomg`+`iorder`-2.
    """

    from ..base.inteq import volterra_weights as b_volterra_weights

    fun_name = 'naginterfaces.library.inteq.volterra_weights'

    _nomg = _b_utils._EngineIntScalarType(
        nomg, {'fun_name': fun_name, 'entity': 'nomg'},
    ).to_ctypes()
    _nomg_py = _b_utils._EngineIntScalarType.to_py(_nomg)
    _omega = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nomg_py,),
    )
    _method = _b_utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=1,
    ).to_ctypes()
    _method_py = _b_utils._EngineCharScalarType.to_py(_method)
    _iorder = _b_utils._EngineIntScalarType(
        iorder, {'fun_name': fun_name, 'entity': 'iorder'},
    ).to_ctypes()
    _iorder_py = _b_utils._EngineIntScalarType.to_py(_iorder)
    if _method_py in ('A', 'a'):
        _vl_p_py = _iorder_py-1
    elif _method_py in ('B', 'b'):
        _vl_p_py = _iorder_py
    else:
        _vl_p_py = 0
    _vl_n_py = _nomg_py+_vl_p_py-1
    _nwt = _b_utils._EngineIntScalarType(
        nwt, {'fun_name': fun_name, 'entity': 'nwt'},
    ).to_ctypes()
    _nwt_py = _b_utils._EngineIntScalarType.to_py(_nwt)
    _sw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_vl_n_py, _nwt_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _return_scalar = b_volterra_weights(
        method,
        _iorder,
        _omega,
        _nomg,
        _sw,
        _nwt,
    )
    _omega_py = _omega
    _lensw_py = _return_scalar
    _sw_py = _sw
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['omega', 'lensw', 'sw'],
    )(
        _omega_py,
        _lensw_py,
        _sw_py,
    )
    return _return_data

def abel_weak_weights(iorder, iq, lenfw):
    u"""
    Generate weights for use in solving weakly singular Abel-type
    equations.

    ``abel_weak_weights`` computes the fractional quadrature weights
    associated with the Backward Differentiation Formulae (BDF) of
    orders 4, 5 and 6.
    These weights can then be used in the solution of weakly singular
    equations of Abel type.

    For full information please refer to the NAG Library document for
    d05by

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d05/d05byf.html

    Parameters
    ----------
    iorder : int
        p, the order of the BDF method to be used.

    iq : int
        Determines the number of weights to be computed. By setting `iq`
        to a value, 2^(`iq`+1) fractional convolution weights are
        computed.

    lenfw : int
        The dimension of the array `wt`.

    Returns
    -------
    wt : float, ndarray, shape (`lenfw`)
        The first 2^(`iq`+1) elements of `wt` contains the fractional
        convolution weights omega_i, for i = 0,1,...,2^(`iq`+1)-1. The
        remainder of the array is used as workspace.

    sw : float, ndarray, shape (2^(`iq`+1)+2*`iorder`-1, 2*`iorder`-1)
        `sw`[i-1,j] contains the fractional starting weights W_{i-1,j},
        for j = 0,1,...,2*`iorder`-2, for i = 1,2,...,N, where N =
        (2^(`iq`+1)+2*`iorder`-1).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lenfw` = *<value>* and `iq` = *<value>*.

            Constraint: `lenfw` >= 2^(`iq`+2).

        (`errno` 1)
            On entry, ldsw = *<value>*, `iq` = *<value>* and `iorder` =
            *<value>*.

            Constraint: ldsw >= 2^(`iq`+1)+2*`iorder`-1.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `iorder` = *<value>*.

            Constraint: 4 <= `iorder` <= 6.
    """

    from ..base.inteq import abel_weak_weights as b_abel_weak_weights

    fun_name = 'naginterfaces.library.inteq.abel_weak_weights'

    _lenfw = _b_utils._EngineIntScalarType(
        lenfw, {'fun_name': fun_name, 'entity': 'lenfw'},
    ).to_ctypes()
    _lenfw_py = _b_utils._EngineIntScalarType.to_py(_lenfw)
    _wt = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lenfw_py,),
    )
    _iq = _b_utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = _b_utils._EngineIntScalarType.to_py(_iq)
    _iorder = _b_utils._EngineIntScalarType(
        iorder, {'fun_name': fun_name, 'entity': 'iorder'},
    ).to_ctypes()
    _iorder_py = _b_utils._EngineIntScalarType.to_py(_iorder)
    _sw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2**_iq_py+1+2*_iorder_py-1, 2*_iorder_py-1),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_abel_weak_weights(
        _iorder,
        _iq,
        _lenfw,
        _wt,
        _sw,
    )
    _wt_py = _wt
    _sw_py = _sw
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['wt', 'sw'],
    )(
        _wt_py,
        _sw_py,
    )
    return _return_data
