# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `omp` Chapter.

``omp`` - OpenMP Utilities

This module contains utilities for controlling the OpenMP environment
for your program.
They are based on OpenMP runtime library routines, although their
functionality varies slightly.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06intro.html
"""

# NAG Copyright 2017-2019.

def set_num_threads(num):
    u"""
    Sets the number of threads for OpenMP parallel regions.

    ``set_num_threads`` sets the number of threads requested for
    subsequent OpenMP parallel regions.

    For full information please refer to the NAG Library document for
    x06aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06aaf.html

    Parameters
    ----------
    num : int
        The number of threads requested for subsequent OpenMP parallel
        regions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 1.
    """

    from ..base.omp import set_num_threads as b_set_num_threads

    b_set_num_threads(
        num,
    )

def get_num_threads():
    u"""
    The number of OpenMP threads in the current team.

    ``get_num_threads`` returns the number of OpenMP threads in the
    current team.

    For full information please refer to the NAG Library document for
    x06ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06abf.html

    Returns
    -------
    nt : int
        The returned number of OpenMP threads in the current team.
    """

    from ..base.omp import get_num_threads as b_get_num_threads

    _return_scalar = b_get_num_threads(
    )
    return _return_scalar

def get_max_threads():
    u"""
    An upper bound on the number of threads in the next parallel region.

    ``get_max_threads`` returns an upper bound on the number of threads
    used for subsequent OpenMP parallel regions.

    For full information please refer to the NAG Library document for
    x06ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06acf.html

    Returns
    -------
    ntmax : int
        The returned upper bound on the number of threads used for
        subsequent OpenMP parallel regions.
    """

    from ..base.omp import get_max_threads as b_get_max_threads

    _return_scalar = b_get_max_threads(
    )
    return _return_scalar

def get_thread_num():
    u"""
    The OpenMP thread number of the calling thread.

    ``get_thread_num`` returns the OpenMP thread number of the calling
    thread.

    For full information please refer to the NAG Library document for
    x06ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06adf.html

    Returns
    -------
    tn : int
        The returned OpenMP thread number of the calling thread.
    """

    from ..base.omp import get_thread_num as b_get_thread_num

    _return_scalar = b_get_thread_num(
    )
    return _return_scalar

def in_parallel():
    u"""
    Tests for an active OpenMP parallel region.

    ``in_parallel`` returns 1 if called from within an active OpenMP
    parallel region, and 0 otherwise.

    For full information please refer to the NAG Library document for
    x06af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06aff.html

    Returns
    -------
    inpar : int
        1 if called from within an active OpenMP parallel region, and 0
        otherwise.
    """

    from ..base.omp import in_parallel as b_in_parallel

    _return_scalar = b_in_parallel(
    )
    return _return_scalar

def set_nested(nesting):
    u"""
    Enables or disables nested OpenMP parallelism.

    ``set_nested`` enables or disables nested OpenMP parallelism.

    For full information please refer to the NAG Library document for
    x06ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06agf.html

    Parameters
    ----------
    nesting : int
        If `nesting` = 0, nesting of OpenMP parallel regions is
        disabled, otherwise it is enabled.
    """

    from ..base.omp import set_nested as b_set_nested

    b_set_nested(
        nesting,
    )

def get_nested():
    u"""
    Tests the status of nested OpenMP parallelism.

    ``get_nested`` returns 1 if nested OpenMP parallelism has been
    enabled, and 0 otherwise.

    For full information please refer to the NAG Library document for
    x06ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06ahf.html

    Returns
    -------
    nesting : int
        1 if nested OpenMP parallelism has been enabled, and 0
        otherwise.
    """

    from ..base.omp import get_nested as b_get_nested

    _return_scalar = b_get_nested(
    )
    return _return_scalar

def using_threaded_impl():
    u"""
    Tests whether a threaded NAG Library is being used.

    ``using_threaded_impl`` returns 1 when using a multithreaded
    implementation of the NAG Library and 0 otherwise.

    For full information please refer to the NAG Library document for
    x06xa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06xaf.html

    Returns
    -------
    smplib : int
        1 if using a multithreaded implementation of the NAG Library and
        0 otherwise.
    """

    from ..base.omp import using_threaded_impl as b_using_threaded_impl

    _return_scalar = b_using_threaded_impl(
    )
    return _return_scalar
