# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `orthog` Chapter.

``orthog`` - Orthogonalization

This module is concerned with the orthogonalization of vectors in a
finite dimensional space.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f05/f05intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections

from ..base import utils as _b_utils
from ..library import _utils

def real_gram_schmidt(a, n1):
    u"""
    Gram--Schmidt orthogonalization of n vectors of order m.

    ``real_gram_schmidt`` applies the Schmidt orthogonalization process
    to n vectors in m-dimensional space, n <= m.

    For full information please refer to the NAG Library document for
    f05aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f05/f05aaf.html

    Parameters
    ----------
    a : float, array-like, shape (m, n2)
        Columns `n1` to n2 contain the vectors to be orthogonalized. The
        vectors are stored by columns in elements 1 to m.

    n1 : int
        The indices of the first and last columns of A to be
        orthogonalized.

    Returns
    -------
    a : float, ndarray, shape (m, n2)
        These vectors are overwritten by the orthonormal vectors.

    cc : float
        Is used to indicate linear dependence of the original vectors.
        The nearer `cc` is to 1.0, the more likely vector `icol` is
        dependent on vectors `n1` to `icol`-1. See [Further Comments].

    icol : int
        The column number corresponding to `cc`. See [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>* and n2 = *<value>*.

            Constraint: `n1` <= n2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.orthog import real_gram_schmidt as b_real_gram_schmidt

    fun_name = 'naginterfaces.library.orthog.real_gram_schmidt'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n2_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _return_scalars = b_real_gram_schmidt(
        _a,
        _m,
        n1,
        _n2,
    )
    _a_py = _a
    _cc_py = _return_scalars.cc
    _icol_py = _return_scalars.icol
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'cc', 'icol'],
    )(
        _a_py,
        _cc_py,
        _icol_py,
    )
    return _return_data
