"""
Deprecated wrappers for the NAG C Library (Mark 26.0) g10 Chapter.

g10 - Smoothing in Statistics.

https://www.nag.com/numeric/CL/nagdoc_cl26/html/G10/g10conts.html

NAG Copyright 2017.
"""
from ctypes import c_double
from ctypes import POINTER

from .util import get_library

from .util import Nag_SmoothFitType
from .util import Integer
from .util import NagError
from .util import Nag_SmoothParamMethods
from .util import Nag_WindowType
from .util import Nag_Boolean
from .util import Nag_Smooth_Type

_library_cl = get_library()

g10abc_ctypes = _library_cl.g10abc
g10abc_ctypes.restype = None
g10abc_ctypes.argtypes = [Nag_SmoothFitType, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def g10abc(mode, n, x, y, weights, rho, yhat, coeff, rss, df, res, h, commar, fail):
    """
    Fit cubic smoothing spline, smoothing parameter given.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g10/g10abc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        x, y, weights, yhat, coeff, rss, df, res, h, commar
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        x, y, weights, yhat, coeff, rss, df, res, h, commar
    """
    from .util import nag_double_type_check_and_cast
    _x = nag_double_type_check_and_cast('g10abc', x, 'x')
    _y = nag_double_type_check_and_cast('g10abc', y, 'y')
    _weights = nag_double_type_check_and_cast('g10abc', weights, 'weights')
    _yhat = nag_double_type_check_and_cast('g10abc', yhat, 'yhat')
    _coeff = nag_double_type_check_and_cast('g10abc', coeff, 'coeff')
    _rss = nag_double_type_check_and_cast('g10abc', rss, 'rss')
    _df = nag_double_type_check_and_cast('g10abc', df, 'df')
    _res = nag_double_type_check_and_cast('g10abc', res, 'res')
    _h = nag_double_type_check_and_cast('g10abc', h, 'h')
    _commar = nag_double_type_check_and_cast('g10abc', commar, 'commar')
    g10abc_ctypes(mode, n, _x, _y, _weights, rho, _yhat, _coeff, _rss, _df, _res, _h, _commar, fail)

g10acc_ctypes = _library_cl.g10acc
g10acc_ctypes.restype = None
g10acc_ctypes.argtypes = [Nag_SmoothParamMethods, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, c_double, Integer, POINTER(NagError)]


def g10acc(method, n, x, y, weights, yhat, coeff, rss, df, res, h, crit, rho, u, tol, maxcal, fail):
    """
    Fit cubic smoothing spline, smoothing parameter estimated.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g10/g10acc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        x, y, weights, yhat, coeff, rss, df, res, h, crit, rho
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        x, y, weights, yhat, coeff, rss, df, res, h, crit, rho
    """
    from .util import nag_double_type_check_and_cast
    _x = nag_double_type_check_and_cast('g10acc', x, 'x')
    _y = nag_double_type_check_and_cast('g10acc', y, 'y')
    _weights = nag_double_type_check_and_cast('g10acc', weights, 'weights')
    _yhat = nag_double_type_check_and_cast('g10acc', yhat, 'yhat')
    _coeff = nag_double_type_check_and_cast('g10acc', coeff, 'coeff')
    _rss = nag_double_type_check_and_cast('g10acc', rss, 'rss')
    _df = nag_double_type_check_and_cast('g10acc', df, 'df')
    _res = nag_double_type_check_and_cast('g10acc', res, 'res')
    _h = nag_double_type_check_and_cast('g10acc', h, 'h')
    _crit = nag_double_type_check_and_cast('g10acc', crit, 'crit')
    _rho = nag_double_type_check_and_cast('g10acc', rho, 'rho')
    g10acc_ctypes(method, n, _x, _y, _weights, _yhat, _coeff, _rss, _df, _res, _h, _crit, _rho, u, tol, maxcal, fail)

g10bbc_ctypes = _library_cl.g10bbc
g10bbc_ctypes.restype = None
g10bbc_ctypes.argtypes = [Integer, POINTER(c_double), Nag_WindowType, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Nag_Boolean, POINTER(c_double), POINTER(NagError)]


def g10bbc(n, x, wtype, window, slo, shi, ns, smooth, t, fcall, rcomm, fail):
    """
    Kernel density estimate using Gaussian kernel (thread safe).
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g10/g10bbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        x, window, slo, shi, smooth, t, rcomm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        x, window, slo, shi, smooth, t, rcomm
    """
    from .util import nag_double_type_check_and_cast
    _x = nag_double_type_check_and_cast('g10bbc', x, 'x')
    _window = nag_double_type_check_and_cast('g10bbc', window, 'window')
    _slo = nag_double_type_check_and_cast('g10bbc', slo, 'slo')
    _shi = nag_double_type_check_and_cast('g10bbc', shi, 'shi')
    _smooth = nag_double_type_check_and_cast('g10bbc', smooth, 'smooth')
    _t = nag_double_type_check_and_cast('g10bbc', t, 't')
    _rcomm = nag_double_type_check_and_cast('g10bbc', rcomm, 'rcomm')
    g10bbc_ctypes(n, _x, wtype, _window, _slo, _shi, ns, _smooth, _t, fcall, _rcomm, fail)

g10cac_ctypes = _library_cl.g10cac
g10cac_ctypes.restype = None
g10cac_ctypes.argtypes = [Nag_Smooth_Type, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def g10cac(smoother, n, y, smooth, rough, fail):
    """
    Compute smoothed data sequence using running median smoothers.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g10/g10cac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        y, smooth, rough
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        y, smooth, rough
    """
    from .util import nag_double_type_check_and_cast
    _y = nag_double_type_check_and_cast('g10cac', y, 'y')
    _smooth = nag_double_type_check_and_cast('g10cac', smooth, 'smooth')
    _rough = nag_double_type_check_and_cast('g10cac', rough, 'rough')
    g10cac_ctypes(smoother, n, _y, _smooth, _rough, fail)

g10zac_ctypes = _library_cl.g10zac
g10zac_ctypes.restype = None
g10zac_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def g10zac(n, x, y, weights, nord, xord, yord, wwt, rss, fail):
    """
    Reorder data to give ordered distinct observations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g10/g10zac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        x, y, weights, nord, xord, yord, wwt, rss
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        x, y, weights, xord, yord, wwt, rss
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nord
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _x = nag_double_type_check_and_cast('g10zac', x, 'x')
    _y = nag_double_type_check_and_cast('g10zac', y, 'y')
    _weights = nag_double_type_check_and_cast('g10zac', weights, 'weights')
    _nord = nag_integer_type_check_and_cast('g10zac', nord, 'nord')
    _xord = nag_double_type_check_and_cast('g10zac', xord, 'xord')
    _yord = nag_double_type_check_and_cast('g10zac', yord, 'yord')
    _wwt = nag_double_type_check_and_cast('g10zac', wwt, 'wwt')
    _rss = nag_double_type_check_and_cast('g10zac', rss, 'rss')
    g10zac_ctypes(n, _x, _y, _weights, _nord, _xord, _yord, _wwt, _rss, fail)

nag_smooth_spline_fit = g10abc
nag_smooth_spline_estim = g10acc
nag_kernel_density_gauss = g10bbc
nag_running_median_smoother = g10cac
nag_order_data = g10zac
__all__ = ['g10abc', 'g10acc', 'g10bbc', 'g10cac', 'g10zac',
           'nag_kernel_density_gauss',
           'nag_order_data', 'nag_running_median_smoother',
           'nag_smooth_spline_estim', 'nag_smooth_spline_fit']
