# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `blast` Chapter.

``blast`` - Further Linear Algebra Support Routines

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with basic linear algebra functions which
perform elementary algebraic operations involving scalars, vectors and
matrices.
Most functions for such operations conform either to the specifications
of the BLAS (Basic Linear Algebra Subprograms) or to the specifications
of the BLAST (Basic Linear Algebra Subprograms Technical) Forum.
This module includes functions from the BLAST specifications. Most
(BLAS) functions for such operations are available in submodule
``blas``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections
import sys as _sys

from ..base import utils

def isum(n, x, incx):
    u"""
    Sum elements of integer vector.

    ``isum`` sums the elements of an integer vector.

    For full information please refer to the NAG Library document for
    f16dl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16dlf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : int, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    sumx : int
        The sum of the elements of x.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.isum'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16dlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, sumx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _sumx = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _sumx,
    )
    utils._handle_cb_exception(excs)
    _sumx_py = utils._EngineIntScalarType.to_py(_sumx)
    return_data = _sumx_py
    return return_data

def imax_val(n, x, incx):
    u"""
    Maximum value and location, integer vector.

    ``imax_val`` computes the largest component of an integer vector,
    along with the index of that component.

    For full information please refer to the NAG Library document for
    f16dn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16dnf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : int, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the largest
        component of x. If `n` <= 0 on input then `k` is returned as 0.

    i : int
        i, the largest component of x. If `n` <= 0 on input then `i` is
        returned as 0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.imax_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16dnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_INT, # Output, i.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _i = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _i,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _i_py = utils._EngineIntScalarType.to_py(_i)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'i'],
    )(
        _k_py,
        _i_py,
    )
    return return_data

def imin_val(n, x, incx):
    u"""
    Minimum value and location, integer vector.

    ``imin_val`` computes the smallest component of an integer vector,
    along with the index of that component.

    For full information please refer to the NAG Library document for
    f16dp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16dpf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : int, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the smallest
        component of x. If `n` <= 0 on input then `k` is returned as 0.

    i : int
        i, the smallest component of x. If `n` <= 0 on input then `i` is
        returned as 0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.imin_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16dpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_INT, # Output, i.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _i = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _i,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _i_py = utils._EngineIntScalarType.to_py(_i)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'i'],
    )(
        _k_py,
        _i_py,
    )
    return return_data

def iamax_val(n, x, incx):
    u"""
    Maximum absolute value and location, integer vector.

    ``iamax_val`` computes, with respect to absolute value, the largest
    component of an integer vector, along with the index of that
    component.

    For full information please refer to the NAG Library document for
    f16dq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16dqf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : int, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the largest
        component of x with respect to absolute value. If `n` <= 0 on
        input then `k` is returned as 0.

    i : int
        i, the largest component of x with respect to absolute value. If
        `n` <= 0 on input then `i` is returned as 0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.iamax_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16dqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_INT, # Output, i.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _i = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _i,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _i_py = utils._EngineIntScalarType.to_py(_i)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'i'],
    )(
        _k_py,
        _i_py,
    )
    return return_data

def iamin_val(n, x, incx):
    u"""
    Minimum absolute value and location, integer vector.

    ``iamin_val`` computes, with respect to absolute value, the smallest
    component of an integer vector, along with the index of that
    component.

    For full information please refer to the NAG Library document for
    f16dr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16drf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : int, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the smallest
        component of x with respect to absolute value. If `n` <= 0 on
        input then `k` is returned as 0.

    i : int
        i, the smallest component of x with respect to absolute value.
        If `n` <= 0 on input then `i` is returned as 0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.iamin_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16drft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_INT, # Output, i.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _i = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _i,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _i_py = utils._EngineIntScalarType.to_py(_i)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'i'],
    )(
        _k_py,
        _i_py,
    )
    return return_data

def ddot(conj, n, alpha, x, incx, beta, y, incy, r):
    u"""
    Dot product of two vectors, allows scaling and accumulation.

    ``ddot`` updates a scalar by a scaled dot product of two real
    vectors.

    For full information please refer to the NAG Library document for
    f16ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16eaf.html

    Parameters
    ----------
    conj : int
        `conj` is not referenced and need not be set. The presence of
        this argument in the BLAST standard is for consistency with the
        interface of the complex variant of this function.

    n : int
        n, the number of elements in x and y.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If alpha = 0.0
        or `n` = 0, `x` is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[(i-1)*`incy`], for i =
        1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[(`n`-i)*|`incy`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        If alpha = 0.0 or `n` = 0, `y` is not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    r : float
        The initial value, r, to be updated. If beta = 0.0, `r` need not
        be set on entry.

    Returns
    -------
    r : float
        The value r, scaled by beta and updated by the scaled dot
        product of x and y.

    Raises
    ------
    NagValueError
        (`errno` -5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` -8)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.ddot'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, conj.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Input/Output, r.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _conj = utils._EngineIntScalarType(
        conj, {'fun_name': fun_name, 'entity': 'conj'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-5,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(-8,),
            entity='incy',
        ).manage()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _r = utils._EngineFloat64ScalarType(
        r, {'fun_name': fun_name, 'entity': 'r'},
    ).to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _conj,
        _n,
        _alpha,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _r,
    )
    utils._handle_cb_exception(excs)
    _r_py = utils._EngineFloat64ScalarType.to_py(_r)
    return_data = _r_py
    return return_data

def daxpby(n, alpha, x, incx, beta, y, incy):
    u"""
    Real weighted vector addition.

    ``daxpby`` computes the sum of two scaled vectors, for real vectors
    and scalars.

    For full information please refer to the NAG Library document for
    f16ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16ecf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[(i-1)*`incy`], for i =
        1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[(`n`-i)*|`incy`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` -4)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` -7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.daxpby'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16ecft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(-7,),
            entity='incy',
        ).manage()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dwaxpby(n, alpha, x, incx, beta, y, incy, w, incw):
    u"""
    Real weighted vector addition preserving input.

    ``dwaxpby`` computes the sum of two scaled vectors, preserving
    input, for real scalars and vectors.

    For full information please refer to the NAG Library document for
    f16eh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16ehf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x, y and w.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[(i-1)*`incy`], for i =
        1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[(`n`-i)*|`incy`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        If beta = 0.0 or `n` = 0, `y` is not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    w : float, shape (1+(`n`-1)*|`incw`|), modified in place
        On entry: if |`incw`| != 1, intermediate elements of `w` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements w_i of the vector w will be stored in `w`
        as follows.

        If `incw` > 0, w_i is in `w`[(i-1)*`incw`], for i = 1,2,...,`n`.

        If `incw` < 0, w_i is in `w`[(`n`-i)*|`incw`|], for i =
        1,2,...,`n`.

        Intermediate elements of `w` are not referenced.

    incw : int
        The increment in the subscripts of `w` between successive
        elements of w.

    Raises
    ------
    NagValueError
        (`errno` -4)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` -7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

        (`errno` -9)
            On entry, error in parameter `incw`.

            Constraint: `incw` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.dwaxpby'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16ehft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, w[1+(n-1)*abs(incw)].
        utils._BYREF_INT, # Input, incw.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(-7,),
            entity='incy',
        ).manage()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _incw = utils._EngineIntScalarType(
        incw, {'fun_name': fun_name, 'entity': 'incw'},
    ).to_ctypes()
    _incw_py = utils._EngineIntScalarType.to_py(_incw)
    if not _incw_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incw != 0.'),
            ],
            errcodes=(-9,),
            entity='incw',
        ).manage()
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(1+(_n_py-1)*abs(_incw_py),),
    )
    _w = _w_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _w,
        _incw,
    )
    utils._handle_cb_exception(excs)
    _w_cls.refresh_py(_w)
    return_data = None
    return return_data

def dsum(n, x, incx):
    u"""
    Sum elements of real vector.

    ``dsum`` sums the elements of a real vector.

    For full information please refer to the NAG Library document for
    f16el

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16elf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    sumx : float
        The sum of the elements of x.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.dsum'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16elft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Output, sumx.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _sumx = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _sumx,
    )
    utils._handle_cb_exception(excs)
    _sumx_py = utils._EngineFloat64ScalarType.to_py(_sumx)
    return_data = _sumx_py
    return return_data

def zaxpby(n, alpha, x, incx, beta, y, incy):
    u"""
    Complex weighted vector addition.

    ``zaxpby`` computes the sum of two scaled vectors, for complex
    scalars and vectors.

    For full information please refer to the NAG Library document for
    f16gc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16gcf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[(i-1)*`incy`], for i =
        1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[(`n`-i)*|`incy`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` -4)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` -7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.zaxpby'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16gcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(-7,),
            entity='incy',
        ).manage()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zwaxpby(n, alpha, x, incx, beta, y, incy, w, incw):
    u"""
    Complex weighted vector addition preserving input.

    ``zwaxpby`` computes the sum of two scaled vectors, preserving
    input, for complex scalars and vectors.

    For full information please refer to the NAG Library document for
    f16gh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16ghf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x, y and w.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[(i-1)*`incy`], for i =
        1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[(`n`-i)*|`incy`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        If beta = 0.0 or `n` = 0, `y` is not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    w : complex, shape (1+(`n`-1)*|`incw`|), modified in place
        On entry: if |`incw`| != 1, intermediate elements of `w` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements w_i of the vector w will be stored in `w`
        as follows.

        If `incw` > 0, w_i is in `w`[(i-1)*`incw`], for i = 1,2,...,`n`.

        If `incw` < 0, w_i is in `w`[(`n`-i)*|`incw`|], for i =
        1,2,...,`n`.

        Intermediate elements of `w` are not referenced.

    incw : int
        The increment in the subscripts of `w` between successive
        elements of w.

    Raises
    ------
    NagValueError
        (`errno` -4)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` -7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

        (`errno` -9)
            On entry, error in parameter `incw`.

            Constraint: `incw` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.zwaxpby'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16ghft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, w[1+(n-1)*abs(incw)].
        utils._BYREF_INT, # Input, incw.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(-7,),
            entity='incy',
        ).manage()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _incw = utils._EngineIntScalarType(
        incw, {'fun_name': fun_name, 'entity': 'incw'},
    ).to_ctypes()
    _incw_py = utils._EngineIntScalarType.to_py(_incw)
    if not _incw_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incw != 0.'),
            ],
            errcodes=(-9,),
            entity='incw',
        ).manage()
    _w_cls = utils._EngineComplex128ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(1+(_n_py-1)*abs(_incw_py),),
    )
    _w = _w_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _w,
        _incw,
    )
    utils._handle_cb_exception(excs)
    _w_cls.refresh_py(_w)
    return_data = None
    return return_data

def zsum(n, x, incx):
    u"""
    Sum elements of complex vector.

    ``zsum`` sums the elements of a complex vector.

    For full information please refer to the NAG Library document for
    f16gl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16glf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    sx : complex
        The sum of the elements of the complex vector x.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.zsum'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16glft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Output, sx.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _sx = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _sx,
    )
    utils._handle_cb_exception(excs)
    _sx_py = utils._EngineComplex128ScalarType.to_py(_sx)
    return_data = _sx_py
    return return_data

def dmax_val(n, x, incx):
    u"""
    Maximum value and location, real vector.

    ``dmax_val`` computes the largest component of a real vector, along
    with the index of that component.

    For full information please refer to the NAG Library document for
    f16jn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16jnf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the largest
        component of x. If `n` <= 0 on input then `k` is returned as 0.

    r : float
        r, the largest component of x. If `n` <= 0 on input then `r` is
        returned as 0.0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.dmax_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16jnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Output, r.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _r = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _r,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _r_py = utils._EngineFloat64ScalarType.to_py(_r)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'r'],
    )(
        _k_py,
        _r_py,
    )
    return return_data

def dmin_val(n, x, incx):
    u"""
    Minimum value and location, real vector.

    ``dmin_val`` computes the smallest component of a real vector, along
    with the index of that component.

    For full information please refer to the NAG Library document for
    f16jp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16jpf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the smallest
        component of x. If `n` <= 0 on input then `k` is returned as 0.

    r : float
        r, the smallest component of x. If `n` <= 0 on input then `r` is
        returned as 0.0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.dmin_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16jpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Output, r.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _r = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _r,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _r_py = utils._EngineFloat64ScalarType.to_py(_r)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'r'],
    )(
        _k_py,
        _r_py,
    )
    return return_data

def damax_val(n, x, incx):
    u"""
    Maximum absolute value and location, real vector.

    ``damax_val`` computes, with respect to absolute value, the largest
    component of a real vector, along with the index of that component.

    For full information please refer to the NAG Library document for
    f16jq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16jqf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the largest
        component of x with respect to absolute value. If `n` <= 0 on
        input then `k` is returned as 0.

    r : float
        r, the largest component of x with respect to absolute value. If
        `n` <= 0 on input then `r` is returned as 0.0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.damax_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16jqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Output, r.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _r = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _r,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _r_py = utils._EngineFloat64ScalarType.to_py(_r)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'r'],
    )(
        _k_py,
        _r_py,
    )
    return return_data

def damin_val(n, x, incx):
    u"""
    Minimum absolute value and location, real vector.

    ``damin_val`` computes, with respect to absolute value, the smallest
    component of a real vector, along with the index of that component.

    For full information please refer to the NAG Library document for
    f16jr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16jrf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the smallest
        component of x with respect to absolute value. If `n` <= 0 on
        input then `k` is returned as 0.

    r : float
        r, the smallest component of x with respect to absolute value.
        If `n` <= 0 on input then `r` is returned as 0.0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.damin_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16jrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Output, r.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _r = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _r,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _r_py = utils._EngineFloat64ScalarType.to_py(_r)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'r'],
    )(
        _k_py,
        _r_py,
    )
    return return_data

def zamax_val(n, x, incx):
    u"""
    Maximum absolute value and location, complex vector.

    ``zamax_val`` computes, with respect to absolute value, the largest
    component of a complex vector, along with the index of that
    component.

    For full information please refer to the NAG Library document for
    f16js

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16jsf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the largest
        component of x with respect to absolute value. If `n` <= 0 on
        input then `k` is returned as 0.

    r : float
        r, the largest component of x with respect to absolute value. If
        `n` <= 0 on input then `r` is returned as 0.0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.zamax_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16jsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Output, r.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _r = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _r,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _r_py = utils._EngineFloat64ScalarType.to_py(_r)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'r'],
    )(
        _k_py,
        _r_py,
    )
    return return_data

def zamin_val(n, x, incx):
    u"""
    Minimum absolute value and location, complex vector.

    ``zamin_val`` computes, with respect to absolute value, the smallest
    component of a complex vector, along with the index of that
    component.

    For full information please refer to the NAG Library document for
    f16jt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16jtf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[(i-1)*`incx`], for i =
        1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[(`n`-i)*|`incx`|], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced. If `n` = 0, `x`
        is not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    k : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        k, the index, from the set {1,2,...,`n`}, of the smallest
        component of x with respect to absolute value. If `n` <= 0 on
        input then `k` is returned as 0.

    r : float
        r, the smallest component of x with respect to absolute value.
        If `n` <= 0 on input then `r` is returned as 0.0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.zamin_val'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16jtft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Output, r.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(-3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _r = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _k,
        _r,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _r_py = utils._EngineFloat64ScalarType.to_py(_r)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'r'],
    )(
        _k_py,
        _r_py,
    )
    return return_data

def dgb_norm(inorm, m, n, kl, ku, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real band matrix.

    ``dgb_norm`` calculates the value of the 1-norm, the infinity-norm,
    the Frobenius norm or the maximum absolute value of the elements of
    a real m by n band matrix stored in banded form.

    For full information please refer to the NAG Library document for
    f16rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16rbf.html

    Parameters
    ----------
    inorm : int
        Specifies the value to be returned.

        `inorm` = 171

          The 1-norm.

        `inorm` = 173

          The 2-norm of a row or column vector.

        `inorm` = 174

          The Frobenius (or Euclidean) norm.

        `inorm` = 175

          The infinity-norm.

        `inorm` = 177

          The value max_{i,j}(|a_{ij}|) (not a norm).

    m : int
        m, the number of rows of the matrix A. If `m` <= 0 on input,
        ``dgb_norm`` returns 0.

    n : int
        n, the number of columns of the matrix A. If `n` <= 0 on input,
        ``dgb_norm`` returns 0.

    kl : int
        k_l, the number of subdiagonals within the band of A. If `kl` <=
        0 on input, ``dgb_norm`` returns 0.

    ku : int
        k_u, the number of superdiagonals within the band of A. If `ku`
        <= 0 on input, ``dgb_norm`` returns 0.

    ab : float, shape (`kl`+`ku`+1, `n`)
        The m by n band matrix A.

    Returns
    -------
    nrm : float
        The value of the required norm of the matrix supplied in `ab`.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.dgb_norm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16rbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, inorm.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._EngineFloat64ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._BYREF_INT, # Output, ifail.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _inorm = utils._EngineIntScalarType(
        inorm, {'fun_name': fun_name, 'entity': 'inorm'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _ldab_py = utils._EngineFloat64ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ab = utils._EngineFloat64ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _inorm,
        _m,
        _n,
        _kl,
        _ku,
        _ab,
        _ldab,
        _nrm,
        _ierr,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        if _ierr_py == 999:
            _err_class = utils._EngineAllocationErrorExit(
                fun_name=fun_name,
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
        return_data = _nrm_py
    _err_class.manage()
    return return_data

def zgb_norm(inorm, m, n, kl, ku, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex band matrix.

    ``zgb_norm`` calculates the value of the 1-norm, the infinity-norm,
    the Frobenius norm or the maximum absolute value of the elements of
    a complex m by n band matrix stored in banded packed form.

    For full information please refer to the NAG Library document for
    f16ub

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f16/f16ubf.html

    Parameters
    ----------
    inorm : int
        Specifies the value to be returned.

        `inorm` = 171

          The 1-norm.

        `inorm` = 173

          The 2-norm of a row or column vector.

        `inorm` = 174

          The Frobenius (or Euclidean) norm.

        `inorm` = 175

          The infinity-norm.

        `inorm` = 177

          The value max_{i,j}(|a_{ij}|) (not a norm).

    m : int
        m, the number of rows of the matrix A. If `m` <= 0 on input,
        ``zgb_norm`` returns 0.

    n : int
        n, the number of columns of the matrix A. If `n` <= 0 on input,
        ``zgb_norm`` returns 0.

    kl : int
        k_l, the number of subdiagonals within the band of A. If `kl` <=
        0 on input, ``zgb_norm`` returns 0.

    ku : int
        k_u, the number of superdiagonals within the band of A. If `ku`
        <= 0 on input, ``zgb_norm`` returns 0.

    ab : complex, shape (`kl`+`ku`+1, `n`)
        The m by n band matrix A.

    Returns
    -------
    nrm : float
        The value of the required norm of the matrix supplied in `ab`.
    """

    excs = []
    fun_name = 'naginterfaces.base.blast.zgb_norm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f16ubft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, inorm.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._EngineComplex128ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._BYREF_INT, # Output, ifail.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _inorm = utils._EngineIntScalarType(
        inorm, {'fun_name': fun_name, 'entity': 'inorm'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _ldab_py = utils._EngineComplex128ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ab = utils._EngineComplex128ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _inorm,
        _m,
        _n,
        _kl,
        _ku,
        _ab,
        _ldab,
        _nrm,
        _ierr,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        if _ierr_py == 999:
            _err_class = utils._EngineAllocationErrorExit(
                fun_name=fun_name,
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
        return_data = _nrm_py
    _err_class.manage()
    return return_data
