# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `blas` Chapter.

``blas`` - Linear Algebra Support Routines

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with basic linear algebra functions which
perform elementary algebraic operations involving scalars, vectors and
matrices.
It includes functions which conform to the specifications of the BLAS
(Basic Linear Algebra Subprograms).

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections
import sys as _sys

from ..base import utils

def drotg(a, b):
    u"""
    Generate real plane rotation.

    ``drotg`` generates a real Givens plane rotation.

    For full information please refer to the NAG Library document for
    f06aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06aaf.html

    Parameters
    ----------
    a : float
        The value a, the first element of the vector which determines
        the rotation.

    b : float
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : float
        The value d.

    b : float
        The value z, from which c and s can be reconstructed.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.drotg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input/Output, a.
        utils._BYREF_FLOAT64, # Input/Output, b.
        utils._BYREF_FLOAT64, # Output, c.
        utils._BYREF_FLOAT64, # Output, s.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType.empty_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _a,
        _b,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _a_py = utils._EngineFloat64ScalarType.to_py(_a)
    _b_py = utils._EngineFloat64ScalarType.to_py(_b)
    _c_py = utils._EngineFloat64ScalarType.to_py(_c)
    _s_py = utils._EngineFloat64ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return return_data

def drotgc(a, b):
    u"""
    Generate real plane rotation, storing tangent.

    ``drotgc`` generates a real Givens plane rotation and the tangent of
    that rotation.

    For full information please refer to the NAG Library document for
    f06ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06baf.html

    Parameters
    ----------
    a : float
        The value a, the first element of the vector which determines
        the rotation.

    b : float
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : float
        The value d.

    b : float
        The value t, the tangent of the rotation.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.drotgc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input/Output, a.
        utils._BYREF_FLOAT64, # Input/Output, b.
        utils._BYREF_FLOAT64, # Output, c.
        utils._BYREF_FLOAT64, # Output, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType.empty_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _a_py = utils._EngineFloat64ScalarType.to_py(_a)
    _b_py = utils._EngineFloat64ScalarType.to_py(_b)
    _c_py = utils._EngineFloat64ScalarType.to_py(_c)
    _s_py = utils._EngineFloat64ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return return_data

def dcsg(t):
    u"""
    Recover cosine and sine from given real tangent.

    ``dcsg`` reconstructs the parameters c and s of a real plane
    rotation from the tangent of that rotation.

    For full information please refer to the NAG Library document for
    f06bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bcf.html

    Parameters
    ----------
    t : float
        The value t, the tangent of the rotation.

    Returns
    -------
    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dcsg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, c.
        utils._BYREF_FLOAT64, # Output, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _t = utils._EngineFloat64ScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType.empty_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _t,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _c_py = utils._EngineFloat64ScalarType.to_py(_c)
    _s_py = utils._EngineFloat64ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's'],
    )(
        _c_py,
        _s_py,
    )
    return return_data

def drotj(job, x, y, z):
    u"""
    Generate real Jacobi plane rotation.

    ``drotj`` generates a real Jacobi plane rotation.

    For full information please refer to the NAG Library document for
    f06be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bef.html

    Parameters
    ----------
    job : str, length 1
        Specifies the property which determines the precise form of the
        rotation.

        `job` = 'B'

          c >= 1/sqrt(2).

        `job` = 'S'

          0 <= c <= 1/sqrt(2).

        `job` = 'M'

          |a| >= |b|.

    x : float
        The value x, the (1,1) element of the input matrix.

    y : float
        The value y, the (1,2) or (2,1) element of the input matrix.

    z : float
        The value z. the (2,2) element of the input matrix.

    Returns
    -------
    x : float
        The value a.

    y : float
        The value t, the tangent of the rotation.

    z : float
        The value b.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `job`.

            Constraint: `job` = 'B', 'S' or 'M'.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.drotj'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, job(len=1).
        utils._BYREF_FLOAT64, # Input/Output, x.
        utils._BYREF_FLOAT64, # Input/Output, y.
        utils._BYREF_FLOAT64, # Input/Output, z.
        utils._BYREF_FLOAT64, # Output, c.
        utils._BYREF_FLOAT64, # Output, s.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of job.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _job = utils._EngineCharScalarType(
        job, {'fun_name': fun_name, 'entity': 'job'}, exp_chrlen=1,
    ).to_ctypes()
    _job_py = utils._EngineCharScalarType.to_py(_job)
    if _job_py not in ('B', 'b', 'S', 's', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('job in (\'B\', \'b\', \'S\', \'s\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='job',
        ).manage()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _z = utils._EngineFloat64ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType.empty_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _job,
        _x,
        _y,
        _z,
        _c,
        _s,
        _job._length_,
    )
    utils._handle_cb_exception(excs)
    _x_py = utils._EngineFloat64ScalarType.to_py(_x)
    _y_py = utils._EngineFloat64ScalarType.to_py(_y)
    _z_py = utils._EngineFloat64ScalarType.to_py(_z)
    _c_py = utils._EngineFloat64ScalarType.to_py(_c)
    _s_py = utils._EngineFloat64ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'z', 'c', 's'],
    )(
        _x_py,
        _y_py,
        _z_py,
        _c_py,
        _s_py,
    )
    return return_data

def drot2(x, y, z, c, s):
    u"""
    Apply real similarity rotation to 2 by 2 symmetric matrix.

    ``drot2`` applies a real similarity rotation to a 2 by 2 real
    symmetric matrix.

    For full information please refer to the NAG Library document for
    f06bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bhf.html

    Parameters
    ----------
    x : float
        The value x, the (1,1) element of the input matrix.

    y : float
        The value y, the (1,2) or (2,1) element of the input matrix.

    z : float
        The value z, the (2,2) element of the input matrix.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Returns
    -------
    x : float
        The transformed value x.

    y : float
        The transformed value y.

    z : float
        The transformed value z.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.drot2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input/Output, x.
        utils._BYREF_FLOAT64, # Input/Output, y.
        utils._BYREF_FLOAT64, # Input/Output, z.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _z = utils._EngineFloat64ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _z,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_py = utils._EngineFloat64ScalarType.to_py(_x)
    _y_py = utils._EngineFloat64ScalarType.to_py(_y)
    _z_py = utils._EngineFloat64ScalarType.to_py(_z)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'z'],
    )(
        _x_py,
        _y_py,
        _z_py,
    )
    return return_data

def ddiv(a, b):
    u"""
    Compute quotient of two real scalars, with overflow flag.

    ``ddiv`` computes the quotient of two real scalars.

    For full information please refer to the NAG Library document for
    f06bl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06blf.html

    Parameters
    ----------
    a : float
        The value a.

    b : float
        The value b.

    Returns
    -------
    div : float
        The quotient of the two real scalars.

    fail : bool
        True if a/b would overflow (in which case |q| = flmax) or a = b
        = 0 (in which case q = 0); otherwise False.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ddiv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06blft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Output, div.
        utils._BYREF_LOGICAL, # Output, fail.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _div = utils._EngineFloat64ScalarType.empty_ctypes()
    _fail = utils._EngineLogicalScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _div,
        _fail,
    )
    utils._handle_cb_exception(excs)
    _div_py = utils._EngineFloat64ScalarType.to_py(_div)
    _fail_py = utils._EngineLogicalScalarType.to_py(_fail)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['div', 'fail'],
    )(
        _div_py,
        _fail_py,
    )
    return return_data

def dnorm(scal, ssq):
    u"""
    Compute Euclidean norm from scaled form.

    ``dnorm`` completes the safe computation of the Euclidean length of
    a vector, following a call to ``dssq`` or ``dzssq``.

    For full information please refer to the NAG Library document for
    f06bm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bmf.html

    Parameters
    ----------
    scal : float
        The scaling factor alpha, returned by ``dssq`` or ``dzssq``.

    ssq : float
        The scaled sum of squares xi, returned by ``dssq`` or ``dzssq``.

    Returns
    -------
    nrm2 : float
        The Euclidean length of the vector.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `scal`.

            Constraint: `scal` >= 0.0.

        (`errno` 2)
            On entry, error in parameter `ssq`.

            Constraint: `ssq` >= 1.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dnorm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06bmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, scal.
        utils._BYREF_FLOAT64, # Input, ssq.
        utils._BYREF_FLOAT64, # Output, nrm2.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _scal = utils._EngineFloat64ScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'},
    ).to_ctypes()
    _scal_py = utils._EngineFloat64ScalarType.to_py(_scal)
    if not _scal_py >= 0.0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('scal >= 0.0.'),
            ],
            errcodes=(1,),
            entity='scal',
        ).manage()
    _ssq = utils._EngineFloat64ScalarType(
        ssq, {'fun_name': fun_name, 'entity': 'ssq'},
    ).to_ctypes()
    _ssq_py = utils._EngineFloat64ScalarType.to_py(_ssq)
    if not _ssq_py >= 1.0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('ssq >= 1.0.'),
            ],
            errcodes=(2,),
            entity='ssq',
        ).manage()
    _nrm2 = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _scal,
        _ssq,
        _nrm2,
    )
    utils._handle_cb_exception(excs)
    _nrm2_py = utils._EngineFloat64ScalarType.to_py(_nrm2)
    return_data = _nrm2_py
    return return_data

def dpyth(a, b):
    u"""
    Compute square root of (a^2+b^2), real a and b.

    ``dpyth`` returns the value sqrt(a^2+b^2).

    For full information please refer to the NAG Library document for
    f06bn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bnf.html

    Parameters
    ----------
    a : float
        The value a.

    b : float
        The value b.

    Returns
    -------
    p : float
        The value sqrt(a^2+b^2).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dpyth'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06bnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Output, p.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _p,
    )
    utils._handle_cb_exception(excs)
    _p_py = utils._EngineFloat64ScalarType.to_py(_p)
    return_data = _p_py
    return return_data

def deig2(a, b, c):
    u"""
    Compute eigenvalue of 2 by 2 real symmetric matrix.

    ``deig2`` returns an eigenvalue of a 2 by 2 real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06bp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bpf.html

    Parameters
    ----------
    a : float
        The value a, the (1,1) element of the input matrix.

    b : float
        The value b, the (1,2) or (2,1) element of the input matrix.

    c : float
        The value c, the (2,2) element of the input matrix.

    Returns
    -------
    ev : float
        The eigenvalue.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.deig2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06bpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Output, ev.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _ev = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _c,
        _ev,
    )
    utils._handle_cb_exception(excs)
    _ev_py = utils._EngineFloat64ScalarType.to_py(_ev)
    return_data = _ev_py
    return return_data

def zrotgc(a, b):
    u"""
    Generate complex plane rotation, storing tangent, real cosine.

    ``zrotgc`` generates a complex Givens plane rotation having real
    cosine.

    For full information please refer to the NAG Library document for
    f06ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06caf.html

    Parameters
    ----------
    a : complex
        The value a, the first element of the vector which determines
        the rotation.

    b : complex
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : complex
        The value d.

    b : complex
        The value t, the tangent of the rotation.

    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zrotgc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input/Output, a.
        utils._BYREF_COMPLEX128, # Input/Output, b.
        utils._BYREF_FLOAT64, # Output, c.
        utils._BYREF_COMPLEX128, # Output, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineComplex128ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineComplex128ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType.empty_ctypes()
    _s = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _a_py = utils._EngineComplex128ScalarType.to_py(_a)
    _b_py = utils._EngineComplex128ScalarType.to_py(_b)
    _c_py = utils._EngineFloat64ScalarType.to_py(_c)
    _s_py = utils._EngineComplex128ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return return_data

def zrotgs(a, b):
    u"""
    Generate complex plane rotation, storing tangent, real sine.

    ``zrotgs`` generates a complex Givens plane rotation having real
    sine.

    For full information please refer to the NAG Library document for
    f06cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06cbf.html

    Parameters
    ----------
    a : complex
        The value a, the first element of the vector which determines
        the rotation.

    b : complex
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : complex
        The value d.

    b : complex
        The value t, the tangent of the rotation.

    c : complex
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zrotgs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input/Output, a.
        utils._BYREF_COMPLEX128, # Input/Output, b.
        utils._BYREF_COMPLEX128, # Output, c.
        utils._BYREF_FLOAT64, # Output, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineComplex128ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineComplex128ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _c = utils._EngineComplex128ScalarType.empty_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _a_py = utils._EngineComplex128ScalarType.to_py(_a)
    _b_py = utils._EngineComplex128ScalarType.to_py(_b)
    _c_py = utils._EngineComplex128ScalarType.to_py(_c)
    _s_py = utils._EngineFloat64ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return return_data

def zcsg(t):
    u"""
    Recover cosine and sine from given complex tangent, real cosine.

    ``zcsg`` reconstructs the parameters c (real) and s (complex) of a
    complex plane rotation from the tangent of that rotation.

    For full information please refer to the NAG Library document for
    f06cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ccf.html

    Parameters
    ----------
    t : complex
        The value t, the tangent of the rotation.

    Returns
    -------
    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zcsg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, t.
        utils._BYREF_FLOAT64, # Output, c.
        utils._BYREF_COMPLEX128, # Output, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _t = utils._EngineComplex128ScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType.empty_ctypes()
    _s = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _t,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _c_py = utils._EngineFloat64ScalarType.to_py(_c)
    _s_py = utils._EngineComplex128ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's'],
    )(
        _c_py,
        _s_py,
    )
    return return_data

def zcsgs(t):
    u"""
    Recover cosine and sine from given complex tangent, real sine.

    ``zcsgs`` reconstructs the parameters c (complex) and s (real) of a
    complex plane rotation from the tangent of that rotation.

    For full information please refer to the NAG Library document for
    f06cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06cdf.html

    Parameters
    ----------
    t : complex
        The value t, the tangent of the rotation.

    Returns
    -------
    c : complex
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zcsgs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06cdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, t.
        utils._BYREF_COMPLEX128, # Output, c.
        utils._BYREF_FLOAT64, # Output, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _t = utils._EngineComplex128ScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _c = utils._EngineComplex128ScalarType.empty_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _t,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _c_py = utils._EngineComplex128ScalarType.to_py(_c)
    _s_py = utils._EngineFloat64ScalarType.to_py(_s)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's'],
    )(
        _c_py,
        _s_py,
    )
    return return_data

def zrot2(x, y, z, c, s):
    u"""
    Apply complex similarity rotation to 2 by 2 Hermitian matrix.

    ``zrot2`` applies a complex similarity rotation having real cosine
    and complex sine to a 2 by 2 complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06ch

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06chf.html

    Parameters
    ----------
    x : complex
        The value x, the (1,1) element of the input matrix.

    y : complex
        The value y, the (1,2) element of the input matrix.

    z : complex
        The value z, the (2,2) element of the input matrix.

    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Returns
    -------
    x : complex
        The transformed value x.

    y : complex
        The transformed value y.

    z : complex
        The transformed value z.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zrot2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06chft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input/Output, x.
        utils._BYREF_COMPLEX128, # Input/Output, y.
        utils._BYREF_COMPLEX128, # Input/Output, z.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_COMPLEX128, # Input, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _x = utils._EngineComplex128ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _y = utils._EngineComplex128ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    _z = utils._EngineComplex128ScalarType(
        z, {'fun_name': fun_name, 'entity': 'z'},
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineComplex128ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _z,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_py = utils._EngineComplex128ScalarType.to_py(_x)
    _y_py = utils._EngineComplex128ScalarType.to_py(_y)
    _z_py = utils._EngineComplex128ScalarType.to_py(_z)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'z'],
    )(
        _x_py,
        _y_py,
        _z_py,
    )
    return return_data

def zdiv(a, b):
    u"""
    Compute quotient of two complex scalars, with overflow flag.

    ``zdiv`` computes the quotient of two complex scalars.

    For full information please refer to the NAG Library document for
    f06cl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06clf.html

    Parameters
    ----------
    a : complex
        The value a.

    b : complex
        The value b.

    Returns
    -------
    sdiv : complex
        The value q.

    fail : bool
        True if a/b would overflow (in which case |Re(q)| = |Im(q)| =
        flmax) or a = b = 0 (in which case q = 0); otherwise False.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdiv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06clft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, a.
        utils._BYREF_COMPLEX128, # Input, b.
        utils._BYREF_COMPLEX128, # Output, sdiv.
        utils._BYREF_LOGICAL, # Output, fail.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _a = utils._EngineComplex128ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineComplex128ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _sdiv = utils._EngineComplex128ScalarType.empty_ctypes()
    _fail = utils._EngineLogicalScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _sdiv,
        _fail,
    )
    utils._handle_cb_exception(excs)
    _sdiv_py = utils._EngineComplex128ScalarType.to_py(_sdiv)
    _fail_py = utils._EngineLogicalScalarType.to_py(_fail)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['sdiv', 'fail'],
    )(
        _sdiv_py,
        _fail_py,
    )
    return return_data

def iload(n, con, x, incx):
    u"""
    Broadcast scalar into integer vector.

    ``iload`` broadcasts an integer scalar into an integer vector.

    For full information please refer to the NAG Library document for
    f06db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06dbf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    con : int
        The scalar alpha.

    x : int, shape (1+(`n`-1)*`incx`), modified in place
        On entry: if |`incx`| != 1, intermediate elements of `x` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements x_i of the vector x will be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.iload'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06dbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, con.
        utils._EngineIntArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _con = utils._EngineIntScalarType(
        con, {'fun_name': fun_name, 'entity': 'con'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _con,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def icopy(n, x, incx, y, incy):
    u"""
    Copy integer vector.

    ``icopy`` copies an integer vector.

    For full information please refer to the NAG Library document for
    f06df

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06dff.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : int, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : int, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: if |`incy`| != 1, intermediate elements of `y` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements y_i of the vector y will be stored in `y`
        as follows.

        If `incy` > 0, y_i will be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i will be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.icopy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06dfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineIntArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineIntArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def ddot(n, x, incx, y, incy):
    u"""
    Dot product of two real vectors.

    ``ddot`` computes the scalar product of two real vectors.

    For full information please refer to the NAG Library document for
    f06ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06eaf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Returns
    -------
    d : float
        The scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ddot'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Output, d.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _d = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
        _d,
    )
    utils._handle_cb_exception(excs)
    _d_py = utils._EngineFloat64ScalarType.to_py(_d)
    return_data = _d_py
    return return_data

def daxpy(n, alpha, x, incx, y, incy):
    u"""
    Add scalar times real vector to real vector.

    ``daxpy`` adds a scaled real vector to an unscaled real vector.

    For full information please refer to the NAG Library document for
    f06ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ecf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.daxpy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ecft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dscal(n, alpha, x, incx):
    u"""
    Multiply real vector by scalar.

    ``dscal`` multiplies a real vector by a real scalar.

    For full information please refer to the NAG Library document for
    f06ed

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06edf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the vector alpha x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dscal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06edft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dcopy(n, x, incx, y, incy):
    u"""
    Copy real vector.

    ``dcopy`` makes a copy of a real vector.

    For full information please refer to the NAG Library document for
    f06ef

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06eff.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: if |`incy`| != 1, intermediate elements of `y` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements y_i of the vector y will be stored in `y`
        as follows.

        If `incy` > 0, y_i will be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i will be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dcopy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06efft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dswap(n, x, incx, y, incy):
    u"""
    Swap two real vectors.

    ``dswap`` interchanges two n-element real vectors x and y.

    For full information please refer to the NAG Library document for
    f06eg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06egf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the original vector y stored in the array elements used
        to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the original vector x stored in the array elements used
        to supply the original vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dswap'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06egft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dnrm2(n, x, incx):
    u"""
    Compute Euclidean norm of real vector.

    ``dnrm2`` returns the Euclidean norm of the n-element real vector x.

    For full information please refer to the NAG Library document for
    f06ej

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ejf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    nrm : float
        The Euclidean norm.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dnrm2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ejft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Output, nrm.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _nrm,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dasum(n, x, incx):
    u"""
    Sum absolute values of real vector elements.

    ``dasum`` returns the 1-norm of the n-element real vector x.

    For full information please refer to the NAG Library document for
    f06ek

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ekf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    sumx : float
        The 1-norm of the n-element real vector x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dasum'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ekft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Output, sumx.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _sumx = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _sumx,
    )
    utils._handle_cb_exception(excs)
    _sumx_py = utils._EngineFloat64ScalarType.to_py(_sumx)
    return_data = _sumx_py
    return return_data

def drot(n, x, incx, y, incy, c, s):
    u"""
    Apply real plane rotation.

    ``drot`` applies a real plane rotation to two real vectors.

    For full information please refer to the NAG Library document for
    f06ep

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06epf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the transformed vector x stored in the array elements
        used to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the transformed vector y stored in the array elements
        used to supply the original vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.drot'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06epft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, s.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def ddoti(nz, x, indx, y):
    u"""
    Dot product of a real sparse and a full vector.

    ``ddoti`` computes the scalar product of a sparse real vector,
    stored in compressed form, with a real vector.

    For full information please refer to the NAG Library document for
    f06er

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06erf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    x : float, shape (`nz`)
        The nonzero elements of the sparse vector x.

    indx : int, shape (`nz`)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,`nz`.

    y : float, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    d : float
        The scalar product of the sparse real vector, x, with the real
        vector y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ddoti'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06erft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineFloat64ArrayType.c_type, # Input, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineFloat64ArrayType.c_type, # Input, y[max(indx)].
        utils._BYREF_FLOAT64, # Output, d.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    ).to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    ).to_ctypes()
    _d = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _x,
        _indx,
        _y,
        _d,
    )
    utils._handle_cb_exception(excs)
    _d_py = utils._EngineFloat64ScalarType.to_py(_d)
    return_data = _d_py
    return return_data

def daxpyi(nz, a, x, indx, y):
    u"""
    Add scalar times real sparse vector to a full vector.

    ``daxpyi`` adds a scaled sparse real vector, stored in compressed
    form, to an unscaled real vector.

    For full information please refer to the NAG Library document for
    f06et

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06etf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    a : float
        The scalar alpha.

    x : float, shape (`nz`)
        The nonzero elements of the sparse vector x.

    indx : int, shape (`nz`)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,`nz`.

    y : float, shape (max(`indx`)), modified in place
        On entry: the vector y. Only elements corresponding to indices
        in `indx` are accessed.

        On exit: the updated vector y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.daxpyi'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06etft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._BYREF_FLOAT64, # Input, a.
        utils._EngineFloat64ArrayType.c_type, # Input, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[max(indx)].
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    ).to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _a,
        _x,
        _indx,
        _y,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dgthr(nz, y, x, indx):
    u"""
    Gather real sparse vector.

    ``dgthr`` gathers specified (usually nonzero) elements of a real
    vector y in full storage form into a sparse real vector x in
    compressed form.

    For full information please refer to the NAG Library document for
    f06eu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06euf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the compressed sparse vector x.

    y : float, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    x : float, shape (`nz`), modified in place
        On exit: the compressed vector x.

    indx : int, shape (`nz`)
        `indx`[i-1] must contain the index `y`[i-1], for i =
        1,2,...,`nz`, which is to be gathered into x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgthr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06euft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineFloat64ArrayType.c_type, # Input, y[max(indx)].
        utils._EngineFloat64ArrayType.c_type, # Output, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _y,
        _x,
        _indx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dgthrz(nz, y, x, indx):
    u"""
    Gather and set to zero real sparse vector.

    ``dgthrz`` gathers specified (usually nonzero) elements of a real
    vector y in full storage form into a sparse real vector x in
    compressed form.
    The specified elements of y are set to zero.

    For full information please refer to the NAG Library document for
    f06ev

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06evf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the compressed sparse vector x.

    y : float, shape (max(`indx`)), modified in place
        On entry: the vector y. Only elements corresponding to indices
        in `indx` are accessed.

        On exit: the elements of y corresponding to indices in `indx`
        are set to zero.

    x : float, shape (`nz`), modified in place
        On exit: the compressed vector x.

    indx : int, shape (`nz`)
        `indx`[i-1] must contain the index `y`[i-1], for i =
        1,2,...,`nz`, which is to be gathered into x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `indx`.

            Constraint: the indices must be distinct.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgthrz'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06evft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[max(indx)].
        utils._EngineFloat64ArrayType.c_type, # Output, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    )
    _y = _y_cls.to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _y,
        _x,
        _indx,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dsctr(nz, x, indx, y):
    u"""
    Scatter real sparse vector.

    ``dsctr`` scatters the elements of a sparse real vector x stored in
    compressed form, into a real vector y in full storage form.

    For full information please refer to the NAG Library document for
    f06ew

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ewf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    x : float, shape (`nz`)
        The nonzero elements of the sparse vector x.

    indx : int, shape (`nz`)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,`nz`.

    y : float, shape (max(`indx`)), modified in place
        On entry: the vector y.

        On exit: the vector y, with the elements corresponding to
        indices in `indx` altered.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsctr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ewft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineFloat64ArrayType.c_type, # Input, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[max(indx)].
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    ).to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _x,
        _indx,
        _y,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def droti(nz, x, indx, y, c, s):
    u"""
    Apply plane rotation to a real sparse and a full vector.

    ``droti`` applies a real plane rotation to a sparse real vector and
    a real vector.

    For full information please refer to the NAG Library document for
    f06ex

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06exf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    x : float, shape (`nz`), modified in place
        On entry: the nonzero elements of the sparse vector x.

        On exit: the transformed vector x.

    indx : int, shape (`nz`)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,`nz`.

    y : float, shape (max(`indx`)), modified in place
        On entry: the vector y. Only the elements corresponding to
        indices in `indx` are referenced.

        On exit: the transformed vector y. Only elements corresponding
        to indices in `indx` are altered.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.droti'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06exft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[max(indx)].
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, s.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    )
    _x = _x_cls.to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    )
    _y = _y_cls.to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _x,
        _indx,
        _y,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dvcos(n, j, tolx, x, incx, toly, y, incy):
    u"""
    Compute cosine of angle between two real vectors.

    ``dvcos`` computes the cosine of the angle between two real vectors.

    For full information please refer to the NAG Library document for
    f06fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06faf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    j : int
        If the vector y is supplied in `y`, `j` should be set to 0.
        Otherwise, `j` specifies the index j of the unit vector e_j to
        be used as y.

    tolx : float
        The value tolx, used to determine whether ||x||_2 is effectively
        zero.

        If `tolx` is negative, the value zero is used.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    toly : float
        The value toly, used to determine whether ||y||_2 is effectively
        zero.

        If `toly` is negative, the value zero is used.

    y : float, shape (1+(`n`-1)*|`incy`|)
        If 1 <= `j` <= `n`, `y` is not referenced. Otherwise, `y` holds
        the vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Returns
    -------
    vcos : float
        The cosine of the angle between x and y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dvcos'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06faft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, j.
        utils._BYREF_FLOAT64, # Input, tolx.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, toly.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Output, vcos.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _j = utils._EngineIntScalarType(
        j, {'fun_name': fun_name, 'entity': 'j'},
    ).to_ctypes()
    _tolx = utils._EngineFloat64ScalarType(
        tolx, {'fun_name': fun_name, 'entity': 'tolx'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _toly = utils._EngineFloat64ScalarType(
        toly, {'fun_name': fun_name, 'entity': 'toly'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _vcos = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _j,
        _tolx,
        _x,
        _incx,
        _toly,
        _y,
        _incy,
        _vcos,
    )
    utils._handle_cb_exception(excs)
    _vcos_py = utils._EngineFloat64ScalarType.to_py(_vcos)
    return_data = _vcos_py
    return return_data

def dload(n, con, x, incx):
    u"""
    Broadcast scalar into real vector.

    ``dload`` broadcasts a real scalar into a real vector.

    For full information please refer to the NAG Library document for
    f06fb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fbf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    con : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: if |`incx`| != 1, intermediate elements of `x` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements x_i of the vector x will be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dload'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, con.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _con = utils._EngineFloat64ScalarType(
        con, {'fun_name': fun_name, 'entity': 'con'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _con,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def ddscl(n, d, incd, x, incx):
    u"""
    Multiply real vector by diagonal matrix.

    ``ddscl`` multiplies a real vector by a real diagonal matrix.

    For full information please refer to the NAG Library document for
    f06fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fcf.html

    Parameters
    ----------
    n : int
        n, the number of elements in d and x.

    d : float, shape (1+(`n`-1)*|`incd`|)
        The vector d.

        If `incd` > 0, d_i must be stored in `d`[(i-1)*`incd`], for i =
        1,2,...,`n`.

        If `incd` < 0, d_i must be stored in `d`[ 1-(`n`-i)*`incd` -1],
        for i = 1,2,...,`n`.

    incd : int
        The increment in the subscripts of `d` between successive
        elements of d.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the array `x` must contain the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[ 1+(i-1)*`incx` -1],
        for i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ddscl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, d[1+(n-1)*abs(incd)].
        utils._BYREF_INT, # Input, incd.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incd = utils._EngineIntScalarType(
        incd, {'fun_name': fun_name, 'entity': 'incd'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incd_py = utils._EngineIntScalarType.to_py(_incd)
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(1+(_n_py-1)*abs(_incd_py),),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _d,
        _incd,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def axpzy(n, alpha, x, incx, y, incy):
    u"""
    Multiply real vector by scalar, preserving input vector.

    ``axpzy`` multiplies a real vector by a scalar, preserving the input
    vector.

    For full information please refer to the NAG Library document for
    f06fd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fdf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: if |`incy`| != 1, intermediate elements of `y` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements y_i of the vector y will be stored in `y`
        as follows.

        If `incy` > 0, y_i will be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i will be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.axpzy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def drscl(n, alpha, x, incx):
    u"""
    Multiply real vector by reciprocal of scalar.

    ``drscl`` multiplies a real vector by the reciprocal of a scalar.

    For full information please refer to the NAG Library document for
    f06fe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fef.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the updated vector x, stored in the same array elements
        used to supply the original vector.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `alpha`.

            Constraint: `alpha` != 0.0.

        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.drscl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06feft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
    if not _alpha_py != 0.0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('alpha != 0.0.'),
            ],
            errcodes=(2,),
            entity='alpha',
        ).manage()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dnegv(n, x, incx):
    u"""
    Negate real vector.

    ``dnegv`` negates a real vector.

    For full information please refer to the NAG Library document for
    f06fg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fgf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the vector -x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dnegv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dssq(n, x, incx, scal, sumsq):
    u"""
    Update Euclidean norm of real vector in scaled form.

    ``dssq`` updates the Euclidean norm of real vector in scaled form.

    For full information please refer to the NAG Library document for
    f06fj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fjf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    scal : float
        The scaling factor alpha. On the first call to ``dssq`` `scal` =
        0.0.

    sumsq : float
        The scaled sum of squares xi. On the first call to ``dssq``
        `sumsq` = 1.0.

    Returns
    -------
    scal : float
        The updated scaling factor alpha~ = max_i(alpha,|x_i|).

    sumsq : float
        The updated scaled sum of squares xi~, satisfying: 1 <= xi~ <=
        xi+n.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

        (`errno` 4)
            On entry, error in parameter `scal`.

            Constraint: `scal` >= 0.0.

        (`errno` 5)
            On entry, error in parameter `sumsq`.

            Constraint: `sumsq` >= 1.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dssq'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input/Output, scal.
        utils._BYREF_FLOAT64, # Input/Output, sumsq.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _scal = utils._EngineFloat64ScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'},
    ).to_ctypes()
    _scal_py = utils._EngineFloat64ScalarType.to_py(_scal)
    if not _scal_py >= 0.0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('scal >= 0.0.'),
            ],
            errcodes=(4,),
            entity='scal',
        ).manage()
    _sumsq = utils._EngineFloat64ScalarType(
        sumsq, {'fun_name': fun_name, 'entity': 'sumsq'},
    ).to_ctypes()
    _sumsq_py = utils._EngineFloat64ScalarType.to_py(_sumsq)
    if not _sumsq_py >= 1.0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('sumsq >= 1.0.'),
            ],
            errcodes=(5,),
            entity='sumsq',
        ).manage()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _scal,
        _sumsq,
    )
    utils._handle_cb_exception(excs)
    _scal_py = utils._EngineFloat64ScalarType.to_py(_scal)
    _sumsq_py = utils._EngineFloat64ScalarType.to_py(_sumsq)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['scal', 'sumsq'],
    )(
        _scal_py,
        _sumsq_py,
    )
    return return_data

def dnrm2w(n, w, incw, x, incx):
    u"""
    Compute weighted Euclidean norm of real vector.

    ``dnrm2w`` computes the weighted Euclidean norm of a real vector.

    For full information please refer to the NAG Library document for
    f06fk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fkf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    w : float, shape (1+(`n`-1)*|`incw`|)
        w, the vector of weights.

        If `incw` > 0, w_i must be stored in `w`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incw` < 0, w_i must be stored in `w`[1-(`n`-i)*`incw`-1],
        for i = 1,2,...,`n`.

    incw : int
        The increment in the subscripts of `w` between successive
        elements of w.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    nrm : float
        The weighted Euclidean norm of the real vector, x.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `w`.

            Constraint: the elements of the array must not be negative.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dnrm2w'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, w[1+(n-1)*abs(incw)].
        utils._BYREF_INT, # Input, incw.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Output, nrm.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incw = utils._EngineIntScalarType(
        incw, {'fun_name': fun_name, 'entity': 'incw'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incw_py = utils._EngineIntScalarType.to_py(_incw)
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(1+(_n_py-1)*abs(_incw_py),),
    )
    _w = _w_cls.to_ctypes()
    if not _w_cls.is_nonnegative(1-1, max(1, 1+(_n_py-1)*abs(_incw_py))-1):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('w[0:max(1, 1+(n-1)*abs(incw))] must not be negative.'),
            ],
            errcodes=(2,),
            entity='w',
        ).manage()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _w,
        _incw,
        _x,
        _incx,
        _nrm,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def darang(n, x, incx):
    u"""
    Elements of real vector with largest and smallest absolute value.

    ``darang`` returns the absolutely largest and absolutely smallest
    values from a real vector.

    For full information please refer to the NAG Library document for
    f06fl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06flf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    xmax : float
        The value x_max = max_i|x_i|.

    xmin : float
        The value x_min = min_i|x_i|.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.darang'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06flft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Output, xmax.
        utils._BYREF_FLOAT64, # Output, xmin.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType.empty_ctypes()
    _xmin = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _xmax,
        _xmin,
    )
    utils._handle_cb_exception(excs)
    _xmax_py = utils._EngineFloat64ScalarType.to_py(_xmax)
    _xmin_py = utils._EngineFloat64ScalarType.to_py(_xmin)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['xmax', 'xmin'],
    )(
        _xmax_py,
        _xmin_py,
    )
    return return_data

def drots(n, x, incx, y, incy, c, s):
    u"""
    Apply real symmetric plane rotation to two vectors.

    ``drots`` applies a real symmetric plane rotation to two real
    vectors.

    For full information please refer to the NAG Library document for
    f06fp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fpf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the original vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the transformed vector x stored in the same elements
        used to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the original vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the transformed vector y stored in the same elements
        used to supply the original vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.drots'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dsrotg(pivot, direct, n, alpha, x, incx, c, s):
    u"""
    Generate sequence of real plane rotations.

    ``dsrotg`` generates a sequence of real plane rotations.

    For full information please refer to the NAG Library document for
    f06fq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fqf.html

    Parameters
    ----------
    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'F' (fixed pivot)

          P_k rotates the (1,k+1) plane if `direct` = 'F', or the
          (k,n+1) plane if `direct` = 'B'.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_n...P_2 P_1.

        `direct` = 'B' (backward sequence)

          P = P_1 P_2...P_n.

    n : int
        n, the number of elements in x.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the sequence of plane rotations.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    c : float, shape (`n`), modified in place
        On exit: the values c_k, the cosines of the rotations.

    s : float, shape (`n`), modified in place
        On exit: the values s_k, the sines of the rotations.

    Returns
    -------
    alpha : float
        The scalar beta.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V' or 'F'.

        (`errno` 2)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 6)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsrotg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input/Output, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Output, c[n].
        utils._EngineFloat64ArrayType.c_type, # Output, s[n].
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'F', 'f'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'F\', \'f\').'),
            ],
            errcodes=(1,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='direct',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(6,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _pivot,
        _direct,
        _n,
        _alpha,
        _x,
        _incx,
        _c,
        _s,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
    _x_cls.refresh_py(_x)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    return_data = _alpha_py
    return return_data

def dnhousg(n, alpha, x, incx, tol):
    u"""
    Generate real elementary reflection, NAG style.

    ``dnhousg`` generates a real elementary reflection in the NAG (as
    opposed to LINPACK) style.

    For full information please refer to the NAG Library document for
    f06fr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06frf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and z.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the real elementary reflection.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    tol : float
        The value tol.

    Returns
    -------
    alpha : float
        The scalar beta.

    zeta : float
        The scalar zeta.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dnhousg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06frft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input/Output, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Output, zeta.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _zeta = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _tol,
        _zeta,
    )
    utils._handle_cb_exception(excs)
    _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
    _x_cls.refresh_py(_x)
    _zeta_py = utils._EngineFloat64ScalarType.to_py(_zeta)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'zeta'],
    )(
        _alpha_py,
        _zeta_py,
    )
    return return_data

def dlhousg(n, alpha, x, incx, tol):
    u"""
    Generate real elementary reflection, LINPACK style.

    ``dlhousg`` generates a real elementary reflection in the LINPACK
    (as opposed to NAG) style.

    For full information please refer to the NAG Library document for
    f06fs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fsf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and z.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the real elementary reflection.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    tol : float
        The value tol.

        If `tol` is not in the range (0,1), the value 0 is used for tol.

    Returns
    -------
    alpha : float
        The scalar beta.

    z1 : float
        The scalar zeta.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlhousg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input/Output, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Output, z1.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _z1 = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _tol,
        _z1,
    )
    utils._handle_cb_exception(excs)
    _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
    _x_cls.refresh_py(_x)
    _z1_py = utils._EngineFloat64ScalarType.to_py(_z1)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'z1'],
    )(
        _alpha_py,
        _z1_py,
    )
    return return_data

def dnhous(n, delta, y, incy, zeta, z, incz):
    u"""
    Apply real elementary reflection, NAG style.

    ``dnhous`` applies a NAG (as opposed to LINPACK) style real
    elementary reflection to a real vector.

    For full information please refer to the NAG Library document for
    f06ft

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ftf.html

    Parameters
    ----------
    n : int
        n, the number of elements in y and z.

    delta : float
        The original scalar delta.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the original vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1) *`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i) *`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the transformed stored in the same array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    zeta : float
        The scalar zeta, as returned by ``dnhousg``.

        If zeta = 0, P is assumed to be the unit matrix and the
        transformation is skipped.

    z : float, shape (1+(`n`-1)*|`incz`|)
        The vector z, as returned by ``dnhousg``.

        If `incz` > 0, z_i must be stored in `z`[1+(i-1)*`incz`-1], for
        i = 1,2,...,`n`.

        If `incz` < 0, z_i must be stored in `z`[1-(`n`-i)*`incz`-1],
        for i = 1,2,...,`n`.

    incz : int
        The increment in the subscripts of `z` between successive
        elements of z.

    Returns
    -------
    delta : float
        The transformed scalar delta.

    Raises
    ------
    NagValueError
        (`errno` 5)
            On entry, error in parameter `zeta`.

            Constraint: `zeta` = 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dnhous'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ftft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input/Output, delta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Input, zeta.
        utils._EngineFloat64ArrayType.c_type, # Input, z[1+(n-1)*abs(incz)].
        utils._BYREF_INT, # Input, incz.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _delta = utils._EngineFloat64ScalarType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _zeta = utils._EngineFloat64ScalarType(
        zeta, {'fun_name': fun_name, 'entity': 'zeta'},
    ).to_ctypes()
    _zeta_py = utils._EngineFloat64ScalarType.to_py(_zeta)
    if _n_py == 0:
        if not _zeta_py == 0.0:
            utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    utils._lapack_constraint_err_text('zeta = 0.0.'),
                ],
                errcodes=(5,),
                entity='zeta',
            ).manage()
    _incz = utils._EngineIntScalarType(
        incz, {'fun_name': fun_name, 'entity': 'incz'},
    ).to_ctypes()
    _incz_py = utils._EngineIntScalarType.to_py(_incz)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(1+(_n_py-1)*abs(_incz_py),),
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _delta,
        _y,
        _incy,
        _zeta,
        _z,
        _incz,
    )
    utils._handle_cb_exception(excs)
    _delta_py = utils._EngineFloat64ScalarType.to_py(_delta)
    _y_cls.refresh_py(_y)
    return_data = _delta_py
    return return_data

def dlhous(n, z, incz, z1, alpha, x, incx):
    u"""
    Apply real elementary reflection, LINPACK style.

    ``dlhous`` applies a LINPACK (as opposed to NAG) style real
    elementary reflection to a real vector.

    For full information please refer to the NAG Library document for
    f06fu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fuf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and z.

    z : float, shape (1+(`n`-1)*|`incz`|)
        The vector z, as returned by ``dlhousg``.

        If `incz` > 0, z_i must be stored in `z`[1+(i-1)*`incz`-1], for
        i = 1,2,...,`n`.

        If `incz` < 0, z_i must be stored in `z`[1-(`n`-i)*`incz`-1],
        for i = 1,2,...,`n`.

    incz : int
        The increment in the subscripts of `z` between successive
        elements of z.

    z1 : float
        The scalar zeta, as returned by ``dlhousg``.

        If zeta = 0, P is assumed to be the unit matrix and the
        transformation is skipped.

    alpha : float
        The original scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the original vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the transformed vector x stored in the same array
        elements used to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    alpha : float
        The transformed scalar alpha.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlhous'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06fuft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, z[1+(n-1)*abs(incz)].
        utils._BYREF_INT, # Input, incz.
        utils._BYREF_FLOAT64, # Input, z1.
        utils._BYREF_FLOAT64, # Input/Output, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incz = utils._EngineIntScalarType(
        incz, {'fun_name': fun_name, 'entity': 'incz'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incz_py = utils._EngineIntScalarType.to_py(_incz)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(1+(_n_py-1)*abs(_incz_py),),
    ).to_ctypes()
    _z1 = utils._EngineFloat64ScalarType(
        z1, {'fun_name': fun_name, 'entity': 'z1'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _z,
        _incz,
        _z1,
        _alpha,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
    _x_cls.refresh_py(_x)
    return_data = _alpha_py
    return return_data

def zdotu(n, x, incx, y, incy):
    u"""
    Dot product of two complex vectors, unconjugated.

    ``zdotu`` computes the scalar product of two complex vectors.

    For full information please refer to the NAG Library document for
    f06ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gaf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Returns
    -------
    zdu : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdotu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_COMPLEX128, # Output, zdu.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _zdu = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
        _zdu,
    )
    utils._handle_cb_exception(excs)
    _zdu_py = utils._EngineComplex128ScalarType.to_py(_zdu)
    return_data = _zdu_py
    return return_data

def zdotc(n, x, incx, y, incy):
    u"""
    Dot product of two complex vectors, conjugated.

    ``zdotc`` computes the scalar product of two complex vectors.

    For full information please refer to the NAG Library document for
    f06gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gbf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Returns
    -------
    zdc : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdotc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_COMPLEX128, # Output, zdc.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _zdc = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
        _zdc,
    )
    utils._handle_cb_exception(excs)
    _zdc_py = utils._EngineComplex128ScalarType.to_py(_zdc)
    return_data = _zdc_py
    return return_data

def zaxpy(n, alpha, x, incx, y, incy):
    u"""
    Add scalar times complex vector to complex vector.

    ``zaxpy`` adds a scaled complex vector to an unscaled complex
    vector.

    For full information please refer to the NAG Library document for
    f06gc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gcf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the updated vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zaxpy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zscal(n, alpha, x, incx):
    u"""
    Multiply complex vector by complex scalar.

    ``zscal`` scales a complex vector by a complex scalar.

    For full information please refer to the NAG Library document for
    f06gd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gdf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the vector alpha x stored in the array elements used to
        supply the original vector x(y).

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zscal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zcopy(n, x, incx, y, incy):
    u"""
    Copy complex vector.

    ``zcopy`` copies a complex vector to a complex vector.

    For full information please refer to the NAG Library document for
    f06gf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gff.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: if |`incy`| != 1, intermediate elements of `y` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements y_i of the vector y will be stored in `y`
        as follows.

        If `incy` > 0, y_i will be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i will be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zcopy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zswap(n, x, incx, y, incy):
    u"""
    Swap two complex vectors.

    ``zswap`` interchanges two n-element complex vectors x and y.

    For full information please refer to the NAG Library document for
    f06gg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ggf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the original vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` <= 0, x_i must be stored in `x`[1+(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        On exit: the original vector y stored in the array elements used
        to store the original vector x. Intermediate elements of `x` are
        unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the original vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` <= 0, y_i must be stored in `y`[1+(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        On exit: the original vector x stored in the array elements used
        to store the original vector y. Intermediate elements of `y` are
        unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zswap'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ggft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zdotui(nz, x, indx, y):
    u"""
    Dot product of a complex sparse and a full vector, unconjugated.

    ``zdotui`` computes the scalar product of an unconjugated sparse
    complex vector with a complex vector.

    For full information please refer to the NAG Library document for
    f06gr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06grf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    x : complex, shape (`nz`)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, shape (`nz`)
        `indx` must contain the set of indices J.

    y : complex, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    zdu : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdotui'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06grft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineComplex128ArrayType.c_type, # Input, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineComplex128ArrayType.c_type, # Input, y[max(indx)].
        utils._BYREF_COMPLEX128, # Output, zdu.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    ).to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    ).to_ctypes()
    _zdu = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _x,
        _indx,
        _y,
        _zdu,
    )
    utils._handle_cb_exception(excs)
    _zdu_py = utils._EngineComplex128ScalarType.to_py(_zdu)
    return_data = _zdu_py
    return return_data

def zdotci(nz, x, indx, y):
    u"""
    Dot product of a complex sparse and a full vector, conjugated.

    ``zdotci`` computes the scalar product of a conjugated sparse
    complex vector with a complex vector.

    For full information please refer to the NAG Library document for
    f06gs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gsf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    x : complex, shape (`nz`)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, shape (`nz`)
        `indx` must contain the set of indices J.

    y : complex, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    zdc : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdotci'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineComplex128ArrayType.c_type, # Input, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineComplex128ArrayType.c_type, # Input, y[max(indx)].
        utils._BYREF_COMPLEX128, # Output, zdc.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    ).to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    ).to_ctypes()
    _zdc = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _x,
        _indx,
        _y,
        _zdc,
    )
    utils._handle_cb_exception(excs)
    _zdc_py = utils._EngineComplex128ScalarType.to_py(_zdc)
    return_data = _zdc_py
    return return_data

def zaxpyi(nz, a, x, indx, y):
    u"""
    Add scalar times complex sparse vector to a full vector.

    ``zaxpyi`` adds a scaled sparse complex vector to an unscaled
    complex vector.

    For full information please refer to the NAG Library document for
    f06gt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gtf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    a : complex
        The scalar alpha.

    x : complex, shape (`nz`)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, shape (`nz`)
        The indices of the elements in the compressed vector x.

    y : complex, shape (max(`indx`)), modified in place
        On entry: the vector y. Only elements corresponding to indices
        in `indx` are accessed.

        On exit: the updated vector y.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `indx`.

            Constraint: the indices must be distinct.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zaxpyi'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gtft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._BYREF_COMPLEX128, # Input, a.
        utils._EngineComplex128ArrayType.c_type, # Input, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[max(indx)].
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _a = utils._EngineComplex128ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    ).to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _a,
        _x,
        _indx,
        _y,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zgthr(nz, y, x, indx):
    u"""
    Gather complex sparse vector.

    ``zgthr`` gathers specified (usually nonzero) elements of a complex
    vector y in full storage form into a sparse complex vector x in
    compressed form.

    For full information please refer to the NAG Library document for
    f06gu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06guf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    y : complex, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    x : complex, shape (`nz`), modified in place
        On exit: the compressed vector x.

    indx : int, shape (`nz`)
        `indx` must contain the set of indices J.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgthr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06guft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineComplex128ArrayType.c_type, # Input, y[max(indx)].
        utils._EngineComplex128ArrayType.c_type, # Output, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _y,
        _x,
        _indx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zgthrz(nz, y, x, indx):
    u"""
    Gather and set to zero complex sparse vector.

    ``zgthrz`` gathers specified (usually nonzero) elements of a complex
    vector y in full storage form into a sparse complex vector x in
    compressed form.
    The specified elements of y are set to zero.

    For full information please refer to the NAG Library document for
    f06gv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gvf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    y : complex, shape (max(`indx`)), modified in place
        On entry: the vector y. Only elements corresponding to indices
        in `indx` are accessed.

        On exit: the elements of y corresponding to indices in `indx`
        are set to zero.

    x : complex, shape (`nz`), modified in place
        On exit: the compressed vector x.

    indx : int, shape (`nz`)
        `indx` must contain the set of indices J.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgthrz'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gvft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[max(indx)].
        utils._EngineComplex128ArrayType.c_type, # Output, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    )
    _y = _y_cls.to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _nz,
        _y,
        _x,
        _indx,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zsctr(nz, x, indx, y):
    u"""
    Scatter complex sparse vector.

    ``zsctr`` scatters the elements of a sparse complex vector x stored
    in compressed form, into a complex vector y in full storage form.

    For full information please refer to the NAG Library document for
    f06gw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gwf.html

    Parameters
    ----------
    nz : int
        The number of nonzeros in the sparse vector x.

    x : complex, shape (`nz`)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, shape (`nz`)
        The indices of the elements in the compressed vector x.

    y : complex, shape (max(`indx`)), modified in place
        On entry: the vector y.

        On exit: the vector y, with the elements corresponding to
        indices in `indx` altered.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `indx`.

            Constraint: the indices must be distinct.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsctr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06gwft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nz.
        utils._EngineComplex128ArrayType.c_type, # Input, x[nz].
        utils._EngineIntArrayType.c_type, # Input, indx[nz].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[max(indx)].
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nz = utils._EngineIntScalarType(
        nz, {'fun_name': fun_name, 'entity': 'nz'},
    ).to_ctypes()
    _nz_py = utils._EngineIntScalarType.to_py(_nz)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nz_py,),
    ).to_ctypes()
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_nz_py,),
    )
    _indx = _indx_cls.to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_indx_cls.get_max(0, _nz_py-1),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _nz,
        _x,
        _indx,
        _y,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zload(n, con, x, incx):
    u"""
    Broadcast scalar into complex vector.

    ``zload`` broadcasts a complex scalar into a complex vector.

    For full information please refer to the NAG Library document for
    f06hb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hbf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    con : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: if |`incx`| != 1, intermediate elements of `x` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements x_i of the vector x will be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zload'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, con.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _con = utils._EngineComplex128ScalarType(
        con, {'fun_name': fun_name, 'entity': 'con'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _con,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zdscl(n, d, incd, x, incx):
    u"""
    Multiply complex vector by complex diagonal matrix.

    ``zdscl`` multiplies a complex vector by a complex diagonal matrix.

    For full information please refer to the NAG Library document for
    f06hc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hcf.html

    Parameters
    ----------
    n : int
        n, the number of elements in d and x.

    d : complex, shape (1+(`n`-1)*|`incd`|)
        The vector d.

        If `incd` > 0, d_i must be stored in `d`[(i-1)*`incd`], for i =
        1,2,...,`n`.

        If `incd` < 0, d_i must be stored in `d`[ 1-(`n`-i)*`incd` -1],
        for i = 1,2,...,`n`.

    incd : int
        The increment in the subscripts of `d` between successive
        elements of d.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the array `x` must contain the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[ 1+(i-1)*`incx` -1],
        for i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdscl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, d[1+(n-1)*abs(incd)].
        utils._BYREF_INT, # Input, incd.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incd = utils._EngineIntScalarType(
        incd, {'fun_name': fun_name, 'entity': 'incd'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incd_py = utils._EngineIntScalarType.to_py(_incd)
    _d = utils._EngineComplex128ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(1+(_n_py-1)*abs(_incd_py),),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _d,
        _incd,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zaxpzy(n, alpha, x, incx, y, incy):
    u"""
    Multiply complex vector by complex scalar, preserving input vector.

    ``zaxpzy`` multiplies a complex vector by a scalar, preserving the
    input vector.

    For full information please refer to the NAG Library document for
    f06hd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hdf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: if |`incy`| != 1, intermediate elements of `y` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements y_i of the vector y will be stored in `y`
        as follows.

        If `incy` > 0, y_i will be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i will be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zaxpzy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def znegv(n, x, incx):
    u"""
    Negate complex vector.

    ``znegv`` negates a complex vector.

    For full information please refer to the NAG Library document for
    f06hg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hgf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the vector -x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.znegv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zrot(n, cx, incx, cy, incy, c, s):
    u"""
    Apply plane rotation with real cosine and complex sine.

    ``zrot`` applies a plane rotation with a real cosine and complex
    sine to two complex vectors.

    For full information please refer to the NAG Library document for
    f06hm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hmf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    cx : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `cx`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `cx`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `cx` are not referenced.

        On exit: the transformed vector x stored in the array elements
        used to supply the original vector x.

        Intermediate elements of `cx` are unchanged.

    incx : int
        The increment in the subscripts of `cx` between successive
        elements of x.

    cy : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `cy`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `cy`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `cy` are not referenced.

        On exit: the transformed vector y.

        Intermediate elements of `cy` are unchanged.

    incy : int
        The increment in the subscripts of `cy` between successive
        elements of y.

    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zrot'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, cx[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, cy[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_COMPLEX128, # Input, s.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _cx_cls = utils._EngineComplex128ArrayType(
        cx, {'fun_name': fun_name, 'entity': 'cx'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _cx = _cx_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _cy_cls = utils._EngineComplex128ArrayType(
        cy, {'fun_name': fun_name, 'entity': 'cy'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _cy = _cy_cls.to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineComplex128ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _cx,
        _incx,
        _cy,
        _incy,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _cx_cls.refresh_py(_cx)
    _cy_cls.refresh_py(_cy)
    return_data = None
    return return_data

def zcrot(n, x, incx, y, incy, c, s):
    u"""
    Apply complex plane rotation.

    ``zcrot`` applies a complex plane rotation to two complex vectors.

    For full information please refer to the NAG Library document for
    f06hp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hpf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the transformed vector x stored in the array elements
        used to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the transformed vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    c : complex
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zcrot'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_COMPLEX128, # Input, c.
        utils._BYREF_COMPLEX128, # Input, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _c = utils._EngineComplex128ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineComplex128ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zsrotg(pivot, direct, n, alpha, x, incx, c, s):
    u"""
    Generate sequence of complex plane rotations.

    ``zsrotg`` generates a sequence of complex plane rotations.

    For full information please refer to the NAG Library document for
    f06hq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hqf.html

    Parameters
    ----------
    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'F' (fixed pivot)

          P_k rotates the (1,k+1) plane if `direct` = 'F', or the
          (k,n+1) plane if `direct` = 'B'.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_n...P_2 P_1.

        `direct` = 'B' (backward sequence)

          P = P_1 P_2...P_n.

    n : int
        n, the number of elements in x.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the plane rotations.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    c : float, shape (`n`), modified in place
        On exit: the values c_k, the cosines of the rotations.

    s : complex, shape (`n`), modified in place
        On exit: the values s_k, the sines of the rotations.

    Returns
    -------
    alpha : complex
        The scalar beta.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V' or 'F'.

        (`errno` 2)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 6)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsrotg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input/Output, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Output, c[n].
        utils._EngineComplex128ArrayType.c_type, # Output, s[n].
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'F', 'f'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'F\', \'f\').'),
            ],
            errcodes=(1,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='direct',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(6,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _pivot,
        _direct,
        _n,
        _alpha,
        _x,
        _incx,
        _c,
        _s,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _alpha_py = utils._EngineComplex128ScalarType.to_py(_alpha)
    _x_cls.refresh_py(_x)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    return_data = _alpha_py
    return return_data

def zhousg(n, alpha, x, incx, tol):
    u"""
    Generate complex elementary reflection.

    ``zhousg`` generates a complex elementary reflection.

    For full information please refer to the NAG Library document for
    f06hr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hrf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and z.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the complex elementary reflection.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    tol : float
        The value tol.

    Returns
    -------
    alpha : complex
        The scalar beta.

    theta : complex
        The scalar theta.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhousg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06hrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input/Output, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_COMPLEX128, # Output, theta.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _theta = utils._EngineComplex128ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _tol,
        _theta,
    )
    utils._handle_cb_exception(excs)
    _alpha_py = utils._EngineComplex128ScalarType.to_py(_alpha)
    _x_cls.refresh_py(_x)
    _theta_py = utils._EngineComplex128ScalarType.to_py(_theta)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'theta'],
    )(
        _alpha_py,
        _theta_py,
    )
    return return_data

def zhous(n, delta, y, incy, theta, z, incz):
    u"""
    Apply complex elementary reflection.

    ``zhous`` applies a complex elementary reflection to a complex
    vector.

    For full information please refer to the NAG Library document for
    f06ht

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06htf.html

    Parameters
    ----------
    n : int
        n, the number of elements in y and z.

    delta : complex
        The original scalar delta.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the original vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1) *`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i) *`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the transformed stored in the same array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    theta : complex
        The value theta, as returned by ``zhousg``.

        If theta = 0, P is assumed to be the unit matrix and the
        transformation is skipped.

    z : complex, shape (1+(`n`-1)*|`incz`|)
        The vector z, as returned by ``zhousg``.

        If `incz` > 0, z_i must be stored in `z`[1+(i-1)*`incz`-1], for
        i = 1,2,...,`n`.

        If `incz` < 0, z_i must be stored in `z`[1-(`n`-i)*`incz`-1],
        for i = 1,2,...,`n`.

    incz : int
        The increment in the subscripts of `z` between successive
        elements of z.

    Returns
    -------
    delta : complex
        The transformed scalar delta.

    Raises
    ------
    NagValueError
        (`errno` 5)
            On entry, error in parameter `theta`.

            Constraint: Re(`theta`) <= 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhous'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06htft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input/Output, delta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_COMPLEX128, # Input, theta.
        utils._EngineComplex128ArrayType.c_type, # Input, z[1+(n-1)*abs(incz)].
        utils._BYREF_INT, # Input, incz.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _delta = utils._EngineComplex128ScalarType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _theta = utils._EngineComplex128ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _theta_py = utils._EngineComplex128ScalarType.to_py(_theta)
    if _n_py == 0:
        if not _theta_py.real <= 0.0:
            utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    utils._lapack_constraint_err_text('theta.real <= 0.0.'),
                ],
                errcodes=(5,),
                entity='theta',
            ).manage()
    _incz = utils._EngineIntScalarType(
        incz, {'fun_name': fun_name, 'entity': 'incz'},
    ).to_ctypes()
    _incz_py = utils._EngineIntScalarType.to_py(_incz)
    _z = utils._EngineComplex128ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(1+(_n_py-1)*abs(_incz_py),),
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _delta,
        _y,
        _incy,
        _theta,
        _z,
        _incz,
    )
    utils._handle_cb_exception(excs)
    _delta_py = utils._EngineComplex128ScalarType.to_py(_delta)
    _y_cls.refresh_py(_y)
    return_data = _delta_py
    return return_data

def zdscal(n, alpha, x, incx):
    u"""
    Multiply complex vector by real scalar.

    ``zdscal`` scales a complex vector by a real scalar.

    For full information please refer to the NAG Library document for
    f06jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jdf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    alpha : float
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the vector alpha x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdscal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06jdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dznrm2(n, x, incx):
    u"""
    Compute Euclidean norm of complex vector.

    ``dznrm2`` computes the Euclidean norm of a complex vector.

    For full information please refer to the NAG Library document for
    f06jj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jjf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    nrm : float
        The Euclidean norm of the complex vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dznrm2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06jjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Output, nrm.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _nrm,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dzasum(n, x, incx):
    u"""
    Sum absolute values of complex vector elements.

    ``dzasum`` returns the sum of the absolute values of the real and
    imaginary parts of the elements in a complex vector.

    For full information please refer to the NAG Library document for
    f06jk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jkf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    sumx : float
        The sum of the absolute values of the real and imaginary parts
        of the elements in the complex vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dzasum'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06jkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Output, sumx.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _sumx = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _sumx,
    )
    utils._handle_cb_exception(excs)
    _sumx_py = utils._EngineFloat64ScalarType.to_py(_sumx)
    return_data = _sumx_py
    return return_data

def idamax(n, x, incx):
    u"""
    Index, real vector element with largest absolute value.

    ``idamax`` computes the index of the absolutely largest component of
    a real vector.

    For full information please refer to the NAG Library document for
    f06jl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jlf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    imax : int
        The index of the absolutely largest component of the real
        vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.idamax'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06jlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, imax.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _imax = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _imax,
    )
    utils._handle_cb_exception(excs)
    _imax_py = utils._EngineIntScalarType.to_py(_imax)
    return_data = _imax_py
    return return_data

def izamax(n, x, incx):
    u"""
    Index, complex vector element with largest absolute value.

    ``izamax`` computes the index of the absolutely largest component of
    a complex vector.

    For full information please refer to the NAG Library document for
    f06jm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jmf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Returns
    -------
    imax : int
        The index of the absolutely largest component of a complex
        vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.izamax'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06jmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_INT, # Output, imax.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _imax = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _n,
        _x,
        _incx,
        _imax,
    )
    utils._handle_cb_exception(excs)
    _imax_py = utils._EngineIntScalarType.to_py(_imax)
    return_data = _imax_py
    return return_data

def zddscl(n, d, incd, x, incx):
    u"""
    Multiply complex vector by real diagonal matrix.

    ``zddscl`` multiplies a complex vector by a real diagonal matrix.

    For full information please refer to the NAG Library document for
    f06kc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kcf.html

    Parameters
    ----------
    n : int
        n, the number of elements in d and x.

    d : float, shape (1+(`n`-1)*|`incd`|)
        The vector d.

        If `incd` > 0, d_i must be stored in `d`[(i-1)*`incd`], for i =
        1,2,...,`n`.

        If `incd` < 0, d_i must be stored in `d`[ 1-(`n`-i)*`incd` -1],
        for i = 1,2,...,`n`.

    incd : int
        The increment in the subscripts of `d` between successive
        elements of d.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the array `x` must contain the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[ 1+(i-1)*`incx` -1],
        for i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zddscl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06kcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, d[1+(n-1)*abs(incd)].
        utils._BYREF_INT, # Input, incd.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incd = utils._EngineIntScalarType(
        incd, {'fun_name': fun_name, 'entity': 'incd'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incd_py = utils._EngineIntScalarType.to_py(_incd)
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(1+(_n_py-1)*abs(_incd_py),),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _d,
        _incd,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zdaxpzy(n, alpha, x, incx, y, incy):
    u"""
    Multiply complex vector by real scalar, preserving input vector.

    ``zdaxpzy`` multiplies a complex vector by a real scalar, preserving
    the input vector.

    For full information please refer to the NAG Library document for
    f06kd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kdf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    alpha : float
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: if |`incy`| != 1, intermediate elements of `y` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements y_i of the vector y will be stored in `y`
        as follows.

        If `incy` > 0, y_i will be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i will be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdaxpzy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06kdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zdrscl(n, alpha, x, incx):
    u"""
    Multiply complex vector by reciprocal of real scalar.

    ``zdrscl`` multiplies a complex vector by the reciprocal of a real
    scalar.

    For full information please refer to the NAG Library document for
    f06ke

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kef.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    alpha : float
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the updated vector x, stored in the same array elements
        used to supply the original vector.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `alpha`.

            Constraint: `alpha` != 0.0.

        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdrscl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06keft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
    if not _alpha_py != 0.0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('alpha != 0.0.'),
            ],
            errcodes=(2,),
            entity='alpha',
        ).manage()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zdcopy(n, x, incx, y, incy):
    u"""
    Copy real vector to complex vector.

    ``zdcopy`` copies a real vector to a complex vector.

    For full information please refer to the NAG Library document for
    f06kf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kff.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: if |`incy`| != 1, intermediate elements of `y` may
        contain values and will not be referenced; the other elements
        will be overwritten and need not be set.

        On exit: the elements y_i of the vector y will be stored in `y`
        as follows.

        If `incy` > 0, y_i will be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i will be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdcopy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06kfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dzssq(n, x, incx, scal, sumsq):
    u"""
    Update Euclidean norm of complex vector in scaled form.

    ``dzssq`` updates the Euclidean norm of complex vector in scaled
    form.

    For full information please refer to the NAG Library document for
    f06kj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kjf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : complex, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    scal : float
        The scaling factor alpha. On the first call to ``dzssq`` `scal`
        = 0.0.

    sumsq : float
        The scaled sum of squares xi. On the first call to ``dzssq``
        `sumsq` = 1.0.

    Returns
    -------
    scal : float
        The updated scaling factor alpha~ =
        max_i(alpha,|Re(x_i)|,|Im(x_i)|).

    sumsq : float
        The updated scaled sum of squares xi~, satisfying: 1 <= xi~ <=
        xi+2n.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

        (`errno` 4)
            On entry, error in parameter `scal`.

            Constraint: `scal` >= 0.

        (`errno` 5)
            On entry, error in parameter `sumsq`.

            Constraint: `sumsq` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dzssq'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06kjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input/Output, scal.
        utils._BYREF_FLOAT64, # Input/Output, sumsq.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _scal = utils._EngineFloat64ScalarType(
        scal, {'fun_name': fun_name, 'entity': 'scal'},
    ).to_ctypes()
    _scal_py = utils._EngineFloat64ScalarType.to_py(_scal)
    if not _scal_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('scal >= 0.'),
            ],
            errcodes=(4,),
            entity='scal',
        ).manage()
    _sumsq = utils._EngineFloat64ScalarType(
        sumsq, {'fun_name': fun_name, 'entity': 'sumsq'},
    ).to_ctypes()
    _sumsq_py = utils._EngineFloat64ScalarType.to_py(_sumsq)
    if not _sumsq_py >= 1:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('sumsq >= 1.'),
            ],
            errcodes=(5,),
            entity='sumsq',
        ).manage()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _scal,
        _sumsq,
    )
    utils._handle_cb_exception(excs)
    _scal_py = utils._EngineFloat64ScalarType.to_py(_scal)
    _sumsq_py = utils._EngineFloat64ScalarType.to_py(_sumsq)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['scal', 'sumsq'],
    )(
        _scal_py,
        _sumsq_py,
    )
    return return_data

def idrank(n, x, incx, tol):
    u"""
    Last non-negligible element of real vector.

    ``idrank`` determines the index of the first negligible element of a
    real vector.

    For full information please refer to the NAG Library document for
    f06kl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06klf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    x : float, shape (1+(`n`-1)*`incx`)
        The n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    tol : float
        The value tol.

    Returns
    -------
    k : int
        The index of the first negligible element of the real vector.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.idrank'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06klft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Output, k.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(3,),
            entity='incx',
        ).manage()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _tol,
        _k,
    )
    utils._handle_cb_exception(excs)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    return_data = _k_py
    return return_data

def zdrot(n, x, incx, y, incy, c, s):
    u"""
    Apply real plane rotation to two complex vectors.

    ``zdrot`` applies a real plane rotation to two complex vectors.

    For full information please refer to the NAG Library document for
    f06kp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kpf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x and y.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the transformed vector x stored in the array elements
        used to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

        On exit: the transformed vector y.

        Intermediate elements of `y` are unchanged.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zdrot'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06kpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, s.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _incx,
        _y,
        _incy,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dgemv(trans, m, n, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, real rectangular matrix.

    ``dgemv`` computes the matrix-vector product for a real general
    matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06paf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T' or 'C'

          y←alpha A^Tx+beta y.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    alpha : float
        The scalar alpha.

    a : float, shape (`m`, `n`)
        The m by n matrix A.

    x : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`n`-1)*|`incx`|; if `trans` in ('T',
        'C'): 1+(`m`-1)*|`incx`|; otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`n`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
          for i = 1,2,...,`n`.

        If `trans` = 'T' or 'C',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`m`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`m`-i)*`incx`-1],
          for i = 1,2,...,`m`.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`m`-1)*|`incy`|; if `trans` in ('T',
        'C'): 1+(`n`-1)*|`incy`|; otherwise: 0.

        On entry: the vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`m`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`m`-i)*`incy`-1],
          for i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`n`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
          for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 11)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgemv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06paft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, x[*].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[*].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='trans',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    if _trans_py in ('N', 'n'):
        _x_dim1 = 1+(_n_py-1)*abs(_incx_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _x_dim1 = 1+(_m_py-1)*abs(_incx_py)
    else:
        _x_dim1 = 0
    exp_shape_x = (_x_dim1,)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(11,),
            entity='incy',
        ).manage()
    if _trans_py in ('N', 'n'):
        _y_dim1 = 1+(_m_py-1)*abs(_incy_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _y_dim1 = 1+(_n_py-1)*abs(_incy_py)
    else:
        _y_dim1 = 0
    exp_shape_y = (_y_dim1,)
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=exp_shape_y,
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0)
    _routine_name = utils._EngineRoutineName(*b'f06pa ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _trans,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dgbmv(trans, m, n, kl, ku, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, real rectangular band matrix.

    ``dgbmv`` computes the matrix-vector product for a real general band
    matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pbf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T' or 'C'

          y←alpha A^Tx+beta y.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    alpha : float
        The scalar alpha.

    a : float, shape (`kl`+`ku`+1, `n`)
        The m by n band matrix A.

    x : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`n`-1)*|`incx`|; if `trans` in ('T',
        'C'): 1+(`m`-1)*|`incx`|; otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`n`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
          for i = 1,2,...,`n`.

        If `trans` = 'T' or 'C',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`m`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`m`-i)*`incx`-1],
          for i = 1,2,...,`m`.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`m`-1)*|`incy`|; if `trans` in ('T',
        'C'): 1+(`n`-1)*|`incy`|; otherwise: 0.

        On entry: the vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`m`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`m`-i)*`incy`-1],
          for i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`n`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
          for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` 10)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 13)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgbmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, x[*].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[*].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='trans',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    if not _kl_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('kl >= 0.'),
            ],
            errcodes=(4,),
            entity='kl',
        ).manage()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    if not _ku_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('ku >= 0.'),
            ],
            errcodes=(5,),
            entity='ku',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(10,),
            entity='incx',
        ).manage()
    if _trans_py in ('N', 'n'):
        _x_dim1 = 1+(_n_py-1)*abs(_incx_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _x_dim1 = 1+(_m_py-1)*abs(_incx_py)
    else:
        _x_dim1 = 0
    exp_shape_x = (_x_dim1,)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(13,),
            entity='incy',
        ).manage()
    if _trans_py in ('N', 'n'):
        _y_dim1 = 1+(_m_py-1)*abs(_incy_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _y_dim1 = 1+(_n_py-1)*abs(_incy_py)
    else:
        _y_dim1 = 0
    exp_shape_y = (_y_dim1,)
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=exp_shape_y,
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0)
    _routine_name = utils._EngineRoutineName(*b'f06pb ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _trans,
        _m,
        _n,
        _kl,
        _ku,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dsymv(uplo, n, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, real symmetric matrix.

    ``dsymv`` the matrix-vector product for a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06pc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pcf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    a : float, shape (`n`, `n`)
        The n by n symmetric matrix A.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y, if `beta` = 0.0 `y` need not
        be set.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 7)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 10)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsymv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(7,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(10,),
            entity='incy',
        ).manage()
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1)
    _routine_name = utils._EngineRoutineName(*b'f06pc ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dsbmv(uplo, n, k, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, real symmetric band matrix.

    ``dsbmv`` computes the matrix-vector product for a real symmetric
    band matrix.

    For full information please refer to the NAG Library document for
    f06pd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    alpha : float
        The scalar alpha.

    a : float, shape (`k`+1, `n`)
        The n by n symmetric band matrix A.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y, if `beta` = 0.0 `y` need not
        be set.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 11)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsbmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(3,),
            entity='k',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(11,),
            entity='incy',
        ).manage()
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1)
    _routine_name = utils._EngineRoutineName(*b'f06pd ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dspmv(uplo, n, alpha, ap, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, real symmetric packed matrix.

    ``dspmv`` computes the matrix-vector product for a real symmetric
    matrix stored in packed form.

    For full information please refer to the NAG Library document for
    f06pe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    ap : float, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : float
        The scalar beta.

    y : float, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y, if `beta` = 0.0 `y` need not
        be set.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 6)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 9)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dspmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06peft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _ap = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(6,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(9,),
            entity='incy',
        ).manage()
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1)
    _routine_name = utils._EngineRoutineName(*b'f06pe ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _ap,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def dtrmv(uplo, trans, diag, n, a, x, incx):
    u"""
    Matrix-vector product, real triangular matrix.

    ``dtrmv`` computes the matrix-vector product for a real triangular
    matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pff.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T' or 'C'

          x←A^Tx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    a : float, shape (`n`, `n`)
        The n by n triangular matrix A.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtrmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pf ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dtbmv(uplo, trans, diag, n, k, a, x, incx):
    u"""
    Matrix-vector product, real triangular band matrix.

    ``dtbmv`` computes the matrix-vector product for a real triangular
    band matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T' or 'C'

          x←A^Tx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : float, shape (`k`+1, `n`)
        The n by n triangular band matrix A

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtbmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(9,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pg ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _k,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dtpmv(uplo, trans, diag, n, ap, x, incx):
    u"""
    Matrix-vector product, real triangular packed matrix.

    ``dtpmv`` computes the matrix-vector product for a real triangular
    matrix, or its transpose, stored in packed form.

    For full information please refer to the NAG Library document for
    f06ph

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06phf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T' or 'C'

          x←A^Tx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    ap : float, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 7)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtpmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06phft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _ap = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(7,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06ph ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _ap,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dtrsv(uplo, trans, diag, n, a, x, incx):
    u"""
    System of equations, real triangular matrix.

    ``dtrsv`` solves a real triangular system of equations with a single
    right-hand side.

    For full information please refer to the NAG Library document for
    f06pj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T' or 'C'

          x←A^invTx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    a : float, shape (`n`, `n`)
        The n by n triangular matrix A.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtrsv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pj ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dtbsv(uplo, trans, diag, n, k, a, x, incx):
    u"""
    System of equations, real triangular band matrix.

    ``dtbsv`` solves a real triangular banded system of equations with a
    single right-hand side.

    For full information please refer to the NAG Library document for
    f06pk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pkf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T' or 'C'

          x←A^invTx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : float, shape (`k`+1, `n`)
        The n by n triangular band matrix A

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtbsv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(9,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pk ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _k,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dtpsv(uplo, trans, diag, n, ap, x, incx):
    u"""
    System of equations, real triangular packed matrix.

    ``dtpsv`` solves a real triangular system of equations, stored in
    packed form, with a single right-hand side.

    For full information please refer to the NAG Library document for
    f06pl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06plf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T' or 'C'

          x←A^invTx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    ap : float, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : float, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 7)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtpsv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06plft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _ap = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(7,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pl ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _ap,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def dger(m, n, alpha, x, incx, y, incy, a):
    u"""
    Rank-1 update, real rectangular matrix.

    ``dger`` computes the rank-1 update of a real general matrix.

    For full information please refer to the NAG Library document for
    f06pm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pmf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`m`-1)*|`incx`|)
        The m element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`m`.

        If `incx` < 0, x_i must be stored in `x`[1-(`m`-i)*`incx`], for
        i = 1,2,...,`m`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    a : float, shape (`m`, `n`), modified in place
        On entry: the m by n matrix A.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dger'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(m-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(1,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_m_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(7,),
            entity='incy',
        ).manage()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pm ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _m,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _lda,
        _routine_name,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dsyr(uplo, n, alpha, x, incx, a):
    u"""
    Rank-1 update, real symmetric matrix.

    ``dsyr`` computes the rank-1 update of a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06pp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ppf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix A.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsyr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ppft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pp ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _a,
        _lda,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dspr(uplo, n, alpha, x, incx, ap):
    u"""
    Rank-1 update, real symmetric packed matrix.

    ``dspr`` computes the rank-1 update of a real symmetric matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06pq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pqf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    ap : float, shape (`n`*(`n`+1)/2), modified in place
        On entry: the n by n symmetric matrix A, packed by columns.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dspr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06pqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _ap_cls = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pq ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _ap,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _ap_cls.refresh_py(_ap)
    return_data = None
    return return_data

def dsyr2(uplo, n, alpha, x, incx, y, incy, a):
    u"""
    Rank-2 update, real symmetric matrix.

    ``dsyr2`` computes the rank-2 update of a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06pr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06prf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix A.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsyr2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06prft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(7,),
            entity='incy',
        ).manage()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06pr ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _lda,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dspr2(uplo, n, alpha, x, incx, y, incy, ap):
    u"""
    Rank-2 update, real symmetric packed matrix.

    ``dspr2`` computes the rank-2 update of a real symmetric matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06ps

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06psf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    ap : float, shape (`n`*(`n`+1)/2), modified in place
        On entry: the n by n symmetric matrix A, packed by columns.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dspr2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06psft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(7,),
            entity='incy',
        ).manage()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _ap_cls = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06ps ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _ap,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _ap_cls.refresh_py(_ap)
    return_data = None
    return return_data

def dmcopy(matrix, m, n, a, b):
    u"""
    Matrix copy, real rectangular or trapezoidal matrix.

    ``dmcopy`` performs the matrix-copy operation

        B←A

    where A and B are m by n real general or trapezoidal matrices.

    For full information please refer to the NAG Library document for
    f06qf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qff.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrices A and B.

    n : int
        n, the number of columns of the matrices A and B.

    a : float, shape (`m`, `n`)
        The m by n general or trapezoidal matrix A.

    b : float, shape (`m`, `n`), modified in place
        On exit: the m by n general or trapezoidal matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dmcopy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, matrix(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of matrix.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _matrix = utils._EngineCharScalarType(
        matrix, {'fun_name': fun_name, 'entity': 'matrix'}, exp_chrlen=1,
    ).to_ctypes()
    _matrix_py = utils._EngineCharScalarType.to_py(_matrix)
    if _matrix_py not in ('G', 'g', 'U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('matrix in (\'G\', \'g\', \'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='matrix',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _matrix,
        _m,
        _n,
        _a,
        _lda,
        _b,
        _ldb,
        _matrix._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def dmload(matrix, m, n, con, diag, a):
    u"""
    Matrix initialization, real rectangular matrix.

    ``dmload`` forms the real m by n rectangular or trapezoidal matrix A
    given by

        a_{ij} = {[table omitted].

    For full information please refer to the NAG Library document for
    f06qh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qhf.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    con : float
        The value to be assigned to the off-diagonal elements of A.

    diag : float
        The value to be assigned to the diagonal elements of A.

    a : float, shape (`m`, `n`), modified in place
        On exit: the m by n general or trapezoidal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dmload'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, matrix(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, con.
        utils._BYREF_FLOAT64, # Input, diag.
        utils._EngineFloat64ArrayType.c_type, # Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of matrix.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _matrix = utils._EngineCharScalarType(
        matrix, {'fun_name': fun_name, 'entity': 'matrix'}, exp_chrlen=1,
    ).to_ctypes()
    _matrix_py = utils._EngineCharScalarType.to_py(_matrix)
    if _matrix_py not in ('G', 'g', 'U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('matrix in (\'G\', \'g\', \'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='matrix',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _con = utils._EngineFloat64ScalarType(
        con, {'fun_name': fun_name, 'entity': 'con'},
    ).to_ctypes()
    _diag = utils._EngineFloat64ScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _matrix,
        _m,
        _n,
        _con,
        _diag,
        _a,
        _lda,
        _matrix._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dgeap(side, trans, n, perm, k, b):
    u"""
    Permute rows or columns, real rectangular matrix, permutations
    represented by an integer array.

    ``dgeap`` permutes the rows or columns of a real rectangular matrix
    using an int array of permutations.

    For full information please refer to the NAG Library document for
    f06qj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    n : int
        n, the number of interchanges in the representation of P.

    perm : int, shape (`n`)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(`n`, m); if
        `side`='R': `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R':
        max(`n`, max(`perm`)); otherwise: 0.

        On entry: the original matrix B; B is m by k if `side` = 'L', or
        k by m if `side` = 'R'.

        On exit: the permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgeap'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, perm[n].
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _perm_cls = utils._EngineIntArrayType(
        perm, {'fun_name': fun_name, 'entity': 'perm'},
        exp_shape=(_n_py,),
    )
    _perm = _perm_cls.to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _vl_m_py = _perm_cls.get_max(0, max(1, _n_py)-1)
    if _side_py in ('L', 'l'):
        _b_dim1 = max(_n_py, _vl_m_py)
    elif _side_py in ('R', 'r'):
        _b_dim1 = _k_py
    else:
        _b_dim1 = 0
    if _side_py in ('L', 'l'):
        _b_dim2 = _k_py
    elif _side_py in ('R', 'r'):
        _b_dim2 = max(_n_py, _perm_cls.get_max(0, _n_py-1))
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _trans,
        _n,
        _perm,
        _k,
        _b,
        _ldb,
        _side._length_,
        _trans._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def dgeapr(side, trans, n, perm, k, b):
    u"""
    Permute rows or columns, real rectangular matrix, permutations
    represented by a real array.

    ``dgeapr`` permutes the rows or columns of a real rectangular matrix
    using a double array of permutations.

    For full information please refer to the NAG Library document for
    f06qk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qkf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    n : int
        n, the number of interchanges in the representation of P.

    perm : float, shape (`n`)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(`n`, m); if
        `side`='R': `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R':
        max(`n`, int(max(`perm`))); otherwise: 0.

        On entry: the original matrix B; B is m by k if `side` = 'L', or
        k by m if `side` = 'R'.

        On exit: the permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgeapr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, perm[n].
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _perm_cls = utils._EngineFloat64ArrayType(
        perm, {'fun_name': fun_name, 'entity': 'perm'},
        exp_shape=(_n_py,),
    )
    _perm = _perm_cls.to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _vl_m_py = int(_perm_cls.get_max(0, max(1, _n_py)-1))
    if _side_py in ('L', 'l'):
        _b_dim1 = max(_n_py, _vl_m_py)
    elif _side_py in ('R', 'r'):
        _b_dim1 = _k_py
    else:
        _b_dim1 = 0
    if _side_py in ('L', 'l'):
        _b_dim2 = _k_py
    elif _side_py in ('R', 'r'):
        _b_dim2 = max(_n_py, int(_perm_cls.get_max(0, _n_py-1)))
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _trans,
        _n,
        _perm,
        _k,
        _b,
        _ldb,
        _side._length_,
        _trans._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def dsysrc(uplo, pivot, direct, n, k1, k2, c, s, a):
    u"""
    Orthogonal similarity transformation of real symmetric matrix as a
    sequence of plane rotations.

    ``dsysrc`` performs an orthogonal similarity transformation (as a
    sequence of plane rotations) of a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06qm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qmf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    n : int
        n, the order of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix A.

        On exit: the transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsysrc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input, s[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'T', 't', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'T\', \'t\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(3,),
            entity='direct',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _pivot,
        _direct,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _uplo._length_,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dutr1(n, alpha, x, incx, y, incy, a, c, s):
    u"""
    QR factorization by sequence of plane rotations, rank-1 update of
    real upper triangular matrix.

    ``dutr1`` performs a QR factorization (as a sequence of plane
    rotations) of a real upper triangular matrix that has been modified
    by a rank-1 update.

    For full information please refer to the NAG Library document for
    f06qp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qpf.html

    Parameters
    ----------
    n : int
        n, the order of the matrices U and R.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the sequence of plane rotations.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : float, shape (1+(`n`-1)*`incy`)
        The n-element vector y. y_i must be stored in
        `y`[1+(i-1)*`incy`], for i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U.

        On exit: the upper triangular matrix R.

    c : float, shape (`n`-1), modified in place
        On exit: the cosines of the rotations Q_k, for k = 1,2,...,n-1.

    s : float, shape (`n`-1), modified in place
        On exit: the sines of the rotations Q_k, for k = 1,2,...,n-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

        (`errno` 6)
            On entry, error in parameter `incy`.

            Constraint: `incy` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dutr1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input, y[1+(n-1)*incy].
        utils._BYREF_INT, # Input, incy.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Output, c[n-1].
        utils._EngineFloat64ArrayType.c_type, # Output, s[n-1].
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(1,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy > 0.'),
            ],
            errcodes=(6,),
            entity='incy',
        ).manage()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*_incy_py,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py-1,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py-1,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _lda,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _a_cls.refresh_py(_a)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    return_data = None
    return return_data

def dutupd(n, alpha, x, incx, a, c, s):
    u"""
    QR factorization by sequence of plane rotations, real upper
    triangular matrix augmented by a full row.

    ``dutupd`` performs a QR factorization (as a sequence of plane
    rotations) of a real upper triangular matrix that has been augmented
    by a full row.

    For full information please refer to the NAG Library document for
    f06qq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qqf.html

    Parameters
    ----------
    n : int
        n, the order of the matrices U and R.

    alpha : float
        The scalar alpha.

    x : float, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`-1], for i = 1,2,...,`n`.

        On exit: the referenced elements are overwritten by the tangents
        of the rotations Q_k, for k = 1,2,...,n.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U.

        On exit: the upper triangular matrix R.

    c : float, shape (`n`), modified in place
        On exit: the values c_k, the cosines of the rotations Q_k, for k
        = 1,2,...,n.

    s : float, shape (`n`), modified in place
        On exit: the values s_k, the sines of the rotations Q_k, for k =
        1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dutupd'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Output, c[n].
        utils._EngineFloat64ArrayType.c_type, # Output, s[n].
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(1,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _a,
        _lda,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _a_cls.refresh_py(_a)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    return_data = None
    return return_data

def duhqr(side, n, k1, k2, c, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, real upper
    Hessenberg matrix.

    ``duhqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a real upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06qr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qrf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    n : int
        n, the order of the matrix H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1), modified in place
        On exit: `c`[k] holds c_k, the cosine of the rotation P_k, for k
        = k_1,...,k_2-1.

    s : float, shape (`k2`-1), modified in place
        On entry: the nonzero subdiagonal elements of H: `s`[k] must
        hold h_{k+1,k}, for k = k_1,...,k_2-1.

        On exit: `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, shape (`n`, `n`), modified in place
        On entry: the upper triangular part of the n by n upper
        Hessenberg matrix H.

        On exit: the upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.duhqr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Output, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dusqr(side, n, k1, k2, c, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, real upper
    spiked matrix.

    ``dusqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a real upper spiked matrix.

    For full information please refer to the NAG Library document for
    f06qs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qsf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    n : int
        n, the order of the matrix H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1), modified in place
        On exit: `c`[k] holds c_k, the cosine of the rotation P_k, for k
        = k_1,...,k_2-1.

    s : float, shape (`k2`-1), modified in place
        On entry: the nonzero elements of the spike of H: `s`[k] must
        hold h_{k_2,k} if `side` = 'L', and h_{k+1,k_1} if `side` = 'R',
        for k = k_1,...,k_2-1.

        On exit: `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, shape (`n`, `n`), modified in place
        On entry: the upper triangular part of the n by n upper spiked
        matrix H.

        On exit: the upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dusqr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Output, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dutsqr(side, n, k1, k2, c, s, a):
    u"""
    QR factorization of UP or RQ factorization of PU, U real upper
    triangular, P a sequence of plane rotations.

    ``dutsqr`` performs a QR or RQ factorization of the product of a
    real upper triangular matrix and a real matrix of plane rotations.

    For full information please refer to the NAG Library document for
    f06qt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qtf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether P is applied from the left or the right in the
        transformation.

        `side` = 'L'

          P is applied from the left.

        `side` = 'R'

          P is applied from the right.

    n : int
        n, the order of the matrices U and R.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1), modified in place
        On entry: `c`[k] must hold the cosine of the rotation P_k, for k
        = k_1,...,k_2-1.

        On exit: `c`[k] holds the cosine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1), modified in place
        On entry: `s`[k] must hold the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

        On exit: `s`[k] holds the sine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U.

        On exit: the upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dutsqr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qtft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    )
    _c = _c_cls.to_ctypes()
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dutsrh(side, n, k1, k2, c, s, a):
    u"""
    Compute upper Hessenberg matrix by sequence of plane rotations, real
    upper triangular matrix.

    ``dutsrh`` transforms a real upper triangular matrix to an upper
    Hessenberg matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06qv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qvf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    n : int
        n, the order of the matrices U and H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1), modified in place
        On entry: `s`[k] must hold s_k, the sine of the rotation P_k,
        for k = k_1,...,k_2-1.

        On exit: `s`[k] holds h_{k+1,k}, the subdiagonal element of H,
        for k = k_1,...,k_2-1.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U.

        On exit: the upper triangular part of the upper Hessenberg
        matrix H.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dutsrh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qvft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dutsrs(side, n, k1, k2, c, s, a):
    u"""
    Compute upper spiked matrix by sequence of plane rotations, real
    upper triangular matrix.

    ``dutsrs`` transforms a real upper triangular matrix to an upper
    spiked matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06qw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qwf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    n : int
        n, the order of the matrices U and H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1), modified in place
        On entry: `s`[k] must hold s_k, the sine of the rotation P_k,
        for k = k_1,...,k_2-1.

        On exit: `s`[k] holds a nonzero element of the spike of H:
        h_{k_2,k} if `side` = 'L', or h_{k+1,k_1} if `side` = 'R', for k
        = k_1,...,k_2-1.

    a : float, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U.

        On exit: the upper triangular part of the upper spiked matrix H.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dutsrs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qwft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dgesrc(side, pivot, direct, m, n, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, real rectangular matrix.

    ``dgesrc`` applies a sequence of plane rotations to a real
    rectangular matrix.

    For full information please refer to the NAG Library document for
    f06qx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qxf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

        If `m` < 1, an immediate return is effected.

    n : int
        n, the number of columns of the matrix A.

        If `n` < 1, an immediate return is effected.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, shape (`m`, `n`), modified in place
        On entry: the m by n matrix A.

        On exit: the transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgesrc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06qxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input, s[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'T', 't', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'T\', \'t\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(3,),
            entity='direct',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(4,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(5,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _pivot,
        _direct,
        _m,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dlange(norm, m, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real general matrix.

    ``dlange`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real m by n matrix.

    For full information please refer to the NAG Library document for
    f06ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06raf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    m : int
        m, the number of rows of the matrix A.

        When `m` = 0, ``dlange`` is set to zero.

    n : int
        n, the number of columns of the matrix A.

        When `n` = 0, ``dlange`` is set to zero.

    a : float, shape (`m`, `n`)
        The m by n matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real m by n
        matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlange'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06raft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _m,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlangb(norm, n, kl, ku, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real band matrix.

    ``dlangb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n band matrix.

    For full information please refer to the NAG Library document for
    f06rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rbf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlangb`` returns zero.

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    ab : float, shape (`kl`+`ku`+1, `n`)
        The n by n band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlangb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._EngineFloat64ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    if not _kl_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('kl >= 0.'),
            ],
            errcodes=(3,),
            entity='kl',
        ).manage()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    if not _ku_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('ku >= 0.'),
            ],
            errcodes=(4,),
            entity='ku',
        ).manage()
    _ldab_py = utils._EngineFloat64ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _ab = utils._EngineFloat64ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _kl,
        _ku,
        _ab,
        _ldab,
        _nrm,
        _work,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlansy(norm, uplo, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric matrix.

    ``dlansy`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric matrix.

    For full information please refer to the NAG Library document for
    f06rc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rcf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlansy`` returns zero.

    a : float, shape (`n`, `n`)
        The n by n symmetric matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlansy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlansp(norm, uplo, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric matrix, packed storage.

    ``dlansp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06rd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rdf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlansp`` returns zero.

    ap : float, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlansp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _ap = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _ap,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlansb(norm, uplo, n, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric band matrix.

    ``dlansb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric band matrix.

    For full information please refer to the NAG Library document for
    f06re

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ref.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlansb`` returns zero.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : float, shape (`k`+1, `n`)
        The n by n symmetric band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlansb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06reft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _ldab_py = utils._EngineFloat64ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _ab = utils._EngineFloat64ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _k,
        _ab,
        _ldab,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlantr(norm, uplo, diag, m, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real trapezoidal/triangular matrix.

    ``dlantr`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real m by n trapezoidal matrix (triangular if m = n).

    For full information please refer to the NAG Library document for
    f06rj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rjf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower trapezoidal.

        `uplo` = 'U'

          A is upper trapezoidal.

        `uplo` = 'L'

          A is lower trapezoidal.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix A.

        When `m` = 0, ``dlantr`` is set to zero.

    n : int
        n, the number of columns of the matrix A.

        When `n` = 0, ``dlantr`` is set to zero.

    a : float, shape (`m`, `n`)
        The m by n trapezoidal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real m by n
        trapezoidal matrix (triangular if m = n).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlantr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(4,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(5,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _diag,
        _m,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlantp(norm, uplo, diag, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real triangular matrix, packed storage.

    ``dlantp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n triangular matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06rk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rkf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlantp`` returns zero.

    ap : float, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        triangular matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlantp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _ap = utils._EngineFloat64ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _diag,
        _n,
        _ap,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlantb(norm, uplo, diag, n, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real triangular band matrix.

    ``dlantb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n triangular band matrix.

    For full information please refer to the NAG Library document for
    f06rl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rlf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlantb`` returns zero.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : float, shape (`k`+1, `n`)
        The n by n triangular band matrix A

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        triangular band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlantb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _ldab_py = utils._EngineFloat64ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _ab = utils._EngineFloat64ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _diag,
        _n,
        _k,
        _ab,
        _ldab,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlanhs(norm, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real upper Hessenberg matrix.

    ``dlanhs`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06rm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rmf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlanhs`` returns zero.

    a : float, shape (`n`, `n`)
        The n by n upper Hessenberg matrix A; elements of the array
        below the first subdiagonal are not referenced.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        upper Hessenberg matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlanhs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlangt(norm, n, dl, d, du):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real tridiagonal matrix.

    ``dlangt`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06rn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rnf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlangt`` returns zero.

    dl : float, shape (`n`-1)
        The (n-1) subdiagonal elements of A.

    d : float, shape (`n`)
        The n diagonal elements of A.

    du : float, shape (`n`-1)
        The (n-1) superdiagonal elements of A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlangt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, dl[n-1].
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._EngineFloat64ArrayType.c_type, # Input, du[n-1].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _dl = utils._EngineFloat64ArrayType(
        dl, {'fun_name': fun_name, 'entity': 'dl'},
        exp_shape=(_n_py-1,),
    ).to_ctypes()
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _du = utils._EngineFloat64ArrayType(
        du, {'fun_name': fun_name, 'entity': 'du'},
        exp_shape=(_n_py-1,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _dl,
        _d,
        _du,
        _nrm,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dlanst(norm, n, d, e):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric tridiagonal matrix.

    ``dlanst`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06rp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rpf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlanst`` returns zero.

    d : float, shape (`n`)
        The n diagonal elements of the tridiagonal matrix A.

    e : float, shape (`n`-1)
        The (n-1) subdiagonal or superdiagonal elements of the
        tridiagonal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlanst'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06rpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._EngineFloat64ArrayType.c_type, # Input, e[n-1].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _e = utils._EngineFloat64ArrayType(
        e, {'fun_name': fun_name, 'entity': 'e'},
        exp_shape=(_n_py-1,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _d,
        _e,
        _nrm,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zgemv(trans, m, n, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, complex rectangular matrix.

    ``zgemv`` computes the matrix-vector product for a complex general
    matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06saf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T'

          y←alpha A^Tx+beta y.

        `trans` = 'C'

          y←alpha A^Hx+beta y.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    alpha : complex
        The scalar alpha.

    a : complex, shape (`m`, `n`)
        The m by n matrix A.

    x : complex, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`n`-1)*|`incx`|; if `trans` in ('T',
        'C'): 1+(`m`-1)*|`incx`|; otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`n`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
          for i = 1,2,...,`n`.

        If `trans` = 'T' or 'C',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`m`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`m`-i)*`incx`-1],
          for i = 1,2,...,`m`.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`m`-1)*|`incy`|; if `trans` in ('T',
        'C'): 1+(`n`-1)*|`incy`|; otherwise: 0.

        On entry: the vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`m`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`m`-i)*`incy`-1],
          for i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`n`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
          for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 11)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgemv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06saft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, x[*].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[*].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='trans',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    if _trans_py in ('N', 'n'):
        _x_dim1 = 1+(_n_py-1)*abs(_incx_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _x_dim1 = 1+(_m_py-1)*abs(_incx_py)
    else:
        _x_dim1 = 0
    exp_shape_x = (_x_dim1,)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(11,),
            entity='incy',
        ).manage()
    if _trans_py in ('N', 'n'):
        _y_dim1 = 1+(_m_py-1)*abs(_incy_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _y_dim1 = 1+(_n_py-1)*abs(_incy_py)
    else:
        _y_dim1 = 0
    exp_shape_y = (_y_dim1,)
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=exp_shape_y,
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0)
    _routine_name = utils._EngineRoutineName(*b'f06sa ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _trans,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zgbmv(trans, m, n, kl, ku, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, complex rectangular band matrix.

    ``zgbmv`` computes the matrix-vector product for a complex general
    band matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sbf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T'

          y←alpha A^Tx+beta y.

        `trans` = 'C'

          y←alpha A^Hx+beta y.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    alpha : complex
        The scalar alpha.

    a : complex, shape (`kl`+`ku`+1, `n`)
        The m by n band matrix A.

    x : complex, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`n`-1)*|`incx`|; if `trans` in ('T',
        'C'): 1+(`m`-1)*|`incx`|; otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`n`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
          for i = 1,2,...,`n`.

        If `trans` = 'T' or 'C',

        - if `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1],
          for i = 1,2,...,`m`;

        - if `incx` < 0, x_i must be stored in `x`[1-(`m`-i)*`incx`-1],
          for i = 1,2,...,`m`.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `trans`='N': 1+(`m`-1)*|`incy`|; if `trans` in ('T',
        'C'): 1+(`n`-1)*|`incy`|; otherwise: 0.

        On entry: the vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`m`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`m`-i)*`incy`-1],
          for i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1],
          for i = 1,2,...,`n`;

        - if `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
          for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` 10)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 13)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgbmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06sbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, x[*].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[*].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='trans',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    if not _kl_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('kl >= 0.'),
            ],
            errcodes=(4,),
            entity='kl',
        ).manage()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    if not _ku_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('ku >= 0.'),
            ],
            errcodes=(5,),
            entity='ku',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda_py = utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(10,),
            entity='incx',
        ).manage()
    if _trans_py in ('N', 'n'):
        _x_dim1 = 1+(_n_py-1)*abs(_incx_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _x_dim1 = 1+(_m_py-1)*abs(_incx_py)
    else:
        _x_dim1 = 0
    exp_shape_x = (_x_dim1,)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(13,),
            entity='incy',
        ).manage()
    if _trans_py in ('N', 'n'):
        _y_dim1 = 1+(_m_py-1)*abs(_incy_py)
    elif _trans_py in ('T', 't', 'C', 'c'):
        _y_dim1 = 1+(_n_py-1)*abs(_incy_py)
    else:
        _y_dim1 = 0
    exp_shape_y = (_y_dim1,)
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=exp_shape_y,
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0)
    _routine_name = utils._EngineRoutineName(*b'f06sb ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _trans,
        _m,
        _n,
        _kl,
        _ku,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zhemv(uplo, n, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, complex Hermitian matrix.

    ``zhemv`` computes the matrix-vector product for a complex Hermitian
    matrix.

    For full information please refer to the NAG Library document for
    f06sc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06scf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    a : complex, shape (`n`, `n`)
        The n by n Hermitian matrix A.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y, if `beta` = 0.0, `y` need not
        be set.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 7)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 10)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhemv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06scft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(7,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(10,),
            entity='incy',
        ).manage()
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1)
    _routine_name = utils._EngineRoutineName(*b'f06sc ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zhbmv(uplo, n, k, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, complex Hermitian band matrix.

    ``zhbmv`` computes the matrix-vector product for a complex Hermitian
    band matrix.

    For full information please refer to the NAG Library document for
    f06sd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    alpha : complex
        The scalar alpha.

    a : complex, shape (`k`+1, `n`)
        The n by n Hermitian band matrix A.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y, if `beta` = 0.0, `y` need not
        be set.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 11)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhbmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06sdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(3,),
            entity='k',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda_py = utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(11,),
            entity='incy',
        ).manage()
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1)
    _routine_name = utils._EngineRoutineName(*b'f06sd ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zhpmv(uplo, n, alpha, ap, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, complex Hermitian packed matrix.

    ``zhpmv`` computes the matrix-vector product for a complex Hermitian
    matrix stored in packed form.

    For full information please refer to the NAG Library document for
    f06se

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    ap : complex, shape (`n`*(`n`+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y, if `beta` = 0.0, `y` need not
        be set.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 6)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 9)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhpmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06seft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _ap = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(6,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(9,),
            entity='incy',
        ).manage()
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
        check_contents=_beta_py != 0.0 or abs(_incy_py) > 1,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=_beta_py != 0.0 or abs(_incy_py) > 1)
    _routine_name = utils._EngineRoutineName(*b'f06se ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _ap,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def ztrmv(uplo, trans, diag, n, a, x, incx):
    u"""
    Matrix-vector product, complex triangular matrix.

    ``ztrmv`` computes the matrix-vector product for a complex
    triangular matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sff.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T'

          x←A^Tx.

        `trans` = 'C'

          x←A^Hx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    a : complex, shape (`n`, `n`)
        The n by n triangular matrix A.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztrmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06sfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sf ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def ztbmv(uplo, trans, diag, n, k, a, x, incx):
    u"""
    Matrix-vector product, complex triangular band matrix.

    ``ztbmv`` computes the matrix-vector product for a complex
    triangular band matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T'

          x←A^Tx.

        `trans` = 'C'

          x←A^Hx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : complex, shape (`k`+1, `n`)
        The n by n triangular band matrix A

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztbmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06sgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _lda_py = utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(9,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sg ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _k,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def ztpmv(uplo, trans, diag, n, ap, x, incx):
    u"""
    Matrix-vector product, complex triangular packed matrix.

    ``ztpmv`` computes the matrix-vector product for a complex
    triangular matrix, its transpose or its conjugate transpose, stored
    in packed form.

    For full information please refer to the NAG Library document for
    f06sh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06shf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T'

          x←A^Tx.

        `trans` = 'C'

          x←A^Hx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    ap : complex, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 7)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztpmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06shft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _ap = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(7,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sh ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _ap,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def ztrsv(uplo, trans, diag, n, a, x, incx):
    u"""
    System of equations, complex triangular matrix.

    ``ztrsv`` solves a complex triangular system of equations with a
    single right-hand side.

    For full information please refer to the NAG Library document for
    f06sj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T'

          x←A^invTx.

        `trans` = 'C'

          x←A^invHx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    a : complex, shape (`n`, `n`)
        The n by n triangular matrix A.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 8)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztrsv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06sjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(8,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sj ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def ztbsv(uplo, trans, diag, n, k, a, x, incx):
    u"""
    System of equations, complex triangular band matrix.

    ``ztbsv`` solves a complex triangular banded system of equations
    with a single right-hand side.

    For full information please refer to the NAG Library document for
    f06sk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06skf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T'

          x←A^invTx.

        `trans` = 'C'

          x←A^invHx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : complex, shape (`k`+1, `n`)
        The n by n triangular band matrix A

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztbsv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06skft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _lda_py = utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(9,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sk ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _k,
        _a,
        _lda,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def ztpsv(uplo, trans, diag, n, ap, x, incx):
    u"""
    System of equations, complex triangular packed matrix.

    ``ztpsv`` solves a complex triangular system of equations, stored in
    packed form, with a single right-hand side.

    For full information please refer to the NAG Library document for
    f06sl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06slf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T'

          x←A^invTx.

        `trans` = 'C'

          x←A^invHx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

    ap : complex, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : complex, shape (1+(`n`-1)*|`incx`|), modified in place
        On entry: the vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 7)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztpsv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06slft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _ap = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(7,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    )
    _x = _x_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sl ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _diag,
        _n,
        _ap,
        _x,
        _incx,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    return_data = None
    return return_data

def zgeru(m, n, alpha, x, incx, y, incy, a):
    u"""
    Rank-1 update, complex rectangular matrix, unconjugated vector.

    ``zgeru`` computes the rank-1 update of a complex general matrix
    using an unconjugated vector.

    For full information please refer to the NAG Library document for
    f06sm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06smf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`m`-1)*|`incx`|)
        The m element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`m`.

        If `incx` < 0, x_i must be stored in `x`[1-(`m`-i)*`incx`-1],
        for i = 1,2,...,`m`.

        Intermediate elements of X are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    a : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix A.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgeru'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06smft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(m-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(1,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_m_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(7,),
            entity='incy',
        ).manage()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sm ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _m,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _lda,
        _routine_name,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zgerc(m, n, alpha, x, incx, y, incy, a):
    u"""
    Rank-1 update, complex rectangular matrix, conjugated vector.

    ``zgerc`` computes the rank-1 update of a complex general matrix
    using a conjugated vector.

    For full information please refer to the NAG Library document for
    f06sn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06snf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`m`-1)*|`incx`|)
        The m element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`-1], for
        i = 1,2,...,`m`.

        If `incx` < 0, x_i must be stored in `x`[1-(`m`-i)*`incx`-1],
        for i = 1,2,...,`m`.

        Intermediate elements of X are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    a : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix A.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgerc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06snft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(m-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(1,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_m_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(7,),
            entity='incy',
        ).manage()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sn ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _m,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _lda,
        _routine_name,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zher(uplo, n, alpha, x, incx, a):
    u"""
    Rank-1 update, complex Hermitian matrix.

    ``zher`` computes the rank-1 update of a complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06sp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06spf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n Hermitian matrix A.

        On exit: the updated matrix A. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zher'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06spft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sp ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _a,
        _lda,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zhpr(uplo, n, alpha, x, incx, ap):
    u"""
    Rank-1 update, complex Hermitian packed matrix.

    ``zhpr`` computes the rank-1 update of a complex Hermitian matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06sq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sqf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : float
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    ap : complex, shape (`n`*(`n`+1)/2), modified in place
        On entry: the n by n Hermitian matrix A, packed by columns.

        On exit: the updated matrix A. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhpr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06sqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _ap_cls = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sq ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _ap,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _ap_cls.refresh_py(_ap)
    return_data = None
    return return_data

def zher2(uplo, n, alpha, x, incx, y, incy, a):
    u"""
    Rank-2 update, complex Hermitian matrix.

    ``zher2`` computes the rank-2 update of a complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06sr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06srf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n Hermitian matrix A.

        On exit: the updated matrix A. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zher2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06srft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(7,),
            entity='incy',
        ).manage()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06sr ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _lda,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zhpr2(uplo, n, alpha, x, incx, y, incy, ap):
    u"""
    Rank-2 update, complex Hermitian packed matrix.

    ``zhpr2`` computes the rank-2 update of a complex Hermitian matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06ss

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ssf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*|`incy`|)
        The n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`], for i
        = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`], for
        i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    ap : complex, shape (`n`*(`n`+1)/2), modified in place
        On entry: the n by n Hermitian matrix A, packed by columns.

        On exit: the updated matrix A. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 7)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhpr2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ssft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(7,),
            entity='incy',
        ).manage()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    ).to_ctypes()
    _ap_cls = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06ss ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _ap,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _ap_cls.refresh_py(_ap)
    return_data = None
    return return_data

def zsymv(uplo, n, alpha, a, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, complex symmetric matrix.

    ``zsymv`` performs the matrix-vector operation

        y←alpha Ax+beta y,

    where A is an n by n complex symmetric matrix, x and y are n-element
    complex vectors, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06ta

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06taf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    a : complex, shape (`n`, `n`)
        The n by n symmetric matrix A.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 7)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 10)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsymv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06taft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(7,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(10,),
            entity='incy',
        ).manage()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06ta ')
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _alpha,
        _a,
        _lda,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zsyr(uplo, n, alpha, x, incx, a):
    u"""
    Rank-1 update, complex symmetric matrix.

    ``zsyr`` performs the symmetric rank-1 update operation

        A←alpha xx^T+A,

    where A is an n by n complex symmetric matrix, x is an n-element
    complex vector, and alpha is a complex scalar.

    For full information please refer to the NAG Library document for
    f06tb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tbf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix A.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsyr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06tb ')
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _a,
        _lda,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zspmv(uplo, n, alpha, ap, x, incx, beta, y, incy):
    u"""
    Matrix-vector product, complex symmetric packed matrix.

    ``zspmv`` performs the matrix-vector operation

        y←alpha Ax+beta y

    where A is an n by n complex symmetric matrix stored in packed form,
    x and y are n-element complex vectors, and alpha and beta are
    complex scalars.

    For full information please refer to the NAG Library document for
    f06tc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tcf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    ap : complex, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    beta : complex
        The scalar beta.

    y : complex, shape (1+(`n`-1)*|`incy`|), modified in place
        On entry: the n-element vector y.

        If `incy` > 0, y_i must be stored in `y`[1+(i-1)*`incy`-1], for
        i = 1,2,...,`n`.

        If `incy` < 0, y_i must be stored in `y`[1-(`n`-i)*`incy`-1],
        for i = 1,2,...,`n`.

        On exit: the updated vector y stored in the array elements used
        to supply the original vector y.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 6)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

        (`errno` 9)
            On entry, error in parameter `incy`.

            Constraint: `incy` != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zspmv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[1+(n-1)*abs(incy)].
        utils._BYREF_INT, # Input, incy.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _ap = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(6,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy != 0.'),
            ],
            errcodes=(9,),
            entity='incy',
        ).manage()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*abs(_incy_py),),
    )
    _y = _y_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06tc ')
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _alpha,
        _ap,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _y_cls.refresh_py(_y)
    return_data = None
    return return_data

def zspr(uplo, n, alpha, x, incx, ap):
    u"""
    Rank-1 update, complex symmetric packed matrix.

    ``zspr`` performs the symmetric rank-1 update operation

        A←alpha xx^T+A,

    where A is an n by n complex symmetric matrix, stored in packed
    form, x is an n-element complex vector, and alpha is a complex
    scalar.

    For full information please refer to the NAG Library document for
    f06td

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*|`incx`|)
        The n-element vector x.

        If `incx` > 0, x_i must be stored in `x`[1+(i-1)*`incx`], for i
        = 1,2,...,`n`.

        If `incx` < 0, x_i must be stored in `x`[1-(`n`-i)*`incx`], for
        i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    ap : complex, shape (`n`*(`n`+1)/2), modified in place
        On entry: the n by n symmetric matrix A, packed by columns.

        On exit: the updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `incx`.

            Constraint: `incx` != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zspr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, x[1+(n-1)*abs(incx)].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, ap[n*(n+1)//2].
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py != 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx != 0.'),
            ],
            errcodes=(5,),
            entity='incx',
        ).manage()
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*abs(_incx_py),),
    ).to_ctypes()
    _ap_cls = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _ap = _ap_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06td ')
    engine_ctypes(
        _en_data,
        _uplo,
        _n,
        _alpha,
        _x,
        _incx,
        _ap,
        _routine_name,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _ap_cls.refresh_py(_ap)
    return_data = None
    return return_data

def zmcopy(matrix, m, n, a, b):
    u"""
    Matrix copy, complex rectangular or trapezoidal matrix.

    ``zmcopy`` performs the matrix-copy operation

        B←A

    where A and B are m by n complex general or trapezoidal matrices.

    For full information please refer to the NAG Library document for
    f06tf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tff.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrices A and B.

    n : int
        n, the number of columns of the matrices A and B.

    a : complex, shape (`m`, `n`)
        The m by n general or trapezoidal matrix A.

    b : complex, shape (`m`, `n`), modified in place
        On exit: the m by n general or trapezoidal matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zmcopy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, matrix(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of matrix.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _matrix = utils._EngineCharScalarType(
        matrix, {'fun_name': fun_name, 'entity': 'matrix'}, exp_chrlen=1,
    ).to_ctypes()
    _matrix_py = utils._EngineCharScalarType.to_py(_matrix)
    if _matrix_py not in ('G', 'g', 'U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('matrix in (\'G\', \'g\', \'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='matrix',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _matrix,
        _m,
        _n,
        _a,
        _lda,
        _b,
        _ldb,
        _matrix._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def zmload(matrix, m, n, con, diag, a):
    u"""
    Matrix initialization, complex rectangular matrix.

    ``zmload`` forms the complex m by n rectangular or trapezoidal
    matrix A given by

        a_{ij} = {[table omitted].

    For full information please refer to the NAG Library document for
    f06th

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06thf.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    con : complex
        The value to be assigned to the off-diagonal elements of A.

    diag : complex
        The value to be assigned to the diagonal elements of A.

    a : complex, shape (`m`, `n`), modified in place
        On exit: the m by n general or trapezoidal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zmload'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06thft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, matrix(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, con.
        utils._BYREF_COMPLEX128, # Input, diag.
        utils._EngineComplex128ArrayType.c_type, # Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of matrix.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _matrix = utils._EngineCharScalarType(
        matrix, {'fun_name': fun_name, 'entity': 'matrix'}, exp_chrlen=1,
    ).to_ctypes()
    _matrix_py = utils._EngineCharScalarType.to_py(_matrix)
    if _matrix_py not in ('G', 'g', 'U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('matrix in (\'G\', \'g\', \'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='matrix',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _con = utils._EngineComplex128ScalarType(
        con, {'fun_name': fun_name, 'entity': 'con'},
    ).to_ctypes()
    _diag = utils._EngineComplex128ScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _matrix,
        _m,
        _n,
        _con,
        _diag,
        _a,
        _lda,
        _matrix._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zhesrc(uplo, pivot, direct, n, k1, k2, c, s, a):
    u"""
    Unitary similarity transformation of Hermitian matrix as a sequence
    of plane rotations.

    ``zhesrc`` performs a Unitary similarity transformation (as a
    sequence of plane rotations) of a complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06tm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tmf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    n : int
        n, the order of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n Hermitian matrix A.

        On exit: the transformed matrix A. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhesrc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'T', 't', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'T\', \'t\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(3,),
            entity='direct',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _uplo,
        _pivot,
        _direct,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _uplo._length_,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zutr1(n, alpha, x, incx, y, incy, a, c, s):
    u"""
    QR factorization by sequence of plane rotations, rank-1 update of
    complex upper triangular matrix.

    ``zutr1`` performs a QR factorization (as a sequence of plane
    rotations) of a complex upper triangular matrix that has been
    modified by a rank-1 update.

    For full information please refer to the NAG Library document for
    f06tp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tpf.html

    Parameters
    ----------
    n : int
        n, the order of the matrices U and R.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the sequence of plane rotations.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    y : complex, shape (1+(`n`-1)*`incy`)
        The n-element vector y. y_i must be stored in
        `y`[1+(i-1)*`incy`], for i = 1,2,...,`n`.

        Intermediate elements of `y` are not referenced.

    incy : int
        The increment in the subscripts of `y` between successive
        elements of y.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U. The imaginary
        parts of the diagonal elements must be zero.

        On exit: the upper triangular matrix R. The imaginary parts of
        the diagonal elements must be zero.

    c : float, shape (`n`-1), modified in place
        On exit: the cosines of the rotations Q_k, for k = 1,2,...,n-1.

    s : complex, shape (`n`), modified in place
        On exit: the sines of the rotations Q_k, for k = 1,2,...,n-1;
        `s`[n] holds d_n, the nth diagonal element of D.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

        (`errno` 6)
            On entry, error in parameter `incy`.

            Constraint: `incy` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zutr1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input, y[1+(n-1)*incy].
        utils._BYREF_INT, # Input, incy.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Output, c[n-1].
        utils._EngineComplex128ArrayType.c_type, # Output, s[n].
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(1,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _incy = utils._EngineIntScalarType(
        incy, {'fun_name': fun_name, 'entity': 'incy'},
    ).to_ctypes()
    _incy_py = utils._EngineIntScalarType.to_py(_incy)
    if not _incy_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incy > 0.'),
            ],
            errcodes=(6,),
            entity='incy',
        ).manage()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(1+(_n_py-1)*_incy_py,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py-1,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _lda,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _a_cls.refresh_py(_a)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    return_data = None
    return return_data

def zutupd(n, alpha, x, incx, a, c, s):
    u"""
    QR factorization by sequence of plane rotations, complex upper
    triangular matrix augmented by a full row.

    ``zutupd`` performs a QR factorization (as a sequence of plane
    rotations) of a complex upper triangular matrix that has been
    augmented by a full row.

    For full information please refer to the NAG Library document for
    f06tq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tqf.html

    Parameters
    ----------
    n : int
        n, the order of the matrices U and R.

    alpha : complex
        The scalar alpha.

    x : complex, shape (1+(`n`-1)*`incx`), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*`incx`], for i = 1,2,...,`n`.

        Intermediate elements of `x` are not referenced.

        On exit: the referenced elements are overwritten by details of
        the sequence of plane rotations.

    incx : int
        The increment in the subscripts of `x` between successive
        elements of x.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U.

        On exit: the upper triangular matrix R.

    c : float, shape (`n`), modified in place
        On exit: the values c_k, the cosines of the rotations Q_k, for k
        = 1,2,...,n.

    s : complex, shape (`n`), modified in place
        On exit: the values s_k, the sines of the rotations Q_k, for k =
        1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `incx`.

            Constraint: `incx` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zutupd'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[1+(n-1)*incx].
        utils._BYREF_INT, # Input, incx.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Output, c[n].
        utils._EngineComplex128ArrayType.c_type, # Output, s[n].
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(1,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    if not _incx_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('incx > 0.'),
            ],
            errcodes=(4,),
            entity='incx',
        ).manage()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(1+(_n_py-1)*_incx_py,),
    )
    _x = _x_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _s = _s_cls.to_ctypes(can_reference_contents=False)
    engine_ctypes(
        _en_data,
        _n,
        _alpha,
        _x,
        _incx,
        _a,
        _lda,
        _c,
        _s,
    )
    utils._handle_cb_exception(excs)
    _x_cls.refresh_py(_x)
    _a_cls.refresh_py(_a)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    return_data = None
    return return_data

def zuhqr(side, n, k1, k2, c, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, complex upper
    Hessenberg matrix.

    ``zuhqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a complex upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06tr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06trf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    n : int
        n, the order of the matrix H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : complex, shape (`k2`), modified in place
        On exit: `c`[k] holds c_k, the cosine of the rotation P_k, for k
        = k_1,...,k_2-1; `c`[k_2] holds d_k_2, the k_2th diagonal
        element of D, if `side` = 'L', or d_k_1, the k_1th diagonal
        element of D, if `side` = 'R'.

    s : float, shape (`k2`-1), modified in place
        On entry: the nonzero subdiagonal elements of H: `s`[k] must
        hold h_{k+1,k}, for k = k_1,...,k_2-1.

        On exit: `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the upper triangular part of the n by n upper
        Hessenberg matrix H.

        On exit: the upper triangular matrix R. The imaginary parts of
        the diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zuhqr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06trft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineComplex128ArrayType.c_type, # Output, c[k2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zusqr(side, n, k1, k2, c, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, complex upper
    spiked matrix.

    ``zusqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a complex upper spiked matrix.

    For full information please refer to the NAG Library document for
    f06ts

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tsf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    n : int
        n, the order of the matrix H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1), modified in place
        On exit: `c`[k] holds c_k, the cosine of the rotation P_k, for k
        = k_1,...,k_2-1.

    s : complex, shape (`k2`), modified in place
        On entry: the nonzero elements of the spike of H: `s`[k] must
        hold h_{k_2,k} if `side` = 'L', and h_{k+1,k_1} if `side` = 'R',
        for k = k_1,...,k_2-1.

        On exit: `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1; `s`[k_2] holds d_k_2, the k_2th diagonal element
        of D, if `side` = 'L', or d_k_1, the k_1th diagonal element of
        D, if `side` = 'R'.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the upper triangular part of the n by n upper spiked
        matrix H. The imaginary parts of the diagonal elements must be
        zero, except for the (k_2,k_2) element if `side` = 'L', or the
        (k_1,k_1) element if `side` = 'R'.

        On exit: the upper triangular matrix R. The imaginary parts of
        the diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zusqr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Output, c[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, s[k2].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _s_cls = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zutsqr(side, n, k1, k2, c, s, a):
    u"""
    QR factorization of UP or RQ factorization of PU, U complex upper
    triangular, P a sequence of plane rotations.

    ``zutsqr`` performs a QR or RQ factorization of the product of a
    complex upper triangular matrix and a complex matrix of plane
    rotations.

    For full information please refer to the NAG Library document for
    f06tt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ttf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether P is applied from the left or the right in the
        transformation.

        `side` = 'L'

          P is applied from the left.

        `side` = 'R'

          P is applied from the right.

    n : int
        n, the order of the matrices U and R.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1), modified in place
        On entry: `c`[k] must hold the cosine of the rotation P_k, for k
        = k_1,...,k_2-1.

        On exit: `c`[k] holds the cosine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    s : complex, shape (`k2`-1), modified in place
        On entry: `s`[k] must hold the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

        On exit: `s`[k] holds the sine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U.

        On exit: the upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zutsqr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ttft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    )
    _c = _c_cls.to_ctypes()
    _s_cls = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zutsrh(side, n, k1, k2, c, s, a):
    u"""
    Compute upper Hessenberg matrix by sequence of plane rotations,
    complex upper triangular matrix.

    ``zutsrh`` transforms a complex upper triangular matrix to an upper
    Hessenberg matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06tv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tvf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    n : int
        n, the order of the matrices U and H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : complex, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1), modified in place
        On entry: `s`[k] must hold s_k, the sine of the rotation P_k,
        for k = k_1,...,k_2-1.

        On exit: `s`[k] holds h_{k+1,k}, the subdiagonal element of H,
        for k = k_1,...,k_2-1.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U. The imaginary
        parts of the diagonal elements must be zero.

        On exit: the upper triangular part of the upper Hessenberg
        matrix H.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zutsrh'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tvft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineComplex128ArrayType.c_type, # Input, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zutsrs(side, n, k1, k2, c, s, a):
    u"""
    Compute upper spiked matrix by sequence of plane rotations, complex
    upper triangular matrix.

    ``zutsrs`` transforms a complex upper triangular matrix to an upper
    spiked matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06tw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06twf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    n : int
        n, the order of the matrices U and H.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, shape (`k2`-1), modified in place
        On entry: `s`[k] must hold s_k, the sine of the rotation P_k,
        for k = k_1,...,k_2-1.

        On exit: `s`[k] holds a nonzero element of the spike of H:
        h_{k_2,k} if `side` = 'L', or h_{k+1,k_1} if `side` = 'R', for k
        = k_1,...,k_2-1.

    a : complex, shape (`n`, `n`), modified in place
        On entry: the n by n upper triangular matrix U. The imaginary
        parts of the diagonal elements must be zero.

        On exit: the upper triangular part of the upper spiked matrix H.
        The imaginary parts of the diagonal elements are set to zero
        except for the (k_2,k_2) element if `side` = 'L', or the
        (k_1,k_1) element if `side` = 'R'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zutsrs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06twft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s_cls = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    )
    _s = _s_cls.to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
    )
    utils._handle_cb_exception(excs)
    _s_cls.refresh_py(_s)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zgesrc(side, pivot, direct, m, n, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, complex rectangular matrix, real
    cosine and complex sine.

    ``zgesrc`` applies to a complex rectangular matrix a sequence of
    plane rotations having real cosines and complex sines.

    For full information please refer to the NAG Library document for
    f06tx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06txf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix A.

        On exit: the transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgesrc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06txft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'T', 't', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'T\', \'t\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(3,),
            entity='direct',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(4,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(5,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s = utils._EngineComplex128ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _pivot,
        _direct,
        _m,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zgesrs(side, pivot, direct, m, n, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, complex rectangular matrix,
    complex cosine and real sine.

    ``zgesrs`` applies to a complex rectangular matrix a sequence of
    plane rotations having real sines and complex cosines.

    For full information please refer to the NAG Library document for
    f06ty

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tyf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : complex, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix A.

        On exit: the transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgesrs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06tyft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineComplex128ArrayType.c_type, # Input, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'T', 't', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'T\', \'t\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(3,),
            entity='direct',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(4,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(5,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _pivot,
        _direct,
        _m,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def zlange(norm, m, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex general matrix.

    ``zlange`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex m by n matrix.

    For full information please refer to the NAG Library document for
    f06ua

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uaf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    m : int
        m, the number of rows of the matrix A.

        When `m` = 0, ``zlange`` is set to zero.

    n : int
        n, the number of columns of the matrix A.

        When `n` = 0, ``zlange`` is set to zero.

    a : complex, shape (`m`, `n`)
        The m by n matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex m
        by n matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlange'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06uaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(2,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _m,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlangb(norm, n, kl, ku, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex band matrix.

    ``zlangb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n band matrix.

    For full information please refer to the NAG Library document for
    f06ub

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ubf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlangb`` returns zero.

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    ab : complex, shape (`kl`+`ku`+1, `n`)
        The n by n band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlangb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ubft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kl.
        utils._BYREF_INT, # Input, ku.
        utils._EngineComplex128ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _kl = utils._EngineIntScalarType(
        kl, {'fun_name': fun_name, 'entity': 'kl'},
    ).to_ctypes()
    _kl_py = utils._EngineIntScalarType.to_py(_kl)
    if not _kl_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('kl >= 0.'),
            ],
            errcodes=(3,),
            entity='kl',
        ).manage()
    _ku = utils._EngineIntScalarType(
        ku, {'fun_name': fun_name, 'entity': 'ku'},
    ).to_ctypes()
    _ku_py = utils._EngineIntScalarType.to_py(_ku)
    if not _ku_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('ku >= 0.'),
            ],
            errcodes=(4,),
            entity='ku',
        ).manage()
    _ldab_py = utils._EngineComplex128ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _ab = utils._EngineComplex128ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_kl_py+_ku_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _kl,
        _ku,
        _ab,
        _ldab,
        _nrm,
        _work,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlanhe(norm, uplo, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian matrix.

    ``zlanhe`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06uc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ucf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a Hermitian matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanhe`` returns zero.

    a : complex, shape (`n`, `n`)
        The n by n Hermitian matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlanhe'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ucft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlanhp(norm, uplo, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian matrix, packed storage.

    ``zlanhp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06ud

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06udf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a Hermitian matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanhp`` returns zero.

    ap : complex, shape (`n`*(`n`+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlanhp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06udft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _ap = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _ap,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlanhb(norm, uplo, n, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian band matrix.

    ``zlanhb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian band matrix.

    For full information please refer to the NAG Library document for
    f06ue

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uef.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a Hermitian matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanhb`` returns zero.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : complex, shape (`k`+1, `n`)
        The n by n Hermitian band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlanhb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ueft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineComplex128ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _ldab_py = utils._EngineComplex128ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _ab = utils._EngineComplex128ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _k,
        _ab,
        _ldab,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlansy(norm, uplo, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex symmetric matrix.

    ``zlansy`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n symmetric matrix.

    For full information please refer to the NAG Library document for
    f06uf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uff.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlansy`` returns zero.

    a : complex, shape (`n`, `n`)
        The n by n symmetric matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n symmetric matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlansy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ufft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlansp(norm, uplo, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex symmetric matrix, packed storage.

    ``zlansp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n symmetric matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06ug

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ugf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlansp`` returns zero.

    ap : complex, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n symmetric matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlansp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ugft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _ap = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _ap,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlansb(norm, uplo, n, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex symmetric band matrix.

    ``zlansb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n symmetric band matrix.

    For full information please refer to the NAG Library document for
    f06uh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uhf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlansb`` returns zero.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : complex, shape (`k`+1, `n`)
        The n by n symmetric band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n symmetric band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlansb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06uhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineComplex128ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _ldab_py = utils._EngineComplex128ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _ab = utils._EngineComplex128ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _n,
        _k,
        _ab,
        _ldab,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlantr(norm, uplo, diag, m, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex trapezoidal/triangular matrix.

    ``zlantr`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex m by n trapezoidal matrix (triangular if m = n).

    For full information please refer to the NAG Library document for
    f06uj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ujf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower trapezoidal.

        `uplo` = 'U'

          A is upper trapezoidal.

        `uplo` = 'L'

          A is lower trapezoidal.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix A.

        When `m` = 0, ``zlantr`` is set to zero.

    n : int
        n, the number of columns of the matrix A.

        When `n` = 0, ``zlantr`` is set to zero.

    a : complex, shape (`m`, `n`)
        The m by n trapezoidal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex m
        by n trapezoidal matrix (triangular if m = n).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlantr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ujft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(4,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(5,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _diag,
        _m,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlantp(norm, uplo, diag, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex triangular matrix, packed storage.

    ``zlantp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n triangular matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06uk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ukf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlantp`` returns zero.

    ap : complex, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n triangular matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlantp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ukft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, ap[n*(n+1)//2].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _ap = utils._EngineComplex128ArrayType(
        ap, {'fun_name': fun_name, 'entity': 'ap'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _diag,
        _n,
        _ap,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlantb(norm, uplo, diag, n, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex triangular band matrix.

    ``zlantb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n triangular band matrix.

    For full information please refer to the NAG Library document for
    f06ul

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ulf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlantb`` returns zero.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : complex, shape (`k`+1, `n`)
        The n by n triangular band matrix A

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n triangular band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlantb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ulft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineComplex128ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(3,),
            entity='diag',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _ldab_py = utils._EngineComplex128ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _ab = utils._EngineComplex128ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_k_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _uplo,
        _diag,
        _n,
        _k,
        _ab,
        _ldab,
        _nrm,
        _work,
        _norm._length_,
        _uplo._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlanhs(norm, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hessenberg matrix.

    ``zlanhs`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06um

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06umf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanhs`` returns zero.

    a : complex, shape (`n`, `n`)
        The n by n upper Hessenberg matrix A; elements of the array
        below the first subdiagonal are not referenced.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n upper Hessenberg matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlanhs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06umft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _a,
        _lda,
        _nrm,
        _work,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlangt(norm, n, dl, d, du):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex tridiagonal matrix.

    ``zlangt`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06un

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06unf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlangt`` returns zero.

    dl : complex, shape (`n`-1)
        The (n-1) subdiagonal elements of A.

    d : complex, shape (`n`)
        The n diagonal elements of A.

    du : complex, shape (`n`-1)
        The (n-1) superdiagonal elements of A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlangt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06unft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, dl[n-1].
        utils._EngineComplex128ArrayType.c_type, # Input, d[n].
        utils._EngineComplex128ArrayType.c_type, # Input, du[n-1].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _dl = utils._EngineComplex128ArrayType(
        dl, {'fun_name': fun_name, 'entity': 'dl'},
        exp_shape=(_n_py-1,),
    ).to_ctypes()
    _d = utils._EngineComplex128ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _du = utils._EngineComplex128ArrayType(
        du, {'fun_name': fun_name, 'entity': 'du'},
        exp_shape=(_n_py-1,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _dl,
        _d,
        _du,
        _nrm,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zlanht(norm, n, d, e):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian tridiagonal matrix.

    ``zlanht`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06up

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06upf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanht`` returns zero.

    d : float, shape (`n`)
        The n diagonal elements of the tridiagonal matrix A.

    e : complex, shape (`n`-1)
        The (n-1) subdiagonal or superdiagonal elements of the
        tridiagonal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlanht'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06upft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._EngineComplex128ArrayType.c_type, # Input, e[n-1].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(2,),
            entity='n',
        ).manage()
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _e = utils._EngineComplex128ArrayType(
        e, {'fun_name': fun_name, 'entity': 'e'},
        exp_shape=(_n_py-1,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _d,
        _e,
        _nrm,
        _norm._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def zgeap(side, trans, n, perm, k, b):
    u"""
    Permute rows or columns, complex rectangular matrix, permutations
    represented by an integer array.

    ``zgeap`` permutes the rows or columns of a complex rectangular
    matrix using an int array of permutations.

    For full information please refer to the NAG Library document for
    f06vj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06vjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    n : int
        n, the number of interchanges in the representation of P.

    perm : int, shape (`n`)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : complex, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(`n`, m); if
        `side`='R': `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R':
        max(`n`, max(`perm`)); otherwise: 0.

        On entry: the original matrix B; B is m by k if `side` = 'L', or
        k by m if `side` = 'R'.

        On exit: the permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgeap'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06vjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, perm[n].
        utils._BYREF_INT, # Input, k.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _perm_cls = utils._EngineIntArrayType(
        perm, {'fun_name': fun_name, 'entity': 'perm'},
        exp_shape=(_n_py,),
    )
    _perm = _perm_cls.to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _vl_m_py = _perm_cls.get_max(0, max(1, _n_py)-1)
    if _side_py in ('L', 'l'):
        _b_dim1 = max(_n_py, _vl_m_py)
    elif _side_py in ('R', 'r'):
        _b_dim1 = _k_py
    else:
        _b_dim1 = 0
    if _side_py in ('L', 'l'):
        _b_dim2 = _k_py
    elif _side_py in ('R', 'r'):
        _b_dim2 = max(_n_py, _perm_cls.get_max(0, _n_py-1))
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _trans,
        _n,
        _perm,
        _k,
        _b,
        _ldb,
        _side._length_,
        _trans._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def zgeapr(side, trans, n, perm, k, b):
    u"""
    Permute rows or columns, complex rectangular matrix, permutations
    represented by a real array.

    ``zgeapr`` permutes the rows or columns of a complex rectangular
    matrix using a double array of permutations.

    For full information please refer to the NAG Library document for
    f06vk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06vkf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    n : int
        n, the number of interchanges in the representation of P.

    perm : float, shape (`n`)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : complex, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(`n`, m); if
        `side`='R': `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R':
        max(`n`, int(max(`perm`))); otherwise: 0.

        On entry: the original matrix B; B is m by k if `side` = 'L', or
        k by m if `side` = 'R'.

        On exit: the permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgeapr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06vkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, perm[n].
        utils._BYREF_INT, # Input, k.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _perm_cls = utils._EngineFloat64ArrayType(
        perm, {'fun_name': fun_name, 'entity': 'perm'},
        exp_shape=(_n_py,),
    )
    _perm = _perm_cls.to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _vl_m_py = int(_perm_cls.get_max(0, max(1, _n_py)-1))
    if _side_py in ('L', 'l'):
        _b_dim1 = max(_n_py, _vl_m_py)
    elif _side_py in ('R', 'r'):
        _b_dim1 = _k_py
    else:
        _b_dim1 = 0
    if _side_py in ('L', 'l'):
        _b_dim2 = _k_py
    elif _side_py in ('R', 'r'):
        _b_dim2 = max(_n_py, int(_perm_cls.get_max(0, _n_py-1)))
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _trans,
        _n,
        _perm,
        _k,
        _b,
        _ldb,
        _side._length_,
        _trans._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def zsgesr(side, pivot, direct, m, n, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, complex rectangular matrix, real
    cosine and sine.

    ``zsgesr`` applies to a complex rectangular matrix a sequence of
    plane rotations having real cosines and complex sines.

    For full information please refer to the NAG Library document for
    f06vx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06vxf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix A.

        On exit: the transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsgesr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06vxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, pivot(len=1).
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k2.
        utils._EngineFloat64ArrayType.c_type, # Input, c[k2-1].
        utils._EngineFloat64ArrayType.c_type, # Input, s[k2-1].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pivot.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _pivot = utils._EngineCharScalarType(
        pivot, {'fun_name': fun_name, 'entity': 'pivot'}, exp_chrlen=1,
    ).to_ctypes()
    _pivot_py = utils._EngineCharScalarType.to_py(_pivot)
    if _pivot_py not in ('V', 'v', 'T', 't', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('pivot in (\'V\', \'v\', \'T\', \'t\', \'B\', \'b\').'),
            ],
            errcodes=(2,),
            entity='pivot',
        ).manage()
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _direct_py = utils._EngineCharScalarType.to_py(_direct)
    if _direct_py not in ('F', 'f', 'B', 'b'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('direct in (\'F\', \'f\', \'B\', \'b\').'),
            ],
            errcodes=(3,),
            entity='direct',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(4,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(5,),
            entity='n',
        ).manage()
    _k1 = utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k2 = utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = utils._EngineIntScalarType.to_py(_k2)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _s = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_k2_py-1,),
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _pivot,
        _direct,
        _m,
        _n,
        _k1,
        _k2,
        _c,
        _s,
        _a,
        _lda,
        _side._length_,
        _pivot._length_,
        _direct._length_,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    return_data = None
    return return_data

def dlansf(norm, transr, uplo, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric matrix, Rectangular Full Packed format.

    ``dlansf`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real symmetric matrix A stored in Rectangular Full Packed (RFP)
    format.

    For full information please refer to the NAG Library document for
    f06wa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06waf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'T'

          The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlansf`` returns zero.

    a : float, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n symmetric matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    nrm : float
        The infinity-norm, the Frobenius norm, or the maximum absolute
        value of the elements of the real symmetric matrix A stored in
        Rectangular Full Packed (RFP) format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dlansf'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06waft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, transr(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[n*(n+1)//2].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _transr = utils._EngineCharScalarType(
        transr, {'fun_name': fun_name, 'entity': 'transr'}, exp_chrlen=1,
    ).to_ctypes()
    _transr_py = utils._EngineCharScalarType.to_py(_transr)
    if _transr_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transr in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(2,),
            entity='transr',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(3,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _call_vendor,
        _en_data,
        _norm,
        _transr,
        _uplo,
        _n,
        _a,
        _nrm,
        _work,
        _norm._length_,
        _transr._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def dtfsm(transr, side, uplo, trans, diag, m, n, alpha, a, b):
    u"""
    Solves a system of equations with multiple right-hand sides, real
    triangular coefficient matrix, Rectangular Full Packed format.

    ``dtfsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a real triangular matrix stored in Rectangular Full
    Packed (RFP) format, B is an m by n real matrix, and alpha is a real
    scalar. A^inv denotes (A^T)^-1 or equivalently (A^-1)^T.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06wb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wbf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'T'

          The matrix A is stored in transposed RFP format.

    side : str, length 1
        Specifies whether B is operated on from the left or the right,
        or similarly whether A (or its transpose) appears to the left or
        right of the solution matrix in the linear system to be solved.

        `side` = 'L'

          B is pre-multiplied from the left. The system to be solved has
          the form AX = alpha B or A^TX = alpha B.

        `side` = 'R'

          B is post-multiplied from the right. The system to be solved
          has the form XA = alpha B or XA^T = alpha B.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies whether the operation involves A^-1 or A^invT, i.e.,
        whether or not A is transposed in the linear system to be
        solved.

        `trans` = 'N'

          The operation involves A^-1, i.e., A is not transposed.

        `trans` = 'T'

          The operation involves A^invT, i.e., A is transposed.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements of A are stored explicitly.

        `diag` = 'U'

          The diagonal elements of A are assumed to be 1, the
          corresponding elements of `a` are not referenced.

    m : int
        m, the number of rows of the matrix B.

    n : int
        n, the number of columns of the matrix B.

    alpha : float
        The scalar alpha.

    a : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `side`='L': `m`*(`m`+1)/2; if `side`='R':
        `n`*(`n`+1)/2; otherwise: 0.

        A, the m by m triangular matrix A if `side` = 'L' or the n by n
        triangular matrix A if `side` = 'R', stored in RFP format (as
        specified by `transr`). The storage format is described in
        detail in [the F07 Introduction]. If `alpha` = 0.0, `a` is not
        referenced.

    b : float, shape (`m`, `n`), modified in place
        On entry: the m by n matrix B.

        If `alpha` = 0, `b` need not be set.

        On exit: the updated matrix B, or similarly the solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` 2)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 5)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 6)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 7)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtfsm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06wbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, transr(len=1).
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _transr = utils._EngineCharScalarType(
        transr, {'fun_name': fun_name, 'entity': 'transr'}, exp_chrlen=1,
    ).to_ctypes()
    _transr_py = utils._EngineCharScalarType.to_py(_transr)
    if _transr_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transr in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(1,),
            entity='transr',
        ).manage()
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(2,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(3,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(4,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(5,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(6,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(7,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py*(_m_py+1)//2
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py*(_n_py+1)//2
    else:
        _a_dim1 = 0
    exp_shape_a = (_a_dim1,)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _transr,
        _side,
        _uplo,
        _trans,
        _diag,
        _m,
        _n,
        _alpha,
        _a,
        _b,
        _ldb,
        _transr._length_,
        _side._length_,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def dsfrk(transr, uplo, trans, n, k, alpha, a, beta, c):
    u"""
    Rank-k update of a real symmetric matrix, Rectangular Full Packed
    format.

    ``dsfrk`` performs one of the symmetric rank-k update operations

        C←alpha AA^T+beta C or C←alpha A^TA+beta C,

    where A is a real matrix, C is an n by n real symmetric matrix
    stored in Rectangular Full Packed (RFP) format, and alpha and beta
    are real scalars.

    For full information please refer to the NAG Library document for
    f06wc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wcf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of C is normal or
        transposed.

        `transr` = 'N'

          The matrix C is stored in normal RFP format.

        `transr` = 'T'

          The matrix C is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored in RFP format.

        `uplo` = 'U'

          The upper triangular part of C is stored in RFP format.

        `uplo` = 'L'

          The lower triangular part of C is stored in RFP format.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^T+beta C.

        `trans` = 'T'

          C←alpha A^TA+beta C.

    n : int
        n, the order of the matrix C.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'T'.

    alpha : float
        The scalar alpha.

    a : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': `n`;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'. If `alpha` = 0.0, `a` is not referenced.

    beta : float
        The scalar beta.

    c : float, shape (`n`*(`n`+1)/2), modified in place
        On entry: the upper or lower triangular part (as specified by
        `uplo`) of the n by n symmetric matrix C, stored in RFP format
        (as specified by `transr`). The storage format is described in
        detail in [the F07 Introduction].

        On exit: the updated matrix C, that is its upper or lower
        triangular part stored in RFP format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsfrk'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06wcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, transr(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[n*(n+1)//2].
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _transr = utils._EngineCharScalarType(
        transr, {'fun_name': fun_name, 'entity': 'transr'}, exp_chrlen=1,
    ).to_ctypes()
    _transr_py = utils._EngineCharScalarType.to_py(_transr)
    if _transr_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transr in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(1,),
            entity='transr',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(3,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = _n_py
    elif _trans_py in ('T', 't'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _trans_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _trans_py in ('T', 't'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _c = _c_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _transr,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _beta,
        _c,
        _transr._length_,
        _uplo._length_,
        _trans._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def zlanhf(norm, transr, uplo, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian matrix, Rectangular Full Packed format.

    ``zlanhf`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex Hermitian matrix A stored in Rectangular Full Packed (RFP)
    format.

    For full information please refer to the NAG Library document for
    f06wn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wnf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'C'

          The conjugate transpose of the RFP representation of the
          matrix A is stored.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanhf`` returns zero.

    a : complex, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n Hermitian matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    nrm : float
        The infinity-norm, the Frobenius norm, or the maximum absolute
        value of the elements of the complex Hermitian matrix A stored
        in Rectangular Full Packed (RFP) format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zlanhf'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06wnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, transr(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[n*(n+1)//2].
        utils._BYREF_FLOAT64, # Output, nrm.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
    )

    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _norm_py = utils._EngineCharScalarType.to_py(_norm)
    if _norm_py not in ('1', '1', 'O', 'o', 'I', 'i', 'F', 'f', 'E', 'e', 'M', 'm'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('norm in (\'1\', \'1\', \'O\', \'o\', \'I\', \'i\', \'F\', \'f\', \'E\', \'e\', \'M\', \'m\').'),
            ],
            errcodes=(1,),
            entity='norm',
        ).manage()
    _transr = utils._EngineCharScalarType(
        transr, {'fun_name': fun_name, 'entity': 'transr'}, exp_chrlen=1,
    ).to_ctypes()
    _transr_py = utils._EngineCharScalarType.to_py(_transr)
    if _transr_py not in ('N', 'n', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transr in (\'N\', \'n\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='transr',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(3,),
            entity='uplo',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    ).to_ctypes()
    _nrm = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    engine_ctypes(
        _call_vendor,
        _en_data,
        _norm,
        _transr,
        _uplo,
        _n,
        _a,
        _nrm,
        _work,
        _norm._length_,
        _transr._length_,
        _uplo._length_,
    )
    utils._handle_cb_exception(excs)
    _nrm_py = utils._EngineFloat64ScalarType.to_py(_nrm)
    return_data = _nrm_py
    return return_data

def ztfsm(transr, side, uplo, trans, diag, m, n, alpha, a, b):
    u"""
    Solves system of equations with multiple right-hand sides, complex
    triangular coefficient matrix, Rectangular Full Packed format.

    ``ztfsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a complex triangular matrix stored in Rectangular Full
    Packed (RFP) format, B is an m by n complex matrix, and alpha is a
    complex scalar. A^invH denotes (A^H)^-1 or equivalently (A^-1)^H.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06wp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wpf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'C'

          The conjugate transpose of the RFP representation of the
          matrix A is stored.

    side : str, length 1
        Specifies whether B is operated on from the left or the right,
        or similarly whether A (or its transpose) appears to the left or
        right of the solution matrix in the linear system to be solved.

        `side` = 'L'

          B is pre-multiplied from the left. The system to be solved has
          the form AX = alpha B or A^HX = alpha B.

        `side` = 'R'

          B is post-multiplied from the right. The system to be solved
          has the form XA = alpha B or XA^H = alpha B.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies whether the operation involves A^-1 or A^invH, i.e.,
        whether or not A is transpose conjugated in the linear system to
        be solved.

        `trans` = 'N'

          The operation involves A^-1, i.e., A is not transpose
          conjugated.

        `trans` = 'C'

          The operation involves A^invH, i.e., A is transpose
          conjugated.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements of A are stored explicitly.

        `diag` = 'U'

          The diagonal elements of A are assumed to be 1, the
          corresponding elements of `a` are not referenced.

    m : int
        m, the number of rows of the matrix B.

    n : int
        n, the number of columns of the matrix B.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:)
        Note: the required length for this argument is determined as
        follows: if `side`='L': `m`*(`m`+1)/2; if `side`='R':
        `n`*(`n`+1)/2; otherwise: 0.

        A, the m by m triangular matrix A if `side` = 'L' or the n by n
        triangular matrix A if `side` = 'R', stored in RFP format (as
        specified by `transr`). The storage format is described in
        detail in [the F07 Introduction]. If `alpha` = 0.0, `a` is not
        referenced.

    b : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix B.

        If `alpha` = 0, `b` need not be set.

        On exit: the updated matrix B, or similarly the solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` 2)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 5)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 6)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 7)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztfsm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06wpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, transr(len=1).
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _transr = utils._EngineCharScalarType(
        transr, {'fun_name': fun_name, 'entity': 'transr'}, exp_chrlen=1,
    ).to_ctypes()
    _transr_py = utils._EngineCharScalarType.to_py(_transr)
    if _transr_py not in ('N', 'n', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transr in (\'N\', \'n\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='transr',
        ).manage()
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(2,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(3,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'C\', \'c\').'),
            ],
            errcodes=(4,),
            entity='trans',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(5,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(6,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(7,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py*(_m_py+1)//2
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py*(_n_py+1)//2
    else:
        _a_dim1 = 0
    exp_shape_a = (_a_dim1,)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _transr,
        _side,
        _uplo,
        _trans,
        _diag,
        _m,
        _n,
        _alpha,
        _a,
        _b,
        _ldb,
        _transr._length_,
        _side._length_,
        _uplo._length_,
        _trans._length_,
        _diag._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def zhfrk(transr, uplo, trans, n, k, alpha, a, beta, c):
    u"""
    Rank-k update of a complex Hermitian matrix, Rectangular Full Packed
    format.

    ``zhfrk`` performs one of the Hermitian rank-k update operations

        C←alpha AA^H+beta C or C←alpha A^HA+beta C,

    where A is a complex matrix, C is an n by n complex Hermitian matrix
    stored in Rectangular Full Packed (RFP) format, and alpha and beta
    are real scalars.

    For full information please refer to the NAG Library document for
    f06wq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wqf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of C or its
        conjugate transpose is stored.

        `transr` = 'N'

          The matrix C is stored in normal RFP format.

        `transr` = 'C'

          The conjugate transpose of the RFP representation of the
          matrix C is stored.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored in RFP format.

        `uplo` = 'U'

          The upper triangular part of C is stored in RFP format.

        `uplo` = 'L'

          The lower triangular part of C is stored in RFP format.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^H+beta C.

        `trans` = 'C'

          C←alpha A^HA+beta C.

    n : int
        n, the order of the matrix C.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'C'.

    alpha : float
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': max(1, `n`); if
        `trans`='C': max(1, `k`); otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': max(1, `k`); if
        `trans`='C': max(1, `n`); otherwise: 1.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'. If `alpha` = 0.0, `a` is not referenced.

    beta : float
        The scalar beta.

    c : complex, shape (`n`*(`n`+1)/2), modified in place
        On entry: the upper or lower triangular part (as specified by
        `uplo`) of the n by n Hermitian matrix C, stored in RFP format
        (as specified by `transr`). The storage format is described in
        detail in [the F07 Introduction].

        On exit: the updated matrix C, that is its upper or lower
        triangular part stored in RFP format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhfrk'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06wqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, transr(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[n*(n+1)//2].
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _transr = utils._EngineCharScalarType(
        transr, {'fun_name': fun_name, 'entity': 'transr'}, exp_chrlen=1,
    ).to_ctypes()
    _transr_py = utils._EngineCharScalarType.to_py(_transr)
    if _transr_py not in ('N', 'n', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transr in (\'N\', \'n\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='transr',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'C\', \'c\').'),
            ],
            errcodes=(3,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = max(1, _n_py)
    elif _trans_py in ('C', 'c'):
        _a_dim1 = max(1, _k_py)
    else:
        _a_dim1 = 1
    if _trans_py in ('N', 'n'):
        _a_dim2 = max(1, _k_py)
    elif _trans_py in ('C', 'c'):
        _a_dim2 = max(1, _n_py)
    else:
        _a_dim2 = 1
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py*(_n_py+1)//2,),
    )
    _c = _c_cls.to_ctypes()
    engine_ctypes(
        _call_vendor,
        _en_data,
        _transr,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _beta,
        _c,
        _transr._length_,
        _uplo._length_,
        _trans._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def dgemm(transa, transb, m, n, k, alpha, a, b, beta, c):
    u"""
    Matrix-matrix product, two real rectangular matrices.

    ``dgemm`` performs one of the matrix-matrix operations

        [table omitted]

    where A, B and C are real matrices, and alpha and beta are real
    scalars; C is always m by n.

    For full information please refer to the NAG Library document for
    f06ya

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yaf.html

    Parameters
    ----------
    transa : str, length 1
        Specifies whether the operation involves A or A^T.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T' or 'C'

          The operation involves A^T.

    transb : str, length 1
        Specifies whether the operation involves B or B^T.

        `transb` = 'N'

          The operation involves B.

        `transb` = 'T' or 'C'

          The operation involves B^T.

    m : int
        m, the number of rows of the matrix C; the number of rows of A
        if `transa` = 'N', or the number of columns of A if `transa` =
        'T' or 'C'.

    n : int
        n, the number of columns of the matrix C; the number of columns
        of B if `transb` = 'N', or the number of rows of B if `transb` =
        'T' or 'C'.

    k : int
        k, the number of columns of A if `transa` = 'N', or the number
        of rows of A if `transa` = 'T' or 'C'; the number of rows of B
        if `transb` = 'N', or the number of columns of B if `transb` =
        'T' or 'C'.

    alpha : float
        The scalar alpha.

    a : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transa`='N': `m`; if `transa` in
        ('T', 'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transa`='N': `k`; if `transa` in
        ('T', 'C'): `m`; otherwise: 0.

        The matrix A; A is m by k if `transa` = 'N', or k by m if
        `transa` = 'T' or 'C'.

    b : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transb`='N': `k`; if `transb` in
        ('T', 'C'): `n`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transb`='N': `n`; if `transb` in
        ('T', 'C'): `k`; otherwise: 0.

        The matrix B; B is k by n if `transb` = 'N', or n by k if
        `transb` = 'T' or 'C'.

    beta : float
        The scalar beta.

    c : float, shape (`m`, `n`), modified in place
        On entry: the m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `transb`.

            Constraint: `transb` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dgemm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06yaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, transa(len=1).
        utils._BYREF_STRING, # Input, transb(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transa.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _transa = utils._EngineCharScalarType(
        transa, {'fun_name': fun_name, 'entity': 'transa'}, exp_chrlen=1,
    ).to_ctypes()
    _transa_py = utils._EngineCharScalarType.to_py(_transa)
    if _transa_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transa in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='transa',
        ).manage()
    _transb = utils._EngineCharScalarType(
        transb, {'fun_name': fun_name, 'entity': 'transb'}, exp_chrlen=1,
    ).to_ctypes()
    _transb_py = utils._EngineCharScalarType.to_py(_transb)
    if _transb_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transb in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='transb',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(3,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _transa_py in ('N', 'n'):
        _a_dim1 = _m_py
    elif _transa_py in ('T', 't', 'C', 'c'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _transa_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _transa_py in ('T', 't', 'C', 'c'):
        _a_dim2 = _m_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if _transb_py in ('N', 'n'):
        _b_dim1 = _k_py
    elif _transb_py in ('T', 't', 'C', 'c'):
        _b_dim1 = _n_py
    else:
        _b_dim1 = 0
    if _transb_py in ('N', 'n'):
        _b_dim2 = _n_py
    elif _transb_py in ('T', 't', 'C', 'c'):
        _b_dim2 = _k_py
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=_beta_py != 0.0,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=_beta_py != 0.0)
    _routine_name = utils._EngineRoutineName(*b'f06ya ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _transa,
        _transb,
        _m,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _transa._length_,
        _transb._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def dsymm(side, uplo, m, n, alpha, a, b, beta, c):
    u"""
    Matrix-matrix product, one real symmetric matrix, one real
    rectangular matrix.

    ``dsymm`` performs one of the matrix-matrix operations

        C←alpha AB+beta C or C←alpha BA+beta C,

    where A is a real symmetric matrix, B and C are m by n real
    matrices, and alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06yc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ycf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    m : int
        m, the number of rows of the matrices B and C; the order of A if
        `side` = 'L'.

    n : int
        n, the number of columns of the matrices B and C; the order of A
        if `side` = 'R'.

    alpha : float
        The scalar alpha.

    a : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        The symmetric matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : float, shape (`m`, `n`)
        The m by n matrix B.

    beta : float
        The scalar beta.

    c : float, shape (`m`, `n`), modified in place
        On entry: the m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsymm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ycft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(3,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py
    else:
        _a_dim1 = 0
    if _side_py in ('L', 'l'):
        _a_dim2 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=_beta_py != 0.0,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=_beta_py != 0.0)
    _routine_name = utils._EngineRoutineName(*b'f06yc ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _side,
        _uplo,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _side._length_,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def dtrmm(side, uplo, transa, diag, m, n, alpha, a, b):
    u"""
    Matrix-matrix product, one real triangular matrix, one real
    rectangular matrix.

    ``dtrmm`` performs one of the matrix-matrix operations

        [table omitted]

    where B is an m by n real matrix, A is a real triangular matrix, and
    alpha is a real scalar.

    For full information please refer to the NAG Library document for
    f06yf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yff.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A or A^T.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T' or 'C'

          The operation involves A^T.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    n : int
        n, the number of columns of the matrix B; the order of A if
        `side` = 'R'.

    alpha : float
        The scalar alpha.

    a : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : float, shape (`m`, `n`), modified in place
        On entry: the m by n matrix B.

        If `alpha` = 0, `b` need not be set.

        On exit: the updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtrmm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06yfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, transa(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transa.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _transa = utils._EngineCharScalarType(
        transa, {'fun_name': fun_name, 'entity': 'transa'}, exp_chrlen=1,
    ).to_ctypes()
    _transa_py = utils._EngineCharScalarType.to_py(_transa)
    if _transa_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transa in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(3,),
            entity='transa',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(4,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(5,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(6,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py
    else:
        _a_dim1 = 0
    if _side_py in ('L', 'l'):
        _a_dim2 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06yf ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _side,
        _uplo,
        _transa,
        _diag,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _routine_name,
        _side._length_,
        _uplo._length_,
        _transa._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def dtrsm(side, uplo, transa, diag, m, n, alpha, a, b):
    u"""
    Solves a system of equations with multiple right-hand sides, real
    triangular coefficient matrix.

    ``dtrsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a real triangular matrix, B is an m by n real matrix, and
    alpha is a real scalar. A^inv denotes (A^T)^-1 or equivalently
    (A^-1)^T.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06yj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A^-1 or A^invT.

        `transa` = 'N'

          The operation involves A^-1.

        `transa` = 'T' or 'C'

          The operation involves A^invT.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    n : int
        n, the number of columns of the matrix B; the order of A if
        `side` = 'R'.

    alpha : float
        The scalar alpha.

    a : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : float, shape (`m`, `n`), modified in place
        On entry: the m by n matrix B.

        If `alpha` = 0, `b` need not be set.

        On exit: the updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dtrsm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06yjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, transa(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transa.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _transa = utils._EngineCharScalarType(
        transa, {'fun_name': fun_name, 'entity': 'transa'}, exp_chrlen=1,
    ).to_ctypes()
    _transa_py = utils._EngineCharScalarType.to_py(_transa)
    if _transa_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transa in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(3,),
            entity='transa',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(4,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(5,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(6,),
            entity='n',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py
    else:
        _a_dim1 = 0
    if _side_py in ('L', 'l'):
        _a_dim2 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06yj ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _side,
        _uplo,
        _transa,
        _diag,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _routine_name,
        _side._length_,
        _uplo._length_,
        _transa._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def dsyrk(uplo, trans, n, k, alpha, a, beta, c):
    u"""
    Rank-k update of a real symmetric matrix.

    ``dsyrk`` performs one of the symmetric rank-k update operations

        C←alpha AA^T+beta C or C←alpha A^TA+beta C,

    where A is a real matrix, C is an n by n real symmetric matrix, and
    alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06yp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ypf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^T+beta C.

        `trans` = 'T' or 'C'

          C←alpha A^TA+beta C.

    n : int
        n, the order of the matrix C; the number of rows of A if `trans`
        = 'N', or the number of columns of A if `trans` = 'T' or 'C'.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'T' or 'C'.

    alpha : float
        The scalar alpha.

    a : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans` in ('T',
        'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans` in ('T',
        'C'): `n`; otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T' or 'C'.

    beta : float
        The scalar beta.

    c : float, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix C.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsyrk'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ypft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = _n_py
    elif _trans_py in ('T', 't', 'C', 'c'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _trans_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _trans_py in ('T', 't', 'C', 'c'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06yp ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def dsyr2k(uplo, trans, n, k, alpha, a, b, beta, c):
    u"""
    Rank-2k update of a real symmetric matrix.

    ``dsyr2k`` performs one of the symmetric rank-2k update operations

        C←alpha AB^T+alpha BA^T+beta C or C←alpha A^TB+alpha B^TA+beta
        C,

    where A and B are real matrices, C is an n by n real symmetric
    matrix, and alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06yr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AB^T+alpha BA^T+beta C.

        `trans` = 'T' or 'C'

          C←alpha A^TB+alpha B^TA+beta C.

    n : int
        n, the order of the matrix C; the number of rows of A and B if
        `trans` = 'N', or the number of columns of A and B if `trans` =
        'T' or 'C'.

    k : int
        k, the number of columns of A and B if `trans` = 'N', or the
        number of rows of A and B if `trans` = 'T' or 'C'.

    alpha : float
        The scalar alpha.

    a : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans` in ('T',
        'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans` in ('T',
        'C'): `n`; otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T' or 'C'.

    b : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans` in ('T',
        'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans` in ('T',
        'C'): `n`; otherwise: 0.

        The matrix B; B is n by k if `trans` = 'N', or k by n if `trans`
        = 'T' or 'C'.

    beta : float
        The scalar beta.

    c : float, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix C.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.dsyr2k'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06yrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = _n_py
    elif _trans_py in ('T', 't', 'C', 'c'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _trans_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _trans_py in ('T', 't', 'C', 'c'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _b_dim1 = _n_py
    elif _trans_py in ('T', 't', 'C', 'c'):
        _b_dim1 = _k_py
    else:
        _b_dim1 = 0
    if _trans_py in ('N', 'n'):
        _b_dim2 = _k_py
    elif _trans_py in ('T', 't', 'C', 'c'):
        _b_dim2 = _n_py
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06yr ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def zgemm(transa, transb, m, n, k, alpha, a, b, beta, c):
    u"""
    Matrix-matrix product, two complex rectangular matrices.

    ``zgemm`` performs one of the matrix-matrix operations

        [table omitted]

    where A, B and C are complex matrices, and alpha and beta are
    complex scalars; C is always m by n.

    For full information please refer to the NAG Library document for
    f06za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zaf.html

    Parameters
    ----------
    transa : str, length 1
        Specifies whether the operation involves A, A^T or A^H.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T'

          The operation involves A^T.

        `transa` = 'C'

          The operation involves A^H.

    transb : str, length 1
        Specifies whether the operation involves B, B^T or B^H.

        `transb` = 'N'

          The operation involves B.

        `transb` = 'T'

          The operation involves B^T.

        `transb` = 'C'

          The operation involves B^H.

    m : int
        m, the number of rows of the matrix C; the number of rows of A
        if `transa` = 'N', or the number of columns of A if `transa` =
        'T' or 'C'.

    n : int
        n, the number of columns of the matrix C; the number of columns
        of B if `transb` = 'N', or the number of rows of B if `transb` =
        'T' or 'C'.

    k : int
        k, the number of columns of A if `transa` = 'N', or the number
        of rows of A if `transa` = 'T' or 'C'; the number of rows of B
        if `transb` = 'N', or the number of columns of B if `transb` =
        'T' or 'C'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transa`='N': `m`; if `transa` in
        ('T', 'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transa`='N': `k`; if `transa` in
        ('T', 'C'): `m`; otherwise: 0.

        The matrix A; A is m by k if `transa` = 'N', or k by m if
        `transa` = 'T' or 'C'.

    b : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transb`='N': `k`; if `transb` in
        ('T', 'C'): `n`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transb`='N': `n`; if `transb` in
        ('T', 'C'): `k`; otherwise: 0.

        The matrix B; B is k by n if `transb` = 'N', or n by k if
        `transb` = 'T' or 'C'.

    beta : complex
        The scalar beta.

    c : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `transb`.

            Constraint: `transb` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zgemm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, transa(len=1).
        utils._BYREF_STRING, # Input, transb(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transa.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transb.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _transa = utils._EngineCharScalarType(
        transa, {'fun_name': fun_name, 'entity': 'transa'}, exp_chrlen=1,
    ).to_ctypes()
    _transa_py = utils._EngineCharScalarType.to_py(_transa)
    if _transa_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transa in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(1,),
            entity='transa',
        ).manage()
    _transb = utils._EngineCharScalarType(
        transb, {'fun_name': fun_name, 'entity': 'transb'}, exp_chrlen=1,
    ).to_ctypes()
    _transb_py = utils._EngineCharScalarType.to_py(_transb)
    if _transb_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transb in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='transb',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(3,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(5,),
            entity='k',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _transa_py in ('N', 'n'):
        _a_dim1 = _m_py
    elif _transa_py in ('T', 't', 'C', 'c'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _transa_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _transa_py in ('T', 't', 'C', 'c'):
        _a_dim2 = _m_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if _transb_py in ('N', 'n'):
        _b_dim1 = _k_py
    elif _transb_py in ('T', 't', 'C', 'c'):
        _b_dim1 = _n_py
    else:
        _b_dim1 = 0
    if _transb_py in ('N', 'n'):
        _b_dim2 = _n_py
    elif _transb_py in ('T', 't', 'C', 'c'):
        _b_dim2 = _k_py
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=_beta_py != 0.0,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=_beta_py != 0.0)
    _routine_name = utils._EngineRoutineName(*b'f06za ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _transa,
        _transb,
        _m,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _transa._length_,
        _transb._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def zhemm(side, uplo, m, n, alpha, a, b, beta, c):
    u"""
    Matrix-matrix product, one complex Hermitian matrix, one complex
    rectangular matrix.

    ``zhemm`` performs one of the matrix-matrix operations

        C←alpha AB+beta C or C←alpha BA+beta C,

    where A is a complex Hermitian matrix, B and C are m by n complex
    matrices, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zcf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    m : int
        m, the number of rows of the matrices B and C; the order of A if
        `side` = 'L'.

    n : int
        n, the number of columns of the matrices B and C; the order of A
        if `side` = 'R'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        The Hermitian matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, shape (`m`, `n`)
        The m by n matrix B.

    beta : complex
        The scalar beta.

    c : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zhemm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(3,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py
    else:
        _a_dim1 = 0
    if _side_py in ('L', 'l'):
        _a_dim2 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=_beta_py != 0.0,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=_beta_py != 0.0)
    _routine_name = utils._EngineRoutineName(*b'f06zc ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _side,
        _uplo,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _side._length_,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def ztrmm(side, uplo, transa, diag, m, n, alpha, a, b):
    u"""
    Matrix-matrix product, one complex triangular matrix, one complex
    rectangular matrix.

    ``ztrmm`` performs one of the matrix-matrix operations

        [table omitted]

    where B is an m by n complex matrix, A is a complex triangular
    matrix, and alpha is a complex scalar.

    For full information please refer to the NAG Library document for
    f06zf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zff.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A, A^T or A^H.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T'

          The operation involves A^T.

        `transa` = 'C'

          The operation involves A^H.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    n : int
        n, the number of columns of the matrix B; the order of A if
        `side` = 'R'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix B.

        If `alpha` = 0, `b` need not be set.

        On exit: the updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztrmm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, transa(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transa.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _transa = utils._EngineCharScalarType(
        transa, {'fun_name': fun_name, 'entity': 'transa'}, exp_chrlen=1,
    ).to_ctypes()
    _transa_py = utils._EngineCharScalarType.to_py(_transa)
    if _transa_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transa in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(3,),
            entity='transa',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(4,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(5,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(6,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py
    else:
        _a_dim1 = 0
    if _side_py in ('L', 'l'):
        _a_dim2 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06zf ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _side,
        _uplo,
        _transa,
        _diag,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _routine_name,
        _side._length_,
        _uplo._length_,
        _transa._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def ztrsm(side, uplo, transa, diag, m, n, alpha, a, b):
    u"""
    Solves system of equations with multiple right-hand sides, complex
    triangular coefficient matrix.

    ``ztrsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a complex triangular matrix, B is an m by n complex
    matrix, and alpha is a complex scalar. A^inv denotes (A^T)^-1 or
    equivalently (A^-1)^T; A^invH denotes (A^H)^-1 or equivalently
    (A^-1)^H.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06zj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A^-1, A^invT or A^invH.

        `transa` = 'N'

          The operation involves A^-1.

        `transa` = 'T'

          The operation involves A^invT.

        `transa` = 'C'

          The operation involves A^invH.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    n : int
        n, the number of columns of the matrix B; the order of A if
        `side` = 'R'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix B.

        If `alpha` = 0, `b` need not be set.

        On exit: the updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.ztrsm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, transa(len=1).
        utils._BYREF_STRING, # Input, diag(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transa.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of diag.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _transa = utils._EngineCharScalarType(
        transa, {'fun_name': fun_name, 'entity': 'transa'}, exp_chrlen=1,
    ).to_ctypes()
    _transa_py = utils._EngineCharScalarType.to_py(_transa)
    if _transa_py not in ('N', 'n', 'T', 't', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('transa in (\'N\', \'n\', \'T\', \'t\', \'C\', \'c\').'),
            ],
            errcodes=(3,),
            entity='transa',
        ).manage()
    _diag = utils._EngineCharScalarType(
        diag, {'fun_name': fun_name, 'entity': 'diag'}, exp_chrlen=1,
    ).to_ctypes()
    _diag_py = utils._EngineCharScalarType.to_py(_diag)
    if _diag_py not in ('N', 'n', 'U', 'u'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('diag in (\'N\', \'n\', \'U\', \'u\').'),
            ],
            errcodes=(4,),
            entity='diag',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(5,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(6,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py
    else:
        _a_dim1 = 0
    if _side_py in ('L', 'l'):
        _a_dim2 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06zj ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _side,
        _uplo,
        _transa,
        _diag,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _routine_name,
        _side._length_,
        _uplo._length_,
        _transa._length_,
        _diag._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _b_cls.refresh_py(_b)
    return_data = None
    return return_data

def zherk(uplo, trans, n, k, alpha, a, beta, c):
    u"""
    Rank-k update of a complex Hermitian matrix.

    ``zherk`` performs one of the Hermitian rank-k update operations

        C←alpha AA^H+beta C or C←alpha A^HA+beta C

    where A is a complex matrix, C is an n by n complex Hermitian
    matrix, and alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06zp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zpf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^H+beta C.

        `trans` = 'C'

          C←alpha A^HA+beta C.

    n : int
        n, the order of the matrix C; the number of rows of A if `trans`
        = 'N', or the number of columns of A if `trans` = 'C'.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'C'.

    alpha : float
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans`='C': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='C': `n`;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'.

    beta : float
        The scalar beta.

    c : complex, shape (`n`, `n`), modified in place
        On entry: the n by n Hermitian matrix C.

        On exit: the updated matrix C. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zherk'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = _n_py
    elif _trans_py in ('C', 'c'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _trans_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _trans_py in ('C', 'c'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06zp ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def zher2k(uplo, trans, n, k, alpha, a, b, beta, c):
    u"""
    Rank-2k update of a complex Hermitian matrix.

    ``zher2k`` performs one of the Hermitian rank-2k update operations

        C←alpha AB^H+alpha_BA^H+beta C or C←alpha A^HB+alpha_B^HA+beta
        C,

    where A and B are complex matrices, C is an n by n complex Hermitian
    matrix, alpha is a complex scalar, and beta is a real scalar.

    For full information please refer to the NAG Library document for
    f06zr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AB^H+alpha_BA^H+beta C.

        `trans` = 'C'

          C←alpha A^HB+alpha_B^HA+beta C.

    n : int
        n, the order of the matrix C; the number of rows of A if `trans`
        = 'N', or the number of columns of A if `trans` = 'C'.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'C'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans`='C': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='C': `n`;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'.

    b : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans`='C': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='C': `n`;
        otherwise: 0.

        The matrix B; B is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'.

    beta : float
        The scalar beta.

    c : complex, shape (`n`, `n`), modified in place
        On entry: the n by n Hermitian matrix C.

        On exit: the updated matrix C. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zher2k'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'C', 'c'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'C\', \'c\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = _n_py
    elif _trans_py in ('C', 'c'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _trans_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _trans_py in ('C', 'c'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _b_dim1 = _n_py
    elif _trans_py in ('C', 'c'):
        _b_dim1 = _k_py
    else:
        _b_dim1 = 0
    if _trans_py in ('N', 'n'):
        _b_dim2 = _k_py
    elif _trans_py in ('C', 'c'):
        _b_dim2 = _n_py
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06zr ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def zsymm(side, uplo, m, n, alpha, a, b, beta, c):
    u"""
    Matrix-matrix product, one complex symmetric matrix, one complex
    rectangular matrix.

    ``zsymm`` performs one of the matrix-matrix operations

        C←alpha AB+beta C or C←alpha BA+beta C,

    where A is a complex symmetric matrix, B and C are m by n complex
    matrices, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ztf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    m : int
        m, the number of rows of the matrices B and C; the order of A if
        `side` = 'L'.

    n : int
        n, the number of columns of the matrices B and C; the order of A
        if `side` = 'R'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': `n`;
        otherwise: 0.

        The symmetric matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, shape (`m`, `n`)
        The m by n matrix B.

    beta : complex
        The scalar beta.

    c : complex, shape (`m`, `n`), modified in place
        On entry: the m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsymm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06ztft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, b[ldb, n].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _side_py = utils._EngineCharScalarType.to_py(_side)
    if _side_py not in ('L', 'l', 'R', 'r'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('side in (\'L\', \'l\', \'R\', \'r\').'),
            ],
            errcodes=(1,),
            entity='side',
        ).manage()
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(2,),
            entity='uplo',
        ).manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if not _m_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('m >= 0.'),
            ],
            errcodes=(3,),
            entity='m',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(4,),
            entity='n',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _side_py in ('L', 'l'):
        _a_dim1 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim1 = _n_py
    else:
        _a_dim1 = 0
    if _side_py in ('L', 'l'):
        _a_dim2 = _m_py
    elif _side_py in ('R', 'r'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _beta_py = utils._EngineComplex128ScalarType.to_py(_beta)
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=_beta_py != 0.0,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=_beta_py != 0.0)
    _routine_name = utils._EngineRoutineName(*b'f06zt ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _side,
        _uplo,
        _m,
        _n,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _side._length_,
        _uplo._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def zsyrk(uplo, trans, n, k, alpha, a, beta, c):
    u"""
    Rank-k update of a complex symmetric matrix.

    ``zsyrk`` performs one of the symmetric rank-k update operations

        C←alpha AA^T+beta C or C←alpha A^TA+beta C,

    where A is a complex matrix, C is an n by n complex symmetric
    matrix, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zuf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^T+beta C.

        `trans` = 'T'

          C←alpha A^TA+beta C.

    n : int
        n, the order of the matrix C; the number of rows of A and B if
        `trans` = 'N', or the number of columns of A and B if `trans` =
        'T'.

    k : int
        k, the number of columns of A and B if `trans` = 'N', or the
        number of rows of A and B if `trans` = 'T'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': `n`;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'.

    beta : complex
        The scalar beta.

    c : complex, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix C.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsyrk'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zuft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = _n_py
    elif _trans_py in ('T', 't'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _trans_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _trans_py in ('T', 't'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06zu ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def zsyr2k(uplo, trans, n, k, alpha, a, b, beta, c):
    u"""
    Rank-2k update of a complex symmetric matrix.

    ``zsyr2k`` performs one of the symmetric rank-2k update operations

        C←alpha AB^T+alpha BA^T+beta C or C←alpha A^TB+alpha B^TA+beta
        C,

    where A and B are complex matrices, C is an n by n complex symmetric
    matrix, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AB^T+alpha BA^T+beta C.

        `trans` = 'T'

          C←alpha A^TB+alpha B^TA+beta C.

    n : int
        n, the order of the matrix C; the number of rows of A and B if
        `trans` = 'N', or the number of columns of A and B if `trans` =
        'T'.

    k : int
        k, the number of columns of A and B if `trans` = 'N', or the
        number of rows of A and B if `trans` = 'T'.

    alpha : complex
        The scalar alpha.

    a : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': `n`;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'.

    b : complex, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': `n`; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': `n`;
        otherwise: 0.

        The matrix B; B is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'.

    beta : complex
        The scalar beta.

    c : complex, shape (`n`, `n`), modified in place
        On entry: the n by n symmetric matrix C.

        On exit: the updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.blas.zsyr2k'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f06zwft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_LOGICAL, # Input, call_vendor.
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_COMPLEX128, # Input, alpha.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineComplex128ArrayType.c_type, # Input, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_COMPLEX128, # Input, beta.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_STRING, # Input, routine_name(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (c, 'c')),
    )
    _call_vendor = utils._EngineState._literal_true_cint
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _uplo_py = utils._EngineCharScalarType.to_py(_uplo)
    if _uplo_py not in ('U', 'u', 'L', 'l'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('uplo in (\'U\', \'u\', \'L\', \'l\').'),
            ],
            errcodes=(1,),
            entity='uplo',
        ).manage()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = utils._EngineCharScalarType.to_py(_trans)
    if _trans_py not in ('N', 'n', 'T', 't'):
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('trans in (\'N\', \'n\', \'T\', \'t\').'),
            ],
            errcodes=(2,),
            entity='trans',
        ).manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n >= 0.'),
            ],
            errcodes=(3,),
            entity='n',
        ).manage()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    if not _k_py >= 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('k >= 0.'),
            ],
            errcodes=(4,),
            entity='k',
        ).manage()
    _alpha = utils._EngineComplex128ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _a_dim1 = _n_py
    elif _trans_py in ('T', 't'):
        _a_dim1 = _k_py
    else:
        _a_dim1 = 0
    if _trans_py in ('N', 'n'):
        _a_dim2 = _k_py
    elif _trans_py in ('T', 't'):
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_a_dim1, _a_dim2)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if _trans_py in ('N', 'n'):
        _b_dim1 = _n_py
    elif _trans_py in ('T', 't'):
        _b_dim1 = _k_py
    else:
        _b_dim1 = 0
    if _trans_py in ('N', 'n'):
        _b_dim2 = _k_py
    elif _trans_py in ('T', 't'):
        _b_dim2 = _n_py
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineComplex128ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineComplex128ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'f06zw ')
    engine_ctypes(
        _call_vendor,
        _en_data,
        _uplo,
        _trans,
        _n,
        _k,
        _alpha,
        _a,
        _lda,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _routine_name,
        _uplo._length_,
        _trans._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data
