# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `mip` Chapter.

``mip`` - Operations Research

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module provides functions to solve certain integer programming,
transportation and shortest path problems.
Additionally 'best subset' functions are included.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/hintro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections
import sys as _sys

from ..base import utils

def ilp_dense(itmax, msglvl, n, m, a, bl, bu, intvar, cvec, maxnod, intfst, maxdpt, toliv, tolfes, bigbnd, x, comm, io_manager=None):
    u"""
    Integer LP problem (dense).

    ``ilp_dense`` solves 'zero-one', 'general', 'mixed' or 'all' integer
    programming problems using a branch and bound method.
    The function may also be used to find either the first integer
    solution or the optimum integer solution.
    It is not intended for large sparse problems.

    For full information please refer to the NAG Library document for
    h02bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02bbf.html

    Parameters
    ----------
    itmax : int
        An upper bound on the number of iterations for each LP problem.

    msglvl : int
        The amount of printout produced by ``ilp_dense``, as indicated
        below (see [Description of Printed Output] for a description of
        the printed output). All output is written to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

        [table omitted]

    n : int
        n, the number of variables.

    m : int
        m, the number of general linear constraints.

    a : float, shape (`m`, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `m`>0: `n`; if `m`=0: 1; otherwise: 0.

        The ith row of `a` must contain the coefficients of the ith
        general constraint, for i = 1,2,...,m.

        If `m` = 0 then the array `a` is not referenced.

    bl : float, shape (`n`+`m`)
        `bl` must contain the lower bounds and `bu` the upper bounds,
        for all the constraints in the following order. The first n
        elements of each array must contain the bounds on the variables,
        and the next m elements the bounds for the general linear
        constraints (if any). To specify a nonexistent lower bound
        (i.e., l_j = - infinity), set `bl`[j] <= -`bigbnd` and to
        specify a nonexistent upper bound (i.e., u_j = + infinity), set
        `bu`[j] >= `bigbnd`. To specify the jth constraint as an
        equality, set `bl`[j] = `bu`[j] = beta, say, where |beta| <
        `bigbnd`.

    bu : float, shape (`n`+`m`)
        `bl` must contain the lower bounds and `bu` the upper bounds,
        for all the constraints in the following order. The first n
        elements of each array must contain the bounds on the variables,
        and the next m elements the bounds for the general linear
        constraints (if any). To specify a nonexistent lower bound
        (i.e., l_j = - infinity), set `bl`[j] <= -`bigbnd` and to
        specify a nonexistent upper bound (i.e., u_j = + infinity), set
        `bu`[j] >= `bigbnd`. To specify the jth constraint as an
        equality, set `bl`[j] = `bu`[j] = beta, say, where |beta| <
        `bigbnd`.

    intvar : int, shape (`n`)
        Indicates which are the integer variables in the problem. For
        example, if x_j is an integer variable then `intvar`[j] must be
        set to 1, and 0 otherwise.

    cvec : float, shape (`n`)
        The coefficients c_j of the objective function F(x) = c_1
        x_1+c_2 x_2+...+c_n x_n. The function attempts to find a minimum
        of F(x). If a maximum of F(x) is desired, `cvec`[j] should be
        set to -c_j, for j = 1,2,...,n, so that the function will find a
        minimum of -F(x).

    maxnod : int
        The maximum number of nodes that are to be searched in order to
        find a solution (optimum integer solution). If `maxnod` <= 0 and
        `intfst` <= 0, then the B&B tree search is continued until all
        the nodes have been investigated.

    intfst : int
        Specifies whether to terminate the B&B tree search after the
        first integer solution (if any) is obtained. If `intfst` > 0
        then the B&B tree search is terminated at node k say, which
        contains the first integer solution. For `maxnod` > 0 this
        applies only if k <= `maxnod`.

    maxdpt : int
        The maximum depth of the B&B tree used for branch and bound.

    toliv : float
        The integer feasibility tolerance; i.e., an integer variable is
        considered to take an integer value if its violation does not
        exceed `toliv`. For example, if the integer variable x_j is near
        unity then x_j is considered to be integer only if (1-`toliv`)
        <= x_j <= (1+`toliv`).

    tolfes : float
        The maximum acceptable absolute violation in each constraint at
        a 'feasible' point (feasibility tolerance); i.e., a constraint
        is considered satisfied if its violation does not exceed
        `tolfes`.

    bigbnd : float
        The 'infinite' bound size in the definition of the problem
        constraints. More precisely, any upper bound greater than or
        equal to `bigbnd` will be regarded as + infinity and any lower
        bound less than or equal to -`bigbnd` will be regarded as -
        infinity.

    x : float, shape (`n`), modified in place
        On entry: an initial estimate of the original LP solution.

        On exit: with no exception or warning is raised, `x` contains a
        solution which will be an estimate of either the optimum integer
        solution or the first integer solution, depending on the value
        of `intfst`. If `errno` = 9, then `x` contains a solution which
        will be an estimate of the best integer solution that was
        obtained by searching `maxnod` nodes.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.lp_solve'.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    itmax : int
        Unchanged if on entry `itmax` > 0, else `itmax` = max(50,
        5*(`n`+`m`)).

    toliv : float
        Unchanged if on entry `toliv` > 0.0, else `toliv` = 10^(-5).

    tolfes : float
        Unchanged if on entry `tolfes` > 0.0, else `tolfes` =
        sqrt(epsilon) (where epsilon is the machine precision).

    bigbnd : float
        Unchanged if on entry `bigbnd` > 0.0, else `bigbnd` = 10^20.

    objmip : float
        With the function exits successfully or `errno` = 9, `objmip`
        contains the value of the objective function for the IP
        solution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The problem does not have a feasible integer solution.

        (`errno` 2)
            The LP solution is unbounded.

        (`errno` 3)
            The LP does not have a feasible solution.

        (`errno` 4)
            Iteration limit reached without finding a solution.

        (`errno` 6)
            On entry, there were *<value>* infinite or inconsistent
            bounds given in arrays `bl` and `bu`. For further advisory
            details set `msglvl` > 2.

        (`errno` 6)
            On entry, the dimension of `comm`['iwork'] is too small:
            liwork = *<value>*. `comm`['iwork'] must be of dimension (at
            least) *<value>*.

        (`errno` 6)
            On entry, the dimension of `comm`['rwork'] is too small:
            lrwork = *<value>*. `comm`['rwork'] must be of dimension (at
            least) *<value>*.

        (`errno` 6)
            On entry, `maxdpt` = *<value>*.

            Constraint: `maxdpt` >= 2.

        (`errno` 6)
            On entry, i = *<value>*, `intvar`[i-1] = *<value>*.

            Constraint: `intvar` = 0 or 1.

        (`errno` 6)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 8)
            Increase `maxdpt` and rerun ``ilp_dense``.

        (`errno` 8)
            `maxdpt` is too small to solve the problem: `maxdpt` =
            *<value>*.

        (`errno` 10)
            No feasible solution was found for the number of nodes
            investigated in the B&B tree.

        (`errno` 11)
            Not enough workspace to solve problem.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            Search of a branch was terminated due to iteration limit.

        (`errno` 9)
            The IP solution returned is the best solution for the number
            of nodes investigated in the B&B tree.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.mip.ilp_dense_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.ilp_dense'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h02bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input/Output, itmax.
        utils._BYREF_INT, # Input, msglvl.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, bl[n+m].
        utils._EngineFloat64ArrayType.c_type, # Input, bu[n+m].
        utils._EngineIntArrayType.c_type, # Input, intvar[n].
        utils._EngineFloat64ArrayType.c_type, # Input, cvec[n].
        utils._BYREF_INT, # Input, maxnod.
        utils._BYREF_INT, # Input, intfst.
        utils._BYREF_INT, # Input, maxdpt.
        utils._BYREF_FLOAT64, # Input/Output, toliv.
        utils._BYREF_FLOAT64, # Input/Output, tolfes.
        utils._BYREF_FLOAT64, # Input/Output, bigbnd.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, objmip.
        utils._EngineIntArrayType.c_type, # Output, iwork[liwork].
        utils._BYREF_INT, # Input, liwork.
        utils._EngineFloat64ArrayType.c_type, # Output, rwork[lrwork].
        utils._BYREF_INT, # Input, lrwork.
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[120].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[610].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[475].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _itmax = utils._EngineIntScalarType(
        itmax, {'fun_name': fun_name, 'entity': 'itmax'},
    ).to_ctypes()
    _msglvl = utils._EngineIntScalarType(
        msglvl, {'fun_name': fun_name, 'entity': 'msglvl'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _m_py > 0:
        _a_dim2 = _n_py
    elif _m_py == 0:
        _a_dim2 = 1
    else:
        _a_dim2 = 0
    exp_shape_a = (_m_py, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _bl = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_n_py+_m_py,),
    ).to_ctypes()
    _bu = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_n_py+_m_py,),
    ).to_ctypes()
    _intvar = utils._EngineIntArrayType(
        intvar, {'fun_name': fun_name, 'entity': 'intvar'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _cvec = utils._EngineFloat64ArrayType(
        cvec, {'fun_name': fun_name, 'entity': 'cvec'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _maxnod = utils._EngineIntScalarType(
        maxnod, {'fun_name': fun_name, 'entity': 'maxnod'},
    ).to_ctypes()
    _intfst = utils._EngineIntScalarType(
        intfst, {'fun_name': fun_name, 'entity': 'intfst'},
    ).to_ctypes()
    _maxdpt = utils._EngineIntScalarType(
        maxdpt, {'fun_name': fun_name, 'entity': 'maxdpt'},
    ).to_ctypes()
    _toliv = utils._EngineFloat64ScalarType(
        toliv, {'fun_name': fun_name, 'entity': 'toliv'},
    ).to_ctypes()
    _tolfes = utils._EngineFloat64ScalarType(
        tolfes, {'fun_name': fun_name, 'entity': 'tolfes'},
    ).to_ctypes()
    _bigbnd = utils._EngineFloat64ScalarType(
        bigbnd, {'fun_name': fun_name, 'entity': 'bigbnd'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _objmip = utils._EngineFloat64ScalarType.empty_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('iwork', 'rwork'),
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _maxdpt_py = utils._EngineIntScalarType.to_py(_maxdpt)
    _iwork = utils._EngineIntArrayType.empty_ctypes((25+_n_py+_m_py)*_maxdpt_py+5*_n_py+_m_py+4)
    comm['iwork'] = _iwork
    _liwork_py = utils._EngineIntArrayType.get_shape(comm['iwork'], {'fun_name': fun_name, 'entity': 'comm[\'iwork\']'}, exp_rank=1, validate=False)[0]
    _liwork = utils._EngineIntScalarType(
        _liwork_py, {'fun_name': fun_name, 'entity': '_liwork'},
    ).to_ctypes()
    _rwork = utils._EngineFloat64ArrayType.empty_ctypes(_maxdpt_py*(_n_py+1)+2*min(_n_py, _m_py+1)**2+14*_n_py+12*_m_py)
    comm['rwork'] = _rwork
    _lrwork_py = utils._EngineFloat64ArrayType.get_shape(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, exp_rank=1, validate=False)[0]
    _lrwork = utils._EngineIntScalarType(
        _lrwork_py, {'fun_name': fun_name, 'entity': '_lrwork'},
    ).to_ctypes()
    _lwsav = comm['lwsav']
    _iwsav = comm['iwsav']
    _rwsav = comm['rwsav']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _itmax,
        _msglvl,
        _n,
        _m,
        _a,
        _lda,
        _bl,
        _bu,
        _intvar,
        _cvec,
        _maxnod,
        _intfst,
        _maxdpt,
        _toliv,
        _tolfes,
        _bigbnd,
        _x,
        _objmip,
        _iwork,
        _liwork,
        _rwork,
        _lrwork,
        _lwsav,
        _iwsav,
        _rwsav,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The problem does not have a feasible integer solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The LP solution is unbounded.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The LP does not have a feasible solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Iteration limit reached without finding a solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, there were {:s} infinite or inconsistent bounds'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'given in arrays bl and bu.',
                    'For further advisory details set msglvl > 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the dimension of comm[\'iwork\'] is too small:',
                    'liwork = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'comm[\'iwork\'] must be of dimension (at least) {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the dimension of comm[\'rwork\'] is too small:',
                    'lrwork = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'comm[\'rwork\'] must be of dimension (at least) {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxdpt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxdpt >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, intvar[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: intvar = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Search of a branch was terminated due to iteration limit.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Increase maxdpt and rerun ilp_dense.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'maxdpt is too small to solve the problem:',
                    'maxdpt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The IP solution returned is the best solution for',
                    'the number of nodes investigated in the B&B tree.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No feasible solution was found for the number of nodes',
                    'investigated in the B&B tree.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Not enough workspace to solve problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _itmax_py = utils._EngineIntScalarType.to_py(_itmax)
        _toliv_py = utils._EngineFloat64ScalarType.to_py(_toliv)
        _tolfes_py = utils._EngineFloat64ScalarType.to_py(_tolfes)
        _bigbnd_py = utils._EngineFloat64ScalarType.to_py(_bigbnd)
        _x_cls.refresh_py(_x)
        _objmip_py = utils._EngineFloat64ScalarType.to_py(_objmip)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itmax', 'toliv', 'tolfes', 'bigbnd', 'objmip'],
        )(
            _itmax_py,
            _toliv_py,
            _tolfes_py,
            _bigbnd_py,
            _objmip_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def ilp_mpsx(infile, maxn, maxm, optim, xbldef, xbudef, maxdpt, msglvl, x, crname, iwork, liwork, rwork, lrwork, comm, io_manager=None):
    u"""
    Interpret MPSX data file defining IP or LP problem, optimize and
    print solution.

    ``ilp_mpsx`` solves linear or integer programming problems specified
    in MPSX input format.
    It is not intended for large sparse problems.

    For full information please refer to the NAG Library document for
    h02bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02bff.html

    Parameters
    ----------
    infile : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        associated with the MPSX data file.

    maxn : int
        An upper limit for the number of variables in the problem.

    maxm : int
        An upper limit for the number of constraints (including the
        objective) in the problem.

    optim : str, length 3
        Specifies the direction of the optimization. `optim` must be set
        to 'MIN' for minimization and to 'MAX' for maximization.

    xbldef : float
        The default lower bound to be used for the variables in the
        problem when none is specified in the BOUNDS section of the MPSX
        data file. For a standard LP or IP problem `xbldef` would
        normally be set to zero.

    xbudef : float
        The default upper bound to be used for the variables in the
        problem when none is specified in the BOUNDS section of the MPSX
        data file. For a standard LP or IP problem `xbudef` would
        normally be set to 'infinity' (i.e., `xbudef` >= 10^20).

    maxdpt : int
        For an IP problem, `maxdpt` must specify the maximum depth of
        the branch and bound tree.

    msglvl : int
        The amount of printout produced by ``opt.lp_solve`` or
        ``ilp_dense``, as indicated below. For a description of the
        printed output see [Further Comments] or [Parameters h02bbf] (as
        appropriate). All output is written to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

        For an LP problem (``opt.lp_solve``):

        [table omitted]

        For an IP problem (``ilp_dense``):

        [table omitted]

    x : float, shape (`maxn`), modified in place
        On exit: the solution to the problem, stored in
        `x`[0],`x`[1],...,`x`[`n`-1]. `x`[i-1] is the value of the
        variable whose MPSX name is stored in `crname`[i-1], for i =
        1,2,...,`n`.

    crname : str, length 8, shape (`maxn`+`maxm`), modified in place
        On exit: the first `n` elements contain the MPSX names for the
        variables in the problem.

    iwork : int, shape (`liwork`), modified in place
        On exit: the first (`n`+`m`) elements contain ISTATE (the status
        of the constraints in the working set at the solution). Further
        details can be found in [Parameters] (as appropriate).

    liwork : int
        The dimension of the array `iwork`.

    rwork : float, shape (`lrwork`), modified in place
        On exit: the first (`n`+`m`) elements contain BL (the lower
        bounds), the next (`n`+`m`) elements contain BU (the upper
        bounds) and the next (`n`+`m`) elements contain CLAMDA (the
        Lagrange-multipliers). Further details can be found in
        [Parameters] (as appropriate). Note that for an IP problem the
        contents of BL and BU may not be the same as those originally
        specified in the MPSX data file and/or via the arguments
        `xbldef` and `xbudef`.

    lrwork : int
        The dimension of the array `rwork`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.lp_solve'.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    n : int
        n, the actual number of variables in the problem.

    m : int
        m, the actual number of general linear constraints in the
        problem.

    Raises
    ------
    NagValueError
        (`errno` i < 0)
            Either `maxm` and/or `maxn` are too small or the MPSX data
            file is nonstandard and/or corrupt.

        (`errno` 2)
            The LP solution is unbounded.

        (`errno` 3)
            The LP does not have a feasible solution.

        (`errno` 4)
            Iteration limit reached without finding a solution.

        (`errno` 5)
            On entry, not enough real workspace to solve problem:
            `lrwork` = *<value>* `lrwork` must be at least *<value>*.

        (`errno` 5)
            On entry, not enough integer workspace to solve problem:
            `liwork` = *<value>* `liwork` must be at least *<value>*.

        (`errno` 5)
            On entry, `maxdpt` = *<value>*.

            Constraint: `maxdpt` >= 2.

        (`errno` 5)
            On entry, not enough real workspace to read data file:
            `lrwork` = *<value>*.

        (`errno` 5)
            On entry, not enough integer workspace to read data file:
            `liwork` = *<value>*.

        (`errno` 5)
            On entry, `optim` = *<value>*.

            Constraint: `optim` = 'MIN' or 'MAX'.

        (`errno` 5)
            On entry, `xbldef` = *<value>* and `xbudef` = *<value>*.

            Constraint: `xbldef` <= `xbudef`.

        (`errno` 5)
            On entry, `maxm` = *<value>*.

            Constraint: `maxm` >= 1.

        (`errno` 5)
            On entry, `maxn` = *<value>*.

            Constraint: `maxn` >= 1.

        (`errno` 5)
            On entry, `infile` = *<value>*.

            Constraint: 0 <= `infile` <= 99.

        (`errno` 6)
            A serious error has occurred. Check all function calls and
            array dimensions.

        (`errno` 8)
            Increase `maxdpt` and rerun ``ilp_mpsx``.

        (`errno` 8)
            `maxdpt` is too small to solve the problem: `maxdpt` =
            *<value>*.

        (`errno` 9)
            The IP solution returned is the best solution for the number
            of nodes investigated in the branch and bound tree.

        (`errno` 10)
            No feasible solution was found for the number of nodes
            investigated in the branch and bound tree.

        (`errno` 11)
            Not enough workspace to solve problem.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Weak LP solution.

        (`errno` 1)
            The problem does not have a feasible integer solution.

        (`errno` 7)
            Search of a branch was terminated due to iteration limit.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.ilp_mpsx'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_read_rech_py(_, en_data, nin, recin, errbuf, ifail, recin_len, errbuf_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_read_rec(
                en_data, utils._EngineIntScalarType.to_py(nin.contents), recin, errbuf, recin_len, io_manager, excs, ifail.contents,
            )
        except:
            utils._read_err(
                _sys.exc_info(),
                excs,
                utils.FileObjManager._read_err,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h02bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._READ_REC_FUNCTYPE,
        utils._READ_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, infile.
        utils._BYREF_INT, # Input, maxn.
        utils._BYREF_INT, # Input, maxm.
        utils._BYREF_STRING, # Input, optim(len=3).
        utils._BYREF_FLOAT64, # Input, xbldef.
        utils._BYREF_FLOAT64, # Input, xbudef.
        utils._BYREF_INT, # Input, maxdpt.
        utils._BYREF_INT, # Input, msglvl.
        utils._BYREF_INT, # Output, n.
        utils._BYREF_INT, # Output, m.
        utils._EngineFloat64ArrayType.c_type, # Output, x[maxn].
        utils._EngineCharArrayType.c_type, # Output, crname(len=8)[maxn+maxm].
        utils._EngineIntArrayType.c_type, # Output, iwork[liwork].
        utils._BYREF_INT, # Input, liwork.
        utils._EngineFloat64ArrayType.c_type, # Output, rwork[lrwork].
        utils._BYREF_INT, # Input, lrwork.
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[120].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[610].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[475].
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optim.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of crname.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _read_rech = utils._READ_RECH_FUNCTYPE(cb_read_rech_py)
    _infile = utils._EngineIntScalarType(
        infile, {'fun_name': fun_name, 'entity': 'infile'},
    ).to_ctypes()
    _maxn = utils._EngineIntScalarType(
        maxn, {'fun_name': fun_name, 'entity': 'maxn'},
    ).to_ctypes()
    _maxm = utils._EngineIntScalarType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
    ).to_ctypes()
    _optim = utils._EngineCharScalarType(
        optim, {'fun_name': fun_name, 'entity': 'optim'}, exp_chrlen=3,
    ).to_ctypes()
    _xbldef = utils._EngineFloat64ScalarType(
        xbldef, {'fun_name': fun_name, 'entity': 'xbldef'},
    ).to_ctypes()
    _xbudef = utils._EngineFloat64ScalarType(
        xbudef, {'fun_name': fun_name, 'entity': 'xbudef'},
    ).to_ctypes()
    _maxdpt = utils._EngineIntScalarType(
        maxdpt, {'fun_name': fun_name, 'entity': 'maxdpt'},
    ).to_ctypes()
    _msglvl = utils._EngineIntScalarType(
        msglvl, {'fun_name': fun_name, 'entity': 'msglvl'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType.empty_ctypes()
    _m = utils._EngineIntScalarType.empty_ctypes()
    _maxn_py = utils._EngineIntScalarType.to_py(_maxn)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_maxn_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _maxm_py = utils._EngineIntScalarType.to_py(_maxm)
    _crname_cls = utils._EngineCharArrayType(
        crname, {'fun_name': fun_name, 'entity': 'crname'}, exp_chrlen=8,
        exp_shape=(_maxn_py+_maxm_py,),
        check_contents=False,
    )
    _crname = _crname_cls.to_ctypes(can_reference_contents=False)
    _crname_chrlen = _crname_cls.chrlen
    _liwork = utils._EngineIntScalarType(
        liwork, {'fun_name': fun_name, 'entity': 'liwork'},
    ).to_ctypes()
    _liwork_py = utils._EngineIntScalarType.to_py(_liwork)
    _iwork_cls = utils._EngineIntArrayType(
        iwork, {'fun_name': fun_name, 'entity': 'iwork'},
        exp_shape=(_liwork_py,),
        check_contents=False,
    )
    _iwork = _iwork_cls.to_ctypes(can_reference_contents=False)
    _lrwork = utils._EngineIntScalarType(
        lrwork, {'fun_name': fun_name, 'entity': 'lrwork'},
    ).to_ctypes()
    _lrwork_py = utils._EngineIntScalarType.to_py(_lrwork)
    _rwork_cls = utils._EngineFloat64ArrayType(
        rwork, {'fun_name': fun_name, 'entity': 'rwork'},
        exp_shape=(_lrwork_py,),
        check_contents=False,
    )
    _rwork = _rwork_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _lwsav = comm['lwsav']
    _iwsav = comm['iwsav']
    _rwsav = comm['rwsav']
    _routine_name = utils._EngineRoutineName(*b'h02bf ')
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        utils._UNUSED_READ_REC,
        _read_rech,
        _infile,
        _maxn,
        _maxm,
        _optim,
        _xbldef,
        _xbudef,
        _maxdpt,
        _msglvl,
        _n,
        _m,
        _x,
        _crname,
        _iwork,
        _liwork,
        _rwork,
        _lrwork,
        _lwsav,
        _iwsav,
        _rwsav,
        _routine_name,
        _errbuf,
        _ierr,
        _optim._length_,
        _crname_chrlen,
        _routine_name._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py < 0:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either maxm and/or maxn are too small',
                    'or the MPSX data file is nonstandard and/or corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Weak LP solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The problem does not have a feasible integer solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The LP solution is unbounded.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The LP does not have a feasible solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Iteration limit reached without finding a solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, not enough real workspace to solve problem:',
                    'lrwork = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lrwork must be at least {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, not enough integer workspace to solve problem:',
                    'liwork = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'liwork must be at least {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxdpt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxdpt >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, not enough real workspace to read data file:',
                    'lrwork = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, not enough integer workspace to read data file:',
                    'liwork = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, optim = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: optim = \'MIN\' or \'MAX\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xbldef = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xbudef = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xbldef <= xbudef.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxm >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, infile = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= infile <= 99.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred.',
                    'Check all function calls and array dimensions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Search of a branch was terminated due to iteration limit.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Increase maxdpt and rerun ilp_mpsx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'maxdpt is too small to solve the problem:',
                    'maxdpt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The IP solution returned is the best solution for',
                    'the number of nodes investigated in the branch and bound tree.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No feasible solution was found for the number of nodes',
                    'investigated in the branch and bound tree.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Not enough workspace to solve problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _crname_cls.refresh_py(_crname)
        _iwork_cls.refresh_py(_iwork)
        _rwork_cls.refresh_py(_rwork)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['n', 'm'],
        )(
            _n_py,
            _m_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def ilp_mpsx_convert(infile, maxn, maxm, optim, xbldef, xbudef, nmobj, nmrhs, nmrng, nmbnd, mpslst, a, bl, bu, cvec, x, intvar, crname, io_manager=None):
    u"""
    Convert MPSX data file defining IP or LP problem to format required
    by ``ilp_dense`` or ``opt.lp_solve``.

    ``ilp_mpsx_convert`` reads data for a linear or integer programming
    problem from an external file which is in standard or compatible
    MPSX input format.

    For full information please refer to the NAG Library document for
    h02bu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02buf.html

    Parameters
    ----------
    infile : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        associated with the MPSX data file.

    maxn : int
        An upper limit for the number of variables in the problem.

    maxm : int
        An upper limit for the number of constraints (including the
        objective) in the problem.

    optim : str, length 3
        Specifies the direction of the optimization. `optim` must be set
        to 'MIN' for minimization and to 'MAX' for maximization.

    xbldef : float
        The default lower bound to be used for the variables in the
        problem when none is specified in the BOUNDS section of the MPSX
        data file. For a standard LP or IP problem `xbldef` would
        normally be set to zero.

    xbudef : float
        The default upper bound to be used for the variables in the
        problem when none is specified in the BOUNDS section of the MPSX
        data file. For a standard LP or IP problem `xbudef` would
        normally be set to 'infinity' (i.e., `xbudef` >= 10^20).

    nmobj : str, length 8
        Either the name of the objective function to be used for the
        optimization, or blank (in which case the first objective (free)
        row in the file is used).

    nmrhs : str, length 8
        Either the name of the RHS set to be used for the optimization,
        or blank (in which case the first RHS set is used).

    nmrng : str, length 8
        Either the name of the RANGE set to be used for the
        optimization, or blank (in which case the first RANGE set (if
        any) is used).

    nmbnd : str, length 8
        Either the name of the BOUNDS set to be used for the
        optimization, or blank (in which case the first BOUNDS set (if
        any) is used).

    mpslst : bool
        If `mpslst` = True, then a listing of the input data is sent to
        the file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`). This can be
        useful for debugging the MPSX data file.

    a : float, shape (`maxm`, `maxn`), modified in place
        On exit: A, the matrix of general linear constraints.

    bl : float, shape (`maxn`+`maxm`), modified in place
        On exit: l, the lower bounds for all the variables and
        constraints in the following order. The first `n` elements of
        `bl` contain the bounds on the variables and the next `m`
        elements contain the bounds for the general linear constraints
        (if any). Note that an 'infinite' lower bound is indicated by
        `bl`[j-1] = -1.0e+20 and an equality constraint by `bl`[j-1] =
        `bu`[j-1].

    bu : float, shape (`maxn`+`maxm`), modified in place
        On exit: u, the upper bounds for all the variables and
        constraints in the following order. The first `n` elements of
        `bu` contain the bounds on the variables and the next `m`
        elements contain the bounds for the general linear constraints
        (if any). Note that an 'infinite' upper bound is indicated by
        `bu`[j-1] = 1.0e+20 and an equality constraint by `bu`[j-1] =
        `bl`[j-1].

    cvec : float, shape (`maxn`), modified in place
        On exit: c, the coefficients of the objective function. The
        signs of these coefficients are determined by the problem
        (either LP or IP) and the direction of the optimization (see
        `optim` above).

    x : float, shape (`maxn`), modified in place
        On exit: an initial estimate of the solution to the problem.
        More precisely, `x`[j-1] = 1.0 if j is odd and 0.0 otherwise,
        for j = 1,2,...,`n`.

    intvar : int, shape (`maxn`), modified in place
        On exit: indicates which are the integer variables in the
        problem. More precisely, `intvar`[k-1] = 1 if x_k is an integer
        variable, and 0 otherwise, for k = 1,2,...,`n`.

    crname : str, length 8, shape (`maxn`+`maxm`), modified in place
        On exit: the MPSX names of all the variables and constraints in
        the problem in the following order. The first `n` elements
        contain the MPSX names for the variables and the next `m`
        elements contain the MPSX names for the general linear
        constraints (if any).

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nmobj : str, length 8
        The name of the objective row as defined in the MPSX data file.

    nmrhs : str, length 8
        The name of the RHS set read in the MPSX data file.

    nmrng : str, length 8
        The name of the RANGE set read in the MPSX data file. This is
        blank if the MPSX data file does not have a RANGE set.

    nmbnd : str, length 8
        The name of the BOUNDS set read in the MPSX data file. This is
        blank if the MPSX data file does not have a BOUNDS set.

    n : int
        n, the actual number of variables in the problem.

    m : int
        m, the actual number of general linear constraints in the
        problem.

    nmprob : str, length 8
        The name of the problem as defined in the MPSX data file.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Too many rows. Limit is *<value>*, but the actual number
            required is *<value>*.

        (`errno` 2)
            Too many columns. Limit is *<value>*, but the actual number
            required is *<value>*.

        (`errno` 3)
            No objective function row found.

        (`errno` 4)
            No rows specified in the ROWS section.

        (`errno` 5)
            Illegal constraint type *<value>* was detected in the ROWS
            section.

        (`errno` 6)
            Row name with leading blank or non-alphanumeric character
            *<value>*.

        (`errno` 7)
            Column name with leading blank or non-alphanumeric character
            *<value>*.

        (`errno` 8)
            Illegal bound type *<value>*.

        (`errno` 9)
            Column name *<value>* is not defined in the COLUMNS section.

        (`errno` 10)
            The last line must be the ENDATA indicator line.

        (`errno` 11)
            Line *<value>* is not a comment nor a valid line.

        (`errno` 12)
            Row name *<value>* is not defined in the ROWS section.

        (`errno` 13)
            No columns specified in the COLUMNS section.

        (`errno` 14)
            On entry, `optim` = *<value>*.

            Constraint: `optim` = 'MIN' or 'MAX'.

        (`errno` 14)
            On entry, `xbldef` = *<value>* and `xbudef` = *<value>*.

            Constraint: `xbldef` <= `xbudef`.

        (`errno` 14)
            On entry, `maxm` = *<value>*.

            Constraint: `maxm` >= 1.

        (`errno` 14)
            On entry, `maxn` = *<value>*.

            Constraint: `maxn` >= 1.

        (`errno` 14)
            On entry, `infile` = *<value>*.

            Constraint: 0 <= `infile` <= 99.

        (`errno` 15)
            Integer marker is not in a correct position.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.ilp_mpsx_convert'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_read_rech_py(_, en_data, nin, recin, errbuf, ifail, recin_len, errbuf_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_read_rec(
                en_data, utils._EngineIntScalarType.to_py(nin.contents), recin, errbuf, recin_len, io_manager, excs, ifail.contents,
            )
        except:
            utils._read_err(
                _sys.exc_info(),
                excs,
                utils.FileObjManager._read_err,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h02buft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._READ_REC_FUNCTYPE,
        utils._READ_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, infile.
        utils._BYREF_INT, # Input, maxn.
        utils._BYREF_INT, # Input, maxm.
        utils._BYREF_STRING, # Input, optim(len=3).
        utils._BYREF_FLOAT64, # Input, xbldef.
        utils._BYREF_FLOAT64, # Input, xbudef.
        utils._BYREF_STRING, # Input/Output, nmobj(len=8).
        utils._BYREF_STRING, # Input/Output, nmrhs(len=8).
        utils._BYREF_STRING, # Input/Output, nmrng(len=8).
        utils._BYREF_STRING, # Input/Output, nmbnd(len=8).
        utils._BYREF_LOGICAL, # Input, mpslst.
        utils._BYREF_INT, # Output, n.
        utils._BYREF_INT, # Output, m.
        utils._EngineFloat64ArrayType.c_type, # Output, a[maxm, maxn].
        utils._EngineFloat64ArrayType.c_type, # Output, bl[maxn+maxm].
        utils._EngineFloat64ArrayType.c_type, # Output, bu[maxn+maxm].
        utils._EngineFloat64ArrayType.c_type, # Output, cvec[maxn].
        utils._EngineFloat64ArrayType.c_type, # Output, x[maxn].
        utils._EngineIntArrayType.c_type, # Output, intvar[maxn].
        utils._EngineCharArrayType.c_type, # Output, crname(len=8)[maxn+maxm].
        utils._BYREF_STRING, # Output, nmprob(len=8).
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optim.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of nmobj.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of nmrhs.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of nmrng.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of nmbnd.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of crname.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of nmprob.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _read_rech = utils._READ_RECH_FUNCTYPE(cb_read_rech_py)
    _infile = utils._EngineIntScalarType(
        infile, {'fun_name': fun_name, 'entity': 'infile'},
    ).to_ctypes()
    _maxn = utils._EngineIntScalarType(
        maxn, {'fun_name': fun_name, 'entity': 'maxn'},
    ).to_ctypes()
    _maxm = utils._EngineIntScalarType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
    ).to_ctypes()
    _optim = utils._EngineCharScalarType(
        optim, {'fun_name': fun_name, 'entity': 'optim'}, exp_chrlen=3,
    ).to_ctypes()
    _xbldef = utils._EngineFloat64ScalarType(
        xbldef, {'fun_name': fun_name, 'entity': 'xbldef'},
    ).to_ctypes()
    _xbudef = utils._EngineFloat64ScalarType(
        xbudef, {'fun_name': fun_name, 'entity': 'xbudef'},
    ).to_ctypes()
    _nmobj = utils._EngineCharScalarType(
        nmobj, {'fun_name': fun_name, 'entity': 'nmobj'}, exp_chrlen=8,
    ).to_ctypes()
    _nmrhs = utils._EngineCharScalarType(
        nmrhs, {'fun_name': fun_name, 'entity': 'nmrhs'}, exp_chrlen=8,
    ).to_ctypes()
    _nmrng = utils._EngineCharScalarType(
        nmrng, {'fun_name': fun_name, 'entity': 'nmrng'}, exp_chrlen=8,
    ).to_ctypes()
    _nmbnd = utils._EngineCharScalarType(
        nmbnd, {'fun_name': fun_name, 'entity': 'nmbnd'}, exp_chrlen=8,
    ).to_ctypes()
    _mpslst = utils._EngineLogicalScalarType(
        mpslst, {'fun_name': fun_name, 'entity': 'mpslst'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType.empty_ctypes()
    _m = utils._EngineIntScalarType.empty_ctypes()
    _maxm_py = utils._EngineIntScalarType.to_py(_maxm)
    _maxn_py = utils._EngineIntScalarType.to_py(_maxn)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_maxm_py, _maxn_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _bl_cls = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_maxn_py+_maxm_py,),
        check_contents=False,
    )
    _bl = _bl_cls.to_ctypes(can_reference_contents=False)
    _bu_cls = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_maxn_py+_maxm_py,),
        check_contents=False,
    )
    _bu = _bu_cls.to_ctypes(can_reference_contents=False)
    _cvec_cls = utils._EngineFloat64ArrayType(
        cvec, {'fun_name': fun_name, 'entity': 'cvec'},
        exp_shape=(_maxn_py,),
        check_contents=False,
    )
    _cvec = _cvec_cls.to_ctypes(can_reference_contents=False)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_maxn_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _intvar_cls = utils._EngineIntArrayType(
        intvar, {'fun_name': fun_name, 'entity': 'intvar'},
        exp_shape=(_maxn_py,),
        check_contents=False,
    )
    _intvar = _intvar_cls.to_ctypes(can_reference_contents=False)
    _crname_cls = utils._EngineCharArrayType(
        crname, {'fun_name': fun_name, 'entity': 'crname'}, exp_chrlen=8,
        exp_shape=(_maxn_py+_maxm_py,),
        check_contents=False,
    )
    _crname = _crname_cls.to_ctypes(can_reference_contents=False)
    _crname_chrlen = _crname_cls.chrlen
    _nmprob = utils._EngineCharScalarType.empty_ctypes(chrlen=8)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        utils._UNUSED_READ_REC,
        _read_rech,
        _infile,
        _maxn,
        _maxm,
        _optim,
        _xbldef,
        _xbudef,
        _nmobj,
        _nmrhs,
        _nmrng,
        _nmbnd,
        _mpslst,
        _n,
        _m,
        _a,
        _bl,
        _bu,
        _cvec,
        _x,
        _intvar,
        _crname,
        _nmprob,
        _iwork,
        _errbuf,
        _ierr,
        _optim._length_,
        _nmobj._length_,
        _nmrhs._length_,
        _nmrng._length_,
        _nmbnd._length_,
        _crname_chrlen,
        _nmprob._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Too many rows. Limit is {:s}, but the actual number'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'required is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Too many columns. Limit is {:s}, but the actual number'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'required is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No objective function row found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No rows specified in the ROWS section.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Illegal constraint type \'{:s}\' was detected in the ROWS section.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Row name with leading blank or non-alphanumeric character',
                    '\'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Column name with leading blank or non-alphanumeric character',
                    '\'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Illegal bound type \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Column name \'{:s}\' is not defined in the COLUMNS section.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The last line must be the ENDATA indicator line.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Line {:s} is not a comment nor a valid line.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Row name \'{:s}\' is not defined in the ROWS section.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No columns specified in the COLUMNS section.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, optim = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: optim = \'MIN\' or \'MAX\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xbldef = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xbudef = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xbldef <= xbudef.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxm >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, infile = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= infile <= 99.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 15 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Integer marker is not in a correct position.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nmobj_py = utils._EngineCharScalarType.to_py(_nmobj)
        _nmrhs_py = utils._EngineCharScalarType.to_py(_nmrhs)
        _nmrng_py = utils._EngineCharScalarType.to_py(_nmrng)
        _nmbnd_py = utils._EngineCharScalarType.to_py(_nmbnd)
        _a_cls.refresh_py(_a)
        _bl_cls.refresh_py(_bl)
        _bu_cls.refresh_py(_bu)
        _cvec_cls.refresh_py(_cvec)
        _x_cls.refresh_py(_x)
        _intvar_cls.refresh_py(_intvar)
        _crname_cls.refresh_py(_crname)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        _nmprob_py = utils._EngineCharScalarType.to_py(_nmprob)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nmobj', 'nmrhs', 'nmrng', 'nmbnd', 'n', 'm', 'nmprob'],
        )(
            _nmobj_py,
            _nmrhs_py,
            _nmrng_py,
            _nmbnd_py,
            _n_py,
            _m_py,
            _nmprob_py,
        )
    _err_class.manage()
    return return_data

def ilp_print(n, m, a, bl, bu, x, clamda, istate, crname, comm, io_manager=None):
    u"""
    Print IP or LP solutions with user-specified names for rows and
    columns.

    ``ilp_print`` prints the solution to a linear or integer programming
    problem computed by ``opt.lp_solve`` or ``ilp_dense`` and
    ``ilp_info``, with user-supplied names for the rows and columns.

    For full information please refer to the NAG Library document for
    h02bv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02bvf.html

    Parameters
    ----------
    n : int
        The number of variables, as returned by ``ilp_mpsx_convert``.

    m : int
        The number of general linear constraints, as returned by
        ``ilp_mpsx_convert``.

    a : float, shape (`m`, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `m`>0: `n`; if `m`=0: 1; otherwise: 0.

        The matrix of general linear constraints, as returned by
        ``ilp_mpsx_convert``.

    bl : float, shape (`n`+`m`)
        The lower bounds for all the constraints, as returned by
        ``opt.lp_solve`` or ``ilp_info``.

    bu : float, shape (`n`+`m`)
        The upper bounds for all the constraints, as returned by
        ``opt.lp_solve`` or ``ilp_info``.

    x : float, shape (`n`)
        The solution to the problem, as returned by ``opt.lp_solve`` or
        ``ilp_dense``.

    clamda : float, shape (`n`+`m`)
        The Lagrange-multipliers (reduced costs) for each constraint
        with respect to the working set, as returned by ``opt.lp_solve``
        or ``ilp_info``.

    istate : int, shape (`n`+`m`)
        The status of every constraint in the working set at the
        solution, as returned by ``opt.lp_solve`` or ``ilp_info``.

    crname : str, length 8, shape (`n`+`m`)
        The user-defined names for all the variables and constraints, as
        returned by ``ilp_mpsx_convert``.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.lp_solve'.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.ilp_print'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h02bvft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, bl[n+m].
        utils._EngineFloat64ArrayType.c_type, # Input, bu[n+m].
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, clamda[n+m].
        utils._EngineIntArrayType.c_type, # Input, istate[n+m].
        utils._EngineCharArrayType.c_type, # Input, crname(len=8)[n+m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[475].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of crname.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _m_py > 0:
        _a_dim2 = _n_py
    elif _m_py == 0:
        _a_dim2 = 1
    else:
        _a_dim2 = 0
    exp_shape_a = (_m_py, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _bl = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_n_py+_m_py,),
    ).to_ctypes()
    _bu = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_n_py+_m_py,),
    ).to_ctypes()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _clamda = utils._EngineFloat64ArrayType(
        clamda, {'fun_name': fun_name, 'entity': 'clamda'},
        exp_shape=(_n_py+_m_py,),
    ).to_ctypes()
    _istate = utils._EngineIntArrayType(
        istate, {'fun_name': fun_name, 'entity': 'istate'},
        exp_shape=(_n_py+_m_py,),
    ).to_ctypes()
    _crname_cls = utils._EngineCharArrayType(
        crname, {'fun_name': fun_name, 'entity': 'crname'}, exp_chrlen=8,
        exp_shape=(_n_py+_m_py,),
    )
    _crname = _crname_cls.to_ctypes()
    _crname_chrlen = _crname_cls.chrlen
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('rwsav',),
    )
    _rwsav = comm['rwsav']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _n,
        _m,
        _a,
        _lda,
        _bl,
        _bu,
        _x,
        _clamda,
        _istate,
        _crname,
        _rwsav,
        _errbuf,
        _ierr,
        _crname_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def ilp_info(n, m, bl, bu, clamda, istate, comm):
    u"""
    Integer programming solution, extracts further information on
    solution obtained by ``ilp_dense``.

    ``ilp_info`` extracts more information associated with the solution
    of an integer programming problem computed by ``ilp_dense``.

    For full information please refer to the NAG Library document for
    h02bz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02bzf.html

    Parameters
    ----------
    n : int
        This must be the same argument `n` as supplied to ``ilp_dense``.

    m : int
        This must be the same argument `m` as supplied to ``ilp_dense``.

    bl : float, shape (`n`+`m`), modified in place
        On exit: if ``ilp_dense`` exits with the function exits
        successfully or `errno` = 7 or 9, the values in the array `bl`
        contain the lower bounds imposed on the integer solution for all
        the constraints. The first `n` elements contain the lower bounds
        on the variables, and the next `m` elements contain the lower
        bounds for the general linear constraints (if any).

    bu : float, shape (`n`+`m`), modified in place
        On exit: if ``ilp_dense`` exits with the function exits
        successfully or `errno` = 7 or 9, the values in the array `bu`
        contain the upper bounds imposed on the integer solution for all
        the constraints. The first `n` elements contain the upper bounds
        on the variables, and the next `m` elements contain the upper
        bounds for the general linear constraints (if any).

    clamda : float, shape (`n`+`m`), modified in place
        On exit: if ``ilp_dense`` exits with the function exits
        successfully or `errno` = 7 or 9, the values in the array
        `clamda` contain the values of the Lagrange-multipliers for each
        constraint with respect to the current working set. The first
        `n` elements contain the multipliers (reduced costs) for the
        bound constraints on the variables, and the next `m` elements
        contain the multipliers (shadow costs) for the general linear
        constraints (if any).

    istate : int, shape (`n`+`m`), modified in place
        On exit: if ``ilp_dense`` exits with the function exits
        successfully or `errno` = 7 or 9, the values in the array
        `istate` indicate the status of the constraints in the working
        set at an integer solution. Otherwise, `istate` indicates the
        composition of the working set at the final iterate. The
        significance of each possible value of `istate`[j] is as
        follows.

        [table omitted]

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ilp_dense``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.ilp_info'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h02bzft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Output, bl[n+m].
        utils._EngineFloat64ArrayType.c_type, # Output, bu[n+m].
        utils._EngineFloat64ArrayType.c_type, # Output, clamda[n+m].
        utils._EngineIntArrayType.c_type, # Output, istate[n+m].
        utils._EngineIntArrayType.c_type, # Input, iwork[liwork].
        utils._BYREF_INT, # Input, liwork.
        utils._EngineFloat64ArrayType.c_type, # Input, rwork[lrwork].
        utils._BYREF_INT, # Input, lrwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _bl_cls = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_n_py+_m_py,),
        check_contents=False,
    )
    _bl = _bl_cls.to_ctypes(can_reference_contents=False)
    _bu_cls = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_n_py+_m_py,),
        check_contents=False,
    )
    _bu = _bu_cls.to_ctypes(can_reference_contents=False)
    _clamda_cls = utils._EngineFloat64ArrayType(
        clamda, {'fun_name': fun_name, 'entity': 'clamda'},
        exp_shape=(_n_py+_m_py,),
        check_contents=False,
    )
    _clamda = _clamda_cls.to_ctypes(can_reference_contents=False)
    _istate_cls = utils._EngineIntArrayType(
        istate, {'fun_name': fun_name, 'entity': 'istate'},
        exp_shape=(_n_py+_m_py,),
        check_contents=False,
    )
    _istate = _istate_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iwork', 'rwork'),
    )
    _liwork_py = utils._EngineIntArrayType.get_shape(comm['iwork'], {'fun_name': fun_name, 'entity': 'comm[\'iwork\']'}, exp_rank=1, validate=False)[0]
    _liwork = utils._EngineIntScalarType(
        _liwork_py, {'fun_name': fun_name, 'entity': '_liwork'},
    ).to_ctypes()
    _iwork = comm['iwork']
    _lrwork_py = utils._EngineFloat64ArrayType.get_shape(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, exp_rank=1, validate=False)[0]
    _lrwork = utils._EngineIntScalarType(
        _lrwork_py, {'fun_name': fun_name, 'entity': '_lrwork'},
    ).to_ctypes()
    _rwork = comm['rwork']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _bl,
        _bu,
        _clamda,
        _istate,
        _iwork,
        _liwork,
        _rwork,
        _lrwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _bl_cls.refresh_py(_bl)
        _bu_cls.refresh_py(_bu)
        _clamda_cls.refresh_py(_clamda)
        _istate_cls.refresh_py(_istate)
        return_data = None
    _err_class.manage()
    return return_data

def iqp_dense(n, nclin, a, bl, bu, cvec, h, qphess, intvar, lintvr, mdepth, istate, xs, ax, clamda, strtgy, monit, comm, data=None, io_manager=None):
    u"""
    Integer QP problem (dense).

    Note: this function uses optional algorithmic parameters, see also:
    ``iqp_dense_optfile``, ``iqp_dense_optstr``.

    ``iqp_dense`` solves general quadratic programming problems with
    integer constraints on the variables.
    It is not intended for large sparse problems.

    For full information please refer to the NAG Library document for
    h02cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02cbf.html

    Parameters
    ----------
    n : int
        n, the number of variables.

    nclin : int
        m_L, the number of general linear constraints.

    a : float, shape (`nclin`, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `nclin`>0: `n`; if `nclin`=0: 1;
        otherwise: 0.

        The ith row of `a` must contain the coefficients of the ith
        general linear constraint, for i = 1,2,...,m_L.

        If `nclin` = 0, the array `a` is not referenced.

    bl : float, shape (`n`+`nclin`), modified in place
        `bl` must contain the lower bounds and `bu` the upper bounds,
        for all the constraints in the following order. The first n
        elements of each array must contain the bounds on the variables,
        and the next m_L elements the bounds for the general linear
        constraints (if any). To specify a nonexistent lower bound
        (i.e., l_j = - infinity), set `bl`[j] <= -bigbnd, and to specify
        a nonexistent upper bound (i.e., u_j = + infinity), set `bu`[j]
        >= bigbnd; the default value of bigbnd is 10^20, but this may be
        changed by the 'Infinite Bound Size'. To specify the jth
        constraint as an equality, set `bl`[j] = `bu`[j] = beta, say,
        where |beta| < bigbnd.

    bu : float, shape (`n`+`nclin`), modified in place
        `bl` must contain the lower bounds and `bu` the upper bounds,
        for all the constraints in the following order. The first n
        elements of each array must contain the bounds on the variables,
        and the next m_L elements the bounds for the general linear
        constraints (if any). To specify a nonexistent lower bound
        (i.e., l_j = - infinity), set `bl`[j] <= -bigbnd, and to specify
        a nonexistent upper bound (i.e., u_j = + infinity), set `bu`[j]
        >= bigbnd; the default value of bigbnd is 10^20, but this may be
        changed by the 'Infinite Bound Size'. To specify the jth
        constraint as an equality, set `bl`[j] = `bu`[j] = beta, say,
        where |beta| < bigbnd.

    cvec : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `comm`['iwsav'][364]<1 or `comm`['iwsav'][364] in
        (2, 4, 6): `n`; otherwise: 1.

        The coefficients of the explicit linear term of the objective
        function when the problem is of type LP, QP2 (the default) and
        QP4.

        If the problem is of type FP, QP1, or QP3, `cvec` is not
        referenced.

    h : float, shape (ldh, tdh)
        May be used to store the quadratic term H of the QP objective
        function if desired. In some cases, you need not use `h` to
        store H explicitly (see the specification of `qphess`). The
        elements of `h` are referenced only by `qphess`. The number of
        rows of H is denoted by m, whose default value is n. (The
        'Hessian Rows' may be used to specify a value of m < n.)

        If the default version of `qphess` is used and the problem is of
        type QP1 or QP2 (the default), the first m rows and columns of
        `h` must contain the leading m by m rows and columns of the
        symmetric Hessian matrix H.

        Only the diagonal and upper triangular elements of the leading m
        rows and columns of `h` are referenced.

        The remaining elements need not be assigned.

        If the default version of `qphess` is used and the problem is of
        type QP3 or QP4, the first m rows of `h` must contain an m by n
        upper trapezoidal factor of the symmetric Hessian matrix H^TH.

        The factor need not be of full rank, i.e., some of the diagonal
        elements may be zero.

        However, as a general rule, the larger the dimension of the
        leading nonsingular sub-matrix of `h`, the fewer iterations will
        be required.

        Elements outside the upper trapezoidal part of the first m rows
        of `h` need not be assigned.

        In other situations, it may be desirable to compute Hx or H^THx
        without accessing `h` -- for example, if H or H^TH is sparse or
        has special structure.

        The arguments `h` and ldh may then refer to any convenient
        array.

        If the problem is of type FP or LP, `h` is not referenced.

    qphess : None or callable qphess(n, jthcol, h, x, hx, data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        In general, you need not provide a version of `qphess`, because
        supplying None will cause a 'default' function to be used.

        However, the algorithm of ``iqp_dense`` requires only the
        product of H or H^TH and a vector x; and in some cases you may
        obtain increased efficiency by providing a version of `qphess`
        that avoids the need to define the elements of the matrices H or
        H^TH explicitly. `qphess` is not referenced if the problem is of
        type FP or LP, in which case `qphess` may be None.

        Parameters
        ~~~~~~~~~~
        n : int
            This is the same argument n as supplied to ``iqp_dense``.

        jthcol : int
            Specifies whether or not the vector x is a column of the
            identity matrix.

            `jthcol` = j > 0

              The vector x is the jth column of the identity matrix, and
              hence Hx or `h`^THx is the jth column of `h` or `h`^TH,
              respectively, which may in some cases require very little
              computation and `qphess` may be coded to take advantage of
              this. However special code is not necessary because x is
              always stored explicitly in the array `x`.

            `jthcol` = 0

              x has no special form.

        h : float, ndarray, shape (ldh, tdh)
            This is the same argument h as supplied to ``iqp_dense``.

        x : float, ndarray, shape (`n`)
            The vector x.

        hx : float, ndarray, shape (`n`), to be modified in place
            On exit: the product Hx if the problem is of type QP1 or QP2
            (the default), or the product `h`^THx if the problem is of
            type QP3 or QP4.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    intvar : int, shape (`lintvr`)
        `intvar`[i] must contain the index of the solution vector x
        which is required to be integer. For example, if x_1 and x_3 are
        constrained to take integer values then `intvar`[0] might be set
        to 1 and `intvar`[1] to 3. The order in which the indices are
        specified is important, since this determines the order in which
        the sub-problems are generated. As a rule-of-thumb, the
        important variables should always be specified first. Thus, in
        the above example, if x_3 relates to a more important quantity
        than x_1, then it might be advantageous to set `intvar`[0] = 3
        and `intvar`[1] = 1. If k is the smallest integer such that
        `intvar`[k] is less than or equal to zero then ``iqp_dense``
        assumes that k-1 variables are constrained to be integer;
        components `intvar`[k+1], ..., `intvar`[`lintvr`-1] are not
        referenced.

    lintvr : int
        Often `lintvr` is the number of variables that are constrained
        to be integer.

    mdepth : int
        The maximum depth that ``iqp_dense`` may insert before admitting
        failure.

    istate : int, shape (`n`+`nclin`), modified in place
        On entry: need not be set if the (default) option 'Cold Start'
        is used.

        If the option 'Warm Start' has been chosen, `istate` specifies
        the desired status of the constraints at the start of the
        feasibility phase.

        More precisely, the first n elements of `istate` refer to the
        upper and lower bounds on the variables, and the next m_L
        elements refer to the general linear constraints (if any).

        Possible values for `istate`[j] are as follows:

        [table omitted]

        The values -2, -1 and 4 are also acceptable but will be reset to
        zero by the function.

        If ``iqp_dense`` has been called previously with the same values
        of `n` and `nclin`, `istate` already contains satisfactory
        information. (See also the description of the option 'Warm
        Start'.) The function also adjusts (if necessary) the values
        supplied in `xs` to be consistent with `istate`.

        On exit: the status of the constraints in the working set at the
        point returned in `xs`. The significance of each possible value
        of `istate`[j] is as follows:

        [table omitted]

    xs : float, shape (`n`), modified in place
        On entry: an initial estimate of the solution.

        On exit: the point at which ``iqp_dense`` terminated. If the
        function exits successfully or `errno` = 1 or 3, `xs` contains
        an estimate of the solution.

    ax : float, shape (max(1, `nclin`)), modified in place
        On exit: the final values of the linear constraints Ax.

        If `nclin` = 0, `ax` is not referenced.

    clamda : float, shape (`n`+`nclin`), modified in place
        On exit: the values of the Lagrange-multipliers for each
        constraint with respect to the current working set. The first n
        elements contain the multipliers for the bound constraints on
        the variables, and the next m_L elements contain the multipliers
        for the general linear constraints (if any). If `istate`[j] = 0
        (i.e., constraint j is not in the working set), `clamda`[j] is
        zero. If x is optimal, `clamda`[j] should be non-negative if
        `istate`[j] = 1, non-positive if `istate`[j] = 2 and zero if
        `istate`[j] = 4.

    strtgy : int
        Determines a branching strategy to be used throughout the
        computation, as follows:

        [table omitted]

    monit : None or callable (bstval, halt, count) = monit(intfnd,
        nodes, depth, obj, x, bstval, bstsol, bl, bu, n, halt, count,
        data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` may be used to print out intermediate output and to
        affect the course of the computation.

        Specifically, it allows you to specify a realistic value for the
        cut-off value (see [Description]) and to terminate the
        algorithm.

        If you do not require any intermediate output, have no estimate
        of the cut-off value and require an exhaustive tree search then
        `monit` may be None.

        Parameters
        ~~~~~~~~~~
        intfnd : int
            Specifies the number of integer solutions obtained so far.

        nodes : int
            Specifies the number of nodes (sub-problems) solved so far.

        depth : int
            Specifies the depth in the tree of sub-problems the
            algorithm has now reached.

        obj : float
            Specifies the value of the objective function of the end of
            the latest sub-problem.

        x : float, ndarray, shape (`n`)
            Specifies the values of the independent variables at the end
            of the latest sub-problem.

        bstval : float
            Normally specifies the value of the best integer solution
            found so far.

        bstsol : float, ndarray, shape (`n`)
            Specifies the solution vector which gives rise to the best
            integer solution value so far discovered.

        bl : float, ndarray, shape (`n`)
            `bl`[i] specifies the current lower bounds on the variable
            x_i.

        bu : float, ndarray, shape (`n`)
            `bu`[i] specifies the current upper bounds on the variable
            x_i.

        n : int
            Specifies the number of variables.

        halt : bool
            Will have the value False.

        count : int
            Unchanged from previous call.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        bstval : float
            May be set to a cut-off value if you are an experienced user
            as follows. Before an integer solution has been found
            `bstval` will be set by ``iqp_dense`` to the largest machine
            representable number (see ``machine.real_largest``). If you
            know that the solution being sought is a much smaller
            number, then `bstval` may be set to this number as a cut-off
            value (see [Description]). Beware of setting `bstval` too
            small, since then no integer solutions will be discovered.
            Also, on entry to `monit`, `bstval` should be set using a
            statement of the form

            ::

              if intfnd == 0: bstval = cut_off_value

            Setting `bstval` in this way will not prevent the normal
            operation of the algorithm when subsequent integer solutions
            are found. It would be a grievous mistake to unconditionally
            set `bstval` and if you have any doubts whatsoever about the
            correct use of this argument then you are strongly
            recommended to leave it unchanged.

        halt : bool
            If `halt` is set to True, ``opt.qp_dense_solve`` will be
            brought to a halt with `errno` = -1. This facility may be
            useful if you are content with any integer solution, or with
            any integer solution that fits certain criteria. Under these
            circumstances setting `halt` = True can save considerable
            unnecessary computation.

        count : int
            May be used by you to save the last value of `intfnd`. If a
            subsequent call of `monit` has a value of `intfnd` which is
            greater than `count`, then you know that a new integer
            solution has been found at this node.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.qp_dense_solve'.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    obj : float
        The value of the objective function at x if x is feasible, or
        the sum of infeasibilities at x otherwise. If the problem is of
        type FP and x is feasible, `obj` is set to zero.

    Other Parameters
    ----------------
    'Check Frequency' : int
        Default = 50

        Every ith iteration, a numerical test is made to see if the
        current solution x satisfies the constraints in the working set.
        If the largest residual of the constraints in the working set is
        judged to be too large, the current working set is refactorized
        and the variables are recomputed to satisfy the constraints more
        accurately.
        If i <= 0, the default value is used.

    'Cold Start' : valueless
        Default

        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``iqp_dense`` chooses the initial working
        set based on the values of the variables and constraints at the
        initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate` (see [Parameters] for the
        definition of this array). ``iqp_dense`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``iqp_dense`` is called repeatedly to solve related problems.

    'Warm Start' : valueless
        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``iqp_dense`` chooses the initial working
        set based on the values of the variables and constraints at the
        initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate` (see [Parameters] for the
        definition of this array). ``iqp_dense`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``iqp_dense`` is called repeatedly to solve related problems.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used in conjunction with the option 'Cold Start'
        (the default value) when ``iqp_dense`` selects an initial
        working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Expand Frequency' : int
        Default = 5

        This option is part of an anti-cycling procedure designed to
        guarantee progress even on highly degenerate problems.

        The strategy is to force a positive step at every iteration, at
        the expense of violating the constraints by a small amount.
        Suppose that the value of the option 'Feasibility Tolerance' is
        delta.
        Over a period of i iterations, the feasibility tolerance
        actually used by ``iqp_dense`` increases from 0.5 delta to delta
        (in steps of 0.5 delta/i).

        At certain stages the following 'resetting procedure' is used to
        remove constraint infeasibilities.
        First, all variables whose upper or lower bounds are in the
        working set are moved exactly onto their bounds.
        A count is kept of the number of nontrivial adjustments made.
        If the count is positive, iterative refinement is used to give
        variables that satisfy the working set to (essentially) machine
        precision.
        Finally, the working feasibility tolerance is reinitialized to
        0.5 delta.

        If a problem requires more than i iterations, the resetting
        procedure is invoked and a new cycle of i iterations is started
        with i incremented by 10. (The decision to resume the
        feasibility phase or optimality phase is based on comparing any
        constraint infeasibilities with delta.)

        The resetting procedure is also invoked when ``iqp_dense``
        reaches an apparently optimal, infeasible or unbounded solution,
        unless this situation has already occurred twice.
        If any nontrivial adjustments are made, iterations are
        continued.

        If i <= 0, the default value is used.
        If i >= 9999999, no anti-cycling procedure is invoked.

    'Feasibility Phase Iteration Limit' : int
        Default = max(50, 5*(n+m_L))

        The scalars i_1 and i_2 specify the maximum number of iterations
        allowed in the feasibility and optimality phases. 'Optimality
        Phase Iteration Limit' is equivalent to 'Iteration Limit'.
        Setting i_1 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.
        If i_1 < 0 or i_2 < 0, the default value is used.

    'Optimality Phase Iteration Limit' : int
        Default = max(50, 5*(n+m_L))

        The scalars i_1 and i_2 specify the maximum number of iterations
        allowed in the feasibility and optimality phases. 'Optimality
        Phase Iteration Limit' is equivalent to 'Iteration Limit'.
        Setting i_1 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.
        If i_1 < 0 or i_2 < 0, the default value is used.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        If r >= epsilon, r defines the maximum acceptable absolute
        violation in each constraint at a 'feasible' point.
        For example, if the variables and the coefficients in the
        general constraints are of order unity, and the latter are
        correct to about 6 decimal digits, it would be appropriate to
        specify r as 10^(-6).
        If 0 <= r < epsilon, the default value is used.

        ``iqp_dense`` attempts to find a feasible solution before
        optimizing the objective function.
        If the sum of infeasibilities cannot be reduced to zero, the
        'Minimum Sum of Infeasibilities' can be used to find the minimum
        value of the sum.
        Let Sinf be the corresponding sum of infeasibilities.
        If Sinf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Note that a 'feasible solution' is a solution that satisfies the
        current constraints to within the tolerance r.

    'Hessian Rows' : int
        Default = n

        Note that this option does not apply to problems of type FP or
        LP.

        This specifies m, the number of rows of the Hessian matrix H.
        The default value of m is n, the number of variables of the
        problem.

        If the problem is of type QP, m will usually be n, the number of
        variables.
        However, a value of m less than n is appropriate for QP3 or QP4
        if `h` is an upper trapezoidal matrix with m rows.
        Similarly, m may be used to define the dimension of a leading
        block of nonzeros in the Hessian matrices of QP1 or QP2, in
        which case the last n-m rows and columns of `h` are assumed to
        be zero.
        In the QP case, m should not be greater than n; if it is, the
        last m-n rows of `h` are ignored.

        If i < 0 or i > n, the default value is used.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r <= 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that will be considered a step to an unbounded solution. (Note
        that an unbounded solution can occur only when the Hessian is
        not positive definite.) If the change in x during an iteration
        would exceed the value of r, the objective function is
        considered to be unbounded below in the feasible region.
        If r <= 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 5*(n+m_L))

        See option 'Feasibility Phase Iteration Limit'.

    'Iters' : int
        Default = max(50, 5*(n+m_L))

        See option 'Feasibility Phase Iteration Limit'.

    'Itns' : int
        Default = max(50, 5*(n+m_L))

        See option 'Feasibility Phase Iteration Limit'.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Maximum Degrees of Freedom' : int
        Default = n

        Note that this option does not apply to problems of type FP or
        LP.

        This places a limit on the storage allocated for the triangular
        factor R of the reduced Hessian H_R.
        Ideally, i should be set slightly larger than the value of n_R
        expected at the solution.
        It need not be larger than m_`n`+1, where m_`n` is the number of
        variables that appear nonlinearly in the quadratic objective
        function.
        For many problems it can be much smaller than m_`n`.

        For quadratic problems, a minimizer may lie on any number of
        constraints, so that n_R may vary between 1 and n.
        The default value of i is therefore the number of variables n.
        If 'Hessian Rows' m is specified, the default value of i is the
        same number, m.

    'Minimum Sum of Infeasibilities' : str
        Default = 'NO'

        If no feasible point exists for the constraints, this option is
        used to control whether or not ``iqp_dense`` will calculate a
        point that minimizes the constraint violations.
        If 'Minimum Sum of Infeasibilities' = '[N]O', ``iqp_dense`` will
        terminate as soon as it is evident that no feasible point exists
        for the constraints.
        The final point will generally not be the point at which the sum
        of infeasibilities is minimized.
        If 'Minimum Sum of Infeasibilities' = '[Y]ES', ``iqp_dense``
        will continue until the sum of infeasibilities is minimized.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Print Level' >= 5, monitoring information
        produced by ``iqp_dense`` at every iteration is sent to a file
        with logical unit number i.
        If i < 0 and/or 'Print Level' < 5, no monitoring information is
        produced.

    'Optimality Tolerance' : float
        Default = epsilon^0.8

        If r >= epsilon, r defines the tolerance used to determine if
        the bounds and general constraints have the right 'sign' for the
        solution to be judged to be optimal.

        If 0 <= r < epsilon, the default value is used.

    'Print Level' : int
        Default = 10

        The value of i controls the amount of printout produced by
        ``iqp_dense``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each iteration and the
        final solution) and [Monitoring Information] (monitoring
        information at each iteration).
        If i < 0, the default value is used.

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the 'Monitoring File':

        [table omitted]

        If 'Print Level' >= 5 and the unit number defined by 'Monitoring
        File' is the advisory unit number, then the summary output is
        suppressed.

    'Problem Type' : str
        Default = QP2

        This option specifies the type of objective function to be
        minimized during the optimality phase.
        The following are the five optional keywords and the dimensions
        of the arrays that must be specified in order to define the
        objective function:

        [table omitted]

        For problems of type FP, the objective function is omitted and
        neither `h` nor `cvec` are referenced.

        The following keywords are also acceptable.

        [table omitted]

        In addition, the keyword QP is equivalent to the default option
        QP2.

        If `h` = 0, i.e., the objective function is purely linear, the
        efficiency of ``iqp_dense`` may be increased by specifying a as
        LP.

    'Rank Tolerance' : float
        Default = 100 epsilon

        Note that this option does not apply to problems of type FP or
        LP.

        This argument enables you to control the condition number of the
        triangular factor R (see [Algorithmic Details]).
        If rho_i denotes the function rho_i =
        max{|R_11|,|R_22|,...,|R_{ii}|}, the dimension of R is defined
        to be smallest index i such that |R_{i+1,i+1}| <=
        sqrt(r)|rho_{i+1}|.
        If r <= 0, the default value is used.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Invalid input argument for basic problem.

        (`errno` 1)
            On entry, `strtgy` = *<value>*.

            Constraint: `strtgy` = 0, 1, 2 or 3.

        (`errno` 1)
            On entry, `mdepth` = *<value>*.

            Constraint: `mdepth` >= 1.

        (`errno` 1)
            On entry, `lintvr` = *<value>*.

            Constraint: `lintvr` >= 1.

        (`errno` 1)
            On entry, `nclin` = *<value>*.

            Constraint: `nclin` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            No integer solutions found.

        (`errno` 3)
            Need to increase depth, `mdepth` = *<value>*.

        (`errno` 4)
            Basic problem (without integer constraints) unbounded.

        (`errno` 5)
            Basic problem is infeasible.

        (`errno` 6)
            Basic problem requires too many iterations.

        (`errno` 7)
            Basic problem has a reduced Hessian exceeding assigned
            dimensions.

        (`errno` 9)
            Error in designated problem type.

        (`errno` 12)
            Unexpected error in NAG function -- contact NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` -1)
            Halted at your request.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.iqp_dense'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_qphess_name = fun_name + ':qphess'

    qphessh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, jthcol.
        utils._EngineFloat64ArrayType.c_type, # Input, h[ldh, tdh].
        utils._BYREF_INT, # Input, ldh.
        utils._BYREF_INT, # Input, tdh.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, hx[n].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[610].
    )

    if qphess is None:
        qphess_auxil_ctypes = utils._EngineState._get_symbol('e04nfut')
        qphess_auxil_ctypes.restype = None
        qphess_auxil_ctypes.argtypes = (
            utils._BYREF_ENGINEDATA,
            utils._BYREF_INT, # Input, n.
            utils._BYREF_INT, # Input, jthcol.
            utils._EngineFloat64ArrayType.c_type, # Input, h[ldh, tdh].
            utils._BYREF_INT, # Input, ldh.
            utils._BYREF_INT, # Input, tdh.
            utils._EngineFloat64ArrayType.c_type, # Input, x[n].
            utils._EngineFloat64ArrayType.c_type, # Output, hx[n].
            utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
            utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
            utils._EngineIntArrayType.c_type, # Input/Output, iwsav[610].
        )

    def cb_qphessh(_qphess, en_data, n, jthcol, h, ldh, tdh, x, hx, iuser, ruser, iwsav):
        """Helper for qphess."""
        try:
            if qphess is None:
                qphess_auxil_ctypes(
                    en_data,
                    n,
                    jthcol,
                    h,
                    ldh,
                    tdh,
                    x,
                    hx,
                    iuser,
                    ruser,
                    iwsav,
                )
                return
            _n_qphessh = utils._EngineIntScalarType.to_py(n.contents)
            _jthcol_qphessh = utils._EngineIntScalarType.to_py(jthcol.contents)
            _ldh_qphessh = utils._EngineIntScalarType.to_py(ldh.contents)
            _tdh_qphessh = utils._EngineIntScalarType.to_py(tdh.contents)
            _h_qphessh = utils._EngineFloat64ArrayType.from_pointer(
                h,
                (_ldh_qphessh, _tdh_qphessh),
                sorder=_m_sorder,
            )
            _x_qphessh = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_n_qphessh,),
            )
            _hx_qphessh = utils._EngineFloat64ArrayType.from_pointer(
                hx,
                (_n_qphessh,),
                can_reference_contents=False,
            )
            if data is not None:
                qphess(
                    _n_qphessh,
                    _jthcol_qphessh,
                    _h_qphessh,
                    _x_qphessh,
                    _hx_qphessh,
                    data,
                )
            else:
                qphess(
                    _n_qphessh,
                    _jthcol_qphessh,
                    _h_qphessh,
                    _x_qphessh,
                    _hx_qphessh,
                )
            utils._unlink_pointer(h)
            utils._unlink_pointer(x)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _hx_qphessh, {'fun_name': cb_qphess_name, 'returned': 'hx'}, hx,
            )
            utils._unlink_pointer(hx)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_monit_name = fun_name + ':monit'

    monith = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, intfnd.
        utils._BYREF_INT, # Input, nodes.
        utils._BYREF_INT, # Input, depth.
        utils._BYREF_FLOAT64, # Input, obj.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Input/Output, bstval.
        utils._EngineFloat64ArrayType.c_type, # Input, bstsol[n].
        utils._EngineFloat64ArrayType.c_type, # Input, bl[n].
        utils._EngineFloat64ArrayType.c_type, # Input, bu[n].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_LOGICAL, # Input/Output, halt.
        utils._BYREF_INT, # Input/Output, count.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_monith(_monit, en_data, intfnd, nodes, depth, obj, x, bstval, bstsol, bl, bu, n, halt, count, _iuser, _ruser):
        """Helper for monit."""
        if monit is None:
            return
        try:
            _intfnd_monith = utils._EngineIntScalarType.to_py(intfnd.contents)
            _nodes_monith = utils._EngineIntScalarType.to_py(nodes.contents)
            _depth_monith = utils._EngineIntScalarType.to_py(depth.contents)
            _obj_monith = utils._EngineFloat64ScalarType.to_py(obj.contents)
            _n_monith = utils._EngineIntScalarType.to_py(n.contents)
            _x_monith = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_n_monith,),
            )
            _bstval_monith = utils._EngineFloat64ScalarType.to_py(bstval.contents)
            _bstsol_monith = utils._EngineFloat64ArrayType.from_pointer(
                bstsol,
                (_n_monith,),
            )
            _bl_monith = utils._EngineFloat64ArrayType.from_pointer(
                bl,
                (_n_monith,),
            )
            _bu_monith = utils._EngineFloat64ArrayType.from_pointer(
                bu,
                (_n_monith,),
            )
            _halt_monith = utils._EngineLogicalScalarType.to_py(halt.contents)
            _count_monith = utils._EngineIntScalarType.to_py(count.contents)
            if data is not None:
                _bstval_monith, _halt_monith, _count_monith = monit(
                    _intfnd_monith,
                    _nodes_monith,
                    _depth_monith,
                    _obj_monith,
                    _x_monith,
                    _bstval_monith,
                    _bstsol_monith,
                    _bl_monith,
                    _bu_monith,
                    _n_monith,
                    _halt_monith,
                    _count_monith,
                    data,
                )
            else:
                _bstval_monith, _halt_monith, _count_monith = monit(
                    _intfnd_monith,
                    _nodes_monith,
                    _depth_monith,
                    _obj_monith,
                    _x_monith,
                    _bstval_monith,
                    _bstsol_monith,
                    _bl_monith,
                    _bu_monith,
                    _n_monith,
                    _halt_monith,
                    _count_monith,
                )
            utils._unlink_pointer(x)
            utils._EngineFloat64ScalarType(
                _bstval_monith, {'fun_name': cb_monit_name, 'returned': 'bstval'},
            ).refresh_ctypes_buffer(bstval)
            utils._unlink_pointer(bstsol)
            utils._unlink_pointer(bl)
            utils._unlink_pointer(bu)
            utils._EngineLogicalScalarType(
                _halt_monith, {'fun_name': cb_monit_name, 'returned': 'halt'},
            ).refresh_ctypes_buffer(halt)
            utils._EngineIntScalarType(
                _count_monith, {'fun_name': cb_monit_name, 'returned': 'count'},
            ).refresh_ctypes_buffer(count)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('h02cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nclin.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, bl[n+nclin].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, bu[n+nclin].
        utils._EngineFloat64ArrayType.c_type, # Input, cvec[*].
        utils._EngineFloat64ArrayType.c_type, # Input, h[ldh, tdh].
        utils._BYREF_INT, # Input, ldh.
        utils._BYREF_INT, # Input, tdh.
        utils._UNUSED_FUNCTYPE,
        qphessh,
        utils._EngineIntArrayType.c_type, # Input, intvar[lintvr].
        utils._BYREF_INT, # Input, lintvr.
        utils._BYREF_INT, # Input, mdepth.
        utils._EngineIntArrayType.c_type, # Input/Output, istate[n+nclin].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xs[n].
        utils._BYREF_FLOAT64, # Output, obj.
        utils._EngineFloat64ArrayType.c_type, # Output, ax[max(1, nclin)].
        utils._EngineFloat64ArrayType.c_type, # Output, clamda[n+nclin].
        utils._BYREF_INT, # Input, strtgy.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._BYREF_INT, # Input, liwrk.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wrk.
        utils._BYREF_INT, # Input, lwrk.
        utils._UNUSED_FUNCTYPE,
        monith,
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[120].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[610].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[475].
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (h, 'h')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nclin = utils._EngineIntScalarType(
        nclin, {'fun_name': fun_name, 'entity': 'nclin'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _nclin_py = utils._EngineIntScalarType.to_py(_nclin)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _nclin_py > 0:
        _a_dim2 = _n_py
    elif _nclin_py == 0:
        _a_dim2 = 1
    else:
        _a_dim2 = 0
    exp_shape_a = (_nclin_py, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _bl_cls = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_n_py+_nclin_py,),
    )
    _bl = _bl_cls.to_ctypes()
    _bu_cls = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_n_py+_nclin_py,),
    )
    _bu = _bu_cls.to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _iwsav = comm['iwsav']
    if utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (364,), validate=False) < 1 or utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (364,), validate=False) in (2, 4, 6):
        _cvec_dim1 = _n_py
    else:
        _cvec_dim1 = 1
    exp_shape_cvec = (_cvec_dim1,)
    _cvec = utils._EngineFloat64ArrayType(
        cvec, {'fun_name': fun_name, 'entity': 'cvec'},
        exp_shape=exp_shape_cvec,
    ).to_ctypes()
    _ldh = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_shape(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2)[0]), {'fun_name': fun_name, 'entity': '_ldh'},
    ).to_ctypes()
    _tdh_py = utils._EngineFloat64ArrayType.get_shape(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2)[1]
    _tdh = utils._EngineIntScalarType(
        _tdh_py, {'fun_name': fun_name, 'entity': '_tdh'},
    ).to_ctypes()
    _ldh_py = utils._EngineIntScalarType.to_py(_ldh)
    _h = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_ldh_py, _tdh_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    utils._check_callable(qphess, {'fun_name': fun_name, 'entity': 'qphess'}, allow_none=True)
    _qphess = utils._UNUSED_CALLBACK
    _qphessh = utils._cast_callback(
        qphessh, cb_qphessh,
        {'fun_name': fun_name, 'entity': 'qphess'},
    )
    _lintvr = utils._EngineIntScalarType(
        lintvr, {'fun_name': fun_name, 'entity': 'lintvr'},
    ).to_ctypes()
    _lintvr_py = utils._EngineIntScalarType.to_py(_lintvr)
    _intvar = utils._EngineIntArrayType(
        intvar, {'fun_name': fun_name, 'entity': 'intvar'},
        exp_shape=(_lintvr_py,),
    ).to_ctypes()
    _mdepth = utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    _istate_cls = utils._EngineIntArrayType(
        istate, {'fun_name': fun_name, 'entity': 'istate'},
        exp_shape=(_n_py+_nclin_py,),
    )
    _istate = _istate_cls.to_ctypes()
    _xs_cls = utils._EngineFloat64ArrayType(
        xs, {'fun_name': fun_name, 'entity': 'xs'},
        exp_shape=(_n_py,),
    )
    _xs = _xs_cls.to_ctypes()
    _obj = utils._EngineFloat64ScalarType.empty_ctypes()
    _ax_cls = utils._EngineFloat64ArrayType(
        ax, {'fun_name': fun_name, 'entity': 'ax'},
        exp_shape=(max(1, _nclin_py),),
        check_contents=False,
    )
    _ax = _ax_cls.to_ctypes(can_reference_contents=False)
    _clamda_cls = utils._EngineFloat64ArrayType(
        clamda, {'fun_name': fun_name, 'entity': 'clamda'},
        exp_shape=(_n_py+_nclin_py,),
        check_contents=False,
    )
    _clamda = _clamda_cls.to_ctypes(can_reference_contents=False)
    _strtgy = utils._EngineIntScalarType(
        strtgy, {'fun_name': fun_name, 'entity': 'strtgy'},
    ).to_ctypes()
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _liwrk = utils._EngineIntScalarType.c_type(0)
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwrk = utils._EngineIntScalarType.c_type(0)
    utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = utils._UNUSED_CALLBACK
    _monith = utils._cast_callback(
        monith, cb_monith,
        {'fun_name': fun_name, 'entity': 'monit'},
    )
    _iuser = None
    _ruser = None
    _lwsav = comm['lwsav']
    _rwsav = comm['rwsav']
    _routine_name = utils._EngineRoutineName(*b'h02cb ')
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _n,
        _nclin,
        _a,
        _lda,
        _bl,
        _bu,
        _cvec,
        _h,
        _ldh,
        _tdh,
        _qphess,
        _qphessh,
        _intvar,
        _lintvr,
        _mdepth,
        _istate,
        _xs,
        _obj,
        _ax,
        _clamda,
        _strtgy,
        _iwrk,
        _liwrk,
        _wrk,
        _lwrk,
        _monit,
        _monith,
        _iuser,
        _ruser,
        _lwsav,
        _iwsav,
        _rwsav,
        _routine_name,
        _errbuf,
        _ierr,
        _routine_name._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Halted at your request.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid input argument for basic problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, strtgy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: strtgy = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mdepth = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mdepth >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lintvr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lintvr >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nclin = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nclin >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No integer solutions found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Need to increase depth, mdepth = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Basic problem (without integer constraints) unbounded.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Basic problem is infeasible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Basic problem requires too many iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Basic problem has a reduced Hessian exceeding assigned dimensions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Error in designated problem type.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unexpected error in NAG function -- contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _bl_cls.refresh_py(_bl)
        _bu_cls.refresh_py(_bu)
        _istate_cls.refresh_py(_istate)
        _xs_cls.refresh_py(_xs)
        _ax_cls.refresh_py(_ax)
        _clamda_cls.refresh_py(_clamda)
        _obj_py = utils._EngineFloat64ScalarType.to_py(_obj)
        return_data = _obj_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['obj'],
            )(
                _obj_py,
            )
    _err_class.manage()
    return return_data

def iqp_dense_optfile(ioptns, comm, io_manager=None):
    u"""
    Read option values for ``iqp_dense`` from external file.

    ``iqp_dense_optfile`` may be used to supply options to ``iqp_dense``
    from an external file.

    For full information please refer to the NAG Library document for
    h02cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02ccf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.qp_dense_solve'.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ioptns` = *<value>*. Constraint: 0 <= `ioptns` <=
            99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.iqp_dense_optfile'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_read_rech_py(_, en_data, nin, recin, errbuf, ifail, recin_len, errbuf_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_read_rec(
                en_data, utils._EngineIntScalarType.to_py(nin.contents), recin, errbuf, recin_len, io_manager, excs, ifail.contents,
            )
        except:
            utils._read_err(
                _sys.exc_info(),
                excs,
                utils.FileObjManager._read_err,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e04ngft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._READ_REC_FUNCTYPE,
        utils._READ_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, ioptns.
        utils._BYREF_LOGICAL, # Input, dlist.
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[120].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[610].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[475].
        utils._BYREF_INT, # Output, inform.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _read_rech = utils._READ_RECH_FUNCTYPE(cb_read_rech_py)
    _ioptns = utils._EngineIntScalarType(
        ioptns, {'fun_name': fun_name, 'entity': 'ioptns'},
    ).to_ctypes()
    _dlist = utils._EngineLogicalScalarType(
        False, {'fun_name': fun_name, 'entity': '_dlist'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _lwsav = comm['lwsav']
    _iwsav = comm['iwsav']
    _rwsav = comm['rwsav']
    _inform = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        utils._UNUSED_READ_REC,
        _read_rech,
        _ioptns,
        _dlist,
        _lwsav,
        _iwsav,
        _rwsav,
        _inform,
    )
    utils._handle_cb_exception(excs)
    _inform_py = utils._EngineIntScalarType.to_py(_inform)
    if _inform_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ioptns_py = utils._EngineIntScalarType.to_py(_ioptns)
        if _inform_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ioptns = {:s}. Constraint: 0 <= ioptns <= 99.'.format(
                        utils._STR_FMT_INT.format(_ioptns_py),
                    ),
                ],
                errcodes=(_inform_py,),
            )
        elif _inform_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Begin was found, but end-of-file was found before End was found.',
                ],
                errcodes=(_inform_py,),
            )
        elif _inform_py == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'End-of-file was found before Begin was found.',
                ],
                errcodes=(_inform_py,),
            )
        elif _inform_py == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or more lines of the options file is invalid.',
                ],
                errcodes=(_inform_py,),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _inform_py),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def iqp_dense_optstr(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``iqp_dense``.

    ``iqp_dense_optstr`` may be used to supply individual options to
    ``iqp_dense``.

    For full information please refer to the NAG Library document for
    h02cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02cdf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description]
        above and in [Optional Parameters]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.qp_dense_solve'.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.iqp_dense_optstr'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e04nhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._BYREF_LOGICAL, # Input, dlist.
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[120].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[610].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[475].
        utils._BYREF_INT, # Output, inform.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _dlist = utils._EngineLogicalScalarType(
        False, {'fun_name': fun_name, 'entity': '_dlist'},
    ).to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'h02cd ')
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _lwsav = comm['lwsav']
    _iwsav = comm['iwsav']
    _rwsav = comm['rwsav']
    _inform = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _optstr,
        _dlist,
        _routine_name,
        _lwsav,
        _iwsav,
        _rwsav,
        _inform,
        _optstr._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _inform_py = utils._EngineIntScalarType.to_py(_inform)
    if _inform_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _inform_py == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The supplied option string is invalid. Supplied value was:',
                    '{:s}.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_inform_py,),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _inform_py),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def iqp_sparse(n, m, nnz, iobj, ncolh, qphx, a, ha, ka, bl, bu, start, names, nname, crname, ns, xs, intvar, lintvr, mdepth, istate, clamda, strtgy, leniz, lenz, monit, comm, data=None, io_manager=None):
    u"""
    Integer LP or QP problem (sparse), using
    ``opt.qpconvex1_sparse_solve``.

    Note: this function uses optional algorithmic parameters, see also:
    ``iqp_sparse_optfile``, ``iqp_sparse_optstr``.

    ``iqp_sparse`` obtains integer solutions to sparse linear
    programming and quadratic programming problems.

    For full information please refer to the NAG Library document for
    h02ce

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02cef.html

    Parameters
    ----------
    n : int
        n, the number of variables (excluding slacks). This is the
        number of columns in the linear constraint matrix A.

    m : int
        m, the number of general linear constraints (or slacks). This is
        the number of rows in A, including the free row (if any; see
        `iobj`).

    nnz : int
        The number of nonzero elements in A.

    iobj : int
        If `iobj` > 0, row `iobj` of A is a free row containing the
        nonzero elements of the vector c appearing in the linear
        objective term c^Tx.

        If `iobj` = 0, there is no free row, i.e., the problem is either
        an FP problem (in which case `iobj` must be set to zero), or a
        QP problem with c = 0.

    ncolh : int
        n_H, the number of leading nonzero columns of the Hessian matrix
        H. For FP and LP problems, `ncolh` must be set to zero.

    qphx : None or callable qphx(nstate, ncolh, x, hx, data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        For QP problems, you must supply a version of `qphx` to compute
        the matrix product Hx.

        If H has rows and columns consisting entirely of zeros, it is
        most efficient to order the variables x = (y z)^T so that

            Hx = [table omitted](y;z) = (H_1 y;0),

        where the nonlinear variables y appear first as shown.

        For LP problems, `qphx` will never be called by ``iqp_sparse``.

        Parameters
        ~~~~~~~~~~
        nstate : int
            If `nstate` = 1, then ``iqp_sparse`` is calling `qphx` for
            the first time on a sub-problem. This argument setting
            allows you to save computation time if certain data must be
            read or calculated only once.

            If `nstate` >= 2, then ``iqp_sparse`` is calling `qphx` for
            the last time.

            This argument setting allows you to perform some additional
            computation on the final sub-problem solution.

            In general, the last call to `qphx` is made with `nstate` =
            2+errno (see [Errors]).

            Otherwise, `nstate` = 0.

        ncolh : int
            This is the same argument `ncolh` as supplied to
            ``iqp_sparse``.

        x : float, ndarray, shape (`ncolh`)
            The first `ncolh` elements of the vector x.

        hx : float, ndarray, shape (`ncolh`), to be modified in place
            On exit: the product Hx.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    a : float, shape (`nnz`), modified in place
        On entry: the nonzero elements of A, ordered by increasing
        column index. Note that multiple elements with the same row and
        column indices are not allowed.

        On exit: used as internal workspace prior to being restored and
        hence is unchanged.

    ha : int, shape (`nnz`)
        `ha`[i] must contain the row index of the nonzero element stored
        in `a`[i], for i = 1,2,...,`nnz`. Note that the row indices for
        a column may be supplied in any order.

    ka : int, shape (`n`+1)
        `ka`[j] must contain the index in `a` of the start of the jth
        column, for j = 1,2,...,`n`. To specify the jth column as empty,
        set `ka`[j] = `ka`[j+1]. Note that the first and last elements
        of `ka` must be such that `ka`[0] = 1 and `ka`[`n`] = `nnz`+1.

    bl : float, shape (`n`+`m`), modified in place
        l, the lower bounds for all the variables and general
        constraints, in the following order. The first `n` elements of
        `bl` must contain the bounds on the variables x, and the next
        `m` elements the bounds for the general linear constraints Ax
        (or slacks s) and the free row (if any). To specify a
        nonexistent lower bound (i.e., l_j = - infinity), set `bl`[j] <=
        -bigbnd, where bigbnd is the value of the option 'Infinite Bound
        Size' (default value = 10^20). To specify the jth constraint as
        an equality, set `bl`[j] = `bu`[j] = beta, say, where |beta| <
        bigbnd. Note that the lower bound corresponding to the free row
        must be set to - infinity and stored in `bl`[`n`+`iobj`-1].

    bu : float, shape (`n`+`m`), modified in place
        On entry: u, the upper bounds for all the variables and general
        constraints, in the following order. The first `n` elements of
        `bl` must contain the bounds on the variables x, and the next
        `m` elements the bounds for the general linear constraints Ax
        (or slacks s) and the free row (if any). To specify a
        nonexistent upper bound (i.e., u_j = + infinity), set `bu`[j] >=
        bigbnd. Note that the upper bound corresponding to the free row
        must be set to + infinity and stored in `bu`[`n`+`iobj`-1].

        On exit: used as internal workspace prior to being restored and
        hence is unchanged.

    start : str, length 1
        Indicates how a starting basis is to be obtained.

        `start` = 'C'

          An internal crash procedure will be used to choose an initial
          basis matrix B.

        `start` = 'W'

          A basis is already defined in `istate` (probably from a
          previous call).

    names : str, length 8, shape (5)
        A set of names associated with the so-called MPSX form of the
        problem.

        `names`[0]

          Must contain the name for the problem (or be blank).

        `names`[1]

          Must contain the name for the free row (or be blank).

        `names`[2]

          Must contain the name for the constraint right-hand side (or
          be blank).

        `names`[3]

          Must contain the name for the ranges (or be blank).

        `names`[4]

          Must contain the name for the bounds (or be blank).

        (These names are used in the monitoring file output; see
        [Monitoring Information].)

    nname : int
        The number of column and row names supplied in the array
        `names`.

        `nname` = 1

          There are no names. Default names will be used in the printed
          output.

        `nname` = `n`+`m`

          All names must be supplied.

    crname : str, length 8, shape (`nname`)
        The optional column and row names.

        If `nname` = 1, `crname` is not referenced and the printed
        output will use default names for the columns and rows.

        If `nname` = `n`+`m`, the first `n` elements must contain the
        names for the columns and the next `m` elements must contain the
        names for the rows.

        Note that the name for the free row (if any) must be stored in
        `crname`[`n`+`iobj`].

    ns : int
        n_S, the number of superbasics. For QP problems, `ns` need not
        be specified if `start` = 'C', but must retain its value from a
        previous call when `start` = 'W'. For FP and LP problems, `ns`
        need not be initialized.

    xs : float, shape (`n`+`m`), modified in place
        On entry: the initial values of the variables and slacks (x,s).
        (See the description for `istate`.)

        On exit: `xs`[i-1] contains the final value of x_i, for i =
        1,2,...,n.

    intvar : int, shape (`lintvr`)
        Specifies which components of the solution vector x are
        constrained to be integer. Specifically, if k elements of x are
        required to take integer values then `intvar`[i] = l_i, for i =
        1,2,...,k, where l_i is the integer index such that x_l_i is
        integer. If k < `lintvr` then `intvar`[k+1] must be set to -1 to
        signal the end of the integer variable indices.

        The order in which the indices of those components of x required
        to be integer is presented determines the order in which the
        sub-problems are treated and solved.

        As such it can be a powerful tool to assist the function in
        achieving a solution efficiently.

        The general advice is to enter the important integer variables
        in the model early in `intvar`; secondary or less important
        variables should be entered near the end of the list.

        However some experimentation might be required to find the
        optimal order.

    lintvr : int
        k, the number of components of x required to be integer. If k =
        0, then `lintvr` must be set to 1 and `intvar`[0] set to -1.

    mdepth : int
        Specifies the maximum depth the tree of sub-problems may be
        developed.

    istate : int, shape (`n`+`m`), modified in place
        On entry: if `start` = 'C', the first `n` elements of `istate`
        and `xs` must specify the initial states and values,
        respectively, of the variables x. (The slacks s need not be
        initialized.) An internal crash procedure is then used to select
        an initial basis matrix B. The initial basis matrix will be
        triangular (neglecting certain small elements in each column).
        It is chosen from various rows and columns of columns of (A-I).
        Possible values for `istate`[j] are as follows:

        [table omitted]

        If nothing special is known about the problem, or there is no
        wish to provide special information, you may set `istate`[j] = 0
        and `xs`[j] = 0.0, for j = 1,2,...,`n`. All variables will then
        be eligible for the initial basis. Less trivially, to say that
        the jth variable will probably be equal to one of its bounds,
        set `istate`[j] = 4 and `xs`[j] = `bl`[j] or `istate`[j] = 5 and
        `xs`[j] = `bu`[j] as appropriate.

        Following the crash procedure, variables for which `istate`[j] =
        2 are made superbasic.

        Other variables not selected for the basis are then made
        nonbasic at the value `xs`[j] if `bl`[j] <= `xs`[j] <= `bu`[j],
        or at the value `bl`[j] or `bu`[j] closest to `xs`[j].

        If `start` = 'W', `istate` and `xs` must specify the initial
        states and values, respectively, of the variables and slacks
        (x,s).

        If ``iqp_sparse`` has been called previously with the same
        values of `n` and `m`, `istate` already contains satisfactory
        information.

        On exit: the final states of the variables and slacks (x,s) from
        the solution of the last sub-problem tackled. The significance
        of each possible value of `istate`[j] is as follows:

        [table omitted]

        If Ninf = 0 (see [Further Comments]), basic and superbasic
        variables may be outside their bounds by as much as the value of
        the option 'Feasibility Tolerance' (default value = max(10^(-6),
        sqrt(epsilon)), where epsilon is the machine precision). Note
        that unless the option 'Scale Option' = 0 (default value = 2) is
        specified, the 'Feasibility Tolerance' applies to the variables
        of the scaled problem. In this case, the variables of the
        original problem may be as much as 0.1 outside their bounds, but
        this is unlikely unless the problem is very badly scaled.

        Very occasionally some nonbasic variables may be outside their
        bounds by as much as the 'Feasibility Tolerance', and there may
        be some nonbasic variables for which `xs`[j] lies strictly
        between its bounds.

        If Ninf > 0, some basic and superbasic variables may be outside
        their bounds by an arbitrary amount (bounded by Sinf (see
        [Further Comments]) if 'Scale Option' = 0).

    clamda : float, shape (`n`+`m`), modified in place
        On exit: a set of Lagrange-multipliers for the bounds on the
        variables and the general constraints. More precisely, the first
        `n` elements contain the multipliers (reduced costs) for the
        bounds on the variables, and the next `m` elements contain the
        multipliers (shadow prices) for the general linear constraints.

    strtgy : int
        Defines the branching strategy adopted by the function.

        `strtgy` = 0

          Each sub-problem first explored imposes a tighter upper bound
          on the component of x.

        `strtgy` = 1

          Each sub-problem first explored imposes a tighter lower bound
          on the component of x.

        `strtgy` = 2

          Each branch explored imposes a tighter upper bound on the
          component of x if its fractional part is less than 0.5,
          otherwise it imposes a tighter lower bound.

        `strtgy` = 3

          Random choice is made between first exploring a tighter lower
          bound or a tighter upper bound sub-problem.

    leniz : int
        The dimension of the internal workspace array iz.

    lenz : int
        The dimension of the internal workspace array z.

    monit : None or callable (bstval, halt) = monit(intfnd, nodes,
        depth, obj, x, bstval, bstsol, bl, bu, n, halt, data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        To provide feed-back on the progress of the branch and bound
        process.

        Additionally `monit` provides, via its argument `halt`, the
        ability to terminate the process. (You might choose to do this
        when an integer solution is found, rather than search for a
        better solution.) If you do not require any intermediate output
        then `monit` may be None.

        Parameters
        ~~~~~~~~~~
        intfnd : int
            Contains the number of integer solutions obtained so far.

        nodes : int
            Contains the number of nodes (sub-problems) solved so far.

        depth : int
            Contains the depth reached in the tree of problems.

        obj : float
            Contains the solution value to the sub-problem at this node.

        x : float, ndarray, shape (`n`)
            Contains the solution vector to the sub-problem at this
            node.

        bstval : float
            Contains the value of the objective function corresponding
            to the best integer solution obtained so far. If no integer
            solution has been found `bstval` contains the largest
            machine representable number (see ``machine.real_largest``).

        bstsol : float, ndarray, shape (`n`)
            Contains the value of the best integer solution obtained so
            far.

        bl : float, ndarray, shape (`n`)
            Contains the current lower bounds on the variables at this
            point.

        bu : float, ndarray, shape (`n`)
            Contains the current upper bounds on the variables at this
            point.

        n : int
            Contains the number of variables in the minimization
            problem.

        halt : bool
            Will have the value False.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        bstval : float
            May be set to a cut-off value, if you are a sophisticated
            user, as follows. Before an integer solution has been found
            `bstval` will be set by ``iqp_sparse`` to the largest
            machine representable number (see ``machine.real_largest``).
            If you know that the solution being sought is a much smaller
            number, then `bstval` may be set to this number as a cut-off
            value (see [Description]). Beware of setting `bstval` too
            small, since then no integer solutions will be discovered.
            Also, on entry to `monit`, `bstval` should be set using a
            statement of the form

            ::

              if intfnd == 0: bstval = cut_off_value

            Setting `bstval` in this way will not prevent the normal
            operation of the algorithm when subsequent integer solutions
            are found. It would be a grievous mistake to unconditionally
            set `bstval` and if you have any doubts whatsoever about the
            correct use of this argument then you are strongly
            recommended to leave it unchanged.

        halt : bool
            If `halt` is set to True, ``opt.qpconvex1_sparse_solve``
            will be brought to a halt with errno exit -1. This facility
            may be useful if you are content with any integer solution,
            or with any integer solution that fits certain criteria.
            Under these circumstances setting `halt` = True can save
            considerable unnecessary computation.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.qpconvex1_sparse_solve'.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ns : int
        The final number of superbasics. This will be zero for FP and LP
        problems.

    miniz : int
        The minimum value of `leniz` required to start solving the
        problem. If `errno` = 14, ``iqp_sparse`` may be called again
        with `leniz` suitably larger than `miniz`. (The bigger the
        better, since it is not certain how much workspace the basis
        factors need.)

    minz : int
        The minimum value of `lenz` required to start solving the
        problem. If `errno` = 15, ``iqp_sparse`` may be called again
        with `lenz` suitably larger than `minz`. (The bigger the better,
        since it is not certain how much workspace the basis factors
        need.)

    obj : float
        The value of the objective function.

        If Ninf = 0, `obj` includes the quadratic objective term
        1/2x^THx (if any).

        If Ninf > 0, `obj` is just the linear objective term c^Tx (if
        any).

        For FP problems, `obj` is set to zero.

    Other Parameters
    ----------------
    'Check Frequency' : int
        Default = 60

        Every ith iteration after the most recent basis factorization, a
        numerical test is made to see if the current solution (x,s)
        satisfies the linear constraints Ax-s = 0.
        If the largest element of the residual vector r = Ax-s is judged
        to be too large, the current basis is refactorized and the basic
        variables recomputed to satisfy the constraints more accurately.
        If i < 0, the default value is used.
        If i = 0, the value i = 99999999 is used and effectively no
        checks are made.

    'Crash Option' : int
        Default = 2

        Note that this option does not apply when `start` = 'W' (see
        [Parameters]).

        If `start` = 'C', an internal crash procedure is used to select
        an initial basis from various rows and columns of the constraint
        matrix (A-I).
        The value of i determines which rows and columns are initially
        eligible for the basis, and how many times the crash procedure
        is called.
        If i = 0, the all-slack basis B = -I is chosen.
        If i = 1, the crash procedure is called once (looking for a
        triangular basis in all rows and columns of the linear
        constraint matrix A).
        If i = 2, the crash procedure is called twice (looking at any
        equality constraints first followed by any inequality
        constraints).
        If i < 0 or i > 2, the default value is used.

        If i = 1 or 2, certain slacks on inequality rows are selected
        for the basis first. (If i = 2, numerical values are used to
        exclude slacks that are close to a bound.) The crash procedure
        then makes several passes through the columns of A, searching
        for a basis matrix that is essentially triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

    'Crash Tolerance' : float
        Default = 0.1

        This value allows the crash procedure to ignore certain 'small'
        nonzero elements in the constraint matrix A while searching for
        a triangular basis.
        For each column of A, if a_max is the largest element in the
        column, other nonzeros in that column are ignored if they are
        less than (or equal to) a_max*r.

        When r > 0, the basis obtained by the crash procedure may not be
        strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis with more column
        variables and fewer (arbitrary) slacks.
        A feasible solution may be reached earlier for some problems.
        If r < 0 or r >= 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Expand Frequency' : int
        Default = 10000

        This option is part of an anti-cycling procedure (see
        [Miscellaneous]) designed to allow progress even on highly
        degenerate problems.

        For LP problems, the strategy is to force a positive step at
        every iteration, at the expense of violating the constraints by
        a small amount.
        Suppose that the value of the option 'Feasibility Tolerance' is
        delta.
        Over a period of i iterations, the feasibility tolerance
        actually used by ``iqp_sparse`` increases from 0.5 delta to
        delta (in steps of 0.5 delta/i).

        For QP problems, the same procedure is used for iterations in
        which there is only one superbasic variable. (Cycling can only
        occur when the current solution is at a vertex of the feasible
        region.) Thus, zero steps are allowed if there is more than one
        superbasic variable, but otherwise positive steps are enforced.

        Increasing the value of i helps reduce the number of slightly
        infeasible nonbasic basic variables (most of which are
        eliminated during the resetting procedure).
        However, it also diminishes the freedom to choose a large pivot
        element (see the description of the option 'Pivot Tolerance').

        If i < 0, the default value is used.
        If i = 0, the value i = 99999999 is used and effectively no
        anti-cycling procedure is invoked.

    'Factorization Frequency' : int
        Default = 100

        If i > 0, at most i basis changes will occur between
        factorizations of the basis matrix.
        For LP problems, the basis factors are usually updated at every
        iteration.
        For QP problems, fewer basis updates will occur as the solution
        is approached.
        The number of iterations between basis factorizations will
        therefore increase.
        During these iterations a test is made regularly according to
        the value of option 'Check Frequency' to ensure that the linear
        constraints Ax-s = 0 are satisfied.
        If necessary, the basis will be refactorized before the limit of
        i updates is reached.
        If i <= 0, the default value is used.

    'Feasibility Tolerance' : float
        Default = max(10^(-6), sqrt(epsilon))

        If r >= epsilon, r defines the maximum acceptable absolute
        violation in each constraint at a 'feasible' point (including
        slack variables).
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about five decimal digits, it would be appropriate to specify r
        as 10^(-5).
        If r < epsilon, the default value is used.

        ``iqp_sparse`` attempts to find a feasible solution before
        optimizing the objective function.
        If the sum of infeasibilities cannot be reduced to zero, the
        problem is assumed to be infeasible.
        Let Sinf be the corresponding sum of infeasibilities.
        If Sinf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.
        Note that the function does not attempt to find the minimum
        value of Sinf.

        If the constraints and variables have been scaled (see the
        description of the option 'Scale Option'), then feasibility is
        defined in terms of the scaled problem (since it is more likely
        to be meaningful).

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r <= 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that will be considered a step to an unbounded solution. (Note
        that an unbounded solution can occur only when the Hessian is
        not positive definite.) If the change in x during an iteration
        would exceed the value of r, the objective function is
        considered to be unbounded below in the feasible region.
        If r <= 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 5*(n+m))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        Setting i = 0 and 'Print Level' > 0 means that the workspace
        needed to start solving the problem will be computed and
        printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 5*(n+m))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        Setting i = 0 and 'Print Level' > 0 means that the workspace
        needed to start solving the problem will be computed and
        printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 5*(n+m))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        Setting i = 0 and 'Print Level' > 0 means that the workspace
        needed to start solving the problem will be computed and
        printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'LU Factor Tolerance' : float
        Default = 100.0

        The values of r_1 and r_2 affect the stability and sparsity of
        the basis factorization B = LU, during refactorization and
        updates respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        For large and relatively dense problems, setting r_1 and r_2 to
        25 (say) may give a marked improvement in sparsity without
        impairing stability to a serious degree.
        Note that for band matrices it may be necessary to set r_1 in
        the range 1 <= r_1 < 2 in order to achieve stability.
        If r_1 < 1 or r_2 < 1, the default value is used.

    'LU Update Tolerance' : float
        Default = 10.0

        The values of r_1 and r_2 affect the stability and sparsity of
        the basis factorization B = LU, during refactorization and
        updates respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        For large and relatively dense problems, setting r_1 and r_2 to
        25 (say) may give a marked improvement in sparsity without
        impairing stability to a serious degree.
        Note that for band matrices it may be necessary to set r_1 in
        the range 1 <= r_1 < 2 in order to achieve stability.
        If r_1 < 1 or r_2 < 1, the default value is used.

    'LU Singularity Tolerance' : float
        Default = epsilon^0.67

        If r > 0, r defines the singularity tolerance used to guard
        against ill-conditioned basis matrices.
        Whenever the basis is refactorized, the diagonal elements of U
        are tested as follows.
        If |u_{jj}| <= r or |u_{jj}| < r*max_i|u_{ij}|, the jth column
        of the basis is replaced by the corresponding slack variable.
        If r <= 0, the default value is used.

    'Minimize' : valueless
        Default

        This option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j (see [The Main Iteration]) will be reversed.

    'Maximize' : valueless
        This option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j (see [The Main Iteration]) will be reversed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Print Level' > 0, monitoring information produced
        by ``iqp_sparse`` is sent to a file with logical unit number i.
        If i < 0 and/or 'Print Level' = 0, the default value is used and
        hence no monitoring information is produced.

    'Optimality Tolerance' : float
        Default = max(10^(-6), sqrt(epsilon))

        If r >= epsilon, r is used to judge the size of the reduced
        gradients d_j = g_j-pi^Ta_j.
        By definition, the reduced gradients for basic variables are
        always zero.
        Optimality is declared if the reduced gradients for any nonbasic
        variables at their lower or upper bounds satisfy -r*max(1,
        ||pi||) <= d_j <= r*max(1, ||pi||), and if |d_j| <= r*max(1,
        ||pi||) for any superbasic variables.
        If r < epsilon, the default value is used.

    'Partial Price' : int
        Default = 10

        Note that this option does not apply to QP problems.

        This option is recommended for large FP or LP problems that have
        significantly more variables than constraints (i.e., n≫m).
        It reduces the work required for each pricing operation.
        If i = 1, all columns of the constraint matrix (A-I) are
        searched.
        If i > 1, A and -I are partitioned to give i roughly equal
        segments A_j,K_j, for j = 1,2,...,p (modulo p).
        If the previous pricing search was successful on
        A_{j-1},K_{j-1}, the next search begins on the segments A_j,K_j.
        If a reduced gradient is found that is larger than some dynamic
        tolerance, the variable with the largest such reduced gradient
        (of appropriate sign) is selected to enter the basis.
        If nothing is found, the search continues on the next segments
        A_{j+1},K_{j+1}, and so on.
        If i <= 0, the default value is used.

    'Pivot Tolerance' : float
        Default = epsilon^0.67

        If r > 0, r is used to prevent columns entering the basis if
        they would cause the basis to become almost singular.
        If r <= 0, the default value is used.

    'Print Level' : int
        Default = 10

        The value of i controls the amount of printout produced by
        ``iqp_sparse``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each iteration and the
        final solution) and [Monitoring Information] (monitoring
        information at each iteration).
        Note that the summary output will not exceed 80 characters per
        line and that the monitoring information will not exceed 120
        characters per line.
        If i < 0, the default value is used.
        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the 'Monitoring File':

        [table omitted]

        If 'Print Level' > 0 and the unit number defined by 'Monitoring
        File' is the advisory unit number, then the summary output is
        suppressed.

    'Rank Tolerance' : float
        Default = 100 epsilon

        See above.

    'Scale Option' : int
        Default = 2

        This option enables you to scale the variables and constraints
        using an iterative procedure due to Fourer, which attempts to
        compute row scales r_i and column scales c_j such that the
        scaled matrix coefficients a__{ij} = a_{ij}*(c_j/r_i) are as
        close as possible to unity.
        This may improve the overall efficiency of the function on some
        problems. (The lower and upper bounds on the variables and
        slacks for the scaled problem are redefined as l__j = l_j/c_j
        and u__j = u_j/c_j respectively, where c_j = r_{j-n} if j > n.)

        If i = 0, no scaling is performed.
        If i = 1, all rows and columns of the constraint matrix A are
        scaled.
        If i = 2, an additional scaling is performed that may be helpful
        when the solution x is large; it takes into account columns of
        (A-I) that are fixed or have positive lower bounds or negative
        upper bounds.
        If i < 0 or i > 2, the default value is used.

    'Scale Tolerance' : float
        Default = 0.9

        Note that this option does not apply when 'Scale Option' = 0.

        If 0 < r < 1, r is used to control the number of scaling passes
        to be made through the constraint matrix A.
        At least 3 (and at most 10) passes will be made.
        More precisely, let a_p denote the largest column ratio (i.e.,
        'biggest'element/'smallest'element in some sense) after the pth
        scaling pass through A.
        The scaling procedure is terminated if a_p >= a_{p-1}*r for some
        p >= 3.
        Thus, increasing the value of r from 0.9 to 0.99 (say) will
        probably increase the number of passes through A.
        If r <= 0 or r >= 1, the default value is used.

    'Superbasics Limit' : int
        Default = min(n_H+1, n)

        Note that this option does not apply to FP or LP problems.

        The value of i specifies 'how nonlinear' you expect the QP
        problem to be.
        If i <= 0, the default value is used.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `strtgy` = *<value>*.

            Constraint: `strtgy` = 0, 1, 2 or 3.

        (`errno` 1)
            On entry, `mdepth` = *<value>*.

            Constraint: `mdepth` >= 1.

        (`errno` 1)
            On entry, `lintvr` = *<value>*.

            Constraint: `lintvr` >= 1.

        (`errno` 1)
            On entry, i = *<value>*, `istate`[i-1] = *<value>*.

            Constraint: 0 <= `istate`[i] <= 3 with `start` = 'W'.

        (`errno` 1)
            On entry, i = *<value>*, `istate`[i-1] = *<value>*.

            Constraint: 0 <= `istate`[i] <= 5 with `start` = 'C'.

        (`errno` 1)
            On entry, `nname` = *<value>*, `n` = *<value>* and `m` =
            *<value>*.

            Constraint: `nname` = 1 or `nname` = `n`+`m`.

        (`errno` 1)
            On entry, `start` = *<value>*.

            Constraint: `start` = 'C' or 'W'.

        (`errno` 1)
            On entry, `ka`[i] = *<value>*, `ka`[i-1] = *<value>* i =
            *<value>* and `m` = *<value>*.

            Constraint: 0 <= `ka`[i]-`ka`[i-1] <= `m`.

        (`errno` 1)
            On entry, i = *<value>* and `ka`[i-1] = *<value>*.

            Constraint: `ka`[i-1] >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*, `ka`[`n`] = *<value>* and `nnz` =
            *<value>*.

            Constraint: `ka`[`n`] = `nnz`+1.

        (`errno` 1)
            On entry, `ka`[0] = *<value>*.

            Constraint: `ka`[0] = 1.

        (`errno` 1)
            On entry, i = *<value>*, `ha`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `ha`[i] <= `m`.

        (`errno` 1)
            On entry, `ncolh` = *<value>* and `n` = *<value>*.

            Constraint: 0 <= `ncolh` <= `n`.

        (`errno` 1)
            On entry, `iobj` = *<value>* and `m` = *<value>*.

            Constraint: 0 <= `iobj` <= `m`.

        (`errno` 1)
            On entry, `nnz` = *<value>*, `n` = *<value>* and `m` =
            *<value>*.

            Constraint: 0 <= `nnz` <= `n`*`m`

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            No integer solutions found.

        (`errno` 3)
            Need to increase depth, `mdepth` = *<value>*.

        (`errno` 4)
            The problem is unbounded (or badly scaled).

        (`errno` 5)
            The problem is infeasible.

        (`errno` 6)
            Too many iterations. Try increasing Iteration Limit.

        (`errno` 7)
            Reduced Hessian exceeds assigned dimension.

        (`errno` 8)
            Hessian appears to be indefinite.

        (`errno` 9)
            Invalid input argument for internal call to local optimizer.

        (`errno` 10)
            Numerical error trying to satisfy constraints. The basis is
            very ill-conditioned.

        (`errno` 13)
            Singular basis after 15 attempts to factorize.

        (`errno` 16)
            Unexpected error in NAG function -- contact NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` -1)
            Halted at your request.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.iqp_sparse'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_qphx_name = fun_name + ':qphx'

    qphxh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nstate.
        utils._BYREF_INT, # Input, ncolh.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ncolh].
        utils._EngineFloat64ArrayType.c_type, # Output, hx[ncolh].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_qphxh(_qphx, en_data, nstate, ncolh, x, hx, _iuser, _ruser):
        """Helper for qphx."""
        if qphx is None:
            return
        try:
            _nstate_qphxh = utils._EngineIntScalarType.to_py(nstate.contents)
            _ncolh_qphxh = utils._EngineIntScalarType.to_py(ncolh.contents)
            _x_qphxh = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_ncolh_qphxh,),
            )
            _hx_qphxh = utils._EngineFloat64ArrayType.from_pointer(
                hx,
                (_ncolh_qphxh,),
                can_reference_contents=False,
            )
            if data is not None:
                qphx(
                    _nstate_qphxh,
                    _ncolh_qphxh,
                    _x_qphxh,
                    _hx_qphxh,
                    data,
                )
            else:
                qphx(
                    _nstate_qphxh,
                    _ncolh_qphxh,
                    _x_qphxh,
                    _hx_qphxh,
                )
            utils._unlink_pointer(x)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _hx_qphxh, {'fun_name': cb_qphx_name, 'returned': 'hx'}, hx,
            )
            utils._unlink_pointer(hx)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_monit_name = fun_name + ':monit'

    monith = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, intfnd.
        utils._BYREF_INT, # Input, nodes.
        utils._BYREF_INT, # Input, depth.
        utils._BYREF_FLOAT64, # Input, obj.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Input/Output, bstval.
        utils._EngineFloat64ArrayType.c_type, # Input, bstsol[n].
        utils._EngineFloat64ArrayType.c_type, # Input, bl[n].
        utils._EngineFloat64ArrayType.c_type, # Input, bu[n].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_LOGICAL, # Input/Output, halt.
        utils._BYREF_INT, # Input/Output, count.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_monith(_monit, en_data, intfnd, nodes, depth, obj, x, bstval, bstsol, bl, bu, n, halt, count, _iuser, _ruser):
        """Helper for monit."""
        if monit is None:
            return
        try:
            _intfnd_monith = utils._EngineIntScalarType.to_py(intfnd.contents)
            _nodes_monith = utils._EngineIntScalarType.to_py(nodes.contents)
            _depth_monith = utils._EngineIntScalarType.to_py(depth.contents)
            _obj_monith = utils._EngineFloat64ScalarType.to_py(obj.contents)
            _n_monith = utils._EngineIntScalarType.to_py(n.contents)
            _x_monith = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_n_monith,),
            )
            _bstval_monith = utils._EngineFloat64ScalarType.to_py(bstval.contents)
            _bstsol_monith = utils._EngineFloat64ArrayType.from_pointer(
                bstsol,
                (_n_monith,),
            )
            _bl_monith = utils._EngineFloat64ArrayType.from_pointer(
                bl,
                (_n_monith,),
            )
            _bu_monith = utils._EngineFloat64ArrayType.from_pointer(
                bu,
                (_n_monith,),
            )
            _halt_monith = utils._EngineLogicalScalarType.to_py(halt.contents)
            if data is not None:
                _bstval_monith, _halt_monith = monit(
                    _intfnd_monith,
                    _nodes_monith,
                    _depth_monith,
                    _obj_monith,
                    _x_monith,
                    _bstval_monith,
                    _bstsol_monith,
                    _bl_monith,
                    _bu_monith,
                    _n_monith,
                    _halt_monith,
                    data,
                )
            else:
                _bstval_monith, _halt_monith = monit(
                    _intfnd_monith,
                    _nodes_monith,
                    _depth_monith,
                    _obj_monith,
                    _x_monith,
                    _bstval_monith,
                    _bstsol_monith,
                    _bl_monith,
                    _bu_monith,
                    _n_monith,
                    _halt_monith,
                )
            utils._unlink_pointer(x)
            utils._EngineFloat64ScalarType(
                _bstval_monith, {'fun_name': cb_monit_name, 'returned': 'bstval'},
            ).refresh_ctypes_buffer(bstval)
            utils._unlink_pointer(bstsol)
            utils._unlink_pointer(bl)
            utils._unlink_pointer(bu)
            utils._EngineLogicalScalarType(
                _halt_monith, {'fun_name': cb_monit_name, 'returned': 'halt'},
            ).refresh_ctypes_buffer(halt)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('h02ceft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, nnz.
        utils._BYREF_INT, # Input, iobj.
        utils._BYREF_INT, # Input, ncolh.
        utils._UNUSED_FUNCTYPE,
        qphxh,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, ha[nnz].
        utils._EngineIntArrayType.c_type, # Input, ka[n+1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, bl[n+m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, bu[n+m].
        utils._BYREF_STRING, # Input, start(len=1).
        utils._EngineCharArrayType.c_type, # Input, names(len=8)[5].
        utils._BYREF_INT, # Input, nname.
        utils._EngineCharArrayType.c_type, # Input, crname(len=8)[nname].
        utils._BYREF_INT, # Input/Output, ns.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xs[n+m].
        utils._EngineIntArrayType.c_type, # Input, intvar[lintvr].
        utils._BYREF_INT, # Input, lintvr.
        utils._BYREF_INT, # Input, mdepth.
        utils._EngineIntArrayType.c_type, # Input/Output, istate[n+m].
        utils._BYREF_INT, # Output, miniz.
        utils._BYREF_INT, # Output, minz.
        utils._BYREF_FLOAT64, # Output, obj.
        utils._EngineFloat64ArrayType.c_type, # Output, clamda[n+m].
        utils._BYREF_INT, # Input, strtgy.
        utils._EngineIntArrayType.c_type, # Workspace, iz.
        utils._BYREF_INT, # Input, leniz.
        utils._EngineFloat64ArrayType.c_type, # Workspace, z.
        utils._BYREF_INT, # Input, lenz.
        utils._UNUSED_FUNCTYPE,
        monith,
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[20].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[380].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[285].
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of start.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of names.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of crname.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _iobj = utils._EngineIntScalarType(
        iobj, {'fun_name': fun_name, 'entity': 'iobj'},
    ).to_ctypes()
    _ncolh = utils._EngineIntScalarType(
        ncolh, {'fun_name': fun_name, 'entity': 'ncolh'},
    ).to_ctypes()
    utils._check_callable(qphx, {'fun_name': fun_name, 'entity': 'qphx'}, allow_none=True)
    _qphx = utils._UNUSED_CALLBACK
    _qphxh = utils._cast_callback(
        qphxh, cb_qphxh,
        {'fun_name': fun_name, 'entity': 'qphx'},
    )
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _ha = utils._EngineIntArrayType(
        ha, {'fun_name': fun_name, 'entity': 'ha'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ka = utils._EngineIntArrayType(
        ka, {'fun_name': fun_name, 'entity': 'ka'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _bl_cls = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_n_py+_m_py,),
    )
    _bl = _bl_cls.to_ctypes()
    _bu_cls = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_n_py+_m_py,),
    )
    _bu = _bu_cls.to_ctypes()
    _start = utils._EngineCharScalarType(
        start, {'fun_name': fun_name, 'entity': 'start'}, exp_chrlen=1,
    ).to_ctypes()
    _names_cls = utils._EngineCharArrayType(
        names, {'fun_name': fun_name, 'entity': 'names'}, exp_chrlen=8,
        exp_shape=(5,),
    )
    _names = _names_cls.to_ctypes()
    _names_chrlen = _names_cls.chrlen
    _nname = utils._EngineIntScalarType(
        nname, {'fun_name': fun_name, 'entity': 'nname'},
    ).to_ctypes()
    _nname_py = utils._EngineIntScalarType.to_py(_nname)
    _crname_cls = utils._EngineCharArrayType(
        crname, {'fun_name': fun_name, 'entity': 'crname'}, exp_chrlen=8,
        exp_shape=(_nname_py,),
    )
    _crname = _crname_cls.to_ctypes()
    _crname_chrlen = _crname_cls.chrlen
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _xs_cls = utils._EngineFloat64ArrayType(
        xs, {'fun_name': fun_name, 'entity': 'xs'},
        exp_shape=(_n_py+_m_py,),
    )
    _xs = _xs_cls.to_ctypes()
    _lintvr = utils._EngineIntScalarType(
        lintvr, {'fun_name': fun_name, 'entity': 'lintvr'},
    ).to_ctypes()
    _lintvr_py = utils._EngineIntScalarType.to_py(_lintvr)
    _intvar = utils._EngineIntArrayType(
        intvar, {'fun_name': fun_name, 'entity': 'intvar'},
        exp_shape=(_lintvr_py,),
    ).to_ctypes()
    _mdepth = utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    _istate_cls = utils._EngineIntArrayType(
        istate, {'fun_name': fun_name, 'entity': 'istate'},
        exp_shape=(_n_py+_m_py,),
    )
    _istate = _istate_cls.to_ctypes()
    _miniz = utils._EngineIntScalarType.empty_ctypes()
    _minz = utils._EngineIntScalarType.empty_ctypes()
    _obj = utils._EngineFloat64ScalarType.empty_ctypes()
    _clamda_cls = utils._EngineFloat64ArrayType(
        clamda, {'fun_name': fun_name, 'entity': 'clamda'},
        exp_shape=(_n_py+_m_py,),
        check_contents=False,
    )
    _clamda = _clamda_cls.to_ctypes(can_reference_contents=False)
    _strtgy = utils._EngineIntScalarType(
        strtgy, {'fun_name': fun_name, 'entity': 'strtgy'},
    ).to_ctypes()
    _iz = utils._EngineIntArrayType.empty_ctypes(0)
    _leniz = utils._EngineIntScalarType(
        leniz, {'fun_name': fun_name, 'entity': 'leniz'},
    ).to_ctypes()
    _z = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lenz = utils._EngineIntScalarType(
        lenz, {'fun_name': fun_name, 'entity': 'lenz'},
    ).to_ctypes()
    utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = utils._UNUSED_CALLBACK
    _monith = utils._cast_callback(
        monith, cb_monith,
        {'fun_name': fun_name, 'entity': 'monit'},
    )
    _iuser = None
    _ruser = None
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _lwsav = comm['lwsav']
    _iwsav = comm['iwsav']
    _rwsav = comm['rwsav']
    _routine_name = utils._EngineRoutineName(*b'h02ce ')
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _n,
        _m,
        _nnz,
        _iobj,
        _ncolh,
        _qphx,
        _qphxh,
        _a,
        _ha,
        _ka,
        _bl,
        _bu,
        _start,
        _names,
        _nname,
        _crname,
        _ns,
        _xs,
        _intvar,
        _lintvr,
        _mdepth,
        _istate,
        _miniz,
        _minz,
        _obj,
        _clamda,
        _strtgy,
        _iz,
        _leniz,
        _z,
        _lenz,
        _monit,
        _monith,
        _iuser,
        _ruser,
        _lwsav,
        _iwsav,
        _rwsav,
        _routine_name,
        _errbuf,
        _ierr,
        _start._length_,
        _names_chrlen,
        _crname_chrlen,
        _routine_name._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99966:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Halted at your request.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, strtgy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: strtgy = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mdepth = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mdepth >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lintvr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lintvr >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, istate[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= istate[i] <= 3',
                    'with start = \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, istate[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= istate[i] <= 5',
                    'with start = \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nname = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: nname = 1 or nname = n+m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, start = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: start = \'C\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ka[i] = {:s}, ka[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'i = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: 0 <= ka[i]-ka[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and ka[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ka[i-1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}, ka[n] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ka[n] = nnz+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ka[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ka[0] = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ha[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= ha[i] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncolh = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= ncolh <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iobj = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= iobj <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 <= nnz <= n*m',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99967:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No integer solutions found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99965:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Need to increase depth, mdepth = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The problem is unbounded (or badly scaled).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99977:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The problem is infeasible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99976:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Too many iterations. Try increasing Iteration Limit.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99975:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Reduced Hessian exceeds assigned dimension.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99974:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Hessian appears to be indefinite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99973:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid input argument for internal call to local optimizer.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99972:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Numerical error trying to satisfy constraints. The basis is very',
                    'ill-conditioned.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 99971:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Singular basis after 15 attempts to factorize.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 16 and _ifmt == 99968:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unexpected error in NAG function -- contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _bl_cls.refresh_py(_bl)
        _bu_cls.refresh_py(_bu)
        _ns_py = utils._EngineIntScalarType.to_py(_ns)
        _xs_cls.refresh_py(_xs)
        _istate_cls.refresh_py(_istate)
        _clamda_cls.refresh_py(_clamda)
        _miniz_py = utils._EngineIntScalarType.to_py(_miniz)
        _minz_py = utils._EngineIntScalarType.to_py(_minz)
        _obj_py = utils._EngineFloat64ScalarType.to_py(_obj)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ns', 'miniz', 'minz', 'obj'],
        )(
            _ns_py,
            _miniz_py,
            _minz_py,
            _obj_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def iqp_sparse_optfile(ioptns, comm, io_manager=None):
    u"""
    Read option values for ``iqp_sparse`` from external file.

    ``iqp_sparse_optfile`` may be used to supply options to
    ``iqp_sparse`` from an external file.

    For full information please refer to the NAG Library document for
    h02cf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02cff.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.qpconvex1_sparse_solve'.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ioptns` = *<value>*. Constraint: 0 <= `ioptns` <=
            99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            The supplied option string is invalid. Check that the
            keywords are neither ambiguous nor misspelt. The option
            string is *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.iqp_sparse_optfile'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_read_rech_py(_, en_data, nin, recin, errbuf, ifail, recin_len, errbuf_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_read_rec(
                en_data, utils._EngineIntScalarType.to_py(nin.contents), recin, errbuf, recin_len, io_manager, excs, ifail.contents,
            )
        except:
            utils._read_err(
                _sys.exc_info(),
                excs,
                utils.FileObjManager._read_err,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e04nlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._READ_REC_FUNCTYPE,
        utils._READ_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, ioptns.
        utils._BYREF_LOGICAL, # Input, dlist.
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[20].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[380].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[285].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, inform.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _read_rech = utils._READ_RECH_FUNCTYPE(cb_read_rech_py)
    _ioptns = utils._EngineIntScalarType(
        ioptns, {'fun_name': fun_name, 'entity': 'ioptns'},
    ).to_ctypes()
    _dlist = utils._EngineLogicalScalarType(
        False, {'fun_name': fun_name, 'entity': '_dlist'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _lwsav = comm['lwsav']
    _iwsav = comm['iwsav']
    _rwsav = comm['rwsav']
    _errbuf = utils._EngineErrbuf()
    _inform = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        utils._UNUSED_READ_REC,
        _read_rech,
        _ioptns,
        _dlist,
        _lwsav,
        _iwsav,
        _rwsav,
        _errbuf,
        _inform,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _inform_py = utils._EngineIntScalarType.to_py(_inform)
    if _inform_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _inform_py)
        _ioptns_py = utils._EngineIntScalarType.to_py(_ioptns)
        if _inform_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ioptns = {:s}. Constraint: 0 <= ioptns <= 99.'.format(
                        utils._STR_FMT_INT.format(_ioptns_py),
                    ),
                ],
                errcodes=(_inform_py, _inform_py, _ifmt),
            )
        elif _inform_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Begin was found, but end-of-file was found before End was found.',
                ],
                errcodes=(_inform_py, _inform_py, _ifmt),
            )
        elif _inform_py == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'End-of-file was found before Begin was found.',
                ],
                errcodes=(_inform_py, _inform_py, _ifmt),
            )
        elif _inform_py == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The supplied option string is invalid.',
                    'Check that the keywords are neither ambiguous nor misspelt.',
                    'The option string is \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_inform_py, _inform_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _inform_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def iqp_sparse_optstr(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``iqp_sparse``.

    ``iqp_sparse_optstr`` may be used to supply individual options to
    ``iqp_sparse``.

    For full information please refer to the NAG Library document for
    h02cg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02cgf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description]
        above and in [Optional Parameters]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt.nlp1_init`` with rname = 'opt.qpconvex1_sparse_solve'.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.iqp_sparse_optstr'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e04nmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._BYREF_LOGICAL, # Input, dlist.
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineLogicalArrayType.c_type, # Input/Output, lwsav[20].
        utils._EngineIntArrayType.c_type, # Input/Output, iwsav[380].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rwsav[285].
        utils._BYREF_INT, # Output, inform.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _dlist = utils._EngineLogicalScalarType(
        False, {'fun_name': fun_name, 'entity': '_dlist'},
    ).to_ctypes()
    _routine_name = utils._EngineRoutineName(*b'h02cg ')
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _lwsav = comm['lwsav']
    _iwsav = comm['iwsav']
    _rwsav = comm['rwsav']
    _inform = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _optstr,
        _dlist,
        _routine_name,
        _lwsav,
        _iwsav,
        _rwsav,
        _inform,
        _optstr._length_,
        _routine_name._length_,
    )
    utils._handle_cb_exception(excs)
    _inform_py = utils._EngineIntScalarType.to_py(_inform)
    if _inform_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _inform_py == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The supplied option string is invalid. Supplied value was:',
                    '{:s}.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_inform_py,),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _inform_py),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def sqp(n, nclin, ncnln, a, d, ax, bl, bu, varcon, x, confun, c, cjac, objfun, objgrd, maxit, acc, comm, data=None, io_manager=None):
    u"""
    Mixed integer nonlinear programming.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``.

    ``sqp`` solves general nonlinear programming problems with integer
    constraints on some of the variables.

    For full information please refer to the NAG Library document for
    h02da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02daf.html

    Parameters
    ----------
    n : int
        n, the total number of variables, n_c+n_i.

    nclin : int
        n_l, the number of general linear constraints defined by A and
        d.

    ncnln : int
        n_N, the number of constraints supplied by c(x).

    a : None or float, shape (`nclin`, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `nclin`>0: `n`; otherwise: 0.

        The ith row of `a` must contain the coefficients of the ith
        general linear constraint, for i = 1,2,...,n_l. Any equality
        constraints must be specified first.

        If `nclin` = 0, the array `a` is not referenced.

    d : None or float, shape (`nclin`)
        d_i, the constant for the ith linear constraint.

        If `nclin` = 0, the array `d` is not referenced.

    ax : None or float, shape (`nclin`), modified in place
        On exit, if not None on entry: the final residuals of the linear
        constraints Ax-d.

        If `nclin` = 0, `ax` is not referenced.

    bl : float, shape (`n`)
        `bl` must contain the lower bounds, l_i, and `bu` the upper
        bounds, u_i, for the variables; bounds on integer variables are
        rounded, bounds on binary variables need not be supplied.

    bu : float, shape (`n`)
        `bl` must contain the lower bounds, l_i, and `bu` the upper
        bounds, u_i, for the variables; bounds on integer variables are
        rounded, bounds on binary variables need not be supplied.

    varcon : int, shape (`n`+`nclin`+`ncnln`)
        `varcon` indicates the nature of each variable and constraint in
        the problem. The first n elements of the array must describe the
        nature of the variables, the next n_L elements the nature of the
        general linear constraints (if any) and the next n_N elements
        the general constraints (if any).

        `varcon`[j-1] = 0

          A continuous variable.

        `varcon`[j-1] = 1

          A binary variable.

        `varcon`[j-1] = 2

          An integer variable.

        `varcon`[j-1] = 3

          An equality constraint.

        `varcon`[j-1] = 4

          An inequality constraint.

    x : float, shape (`n`), modified in place
        On entry: an initial estimate of the solution, which need not be
        feasible. Values corresponding to integer variables are rounded;
        if an initial value less than half is supplied for a binary
        variable the value zero is used, otherwise the value one is
        used.

        On exit: the final estimate of the solution.

    confun : None or callable confun(mode, ncnln, n, varcon, x, c, cjac,
        nstate, data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the constraint functions supplied by
        c(x) and their Jacobian at x. If all constraints are supplied by
        A and d (i.e., `ncnln` = 0), `confun` will never be called by
        ``sqp`` and `confun` may be None. If `ncnln` > 0, the first call
        to `confun` will occur after the first call to `objfun`.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0

              Elements of `c` containing continuous variables.

            `mode` = 1

              Elements of `cjac` containing continuous variables.

        ncnln : int
            The number of constraints supplied by c(x), n_N.

        n : int
            n, the total number of variables, n_c+n_i.

        varcon : int, ndarray, shape (lvarcon)
            The array `varcon` as supplied to ``sqp``.

        x : float, ndarray, shape (`n`)
            The vector of variables at which the objective function
            and/or all continuous elements of its gradient are to be
            evaluated.

        c : float, ndarray, shape (`ncnln`), to be modified in place
            On exit: must contain `ncnln` constraint values, with the
            value of the jth constraint c_j(x) in `c`[j-1].

        cjac : float, ndarray, shape (`ncnln`, `n`), to be modified in
        place
            On entry: continuous elements of `cjac` are set to the value
            of NaN.

            On exit: the ith row of `cjac` must contain elements of
            dc_i/dx_j for each continuous variable x_j.

        nstate : int
            If `nstate` = 1, ``sqp`` is calling `confun` for the first
            time. This argument setting allows you to save computation
            time if certain data must be read or calculated only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    c : None or float, shape (`ncnln`), modified in place
        On exit, if not None on entry: if `ncnln` > 0, `c`[j-1] contains
        the value of the jth constraint function c_j(x) at the final
        iterate, for j = 1,2,...,`ncnln`.

        If `ncnln` = 0, the array `c` is not referenced.

    cjac : None or float, shape (`ncnln`, `n`), modified in place
        On exit, if not None on entry: if `ncnln` > 0, `cjac` contains
        the Jacobian matrix of the constraint functions at the final
        iterate, i.e., `cjac`[i-1,j-1] contains the partial derivative
        of the ith constraint function with respect to the jth variable,
        for j = 1,2,...,`n`, for i = 1,2,...,`ncnln`. (See the
        discussion of argument `cjac` under `confun`.)

        If `ncnln` = 0, the array `cjac` is not referenced.

    objfun : callable objmip = objfun(mode, n, varcon, x, objgrd,
        nstate, data=None)
        `objfun` must calculate the objective function f(x) and its
        gradient for a specified n-element vector x.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0

              The objective function value, `objmip`.

            `mode` = 1

              The continuous elements of `objgrd`.

        n : int
            n, the total number of variables, n_c+n_i.

        varcon : int, ndarray, shape (lvarcon)
            The array `varcon` as supplied to ``sqp``.

        x : float, ndarray, shape (`n`)
            The vector of variables at which the objective function
            and/or all continuous elements of its gradient are to be
            evaluated.

        objgrd : float, ndarray, shape (`n`), to be modified in place
            On entry: continuous elements of `objgrd` are set to the
            value of NaN.

            On exit: must contain the gradient vector of the objective
            function if `mode` = 1, with `objgrd`[j-1] containing the
            partial derivative of f with respect to continuous variable
            x_j; otherwise `objgrd` is not referenced.

        nstate : int
            If `nstate` = 1, ``sqp`` is calling `objfun` for the first
            time. This argument setting allows you to save computation
            time if certain data must be read or calculated only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        objmip : float
            Must be set to the objective function value, f.

    objgrd : float, shape (`n`), modified in place
        On exit: the objective function gradient at the solution.

    maxit : int
        The maximum number of iterations within which to find a
        solution. If `maxit` is less than or equal to zero, the
        suggested value below is used.

    acc : float
        The requested accuracy for determining feasible points during
        iterations and for halting the method when the predicted
        improvement in objective function is less than `acc`. If `acc`
        is less than or equal to epsilon (epsilon being the machine
        precision as given by ``machine.precision``), the below
        suggested value is used.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``optset``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    objmip : float
        With no exception or warning is raised, `objmip` contains the
        value of the objective function for the MINLP solution.

    Other Parameters
    ----------------
    'Branch Bound Steps' : int
        Default = 500

        Maximum number of branch-and-bound steps for solving the mixed
        integer quadratic problems.

    'Branching Rule' : str
        Default = 'Maximum'

        Branching rule for branch and bound search.

        'Branching Rule' = 'Maximum'
          Maximum fractional branching.

        'Branching Rule' = 'Minimum'
          Minimum fractional branching.

    'Check Gradients' : str
        Default = 'No'

        Perform an internal check of supplied objective and constraint
        gradients.
        It is advisable to set 'Check Gradients' = 'Yes' during code
        development to avoid difficulties associated with incorrect
        user-supplied gradients.

    'Continuous Trust Radius' : float
        Default = 10.0

        Initial continuous trust region radius, Delta_0^c; the initial
        trial step d in R^(n_c) for the SQP approximation must satisfy
        ||d||_infinity <= Delta_0^c.

    'Descent' : float
        Default = 0.05

        Initial descent parameter, delta, larger values of delta allow
        penalty option sigma to increase faster which can lead to faster
        convergence.

    'Descent Factor' : float
        Default = 0.1

        Factor for decreasing the internal descent parameter, delta,
        between iterations.

    'Feasible Steps' : int
        Default = 10

        Maximum number of feasible steps without improvements, where
        feasibility is measured by ||g(x)||_infinity <= sqrt(`acc`).

    'Improved Bounds' : str
        Default = 'No'

        Calculate improved bounds in case of 'Best of all' node
        selection strategy.

    'Integer Trust Radius' : float
        Default = 10.0

        Initial integer trust region radius, Delta_0^i; the initial
        trial step e in R^(n_i) for the SQP approximation must satisfy
        ||e||_infinity <= Delta_0^i.

    'Maximum Restarts' : int
        Default = 2

        Maximum number of restarts that allow the mixed integer SQP
        algorithm to return to a better solution.
        Setting a value higher than the default might lead to better
        results at the expense of function evaluations.

    'Minor Print Level' : int
        Default = 0

        Print level of the subproblem solver.
        Active only if 'Print Level' != 0.

    'Modify Hessian' : str
        Default = 'Yes'

        Modify the Hessian approximation in an attempt to get more
        accurate search directions.
        Calculation time is increased when the number of integer
        variables is large.

    'Node Selection' : str
        Default = 'Depth First'

        Node selection strategy for branch and bound.

        'Node Selection' = 'Best of all'
          Large tree search; this method is the slowest as it solves all
          subproblem QPs independently.

        'Node Selection' = 'Best of two'
          Uses warm starts and less memory.

        'Node Selection' = 'Depth first'
          Uses more warm starts. If warm starts are applied, they can
          speed up the solution of mixed integer subproblems
          significantly when solving almost identical QPs.

    'Non Monotone' : int
        Default = 10

        Maximum number of successive iterations considered for the
        non-monotone trust region algorithm, allowing the penalty
        function to increase.

    'Objective Scale Bound' : float
        Default = 1.0

        When 'Scale Objective Function' > 0 internally scale absolute
        function values greater than 1.0 or 'Objective Scale Bound'.

    'Penalty' : float
        Default = 1000.0

        Initial penalty option, sigma.

    'Penalty Factor' : float
        Default = 10.0

        Factor for increasing penalty option sigma when the trust
        regions cannot be enlarged at a trial step.

    'Print Level' : int
        Default = 0

        Specifies the desired output level of printing.

        'Print Level' = '0'
          No output.

        'Print Level' = '1'
          Final convergence analysis.

        'Print Level' = '2'
          One line of intermediate results per iteration.

        'Print Level' = '3'
          Detailed information printed per iteration.

    'QP Accuracy' : float
        Default = 1.0e-10

        Termination tolerance of the relaxed quadratic program
        subproblems.

    'Scale Continuous Variables' : str
        Default = 'Yes'

        Internally scale continuous variables values.

    'Scale Objective Function' : int
        Default = 1

        Internally scale objective function values.

        'Scale Objective Function' = '0'
          No scaling.

        'Scale Objective Function' = '1'
          Scale absolute values greater than 'Objective Scale Bound'.

    'Warm Starts' : int
        Default = 100

        Maximum number of warm starts within the mixed integer QP
        solver, see 'Node Selection'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            On entry, `nclin` = *<value>*.

            Constraint: `nclin` >= 0.

        (`errno` 3)
            On entry, `ncnln` = *<value>*.

            Constraint: `ncnln` >= 0.

        (`errno` 5)
            On entry, `bl`[*<value>*] > `bu`[*<value>*].

            Constraint: `bl`[i-1] <= `bu`[i-1], for i = 1,2,...,`n`.

        (`errno` 6)
            On entry, `varcon`[*<value>*] = *<value>*.

            Constraint: `varcon`[i-1] = 0, 1 or 2, for i = 1,2,...,`n`.

        (`errno` 7)
            On entry, `varcon`[*<value>*] = *<value>*.

            Constraint: `varcon`[i-1] = 3 or 4, for i =
            `n`+1,...,`n`+`nclin`+`ncnln`.

        (`errno` 8)
            The supplied `objfun` returned a NaN value.

        (`errno` 9)
            The supplied `confun` returned a NaN value.

        (`errno` 10)
            On entry, the option arrays `comm`['iopts'] and
            `comm`['opts'] have either not been initialized or been
            corrupted.

        (`errno` 11)
            On entry, there are no binary or integer variables.

        (`errno` 12)
            On entry, linear equality constraints do not precede linear
            inequality constraints.

        (`errno` 13)
            On entry, nonlinear equality constraints do not precede
            nonlinear inequality constraints.

        (`errno` 81)
            One or more objective gradients appear to be incorrect.

        (`errno` 91)
            One or more constraint gradients appear to be incorrect.

        (`errno` 1002)
            More than the maximum number of feasible steps without
            improvement in the objective function. If the maximum number
            of feasible steps is small, say less than 5, try increasing
            it. Option 'Feasible Steps' = *<value>*.

        (`errno` 1003)
            Penalty parameter tends to infinity in an underlying
            mixed-integer quadratic program; the problem may be
            infeasible. If sigma is relatively low value, try a higher
            one, for example 10^20.

            Option 'Penalty' = *<value>*.

        (`errno` 1004)
            Termination at an infeasible iterate; if the problem is
            feasible, try a different starting value.

        (`errno` 1005)
            Termination with zero integer trust region for integer
            variables; try a different starting value.

            Option 'Integer Trust Radius' = *<value>*.

        (`errno` 1008)
            The optimization failed due to numerical difficulties. Set
            option 'Print Level' = 3 for more information.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1001)
            On entry, `maxit` = *<value>*. Exceeded the maximum number
            of iterations.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            The optimization halted because you set `mode` negative in
            `objfun` or `mode` negative in `confun`, to *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.mip.sqp_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.sqp'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_confun_name = fun_name + ':confun'

    confunh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, mode.
        utils._BYREF_INT, # Input, ncnln.
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, varcon[lvarcon].
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, c[ncnln].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, cjac[ncnln, n].
        utils._BYREF_INT, # Input, nstate.
        utils._BYREF_INT, # Input, lvarcon.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    if confun is None:
        confun_auxil_ctypes = utils._EngineState._get_symbol('h02ddmt')
        confun_auxil_ctypes.restype = None
        confun_auxil_ctypes.argtypes = (
            utils._BYREF_ENGINEDATA,
            utils._BYREF_INT, # Input/Output, mode.
            utils._BYREF_INT, # Input, ncnln.
            utils._BYREF_INT, # Input, n.
            utils._EngineIntArrayType.c_type, # Input, varcon[lvarcon].
            utils._EngineFloat64ArrayType.c_type, # Input, x[n].
            utils._EngineFloat64ArrayType.c_type, # Output, c[ncnln].
            utils._EngineFloat64ArrayType.c_type, # Input/Output, cjac[ncnln, n].
            utils._BYREF_INT, # Input, nstate.
            utils._BYREF_INT, # Input, lvarcon.
            utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
            utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        )

    def cb_confunh(_confun, en_data, mode, ncnln, n, varcon, x, c, cjac, nstate, lvarcon, iuser, ruser):
        """Helper for confun."""
        try:
            if confun is None:
                confun_auxil_ctypes(
                    en_data,
                    mode,
                    ncnln,
                    n,
                    varcon,
                    x,
                    c,
                    cjac,
                    nstate,
                    lvarcon,
                    iuser,
                    ruser,
                )
                return
            _mode_confunh = utils._EngineIntScalarType.to_py(mode.contents)
            _ncnln_confunh = utils._EngineIntScalarType.to_py(ncnln.contents)
            _n_confunh = utils._EngineIntScalarType.to_py(n.contents)
            _lvarcon_confunh = utils._EngineIntScalarType.to_py(lvarcon.contents)
            _varcon_confunh = utils._EngineIntArrayType.from_pointer(
                varcon,
                (_lvarcon_confunh,),
            )
            _x_confunh = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_n_confunh,),
            )
            _c_confunh = utils._EngineFloat64ArrayType.from_pointer(
                c,
                (_ncnln_confunh,),
                can_reference_contents=False,
            )
            _cjac_confunh = utils._EngineFloat64ArrayType.from_pointer(
                cjac,
                (_ncnln_confunh, _n_confunh),
                sorder=_m_sorder,
            )
            _nstate_confunh = utils._EngineIntScalarType.to_py(nstate.contents)
            if data is not None:
                confun(
                    _mode_confunh,
                    _ncnln_confunh,
                    _n_confunh,
                    _varcon_confunh,
                    _x_confunh,
                    _c_confunh,
                    _cjac_confunh,
                    _nstate_confunh,
                    data,
                )
            else:
                confun(
                    _mode_confunh,
                    _ncnln_confunh,
                    _n_confunh,
                    _varcon_confunh,
                    _x_confunh,
                    _c_confunh,
                    _cjac_confunh,
                    _nstate_confunh,
                )
            utils._unlink_pointer(varcon)
            utils._unlink_pointer(x)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _c_confunh, {'fun_name': cb_confun_name, 'returned': 'c'}, c,
            )
            utils._unlink_pointer(c)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _cjac_confunh, {'fun_name': cb_confun_name, 'returned': 'cjac'}, cjac,
            )
            utils._unlink_pointer(cjac)
        except:
            excs.append(_sys.exc_info())
            mode.contents.value = -1
            if excs[-1][0] is not utils.UserCallbackTerminate:
                en_data.contents.hlperr = utils._IERR_HELPER

    cb_objfun_name = fun_name + ':objfun'

    objfunh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, mode.
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, varcon[lvarcon].
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Output, objmip.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, objgrd[n].
        utils._BYREF_INT, # Input, nstate.
        utils._BYREF_INT, # Input, lvarcon.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_objfunh(_objfun, en_data, mode, n, varcon, x, objmip, objgrd, nstate, lvarcon, _iuser, _ruser):
        """Helper for objfun."""
        try:
            _mode_objfunh = utils._EngineIntScalarType.to_py(mode.contents)
            _n_objfunh = utils._EngineIntScalarType.to_py(n.contents)
            _lvarcon_objfunh = utils._EngineIntScalarType.to_py(lvarcon.contents)
            _varcon_objfunh = utils._EngineIntArrayType.from_pointer(
                varcon,
                (_lvarcon_objfunh,),
            )
            _x_objfunh = utils._EngineFloat64ArrayType.from_pointer(
                x,
                (_n_objfunh,),
            )
            _objgrd_objfunh = utils._EngineFloat64ArrayType.from_pointer(
                objgrd,
                (_n_objfunh,),
            )
            _nstate_objfunh = utils._EngineIntScalarType.to_py(nstate.contents)
            if data is not None:
                _objmip_objfunh = objfun(
                    _mode_objfunh,
                    _n_objfunh,
                    _varcon_objfunh,
                    _x_objfunh,
                    _objgrd_objfunh,
                    _nstate_objfunh,
                    data,
                )
            else:
                _objmip_objfunh = objfun(
                    _mode_objfunh,
                    _n_objfunh,
                    _varcon_objfunh,
                    _x_objfunh,
                    _objgrd_objfunh,
                    _nstate_objfunh,
                )
            utils._unlink_pointer(varcon)
            utils._unlink_pointer(x)
            utils._EngineFloat64ScalarType(
                _objmip_objfunh, {'fun_name': cb_objfun_name, 'returned': 'objmip'},
            ).refresh_ctypes_buffer(objmip)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _objgrd_objfunh, {'fun_name': cb_objfun_name, 'returned': 'objgrd'}, objgrd,
            )
            utils._unlink_pointer(objgrd)
        except:
            excs.append(_sys.exc_info())
            mode.contents.value = -1
            if excs[-1][0] is not utils.UserCallbackTerminate:
                en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('h02daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nclin.
        utils._BYREF_INT, # Input, ncnln.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, *].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input, d[nclin].
        utils._EngineFloat64ArrayType.c_type, # Output, ax[nclin].
        utils._EngineFloat64ArrayType.c_type, # Input, bl[n].
        utils._EngineFloat64ArrayType.c_type, # Input, bu[n].
        utils._EngineIntArrayType.c_type, # Input, varcon[n+nclin+ncnln].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._UNUSED_FUNCTYPE,
        confunh,
        utils._EngineFloat64ArrayType.c_type, # Output, c[ncnln].
        utils._EngineFloat64ArrayType.c_type, # Output, cjac[ncnln, n].
        utils._UNUSED_FUNCTYPE,
        objfunh,
        utils._EngineFloat64ArrayType.c_type, # Output, objgrd[n].
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, acc.
        utils._BYREF_FLOAT64, # Output, objmip.
        utils._EngineIntArrayType.c_type, # Input, iopts[liopts].
        utils._BYREF_INT, # Input, liopts.
        utils._EngineFloat64ArrayType.c_type, # Input, opts[lopts].
        utils._BYREF_INT, # Input, lopts.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (cjac, 'cjac')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nclin = utils._EngineIntScalarType(
        nclin, {'fun_name': fun_name, 'entity': 'nclin'},
    ).to_ctypes()
    _ncnln = utils._EngineIntScalarType(
        ncnln, {'fun_name': fun_name, 'entity': 'ncnln'},
    ).to_ctypes()
    _lda = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)), {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _nclin_py = utils._EngineIntScalarType.to_py(_nclin)
    if _nclin_py > 0:
        if a is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'a'},
                'must not be None when nclin > 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _nclin_py > 0:
        _a_dim2 = _n_py
    else:
        _a_dim2 = 0
    exp_shape_a = (_nclin_py, _a_dim2)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=exp_shape_a,
        mutual_sorder=_m_sorder,
        check_contents=_nclin_py > 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_nclin_py > 0)
    if _nclin_py > 0:
        if d is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'd'},
                'must not be None when nclin > 0',
            )
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_nclin_py,),
        check_contents=_nclin_py > 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_nclin_py > 0)
    if _nclin_py > 0:
        if ax is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ax'},
                'must not be None when nclin > 0',
            )
    _ax_cls = utils._EngineFloat64ArrayType(
        ax, {'fun_name': fun_name, 'entity': 'ax'},
        exp_shape=(_nclin_py,),
        check_contents=False,
        allow_none=True,
    )
    _ax = _ax_cls.to_ctypes(can_reference_contents=False)
    _bl = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _bu = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ncnln_py = utils._EngineIntScalarType.to_py(_ncnln)
    _varcon = utils._EngineIntArrayType(
        varcon, {'fun_name': fun_name, 'entity': 'varcon'},
        exp_shape=(_n_py+_nclin_py+_ncnln_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    if _ncnln_py != 0:
        if confun is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'confun'},
                'must not be None when ncnln != 0',
            )
    utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = utils._UNUSED_CALLBACK
    _confunh = utils._cast_callback(
        confunh, cb_confunh,
        {'fun_name': fun_name, 'entity': 'confun'},
    )
    if _ncnln_py > 0:
        if c is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when ncnln > 0',
            )
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ncnln_py,),
        check_contents=False,
        allow_none=True,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    if _ncnln_py > 0:
        if cjac is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'cjac'},
                'must not be None when ncnln > 0',
            )
    _cjac_cls = utils._EngineFloat64ArrayType(
        cjac, {'fun_name': fun_name, 'entity': 'cjac'},
        exp_shape=(_ncnln_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _cjac = _cjac_cls.to_ctypes(can_reference_contents=False)
    utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = utils._UNUSED_CALLBACK
    _objfunh = utils._cast_callback(
        objfunh, cb_objfunh,
        {'fun_name': fun_name, 'entity': 'objfun'},
    )
    _objgrd_cls = utils._EngineFloat64ArrayType(
        objgrd, {'fun_name': fun_name, 'entity': 'objgrd'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _objgrd = _objgrd_cls.to_ctypes(can_reference_contents=False)
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _acc = utils._EngineFloat64ScalarType(
        acc, {'fun_name': fun_name, 'entity': 'acc'},
    ).to_ctypes()
    _objmip = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iopts', 'opts'),
    )
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    _iopts = comm['iopts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _opts = comm['opts']
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _n,
        _nclin,
        _ncnln,
        _a,
        _lda,
        _d,
        _ax,
        _bl,
        _bu,
        _varcon,
        _x,
        _confun,
        _confunh,
        _c,
        _cjac,
        _objfun,
        _objfunh,
        _objgrd,
        _maxit,
        _acc,
        _objmip,
        _iopts,
        _liopts,
        _opts,
        _lopts,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py < 0 and _ifmt == 1515:
            _err_class = utils._UserCallbackExit(
                fun_name=fun_name,
                msg_lines=[
                    'The optimization halted because you set mode negative in objfun or',
                    'mode negative in confun, to {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
                cb_excs=excs,
            )
        elif _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nclin = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nclin >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncnln = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncnln >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bl[{:s}] > bu[{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: bl[i-1] <= bu[i-1], for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, varcon[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: varcon[i-1] = 0, 1 or 2, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, varcon[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: varcon[i-1] = 3 or 4, for i = n+1,...,n+nclin+ncnln.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The supplied objfun returned a NaN value.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The supplied confun returned a NaN value.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the option arrays comm[\'iopts\'] and comm[\'opts\'] have either',
                    'not been initialized or been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, there are no binary or integer variables.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, linear equality constraints do not precede linear inequality',
                    'constraints.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nonlinear equality constraints do not precede nonlinear inequality',
                    'constraints.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or more objective gradients appear to be incorrect.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or more constraint gradients appear to be incorrect.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1001 and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Exceeded the maximum number of iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1002 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'More than the maximum number of feasible steps without improvement in the',
                    'objective function. If the maximum number of feasible steps is small, say',
                    'less than 5, try increasing it.',
                    'Option \'Feasible Steps\' = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1003 and _ifmt == 1003:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Penalty parameter tends to infinity in an underlying mixed-integer quadratic',
                    'program; the problem may be infeasible.',
                    'If sigma is relatively low value, try a higher one, for example 10^20.',
                    'Option \'Penalty\' = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1004 and _ifmt == 1004:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Termination at an infeasible iterate; if the problem is feasible,',
                    'try a different starting value.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1005 and _ifmt == 1005:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Termination with zero integer trust region for integer variables;',
                    'try a different starting value.',
                    'Option \'Integer Trust Radius\' = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1008 and _ifmt == 1008:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The optimization failed due to numerical difficulties.',
                    'Set option \'Print Level\' = 3 for more information.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        if ax is not None:
            _ax_cls.refresh_py(_ax)
        _x_cls.refresh_py(_x)
        if c is not None:
            _c_cls.refresh_py(_c)
        if cjac is not None:
            _cjac_cls.refresh_py(_cjac)
        _objgrd_cls.refresh_py(_objgrd)
        _objmip_py = utils._EngineFloat64ScalarType.to_py(_objmip)
        return_data = _objmip_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['objmip'],
            )(
                _objmip_py,
            )
    _err_class.manage()
    return return_data

def optset(optstr, comm):
    u"""
    Option setting routine for ``sqp``.

    ``optset`` either initializes or resets the option arrays or sets a
    single option for supported problem solving functions in submodule
    ``mip``.

    Currently, only ``sqp`` is supported.

    For full information please refer to the NAG Library document for
    h02zk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02zkf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option to be set.

        Initialize = function name

          Initialize the option arrays `comm`['iopts'] and
          `comm`['opts'] for use with function function name, where
          function name is the name associated with the function of
          interest.

        Defaults

          Resets all options to their default values.

        option = optval

          See [Optional Parameters] for details of valid values for
          option and optval. The equals sign (=) delimiter must be used
          to separate the option from its optval value.

        `optstr` is case insensitive.

        Each token in the option and optval component must be separated
        by at least one space.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 12)
            On entry, the expected delimiter '=' was not found in
            `optstr`: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to an
            integer: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to a real:
            `optstr` = *<value>*.

        (`errno` 14)
            On entry, attempting to initialize the option arrays but
            specified function name was not valid: name = *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 17)
            On entry, the optval supplied for the character option is
            not valid.

            `optstr` = *<value>*.

        (`errno` 21)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

    See Also
    --------
    :meth:`naginterfaces.library.examples.mip.sqp_ex.main`
    """

    from ..base import _nag_e_consts

    excs = []
    fun_name = 'naginterfaces.base.mip.optset'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h02zkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._EngineIntArrayType.c_type, # Input/Output, iopts[liopts].
        utils._BYREF_INT, # Input, liopts.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, opts[lopts].
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('iopts', 'opts'),
    )
    if _comm_needs_init:
        if _comm_needs_init:
            _iopts_dim1 = 200
        else:
            _iopts_dim1 = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
        _iopts = utils._EngineIntArrayType.empty_ctypes(_iopts_dim1)
        comm['iopts'] = _iopts
    else:
        _iopts = comm['iopts']
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    if _comm_needs_init:
        if _comm_needs_init:
            _opts_dim1 = 100
        else:
            _opts_dim1 = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
        _opts = utils._EngineFloat64ArrayType.empty_ctypes(_opts_dim1)
        comm['opts'] = _opts
    else:
        _opts = comm['opts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _optstr,
        _iopts,
        _liopts,
        _opts,
        _lopts,
        _errbuf,
        _ierr,
        _optstr._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _ierr_py == _nag_e_consts.IERR_OPT_UNKNOWN and _ifmt == _nag_e_consts.IFMT_OPT_UNKNOWN:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the option in optstr was not recognized:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_NO_EQUAL and _ifmt == _nag_e_consts.IFMT_OPT_NO_EQUAL:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the expected delimiter \'=\' was not found in optstr:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_NON_NUMERIC and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert the specified optval to an integer:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_NON_NUMERIC and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert the specified optval to a real:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_RNAME and _ifmt == _nag_e_consts.IFMT_OPT_RNAME:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, attempting to initialize the option arrays but',
                    'specified function name was not valid:',
                    'name = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_IVALUE and _ifmt == _nag_e_consts.IERR_OPT_IVALUE:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the integer option is not',
                    'valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_RVALUE and _ifmt == _nag_e_consts.IERR_OPT_RVALUE:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the real option is not valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPT_CVALUE and _ifmt == _nag_e_consts.IERR_OPT_CVALUE:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the character option is not',
                    'valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_NOT_INIT and _ifmt == _nag_e_consts.IFMT_OPTS_NOT_INIT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT and _ifmt == _nag_e_consts.IFMT_OPTS_NOT_LINKED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT and _ifmt == _nag_e_consts.IFMT_OPTS_CORRUPT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def optget(optstr, comm):
    u"""
    Option getting routine for ``sqp``.

    ``optget`` is used to query the value of options available to
    supported problem solving functions in submodule ``mip``.

    Currently, only ``sqp`` is supported.

    For full information please refer to the NAG Library document for
    h02zl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h02zlf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option whose current value is required.
        See [Optional Parameters] for information on valid options. In
        addition, the following is a valid option:

        Identify

          ``optget`` returns in `cvalue` the function name supplied to
          ``optset`` when the option arrays `comm`['iopts'] and
          `comm`['opts'] were initialized.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``optset``.

    Returns
    -------
    ivalue : int
        If the option supplied in `optstr` is an integer valued
        parameter, `ivalue` will hold its current value.

    rvalue : float
        If the option supplied in `optstr` is a real valued parameter,
        `rvalue` will hold its current value.

    cvalue : str
        If the option supplied in `optstr` is a character valued
        parameter, `cvalue` will hold its current value. `cvalue` will
        also contain additional information for some integer and real
        valued parameters, as indicated by `optype`.

    optype : int
        Indicates whether the option supplied in `optstr` is an integer,
        real or character valued parameter and hence which of `ivalue`,
        `rvalue` or `cvalue` holds the current value.

        `optype` = 1

          `optstr` is an integer valued option, its current value has
          been returned in `ivalue`.

        `optype` = 2

          `optstr` is a real valued option, its current value has been
          returned in `rvalue`.

        `optype` = 3

          `optstr` is a character valued option, its current value has
          been returned in `cvalue`.

        `optype` = 4

          `optstr` is an integer valued option, its current value has
          been returned in `ivalue`. Additional information has been
          returned in `cvalue`.

        `optype` = 5

          `optstr` is a real valued option, its current value has been
          returned in `rvalue`. Additional information has been returned
          in `cvalue`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` has not been recognized.

        (`errno` 61)
            The arrays `comm`['iopts'] and `comm`['opts'] have either
            not been initialized, have become corrupted, or are not
            compatible with this option setting function.
    """

    from ..base import _nag_e_consts
    from ..base import _nag_h_consts

    excs = []
    fun_name = 'naginterfaces.base.mip.optget'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h02zlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._BYREF_INT, # Output, ivalue.
        utils._BYREF_FLOAT64, # Output, rvalue.
        utils._BYREF_STRING, # Output, cvalue(len=*).
        utils._BYREF_INT, # Output, optype.
        utils._EngineIntArrayType.c_type, # Input, iopts[liopts].
        utils._BYREF_INT, # Input, liopts.
        utils._EngineFloat64ArrayType.c_type, # Input, opts[lopts].
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of cvalue.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _ivalue = utils._EngineIntScalarType.empty_ctypes()
    _rvalue = utils._EngineFloat64ScalarType.empty_ctypes()
    _cvalue = utils._EngineCharScalarType.empty_ctypes(chrlen=32)
    _optype = utils._EngineIntScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iopts', 'opts'),
    )
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    _iopts = comm['iopts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _opts = comm['opts']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _optstr,
        _ivalue,
        _rvalue,
        _cvalue,
        _optype,
        _iopts,
        _liopts,
        _opts,
        _lopts,
        _errbuf,
        _ierr,
        _optstr._length_,
        _cvalue._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_e_consts.IERR_OPT_UNKNOWN and _ifmt == _nag_e_consts.IFMT_OPT_UNKNOWN:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the option in optstr has not been recognized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER and _ifmt == _nag_e_consts.IFMT_OPTS_NOT_INIT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The arrays comm[\'iopts\'] and comm[\'opts\'] have either not been initialized,',
                    'have become corrupted, or are not compatible with this option',
                    'setting function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER and _ifmt == _nag_e_consts.IFMT_OPTS_TOO_SMALL:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The arrays comm[\'iopts\'] and comm[\'opts\'] have either not been initialized,',
                    'have become corrupted, or are not compatible with this option',
                    'setting function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER and _ifmt == _nag_e_consts.IFMT_OPT_NOMATCH:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The arrays comm[\'iopts\'] and comm[\'opts\'] have either not been initialized,',
                    'have become corrupted, or are not compatible with this option',
                    'setting function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER and _ifmt == _nag_e_consts.IFMT_OPTS_NOT_LINKED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The arrays comm[\'iopts\'] and comm[\'opts\'] have either not been initialized,',
                    'have become corrupted, or are not compatible with this option',
                    'setting function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER and _ifmt == _nag_e_consts.IFMT_OPTS_CORRUPT:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The arrays comm[\'iopts\'] and comm[\'opts\'] have either not been initialized,',
                    'have become corrupted, or are not compatible with this option',
                    'setting function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ivalue_py = utils._EngineIntScalarType.to_py(_ivalue)
        _rvalue_py = utils._EngineFloat64ScalarType.to_py(_rvalue)
        _optype_py = utils._EngineIntScalarType.to_py(_optype)
        if _optype_py == _nag_e_consts.OPT_TYPE_IDENTIFY:
            _optype_py = _nag_e_consts.OPT_TYPE_CHAR
            if _ivalue_py == _nag_h_consts.MINLP_SOLVER_ID:
                _cvalue_py = 'h02da'
            else:
                utils._EngineErrorExit(
                    fun_name=fun_name,
                    msg_lines=[
                        'The communication structure comm has either not been initialized,',
                        'has become corrupted, or is not compatible with this option',
                        'setting function.',
                    ],
                    errcodes=(_nag_e_consts.IERR_OPTS_CORRUPT_GETTER,),
                ).manage()
            _ivalue_py = 0
        else:
            _cvalue_py = utils._EngineCharScalarType.to_py(_cvalue)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ivalue', 'rvalue', 'cvalue', 'optype'],
        )(
            _ivalue_py,
            _rvalue_py,
            _cvalue_py,
            _optype_py,
        )
    _err_class.manage()
    return return_data

def transportation(kost, ma, mb, m, k15, maxit, k6, k8, k11, k12):
    u"""
    Transportation problem, modified 'stepping stone' method.

    ``transportation`` solves the classical Transportation ('Hitchcock')
    problem.

    For full information please refer to the NAG Library document for
    h03ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h03abf.html

    Parameters
    ----------
    kost : int, shape (`ma`, `mb`)
        The coefficients c_{ij}, for j = 1,2,...,m_b, for i =
        1,2,...,m_a.

    ma : int
        The number of sources, m_a.

    mb : int
        The number of destinations, m_b.

    m : int
        The value of m_a+m_b.

    k15 : int, shape (`m`), modified in place
        On entry: `k15`[i] must be set to the availabilities A_i, for i
        = 1,2,...,m_a; and `k15`[m_a+j] must be set to the requirements
        B_j, for j = 1,2,...,m_b.

        On exit: the contents of `k15` are undefined.

    maxit : int
        The maximum number of iterations allowed.

    k6 : int, shape (`m`), modified in place
        On exit: `k6`[k-1], for k = 1,2,...,m_a+m_b-1, contains the
        source indices of the optimal solution (see `k11`).

    k8 : int, shape (`m`), modified in place
        On exit: `k8`[k-1], for k = 1,2,...,m_a+m_b-1, contains the
        destination indices of the optimal solution (see `k11`).

    k11 : int, shape (`m`), modified in place
        On exit: `k11`[k-1], for k = 1,2,...,m_a+m_b-1, contains the
        optimal quantities x_{ij} which, sent from source i = `k6`[k-1]
        to destination j = `k8`[k-1], minimize z.

    k12 : int, shape (`m`), modified in place
        On exit: `k12`[k-1], for k = 1,2,...,m_a+m_b-1, contains the
        unit cost c_{ij} associated with the route from source i =
        `k6`[k-1] to destination j = `k8`[k-1].

    Returns
    -------
    numit : int
        The number of iterations performed.

    z : float
        The value of the minimized total cost.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, the sum of the availabilities does not equal the
            sum of the requirements.

        (`errno` 2)
            During computations `maxit` has been exceeded.

        (`errno` 3)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 4)
            On entry, `m` = *<value>*, `ma` = *<value>* and `mb` =
            *<value>*.

            Constraint: `m` = `ma`+`mb`.

        (`errno` 4)
            On entry, `ma` = *<value>* and ldkost = *<value>*.

            Constraint: `ma` <= ldkost.

        (`errno` 4)
            On entry, `mb` = *<value>*.

            Constraint: `mb` > 0.

        (`errno` 4)
            On entry, `ma` = *<value>*.

            Constraint: `ma` > 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.transportation'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h03abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, kost[ldkost, mb].
        utils._BYREF_INT, # Input, ldkost.
        utils._BYREF_INT, # Input, ma.
        utils._BYREF_INT, # Input, mb.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input/Output, k15[m].
        utils._BYREF_INT, # Input, maxit.
        utils._EngineIntArrayType.c_type, # Workspace, k7.
        utils._EngineIntArrayType.c_type, # Workspace, k9.
        utils._BYREF_INT, # Output, numit.
        utils._EngineIntArrayType.c_type, # Output, k6[m].
        utils._EngineIntArrayType.c_type, # Output, k8[m].
        utils._EngineIntArrayType.c_type, # Output, k11[m].
        utils._EngineIntArrayType.c_type, # Output, k12[m].
        utils._BYREF_FLOAT64, # Output, z.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((kost, 'kost'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldkost_py = utils._EngineIntArrayType.get_nd(kost, {'fun_name': fun_name, 'entity': 'kost'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldkost = utils._EngineIntScalarType(
        _ldkost_py, {'fun_name': fun_name, 'entity': '_ldkost'},
    ).to_ctypes()
    _mb = utils._EngineIntScalarType(
        mb, {'fun_name': fun_name, 'entity': 'mb'},
    ).to_ctypes()
    _ma = utils._EngineIntScalarType(
        ma, {'fun_name': fun_name, 'entity': 'ma'},
    ).to_ctypes()
    _ma_py = utils._EngineIntScalarType.to_py(_ma)
    _mb_py = utils._EngineIntScalarType.to_py(_mb)
    _kost = utils._EngineIntArrayType(
        kost, {'fun_name': fun_name, 'entity': 'kost'},
        exp_shape=(_ma_py, _mb_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _k15_cls = utils._EngineIntArrayType(
        k15, {'fun_name': fun_name, 'entity': 'k15'},
        exp_shape=(_m_py,),
    )
    _k15 = _k15_cls.to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _k7 = utils._EngineIntArrayType.empty_ctypes(0)
    _k9 = utils._EngineIntArrayType.empty_ctypes(0)
    _numit = utils._EngineIntScalarType.empty_ctypes()
    _k6_cls = utils._EngineIntArrayType(
        k6, {'fun_name': fun_name, 'entity': 'k6'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _k6 = _k6_cls.to_ctypes(can_reference_contents=False)
    _k8_cls = utils._EngineIntArrayType(
        k8, {'fun_name': fun_name, 'entity': 'k8'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _k8 = _k8_cls.to_ctypes(can_reference_contents=False)
    _k11_cls = utils._EngineIntArrayType(
        k11, {'fun_name': fun_name, 'entity': 'k11'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _k11 = _k11_cls.to_ctypes(can_reference_contents=False)
    _k12_cls = utils._EngineIntArrayType(
        k12, {'fun_name': fun_name, 'entity': 'k12'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _k12 = _k12_cls.to_ctypes(can_reference_contents=False)
    _z = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _kost,
        _ldkost,
        _ma,
        _mb,
        _m,
        _k15,
        _maxit,
        _k7,
        _k9,
        _numit,
        _k6,
        _k8,
        _k11,
        _k12,
        _z,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the sum of the availabilities does not',
                    'equal the sum of the requirements.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'During computations maxit has been exceeded.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}, ma = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'mb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m = ma+mb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ma = {:s} and ldkost = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ma <= ldkost.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mb > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ma = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ma > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _k15_cls.refresh_py(_k15)
        _k6_cls.refresh_py(_k6)
        _k8_cls.refresh_py(_k8)
        _k11_cls.refresh_py(_k11)
        _k12_cls.refresh_py(_k12)
        _numit_py = utils._EngineIntScalarType.to_py(_numit)
        _z_py = utils._EngineFloat64ScalarType.to_py(_z)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['numit', 'z'],
        )(
            _numit_py,
            _z_py,
        )
    _err_class.manage()
    return return_data

def shortestpath(n, ns, ne, direct, nnz, d, irow, icol, path):
    u"""
    Shortest path problem, Dijkstra's algorithm.

    ``shortestpath`` finds the shortest path through a directed or
    undirected acyclic network using Dijkstra's algorithm.

    For full information please refer to the NAG Library document for
    h03ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h03adf.html

    Parameters
    ----------
    n : int
        n, the number of vertices.

    ns : int
        n_s and n_e, the labels of the first and last vertices,
        respectively, between which the shortest path is sought.

    ne : int
        n_s and n_e, the labels of the first and last vertices,
        respectively, between which the shortest path is sought.

    direct : bool
        Indicates whether the network is directed or undirected.

        `direct` = True

          The network is directed.

        `direct` = False

          The network is undirected.

    nnz : int
        The number of nonzero elements in the distance matrix D.

    d : float, shape (`nnz`)
        The nonzero elements of the distance matrix D, ordered by
        increasing row index and increasing column index within each
        row. More precisely, `d`[k] must contain the value of the
        nonzero element with indices (`irow`[k],`icol`[k]); this is the
        length of the arc from the vertex with label `irow`[k] to the
        vertex with label `icol`[k]. Elements with the same row and
        column indices are not allowed. If `direct` = False, then only
        those nonzero elements in the strict upper triangle of `d` need
        be supplied since d_{ij} = d_{ji}. (``sparse.real_gen_sort`` may
        be used to sort the elements of an arbitrarily ordered matrix
        into the required form. This is illustrated in [Example].)

    irow : int, shape (`nnz`), modified in place
        `irow`[k] and `icol`[k] must contain the row and column indices,
        respectively, for the nonzero element stored in `d`[k].

    icol : int, shape (`nnz`)
        `irow`[k] and `icol`[k] must contain the row and column indices,
        respectively, for the nonzero element stored in `d`[k].

    path : int, shape (`n`), modified in place
        On exit: contains details of the shortest path between the
        specified vertices n_s and n_e. More precisely, `ns` =
        `path`[0]→`path`[1]→...→`path`[p] = `ne` for some p <= n. The
        remaining (n-p) elements are set to zero.

    Returns
    -------
    splen : float
        The length of the shortest path between the specified vertices
        n_s and n_e.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>* and `ne` = *<value>*.

            Constraint: `ns` != `ne`.

        (`errno` 1)
            On entry, `ne` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `ne` <= `n`.

        (`errno` 1)
            On entry, `ns` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `ns` <= `n`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: if `direct` = False, 1 <= `nnz` <=
            `n`*(`n`-1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: if `direct` = True, 1 <= `nnz` <= `n`*(`n`-1).

        (`errno` 3)
            On entry, k = *<value>*, `irow`[k-1] = *<value>*,
            `icol`[k-1] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[k-1] <= `n`, 1 <= `icol`[k-1] <=
            `n`; `icol`[k-1] != `irow`[k-1] when `direct` = True.

        (`errno` 4)
            On entry, k = *<value>*, `irow`[k-1] = *<value>*,
            `icol`[k-1] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[k-1] < `icol`[k-1] <= `n` when
            `direct` = False.

        (`errno` 5)
            On entry, k = *<value>*, `d`[k-1] = *<value>*.

            Constraint: `d`[k-1] > 0.0.

        (`errno` 6)
            On entry, k = *<value>*, `irow`[k-1] = *<value>*,
            `irow`[k-2] = *<value>*, `icol`[k-1] = *<value>*,
            `icol`[k-2] = *<value>*.

            Constraints: `irow`[k-2] < `irow`[k-1] or `irow`[k-2] =
            `irow`[k-1] and `icol`[k-2] < `icol`[k-1].

        (`errno` 7)
            On entry, k = *<value>* `irow`[k-1] = *<value>*, `irow`[k-2]
            = *<value>* `icol`[k-1] = *<value>*, `icol`[k-2] =
            *<value>*.

            Constraint: `irow`[k-1] != `irow`[k-2] or `icol`[k-1] !=
            `icol`[k-2].

        (`errno` 8)
            On entry, `ns` = *<value>* and `ne` = *<value>*.

            No connected network exists between vertices `ns` and `ne`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.shortestpath'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h03adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ns.
        utils._BYREF_INT, # Input, ne.
        utils._BYREF_LOGICAL, # Input, direct.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, d[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Output, splen.
        utils._EngineIntArrayType.c_type, # Output, path[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ne = utils._EngineIntScalarType(
        ne, {'fun_name': fun_name, 'entity': 'ne'},
    ).to_ctypes()
    _direct = utils._EngineLogicalScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _splen = utils._EngineFloat64ScalarType.empty_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _path_cls = utils._EngineIntArrayType(
        path, {'fun_name': fun_name, 'entity': 'path'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _path = _path_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _ns,
        _ne,
        _direct,
        _nnz,
        _d,
        _irow,
        _icol,
        _splen,
        _path,
        _iwork,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and ne = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns != ne.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ne = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ne <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ns <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if direct = False, 1 <= nnz <= n*(n-1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if direct = True, 1 <= nnz <= n*(n-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}, irow[k-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'icol[k-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: 1 <= irow[k-1] <= n, 1 <= icol[k-1] <= n; icol[k-1] != irow[k-1]',
                    'when direct = True.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}, irow[k-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'icol[k-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: 1 <= irow[k-1] < icol[k-1] <= n',
                    'when direct = False.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}, d[k-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: d[k-1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}, irow[k-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'irow[k-2] = {:s}, icol[k-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'icol[k-2] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[5]),
                    ),
                    'Constraints: irow[k-2] < irow[k-1] or',
                    'irow[k-2] = irow[k-1] and icol[k-2] < icol[k-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[k-1] = {:s}, irow[k-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'icol[k-1] = {:s}, icol[k-2] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[5]),
                    ),
                    'Constraint: irow[k-1] != irow[k-2] or icol[k-1] != icol[k-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and ne = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'No connected network exists between vertices ns and ne.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irow_cls.refresh_py(_irow)
        _path_cls.refresh_py(_path)
        _splen_py = utils._EngineFloat64ScalarType.to_py(_splen)
        return_data = _splen_py
    _err_class.manage()
    return return_data

def tsp_simann(nc, dm, bound, targc, path, alg_stats, statecomm):
    u"""
    Travelling Salesman Problem, simulated annealing.

    ``tsp_simann`` calculates an approximate solution to a symmetric
    travelling salesman problem using simulated annealing via a
    configuration free interface.

    For full information please refer to the NAG Library document for
    h03bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h03bbf.html

    Parameters
    ----------
    nc : int
        The number of cities. In the trivial cases `nc` = 1, 2 or 3, the
        function returns the optimal solution immediately with `tmode` =
        0 (provided the relevant distance matrix entries are not
        negative).

    dm : float, shape (`nc`, `nc`)
        The distance matrix; each `dm`[i-1,j-1] is the effective cost or
        weight between nodes i and j. Only the strictly upper half of
        the matrix is referenced.

    bound : float
        A lower bound on the solution. If the optimum is unknown set
        `bound` to zero or a negative value; the function will then
        calculate the minimum spanning tree for `dm` and use this as a
        lower bound (returned in `alg_stats`[5]). If an optimal value
        for the cost is known then this should be used for the lower
        bound. A detailed discussion of relaxations for lower bounds,
        including the minimal spanning tree, can be found in Reinelt.

    targc : float
        A measure of how close an approximation needs to be to the lower
        bound. The function terminates when a cost is found less than or
        equal to `bound`+`targc`. This argument is useful when an
        optimal value for the cost is known and supplied in `bound`. It
        may be sufficient to obtain a path that is close enough (in
        terms of cost) to the optimal path; this allows the algorithm to
        terminate at that point and avoid further computation in
        attempting to find a better path.

        If `targc` < 0, `targc` = 0 is assumed.

    path : int, shape (`nc`), modified in place
        On exit: the best path discovered by the simulation. That is,
        `path` contains the city indices in path order.

    alg_stats : float, shape (6), modified in place
        On exit: an array of metrics collected during the initial
        search. These could be used as a basis for future optimization.
        If an exception is raised, the elements of `alg_stats` are set
        to zero; the first five elements are also set to zero in the
        trival cases `nc` = 1, 2 or 3.

        `alg_stats`[0]

          Mean delta.

        `alg_stats`[1]

          Standard deviation of deltas.

        `alg_stats`[2]

          Cost at end of initial search phase.

        `alg_stats`[3]

          Best cost encountered during search phase.

        `alg_stats`[4]

          Initial system temperature. At the end of stage 1 of the
          algorithm, this is a function of the mean and variance of the
          deltas, and of the distance from best cost to the lower bound.
          It is a measure of the initial acceptance criteria for stage
          2. The larger this value, the more iterations it will take to
          geometrically reduce it during stage 2 until the system is
          cooled (below a threshold value).

        `alg_stats`[5]

          The lower bound used, which will be that computed internally
          when `bound` <= 0 on input. Subsequent calls with different
          random states can set `bound` to the value returned in
          `alg_stats`[5] to avoid recomputation of the minimal spanning
          tree.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``rand.init_repeat`` or ``rand.init_nonrepeat``.

    Returns
    -------
    cost : float
        The cost or weight of `path`.

    tmode : int
        The termination mode of the function:

        `tmode` = 0

          Optimal solution found, `cost` = `bound`.

        `tmode` = 1

          System temperature cooled. The algorithm returns a `path` and
          associated `cost` that does not attain, nor lie within `targc`
          of, the `bound`. This could be a sufficiently good
          approximation to the optimal `path`, particularly when
          `bound`+`targc` lies below the optimal `cost`.

        `tmode` = 2

          Halted by `cost` falling within the desired `targc` range of
          the `bound`.

        `tmode` = 3

          System stalled following lack of improvement.

        `tmode` = 4

          Initial search failed to find a single improvement (the
          solution could be optimal).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nc` = *<value>*.

            Constraint: `nc` >= 1.

        (`errno` 2)
            On entry, the strictly upper triangle of `dm` had a negative
            element.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.tsp_simann'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h03bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nc.
        utils._EngineFloat64ArrayType.c_type, # Input, dm[nc, nc].
        utils._BYREF_FLOAT64, # Input, bound.
        utils._BYREF_FLOAT64, # Input, targc.
        utils._EngineIntArrayType.c_type, # Output, path[nc].
        utils._BYREF_FLOAT64, # Output, cost.
        utils._BYREF_INT, # Output, tmode.
        utils._EngineFloat64ArrayType.c_type, # Output, alg_stats[6].
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dm, 'dm'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nc = utils._EngineIntScalarType(
        nc, {'fun_name': fun_name, 'entity': 'nc'},
    ).to_ctypes()
    _nc_py = utils._EngineIntScalarType.to_py(_nc)
    _dm = utils._EngineFloat64ArrayType(
        dm, {'fun_name': fun_name, 'entity': 'dm'},
        exp_shape=(_nc_py, _nc_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _bound = utils._EngineFloat64ScalarType(
        bound, {'fun_name': fun_name, 'entity': 'bound'},
    ).to_ctypes()
    _targc = utils._EngineFloat64ScalarType(
        targc, {'fun_name': fun_name, 'entity': 'targc'},
    ).to_ctypes()
    _path_cls = utils._EngineIntArrayType(
        path, {'fun_name': fun_name, 'entity': 'path'},
        exp_shape=(_nc_py,),
        check_contents=False,
    )
    _path = _path_cls.to_ctypes(can_reference_contents=False)
    _cost = utils._EngineFloat64ScalarType.empty_ctypes()
    _tmode = utils._EngineIntScalarType.empty_ctypes()
    _alg_stats_cls = utils._EngineFloat64ArrayType(
        alg_stats, {'fun_name': fun_name, 'entity': 'alg_stats'},
        exp_shape=(6,),
        check_contents=False,
    )
    _alg_stats = _alg_stats_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nc,
        _dm,
        _bound,
        _targc,
        _path,
        _cost,
        _tmode,
        _alg_stats,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nc = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nc_py),
                    ),
                    'Constraint: nc >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the strictly upper triangle of dm had a negative element.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _path_cls.refresh_py(_path)
        _alg_stats_cls.refresh_py(_alg_stats)
        _cost_py = utils._EngineFloat64ScalarType.to_py(_cost)
        _tmode_py = utils._EngineIntScalarType.to_py(_tmode)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['cost', 'tmode'],
        )(
            _cost_py,
            _tmode_py,
        )
    _err_class.manage()
    return return_data

def best_subset_given_size_revcomm(irevcm, mincr, m, ip, nbest, drop, lz, z, la, a, bscore, bz, mincnt, gamma, acc, comm, io_manager=None):
    u"""
    Best n subsets of size p (reverse communication).

    Given a set of m features and a scoring mechanism for any subset of
    those features, ``best_subset_given_size_revcomm`` selects the best
    n subsets of size p using a reverse communication branch and bound
    algorithm.

    For full information please refer to the NAG Library document for
    h05aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h05aaf.html

    Parameters
    ----------
    irevcm : int
        On initial entry: must be set to 0.

        On intermediate entry: `irevcm` must remain unchanged.

    mincr : int
        Flag indicating whether the scoring function f is increasing or
        decreasing.

        `mincr` = 1

          f(S_i) <= f(S_j), i.e., the subsets with the largest score
          will be selected.

        `mincr` = 0

          f(S_i) >= f(S_j), i.e., the subsets with the smallest score
          will be selected.

        For all S_j⊆Omega and S_i⊆S_j.

    m : int
        m, the number of features in the full feature set.

    ip : int
        p, the number of features in the subset of interest.

    nbest : int
        n, the maximum number of best subsets required. The actual
        number of subsets returned is given by `la` on final exit. If on
        final exit `la` != `nbest` then `errno` = 53 is returned.

    drop : int
        On initial entry: `drop` need not be set.

        On intermediate entry: `drop` must remain unchanged.

    lz : int
        On initial entry: `lz` need not be set.

        On intermediate entry: `lz` must remain unchanged.

    z : int, shape (`m`-`ip`), modified in place
        On initial entry: `z` need not be set.

        On intermediate exit: `z`[i-1], for i = 1,2,...,`lz`, contains
        the list of features which, along with those specified in `a`,
        define the subsets whose score is required. See `irevcm` for
        additional details.

        On intermediate entry: `z` must remain unchanged.

        On final exit: `z` is undefined.

    la : int
        On initial entry: `la` need not be set.

        On intermediate entry: `la` must remain unchanged.

    a : int, shape (max(`nbest`, `m`)), modified in place
        On initial entry: `a` need not be set.

        On intermediate exit: `a`[j-1], for j = 1,2,...,`la`, contains
        the list of features which, along with those specified in `z`,
        define the subsets whose score is required. See `irevcm` for
        additional details.

        On intermediate entry: `a` must remain unchanged.

        On final exit: `a` is undefined.

    bscore : float, shape (max(`nbest`, `m`)), modified in place
        On initial entry: `bscore` need not be set.

        On intermediate exit: `bscore` is undefined.

        On intermediate entry: `bscore`[j-1] must hold the score for the
        jth subset as described in `irevcm`.

        On final exit: holds the score for the `la` best subsets
        returned in `bz`.

    bz : int, shape (`m`-`ip`, `nbest`), modified in place
        On initial entry: `bz` need not be set.

        On intermediate exit: `bz` is used for storage between calls to
        ``best_subset_given_size_revcomm``.

        On intermediate entry: `bz` must remain unchanged.

        On final exit: the jth best subset is constructed by dropping
        the features specified in `bz`[i-1,j-1], for j = 1,2,...,`la`,
        for i = 1,2,...,`m`-`ip`, from the set of all features, Omega.
        The score for the jth best subset is given in `bscore`[j-1].

    mincnt : int
        k, the minimum number of times the effect of each feature, x_i,
        must have been observed before f(S-{x_i}) is estimated from f(S)
        as opposed to being calculated directly.

        If k = 0 then f(S-{x_i}) is never estimated.

        If `mincnt` < 0 then k is set to 1.

    gamma : float
        gamma, the scaling factor used when estimating scores. If
        `gamma` < 0 then gamma = 1 is used.

    acc : float, shape (2)
        A measure of the accuracy of the scoring function, f.

        Letting a_i = epsilon_1*|f(S_i)|+epsilon_2, then when confirming
        whether the scoring function is strictly increasing or
        decreasing (as described in `mincr`), or when assessing whether
        a node defined by subset S_i can be trimmed, then any values in
        the range f(S_i)+/-a_i are treated as being numerically
        equivalent.

        If 0 <= `acc`[0] <= 1 then epsilon_1 = `acc`[0], otherwise
        epsilon_1 = 0.

        If `acc`[1] >= 0 then epsilon_2 = `acc`[1], otherwise epsilon_2
        = 0.

        In most situations setting both epsilon_1 and epsilon_2 to zero
        should be sufficient.

        Using a nonzero value, when one is not required, can
        significantly increase the number of subsets that need to be
        evaluated.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    irevcm : int
        On intermediate exit: `irevcm` = 1 and before re-entry the
        scores associated with `la` subsets must be calculated and
        returned in `bscore`.

        The `la` subsets are constructed as follows:

        `drop` = 1

          The jth subset is constructed by dropping the features
          specified in the first `lz` elements of `z` and the single
          feature given in `a`[j-1] from the full set of features,
          Omega. The subset will therefore contain `m`-`lz`-1 features.

        `drop` = 0

          The jth subset is constructed by adding the features specified
          in the first `lz` elements of `z` and the single feature
          specified in `a`[j-1] to the empty set, ∅. The subset will
          therefore contain `lz`+1 features.

        In both cases the individual features are referenced by the
        integers 1 to `m` with 1 indicating the first feature, 2 the
        second, etc., for some arbitrary ordering of the features.

        The same ordering must be used in all calls to
        ``best_subset_given_size_revcomm``.

        If `la` = 0, the score for a single subset should be returned.

        This subset is constructed by adding or removing only those
        features specified in the first `lz` elements of `z`.

        If `lz` = 0, this subset will either be Omega or ∅.

        The score associated with the jth subset must be returned in
        `bscore`[j-1].

        On final exit: `irevcm` = 0, and the algorithm has terminated.

    drop : int
        On intermediate exit: flag indicating whether the intermediate
        subsets should be constructed by dropping features from the full
        set (`drop` = 1) or adding features to the empty set (`drop` =
        0). See `irevcm` for details.

        On final exit: `drop` is undefined.

    lz : int
        On intermediate exit: the number of features stored in `z`.

        On final exit: `lz` is undefined.

    la : int
        On intermediate exit: if `la` > 0, the number of subsets for
        which a score must be returned.

        If `la` = 0, the score for a single subset should be returned.

        See `irevcm` for additional details.

        On final exit: the number of best subsets returned.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0 or 1.

        (`errno` 21)
            On entry, `mincr` = *<value>*.

            Constraint: `mincr` = 0 or 1.

        (`errno` 22)
            `mincr` has changed between calls.

            On intermediate entry, `mincr` = *<value>*.

            On initial entry, `mincr` = *<value>*.

        (`errno` 31)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 32)
            `m` has changed between calls.

            On intermediate entry, `m` = *<value>*.

            On initial entry, `m` = *<value>*.

        (`errno` 41)
            On entry, `ip` = *<value>* and `m` = *<value>*.

            Constraint: 1 <= `ip` <= `m`.

        (`errno` 42)
            `ip` has changed between calls.

            On intermediate entry, `ip` = *<value>*.

            On initial entry, `ip` = *<value>*.

        (`errno` 51)
            On entry, `nbest` = *<value>*.

            Constraint: `nbest` >= 1.

        (`errno` 52)
            `nbest` has changed between calls.

            On intermediate entry, `nbest` = *<value>*.

            On initial entry, `nbest` = *<value>*.

        (`errno` 61)
            `drop` has changed between calls.

            On intermediate entry, `drop` = *<value>*.

            On initial entry, `drop` = *<value>*.

        (`errno` 71)
            `lz` has changed between calls.

            On entry, `lz` = *<value>*.

            On previous exit, `lz` = *<value>*.

        (`errno` 91)
            `la` has changed between calls.

            On entry, `la` = *<value>*.

            On previous exit, `la` = *<value>*.

        (`errno` 111)
            `bscore`[*<value>*] = *<value>*, which is inconsistent with
            the score for the parent node. Score for the parent node is
            *<value>*.

        (`errno` 131)
            `mincnt` has changed between calls.

            On intermediate entry, `mincnt` = *<value>*.

            On initial entry, `mincnt` = *<value>*.

        (`errno` 141)
            `gamma` has changed between calls.

            On intermediate entry, `gamma` = *<value>*.

            On initial entry, `gamma` = *<value>*.

        (`errno` 151)
            `acc`[0] has changed between calls.

            On intermediate entry, `acc`[0] = *<value>*.

            On initial entry, `acc`[0] = *<value>*.

        (`errno` 152)
            `acc`[1] has changed between calls.

            On intermediate entry, `acc`[1] = *<value>*.

            On initial entry, `acc`[1] = *<value>*.

        (`errno` 161)
            `comm`['icomm'] has been corrupted between calls.

        (`errno` 181)
            `comm`['rcomm'] has been corrupted between calls.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 53)
            On entry, `nbest` = *<value>*.

            But only *<value>* best subsets could be calculated.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.best_subset_given_size_revcomm'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h05aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._BYREF_INT, # Input, mincr.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, nbest.
        utils._BYREF_INT, # Input/Output, drop.
        utils._BYREF_INT, # Input/Output, lz.
        utils._EngineIntArrayType.c_type, # Input/Output, z[m-ip].
        utils._BYREF_INT, # Input/Output, la.
        utils._EngineIntArrayType.c_type, # Input/Output, a[max(nbest, m)].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, bscore[max(nbest, m)].
        utils._EngineIntArrayType.c_type, # Input/Output, bz[m-ip, nbest].
        utils._BYREF_INT, # Input, mincnt.
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._EngineFloat64ArrayType.c_type, # Input, acc[2].
        utils._EngineIntArrayType.c_type, # Input/Output, icomm[licomm].
        utils._BYREF_INT, # Input, licomm.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._BYREF_INT, # Input, monitor.
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_INT, # Input, mwidth.
        utils._BYREF_INT, # Input, csep.
        utils._BYREF_INT, # Input, cwidth.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((bz, 'bz'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    _mincr = utils._EngineIntScalarType(
        mincr, {'fun_name': fun_name, 'entity': 'mincr'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _nbest = utils._EngineIntScalarType(
        nbest, {'fun_name': fun_name, 'entity': 'nbest'},
    ).to_ctypes()
    _drop = utils._EngineIntScalarType(
        drop, {'fun_name': fun_name, 'entity': 'drop'},
    ).to_ctypes()
    _lz = utils._EngineIntScalarType(
        lz, {'fun_name': fun_name, 'entity': 'lz'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _z_cls = utils._EngineIntArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_m_py-_ip_py,),
    )
    _z = _z_cls.to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _nbest_py = utils._EngineIntScalarType.to_py(_nbest)
    _a_cls = utils._EngineIntArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(max(_nbest_py, _m_py),),
    )
    _a = _a_cls.to_ctypes()
    _bscore_cls = utils._EngineFloat64ArrayType(
        bscore, {'fun_name': fun_name, 'entity': 'bscore'},
        exp_shape=(max(_nbest_py, _m_py),),
    )
    _bscore = _bscore_cls.to_ctypes()
    _bz_cls = utils._EngineIntArrayType(
        bz, {'fun_name': fun_name, 'entity': 'bz'},
        exp_shape=(_m_py-_ip_py, _nbest_py),
        mutual_sorder=_m_sorder,
    )
    _bz = _bz_cls.to_ctypes()
    _mincnt = utils._EngineIntScalarType(
        mincnt, {'fun_name': fun_name, 'entity': 'mincnt'},
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _acc = utils._EngineFloat64ArrayType(
        acc, {'fun_name': fun_name, 'entity': 'acc'},
        exp_shape=(2,),
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('icomm', 'rcomm'),
    )
    _mincnt_py = utils._EngineIntScalarType.to_py(_mincnt)
    if _comm_needs_init:
        if _mincnt_py == 0:
            _icomm_dim1 = 2*max(_nbest_py, _m_py)+_m_py*(_m_py+2)+(_m_py+1)*max(_m_py-_ip_py, 1)+27
        else:
            _icomm_dim1 = 2*max(_nbest_py, _m_py)+_m_py*(_m_py+3)+(2*_m_py+1)*max(_m_py-_ip_py, 1)+25
        _icomm = utils._EngineIntArrayType.empty_ctypes(_icomm_dim1)
        comm['icomm'] = _icomm
    else:
        _icomm = comm['icomm']
    _licomm_py = utils._EngineIntArrayType.get_shape(comm['icomm'], {'fun_name': fun_name, 'entity': 'comm[\'icomm\']'}, exp_rank=1, validate=False)[0]
    _licomm = utils._EngineIntScalarType(
        _licomm_py, {'fun_name': fun_name, 'entity': '_licomm'},
    ).to_ctypes()
    if _comm_needs_init:
        if _mincnt_py == 0:
            _rcomm_dim1 = 9+_nbest_py+_m_py*max(_m_py-_ip_py, 1)
        else:
            _rcomm_dim1 = 8+_m_py+_nbest_py+_m_py*max(_m_py-_ip_py, 1)
        _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_rcomm_dim1)
        comm['rcomm'] = _rcomm
    else:
        _rcomm = comm['rcomm']
    _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, exp_rank=1, validate=False)[0]
    _lrcomm = utils._EngineIntScalarType(
        _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
    ).to_ctypes()
    _monitor = utils._EngineIntScalarType.c_type(0)
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _mwidth = utils._EngineIntScalarType.c_type(80)
    _csep = utils._EngineIntScalarType.c_type(5)
    _mwidth_py = utils._EngineIntScalarType.to_py(_mwidth)
    _csep_py = utils._EngineIntScalarType.to_py(_csep)
    _cwidth = utils._EngineIntScalarType(
        (_mwidth_py-_csep_py)//2, {'fun_name': fun_name, 'entity': '_cwidth'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _irevcm,
        _mincr,
        _m,
        _ip,
        _nbest,
        _drop,
        _lz,
        _z,
        _la,
        _a,
        _bscore,
        _bz,
        _mincnt,
        _gamma,
        _acc,
        _icomm,
        _licomm,
        _rcomm,
        _lrcomm,
        _monitor,
        _nadv,
        _mwidth,
        _csep,
        _cwidth,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 0 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mincr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mincr = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'mincr has changed between calls.',
                    'On intermediate entry, mincr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, mincr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'm has changed between calls.',
                    'On intermediate entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ip <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _ierr_fl = 41
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ip <= m.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 43 and _ifmt == 43:
            _ierr_fl = 42
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ip has changed between calls.',
                    'On intermediate entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nbest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nbest >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'nbest has changed between calls.',
                    'On intermediate entry, nbest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, nbest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 53 and _ifmt == 53:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nbest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'But only {:s} best subsets could be calculated.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'drop has changed between calls.',
                    'On intermediate entry, drop = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, drop = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'lz has changed between calls.',
                    'On entry, lz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On previous exit, lz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'la has changed between calls.',
                    'On entry, la = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On previous exit, la = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'bscore[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'which is inconsistent with the score for the parent node.',
                    'Score for the parent node is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _ierr_fl = 111
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'bscore[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'which is inconsistent with the score for the parent node.',
                    'Score for the parent node is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 131 and _ifmt == 131:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'mincnt has changed between calls.',
                    'On intermediate entry, mincnt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, mincnt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 141 and _ifmt == 141:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'gamma has changed between calls.',
                    'On intermediate entry, gamma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, gamma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 151 and _ifmt == 151:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'acc[0] has changed between calls.',
                    'On intermediate entry, acc[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, acc[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 152 and _ifmt == 152:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'acc[1] has changed between calls.',
                    'On intermediate entry, acc[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, acc[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 161 and _ifmt == 161:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'icomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 162 and _ifmt == 162:
            _ierr_fl = 161
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'icomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 181 and _ifmt == 181:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'rcomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _drop_py = utils._EngineIntScalarType.to_py(_drop)
        _lz_py = utils._EngineIntScalarType.to_py(_lz)
        _z_cls.refresh_py(_z)
        _la_py = utils._EngineIntScalarType.to_py(_la)
        _a_cls.refresh_py(_a)
        _bscore_cls.refresh_py(_bscore)
        _bz_cls.refresh_py(_bz)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['irevcm', 'drop', 'lz', 'la'],
        )(
            _irevcm_py,
            _drop_py,
            _lz_py,
            _la_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def best_subset_given_size(mincr, m, ip, nbest, bscore, bz, f, mincnt, gamma, acc, data=None, io_manager=None):
    u"""
    Best n subsets of size p (direct communication).

    Given a set of m features and a scoring mechanism for any subset of
    those features, ``best_subset_given_size`` selects the best n
    subsets of size p using a direct communication branch and bound
    algorithm.

    For full information please refer to the NAG Library document for
    h05ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/h/h05abf.html

    Parameters
    ----------
    mincr : int
        Flag indicating whether the scoring function f is increasing or
        decreasing.

        `mincr` = 1

          f(S_i) <= f(S_j), i.e., the subsets with the largest score
          will be selected.

        `mincr` = 0

          f(S_i) >= f(S_j), i.e., the subsets with the smallest score
          will be selected.

        For all S_j⊆Omega and S_i⊆S_j.

    m : int
        m, the number of features in the full feature set.

    ip : int
        p, the number of features in the subset of interest.

    nbest : int
        n, the maximum number of best subsets required. The actual
        number of subsets returned is given by `la` on final exit. If on
        final exit `la` != `nbest` then `errno` = 42 is returned.

    bscore : float, shape (`nbest`), modified in place
        On exit: holds the score for the `la` best subsets returned in
        `bz`.

    bz : int, shape (`m`-`ip`, `nbest`), modified in place
        On exit: the jth best subset is constructed by dropping the
        features specified in `bz`[i-1,j-1], for j = 1,2,...,`la`, for i
        = 1,2,...,`m`-`ip`, from the set of all features, Omega. The
        score for the jth best subset is given in `bscore`[j-1].

    f : callable f(m, drop, lz, z, la, a, score, data=None)
        `f` must evaluate the scoring function f.

        Parameters
        ~~~~~~~~~~
        m : int
            m = |Omega|, the number of features in the full feature set.

        drop : int
            Flag indicating whether the intermediate subsets should be
            constructed by dropping features from the full set (`drop` =
            1) or adding features to the empty set (`drop` = 0). See
            `score` for additional details.

        lz : int
            The number of features stored in `z`.

        z : int, ndarray, shape (`lz`)
            `z`[i-1], for i = 1,2,...,`lz`, contains the list of
            features which, along with those specified in `a`, define
            the subsets whose score is required. See `score` for
            additional details.

        la : int
            If `la` > 0, the number of subsets for which a score must be
            returned.

            If `la` = 0, the score for a single subset should be
            returned.

            See `score` for additional details.

        a : int, ndarray, shape (`la`)
            `a`[j-1], for j = 1,2,...,`la`, contains the list of
            features which, along with those specified in `z`, define
            the subsets whose score is required. See `score` for
            additional details.

        score : float, ndarray, shape (max(`la`, 1)), to be modified in
        place
            On exit: the value f(S_j), for j = 1,2,...,`la`, the score
            associated with the jth subset. S_j is constructed as
            follows:

            `drop` = 1

              S_j is constructed by dropping the features specified in
              the first `lz` elements of `z` and the single feature
              given in `a`[j-1] from the full set of features, Omega.
              The subset will therefore contain `m`-`lz`-1 features.

            `drop` = 0

              S_j is constructed by adding the features specified in the
              first `lz` elements of `z` and the single feature
              specified in `a`[j-1] to the empty set, ∅. The subset will
              therefore contain `lz`+1 features.

            In both cases the individual features are referenced by the
            integers 1 to `m` with 1 indicating the first feature, 2 the
            second, etc., for some arbitrary ordering of the features,
            chosen by you prior to calling ``best_subset_given_size``.

            For example, 1 might refer to the first variable in a
            particular set of data, 2 the second, etc..

            If `la` = 0, the score for a single subset should be
            returned.

            This subset is constructed by adding or removing only those
            features specified in the first `lz` elements of `z`.

            If `lz` = 0, this subset will either be Omega or ∅.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    mincnt : int
        k, the minimum number of times the effect of each feature, x_i,
        must have been observed before f(S-{x_i}) is estimated from f(S)
        as opposed to being calculated directly.

        If k = 0 then f(S-{x_i}) is never estimated.

        If `mincnt` < 0 then k is set to 1.

    gamma : float
        gamma, the scaling factor used when estimating scores. If
        `gamma` < 0 then gamma = 1 is used.

    acc : float, shape (2)
        A measure of the accuracy of the scoring function, f.

        Letting a_i = epsilon_1*|f(S_i)|+epsilon_2, then when confirming
        whether the scoring function is strictly increasing or
        decreasing (as described in `mincr`), or when assessing whether
        a node defined by subset S_i can be trimmed, then any values in
        the range f(S_i)+/-a_i are treated as being numerically
        equivalent.

        If 0 <= `acc`[0] <= 1 then epsilon_1 = `acc`[0], otherwise
        epsilon_1 = 0.

        If `acc`[1] >= 0 then epsilon_2 = `acc`[1], otherwise epsilon_2
        = 0.

        In most situations setting both epsilon_1 and epsilon_2 to zero
        should be sufficient.

        Using a nonzero value, when one is not required, can
        significantly increase the number of subsets that need to be
        evaluated.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    la : int
        The number of best subsets returned.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `mincr` = *<value>*.

            Constraint: `mincr` = 0 or 1.

        (`errno` 21)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 31)
            On entry, `ip` = *<value>* and `m` = *<value>*.

            Constraint: 1 <= `ip` <= `m`.

        (`errno` 41)
            On entry, `nbest` = *<value>*.

            Constraint: `nbest` >= 1.

        (`errno` 81)
            On exit from `f`, `score`[*<value>*] = *<value>*, which is
            inconsistent with the score for the parent node. Score for
            the parent node is *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 42)
            On entry, `nbest` = *<value>*.

            But only *<value>* best subsets could be calculated.

    NagCallbackTerminateWarning
        (`errno` 82)
            A nonzero value for `info` has been returned: `info` =
            *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.mip.best_subset_given_size'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_f_name = fun_name + ':f'

    fh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, drop.
        utils._BYREF_INT, # Input, lz.
        utils._EngineIntArrayType.c_type, # Input, z[lz].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, a[la].
        utils._EngineFloat64ArrayType.c_type, # Output, score[max(la, 1)].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Input/Output, info.
    )

    def cb_fh(_f, en_data, m, drop, lz, z, la, a, score, _iuser, _ruser, info):
        """Helper for f."""
        try:
            _m_fh = utils._EngineIntScalarType.to_py(m.contents)
            _drop_fh = utils._EngineIntScalarType.to_py(drop.contents)
            _lz_fh = utils._EngineIntScalarType.to_py(lz.contents)
            _z_fh = utils._EngineIntArrayType.from_pointer(
                z,
                (_lz_fh,),
            )
            _la_fh = utils._EngineIntScalarType.to_py(la.contents)
            _a_fh = utils._EngineIntArrayType.from_pointer(
                a,
                (_la_fh,),
            )
            _score_fh = utils._EngineFloat64ArrayType.from_pointer(
                score,
                (max(_la_fh, 1),),
                can_reference_contents=False,
            )
            if data is not None:
                f(
                    _m_fh,
                    _drop_fh,
                    _lz_fh,
                    _z_fh,
                    _la_fh,
                    _a_fh,
                    _score_fh,
                    data,
                )
            else:
                f(
                    _m_fh,
                    _drop_fh,
                    _lz_fh,
                    _z_fh,
                    _la_fh,
                    _a_fh,
                    _score_fh,
                )
            utils._unlink_pointer(z)
            utils._unlink_pointer(a)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _score_fh, {'fun_name': cb_f_name, 'returned': 'score'}, score,
            )
            utils._unlink_pointer(score)
        except:
            excs.append(_sys.exc_info())
            info.contents.value = -1
            if excs[-1][0] is not utils.UserCallbackTerminate:
                en_data.contents.hlperr = utils._IERR_HELPER

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('h05abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mincr.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, nbest.
        utils._BYREF_INT, # Output, la.
        utils._EngineFloat64ArrayType.c_type, # Output, bscore[nbest].
        utils._EngineIntArrayType.c_type, # Output, bz[m-ip, nbest].
        utils._UNUSED_FUNCTYPE,
        fh,
        utils._BYREF_INT, # Input, mincnt.
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._EngineFloat64ArrayType.c_type, # Input, acc[2].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Input, monitor.
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_INT, # Input, mwidth.
        utils._BYREF_INT, # Input, csep.
        utils._BYREF_INT, # Input, cwidth.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((bz, 'bz'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mincr = utils._EngineIntScalarType(
        mincr, {'fun_name': fun_name, 'entity': 'mincr'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _nbest = utils._EngineIntScalarType(
        nbest, {'fun_name': fun_name, 'entity': 'nbest'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType.empty_ctypes()
    _nbest_py = utils._EngineIntScalarType.to_py(_nbest)
    _bscore_cls = utils._EngineFloat64ArrayType(
        bscore, {'fun_name': fun_name, 'entity': 'bscore'},
        exp_shape=(_nbest_py,),
        check_contents=False,
    )
    _bscore = _bscore_cls.to_ctypes(can_reference_contents=False)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _bz_cls = utils._EngineIntArrayType(
        bz, {'fun_name': fun_name, 'entity': 'bz'},
        exp_shape=(_m_py-_ip_py, _nbest_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _bz = _bz_cls.to_ctypes(can_reference_contents=False)
    utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = utils._UNUSED_CALLBACK
    _fh = utils._cast_callback(
        fh, cb_fh,
        {'fun_name': fun_name, 'entity': 'f'},
    )
    _mincnt = utils._EngineIntScalarType(
        mincnt, {'fun_name': fun_name, 'entity': 'mincnt'},
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _acc = utils._EngineFloat64ArrayType(
        acc, {'fun_name': fun_name, 'entity': 'acc'},
        exp_shape=(2,),
    ).to_ctypes()
    _iuser = None
    _ruser = None
    _monitor = utils._EngineIntScalarType.c_type(0)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _mwidth = utils._EngineIntScalarType.c_type(80)
    _csep = utils._EngineIntScalarType.c_type(5)
    _mwidth_py = utils._EngineIntScalarType.to_py(_mwidth)
    _csep_py = utils._EngineIntScalarType.to_py(_csep)
    _cwidth = utils._EngineIntScalarType(
        (_mwidth_py-_csep_py)//2, {'fun_name': fun_name, 'entity': '_cwidth'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mincr,
        _m,
        _ip,
        _nbest,
        _la,
        _bscore,
        _bz,
        _f,
        _fh,
        _mincnt,
        _gamma,
        _acc,
        _iuser,
        _ruser,
        _monitor,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _nadv,
        _mwidth,
        _csep,
        _cwidth,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 21 and _ifmt == 21:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mincr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mincr = 0 or 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _ierr_fl = 31
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ip <= m.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _ierr_fl = 31
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ip <= m.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _ierr_fl = 41
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nbest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nbest >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 53 and _ifmt == 53:
            _ierr_fl = 42
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nbest = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'But only {:s} best subsets could be calculated.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _ierr_fl = 81
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On exit from f,',
                    'score[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'which is inconsistent with the score for the parent node.',
                    'Score for the parent node is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _ierr_fl = 81
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On exit from f,',
                    'score[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'which is inconsistent with the score for the parent node.',
                    'Score for the parent node is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _err_class = utils._UserCallbackExit(
                fun_name=fun_name,
                msg_lines=[
                    'A nonzero value for info has been returned: info = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
                cb_excs=excs,
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _bscore_cls.refresh_py(_bscore)
        _bz_cls.refresh_py(_bz)
        _la_py = utils._EngineIntScalarType.to_py(_la)
        return_data = _la_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['la'],
            )(
                _la_py,
            )
    _err_class.manage()
    return return_data
