# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `sparse` Chapter.

``sparse`` - Large Scale Linear Systems

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module provides functions for the solution of large sparse systems
of simultaneous linear equations.
These include iterative methods for real nonsymmetric and symmetric,
complex non-Hermitian and Hermitian linear systems and direct methods
for general real linear systems.
Further direct methods are currently available in submodule ``matop``
and submodule ``linsys``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections
import sys as _sys

from ..base import utils

def real_gen_basic_setup(method, precon, norm, weight, iterm, n, m, tol, maxitn, anorm, sigmax, monit):
    u"""
    Real sparse nonsymmetric linear systems, setup for
    ``real_gen_basic_solver``.

    ``real_gen_basic_setup`` is a setup function, the first in a suite
    of three functions for the iterative solution of a real general
    (nonsymmetric) system of simultaneous linear equations.
    ``real_gen_basic_setup`` must be called before
    ``real_gen_basic_solver``, the iterative solver. The third function
    in the suite, ``real_gen_basic_diag``, can be used to return
    additional information about the computation.

    These functions are suitable for the solution of large sparse
    general (nonsymmetric) systems of equations.

    For full information please refer to the NAG Library document for
    f11bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11bdf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'P'

          Preconditioning.

    norm : str, length 1
        Defines the matrix and vector norm to be used in the termination
        criteria.

        `norm` = '1'

          l_1 norm.

        `norm` = 'I'

          l_infinity norm.

        `norm` = '2'

          l_2 norm.

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the computation of the vector norms required in
        termination criterion (2) (`iterm` = 1): ||v||_p^(w) =
        ||v^(w)||_p, where v_i^(w) = w_i v_i, for i = 1,2,...,n. The
        suffix p = 1,2, infinity denotes the vector norm used, as
        specified by the argument `norm`. Note that weights cannot be
        used when `iterm` = 2, i.e., when criterion (3) is used.

        `weight` = 'W'

          User-supplied weights are to be used and must be supplied on
          initial entry to ``real_gen_basic_solver``.

        `weight` = 'N'

          All weights are implicitly set equal to one. Weights do not
          need to be supplied on initial entry to
          ``real_gen_basic_solver``.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

          Use the termination criterion defined in (2).

        `iterm` = 2

          Use the termination criterion defined in (3).

    n : int
        n, the order of the matrix A.

    m : int
        If `method` = 'RGMRES', `m` is the dimension m of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The tolerance tau for the termination criterion. If `tol` <=
        0.0, tau = max(sqrt(epsilon), sqrt(n)*epsilon) is used, where
        epsilon is the machine precision. Otherwise tau = max(`tol`,
        10*epsilon, sqrt(n)*epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', then
        ||A||_1 = ||A||_infinity is estimated internally by
        ``real_gen_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

    sigmax : float
        If `iterm` = 2, the largest singular value sigma_1 of the
        preconditioned iteration matrix; otherwise, `sigmax` is not
        referenced.

        If `sigmax` <= 0.0, `iterm` = 2 and `method` = 'RGMRES', the
        value of sigma_1 will be estimated internally.

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``real_gen_basic_solver``: if `method` = 'CGS' or
        'TFQMR', a monitoring step is executed every `monit` iterations;
        otherwise, a monitoring step is executed every `monit`
        super-iterations (groups of up to m or l iterations for RGMRES
        or BI-CGSTAB(l), respectively).

        There are some additional computational costs involved in
        monitoring the solution and residual vectors when the
        BI-CGSTAB(l) method is used with l > 1.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -12)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -11)
            On entry, `iterm` = 2, `method` = *<value>* and `sigmax` =
            *<value>*.

            Constraint: if `iterm` = 2 and `method` = 'CGS' or
            'BICGSTAB', `sigmax` > 0.0.

        (`errno` -10)
            On entry, `iterm` = 1, `norm` = '2' and `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2', `anorm` > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `m` = *<value>* and `method` = *<value>*.

            Constraint: if `method` = 'RGMRES' or 'BICGSTAB', `m` > 0.

        (`errno` -7)
            On entry, `m` = *<value>*, `n` = *<value>* and `method` =
            *<value>*.

            Constraint: if `method` = 'RGMRES', `m` <= min(`n`, 50). If
            `method` = 'BICGSTAB', `m` <= min(`n`, 10).

        (`errno` -6)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -5)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -5)
            On entry, `iterm` = 2 and `method` = 'TFQMR'.

            Constraint: if `iterm` = 2, `method` != 'TFQMR'.

        (`errno` -5)
            On entry, `iterm` = 2 and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2'.

        (`errno` -5)
            On entry, `iterm` = 2 and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -4)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -3)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS', 'BICGSTAB' or
            'TFQMR'.

        (`errno` 1)
            ``real_gen_basic_setup`` has been called out of sequence:
            either ``real_gen_basic_setup`` has been called twice or
            ``real_gen_basic_solver`` has not terminated its current
            task.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineFloat64ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType.c_type(100)
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(_lwork_py)
    _comm_py = {}
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _norm,
        _weight,
        _iterm,
        _n,
        _m,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    if _ierr.value == utils._IERR_SUCCESS:
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        _lwork = utils._EngineIntScalarType(
            _lwreq_py, {'fun_name': fun_name, 'entity': '_lwork'},
        ).to_ctypes()
        _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
        _work = utils._EngineFloat64ArrayType.empty_ctypes(_lwork_py)
        _comm_py['work'] = _work
        engine_ctypes(
            _en_data,
            _method,
            _precon,
            _norm,
            _weight,
            _iterm,
            _n,
            _m,
            _tol,
            _maxitn,
            _anorm,
            _sigmax,
            _monit,
            _lwreq,
            _work,
            _lwork,
            _errbuf,
            _ierr,
            _method._length_,
            _precon._length_,
            _norm._length_,
            _weight._length_,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _monit_py = utils._EngineIntScalarType.to_py(_monit)
        _maxitn_py = utils._EngineIntScalarType.to_py(_maxitn)
        _method_py = utils._EngineCharScalarType.to_py(_method)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _tol_py = utils._EngineFloat64ScalarType.to_py(_tol)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _iterm_py = utils._EngineIntScalarType.to_py(_iterm)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        _weight_py = utils._EngineCharScalarType.to_py(_weight)
        _precon_py = utils._EngineCharScalarType.to_py(_precon)
        if _ierr_py == -12 and _ifmt == -12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_monit_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -11 and _ifmt == -11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2, method = \'{:s}\' and'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'sigmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_sigmax_py),
                    ),
                    'Constraint: if iterm = 2 and method = \'CGS\' or \'BICGSTAB\', sigmax > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == -10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 1, norm = \'2\' and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == -9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_tol_py),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and method = \'{:s}\'.'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\' or \'BICGSTAB\', m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\', m <= min(n, 50).',
                    'If method = \'BICGSTAB\', m <= min(n, 10).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == -6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == -5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT.format(_iterm_py),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and method = \'TFQMR\'.',
                    'Constraint: if iterm = 2, method != \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == -4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_precon_py),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\', \'BICGSTAB\' or \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_gen_basic_setup has been called out of sequence:',
                    'either real_gen_basic_setup has been called twice or',
                    'real_gen_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def real_gen_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Real sparse nonsymmetric linear systems, preconditioned RGMRES, CGS,
    Bi-CGSTAB or TFQMR method.

    ``real_gen_basic_solver`` is an iterative solver for a real general
    (nonsymmetric) system of simultaneous linear equations;
    ``real_gen_basic_solver`` is the second in a suite of three
    functions, where the first function, ``real_gen_basic_setup``, must
    be called prior to ``real_gen_basic_solver`` to set up the suite,
    and the third function in the suite, ``real_gen_basic_diag``, can be
    used to return additional information about the computation.

    These functions are suitable for the solution of large sparse
    general (nonsymmetric) systems of equations.

    For full information please refer to the NAG Library document for
    f11be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11bef.html

    Parameters
    ----------
    irevcm : int
        On initial entry: `irevcm` = 0, otherwise an error condition
        will be raised.

        On intermediate entry: must either be unchanged from its
        previous exit value, or can have one of the following values.

        `irevcm` = 5

          Tidy termination: the computation will terminate at the end of
          the current iteration. Further reverse communication exits may
          occur depending on when the termination request is issued.
          ``real_gen_basic_solver`` will then return with the
          termination code `irevcm` = 4. Note that before calling
          ``real_gen_basic_solver`` with `irevcm` = 5 the calling
          program must have performed the tasks required by the value of
          `irevcm` returned by the previous call to
          ``real_gen_basic_solver``, otherwise subsequently returned
          values may be invalid.

        `irevcm` = 6

          Immediate termination: ``real_gen_basic_solver`` will return
          immediately with termination code `irevcm` = 4 and with any
          useful information available. This includes the last iterate
          of the solution. The residual vector is generally not
          available.

          Immediate termination may be useful, for example, when errors
          are detected during matrix-vector multiplication or during the
          solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : float, shape (n), modified in place
        On initial entry: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        On intermediate entry: must remain unchanged.

        On intermediate exit: the returned value of `irevcm` determines
        the contents of `u` in the following way:

        - if `irevcm` = -1, 1 or 2, `u` holds the vector u on which the
          operation specified by `irevcm` is to be carried out;

        - if `irevcm` = 3, `u` holds the current iterate of the solution
          vector.

        On final exit: if `errno` = 3 or `errno` < 0, the array `u` is
        unchanged from the initial entry to ``real_gen_basic_solver``.

        If `errno` = 1, the array `u` is unchanged from the last entry
        to ``real_gen_basic_solver``.

        Otherwise, `u` holds the last available iterate of the solution
        of the system of equations, for all returned values of errno.

    v : float, shape (n), modified in place
        On initial entry: the right-hand side b of the system of
        equations Ax = b.

        On intermediate entry: the returned value of `irevcm` determines
        the contents of `v` in the following way:

        - if `irevcm` = -1, 1 or 2, `v` must store the vector v, the
          result of the operation specified by the value of `irevcm`
          returned by the previous call to ``real_gen_basic_solver``;

        - if `irevcm` = 3, `v` must remain unchanged.

        On intermediate exit: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        On final exit: if `errno` = 3 or `errno` < 0, the array `v` is
        unchanged from the initial entry to ``real_gen_basic_solver``.

        If `errno` = 1, the array `v` is unchanged from the last entry
        to ``real_gen_basic_solver``.

        If the function exits successfully or `errno` = 2, the array `v`
        contains the true residual vector of the system of equations
        (see also [Errors]).

        Otherwise, `v` stores the last available iterate of the residual
        vector unless `errno` = 8 is returned on last entry, in which
        case `v` is set to 0.0.

    wgt : float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Description] and [Parameters]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_gen_basic_setup``.

    Returns
    -------
    irevcm : int
        On intermediate exit: has the following meanings.

        `irevcm` = -1

          The calling program must compute the matrix-vector product v =
          A^Tu, where u and v are stored in `u` and `v`, respectively;
          RGMRES, CGS and BI-CGSTAB(l) methods return `irevcm` = -1 only
          if the matrix norm ||A||_1 or ||A||_infinity is estimated
          internally using Higham's method. This can only happen if
          iterm = 1 in ``real_gen_basic_setup``.

        `irevcm` = 1

          The calling program must compute the matrix-vector product v =
          Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

          The calling program must solve the preconditioning equation Mv
          = u, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 3

          Monitoring step: the solution and residual at the current
          iteration are returned in the arrays `u` and `v`,
          respectively. No action by the calling program is required.
          ``real_gen_basic_diag`` can be called at this step to return
          additional information.

        On final exit: `irevcm` = 4: ``real_gen_basic_solver`` has
        completed its tasks. The value of errno determines whether the
        iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``real_gen_basic_setup`` was not called before
            calling ``real_gen_basic_solver`` or it has returned an
            error.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``real_gen_basic_solver`` has already completed its tasks.
            You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested tidy termination. The required accuracy has
            not been achieved. However, a reasonable accuracy may have
            been achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithm breakdown at iteration no. *<value>*.

        (`errno` 8)
            User-requested immediate termination.

        (`errno` 10)
            The weights in array `wgt` are all zero.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, u[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _vl_n_py = int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (8,), validate=False))
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
    ).to_ctypes()
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_gen_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The required accuracy has not been achieved.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either real_gen_basic_setup was not called before calling real_gen_basic_solver',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithm breakdown at iteration no. {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def real_gen_basic_diag(comm):
    u"""
    Real sparse nonsymmetric linear systems, diagnostic for
    ``real_gen_basic_solver``.

    ``real_gen_basic_diag`` is the third in a suite of three functions
    for the iterative solution of a real general (nonsymmetric) system
    of simultaneous linear equations. ``real_gen_basic_diag`` returns
    information about the computations during an iteration and/or after
    this has been completed.
    The first function of the suite, ``real_gen_basic_setup``, is a
    setup function; the second function, ``real_gen_basic_solver``, is
    the iterative solver itself.

    These three functions are suitable for the solution of large sparse
    general (nonsymmetric) systems of equations.

    For full information please refer to the NAG Library document for
    f11bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11bff.html

    Parameters
    ----------
    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_gen_basic_setup``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``real_gen_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``real_gen_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``real_gen_basic_solver``.

    anorm : float
        If iterm = 1 in the previous call to ``real_gen_basic_setup``,
        then `anorm` contains ||A||_p, where p = 1, 2 or infinity,
        either supplied or, in the case of 1 or infinity, estimated by
        ``real_gen_basic_solver``; otherwise `anorm` = 0.0.

    sigmax : float
        If iterm = 2 in the previous call to ``real_gen_basic_setup``,
        the current estimate of the largest singular value sigma_1(A_)
        of the preconditioned iteration matrix, either when it has been
        supplied to ``real_gen_basic_setup`` or it has been estimated by
        ``real_gen_basic_solver`` (see also [Description] and
        [Parameters]); otherwise, `sigmax` = 0.0 is returned.

    Raises
    ------
    NagValueError
        (`errno` 1)
            ``real_gen_basic_diag`` has been called out of sequence.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._EngineFloat64ArrayType.c_type, # Input, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_gen_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_basic_setup(method, precon, norm, weight, iterm, n, m, tol, maxitn, anorm, sigmax, monit, lwork):
    u"""
    Complex sparse non-Hermitian linear systems, setup for
    ``complex_gen_basic_solver``.

    ``complex_gen_basic_setup`` is a setup function, the first in a
    suite of three functions for the iterative solution of a complex
    general (non-Hermitian) system of simultaneous linear equations.
    ``complex_gen_basic_setup`` must be called before
    ``complex_gen_basic_solver``, the iterative solver.
    The third function in the suite, ``complex_gen_basic_diag``, can be
    used to return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    general (non-Hermitian) systems of equations.

    For full information please refer to the NAG Library document for
    f11br

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11brf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'P'

          Preconditioning.

    norm : str, length 1
        Defines the matrix and vector norm to be used in the termination
        criteria.

        `norm` = '1'

          l_1 norm.

        `norm` = 'I'

          l_infinity norm.

        `norm` = '2'

          l_2 norm.

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the computation of the vector norms required in
        termination criterion (2) (`iterm` = 1): ||v||_p^(w) =
        ||v^(w)||_p, where v_i^(w) = w_i v_i, for i = 1,2,...,n. The
        suffix p = 1,2, infinity denotes the vector norm used, as
        specified by the argument `norm`. Note that weights cannot be
        used when `iterm` = 2, i.e., when criterion (3) is used.

        `weight` = 'W'

          User-supplied weights are to be used and must be supplied on
          initial entry to ``complex_gen_basic_solver``.

        `weight` = 'N'

          All weights are implicitly set equal to one. Weights do not
          need to be supplied on initial entry to
          ``complex_gen_basic_solver``.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

          Use the termination criterion defined in (2).

        `iterm` = 2

          Use the termination criterion defined in (3).

    n : int
        n, the order of the matrix A.

    m : int
        If `method` = 'RGMRES', `m` is the dimension m of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The tolerance tau for the termination criterion. If `tol` <=
        0.0, tau = max(sqrt(epsilon), sqrt(n)*epsilon) is used, where
        epsilon is the machine precision. Otherwise tau = max(`tol`,
        10*epsilon, sqrt(n)*epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', then
        ||A||_1 = ||A||_infinity is estimated internally by
        ``complex_gen_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

    sigmax : float
        If `iterm` = 2, the largest singular value sigma_1 of the
        preconditioned iteration matrix; otherwise, `sigmax` is not
        referenced.

        If `sigmax` <= 0.0, `iterm` = 2 and `method` = 'RGMRES', the
        value of sigma_1 will be estimated internally.

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``complex_gen_basic_solver``: if `method` = 'CGS' or
        'TFQMR', a monitoring step is executed every `monit` iterations;
        otherwise, a monitoring step is executed every `monit`
        super-iterations (groups of up to m or l iterations for RGMRES
        or BI-CGSTAB(l), respectively).

        There are some additional computational costs involved in
        monitoring the solution and residual vectors when the
        BI-CGSTAB(l) method is used with l > 1.

    lwork : int
        The dimension of the array `comm`['work'].

    Returns
    -------
    lwreq : int
        The minimum amount of workspace required by
        ``complex_gen_basic_solver``. (See also [Parameters].)

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -15)
            On entry, `lwork` = *<value>*.

            Constraint: `lwork` >= 120.

        (`errno` -12)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -11)
            On entry, `iterm` = 2, `method` = *<value>* and `sigmax` =
            *<value>*.

            Constraint: if `iterm` = 2 and `method` = 'CGS' or
            'BICGSTAB', `sigmax` > 0.0.

        (`errno` -10)
            On entry, `iterm` = 1, `norm` = '2' and `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2', `anorm` > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `m` = *<value>*, `n` = *<value>* and `method` =
            *<value>*.

            Constraint: if `method` = 'RGMRES', `m` <= min(`n`, 50). If
            `method` = 'BICGSTAB', `m` <= min(`n`, 10).

        (`errno` -7)
            On entry, `m` = *<value>* and `method` = *<value>*.

            Constraint: if `method` = 'RGMRES' or 'BICGSTAB', `m` > 0.

        (`errno` -6)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -5)
            On entry, `iterm` = 2 and `method` = 'TFQMR'.

            Constraint: if `iterm` = 2, `method` != 'TFQMR'.

        (`errno` -5)
            On entry, `iterm` = 2 and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2'.

        (`errno` -5)
            On entry, `iterm` = 2 and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -5)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -4)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -3)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS', 'BICGSTAB' or
            'TFQMR'.

        (`errno` 1)
            ``complex_gen_basic_setup`` has been called out of sequence:
            either ``complex_gen_basic_setup`` has been called twice or
            ``complex_gen_basic_solver`` has not terminated its current
            task.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11brft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineComplex128ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType(
        lwork, {'fun_name': fun_name, 'entity': 'lwork'},
    ).to_ctypes()
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work = utils._EngineComplex128ArrayType.empty_ctypes(_lwork_py)
    _comm_py = {}
    _comm_py['work'] = _work
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _norm,
        _weight,
        _iterm,
        _n,
        _m,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _monit_py = utils._EngineIntScalarType.to_py(_monit)
        _maxitn_py = utils._EngineIntScalarType.to_py(_maxitn)
        _method_py = utils._EngineCharScalarType.to_py(_method)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _tol_py = utils._EngineFloat64ScalarType.to_py(_tol)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        _weight_py = utils._EngineCharScalarType.to_py(_weight)
        _iterm_py = utils._EngineIntScalarType.to_py(_iterm)
        _precon_py = utils._EngineCharScalarType.to_py(_precon)
        if _ierr_py == -15 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwork = {:s}.'.format(
                        utils._STR_FMT_INT.format(_lwork_py),
                    ),
                    'Constraint: lwork >= 120.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -12 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_monit_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -11 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2, method = \'{:s}\' and'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'sigmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_sigmax_py),
                    ),
                    'Constraint: if iterm = 2 and method = \'CGS\' or \'BICGSTAB\', sigmax > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 1, norm = \'2\' and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_tol_py),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}, n = {:s} and'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\', m <= min(n, 50).',
                    'If method = \'BICGSTAB\', m <= min(n, 10).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and method = \'{:s}\'.'.format(
                        utils._STR_FMT_INT.format(_m_py),
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: if method = \'RGMRES\' or \'BICGSTAB\', m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and method = \'TFQMR\'.',
                    'Constraint: if iterm = 2, method != \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99919:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99929:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT.format(_iterm_py),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_precon_py),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\', \'BICGSTAB\' or \'TFQMR\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_gen_basic_setup has been called out of sequence:',
                    'either complex_gen_basic_setup has been called twice or',
                    'complex_gen_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['lwreq', 'comm'],
        )(
            _lwreq_py,
            _comm_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Complex sparse non-Hermitian linear systems, preconditioned RGMRES,
    CGS, Bi-CGSTAB or TFQMR method.

    ``complex_gen_basic_solver`` is an iterative solver for a complex
    general (non-Hermitian) system of simultaneous linear equations;
    ``complex_gen_basic_solver`` is the second in a suite of three
    functions, where the first function, ``complex_gen_basic_setup``,
    must be called prior to ``complex_gen_basic_solver`` to set up the
    suite, and the third function in the suite,
    ``complex_gen_basic_diag``, can be used to return additional
    information about the computation.

    These three functions are suitable for the solution of large sparse
    general (non-Hermitian) systems of equations.

    For full information please refer to the NAG Library document for
    f11bs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11bsf.html

    Parameters
    ----------
    irevcm : int
        On initial entry: `irevcm` = 0, otherwise an error condition
        will be raised.

        On intermediate entry: must either be unchanged from its
        previous exit value, or can have one of the following values.

        `irevcm` = 5

          Tidy termination: the computation will terminate at the end of
          the current iteration. Further reverse communication exits may
          occur depending on when the termination request is issued.
          ``complex_gen_basic_solver`` will then return with the
          termination code `irevcm` = 4. Note that before calling
          ``complex_gen_basic_solver`` with `irevcm` = 5 the calling
          program must have performed the tasks required by the value of
          `irevcm` returned by the previous call to
          ``complex_gen_basic_solver``, otherwise subsequently returned
          values may be invalid.

        `irevcm` = 6

          Immediate termination: ``complex_gen_basic_solver`` will
          return immediately with termination code `irevcm` = 4 and with
          any useful information available. This includes the last
          iterate of the solution.

          Immediate termination may be useful, for example, when errors
          are detected during matrix-vector multiplication or during the
          solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : complex, shape (n), modified in place
        On initial entry: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        On intermediate entry: must remain unchanged.

        On intermediate exit: the returned value of `irevcm` determines
        the contents of `u` as follows.

        If `irevcm` = -1, 1 or 2, `u` holds the vector u on which the
        operation specified by `irevcm` is to be carried out.

        If `irevcm` = 3, `u` holds the current iterate of the solution
        vector.

        On final exit: if `errno` = 3 or `errno` < 0, the array `u` is
        unchanged from the initial entry to
        ``complex_gen_basic_solver``.

        If `errno` = 1, the array `u` is unchanged from the last entry
        to ``complex_gen_basic_solver``.

        Otherwise, `u` holds the last available iterate of the solution
        of the system of equations, for all returned values of errno.

    v : complex, shape (n), modified in place
        On initial entry: the right-hand side b of the system of
        equations Ax = b.

        On intermediate entry: the returned value of `irevcm` determines
        the contents of `v` as follows.

        If `irevcm` = -1, 1 or 2, `v` must store the vector v, the
        result of the operation specified by the value of `irevcm`
        returned by the previous call to ``complex_gen_basic_solver``.

        If `irevcm` = 3, `v` must remain unchanged.

        On intermediate exit: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        On final exit: if `errno` = 3 or `errno` < 0, the array `v` is
        unchanged from the initial entry to
        ``complex_gen_basic_solver``.

        If `errno` = 1, the array `v` is unchanged from the last entry
        to ``complex_gen_basic_solver``.

        If the function exits successfully or `errno` = 2, the array `v`
        contains the true residual vector of the system of equations
        (see also [Errors]).

        Otherwise, `v` stores the last available iterate of the residual
        vector unless `errno` = 8 is returned on last entry, in which
        case `v` is set to 0.0.

    wgt : float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Description] and [Parameters]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_gen_basic_setup``.

    Returns
    -------
    irevcm : int
        On intermediate exit: has the following meanings.

        `irevcm` = -1

          The calling program must compute the matrix-vector product v =
          A^Hu, where u and v are stored in `u` and `v`, respectively;
          RGMRES, CGS and BI-CGSTAB(l) methods return `irevcm` = -1 only
          if the matrix norm ||A||_1 or ||A||_infinity is estimated
          internally using Higham's method. This can only happen if
          iterm = 1 in ``complex_gen_basic_setup``.

        `irevcm` = 1

          The calling program must compute the matrix-vector product v =
          Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

          The calling program must solve the preconditioning equation Mv
          = u, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 3

          Monitoring step: the solution and residual at the current
          iteration are returned in the arrays `u` and `v`,
          respectively. No action by the calling program is required.
          ``complex_gen_basic_diag`` can be called at this step to
          return additional information.

        On final exit: `irevcm` = 4: ``complex_gen_basic_solver`` has
        completed its tasks. The value of errno determines whether the
        iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -6)
            On entry, lwork = *<value>*.

            Constraint: lwork >= lwreq, where lwreq is returned by
            ``complex_gen_basic_setup``.

        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``complex_gen_basic_setup`` was not called before
            calling ``complex_gen_basic_solver`` or it has returned an
            error.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``complex_gen_basic_solver`` has already completed its
            tasks. You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested termination: the required accuracy could not
            be obtained. However, a reasonable accuracy may have been
            achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithm breakdown at iteration no. *<value>*.

        (`errno` 8)
            User-requested immediate termination.

        (`errno` 10)
            The weights in array `wgt` are all zero.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11bsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, u[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _vl_n_py = int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (9,), validate=False).real)
    _u_cls = utils._EngineComplex128ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineComplex128ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
    ).to_ctypes()
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -6 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwork = {:s}.'.format(
                        utils._STR_FMT_INT.format(_lwork_py),
                    ),
                    'Constraint: lwork >= lwreq, where lwreq is returned by complex_gen_basic_setup.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_gen_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested termination: the required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either complex_gen_basic_setup was not called before calling complex_gen_basic_solver',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithm breakdown at iteration no. {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def complex_gen_basic_diag(comm):
    u"""
    Complex sparse non-Hermitian linear systems, diagnostic for
    ``complex_gen_basic_solver``.

    ``complex_gen_basic_diag`` is the third in a suite of three
    functions for the iterative solution of a complex general
    (non-Hermitian) system of simultaneous linear equations.
    ``complex_gen_basic_diag`` returns information about the
    computations during an iteration and/or after this has been
    completed.
    The first function of the suite, ``complex_gen_basic_setup``, is a
    setup function; the second function, ``complex_gen_basic_solver``,
    is the iterative solver itself.

    These three functions are suitable for the solution of large sparse
    general (non-Hermitian) systems of equations.

    For full information please refer to the NAG Library document for
    f11bt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11btf.html

    Parameters
    ----------
    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_gen_basic_setup``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``complex_gen_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``complex_gen_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``complex_gen_basic_solver``.

    anorm : float
        If iterm = 1 in the previous call to
        ``complex_gen_basic_setup``, then `anorm` contains ||A||_p,
        where p = 1, 2 or infinity, either supplied or, in the case of 1
        or infinity, estimated by ``complex_gen_basic_solver``;
        otherwise `anorm` = 0.0.

    sigmax : float
        If iterm = 2 in the previous call to
        ``complex_gen_basic_setup``, the current estimate of the largest
        singular value sigma_1(A_) of the preconditioned iteration
        matrix when it is used by the termination criterion in
        ``complex_gen_basic_solver``, either when it has been supplied
        to ``complex_gen_basic_setup`` or it has been estimated by
        ``complex_gen_basic_solver`` (see also [Description] and
        [Parameters]); otherwise, `sigmax` = 0.0 is returned.

    Raises
    ------
    NagValueError
        (`errno` -7)
            On entry, `lwork` = *<value>*.

            Constraint: `lwork` >= 120.

        (`errno` 1)
            ``complex_gen_basic_diag`` has been called out of sequence.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11btft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
        if _ierr_py == -7 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwork = {:s}.'.format(
                        utils._STR_FMT_INT.format(_lwork_py),
                    ),
                    'Constraint: lwork >= 120.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_gen_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
        )
    _err_class.manage()
    return return_data

def real_gen_precon_ilu(n, nnz, a, la, irow, icol, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag):
    u"""
    Real sparse nonsymmetric linear systems, incomplete LU
    factorization.

    ``real_gen_precon_ilu`` computes an incomplete LU factorization of a
    real sparse nonsymmetric matrix, represented in coordinate storage
    format.
    This factorization may be used as a preconditioner in combination
    with ``real_gen_solve_ilu`` or ``real_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11daf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`la`), modified in place
        On entry: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``real_gen_sort`` may be used to
        order the elements in this way.

        On exit: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrix C. Matrix elements are ordered by increasing row
        index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        On entry: the row indices of the nonzero elements supplied in
        `a`.

        On exit: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        On entry: the column indices of the nonzero elements supplied in
        `a`.

        On exit: the column indices of the nonzero elements returned in
        `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

          No pivoting is carried out.

        `pstrat` = 'U'

          Pivoting is carried out according to the user-defined input
          values of `ipivp` and `ipivq`.

        `pstrat` = 'P'

          Partial pivoting by columns for stability is carried out.

        `pstrat` = 'C'

          Complete pivoting by rows for sparsity, and by columns for
          stability, is carried out.

    milu : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row-sums (see [Further Comments]).

        `milu` = 'M'

          The factorization is modified.

        `milu` = 'N'

          The factorization is not modified.

    ipivp : int, shape (`n`), modified in place
        On entry: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1] must
        specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the pivot indices. If `ipivp`[k-1] = i and `ipivq`[k-1]
        = j then the element in row i and column j was used as the pivot
        at elimination stage k.

    ipivq : int, shape (`n`), modified in place
        On entry: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1] must
        specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the pivot indices. If `ipivp`[k-1] = i and `ipivq`[k-1]
        = j then the element in row i and column j was used as the pivot
        at elimination stage k.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    idiag : int, shape (`n`), modified in place
        On exit: `idiag`[i-1], for i = 1,2,...,`n`, holds the index of
        arrays `a`, `irow` and `icol` which holds the diagonal element
        in row i of the matrix C.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the matrix C.

    npivm : int
        If `npivm` > 0 it gives the number of pivots which were modified
        during the factorization to ensure that M exists.

        If `npivm` = -1 no pivot modifications were required, but a
        local restart occurred (see [Algorithmic Details]).

        The quality of the preconditioner will generally depend on the
        returned value of `npivm`.

        If `npivm` is large the preconditioner may not be satisfactory.

        In this case it may be advantageous to call
        ``real_gen_precon_ilu`` again with an increased value of
        `lfill`, a reduced value of `dtol`, or set `pstrat` = 'C'.

        See also [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `milu` = *<value>*.

            Constraint: `milu` = 'M' or 'N'.

        (`errno` 1)
            On entry, `pstrat` = *<value>*.

            Constraint: `pstrat` = 'N', 'U', 'P' or 'C'.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` lies outside the
            range [1,`n`].

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            reducing `lfill`, or increasing `dtol`.

        (`errno` 5)
            A serious error has occurred in an internal call to an
            auxiliary function. Check all function calls and array
            sizes. Seek expert help.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._BYREF_STRING, # Input, milu(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[n].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _milu = utils._EngineCharScalarType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat._length_,
        _milu._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: milu = \'M\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pstrat = \'N\', \'U\', \'P\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call to an auxiliary function.',
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def real_gen_precon_ilu_solve(trans, n, a, la, irow, icol, ipivp, ipivq, istr, idiag, check, y, x):
    u"""
    Solution of linear system involving incomplete LU preconditioning
    matrix generated by ``real_gen_precon_ilu``.

    ``real_gen_precon_ilu_solve`` solves a system of linear equations
    involving the incomplete LU preconditioning matrix generated by
    ``real_gen_precon_ilu``.

    For full information please refer to the NAG Library document for
    f11db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dbf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

          Mx = y is solved.

        `trans` = 'T'

          M^Tx = y is solved.

    n : int
        n, the order of the matrix M. This must be the same value as was
        supplied in the preceding call to ``real_gen_precon_ilu``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_gen_precon_ilu``.

    la : int
        This must be the same value returned by the preceding call to
        ``real_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

          Checks are carried on the values of `n`, `irow`, `icol`,
          `ipivp`, `ipivq`, `istr` and `idiag`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipivq`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivq`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivq`[i-1] >= 1 and `ipivq`[i-1] <= `n`.

        (`errno` 3)
            On entry, `ipivp`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivp`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivp`[i-1] >= 1 and `ipivp`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, `idiag`[i-1] appears to be incorrect: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*, and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_ilu_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivq[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivq[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivq[i-1] >= 1 and ipivq[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivp[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivp[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivp[i-1] >= 1 and ipivp[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idiag[i-1] appears to be incorrect:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_gen_solve_ilu(method, n, nnz, a, la, irow, icol, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of real sparse nonsymmetric linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, preconditioner computed by
    ``real_gen_precon_ilu``.

    ``real_gen_solve_ilu`` solves a real sparse nonsymmetric system of
    linear equations, represented in coordinate storage format, using a
    restarted generalized minimal residual (RGMRES), conjugate gradient
    squared (CGS), stabilized bi-conjugate gradient (BI-CGSTAB), or
    transpose-free quasi-minimal residual (TFQMR) method, with
    incomplete LU preconditioning.

    For full information please refer to the NAG Library document for
    f11dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dcf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    n : int
        n, the order of the matrix A. This must be the same value as was
        supplied in the preceding call to ``real_gen_precon_ilu``.

    nnz : int
        The number of nonzero elements in the matrix A. This must be the
        same value as was supplied in the preceding call to
        ``real_gen_precon_ilu``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_gen_precon_ilu``.

    la : int
        This must be the same value as was supplied in the preceding
        call to ``real_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    b : float, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method; otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 1 and `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            A serious error, code *<value>*, has occurred in an internal
            call. Check all function calls and array sizes. Seek expert
            help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_solve_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m >= 1 and m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_gen_precon_ssor_solve(trans, n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to real sparse nonsymmetric matrix.

    ``real_gen_precon_ssor_solve`` solves a system of linear equations
    involving the preconditioning matrix corresponding to SSOR applied
    to a real sparse nonsymmetric matrix, represented in coordinate
    storage format.

    For full information please refer to the NAG Library document for
    f11dd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11ddf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

          Mx = y is solved.

        `trans` = 'T'

          M^Tx = y is solved.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``real_gen_sort`` may be used to order the elements
        in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    rdiag : float, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

          Checks are carried on the values of `n`, `nnz`, `irow`, `icol`
          and `omega`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: 1 <= `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11ddft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineFloat64ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_gen_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, m, tol, maxitn, x):
    u"""
    Solution of real sparse nonsymmetric linear system, RGMRES, CGS,
    Bi-CGSTAB, or TFQMR method, Jacobi or SSOR preconditioner (Black
    Box).

    ``real_gen_solve_jacssor`` solves a real sparse nonsymmetric system
    of linear equations, represented in coordinate storage format, using
    a restarted generalized minimal residual (RGMRES), conjugate
    gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, without preconditioning, with Jacobi, or with SSOR
    preconditioning.

    For full information please refer to the NAG Library document for
    f11de

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11def.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'J'

          Jacobi.

        `precon` = 'S'

          Symmetric successive-over-relaxation.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements of the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``real_gen_sort`` may be used to order the elements
        in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized and is not referenced.

    b : float, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 1 and `m` <= MIN(`n`,*<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: `omega` > 0.0 and `omega` < 2.0.

        (`errno` 1)
            On entry, `precon` != 'N', 'J' or 'S': `precon` = *<value>*.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            A serious error has occurred in an internal call: IREVCM =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

        (`errno` 7)
            A serious error has occurred in an internal call: errno =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11deft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m >= 1 and m <= MIN(n,{:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: omega > 0.0 and omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon != \'N\', \'J\' or \'S\': precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'IREVCM = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'errno = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_gen_precon_bdilu(n, nnz, a, la, irow, icol, nb, istb, indb, lindb, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag, npivm):
    u"""
    Real sparse nonsymmetric linear system, incomplete LU factorization
    of local or overlapping diagonal blocks.

    ``real_gen_precon_bdilu`` computes a block diagonal incomplete LU
    factorization of a real sparse nonsymmetric matrix, represented in
    coordinate storage format.
    The diagonal blocks may be composed of arbitrary rows and the
    corresponding columns, and may overlap.
    This factorization can be used to provide a block Jacobi or additive
    Schwarz preconditioner, for use in combination with
    ``real_gen_solve_bdilu`` or ``real_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11df

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dff.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`la`), modified in place
        On entry: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``real_gen_sort`` may be used to
        order the elements in this way.

        On exit: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrices C_b, for b = 1,2,...,`nb` stored consecutively.
        Within each block the matrix elements are ordered by increasing
        row index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

        Note: the minimum value for `la` is only appropriate if `lfill`
        and `dtol` are set such that minimal fill-in occurs. If this is
        not the case then we recommend that `la` is set much larger than
        the minimum value indicated in the constraint.

    irow : int, shape (`la`), modified in place
        On entry: the row indices of the nonzero elements supplied in
        `a`.

        On exit: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        On entry: the column indices of the nonzero elements supplied in
        `a`.

        On exit: the column indices of the nonzero elements returned in
        `a`.

    nb : int
        The number of diagonal blocks to factorize.

    istb : int, shape (`nb`+1)
        `istb`[b-1], for b = 1,2,...,`nb`, holds the indices in arrays
        `indb`, `ipivp`, `ipivq` and `idiag` that, on successful exit
        from this function, define block b. `istb`[`nb`] holds the sum
        of the number of rows in all blocks plus `istb`[0].

    indb : int, shape (`lindb`)
        `indb` must hold the row indices appearing in each diagonal
        block, stored consecutively. Thus the elements
        `indb`[`istb`[b-1]-1] to `indb`[`istb`[b]-2] are the row indices
        in the bth block, for b = 1,2,...,`nb`.

    lindb : int
        The dimension of the arrays `indb`, `ipivp`, `ipivq` and
        `idiag`.

    lfill : int, shape (`nb`)
        If `lfill`[b-1] >= 0 its value is the maximum level of fill
        allowed in the decomposition of the block (see [Further
        Comments]). A negative value of `lfill`[b-1] indicates that
        `dtol`[b-1] will be used to control the fill in the block
        instead.

    dtol : float, shape (`nb`)
        If `lfill`[b-1] < 0 then `dtol`[b-1] is used as a drop tolerance
        in the block to control the fill-in (see [Further Comments]);
        otherwise `dtol`[b-1] is not referenced.

    pstrat : str, length 1, shape (`nb`)
        `pstrat`[b-1], for b = 1,2,...,`nb`, specifies the pivoting
        strategy to be adopted in the block as follows:

        `pstrat`[b-1] = 'N'

          No pivoting is carried out.

        `pstrat`[b-1] = 'U'

          Pivoting is carried out according to the user-defined input
          values of `ipivp` and `ipivq`.

        `pstrat`[b-1] = 'P'

          Partial pivoting by columns for stability is carried out.

        `pstrat`[b-1] = 'C'

          Complete pivoting by rows for sparsity, and by columns for
          stability, is carried out.

    milu : str, length 1, shape (`nb`)
        `milu`[b-1], for b = 1,2,...,`nb`, indicates whether or not the
        factorization in the block should be modified to preserve
        row-sums (see [Further Comments]).

        `milu`[b-1] = 'M'

          The factorization is modified.

        `milu`[b-1] = 'N'

          The factorization is not modified.

    ipivp : int, shape (`lindb`), modified in place
        On entry: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    ipivq : int, shape (`lindb`), modified in place
        On entry: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    istr : int, shape (`lindb`+1), modified in place
        On exit: `istr`[`istb`[b-1]+k-2], gives the index in the arrays
        `a`, `irow` and `icol` of row k of the matrix C_b, for k =
        1,2,...,`istb`[b]-`istb`[b-1], for b = 1,2,...,`nb`.

        `istr`[`istb`[`nb`]-1] contains `nnz`+`nnzc`+1.

    idiag : int, shape (`lindb`), modified in place
        On exit: `idiag`[`istb`[b-1]+k-2], gives the index in the arrays
        `a`, `irow` and `icol` of the diagonal element in row k of the
        matrix C_b, for k = 1,2,...,`istb`[b]-`istb`[b-1], for b =
        1,2,...,`nb`.

    npivm : int, shape (`nb`), modified in place
        On exit: if `npivm`[b-1] > 0 it gives the number of pivots which
        were modified during the factorization to ensure that M_b
        exists.

        If `npivm`[b-1] = -1 no pivot modifications were required, but a
        local restart occurred (see [Further Comments f11daf]). The
        quality of the preconditioner will generally depend on the
        returned values of `npivm`[b-1], for b = 1,2,...,`nb`.

        If `npivm`[b-1] is large, for some block, the preconditioner may
        not be satisfactory.

        In this case it may be advantageous to call
        ``real_gen_precon_bdilu`` again with an increased value of
        `lfill`[b-1], a reduced value of `dtol`[b-1], or `pstrat`[b-1] =
        'C'.

    Returns
    -------
    nnzc : int
        The sum total number of nonzero elements in the matrices C_b,
        for b = 1,2,...,`nb`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `dtol`[*<value>*] = *<value>*.

            Constraint: `dtol`[b-1] >= 0.0, for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `pstrat`[*<value>*] = *<value>*.

            Constraint: `pstrat`[b-1] = 'N', 'U', 'P' or 'C' for all b.

        (`errno` 1)
            On entry, `milu`[*<value>*] = *<value>*.

            Constraint: `milu`[b-1] = 'M' or 'N' for all b.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[j-1] <= `n`, for j = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* was repeated.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* was repeated.

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large.

            The decomposition has been terminated before completion.

            Either increase `la`, or reduce the fill by reducing
            `lfill`, or increasing `dtol`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input, lfill[nb].
        utils._EngineFloat64ArrayType.c_type, # Input, dtol[nb].
        utils._EngineCharArrayType.c_type, # Input, pstrat(len=1)[nb].
        utils._EngineCharArrayType.c_type, # Input, milu(len=1)[nb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Output, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[lindb].
        utils._BYREF_INT, # Output, nnzc.
        utils._EngineIntArrayType.c_type, # Output, npivm[nb].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _lfill = utils._EngineIntArrayType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _dtol = utils._EngineFloat64ArrayType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _pstrat_cls = utils._EngineCharArrayType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _pstrat = _pstrat_cls.to_ctypes()
    _pstrat_chrlen = _pstrat_cls.chrlen
    _milu_cls = utils._EngineCharArrayType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _milu = _milu_cls.to_ctypes()
    _milu_chrlen = _milu_cls.chrlen
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm_cls = utils._EngineIntArrayType(
        npivm, {'fun_name': fun_name, 'entity': 'npivm'},
        exp_shape=(_nb_py,),
        check_contents=False,
    )
    _npivm = _npivm_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat_chrlen,
        _milu_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1014:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: dtol[b-1] >= 0.0, for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: pstrat[b-1] = \'N\', \'U\', \'P\' or \'C\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: milu[b-1] = \'M\' or \'N\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[j-1] <= n, for j = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1211:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1212:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1213:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _npivm_cls.refresh_py(_npivm)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        return_data = _nnzc_py
    _err_class.manage()
    return return_data

def real_gen_solve_bdilu(method, n, nnz, a, la, irow, icol, nb, istb, indb, lindb, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of real sparse nonsymmetric linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, incomplete LU block diagonal
    preconditioner computed by ``real_gen_precon_bdilu``.

    ``real_gen_solve_bdilu`` solves a real sparse nonsymmetric system of
    linear equations, represented in coordinate storage format, using a
    restarted generalized minimal residual (RGMRES), conjugate gradient
    squared (CGS), stabilized bi-conjugate gradient (BI-CGSTAB), or
    transpose-free quasi-minimal residual (TFQMR) method, with block
    Jacobi or additive Schwarz preconditioning.

    For full information please refer to the NAG Library document for
    f11dg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dgf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    n : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    nnz : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    a : float, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    la : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    nb : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    istb : int, shape (`nb`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    indb : int, shape (`lindb`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    lindb : int
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    ipivp : int, shape (`lindb`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    ipivq : int, shape (`lindb`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    istr : int, shape (`lindb`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    idiag : int, shape (`lindb`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``real_gen_precon_bdilu``.

        The arrays `istb`, `indb` and `a` together with the the scalars
        `n`, `nnz`, `la`, `nb` and `lindb` must be the same values that
        were supplied in the preceding call to
        ``real_gen_precon_bdilu``.

    b : float, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),sqrt(n)epsilon) is
        used, where epsilon is the machine precision. Otherwise tau =
        max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: if `method` = 'RGMRES', 1 <= `m` <= min(`n`,
            *<value>*).

            If `method` = 'BICGSTAB', 1 <= `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``real_gen_precon_bdilu`` and ``real_gen_solve_bdilu``.

        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy may have been achieved. You should check
            the output value of `rnorm` for acceptability. This error
            code usually implies that your problem has been fully and
            satisfactorily solved to within or close to the accuracy
            available on your system. Further iterations are unlikely to
            improve on this situation.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_solve_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Input, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[lindb].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1019:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'RGMRES\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'If method = \'BICGSTAB\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1023:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to real_gen_precon_bdilu and real_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained. However a reasonable accuracy',
                    'may have been achieved. You should check the output value of rnorm for',
                    'acceptability. This error code usually implies that your problem has been',
                    'fully and satisfactorily solved to within or close to the accuracy available',
                    'on your system. Further iterations are unlikely to improve on this',
                    'situation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1510:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def real_gen_precon_jacobi(store, trans, init, niter, n, nnz, a, irow, icol, check, b, x, diag):
    u"""
    Real, sparse, symmetric or nonsymmetric, linear systems, line Jacobi
    preconditioner.

    ``real_gen_precon_jacobi`` computes the approximate solution of a
    real, symmetric or nonsymmetric, sparse system of linear equations
    applying a number of Jacobi iterations.
    It is expected that ``real_gen_precon_jacobi`` will be used as a
    preconditioner for the iterative solution of real sparse systems of
    equations.

    For full information please refer to the NAG Library document for
    f11dk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dkf.html

    Parameters
    ----------
    store : str, length 1
        Specifies whether the matrix A is stored using symmetric
        coordinate storage (SCS) (applicable only to a symmetric matrix
        A) or coordinate storage (CS) (applicable to both symmetric and
        non-symmetric matrices).

        `store` = 'N'

          The complete matrix A is stored in CS format.

        `store` = 'S'

          The lower triangle of the symmetric matrix A is stored in SCS
          format.

    trans : str, length 1
        If `store` = 'N', specifies whether the approximate solution of
        Ax = b or of A^Tx = b is required.

        `trans` = 'N'

          The approximate solution of Ax = b is calculated.

        `trans` = 'T'

          The approximate solution of A^Tx = b is calculated.

    init : str, length 1
        On first entry, `init` should be set to 'I', unless the diagonal
        elements of A are already stored in the array `diag`. If `diag`
        already contains the diagonal of A, it must be set to 'N'.

        `init` = 'N'

          `diag` must contain the diagonal of A.

        `init` = 'I'

          `diag` will store the diagonal of A on exit.

    niter : int
        The number of Jacobi iterations requested.

    n : int
        n, the order of the matrix A.

    nnz : int
        If `store` = 'N', the number of nonzero elements in the matrix
        A.

        If `store` = 'S', the number of nonzero elements in the lower
        triangle of the matrix A.

    a : float, shape (`nnz`)
        If `store` = 'N', the nonzero elements in the matrix A (CS
        format).

        If `store` = 'S', the nonzero elements in the lower triangle of
        the matrix A (SCS format).

        In both cases, the elements of either A or of its lower triangle
        must be ordered by increasing row index and by increasing column
        index within each row.

        Multiple entries for the same row and columns indices are not
        permitted.

        The function ``real_gen_sort`` or ``real_symm_sort`` may be used
        to reorder the elements in this way for CS and SCS storage,
        respectively.

    irow : int, shape (`nnz`)
        If `store` = 'N', the row indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    icol : int, shape (`nnz`)
        If `store` = 'N', the column indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    check : str, length 1
        Specifies whether or not the CS or SCS representation of the
        matrix A should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `nnz`, `irow`,
          `icol`; if `init` = 'N', `diag` is also checked.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    b : float, shape (`n`)
        The right-hand side vector b.

    x : float, shape (`n`), modified in place
        On exit: the approximate solution vector x_`niter`.

    diag : float, shape (`n`), modified in place
        On entry: if `init` = 'N', the diagonal elements of A.

        On exit: if `init` = 'N', unchanged on exit.

        If `init` = 'I', the diagonal elements of A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `niter` = *<value>*.

            Constraint: `niter` >= 1.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'N' or 'C'.

        (`errno` 1)
            On entry, `init` = *<value>*.

            Constraint: `init` = 'N' or 'I'.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `store` = *<value>*.

            Constraint: `store` = 'N' or 'S'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `n`.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[I-1] >= 1 and `irow`[I-1] <= `n`.

        (`errno` 4)
            On entry, the element `diag`[I-1] is zero: I = *<value>*.

        (`errno` 5)
            On entry, the diagonal element of the Ith row is zero or
            missing: I = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_precon_jacobi'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, store(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, init(len=1).
        utils._BYREF_INT, # Input, niter.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, diag[n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of store.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of init.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _store = utils._EngineCharScalarType(
        store, {'fun_name': fun_name, 'entity': 'store'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _init = utils._EngineCharScalarType(
        init, {'fun_name': fun_name, 'entity': 'init'}, exp_chrlen=1,
    ).to_ctypes()
    _niter = utils._EngineIntScalarType(
        niter, {'fun_name': fun_name, 'entity': 'niter'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _diag_cls = utils._EngineFloat64ArrayType(
        diag, {'fun_name': fun_name, 'entity': 'diag'},
        exp_shape=(_n_py,),
    )
    _diag = _diag_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _store,
        _trans,
        _init,
        _niter,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _b,
        _x,
        _diag,
        _work,
        _errbuf,
        _ierr,
        _store._length_,
        _trans._length_,
        _init._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, niter = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: niter >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'N\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: init = \'N\' or \'I\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, store = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: store = \'N\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[I-1] >= 1 and irow[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the element diag[I-1] is zero:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the diagonal element of the Ith row is zero or missing:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _diag_cls.refresh_py(_diag)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_precon_ilu(n, nnz, a, la, irow, icol, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag):
    u"""
    Complex sparse non-Hermitian linear systems, incomplete LU
    factorization.

    ``complex_gen_precon_ilu`` computes an incomplete LU factorization
    of a complex sparse non-Hermitian matrix, represented in coordinate
    storage format.
    This factorization may be used as a preconditioner in combination
    with ``complex_gen_solve_ilu`` or ``complex_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11dn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dnf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`la`), modified in place
        On entry: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``complex_gen_sort`` may be used to
        order the elements in this way.

        On exit: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrix C. Matrix elements are ordered by increasing row
        index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        On entry: the row indices of the nonzero elements supplied in
        `a`.

        On exit: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        On entry: the column indices of the nonzero elements supplied in
        `a`.

        On exit: the column indices of the nonzero elements returned in
        `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

          No pivoting is carried out.

        `pstrat` = 'U'

          Pivoting is carried out according to the user-defined input
          values of `ipivp` and `ipivq`.

        `pstrat` = 'P'

          Partial pivoting by columns for stability is carried out.

        `pstrat` = 'C'

          Complete pivoting by rows for sparsity, and by columns for
          stability, is carried out.

    milu : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row-sums (see [Further Comments]).

        `milu` = 'M'

          The factorization is modified.

        `milu` = 'N'

          The factorization is not modified.

    ipivp : int, shape (`n`), modified in place
        On entry: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1] must
        specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the pivot indices. If `ipivp`[k-1] = i and `ipivq`[k-1]
        = j then the element in row i and column j was used as the pivot
        at elimination stage k.

    ipivq : int, shape (`n`), modified in place
        On entry: if `pstrat` = 'U', `ipivp`[k-1] and `ipivq`[k-1] must
        specify the row and column indices of the element used as a
        pivot at elimination stage k. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the pivot indices. If `ipivp`[k-1] = i and `ipivq`[k-1]
        = j then the element in row i and column j was used as the pivot
        at elimination stage k.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    idiag : int, shape (`n`), modified in place
        On exit: `idiag`[i-1], for i = 1,2,...,`n`, holds the index of
        arrays `a`, `irow` and `icol` which holds the diagonal element
        in row i of the matrix C.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the matrix C.

    npivm : int
        If `npivm` > 0 it gives the number of pivots which were modified
        during the factorization to ensure that M exists.

        If `npivm` = -1 no pivot modifications were required, but a
        local restart occurred (see [Algorithmic Details]).

        The quality of the preconditioner will generally depend on the
        returned value of `npivm`.

        If `npivm` is large the preconditioner may not be satisfactory.

        In this case it may be advantageous to call
        ``complex_gen_precon_ilu`` again with an increased value of
        `lfill`, a reduced value of `dtol`, or set `pstrat` = 'C'.

        See also [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `milu` != 'M' or 'N': `milu` = *<value>*.

        (`errno` 1)
            On entry, `pstrat` != 'N', 'U', 'P' or 'C': `pstrat` =
            *<value>*.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `n`, for i = 1,2,...,`nnz`

        (`errno` 2)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivq` lies outside the
            range [1,`n`].

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipivp` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            reducing `lfill`, or increasing `dtol`.

        (`errno` 5)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._BYREF_STRING, # Input, milu(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[n].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _milu = utils._EngineCharScalarType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat._length_,
        _milu._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu != \'M\' or \'N\': milu = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat != \'N\', \'U\', \'P\' or \'C\': pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= n, for i = 1,2,...,nnz',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, irow[I-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivq lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipivp lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_precon_ilu_solve(trans, n, a, la, irow, icol, ipivp, ipivq, istr, idiag, check, y, x):
    u"""
    Solution of complex linear system involving incomplete LU
    preconditioning matrix generated by ``complex_gen_precon_ilu``.

    ``complex_gen_precon_ilu_solve`` solves a system of complex linear
    equations involving the incomplete LU preconditioning matrix
    generated by ``complex_gen_precon_ilu``.

    For full information please refer to the NAG Library document for
    f11dp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dpf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

          Mx = y is solved.

        `trans` = 'T'

          M^Tx = y is solved.

    n : int
        n, the order of the matrix M. This must be the same value as was
        supplied in the preceding call to ``complex_gen_precon_ilu``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_gen_precon_ilu``.

    la : int
        This must be the same value supplied in the preceding call to
        ``complex_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

          Checks are carried on the values of `n`, `irow`, `icol`,
          `ipivp`, `ipivq`, `istr` and `idiag`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipivq`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivq`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivq`[i-1] >= 1 and `ipivq`[i-1] <= `n`.

        (`errno` 3)
            On entry, `ipivp`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipivp`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipivp`[i-1] >= 1 and `ipivp`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, `idiag`[i-1] appears to be incorrect: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*, and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_ilu_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivq[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivq[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivq[i-1] >= 1 and ipivq[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipivp[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipivp[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipivp[i-1] >= 1 and ipivp[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idiag[i-1] appears to be incorrect:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_solve_ilu(method, n, nnz, a, la, irow, icol, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of complex sparse non-Hermitian linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, preconditioner computed by
    ``complex_gen_precon_ilu`` (Black Box).

    ``complex_gen_solve_ilu`` solves a complex sparse non-Hermitian
    system of linear equations, represented in coordinate storage
    format, using a restarted generalized minimal residual (RGMRES),
    conjugate gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, with incomplete LU preconditioning.

    For full information please refer to the NAG Library document for
    f11dq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dqf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    n : int
        n, the order of the matrix A. This must be the same value as was
        supplied in the preceding call to ``complex_gen_precon_ilu``.

    nnz : int
        The number of nonzero elements in the matrix A. This must be the
        same value as was supplied in the preceding call to
        ``complex_gen_precon_ilu``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_gen_precon_ilu``.

    la : int
        This must be the same value as was supplied in the preceding
        call to ``complex_gen_precon_ilu``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivp : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    ipivq : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    idiag : int, shape (`n`)
        The values returned in arrays `irow`, `icol`, `ipivp`, `ipivq`,
        `istr` and `idiag` by a previous call to
        ``complex_gen_precon_ilu``.

        `ipivp` and `ipivq` are restored on exit.

    b : complex, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 1 and `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*, and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_solve_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[n].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[n].
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_n_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_n_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m >= 1 and m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_gen_precon_ssor_solve(trans, n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to complex sparse non-Hermitian matrix.

    ``complex_gen_precon_ssor_solve`` solves a system of linear
    equations involving the preconditioning matrix corresponding to SSOR
    applied to a complex sparse non-Hermitian matrix, represented in
    coordinate storage format.

    For full information please refer to the NAG Library document for
    f11dr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11drf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix M is transposed.

        `trans` = 'N'

          Mx = y is solved.

        `trans` = 'T'

          M^Hx = y is solved.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``complex_gen_sort`` may be used to order the
        elements in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in `a`.

    rdiag : complex, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix M
        should be checked.

        `check` = 'C'

          Checks are carried on the values of `n`, `nnz`, `irow`, `icol`
          and `omega`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` != 'C' or 'N': `check` = *<value>*.

        (`errno` 1)
            On entry, `trans` != 'N' or 'T': `trans` = *<value>*.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11drft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineComplex128ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineComplex128ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check != \'C\' or \'N\': check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans != \'N\' or \'T\': trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, m, tol, maxitn, x):
    u"""
    Solution of complex sparse non-Hermitian linear system, RGMRES, CGS,
    Bi-CGSTAB or TFQMR method, Jacobi or SSOR preconditioner Black Box.

    ``complex_gen_solve_jacssor`` solves a complex sparse non-Hermitian
    system of linear equations, represented in coordinate storage
    format, using a restarted generalized minimal residual (RGMRES),
    conjugate gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, without preconditioning, with Jacobi, or with SSOR
    preconditioning.

    For full information please refer to the NAG Library document for
    f11ds

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dsf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'J'

          Jacobi.

        `precon` = 'S'

          Symmetric successive-over-relaxation (SSOR).

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements of the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``complex_gen_sort`` may be used to order the
        elements in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized and is not referenced.

    b : complex, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: 0 < `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0

        (`errno` 1)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N', 'J' or 'S'.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.

        (`errno` 7)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 < m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: precon = \'N\', \'J\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_gen_precon_bdilu(n, nnz, a, la, irow, icol, nb, istb, indb, lindb, lfill, dtol, pstrat, milu, ipivp, ipivq, istr, idiag, npivm):
    u"""
    Complex, sparse, non-Hermitian linear system, incomplete LU
    factorization of local or overlapping diagonal blocks.

    ``complex_gen_precon_bdilu`` computes a block diagonal incomplete LU
    factorization of a complex sparse non-Hermitian matrix, represented
    in coordinate storage format.
    The diagonal blocks may be composed of arbitrary rows and the
    corresponding columns, and may overlap.
    This factorization can be used to provide a block Jacobi or additive
    Schwarz preconditioner, for use in combination with
    ``complex_gen_solve_bdilu`` or ``complex_gen_basic_solver``.

    For full information please refer to the NAG Library document for
    f11dt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dtf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`la`), modified in place
        On entry: the nonzero elements in the matrix A, ordered by
        increasing row index, and by increasing column index within each
        row. Multiple entries for the same row and column indices are
        not permitted. The function ``complex_gen_sort`` may be used to
        order the elements in this way.

        On exit: the first `nnz` entries of `a` contain the nonzero
        elements of A and the next `nnzc` entries contain the elements
        of the matrices C_b, for b = 1,2,...,`nb` stored consecutively.
        Within each block the matrix elements are ordered by increasing
        row index, and by increasing column index within each row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

        Note: the minimum value for `la` is only appropriate if `lfill`
        and `dtol` are set such that minimal fill-in occurs. If this is
        not the case then we recommend that `la` is set much larger than
        the minimum value indicated in the constraint.

    irow : int, shape (`la`), modified in place
        On entry: the row indices of the nonzero elements supplied in
        `a`.

        On exit: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        On entry: the column indices of the nonzero elements supplied in
        `a`.

        On exit: the column indices of the nonzero elements returned in
        `a`.

    nb : int
        The number of diagonal blocks to factorize.

    istb : int, shape (`nb`+1)
        `istb`[b-1], for b = 1,2,...,`nb`, holds the indices in arrays
        `indb`, `ipivp`, `ipivq` and `idiag` that, on successful exit
        from this function, define block b. `istb`[`nb`] holds the sum
        of the number of rows in all blocks plus `istb`[0].

    indb : int, shape (`lindb`)
        `indb` must hold the row indices appearing in each diagonal
        block, stored consecutively. Thus the elements
        `indb`[`istb`[b-1]-1] to `indb`[`istb`[b]-2] are the row indices
        in the bth block, for b = 1,2,...,`nb`.

    lindb : int
        The dimension of the arrays `indb`, `ipivp`, `ipivq` and
        `idiag`.

    lfill : int, shape (`nb`)
        If `lfill`[b-1] >= 0 its value is the maximum level of fill
        allowed in the decomposition of the block b (see [Further
        Comments]). A negative value of `lfill`[b-1] indicates that
        `dtol`[b-1] will be used to control the fill in block b instead.

    dtol : float, shape (`nb`)
        If `lfill`[b-1] < 0 then `dtol`[b-1] is used as a drop tolerance
        in block b to control the fill-in (see [Further Comments]);
        otherwise `dtol`[b-1] is not referenced.

    pstrat : str, length 1, shape (`nb`)
        `pstrat`[b-1], for b = 1,2,...,`nb`, specifies the pivoting
        strategy to be adopted in the block as follows:

        `pstrat`[b-1] = 'N'

          No pivoting is carried out.

        `pstrat`[b-1] = 'U'

          Pivoting is carried out according to the user-defined input
          values of `ipivp` and `ipivq`.

        `pstrat`[b-1] = 'P'

          Partial pivoting by columns for stability is carried out.

        `pstrat`[b-1] = 'C'

          Complete pivoting by rows for sparsity, and by columns for
          stability, is carried out.

    milu : str, length 1, shape (`nb`)
        `milu`[b-1], for b = 1,2,...,`nb`, indicates whether or not the
        factorization in block b should be modified to preserve row-sums
        (see [Further Comments]).

        `milu`[b-1] = 'M'

          The factorization is modified.

        `milu`[b-1] = 'N'

          The factorization is not modified.

    ipivp : int, shape (`lindb`), modified in place
        On entry: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    ipivq : int, shape (`lindb`), modified in place
        On entry: if `pstrat`[b-1] = 'U', `ipivp`[`istb`[b-1]+k-2] and
        `ipivq`[`istb`[b-1]+k-2] must specify the row and column indices
        of the element used as a pivot at elimination stage k of the
        factorization of the block. Otherwise `ipivp` and `ipivq` need
        not be initialized.

        On exit: the row and column indices of the pivot elements,
        arranged consecutively for each block, as for `indb`. If
        `ipivp`[`istb`[b-1]+k-2] = i and `ipivq`[`istb`[b-1]+k-2] = j,
        the element in row i and column j of A_b was used as the pivot
        at elimination stage k.

    istr : int, shape (`lindb`+1), modified in place
        On exit: `istr`[`istb`[b-1]+k-2], gives the index in the arrays
        `a`, `irow` and `icol` of row k of the matrix C_b, for k =
        1,2,...,`istb`[b]-`istb`[b-1], for b = 1,2,...,`nb`.

        `istr`[`istb`[`nb`]-1] contains `nnz`+`nnzc`+1.

    idiag : int, shape (`lindb`), modified in place
        On exit: `idiag`[`istb`[b-1]+k-2], gives the index in the arrays
        `a`, `irow` and `icol` of the diagonal element in row k of the
        matrix C_b, for k = 1,2,...,`istb`[b]-`istb`[b-1], for b =
        1,2,...,`nb`.

    npivm : int, shape (`nb`), modified in place
        On exit: if `npivm`[b-1] > 0 it gives the number of pivots which
        were modified during the factorization to ensure that M_b
        exists.

        If `npivm`[b-1] = -1 no pivot modifications were required, but a
        local restart occurred (see [Further Comments f11dnf]). The
        quality of the preconditioner will generally depend on the
        returned values of `npivm`[b-1], for b = 1,2,...,`nb`.

        If `npivm`[b-1] is large, for some b, the preconditioner may not
        be satisfactory.

        In this case it may be advantageous to call
        ``complex_gen_precon_bdilu`` again with an increased value of
        `lfill`[b-1], a reduced value of `dtol`[b-1], or `pstrat`[b-1] =
        'C'.

    Returns
    -------
    nnzc : int
        The sum total number of nonzero elements in the matrices C_b,
        for b = 1,2,...,`nb`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `dtol`[*<value>*] = *<value>*.

            Constraint: `dtol`[b-1] >= 0.0, for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `pstrat`[*<value>*] = *<value>*.

            Constraint: `pstrat`[b-1] = 'N', 'U', 'P' or 'C' for all b.

        (`errno` 1)
            On entry, `milu`[*<value>*] = *<value>*.

            Constraint: `milu`[b-1] = 'M' or 'N' for all b.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[j-1] <= `n`, for j = 1,2,...,`nnz`.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivp` for block
            *<value>* was repeated.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* lies outside its range.

        (`errno` 3)
            On entry, the user-supplied value of `ipivq` for block
            *<value>* was repeated.

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large.

            The decomposition has been terminated before completion.

            Either increase `la`, or reduce the fill by reducing
            `lfill`, or increasing `dtol`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dtft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input, lfill[nb].
        utils._EngineFloat64ArrayType.c_type, # Input, dtol[nb].
        utils._EngineCharArrayType.c_type, # Input, pstrat(len=1)[nb].
        utils._EngineCharArrayType.c_type, # Input, milu(len=1)[nb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Output, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Output, idiag[lindb].
        utils._BYREF_INT, # Output, nnzc.
        utils._EngineIntArrayType.c_type, # Output, npivm[nb].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of milu.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _lfill = utils._EngineIntArrayType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _dtol = utils._EngineFloat64ArrayType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _pstrat_cls = utils._EngineCharArrayType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _pstrat = _pstrat_cls.to_ctypes()
    _pstrat_chrlen = _pstrat_cls.chrlen
    _milu_cls = utils._EngineCharArrayType(
        milu, {'fun_name': fun_name, 'entity': 'milu'}, exp_chrlen=1,
        exp_shape=(_nb_py,),
    )
    _milu = _milu_cls.to_ctypes()
    _milu_chrlen = _milu_cls.chrlen
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _idiag_cls = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
        check_contents=False,
    )
    _idiag = _idiag_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm_cls = utils._EngineIntArrayType(
        npivm, {'fun_name': fun_name, 'entity': 'npivm'},
        exp_shape=(_nb_py,),
        check_contents=False,
    )
    _npivm = _npivm_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _lfill,
        _dtol,
        _pstrat,
        _milu,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _pstrat_chrlen,
        _milu_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1014:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: dtol[b-1] >= 0.0, for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: pstrat[b-1] = \'N\', \'U\', \'P\' or \'C\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, milu[{:s}] = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: milu[b-1] = \'M\' or \'N\' for all b.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[j-1] <= n, for j = 1,2,...,nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1211:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivp for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1212:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lies outside its range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1213:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the user-supplied value of ipivq for block {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'was repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by reducing lfill,',
                    'or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _istr_cls.refresh_py(_istr)
        _idiag_cls.refresh_py(_idiag)
        _npivm_cls.refresh_py(_npivm)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        return_data = _nnzc_py
    _err_class.manage()
    return return_data

def complex_gen_solve_bdilu(method, n, nnz, a, la, irow, icol, nb, istb, indb, lindb, ipivp, ipivq, istr, idiag, b, m, tol, maxitn, x):
    u"""
    Solution of complex, sparse, non-Hermitian linear system, RGMRES,
    CGS, Bi-CGSTAB or TFQMR method, incomplete LU block diagonal
    preconditioner computed by ``complex_gen_precon_bdilu``.

    ``complex_gen_solve_bdilu`` solves a complex sparse non-Hermitian
    system of linear equations, represented in coordinate storage
    format, using a restarted generalized minimal residual (RGMRES),
    conjugate gradient squared (CGS), stabilized bi-conjugate gradient
    (BI-CGSTAB), or transpose-free quasi-minimal residual (TFQMR)
    method, with block Jacobi or additive Schwarz preconditioning.

    For full information please refer to the NAG Library document for
    f11du

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11duf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'RGMRES'

          Restarted generalized minimum residual method.

        `method` = 'CGS'

          Conjugate gradient squared method.

        `method` = 'BICGSTAB'

          Bi-conjugate gradient stabilized (l) method.

        `method` = 'TFQMR'

          Transpose-free quasi-minimal residual method.

    n : int
        The value of `n` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    nnz : int
        The value of `nnz` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    a : complex, shape (`la`)
        The value of `a` returned by a previous call to
        ``complex_gen_precon_bdilu``

    la : int
        The value of `la` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    irow : int, shape (`la`)
        The value of `irow` returned by a previous call to
        ``complex_gen_precon_bdilu``

    icol : int, shape (`la`)
        The value of `icol` returned by a previous call to
        ``complex_gen_precon_bdilu``

    nb : int
        The value of `nb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    istb : int, shape (`nb`+1)
        The value of `istb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    indb : int, shape (`lindb`)
        The value of `indb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    lindb : int
        The value of `lindb` as supplied to a previous call to
        ``complex_gen_precon_bdilu``

    ipivp : int, shape (`lindb`), modified in place
        The value of `ipivp` returned by a previous call to
        ``complex_gen_precon_bdilu``

    ipivq : int, shape (`lindb`), modified in place
        The value of `ipivq` returned by a previous call to
        ``complex_gen_precon_bdilu``

    istr : int, shape (`lindb`+1)
        The value of `istr` returned by a previous call to
        ``complex_gen_precon_bdilu``

    idiag : int, shape (`lindb`)
        The value of `idiag` returned by a previous call to
        ``complex_gen_precon_bdilu``

    b : complex, shape (`n`)
        The right-hand side vector b.

    m : int
        If `method` = 'RGMRES', `m` is the dimension of the restart
        subspace.

        If `method` = 'BICGSTAB', `m` is the order l of the polynomial
        BI-CGSTAB method.

        Otherwise, `m` is not referenced.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),sqrt(n)epsilon) is
        used, where epsilon is the machine precision. Otherwise tau =
        max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nb` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nb` <= `n`.

        (`errno` 1)
            On entry, `istb`[0] = *<value>*.

            Constraint: `istb`[0] >= 1.

        (`errno` 1)
            On entry, for b = *<value>*, `istb`[b] = *<value>* and
            `istb`[b-1] = *<value>*.

            Constraint: `istb`[b] > `istb`[b-1], for b = 1,2,...,`nb`.

        (`errno` 1)
            On entry, `indb`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `indb`[m-1] <= `n`, for m =
            1,2,...,`istb`[`nb`]-1

        (`errno` 1)
            On entry, `lindb` = *<value>*, `istb`[`nb`]-1 = *<value>*
            and `nb` = *<value>*.

            Constraint: `lindb` >= `istb`[`nb`]-1.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'RGMRES', 'CGS' or 'BICGSTAB'.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: if `method` = 'RGMRES', 1 <= `m` <= min(`n`,
            *<value>*).

            If `method` = 'BICGSTAB', 1 <= `m` <= min(`n`, *<value>*).

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 2)
            On entry, `irow`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 2)
            On entry, `icol`[*<value>*] = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `n`, for i = 1,2,...,`nnz`.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 2)
            On entry, element *<value>* of `a` was out of order.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 2)
            On entry, location *<value>* of (`irow`,`icol`) was a
            duplicate.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 3)
            The CS representation of the preconditioner is invalid.

            Check that `a`, `irow`, `icol`, `ipivp`, `ipivq`, `istr` and
            `idiag` have not been corrupted between calls to
            ``complex_gen_precon_bdilu`` and
            ``complex_gen_solve_bdilu``.

        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            Algorithmic breakdown. A solution is returned, although it
            is possible that it is completely inaccurate.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_solve_bdilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11duft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._BYREF_INT, # Input, nb.
        utils._EngineIntArrayType.c_type, # Input, istb[nb+1].
        utils._EngineIntArrayType.c_type, # Input, indb[lindb].
        utils._BYREF_INT, # Input, lindb.
        utils._EngineIntArrayType.c_type, # Input/Output, ipivp[lindb].
        utils._EngineIntArrayType.c_type, # Input/Output, ipivq[lindb].
        utils._EngineIntArrayType.c_type, # Input, istr[lindb+1].
        utils._EngineIntArrayType.c_type, # Input, idiag[lindb].
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _istb = utils._EngineIntArrayType(
        istb, {'fun_name': fun_name, 'entity': 'istb'},
        exp_shape=(_nb_py+1,),
    ).to_ctypes()
    _lindb = utils._EngineIntScalarType(
        lindb, {'fun_name': fun_name, 'entity': 'lindb'},
    ).to_ctypes()
    _lindb_py = utils._EngineIntScalarType.to_py(_lindb)
    _indb = utils._EngineIntArrayType(
        indb, {'fun_name': fun_name, 'entity': 'indb'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _ipivp_cls = utils._EngineIntArrayType(
        ipivp, {'fun_name': fun_name, 'entity': 'ipivp'},
        exp_shape=(_lindb_py,),
    )
    _ipivp = _ipivp_cls.to_ctypes()
    _ipivq_cls = utils._EngineIntArrayType(
        ipivq, {'fun_name': fun_name, 'entity': 'ipivq'},
        exp_shape=(_lindb_py,),
    )
    _ipivq = _ipivq_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_lindb_py+1,),
    ).to_ctypes()
    _idiag = utils._EngineIntArrayType(
        idiag, {'fun_name': fun_name, 'entity': 'idiag'},
        exp_shape=(_lindb_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _nb,
        _istb,
        _indb,
        _lindb,
        _ipivp,
        _ipivq,
        _istr,
        _idiag,
        _b,
        _m,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nb <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istb[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istb[0] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1017:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, for b = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[b] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'istb[b-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istb[b] > istb[b-1], for b = 1,2,...,nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indb[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= indb[m-1] <= n, for m = 1,2,...,istb[nb]-1',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1019:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lindb = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'istb[nb]-1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: lindb >= istb[nb]-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'RGMRES\', \'CGS\' or \'BICGSTAB\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1021:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if method = \'RGMRES\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'If method = \'BICGSTAB\', 1 <= m <= min(n, {:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1022:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1023:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irow[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icol[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= n, for i = 1,2,...,nnz.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element {:s} of a was out of order.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, location {:s} of (irow,icol) was a duplicate.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1210:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The CS representation of the preconditioner is invalid.',
                    'Check that a, irow, icol, ipivp, ipivq, istr and idiag',
                    'have not been corrupted between calls to complex_gen_precon_bdilu and complex_gen_solve_bdilu.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1310:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1410:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1510:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Algorithmic breakdown. A solution is returned, although',
                    'it is possible that it is completely inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipivp_cls.refresh_py(_ipivp)
        _ipivq_cls.refresh_py(_ipivq)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def complex_gen_precon_jacobi(store, trans, init, niter, n, nnz, a, irow, icol, check, b, x, diag):
    u"""
    Complex, sparse, Hermitian or non-Hermitian, linear systems, line
    Jacobi preconditioner.

    ``complex_gen_precon_jacobi`` computes the approximate solution of a
    complex, Hermitian or non-Hermitian, sparse system of linear
    equations applying a number of Jacobi iterations.
    It is expected that ``complex_gen_precon_jacobi`` will be used as a
    preconditioner for the iterative solution of complex sparse systems
    of equations.

    For full information please refer to the NAG Library document for
    f11dx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11dxf.html

    Parameters
    ----------
    store : str, length 1
        Specifies whether the matrix A is stored using symmetric
        coordinate storage (SCS) (applicable only to a Hermitian matrix
        A) or coordinate storage (CS) (applicable to both Hermitian and
        non-Hermitian matrices).

        `store` = 'N'

          The complete matrix A is stored in CS format.

        `store` = 'S'

          The lower triangle of the Hermitian matrix A is stored in SCS
          format.

    trans : str, length 1
        If `store` = 'N', specifies whether the approximate solution of
        Ax = b or of A^Hx = b is required.

        `trans` = 'N'

          The approximate solution of Ax = b is calculated.

        `trans` = 'T'

          The approximate solution of A^Hx = b is calculated.

    init : str, length 1
        On first entry, `init` should be set to 'I', unless the diagonal
        elements of A are already stored in the array `diag`. If `diag`
        already contains the diagonal of A, it must be set to 'N'.

        `init` = 'N'

          `diag` must contain the diagonal of A.

        `init` = 'I'

          `diag` will store the diagonal of A on exit.

    niter : int
        The number of Jacobi iterations requested.

    n : int
        n, the order of the matrix A.

    nnz : int
        If `store` = 'N', the number of nonzero elements in the matrix
        A.

        If `store` = 'S', the number of nonzero elements in the lower
        triangle of the matrix A.

    a : complex, shape (`nnz`)
        If `store` = 'N', the nonzero elements in the matrix A (CS
        format).

        If `store` = 'S', the nonzero elements in the lower triangle of
        the matrix A (SCS format).

        In both cases, the elements of either A or of its lower triangle
        must be ordered by increasing row index and by increasing column
        index within each row.

        Multiple entries for the same row and columns indices are not
        permitted.

        The function ``complex_gen_sort`` or ``complex_herm_sort`` may
        be used to reorder the elements in this way for CS and SCS
        storage, respectively.

    irow : int, shape (`nnz`)
        If `store` = 'N', the row indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    icol : int, shape (`nnz`)
        If `store` = 'N', the column indices of the nonzero elements
        supplied in `a`.

        If `store` = 'S', the row and column indices of the nonzero
        elements of the lower triangle of the matrix A supplied in `a`.

    check : str, length 1
        Specifies whether or not the CS or SCS representation of the
        matrix A should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `nnz`, `irow`,
          `icol`; if `init` = 'N', `diag` is also checked.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    b : complex, shape (`n`)
        The right-hand side vector b.

    x : complex, shape (`n`), modified in place
        On exit: the approximate solution vector x_`niter`.

    diag : complex, shape (`n`), modified in place
        On entry: if `init` = 'N', the diagonal elements of A.

        On exit: if `init` = 'N', unchanged on exit.

        If `init` = 'I', the diagonal elements of A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `niter` = *<value>*.

            Constraint: `niter` >= 1.

        (`errno` 1)
            On entry, `check` != 'N' or 'C': `check` = *<value>*.

        (`errno` 1)
            On entry, `init` != 'N' or 'I': `init` = *<value>*.

        (`errno` 1)
            On entry, `trans` != 'N' or 'T': `trans` = *<value>*.

        (`errno` 1)
            On entry, `store` != 'N' or 'S': `store` = *<value>*.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[I-1] >= 1 and `irow`[I-1] <= `n`.

        (`errno` 4)
            On entry, the element `diag`[I-1] is zero: I = *<value>*.

        (`errno` 5)
            On entry, the diagonal element of the Ith row is zero or
            missing: I = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_precon_jacobi'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11dxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, store(len=1).
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_STRING, # Input, init(len=1).
        utils._BYREF_INT, # Input, niter.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, diag[n].
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of store.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of init.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _store = utils._EngineCharScalarType(
        store, {'fun_name': fun_name, 'entity': 'store'}, exp_chrlen=1,
    ).to_ctypes()
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _init = utils._EngineCharScalarType(
        init, {'fun_name': fun_name, 'entity': 'init'}, exp_chrlen=1,
    ).to_ctypes()
    _niter = utils._EngineIntScalarType(
        niter, {'fun_name': fun_name, 'entity': 'niter'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _diag_cls = utils._EngineComplex128ArrayType(
        diag, {'fun_name': fun_name, 'entity': 'diag'},
        exp_shape=(_n_py,),
    )
    _diag = _diag_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _store,
        _trans,
        _init,
        _niter,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _b,
        _x,
        _diag,
        _work,
        _errbuf,
        _ierr,
        _store._length_,
        _trans._length_,
        _init._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, niter = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: niter >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check != \'N\' or \'C\': check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init != \'N\' or \'I\': init = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans != \'N\' or \'T\': trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, store != \'N\' or \'S\': store = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[I-1] >= 1 and irow[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the element diag[I-1] is zero:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the diagonal element of the Ith row is zero or missing:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _diag_cls.refresh_py(_diag)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_basic_setup(method, precon, sigcmp, norm, weight, iterm, n, tol, maxitn, anorm, sigmax, sigtol, maxits, monit, lwork):
    u"""
    Real sparse symmetric linear systems, setup for
    ``real_symm_basic_solver``.

    ``real_symm_basic_setup`` is a setup function, the first in a suite
    of three functions for the iterative solution of a symmetric system
    of simultaneous linear equations. ``real_symm_basic_setup`` must be
    called before the iterative solver, ``real_symm_basic_solver``.
    The third function in the suite, ``real_symm_basic_diag``, can be
    used to return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    symmetric systems of equations.

    For full information please refer to the NAG Library document for
    f11gd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11gdf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'CG'

          Conjugate gradient method (CG).

        `method` = 'SYMMLQ'

          Lanczos method (SYMMLQ).

        `method` = 'MINRES'

          Minimum residual method (MINRES).

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'P'

          Preconditioning.

    sigcmp : str, length 1
        Determines whether an estimate of sigma_1(A_) =
        ||E^-1AE^invTH||_2, the largest singular value of the
        preconditioned matrix of the coefficients, is to be computed
        using the bisection method on the sequence of tridiagonal
        matrices {T_k} generated during the iteration. Note that A_= A
        when a preconditioner is not used.

        If `sigmax` > 0.0 , i.e., when sigma_1(A_) is supplied, the
        value of `sigcmp` is ignored.

        `sigcmp` = 'S'

          sigma_1(A_) is to be computed using the bisection method.

        `sigcmp` = 'N'

          The bisection method is not used.

          If the termination criterion (3) is used, requiring
          sigma_1(A_), an inexpensive estimate is computed and used (see
          [Description]).

        It is not used if `method` = 'MINRES'.

    norm : str, length 1
        If `method` = 'CG' or 'SYMMLQ', `norm` defines the matrix and
        vector norm to be used in the termination criteria.

        `norm` = '1'

          Use the l_1 norm.

        `norm` = 'I'

          Use the l_infinity norm.

        `norm` = '2'

          Use the l_2 norm.

        It has no effect if `method` = 'MINRES'.

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the vector norms used in the computation of termination
        criterion (2) (`iterm` = 1): ||v||_p^(w) = ||v^(w)||_p, where
        v_i^(w) = w_i v_i, for i = 1,2,...,n. The suffix p = 1,2,
        infinity denotes the vector norm used, as specified by the
        argument `norm`. Note that weights cannot be used when `iterm` =
        2, i.e., when criterion (3) is used.

        `weight` = 'W'

          User-supplied weights are to be used and must be supplied on
          initial entry to ``real_symm_basic_solver``.

        `weight` = 'N'

          All weights are implicitly set equal to one. Weights do not
          need to be supplied on initial entry to
          ``real_symm_basic_solver``.

        It has no effect if `method` = 'MINRES'.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

          Use the termination criterion defined in (2) (both conjugate
          gradient and Lanczos (SYMMLQ) methods).

        `iterm` = 2

          Use the termination criterion defined in (3) (Lanczos method
          (SYMMLQ) only).

        It has no effect if `method` = 'MINRES'.

    n : int
        n, the order of the matrix A.

    tol : float
        The tolerance tau for the termination criterion.

        If `tol` <= 0.0, tau = max(sqrt(epsilon), sqrt(n)epsilon) is
        used, where epsilon is the machine precision.

        Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', ||A||_1
        = ||A||_infinity is estimated internally by
        ``real_symm_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

        It has no effect if `method` = 'MINRES'.

    sigmax : float
        If `sigmax` > 0.0, the value of sigma_1(A_) =
        ||E^-1AE^invTH||_2.

        If `sigmax` <= 0.0, sigma_1(A_) is estimated by
        ``real_symm_basic_solver`` when either `sigcmp` = 'S' or
        termination criterion (3) (`iterm` = 2) is employed, though it
        will be used only in the latter case.

        Otherwise, or if `method` = 'MINRES', `sigmax` is not
        referenced.

    sigtol : float
        The tolerance used in assessing the convergence of the estimate
        of sigma_1(A_) = ||A_||_2 when the bisection method is used.

        If `sigtol` <= 0.0, the default value `sigtol` = 0.01 is used.

        The actual value used is max(`sigtol`, epsilon).

        If `sigcmp` = 'N' or `sigmax` > 0.0, `sigtol` is not referenced.

        It has no effect if `method` = 'MINRES'.

    maxits : int
        The maximum iteration number k = `maxits` for which sigma_1(T_k)
        is computed by bisection (see also [Description]). If `sigcmp` =
        'N' or `sigmax` > 0.0, or if `method` = 'MINRES', `maxits` is
        not referenced.

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``real_symm_basic_solver``: the current solution and
        residual iterates will be returned by ``real_symm_basic_solver``
        and a call to ``real_symm_basic_diag`` made possible every
        `monit` iterations, starting from the (`monit`)th. Otherwise, no
        monitoring takes place.

        There are some additional computational costs involved in
        monitoring the solution and residual vectors when the Lanczos
        method (SYMMLQ) is used.

    lwork : int
        The dimension of the array `comm`['work'].

    Returns
    -------
    lwreq : int
        The minimum amount of workspace required by
        ``real_symm_basic_solver``. (See also [Parameters].)

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -14)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -13)
            On entry, `sigcmp` = 'S', `sigmax` <= 0.0, `maxits` =
            *<value>* and `maxitn` = *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `maxits`
            <= `maxitn`.

        (`errno` -13)
            On entry, `sigcmp` = 'S', `sigmax` <= 0.0 and `maxits` =
            *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `maxits`
            >= 1.

        (`errno` -12)
            On entry, `sigcmp` = 'S', `sigmax` <= 0.0 and `sigtol` =
            *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `sigtol`
            < 1.0.

        (`errno` -10)
            On entry, `iterm` = 1, `norm` = '2' and `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2', `anorm` > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -6)
            On entry, `iterm` = 2 and `method` = 'CG'.

            Constraint: if `iterm` = 2, `method` != 'CG'.

        (`errno` -6)
            On entry, `iterm` = 2 and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2'.

        (`errno` -6)
            On entry, `iterm` = 2 and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -6)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -5)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -4)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -3)
            On entry, `sigcmp` = *<value>*.

            Constraint: `sigcmp` = 'S' or 'N'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG', 'SYMMLQ' or 'MINRES'.

        (`errno` 1)
            ``real_symm_basic_setup`` has been called out of sequence:
            either ``real_symm_basic_setup`` has been called twice or
            ``real_symm_basic_solver`` has not terminated its current
            task.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, sigcmp(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_FLOAT64, # Input, sigtol.
        utils._BYREF_INT, # Input, maxits.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineFloat64ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of sigcmp.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _sigcmp = utils._EngineCharScalarType(
        sigcmp, {'fun_name': fun_name, 'entity': 'sigcmp'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _sigtol = utils._EngineFloat64ScalarType(
        sigtol, {'fun_name': fun_name, 'entity': 'sigtol'},
    ).to_ctypes()
    _maxits = utils._EngineIntScalarType(
        maxits, {'fun_name': fun_name, 'entity': 'maxits'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType(
        lwork, {'fun_name': fun_name, 'entity': 'lwork'},
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(120)
    _comm_py = {}
    _comm_py['work'] = _work
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _sigcmp,
        _norm,
        _weight,
        _iterm,
        _n,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _sigtol,
        _maxits,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _sigcmp._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _monit_py = utils._EngineIntScalarType.to_py(_monit)
        _maxitn_py = utils._EngineIntScalarType.to_py(_maxitn)
        _maxits_py = utils._EngineIntScalarType.to_py(_maxits)
        _sigtol_py = utils._EngineFloat64ScalarType.to_py(_sigtol)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _tol_py = utils._EngineFloat64ScalarType.to_py(_tol)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        _weight_py = utils._EngineCharScalarType.to_py(_weight)
        _iterm_py = utils._EngineIntScalarType.to_py(_iterm)
        _sigcmp_py = utils._EngineCharScalarType.to_py(_sigcmp)
        _precon_py = utils._EngineCharScalarType.to_py(_precon)
        _method_py = utils._EngineCharScalarType.to_py(_method)
        if _ierr_py == -14 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_monit_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -13 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'S\', sigmax <= 0.0,',
                    'maxits = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxits_py),
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0,',
                    'maxits <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -13 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'S\', sigmax <= 0.0 and',
                    'maxits = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxits_py),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0,',
                    'maxits >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -12 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'S\', sigmax <= 0.0 and',
                    'sigtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_sigtol_py),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0,',
                    'sigtol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 1, norm = \'2\' and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT.format(_maxitn_py),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_tol_py),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and method = \'CG\'.',
                    'Constraint: if iterm = 2, method != \'CG\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99919:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99929:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = 2 and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT.format(_iterm_py),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_weight_py),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_sigcmp_py),
                    ),
                    'Constraint: sigcmp = \'S\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_precon_py),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_method_py),
                    ),
                    'Constraint: method = \'CG\', \'SYMMLQ\' or \'MINRES\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_symm_basic_setup has been called out of sequence:',
                    'either real_symm_basic_setup has been called twice or',
                    'real_symm_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['lwreq', 'comm'],
        )(
            _lwreq_py,
            _comm_py,
        )
    _err_class.manage()
    return return_data

def real_symm_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Real sparse symmetric linear systems, preconditioned conjugate
    gradient or Lanczos method or the MINRES algorithm.

    ``real_symm_basic_solver`` is an iterative solver for a symmetric
    system of simultaneous linear equations; ``real_symm_basic_solver``
    is the second in a suite of three functions, where the first
    function, ``real_symm_basic_setup``, must be called prior to
    ``real_symm_basic_solver`` to set up the suite, and the third
    function in the suite, ``real_symm_basic_diag``, can be used to
    return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    symmetric systems of equations.

    For full information please refer to the NAG Library document for
    f11ge

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11gef.html

    Parameters
    ----------
    irevcm : int
        On initial entry: `irevcm` = 0, otherwise an error condition
        will be raised.

        On intermediate entry: must either be unchanged from its
        previous exit value, or can have one of the following values.

        `irevcm` = 5

          Tidy termination: the computation will terminate at the end of
          the current iteration. Further reverse communication exits may
          occur depending on when the termination request is issued.
          ``real_symm_basic_solver`` will then return with the
          termination code `irevcm` = 4. Note that before calling
          ``real_symm_basic_solver`` with `irevcm` = 5 the calling
          program must have performed the tasks required by the value of
          `irevcm` returned by the previous call to
          ``real_symm_basic_solver``, otherwise subsequently returned
          values may be invalid.

        `irevcm` = 6

          Immediate termination: ``real_symm_basic_solver`` will return
          immediately with termination code `irevcm` = 4 and with any
          useful information available. This includes the last iterate
          of the solution and, for conjugate gradient only, the last
          iterate of the residual vector. The residual vector is
          generally not available when the Lanczos method (SYMMLQ) is
          used. ``real_symm_basic_solver`` will then return with the
          termination code `irevcm` = 4.

          Immediate termination may be useful, for example, when errors
          are detected during matrix-vector multiplication or during the
          solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : float, shape (n), modified in place
        On initial entry: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        On intermediate entry: must remain unchanged.

        On intermediate exit: the returned value of `irevcm` determines
        the contents of `u` in the following way.

        If `irevcm` = 1 or 2, `u` holds the vector u on which the
        operation specified by `irevcm` is to be carried out.

        If `irevcm` = 3, `u` holds the current iterate of the solution
        vector.

        On final exit: if `errno` = 3 or -i, the array `u` is unchanged
        from the initial entry to ``real_symm_basic_solver``. If `errno`
        = 1, the array `u` is unchanged from the last entry to
        ``real_symm_basic_solver``. Otherwise, `u` holds the last
        iterate of the solution of the system of equations, for all
        returned values of errno.

    v : float, shape (n), modified in place
        On initial entry: the right-hand side b of the system of
        equations Ax = b.

        On intermediate entry: the returned value of `irevcm` determines
        the contents of `v` in the following way.

        If `irevcm` = 1 or 2, `v` must store the vector v, the result of
        the operation specified by the value of `irevcm` returned by the
        previous call to ``real_symm_basic_solver``.

        If `irevcm` = 3, `v` must remain unchanged.

        On intermediate exit: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        On final exit: if `errno` = 3 or -i, the array `v` is unchanged
        from the last entry to ``real_symm_basic_solver``. If `errno` =
        1, the array `v` is unchanged from the initial entry to
        ``real_symm_basic_solver``. If the function exits successfully
        or `errno` = 2, the array `v` contains the true residual vector
        of the system of equations (see also [Errors]). Otherwise, `v`
        stores the last iterate of the residual vector unless the
        Lanczos method (SYMMLQ) was used and `errno` >= 5, in which case
        `v` is set to 0.0.

    wgt : float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Description] and [Parameters]).

        Weights are NOT used in the MINRES algorithm.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_symm_basic_setup``.

    Returns
    -------
    irevcm : int
        On intermediate exit: has the following meanings.

        `irevcm` = 1

          The calling program must compute the matrix-vector product v =
          Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

          The calling program must solve the preconditioning equation Mv
          = u, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 3

          Monitoring step: the solution and residual at the current
          iteration are returned in the arrays `u` and `v`,
          respectively. No action by the calling program is required. To
          return additional information ``real_symm_basic_diag`` can be
          called at this step.

        On final exit: if `irevcm` = 4, ``real_symm_basic_solver`` has
        completed its tasks. The value of errno determines whether the
        iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``real_symm_basic_setup`` was not called before
            calling this function or it has returned an error.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 9)
            The matrix of the coefficients A appears to be singular. The
            computation cannot continue.

        (`errno` 10)
            The weights in array `wgt` are all zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``real_symm_basic_solver`` has already completed its tasks.
            You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested termination: the required accuracy could not
            be obtained. However, a reasonable accuracy may have been
            achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 7)
            The matrix of the coefficients A appears not to be positive
            definite. The computation cannot continue.

        (`errno` 8)
            User-requested immediate termination.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11geft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, u[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    if _lwork_py >= 20:
        if int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (3,), validate=False)) == 1739512 and int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (19,), validate=False)) == 840653:
            _vl_n_py = int(utils._EngineFloat64ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (9,), validate=False))
        else:
            _vl_n_py = 0
    else:
        _vl_n_py = 0
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_symm_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested termination: the required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either real_symm_basic_setup was not called before calling this function',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears to be singular.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def real_symm_basic_diag(comm):
    u"""
    Real sparse symmetric linear systems, diagnostic for
    ``real_symm_basic_solver``.

    ``real_symm_basic_diag`` is the third in a suite of three functions
    for the iterative solution of a symmetric system of simultaneous
    linear equations. ``real_symm_basic_diag`` returns information about
    the computations during an iteration and/or after this has been
    completed.
    The first function of the suite, ``real_symm_basic_setup``, is a
    setup function, the second function, ``real_symm_basic_solver`` is
    the proper iterative solver.

    These three functions are suitable for the solution of large sparse
    symmetric systems of equations.

    For full information please refer to the NAG Library document for
    f11gf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11gff.html

    Parameters
    ----------
    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_symm_basic_setup``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``real_symm_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``real_symm_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``real_symm_basic_solver``.

    anorm : float
        For CG and SYMMLQ methods, the norm ||A||_1 = ||A||_infinity
        when either it has been supplied to ``real_symm_basic_setup`` or
        it has been estimated by ``real_symm_basic_solver`` (see also
        [Description] and [Parameters]). Otherwise, `anorm` = 0.0 is
        returned.

        For MINRES method, an estimate of the infinity norm of the
        preconditioned matrix operator.

    sigmax : float
        For CG and SYMMLQ methods, the current estimate of the largest
        singular value sigma_1(A_) of the preconditioned iteration
        matrix A_= E^-1AE^invT, when either it has been supplied to
        ``real_symm_basic_setup`` or it has been estimated by
        ``real_symm_basic_solver`` (see also [Description] and
        [Parameters]). Note that if `its` < `itn` then `sigmax` contains
        the final estimate. If, on final exit from
        ``real_symm_basic_solver``, `its` = `itn`, the estimation of
        sigma_1(A_) may have not converged; in this case you should look
        at the value returned in `sigerr`. Otherwise, `sigmax` = 0.0 is
        returned.

        For MINRES method, an estimate of the final transformed
        residual.

    its : int
        For CG and SYMMLQ methods, the number of iterations employed so
        far in the computation of the estimate of sigma_1(A_), the
        largest singular value of the preconditioned matrix A_=
        E^-1AE^invTH, when sigma_1(A_) has been estimated by
        ``real_symm_basic_solver`` using the bisection method (see also
        [Description], [Parameters] and [Further Comments]). Otherwise,
        `its` = 0 is returned.

    sigerr : float
        For CG and SYMMLQ methods, if sigma_1(A_) has been estimated by
        ``real_symm_basic_solver`` using bisection,

            `sigerr` = max(|sigma_1^(k)-sigma_1^(k-1)|/sigma_1^(k),
            |sigma_1^(k)-sigma_1^(k-2)|/sigma_1^(k)),

        where k = `its` denotes the iteration number. The estimation has
        converged if `sigerr` <= sigtol where sigtol is an input
        argument to ``real_symm_basic_setup``. Otherwise, `sigerr` = 0.0
        is returned.

        For MINRES method, an estimate of the condition number of the
        preconditioned matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            ``real_symm_basic_diag`` has been called out of sequence.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._BYREF_INT, # Output, its.
        utils._BYREF_FLOAT64, # Output, sigerr.
        utils._EngineFloat64ArrayType.c_type, # Input, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    _its = utils._EngineIntScalarType.empty_ctypes()
    _sigerr = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _its,
        _sigerr,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'real_symm_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _its_py = utils._EngineIntScalarType.to_py(_its)
        _sigerr_py = utils._EngineFloat64ScalarType.to_py(_sigerr)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax', 'its', 'sigerr'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
            _its_py,
            _sigerr_py,
        )
    _err_class.manage()
    return return_data

def complex_herm_basic_setup(method, precon, sigcmp, norm, weight, iterm, n, tol, maxitn, anorm, sigmax, sigtol, maxits, monit, lwork):
    u"""
    Complex sparse Hermitian linear systems, setup for
    ``complex_herm_basic_solver``.

    ``complex_herm_basic_setup`` is a setup function, the first in a
    suite of three functions for the iterative solution of a complex
    Hermitian system of simultaneous linear equations.
    ``complex_herm_basic_setup`` must be called before
    ``complex_herm_basic_solver``, the iterative solver.
    The third function in the suite, ``complex_herm_basic_diag``, can be
    used to return additional information about the computation.

    These three functions are suitable for the solution of large sparse
    complex Hermitian systems of equations.

    For full information please refer to the NAG Library document for
    f11gr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11grf.html

    Parameters
    ----------
    method : str
        The iterative method to be used.

        `method` = 'CG'

          Conjugate gradient method.

        `method` = 'SYMMLQ'

          Lanczos method (SYMMLQ).

    precon : str, length 1
        Determines whether preconditioning is used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'P'

          Preconditioning.

    sigcmp : str, length 1
        Determines whether an estimate of sigma_1(A_) =
        ||E^-1AE^invTH||_2, the largest singular value of the
        preconditioned matrix of the coefficients, is to be computed
        using the bisection method on the sequence of tridiagonal
        matrices {T_k} generated during the iteration. Note that A_= A
        when a preconditioner is not used.

        If `sigmax` > 0.0 , i.e., when sigma_1(A_) is supplied, the
        value of `sigcmp` is ignored.

        `sigcmp` = 'S'

          sigma_1(A_) is to be computed using the bisection method.

        `sigcmp` = 'N'

          The bisection method is not used.

          If the termination criterion (3) is used, requiring
          sigma_1(A_), an inexpensive estimate is computed and used (see
          [Description]).

    norm : str, length 1
        Defines the matrix and vector norm to be used in the termination
        criteria.

        `norm` = '1'

          Use the l_1 norm.

        `norm` = 'I'

          Use the l_infinity norm.

        `norm` = '2'

          Use the l_2 norm.

    weight : str, length 1
        Specifies whether a vector w of user-supplied weights is to be
        used in the vector norms used in the computation of termination
        criterion (2) (`iterm` = 1): ||v||_p^(w) = ||v^(w)||_p, where
        v_i^(w) = w_i v_i, for i = 1,2,...,n. The suffix p = 1,2,
        infinity denotes the vector norm used, as specified by the
        argument `norm`. Note that weights cannot be used when `iterm` =
        2, i.e., when criterion (3) is used.

        `weight` = 'W'

          User-supplied weights are to be used and must be supplied on
          initial entry to ``complex_herm_basic_solver``.

        `weight` = 'N'

          All weights are implicitly set equal to one. Weights do not
          need to be supplied on initial entry to
          ``complex_herm_basic_solver``.

    iterm : int
        Defines the termination criterion to be used.

        `iterm` = 1

          Use the termination criterion defined in (2) (both conjugate
          gradient and Lanczos (SYMMLQ) methods).

        `iterm` = 2

          Use the termination criterion defined in (3) (Lanczos method
          (SYMMLQ) only).

    n : int
        n, the order of the matrix A.

    tol : float
        The tolerance tau for the termination criterion.

        If `tol` <= 0.0, tau = max(sqrt(epsilon), sqrt(n)epsilon) is
        used, where epsilon is the machine precision.

        Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon) is used.

    maxitn : int
        The maximum number of iterations.

    anorm : float
        If `anorm` > 0.0, the value of ||A||_p to be used in the
        termination criterion (2) (`iterm` = 1).

        If `anorm` <= 0.0, `iterm` = 1 and `norm` = '1' or 'I', ||A||_1
        = ||A||_infinity is estimated internally by
        ``complex_herm_basic_solver``.

        If `iterm` = 2, `anorm` is not referenced.

    sigmax : float
        If `sigmax` > 0.0, the value of sigma_1(A_) =
        ||E^-1AE^invTH||_2.

        If `sigmax` <= 0.0, sigma_1(A_) is estimated by
        ``complex_herm_basic_solver`` when either `sigcmp` = 'S' or
        termination criterion (3) (`iterm` = 2) is employed, though it
        will be used only in the latter case.

        Otherwise, `sigmax` is not referenced.

    sigtol : float
        The tolerance used in assessing the convergence of the estimate
        of sigma_1(A_) = ||A_||_2 when the bisection method is used.

        If `sigtol` <= 0.0, the default value `sigtol` = 0.01 is used.

        The actual value used is max(`sigtol`, epsilon).

        If `sigcmp` = 'N' or `sigmax` > 0.0, `sigtol` is not referenced.

    maxits : int
        The maximum iteration number k = `maxits` for which sigma_1(T_k)
        is computed by bisection (see also [Description]). If `sigcmp` =
        'N' or `sigmax` > 0.0, `maxits` is not referenced.

    monit : int
        If `monit` > 0, the frequency at which a monitoring step is
        executed by ``complex_herm_basic_solver``: the current solution
        and residual iterates will be returned by
        ``complex_herm_basic_solver`` and a call to
        ``complex_herm_basic_diag`` made possible every `monit`
        iterations, starting from iteration number `monit`. Otherwise,
        no monitoring takes place. There are some additional
        computational costs involved in monitoring the solution and
        residual vectors when the Lanczos method (SYMMLQ) is used.

    lwork : int
        The dimension of the array `comm`['work'].

    Returns
    -------
    lwreq : int
        The minimum amount of workspace required by
        ``complex_herm_basic_solver``. (See also [Parameters].)

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -14)
            On entry, `monit` = *<value>* and `maxitn` = *<value>*.

            Constraint: `monit` <= `maxitn`.

        (`errno` -13)
            On entry, `maxits` = *<value>*, `maxitn` = *<value>*,
            `sigcmp` = *<value>*, and `sigmax` = *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, 1 <=
            `maxits` <= `maxitn`.

        (`errno` -12)
            On entry, `sigcmp` = *<value>*, `sigtol` = *<value>*, and
            `sigmax` = *<value>*.

            Constraint: if `sigcmp` = 'S' and `sigmax` <= 0.0, `sigtol`
            < 1.0.

        (`errno` -10)
            On entry, `norm` = *<value>*, `iterm` = *<value>*, and
            `anorm` = *<value>*.

            Constraint: if `iterm` = 1 and `norm` = '2' or 'I', `anorm`
            > 0.0.

        (`errno` -9)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` > 0.

        (`errno` -8)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` -7)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` -6)
            On entry, `iterm` = *<value>*.

            Constraint: `iterm` = 1 or 2.

        (`errno` -6)
            On entry, `iterm` = *<value>* and `method` = *<value>*.

            Constraint: if `iterm` = 2, `method` = 'SYMMLQ'.

        (`errno` -6)
            On entry, `iterm` = *<value>* and `norm` = *<value>*.

            Constraint: if `iterm` = 2, `norm` = '2' or 'I'.

        (`errno` -5)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'N' or 'W'.

        (`errno` -5)
            On entry, `iterm` = *<value>* and `weight` = *<value>*.

            Constraint: if `iterm` = 2, `weight` = 'N'.

        (`errno` -4)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'I' or '2'.

        (`errno` -3)
            On entry, `sigcmp` = *<value>*.

            Constraint: `sigcmp` = 'N' or 'S'.

        (`errno` -2)
            On entry, `precon` = *<value>*.

            Constraint: `precon` = 'N' or 'P'.

        (`errno` -1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            ``complex_herm_basic_setup`` has been called out of
            sequence: either ``complex_herm_basic_setup`` has been
            called twice or ``complex_herm_basic_solver`` has not
            terminated its current task.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_basic_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11grft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_STRING, # Input, sigcmp(len=1).
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, iterm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Input, sigmax.
        utils._BYREF_FLOAT64, # Input, sigtol.
        utils._BYREF_INT, # Input, maxits.
        utils._BYREF_INT, # Input, monit.
        utils._BYREF_INT, # Output, lwreq.
        utils._EngineComplex128ArrayType.c_type, # Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of sigcmp.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _sigcmp = utils._EngineCharScalarType(
        sigcmp, {'fun_name': fun_name, 'entity': 'sigcmp'}, exp_chrlen=1,
    ).to_ctypes()
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _iterm = utils._EngineIntScalarType(
        iterm, {'fun_name': fun_name, 'entity': 'iterm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _sigmax = utils._EngineFloat64ScalarType(
        sigmax, {'fun_name': fun_name, 'entity': 'sigmax'},
    ).to_ctypes()
    _sigtol = utils._EngineFloat64ScalarType(
        sigtol, {'fun_name': fun_name, 'entity': 'sigtol'},
    ).to_ctypes()
    _maxits = utils._EngineIntScalarType(
        maxits, {'fun_name': fun_name, 'entity': 'maxits'},
    ).to_ctypes()
    _monit = utils._EngineIntScalarType(
        monit, {'fun_name': fun_name, 'entity': 'monit'},
    ).to_ctypes()
    _lwreq = utils._EngineIntScalarType.empty_ctypes()
    _lwork = utils._EngineIntScalarType(
        lwork, {'fun_name': fun_name, 'entity': 'lwork'},
    ).to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(120)
    _comm_py = {}
    _comm_py['work'] = _work
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _sigcmp,
        _norm,
        _weight,
        _iterm,
        _n,
        _tol,
        _maxitn,
        _anorm,
        _sigmax,
        _sigtol,
        _maxits,
        _monit,
        _lwreq,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _sigcmp._length_,
        _norm._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -14 and _ifmt == -14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monit = {:s} and maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: monit <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -13 and _ifmt == -13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxits = {:s}, maxitn = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'sigcmp = \'{:s}\', and sigmax = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0, 1 <= maxits <= maxitn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -12 and _ifmt == -12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'{:s}\', sigtol = {:s},'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and sigmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if sigcmp = \'S\' and sigmax <= 0.0, sigtol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -10 and _ifmt == -10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\', iterm = {:s},'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if iterm = 1 and norm = \'2\' or \'I\', anorm > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -9 and _ifmt == -9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -8 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -7 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == -6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iterm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s} and method = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if iterm = 2, method = \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -6 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s} and norm = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if iterm = 2, norm = \'2\' or \'I\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == -5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'N\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -5 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iterm = {:s} and weight = \'{:s}\'.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if iterm = 2, weight = \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -4 and _ifmt == -4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: norm = \'1\', \'I\' or \'2\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -3 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigcmp = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigcmp = \'N\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -2 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: precon = \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_herm_basic_setup has been called out of sequence:',
                    'either complex_herm_basic_setup has been called twice or',
                    'complex_herm_basic_solver has not terminated its current task.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lwreq_py = utils._EngineIntScalarType.to_py(_lwreq)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['lwreq', 'comm'],
        )(
            _lwreq_py,
            _comm_py,
        )
    _err_class.manage()
    return return_data

def complex_herm_basic_solver(irevcm, u, v, wgt, comm):
    u"""
    Complex sparse Hermitian linear systems, preconditioned conjugate
    gradient or Lanczos.

    ``complex_herm_basic_solver`` is an iterative solver for a complex
    Hermitian system of simultaneous linear equations;
    ``complex_herm_basic_solver`` is the second in a suite of three
    functions, where the first function, ``complex_herm_basic_setup``,
    must be called prior to ``complex_herm_basic_solver`` to set up the
    suite, and the third function in the suite,
    ``complex_herm_basic_diag``, can be used to return additional
    information about the computation.

    These three functions are suitable for the solution of large sparse
    complex Hermitian systems of equations.

    For full information please refer to the NAG Library document for
    f11gs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11gsf.html

    Parameters
    ----------
    irevcm : int
        On initial entry: `irevcm` = 0, otherwise an error condition
        will be raised.

        On intermediate entry: `irevcm` must either be unchanged from
        its previous exit value, or can have one of the following
        values.

        `irevcm` = 5

          Tidy termination: the computation will terminate at the end of
          the current iteration. Further reverse communication exits may
          occur depending on when the termination request is issued.
          ``complex_herm_basic_solver`` will then return with the
          termination code `irevcm` = 4. Note that before calling
          ``complex_herm_basic_solver`` with `irevcm` = 5 the calling
          program must have performed the tasks required by the value of
          `irevcm` returned by the previous call to
          ``complex_herm_basic_solver``, otherwise subsequently returned
          values may be invalid.

        `irevcm` = 6

          Immediate termination: ``complex_herm_basic_solver`` will
          return immediately with termination code `irevcm` = 4 and with
          any useful information available. This includes the last
          iterate of the solution and, for conjugate gradient only, the
          last iterate of the residual vector. The residual vector is
          generally not available when the Lanczos method (SYMMLQ) is
          used. ``complex_herm_basic_solver`` will then return with the
          termination code `irevcm` = 4.

          Immediate termination may be useful, for example, when errors
          are detected during matrix-vector multiplication or during the
          solution of the preconditioning equation.

        Changing `irevcm` to any other value between calls will result
        in an error.

    u : complex, shape (n), modified in place
        On initial entry: an initial estimate, x_0, of the solution of
        the system of equations Ax = b.

        On intermediate entry: must remain unchanged.

        On intermediate exit: the returned value of `irevcm` determines
        the contents of `u` in the following way.

        If `irevcm` = 1 or 2, `u` holds the vector u on which the
        operation specified by `irevcm` is to be carried out.

        If `irevcm` = 3, `u` holds the current iterate of the solution
        vector.

        On final exit: if `errno` = 3 or -i, the array `u` is unchanged
        from the initial entry to ``complex_herm_basic_solver``. If
        `errno` = 1, the array `u` is unchanged from the last entry to
        ``complex_herm_basic_solver``. Otherwise, `u` holds the last
        iterate of the solution of the system of equations, for all
        returned values of errno.

    v : complex, shape (n), modified in place
        On initial entry: the right-hand side b of the system of
        equations Ax = b.

        On intermediate entry: the returned value of `irevcm` determines
        the contents of `v` in the following way.

        If `irevcm` = 1 or 2, `v` must store the vector v, the result of
        the operation specified by the value of `irevcm` returned by the
        previous call to ``complex_herm_basic_solver``

        If `irevcm` = 3, `v` must remain unchanged.

        On intermediate exit: if `irevcm` = 3, `v` holds the current
        iterate of the residual vector. Note that this is an
        approximation to the true residual vector. Otherwise, it does
        not contain any useful information.

        On final exit: if `errno` = 3 or 0, the array `v` is unchanged
        from the last entry to ``complex_herm_basic_solver``. If `errno`
        = 1, the array `v` is unchanged from the initial entry to
        ``complex_herm_basic_solver``. If the function exits
        successfully or `errno` = 2, the array `v` contains the true
        residual vector of the system of equations (see also [Errors]).
        Otherwise, `v` stores the last iterate of the residual vector
        unless the Lanczos method (SYMMLQ) was used and `errno` >= 5, in
        which case `v` is set to 0.0.

    wgt : float, shape (n)
        The user-supplied weights, if these are to be used in the
        computation of the vector norms in the termination criterion
        (see [Description] and [Parameters]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_herm_basic_setup``.

    Returns
    -------
    irevcm : int
        On intermediate exit: has the following meanings.

        `irevcm` = 1

          The calling program must compute the matrix-vector product v =
          Au, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 2

          The calling program must solve the preconditioning equation Mv
          = u, where u and v are stored in `u` and `v`, respectively.

        `irevcm` = 3

          Monitoring step: the solution and residual at the current
          iteration are returned in the arrays `u` and `v`,
          respectively. No action by the calling program is required. To
          return additional information ``complex_herm_basic_diag`` can
          be called at this step.

        On final exit: if `irevcm` = 4, ``complex_herm_basic_solver``
        has completed its tasks. The value of errno determines whether
        the iteration has been successfully completed, errors have been
        detected or the calling program has requested termination.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: either `irevcm` must be unchanged from its
            previous exit value or `irevcm` = 5 or 6.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 3)
            Either ``complex_herm_basic_setup`` was not called before
            calling ``complex_herm_basic_solver`` or it has returned an
            error.

        (`errno` 9)
            The matrix of the coefficients A appears to be singular. The
            computation cannot continue.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            ``complex_herm_basic_solver`` has already completed its
            tasks. You need to set a new problem.

        (`errno` 2)
            The required accuracy could not be obtained. However, a
            reasonable accuracy may have been achieved.

        (`errno` 2)
            User-requested termination: the required accuracy could not
            be obtained. However, a reasonable accuracy may have been
            achieved.

        (`errno` 4)
            User-requested tidy termination. The solution has not
            converged after *<value>* iterations.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients A appears not to be positive
            definite. The computation cannot continue.

        (`errno` 8)
            User-requested immediate termination.

        (`errno` 10)
            The weights in array `wgt` are all zero.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_basic_solver'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, u[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, v[*].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[*].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    if _lwork_py >= 20:
        if int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (4,), validate=False).real) == 1739512 and int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (19,), validate=False).real) == 840653:
            _vl_n_py = int(utils._EngineComplex128ArrayType.get_element(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, (10,), validate=False).real)
        else:
            _vl_n_py = 0
    else:
        _vl_n_py = 0
    _u_cls = utils._EngineComplex128ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_vl_n_py,),
    )
    _u = _u_cls.to_ctypes()
    _v_cls = utils._EngineComplex128ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_vl_n_py,),
    )
    _v = _v_cls.to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_vl_n_py,),
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _u,
        _v,
        _wgt,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 4 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 99909:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On intermediate re-entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: either irevcm must be unchanged from its previous exit value or',
                    'irevcm = 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == -1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On initial entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_herm_basic_solver has already completed its tasks. You need to set a new problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested termination: the required accuracy could not be obtained.',
                    'However, a reasonable accuracy may have been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either complex_herm_basic_setup was not called before calling complex_herm_basic_solver',
                    'or it has returned an error.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested tidy termination.',
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User-requested immediate termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients A appears to be singular.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weights in array wgt are all zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _u_cls.refresh_py(_u)
        _v_cls.refresh_py(_v)
        return_data = _irevcm_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['irevcm'],
            )(
                _irevcm_py,
            )
    _err_class.manage()
    return return_data

def complex_herm_basic_diag(comm):
    u"""
    Complex sparse Hermitian linear systems, diagnostic for
    ``complex_herm_basic_solver``.

    ``complex_herm_basic_diag`` is the third in a suite of three
    functions for the iterative solution of a complex Hermitian system
    of simultaneous linear equations. ``complex_herm_basic_diag``
    returns information about the computations during an iteration
    and/or after this has been completed.
    The first function of the suite, ``complex_herm_basic_setup``, is a
    setup function, the second function, ``complex_herm_basic_solver``
    is the proper iterative solver.

    These three functions are suitable for the solution of large sparse
    complex Hermitian systems of equations.

    For full information please refer to the NAG Library document for
    f11gt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11gtf.html

    Parameters
    ----------
    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``complex_herm_basic_solver``.

    Returns
    -------
    itn : int
        The number of iterations carried out by
        ``complex_herm_basic_solver``.

    stplhs : float
        The current value of the left-hand side of the termination
        criterion used by ``complex_herm_basic_solver``.

    stprhs : float
        The current value of the right-hand side of the termination
        criterion used by ``complex_herm_basic_solver``.

    anorm : float
        The norm ||A||_1 = ||A||_infinity when either it has been
        supplied to ``complex_herm_basic_setup`` or it has been
        estimated by ``complex_herm_basic_solver`` (see also
        [Description] and [Parameters]).

        Otherwise, `anorm` = 0.0 is returned.

    sigmax : float
        The current estimate of the largest singular value sigma_1(A_)
        of the preconditioned iteration matrix A_= E^-1AE^(-H), when
        either it has been supplied to ``complex_herm_basic_setup`` or
        it has been estimated by ``complex_herm_basic_solver`` (see also
        [Description] and [Parameters]). Note that if `its` < `itn` then
        `sigmax` contains the final estimate. If, on final exit from
        ``complex_herm_basic_solver``, `its` = `itn`, the estimation of
        sigma_1(A_) may have not converged: in this case you should look
        at the value returned in `sigerr`. Otherwise, `sigmax` = 0.0 is
        returned.

    its : int
        The number of iterations employed so far in the computation of
        the estimate of sigma_1(A_), the largest singular value of the
        preconditioned matrix A_= E^-1AE^(-H), when sigma_1(A_) has been
        estimated by ``complex_herm_basic_solver`` using the bisection
        method (see also [Description], [Parameters] and [Further
        Comments]). Otherwise, `its` = 0 is returned.

    sigerr : float
        If sigma_1(A_) has been estimated by
        ``complex_herm_basic_solver`` using bisection,

            `sigerr` = max(|sigma_1^(k)-sigma_1^(k-1)|/sigma_1^(k),
            |sigma_1^(k)-sigma_1^(k-2)|/sigma_1^(k)),

        where k = `its` denotes the iteration number. The estimation has
        converged if `sigerr` <= sigtol where sigtol is an input
        argument to ``complex_herm_basic_setup``.

        Otherwise, `sigerr` = 0.0 is returned.

    Raises
    ------
    NagValueError
        (`errno` 1)
            ``complex_herm_basic_diag`` has been called out of sequence.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_basic_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11gtft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, itn.
        utils._BYREF_FLOAT64, # Output, stplhs.
        utils._BYREF_FLOAT64, # Output, stprhs.
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_FLOAT64, # Output, sigmax.
        utils._BYREF_INT, # Output, its.
        utils._BYREF_FLOAT64, # Output, sigerr.
        utils._EngineComplex128ArrayType.c_type, # Input, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _stplhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _stprhs = utils._EngineFloat64ScalarType.empty_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _sigmax = utils._EngineFloat64ScalarType.empty_ctypes()
    _its = utils._EngineIntScalarType.empty_ctypes()
    _sigerr = utils._EngineFloat64ScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('work',),
    )
    _lwork_py = utils._EngineComplex128ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _work = comm['work']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itn,
        _stplhs,
        _stprhs,
        _anorm,
        _sigmax,
        _its,
        _sigerr,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'complex_herm_basic_diag has been called out of sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        _stplhs_py = utils._EngineFloat64ScalarType.to_py(_stplhs)
        _stprhs_py = utils._EngineFloat64ScalarType.to_py(_stprhs)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _sigmax_py = utils._EngineFloat64ScalarType.to_py(_sigmax)
        _its_py = utils._EngineIntScalarType.to_py(_its)
        _sigerr_py = utils._EngineFloat64ScalarType.to_py(_sigerr)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itn', 'stplhs', 'stprhs', 'anorm', 'sigmax', 'its', 'sigerr'],
        )(
            _itn_py,
            _stplhs_py,
            _stprhs_py,
            _anorm_py,
            _sigmax_py,
            _its_py,
            _sigerr_py,
        )
    _err_class.manage()
    return return_data

def real_symm_precon_ichol(n, nnz, a, la, irow, icol, lfill, dtol, mic, dscale, pstrat, ipiv, istr):
    u"""
    Real sparse symmetric matrix, incomplete Cholesky factorization.

    ``real_symm_precon_ichol`` computes an incomplete Cholesky
    factorization of a real sparse symmetric matrix, represented in
    symmetric coordinate storage format.
    This factorization may be used as a preconditioner in combination
    with ``real_symm_solve_ichol`` or ``real_symm_basic_solver``.

    For full information please refer to the NAG Library document for
    f11ja

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jaf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : float, shape (`la`), modified in place
        On entry: the nonzero elements in the lower triangular part of
        the matrix A, ordered by increasing row index, and by increasing
        column index within each row. Multiple entries for the same row
        and column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

        On exit: the first `nnz` elements of `a` contain the nonzero
        elements of A and the next `nnzc` elements contain the elements
        of the lower triangular matrix C. Matrix elements are ordered by
        increasing row index, and by increasing column index within each
        row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        On entry: the row indices of the nonzero elements supplied in
        `a`.

        On exit: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        On entry: the column indices of the nonzero elements supplied in
        `a`.

        On exit: the column indices of the nonzero elements returned in
        `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    mic : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row sums (see [Further Comments]).

        `mic` = 'M'

          The factorization is modified.

        `mic` = 'N'

          The factorization is not modified.

    dscale : float
        The diagonal scaling parameter. All diagonal elements are
        multiplied by the factor (1+`dscale`) at the start of the
        factorization. This can be used to ensure that the
        preconditioner is positive definite. See [Further Comments].

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

          No pivoting is carried out.

        `pstrat` = 'M'

          Diagonal pivoting aimed at minimizing fill-in is carried out,
          using the Markowitz strategy.

        `pstrat` = 'U'

          Diagonal pivoting is carried out according to the user-defined
          input value of `ipiv`.

    ipiv : int, shape (`n`), modified in place
        On entry: if `pstrat` = 'U', `ipiv`[i-1] must specify the row
        index of the diagonal element used as a pivot at elimination
        stage i. Otherwise `ipiv` need not be initialized.

        On exit: the pivot indices. If `ipiv`[i-1] = j then the diagonal
        element in row j was used as the pivot at elimination stage i.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the lower triangular matrix C.

    npivm : int
        The number of pivots which were modified during the
        factorization to ensure that M was positive definite. The
        quality of the preconditioner will generally depend on the
        returned value of `npivm`. If `npivm` is large the
        preconditioner may not be satisfactory. In this case it may be
        advantageous to call ``real_symm_precon_ichol`` again with an
        increased value of either `lfill` or `dscale`. See also [Further
        Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `pstrat` != 'N', 'U' or 'M': `pstrat` = *<value>*.

        (`errno` 1)
            On entry, `mic` != 'M' or 'N': `mic` = *<value>*.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            setting `pstrat` = 'M', reducing `lfill`, or increasing
            `dtol`.

        (`errno` 5)
            A serious error has occurred in an internal call. Check all
            function calls and array sizes. Seek expert help.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_precon_ichol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, mic(len=1).
        utils._BYREF_FLOAT64, # Input, dscale.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mic.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _mic = utils._EngineCharScalarType(
        mic, {'fun_name': fun_name, 'entity': 'mic'}, exp_chrlen=1,
    ).to_ctypes()
    _dscale = utils._EngineFloat64ScalarType(
        dscale, {'fun_name': fun_name, 'entity': 'dscale'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _mic,
        _dscale,
        _pstrat,
        _ipiv,
        _istr,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _mic._length_,
        _pstrat._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat != \'N\', \'U\' or \'M\': pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mic != \'M\' or \'N\': mic = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by setting pstrat = \'M\',',
                    'reducing lfill, or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call.',
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipiv_cls.refresh_py(_ipiv)
        _istr_cls.refresh_py(_istr)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def real_symm_precon_ichol_solve(n, a, la, irow, icol, ipiv, istr, check, y, x):
    u"""
    Solution of linear system involving incomplete Cholesky
    preconditioning matrix generated by ``real_symm_precon_ichol``.

    ``real_symm_precon_ichol_solve`` solves a system of linear equations
    involving the incomplete Cholesky preconditioning matrix generated
    by ``real_symm_precon_ichol``.

    For full information please refer to the NAG Library document for
    f11jb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jbf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix M. This must be the same value as was
        supplied in the preceding call to ``real_symm_precon_ichol``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_symm_precon_ichol``.

    la : int
        This must be the same value returned by the preceding call to
        ``real_symm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `irow`, `icol`,
          `ipiv` and `istr`.

        `check` = 'N'

          No checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipiv`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipiv`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipiv`[i-1] >= 1 and `ipiv`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_precon_ichol_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipiv[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipiv[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipiv[i-1] >= 1 and ipiv[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[i-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_solve_ichol(method, n, nnz, a, la, irow, icol, ipiv, istr, b, tol, maxitn, x):
    u"""
    Solution of real sparse symmetric linear system, conjugate
    gradient/Lanczos method, preconditioner computed by
    ``real_symm_precon_ichol`` (Black Box).

    ``real_symm_solve_ichol`` solves a real sparse symmetric system of
    linear equations, represented in symmetric coordinate storage
    format, using a conjugate gradient or Lanczos method, with
    incomplete Cholesky preconditioning.

    For full information please refer to the NAG Library document for
    f11jc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jcf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

          Conjugate gradient method.

        `method` = 'SYMMLQ'

          Lanczos method (SYMMLQ).

    n : int
        n, the order of the matrix A. This must be the same value as was
        supplied in the preceding call to ``real_symm_precon_ichol``.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A. This must be the same value as was supplied in the
        preceding call to ``real_symm_precon_ichol``.

    a : float, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``real_symm_precon_ichol``.

    la : int
        This must be the same value as was supplied in the preceding
        call to ``real_symm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``real_symm_precon_ichol``.

    b : float, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*,
            `irow`[i-1] = *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `irow`[i-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The SCS representation of the preconditioner is invalid.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error has occurred in an internal call: IREVCM =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

        (`errno` 8)
            A serious error has occurred in an internal call: errno =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy has been achieved.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_solve_ichol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[i-1] = {:s}, irow[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The SCS representation of the preconditioner is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However a reasonable accuracy has been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'IREVCM = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'errno = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def real_symm_precon_ssor_solve(n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to real sparse symmetric matrix.

    ``real_symm_precon_ssor_solve`` solves a system of linear equations
    involving the preconditioning matrix corresponding to SSOR applied
    to a real sparse symmetric matrix, represented in symmetric
    coordinate storage format.

    For full information please refer to the NAG Library document for
    f11jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jdf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        A.

    a : float, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    rdiag : float, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `nnz`, `irow`,
          `icol` and `omega`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    y : float, shape (`n`)
        The right-hand side vector y.

    x : float, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` != 'C' or 'N': `check` = *<value>*.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineFloat64ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check != \'C\' or \'N\': check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, tol, maxitn, x):
    u"""
    Solution of real sparse symmetric linear system, conjugate
    gradient/Lanczos method, Jacobi or SSOR preconditioner (Black Box).

    ``real_symm_solve_jacssor`` solves a real sparse symmetric system of
    linear equations, represented in symmetric coordinate storage
    format, using a conjugate gradient or Lanczos method, without
    preconditioning, with Jacobi or with SSOR preconditioning.

    For full information please refer to the NAG Library document for
    f11je

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jef.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

          Conjugate gradient method.

        `method` = 'SYMMLQ'

          Lanczos method (SYMMLQ).

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'J'

          Jacobi.

        `precon` = 'S'

          Symmetric successive-over-relaxation (SSOR).

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements of the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized.

    b : float, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : float, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: `omega` > 0.0 and `omega` < 2.0.

        (`errno` 1)
            On entry, `precon` != 'N', 'J' or 'S': `precon` = *<value>*.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy has been achieved.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error has occurred in an internal call: IREVCM =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

        (`errno` 8)
            A serious error has occurred in an internal call: errno =
            *<value>*. Check all function calls and array sizes. Seek
            expert help.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineFloat64ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: omega > 0.0 and omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon != \'N\', \'J\' or \'S\': precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However a reasonable accuracy has been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'IREVCM = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call:',
                    'errno = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def complex_herm_precon_ichol(n, nnz, a, la, irow, icol, lfill, dtol, mic, dscale, pstrat, ipiv, istr):
    u"""
    Complex sparse Hermitian matrix, incomplete Cholesky factorization.

    ``complex_herm_precon_ichol`` computes an incomplete Cholesky
    factorization of a complex sparse Hermitian matrix, represented in
    symmetric coordinate storage format.
    This factorization may be used as a preconditioner in combination
    with ``complex_herm_solve_ilu``.

    For full information please refer to the NAG Library document for
    f11jn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jnf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`la`), modified in place
        On entry: the nonzero elements in the lower triangular part of
        the matrix A, ordered by increasing row index, and by increasing
        column index within each row. Multiple entries for the same row
        and column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

        On exit: the first `nnz` elements of `a` contain the nonzero
        elements of A and the next `nnzc` elements contain the elements
        of the lower triangular matrix C. Matrix elements are ordered by
        increasing row index, and by increasing column index within each
        row.

    la : int
        These arrays must be of sufficient size to store both A (`nnz`
        elements) and C (`nnzc` elements).

    irow : int, shape (`la`), modified in place
        On entry: the row indices of the nonzero elements supplied in
        `a`.

        On exit: the row indices of the nonzero elements returned in
        `a`.

    icol : int, shape (`la`), modified in place
        On entry: the column indices of the nonzero elements supplied in
        `a`.

        On exit: the column indices of the nonzero elements returned in
        `a`.

    lfill : int
        If `lfill` >= 0 its value is the maximum level of fill allowed
        in the decomposition (see [Further Comments]). A negative value
        of `lfill` indicates that `dtol` will be used to control the
        fill instead.

    dtol : float
        If `lfill` < 0, `dtol` is used as a drop tolerance to control
        the fill-in (see [Further Comments]); otherwise `dtol` is not
        referenced.

    mic : str, length 1
        Indicates whether or not the factorization should be modified to
        preserve row sums (see [Further Comments]).

        `mic` = 'M'

          The factorization is modified.

        `mic` = 'N'

          The factorization is not modified.

    dscale : float
        The diagonal scaling parameter. All diagonal elements are
        multiplied by the factor (1.0+`dscale`) at the start of the
        factorization. This can be used to ensure that the
        preconditioner is positive definite. See also [Further
        Comments].

    pstrat : str, length 1
        Specifies the pivoting strategy to be adopted.

        `pstrat` = 'N'

          No pivoting is carried out.

        `pstrat` = 'M'

          Diagonal pivoting aimed at minimizing fill-in is carried out,
          using the Markowitz strategy.

        `pstrat` = 'U'

          Diagonal pivoting is carried out according to the user-defined
          input array `ipiv`.

    ipiv : int, shape (`n`), modified in place
        On entry: if `pstrat` = 'U', `ipiv`[i-1] must specify the row
        index of the diagonal element to be used as a pivot at
        elimination stage i. Otherwise `ipiv` need not be initialized.

        On exit: the pivot indices. If `ipiv`[i-1] = j, the diagonal
        element in row j was used as the pivot at elimination stage i.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix C. `istr`[`n`]-1 is the address of the last nonzero
        element in C plus one.

    Returns
    -------
    nnzc : int
        The number of nonzero elements in the lower triangular matrix C.

    npivm : int
        The number of pivots which were modified during the
        factorization to ensure that M was positive definite. The
        quality of the preconditioner will generally depend on the
        returned value of `npivm`. If `npivm` is large the
        preconditioner may not be satisfactory. In this case it may be
        advantageous to call ``complex_herm_precon_ichol`` again with an
        increased value of either `lfill` or `dscale`. See also [Further
        Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `pstrat` != 'N', 'U' or 'M': `pstrat` = *<value>*.

        (`errno` 1)
            On entry, `mic` != 'M' or 'N': `mic` = *<value>*.

        (`errno` 1)
            On entry, `dtol` = *<value>*.

            Constraint: `dtol` >= 0.0

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[I-1] >= 1 and `irow`[I-1] <= `n`.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` is repeated.

        (`errno` 3)
            On entry, a user-supplied value of `ipiv` lies outside the
            range [1,`n`].

        (`errno` 4)
            The number of nonzero entries in the decomposition is too
            large. The decomposition has been terminated before
            completion. Either increase `la`, or reduce the fill by
            setting `pstrat` = 'M', reducing `lfill`, or increasing
            `dtol`.

        (`errno` 5)
            A serious error has occurred in an internal call. Check all
            function calls and array sizes. Seek expert help.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_precon_ichol'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input/Output, irow[la].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[la].
        utils._BYREF_INT, # Input, lfill.
        utils._BYREF_FLOAT64, # Input, dtol.
        utils._BYREF_STRING, # Input, mic(len=1).
        utils._BYREF_FLOAT64, # Input, dscale.
        utils._BYREF_STRING, # Input, pstrat(len=1).
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._BYREF_INT, # Output, nnzc.
        utils._BYREF_INT, # Output, npivm.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mic.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pstrat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _lfill = utils._EngineIntScalarType(
        lfill, {'fun_name': fun_name, 'entity': 'lfill'},
    ).to_ctypes()
    _dtol = utils._EngineFloat64ScalarType(
        dtol, {'fun_name': fun_name, 'entity': 'dtol'},
    ).to_ctypes()
    _mic = utils._EngineCharScalarType(
        mic, {'fun_name': fun_name, 'entity': 'mic'}, exp_chrlen=1,
    ).to_ctypes()
    _dscale = utils._EngineFloat64ScalarType(
        dscale, {'fun_name': fun_name, 'entity': 'dscale'},
    ).to_ctypes()
    _pstrat = utils._EngineCharScalarType(
        pstrat, {'fun_name': fun_name, 'entity': 'pstrat'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _nnzc = utils._EngineIntScalarType.empty_ctypes()
    _npivm = utils._EngineIntScalarType.empty_ctypes()
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _lfill,
        _dtol,
        _mic,
        _dscale,
        _pstrat,
        _ipiv,
        _istr,
        _nnzc,
        _npivm,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _mic._length_,
        _pstrat._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pstrat != \'N\', \'U\' or \'M\': pstrat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mic != \'M\' or \'N\': mic = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dtol >= 0.0',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[I-1] >= 1 and irow[I-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv is repeated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a user-supplied value of ipiv lies outside the range [1,n].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of nonzero entries in the decomposition is too large.',
                    'The decomposition has been terminated before completion.',
                    'Either increase la, or reduce the fill by setting pstrat = \'M\',',
                    'reducing lfill, or increasing dtol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error has occurred in an internal call.',
                    'Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _ipiv_cls.refresh_py(_ipiv)
        _istr_cls.refresh_py(_istr)
        _nnzc_py = utils._EngineIntScalarType.to_py(_nnzc)
        _npivm_py = utils._EngineIntScalarType.to_py(_npivm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nnzc', 'npivm'],
        )(
            _nnzc_py,
            _npivm_py,
        )
    _err_class.manage()
    return return_data

def complex_herm_precon_ilu_solve(n, a, la, irow, icol, ipiv, istr, check, y, x):
    u"""
    Solution of complex linear system involving incomplete Cholesky
    preconditioning matrix generated by ``complex_herm_precon_ichol``.

    ``complex_herm_precon_ilu_solve`` solves a system of complex linear
    equations involving the incomplete Cholesky preconditioning matrix
    generated by ``complex_herm_precon_ichol``.

    For full information please refer to the NAG Library document for
    f11jp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jpf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix M. This must be the same value as was
        supplied in the preceding call to ``complex_herm_precon_ichol``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_herm_precon_ichol``.

    la : int
        This must be the same value supplied in the preceding call to
        ``complex_herm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `irow`, `icol`,
          `ipiv` and `istr`.

        `check` = 'N'

          None of these checks are carried out.

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ipiv`[i-1] is a repeated value: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `ipiv`[i-1] = *<value>*, `n` =
            *<value>*.

            Constraint: `ipiv`[i-1] >= 1 and `ipiv`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr`[i-1] is inconsistent with `irow`: i =
            *<value>*.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_precon_ilu_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _check,
        _y,
        _x,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ipiv[i-1] is a repeated value:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ipiv[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ipiv[i-1] >= 1 and ipiv[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr[i-1] is inconsistent with irow:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_herm_solve_ilu(method, n, nnz, a, la, irow, icol, ipiv, istr, b, tol, maxitn, x):
    u"""
    Solution of complex sparse Hermitian linear system, conjugate
    gradient/Lanczos method, preconditioner computed by
    ``complex_herm_precon_ichol`` (Black Box).

    ``complex_herm_solve_ilu`` solves a complex sparse Hermitian system
    of linear equations, represented in symmetric coordinate storage
    format, using a conjugate gradient or Lanczos method, with
    incomplete Cholesky preconditioning.

    For full information please refer to the NAG Library document for
    f11jq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jqf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

          Conjugate gradient method.

        `method` = 'SYMMLQ'

          Lanczos method (SYMMLQ).

    n : int
        n, the order of the matrix A. This must be the same value as was
        supplied in the preceding call to ``complex_herm_precon_ichol``.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A. This must be the same value as was supplied in the
        preceding call to ``complex_herm_precon_ichol``.

    a : complex, shape (`la`)
        The values returned in the array `a` by a previous call to
        ``complex_herm_precon_ichol``.

    la : int
        This must be the same value as was supplied in the preceding
        call to ``complex_herm_precon_ichol``.

    irow : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    icol : int, shape (`la`)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    ipiv : int, shape (`n`), modified in place
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    istr : int, shape (`n`+1)
        The values returned in arrays `irow`, `icol`, `ipiv` and `istr`
        by a previous call to ``complex_herm_precon_ichol``.

    b : complex, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||_infinity, where k
        is the output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'CG' or 'SYMMLQ'.

        (`errno` 1)
            On entry, `la` = *<value>* and `nnz` = *<value>*.

            Constraint: `la` >= 2*`nnz`.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[i-1],`icol`[i-1]) is a
            duplicate: i = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>*,
            `irow`[i-1] = *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `irow`[i-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, `istr` appears to be invalid.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error, code *<value>*, has occurred in an internal
            call. Check all function calls and array sizes. Seek expert
            help.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy could not be obtained. However a
            reasonable accuracy has been achieved.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_solve_ilu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[la].
        utils._BYREF_INT, # Input, la.
        utils._EngineIntArrayType.c_type, # Input, irow[la].
        utils._EngineIntArrayType.c_type, # Input, icol[la].
        utils._EngineIntArrayType.c_type, # Input/Output, ipiv[n].
        utils._EngineIntArrayType.c_type, # Input, istr[n+1].
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _la = utils._EngineIntScalarType(
        la, {'fun_name': fun_name, 'entity': 'la'},
    ).to_ctypes()
    _la_py = utils._EngineIntScalarType.to_py(_la)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_la_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipiv_cls = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    )
    _ipiv = _ipiv_cls.to_ctypes()
    _istr = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _n,
        _nnz,
        _a,
        _la,
        _irow,
        _icol,
        _ipiv,
        _istr,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _method._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = \'CG\' or \'SYMMLQ\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, la = {:s} and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: la >= 2*nnz.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[i-1],icol[i-1]) is a duplicate:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[i-1] = {:s}, irow[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[i-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istr appears to be invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However a reasonable accuracy has been achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ipiv_cls.refresh_py(_ipiv)
        _x_cls.refresh_py(_x)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def complex_herm_precon_ssor_solve(n, nnz, a, irow, icol, rdiag, omega, check, y, x):
    u"""
    Solution of linear system involving preconditioning matrix generated
    by applying SSOR to complex sparse Hermitian matrix.

    ``complex_herm_precon_ssor_solve`` solves a system of linear
    equations involving the preconditioning matrix corresponding to SSOR
    applied to a complex sparse Hermitian matrix, represented in
    symmetric coordinate storage format.

    For full information please refer to the NAG Library document for
    f11jr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jrf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    rdiag : float, shape (`n`)
        The elements of the diagonal matrix D^-1, where D is the
        diagonal part of A. Note that since A is Hermitian the elements
        of D^-1 are necessarily real.

    omega : float
        The relaxation parameter omega.

    check : str, length 1
        Specifies whether or not the input data should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `nnz`, `irow`,
          `icol` and `omega`.

        `check` = 'N'

          None of these checks are carried out.

    y : complex, shape (`n`)
        The right-hand side vector y.

    x : complex, shape (`n`), modified in place
        On exit: the solution vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>*,
            `irow`[I-1] = *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `irow`[i-1].

        (`errno` 3)
            On entry, I = *<value>*, `irow`[I-1] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`.

        (`errno` 4)
            The matrix A has no diagonal entry in row *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_precon_ssor_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._EngineFloat64ArrayType.c_type, # Input, rdiag[n].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, y[n].
        utils._EngineComplex128ArrayType.c_type, # Output, x[n].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rdiag = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _rdiag,
        _omega,
        _check,
        _y,
        _x,
        _iwork,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s}, irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'irow[I-1] = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def complex_herm_solve_jacssor(method, precon, n, nnz, a, irow, icol, omega, b, tol, maxitn, x, rdiag):
    u"""
    Solution of complex sparse Hermitian linear system, conjugate
    gradient/Lanczos method, Jacobi or SSOR preconditioner (Black Box).

    ``complex_herm_solve_jacssor`` solves a complex sparse Hermitian
    system of linear equations, represented in symmetric coordinate
    storage format, using a conjugate gradient or Lanczos method,
    without preconditioning, with Jacobi or with SSOR preconditioning.

    For full information please refer to the NAG Library document for
    f11js

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11jsf.html

    Parameters
    ----------
    method : str
        Specifies the iterative method to be used.

        `method` = 'CG'

          Conjugate gradient method.

        `method` = 'SYMMLQ'

          Lanczos method (SYMMLQ).

    precon : str, length 1
        Specifies the type of preconditioning to be used.

        `precon` = 'N'

          No preconditioning.

        `precon` = 'J'

          Jacobi.

        `precon` = 'S'

          Symmetric successive-over-relaxation (SSOR).

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements of the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    omega : float
        If `precon` = 'S', `omega` is the relaxation parameter omega to
        be used in the SSOR method. Otherwise `omega` need not be
        initialized.

    b : complex, shape (`n`)
        The right-hand side vector b.

    tol : float
        The required tolerance. Let x_k denote the approximate solution
        at iteration k, and r_k the corresponding residual. The
        algorithm is considered to have converged at iteration k if

            ||r_k||_infinity <=
            tau*(||b||_infinity+||A||_infinity||x_k||_infinity).

        If `tol` <= 0.0, tau = max(sqrt(epsilon),10
        epsilon,sqrt(n)epsilon) is used, where epsilon is the machine
        precision. Otherwise tau = max(`tol`,10 epsilon, sqrt(n)epsilon)
        is used.

    maxitn : int
        The maximum number of iterations allowed.

    x : complex, shape (`n`), modified in place
        On entry: an initial approximation to the solution vector x.

        On exit: an improved approximation to the solution vector x.

    rdiag : float, shape (`n`), modified in place
        On exit: the elements of the diagonal matrix D^-1, where D is
        the diagonal part of A. Note that since A is Hermitian the
        elements of D^-1 are necessarily real.

    Returns
    -------
    rnorm : float
        The final value of the residual norm ||r_k||, where k is the
        output value of `itn`.

    itn : int
        The number of iterations carried out.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxitn` = *<value>*.

            Constraint: `maxitn` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` < 1.0.

        (`errno` 1)
            On entry, `method` != 'CG' or 'SYMMLQ': `method` =
            *<value>*.

        (`errno` 1)
            On entry, `omega` = *<value>*.

            Constraint: 0.0 < `omega` < 2.0.

        (`errno` 1)
            On entry, `precon` != 'N', 'J' or 'S': `precon` = *<value>*.

        (`errno` 1)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 2)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            The matrix A has a zero diagonal entry in row *<value>*.

        (`errno` 3)
            The matrix A has no diagonal entry in row *<value>*.

        (`errno` 4)
            The required accuracy could not be obtained. However, a
            reasonable accuracy has been achieved and further iterations
            could not improve the result.

        (`errno` 5)
            The solution has not converged after *<value>* iterations.

        (`errno` 6)
            The preconditioner appears not to be positive definite. The
            computation cannot continue.

        (`errno` 7)
            The matrix of the coefficients `a` appears not to be
            positive definite. The computation cannot continue.

        (`errno` 8)
            A serious error, code *<value>*, has occurred in an internal
            call. Check all function calls and array sizes. Seek expert
            help.

        (`errno` 9)
            The matrix A has a non-real diagonal entry in row *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_solve_jacssor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11jsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=*).
        utils._BYREF_STRING, # Input, precon(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_FLOAT64, # Input, omega.
        utils._EngineComplex128ArrayType.c_type, # Input, b[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxitn.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_FLOAT64, # Output, rnorm.
        utils._BYREF_INT, # Output, itn.
        utils._EngineFloat64ArrayType.c_type, # Output, rdiag[n].
        utils._EngineComplex128ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of precon.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=None,
    ).to_ctypes()
    _precon = utils._EngineCharScalarType(
        precon, {'fun_name': fun_name, 'entity': 'precon'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _omega = utils._EngineFloat64ScalarType(
        omega, {'fun_name': fun_name, 'entity': 'omega'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _b = utils._EngineComplex128ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxitn = utils._EngineIntScalarType(
        maxitn, {'fun_name': fun_name, 'entity': 'maxitn'},
    ).to_ctypes()
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _rnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _itn = utils._EngineIntScalarType.empty_ctypes()
    _rdiag_cls = utils._EngineFloat64ArrayType(
        rdiag, {'fun_name': fun_name, 'entity': 'rdiag'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _rdiag = _rdiag_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineComplex128ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _precon,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _omega,
        _b,
        _tol,
        _maxitn,
        _x,
        _rnorm,
        _itn,
        _rdiag,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _method._length_,
        _precon._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxitn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxitn >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method != \'CG\' or \'SYMMLQ\': method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, omega = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < omega < 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, precon != \'N\', \'J\' or \'S\': precon = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a zero diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has no diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The required accuracy could not be obtained.',
                    'However, a reasonable accuracy has been achieved and further iterations',
                    'could not improve the result.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The solution has not converged after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The preconditioner appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of the coefficients a appears not to be positive definite.',
                    'The computation cannot continue.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A serious error, code {:s}, has occurred in an internal'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'call. Check all function calls and array sizes. Seek expert help.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A has a non-real diagonal entry in row {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _rdiag_cls.refresh_py(_rdiag)
        _rnorm_py = utils._EngineFloat64ScalarType.to_py(_rnorm)
        _itn_py = utils._EngineIntScalarType.to_py(_itn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rnorm', 'itn'],
        )(
            _rnorm_py,
            _itn_py,
        )
    _err_class.manage()
    return return_data

def direct_real_gen_setup(spec, n, icolzp, irowix, iprm):
    u"""
    Real sparse nonsymmetric linear systems, setup for
    ``direct_real_gen_lu``.

    ``direct_real_gen_setup`` computes a column permutation suitable for
    LU factorization (by ``direct_real_gen_lu``) of a real sparse matrix
    in compressed column (Harwell--Boeing) format and applies it to the
    matrix.
    This function must be called prior to ``direct_real_gen_lu``.

    For full information please refer to the NAG Library document for
    f11md

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mdf.html

    Parameters
    ----------
    spec : str, length 1
        Indicates the permutation to be applied.

        `spec` = 'N'

          The identity permutation is used.

        `spec` = 'U'

          The permutation in the `iprm` array is used, as supplied by
          you.

        `spec` = 'M'

          The permutation computed by the COLAMD algorithm is used

    n : int
        n, the order of the matrix A.

    icolzp : int, shape (max(1, (`n`+1)))
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (max(1, (`icolzp`[max(1, (`n`+1))-1]-1)))
        `irowix`[i-1] contains the row index in A for element A(i). See
        [the F11 Introduction].

    iprm : int, shape (max(1, (7*`n`))), modified in place
        On entry: the first `n` entries contain the column permutation
        if supplied by the user. This will be used if `spec` = 'U', and
        ignored otherwise. If used, it must consist of a permutation of
        all the integers in the range [0,(`n`-1)], the leftmost column
        of the matrix A denoted by 0 and the rightmost by `n`-1.
        Labelling columns in this way, `iprm`[i] = j means that column
        i-1 of A is in position j in A*P_c, where P_r AP_c = LU
        expresses the factorization to be performed.

        On exit: The column permutation given or computed is returned in
        the second `n` entries. The rest of the array contains data
        structures that will be used by other functions in the suite.
        The function computes the column elimination tree for A and a
        post-order permutation on the tree. It then compounds the `iprm`
        permutation given or computed by the COLAMD algorthm with the
        post-order permutation and this permutation is returned in the
        first `n` entries. This whole array is needed by the LU
        factorization function ``direct_real_gen_lu`` and associated
        functions ``direct_real_gen_solve``, ``direct_real_gen_refine``
        and ``direct_real_gen_diag`` and should be passed to them
        unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `spec` = *<value>*.

            Constraint: `spec` = 'N', 'U' or 'M'.

        (`errno` 2)
            Incorrect column permutations in array `iprm`.

        (`errno` 3)
            COLAMD algorithm failed.

        (`errno` 4)
            Incorrect specification of argument `icolzp`.

        (`errno` 5)
            Incorrect specification of argument `irowix`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, spec(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[max(1, (n+1))].
        utils._EngineIntArrayType.c_type, # Input, irowix[max(1, (icolzp[max(1, (n+1))-1]-1))].
        utils._EngineIntArrayType.c_type, # Input/Output, iprm[max(1, (7*n))].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of spec.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _spec = utils._EngineCharScalarType(
        spec, {'fun_name': fun_name, 'entity': 'spec'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(max(1, (_n_py+1)),),
    ).to_ctypes()
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(max(1, (utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, (max(1, (_n_py+1))-1,))-1)),),
    ).to_ctypes()
    _iprm_cls = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(max(1, (7*_n_py)),),
    )
    _iprm = _iprm_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _spec,
        _n,
        _icolzp,
        _irowix,
        _iprm,
        _errbuf,
        _ierr,
        _spec._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _spec_py = utils._EngineCharScalarType.to_py(_spec)
        if _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, spec = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_spec_py),
                    ),
                    'Constraint: spec = \'N\', \'U\' or \'M\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'COLAMD algorithm failed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect specification of argument icolzp.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect specification of argument irowix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iprm_cls.refresh_py(_iprm)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_lu(n, irowix, a, iprm, thresh, nzlmx, nzlumx, nzumx, il, lval, iu, uval):
    u"""
    LU factorization of real sparse matrix.

    ``direct_real_gen_lu`` computes the LU factorization of a real
    sparse matrix in compressed column (Harwell--Boeing),
    column-permuted format.

    For full information please refer to the NAG Library document for
    f11me

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mef.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    irowix : int, shape (:)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (:)
        The array of nonzero values in the sparse matrix A.

    iprm : int, shape (7*`n`), modified in place
        On entry: contains the column permutation which defines the
        permutation P_c and associated data structures as computed by
        function ``direct_real_gen_setup``.

        On exit: part of the array is modified to record the row
        permutation P_r determined by pivoting.

    thresh : float
        The diagonal pivoting threshold, t. At step j of the Gaussian
        elimination, if |A_{jj}| >= t*(max_{i >= j}|A_{ij}|), use A_{jj}
        as a pivot, otherwise use max_{i >= j}|A_{ij}|. A value of t = 1
        corresponds to partial pivoting, a value of t = 0 corresponds to
        always choosing the pivot on the diagonal (unless it is zero).

    nzlmx : int
        Indicates the available size of array `il`. The dimension of
        `il` should be at least 7*`n`+`nzlmx`+4. A good range for
        `nzlmx` that works for many problems is nnz to 8*nnz, where nnz
        is the number of nonzeros in the sparse matrix A. If, on exit,
        `errno` = 2, the given `nzlmx` was too small and you should
        attempt to provide more storage and call the function again.

    nzlumx : int
        Indicates the available size of array `lval`. The dimension of
        `lval` should be at least `nzlumx`.

    nzumx : int
        Indicates the available sizes of arrays `iu` and `uval`. The
        dimension of `iu` should be at least 2*`n`+`nzumx`+1 and the
        dimension of `uval` should be at least `nzumx`. A good range for
        `nzumx` that works for many problems is nnz to 8*nnz, where nnz
        is the number of nonzeros in the sparse matrix A. If, on exit,
        `errno` = 3, the given `nzumx` was too small and you should
        attempt to provide more storage and call the function again.

    il : int, shape (7*`n`+`nzlmx`+4), modified in place
        On exit: encapsulates the sparsity pattern of matrix L.

    lval : float, shape (`nzlumx`), modified in place
        On exit: records the nonzero values of matrix L and some of the
        nonzero values of matrix U.

    iu : int, shape (2*`n`+`nzumx`+1), modified in place
        On exit: encapsulates the sparsity pattern of matrix U.

    uval : float, shape (`nzumx`), modified in place
        On exit: records some of the nonzero values of matrix U.

    Returns
    -------
    nzlumx : int
        If `errno` = 4, the given `nzlumx` was too small and is reset to
        a value that will be sufficient. You should then provide the
        indicated storage and call the function again.

    nnzl : int
        The number of nonzero values in the matrix L.

    nnzu : int
        The number of nonzero values in the matrix U.

    flop : float
        The number of floating-point operations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nzumx` = *<value>*.

            Constraint: `nzumx` >= 1.

        (`errno` 1)
            On entry, `nzlumx` = *<value>*.

            Constraint: `nzlumx` >= 1.

        (`errno` 1)
            On entry, `nzlmx` = *<value>*.

            Constraint: `nzlmx` >= 1.

        (`errno` 1)
            On entry, `thresh` = *<value>*.

            Constraint: 0.0 <= `thresh` <= 1.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            Insufficient `nzlmx`.

        (`errno` 3)
            Insufficient `nzumx`.

        (`errno` 4)
            Insufficient `nzlumx`.

        (`errno` 5)
            The matrix is singular -- no factorization possible.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_lu'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11meft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineIntArrayType.c_type, # Input/Output, iprm[7*n].
        utils._BYREF_FLOAT64, # Input, thresh.
        utils._BYREF_INT, # Input, nzlmx.
        utils._BYREF_INT, # Input/Output, nzlumx.
        utils._BYREF_INT, # Input, nzumx.
        utils._EngineIntArrayType.c_type, # Output, il[7*n+nzlmx+4].
        utils._EngineFloat64ArrayType.c_type, # Output, lval[nzlumx].
        utils._EngineIntArrayType.c_type, # Output, iu[2*n+nzumx+1].
        utils._EngineFloat64ArrayType.c_type, # Output, uval[nzumx].
        utils._BYREF_INT, # Output, nnzl.
        utils._BYREF_INT, # Output, nnzu.
        utils._BYREF_FLOAT64, # Output, flop.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(None,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(None,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _iprm_cls = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    )
    _iprm = _iprm_cls.to_ctypes()
    _thresh = utils._EngineFloat64ScalarType(
        thresh, {'fun_name': fun_name, 'entity': 'thresh'},
    ).to_ctypes()
    _nzlmx = utils._EngineIntScalarType(
        nzlmx, {'fun_name': fun_name, 'entity': 'nzlmx'},
    ).to_ctypes()
    _nzlumx = utils._EngineIntScalarType(
        nzlumx, {'fun_name': fun_name, 'entity': 'nzlumx'},
    ).to_ctypes()
    _nzumx = utils._EngineIntScalarType(
        nzumx, {'fun_name': fun_name, 'entity': 'nzumx'},
    ).to_ctypes()
    _nzlmx_py = utils._EngineIntScalarType.to_py(_nzlmx)
    _il_cls = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(7*_n_py+_nzlmx_py+4,),
        check_contents=False,
    )
    _il = _il_cls.to_ctypes(can_reference_contents=False)
    _nzlumx_py = utils._EngineIntScalarType.to_py(_nzlumx)
    _lval_cls = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(_nzlumx_py,),
        check_contents=False,
    )
    _lval = _lval_cls.to_ctypes(can_reference_contents=False)
    _nzumx_py = utils._EngineIntScalarType.to_py(_nzumx)
    _iu_cls = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(2*_n_py+_nzumx_py+1,),
        check_contents=False,
    )
    _iu = _iu_cls.to_ctypes(can_reference_contents=False)
    _uval_cls = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(_nzumx_py,),
        check_contents=False,
    )
    _uval = _uval_cls.to_ctypes(can_reference_contents=False)
    _nnzl = utils._EngineIntScalarType.empty_ctypes()
    _nnzu = utils._EngineIntScalarType.empty_ctypes()
    _flop = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _irowix,
        _a,
        _iprm,
        _thresh,
        _nzlmx,
        _nzlumx,
        _nzumx,
        _il,
        _lval,
        _iu,
        _uval,
        _nnzl,
        _nnzu,
        _flop,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _nzlumx_py = utils._EngineIntScalarType.to_py(_nzlumx)
        _thresh_py = utils._EngineFloat64ScalarType.to_py(_thresh)
        if _ierr_py == 1 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nzumx = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nzumx_py),
                    ),
                    'Constraint: nzumx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nzlumx = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nzlumx_py),
                    ),
                    'Constraint: nzlumx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nzlmx = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nzlmx_py),
                    ),
                    'Constraint: nzlmx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, thresh = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_thresh_py),
                    ),
                    'Constraint: 0.0 <= thresh <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Insufficient nzlmx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Insufficient nzumx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Insufficient nzlumx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix is singular -- no factorization possible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iprm_cls.refresh_py(_iprm)
        _nzlumx_py = utils._EngineIntScalarType.to_py(_nzlumx)
        _il_cls.refresh_py(_il)
        _lval_cls.refresh_py(_lval)
        _iu_cls.refresh_py(_iu)
        _uval_cls.refresh_py(_uval)
        _nnzl_py = utils._EngineIntScalarType.to_py(_nnzl)
        _nnzu_py = utils._EngineIntScalarType.to_py(_nnzu)
        _flop_py = utils._EngineFloat64ScalarType.to_py(_flop)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nzlumx', 'nnzl', 'nnzu', 'flop'],
        )(
            _nzlumx_py,
            _nnzl_py,
            _nnzu_py,
            _flop_py,
        )
    _err_class.manage()
    return return_data

def direct_real_gen_solve(trans, n, iprm, il, lval, iu, uval, nrhs, b):
    u"""
    Solution of real sparse simultaneous linear equations (coefficient
    matrix already factorized).

    ``direct_real_gen_solve`` solves a real sparse system of linear
    equations with multiple right-hand sides given an LU factorization
    of the sparse matrix computed by ``direct_real_gen_lu``.

    For full information please refer to the NAG Library document for
    f11mf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mff.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether AX = B or A^TX = B is solved.

        `trans` = 'N'

          AX = B is solved.

        `trans` = 'T'

          A^TX = B is solved.

    n : int
        n, the order of the matrix A.

    iprm : int, shape (7*`n`)
        The column permutation which defines P_c, the row permutation
        which defines P_r, plus associated data structures as computed
        by ``direct_real_gen_lu``.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    nrhs : int
        nrhs, the number of right-hand sides in B.

    b : float, shape (`n`, `nrhs`), modified in place
        On entry: the `n` by `nrhs` right-hand side matrix B.

        On exit: the `n` by `nrhs` solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 2)
            Incorrect row permutations in array `iprm`.

        (`errno` 3)
            Incorrect column permutations in array `iprm`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_solve'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, iprm[7*n].
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _iprm = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _iprm,
        _il,
        _lval,
        _iu,
        _uval,
        _nrhs,
        _b,
        _ldb,
        _errbuf,
        _ierr,
        _trans._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _trans_py = utils._EngineCharScalarType.to_py(_trans)
        if _ierr_py == 1 and _ifmt == -8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nrhs_py),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_trans_py),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect row permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_cond(norm, n, il, lval, iu, uval, anorm):
    u"""
    Estimate condition number of real matrix, matrix already factorized
    by ``direct_real_gen_lu``.

    ``direct_real_gen_cond`` computes an estimate of the reciprocal of
    the condition number of a sparse matrix given an LU factorization of
    the matrix computed by ``direct_real_gen_lu``.

    For full information please refer to the NAG Library document for
    f11mg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is to be
        estimated.

        `norm` = '1' or 'O'

          kappa_1(A) is estimated.

        `norm` = 'I'

          kappa_infinity (A) is estimated.

    n : int
        n, the order of the matrix A.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the matrix A.

        If `norm` = 'I', the infinity-norm of the matrix A.

        `anorm` may be computed by calling ``direct_real_gen_norm`` with
        the same value for the argument `norm`.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `anorm` = *<value>*.

            Constraint: `anorm` >= 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'O' or 'I'.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_cond'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_FLOAT64, # Input, anorm.
        utils._BYREF_FLOAT64, # Output, rcond.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType(
        anorm, {'fun_name': fun_name, 'entity': 'anorm'},
    ).to_ctypes()
    _rcond = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _n,
        _il,
        _lval,
        _iu,
        _uval,
        _anorm,
        _rcond,
        _errbuf,
        _ierr,
        _norm._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        if _ierr_py == 1 and _ifmt == -7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, anorm = {:s}.'.format(
                        utils._STR_FMT_FLOAT.format(_anorm_py),
                    ),
                    'Constraint: anorm >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'O\' or \'I\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rcond_py = utils._EngineFloat64ScalarType.to_py(_rcond)
        return_data = _rcond_py
    _err_class.manage()
    return return_data

def direct_real_gen_refine(trans, n, icolzp, irowix, a, iprm, il, lval, iu, uval, nrhs, b, x, ferr, berr):
    u"""
    Refined solution with error bounds of real system of linear
    equations, multiple right-hand sides.

    ``direct_real_gen_refine`` returns error bounds for the solution of
    a real sparse system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B.
    It improves the solution by iterative refinement in standard
    precision, in order to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f11mh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mhf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether AX = B or A^TX = B is solved.

        `trans` = 'N'

          AX = B is solved.

        `trans` = 'T'

          A^TX = B is solved.

    n : int
        n, the order of the matrix A.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (asize)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    iprm : int, shape (7*`n`)
        The column permutation which defines P_c, the row permutation
        which defines P_r, plus associated data structures as computed
        by ``direct_real_gen_lu``.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    nrhs : int
        nrhs, the number of right-hand sides in B.

    b : float, shape (`n`, `nrhs`)
        The n by nrhs right-hand side matrix B.

    x : float, shape (`n`, `nrhs`), modified in place
        On entry: the n by nrhs solution matrix X, as returned by
        ``direct_real_gen_solve``.

        On exit: the n by nrhs improved solution matrix X.

    ferr : float, shape (`nrhs`), modified in place
        On exit: `ferr`[j-1] contains an estimated error bound for the
        jth solution vector, that is, the jth column of X, for j =
        1,2,...,nrhs.

    berr : float, shape (`nrhs`), modified in place
        On exit: `berr`[j-1] contains the component-wise backward error
        bound beta for the jth solution vector, that is, the jth column
        of X, for j = 1,2,...,nrhs.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nrhs` = *<value>*.

            Constraint: `nrhs` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 2)
            Incorrect row permutations in array `iprm`.

        (`errno` 3)
            Incorrect column permutations in array `iprm`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_refine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineIntArrayType.c_type, # Input, iprm[7*n].
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_INT, # Input, nrhs.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, nrhs].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, nrhs].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Output, ferr[nrhs].
        utils._EngineFloat64ArrayType.c_type, # Output, berr[nrhs].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _iprm = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _nrhs = utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _nrhs_py = utils._EngineIntScalarType.to_py(_nrhs)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldx = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _nrhs_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _ferr_cls = utils._EngineFloat64ArrayType(
        ferr, {'fun_name': fun_name, 'entity': 'ferr'},
        exp_shape=(_nrhs_py,),
        check_contents=False,
    )
    _ferr = _ferr_cls.to_ctypes(can_reference_contents=False)
    _berr_cls = utils._EngineFloat64ArrayType(
        berr, {'fun_name': fun_name, 'entity': 'berr'},
        exp_shape=(_nrhs_py,),
        check_contents=False,
    )
    _berr = _berr_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _icolzp,
        _irowix,
        _a,
        _iprm,
        _il,
        _lval,
        _iu,
        _uval,
        _nrhs,
        _b,
        _ldb,
        _x,
        _ldx,
        _ferr,
        _berr,
        _errbuf,
        _ierr,
        _trans._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _trans_py = utils._EngineCharScalarType.to_py(_trans)
        if _ierr_py == 1 and _ifmt == -11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrhs = {:s}.'.format(
                        utils._STR_FMT_INT.format(_nrhs_py),
                    ),
                    'Constraint: nrhs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_trans_py),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect row permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _ferr_cls.refresh_py(_ferr)
        _berr_cls.refresh_py(_berr)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_matmul(trans, n, m, alpha, icolzp, irowix, a, b, beta, c):
    u"""
    Real sparse nonsymmetric matrix-matrix multiply, compressed column
    storage.

    ``direct_real_gen_matmul`` computes a matrix-matrix or transposed
    matrix-matrix product involving a real, square, sparse nonsymmetric
    matrix stored in compressed column (Harwell--Boeing) format.

    For full information please refer to the NAG Library document for
    f11mk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mkf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix A is transposed.

        `trans` = 'N'

          alpha AB+beta C is computed.

        `trans` = 'T'

          alpha A^TB+beta C is computed.

    n : int
        n, the order of the matrix A.

    m : int
        m, the number of columns of matrices B and C.

    alpha : float
        alpha, the scalar factor in the matrix multiplication.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (asize)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    b : float, shape (`n`, `m`)
        The n by m matrix B.

    beta : float
        The scalar factor beta.

    c : float, shape (`n`, `m`), modified in place
        On entry: the n by m matrix C.

        On exit: C is overwritten by alpha AB+beta C or alpha A^TB+beta
        C depending on the value of `trans`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_matmul'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, m].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _ldb = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _ldc = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')), {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _beta_py = utils._EngineFloat64ScalarType.to_py(_beta)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=_beta_py != 0.0,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=_beta_py != 0.0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _m,
        _alpha,
        _icolzp,
        _irowix,
        _a,
        _b,
        _ldb,
        _beta,
        _c,
        _ldc,
        _errbuf,
        _ierr,
        _trans._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _trans_py = utils._EngineCharScalarType.to_py(_trans)
        if _ierr_py == 1 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT.format(_m_py),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_trans_py),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _c_cls.refresh_py(_c)
        return_data = None
    _err_class.manage()
    return return_data

def direct_real_gen_norm(norm, n, icolzp, irowix, a):
    u"""
    1-norm, infinity-norm, largest absolute element, real, square,
    sparse matrix.

    ``direct_real_gen_norm`` computes the 1-norm, the infinity-norm or
    the maximum absolute value of the elements of a real, square, sparse
    matrix which is held in compressed column (Harwell--Boeing) format.

    For full information please refer to the NAG Library document for
    f11ml

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mlf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned in `anorm`.

        `norm` = '1' or 'O'

          The 1-norm ||A||_1 of the matrix is computed, that is max_{1
          <= j <= n}sum_i = 1^n|A_{ij}|.

        `norm` = 'I'

          The infinity-norm ||A||_infinity of the matrix is computed,
          that is max_{1 <= i <= n}sum_j = 1^n|A_{ij}|.

        `norm` = 'M'

          The value max_{1 <= i,j <= n}(|A_{ij}|) (not a norm).

    n : int
        n, the order of the matrix A.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    irowix : int, shape (asize)
        The row index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    Returns
    -------
    anorm : float
        The computed quantity relating the matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1', 'O', 'I' or 'M'.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_norm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, norm(len=1).
        utils._BYREF_FLOAT64, # Output, anorm.
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[*].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of norm.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _norm = utils._EngineCharScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'}, exp_chrlen=1,
    ).to_ctypes()
    _anorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _norm,
        _anorm,
        _n,
        _icolzp,
        _irowix,
        _a,
        _errbuf,
        _ierr,
        _norm._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _norm_py = utils._EngineCharScalarType.to_py(_norm)
        if _ierr_py == 1 and _ifmt == -3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_norm_py),
                    ),
                    'Constraint: norm = \'1\', \'O\', \'I\' or \'M\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _anorm_py = utils._EngineFloat64ScalarType.to_py(_anorm)
        return_data = _anorm_py
    _err_class.manage()
    return return_data

def direct_real_gen_diag(n, icolzp, a, iprm, il, lval, iu, uval):
    u"""
    Real sparse nonsymmetric linear systems, diagnostic for
    ``direct_real_gen_lu``.

    ``direct_real_gen_diag`` computes the reciprocal pivot growth factor
    of an LU factorization of a real sparse matrix in compressed column
    (Harwell--Boeing) format.

    For full information please refer to the NAG Library document for
    f11mm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11mmf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    icolzp : int, shape (`n`+1)
        The new column index array of sparse matrix A. See [the F11
        Introduction].

    a : float, shape (asize)
        The array of nonzero values in the sparse matrix A.

    iprm : int, shape (7*`n`)
        The column permutation which defines P_c, the row permutation
        which defines P_r, plus associated data structures as computed
        by ``direct_real_gen_lu``.

    il : int, shape (:)
        Records the sparsity pattern of matrix L as computed by
        ``direct_real_gen_lu``.

    lval : float, shape (:)
        Records the nonzero values of matrix L and some nonzero values
        of matrix U as computed by ``direct_real_gen_lu``.

    iu : int, shape (:)
        Records the sparsity pattern of matrix U as computed by
        ``direct_real_gen_lu``.

    uval : float, shape (:)
        Records some nonzero values of matrix U as computed by
        ``direct_real_gen_lu``.

    Returns
    -------
    rpg : float
        The reciprocal pivot growth factor
        max_j(||A_j||_infinity/||U_j||_infinity). If the reciprocal
        pivot growth factor is much less than 1, the stability of the LU
        factorization may be poor.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            Incorrect column permutations in array `iprm`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.direct_real_gen_diag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11mmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineFloat64ArrayType.c_type, # Input, a[*].
        utils._EngineIntArrayType.c_type, # Input, iprm[7*n].
        utils._EngineIntArrayType.c_type, # Input, il[*].
        utils._EngineFloat64ArrayType.c_type, # Input, lval[*].
        utils._EngineIntArrayType.c_type, # Input, iu[*].
        utils._EngineFloat64ArrayType.c_type, # Input, uval[*].
        utils._BYREF_FLOAT64, # Output, rpg.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    if _n_py >= 0:
        _vl_asize_py = utils._EngineIntArrayType.get_element(icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, ((_n_py+1)-1,))-1
    else:
        _vl_asize_py = 0
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_vl_asize_py,),
    ).to_ctypes()
    _iprm = utils._EngineIntArrayType(
        iprm, {'fun_name': fun_name, 'entity': 'iprm'},
        exp_shape=(7*_n_py,),
    ).to_ctypes()
    _il = utils._EngineIntArrayType(
        il, {'fun_name': fun_name, 'entity': 'il'},
        exp_shape=(None,),
    ).to_ctypes()
    _lval = utils._EngineFloat64ArrayType(
        lval, {'fun_name': fun_name, 'entity': 'lval'},
        exp_shape=(None,),
    ).to_ctypes()
    _iu = utils._EngineIntArrayType(
        iu, {'fun_name': fun_name, 'entity': 'iu'},
        exp_shape=(None,),
    ).to_ctypes()
    _uval = utils._EngineFloat64ArrayType(
        uval, {'fun_name': fun_name, 'entity': 'uval'},
        exp_shape=(None,),
    ).to_ctypes()
    _rpg = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _icolzp,
        _a,
        _iprm,
        _il,
        _lval,
        _iu,
        _uval,
        _rpg,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == -1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Incorrect column permutations in array iprm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rpg_py = utils._EngineFloat64ScalarType.to_py(_rpg)
        return_data = _rpg_py
    _err_class.manage()
    return return_data

def real_gen_matvec(trans, n, nnz, a, irow, icol, check, x, y):
    u"""
    Real, sparse, nonsymmetric matrix-vector multiply.

    ``real_gen_matvec`` computes a matrix-vector or transposed
    matrix-vector product involving a real sparse nonsymmetric matrix
    stored in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11xa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11xaf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix A is transposed.

        `trans` = 'N'

          y = Ax is computed.

        `trans` = 'T'

          y = A^Tx is computed.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : float, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``real_gen_sort`` may be used to order the elements
        in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

          Checks are carried on the values of `n`, `nnz`, `irow` and
          `icol`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    x : float, shape (`n`)
        The vector x.

    y : float, shape (`n`), modified in place
        On exit: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def real_symm_matvec(n, nnz, a, irow, icol, check, x, y):
    u"""
    Real sparse symmetric matrix-vector multiply.

    ``real_symm_matvec`` computes a matrix-vector product involving a
    real sparse symmetric matrix stored in symmetric coordinate storage
    format.

    For full information please refer to the NAG Library document for
    f11xe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11xef.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        A.

    a : float, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``real_symm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the SCS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `nnz`, `irow` and
          `icol`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    x : float, shape (`n`)
        The vector x.

    y : float, shape (`n`), modified in place
        On exit: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def complex_gen_matvec(trans, n, nnz, a, irow, icol, check, x, y):
    u"""
    Complex sparse non-Hermitian matrix-vector multiply.

    ``complex_gen_matvec`` computes a matrix-vector or conjugate
    transposed matrix-vector product involving a complex sparse
    non-Hermitian matrix stored in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11xn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11xnf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies whether or not the matrix A is conjugate transposed.

        `trans` = 'N'

          y = Ax is computed.

        `trans` = 'T'

          y = A^Hx is computed.

    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the matrix A, ordered by increasing row
        index, and by increasing column index within each row. Multiple
        entries for the same row and column indices are not permitted.
        The function ``complex_gen_sort`` may be used to order the
        elements in this way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the CS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

          Checks are carried on the values of `n`, `nnz`, `irow` and
          `icol`.

        `check` = 'N'

          None of these checks are carried out.

        See also [Further Comments].

    x : complex, shape (`n`)
        The vector x.

    y : complex, shape (`n`), modified in place
        On exit: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `trans` = *<value>*.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`^2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, trans(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, x[n].
        utils._EngineComplex128ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of trans.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _trans = utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _trans,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _trans._length_,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, trans = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: trans = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def complex_herm_matvec(n, nnz, a, irow, icol, check, x, y):
    u"""
    Complex sparse Hermitian matrix-vector multiply.

    ``complex_herm_matvec`` computes a matrix-vector product involving a
    complex sparse Hermitian matrix stored in symmetric coordinate
    storage format.

    For full information please refer to the NAG Library document for
    f11xs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11xsf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the lower triangular part of
        the matrix A.

    a : complex, shape (`nnz`)
        The nonzero elements in the lower triangular part of the matrix
        A, ordered by increasing row index, and by increasing column
        index within each row. Multiple entries for the same row and
        column indices are not permitted. The function
        ``complex_herm_sort`` may be used to order the elements in this
        way.

    irow : int, shape (`nnz`)
        The row indices of the nonzero elements supplied in array `a`.

    icol : int, shape (`nnz`)
        The column indices of the nonzero elements supplied in array
        `a`.

    check : str, length 1
        Specifies whether or not the SCS representation of the matrix A,
        values of `n`, `nnz`, `irow` and `icol` should be checked.

        `check` = 'C'

          Checks are carried out on the values of `n`, `nnz`, `irow` and
          `icol`.

        `check` = 'N'

          None of these checks are carried out.

    x : complex, shape (`n`)
        The vector x.

    y : complex, shape (`n`), modified in place
        On exit: the vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `check` = *<value>*.

            Constraint: `check` = 'C' or 'N'.

        (`errno` 2)
            On entry, `nnz` = *<value>* and `n` = *<value>*.

            Constraint: `nnz` <= `n`*(`n`+1)/2.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, the location (`irow`[I-1],`icol`[I-1]) is a
            duplicate: I = *<value>*.

        (`errno` 3)
            On entry, `a`[i-1] is out of order: i = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 3)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_matvec'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11xsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input, a[nnz].
        utils._EngineIntArrayType.c_type, # Input, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input, icol[nnz].
        utils._BYREF_STRING, # Input, check(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input, x[n].
        utils._EngineComplex128ArrayType.c_type, # Output, y[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of check.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _irow = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _icol = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _check = utils._EngineCharScalarType(
        check, {'fun_name': fun_name, 'entity': 'check'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _check,
        _x,
        _y,
        _errbuf,
        _ierr,
        _check._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, check = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: check = \'C\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nnz <= n*(n+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the location (irow[I-1],icol[I-1]) is a duplicate:',
                    'I = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[i-1] is out of order:',
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'icol[I-1] = {:s} and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def sym_rcm(n, nnz, icolzp, irowix, lopts, mask, perm, info):
    u"""
    Reverse Cuthill--McKee reordering of a sparse symmetric matrix in
    CCS format.

    ``sym_rcm`` reduces the bandwidth of a sparse symmetric matrix
    stored in compressed column storage format using the Reverse
    Cuthill--McKee algorithm.

    For full information please refer to the NAG Library document for
    f11ye

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11yef.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of nonzero elements in the matrix A.

    icolzp : int, shape (`n`+1)
        `icolzp` records the index into `irowix` which starts each new
        column.

    irowix : int, shape (`nnz`)
        The row indices corresponding to the nonzero elements in the
        matrix A.

    lopts : bool, shape (5)
        The options to be used by ``sym_rcm``.

        `lopts`[0] = True

          Row/column i of the matrix A will only be referenced if
          `mask`[i-1] != 0, otherwise `mask` will be ignored.

        `lopts`[1] = True

          The final permutation will not be reversed, that is, the
          Cuthill--McKee ordering will be returned. The bandwidth of the
          non-reversed matrix will be the same but the profile will be
          the same or larger.

        `lopts`[2] = True

          The matrix A will be checked for symmetrical sparsity pattern,
          otherwise not.

        `lopts`[3] = True

          The bandwidth and profile of the unpermuted matrix will be
          calculated, otherwise not.

        `lopts`[4] = True

          The bandwidth and profile of the permuted matrix will be
          calculated, otherwise not.

    mask : None or int, shape (:)
        Note: the required length for this argument is determined as
        follows: if `lopts`[0]= True: `n`; otherwise: 0.

        `mask` is only referenced if `lopts`[0] is True. A value of
        `mask`[i-1] = 0 indicates that the node corresponding to row or
        column i is not to be referenced. A value of `mask`[i-1] != 0
        indicates that the node corresponding to row or column i is to
        be referenced. In particular, rows and columns not referenced
        will not be permuted.

    perm : int, shape (`n`), modified in place
        On exit: this will contain the permutation vector that describes
        the permutation matrix P for the reordering of the matrix A. The
        elements of the permutation matrix P are zero except for the
        unit elements in row i and column `perm`[i-1], i = 1,2,...n.

    info : int, shape (4), modified in place
        On exit: statistics about the matrix A and the permuted matrix.
        The quantities below are calculated using any masking in effect
        otherwise the value zero is returned.

        `info`[0]

          The bandwidth of the matrix A, if `lopts`[3] = True.

        `info`[1]

          The profile of the matrix A, if `lopts`[3] = True.

        `info`[2]

          The bandwidth of the permuted matrix PAP^T, if `lopts`[4] =
          True.

        `info`[3]

          The profile of the permuted matrix PAP^T, if `lopts`[4] =
          True.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 1.

        (`errno` 3)
            On entry, `irowix`[*<value>*] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `irowix`[i-1] <= `n` for all i.

        (`errno` 4)
            On entry, `icolzp`[*<value>*] = *<value>* and `nnz` =
            *<value>*.

            Constraint: 1 <= `icolzp`[i-1] <= `nnz` for all i.

        (`errno` 5)
            On entry, `icolzp`[0] = *<value>*.

            Constraint: `icolzp`[0] = 1.

        (`errno` 5)
            On entry, `icolzp`[`n`] = *<value>* and `nnz` = *<value>*.

            Constraint: `icolzp`[`n`] = `nnz`+1.

        (`errno` 6)
            On entry, the matrix A is not symmetric.

            Element (*<value>*,*<value>*) has no symmetric element.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.sym_rcm'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11yeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nnz.
        utils._EngineIntArrayType.c_type, # Input, icolzp[n+1].
        utils._EngineIntArrayType.c_type, # Input, irowix[nnz].
        utils._EngineLogicalArrayType.c_type, # Input, lopts[5].
        utils._EngineIntArrayType.c_type, # Input, mask[*].
        utils._EngineIntArrayType.c_type, # Output, perm[n].
        utils._EngineIntArrayType.c_type, # Output, info[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _icolzp = utils._EngineIntArrayType(
        icolzp, {'fun_name': fun_name, 'entity': 'icolzp'},
        exp_shape=(_n_py+1,),
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _irowix = utils._EngineIntArrayType(
        irowix, {'fun_name': fun_name, 'entity': 'irowix'},
        exp_shape=(_nnz_py,),
    ).to_ctypes()
    _lopts = utils._EngineLogicalArrayType(
        lopts, {'fun_name': fun_name, 'entity': 'lopts'},
        exp_shape=(5,),
    ).to_ctypes()
    if utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)):
        if mask is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'mask'},
                'must not be None when lopts[0]',
            )
    if utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)):
        _mask_dim1 = _n_py
    else:
        _mask_dim1 = 0
    exp_shape_mask = (_mask_dim1,)
    _mask = utils._EngineIntArrayType(
        mask, {'fun_name': fun_name, 'entity': 'mask'},
        exp_shape=exp_shape_mask,
        check_contents=utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)),
        allow_none=True,
    ).to_ctypes(can_reference_contents=utils._EngineLogicalArrayType.get_element(lopts, {'fun_name': fun_name, 'entity': 'lopts'}, (0,)))
    _perm_cls = utils._EngineIntArrayType(
        perm, {'fun_name': fun_name, 'entity': 'perm'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _perm = _perm_cls.to_ctypes(can_reference_contents=False)
    _info_cls = utils._EngineIntArrayType(
        info, {'fun_name': fun_name, 'entity': 'info'},
        exp_shape=(4,),
        check_contents=False,
    )
    _info = _info_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _icolzp,
        _irowix,
        _lopts,
        _mask,
        _perm,
        _info,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1020:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1030:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irowix[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irowix[i-1] <= n for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1040:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icolzp[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= icolzp[i-1] <= nnz for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1050:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icolzp[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icolzp[0] = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1055:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icolzp[n] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: icolzp[n] = nnz+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1060:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the matrix A is not symmetric.',
                    'Element ({:s},{:s}) has no symmetric'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'element.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _perm_cls.refresh_py(_perm)
        _info_cls.refresh_py(_info)
        return_data = None
    _err_class.manage()
    return return_data

def real_gen_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Real sparse nonsymmetric matrix reorder function.

    ``real_gen_sort`` sorts the nonzero elements of a real sparse
    nonsymmetric matrix, represented in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11zaf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : float, shape (`nnz`), modified in place
        On entry: the nonzero elements of the matrix A. These may be in
        any order and there may be multiple nonzero elements with the
        same row and column indices.

        On exit: the nonzero elements ordered by increasing row index,
        and by increasing column index within each row. Each nonzero
        element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        On entry: the row indices corresponding to the elements supplied
        in the array `a`.

        On exit: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        On entry: the column indices corresponding to the elements
        supplied in the array `a`.

        On exit: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

          Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

          The relevant values in `a` are summed.

        `dup` = 'F'

          The function fails with `errno` = 3 on detecting a duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

          The entries are removed.

        `zer` = 'K'

          The entries are kept.

        `zer` = 'F'

          The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row i and
            column j: i = *<value>*, j = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row i and column j:
            i = *<value>*, j = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_gen_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 6 and _ifmt == 99994:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99992:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99988:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row i and column j:',
                    'i = {:s}, j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 99987:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row i and column j:',
                    'i = {:s}, j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def real_symm_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Real sparse symmetric matrix reorder function.

    ``real_symm_sort`` sorts the nonzero elements of a real sparse
    symmetric matrix, represented in symmetric coordinate storage
    format.

    For full information please refer to the NAG Library document for
    f11zb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11zbf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : float, shape (`nnz`), modified in place
        On entry: the nonzero elements of the lower triangular part of
        the real matrix A. These may be in any order and there may be
        multiple nonzero elements with the same row and column indices.

        On exit: the lower triangular nonzero elements ordered by
        increasing row index, and by increasing column index within each
        row. Each nonzero element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        On entry: the row indices corresponding to the elements supplied
        in the array `a`.

        On exit: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        On entry: the column indices corresponding to the elements
        supplied in the array `a`.

        On exit: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

          Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

          The relevant values in `a` are summed.

        `dup` = 'F'

          The function fails with `errno` = 3 on detecting a duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

          The entries are removed.

        `zer` = 'K'

          The entries are kept.

        `zer` = 'F'

          The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and
            `irow`[i-1] = *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `irow`[i-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row I and
            column J: I = *<value>*, J = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row I and column J:
            I = *<value>*, J = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_symm_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= irow[i-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def real_rect_sort(m, n, nnz, a, irow, icol, istc, store, dup, zer):
    u"""
    Sorts and merges elements of a real sparse rectangular matrix,
    represented in coordinate storage format, and provides the
    compressed column storage format of the result.

    ``real_rect_sort`` sorts the nonzero elements of a real sparse
    rectangular matrix, represented in coordinate storage or compressed
    column storage format.

    For full information please refer to the NAG Library document for
    f11zc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11zcf.html

    Parameters
    ----------
    m : int
        m, the number of rows in the matrix A.

    n : int
        n, the number of columns in the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : float, shape (`nnz`), modified in place
        On entry: the nonzero elements of the matrix A. If `store` = 1,
        the elements may be in any order. If `store` = 2, the elements
        must be ordered by increasing column index. There may be
        multiple nonzero elements with the same row and column indices.

        On exit: the nonzero elements ordered by increasing column
        index, and by increasing row index within each column. Each
        nonzero element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        On entry: the row indices corresponding to the elements supplied
        in the array `a`.

        On exit: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        On entry: if `store` = 1, the column indices corresponding to
        the elements supplied in the array `a`.

        If `store` = 2, `icol` need not be set.

        On exit: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    istc : int, shape (`n`+1), modified in place
        On entry: if `store` = 2, the starting address of each column,
        as supplied in the array `a`.

        If `store` = 1, `istc` need not be set.

        On exit: the starting address of each column, as returned in the
        array `a`. `istc`[`n`]-1 is the address of the last element in
        `a` plus one.

    store : int
        Indicates which storage format the matrix A is represented with
        on entry.

        `store` = 1

          A is represented in coordinate storage (CS) format using `a`,
          `irow` and `icol`.

        `store` = 2

          A is represented in compressed column storage (CCS) format
          using `a`, `irow` and `istc`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

          Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

          The relevant values in `a` are summed.

        `dup` = 'F'

          The function fails with `errno` = 12 on detecting a duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

          The entries are removed.

        `zer` = 'K'

          The entries are kept.

        `zer` = 'F'

          The function fails with `errno` = 13 on detecting a zero.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 4)
            On entry, `store` = *<value>*.

            Constraint: `store` = 1 or 2.

        (`errno` 5)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 6)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 7)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `m`.

        (`errno` 8)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: if `store` = 1, then 1 <= `icol`[i-1] <= `n`.

        (`errno` 9)
            On entry, `istc`[0] = *<value>*.

            Constraint: `istc`[0] = 1.

        (`errno` 10)
            On entry, i = *<value>*, `istc`[i-1] = *<value>* and
            `istc`[i] = *<value>*.

            Constraint: `istc`[i-1] <= `istc`[i], for i = 1,2,...,`n`

        (`errno` 11)
            On entry, `n` = *<value>*, `istc`[`n`] = *<value>* and `nnz`
            = *<value>*.

            Constraint: `istc`[`n`] = `nnz`+1.

        (`errno` 12)
            On entry, a duplicate entry has been found in row *<value>*
            and column *<value>*.

        (`errno` 13)
            On entry, a zero entry has been found in row *<value>* and
            column *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.real_rect_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, istc[n+1].
        utils._BYREF_INT, # Input, store.
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istc_cls = utils._EngineIntArrayType(
        istc, {'fun_name': fun_name, 'entity': 'istc'},
        exp_shape=(_n_py+1,),
    )
    _istc = _istc_cls.to_ctypes()
    _store = utils._EngineIntScalarType(
        store, {'fun_name': fun_name, 'entity': 'store'},
    ).to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _istc,
        _store,
        _dup,
        _zer,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, store = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: store = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= irow[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if store = 1, then 1 <= icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istc[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istc[0] = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, istc[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and istc[i] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istc[i-1] <= istc[i], for i = 1,2,...,n',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}, istc[n] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: istc[n] = nnz+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and column {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and column {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istc_cls.refresh_py(_istc)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def complex_gen_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Complex sparse non-Hermitian matrix reorder function.

    ``complex_gen_sort`` sorts the nonzero elements of a complex sparse
    non-Hermitian matrix, represented in coordinate storage format.

    For full information please refer to the NAG Library document for
    f11zn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11znf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : complex, shape (`nnz`), modified in place
        On entry: the nonzero elements of the matrix A. These may be in
        any order and there may be multiple nonzero elements with the
        same row and column indices.

        On exit: the nonzero elements ordered by increasing row index,
        and by increasing column index within each row. Each nonzero
        element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        On entry: the row indices corresponding to the elements supplied
        in the array `a`.

        On exit: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        On entry: the column indices corresponding to the elements
        supplied in the array `a`.

        On exit: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

          Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

          The relevant values in `a` are summed.

        `dup` = 'F'

          The function fails with `errno` = 3 on detecting a duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

          The entries are removed.

        `zer` = 'K'

          The entries are kept.

        `zer` = 'F'

          The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `icol`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `icol`[i-1] >= 1 and `icol`[i-1] <= `n`.

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row I and
            column J: I = *<value>*, J = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row I and column J:
            I = *<value>*, J = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_gen_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11znft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, icol[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[i-1] >= 1 and icol[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data

def complex_herm_sort(n, nnz, a, irow, icol, dup, zer, istr):
    u"""
    Complex sparse Hermitian matrix reorder function.

    ``complex_herm_sort`` sorts the nonzero elements of a sparse complex
    Hermitian matrix, represented in symmetric coordinate storage
    format.

    For full information please refer to the NAG Library document for
    f11zp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f11/f11zpf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nnz : int
        The number of elements supplied in the array `a`.

    a : complex, shape (`nnz`), modified in place
        On entry: the nonzero elements of the lower triangular part of
        the complex matrix A. These may be in any order and there may be
        multiple nonzero elements with the same row and column indices.

        On exit: the lower triangular nonzero elements ordered by
        increasing row index, and by increasing column index within each
        row. Each nonzero element has a unique row and column index.

    irow : int, shape (`nnz`), modified in place
        On entry: the row indices corresponding to the elements supplied
        in the array `a`.

        On exit: the first `nnz` elements contain the row indices
        corresponding to the elements returned in the array `a`.

    icol : int, shape (`nnz`), modified in place
        On entry: the column indices corresponding to the elements
        supplied in the array `a`.

        On exit: the first `nnz` elements contain the column indices
        corresponding to the elements returned in the array `a`.

    dup : str, length 1
        Indicates how elements in `a` with duplicate row and column
        indices are to be treated.

        `dup` = 'R'

          Duplicate entries are removed, only the first entry is kept.

        `dup` = 'S'

          The relevant values in `a` are summed.

        `dup` = 'F'

          The function fails with `errno` = 3 on detecting a duplicate.

    zer : str, length 1
        Indicates how elements in `a` with zero values are to be
        treated.

        `zer` = 'R'

          The entries are removed.

        `zer` = 'K'

          The entries are kept.

        `zer` = 'F'

          The function fails with `errno` = 4 on detecting a zero.

    istr : int, shape (`n`+1), modified in place
        On exit: `istr`[i-1]-1, for i = 1,2,...,`n`, is the starting
        address in the arrays `a`, `irow` and `icol` of row i of the
        matrix A. `istr`[`n`]-1 is the address of the last element in
        `a` plus one.

    Returns
    -------
    nnz : int
        The number of elements with unique row and column indices.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'R', 'K' or 'F'.

        (`errno` 1)
            On entry, `dup` = *<value>*.

            Constraint: `dup` = 'R', 'S' or 'F'.

        (`errno` 1)
            On entry, `nnz` = *<value>*.

            Constraint: `nnz` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, I = *<value>*, `icol`[I-1] = *<value>* and
            `irow`[I-1] = *<value>*.

            Constraint: `icol`[I-1] >= 1 and `icol`[I-1] <= `irow`[I-1].

        (`errno` 2)
            On entry, i = *<value>*, `irow`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `irow`[i-1] >= 1 and `irow`[i-1] <= `n`.

        (`errno` 3)
            On entry, a duplicate entry has been found in row I and
            column J: I = *<value>*, J = *<value>*.

        (`errno` 4)
            On entry, a zero entry has been found in row I and column J:
            I = *<value>*, J = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.sparse.complex_herm_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f11zpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input/Output, nnz.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, a[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, irow[nnz].
        utils._EngineIntArrayType.c_type, # Input/Output, icol[nnz].
        utils._BYREF_STRING, # Input, dup(len=1).
        utils._BYREF_STRING, # Input, zer(len=1).
        utils._EngineIntArrayType.c_type, # Output, istr[n+1].
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dup.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of zer.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nnz = utils._EngineIntScalarType(
        nnz, {'fun_name': fun_name, 'entity': 'nnz'},
    ).to_ctypes()
    _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
    _a_cls = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nnz_py,),
    )
    _a = _a_cls.to_ctypes()
    _irow_cls = utils._EngineIntArrayType(
        irow, {'fun_name': fun_name, 'entity': 'irow'},
        exp_shape=(_nnz_py,),
    )
    _irow = _irow_cls.to_ctypes()
    _icol_cls = utils._EngineIntArrayType(
        icol, {'fun_name': fun_name, 'entity': 'icol'},
        exp_shape=(_nnz_py,),
    )
    _icol = _icol_cls.to_ctypes()
    _dup = utils._EngineCharScalarType(
        dup, {'fun_name': fun_name, 'entity': 'dup'}, exp_chrlen=1,
    ).to_ctypes()
    _zer = utils._EngineCharScalarType(
        zer, {'fun_name': fun_name, 'entity': 'zer'}, exp_chrlen=1,
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _istr_cls = utils._EngineIntArrayType(
        istr, {'fun_name': fun_name, 'entity': 'istr'},
        exp_shape=(_n_py+1,),
        check_contents=False,
    )
    _istr = _istr_cls.to_ctypes(can_reference_contents=False)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nnz,
        _a,
        _irow,
        _icol,
        _dup,
        _zer,
        _istr,
        _iwork,
        _errbuf,
        _ierr,
        _dup._length_,
        _zer._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zer = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zer = \'R\', \'K\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dup = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dup = \'R\', \'S\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nnz = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nnz >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, icol[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and irow[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: icol[I-1] >= 1 and icol[I-1] <= irow[I-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, irow[i-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: irow[i-1] >= 1 and irow[i-1] <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a duplicate entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a zero entry has been found in row I and column J:',
                    'I = {:s}, J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nnz_py = utils._EngineIntScalarType.to_py(_nnz)
        _a_cls.refresh_py(_a)
        _irow_cls.refresh_py(_irow)
        _icol_cls.refresh_py(_icol)
        _istr_cls.refresh_py(_istr)
        return_data = _nnz_py
    _err_class.manage()
    return return_data
