# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `tsa` Chapter.

``tsa`` - Time Series Analysis

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module provides facilities for investigating and modelling the
statistical structure of series of observations collected at points in
time.
The models may then be used to forecast the series.

The module covers the following models and approaches:

- univariate time series analysis, including autocorrelation functions
  and autoregressive moving average (ARMA) models;

- univariate spectral analysis;

- transfer function (multi-input) modelling, in which one time series is
  dependent on other time series;

- bivariate spectral methods including coherency, gain and input
  response functions;

- vector ARMA models for multivariate time series;

- Kalman filter models (linear and nonlinear);

- GARCH models for volatility;

- inhomogeneous time series.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13intro.html
"""

# NAG Copyright 2017-2019.

from __future__ import absolute_import

import collections as _collections
import sys as _sys

from math import ceil

from ..base import utils

def uni_diff(x, nx, nd, nds, ns, xd):
    u"""
    Univariate time series, seasonal and non-seasonal differencing.

    ``uni_diff`` carries out non-seasonal and seasonal differencing on a
    time series.
    Information which allows the original series to be reconstituted
    from the differenced series is also produced.
    This information is required in time series forecasting.

    For full information please refer to the NAG Library document for
    g13aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13aaf.html

    Parameters
    ----------
    x : float, shape (`nx`)
        The undifferenced time series, x_i, for i = 1,2,...,n.

    nx : int
        n, the number of values in the undifferenced time series.

    nd : int
        d, the order of non-seasonal differencing.

    nds : int
        D, the order of seasonal differencing.

    ns : int
        s, the seasonality.

    xd : float, shape (`nx`), modified in place
        On exit: the differenced values in elements 0 to `nxd`-1, and
        reconstitution data in the remainder of the array.

    Returns
    -------
    nxd : int
        The number of differenced values in the array `xd`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nds` = *<value>*.

            Constraint: if `ns` = 0, `nds` <= 0.

        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 0.

        (`errno` 1)
            On entry, `nds` = *<value>*.

            Constraint: `nds` >= 0.

        (`errno` 1)
            On entry, `nd` = *<value>*.

            Constraint: `nd` >= 0.

        (`errno` 2)
            On entry, `nx` = *<value>*, `nd` = *<value>*, `nds` =
            *<value>* and `ns` = *<value>*.

            Constraint: `nx` > `nd`+(`nds`*`ns`).
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_diff'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, x[nx].
        utils._BYREF_INT, # Input, nx.
        utils._BYREF_INT, # Input, nd.
        utils._BYREF_INT, # Input, nds.
        utils._BYREF_INT, # Input, ns.
        utils._EngineFloat64ArrayType.c_type, # Output, xd[nx].
        utils._BYREF_INT, # Output, nxd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = utils._EngineIntScalarType.to_py(_nx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nx_py,),
    ).to_ctypes()
    _nd = utils._EngineIntScalarType(
        nd, {'fun_name': fun_name, 'entity': 'nd'},
    ).to_ctypes()
    _nds = utils._EngineIntScalarType(
        nds, {'fun_name': fun_name, 'entity': 'nds'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _xd_cls = utils._EngineFloat64ArrayType(
        xd, {'fun_name': fun_name, 'entity': 'xd'},
        exp_shape=(_nx_py,),
        check_contents=False,
    )
    _xd = _xd_cls.to_ctypes(can_reference_contents=False)
    _nxd = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _nx,
        _nd,
        _nds,
        _ns,
        _xd,
        _nxd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nds = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if ns = 0, nds <= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nds = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nds >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nd = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nd >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s}, nd = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nds = {:s} and ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: nx > nd+(nds*ns).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xd_cls.refresh_py(_xd)
        _nxd_py = utils._EngineIntScalarType.to_py(_nxd)
        return_data = _nxd_py
    _err_class.manage()
    return return_data

def uni_autocorr(x, nx, nk, r):
    u"""
    Univariate time series, sample autocorrelation function.

    ``uni_autocorr`` computes the sample autocorrelation function of a
    time series.
    It also computes the sample mean, the sample variance and a
    statistic which may be used to test the hypothesis that the true
    autocorrelation function is zero.

    For full information please refer to the NAG Library document for
    g13ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13abf.html

    Parameters
    ----------
    x : float, shape (`nx`)
        The time series, x_i, for i = 1,2,...,n.

    nx : int
        n, the number of values in the time series.

    nk : int
        K, the number of lags for which the autocorrelations are
        required. The lags range from 1 to K and do not include zero.

    r : float, shape (`nk`), modified in place
        On exit: the sample autocorrelation coefficient relating to lag
        k, for k = 1,2,...,K.

    Returns
    -------
    xm : float
        The sample mean of the input time series.

    xv : float
        The sample variance of the input time series.

    stat : float
        The statistic used to test the hypothesis that the true
        autocorrelation function of the time series is identically zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nx` = *<value>* and `nk` = *<value>*.

            Constraint: `nx` > `nk`.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` > 1.

        (`errno` 1)
            On entry, `nk` = *<value>*.

            Constraint: `nk` > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, all values of `x` are practically identical.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_autocorr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, x[nx].
        utils._BYREF_INT, # Input, nx.
        utils._BYREF_INT, # Input, nk.
        utils._BYREF_FLOAT64, # Output, xm.
        utils._BYREF_FLOAT64, # Output, xv.
        utils._EngineFloat64ArrayType.c_type, # Output, r[nk].
        utils._BYREF_FLOAT64, # Output, stat.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = utils._EngineIntScalarType.to_py(_nx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nx_py,),
    ).to_ctypes()
    _nk = utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _xm = utils._EngineFloat64ScalarType.empty_ctypes()
    _xv = utils._EngineFloat64ScalarType.empty_ctypes()
    _nk_py = utils._EngineIntScalarType.to_py(_nk)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nk_py,),
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _stat = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _nx,
        _nk,
        _xm,
        _xv,
        _r,
        _stat,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 32 and _ifmt == 99997:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s} and nk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nx > nk.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 99998:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nx > 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 99999:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nk > 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 99996:
            _ierr_fl = 2
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all values of x are practically identical.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _r_cls.refresh_py(_r)
        _xm_py = utils._EngineFloat64ScalarType.to_py(_xm)
        _xv_py = utils._EngineFloat64ScalarType.to_py(_xv)
        _stat_py = utils._EngineFloat64ScalarType.to_py(_stat)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['xm', 'xv', 'stat'],
        )(
            _xm_py,
            _xv_py,
            _stat_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_autocorr_part(r, nk, nl, p, v, ar):
    u"""
    Univariate time series, partial autocorrelations from
    autocorrelations.

    ``uni_autocorr_part`` calculates partial autocorrelation
    coefficients given a set of autocorrelation coefficients.
    It also calculates the predictor error variance ratios for
    increasing order of finite lag autoregressive predictor, and the
    autoregressive parameters associated with the predictor of maximum
    order.

    For full information please refer to the NAG Library document for
    g13ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13acf.html

    Parameters
    ----------
    r : float, shape (`nk`)
        The autocorrelation coefficient relating to lag k, for k =
        1,2,...,K.

    nk : int
        K, the number of lags. The lags range from 1 to K and do not
        include zero.

    nl : int
        L, the number of partial autocorrelation coefficients required.

    p : float, shape (`nl`), modified in place
        On exit: `p`[l-1] contains the partial autocorrelation
        coefficient at lag l, p_{l,l}, for l = 1,2,...,`nvl`.

    v : float, shape (`nl`), modified in place
        On exit: `v`[l-1] contains the predictor error variance ratio
        v_l, for l = 1,2,...,`nvl`.

    ar : float, shape (`nl`), modified in place
        On exit: the autoregressive parameters of maximum order, i.e.,
        p_{Lj} if no exception or warning is raised, or p_{l_0-1,j} if
        `errno` = 3, for j = 1,2,...,`nvl`.

    Returns
    -------
    nvl : int
        The number of valid values in each of `p`, `v` and `ar`. Thus in
        the case of premature termination at iteration l_0 (see
        [Description]), `nvl` is returned as l_0-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nk` = *<value>* and `nl` = *<value>*.

            Constraint: `nk` >= `nl`.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` > 0.

        (`errno` 1)
            On entry, `nk` = *<value>*.

            Constraint: `nk` > 0.

        (`errno` 2)
            On entry, the autocorrelation coefficient of lag 1 has
            absolute value greater than or equal to 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The autocorrelation coefficients do not form a positive
            definite sequence.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_autocorr_part'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, r[nk].
        utils._BYREF_INT, # Input, nk.
        utils._BYREF_INT, # Input, nl.
        utils._EngineFloat64ArrayType.c_type, # Output, p[nl].
        utils._EngineFloat64ArrayType.c_type, # Output, v[nl].
        utils._EngineFloat64ArrayType.c_type, # Output, ar[nl].
        utils._BYREF_INT, # Output, nvl.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nk = utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _nk_py = utils._EngineIntScalarType.to_py(_nk)
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nk_py,),
    ).to_ctypes()
    _nl = utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = utils._EngineIntScalarType.to_py(_nl)
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_nl_py,),
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_nl_py,),
        check_contents=False,
    )
    _v = _v_cls.to_ctypes(can_reference_contents=False)
    _ar_cls = utils._EngineFloat64ArrayType(
        ar, {'fun_name': fun_name, 'entity': 'ar'},
        exp_shape=(_nl_py,),
        check_contents=False,
    )
    _ar = _ar_cls.to_ctypes(can_reference_contents=False)
    _nvl = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _r,
        _nk,
        _nl,
        _p,
        _v,
        _ar,
        _nvl,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nk = {:s} and nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nk >= nl.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nl > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nk > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the autocorrelation coefficient of lag 1',
                    'has absolute value greater than or equal to 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The autocorrelation coefficients do not form a',
                    'positive definite sequence.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p_cls.refresh_py(_p)
        _v_cls.refresh_py(_v)
        _ar_cls.refresh_py(_ar)
        _nvl_py = utils._EngineIntScalarType.to_py(_nvl)
        return_data = _nvl_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nvl'],
            )(
                _nvl_py,
            )
    _err_class.manage()
    return return_data

def uni_arima_prelim(mr, r, nk, xv, npar, par, isf):
    u"""
    Univariate time series, preliminary estimation, seasonal ARIMA
    model.

    ``uni_arima_prelim`` calculates preliminary estimates of the
    parameters of an autoregressive integrated moving average (ARIMA)
    model from the autocorrelation function of the appropriately
    differenced times series.

    For full information please refer to the NAG Library document for
    g13ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13adf.html

    Parameters
    ----------
    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    r : float, shape (`nk`)
        The autocorrelations (starting at lag 1), which must have been
        calculated after the time series has been appropriately
        differenced.

    nk : int
        The maximum lag of the autocorrelations in array `r`.

    xv : float
        The series sample variance, calculated after appropriate
        differencing has been applied to the series.

    npar : int
        The exact number of parameters specified in the model by array
        `mr`.

    par : float, shape (`npar`), modified in place
        On exit: the first `npar` elements of `par` contain the
        preliminary estimates of the ARIMA model parameters, in standard
        order.

    isf : int, shape (4), modified in place
        On exit: contains success/failure indicators, one for each of
        the four types of parameter (autoregressive, moving average,
        seasonal autoregressive, seasonal moving average).

        The indicator has the interpretation:

        [table omitted]

    Returns
    -------
    rv : float
        An estimate of the residual variance of the preliminarily
        estimated model.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 2)
            On entry, `nk` = *<value>*.

            Constraint: `nk` >= max(p+q, s*(P+Q)).

        (`errno` 3)
            On entry, `r`[*<value>*] = *<value>*.

            Constraint: -1.0 <= `r`[i] <= 1.0, for i = 1,2,...`nk`.

        (`errno` 4)
            On entry, `xv` = *<value>*.

            Constraint: `xv` > 0.0.

        (`errno` 5)
            On entry, `npar` = *<value>*.

            Constraint: `npar` = p+q+P+Q.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arima_prelim'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._EngineFloat64ArrayType.c_type, # Input, r[nk].
        utils._BYREF_INT, # Input, nk.
        utils._BYREF_FLOAT64, # Input, xv.
        utils._BYREF_INT, # Input, npar.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wa.
        utils._BYREF_INT, # Input, nwa.
        utils._EngineFloat64ArrayType.c_type, # Output, par[npar].
        utils._BYREF_FLOAT64, # Output, rv.
        utils._EngineIntArrayType.c_type, # Output, isf[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _nk = utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _nk_py = utils._EngineIntScalarType.to_py(_nk)
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nk_py,),
    ).to_ctypes()
    _xv = utils._EngineFloat64ScalarType(
        xv, {'fun_name': fun_name, 'entity': 'xv'},
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _wa = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _nwa = utils._EngineIntScalarType.c_type(0)
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par_cls = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _par = _par_cls.to_ctypes(can_reference_contents=False)
    _rv = utils._EngineFloat64ScalarType.empty_ctypes()
    _isf_cls = utils._EngineIntArrayType(
        isf, {'fun_name': fun_name, 'entity': 'isf'},
        exp_shape=(4,),
        check_contents=False,
    )
    _isf = _isf_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _mr,
        _r,
        _nk,
        _xv,
        _npar,
        _wa,
        _nwa,
        _par,
        _rv,
        _isf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nk >= max(p+q, s*(P+Q)).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: -1.0 <= r[i] <= 1.0, for i = 1,2,...nk.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xv = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: xv > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar = p+q+P+Q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Satisfactory parameter estimates could not be obtained',
                    'for all parameter types in the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _par_cls.refresh_py(_par)
        _isf_cls.refresh_py(_isf)
        _rv_py = utils._EngineFloat64ScalarType.to_py(_rv)
        return_data = _rv_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['rv'],
            )(
                _rv_py,
            )
    _err_class.manage()
    return return_data

def uni_arima_estim(mr, par, npar, c, kfc, x, nx, icount, ex, exr, al, iex, g, igh, sd, h, st, ist, piv, kpiv, nit, zsp, kzsp, isf, data=None, io_manager=None):
    u"""
    Univariate time series, estimation, seasonal ARIMA model
    (comprehensive).

    ``uni_arima_estim`` fits a seasonal autoregressive integrated moving
    average (ARIMA) model to an observed time series, using a nonlinear
    least squares procedure incorporating backforecasting. Parameter
    estimates are obtained, together with appropriate standard errors.
    The residual series is returned, and information for use in
    forecasting the time series is produced for use by the functions
    ``uni_arima_update`` and ``uni_arima_forecast_state``.

    The estimation procedure is iterative, starting with initial
    parameter values such as may be obtained using ``uni_arima_prelim``.
    It continues until a specified convergence criterion is satisfied,
    or until a specified number of iterations has been carried out.
    The progress of the procedure can be monitored by means of a
    user-supplied function.

    For full information please refer to the NAG Library document for
    g13ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13aef.html

    Parameters
    ----------
    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    par : float, shape (`npar`), modified in place
        On entry: the initial estimates of the p values of the phi
        parameters, the q values of the theta parameters, the P values
        of the Phi parameters and the Q values of the Theta parameters,
        in that order.

        On exit: the latest values of the estimates of these parameters.

    npar : int
        The total number of phi, theta, Phi and Theta parameters to be
        estimated.

    c : float
        If `kfc` = 0, `c` must contain the expected value, c, of the
        differenced series.

        If `kfc` = 1, `c` must contain an initial estimate of c.

    kfc : int
        Must be set to 1 if the constant, c, is to be estimated and 0 if
        it is to be held fixed at its initial value.

    x : float, shape (`nx`)
        The n values of the original undifferenced time series.

    nx : int
        n, the length of the original undifferenced time series.

    icount : int, shape (6), modified in place
        On exit: size of various output arrays.

        `icount`[0]

          Contains q+(Q*s), the number of backforecasts.

        `icount`[1]

          Contains n-d-(D*s), the number of differenced values.

        `icount`[2]

          Contains d+(D*s), the number of values of reconstitution
          information.

        `icount`[3]

          Contains n+q+(Q*s), the number of values held in each of the
          series `ex`, `exr` and `al`.

        `icount`[4]

          Contains n-d-(D*s)-p-q-P-Q-`kfc`, the number of degrees of
          freedom associated with S.

        `icount`[5]

          Contains `icount`[0]+`npar`+`kfc`, the number of parameters
          being estimated.

        These values are always computed regardless of the exit value of
        errno.

    ex : float, shape (`iex`), modified in place
        On exit: the extended differenced series which is made up of:

        `icount`[0] backforecast values of the differenced series.

        `icount`[1] actual values of the differenced series.

        `icount`[2] values of reconstitution information.

        The total number of these values held in `ex` is `icount`[3].

        If the function exits because of a faulty input parameter, the
        contents of `ex` will be indeterminate.

    exr : float, shape (`iex`), modified in place
        On exit: the values of the model residuals which is made up of:

        `icount`[0] residuals corresponding to the backforecasts in the
        differenced series.

        `icount`[1] residuals corresponding to the actual values in the
        differenced series.

        The remaining `icount`[2] values contain zeros.

        If the function exits with errno holding a value other than 0 or
        9, the contents of `exr` will be indeterminate.

    al : float, shape (`iex`), modified in place
        On exit: the intermediate series which is made up of:

        `icount`[0] intermediate series values corresponding to the
        backforecasts in the differenced series.

        `icount`[1] intermediate series values corresponding to the
        actual values in the differenced series.

        The remaining `icount`[2] values contain zeros.

        If an exception is raised, the contents of `al` will be
        indeterminate.

    iex : int
        The dimension of the arrays `ex`, `exr` and `al`.

    g : float, shape (`igh`), modified in place
        On exit: the latest value of the derivatives of S with respect
        to each of the parameters being estimated (backforecasts, `par`
        parameters, and where relevant the constant -- in that order).
        The contents of `g` will be indeterminate if the function exits
        with a faulty input parameter.

    igh : int
        The dimension of the arrays `g` and `sd`.

        The second dimension of the arrays `h` and `hc`.

    sd : float, shape (`igh`), modified in place
        On exit: the standard deviations corresponding to each of the
        parameters being estimated (backforecasts, `par` parameters, and
        where relevant the constant, in that order).

        If the function exits with errno containing a value other than 0
        or 9, or if the required number of iterations is zero, the
        contents of `sd` will be indeterminate.

    h : float, shape (1+q+(Q*s)+`npar`+`kfc`, `igh`), modified in place
        On exit: the second derivative of S and correlation
        coefficients.

        - the latest values of an approximation to the second derivative
          of S with respect to each of the (q+Q*s+`npar`+`kfc`)
          parameters being estimated (backforecasts, `par` parameters,
          and where relevant the constant -- in that order), and

        - the correlation coefficients relating to each pair of these
          parameters.

        These are held in a matrix defined by the first
        (q+Q*s+`npar`+`kfc`) rows and the first (q+Q*s+`npar`+`kfc`)
        columns of `h`. (Note that `icount`[5] contains the value of
        this expression.) The values of (a) are contained in the upper
        triangle, and the values of (b) in the strictly lower triangle.

        These correlation coefficients are zero during intermediate
        printout using `piv`, and indeterminate if errno contains on
        exit a value other than 0 or 9.

        All the contents of `h` are indeterminate if the required number
        of iterations are zero.

        The (q+(Q*s)+`npar`+`kfc`+1)th row of `h` is used internally as
        workspace.

    st : float, shape (`ist`), modified in place
        On exit: the `nst` values of the state set array. If the
        function exits with errno containing a value other than 0 or 9,
        the contents of `st` will be indeterminate.

    ist : int
        The dimension of the array `st`.

    piv : None or callable piv(mr, par, npar, c, kfc, icount, s, g, h,
        igh, itc, zsp, data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `piv` is used to monitor the progress of the optimization.

        Parameters
        ~~~~~~~~~~
        mr : int, ndarray, shape (7)
            `mr`, defined as for ``uni_arima_estim``.

        par : float, ndarray, shape (`npar`)
            `par`, defined as for ``uni_arima_estim``.

        npar : int
            `npar`, defined as for ``uni_arima_estim``.

        c : float
            `c`, defined as for ``uni_arima_estim``.

        kfc : int
            `kfc`, defined as for ``uni_arima_estim``.

        icount : int, ndarray, shape (6)
            `icount`, defined as for ``uni_arima_estim``.

        s : float
            `s`, defined as for ``uni_arima_estim``.

        g : float, ndarray, shape (`igh`)
            `g`, defined as for ``uni_arima_estim``.

        h : float, ndarray, shape (:, `igh`)
            `h`, defined as for ``uni_arima_estim``.

        igh : int
            `igh`, defined as for ``uni_arima_estim``.

        itc : int
            `itc`, defined as for ``uni_arima_estim``.

        zsp : float, ndarray, shape (4)
            `zsp`, defined as for ``uni_arima_estim``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    kpiv : int
        Must be nonzero if the progress of the optimization is to be
        monitored using `piv`. Otherwise `kpiv` must contain 0.

    nit : int
        The maximum number of iterations to be performed.

    zsp : float, shape (4), modified in place
        On entry: when `kzsp` = 1, the first four elements of `zsp` must
        contain the four values used to guide the search procedure.
        These are as follows.

        `zsp`[0] contains alpha, the value used to constrain the
        magnitude of the search procedure steps.

        `zsp`[1] contains beta, the multiplier which regulates the value
        alpha.

        `zsp`[2] contains delta, the value of the stationarity and
        invertibility test tolerance factor.

        `zsp`[3] contains gamma, the value of the convergence criterion.

        If `kzsp` != 1 on entry, default values for `zsp` are supplied
        by the function.

        These are 0.001, 10.0, 1000.0 and max(100*machine
        precision,0.0000001) respectively.

        On exit: `zsp` contains the values, default or otherwise, used
        by the function.

    kzsp : int
        The value 1 if the function is to use the input values of `zsp`,
        and any other value if the default values of `zsp` are to be
        used.

    isf : int, shape (4), modified in place
        On exit: contains success/failure indicators, one for each of
        the four types of parameter in the model (autoregressive, moving
        average, seasonal autoregressive, seasonal moving average), in
        that order.

        Each indicator has the interpretation:

        [table omitted]

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    c : float
        If `kfc` = 0, `c` is unchanged.

        If `kfc` = 1, `c` contains the latest estimate of c.

        Therefore, if `c` and `kfc` are both zero on entry, there is no
        constant correction.

    s : float
        The residual sum of squares after the latest series of parameter
        estimates has been incorporated into the model. If the function
        exits with a faulty input parameter, `s` contains zero.

    nst : int
        The number of values in the state set array `st`.

    itc : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 3)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 3)
            On entry, `zsp`[3] = *<value>*.

            Constraint: 0.0 <= `zsp`[3] < 1.0.

        (`errno` 3)
            On entry, `zsp`[2] = *<value>*.

            Constraint: `zsp`[2] >= 1.0.

        (`errno` 3)
            On entry, `zsp`[1] = *<value>*.

            Constraint: `zsp`[1] > 1.0.

        (`errno` 3)
            On entry, `zsp`[0] = *<value>*.

            Constraint: `zsp`[0] > 0.0.

        (`errno` 4)
            On entry, `ist` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ist` >= (P*s)+d+(D*s)+q+max(p, Q*s).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            On entry, ldh = *<value>*.

            Constraint: ldh >= q*(Q*s)+`npar`+`kfc`.

        (`errno` 6)
            On entry, `igh` = *<value>*.

            Constraint: `igh` >= q*(Q*s)+`npar`+`kfc`.

        (`errno` 6)
            On entry, `iex` = *<value>*.

            Constraint: `iex` >= q*(Q*s)+`nx`.

        (`errno` 7)
            A failure in the search procedure has occurred.

        (`errno` 8)
            Failure to invert H.

        (`errno` 9)
            Unable to calculate the latest estimates of the
            backforecasts.

        (`errno` 10)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arima_estim'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    pivh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_INT, # Input, kfc.
        utils._EngineIntArrayType.c_type, # Input, icount[6].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Input, g[igh].
        utils._EngineFloat64ArrayType.c_type, # Input, h[ldh, igh].
        utils._BYREF_INT, # Input, ldh.
        utils._BYREF_INT, # Input, igh.
        utils._BYREF_INT, # Input, itc.
        utils._EngineFloat64ArrayType.c_type, # Input, zsp[4].
        utils._BYREF_INT, # Input, nadv.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    if piv is None:
        piv_auxil_ctypes = utils._EngineState._get_symbol('g13afzt')
        piv_auxil_ctypes.restype = None
        piv_auxil_ctypes.argtypes = (
            utils._BYREF_ENGINEDATA,
            utils._PRINT_REC_FUNCTYPE,
            utils._PRINT_RECH_FUNCTYPE,
            utils._EngineIntArrayType.c_type, # Input, mr[7].
            utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
            utils._BYREF_INT, # Input, npar.
            utils._BYREF_FLOAT64, # Input, c.
            utils._BYREF_INT, # Input, kfc.
            utils._EngineIntArrayType.c_type, # Input, icount[6].
            utils._BYREF_FLOAT64, # Input, s.
            utils._EngineFloat64ArrayType.c_type, # Input, g[igh].
            utils._EngineFloat64ArrayType.c_type, # Input, h[ldh, igh].
            utils._BYREF_INT, # Input, ldh.
            utils._BYREF_INT, # Input, igh.
            utils._BYREF_INT, # Input, itc.
            utils._EngineFloat64ArrayType.c_type, # Input, zsp[4].
            utils._BYREF_INT, # Input, nadv.
            utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
            utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        )

    def cb_pivh(_piv, en_data, print_rec, print_rech, mr, par, npar, c, kfc, icount, s, g, h, ldh, igh, itc, zsp, nadv, iuser, ruser):
        """Helper for piv."""
        try:
            if piv is None:
                piv_auxil_ctypes(
                    en_data,
                    print_rec,
                    print_rech,
                    mr,
                    par,
                    npar,
                    c,
                    kfc,
                    icount,
                    s,
                    g,
                    h,
                    ldh,
                    igh,
                    itc,
                    zsp,
                    nadv,
                    iuser,
                    ruser,
                )
                return
            _mr_pivh = utils._EngineIntArrayType.from_pointer(
                mr,
                (7,),
            )
            _npar_pivh = utils._EngineIntScalarType.to_py(npar.contents)
            _par_pivh = utils._EngineFloat64ArrayType.from_pointer(
                par,
                (_npar_pivh,),
            )
            _c_pivh = utils._EngineFloat64ScalarType.to_py(c.contents)
            _kfc_pivh = utils._EngineIntScalarType.to_py(kfc.contents)
            _icount_pivh = utils._EngineIntArrayType.from_pointer(
                icount,
                (6,),
            )
            _s_pivh = utils._EngineFloat64ScalarType.to_py(s.contents)
            _igh_pivh = utils._EngineIntScalarType.to_py(igh.contents)
            _g_pivh = utils._EngineFloat64ArrayType.from_pointer(
                g,
                (_igh_pivh,),
            )
            _ldh_pivh = utils._EngineIntScalarType.to_py(ldh.contents)
            _h_pivh = utils._EngineFloat64ArrayType.from_pointer(
                h,
                (_ldh_pivh, _igh_pivh),
                sorder=_m_sorder,
            )
            _itc_pivh = utils._EngineIntScalarType.to_py(itc.contents)
            _zsp_pivh = utils._EngineFloat64ArrayType.from_pointer(
                zsp,
                (4,),
            )
            if data is not None:
                piv(
                    _mr_pivh,
                    _par_pivh,
                    _npar_pivh,
                    _c_pivh,
                    _kfc_pivh,
                    _icount_pivh,
                    _s_pivh,
                    _g_pivh,
                    _h_pivh,
                    _igh_pivh,
                    _itc_pivh,
                    _zsp_pivh,
                    data,
                )
            else:
                piv(
                    _mr_pivh,
                    _par_pivh,
                    _npar_pivh,
                    _c_pivh,
                    _kfc_pivh,
                    _icount_pivh,
                    _s_pivh,
                    _g_pivh,
                    _h_pivh,
                    _igh_pivh,
                    _itc_pivh,
                    _zsp_pivh,
                )
            utils._unlink_pointer(mr)
            utils._unlink_pointer(par)
            utils._unlink_pointer(icount)
            utils._unlink_pointer(g)
            utils._unlink_pointer(h)
            utils._unlink_pointer(zsp)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g13aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input/Output, c.
        utils._BYREF_INT, # Input, kfc.
        utils._EngineFloat64ArrayType.c_type, # Input, x[nx].
        utils._BYREF_INT, # Input, nx.
        utils._EngineIntArrayType.c_type, # Output, icount[6].
        utils._EngineFloat64ArrayType.c_type, # Output, ex[iex].
        utils._EngineFloat64ArrayType.c_type, # Output, exr[iex].
        utils._EngineFloat64ArrayType.c_type, # Output, al[iex].
        utils._BYREF_INT, # Input, iex.
        utils._BYREF_FLOAT64, # Output, s.
        utils._EngineFloat64ArrayType.c_type, # Output, g[igh].
        utils._BYREF_INT, # Input, igh.
        utils._EngineFloat64ArrayType.c_type, # Output, sd[igh].
        utils._EngineFloat64ArrayType.c_type, # Output, h[ldh, igh].
        utils._BYREF_INT, # Input, ldh.
        utils._EngineFloat64ArrayType.c_type, # Output, st[ist].
        utils._BYREF_INT, # Input, ist.
        utils._BYREF_INT, # Output, nst.
        utils._UNUSED_FUNCTYPE,
        pivh,
        utils._BYREF_INT, # Input, kpiv.
        utils._BYREF_INT, # Input, nit.
        utils._BYREF_INT, # Output, itc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, zsp[4].
        utils._BYREF_INT, # Input, kzsp.
        utils._EngineIntArrayType.c_type, # Output, isf[4].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wa.
        utils._BYREF_INT, # Input, iwa.
        utils._EngineFloat64ArrayType.c_type, # Workspace, hc.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Input, nadv.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((h, 'h'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par_cls = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    )
    _par = _par_cls.to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _kfc = utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = utils._EngineIntScalarType.to_py(_nx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nx_py,),
    ).to_ctypes()
    _icount_cls = utils._EngineIntArrayType(
        icount, {'fun_name': fun_name, 'entity': 'icount'},
        exp_shape=(6,),
        check_contents=False,
    )
    _icount = _icount_cls.to_ctypes(can_reference_contents=False)
    _iex = utils._EngineIntScalarType(
        iex, {'fun_name': fun_name, 'entity': 'iex'},
    ).to_ctypes()
    _iex_py = utils._EngineIntScalarType.to_py(_iex)
    _ex_cls = utils._EngineFloat64ArrayType(
        ex, {'fun_name': fun_name, 'entity': 'ex'},
        exp_shape=(_iex_py,),
        check_contents=False,
    )
    _ex = _ex_cls.to_ctypes(can_reference_contents=False)
    _exr_cls = utils._EngineFloat64ArrayType(
        exr, {'fun_name': fun_name, 'entity': 'exr'},
        exp_shape=(_iex_py,),
        check_contents=False,
    )
    _exr = _exr_cls.to_ctypes(can_reference_contents=False)
    _al_cls = utils._EngineFloat64ArrayType(
        al, {'fun_name': fun_name, 'entity': 'al'},
        exp_shape=(_iex_py,),
        check_contents=False,
    )
    _al = _al_cls.to_ctypes(can_reference_contents=False)
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    _igh = utils._EngineIntScalarType(
        igh, {'fun_name': fun_name, 'entity': 'igh'},
    ).to_ctypes()
    _igh_py = utils._EngineIntScalarType.to_py(_igh)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_igh_py,),
        check_contents=False,
    )
    _g = _g_cls.to_ctypes(can_reference_contents=False)
    _sd_cls = utils._EngineFloat64ArrayType(
        sd, {'fun_name': fun_name, 'entity': 'sd'},
        exp_shape=(_igh_py,),
        check_contents=False,
    )
    _sd = _sd_cls.to_ctypes(can_reference_contents=False)
    _ldh_py = utils._EngineFloat64ArrayType.get_nd(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldh = utils._EngineIntScalarType(
        _ldh_py, {'fun_name': fun_name, 'entity': '_ldh'},
    ).to_ctypes()
    _kfc_py = utils._EngineIntScalarType.to_py(_kfc)
    _vl_q_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vu_q_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _vl_s_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(1+_vl_q_py+(_vu_q_py*_vl_s_py)+_npar_py+_kfc_py, _igh_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _h = _h_cls.to_ctypes(can_reference_contents=False)
    _ist = utils._EngineIntScalarType(
        ist, {'fun_name': fun_name, 'entity': 'ist'},
    ).to_ctypes()
    _ist_py = utils._EngineIntScalarType.to_py(_ist)
    _st_cls = utils._EngineFloat64ArrayType(
        st, {'fun_name': fun_name, 'entity': 'st'},
        exp_shape=(_ist_py,),
        check_contents=False,
    )
    _st = _st_cls.to_ctypes(can_reference_contents=False)
    _nst = utils._EngineIntScalarType.empty_ctypes()
    utils._check_callable(piv, {'fun_name': fun_name, 'entity': 'piv'}, allow_none=True)
    _piv = utils._UNUSED_CALLBACK
    _pivh = utils._cast_callback(
        pivh, cb_pivh,
        {'fun_name': fun_name, 'entity': 'piv'},
    )
    _kpiv = utils._EngineIntScalarType(
        kpiv, {'fun_name': fun_name, 'entity': 'kpiv'},
    ).to_ctypes()
    _nit = utils._EngineIntScalarType(
        nit, {'fun_name': fun_name, 'entity': 'nit'},
    ).to_ctypes()
    _itc = utils._EngineIntScalarType.empty_ctypes()
    _zsp_cls = utils._EngineFloat64ArrayType(
        zsp, {'fun_name': fun_name, 'entity': 'zsp'},
        exp_shape=(4,),
    )
    _zsp = _zsp_cls.to_ctypes()
    _kzsp = utils._EngineIntScalarType(
        kzsp, {'fun_name': fun_name, 'entity': 'kzsp'},
    ).to_ctypes()
    _isf_cls = utils._EngineIntArrayType(
        isf, {'fun_name': fun_name, 'entity': 'isf'},
        exp_shape=(4,),
        check_contents=False,
    )
    _isf = _isf_cls.to_ctypes(can_reference_contents=False)
    _wa = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iwa = utils._EngineIntScalarType.c_type(0)
    _hc = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _mr,
        _par,
        _npar,
        _c,
        _kfc,
        _x,
        _nx,
        _icount,
        _ex,
        _exr,
        _al,
        _iex,
        _s,
        _g,
        _igh,
        _sd,
        _h,
        _ldh,
        _st,
        _ist,
        _nst,
        _piv,
        _pivh,
        _kpiv,
        _nit,
        _itc,
        _zsp,
        _kzsp,
        _isf,
        _wa,
        _iwa,
        _hc,
        _iuser,
        _ruser,
        _nadv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kfc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kfc = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar = p+q+P+Q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model is over-parameterised.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[3] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= zsp[3] < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zsp[2] >= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zsp[1] > 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zsp[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ist = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ist >= (P*s)+d+(D*s)+q+max(p, Q*s).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldh = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ldh >= q*(Q*s)+npar+kfc.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99987:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, igh = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: igh >= q*(Q*s)+npar+kfc.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iex = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iex >= q*(Q*s)+nx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A failure in the search procedure has occurred.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99981:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to invert H.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the latest estimates of the backforecasts.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Satisfactory parameter estimates could not be obtained',
                    'for all parameter types in the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _par_cls.refresh_py(_par)
        _c_py = utils._EngineFloat64ScalarType.to_py(_c)
        _icount_cls.refresh_py(_icount)
        _ex_cls.refresh_py(_ex)
        _exr_cls.refresh_py(_exr)
        _al_cls.refresh_py(_al)
        _g_cls.refresh_py(_g)
        _sd_cls.refresh_py(_sd)
        _h_cls.refresh_py(_h)
        _st_cls.refresh_py(_st)
        _zsp_cls.refresh_py(_zsp)
        _isf_cls.refresh_py(_isf)
        _s_py = utils._EngineFloat64ScalarType.to_py(_s)
        _nst_py = utils._EngineIntScalarType.to_py(_nst)
        _itc_py = utils._EngineIntScalarType.to_py(_itc)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['c', 's', 'nst', 'itc'],
        )(
            _c_py,
            _s_py,
            _nst_py,
            _itc_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_arima_estim_easy(mr, par, npar, c, kfc, x, nx, sd, nppc, cm, st, kpiv, nit, isf, res, ires, io_manager=None):
    u"""
    Univariate time series, estimation, seasonal ARIMA model
    (easy-to-use).

    ``uni_arima_estim_easy`` is an easy-to-use version of
    ``uni_arima_estim``.
    It fits a seasonal autoregressive integrated moving average (ARIMA)
    model to an observed time series, using a nonlinear least squares
    procedure incorporating backforecasting. Parameter estimates are
    obtained, together with appropriate standard errors.
    The residual series is returned, and information for use in
    forecasting the time series is produced for use in
    ``uni_arima_update`` and ``uni_arima_forecast_state``.

    The estimation procedure is iterative, starting with initial
    parameter values such as may be obtained using ``uni_arima_prelim``.
    It continues until a specified convergence criterion is satisfied or
    until a specified number of iterations have been carried out.
    The progress of the iteration can be monitored by means of an
    optional printing facility.

    For full information please refer to the NAG Library document for
    g13af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13aff.html

    Parameters
    ----------
    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    par : float, shape (`npar`), modified in place
        On entry: the initial estimates of the p values of the phi
        parameters, the q values of the theta parameters, the P values
        of the Phi parameters and the Q values of the Theta parameters,
        in that order.

        On exit: contains the latest values of the estimates of these
        parameters.

    npar : int
        The total number of phi, theta, Phi and Theta parameters to be
        estimated.

    c : float
        If `kfc` = 0, `c` must contain the expected value, c, of the
        differenced series.

        If `kfc` = 1, `c` must contain an initial estimate of c.

        Therefore, if `c` and `kfc` are both zero on entry, there is no
        constant correction.

    kfc : int
        Must be set to 1 if the constant, c, is to be estimated and 0 if
        it is to be held fixed at its initial value.

    x : float, shape (`nx`)
        The n values of the original undifferenced time series.

    nx : int
        n, the length of the original undifferenced time series.

    sd : float, shape (`nppc`), modified in place
        On exit: the standard deviations corresponding to the parameters
        in the model (p autoregressive, q moving average, P seasonal
        autoregressive, Q seasonal moving average and c, if estimated,
        in that order). If the function exits with errno containing a
        value other than 0 or 9, or if the required number of iterations
        is zero, the contents of `sd` will be indeterminate.

    nppc : int
        The number of phi, theta, Phi, Theta and c parameters to be
        estimated. `nppc` = p+q+P+Q+1 if the constant is being estimated
        and `nppc` = p+q+P+Q if not.

    cm : float, shape (`nppc`, `nppc`), modified in place
        On exit: the correlation coefficients associated with each pair
        of the `nppc` parameters. These are held in the first `nppc`
        rows and the first `nppc` columns of `cm`. These correlation
        coefficients are indeterminate if errno contains on exit a value
        other than 0 or 9, or if the required number of iterations is
        zero.

    st : float, shape (`nx`), modified in place
        On exit: the value of the state set in its first `nst` elements.
        If the function exits with errno containing a value other than 0
        or 9, the contents of `st` will be indeterminate.

    kpiv : int
        Must be nonzero if the progress of the optimization is to be
        monitored using the built-in printing facility. Otherwise `kpiv`
        must contain zero. If selected, monitoring output will be sent
        to the file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`). For each
        iteration, the heading

        ::

          G13AFZ MONITORING OUTPUT - ITERATION n

        followed by the argument values, and residual sum of squares,
        are printed.

    nit : int
        The maximum number of iterations to be performed.

    isf : int, shape (4), modified in place
        On exit: the first four elements of `isf` contain
        success/failure indicators, one for each of the four types of
        parameter in the model (autoregressive, moving average, seasonal
        autoregressive, seasonal moving average), in that order.

        Each indicator has the interpretation:

        [table omitted]

    res : float, shape (`ires`), modified in place
        On exit: the first `nres` elements of `res` contain the model
        residuals derived from the differenced series. If the function
        exits with errno holding a value other than 0 or 9, these
        elements of `res` will be indeterminate. The rest of the array
        `res` is used as workspace.

    ires : int
        The dimension of the array `res`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    c : float
        If `kfc` = 0, `c` is unchanged.

        If `kfc` = 1, `c` contains the latest estimate of c.

    s : float
        The residual sum of squares after the latest series of parameter
        estimates has been incorporated into the model. If the function
        exits with a faulty input parameter, `s` contains zero.

    ndf : int
        The number of degrees of freedom associated with `s`,

        `ndf` = n-d-D*s-p-q-P-Q-`kfc`.

    nst : int
        The size of the state set. `nst` = P*s+D*s+d+q+max(p, Q*s).

        `nst` should be used subsequently in ``uni_arima_update`` and
        ``uni_arima_forecast_state`` as the dimension of `st`.

    itc : int
        The number of iterations performed.

    nres : int
        The number of model residuals returned in `res`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 3)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 4)
            On entry, `nx` = *<value>* and minimum required size for the
            state set array = *<value>*.

        (`errno` 5)
            On entry, `ires` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ires` >=
            15*Q'+11n+13*`nppc`+8*P'+12+2*(Q'+`nppc`)^2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            A failure in the search procedure has occurred.

        (`errno` 8)
            The inversion of the Hessian matrix in the calculation of
            the covariance matrix of the parameter estimates has failed.

        (`errno` 9)
            Failure whilst calculating backforecasts.

        (`errno` 10)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arima_estim_easy'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input/Output, c.
        utils._BYREF_INT, # Input, kfc.
        utils._EngineFloat64ArrayType.c_type, # Input, x[nx].
        utils._BYREF_INT, # Input, nx.
        utils._BYREF_FLOAT64, # Output, s.
        utils._BYREF_INT, # Output, ndf.
        utils._EngineFloat64ArrayType.c_type, # Output, sd[nppc].
        utils._BYREF_INT, # Input, nppc.
        utils._EngineFloat64ArrayType.c_type, # Output, cm[ldcm, nppc].
        utils._BYREF_INT, # Input, ldcm.
        utils._EngineFloat64ArrayType.c_type, # Output, st[nx].
        utils._BYREF_INT, # Output, nst.
        utils._BYREF_INT, # Input, kpiv.
        utils._BYREF_INT, # Input, nit.
        utils._BYREF_INT, # Output, itc.
        utils._EngineIntArrayType.c_type, # Output, isf[4].
        utils._EngineFloat64ArrayType.c_type, # Output, res[ires].
        utils._BYREF_INT, # Input, ires.
        utils._BYREF_INT, # Output, nres.
        utils._BYREF_INT, # Input, nadv.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((cm, 'cm'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par_cls = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    )
    _par = _par_cls.to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _kfc = utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = utils._EngineIntScalarType.to_py(_nx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nx_py,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    _ndf = utils._EngineIntScalarType.empty_ctypes()
    _nppc = utils._EngineIntScalarType(
        nppc, {'fun_name': fun_name, 'entity': 'nppc'},
    ).to_ctypes()
    _nppc_py = utils._EngineIntScalarType.to_py(_nppc)
    _sd_cls = utils._EngineFloat64ArrayType(
        sd, {'fun_name': fun_name, 'entity': 'sd'},
        exp_shape=(_nppc_py,),
        check_contents=False,
    )
    _sd = _sd_cls.to_ctypes(can_reference_contents=False)
    _ldcm_py = utils._EngineFloat64ArrayType.get_nd(cm, {'fun_name': fun_name, 'entity': 'cm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcm = utils._EngineIntScalarType(
        _ldcm_py, {'fun_name': fun_name, 'entity': '_ldcm'},
    ).to_ctypes()
    _cm_cls = utils._EngineFloat64ArrayType(
        cm, {'fun_name': fun_name, 'entity': 'cm'},
        exp_shape=(_nppc_py, _nppc_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cm = _cm_cls.to_ctypes(can_reference_contents=False)
    _st_cls = utils._EngineFloat64ArrayType(
        st, {'fun_name': fun_name, 'entity': 'st'},
        exp_shape=(_nx_py,),
        check_contents=False,
    )
    _st = _st_cls.to_ctypes(can_reference_contents=False)
    _nst = utils._EngineIntScalarType.empty_ctypes()
    _kpiv = utils._EngineIntScalarType(
        kpiv, {'fun_name': fun_name, 'entity': 'kpiv'},
    ).to_ctypes()
    _nit = utils._EngineIntScalarType(
        nit, {'fun_name': fun_name, 'entity': 'nit'},
    ).to_ctypes()
    _itc = utils._EngineIntScalarType.empty_ctypes()
    _isf_cls = utils._EngineIntArrayType(
        isf, {'fun_name': fun_name, 'entity': 'isf'},
        exp_shape=(4,),
        check_contents=False,
    )
    _isf = _isf_cls.to_ctypes(can_reference_contents=False)
    _ires = utils._EngineIntScalarType(
        ires, {'fun_name': fun_name, 'entity': 'ires'},
    ).to_ctypes()
    _ires_py = utils._EngineIntScalarType.to_py(_ires)
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_ires_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _nres = utils._EngineIntScalarType.empty_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _mr,
        _par,
        _npar,
        _c,
        _kfc,
        _x,
        _nx,
        _s,
        _ndf,
        _sd,
        _nppc,
        _cm,
        _ldcm,
        _st,
        _nst,
        _kpiv,
        _nit,
        _itc,
        _isf,
        _res,
        _ires,
        _nres,
        _nadv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kfc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kfc = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar = p+q+P+Q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model is over-parameterised.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s} and minimum required size for the state'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'set array = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ires = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ires >= 15*Q\'+11n+13*nppc+8*P\'+12+2*(Q\'+nppc)^2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A failure in the search procedure has occurred.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99989:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The inversion of the Hessian matrix in the calculation of the covariance',
                    'matrix of the parameter estimates has failed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure whilst calculating backforecasts.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99987:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Satisfactory parameter estimates could not be obtained',
                    'for all parameter types in the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _par_cls.refresh_py(_par)
        _c_py = utils._EngineFloat64ScalarType.to_py(_c)
        _sd_cls.refresh_py(_sd)
        _cm_cls.refresh_py(_cm)
        _st_cls.refresh_py(_st)
        _isf_cls.refresh_py(_isf)
        _res_cls.refresh_py(_res)
        _s_py = utils._EngineFloat64ScalarType.to_py(_s)
        _ndf_py = utils._EngineIntScalarType.to_py(_ndf)
        _nst_py = utils._EngineIntScalarType.to_py(_nst)
        _itc_py = utils._EngineIntScalarType.to_py(_itc)
        _nres_py = utils._EngineIntScalarType.to_py(_nres)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['c', 's', 'ndf', 'nst', 'itc', 'nres'],
        )(
            _c_py,
            _s_py,
            _ndf_py,
            _nst_py,
            _itc_py,
            _nres_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_arima_update(st, nst, mr, par, npar, c, anx, nuv, anexr):
    u"""
    Univariate time series, update state set for forecasting.

    ``uni_arima_update`` accepts a series of new observations of a time
    series, the model of which is already fully specified, and updates
    the 'state set' information for use in constructing further
    forecasts.
    The previous specifications of the time series model should have
    been obtained by using ``uni_arima_estim`` or
    ``uni_arima_estim_easy`` to estimate the relevant parameters.
    The supplied state set will originally have been produced by
    ``uni_arima_estim`` or ``uni_arima_estim_easy``, but may since have
    been updated by earlier calls to ``uni_arima_update``.

    A set of residuals corresponding to the new observations is
    returned.
    These may be of use in checking that the new observations conform to
    the previously fitted model.

    For full information please refer to the NAG Library document for
    g13ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13agf.html

    Parameters
    ----------
    st : float, shape (`nst`), modified in place
        On entry: the state set derived from ``uni_arima_estim`` or
        ``uni_arima_estim_easy``, or as modified using earlier calls of
        ``uni_arima_update``.

        On exit: the updated values of the state set.

    nst : int
        The number of values in the state set array `st`.

    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, shape (`npar`)
        The estimates of the p values of the phi parameters, the q
        values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters in the model
        -- in that order, using the usual notation.

    npar : int
        The number of phi, theta, Phi and Theta parameters in the model.

    c : float
        The constant to be subtracted from the differenced data.

    anx : float, shape (`nuv`)
        The new undifferenced observations which are to be used to
        update `st`.

    nuv : int
        k, the number of new observations in `anx`.

    anexr : float, shape (`nuv`), modified in place
        On exit: the residuals corresponding to the new observations in
        `anx`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` = p+q+P+Q.

        (`errno` 2)
            On entry, `nst` = *<value>*.

            Constraint: `nst` = P*s+D*s+d+q+max(Q*s, p).

        (`errno` 3)
            On entry, `nuv` = *<value>*.

            Constraint: `nuv` > 0.

        (`errno` 4)
            On entry, nwa = *<value>* and the minimum size required =
            *<value>*.

            Constraint: nwa >= 4*`npar`+3*`nst`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arima_update'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, st[nst].
        utils._BYREF_INT, # Input, nst.
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input, c.
        utils._EngineFloat64ArrayType.c_type, # Input, anx[nuv].
        utils._BYREF_INT, # Input, nuv.
        utils._EngineFloat64ArrayType.c_type, # Output, anexr[nuv].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wa.
        utils._BYREF_INT, # Input, nwa.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _nst = utils._EngineIntScalarType(
        nst, {'fun_name': fun_name, 'entity': 'nst'},
    ).to_ctypes()
    _nst_py = utils._EngineIntScalarType.to_py(_nst)
    _st_cls = utils._EngineFloat64ArrayType(
        st, {'fun_name': fun_name, 'entity': 'st'},
        exp_shape=(_nst_py,),
    )
    _st = _st_cls.to_ctypes()
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _nuv = utils._EngineIntScalarType(
        nuv, {'fun_name': fun_name, 'entity': 'nuv'},
    ).to_ctypes()
    _nuv_py = utils._EngineIntScalarType.to_py(_nuv)
    _anx = utils._EngineFloat64ArrayType(
        anx, {'fun_name': fun_name, 'entity': 'anx'},
        exp_shape=(_nuv_py,),
    ).to_ctypes()
    _anexr_cls = utils._EngineFloat64ArrayType(
        anexr, {'fun_name': fun_name, 'entity': 'anexr'},
        exp_shape=(_nuv_py,),
        check_contents=False,
    )
    _anexr = _anexr_cls.to_ctypes(can_reference_contents=False)
    _wa = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _nwa = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _st,
        _nst,
        _mr,
        _par,
        _npar,
        _c,
        _anx,
        _nuv,
        _anexr,
        _wa,
        _nwa,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar = p+q+P+Q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nst = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nst = P*s+D*s+d+q+max(Q*s, p).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nuv = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nuv > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nwa = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nwa >= 4*npar+3*nst.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _st_cls.refresh_py(_st)
        _anexr_cls.refresh_py(_anexr)
        return_data = None
    _err_class.manage()
    return return_data

def uni_arima_forecast_state(st, nst, mr, par, npar, c, rms, nfv, fva, fsd):
    u"""
    Univariate time series, forecasting from state set.

    ``uni_arima_forecast_state`` produces forecasts of a time series,
    given a time series model which has already been fitted to the time
    series using ``uni_arima_estim`` or ``uni_arima_estim_easy``.
    The original observations are not required, since
    ``uni_arima_forecast_state`` uses as input either the original state
    set produced by ``uni_arima_estim`` or ``uni_arima_estim_easy`` or
    the state set updated by a series of new observations using
    ``uni_arima_update``.
    Standard errors of the forecasts are also provided.

    For full information please refer to the NAG Library document for
    g13ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ahf.html

    Parameters
    ----------
    st : float, shape (`nst`)
        The state set derived from ``uni_arima_estim`` or
        ``uni_arima_estim_easy`` originally, or as modified using
        earlier calls of ``uni_arima_update``.

    nst : int
        The number of values in the state set array `st`.

    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, shape (`npar`)
        The estimates of the p values of the phi parameters, the q
        values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters which
        specify the model and which were output originally by
        ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    npar : int
        The number of phi, theta, Phi and Theta parameters in the model.

    c : float
        c, the value of the model constant. This will have been output
        by ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    rms : float
        V, the residual variance associated with the model.

        If ``uni_arima_estim_easy`` was used to estimate the model,
        `rms` should be set to s/ndf, where s and ndf were output by
        ``uni_arima_estim_easy``.

        If ``uni_arima_estim`` was used to estimate the model, `rms`
        should be set to s/icount[4], where s and icount[4] were output
        by ``uni_arima_estim``.

    nfv : int
        L, the required number of forecasts.

    fva : float, shape (`nfv`), modified in place
        On exit: `nfv` forecast values relating to the original
        undifferenced series.

    fsd : float, shape (`nfv`), modified in place
        On exit: the standard errors associated with each of the `nfv`
        forecast values in `fva`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` = p+q+P+Q.

        (`errno` 2)
            On entry, `nst` = *<value>*.

            Constraint: `nst` = P*s+D*s+d+q+max(Q*s, p).

        (`errno` 3)
            On entry, `nfv` = *<value>*.

            Constraint: `nfv` > 0.

        (`errno` 4)
            On entry, nwa = *<value>* and the minimum size required =
            *<value>*.

            Constraint: nwa >= 4*`npar`+3*`nst`.

        (`errno` 5)
            On entry, `rms` = *<value>*.

            Constraint: `rms` >= 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arima_forecast_state'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._EngineFloat64ArrayType.c_type, # Input, st[nst].
        utils._BYREF_INT, # Input, nst.
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_FLOAT64, # Input, rms.
        utils._BYREF_INT, # Input, nfv.
        utils._EngineFloat64ArrayType.c_type, # Output, fva[nfv].
        utils._EngineFloat64ArrayType.c_type, # Output, fsd[nfv].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wa.
        utils._BYREF_INT, # Input, nwa.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _nst = utils._EngineIntScalarType(
        nst, {'fun_name': fun_name, 'entity': 'nst'},
    ).to_ctypes()
    _nst_py = utils._EngineIntScalarType.to_py(_nst)
    _st = utils._EngineFloat64ArrayType(
        st, {'fun_name': fun_name, 'entity': 'st'},
        exp_shape=(_nst_py,),
    ).to_ctypes()
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _rms = utils._EngineFloat64ScalarType(
        rms, {'fun_name': fun_name, 'entity': 'rms'},
    ).to_ctypes()
    _nfv = utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _nfv_py = utils._EngineIntScalarType.to_py(_nfv)
    _fva_cls = utils._EngineFloat64ArrayType(
        fva, {'fun_name': fun_name, 'entity': 'fva'},
        exp_shape=(_nfv_py,),
        check_contents=False,
    )
    _fva = _fva_cls.to_ctypes(can_reference_contents=False)
    _fsd_cls = utils._EngineFloat64ArrayType(
        fsd, {'fun_name': fun_name, 'entity': 'fsd'},
        exp_shape=(_nfv_py,),
        check_contents=False,
    )
    _fsd = _fsd_cls.to_ctypes(can_reference_contents=False)
    _wa = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _nwa = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _st,
        _nst,
        _mr,
        _par,
        _npar,
        _c,
        _rms,
        _nfv,
        _fva,
        _fsd,
        _wa,
        _nwa,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar = p+q+P+Q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nst = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nst = P*s+D*s+d+q+max(Q*s, p).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfv = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nfv > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nwa = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nwa >= 4*npar+3*nst.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rms = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rms >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _fva_cls.refresh_py(_fva)
        _fsd_cls.refresh_py(_fsd)
        return_data = None
    _err_class.manage()
    return return_data

def uni_arima_forcecast(mr, par, npar, c, kfc, x, nx, st, ist, nfv, fva, fsd, ifv, isf):
    u"""
    Univariate time series, state set and forecasts, from fully
    specified seasonal ARIMA model.

    ``uni_arima_forcecast`` applies a fully specified seasonal ARIMA
    model to an observed time series, generates the state set for
    forecasting and (optionally) derives a specified number of forecasts
    together with their standard deviations.

    For full information please refer to the NAG Library document for
    g13aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ajf.html

    Parameters
    ----------
    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, shape (`npar`)
        The p values of the phi parameters, the q values of the theta
        parameters, the P values of the Phi parameters, and the Q values
        of the Theta parameters, in that order.

    npar : int
        The exact number of phi, theta, Phi and Theta parameters.

    c : float
        c, the expected value of the differenced series (i.e., c is the
        constant correction). Where there is no constant term, `c` must
        be set to 0.0.

    kfc : int
        Must be set to 0 if `c` was not estimated, and 1 if `c` was
        estimated. This is irrespective of whether or not `c` = 0.0. The
        only effect is that the residual degrees of freedom are one
        greater when `kfc` = 0. Assuming the supplied time series to be
        the same as that to which the model was originally fitted, this
        ensures an unbiased estimate of the residual mean-square.

    x : float, shape (`nx`)
        The n values of the original undifferenced time series.

    nx : int
        n, the length of the original undifferenced time series.

    st : float, shape (`ist`), modified in place
        On exit: the `nst` values of the state set.

    ist : int
        The dimension of the array `st`.

    nfv : int
        The required number of forecasts. If `nfv` <= 0, no forecasts
        will be computed.

    fva : float, shape (`ifv`), modified in place
        On exit: if `nfv` > 0, `fva` contains the `nfv` forecast values
        relating to the original undifferenced time series.

    fsd : float, shape (`ifv`), modified in place
        On exit: if `nfv` > 0, `fsd` contains the estimated standard
        errors of the `nfv` forecast values.

    ifv : int
        The dimension of the arrays `fva` and `fsd`.

    isf : int, shape (4), modified in place
        On exit: contains validity indicators, one for each of the four
        possible parameter types in the model (autoregressive, moving
        average, seasonal autoregressive, seasonal moving average), in
        that order.

        Each indicator has the interpretation:

        [table omitted]

    Returns
    -------
    rms : float
        The residual variance (mean square) associated with the model.

    nst : int
        The number of values in the state set array `st`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 4)
            On entry, `ist` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ist` >= (P*s)+d+(D*s)+q+max(p, Q*s).

        (`errno` 5)
            Unable to calculate the latest estimates of the
            backforecasts.

        (`errno` 6)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

        (`errno` 7)
            On entry, `ifv` = *<value>* and `nfv` = *<value>*.

            Constraint: `ifv` >= max(1, `nfv`).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arima_forcecast'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ajft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input, c.
        utils._BYREF_INT, # Input, kfc.
        utils._EngineFloat64ArrayType.c_type, # Input, x[nx].
        utils._BYREF_INT, # Input, nx.
        utils._BYREF_FLOAT64, # Output, rms.
        utils._EngineFloat64ArrayType.c_type, # Output, st[ist].
        utils._BYREF_INT, # Input, ist.
        utils._BYREF_INT, # Output, nst.
        utils._BYREF_INT, # Input, nfv.
        utils._EngineFloat64ArrayType.c_type, # Output, fva[ifv].
        utils._EngineFloat64ArrayType.c_type, # Output, fsd[ifv].
        utils._BYREF_INT, # Input, ifv.
        utils._EngineIntArrayType.c_type, # Output, isf[4].
        utils._EngineFloat64ArrayType.c_type, # Workspace, w.
        utils._BYREF_INT, # Input, iw.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    ).to_ctypes()
    _c = utils._EngineFloat64ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _kfc = utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = utils._EngineIntScalarType.to_py(_nx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nx_py,),
    ).to_ctypes()
    _rms = utils._EngineFloat64ScalarType.empty_ctypes()
    _ist = utils._EngineIntScalarType(
        ist, {'fun_name': fun_name, 'entity': 'ist'},
    ).to_ctypes()
    _ist_py = utils._EngineIntScalarType.to_py(_ist)
    _st_cls = utils._EngineFloat64ArrayType(
        st, {'fun_name': fun_name, 'entity': 'st'},
        exp_shape=(_ist_py,),
        check_contents=False,
    )
    _st = _st_cls.to_ctypes(can_reference_contents=False)
    _nst = utils._EngineIntScalarType.empty_ctypes()
    _nfv = utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _ifv = utils._EngineIntScalarType(
        ifv, {'fun_name': fun_name, 'entity': 'ifv'},
    ).to_ctypes()
    _ifv_py = utils._EngineIntScalarType.to_py(_ifv)
    _fva_cls = utils._EngineFloat64ArrayType(
        fva, {'fun_name': fun_name, 'entity': 'fva'},
        exp_shape=(_ifv_py,),
        check_contents=False,
    )
    _fva = _fva_cls.to_ctypes(can_reference_contents=False)
    _fsd_cls = utils._EngineFloat64ArrayType(
        fsd, {'fun_name': fun_name, 'entity': 'fsd'},
        exp_shape=(_ifv_py,),
        check_contents=False,
    )
    _fsd = _fsd_cls.to_ctypes(can_reference_contents=False)
    _isf_cls = utils._EngineIntArrayType(
        isf, {'fun_name': fun_name, 'entity': 'isf'},
        exp_shape=(4,),
        check_contents=False,
    )
    _isf = _isf_cls.to_ctypes(can_reference_contents=False)
    _w = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iw = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _mr,
        _par,
        _npar,
        _c,
        _kfc,
        _x,
        _nx,
        _rms,
        _st,
        _ist,
        _nst,
        _nfv,
        _fva,
        _fsd,
        _ifv,
        _isf,
        _w,
        _iw,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kfc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kfc = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar = p+q+P+Q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The model is over-parameterised.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ist = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ist >= (P*s)+d+(D*s)+q+max(p, Q*s).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the latest estimates of the backforecasts.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Satisfactory parameter estimates could not be obtained',
                    'for all parameter types in the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ifv = {:s} and nfv = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ifv >= max(1, nfv).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _st_cls.refresh_py(_st)
        _fva_cls.refresh_py(_fva)
        _fsd_cls.refresh_py(_fsd)
        _isf_cls.refresh_py(_isf)
        _rms_py = utils._EngineFloat64ScalarType.to_py(_rms)
        _nst_py = utils._EngineIntScalarType.to_py(_nst)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rms', 'nst'],
        )(
            _rms_py,
            _nst_py,
        )
    _err_class.manage()
    return return_data

def uni_smooth_exp(mode, itype, p, param, n, y, k, init, nf, fv, fse, yhat, res, r):
    u"""
    Univariate time series, exponential smoothing.

    ``uni_smooth_exp`` performs exponential smoothing using either
    single exponential, double exponential or a Holt--Winters method.

    For full information please refer to the NAG Library document for
    g13am

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13amf.html

    Parameters
    ----------
    mode : int
        Indicates if ``uni_smooth_exp`` is continuing from a previous
        call or, if not, how the initial values are computed.

        `mode` = 0

          Required values for m_0, r_0 and s_{-j}, for j = 0,1,...,p-1,
          are supplied in `init`.

        `mode` = 1

          ``uni_smooth_exp`` continues from a previous call using values
          that are supplied in `r`.

        `mode` = 2

          Required values for m_0, r_0 and s_{-j}, for j = 0,1,...,p-1,
          are estimated using the first k observations.

    itype : int
        The smoothing function.

        `itype` = 1

          Single exponential.

        `itype` = 2

          Brown double exponential.

        `itype` = 3

          Linear Holt.

        `itype` = 4

          Additive Holt--Winters.

        `itype` = 5

          Multiplicative Holt--Winters.

    p : int
        If `itype` = 4 or 5, the seasonal order, p, otherwise `p` is not
        referenced.

    param : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2): 1; if `itype`=3: 3; if `itype` in
        (4, 5): 4; otherwise: 0.

        The smoothing parameters.

        If `itype` = 1 or 2, `param`[0] = alpha and any remaining
        elements of `param` are not referenced.

        If `itype` = 3, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = phi and any remaining elements of `param` are not
        referenced.

        If `itype` = 4 or 5, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = beta and `param`[3] = phi.

    n : int
        The number of observations in the series.

    y : float, shape (`n`)
        The time series.

    k : int
        If `mode` = 2, the number of observations used to initialize the
        smoothing.

        If `mode` != 2, `k` is not referenced.

    init : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `itype`=1: 1; if `itype` in (2, 3): 2; if `itype` in
        (4, 5): 2+`p`; otherwise: 0.

        On entry: if `mode` = 0, the initial values for m_0, r_0 and
        s_{-j}, for j = 0,1,...,p-1, used to initialize the smoothing.

        If `itype` = 1, `init`[0] = m_0 and the remaining elements of
        `init` are not referenced.

        If `itype` = 2 or 3, `init`[0] = m_0 and `init`[1] = r_0 and the
        remaining elements of `init` are not referenced.

        If `itype` = 4 or 5, `init`[0] = m_0, `init`[1] = r_0 and
        `init`[2] to `init`[p+1] hold the values for s_{-j}, for j =
        0,1,...,p-1.

        The remaining elements of `init` are not referenced.

        On exit: if `mode` != 1, the values used to initialize the
        smoothing. These are in the same order as described above.

    nf : int
        The number of forecasts required beyond the end of the series.
        Note, the one step ahead forecast is always produced.

    fv : float, shape (`nf`), modified in place
        On exit: y^_{t+f}, for f = 1,2,...,`nf`, the next `nf` step
        forecasts. Where t = `n`, if `mode` != 1, else t is the total
        number of smoothed and forecast values already produced.

    fse : float, shape (`nf`), modified in place
        On exit: the forecast standard errors for the values given in
        `fv`.

    yhat : float, shape (`n`), modified in place
        On exit: y^_{t+1}, for t = 1,2,...,`n`, the one step ahead
        forecast values, with `yhat`[i-1] being the one step ahead
        forecast of `y`[i-2].

    res : float, shape (`n`), modified in place
        On exit: the residuals, (y_{t+1}-y^_{t+1}), for t = 1,2,...,`n`.

    r : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2, 3): 13; if `itype` in (4, 5):
        13+`p`; otherwise: 0.

        On entry: if `mode` = 1, `r` must contain the values as returned
        by a previous call to ``rand.times_smooth_exp`` or
        ``uni_smooth_exp``, `r` need not be set otherwise.

        If `itype` = 1, 2 or 3, only the first 13 elements of `r` are
        referenced, otherwise the first 13+p elements are referenced.

        On exit: the information on the current state of the smoothing.

    Returns
    -------
    dv : float
        The square root of the mean deviation.

    ad : float
        The mean absolute deviation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1, 2, 3, 4 or 5.

        (`errno` 3)
            On entry, `p` = *<value>*.

            Constraint: if `itype` = 4 or 5, `p` > 1.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: 0.0 <= `param`[*<value>*] <= 1.0.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: if `itype` = 2, 0.0 < `param`[*<value>*] <= 1.0.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: `param`[*<value>*] >= 0.0.

        (`errno` 5)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 6)
            A multiplicative Holt--Winters model cannot be used with the
            supplied data.

        (`errno` 7)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: if `mode` = 2 and `itype` = 4 or 5, 1 <= `k` <=
            `n`.

        (`errno` 7)
            On entry, `k` = *<value>*, 2*`p` = *<value>*.

            Constraint: if `mode` = 2 and `itype` = 4 or 5, 2*`p` <=
            `k`.

        (`errno` 9)
            On entry, `nf` = *<value>*.

            Constraint: `nf` >= 0.

        (`errno` 16)
            On entry, the array `r` has not been initialized correctly.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_smooth_exp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13amft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, itype.
        utils._BYREF_INT, # Input, p.
        utils._EngineFloat64ArrayType.c_type, # Input, param[lparam].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, init[linit].
        utils._BYREF_INT, # Input, nf.
        utils._EngineFloat64ArrayType.c_type, # Output, fv[nf].
        utils._EngineFloat64ArrayType.c_type, # Output, fse[nf].
        utils._EngineFloat64ArrayType.c_type, # Output, yhat[n].
        utils._EngineFloat64ArrayType.c_type, # Output, res[n].
        utils._BYREF_FLOAT64, # Output, dv.
        utils._BYREF_FLOAT64, # Output, ad.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lparam.
        utils._BYREF_INT, # Input, linit.
        utils._BYREF_INT, # Input, lr.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _p = utils._EngineIntScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _lparam_py = utils._EngineFloat64ArrayType.get_shape(param, {'fun_name': fun_name, 'entity': 'param'}, exp_rank=1)[0]
    _lparam = utils._EngineIntScalarType(
        _lparam_py, {'fun_name': fun_name, 'entity': '_lparam'},
    ).to_ctypes()
    _itype_py = utils._EngineIntScalarType.to_py(_itype)
    if _itype_py in (1, 2):
        _param_dim1 = 1
    elif _itype_py == 3:
        _param_dim1 = 3
    elif _itype_py in (4, 5):
        _param_dim1 = 4
    else:
        _param_dim1 = 0
    exp_shape_param = (_param_dim1,)
    _param = utils._EngineFloat64ArrayType(
        param, {'fun_name': fun_name, 'entity': 'param'},
        exp_shape=exp_shape_param,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _linit_py = utils._EngineFloat64ArrayType.get_shape(init, {'fun_name': fun_name, 'entity': 'init'}, exp_rank=1)[0]
    _linit = utils._EngineIntScalarType(
        _linit_py, {'fun_name': fun_name, 'entity': '_linit'},
    ).to_ctypes()
    _p_py = utils._EngineIntScalarType.to_py(_p)
    if _itype_py == 1:
        _init_dim1 = 1
    elif _itype_py in (2, 3):
        _init_dim1 = 2
    elif _itype_py in (4, 5):
        _init_dim1 = 2+_p_py
    else:
        _init_dim1 = 0
    exp_shape_init = (_init_dim1,)
    _init_cls = utils._EngineFloat64ArrayType(
        init, {'fun_name': fun_name, 'entity': 'init'},
        exp_shape=exp_shape_init,
    )
    _init = _init_cls.to_ctypes()
    _nf = utils._EngineIntScalarType(
        nf, {'fun_name': fun_name, 'entity': 'nf'},
    ).to_ctypes()
    _nf_py = utils._EngineIntScalarType.to_py(_nf)
    _fv_cls = utils._EngineFloat64ArrayType(
        fv, {'fun_name': fun_name, 'entity': 'fv'},
        exp_shape=(_nf_py,),
        check_contents=False,
    )
    _fv = _fv_cls.to_ctypes(can_reference_contents=False)
    _fse_cls = utils._EngineFloat64ArrayType(
        fse, {'fun_name': fun_name, 'entity': 'fse'},
        exp_shape=(_nf_py,),
        check_contents=False,
    )
    _fse = _fse_cls.to_ctypes(can_reference_contents=False)
    _yhat_cls = utils._EngineFloat64ArrayType(
        yhat, {'fun_name': fun_name, 'entity': 'yhat'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _yhat = _yhat_cls.to_ctypes(can_reference_contents=False)
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _dv = utils._EngineFloat64ScalarType.empty_ctypes()
    _ad = utils._EngineFloat64ScalarType.empty_ctypes()
    _lr_py = utils._EngineFloat64ArrayType.get_shape(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    if _itype_py in (1, 2, 3):
        _r_dim1 = 13
    elif _itype_py in (4, 5):
        _r_dim1 = 13+_p_py
    else:
        _r_dim1 = 0
    exp_shape_r = (_r_dim1,)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=exp_shape_r,
    )
    _r = _r_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _itype,
        _p,
        _param,
        _n,
        _y,
        _k,
        _init,
        _nf,
        _fv,
        _fse,
        _yhat,
        _res,
        _dv,
        _ad,
        _r,
        _lparam,
        _linit,
        _lr,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = 1, 2, 3, 4 or 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if itype = 4 or 5, p > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, param[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0.0 <= param[{:s}] <= 1.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, param[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if itype = 2, 0.0 < param[{:s}] <= 1.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 43 and _ifmt == 43:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, param[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: param[{:s}] >= 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A multiplicative Holt--Winters model cannot be used with the supplied data.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if mode = 2 and itype = 4 or 5, 1 <= k <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 72 and _ifmt == 72:
            _ierr_fl = 7
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}, 2*p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if mode = 2 and itype = 4 or 5, 2*p <= k.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nf >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 16 and _ifmt == 16:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the array r has not been initialized correctly.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _init_cls.refresh_py(_init)
        _fv_cls.refresh_py(_fv)
        _fse_cls.refresh_py(_fse)
        _yhat_cls.refresh_py(_yhat)
        _res_cls.refresh_py(_res)
        _r_cls.refresh_py(_r)
        _dv_py = utils._EngineFloat64ScalarType.to_py(_dv)
        _ad_py = utils._EngineFloat64ScalarType.to_py(_ad)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['dv', 'ad'],
        )(
            _dv_py,
            _ad_py,
        )
    _err_class.manage()
    return return_data

def uni_arima_resid(n, v, mr, m, par, npar, ishow, r, rcm, io_manager=None):
    u"""
    Univariate time series, diagnostic checking of residuals, following
    ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    ``uni_arima_resid`` is a diagnostic checking function suitable for
    use after fitting a Box--Jenkins ARMA model to a univariate time
    series using ``uni_arima_estim`` or ``uni_arima_estim_easy``. The
    residual autocorrelation function is returned along with an estimate
    of its asymptotic standard errors and correlations.
    Also, ``uni_arima_resid`` calculates the Box--Ljung portmanteau
    statistic and its significance level for testing model adequacy.

    For full information please refer to the NAG Library document for
    g13as

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13asf.html

    Parameters
    ----------
    n : int
        n, the number of observations in the residual series.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim``, then `n` must be the value icount[1]
        returned by ``uni_arima_estim``.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim_easy``, then `n` must be the value nres
        returned by ``uni_arima_estim_easy``.

    v : float, shape (`n`)
        `v`[t-1] must contain an estimate of epsilon_t, for t =
        1,2,...,n.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim`` then the actual argument `v` must be
        exr[icount[0]] as returned by ``uni_arima_estim``.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim_easy`` then the actual argument `v` must be
        res as returned by ``uni_arima_estim_easy``.

    mr : int, shape (7)
        The orders vector (p, d, q, P, D, Q, s) as supplied to
        ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    m : int
        The value of m, the number of residual autocorrelations to be
        computed. See [Further Comments] for advice on the value of `m`.

    par : float, shape (`npar`)
        The parameter estimates in the order phi_1,phi_2,...,phi_p,
        theta_1,theta_2,...,theta_q, Phi_1,Phi_2,...,Phi_P,
        Theta_1,Theta_2,...,Theta_Q only.

    npar : int
        The total number of phi, theta, Phi and Theta parameters, i.e.,
        `npar` = p+q+P+Q.

    ishow : int
        Must be nonzero if the residual autocorrelations, their standard
        errors and the portmanteau statistics are to be printed and zero
        otherwise.

        These quantities are available also as output variables in `r`,
        `rcm`, `chi`, `idf` and `siglev`.

    r : float, shape (`m`), modified in place
        On exit: an estimate of the residual autocorrelation coefficient
        at lag l, for l = 1,2,...,m. If `errno` = 3 on exit then all
        elements of `r` are set to zero.

    rcm : float, shape (`m`, `m`), modified in place
        On exit: the estimated standard errors and correlations of the
        elements in the array `r`. The correlation between `r`[i-1] and
        `r`[j-1] is returned as `rcm`[i-1,j-1] except that if i = j then
        `rcm`[i-1,j-1] contains the standard error of `r`[i-1]. If on
        exit, `errno` >= 5, then all off-diagonal elements of `rcm` are
        set to zero and all diagonal elements are set to 1/sqrt(n).

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    chi : float
        The value of the portmanteau statistic, Q_(m). If `errno` = 3 on
        exit then `chi` is returned as zero.

    idf : int
        The number of degrees of freedom of `chi`.

    siglev : float
        The significance level of `chi` based on `idf` degrees of
        freedom. If `errno` = 3 on exit, `siglev` is returned as one.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` < `n`.

        (`errno` 1)
            On entry, `m` = *<value>* and `npar` = *<value>*.

            Constraint: `m` > `npar`.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` != 0 and `npar` = p+q+P+Q.

        (`errno` 1)
            On entry, `mr`[3] = *<value>* and `mr`[5] = *<value>*.

            Constraint: `mr`[3] = 0 and `mr`[5] = 0.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*.

            Constraint: `mr`[6] >= 0.

        (`errno` 1)
            On entry, `mr`[5] = *<value>*.

            Constraint: `mr`[5] >= 0.

        (`errno` 1)
            On entry, `mr`[3] = *<value>*.

            Constraint: `mr`[3] >= 0.

        (`errno` 1)
            On entry, `mr`[2] = *<value>*.

            Constraint: `mr`[2] >= 0.

        (`errno` 1)
            On entry, `mr`[0] = *<value>*.

            Constraint: `mr`[0] >= 0.

        (`errno` 2)
            On entry, the seasonal moving average operator is
            non-invertible.

        (`errno` 2)
            On entry, the seasonal autoregressive operator is
            nonstationary.

        (`errno` 2)
            On entry, the non-seasonal moving average operator is
            non-invertible.

        (`errno` 2)
            On entry, the non-seasonal autoregressive operator is
            nonstationary.

        (`errno` 4)
            Excessive iterations needed to find zeros of determinental
            polynomials.

        (`errno` 5)
            On entry, one or more of the AR operators has a factor in
            common with one or more of the MA operators.

        (`errno` 6)
            The matrix `rcm` could not be computed because one of its
            diagonal elements was found to be non-positive.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, the elements of `v` are nearly identical giving
            near-zero variance.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arima_resid'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13asft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, v[n].
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_INT, # Input, ishow.
        utils._BYREF_INT, # Input, nadv.
        utils._EngineFloat64ArrayType.c_type, # Output, r[m].
        utils._EngineFloat64ArrayType.c_type, # Output, rcm[ldrcm, m].
        utils._BYREF_INT, # Input, ldrcm.
        utils._BYREF_FLOAT64, # Output, chi.
        utils._BYREF_INT, # Output, idf.
        utils._BYREF_FLOAT64, # Output, siglev.
        utils._EngineIntArrayType.c_type, # Workspace, iw.
        utils._BYREF_INT, # Input, liw.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((rcm, 'rcm'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    ).to_ctypes()
    _ishow = utils._EngineIntScalarType(
        ishow, {'fun_name': fun_name, 'entity': 'ishow'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ldrcm_py = utils._EngineFloat64ArrayType.get_nd(rcm, {'fun_name': fun_name, 'entity': 'rcm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrcm = utils._EngineIntScalarType(
        _ldrcm_py, {'fun_name': fun_name, 'entity': '_ldrcm'},
    ).to_ctypes()
    _rcm_cls = utils._EngineFloat64ArrayType(
        rcm, {'fun_name': fun_name, 'entity': 'rcm'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rcm = _rcm_cls.to_ctypes(can_reference_contents=False)
    _chi = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _siglev = utils._EngineFloat64ScalarType.empty_ctypes()
    _iw = utils._EngineIntArrayType.empty_ctypes(0)
    _liw = utils._EngineIntScalarType.c_type(0)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _n,
        _v,
        _mr,
        _m,
        _par,
        _npar,
        _ishow,
        _nadv,
        _r,
        _rcm,
        _ldrcm,
        _chi,
        _idf,
        _siglev,
        _iw,
        _liw,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m < n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m > npar.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar != 0 and npar = p+q+P+Q.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[3] = {:s} and mr[5] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mr[3] = 0 and mr[5] = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[6] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mr[6] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[5] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mr[5] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[3] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mr[3] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[2] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mr[2] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mr[0] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the seasonal moving average operator is non-invertible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the seasonal autoregressive operator is nonstationary.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the non-seasonal moving average operator is non-invertible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the non-seasonal autoregressive operator is nonstationary.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99976:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the elements of v are nearly identical giving near-zero variance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Excessive iterations needed to find zeros of',
                    'determinental polynomials.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99975:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, one or more of the AR operators has a factor in common with one',
                    'or more of the MA operators.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99974:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix rcm could not be computed because',
                    'one of its diagonal elements was found to be non-positive.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _r_cls.refresh_py(_r)
        _rcm_cls.refresh_py(_rcm)
        _chi_py = utils._EngineFloat64ScalarType.to_py(_chi)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _siglev_py = utils._EngineFloat64ScalarType.to_py(_siglev)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['chi', 'idf', 'siglev'],
        )(
            _chi_py,
            _idf_py,
            _siglev_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_means(n, z, m, ngrps, rs, y, mean):
    u"""
    Computes quantities needed for range-mean or standard deviation-mean
    plot.

    ``uni_means`` calculates the range (or standard deviation) and the
    mean for groups of successive time series values.
    It is intended for use in the construction of range-mean plots.

    For full information please refer to the NAG Library document for
    g13au

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13auf.html

    Parameters
    ----------
    n : int
        n, the number of observations in the time series.

    z : float, shape (`n`)
        `z`[t-1] must contain the tth observation Z_t, for t =
        1,2,...,n.

    m : int
        m, the group size.

    ngrps : int
        k, the number of groups.

    rs : str, length 1
        Indicates whether ranges or standard deviations are to be
        calculated.

        `rs` = 'R'

          Ranges are calculated.

        `rs` = 'S'

          Standard deviations are calculated.

    y : float, shape (`ngrps`), modified in place
        On exit: `y`[i-1] contains the range or standard deviation, as
        determined by `rs`, of the ith group of observations, for i =
        1,2,...,k.

    mean : float, shape (`ngrps`), modified in place
        On exit: `mean`[i-1] contains the mean of the ith group of
        observations, for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*, `m` = *<value>* and `ngrps` =
            *<value>*.

            Constraint: `ngrps` = int(`n`/`m`).

        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` >= `m`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 2)
            On entry, `rs` = *<value>*.

            Constraint: `rs` = 'R' or 'S'.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_means'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13auft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, z[n].
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, ngrps.
        utils._BYREF_STRING, # Input, rs(len=1).
        utils._EngineFloat64ArrayType.c_type, # Output, y[ngrps].
        utils._EngineFloat64ArrayType.c_type, # Output, mean[ngrps].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of rs.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ngrps = utils._EngineIntScalarType(
        ngrps, {'fun_name': fun_name, 'entity': 'ngrps'},
    ).to_ctypes()
    _rs = utils._EngineCharScalarType(
        rs, {'fun_name': fun_name, 'entity': 'rs'}, exp_chrlen=1,
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(int((_n_py//_m_py)),),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _mean_cls = utils._EngineFloat64ArrayType(
        mean, {'fun_name': fun_name, 'entity': 'mean'},
        exp_shape=(int((_n_py//_m_py)),),
        check_contents=False,
    )
    _mean = _mean_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _z,
        _m,
        _ngrps,
        _rs,
        _y,
        _mean,
        _errbuf,
        _ierr,
        _rs._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}, m = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and ngrps = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ngrps = int(n/m).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rs = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rs = \'R\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        _mean_cls.refresh_py(_mean)
        return_data = None
    _err_class.manage()
    return return_data

def uni_dickey_fuller_unit(ut_type, p, n, y):
    u"""
    Computes (augmented) Dickey--Fuller unit root test statistic.

    ``uni_dickey_fuller_unit`` returns the (augmented) Dickey--Fuller
    unit root test.

    For full information please refer to the NAG Library document for
    g13aw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13awf.html

    Parameters
    ----------
    ut_type : int
        The type of unit test for which the probability is required.

        `ut_type` = 1

          A unit root test will be performed and tau returned.

        `ut_type` = 2

          A unit root test with drift will be performed and tau_mu
          returned.

        `ut_type` = 3

          A unit root test with drift and deterministic time trend will
          be performed and tau_tau returned.

    p : int
        p, the degree of the autoregressive (AR) component of the
        Dickey--Fuller test statistic. When p > 1 the test is usually
        referred to as the augmented Dickey--Fuller test.

    n : int
        n, the length of the time series.

    y : float, shape (`n`)
        y, the time series.

    Returns
    -------
    ts : float
        The (augmented) Dickey--Fuller test statistic: tau when
        `ut_type` = 1; tau_mu when `ut_type` = 2 or tau_tau when
        `ut_type` = 3.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ut_type` = *<value>*.

            Constraint: `ut_type` = 1, 2 or 3.

        (`errno` 21)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `ut_type` = 1, `n` > 2*`p` if `ut_type` = 2,
            `n` > 2*`p`+1 if `ut_type` = 3, `n` > 2*`p`+2

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 41)
            On entry, the design matrix used in the estimation of beta_1
            is not of full rank, this is usually due to all elements of
            the series being virtually identical. The returned statistic
            is therefore not unique and likely to be meaningless.

        (`errno` 42)
            standard error of the parameter estimate is zero, therefore
            depending on the sign of the parameter estimate, a large
            positive or negative value has been returned.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_dickey_fuller_unit_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_dickey_fuller_unit'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13awft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ut_type.
        utils._BYREF_INT, # Input, p.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Output, ts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ut_type = utils._EngineIntScalarType(
        ut_type, {'fun_name': fun_name, 'entity': 'ut_type'},
    ).to_ctypes()
    _p = utils._EngineIntScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType.c_type(0.0)
    _ts = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ut_type,
        _p,
        _n,
        _y,
        _tol,
        _ts,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ut_type = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ut_type = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: p > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint:',
                    'if ut_type = 1, n > 2*p',
                    'if ut_type = 2, n > 2*p+1',
                    'if ut_type = 3, n > 2*p+2',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the design matrix used in the estimation of beta_1',
                    'is not of full rank, this is usually due to all elements of the series being',
                    'virtually identical. The returned statistic is therefore not unique',
                    'and likely to be meaningless.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'standard error of the parameter estimate is zero,',
                    'therefore depending on the sign of the parameter estimate,',
                    'a large positive or negative value has been returned.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ts_py = utils._EngineFloat64ScalarType.to_py(_ts)
        return_data = _ts_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ts'],
            )(
                _ts_py,
            )
    _err_class.manage()
    return return_data

def multi_filter_arima(y, ny, mr, nmr, par, npar, cy, b, nb):
    u"""
    Multivariate time series, filtering (pre-whitening) by an ARIMA
    model.

    ``multi_filter_arima`` filters a time series by an ARIMA model.

    For full information please refer to the NAG Library document for
    g13ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13baf.html

    Parameters
    ----------
    y : float, shape (`ny`)
        The Q'_y backforecasts, starting with backforecast at time
        1-Q'_y to backforecast at time 0, followed by the time series
        starting at time 1, where Q'_y = `mr`[9]+`mr`[12]*`mr`[13]. If
        there are no backforecasts, either because the ARIMA model for
        the time series is not known, or because it is known but has no
        moving average terms, then the time series starts at the
        beginning of `y`.

    ny : int
        The total number of backforecasts and time series data points in
        array `y`.

    mr : int, shape (`nmr`)
        The orders vector for the filtering model, followed by the
        orders vector for the ARIMA model for the time series if the
        latter is known. The orders appear in the standard sequence
        (p,d,q,P,D,Q,s) as given in [the G13 Introduction]. If the ARIMA
        model for the time series is supplied, the function will assume
        that the first Q'_y values of the array `y` are backforecasts.

    nmr : int
        The number of values specified in the array `mr`. It takes the
        value 7 if no ARIMA model for the time series is supplied but
        otherwise it takes the value 14. Thus `nmr` acts as an indicator
        as to whether backforecasting can be carried out.

    par : float, shape (`npar`)
        The parameters of the filtering model, followed by the
        parameters of the ARIMA model for the time series, if supplied.
        Within each model the parameters are in the standard order of
        non-seasonal AR and MA followed by seasonal AR and MA.

    npar : int
        The total number of parameters held in array `par`.

    cy : None or float
        If the ARIMA model is known (i.e., `nmr` = 14), `cy` must
        specify the constant term of the ARIMA model for the time
        series. If this model is not known (i.e., `nmr` = 7), `cy` is
        not used.

    b : float, shape (`nb`), modified in place
        On exit: the filtered output series. If the ARIMA model for the
        time series was known, and hence Q'_y backforecasts were
        supplied in `y`, then `b` contains Q'_y 'filtered' backforecasts
        followed by the filtered series. Otherwise, the filtered series
        begins at the start of `b` just as the original series began at
        the start of `y`. In either case, if the value of the series at
        time t is held in `y`[t-1], then the filtered value at time t is
        held in `b`[t-1].

    nb : int
        In addition to holding the returned filtered series, `b` is also
        used as an intermediate work array if the ARIMA model for the
        time series was known.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nmr` = *<value>*.

            Constraint: `nmr` = 7 or 14.

        (`errno` 2)
            On entry, the orders vector `mr` is invalid.

        (`errno` 3)
            On entry, `npar` = *<value>*.

            Constraint: `npar` must be inconsistent with `mr`.

        (`errno` 4)
            On entry, `ny` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ny` >= max(1+Q'_y, `npar`).

        (`errno` 6)
            On entry, `nb` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: if `nmr` = 14 then `nb` >= `ny`+max(K_3,
            K_1+K_2), otherwise `nb` >= `ny`.

        (`errno` 7)
            The orders vector for the filtering model is invalid.

        (`errno` 8)
            The orders vector for the ARIMA model is invalid.

        (`errno` 9)
            The initial values of the filtered series are indeterminate
            for the given models.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_filter_arima'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, y[ny].
        utils._BYREF_INT, # Input, ny.
        utils._EngineIntArrayType.c_type, # Input, mr[nmr].
        utils._BYREF_INT, # Input, nmr.
        utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input, cy.
        utils._EngineFloat64ArrayType.c_type, # Output, b[nb].
        utils._BYREF_INT, # Input, nb.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ny = utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _ny_py = utils._EngineIntScalarType.to_py(_ny)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_ny_py,),
    ).to_ctypes()
    _nmr = utils._EngineIntScalarType(
        nmr, {'fun_name': fun_name, 'entity': 'nmr'},
    ).to_ctypes()
    _nmr_py = utils._EngineIntScalarType.to_py(_nmr)
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(_nmr_py,),
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    ).to_ctypes()
    if _nmr_py == 14:
        if cy is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'cy'},
                'must not be None when nmr = 14',
            )
    if cy is None:
        _cy = utils._EngineFloat64ScalarType.c_type(0.0)
    else:
        _cy = utils._EngineFloat64ScalarType(
            cy, {'fun_name': fun_name, 'entity': 'cy'},
            allow_none=True,
        ).to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    if _nmr_py == 14:
        _b_dim1 = _ny_py+max(utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))+utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))*utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)), utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))*utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))+utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))+utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))*utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)))
    elif _nmr_py == 7:
        _b_dim1 = _ny_py
    else:
        _b_dim1 = 0
    exp_shape_b = (_b_dim1,)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _y,
        _ny,
        _mr,
        _nmr,
        _par,
        _npar,
        _cy,
        _b,
        _nb,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nmr = 7 or 14.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar must be inconsistent with mr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ny = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ny >= max(1+Q\'_y, npar).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if nmr = 14 then nb >= ny+max(K_3, K_1+K_2),',
                    'otherwise nb >= ny.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector for the filtering model is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector for the ARIMA model is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The initial values of the filtered series are',
                    'indeterminate for the given models.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def multi_filter_transf(y, ny, mr, nmr, par, npar, cy, b, nb):
    u"""
    Multivariate time series, filtering by a transfer function model.

    ``multi_filter_transf`` filters a time series by a transfer function
    model.

    For full information please refer to the NAG Library document for
    g13bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bbf.html

    Parameters
    ----------
    y : float, shape (`ny`)
        The Q'_y backforecasts starting with backforecast at time 1-Q'_y
        to backforecast at time 0 followed by the time series starting
        at time 1, where Q'_y = `mr`[5]+`mr`[8]*`mr`[9]. If there are no
        backforecasts either because the ARIMA model for the time series
        is not known or because it is known but has no moving average
        terms, then the time series starts at the beginning of `y`.

    ny : int
        The total number of backforecasts and time series data points in
        array `y`.

    mr : int, shape (`nmr`)
        The orders vector for the filtering transfer function model
        followed by the orders vector for the ARIMA model for the time
        series if the latter is known. The transfer function model
        orders appear in the standard form (b,q,p) as given in [the G13
        Introduction]. Note that if the ARIMA model for the time series
        is supplied then the function will assume that the first Q'_y
        values of the array `y` are backforecasts.

    nmr : int
        The number of values supplied in the array `mr`. It takes the
        value 3 if no ARIMA model for the time series is supplied but
        otherwise it takes the value 10. Thus `nmr` acts as an indicator
        as to whether backforecasting can be carried out.

    par : float, shape (`npar`)
        The parameters of the filtering transfer function model followed
        by the parameters of the ARIMA model for the time series. In the
        transfer function model the parameters are in the standard order
        of MA-like followed by AR-like operator parameters. In the ARIMA
        model the parameters are in the standard order of non-seasonal
        AR and MA followed by seasonal AR and MA.

    npar : int
        The total number of parameters held in array `par`.

    cy : None or float
        If the ARIMA model is known (i.e., `nmr` = 10), `cy` must
        specify the constant term of the ARIMA model for the time
        series. If this model is not known (i.e., `nmr` = 3) then `cy`
        is not used.

    b : float, shape (`nb`), modified in place
        On exit: the filtered output series. If the ARIMA model for the
        time series was known, and hence Q'_y backforecasts were
        supplied in `y`, then `b` contains Q'_y 'filtered' backforecasts
        followed by the filtered series. Otherwise, the filtered series
        begins at the start of `b` just as the original series began at
        the start of `y`. In either case, if the value of the series at
        time t is held in `y`[t-1], then the filtered value at time t is
        held in `b`[t-1].

    nb : int
        In addition to holding the returned filtered series, `b` is also
        used as an intermediate work array if the ARIMA model for the
        time series is known.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nb` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: if `nmr` = 10 then `nb` >= max(b, p, q),
            otherwise `nb` >= `ny`.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` must be inconsistent with `mr`.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*, `mr`[7] = *<value>* and
            `mr`[8] = *<value>*.

            Constraint: `mr`[6]+`mr`[7]+`mr`[8] != 0.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*, `mr`[7] = *<value>* and
            `mr`[8] = *<value>*.

            Constraint: `mr`[6]+`mr`[7]+`mr`[8] = 0.

        (`errno` 1)
            On entry, `mr`[9] = 1.

            Constraint: `mr`[9] != 1.

        (`errno` 1)
            On entry, i = *<value>*, `mr`[i-1] = *<value>*.

            Constraint: `mr`[i-1] >= 0.

        (`errno` 1)
            On entry, `nmr` = *<value>*.

            Constraint: `nmr` = 3 or 10.

        (`errno` 2)
            A supplied model has invalid parameters.

        (`errno` 3)
            The supplied time series is too short.

        (`errno` 4)
            The matrix used to solve for starting values for MA is
            singular.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_filter_transf'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, y[ny].
        utils._BYREF_INT, # Input, ny.
        utils._EngineIntArrayType.c_type, # Input, mr[nmr].
        utils._BYREF_INT, # Input, nmr.
        utils._EngineFloat64ArrayType.c_type, # Input, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._BYREF_FLOAT64, # Input, cy.
        utils._EngineFloat64ArrayType.c_type, # Output, b[nb].
        utils._BYREF_INT, # Input, nb.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ny = utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _ny_py = utils._EngineIntScalarType.to_py(_ny)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_ny_py,),
    ).to_ctypes()
    _nmr = utils._EngineIntScalarType(
        nmr, {'fun_name': fun_name, 'entity': 'nmr'},
    ).to_ctypes()
    _nmr_py = utils._EngineIntScalarType.to_py(_nmr)
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(_nmr_py,),
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    ).to_ctypes()
    if _nmr_py == 10:
        if cy is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'cy'},
                'must not be None when nmr = 10',
            )
    if cy is None:
        _cy = utils._EngineFloat64ScalarType.c_type(0.0)
    else:
        _cy = utils._EngineFloat64ScalarType(
            cy, {'fun_name': fun_name, 'entity': 'cy'},
            allow_none=True,
        ).to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    if _nmr_py == 3:
        _b_dim1 = _ny_py
    elif _nmr_py == 10:
        _b_dim1 = _ny_py+max(utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,)), utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,)))
    else:
        _b_dim1 = 0
    exp_shape_b = (_b_dim1,)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _y,
        _ny,
        _mr,
        _nmr,
        _par,
        _npar,
        _cy,
        _b,
        _nb,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if nmr = 10 then nb >= max(b, p, q),',
                    'otherwise nb >= ny.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar must be inconsistent with mr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[6] = {:s}, mr[7] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'mr[8] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: mr[6]+mr[7]+mr[8] != 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[6] = {:s}, mr[7] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'mr[8] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: mr[6]+mr[7]+mr[8] = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mr[9] = 1.',
                    'Constraint: mr[9] != 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, mr[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mr[i-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nmr = 3 or 10.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A supplied model has invalid parameters.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The supplied time series is too short.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix used to solve for starting values for MA is singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def multi_xcorr(x, y, nxy, nl, r):
    u"""
    Multivariate time series, cross-correlations.

    ``multi_xcorr`` calculates cross-correlations between two time
    series.

    For full information please refer to the NAG Library document for
    g13bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bcf.html

    Parameters
    ----------
    x : float, shape (`nxy`)
        The n values of the x series.

    y : float, shape (`nxy`)
        The n values of the y series.

    nxy : int
        n, the length of the time series.

    nl : int
        L, the maximum lag for calculating cross-correlations.

    r : float, shape (`nl`), modified in place
        On exit: `r`[l-1] contains the cross-correlations between the x
        and y series at lags L, r_{xy}(l), for l = 1,2,...,L.

    Returns
    -------
    s : float
        The ratio of the standard deviation of the y series to the
        standard deviation of the x series, s_y/s_x.

    r0 : float
        The cross-correlation between the x and y series at lag zero.

    stat : float
        The statistic for testing for absence of cross-correlation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nxy` = *<value>*.

            Constraint: `nxy` > 1.

        (`errno` 1)
            On entry, `nl` = *<value>* and `nxy` = *<value>*.

            Constraint: 1 <= `nl` < `nxy`.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` >= 1.

        (`errno` 2)
            One or both of the x and y series have zero variance.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_xcorr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, x[nxy].
        utils._EngineFloat64ArrayType.c_type, # Input, y[nxy].
        utils._BYREF_INT, # Input, nxy.
        utils._BYREF_INT, # Input, nl.
        utils._BYREF_FLOAT64, # Output, s.
        utils._BYREF_FLOAT64, # Output, r0.
        utils._EngineFloat64ArrayType.c_type, # Output, r[nl].
        utils._BYREF_FLOAT64, # Output, stat.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nxy = utils._EngineIntScalarType(
        nxy, {'fun_name': fun_name, 'entity': 'nxy'},
    ).to_ctypes()
    _nxy_py = utils._EngineIntScalarType.to_py(_nxy)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nxy_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_nxy_py,),
    ).to_ctypes()
    _nl = utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    _r0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _nl_py = utils._EngineIntScalarType.to_py(_nl)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nl_py,),
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _stat = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _nxy,
        _nl,
        _s,
        _r0,
        _r,
        _stat,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 31 and _ifmt == 31:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nxy > 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s} and nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nl < nxy.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nl >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or both of the x and y series have zero variance.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or both of the x and y series have zero variance.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _r_cls.refresh_py(_r)
        _s_py = utils._EngineFloat64ScalarType.to_py(_s)
        _r0_py = utils._EngineFloat64ScalarType.to_py(_r0)
        _stat_py = utils._EngineFloat64ScalarType.to_py(_stat)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['s', 'r0', 'stat'],
        )(
            _s_py,
            _r0_py,
            _stat_py,
        )
    _err_class.manage()
    return return_data

def multi_transf_prelim(r0, r, nl, nna, s, nwds, wds, isf):
    u"""
    Multivariate time series, preliminary estimation of transfer
    function model.

    ``multi_transf_prelim`` calculates preliminary estimates of the
    parameters of a transfer function model.

    For full information please refer to the NAG Library document for
    g13bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bdf.html

    Parameters
    ----------
    r0 : float
        The cross-correlation between the two series at lag 0,
        r_{xy}(0).

    r : float, shape (`nl`)
        The cross-correlations between the two series at lags 1 to L,
        r_{xy}(l), for l = 1,2,...,L.

    nl : int
        L, the number of lagged cross-correlations in the array `r`.

    nna : int, shape (3)
        The transfer function model orders in the standard form b,q,p.

    s : float
        The ratio of the standard deviation of the y series to that of
        the x series, s_y/s_x.

    nwds : int
        The exact number of parameters in the transfer function model.

    wds : float, shape (`nwds`), modified in place
        On exit: the preliminary estimates of the parameters of the
        transfer function model in the order of q+1 MA-like parameters
        followed by the p AR-like parameters. If the estimation of
        either type of parameter fails then these arguments are set to
        0.0.

    isf : int, shape (2), modified in place
        On exit: indicators of the success of the estimation of MA-like
        and AR-like parameters respectively. A value 0 indicates that
        there are no parameters of that type to be estimated. A value of
        1 or -1 indicates that there are parameters of that type in the
        model and the estimation of that type has been successful or
        unsuccessful respectively. Note that there is always at least
        one MA-like parameter in the model.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nwds` = *<value>*, `nna`[1] = *<value>* and
            `nna`[2] = *<value>*.

            Constraint: `nwds` = `nna`[1]+`nna`[2]+1.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` > 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `r`[i-1] = *<value>*.

            Constraint: -1.0 <= `r`[i] <= 1.0.

        (`errno` 1)
            On entry, `r0` = *<value>*.

            Constraint: -1.0 <= `r0` <= 1.0.

        (`errno` 1)
            On entry, `nl` = *<value>*, `nna`[0] = *<value>*, `nna`[1] =
            *<value>* and `nna`[2] = *<value>*.

            Constraint: `nl` >= max(`nna`[0]+`nna`[1]+`nna`[2], 1).

        (`errno` 1)
            On entry, i = *<value>* and `nna`[i-1] = *<value>*.

            Constraint: `nna`[i-1] >= 0.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_transf_prelim'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, r0.
        utils._EngineFloat64ArrayType.c_type, # Input, r[nl].
        utils._BYREF_INT, # Input, nl.
        utils._EngineIntArrayType.c_type, # Input, nna[3].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_INT, # Input, nwds.
        utils._EngineFloat64ArrayType.c_type, # Output, wds[nwds].
        utils._EngineIntArrayType.c_type, # Output, isf[2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _r0 = utils._EngineFloat64ScalarType(
        r0, {'fun_name': fun_name, 'entity': 'r0'},
    ).to_ctypes()
    _nl = utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = utils._EngineIntScalarType.to_py(_nl)
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nl_py,),
    ).to_ctypes()
    _nna = utils._EngineIntArrayType(
        nna, {'fun_name': fun_name, 'entity': 'nna'},
        exp_shape=(3,),
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _nwds = utils._EngineIntScalarType(
        nwds, {'fun_name': fun_name, 'entity': 'nwds'},
    ).to_ctypes()
    _wds_cls = utils._EngineFloat64ArrayType(
        wds, {'fun_name': fun_name, 'entity': 'wds'},
        exp_shape=(utils._EngineIntArrayType.get_element(nna, {'fun_name': fun_name, 'entity': 'nna'}, (1,))+utils._EngineIntArrayType.get_element(nna, {'fun_name': fun_name, 'entity': 'nna'}, (2,))+1,),
        check_contents=False,
    )
    _wds = _wds_cls.to_ctypes(can_reference_contents=False)
    _isf_cls = utils._EngineIntArrayType(
        isf, {'fun_name': fun_name, 'entity': 'isf'},
        exp_shape=(2,),
        check_contents=False,
    )
    _isf = _isf_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _r0,
        _r,
        _nl,
        _nna,
        _s,
        _nwds,
        _wds,
        _isf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nwds = {:s}, nna[1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nna[2] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: nwds = nna[1]+nna[2]+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and r[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: -1.0 <= r[i] <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: -1.0 <= r0 <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}, nna[0] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nna[1] = {:s} and nna[2] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: nl >= max(nna[0]+nna[1]+nna[2], 1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and nna[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nna[i-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _wds_cls.refresh_py(_wds)
        _isf_cls.refresh_py(_isf)
        return_data = None
    _err_class.manage()
    return return_data

def multi_inputmod_estim(mr, nser, mt, para, npara, kfc, nxxy, xxy, kef, nit, kzsp, zsp, sd, cm, kzef, res, sttf, isttf, kpriv, io_manager=None):
    u"""
    Multivariate time series, estimation of multi-input model.

    ``multi_inputmod_estim`` fits a multi-input model relating one
    output series to the input series with a choice of three different
    estimation criteria: nonlinear least squares, exact likelihood and
    marginal likelihood.
    When no input series are present, ``multi_inputmod_estim`` fits a
    univariate ARIMA model.

    For full information please refer to the NAG Library document for
    g13be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bef.html

    Parameters
    ----------
    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing and the
        seasonal period.

    nser : int
        The total number of input and output series. There may be any
        number of input series (including none), but always one output
        series.

    mt : int, shape (4, `nser`)
        The transfer function model orders b, p and q of each of the
        input series. The order parameters for input series i are held
        in column i. Row 1 holds the value b_i, row 2 holds the value
        q_i and row 3 holds the value p_i. For a simple input, b_i = q_i
        = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, r_i
        = 2 for a transfer function input for which no allowance is to
        be made for pre-observation period effects, and r_i = 3 for a
        transfer function input for which pre-observation period effects
        will be treated by estimation of appropriate nuisance
        parameters.

        When r_i = 1, any nonzero contents of rows 1, 2, and 3 of column
        i are ignored.

    para : float, shape (`npara`), modified in place
        On entry: initial values of the multi-input model parameters.
        These are in order, firstly the ARIMA model parameters: p values
        of phi parameters, q values of theta parameters, P values of Phi
        parameters and Q values of Theta parameters. These are followed
        by initial values of the transfer function model parameters
        omega_0,omega_1,...,omega_q_1, delta_1,delta_2,...,delta_p_1 for
        the first of any input series and similarly for each subsequent
        input series. The final component of `para` is the initial value
        of the constant c, whether it is fixed or is to be estimated.

        On exit: the latest values of the estimates of these parameters.

    npara : int
        The exact number of phi,theta,Phi,Theta, omega,delta and c
        parameters.

    kfc : int
        Must be set to 0 if the constant c is to remain fixed at its
        initial value, and 1 if it is to be estimated.

    nxxy : int
        The (common) length of the original, undifferenced input and
        output time series.

    xxy : float, shape (`nxxy`, `nser`), modified in place
        On entry: the columns of `xxy` must contain the `nxxy` original,
        undifferenced values of each of the input series and the output
        series x_t in that order.

        On exit: if `kzef` = 0, `xxy` remains unchanged on exit.

        If `kzef` != 0, the columns of `xxy` hold the corresponding
        values of the input component series z_t in place of x_t and the
        output noise component n_t in place of y_t, in that order.

    kef : int
        Indicates the likelihood option.

        `kef` = 1

          Gives least squares.

        `kef` = 2

          Gives exact likelihood.

        `kef` = 3

          Gives marginal likelihood.

    nit : int
        The maximum required number of iterations.

        `nit` = 0

          No change is made to any of the model parameters in array
          `para` except that the constant c (if `kfc` = 1) and any omega
          relating to simple input series are estimated. (Apart from
          these, estimates are always derived for the nuisance
          parameters relating to any backforecasts and any
          pre-observation period effects for transfer function inputs.)

    kzsp : int
        Must be set to 1 if the function is to use the input values of
        `zsp` in the minimization procedure, and to any other value if
        the default values of `zsp` are to be used.

    zsp : float, shape (4), modified in place
        On entry: if `kzsp` = 1, then `zsp` must contain the four values
        used to control the strategy of the search procedure.

        `zsp`[0]

          Contains alpha, the value used to constrain the magnitude of
          the search procedure steps.

        `zsp`[1]

          Contains beta, the multiplier which regulates the value of
          alpha.

        `zsp`[2]

          Contains delta, the value of the stationarity and
          invertibility test tolerance factor.

        `zsp`[3]

          Contains gamma, the value of the convergence criterion.

        If `kzsp` != 1 before entry, default values of `zsp` are
        supplied by the function.

        These are 0.01, 10.0, 1000.0 and max(100*machine
        precision,0.0000001), respectively.

        On exit: contains the values, default or otherwise, used by the
        function.

    sd : float, shape (`npara`), modified in place
        On exit: the `npara` values of the standard deviations
        corresponding to each of the parameters in `para`. When the
        constant is fixed its standard deviation is returned as zero.
        When the values of `para` are valid, the values of `sd` are
        usually also valid. However, if an exit value of `errno` = 3, 8
        or 10, then the contents of `sd` will be indeterminate.

    cm : float, shape (`npara`, `npara`), modified in place
        On exit: the first `npara` rows and columns of `cm` contain the
        correlation coefficients relating to each pair of parameters in
        `para`. All coefficients relating to the constant will be zero
        if the constant is fixed. The contents of `cm` will be
        indeterminate under the same conditions as `sd`.

    kzef : int
        Must not be set to 0, if the values of the input component
        series z_t and the values of the output noise component n_t are
        to overwrite the contents of `xxy` on exit, and must be set to 0
        if `xxy` is to remain unchanged.

    res : float, shape (`nxxy`), modified in place
        On exit: the values of the residuals relating to the differenced
        values of the output series. The remainder of the first `nxxy`
        terms in the array will be zero.

    sttf : float, shape (`isttf`), modified in place
        On exit: the `nsttf` values of the state set array.

    isttf : int
        The dimension of the array `sttf`.

    kpriv : int
        Must not be set to 0, if it is required to monitor the course of
        the optimization. The course of the optimization is monitored by
        printing out at each iteration the iteration count (`itc`), the
        residual sum of squares (`s`), the objective function (`d`) and
        a description and value for each of the parameters in the `para`
        array. The descriptions are PHI for phi, THETA for theta, SPHI
        for Phi, STHETA for Theta, OMEGA/SI for omega in a simple input,
        OMEGA for omega in a transfer function input, DELTA for delta
        and CONSTANT for c. In addition SERIES 1, SERIES 2, etc.
        indicate the input series relevant to the OMEGA and DELTA
        parameters.

        `kpriv` must be set to 0 if the print-out of the above
        information is not required.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    itc : int
        The number of iterations carried out.

        `itc` = -1

          Indicates that the only estimates obtained up to this point
          have been for the nuisance parameters relating to
          backforecasts, unless the marginal likelihood option is used,
          in which case estimates have also been obtained for simple
          input coefficients omega and for the constant c (if `kfc` =
          1). This value of `itc` usually indicates a failure in a
          consequent step of estimating transfer function input
          pre-observation period nuisance parameters.

        `itc` = 0

          Indicates that estimates have been obtained up to this point
          for the constant c (if `kfc` = 1), for simple input
          coefficients omega and for the nuisance parameters relating to
          the backforecasts and to transfer function input
          pre-observation period effects.

    s : float
        The residual sum of squares, S, at the latest set of valid
        parameter estimates.

    d : float
        The objective function, D, at the latest set of valid parameter
        estimates.

    ndf : int
        The number of degrees of freedom associated with S.

    nsttf : int
        The number of values in the state set array `sttf`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nser` = 1 and there are no parameters in the
            model.

        (`errno` 1)
            On entry, `nser` = *<value>*.

            Constraint: `nser` >= 1.

        (`errno` 1)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 1)
            On entry, `kef` = *<value>*.

            Constraint: `kef` = 1, 2 or 3.

        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 2)
            On entry, `ndf` = *<value>*.

            Constraint: `ndf` > 0.

        (`errno` 2)
            On entry, `npara` = *<value>* and `kfc` = *<value>*.

            Constraint: `npara`, `kfc`, `mr` and `mt` must be
            consistent.

        (`errno` 2)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

        (`errno` 2)
            The orders vector `mr` is invalid.

        (`errno` 3)
            On entry, or during execution, one or more sets of delta
            parameters do not satisfy the stationarity or invertibility
            conditions.

        (`errno` 4)
            On entry, `zsp`[3] = *<value>*.

            Constraint: 0.0 <= `zsp`[3] < 1.0.

        (`errno` 4)
            On entry, `zsp`[2] = *<value>*.

            Constraint: `zsp`[2] >= 1.0.

        (`errno` 4)
            On entry, `zsp`[1] = *<value>*.

            Constraint: `zsp`[1] > 1.0.

        (`errno` 4)
            On entry, `zsp`[0] = *<value>*.

            Constraint: `zsp`[0] > 0.0.

        (`errno` 8)
            Unable to calculate the latest parameter estimates.

        (`errno` 9)
            Failure in inversion of second derivative matrix.

        (`errno` 10)
            On entry, or during execution, one or more sets of ARIMA
            parameters do not satisfy the stationarity or invertibility
            conditions.

        (`errno` 11)
            On entry, `isttf` too small.

            Constraint: `isttf` >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncg.

        (`errno` 12)
            The routine has failed to converge after *<value>*
            iterations.

        (`errno` 13)
            On entry, `nit` = *<value>*.

            `isttf` is too small and the process failed to converge
            after `nit` iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.multi_inputmod_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_inputmod_estim'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._BYREF_INT, # Input, nser.
        utils._EngineIntArrayType.c_type, # Input, mt[4, nser].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, para[npara].
        utils._BYREF_INT, # Input, npara.
        utils._BYREF_INT, # Input, kfc.
        utils._BYREF_INT, # Input, nxxy.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xxy[ldxxy, nser].
        utils._BYREF_INT, # Input, ldxxy.
        utils._BYREF_INT, # Input, kef.
        utils._BYREF_INT, # Input, nit.
        utils._BYREF_INT, # Input, kzsp.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, zsp[4].
        utils._BYREF_INT, # Output, itc.
        utils._EngineFloat64ArrayType.c_type, # Output, sd[npara].
        utils._EngineFloat64ArrayType.c_type, # Output, cm[ldcm, npara].
        utils._BYREF_INT, # Input, ldcm.
        utils._BYREF_FLOAT64, # Output, s.
        utils._BYREF_FLOAT64, # Output, d.
        utils._BYREF_INT, # Output, ndf.
        utils._BYREF_INT, # Input, kzef.
        utils._EngineFloat64ArrayType.c_type, # Output, res[nxxy].
        utils._EngineFloat64ArrayType.c_type, # Output, sttf[isttf].
        utils._BYREF_INT, # Input, isttf.
        utils._BYREF_INT, # Output, nsttf.
        utils._BYREF_INT, # Input, kpriv.
        utils._BYREF_INT, # Input, nout.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((mt, 'mt'), (xxy, 'xxy'), (cm, 'cm')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _nser = utils._EngineIntScalarType(
        nser, {'fun_name': fun_name, 'entity': 'nser'},
    ).to_ctypes()
    _nser_py = utils._EngineIntScalarType.to_py(_nser)
    _mt = utils._EngineIntArrayType(
        mt, {'fun_name': fun_name, 'entity': 'mt'},
        exp_shape=(4, _nser_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _npara = utils._EngineIntScalarType(
        npara, {'fun_name': fun_name, 'entity': 'npara'},
    ).to_ctypes()
    _npara_py = utils._EngineIntScalarType.to_py(_npara)
    _para_cls = utils._EngineFloat64ArrayType(
        para, {'fun_name': fun_name, 'entity': 'para'},
        exp_shape=(_npara_py,),
    )
    _para = _para_cls.to_ctypes()
    _kfc = utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _nxxy = utils._EngineIntScalarType(
        nxxy, {'fun_name': fun_name, 'entity': 'nxxy'},
    ).to_ctypes()
    _ldxxy_py = utils._EngineFloat64ArrayType.get_nd(xxy, {'fun_name': fun_name, 'entity': 'xxy'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldxxy = utils._EngineIntScalarType(
        _ldxxy_py, {'fun_name': fun_name, 'entity': '_ldxxy'},
    ).to_ctypes()
    _nxxy_py = utils._EngineIntScalarType.to_py(_nxxy)
    _xxy_cls = utils._EngineFloat64ArrayType(
        xxy, {'fun_name': fun_name, 'entity': 'xxy'},
        exp_shape=(_nxxy_py, _nser_py),
        mutual_sorder=_m_sorder,
    )
    _xxy = _xxy_cls.to_ctypes()
    _kef = utils._EngineIntScalarType(
        kef, {'fun_name': fun_name, 'entity': 'kef'},
    ).to_ctypes()
    _nit = utils._EngineIntScalarType(
        nit, {'fun_name': fun_name, 'entity': 'nit'},
    ).to_ctypes()
    _kzsp = utils._EngineIntScalarType(
        kzsp, {'fun_name': fun_name, 'entity': 'kzsp'},
    ).to_ctypes()
    _zsp_cls = utils._EngineFloat64ArrayType(
        zsp, {'fun_name': fun_name, 'entity': 'zsp'},
        exp_shape=(4,),
    )
    _zsp = _zsp_cls.to_ctypes()
    _itc = utils._EngineIntScalarType.empty_ctypes()
    _sd_cls = utils._EngineFloat64ArrayType(
        sd, {'fun_name': fun_name, 'entity': 'sd'},
        exp_shape=(_npara_py,),
        check_contents=False,
    )
    _sd = _sd_cls.to_ctypes(can_reference_contents=False)
    _ldcm_py = utils._EngineFloat64ArrayType.get_nd(cm, {'fun_name': fun_name, 'entity': 'cm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcm = utils._EngineIntScalarType(
        _ldcm_py, {'fun_name': fun_name, 'entity': '_ldcm'},
    ).to_ctypes()
    _cm_cls = utils._EngineFloat64ArrayType(
        cm, {'fun_name': fun_name, 'entity': 'cm'},
        exp_shape=(_npara_py, _npara_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cm = _cm_cls.to_ctypes(can_reference_contents=False)
    _s = utils._EngineFloat64ScalarType.empty_ctypes()
    _d = utils._EngineFloat64ScalarType.empty_ctypes()
    _ndf = utils._EngineIntScalarType.empty_ctypes()
    _kzef = utils._EngineIntScalarType(
        kzef, {'fun_name': fun_name, 'entity': 'kzef'},
    ).to_ctypes()
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_nxxy_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _isttf = utils._EngineIntScalarType(
        isttf, {'fun_name': fun_name, 'entity': 'isttf'},
    ).to_ctypes()
    _vu_p_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))
    _vl_s_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _vl_d_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))
    _vu_d_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))
    _vl_q_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vl_p_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))
    _vu_q_py = utils._EngineIntArrayType.get_element(mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _vl_ncg_py = 0
    for _vl_i_py in range(1, _nser_py+1):
        if utils._EngineIntArrayType.get_element(mt, {'fun_name': fun_name, 'entity': 'mt'}, (3, _vl_i_py-1)) > 1:
            _vl_ncg_py = _vl_ncg_py+utils._EngineIntArrayType.get_element(mt, {'fun_name': fun_name, 'entity': 'mt'}, (0, _vl_i_py-1))+utils._EngineIntArrayType.get_element(mt, {'fun_name': fun_name, 'entity': 'mt'}, (1, _vl_i_py-1))+utils._EngineIntArrayType.get_element(mt, {'fun_name': fun_name, 'entity': 'mt'}, (2, _vl_i_py-1))
    _vl_ncg = utils._EngineIntScalarType(
        _vl_ncg_py, {'fun_name': fun_name, 'entity': '_vl_ncg_py'},
    ).to_ctypes()
    _sttf_cls = utils._EngineFloat64ArrayType(
        sttf, {'fun_name': fun_name, 'entity': 'sttf'},
        exp_shape=((_vu_p_py*_vl_s_py)+_vl_d_py+(_vu_d_py*_vl_s_py)+_vl_q_py+max(_vl_p_py, _vu_q_py*_vl_s_py)+_vl_ncg_py,),
        check_contents=False,
    )
    _sttf = _sttf_cls.to_ctypes(can_reference_contents=False)
    _nsttf = utils._EngineIntScalarType.empty_ctypes()
    _kpriv = utils._EngineIntScalarType(
        kpriv, {'fun_name': fun_name, 'entity': 'kpriv'},
    ).to_ctypes()
    _nout = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _kfc,
        _nxxy,
        _xxy,
        _ldxxy,
        _kef,
        _nit,
        _kzsp,
        _zsp,
        _itc,
        _sd,
        _cm,
        _ldcm,
        _s,
        _d,
        _ndf,
        _kzef,
        _res,
        _sttf,
        _isttf,
        _nsttf,
        _kpriv,
        _nout,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nser = 1 and there are no parameters in the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nser = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nser >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kef = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kef = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kfc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kfc = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ndf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ndf > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npara = {:s} and kfc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: npara, kfc, mr and mt must be consistent.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and mt[3,i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mt[3,i-1] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, or during execution, one or more sets of delta',
                    'parameters do not satisfy the stationarity or invertibility',
                    'conditions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[3] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= zsp[3] < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zsp[2] >= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zsp[1] > 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, zsp[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: zsp[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the latest parameter estimates.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure in inversion of second derivative matrix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, or during execution, one or more sets of ARIMA',
                    'parameters do not satisfy the stationarity or invertibility',
                    'conditions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isttf too small.',
                    'Constraint: isttf >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncg.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99977:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The routine has failed to converge after {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 99976:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'isttf is too small and the process failed to converge after nit iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _para_cls.refresh_py(_para)
        _xxy_cls.refresh_py(_xxy)
        _zsp_cls.refresh_py(_zsp)
        _sd_cls.refresh_py(_sd)
        _cm_cls.refresh_py(_cm)
        _res_cls.refresh_py(_res)
        _sttf_cls.refresh_py(_sttf)
        _itc_py = utils._EngineIntScalarType.to_py(_itc)
        _s_py = utils._EngineFloat64ScalarType.to_py(_s)
        _d_py = utils._EngineFloat64ScalarType.to_py(_d)
        _ndf_py = utils._EngineIntScalarType.to_py(_ndf)
        _nsttf_py = utils._EngineIntScalarType.to_py(_nsttf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itc', 's', 'd', 'ndf', 'nsttf'],
        )(
            _itc_py,
            _s_py,
            _d_py,
            _ndf_py,
            _nsttf_py,
        )
    _err_class.manage()
    return return_data

def multi_inputmod_update(sttf, nsttf, mr, nser, mt, para, npara, nnv, xxyn, kzef, res):
    u"""
    Multivariate time series, update state set for forecasting from
    multi-input model.

    ``multi_inputmod_update`` accepts a series of new observations of an
    output time series and any associated input time series, for which a
    multi-input model is already fully specified, and updates the 'state
    set' information for use in constructing further forecasts.

    The previous specification of the multi-input model will normally
    have been obtained by using ``multi_inputmod_estim`` to estimate the
    relevant transfer function and ARIMA parameters.
    The supplied state set will originally have been produced by
    ``multi_inputmod_estim`` (or possibly ``multi_inputmod_forecast``),
    but may since have been updated by ``multi_inputmod_update``.

    For full information please refer to the NAG Library document for
    g13bg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bgf.html

    Parameters
    ----------
    sttf : float, shape (`nsttf`), modified in place
        On entry: the `nsttf` values in the state set before updating as
        returned by ``multi_inputmod_estim`` or
        ``multi_inputmod_forecast``, or a previous call to
        ``multi_inputmod_update``.

        On exit: the state set values after updating.

    nsttf : int
        The exact number of values in the state set array `sttf` as
        returned by ``multi_inputmod_estim`` or
        ``multi_inputmod_forecast``.

    mr : int, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing, and the
        seasonal period.

    nser : int
        The total number of input and output series. There may be any
        number of input series (including none), but only one output
        series.

    mt : int, shape (4, `nser`)
        The transfer function model orders b, p and q of each of the
        input series. The data for input series i are held in column i.
        Row 1 holds the value b_i, row 2 holds the value q_i and row 3
        holds the value p_i. For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input and
        r_i = 2 or 3 for a transfer function input.

        When r_i = 1 any nonzero contents of rows 1, 2 and 3 of column i
        are ignored.

        The choice of r_i = 2 or r_i = 3 is an option for use in model
        estimation and does not affect the operation of
        ``multi_inputmod_update``.

    para : float, shape (`npara`)
        Estimates of the multi-input model parameters as returned by
        ``multi_inputmod_estim``. These are in order, firstly the ARIMA
        model parameters: p values of phi parameters, q values of theta
        parameters, P values of Phi parameters and Q values of Theta
        parameters. These are followed by the transfer function model
        parameter values omega_0,omega_1,...,omega_q_1,
        delta_1,delta_2,...,delta_p_1 for the first of any input series
        and similarly for each subsequent input series. The final
        component of `para` is the value of the constant c.

    npara : int
        The exact number of phi, theta, Phi, Theta, omega, delta and c
        parameters. (c must be included whether its value was previously
        estimated or was set fixed.)

    nnv : int
        The number of new observation sets being used to update the
        state set, each observation set consisting of a value of the
        output series and the associated values of each of the input
        series at a particular time point.

    xxyn : float, shape (`nnv`, `nser`), modified in place
        On entry: the `nnv` new observation sets being used to update
        the state set. Column i-1 contains the values of input series i,
        for i = 1,2,...,`nser`-1. Column `nser`-1 contains the values of
        the output series. Consecutive rows correspond to increasing
        time sequence.

        On exit: if `kzef` = 0, `xxyn` remains unchanged.

        If `kzef` != 0, the columns of `xxyn` hold the corresponding
        values of the input component series z_t and the output noise
        component n_t in that order.

    kzef : int
        Must not be set to 0, if the values of the input component
        series z_t and the values of the output noise component n_t are
        to overwrite the contents of `xxyn` on exit, and must be set to
        0 if `xxyn` is to remain unchanged on exit.

    res : float, shape (`nnv`), modified in place
        On exit: the values of the residual series a_t corresponding to
        the new observations of the output series.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nsttf` = *<value>*.

            Constraint: `nsttf`, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, `npara` = *<value>*.

            Constraint: `npara`, `mr` and `mt` must be consistent.

        (`errno` 5)
            On entry, the orders vector `mr` is invalid.

        (`errno` 5)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_inputmod_update'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13bgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, sttf[nsttf].
        utils._BYREF_INT, # Input, nsttf.
        utils._EngineIntArrayType.c_type, # Input, mr[7].
        utils._BYREF_INT, # Input, nser.
        utils._EngineIntArrayType.c_type, # Input, mt[4, nser].
        utils._EngineFloat64ArrayType.c_type, # Input, para[npara].
        utils._BYREF_INT, # Input, npara.
        utils._BYREF_INT, # Input, nnv.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xxyn[ldxxyn, nser].
        utils._BYREF_INT, # Input, ldxxyn.
        utils._BYREF_INT, # Input, kzef.
        utils._EngineFloat64ArrayType.c_type, # Output, res[nnv].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((mt, 'mt'), (xxyn, 'xxyn')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nsttf = utils._EngineIntScalarType(
        nsttf, {'fun_name': fun_name, 'entity': 'nsttf'},
    ).to_ctypes()
    _nsttf_py = utils._EngineIntScalarType.to_py(_nsttf)
    _sttf_cls = utils._EngineFloat64ArrayType(
        sttf, {'fun_name': fun_name, 'entity': 'sttf'},
        exp_shape=(_nsttf_py,),
    )
    _sttf = _sttf_cls.to_ctypes()
    _mr = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    ).to_ctypes()
    _nser = utils._EngineIntScalarType(
        nser, {'fun_name': fun_name, 'entity': 'nser'},
    ).to_ctypes()
    _nser_py = utils._EngineIntScalarType.to_py(_nser)
    _mt = utils._EngineIntArrayType(
        mt, {'fun_name': fun_name, 'entity': 'mt'},
        exp_shape=(4, _nser_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _npara = utils._EngineIntScalarType(
        npara, {'fun_name': fun_name, 'entity': 'npara'},
    ).to_ctypes()
    _npara_py = utils._EngineIntScalarType.to_py(_npara)
    _para = utils._EngineFloat64ArrayType(
        para, {'fun_name': fun_name, 'entity': 'para'},
        exp_shape=(_npara_py,),
    ).to_ctypes()
    _nnv = utils._EngineIntScalarType(
        nnv, {'fun_name': fun_name, 'entity': 'nnv'},
    ).to_ctypes()
    _ldxxyn_py = utils._EngineFloat64ArrayType.get_nd(xxyn, {'fun_name': fun_name, 'entity': 'xxyn'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldxxyn = utils._EngineIntScalarType(
        _ldxxyn_py, {'fun_name': fun_name, 'entity': '_ldxxyn'},
    ).to_ctypes()
    _nnv_py = utils._EngineIntScalarType.to_py(_nnv)
    _xxyn_cls = utils._EngineFloat64ArrayType(
        xxyn, {'fun_name': fun_name, 'entity': 'xxyn'},
        exp_shape=(_nnv_py, _nser_py),
        mutual_sorder=_m_sorder,
    )
    _xxyn = _xxyn_cls.to_ctypes()
    _kzef = utils._EngineIntScalarType(
        kzef, {'fun_name': fun_name, 'entity': 'kzef'},
    ).to_ctypes()
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_nnv_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sttf,
        _nsttf,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _nnv,
        _xxyn,
        _ldxxyn,
        _kzef,
        _res,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nsttf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nsttf, mr and mt must be consistent.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npara = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npara, mr and mt must be consistent.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and mt[3,i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mt[3,i-1] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sttf_cls.refresh_py(_sttf)
        _xxyn_cls.refresh_py(_xxyn)
        _res_cls.refresh_py(_res)
        return_data = None
    _err_class.manage()
    return return_data

def multi_inputmod_forecast_state(sttf, nsttf, mr, nser, mt, para, npara, nfv, xxyn, mrx, parx, rmsxy, kzef, fva, fsd):
    u"""
    Multivariate time series, forecasting from state set of multi-input
    model.

    ``multi_inputmod_forecast_state`` produces forecasts of a time
    series (the output series) which depends on one or more other
    (input) series via a multi-input model which will usually have been
    fitted using ``multi_inputmod_estim``.
    The future values of the input series must be supplied.
    The original observations are not required.
    ``multi_inputmod_forecast_state`` uses as input either the original
    state set obtained from ``multi_inputmod_estim``, or the state set
    updated by a series of new observations from
    ``multi_inputmod_update``.
    Standard errors of the forecasts are produced.
    If future values of some of the input series have been obtained as
    forecasts using ARIMA models for those series, this may be allowed
    for in the calculation of the standard errors.

    For full information please refer to the NAG Library document for
    g13bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bhf.html

    Parameters
    ----------
    sttf : float, shape (`nsttf`)
        The `nsttf` values in the state set as returned by
        ``multi_inputmod_estim`` or ``multi_inputmod_update``.

    nsttf : int
        The exact number of values in the state set array `sttf` as
        returned by ``multi_inputmod_estim`` or
        ``multi_inputmod_update``.

    mr : int, shape (7), modified in place
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q give respectively the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing, and the
        seasonal period.

    nser : int
        The total number of input and output series. There may be any
        number of input series (including none), but only one output
        series.

    mt : int, shape (4, `nser`)
        The transfer function orders b, p and q of each of the input
        series. The data for input series i are held in column i. Row 1
        holds the value b_i, row 2 holds the value q_i and row 3 holds
        the value p_i. For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, r_i
        = 2 or 3 for a transfer function input.

        When r_i = 1, any nonzero contents of rows 1, 2 and 3 of column
        i are ignored.

        The choice of r_i = 2 or r_i = 3 is an option for use in model
        estimation and does not affect the operation of
        ``multi_inputmod_forecast_state``.

    para : float, shape (`npara`)
        Estimates of the multi-input model parameters as returned by
        ``multi_inputmod_estim``. These are in order, firstly the ARIMA
        model parameters: p values of phi parameters, q values of theta
        parameters, P values of Phi parameters and Q values of Theta
        parameters. These are followed by the transfer function model
        parameter values omega_0,omega_1,...,omega_q_1,
        delta_1,delta_2,...,delta_p_1 for the first of any input series
        and similar sets of values for any subsequent input series. The
        final component of `para` is the constant c.

    npara : int
        The exact number of phi, theta, Phi, Theta, omega, delta and c
        parameters. (c must be included, whether its value was
        previously estimated or was set fixed).

    nfv : int
        The number of forecast values required.

    xxyn : float, shape (`nfv`, `nser`), modified in place
        On entry: the supplied `nfv` values for each of the input series
        required to produce the `nfv` output series forecasts. Column i
        contains the values for input series i. Column `nser` need not
        be supplied.

        On exit: if `kzef` = 0, then column `nser` of `xxyn` contains
        the output series forecast values (as does `fva`), but `xxyn` is
        otherwise unchanged.

        If `kzef` != 0, then the columns of `xxyn` hold the
        corresponding values of the forecast components z_t for each of
        the input series and the values of the output noise component
        n_t in that order.

    mrx : int, shape (7, `nser`), modified in place
        On entry: the orders array for each of the input series ARIMA
        models. Thus, column i contains values of p, d, q, P, D, Q, s
        for input series i. In the case of those inputs for which no
        ARIMA model is available, the corresponding orders should be set
        to 0. (The model for any input series only affects the standard
        errors of the forecast values.)

        On exit: unchanged, apart from column `nser` which is used for
        workspace.

    parx : float, shape (ldparx, `nser`)
        Values of the parameters (phi, theta, Phi and Theta) for each of
        the input series ARIMA models. Thus column i contains
        `mrx`[0,i-1] values of phi parameters, `mrx`[2,i-1] values of
        theta parameters, `mrx`[3,i-1] values of Phi parameters and
        `mrx`[5,i-1] values of Theta parameters -- in that order.

        Values in the columns relating to those input series for which
        no ARIMA model is available are ignored. (The model for any
        input series only affects the standard errors of the forecast
        values.)

    rmsxy : float, shape (`nser`)
        The estimated residual variances for each input series ARIMA
        model followed by that for the output noise ARIMA model. In the
        case of those inputs for which no ARIMA model is available, or
        when its effects are to be excluded in the calculation of
        forecast standard errors, the corresponding entry of `rmsxy`
        should be set to 0.

    kzef : int
        Must not be set to 0, if the values of the input component
        series z_t and the values of the output noise component n_t are
        to overwrite the contents of `xxyn` on exit, and must be set to
        0 if `xxyn` is to remain unchanged on exit, apart from the
        appearance of the forecast values in column `nser`.

    fva : float, shape (`nfv`), modified in place
        On exit: the required forecast values for the output series.

    fsd : float, shape (`nfv`), modified in place
        On exit: the standard errors for each of the forecast values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nsttf` = *<value>* and the expected value =
            *<value>*.

            Constraint: `nsttf`, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, `npara` = *<value>* and the expected value =
            *<value>*.

            Constraint: `npara`, `mr` and `mt` must be consistent.

        (`errno` 6)
            The orders vector `mr` is invalid.

        (`errno` 6)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.multi_inputmod_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_inputmod_forecast_state'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, sttf[nsttf].
        utils._BYREF_INT, # Input, nsttf.
        utils._EngineIntArrayType.c_type, # Input/Output, mr[7].
        utils._BYREF_INT, # Input, nser.
        utils._EngineIntArrayType.c_type, # Input, mt[4, nser].
        utils._EngineFloat64ArrayType.c_type, # Input, para[npara].
        utils._BYREF_INT, # Input, npara.
        utils._BYREF_INT, # Input, nfv.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xxyn[ldxxyn, nser].
        utils._BYREF_INT, # Input, ldxxyn.
        utils._EngineIntArrayType.c_type, # Input/Output, mrx[7, nser].
        utils._EngineFloat64ArrayType.c_type, # Input, parx[ldparx, nser].
        utils._BYREF_INT, # Input, ldparx.
        utils._EngineFloat64ArrayType.c_type, # Input, rmsxy[nser].
        utils._BYREF_INT, # Input, kzef.
        utils._EngineFloat64ArrayType.c_type, # Output, fva[nfv].
        utils._EngineFloat64ArrayType.c_type, # Output, fsd[nfv].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((mt, 'mt'), (xxyn, 'xxyn'), (mrx, 'mrx'), (parx, 'parx')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nsttf = utils._EngineIntScalarType(
        nsttf, {'fun_name': fun_name, 'entity': 'nsttf'},
    ).to_ctypes()
    _nsttf_py = utils._EngineIntScalarType.to_py(_nsttf)
    _sttf = utils._EngineFloat64ArrayType(
        sttf, {'fun_name': fun_name, 'entity': 'sttf'},
        exp_shape=(_nsttf_py,),
    ).to_ctypes()
    _mr_cls = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    )
    _mr = _mr_cls.to_ctypes()
    _nser = utils._EngineIntScalarType(
        nser, {'fun_name': fun_name, 'entity': 'nser'},
    ).to_ctypes()
    _nser_py = utils._EngineIntScalarType.to_py(_nser)
    _mt = utils._EngineIntArrayType(
        mt, {'fun_name': fun_name, 'entity': 'mt'},
        exp_shape=(4, _nser_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _npara = utils._EngineIntScalarType(
        npara, {'fun_name': fun_name, 'entity': 'npara'},
    ).to_ctypes()
    _npara_py = utils._EngineIntScalarType.to_py(_npara)
    _para = utils._EngineFloat64ArrayType(
        para, {'fun_name': fun_name, 'entity': 'para'},
        exp_shape=(_npara_py,),
    ).to_ctypes()
    _nfv = utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _ldxxyn_py = utils._EngineFloat64ArrayType.get_nd(xxyn, {'fun_name': fun_name, 'entity': 'xxyn'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldxxyn = utils._EngineIntScalarType(
        _ldxxyn_py, {'fun_name': fun_name, 'entity': '_ldxxyn'},
    ).to_ctypes()
    _nfv_py = utils._EngineIntScalarType.to_py(_nfv)
    _xxyn_cls = utils._EngineFloat64ArrayType(
        xxyn, {'fun_name': fun_name, 'entity': 'xxyn'},
        exp_shape=(_nfv_py, _nser_py),
        mutual_sorder=_m_sorder,
    )
    _xxyn = _xxyn_cls.to_ctypes()
    _mrx_cls = utils._EngineIntArrayType(
        mrx, {'fun_name': fun_name, 'entity': 'mrx'},
        exp_shape=(7, _nser_py),
        mutual_sorder=_m_sorder,
    )
    _mrx = _mrx_cls.to_ctypes()
    _ldparx_py = utils._EngineFloat64ArrayType.get_shape(parx, {'fun_name': fun_name, 'entity': 'parx'}, exp_rank=2)[0]
    _ldparx = utils._EngineIntScalarType(
        _ldparx_py, {'fun_name': fun_name, 'entity': '_ldparx'},
    ).to_ctypes()
    _parx = utils._EngineFloat64ArrayType(
        parx, {'fun_name': fun_name, 'entity': 'parx'},
        exp_shape=(_ldparx_py, _nser_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _rmsxy = utils._EngineFloat64ArrayType(
        rmsxy, {'fun_name': fun_name, 'entity': 'rmsxy'},
        exp_shape=(_nser_py,),
    ).to_ctypes()
    _kzef = utils._EngineIntScalarType(
        kzef, {'fun_name': fun_name, 'entity': 'kzef'},
    ).to_ctypes()
    _fva_cls = utils._EngineFloat64ArrayType(
        fva, {'fun_name': fun_name, 'entity': 'fva'},
        exp_shape=(_nfv_py,),
        check_contents=False,
    )
    _fva = _fva_cls.to_ctypes(can_reference_contents=False)
    _fsd_cls = utils._EngineFloat64ArrayType(
        fsd, {'fun_name': fun_name, 'entity': 'fsd'},
        exp_shape=(_nfv_py,),
        check_contents=False,
    )
    _fsd = _fsd_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sttf,
        _nsttf,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _nfv,
        _xxyn,
        _ldxxyn,
        _mrx,
        _parx,
        _ldparx,
        _rmsxy,
        _kzef,
        _fva,
        _fsd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 89998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nsttf = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the expected value = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nsttf, mr and mt must be consistent.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npara = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the expected value = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: npara, mr and mt must be consistent.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and mt[3,i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mt[3,i-1] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _mr_cls.refresh_py(_mr)
        _xxyn_cls.refresh_py(_xxyn)
        _mrx_cls.refresh_py(_mrx)
        _fva_cls.refresh_py(_fva)
        _fsd_cls.refresh_py(_fsd)
        return_data = None
    _err_class.manage()
    return return_data

def multi_inputmod_forecast(mr, nser, mt, para, npara, kfc, nev, nfv, xxy, kzef, rmsxy, mrx, parx, fva, fsd, sttf, isttf, io_manager=None):
    u"""
    Multivariate time series, state set and forecasts from fully
    specified multi-input model.

    ``multi_inputmod_forecast`` produces forecasts of a time series (the
    output series) which depends on one or more other (input) series via
    a previously estimated multi-input model for which the state set
    information is not available.
    The future values of the input series must be supplied.
    In contrast with ``multi_inputmod_forecast_state`` the original past
    values of the input and output series are required.
    Standard errors of the forecasts are produced.
    If future values of some of the input series have been obtained as
    forecasts using ARIMA models for those series, this may be allowed
    for in the calculation of the standard errors.

    For full information please refer to the NAG Library document for
    g13bj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bjf.html

    Parameters
    ----------
    mr : int, shape (7), modified in place
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing and the
        seasonal period.

    nser : int
        The number of input and output series. There may be any number
        of input series (including none), but only one output series.

    mt : int, shape (4, `nser`)
        The transfer function model orders b, p and q of each of the
        input series. The data for input series i is held in column i.
        Row 1 holds the value b_i, row 2 holds the value q_i and row 3
        holds the value p_i.

        For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, and
        r_i = 2 or 3 for a transfer function input.

        The choice r_i = 3 leads to estimation of the pre-period input
        effects as nuisance parameters, and r_i = 2 suppresses this
        estimation.

        This choice may affect the returned forecasts and the state set.

        When r_i = 1, any nonzero contents of rows 1, 2 and 3 of column
        i are ignored.

    para : float, shape (`npara`), modified in place
        On entry: estimates of the multi-input model parameters. These
        are in order, firstly the ARIMA model parameters: p values of
        phi parameters, q values of theta parameters, P values of Phi
        parameters, Q values of Theta parameters.

        These are followed by the transfer function model parameter
        values omega_0,omega_1,...,omega_q_1, delta_1,...,delta_p_1 for
        the first of any input series and similarly for each subsequent
        input series.

        The final component of `para` is the value of the constant c.

        On exit: the parameter values may be updated using an additional
        iteration in the estimation process.

    npara : int
        The exact number of phi, theta, Phi, Theta, omega, delta, c
        parameters, so that `npara` = p+q+P+Q+`nser`+sum(p+q), the
        summation being over all the input series. (c must be included
        whether its value was previously estimated or was set fixed.)

    kfc : int
        Must be set to 1 if the constant was estimated when the model
        was fitted, and 0 if it was held at a fixed value. This only
        affects the degrees of freedom used in calculating the estimated
        residual variance.

    nev : int
        The number of original (undifferenced) values in each of the
        input and output time series.

    nfv : int
        The number of forecast values of the output series required.

    xxy : float, shape ((`nev`+`nfv`), `nser`), modified in place
        On entry: the columns of `xxy` must contain in the first `nev`
        places, the past values of each of the input and output series,
        in that order. In the next `nfv` places, the columns relating to
        the input series (i.e., columns 1 to `nser`-1) contain the
        future values of the input series which are necessary for
        construction of the forecasts of the output series y.

        On exit: if `kzef` = 0 then `xxy` is unchanged except that the
        relevant `nfv` values in the column relating to the output
        series (column `nser`) contain the forecast values (`fva`), but
        if `kzef` != 0 then the columns of `xxy` contain the
        corresponding values of the input component series z_t and the
        values of the output noise component n_t, in that order.

    kzef : int
        Must be set to 0 if the relevant `nfv` values of the forecasts
        (`fva`) are to be held in the output series column (`nser`) of
        `xxy` (which is otherwise unchanged) on exit, and must not be
        set to 0 if the values of the input component series z_t and the
        values of the output noise component n_t are to overwrite the
        contents of `xxy` on exit.

    rmsxy : float, shape (`nser`), modified in place
        On entry: the first (`nser`-1) elements of `rmsxy` must contain
        the estimated residual variance of the input series ARIMA
        models. In the case of those inputs for which no ARIMA model is
        available or its effects are to be excluded in the calculation
        of forecast standard errors, the corresponding entry of `rmsxy`
        should be set to 0.

        On exit: `rmsxy`[`nser`-1] contains the estimated residual
        variance of the output noise ARIMA model which is calculated
        from the supplied series. Otherwise `rmsxy` is unchanged.

    mrx : int, shape (7, `nser`), modified in place
        On entry: the orders array for each of the input series ARIMA
        models. Thus, column i contains values of p, d, q, P, D, Q, s
        for input series i. In the case of those inputs for which no
        ARIMA model is available, the corresponding orders should be set
        to 0.

        On exit: unchanged, except for column `nser` which is used as
        workspace.

    parx : float, shape (ldparx, `nser`)
        Values of the parameters (phi, theta, Phi, and Theta) for each
        of the input series ARIMA models.

        Thus column i contains `mrx`[0,i-1] values of phi, `mrx`[2,i-1]
        values of theta, `mrx`[3,i-1] values of Phi and `mrx`[5,i-1]
        values of Theta, in that order.

        Values in the columns relating to those input series for which
        no ARIMA model is available are ignored.

    fva : float, shape (`nfv`), modified in place
        On exit: the required forecast values for the output series.
        (Note that these are also output in column `nser` of `xxy` if
        `kzef` = 0.)

    fsd : float, shape (`nfv`), modified in place
        On exit: the standard errors for each of the forecast values.

    sttf : float, shape (`isttf`), modified in place
        On exit: the `nsttf` values of the state set based on the first
        `nev` sets of (past) values of the input and output series.

    isttf : int
        The dimension of the array `sttf`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nsttf : int
        The number of values in the state set array `sttf`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nfv` = *<value>*.

            Constraint: `nfv` > 0.

        (`errno` 1)
            On entry, ldxxy = *<value>*, `nev` = *<value>* and `nfv` =
            *<value>*.

            Constraint: ldxxy >= `nev`+`nfv`.

        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 2)
            The orders vector `mr` is invalid.

        (`errno` 2)
            `npara` is inconsistent with `mr` and `mt`.

        (`errno` 2)
            Insufficient degrees of freedom to solve the problem.

        (`errno` 2)
            On entry, `npara` = *<value>* and the expected value =
            *<value>*.

            Constraint: `npara`, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

        (`errno` 3)
            One or more sets of delta parameters do not satisfy the
            stationarity or invertibility conditions.

        (`errno` 8)
            Unable to calculate the latest parameter estimates.

        (`errno` 9)
            Failure in inversion of second derivative matrix.

        (`errno` 10)
            One or more sets of ARIMA parameters do not satisfy the
            stationarity or invertibility conditions.

        (`errno` 11)
            On entry, `isttf` = *<value>*.

            Constraint: `isttf` >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncf.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_inputmod_forecast'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13bjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._EngineIntArrayType.c_type, # Input/Output, mr[7].
        utils._BYREF_INT, # Input, nser.
        utils._EngineIntArrayType.c_type, # Input, mt[4, nser].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, para[npara].
        utils._BYREF_INT, # Input, npara.
        utils._BYREF_INT, # Input, kfc.
        utils._BYREF_INT, # Input, nev.
        utils._BYREF_INT, # Input, nfv.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xxy[ldxxy, nser].
        utils._BYREF_INT, # Input, ldxxy.
        utils._BYREF_INT, # Input, kzef.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rmsxy[nser].
        utils._EngineIntArrayType.c_type, # Input/Output, mrx[7, nser].
        utils._EngineFloat64ArrayType.c_type, # Input, parx[ldparx, nser].
        utils._BYREF_INT, # Input, ldparx.
        utils._EngineFloat64ArrayType.c_type, # Output, fva[nfv].
        utils._EngineFloat64ArrayType.c_type, # Output, fsd[nfv].
        utils._EngineFloat64ArrayType.c_type, # Output, sttf[isttf].
        utils._BYREF_INT, # Input, isttf.
        utils._BYREF_INT, # Output, nsttf.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((mt, 'mt'), (xxy, 'xxy'), (mrx, 'mrx'), (parx, 'parx')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _mr_cls = utils._EngineIntArrayType(
        mr, {'fun_name': fun_name, 'entity': 'mr'},
        exp_shape=(7,),
    )
    _mr = _mr_cls.to_ctypes()
    _nser = utils._EngineIntScalarType(
        nser, {'fun_name': fun_name, 'entity': 'nser'},
    ).to_ctypes()
    _nser_py = utils._EngineIntScalarType.to_py(_nser)
    _mt = utils._EngineIntArrayType(
        mt, {'fun_name': fun_name, 'entity': 'mt'},
        exp_shape=(4, _nser_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _npara = utils._EngineIntScalarType(
        npara, {'fun_name': fun_name, 'entity': 'npara'},
    ).to_ctypes()
    _npara_py = utils._EngineIntScalarType.to_py(_npara)
    _para_cls = utils._EngineFloat64ArrayType(
        para, {'fun_name': fun_name, 'entity': 'para'},
        exp_shape=(_npara_py,),
    )
    _para = _para_cls.to_ctypes()
    _kfc = utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _nev = utils._EngineIntScalarType(
        nev, {'fun_name': fun_name, 'entity': 'nev'},
    ).to_ctypes()
    _nfv = utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _ldxxy_py = utils._EngineFloat64ArrayType.get_nd(xxy, {'fun_name': fun_name, 'entity': 'xxy'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldxxy = utils._EngineIntScalarType(
        _ldxxy_py, {'fun_name': fun_name, 'entity': '_ldxxy'},
    ).to_ctypes()
    _nev_py = utils._EngineIntScalarType.to_py(_nev)
    _nfv_py = utils._EngineIntScalarType.to_py(_nfv)
    _xxy_cls = utils._EngineFloat64ArrayType(
        xxy, {'fun_name': fun_name, 'entity': 'xxy'},
        exp_shape=((_nev_py+_nfv_py), _nser_py),
        mutual_sorder=_m_sorder,
    )
    _xxy = _xxy_cls.to_ctypes()
    _kzef = utils._EngineIntScalarType(
        kzef, {'fun_name': fun_name, 'entity': 'kzef'},
    ).to_ctypes()
    _rmsxy_cls = utils._EngineFloat64ArrayType(
        rmsxy, {'fun_name': fun_name, 'entity': 'rmsxy'},
        exp_shape=(_nser_py,),
    )
    _rmsxy = _rmsxy_cls.to_ctypes()
    _mrx_cls = utils._EngineIntArrayType(
        mrx, {'fun_name': fun_name, 'entity': 'mrx'},
        exp_shape=(7, _nser_py),
        mutual_sorder=_m_sorder,
    )
    _mrx = _mrx_cls.to_ctypes()
    _ldparx_py = utils._EngineFloat64ArrayType.get_shape(parx, {'fun_name': fun_name, 'entity': 'parx'}, exp_rank=2)[0]
    _ldparx = utils._EngineIntScalarType(
        _ldparx_py, {'fun_name': fun_name, 'entity': '_ldparx'},
    ).to_ctypes()
    _parx = utils._EngineFloat64ArrayType(
        parx, {'fun_name': fun_name, 'entity': 'parx'},
        exp_shape=(_ldparx_py, _nser_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _fva_cls = utils._EngineFloat64ArrayType(
        fva, {'fun_name': fun_name, 'entity': 'fva'},
        exp_shape=(_nfv_py,),
        check_contents=False,
    )
    _fva = _fva_cls.to_ctypes(can_reference_contents=False)
    _fsd_cls = utils._EngineFloat64ArrayType(
        fsd, {'fun_name': fun_name, 'entity': 'fsd'},
        exp_shape=(_nfv_py,),
        check_contents=False,
    )
    _fsd = _fsd_cls.to_ctypes(can_reference_contents=False)
    _isttf = utils._EngineIntScalarType(
        isttf, {'fun_name': fun_name, 'entity': 'isttf'},
    ).to_ctypes()
    _isttf_py = utils._EngineIntScalarType.to_py(_isttf)
    _sttf_cls = utils._EngineFloat64ArrayType(
        sttf, {'fun_name': fun_name, 'entity': 'sttf'},
        exp_shape=(_isttf_py,),
        check_contents=False,
    )
    _sttf = _sttf_cls.to_ctypes(can_reference_contents=False)
    _nsttf = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _kfc,
        _nev,
        _nfv,
        _xxy,
        _ldxxy,
        _kzef,
        _rmsxy,
        _mrx,
        _parx,
        _ldparx,
        _fva,
        _fsd,
        _sttf,
        _isttf,
        _nsttf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfv = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nfv > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldxxy = {:s}, nev = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nfv = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ldxxy >= nev+nfv.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kfc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kfc = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 89998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The orders vector mr is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'npara is inconsistent with mr and mt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Insufficient degrees of freedom to solve the problem.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npara = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the expected value = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: npara, mr and mt must be consistent.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and mt[3,i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mt[3,i-1] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or more sets of delta parameters do not satisfy the',
                    'stationarity or invertibility conditions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the latest parameter estimates.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure in inversion of second derivative matrix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or more sets of ARIMA parameters do not satisfy the',
                    'stationarity or invertibility conditions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isttf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isttf >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncf.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _mr_cls.refresh_py(_mr)
        _para_cls.refresh_py(_para)
        _xxy_cls.refresh_py(_xxy)
        _rmsxy_cls.refresh_py(_rmsxy)
        _mrx_cls.refresh_py(_mrx)
        _fva_cls.refresh_py(_fva)
        _fsd_cls.refresh_py(_fsd)
        _sttf_cls.refresh_py(_sttf)
        _nsttf_py = utils._EngineIntScalarType.to_py(_nsttf)
        return_data = _nsttf_py
    _err_class.manage()
    return return_data

def uni_spectrum_lag(nx, mtx, px, iw, mw, ic, nc, c, kc, l, lg, nxg, xg, stats):
    u"""
    Univariate time series, smoothed sample spectrum using rectangular,
    Bartlett, Tukey or Parzen lag window.

    ``uni_spectrum_lag`` calculates the smoothed sample spectrum of a
    univariate time series using one of four lag windows -- rectangular,
    Bartlett, Tukey or Parzen window.

    For full information please refer to the NAG Library document for
    g13ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13caf.html

    Parameters
    ----------
    nx : int
        n, the length of the time series.

    mtx : int
        If covariances are to be calculated by the function (`ic` = 0),
        `mtx` must specify whether the data are to be initially mean or
        trend corrected.

        `mtx` = 0

          For no correction.

        `mtx` = 1

          For mean correction.

        `mtx` = 2

          For trend correction.

    px : float
        If covariances are to be calculated by the function (`ic` = 0),
        `px` must specify the proportion of the data (totalled over both
        ends) to be initially tapered by the split cosine bell taper.

        If covariances are supplied (`ic` != 0), `px` must specify the
        proportion of data tapered before the supplied covariances were
        calculated and after any mean or trend correction. `px` is
        required for the calculation of output statistics.

        A value of 0.0 implies no tapering.

    iw : int
        The choice of lag window.

        `iw` = 1

          Rectangular.

        `iw` = 2

          Bartlett.

        `iw` = 3

          Tukey.

        `iw` = 4

          Parzen.

    mw : int
        M, the 'cut-off' point of the lag window. Windowed covariances
        at lag M or greater are zero.

    ic : int
        Indicates whether covariances are to be calculated in the
        function or supplied in the call to the function.

        `ic` = 0

          Covariances are to be calculated.

        `ic` != 0

          Covariances are to be supplied.

    nc : int
        The number of covariances to be calculated in the function or
        supplied in the call to the function.

    c : float, shape (`nc`), modified in place
        On entry: if `ic` != 0, `c` must contain the `nc` covariances
        for lags from 0 to (`nc`-1), otherwise `c` need not be set.

        On exit: if `ic` = 0, `c` will contain the `nc` calculated
        covariances.

        If `ic` != 0, the contents of `c` will be unchanged.

    kc : int
        If `ic` = 0, `kc` must specify the order of the fast Fourier
        transform (FFT) used to calculate the covariances.

        If `ic` != 0, that is covariances are supplied, `kc` is not
        used.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        Therefore it is also the order of the FFT used to construct the
        sample spectrum from the covariances.

    lg : int
        Indicates whether unlogged or logged spectral estimates and
        confidence limits are required.

        `lg` = 0

          Unlogged.

        `lg` != 0

          Logged.

    nxg : int
        The dimension of the array `xg`.

    xg : float, shape (`nxg`), modified in place
        On entry: if the covariances are to be calculated, then `xg`
        must contain the `nx` data points. If covariances are supplied,
        `xg` may contain any values.

        On exit: contains the `ng` spectral estimates, f^(omega_i), for
        i = 0,1,...,[L/2] in `xg`[0] to `xg`[`ng`-1] respectively
        (logged if `lg` = 1). The elements `xg`[i-1], for i =
        `ng`+1,...,`nxg` contain 0.0.

    stats : float, shape (4), modified in place
        On exit: four associated statistics. These are the degrees of
        freedom in `stats`[0], the lower and upper 95% confidence limit
        factors in `stats`[1] and `stats`[2] respectively (logged if
        `lg` = 1), and the bandwidth in `stats`[3].

    Returns
    -------
    ng : int
        The number of spectral estimates, [L/2]+1, in `xg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nc` = *<value>* and `nx` = *<value>*.

            Constraint: `nc` <= `nx`.

        (`errno` 1)
            On entry, `nc` = *<value>* and `mw` = *<value>*.

            Constraint: `nc` >= `mw`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nx` = *<value>*.

            Constraint: `mw` <= `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `iw` = *<value>*.

            Constraint: `iw` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` <= 1.0.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 0.0.

        (`errno` 1)
            On entry, `nxg` = *<value>* and `l` = *<value>*.

            Constraint: if `ic` != 0, `nxg` >= `l`.

        (`errno` 1)
            On entry, `nxg` = *<value>*, `kc` = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` = 0, `nxg` >= max(`kc`, `l`).

        (`errno` 1)
            On entry, `ic` = 0 and `mtx` > 2: `mtx` = *<value>*.

        (`errno` 1)
            On entry, `ic` = 0 and `mtx` < 0: `mtx` = *<value>*.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>*, `nx` = *<value>* and `nc` =
            *<value>*.

            Constraint: if `ic` = 0, `kc` >= (`nx`+`nc`).

        (`errno` 3)
            On entry, `l` = *<value>* and `mw` = *<value>*.

            Constraint: `l` >= 2*`mw`-1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            One or more spectral estimates are negative.

            Unlogged spectral estimates are returned in `xg`, and the
            degrees of freedom, unloged confidence limit factors and
            bandwidth in `stats`.

        (`errno` 5)
            The calculation of confidence limit factors has failed.

            Spectral estimates (logged if requested) are returned in
            `xg`, and degrees of freedom and bandwidth in `stats`.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_spectrum_lag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nx.
        utils._BYREF_INT, # Input, mtx.
        utils._BYREF_FLOAT64, # Input, px.
        utils._BYREF_INT, # Input, iw.
        utils._BYREF_INT, # Input, mw.
        utils._BYREF_INT, # Input, ic.
        utils._BYREF_INT, # Input, nc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[nc].
        utils._BYREF_INT, # Input, kc.
        utils._BYREF_INT, # Input, l.
        utils._BYREF_INT, # Input, lg.
        utils._BYREF_INT, # Input, nxg.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xg[nxg].
        utils._BYREF_INT, # Output, ng.
        utils._EngineFloat64ArrayType.c_type, # Output, stats[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _mtx = utils._EngineIntScalarType(
        mtx, {'fun_name': fun_name, 'entity': 'mtx'},
    ).to_ctypes()
    _px = utils._EngineFloat64ScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _iw = utils._EngineIntScalarType(
        iw, {'fun_name': fun_name, 'entity': 'iw'},
    ).to_ctypes()
    _mw = utils._EngineIntScalarType(
        mw, {'fun_name': fun_name, 'entity': 'mw'},
    ).to_ctypes()
    _ic = utils._EngineIntScalarType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
    ).to_ctypes()
    _nc = utils._EngineIntScalarType(
        nc, {'fun_name': fun_name, 'entity': 'nc'},
    ).to_ctypes()
    _nc_py = utils._EngineIntScalarType.to_py(_nc)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_nc_py,),
    )
    _c = _c_cls.to_ctypes()
    _kc = utils._EngineIntScalarType(
        kc, {'fun_name': fun_name, 'entity': 'kc'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _lg = utils._EngineIntScalarType(
        lg, {'fun_name': fun_name, 'entity': 'lg'},
    ).to_ctypes()
    _nxg = utils._EngineIntScalarType(
        nxg, {'fun_name': fun_name, 'entity': 'nxg'},
    ).to_ctypes()
    _nxg_py = utils._EngineIntScalarType.to_py(_nxg)
    _xg_cls = utils._EngineFloat64ArrayType(
        xg, {'fun_name': fun_name, 'entity': 'xg'},
        exp_shape=(_nxg_py,),
    )
    _xg = _xg_cls.to_ctypes()
    _ng = utils._EngineIntScalarType.empty_ctypes()
    _stats_cls = utils._EngineFloat64ArrayType(
        stats, {'fun_name': fun_name, 'entity': 'stats'},
        exp_shape=(4,),
        check_contents=False,
    )
    _stats = _stats_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nx,
        _mtx,
        _px,
        _iw,
        _mw,
        _ic,
        _nc,
        _c,
        _kc,
        _l,
        _lg,
        _nxg,
        _xg,
        _ng,
        _stats,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nc = {:s} and nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nc <= nx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nc = {:s} and mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nc >= mw.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s} and nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mw <= nx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mw >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iw = 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxg = {:s} and l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if ic != 0, nxg >= l.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxg = {:s}, kc = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ic = 0, nxg >= max(kc, l).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ic = 0 and mtx > 2: mtx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ic = 0 and mtx < 0: mtx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kc = {:s}, nx = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ic = 0, kc >= (nx+nc).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s} and mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: l >= 2*mw-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or more spectral estimates are negative.',
                    'Unlogged spectral estimates are returned in xg, and the degrees of',
                    'freedom, unloged confidence limit factors and bandwidth in stats.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The calculation of confidence limit factors has failed.',
                    'Spectral estimates (logged if requested) are returned in xg,',
                    'and degrees of freedom and bandwidth in stats.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _c_cls.refresh_py(_c)
        _xg_cls.refresh_py(_xg)
        _stats_cls.refresh_py(_stats)
        _ng_py = utils._EngineIntScalarType.to_py(_ng)
        return_data = _ng_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ng'],
            )(
                _ng_py,
            )
    _err_class.manage()
    return return_data

def uni_spectrum_daniell(nx, mtx, px, mw, pw, l, kc, lg, xg, stats):
    u"""
    Univariate time series, smoothed sample spectrum using spectral
    smoothing by the trapezium frequency (Daniell) window.

    ``uni_spectrum_daniell`` calculates the smoothed sample spectrum of
    a univariate time series using spectral smoothing by the trapezium
    frequency (Daniell) window.

    For full information please refer to the NAG Library document for
    g13cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cbf.html

    Parameters
    ----------
    nx : int
        n, the length of the time series.

    mtx : int
        Whether the data are to be initially mean or trend corrected.

        `mtx` = 0

          For no correction.

        `mtx` = 1

          For mean correction.

        `mtx` = 2

          For trend correction.

    px : float
        The proportion of the data (totalled over both ends) to be
        initially tapered by the split cosine bell taper. (A value of
        0.0 implies no tapering.)

    mw : int
        The value of M which determines the frequency width of the
        smoothing window as 2*pi/M. A value of n implies no smoothing is
        to be carried out.

    pw : float
        p, the shape parameter of the trapezium frequency window.

        A value of 0.0 gives a triangular window, and a value of 1.0 a
        rectangular window.

        If `mw` = `nx`, `pw` is not used.

    l : int
        L, the frequency division of smoothed spectral estimates as
        2*pi/L.

    kc : int
        K, the order of the fast Fourier transform (FFT) used to
        calculate the spectral estimates.

    lg : int
        Indicates whether unlogged or logged spectral estimates and
        confidence limits are required.

        `lg` = 0

          For unlogged.

        `lg` != 0

          For logged.

    xg : float, shape (`kc`), modified in place
        On entry: the n data points.

        On exit: contains the `ng` spectral estimates f^(omega_i), for i
        = 0,1,...,[L/2], in `xg`[0] to `xg`[`ng`-1] (logged if `lg` !=
        0). The elements `xg`[i-1], for i = `ng`+1,...,`kc`, contain
        0.0.

    stats : float, shape (4), modified in place
        On exit: four associated statistics. These are the degrees of
        freedom in `stats`[0], the lower and upper 95% confidence limit
        factors in `stats`[1] and `stats`[2] respectively (logged if
        `lg` != 0), and the bandwidth in `stats`[3].

    Returns
    -------
    ng : int
        The number of spectral estimates, [L/2]+1, in `xg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `px` = *<value>*, `mw` = *<value>* and `nx` =
            *<value>*.

            Constraint: if `pw` > 1.0, `mw` = `nx`.

        (`errno` 1)
            On entry, `px` = *<value>*, `mw` = *<value>* and `nx` =
            *<value>*.

            Constraint: if `pw` < 0.0, `mw` = `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nx` = *<value>*.

            Constraint: `mw` <= `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` <= 1.0.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 0.0.

        (`errno` 1)
            On entry, `mtx` = *<value>*.

            Constraint: `mtx` <= 2.

        (`errno` 1)
            On entry, `mtx` = *<value>*.

            Constraint: `mtx` >= 0.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>* and `l` = *<value>*.

            Constraint: `kc` must be a multiple of `l`.

        (`errno` 2)
            On entry, `kc` = *<value>* and `nx` = *<value>*.

            Constraint: `kc` >= 2*`nx`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            One or more spectral estimates are negative.

            Unlogged spectral estimates are returned in `xg`, and the
            degrees of freedom, unloged confidence limit factors and
            bandwidth in `stats`.

        (`errno` 5)
            The calculation of confidence limit factors has failed.

            Spectral estimates (logged if requested) are returned in
            `xg`, and degrees of freedom and bandwidth in `stats`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_spectrum_daniell'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nx.
        utils._BYREF_INT, # Input, mtx.
        utils._BYREF_FLOAT64, # Input, px.
        utils._BYREF_INT, # Input, mw.
        utils._BYREF_FLOAT64, # Input, pw.
        utils._BYREF_INT, # Input, l.
        utils._BYREF_INT, # Input, kc.
        utils._BYREF_INT, # Input, lg.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xg[kc].
        utils._BYREF_INT, # Output, ng.
        utils._EngineFloat64ArrayType.c_type, # Output, stats[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _mtx = utils._EngineIntScalarType(
        mtx, {'fun_name': fun_name, 'entity': 'mtx'},
    ).to_ctypes()
    _px = utils._EngineFloat64ScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _mw = utils._EngineIntScalarType(
        mw, {'fun_name': fun_name, 'entity': 'mw'},
    ).to_ctypes()
    _pw = utils._EngineFloat64ScalarType(
        pw, {'fun_name': fun_name, 'entity': 'pw'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _kc = utils._EngineIntScalarType(
        kc, {'fun_name': fun_name, 'entity': 'kc'},
    ).to_ctypes()
    _lg = utils._EngineIntScalarType(
        lg, {'fun_name': fun_name, 'entity': 'lg'},
    ).to_ctypes()
    _kc_py = utils._EngineIntScalarType.to_py(_kc)
    _xg_cls = utils._EngineFloat64ArrayType(
        xg, {'fun_name': fun_name, 'entity': 'xg'},
        exp_shape=(_kc_py,),
    )
    _xg = _xg_cls.to_ctypes()
    _ng = utils._EngineIntScalarType.empty_ctypes()
    _stats_cls = utils._EngineFloat64ArrayType(
        stats, {'fun_name': fun_name, 'entity': 'stats'},
        exp_shape=(4,),
        check_contents=False,
    )
    _stats = _stats_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nx,
        _mtx,
        _px,
        _mw,
        _pw,
        _l,
        _kc,
        _lg,
        _xg,
        _ng,
        _stats,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}, mw = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if pw > 1.0, mw = nx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}, mw = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if pw < 0.0, mw = nx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s} and nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mw <= nx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mw >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, px = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: px >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mtx <= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mtx >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kc = {:s} and l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: kc must be a multiple of l.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kc = {:s} and nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: kc >= 2*nx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'One or more spectral estimates are negative.',
                    'Unlogged spectral estimates are returned in xg, and the degrees of',
                    'freedom, unloged confidence limit factors and bandwidth in stats.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The calculation of confidence limit factors has failed.',
                    'Spectral estimates (logged if requested) are returned in xg,',
                    'and degrees of freedom and bandwidth in stats.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _xg_cls.refresh_py(_xg)
        _stats_cls.refresh_py(_stats)
        _ng_py = utils._EngineIntScalarType.to_py(_ng)
        return_data = _ng_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ng'],
            )(
                _ng_py,
            )
    _err_class.manage()
    return return_data

def multi_spectrum_lag(nxy, mtxy, pxy, iw, mw, ish, ic, nc, cxy, cyx, kc, l, nxyg, xg, yg):
    u"""
    Multivariate time series, smoothed sample cross spectrum using
    rectangular, Bartlett, Tukey or Parzen lag window.

    ``multi_spectrum_lag`` calculates the smoothed sample cross spectrum
    of a bivariate time series using one of four lag windows:
    rectangular, Bartlett, Tukey or Parzen.

    For full information please refer to the NAG Library document for
    g13cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ccf.html

    Parameters
    ----------
    nxy : int
        n, the length of the time series x and y.

    mtxy : int
        If cross-covariances are to be calculated by the function (`ic`
        = 0), `mtxy` must specify whether the data is to be initially
        mean or trend corrected.

        `mtxy` = 0

          For no correction.

        `mtxy` = 1

          For mean correction.

        `mtxy` = 2

          For trend correction.

        If cross-covariances are supplied (`ic` != 0), `mtxy` is not
        used.

    pxy : float
        If cross-covariances are to be calculated by the function (`ic`
        = 0), `pxy` must specify the proportion of the data (totalled
        over both ends) to be initially tapered by the split cosine bell
        taper. A value of 0.0 implies no tapering.

        If cross-covariances are supplied (`ic` != 0), `pxy` is not
        used.

    iw : int
        The choice of lag window.

        `iw` = 1

          Rectangular.

        `iw` = 2

          Bartlett.

        `iw` = 3

          Tukey.

        `iw` = 4

          Parzen.

    mw : int
        M, the 'cut-off' point of the lag window, relative to any
        alignment shift that has been applied. Windowed
        cross-covariances at lags (-`mw`+`ish`) or less, and at lags
        (`mw`+`ish`) or greater are zero.

    ish : int
        S, the alignment shift between the x and y series. If x leads y,
        the shift is positive.

    ic : int
        Indicates whether cross-covariances are to be calculated in the
        function or supplied in the call to the function.

        `ic` = 0

          Cross-covariances are to be calculated.

        `ic` != 0

          Cross-covariances are to be supplied.

    nc : int
        The number of cross-covariances to be calculated in the function
        or supplied in the call to the function.

    cxy : float, shape (`nc`), modified in place
        On entry: if `ic` != 0, `cxy` must contain the `nc`
        cross-covariances between values in the y series and earlier
        values in time in the x series, for lags from 0 to (`nc`-1).

        If `ic` = 0, `cxy` need not be set.

        On exit: if `ic` = 0, `cxy` will contain the `nc` calculated
        cross-covariances.

        If `ic` != 0, the contents of `cxy` will be unchanged.

    cyx : float, shape (`nc`), modified in place
        On entry: if `ic` != 0, `cyx` must contain the `nc`
        cross-covariances between values in the y series and later
        values in time in the x series, for lags from 0 to (`nc`-1).

        If `ic` = 0, `cyx` need not be set.

        On exit: if `ic` = 0, `cyx` will contain the `nc` calculated
        cross-covariances.

        If `ic` != 0, the contents of `cyx` will be unchanged.

    kc : int
        If `ic` = 0, `kc` must specify the order of the fast Fourier
        transform (FFT) used to calculate the cross-covariances.

        If `ic` != 0, that is if covariances are supplied, `kc` is not
        used.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        Therefore it is also the order of the FFT used to construct the
        sample spectrum from the cross-covariances.

    nxyg : int
        The dimension of the arrays `xg` and `yg`.

    xg : float, shape (`nxyg`), modified in place
        On entry: if the cross-covariances are to be calculated, then
        `xg` must contain the `nxy` data points of the x series. If
        covariances are supplied, `xg` need not be set.

        On exit: contains the real parts of the `ng` complex spectral
        estimates in elements `xg`[0] to `xg`[`ng`-1], and `xg`[`ng`] to
        `xg`[`nxyg`-1] contain 0.0. The y series leads the x series.

    yg : float, shape (`nxyg`), modified in place
        On entry: if cross-covariances are to be calculated, `yg` must
        contain the `nxy` data points of the y series. If covariances
        are supplied, `yg` need not be set.

        On exit: contains the imaginary parts of the `ng` complex
        spectral estimates in elements `yg`[0] to `yg`[`ng`-1], and
        `yg`[`ng`] to `yg`[`nxyg`-1] contain 0.0. The y series leads the
        x series.

    Returns
    -------
    ng : int
        The number, [`l`/2]+1, of complex spectral estimates, whose
        separate parts are held in `xg` and `yg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ic` = *<value>*, `nxyg` = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` != 0, `nxyg` >= `l`.

        (`errno` 1)
            On entry, `nxyg` = *<value>*, `kc` = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` = 0, `nxyg` >= max(`kc`, `l`).

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: if `ic` = 0, `pxy` <= 1.0.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: if `ic` = 0, `pxy` >= 0.0.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: if `ic` = 0 then `mtxy` != 0, 1 or 2.

        (`errno` 1)
            On entry, `nc` = *<value>* and `nxy` = *<value>*.

            Constraint: `nc` <= `nxy`.

        (`errno` 1)
            On entry, `nc` = *<value>*, `mw` = *<value>* and `ish` =
            *<value>*.

            Constraint: `nc` >= `mw`+abs(`ish`).

        (`errno` 1)
            On entry, `mw` = *<value>*, `ish` = *<value>* and `nxy` =
            *<value>*.

            Constraint: `mw`+abs(`ish`) <= `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `ish` = *<value>* and `mw` = *<value>*.

            Constraint: abs(`ish`) <= `mw`.

        (`errno` 1)
            On entry, `iw` = *<value>*.

            Constraint: `iw` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `nxy` = *<value>*.

            Constraint: `nxy` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>*, `nxy` = *<value>* and `nc` =
            *<value>*.

            Constraint: if `ic` = 0, `kc` >= `nxy`+`nc`.

        (`errno` 3)
            On entry, `l` = *<value>* and `mw` = *<value>*.

            Constraint: `l` >= 2*`mw`-1.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_spectrum_lag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nxy.
        utils._BYREF_INT, # Input, mtxy.
        utils._BYREF_FLOAT64, # Input, pxy.
        utils._BYREF_INT, # Input, iw.
        utils._BYREF_INT, # Input, mw.
        utils._BYREF_INT, # Input, ish.
        utils._BYREF_INT, # Input, ic.
        utils._BYREF_INT, # Input, nc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, cxy[nc].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, cyx[nc].
        utils._BYREF_INT, # Input, kc.
        utils._BYREF_INT, # Input, l.
        utils._BYREF_INT, # Input, nxyg.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xg[nxyg].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, yg[nxyg].
        utils._BYREF_INT, # Output, ng.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nxy = utils._EngineIntScalarType(
        nxy, {'fun_name': fun_name, 'entity': 'nxy'},
    ).to_ctypes()
    _mtxy = utils._EngineIntScalarType(
        mtxy, {'fun_name': fun_name, 'entity': 'mtxy'},
    ).to_ctypes()
    _pxy = utils._EngineFloat64ScalarType(
        pxy, {'fun_name': fun_name, 'entity': 'pxy'},
    ).to_ctypes()
    _iw = utils._EngineIntScalarType(
        iw, {'fun_name': fun_name, 'entity': 'iw'},
    ).to_ctypes()
    _mw = utils._EngineIntScalarType(
        mw, {'fun_name': fun_name, 'entity': 'mw'},
    ).to_ctypes()
    _ish = utils._EngineIntScalarType(
        ish, {'fun_name': fun_name, 'entity': 'ish'},
    ).to_ctypes()
    _ic = utils._EngineIntScalarType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
    ).to_ctypes()
    _nc = utils._EngineIntScalarType(
        nc, {'fun_name': fun_name, 'entity': 'nc'},
    ).to_ctypes()
    _nc_py = utils._EngineIntScalarType.to_py(_nc)
    _cxy_cls = utils._EngineFloat64ArrayType(
        cxy, {'fun_name': fun_name, 'entity': 'cxy'},
        exp_shape=(_nc_py,),
    )
    _cxy = _cxy_cls.to_ctypes()
    _cyx_cls = utils._EngineFloat64ArrayType(
        cyx, {'fun_name': fun_name, 'entity': 'cyx'},
        exp_shape=(_nc_py,),
    )
    _cyx = _cyx_cls.to_ctypes()
    _kc = utils._EngineIntScalarType(
        kc, {'fun_name': fun_name, 'entity': 'kc'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _nxyg = utils._EngineIntScalarType(
        nxyg, {'fun_name': fun_name, 'entity': 'nxyg'},
    ).to_ctypes()
    _ic_py = utils._EngineIntScalarType.to_py(_ic)
    _kc_py = utils._EngineIntScalarType.to_py(_kc)
    _l_py = utils._EngineIntScalarType.to_py(_l)
    if _ic_py == 0:
        _xg_dim1 = max(_kc_py, _l_py)
    else:
        _xg_dim1 = _l_py
    exp_shape_xg = (_xg_dim1,)
    _xg_cls = utils._EngineFloat64ArrayType(
        xg, {'fun_name': fun_name, 'entity': 'xg'},
        exp_shape=exp_shape_xg,
    )
    _xg = _xg_cls.to_ctypes()
    if _ic_py == 0:
        _yg_dim1 = max(_kc_py, _l_py)
    else:
        _yg_dim1 = _l_py
    exp_shape_yg = (_yg_dim1,)
    _yg_cls = utils._EngineFloat64ArrayType(
        yg, {'fun_name': fun_name, 'entity': 'yg'},
        exp_shape=exp_shape_yg,
    )
    _yg = _yg_cls.to_ctypes()
    _ng = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nxy,
        _mtxy,
        _pxy,
        _iw,
        _mw,
        _ish,
        _ic,
        _nc,
        _cxy,
        _cyx,
        _kc,
        _l,
        _nxyg,
        _xg,
        _yg,
        _ng,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ic = {:s}, nxyg = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ic != 0, nxyg >= l.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxyg = {:s}, kc = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ic = 0, nxyg >= max(kc, l).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pxy = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if ic = 0, pxy <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pxy = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if ic = 0, pxy >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if ic = 0 then mtxy != 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if ic = 0 then mtxy != 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nc = {:s} and nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nc <= nxy.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nc = {:s}, mw = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ish = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: nc >= mw+abs(ish).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s}, ish = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: mw+abs(ish) <= nxy.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mw >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ish = {:s} and mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(ish) <= mw.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mw >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iw = 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nxy >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kc = {:s}, nxy = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ic = 0, kc >= nxy+nc.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s} and mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: l >= 2*mw-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _cxy_cls.refresh_py(_cxy)
        _cyx_cls.refresh_py(_cyx)
        _xg_cls.refresh_py(_xg)
        _yg_cls.refresh_py(_yg)
        _ng_py = utils._EngineIntScalarType.to_py(_ng)
        return_data = _ng_py
    _err_class.manage()
    return return_data

def multi_spectrum_daniell(nxy, mtxy, pxy, mw, ish, pw, l, kc, xg, yg):
    u"""
    Multivariate time series, smoothed sample cross spectrum using
    spectral smoothing by the trapezium frequency (Daniell) window.

    ``multi_spectrum_daniell`` calculates the smoothed sample cross
    spectrum of a bivariate time series using spectral smoothing by the
    trapezium frequency (Daniell) window.

    For full information please refer to the NAG Library document for
    g13cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cdf.html

    Parameters
    ----------
    nxy : int
        n, the length of the time series x and y.

    mtxy : int
        Whether the data is to be initially mean or trend corrected.

        `mtxy` = 0

          For no correction.

        `mtxy` = 1

          For mean correction.

        `mtxy` = 2

          For trend correction.

    pxy : float
        The proportion of the data (totalled over both ends) to be
        initially tapered by the split cosine bell taper.

        A value of 0.0 implies no tapering.

    mw : int
        M, the frequency width of the smoothing window as 2*pi/M.

        A value of n implies that no smoothing is to be carried out.

    ish : int
        S, the alignment shift between the x and y series. If x leads y,
        the shift is positive.

    pw : float
        p, the shape parameter of the trapezium frequency window.

        A value of 0.0 gives a triangular window, and a value of 1.0 a
        rectangular window.

        If `mw` = `nxy` then `pw` is not used.

    l : int
        L, the frequency division of smoothed cross spectral estimates
        as 2*pi/L.

    kc : int
        The order of the fast Fourier transform (FFT) used to calculate
        the spectral estimates.

    xg : float, shape (`kc`), modified in place
        On entry: the `nxy` data points of the x series.

        On exit: the real parts of the `ng` cross spectral estimates in
        elements `xg`[0] to `xg`[`ng`-1], and `xg`[`ng`] to `xg`[`kc`-1]
        contain 0.0. The y series leads the x series.

    yg : float, shape (`kc`), modified in place
        On entry: the `nxy` data points of the y series.

        On exit: the imaginary parts of the `ng` cross spectral
        estimates in elements `yg`[0] to `yg`[`ng`-1], and `yg`[`ng`] to
        `yg`[`kc`-1] contain 0.0. The y series leads the x series.

    Returns
    -------
    ng : int
        The number of spectral estimates, [L/2]+1, whose separate parts
        are held in `xg` and `yg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ish` = *<value>* and `l` = *<value>*.

            Constraint: abs(`ish`) < `l`.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `pxy` = *<value>*, `mw` = *<value>* and `nxy` =
            *<value>*.

            Constraint: if `pw` > 1.0, `mw` = `nxy`.

        (`errno` 1)
            On entry, `pxy` = *<value>*, `mw` = *<value>* and `nxy` =
            *<value>*.

            Constraint: if `pw` < 0.0, `mw` = `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nxy` = *<value>*.

            Constraint: `mw` <= `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: `pxy` <= 1.0.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: `pxy` >= 0.0.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: `mtxy` <= 2.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: `mtxy` >= 0.

        (`errno` 1)
            On entry, `nxy` = *<value>*.

            Constraint: `nxy` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>* and `l` = *<value>*.

            Constraint: `kc` must be a multiple of `l`.

        (`errno` 2)
            On entry, `kc` = *<value>* and `nxy` = *<value>*.

            Constraint: `kc` >= 2*`nxy`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_spectrum_daniell'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13cdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nxy.
        utils._BYREF_INT, # Input, mtxy.
        utils._BYREF_FLOAT64, # Input, pxy.
        utils._BYREF_INT, # Input, mw.
        utils._BYREF_INT, # Input, ish.
        utils._BYREF_FLOAT64, # Input, pw.
        utils._BYREF_INT, # Input, l.
        utils._BYREF_INT, # Input, kc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xg[kc].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, yg[kc].
        utils._BYREF_INT, # Output, ng.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nxy = utils._EngineIntScalarType(
        nxy, {'fun_name': fun_name, 'entity': 'nxy'},
    ).to_ctypes()
    _mtxy = utils._EngineIntScalarType(
        mtxy, {'fun_name': fun_name, 'entity': 'mtxy'},
    ).to_ctypes()
    _pxy = utils._EngineFloat64ScalarType(
        pxy, {'fun_name': fun_name, 'entity': 'pxy'},
    ).to_ctypes()
    _mw = utils._EngineIntScalarType(
        mw, {'fun_name': fun_name, 'entity': 'mw'},
    ).to_ctypes()
    _ish = utils._EngineIntScalarType(
        ish, {'fun_name': fun_name, 'entity': 'ish'},
    ).to_ctypes()
    _pw = utils._EngineFloat64ScalarType(
        pw, {'fun_name': fun_name, 'entity': 'pw'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _kc = utils._EngineIntScalarType(
        kc, {'fun_name': fun_name, 'entity': 'kc'},
    ).to_ctypes()
    _kc_py = utils._EngineIntScalarType.to_py(_kc)
    _xg_cls = utils._EngineFloat64ArrayType(
        xg, {'fun_name': fun_name, 'entity': 'xg'},
        exp_shape=(_kc_py,),
    )
    _xg = _xg_cls.to_ctypes()
    _yg_cls = utils._EngineFloat64ArrayType(
        yg, {'fun_name': fun_name, 'entity': 'yg'},
        exp_shape=(_kc_py,),
    )
    _yg = _yg_cls.to_ctypes()
    _ng = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nxy,
        _mtxy,
        _pxy,
        _mw,
        _ish,
        _pw,
        _l,
        _kc,
        _xg,
        _yg,
        _ng,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ish = {:s} and l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(ish) < l.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pxy = {:s}, mw = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if pw > 1.0, mw = nxy.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pxy = {:s}, mw = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if pw < 0.0, mw = nxy.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s} and nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mw <= nxy.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mw = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mw >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pxy = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pxy <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pxy = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pxy >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mtxy <= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mtxy >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nxy >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kc = {:s} and l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: kc must be a multiple of l.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kc = {:s} and nxy = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: kc >= 2*nxy.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xg_cls.refresh_py(_xg)
        _yg_cls.refresh_py(_yg)
        _ng_py = utils._EngineIntScalarType.to_py(_ng)
        return_data = _ng_py
    _err_class.manage()
    return return_data

def multi_spectrum_bivar(xg, yg, xyrg, xyig, ng, stats, ca, calw, caup, sc, sclw, scup):
    u"""
    Multivariate time series, cross amplitude spectrum, squared
    coherency, bounds, univariate and bivariate (cross) spectra.

    For a bivariate time series, ``multi_spectrum_bivar`` calculates the
    cross amplitude spectrum and squared coherency, together with lower
    and upper bounds from the univariate and bivariate (cross) spectra.

    For full information please refer to the NAG Library document for
    g13ce

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cef.html

    Parameters
    ----------
    xg : float, shape (`ng`)
        The `ng` univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, shape (`ng`)
        The `ng` univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, shape (`ng`)
        The real parts, cf(omega), of the `ng` bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, shape (`ng`)
        The imaginary parts, qf(omega), of the `ng` bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        Note: the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    ng : int
        The number of spectral estimates in each of the arrays `xg`,
        `yg`, `xyrg` and `xyig`. It is also the number of cross
        amplitude spectral and squared coherency estimates.

    stats : float, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the
        degrees of freedom, `stats`[1] and `stats`[2] contain the lower
        and upper bound multiplying factors respectively and `stats`[3]
        contains the bandwidth.

    ca : float, shape (`ng`), modified in place
        On exit: the `ng` cross amplitude spectral estimates A^(omega)
        at each frequency of omega.

    calw : float, shape (`ng`), modified in place
        On exit: the `ng` lower bounds for the `ng` cross amplitude
        spectral estimates.

    caup : float, shape (`ng`), modified in place
        On exit: the `ng` upper bounds for the `ng` cross amplitude
        spectral estimates.

    sc : float, shape (`ng`), modified in place
        On exit: the `ng` squared coherency estimates, W^(omega) at each
        frequency omega.

    sclw : float, shape (`ng`), modified in place
        On exit: the `ng` lower bounds for the `ng` squared coherency
        estimates.

    scup : float, shape (`ng`), modified in place
        On exit: the `ng` upper bounds for the `ng` squared coherency
        estimates.

    Returns
    -------
    t : float
        The critical value for the significance of the squared
        coherency, T.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stats`[2] = *<value>*.

            Constraint: `stats`[2] >= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] <= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] > 0.0.

        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, `ng` = *<value>*.

            Constraint: `ng` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

        (`errno` 4)
            A univariate spectral estimate is zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_spectrum_bivar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ceft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, xg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, yg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, xyrg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, xyig[ng].
        utils._BYREF_INT, # Input, ng.
        utils._EngineFloat64ArrayType.c_type, # Input, stats[4].
        utils._EngineFloat64ArrayType.c_type, # Output, ca[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, calw[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, caup[ng].
        utils._BYREF_FLOAT64, # Output, t.
        utils._EngineFloat64ArrayType.c_type, # Output, sc[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, sclw[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, scup[ng].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ng = utils._EngineIntScalarType(
        ng, {'fun_name': fun_name, 'entity': 'ng'},
    ).to_ctypes()
    _ng_py = utils._EngineIntScalarType.to_py(_ng)
    _xg = utils._EngineFloat64ArrayType(
        xg, {'fun_name': fun_name, 'entity': 'xg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _yg = utils._EngineFloat64ArrayType(
        yg, {'fun_name': fun_name, 'entity': 'yg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _xyrg = utils._EngineFloat64ArrayType(
        xyrg, {'fun_name': fun_name, 'entity': 'xyrg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _xyig = utils._EngineFloat64ArrayType(
        xyig, {'fun_name': fun_name, 'entity': 'xyig'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _stats = utils._EngineFloat64ArrayType(
        stats, {'fun_name': fun_name, 'entity': 'stats'},
        exp_shape=(4,),
    ).to_ctypes()
    _ca_cls = utils._EngineFloat64ArrayType(
        ca, {'fun_name': fun_name, 'entity': 'ca'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _ca = _ca_cls.to_ctypes(can_reference_contents=False)
    _calw_cls = utils._EngineFloat64ArrayType(
        calw, {'fun_name': fun_name, 'entity': 'calw'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _calw = _calw_cls.to_ctypes(can_reference_contents=False)
    _caup_cls = utils._EngineFloat64ArrayType(
        caup, {'fun_name': fun_name, 'entity': 'caup'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _caup = _caup_cls.to_ctypes(can_reference_contents=False)
    _t = utils._EngineFloat64ScalarType.empty_ctypes()
    _sc_cls = utils._EngineFloat64ArrayType(
        sc, {'fun_name': fun_name, 'entity': 'sc'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _sc = _sc_cls.to_ctypes(can_reference_contents=False)
    _sclw_cls = utils._EngineFloat64ArrayType(
        sclw, {'fun_name': fun_name, 'entity': 'sclw'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _sclw = _sclw_cls.to_ctypes(can_reference_contents=False)
    _scup_cls = utils._EngineFloat64ArrayType(
        scup, {'fun_name': fun_name, 'entity': 'scup'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _scup = _scup_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _xg,
        _yg,
        _xyrg,
        _xyig,
        _ng,
        _stats,
        _ca,
        _calw,
        _caup,
        _t,
        _sc,
        _sclw,
        _scup,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[2] >= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[1] <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[0] >= 3.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ng = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ng >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A bivariate spectral estimate is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A univariate spectral estimate is negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A univariate spectral estimate is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A calculated value of the squared coherency exceeds 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ca_cls.refresh_py(_ca)
        _calw_cls.refresh_py(_calw)
        _caup_cls.refresh_py(_caup)
        _sc_cls.refresh_py(_sc)
        _sclw_cls.refresh_py(_sclw)
        _scup_cls.refresh_py(_scup)
        _t_py = utils._EngineFloat64ScalarType.to_py(_t)
        return_data = _t_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['t'],
            )(
                _t_py,
            )
    _err_class.manage()
    return return_data

def multi_gain_bivar(xg, yg, xyrg, xyig, ng, stats, gn, gnlw, gnup, ph, phlw, phup):
    u"""
    Multivariate time series, gain, phase, bounds, univariate and
    bivariate (cross) spectra.

    For a bivariate time series, ``multi_gain_bivar`` calculates the
    gain and phase together with lower and upper bounds from the
    univariate and bivariate spectra.

    For full information please refer to the NAG Library document for
    g13cf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cff.html

    Parameters
    ----------
    xg : float, shape (`ng`)
        The `ng` univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, shape (`ng`)
        The `ng` univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, shape (`ng`)
        The real parts, cf(omega), of the `ng` bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, shape (`ng`)
        The imaginary parts, qf(omega), of the `ng` bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        Note: the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    ng : int
        The number of spectral estimates in each of the arrays `xg`,
        `yg`, `xyrg` and `xyig`. It is also the number of gain and phase
        estimates.

    stats : float, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the
        degrees of freedom, `stats`[1] and `stats`[2] contain the lower
        and upper bound multiplying factors respectively and `stats`[3]
        holds the bandwidth.

    gn : float, shape (`ng`), modified in place
        On exit: the `ng` gain estimates, G^(omega), at each frequency
        omega.

    gnlw : float, shape (`ng`), modified in place
        On exit: the `ng` lower bounds for the `ng` gain estimates.

    gnup : float, shape (`ng`), modified in place
        On exit: the `ng` upper bounds for the `ng` gain estimates.

    ph : float, shape (`ng`), modified in place
        On exit: the `ng` phase estimates, phi^(omega), at each
        frequency omega.

    phlw : float, shape (`ng`), modified in place
        On exit: the `ng` lower bounds for the `ng` phase estimates.

    phup : float, shape (`ng`), modified in place
        On exit: the `ng` upper bounds for the `ng` phase estimates.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, `ng` = *<value>*.

            Constraint: `ng` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

        (`errno` 4)
            A univariate spectral estimate is zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_gain_bivar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13cfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, xg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, yg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, xyrg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, xyig[ng].
        utils._BYREF_INT, # Input, ng.
        utils._EngineFloat64ArrayType.c_type, # Input, stats[4].
        utils._EngineFloat64ArrayType.c_type, # Output, gn[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, gnlw[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, gnup[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, ph[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, phlw[ng].
        utils._EngineFloat64ArrayType.c_type, # Output, phup[ng].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ng = utils._EngineIntScalarType(
        ng, {'fun_name': fun_name, 'entity': 'ng'},
    ).to_ctypes()
    _ng_py = utils._EngineIntScalarType.to_py(_ng)
    _xg = utils._EngineFloat64ArrayType(
        xg, {'fun_name': fun_name, 'entity': 'xg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _yg = utils._EngineFloat64ArrayType(
        yg, {'fun_name': fun_name, 'entity': 'yg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _xyrg = utils._EngineFloat64ArrayType(
        xyrg, {'fun_name': fun_name, 'entity': 'xyrg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _xyig = utils._EngineFloat64ArrayType(
        xyig, {'fun_name': fun_name, 'entity': 'xyig'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _stats = utils._EngineFloat64ArrayType(
        stats, {'fun_name': fun_name, 'entity': 'stats'},
        exp_shape=(4,),
    ).to_ctypes()
    _gn_cls = utils._EngineFloat64ArrayType(
        gn, {'fun_name': fun_name, 'entity': 'gn'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _gn = _gn_cls.to_ctypes(can_reference_contents=False)
    _gnlw_cls = utils._EngineFloat64ArrayType(
        gnlw, {'fun_name': fun_name, 'entity': 'gnlw'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _gnlw = _gnlw_cls.to_ctypes(can_reference_contents=False)
    _gnup_cls = utils._EngineFloat64ArrayType(
        gnup, {'fun_name': fun_name, 'entity': 'gnup'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _gnup = _gnup_cls.to_ctypes(can_reference_contents=False)
    _ph_cls = utils._EngineFloat64ArrayType(
        ph, {'fun_name': fun_name, 'entity': 'ph'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _ph = _ph_cls.to_ctypes(can_reference_contents=False)
    _phlw_cls = utils._EngineFloat64ArrayType(
        phlw, {'fun_name': fun_name, 'entity': 'phlw'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _phlw = _phlw_cls.to_ctypes(can_reference_contents=False)
    _phup_cls = utils._EngineFloat64ArrayType(
        phup, {'fun_name': fun_name, 'entity': 'phup'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _phup = _phup_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _xg,
        _yg,
        _xyrg,
        _xyig,
        _ng,
        _stats,
        _gn,
        _gnlw,
        _gnup,
        _ph,
        _phlw,
        _phup,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[0] >= 3.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ng = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ng >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A bivariate spectral estimate is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A univariate spectral estimate is negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A univariate spectral estimate is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A calculated value of the squared coherency exceeds 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _gn_cls.refresh_py(_gn)
        _gnlw_cls.refresh_py(_gnlw)
        _gnup_cls.refresh_py(_gnup)
        _ph_cls.refresh_py(_ph)
        _phlw_cls.refresh_py(_phlw)
        _phup_cls.refresh_py(_phup)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def multi_noise_bivar(xg, yg, xyrg, xyig, ng, stats, l, n, er, rf):
    u"""
    Multivariate time series, noise spectrum, bounds, impulse response
    function and its standard error.

    For a bivariate time series, ``multi_noise_bivar`` calculates the
    noise spectrum together with multiplying factors for the bounds and
    the impulse response function and its standard error, from the
    univariate and bivariate spectra.

    For full information please refer to the NAG Library document for
    g13cg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cgf.html

    Parameters
    ----------
    xg : float, shape (`ng`)
        The `ng` univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, shape (`ng`)
        The `ng` univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, shape (`ng`)
        The real parts, cf(omega), of the `ng` bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, shape (`ng`)
        The imaginary parts, qf(omega), of the `ng` bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        Note: the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    ng : int
        The number of spectral estimates in each of the arrays `xg`,
        `yg`, `xyrg`, `xyig`. It is also the number of noise spectral
        estimates.

    stats : float, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the degree
        of freedom, `stats`[1] and `stats`[2] contain the lower and
        upper bound multiplying factors respectively and `stats`[3]
        contains the bandwidth.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        It is also the order of the FFT used to calculate the impulse
        response function. `l` must relate to the parameter `ng` by the
        relationship.

    n : int
        The number of points in each of the time series x and y. `n`
        should have the same value as nxy in the call of
        ``multi_spectrum_lag`` or ``multi_spectrum_daniell`` which
        calculated the smoothed sample cross spectrum. `n` is used in
        calculating the impulse response function standard error
        (`rfse`).

    er : float, shape (`ng`), modified in place
        On exit: the `ng` estimates of the noise spectrum, f^_{y |
        x}(omega) at each frequency.

    rf : float, shape (`l`), modified in place
        On exit: the impulse response function. Causal responses are
        stored in ascending frequency in `rf`[0] to `rf`[`ng`-1] and
        anticipatory responses are stored in descending frequency in
        `rf`[`ng`] to `rf`[`l`-1].

    Returns
    -------
    erlw : float
        The noise spectrum lower limit multiplying factor.

    erup : float
        The noise spectrum upper limit multiplying factor.

    rfse : float
        The impulse response function standard error.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `stats`[2] = *<value>*.

            Constraint: `stats`[2] >= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] <= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] > 0.0.

        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, `ng` = *<value>*.

            Constraint: `ng` >= 1.

        (`errno` 6)
            On entry, `ng` = *<value>* and `l` = *<value>*.

            Constraint: `ng` = `l`/2+1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

            For this frequency the noise spectrum is set to zero, and
            the contribution to the impulse response function and its
            standard error is set to zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

            For this frequency the noise spectrum is set to zero, and
            the contributions to the impulse response function and its
            standard error are set to zero.

        (`errno` 4)
            A univariate spectral estimate is zero.

            For this frequency the noise spectrum is set to zero and the
            contributions to the impulse response function and its
            standard error are set to zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

            For this frequency the squared coherency is reset to 1.0
            with the consequence that the noise spectrum is zero and the
            contribution to the impulse response function at this
            frequency is zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_noise_bivar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13cgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, xg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, yg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, xyrg[ng].
        utils._EngineFloat64ArrayType.c_type, # Input, xyig[ng].
        utils._BYREF_INT, # Input, ng.
        utils._EngineFloat64ArrayType.c_type, # Input, stats[4].
        utils._BYREF_INT, # Input, l.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Output, er[ng].
        utils._BYREF_FLOAT64, # Output, erlw.
        utils._BYREF_FLOAT64, # Output, erup.
        utils._EngineFloat64ArrayType.c_type, # Output, rf[l].
        utils._BYREF_FLOAT64, # Output, rfse.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ng = utils._EngineIntScalarType(
        ng, {'fun_name': fun_name, 'entity': 'ng'},
    ).to_ctypes()
    _ng_py = utils._EngineIntScalarType.to_py(_ng)
    _xg = utils._EngineFloat64ArrayType(
        xg, {'fun_name': fun_name, 'entity': 'xg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _yg = utils._EngineFloat64ArrayType(
        yg, {'fun_name': fun_name, 'entity': 'yg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _xyrg = utils._EngineFloat64ArrayType(
        xyrg, {'fun_name': fun_name, 'entity': 'xyrg'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _xyig = utils._EngineFloat64ArrayType(
        xyig, {'fun_name': fun_name, 'entity': 'xyig'},
        exp_shape=(_ng_py,),
    ).to_ctypes()
    _stats = utils._EngineFloat64ArrayType(
        stats, {'fun_name': fun_name, 'entity': 'stats'},
        exp_shape=(4,),
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _er_cls = utils._EngineFloat64ArrayType(
        er, {'fun_name': fun_name, 'entity': 'er'},
        exp_shape=(_ng_py,),
        check_contents=False,
    )
    _er = _er_cls.to_ctypes(can_reference_contents=False)
    _erlw = utils._EngineFloat64ScalarType.empty_ctypes()
    _erup = utils._EngineFloat64ScalarType.empty_ctypes()
    _l_py = utils._EngineIntScalarType.to_py(_l)
    _rf_cls = utils._EngineFloat64ArrayType(
        rf, {'fun_name': fun_name, 'entity': 'rf'},
        exp_shape=(_l_py,),
        check_contents=False,
    )
    _rf = _rf_cls.to_ctypes(can_reference_contents=False)
    _rfse = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _xg,
        _yg,
        _xyrg,
        _xyig,
        _ng,
        _stats,
        _l,
        _n,
        _er,
        _erlw,
        _erup,
        _rf,
        _rfse,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[2] >= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[1] <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stats[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stats[0] >= 3.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ng = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ng >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A bivariate spectral estimate is zero.',
                    'For this frequency the noise spectrum is set to zero, and the contribution',
                    'to the impulse response function and its standard error is set to zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A univariate spectral estimate is negative.',
                    'For this frequency the noise spectrum is set to zero, and the contributions',
                    'to the impulse response function and its standard error are set to zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A univariate spectral estimate is zero.',
                    'For this frequency the noise spectrum is set to zero and the contributions',
                    'to the impulse response function and its standard error are set to zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A calculated value of the squared coherency exceeds 1.0.',
                    'For this frequency the squared coherency is reset to 1.0 with the',
                    'consequence that the noise spectrum is zero and the contribution to the',
                    'impulse response function at this frequency is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ng = {:s} and l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ng = l/2+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _er_cls.refresh_py(_er)
        _rf_cls.refresh_py(_rf)
        _erlw_py = utils._EngineFloat64ScalarType.to_py(_erlw)
        _erup_py = utils._EngineFloat64ScalarType.to_py(_erup)
        _rfse_py = utils._EngineFloat64ScalarType.to_py(_rfse)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['erlw', 'erup', 'rfse'],
        )(
            _erlw_py,
            _erup_py,
            _rfse_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def multi_autocorr_part(c0, c, ns, nl, nk, p, v, d, db, w, wb):
    u"""
    Multivariate time series, multiple squared partial autocorrelations.

    ``multi_autocorr_part`` calculates the multivariate partial
    autocorrelation function of a multivariate time series.

    For full information please refer to the NAG Library document for
    g13db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dbf.html

    Parameters
    ----------
    c0 : float, shape (`ns`, `ns`)
        Contains the zero lag cross-covariances between the `ns` series
        as returned by ``multi_corrmat_cross``. (`c0` is assumed to be
        symmetric, upper triangle only is used.)

    c : float, shape (`ns`, `ns`, `nl`)
        Contains the cross-covariances at lags 1 to `nl`.
        `c`[i-1,j-1,k-1] must contain the cross-covariance, c_{ijk}, of
        series i and series j at lag k. Series j leads series i.

    ns : int
        k, the number of time series whose cross-covariances are
        supplied in `c` and `c0`.

    nl : int
        m, the maximum lag for which cross-covariances are supplied in
        `c`.

    nk : int
        The number of lags to which partial auto-correlations are to be
        calculated.

    p : float, shape (`nk`), modified in place
        On exit: the multiple squared partial autocorrelations from lags
        1 to `nvp`; that is, `p`[l-1] contains p_l^2, for l =
        1,2,...,`nvp`. For lags `nvp`+1 to `nk` the elements of `p` are
        set to zero.

    v : float, shape (`nk`), modified in place
        On exit: the prediction error variance ratios from lags 1 to
        `nvp`; that is, `v`[l-1] contains v_l, for l = 1,2,...,`nvp`.
        For lags `nvp`+1 to `nk` the elements of `v` are set to zero.

    d : float, shape (`ns`, `ns`, `nk`), modified in place
        On exit: the prediction error variance matrices at lags 1 to
        `nvp`.

        Element (i,j,k) of `d` contains the prediction error covariance
        of series i and series j at lag k, for k = 1,2,...,`nvp`.

        Series j leads series i; that is, the (i,j)th element of D_k.

        For lags `nvp`+1 to `nk` the elements of `d` are set to zero.

    db : float, shape (`ns`, `ns`), modified in place
        On exit: the backward prediction error variance matrix at lag
        `nvp`.

        `db`[i-1,j-1] contains the backward prediction error covariance
        of series i and series j; that is, the (i,j)th element of the
        G_k, where k = `nvp`.

    w : float, shape (`ns`, `ns`, `nk`), modified in place
        On exit: the prediction coefficient matrices at lags 1 to `nvp`.

        `w`[i-1,j-1,l-1] contains the jth prediction coefficient of
        series i at lag l; that is, the (i,j)th element of Phi_{kl},
        where k = `nvp`, for l = 1,2,...,`nvp`.

        For lags `nvp`+1 to `nk` the elements of `w` are set to zero.

    wb : float, shape (`ns`, `ns`, `nk`), modified in place
        On exit: the backward prediction coefficient matrices at lags 1
        to `nvp`.

        `wb`[i-1,j-1,l-1] contains the jth backward prediction
        coefficient of series i at lag l; that is, the (i,j)th element
        of Psi_{kl}, where k = `nvp`, for l = 1,2,...,`nvp`.

        For lags `nvp`+1 to `nk` the elements of `wb` are set to zero.

    Returns
    -------
    v0 : float
        The lag zero prediction error variance (equal to the determinant
        of `c0`).

    nvp : int
        The maximum lag, L, for which calculation of `p`, `v`, `d`,
        `db`, `w` and `wb` was successful. If the function completes
        successfully `nvp` will equal `nk`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nk` = *<value>* and `nl` = *<value>*.

            Constraint: `nk` <= `nl`.

        (`errno` 1)
            On entry, `nk` = *<value>*.

            Constraint: `nk` >= 1.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` >= 1.

        (`errno` 1)
            On entry, `ns` = *<value>* and ldc0 = *<value>*.

            Constraint: `ns` <= ldc0.

        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            `c0` is not positive definite.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            For `nvp` = *<value>*, at lag k = `nvp`+1, D_k was found not
            to be positive definite.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_autocorr_part'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._EngineFloat64ArrayType.c_type, # Input, c0[ldc0, ns].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc0, ldc0, nl].
        utils._BYREF_INT, # Input, ldc0.
        utils._BYREF_INT, # Input, ns.
        utils._BYREF_INT, # Input, nl.
        utils._BYREF_INT, # Input, nk.
        utils._EngineFloat64ArrayType.c_type, # Output, p[nk].
        utils._BYREF_FLOAT64, # Output, v0.
        utils._EngineFloat64ArrayType.c_type, # Output, v[nk].
        utils._EngineFloat64ArrayType.c_type, # Output, d[ldc0, ldc0, nk].
        utils._EngineFloat64ArrayType.c_type, # Output, db[ldc0, ns].
        utils._EngineFloat64ArrayType.c_type, # Output, w[ldc0, ldc0, nk].
        utils._EngineFloat64ArrayType.c_type, # Output, wb[ldc0, ldc0, nk].
        utils._BYREF_INT, # Output, nvp.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wa.
        utils._BYREF_INT, # Input, iwa.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c0, 'c0'), (c, 'c'), (d, 'd'), (db, 'db'), (w, 'w'), (wb, 'wb')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _ldc0 = utils._EngineIntScalarType(
        max(1, utils._EngineFloat64ArrayType.get_shape(c0, {'fun_name': fun_name, 'entity': 'c0'}, exp_rank=2)[0]), {'fun_name': fun_name, 'entity': '_ldc0'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _c0 = utils._EngineFloat64ArrayType(
        c0, {'fun_name': fun_name, 'entity': 'c0'},
        exp_shape=(_ns_py, _ns_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nl = utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = utils._EngineIntScalarType.to_py(_nl)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ns_py, _ns_py, _nl_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nk = utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _nk_py = utils._EngineIntScalarType.to_py(_nk)
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_nk_py,),
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _v0 = utils._EngineFloat64ScalarType.empty_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_nk_py,),
        check_contents=False,
    )
    _v = _v_cls.to_ctypes(can_reference_contents=False)
    _d_cls = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_ns_py, _ns_py, _nk_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _d = _d_cls.to_ctypes(can_reference_contents=False)
    _db_cls = utils._EngineFloat64ArrayType(
        db, {'fun_name': fun_name, 'entity': 'db'},
        exp_shape=(_ns_py, _ns_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _db = _db_cls.to_ctypes(can_reference_contents=False)
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_ns_py, _ns_py, _nk_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _w = _w_cls.to_ctypes(can_reference_contents=False)
    _wb_cls = utils._EngineFloat64ArrayType(
        wb, {'fun_name': fun_name, 'entity': 'wb'},
        exp_shape=(_ns_py, _ns_py, _nk_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _wb = _wb_cls.to_ctypes(can_reference_contents=False)
    _nvp = utils._EngineIntScalarType.empty_ctypes()
    _wa = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iwa = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _c0,
        _c,
        _ldc0,
        _ns,
        _nl,
        _nk,
        _p,
        _v0,
        _v,
        _d,
        _db,
        _w,
        _wb,
        _nvp,
        _wa,
        _iwa,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nk = {:s} and nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nk <= nl.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nk >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nl = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nl >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and ldc0 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns <= ldc0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'c0 is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'For nvp = {:s}, at lag k = nvp+1, D_k was'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'found not to be positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p_cls.refresh_py(_p)
        _v_cls.refresh_py(_v)
        _d_cls.refresh_py(_d)
        _db_cls.refresh_py(_db)
        _w_cls.refresh_py(_w)
        _wb_cls.refresh_py(_wb)
        _v0_py = utils._EngineFloat64ScalarType.to_py(_v0)
        _nvp_py = utils._EngineIntScalarType.to_py(_nvp)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['v0', 'nvp'],
        )(
            _v0_py,
            _nvp_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def multi_varma_estimate(k, n, ip, iq, mean, par, npar, qq, w, parhld, exact, iprint, cgetol, maxcal, ishow, v, g, cm, io_manager=None):
    u"""
    Multivariate time series, estimation of VARMA model.

    ``multi_varma_estimate`` fits a vector autoregressive moving average
    (VARMA) model to an observed vector of time series using the method
    of Maximum Likelihood (ML).
    Standard errors of parameter estimates are computed along with their
    appropriate correlation matrix.
    The function also calculates estimates of the residual series.

    For full information please refer to the NAG Library document for
    g13dd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ddf.html

    Parameters
    ----------
    k : int
        k, the number of observed time series.

    n : int
        n, the number of observations in each time series.

    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    mean : bool
        `mean` = True, if components of mu have been estimated and
        `mean` = False, if all elements of mu are to be taken as zero.

    par : float, shape (`npar`), modified in place
        On entry: initial parameter estimates read in row by row in the
        order phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q,mu.

        Thus,

        - if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to
          an initial estimate of the (i,j)th element of phi_l, for j =
          1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,p;

        - if `iq` > 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
          equal to an initial estimate of the (i,j)th element of
          theta_l, l = 1,2,...,q and i,j = 1,2,...,k;

        - if `mean` = True, `par`[(p+q)*k*k+i-1] should be set equal to
          an initial estimate of the ith component of mu (mu(i)). (If
          you set `par`[(p+q)*k*k+i-1] to 0.0 then
          ``multi_varma_estimate`` will calculate the mean of the ith
          series and use this as an initial estimate of mu(i).)

        The first p*k*k elements of `par` must satisfy the stationarity
        condition and the next q*k*k elements of `par` must satisfy the
        invertibility condition.

        If in doubt set all elements of `par` to 0.0.

        On exit: if no exception or warning is raised or `errno` >= 4
        then all the elements of `par` will be overwritten by the latest
        estimates of the corresponding ARMA parameters.

    npar : int
        `npar` is the number of initial parameter estimates.

    qq : float, shape (`k`, `k`), modified in place
        On entry: `qq`[i-1,j-1] must be set equal to an initial estimate
        of the (i,j)th element of Sigma. The lower triangle only is
        needed. `qq` must be positive definite. It is strongly
        recommended that on entry the elements of `qq` are of the same
        order of magnitude as at the solution point. If you set
        `qq`[i-1,j-1] = 0.0, for j = 1,2,...,i, for i = 1,2,...,k, then
        ``multi_varma_estimate`` will calculate the covariance matrix
        between the k time series and use this as an initial estimate of
        Sigma.

        On exit: if no exception or warning is raised or `errno` >= 4
        then `qq`[i-1,j-1] will contain the latest estimate of the
        (i,j)th element of Sigma. The lower triangle only is returned.

    w : float, shape (`k`, `n`)
        `w`[i-1,t-1] must be set equal to the ith component of W_t, for
        t = 1,2,...,n, for i = 1,2,...,k.

    parhld : bool, shape (`npar`)
        `parhld`[i-1] must be set to True if `par`[i-1] is to be held
        constant at its input value and False if `par`[i-1] is a free
        parameter, for i = 1,2,...,`npar`.

        If in doubt try setting all elements of `parhld` to False.

    exact : bool
        Must be set equal to True if you wish ``multi_varma_estimate``
        to compute exact maximum likelihood estimates. `exact` must be
        set equal to False if only conditional likelihood estimates are
        required.

    iprint : int
        The frequency with which the automatic monitoring function is to
        be called.

        `iprint` > 0

          The ML search procedure is monitored once every `iprint`
          iterations and just before exit from the search function.

        `iprint` = 0

          The search function is monitored once at the final point.

        `iprint` < 0

          The search function is not monitored at all.

    cgetol : float
        The accuracy to which the solution in `par` and `qq` is
        required.

        If `cgetol` is set to 10^(-l) and on exit no exception or
        warning is raised or `errno` >= 6, then all the elements in
        `par` and `qq` should be accurate to approximately l decimal
        places.

        For most practical purposes the value 10^(-4) should suffice.

        You should be wary of setting `cgetol` too small since the
        convergence criteria may then have become too strict for the
        machine to handle.

        If `cgetol` has been set to a value which is less than the
        machine precision, epsilon, then ``multi_varma_estimate`` will
        use the value 10.0*sqrt(epsilon) instead.

    maxcal : int
        The maximum number of likelihood evaluations to be permitted by
        the search procedure.

    ishow : int
        Specifies which of the following two quantities are to be
        printed.

        - table of maximum likelihood estimates and their standard
          errors (as returned in the output arrays `par`, `qq` and
          `cm`);

        - table of residual series (as returned in the output array
          `v`).

        `ishow` = 0

          None of the above are printed.

        `ishow` = 1

          (i) only is printed.

        `ishow` = 2

          are printed.

    v : float, shape (`k`, `n`), modified in place
        On exit: if no exception or warning is raised or `errno` >= 4
        then `v`[i-1,t-1] will contain an estimate of the ith component
        of epsilon_t, for t = 1,2,...,n, for i = 1,2,...,k,
        corresponding to the final point held in `par` and `qq`.

    g : float, shape (`npar`), modified in place
        On exit: if no exception or warning is raised or `errno` >= 4
        then `g`[i-1] will contain the estimated first derivative of the
        log-likelihood function with respect to the ith element in the
        array `par`. If the gradient cannot be computed then all the
        elements of `g` are returned as zero.

    cm : float, shape (`npar`, `npar`), modified in place
        On exit: if no exception or warning is raised or `errno` >= 4
        then `cm`[i-1,j-1] will contain an estimate of the correlation
        coefficient between the ith and jth elements in the `par` array
        for 1 <= i <= `npar`, 1 <= j <= `npar`. If i = j, then
        `cm`[i-1,j-1] will contain the estimated standard error of
        `par`[i-1]. If the lth component of `par` has been held
        constant, i.e., `parhld`[l-1] was set to True, then the lth row
        and column of `cm` will be set to zero. If the second derivative
        matrix cannot be computed then all the elements of `cm` are
        returned as zero.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    niter : int
        If no exception or warning is raised or `errno` >= 4 then
        `niter` contains the number of iterations performed by the
        search function.

    rlogl : float
        If no exception or warning is raised or `errno` >= 4 then
        `rlogl` contains the value of the log-likelihood function
        corresponding to the final point held in `par` and `qq`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*, `k` = *<value>* and `npar` =
            *<value>*.

            Constraint: `n`*`k` > `npar`+`k`*(`k`+1)/2.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = 0 and `iq` = 0.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` = *<value>*.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `ishow` = *<value>*.

            Constraint: 0 <= `ishow` <= 2.

        (`errno` 2)
            The initial estimate of Sigma is not positive definite.

        (`errno` 2)
            The initial AR parameter estimates are outside the
            stationarity region.

        (`errno` 2)
            The initial MA parameter estimates are outside the
            invertibility region.

        (`errno` 2)
            The starting point is too close to the boundary of the
            admissibility region.

        (`errno` 3)
            The function cannot compute a sufficiently accurate estimate
            of the gradient vector at the user-supplied starting point.

        (`errno` 4)
            There have been `maxcal` log-likelihood evaluations made in
            the function.

            See the function document for further details.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The conditions for a solution have not all been met, but a
            point at which the log-likelihood took a larger value could
            not be found.

            See the function document for further details.

        (`errno` 6)
            The ML solution is so close to the boundary of either the
            stationarity region or the invertibility region that
            ``multi_varma_estimate`` cannot evaluate the Hessian matrix.
            The elements of `cm` are set to zero, as are the elements of
            `g`. All other output quantities are correct.

        (`errno` 7)
            An estimate of the second derivative matrix and the gradient
            vector at the solution point was computed. Either the
            Hessian matrix was found to be too ill-conditioned to be
            evaluated accurately or the gradient vector could not be
            computed to an acceptable degree of accuracy. The elements
            of `cm` are set to zero, as are the elements of `g`. All
            other output quantities are correct.

        (`errno` 8)
            The second-derivative matrix at the solution point is not
            positive definite. The elements of `cm` are set to zero. All
            other output quantities are correct.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_varma_estimate'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ddft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_LOGICAL, # Input, mean.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, par[npar].
        utils._BYREF_INT, # Input, npar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, qq[kmax, k].
        utils._BYREF_INT, # Input, kmax.
        utils._EngineFloat64ArrayType.c_type, # Input, w[kmax, n].
        utils._EngineLogicalArrayType.c_type, # Input, parhld[npar].
        utils._BYREF_LOGICAL, # Input, exact.
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_FLOAT64, # Input, cgetol.
        utils._BYREF_INT, # Input, maxcal.
        utils._BYREF_INT, # Input, ishow.
        utils._BYREF_INT, # Input, lp.
        utils._BYREF_INT, # Output, niter.
        utils._BYREF_FLOAT64, # Output, rlogl.
        utils._EngineFloat64ArrayType.c_type, # Output, v[kmax, n].
        utils._EngineFloat64ArrayType.c_type, # Output, g[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, cm[ldcm, npar].
        utils._BYREF_INT, # Input, ldcm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((qq, 'qq'), (w, 'w'), (v, 'v'), (cm, 'cm')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _mean = utils._EngineLogicalScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'},
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _par_cls = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_npar_py,),
    )
    _par = _par_cls.to_ctypes()
    _kmax_py = utils._EngineFloat64ArrayType.get_shape(qq, {'fun_name': fun_name, 'entity': 'qq'}, exp_rank=2)[0]
    _kmax = utils._EngineIntScalarType(
        _kmax_py, {'fun_name': fun_name, 'entity': '_kmax'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _qq_cls = utils._EngineFloat64ArrayType(
        qq, {'fun_name': fun_name, 'entity': 'qq'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
    )
    _qq = _qq_cls.to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _parhld = utils._EngineLogicalArrayType(
        parhld, {'fun_name': fun_name, 'entity': 'parhld'},
        exp_shape=(_npar_py,),
    ).to_ctypes()
    _exact = utils._EngineLogicalScalarType(
        exact, {'fun_name': fun_name, 'entity': 'exact'},
    ).to_ctypes()
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _cgetol = utils._EngineFloat64ScalarType(
        cgetol, {'fun_name': fun_name, 'entity': 'cgetol'},
    ).to_ctypes()
    _maxcal = utils._EngineIntScalarType(
        maxcal, {'fun_name': fun_name, 'entity': 'maxcal'},
    ).to_ctypes()
    _ishow = utils._EngineIntScalarType(
        ishow, {'fun_name': fun_name, 'entity': 'ishow'},
    ).to_ctypes()
    _lp = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _niter = utils._EngineIntScalarType.empty_ctypes()
    _rlogl = utils._EngineFloat64ScalarType.empty_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _v = _v_cls.to_ctypes(can_reference_contents=False)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _g = _g_cls.to_ctypes(can_reference_contents=False)
    _ldcm_py = utils._EngineFloat64ArrayType.get_nd(cm, {'fun_name': fun_name, 'entity': 'cm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcm = utils._EngineIntScalarType(
        _ldcm_py, {'fun_name': fun_name, 'entity': '_ldcm'},
    ).to_ctypes()
    _cm_cls = utils._EngineFloat64ArrayType(
        cm, {'fun_name': fun_name, 'entity': 'cm'},
        exp_shape=(_npar_py, _npar_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cm = _cm_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _k,
        _n,
        _ip,
        _iq,
        _mean,
        _par,
        _npar,
        _qq,
        _kmax,
        _w,
        _parhld,
        _exact,
        _iprint,
        _cgetol,
        _maxcal,
        _ishow,
        _lp,
        _niter,
        _rlogl,
        _v,
        _g,
        _cm,
        _ldcm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1011:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}, k = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: n*k > npar+k*(k+1)/2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1013:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1014:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1015:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = 0 and iq = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1016:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1018:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxcal = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxcal >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1019:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ishow = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= ishow <= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The initial estimate of Sigma is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1102:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The initial AR parameter estimates are outside the stationarity region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1103:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The initial MA parameter estimates are outside the invertibility region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1104:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The starting point is too close to the boundary of the admissibility region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1201:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The function cannot compute a sufficiently accurate',
                    'estimate of the gradient vector at the user-supplied starting point.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1301:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There have been maxcal log-likelihood evaluations made in the function.',
                    'See the function document for further details.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1401:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The conditions for a solution have not all been met,',
                    'but a point at which the log-likelihood took a larger',
                    'value could not be found.',
                    'See the function document for further details.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1501:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The ML solution is so close to the boundary of either',
                    'the stationarity region or the invertibility region that',
                    'multi_varma_estimate cannot evaluate the Hessian',
                    'matrix. The elements of cm are set to zero, as are',
                    'the elements of g. All other output quantities are correct.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 1601:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'An estimate of the second derivative matrix and',
                    'the gradient vector at the solution point was computed.',
                    'Either the Hessian matrix was found to be too',
                    'ill-conditioned to be evaluated accurately or the gradient',
                    'vector could not be computed to an acceptable degree',
                    'of accuracy. The elements of cm are set to zero,',
                    'as are the elements of g. All other output quantities are correct.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 1701:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The second-derivative matrix at the solution point is not positive definite.',
                    'The elements of cm are set to zero.',
                    'All other output quantities are correct.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _par_cls.refresh_py(_par)
        _qq_cls.refresh_py(_qq)
        _v_cls.refresh_py(_v)
        _g_cls.refresh_py(_g)
        _cm_cls.refresh_py(_cm)
        _niter_py = utils._EngineIntScalarType.to_py(_niter)
        _rlogl_py = utils._EngineFloat64ScalarType.to_py(_rlogl)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['niter', 'rlogl'],
        )(
            _niter_py,
            _rlogl_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def multi_varma_forecast(k, n, z, tr, dord, delta, ip, iq, mean, par, lpar, qq, v, lmax, predz, sefz, ref, lref):
    u"""
    Multivariate time series, forecasts and their standard errors.

    ``multi_varma_forecast`` computes forecasts of a multivariate time
    series.
    It is assumed that a vector ARMA model has already been fitted to
    the appropriately differenced/transformed time series using
    ``multi_varma_estimate``.
    The standard deviations of the forecast errors are also returned.
    A reference vector is set up so that, should future series values
    become available, the forecasts and their standard errors may be
    updated by calling ``multi_varma_update``.

    For full information please refer to the NAG Library document for
    g13dj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13djf.html

    Parameters
    ----------
    k : int
        k, the dimension of the multivariate time series.

    n : int
        n, the number of observations in the series, Z_t, prior to
        differencing.

    z : float, shape (`k`, `n`)
        `z`[i,t] must contain, z_{it}, the ith component of Z_t, for t =
        1,2,...,n, for i = 1,2,...,k.

    tr : str, length 1, shape (`k`)
        `tr`[i-1] indicates whether the ith time series is to be
        transformed, for i = 1,2,...,k.

        `tr`[i-1] = 'N'

          No transformation is used.

        `tr`[i-1] = 'L'

          A log transformation is used.

        `tr`[i-1] = 'S'

          A square root transformation is used.

    dord : int, shape (`k`)
        `dord`[i-1] must specify, d_i, the order of differencing
        required for the ith series.

    delta : float, shape (`k`, d)
        If `dord`[i-1] > 0, then `delta`[i-1,j-1] must be set equal to
        delta_{ij}, for i = 1,2,...,k, for j = 1,2,...,d_i.

        If d = 0, `delta` is not referenced.

    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    mean : str, length 1
        `mean` = 'M', if components of mu have been estimated and `mean`
        = 'Z', if all elements of mu are to be taken as zero.

    par : float, shape (`lpar`)
        Must contain the parameter estimates read in row by row in the
        order phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q, mu.

        Thus,

        - if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to
          an estimate of the (i,j)th element of phi_l, for j =
          1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,p;

        - if `iq` > 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
          equal to an estimate of the (i,j)th element of theta_l, for j
          = 1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,q;

        - if `mean` = 'M', `par`[(p+q)*k*k+i-1] must be set equal to an
          estimate of the ith component of mu, for i = 1,2,...,k.

    lpar : int
        The dimension of the array `par`.

    qq : float, shape (`k`, `k`), modified in place
        On entry: `qq`[i-1,j-1] must contain an estimate of the (i,j)th
        element of Sigma. The lower triangle only is needed.

        On exit: if `errno` != 1, then the upper triangle is set equal
        to the lower triangle.

    v : float, shape (`k`, `n`-d)
        `v`[i-1,t-1] must contain an estimate of the ith component of
        epsilon_{t+d}, for t = 1,2,...,n-d, for i = 1,2,...,k.

        If q = 0, `v` is not used.

    lmax : int
        The number, l_max, of forecasts required.

    predz : float, shape (`k`, `lmax`), modified in place
        On exit: `predz`[i-1,l-1] contains the forecast of z_{i,n+l},
        for l = 1,2,...,l_max, for i = 1,2,...,k.

    sefz : float, shape (`k`, `lmax`), modified in place
        On exit: `sefz`[i-1,l-1] contains an estimate of the standard
        error of the forecast of z_{i,n+l}, for l = 1,2,...,l_max, for i
        = 1,2,...,k.

    ref : float, shape (`lref`), modified in place
        On exit: the reference vector which may be used to update
        forecasts using ``multi_varma_update``. The first
        (`lmax`-1)*`k`*`k` elements contain the psi weight matrices,
        psi_1,psi_2,...,psi_{l_max-1}. The next `k`*`lmax` elements
        contain the forecasts of the transformed series
        Z^_n+1^*,Z^_n+2^*,...,Z^_n+l_max^* and the next `k`*`lmax`
        contain the variances of the forecasts of the transformed
        variables. The last `k` elements are used to store the
        transformations for the series.

    lref : int
        The dimension of the array `ref`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lref` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `lref` >= (`lmax`-1)*`k`*`k`+2*`k`*`lmax`+`k`.

        (`errno` 1)
            On entry, i = *<value>*, `dord`[i-1] = *<value>* and
            `n`-max(`ip`, `iq`) = *<value>*.

            Constraint: 0 <= `dord`[i-1] < `n`-max(`ip`, `iq`).

        (`errno` 1)
            On entry, `lmax` = *<value>*.

            Constraint: `lmax` >= 1.

        (`errno` 1)
            On entry, number of observations = *<value>* and number of
            parameters = *<value>*.

            Constraint: number of observations >= number of parameters.

        (`errno` 1)
            On entry, `lpar` is too small: `lpar` = *<value>* and the
            minimum size required = *<value>*.

        (`errno` 1)
            On entry, `mean` is an invalid character.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `tr`[i-1] is invalid.

            Constraint: `tr`[i-1] = 'N', 'L' or 'S'.

        (`errno` 3)
            On entry, one (or more) of the transformations requested is
            invalid.

        (`errno` 4)
            On entry, the MA parameter matrices are outside the
            invertibility region.

        (`errno` 4)
            On entry, the AR parameter matrices are outside the
            stationarity region.

        (`errno` 4)
            On entry, the covariance matrix `qq` is not positive
            definite.

        (`errno` 5)
            An excessive number of iterations were needed to evaluate
            the eigenvalues of the matrices used to test for
            stationarity and invertibility.

        (`errno` 6)
            The covariance matrix may be nearly non-positive definite.

        (`errno` 7)
            The forecasts will overflow if computed.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_varma_forecast'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13djft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, z[kmax, n].
        utils._BYREF_INT, # Input, kmax.
        utils._EngineCharArrayType.c_type, # Input, tr(len=1)[k].
        utils._EngineIntArrayType.c_type, # Input, dord[k].
        utils._EngineFloat64ArrayType.c_type, # Input, delta[kmax, *].
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, par[lpar].
        utils._BYREF_INT, # Input, lpar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, qq[kmax, k].
        utils._EngineFloat64ArrayType.c_type, # Input, v[kmax, *].
        utils._BYREF_INT, # Input, lmax.
        utils._EngineFloat64ArrayType.c_type, # Output, predz[kmax, lmax].
        utils._EngineFloat64ArrayType.c_type, # Output, sefz[kmax, lmax].
        utils._EngineFloat64ArrayType.c_type, # Output, ref[lref].
        utils._BYREF_INT, # Input, lref.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._BYREF_INT, # Input, liwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of tr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'), (delta, 'delta'), (qq, 'qq'), (v, 'v'), (predz, 'predz'), (sefz, 'sefz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kmax_py = utils._EngineFloat64ArrayType.get_shape(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _kmax = utils._EngineIntScalarType(
        _kmax_py, {'fun_name': fun_name, 'entity': '_kmax'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _tr_cls = utils._EngineCharArrayType(
        tr, {'fun_name': fun_name, 'entity': 'tr'}, exp_chrlen=1,
        exp_shape=(_k_py,),
    )
    _tr = _tr_cls.to_ctypes()
    _tr_chrlen = _tr_cls.chrlen
    _dord_cls = utils._EngineIntArrayType(
        dord, {'fun_name': fun_name, 'entity': 'dord'},
        exp_shape=(_k_py,),
    )
    _dord = _dord_cls.to_ctypes()
    _vl_d_py = _dord_cls.get_max(0, _k_py-1)
    _delta = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_k_py, _vl_d_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _lpar = utils._EngineIntScalarType(
        lpar, {'fun_name': fun_name, 'entity': 'lpar'},
    ).to_ctypes()
    _lpar_py = utils._EngineIntScalarType.to_py(_lpar)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_lpar_py,),
    ).to_ctypes()
    _qq_cls = utils._EngineFloat64ArrayType(
        qq, {'fun_name': fun_name, 'entity': 'qq'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
    )
    _qq = _qq_cls.to_ctypes()
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_k_py, _n_py-_vl_d_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _lmax = utils._EngineIntScalarType(
        lmax, {'fun_name': fun_name, 'entity': 'lmax'},
    ).to_ctypes()
    _lmax_py = utils._EngineIntScalarType.to_py(_lmax)
    _predz_cls = utils._EngineFloat64ArrayType(
        predz, {'fun_name': fun_name, 'entity': 'predz'},
        exp_shape=(_k_py, _lmax_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _predz = _predz_cls.to_ctypes(can_reference_contents=False)
    _sefz_cls = utils._EngineFloat64ArrayType(
        sefz, {'fun_name': fun_name, 'entity': 'sefz'},
        exp_shape=(_k_py, _lmax_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sefz = _sefz_cls.to_ctypes(can_reference_contents=False)
    _lref = utils._EngineIntScalarType(
        lref, {'fun_name': fun_name, 'entity': 'lref'},
    ).to_ctypes()
    _lref_py = utils._EngineIntScalarType.to_py(_lref)
    _ref_cls = utils._EngineFloat64ArrayType(
        ref, {'fun_name': fun_name, 'entity': 'ref'},
        exp_shape=(_lref_py,),
        check_contents=False,
    )
    _ref = _ref_cls.to_ctypes(can_reference_contents=False)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _lwork_py = utils._EngineIntScalarType.to_py(_lwork)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(_lwork_py)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _liwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _k,
        _n,
        _z,
        _kmax,
        _tr,
        _dord,
        _delta,
        _ip,
        _iq,
        _mean,
        _par,
        _lpar,
        _qq,
        _v,
        _lmax,
        _predz,
        _sefz,
        _ref,
        _lref,
        _work,
        _lwork,
        _iwork,
        _liwork,
        _errbuf,
        _ierr,
        _tr_chrlen,
        _mean._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lref = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lref >= (lmax-1)*k*k+2*k*lmax+k.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, dord[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n-max(ip, iq) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 <= dord[i-1] < n-max(ip, iq).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, dord[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n-max(ip, iq) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 <= dord[i-1] < n-max(ip, iq).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lmax = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lmax >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, number of observations = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'number of parameters = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: number of observations >= number of',
                    'parameters.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lpar is too small: lpar = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean is an invalid character.',
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and tr[i-1] is invalid.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tr[i-1] = \'N\', \'L\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, one (or more) of the transformations requested is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the MA parameter matrices are outside the invertibility region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the AR parameter matrices are outside the stationarity region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the covariance matrix qq is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An excessive number of iterations were needed to evaluate the',
                    'eigenvalues of the matrices used to test for stationarity and invertibility.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The covariance matrix may be nearly non-positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The forecasts will overflow if computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _qq_cls.refresh_py(_qq)
        _predz_cls.refresh_py(_predz)
        _sefz_cls.refresh_py(_sefz)
        _ref_cls.refresh_py(_ref)
        return_data = None
    _err_class.manage()
    return return_data

def multi_varma_update(k, lmax, m, mlast, z, ref, lref, v, predz, sefz):
    u"""
    Multivariate time series, updates forecasts and their standard
    errors.

    ``multi_varma_update`` accepts a sequence of new observations in a
    multivariate time series and updates both the forecasts and the
    standard deviations of the forecast errors.
    A call to ``multi_varma_forecast`` must be made prior to calling
    this function in order to calculate the elements of a reference
    vector together with a set of forecasts and their standard errors.
    On a successful exit from ``multi_varma_update`` the reference
    vector is updated so that should future series values become
    available these forecasts may be updated by recalling
    ``multi_varma_update``.

    For full information please refer to the NAG Library document for
    g13dk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dkf.html

    Parameters
    ----------
    k : int
        k, the dimension of the multivariate time series.

    lmax : int
        The number, l_max, of forecasts requested in the call to
        ``multi_varma_forecast``.

    m : int
        m, the number of new observations available since the last call
        to either ``multi_varma_forecast`` or ``multi_varma_update``.
        The number of new observations since the last call to
        ``multi_varma_forecast`` is then `m`+`mlast`.

    mlast : int
        On the first call to ``multi_varma_update``, since calling
        ``multi_varma_forecast``, `mlast` must be set to 0 to indicate
        that no new observations have yet been used to update the
        forecasts; on subsequent calls `mlast` must contain the value of
        `mlast` as output on the previous call to
        ``multi_varma_update``.

    z : float, shape (`k`, `m`)
        `z`[i-1,j-1] must contain the value of z_{i,n+`mlast`+j}, for j
        = 1,2,...,m, for i = 1,2,...,k, and where n is the number of
        observations in the time series in the last call made to
        ``multi_varma_forecast``.

    ref : float, shape (`lref`), modified in place
        On entry: must contain the first
        (`lmax`-1)*`k`*`k`+2*`k`*`lmax`+`k` elements of the reference
        vector as returned on a successful exit from
        ``multi_varma_forecast`` (or a previous call to
        ``multi_varma_update``).

        On exit: the elements of `ref` are updated. The first
        (`lmax`-1)*`k`*`k` elements store the psi weights
        psi_1,psi_2,...,psi_{l_max-1}. The next `k`*`lmax` elements
        contain the forecasts of the transformed series and the next
        `k`*`lmax` elements contain the variances of the forecasts of
        the transformed variables; see ``multi_varma_forecast``. The
        last `k` elements are not updated.

    lref : int
        The dimension of the array `ref`.

    v : float, shape (`k`, `m`), modified in place
        On exit: `v`[i-1,j-1] contains an estimate of the ith component
        of epsilon_{n+`mlast`+j}, for j = 1,2,...,m, for i = 1,2,...,k.

    predz : float, shape (`k`, `lmax`), modified in place
        On entry: nonupdated values are kept intact.

        On exit: `predz`[i-1,j-1] contains the updated forecast of
        z_{i,n+j}, for j = `mlast`+`m`+1,...,l_max, for i = 1,2,...,k.

        The columns of `predz` corresponding to the new observations
        since the last call to either ``multi_varma_forecast`` or
        ``multi_varma_update`` are set equal to the corresponding
        columns of `z`.

    sefz : float, shape (`k`, `lmax`), modified in place
        On entry: nonupdated values are kept intact.

        On exit: `sefz`[i-1,j-1] contains an estimate of the standard
        error of the corresponding element of `predz`, for j =
        `mlast`+`m`+1,...,l_max, for i = 1,2,...,k.

        The columns of `sefz` corresponding to the new observations
        since the last call to either ``multi_varma_forecast`` or
        ``multi_varma_update`` are set equal to zero.

    Returns
    -------
    mlast : int
        Is incremented by m to indicate that `mlast`+`m` observations
        have now been used to update the forecasts since the last call
        to ``multi_varma_forecast``.

        `mlast` must not be changed between calls to
        ``multi_varma_update``, unless a call to
        ``multi_varma_forecast`` has been made between the calls in
        which case `mlast` should be reset to 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lref` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `lref` >= (`lmax`-1)*`k`*`k`+2*`k`*`lmax`+`k`.

        (`errno` 1)
            On entry, `mlast` = *<value>*.

            Constraint: `mlast` >= 0.

        (`errno` 1)
            On entry, `m` = *<value>*, `lmax` = *<value>* and `mlast` =
            *<value>*.

            Constraint: `m` < `lmax`-`mlast`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 1)
            On entry, `lmax` = *<value>*.

            Constraint: `lmax` >= 2.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 2)
            On entry, some of the elements of the array `ref` have been
            corrupted.

        (`errno` 3)
            On entry, one (or more) of the transformations requested is
            invalid.

        (`errno` 4)
            The updated forecasts will overflow if computed.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_varma_update'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, lmax.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input/Output, mlast.
        utils._EngineFloat64ArrayType.c_type, # Input, z[kmax, m].
        utils._BYREF_INT, # Input, kmax.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ref[lref].
        utils._BYREF_INT, # Input, lref.
        utils._EngineFloat64ArrayType.c_type, # Output, v[kmax, m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, predz[kmax, lmax].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, sefz[kmax, lmax].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'), (v, 'v'), (predz, 'predz'), (sefz, 'sefz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _lmax = utils._EngineIntScalarType(
        lmax, {'fun_name': fun_name, 'entity': 'lmax'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _mlast = utils._EngineIntScalarType(
        mlast, {'fun_name': fun_name, 'entity': 'mlast'},
    ).to_ctypes()
    _kmax_py = utils._EngineFloat64ArrayType.get_shape(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _kmax = utils._EngineIntScalarType(
        _kmax_py, {'fun_name': fun_name, 'entity': '_kmax'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_k_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _lref = utils._EngineIntScalarType(
        lref, {'fun_name': fun_name, 'entity': 'lref'},
    ).to_ctypes()
    _lref_py = utils._EngineIntScalarType.to_py(_lref)
    _ref_cls = utils._EngineFloat64ArrayType(
        ref, {'fun_name': fun_name, 'entity': 'ref'},
        exp_shape=(_lref_py,),
    )
    _ref = _ref_cls.to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_k_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _v = _v_cls.to_ctypes(can_reference_contents=False)
    _lmax_py = utils._EngineIntScalarType.to_py(_lmax)
    _predz_cls = utils._EngineFloat64ArrayType(
        predz, {'fun_name': fun_name, 'entity': 'predz'},
        exp_shape=(_k_py, _lmax_py),
        mutual_sorder=_m_sorder,
    )
    _predz = _predz_cls.to_ctypes()
    _sefz_cls = utils._EngineFloat64ArrayType(
        sefz, {'fun_name': fun_name, 'entity': 'sefz'},
        exp_shape=(_k_py, _lmax_py),
        mutual_sorder=_m_sorder,
    )
    _sefz = _sefz_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _k,
        _lmax,
        _m,
        _mlast,
        _z,
        _kmax,
        _ref,
        _lref,
        _v,
        _predz,
        _sefz,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lref = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'the minimum size required = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lref >= (lmax-1)*k*k+2*k*lmax+k.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mlast = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mlast >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'lmax = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'mlast = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m < lmax-mlast.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lmax = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lmax >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array ref have been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, one (or more) of the transformations requested is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The updated forecasts will overflow if computed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _mlast_py = utils._EngineIntScalarType.to_py(_mlast)
        _ref_cls.refresh_py(_ref)
        _v_cls.refresh_py(_v)
        _predz_cls.refresh_py(_predz)
        _sefz_cls.refresh_py(_sefz)
        return_data = _mlast_py
    _err_class.manage()
    return return_data

def multi_diff(k, n, z, tr, dord, delta, w):
    u"""
    Multivariate time series, differences and/or transforms.

    ``multi_diff`` differences and/or transforms a multivariate time
    series. It is intended to be used prior to ``multi_varma_estimate``
    to fit a vector autoregressive moving average (VARMA) model to the
    differenced/transformed series.

    For full information please refer to the NAG Library document for
    g13dl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dlf.html

    Parameters
    ----------
    k : int
        k, the dimension of the multivariate time series.

    n : int
        n, the number of observations in the series, prior to
        differencing.

    z : float, shape (`k`, `n`)
        `z`[i,t] must contain, z_{it}, the ith component of Z_t, for t =
        1,2,...,n, for i = 1,2,...,k.

    tr : str, length 1, shape (`k`)
        `tr`[i-1] indicates whether the ith time series is to be
        transformed, for i = 1,2,...,k.

        `tr`[i-1] = 'N'

          No transformation is used.

        `tr`[i-1] = 'L'

          A log transformation is used.

        `tr`[i-1] = 'S'

          A square root transformation is used.

    dord : int, shape (`k`)
        The order of differencing for each series, d_1,d_2,...,d_k.

    delta : float, shape (`k`, d)
        If `dord`[i-1] > 0, then `delta`[i-1,j-1] must be set equal to
        delta_{ij}, for i = 1,2,...,k, for j = 1,2,...,d_i.

        If d = 0, `delta` is not referenced.

    w : float, shape (`k`, `n`-d), modified in place
        On exit: `w`[i-1,t-1] contains the value of w_{i,t+d}, for t =
        1,2,...,n-d, for i = 1,2,...,k.

    Returns
    -------
    nd : int
        The number of differenced values, n-d, in the series, where d =
        max(`dord`[i-1]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `dord`[i-1] = *<value>* and `n` =
            *<value>*.

            Constraint: `dord` < `n`.

        (`errno` 2)
            On entry, i = *<value>* and `dord`[i-1] = *<value>*.

            Constraint: `dord`[i-1] >= 0.

        (`errno` 3)
            On entry, i = *<value>* and `tr`[i-1] is invalid.

            Constraint: `tr`[i-1] = 'N', 'L' or 'S'.

        (`errno` 4)
            On entry, one (or more) of the transformations requested is
            invalid.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_diff'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, z[ldz, n].
        utils._BYREF_INT, # Input, ldz.
        utils._EngineCharArrayType.c_type, # Input, tr(len=1)[k].
        utils._EngineIntArrayType.c_type, # Input, dord[k].
        utils._EngineFloat64ArrayType.c_type, # Input, delta[lddelta, *].
        utils._BYREF_INT, # Input, lddelta.
        utils._EngineFloat64ArrayType.c_type, # Output, w[ldw, *].
        utils._BYREF_INT, # Input, ldw.
        utils._BYREF_INT, # Output, nd.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of tr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'), (delta, 'delta'), (w, 'w')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldz_py = utils._EngineFloat64ArrayType.get_nd(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldz = utils._EngineIntScalarType(
        _ldz_py, {'fun_name': fun_name, 'entity': '_ldz'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _tr_cls = utils._EngineCharArrayType(
        tr, {'fun_name': fun_name, 'entity': 'tr'}, exp_chrlen=1,
        exp_shape=(_k_py,),
    )
    _tr = _tr_cls.to_ctypes()
    _tr_chrlen = _tr_cls.chrlen
    _dord_cls = utils._EngineIntArrayType(
        dord, {'fun_name': fun_name, 'entity': 'dord'},
        exp_shape=(_k_py,),
    )
    _dord = _dord_cls.to_ctypes()
    _lddelta_py = utils._EngineFloat64ArrayType.get_nd(delta, {'fun_name': fun_name, 'entity': 'delta'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lddelta = utils._EngineIntScalarType(
        _lddelta_py, {'fun_name': fun_name, 'entity': '_lddelta'},
    ).to_ctypes()
    _vl_d_py = _dord_cls.get_max(0, _k_py-1)
    _delta = utils._EngineFloat64ArrayType(
        delta, {'fun_name': fun_name, 'entity': 'delta'},
        exp_shape=(_k_py, _vl_d_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldw_py = utils._EngineFloat64ArrayType.get_nd(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldw = utils._EngineIntScalarType(
        _ldw_py, {'fun_name': fun_name, 'entity': '_ldw'},
    ).to_ctypes()
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_k_py, _n_py-_vl_d_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _w = _w_cls.to_ctypes(can_reference_contents=False)
    _nd = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _k,
        _n,
        _z,
        _ldz,
        _tr,
        _dord,
        _delta,
        _lddelta,
        _w,
        _ldw,
        _nd,
        _work,
        _errbuf,
        _ierr,
        _tr_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, dord[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: dord < n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and dord[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: dord[i-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and tr[i-1] is invalid.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tr[i-1] = \'N\', \'L\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, one (or more) of the transformations requested is invalid.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _w_cls.refresh_py(_w)
        _nd_py = utils._EngineIntScalarType.to_py(_nd)
        return_data = _nd_py
    _err_class.manage()
    return return_data

def multi_corrmat_cross(matrix, k, n, m, w, wmean, r0, r):
    u"""
    Multivariate time series, sample cross-correlation or
    cross-covariance matrices.

    ``multi_corrmat_cross`` calculates the sample cross-correlation (or
    cross-covariance) matrices of a multivariate time series.

    For full information please refer to the NAG Library document for
    g13dm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dmf.html

    Parameters
    ----------
    matrix : str, length 1
        Indicates whether the cross-covariance or cross-correlation
        matrices are to be computed.

        `matrix` = 'V'

          The cross-covariance matrices are computed.

        `matrix` = 'R'

          The cross-correlation matrices are computed.

    k : int
        k, the dimension of the multivariate time series.

    n : int
        n, the number of observations in the series.

    m : int
        m, the number of cross-correlation (or cross-covariance)
        matrices to be computed. If in doubt set `m` = 10. However it
        should be noted that `m` is usually taken to be at most `n`/4.

    w : float, shape (`k`, `n`)
        `w`[i-1,t-1] must contain the observation w_{it}, for t =
        1,2,...,n, for i = 1,2,...,k.

    wmean : float, shape (`k`), modified in place
        On exit: the means, w__i, for i = 1,2,...,k.

    r0 : float, shape (`k`, `k`), modified in place
        On exit: if i != j, then `r0`[i-1,j-1] contains an estimate of
        the (i,j)th element of the cross-correlation (or
        cross-covariance) matrix at lag zero, R^_{ij}(0); if i = j, then
        if `matrix` = 'V', `r0`[i-1,i-1] contains the variance of the
        ith series, C^_{ii}(0), and if `matrix` = 'R', `r0`[i-1,i-1]
        contains the standard deviation of the ith series,
        sqrt(C^_{ii}(0)).

        If `errno` = 2 and `matrix` = 'R', then on exit all the elements
        in `r0` whose computation involves the zero variance are set to
        zero.

    r : float, shape (`k`, `k`, `m`), modified in place
        On exit: `r`[i-1,j-1,l-1] contains an estimate of the (i,j)th
        element of the cross-correlation (or cross-covariance) at lag l,
        R^_{ij}(l), for j = 1,2,...,k, for i = 1,2,...,k, for l =
        1,2,...,m.

        If `errno` = 2 and `matrix` = 'R', then on exit all the elements
        in `r` whose computation involves the zero variance are set to
        zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 1 and `m` < `n`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 1)
            On entry, `matrix` = *<value>*.

            Constraint: `matrix` = 'V' or 'R'.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, at least one of the series is such that all its
            elements are practically identical giving zero (or near
            zero) variance.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_corrmat_cross'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, matrix(len=1).
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, w[kmax, n].
        utils._BYREF_INT, # Input, kmax.
        utils._EngineFloat64ArrayType.c_type, # Output, wmean[k].
        utils._EngineFloat64ArrayType.c_type, # Output, r0[kmax, k].
        utils._EngineFloat64ArrayType.c_type, # Output, r[kmax, kmax, m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of matrix.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((w, 'w'), (r0, 'r0'), (r, 'r')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _matrix = utils._EngineCharScalarType(
        matrix, {'fun_name': fun_name, 'entity': 'matrix'}, exp_chrlen=1,
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _kmax_py = utils._EngineFloat64ArrayType.get_shape(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[0]
    _kmax = utils._EngineIntScalarType(
        _kmax_py, {'fun_name': fun_name, 'entity': '_kmax'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _wmean_cls = utils._EngineFloat64ArrayType(
        wmean, {'fun_name': fun_name, 'entity': 'wmean'},
        exp_shape=(_k_py,),
        check_contents=False,
    )
    _wmean = _wmean_cls.to_ctypes(can_reference_contents=False)
    _r0_cls = utils._EngineFloat64ArrayType(
        r0, {'fun_name': fun_name, 'entity': 'r0'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r0 = _r0_cls.to_ctypes(can_reference_contents=False)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_k_py, _k_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _matrix,
        _k,
        _n,
        _m,
        _w,
        _kmax,
        _wmean,
        _r0,
        _r,
        _errbuf,
        _ierr,
        _matrix._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= 1 and m < n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, matrix = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: matrix = \'V\' or \'R\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one of the series is such that all its elements are',
                    'practically identical giving zero (or near zero) variance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _wmean_cls.refresh_py(_wmean)
        _r0_cls.refresh_py(_r0)
        _r_cls.refresh_py(_r)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def multi_corrmat_partlag(k, n, m, r0, r, parlag, x, pvalue):
    u"""
    Multivariate time series, sample partial lag correlation matrices,
    chi^2 statistics and significance levels.

    ``multi_corrmat_partlag`` calculates the sample partial lag
    correlation matrices of a multivariate time series.
    A set of chi^2-statistics and their significance levels are also
    returned.
    A call to ``multi_corrmat_cross`` is usually made prior to calling
    this function in order to calculate the sample cross-correlation
    matrices.

    For full information please refer to the NAG Library document for
    g13dn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dnf.html

    Parameters
    ----------
    k : int
        k, the dimension of the multivariate time series.

    n : int
        n, the number of observations in each series.

    m : int
        m, the number of partial lag correlation matrices to be
        computed. Note this also specifies the number of sample
        cross-correlation matrices that must be contained in the array
        `r`.

    r0 : float, shape (`k`, `k`)
        If i != j, then `r0`[i-1,j-1] must contain the (i,j)th element
        of the sample cross-correlation matrix at lag zero, R^_{ij}(0).
        If i = j, then `r0`[i-1,i-1] must contain the standard deviation
        of the ith series.

    r : float, shape (`k`, `k`, `m`)
        `r`[i-1,j-1,l-1] must contain the (i,j)th element of the sample
        cross-correlation at lag l, R^_{ij}(l), for j = 1,2,...,k, for i
        = 1,2,...,k, for l = 1,2,...,m, where series j leads series i
        (see [Further Comments]).

    parlag : float, shape (`k`, `k`, `m`), modified in place
        On exit: `parlag`[i-1,j-1,l-1] contains the (i,j)th element of
        the sample partial lag correlation matrix at lag l, P^_{ij}(l),
        for j = 1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,`maxlag`.

    x : float, shape (`m`), modified in place
        On exit: `x`[l-1] contains the chi^2-statistic at lag l, for l =
        1,2,...,`maxlag`.

    pvalue : float, shape (`m`), modified in place
        On exit: `pvalue`[l-1] contains the significance level of the
        corresponding chi^2-statistic in `x`, for l = 1,2,...,`maxlag`.

    Returns
    -------
    maxlag : int
        The maximum lag up to which partial lag correlation matrices
        (along with chi^2-statistics and their significance levels) have
        been successfully computed. On a successful exit `maxlag` will
        equal `m`. If `errno` = 2 on exit, `maxlag` will be less than
        `m`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` < `n`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The recursive equations used to compute the partial lag
            correlation matrices are ill-conditioned (they have been
            computed up to lag *<value>*).
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_corrmat_partlag'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, kmax.
        utils._EngineFloat64ArrayType.c_type, # Input, r0[kmax, k].
        utils._EngineFloat64ArrayType.c_type, # Input, r[kmax, kmax, m].
        utils._BYREF_INT, # Output, maxlag.
        utils._EngineFloat64ArrayType.c_type, # Output, parlag[kmax, kmax, m].
        utils._EngineFloat64ArrayType.c_type, # Output, x[m].
        utils._EngineFloat64ArrayType.c_type, # Output, pvalue[m].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((r0, 'r0'), (r, 'r'), (parlag, 'parlag')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _kmax_py = utils._EngineFloat64ArrayType.get_shape(r0, {'fun_name': fun_name, 'entity': 'r0'}, exp_rank=2)[0]
    _kmax = utils._EngineIntScalarType(
        _kmax_py, {'fun_name': fun_name, 'entity': '_kmax'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _r0 = utils._EngineFloat64ArrayType(
        r0, {'fun_name': fun_name, 'entity': 'r0'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_k_py, _k_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _maxlag = utils._EngineIntScalarType.empty_ctypes()
    _parlag_cls = utils._EngineFloat64ArrayType(
        parlag, {'fun_name': fun_name, 'entity': 'parlag'},
        exp_shape=(_k_py, _k_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _parlag = _parlag_cls.to_ctypes(can_reference_contents=False)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _pvalue_cls = utils._EngineFloat64ArrayType(
        pvalue, {'fun_name': fun_name, 'entity': 'pvalue'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _pvalue = _pvalue_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _k,
        _n,
        _m,
        _kmax,
        _r0,
        _r,
        _maxlag,
        _parlag,
        _x,
        _pvalue,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m < n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The recursive equations used to compute the partial lag correlation',
                    'matrices are ill-conditioned (they have been computed up to lag',
                    '{:s}).'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _parlag_cls.refresh_py(_parlag)
        _x_cls.refresh_py(_x)
        _pvalue_cls.refresh_py(_pvalue)
        _maxlag_py = utils._EngineIntScalarType.to_py(_maxlag)
        return_data = _maxlag_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['maxlag'],
            )(
                _maxlag_py,
            )
    _err_class.manage()
    return return_data

def multi_regmat_partial(k, n, z, m, parlag, se, qq, x, pvalue, loglhd):
    u"""
    Multivariate time series, partial autoregression matrices.

    ``multi_regmat_partial`` calculates the sample partial
    autoregression matrices of a multivariate time series.
    A set of likelihood ratio statistics and their significance levels
    are also returned.
    These quantities are useful for determining whether the series
    follows an autoregressive model and, if so, of what order.

    For full information please refer to the NAG Library document for
    g13dp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dpf.html

    Parameters
    ----------
    k : int
        k, the number of time series.

    n : int
        n, the number of observations in the time series.

    z : float, shape (`k`, `n`)
        `z`[i-1,t-1] must contain the observation w_{it}, for t =
        1,2,...,n, for i = 1,2,...,k.

    m : int
        m, the number of partial autoregression matrices to be computed.
        If in doubt set `m` = 10.

    parlag : float, shape (`k`, `k`, `m`), modified in place
        On exit: `parlag`[i-1,j-1,l-1] contains an estimate of the
        (i,j)th element of the partial autoregression matrix at lag l,
        P^_l(ij), for l = 1,2,...,`maxlag`, i = 1,2,...,k and j =
        1,2,...,k.

    se : float, shape (`k`, `k`, `m`), modified in place
        On exit: `se`[i-1,j-1,l-1] contains an estimate of the standard
        error of the corresponding element in the array `parlag`.

    qq : float, shape (`k`, `k`, `m`), modified in place
        On exit: `qq`[i-1,j-1,l-1] contains an estimate of the (i,j)th
        element of the corresponding variance-covariance matrix
        Sigma^_l, for j = 1,2,...,k, for i = 1,2,...,k, for l =
        1,2,...,`maxlag`.

    x : float, shape (`m`), modified in place
        On exit: `x`[l-1] contains X_l, the likelihood ratio statistic
        at lag l, for l = 1,2,...,`maxlag`.

    pvalue : float, shape (`m`), modified in place
        On exit: `pvalue`[l-1] contains the significance level of the
        statistic in the corresponding element of `x`.

    loglhd : float, shape (`m`), modified in place
        On exit: `loglhd`[l-1] contains an estimate of the maximum of
        the log-likelihood function when an AR(l-1) model has been
        fitted to the series, for l = 1,2,...,`maxlag`.

    Returns
    -------
    maxlag : int
        The maximum lag up to which partial autoregression matrices
        (along with their likelihood ratio statistics and their
        significance levels) have been successfully computed. On a
        successful exit `maxlag` will equal `m`. If `errno` = 2 on exit
        then `maxlag` will be less than `m`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `k` = *<value>*, `m` = *<value>* and `n` =
            *<value>*.

            Constraint: `n`-`m`-(`k`*`m`+1) >= `k`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 4.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The recursive equations used to compute the partial
            autoregression matrices are ill-conditioned. They have been
            computed up to lag *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_regmat_partial'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, z[kmax, n].
        utils._BYREF_INT, # Input, kmax.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Output, maxlag.
        utils._EngineFloat64ArrayType.c_type, # Output, parlag[kmax, kmax, m].
        utils._EngineFloat64ArrayType.c_type, # Output, se[kmax, kmax, m].
        utils._EngineFloat64ArrayType.c_type, # Output, qq[kmax, kmax, m].
        utils._EngineFloat64ArrayType.c_type, # Output, x[m].
        utils._EngineFloat64ArrayType.c_type, # Output, pvalue[m].
        utils._EngineFloat64ArrayType.c_type, # Output, loglhd[m].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'), (parlag, 'parlag'), (se, 'se'), (qq, 'qq')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kmax_py = utils._EngineFloat64ArrayType.get_shape(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _kmax = utils._EngineIntScalarType(
        _kmax_py, {'fun_name': fun_name, 'entity': '_kmax'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _maxlag = utils._EngineIntScalarType.empty_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _parlag_cls = utils._EngineFloat64ArrayType(
        parlag, {'fun_name': fun_name, 'entity': 'parlag'},
        exp_shape=(_k_py, _k_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _parlag = _parlag_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_k_py, _k_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _qq_cls = utils._EngineFloat64ArrayType(
        qq, {'fun_name': fun_name, 'entity': 'qq'},
        exp_shape=(_k_py, _k_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _qq = _qq_cls.to_ctypes(can_reference_contents=False)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _pvalue_cls = utils._EngineFloat64ArrayType(
        pvalue, {'fun_name': fun_name, 'entity': 'pvalue'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _pvalue = _pvalue_cls.to_ctypes(can_reference_contents=False)
    _loglhd_cls = utils._EngineFloat64ArrayType(
        loglhd, {'fun_name': fun_name, 'entity': 'loglhd'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _loglhd = _loglhd_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _k,
        _n,
        _z,
        _kmax,
        _m,
        _maxlag,
        _parlag,
        _se,
        _qq,
        _x,
        _pvalue,
        _loglhd,
        _work,
        _lwork,
        _iwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}, m = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: n-m-(k*m+1) >= k.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The recursive equations used to compute the partial autoregression',
                    'matrices are ill-conditioned. They have been computed up to',
                    'lag {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _parlag_cls.refresh_py(_parlag)
        _se_cls.refresh_py(_se)
        _qq_cls.refresh_py(_qq)
        _x_cls.refresh_py(_x)
        _pvalue_cls.refresh_py(_pvalue)
        _loglhd_cls.refresh_py(_loglhd)
        _maxlag_py = utils._EngineIntScalarType.to_py(_maxlag)
        return_data = _maxlag_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['maxlag'],
            )(
                _maxlag_py,
            )
    _err_class.manage()
    return return_data

def multi_varma_diag(k, n, v, ip, iq, m, par, parhld, qq, ishow, r0, r, rcm, io_manager=None):
    u"""
    Multivariate time series, diagnostic checking of residuals,
    following ``multi_varma_estimate``.

    ``multi_varma_diag`` is a diagnostic checking function suitable for
    use after fitting a vector ARMA model to a multivariate time series
    using ``multi_varma_estimate``.
    The residual cross-correlation matrices are returned along with an
    estimate of their asymptotic standard errors and correlations.
    Also, ``multi_varma_diag`` calculates the modified Li--McLeod
    portmanteau statistic and its significance level for testing model
    adequacy.

    For full information please refer to the NAG Library document for
    g13ds

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dsf.html

    Parameters
    ----------
    k : int
        k, the number of residual time series.

    n : int
        n, the number of observations in each residual series.

    v : float, shape (`k`, `n`)
        `v`[i-1,t-1] must contain an estimate of the ith component of
        epsilon_t, for t = 1,2,...,n, for i = 1,2,...,k.

    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    m : int
        The value of m, the number of residual cross-correlation
        matrices to be computed. See [Further Comments] for advice on
        the choice of `m`.

    par : float, shape ((`ip`+`iq`)*`k`*`k`)
        The parameter estimates read in row by row in the order
        phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q.

        Thus,

        - if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to
          an estimate of the (i,j)th element of phi_l, for i =
          1,2,...,k, for l = 1,2,...,p;

        - if `iq` >= 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
          equal to an estimate of the (i,j)th element of theta_l, for i
          = 1,2,...,k, for l = 1,2,...,q.

        The first p*k*k elements of `par` must satisfy the stationarity
        condition and the next q*k*k elements of `par` must satisfy the
        invertibility condition.

    parhld : bool, shape ((`ip`+`iq`)*`k`*`k`)
        `parhld`[i-1] must be set to True if `par`[i-1] has been held
        constant at a pre-specified value and False if `par`[i-1] is a
        free parameter, for i = 1,2,...,(p+q)*k*k.

    qq : float, shape (`k`, `k`), modified in place
        On entry: `qq`[i-1,j-1] is an efficient estimate of the (i,j)th
        element of Sigma. The lower triangle only is needed.

        On exit: if `errno` != 1, then the upper triangle is set equal
        to the lower triangle.

    ishow : int
        Must be nonzero if the residual cross-correlation matrices
        {r^_{ij}(l)} and their standard errors {se(r^_{ij}(l))}, the
        modified portmanteau statistic with its significance and a
        summary table are to be printed. The summary table indicates
        which elements of the residual correlation matrices are
        significant at the 5% level in either a positive or negative
        direction; i.e., if r^_{ij}(l) > 1.96*se(r^_{ij}(l)) then a '+'
        is printed, if r^_{ij}(l) < -1.96*se(r^_{ij}(l)) then a '-' is
        printed, otherwise a fullstop (.) is printed. The summary table
        is only printed if k <= 6 on entry.

        The residual cross-correlation matrices, their standard errors
        and the modified portmanteau statistic with its significance are
        available also as output variables in `r`, `rcm`, `chi`, `idf`
        and `siglev`.

    r0 : float, shape (`k`, `k`), modified in place
        On exit: if i != j, then `r0`[i-1,j-1] contains an estimate of
        the (i,j)th element of the residual cross-correlation matrix at
        lag zero, R^_0. When i = j, `r0`[i-1,j-1] contains the standard
        deviation of the ith residual series. If `errno` = 3 on exit
        then the first `k` rows and columns of `r0` are set to zero.

    r : float, shape (`k`, `k`, `m`), modified in place
        On exit: `r`[l-1,i-1,j-1] is an estimate of the (i,j)th element
        of the residual cross-correlation matrix at lag l, for l =
        1,2,...,m, for j = 1,2,...,k, for i = 1,2,...,k. If `errno` = 3
        on exit then all elements of `r` are set to zero.

    rcm : float, shape (`m`*`k`*`k`, `m`*`k`*`k`), modified in place
        On exit: the estimated standard errors and correlations of the
        elements in the array `r`. The correlation between
        `r`[l-1,i-1,j-1] and `r`[l_2-1,i_2-1,j_2-1] is returned as
        `rcm`[s,t] where s = (l-1)*k*k+(j-1)*k+i and t =
        (l_2-1)*k*k+(j_2-1)*k+i_2 except that if s = t, then `rcm`[s,t]
        contains the standard error of `r`[l-1,i-1,j-1]. If on exit,
        `errno` >= 5, then all off-diagonal elements of `rcm` are set to
        zero and all diagonal elements are set to 1/sqrt(n).

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    chi : float
        The value of the modified portmanteau statistic, Q_(m)^*. If
        `errno` = 3 on exit then `chi` is returned as zero.

    idf : int
        The number of degrees of freedom of `chi`.

    siglev : float
        The significance level of `chi` based on `idf` degrees of
        freedom. If `errno` = 3 on exit, `siglev` is returned as one.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ldrcm = *<value>*, `m` = *<value>* and `k` =
            *<value>*.

            Constraint: ldrcm >= `m`*`k`*`k`.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` < `n`.

        (`errno` 1)
            On entry, `m` = *<value>*, `ip` = *<value>* and `iq` =
            *<value>*.

            Constraint: `m` > `ip`+`iq`.

        (`errno` 1)
            On entry, `ip` = 0 and `iq` = 0.

            Constraint: `ip` = `iq` = 0 must not hold.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 2)
            On entry, the MA parameter matrices are outside the
            invertibility region.

        (`errno` 2)
            On entry, the AR parameter matrices are outside the
            stationarity region.

        (`errno` 2)
            On entry, the covariance matrix `qq` is not positive
            definite.

        (`errno` 4)
            Excessive iterations needed to find zeros of determinental
            polynomials.

        (`errno` 5)
            On entry, the AR operator has a factor in common with the MA
            operator.

        (`errno` 6)
            The matrix `rcm` could not be computed because one of its
            diagonal elements was found to be non-positive.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one of the residual series in the array
            `v` has near-zero variance.

        (`errno` 3)
            On entry, at least two of the residual series are identical.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_varma_diag'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, v[kmax, n].
        utils._BYREF_INT, # Input, kmax.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, par[(ip+iq)*k*k].
        utils._EngineLogicalArrayType.c_type, # Input, parhld[(ip+iq)*k*k].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, qq[kmax, k].
        utils._BYREF_INT, # Input, ishow.
        utils._BYREF_INT, # Input, nadv.
        utils._EngineFloat64ArrayType.c_type, # Output, r0[kmax, k].
        utils._EngineFloat64ArrayType.c_type, # Output, r[kmax, kmax, m].
        utils._EngineFloat64ArrayType.c_type, # Output, rcm[ldrcm, m*k*k].
        utils._BYREF_INT, # Input, ldrcm.
        utils._BYREF_FLOAT64, # Output, chi.
        utils._BYREF_INT, # Output, idf.
        utils._BYREF_FLOAT64, # Output, siglev.
        utils._EngineIntArrayType.c_type, # Workspace, iw.
        utils._BYREF_INT, # Input, liw.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((v, 'v'), (qq, 'qq'), (r0, 'r0'), (r, 'r'), (rcm, 'rcm')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kmax_py = utils._EngineFloat64ArrayType.get_shape(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2)[0]
    _kmax = utils._EngineIntScalarType(
        _kmax_py, {'fun_name': fun_name, 'entity': '_kmax'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=((_ip_py+_iq_py)*_k_py*_k_py,),
    ).to_ctypes()
    _parhld = utils._EngineLogicalArrayType(
        parhld, {'fun_name': fun_name, 'entity': 'parhld'},
        exp_shape=((_ip_py+_iq_py)*_k_py*_k_py,),
    ).to_ctypes()
    _qq_cls = utils._EngineFloat64ArrayType(
        qq, {'fun_name': fun_name, 'entity': 'qq'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
    )
    _qq = _qq_cls.to_ctypes()
    _ishow = utils._EngineIntScalarType(
        ishow, {'fun_name': fun_name, 'entity': 'ishow'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _r0_cls = utils._EngineFloat64ArrayType(
        r0, {'fun_name': fun_name, 'entity': 'r0'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r0 = _r0_cls.to_ctypes(can_reference_contents=False)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_k_py, _k_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ldrcm_py = utils._EngineFloat64ArrayType.get_nd(rcm, {'fun_name': fun_name, 'entity': 'rcm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrcm = utils._EngineIntScalarType(
        _ldrcm_py, {'fun_name': fun_name, 'entity': '_ldrcm'},
    ).to_ctypes()
    _rcm_cls = utils._EngineFloat64ArrayType(
        rcm, {'fun_name': fun_name, 'entity': 'rcm'},
        exp_shape=(_m_py*_k_py*_k_py, _m_py*_k_py*_k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rcm = _rcm_cls.to_ctypes(can_reference_contents=False)
    _chi = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _siglev = utils._EngineFloat64ScalarType.empty_ctypes()
    _iw = utils._EngineIntArrayType.empty_ctypes(0)
    _liw = utils._EngineIntScalarType.c_type(0)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _k,
        _n,
        _v,
        _kmax,
        _ip,
        _iq,
        _m,
        _par,
        _parhld,
        _qq,
        _ishow,
        _nadv,
        _r0,
        _r,
        _rcm,
        _ldrcm,
        _chi,
        _idf,
        _siglev,
        _iw,
        _liw,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldrcm = {:s}, m = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ldrcm >= m*k*k.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m < n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}, ip = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m > ip+iq.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = 0 and iq = 0.',
                    'Constraint: ip = iq = 0 must not hold.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the MA parameter matrices are outside the invertibility region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the AR parameter matrices are outside the stationarity region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the covariance matrix qq is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99976:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one of the residual series in the array v',
                    'has near-zero variance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99977:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least two of the residual series are identical.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Excessive iterations needed to find zeros of',
                    'determinental polynomials.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99975:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the AR operator has a factor in common with the MA operator.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99974:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix rcm could not be computed because',
                    'one of its diagonal elements was found to be non-positive.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _qq_cls.refresh_py(_qq)
        _r0_cls.refresh_py(_r0)
        _r_cls.refresh_py(_r)
        _rcm_cls.refresh_py(_rcm)
        _chi_py = utils._EngineFloat64ScalarType.to_py(_chi)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _siglev_py = utils._EngineFloat64ScalarType.to_py(_siglev)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['chi', 'idf', 'siglev'],
        )(
            _chi_py,
            _idf_py,
            _siglev_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_arma_roots(k, ip, par, rr, ri, rmod):
    u"""
    Calculates the zeros of a vector autoregressive (or moving average)
    operator.

    ``uni_arma_roots`` calculates the zeros of a vector autoregressive
    (or moving average) operator. This function is likely to be used in
    conjunction with ``rand.times_mv_varma``, ``uni_arima_resid``,
    ``multi_varma_estimate`` or ``multi_varma_diag``.

    For full information please refer to the NAG Library document for
    g13dx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dxf.html

    Parameters
    ----------
    k : int
        k, the dimension of the multivariate time series.

    ip : int
        The number of AR (or MA) parameter matrices, p (or q).

    par : float, shape (`ip`*`k`*`k`)
        The AR (or MA) parameter matrices read in row by row in the
        order phi_1,phi_2,...,phi_p (or theta_1,theta_2,...,theta_q).
        That is, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to the
        (i,j)th element of phi_l, for l = 1,2,...,p (or the (i,j)th
        element of theta_l, for l = 1,2,...,q).

    rr : float, shape (`k`*`ip`), modified in place
        On exit: the real parts of the eigenvalues.

    ri : float, shape (`k`*`ip`), modified in place
        On exit: the imaginary parts of the eigenvalues.

    rmod : float, shape (`k`*`ip`), modified in place
        On exit: the moduli of the eigenvalues.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 2)
            An excessive number of iterations have been required to
            calculate the eigenvalues.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_arma_roots'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13dxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, par[ip*k*k].
        utils._EngineFloat64ArrayType.c_type, # Output, rr[k*ip].
        utils._EngineFloat64ArrayType.c_type, # Output, ri[k*ip].
        utils._EngineFloat64ArrayType.c_type, # Output, rmod[k*ip].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineIntArrayType.c_type, # Workspace, iwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _par = utils._EngineFloat64ArrayType(
        par, {'fun_name': fun_name, 'entity': 'par'},
        exp_shape=(_ip_py*_k_py*_k_py,),
    ).to_ctypes()
    _rr_cls = utils._EngineFloat64ArrayType(
        rr, {'fun_name': fun_name, 'entity': 'rr'},
        exp_shape=(_k_py*_ip_py,),
        check_contents=False,
    )
    _rr = _rr_cls.to_ctypes(can_reference_contents=False)
    _ri_cls = utils._EngineFloat64ArrayType(
        ri, {'fun_name': fun_name, 'entity': 'ri'},
        exp_shape=(_k_py*_ip_py,),
        check_contents=False,
    )
    _ri = _ri_cls.to_ctypes(can_reference_contents=False)
    _rmod_cls = utils._EngineFloat64ArrayType(
        rmod, {'fun_name': fun_name, 'entity': 'rmod'},
        exp_shape=(_k_py*_ip_py,),
        check_contents=False,
    )
    _rmod = _rmod_cls.to_ctypes(can_reference_contents=False)
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _k,
        _ip,
        _par,
        _rr,
        _ri,
        _rmod,
        _work,
        _iwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An excessive number of iterations have been required to',
                    'calculate the eigenvalues.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rr_cls.refresh_py(_rr)
        _ri_cls.refresh_py(_ri)
        _rmod_cls.refresh_py(_rmod)
        return_data = None
    _err_class.manage()
    return return_data

def multi_kalman_sqrt_var(n, m, l, a, b, stq, q, c, r, s, k, h, tol):
    u"""
    Combined measurement and time update, one iteration of Kalman
    filter, time-varying, square root covariance filter.

    ``multi_kalman_sqrt_var`` performs a combined measurement and time
    update of one iteration of the time-varying Kalman filter using a
    square root covariance filter.

    For full information please refer to the NAG Library document for
    g13ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13eaf.html

    Parameters
    ----------
    n : int
        n, the size of the state vector.

    m : int
        m, the size of the observation vector.

    l : int
        l, the dimension of the state noise.

    a : float, shape (`n`, `n`)
        The state transition matrix, A_i.

    b : float, shape (`n`, `l`)
        The noise coefficient matrix B_i.

    stq : bool
        If `stq` = True, the state noise covariance matrix Q_i is
        assumed to be the identity matrix. Otherwise the lower
        triangular Cholesky factor, Q_i^1/2, must be provided in `q`.

    q : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if not((`stq`): `l`; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if not((`stq`): `l`; otherwise: 0.

        If `stq` = False, `q` must contain the lower triangular Cholesky
        factor of the state noise covariance matrix, Q_i^1/2. Otherwise
        `q` is not referenced.

    c : float, shape (`m`, `n`)
        The measurement coefficient matrix, C_i.

    r : float, shape (`m`, `m`)
        The lower triangular Cholesky factor of the measurement noise
        covariance matrix R_i^1/2.

    s : float, shape (`n`, `n`), modified in place
        On entry: the lower triangular Cholesky factor of the state
        covariance matrix, S_i.

        On exit: the lower triangular Cholesky factor of the state
        covariance matrix, S_{i+1}.

    k : float, shape (`n`, `m`), modified in place
        On exit: the Kalman gain matrix, K_i, premultiplied by the state
        transition matrix, A_i, A_i K_i.

    h : float, shape (`m`, `m`), modified in place
        On exit: the lower triangular matrix H_i^1/2.

    tol : float
        The tolerance used to test for the singularity of H_i^1/2. If
        0.0 <= `tol` < m^2*machine precision, then m^2*machine precision
        is used instead. The inverse of the condition number of H^(1/2)
        is estimated by a call to ``lapacklin.dtrcon``. If this estimate
        is less than `tol` then H^(1/2) is assumed to be singular.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            The matrix H_i^1/2 is singular.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_kalman_sqrt_var'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, l.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lds, n].
        utils._BYREF_INT, # Input, lds.
        utils._EngineFloat64ArrayType.c_type, # Input, b[lds, l].
        utils._BYREF_LOGICAL, # Input, stq.
        utils._EngineFloat64ArrayType.c_type, # Input, q[ldq, *].
        utils._BYREF_INT, # Input, ldq.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldm, n].
        utils._BYREF_INT, # Input, ldm.
        utils._EngineFloat64ArrayType.c_type, # Input, r[ldm, m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[lds, n].
        utils._EngineFloat64ArrayType.c_type, # Output, k[lds, m].
        utils._EngineFloat64ArrayType.c_type, # Output, h[ldm, m].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineIntArrayType.c_type, # Workspace, iwk.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (q, 'q'), (c, 'c'), (r, 'r'), (s, 's'), (k, 'k'), (h, 'h')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _lds_py = utils._EngineFloat64ArrayType.get_shape(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _lds = utils._EngineIntScalarType(
        _lds_py, {'fun_name': fun_name, 'entity': '_lds'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _l_py = utils._EngineIntScalarType.to_py(_l)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _l_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _stq = utils._EngineLogicalScalarType(
        stq, {'fun_name': fun_name, 'entity': 'stq'},
    ).to_ctypes()
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _stq_py = utils._EngineLogicalScalarType.to_py(_stq)
    if not _stq_py:
        _q_dim1 = _l_py
    else:
        _q_dim1 = 1
    if not _stq_py:
        _q_dim2 = _l_py
    else:
        _q_dim2 = 0
    exp_shape_q = (_q_dim1, _q_dim2)
    _q = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=exp_shape_q,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldm_py = utils._EngineFloat64ArrayType.get_shape(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _ldm = utils._EngineIntScalarType(
        _ldm_py, {'fun_name': fun_name, 'entity': '_ldm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _s = _s_cls.to_ctypes()
    _k_cls = utils._EngineFloat64ArrayType(
        k, {'fun_name': fun_name, 'entity': 'k'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _k = _k_cls.to_ctypes(can_reference_contents=False)
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _h = _h_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _iwk = utils._EngineIntArrayType.empty_ctypes(0)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _l,
        _a,
        _lds,
        _b,
        _stq,
        _q,
        _ldq,
        _c,
        _ldm,
        _r,
        _s,
        _k,
        _h,
        _tol,
        _iwk,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix H_i^1/2 is singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _s_cls.refresh_py(_s)
        _k_cls.refresh_py(_k)
        _h_cls.refresh_py(_h)
        return_data = None
    _err_class.manage()
    return return_data

def multi_kalman_sqrt_invar(transf, n, m, l, a, b, stq, q, c, r, s, k, h, u, tol):
    u"""
    Combined measurement and time update, one iteration of Kalman
    filter, time-invariant, square root covariance filter.

    ``multi_kalman_sqrt_invar`` performs a combined measurement and time
    update of one iteration of the time-invariant Kalman filter using a
    square root covariance filter.

    For full information please refer to the NAG Library document for
    g13eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ebf.html

    Parameters
    ----------
    transf : str, length 1
        Indicates whether to transform the input matrix pair (A,C) to
        lower observer Hessenberg form. The transformation will only be
        required on the first call to ``multi_kalman_sqrt_invar``.

        `transf` = 'T'

          The matrices in arrays `a` and `c` are transformed to lower
          observer Hessenberg form and the matrices in `b` and `s` are
          transformed as described in [Description].

        `transf` = 'H'

          The matrices in arrays `a`, `c` and `b` should be as returned
          from a previous call to ``multi_kalman_sqrt_invar`` with
          `transf` = 'T'.

    n : int
        n, the size of the state vector.

    m : int
        m, the size of the observation vector.

    l : int
        l, the dimension of the state noise.

    a : float, shape (`n`, `n`), modified in place
        On entry: if `transf` = 'T', the state transition matrix, A.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

        On exit: if `transf` = 'T', the transformed matrix, U*AU^(*T),
        otherwise `a` is unchanged.

    b : float, shape (`n`, `l`), modified in place
        On entry: if `transf` = 'T', the noise coefficient matrix B.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

        On exit: if `transf` = 'T', the transformed matrix, U*B,
        otherwise `b` is unchanged.

    stq : bool
        If `stq` = True, the state noise covariance matrix Q_i is
        assumed to be the identity matrix. Otherwise the lower
        triangular Cholesky factor, Q_i^1/2, must be provided in `q`.

    q : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if not((`stq`): `l`; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if not((`stq`): `l`; otherwise: 0.

        If `stq` = False, `q` must contain the lower triangular Cholesky
        factor of the state noise covariance matrix, Q_i^1/2. Otherwise
        `q` is not referenced.

    c : float, shape (`m`, `n`), modified in place
        On entry: if `transf` = 'T', the measurement coefficient matrix,
        C.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

        On exit: if `transf` = 'T', the transformed matrix, CU^(*T),
        otherwise `c` is unchanged.

    r : float, shape (`m`, `m`)
        The lower triangular Cholesky factor of the measurement noise
        covariance matrix R_i^1/2.

    s : float, shape (`n`, `n`), modified in place
        On entry: if `transf` = 'T' the lower triangular Cholesky factor
        of the state covariance matrix, S_i.

        If `transf` = 'H' the lower triangular Cholesky factor of the
        covariance matrix of the transformed state vector S_i^* as
        returned from a previous call to ``multi_kalman_sqrt_invar``
        with `transf` = 'T'.

        On exit: the lower triangular Cholesky factor of the transformed
        state covariance matrix, S_i+1^*.

    k : float, shape (`n`, `m`), modified in place
        On exit: the Kalman gain matrix for the transformed state vector
        premultiplied by the state transformed transition matrix,
        U*AK_i.

    h : float, shape (`m`, `m`), modified in place
        On exit: the lower triangular matrix H_i^1/2.

    u : float, shape (`n`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transf`='T': `n`; otherwise: 0.

        On exit: if `transf` = 'T' the n by n transformation matrix U*,
        otherwise `u` is not referenced.

    tol : float
        The tolerance used to test for the singularity of H_i^1/2. If
        0.0 <= `tol` < m^2*machine precision, then m^2*machine precision
        is used instead. The inverse of the condition number of H^(1/2)
        is estimated by a call to ``lapacklin.dtrcon``. If this estimate
        is less than `tol` then H^(1/2) is assumed to be singular.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `transf` = *<value>*.

            Constraint: `transf` = 'T' or 'H'.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            The matrix H_i^1/2 is singular.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.multi_kalman_sqrt_invar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ebft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, transf(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, l.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ldb, l].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_LOGICAL, # Input, stq.
        utils._EngineFloat64ArrayType.c_type, # Input, q[ldq, *].
        utils._BYREF_INT, # Input, ldq.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldm, n].
        utils._BYREF_INT, # Input, ldm.
        utils._EngineFloat64ArrayType.c_type, # Input, r[ldr, m].
        utils._BYREF_INT, # Input, ldr.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, s[lds, n].
        utils._BYREF_INT, # Input, lds.
        utils._EngineFloat64ArrayType.c_type, # Output, k[ldk, m].
        utils._BYREF_INT, # Input, ldk.
        utils._EngineFloat64ArrayType.c_type, # Output, h[ldh, m].
        utils._BYREF_INT, # Input, ldh.
        utils._EngineFloat64ArrayType.c_type, # Output, u[ldu, *].
        utils._BYREF_INT, # Input, ldu.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineIntArrayType.c_type, # Workspace, iwk.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of transf.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'), (b, 'b'), (q, 'q'), (c, 'c'), (r, 'r'), (s, 's'), (k, 'k'), (h, 'h'), (u, 'u')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _transf = utils._EngineCharScalarType(
        transf, {'fun_name': fun_name, 'entity': 'transf'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _l_py = utils._EngineIntScalarType.to_py(_l)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py, _l_py),
        mutual_sorder=_m_sorder,
    )
    _b = _b_cls.to_ctypes()
    _stq = utils._EngineLogicalScalarType(
        stq, {'fun_name': fun_name, 'entity': 'stq'},
    ).to_ctypes()
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _stq_py = utils._EngineLogicalScalarType.to_py(_stq)
    if not _stq_py:
        _q_dim1 = _l_py
    else:
        _q_dim1 = 1
    if not _stq_py:
        _q_dim2 = _l_py
    else:
        _q_dim2 = 0
    exp_shape_q = (_q_dim1, _q_dim2)
    _q = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=exp_shape_q,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldm_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldm = utils._EngineIntScalarType(
        _ldm_py, {'fun_name': fun_name, 'entity': '_ldm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _lds_py = utils._EngineFloat64ArrayType.get_nd(s, {'fun_name': fun_name, 'entity': 's'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lds = utils._EngineIntScalarType(
        _lds_py, {'fun_name': fun_name, 'entity': '_lds'},
    ).to_ctypes()
    _s_cls = utils._EngineFloat64ArrayType(
        s, {'fun_name': fun_name, 'entity': 's'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _s = _s_cls.to_ctypes()
    _ldk_py = utils._EngineFloat64ArrayType.get_nd(k, {'fun_name': fun_name, 'entity': 'k'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldk = utils._EngineIntScalarType(
        _ldk_py, {'fun_name': fun_name, 'entity': '_ldk'},
    ).to_ctypes()
    _k_cls = utils._EngineFloat64ArrayType(
        k, {'fun_name': fun_name, 'entity': 'k'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _k = _k_cls.to_ctypes(can_reference_contents=False)
    _ldh_py = utils._EngineFloat64ArrayType.get_nd(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldh = utils._EngineIntScalarType(
        _ldh_py, {'fun_name': fun_name, 'entity': '_ldh'},
    ).to_ctypes()
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _h = _h_cls.to_ctypes(can_reference_contents=False)
    _ldu_py = utils._EngineFloat64ArrayType.get_nd(u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldu = utils._EngineIntScalarType(
        _ldu_py, {'fun_name': fun_name, 'entity': '_ldu'},
    ).to_ctypes()
    _transf_py = utils._EngineCharScalarType.to_py(_transf)
    if _transf_py in ('T', 't'):
        _u_dim2 = _n_py
    else:
        _u_dim2 = 0
    exp_shape_u = (_n_py, _u_dim2)
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=exp_shape_u,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _u = _u_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _iwk = utils._EngineIntArrayType.empty_ctypes(0)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _transf,
        _n,
        _m,
        _l,
        _a,
        _lda,
        _b,
        _ldb,
        _stq,
        _q,
        _ldq,
        _c,
        _ldm,
        _r,
        _ldr,
        _s,
        _lds,
        _k,
        _ldk,
        _h,
        _ldh,
        _u,
        _ldu,
        _tol,
        _iwk,
        _wk,
        _errbuf,
        _ierr,
        _transf._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, transf = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: transf = \'T\' or \'H\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix H_i^1/2 is singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _b_cls.refresh_py(_b)
        _c_cls.refresh_py(_c)
        _s_cls.refresh_py(_s)
        _k_cls.refresh_py(_k)
        _h_cls.refresh_py(_h)
        _u_cls.refresh_py(_u)
        return_data = None
    _err_class.manage()
    return return_data

def kalman_unscented_state_revcom(irevcm, mx, my, y, lx, ly, x, st, n, xt, fxt, ropt, lropt, comm):
    u"""
    Combined time and measurement update, one iteration of the Unscented
    Kalman Filter for a nonlinear state space model, with additive noise
    (reverse communication).

    ``kalman_unscented_state_revcom`` applies the Unscented Kalman
    Filter to a nonlinear state space model, with additive noise.

    ``kalman_unscented_state_revcom`` uses reverse communication for
    evaluating the nonlinear functionals of the state space model.

    For full information please refer to the NAG Library document for
    g13ej

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ejf.html

    Parameters
    ----------
    irevcm : int
        On initial entry: must be set to 0 or 3.

        If `irevcm` = 0, it is assumed that t = 0, otherwise it is
        assumed that t != 0 and that ``kalman_unscented_state_revcom``
        has been called at least once before at an earlier time step.

        On intermediate entry: `irevcm` must remain unchanged.

    mx : int
        m_x, the number of state variables.

    my : int
        m_y, the number of observed variables.

    y : float, shape (`my`)
        y_t, the observed data at the current time point.

    lx : None or float, shape (`mx`, `mx`)
        L_x, such that L_x L_x^T = Sigma_x, i.e., the lower triangular
        part of a Cholesky decomposition of the process noise covariance
        structure. Only the lower triangular part of `lx` is referenced.

        If ldlx = 0, there is no process noise (v_t = 0 for all t) and
        `lx` is not referenced.

        If Sigma_x is time dependent, the value supplied should be for
        time t.

    ly : float, shape (`my`, `my`)
        L_y, such that L_y L_y^T = Sigma_y, i.e., the lower triangular
        part of a Cholesky decomposition of the observation noise
        covariance structure. Only the lower triangular part of `ly` is
        referenced.

        If Sigma_y is time dependent, the value supplied should be for
        time t.

    x : float, shape (`mx`), modified in place
        On initial entry: x^_{t-1} the state vector for the previous
        time point.

        On intermediate exit: when

        `irevcm` = 1

          `x` is unchanged.

        `irevcm` = 2

          x^_t_.

        On intermediate entry: `x` must remain unchanged.

        On final exit: x^_t the updated state vector.

    st : float, shape (`mx`, `mx`), modified in place
        On initial entry: S_t, such that S_{t-1}S_{t-1}^T = P_{t-1},
        i.e., the lower triangular part of a Cholesky decomposition of
        the state covariance matrix at the previous time point. Only the
        lower triangular part of `st` is referenced.

        On intermediate exit: when

        `irevcm` = 1

          `st` is unchanged.

        `irevcm` = 2

          S_t_, the lower triangular part of a Cholesky factorization of
          P_t_.

        On intermediate entry: `st` must remain unchanged.

        On final exit: S_t, the lower triangular part of a Cholesky
        factorization of the updated state covariance matrix.

    n : int
        On initial entry: the value used in the sizing of the `fxt` and
        `xt` arrays. The value of `n` supplied must be at least as big
        as the maximum number of sigma points that the algorithm will
        use. ``kalman_unscented_state_revcom`` allows sigma points to be
        calculated in two ways during the measurement update; they can
        be redrawn or augmented. Which is used is controlled by `ropt`.

        If redrawn sigma points are used, then the maximum number of
        sigma points will be 2*m_x+1, otherwise the maximum number of
        sigma points will be 4*m_x+1.

        On intermediate entry: `n` must remain unchanged.

    xt : float, shape (`mx`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `irevcm`=0: max(`my`, `n`); if
        `irevcm` in (1, 2, 3) and licomm>=10: max(`my`,
        `comm`['icomm'][9]); otherwise: 0.

        On initial entry: need not be set.

        On intermediate exit: X_t when `irevcm` = 1, otherwise Y_t.

        For the jth sigma point, the value for the ith parameter is held
        in `xt`[i-1,j-1], for j = 1,2,...,`n`, for i = 1,2,...,`mx`.

        On intermediate entry: `xt` must remain unchanged.

        On final exit: the contents of `xt` are undefined.

    fxt : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `irevcm` in (1, 2): max(`mx`, `my`);
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `irevcm`=1: `n`+`mx`; if `irevcm`=2:
        `n`+`my`; otherwise: 0.

        On initial entry: need not be set.

        On intermediate exit: the contents of `fxt` are undefined.

        On intermediate entry: F(X_t) when `irevcm` = 1, otherwise
        H(Y_t) for the values of X_t and Y_t held in `xt`.

        For the jth sigma point the value for the ith parameter should
        be held in `fxt`[i-1,j-1], for j = 1,2,...,`n`.

        When `irevcm` = 1, i = 1,2,...,`mx` and when `irevcm` = 2, i =
        1,2,...,`my`.

        On final exit: the contents of `fxt` are undefined.

    ropt : None or float, shape (`lropt`)
        Options. The default value will be used for `ropt`[i-1] if
        `lropt` < i. Setting `lropt` = 0 will use the default values for
        all options and `ropt` need not be set.

        `ropt`[0]

          If set to 1 then the second set of sigma points are redrawn,
          as given by equation [equation]. If set to 2 then the second
          set of sigma points are generated via augmentation, as given
          by equation [equation].

          Default is for the sigma points to be redrawn (i.e., `ropt`[0]
          = 1)

        `ropt`[1]

          kappa_x, value of kappa used when constructing the first set
          of sigma points, X_t.

          Defaults to 3-`mx`.

        `ropt`[2]

          alpha_x, value of alpha used when constructing the first set
          of sigma points, X_t.

          Defaults to 1.

        `ropt`[3]

          beta_x, value of beta used when constructing the first set of
          sigma points, X_t.

          Defaults to 2.

        `ropt`[4]

          Value of kappa used when constructing the second set of sigma
          points, Y_t.

          Defaults to 3-2*`mx` when ldlx != 0 and the second set of
          sigma points are augmented and kappa_x otherwise.

        `ropt`[5]

          Value of alpha used when constructing the second set of sigma
          points, Y_t.

          Defaults to alpha_x.

        `ropt`[6]

          Value of beta used when constructing the second set of sigma
          points, Y_t.

          Defaults to beta_x.

    lropt : int
        Length of the options array `ropt`.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    irevcm : int
        On intermediate exit: `irevcm` = 1 or 2. The value of `irevcm`
        specifies what intermediate values are returned by this function
        and what values the calling program must assign to arguments of
        ``kalman_unscented_state_revcom`` before re-entering the
        routine. Details of the output and required input are given in
        the individual argument descriptions.

        On final exit: `irevcm` = 3

    n : int
        On intermediate exit: the number of sigma points actually being
        used.

        On final exit: reset to its value on initial entry.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0, 1, 2 or 3.

        (`errno` 21)
            On entry, `mx` = *<value>*.

            Constraint: `mx` >= 1.

        (`errno` 22)
            `mx` has changed between calls.

            On intermediate entry, `mx` = *<value>*.

            On initial entry, `mx` = *<value>*.

        (`errno` 31)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 32)
            `my` has changed between calls.

            On intermediate entry, `my` = *<value>*.

            On initial entry, `my` = *<value>*.

        (`errno` 121)
            On entry, augmented sigma points requested, `n` = *<value>*
            and `mx` = *<value>*.

            Constraint: `n` >= *<value>*.

        (`errno` 122)
            On entry, redrawn sigma points requested, `n` = *<value>*
            and `mx` = *<value>*.

            Constraint: `n` >= *<value>*.

        (`errno` 123)
            `n` has changed between calls.

            On intermediate entry, `n` = *<value>*.

            On intermediate exit, `n` = *<value>*.

        (`errno` 171)
            On entry, `ropt`[0] = *<value>*.

            Constraint: `ropt`[0] = 1 or 2.

        (`errno` 172)
            On entry, `ropt`[*<value>*] = *<value>*.

            Constraint: kappa > *<value>*.

        (`errno` 173)
            On entry, `ropt`[*<value>*] = *<value>*.

            Constraint: alpha > 0.

        (`errno` 181)
            On entry, `lropt` = *<value>*.

            Constraint: 0 <= `lropt` <= 7.

        (`errno` 191)
            `comm`['icomm'] has been corrupted between calls.

        (`errno` 211)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 301)
            A weight was negative and it was not possible to downdate
            the Cholesky factorization.

        (`errno` 302)
            Unable to calculate the Kalman gain matrix.

        (`errno` 303)
            Unable to calculate the Cholesky factorization of the
            updated state covariance matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 202)
            On entry, licomm = *<value>* and lrcomm = *<value>*.

            Constraint: licomm >= 30 and lrcomm >=
            30+`my`+`mx`*`my`+2*max(`mx`, `my`).

            The minimum required values for licomm and lrcomm are
            returned in `comm`['icomm'][0] and `comm`['icomm'][1]
            respectively.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.kalman_unscented_state_revcom'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ejft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, irevcm.
        utils._BYREF_INT, # Input, mx.
        utils._BYREF_INT, # Input, my.
        utils._EngineFloat64ArrayType.c_type, # Input, y[my].
        utils._EngineFloat64ArrayType.c_type, # Input, lx[ldlx, mx].
        utils._BYREF_INT, # Input, ldlx.
        utils._EngineFloat64ArrayType.c_type, # Input, ly[ldly, my].
        utils._BYREF_INT, # Input, ldly.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[mx].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, st[ldst, mx].
        utils._BYREF_INT, # Input, ldst.
        utils._BYREF_INT, # Input/Output, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xt[ldxt, *].
        utils._BYREF_INT, # Input, ldxt.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, fxt[ldfxt, *].
        utils._BYREF_INT, # Input, ldfxt.
        utils._EngineFloat64ArrayType.c_type, # Input, ropt[lropt].
        utils._BYREF_INT, # Input, lropt.
        utils._EngineIntArrayType.c_type, # Input/Output, icomm[licomm].
        utils._BYREF_INT, # Input, licomm.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((lx, 'lx'), (ly, 'ly'), (st, 'st'), (xt, 'xt'), (fxt, 'fxt')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _irevcm = utils._EngineIntScalarType(
        irevcm, {'fun_name': fun_name, 'entity': 'irevcm'},
    ).to_ctypes()
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _ldlx_py = utils._EngineFloat64ArrayType.get_nd(lx, {'fun_name': fun_name, 'entity': 'lx'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldlx = utils._EngineIntScalarType(
        _ldlx_py, {'fun_name': fun_name, 'entity': '_ldlx'},
    ).to_ctypes()
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _lx = utils._EngineFloat64ArrayType(
        lx, {'fun_name': fun_name, 'entity': 'lx'},
        exp_shape=(_mx_py, _mx_py),
        mutual_sorder=_m_sorder,
        allow_none=True,
    ).to_ctypes()
    _ldly_py = utils._EngineFloat64ArrayType.get_nd(ly, {'fun_name': fun_name, 'entity': 'ly'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldly = utils._EngineIntScalarType(
        _ldly_py, {'fun_name': fun_name, 'entity': '_ldly'},
    ).to_ctypes()
    _ly = utils._EngineFloat64ArrayType(
        ly, {'fun_name': fun_name, 'entity': 'ly'},
        exp_shape=(_my_py, _my_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_mx_py,),
    )
    _x = _x_cls.to_ctypes()
    _ldst_py = utils._EngineFloat64ArrayType.get_nd(st, {'fun_name': fun_name, 'entity': 'st'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldst = utils._EngineIntScalarType(
        _ldst_py, {'fun_name': fun_name, 'entity': '_ldst'},
    ).to_ctypes()
    _st_cls = utils._EngineFloat64ArrayType(
        st, {'fun_name': fun_name, 'entity': 'st'},
        exp_shape=(_mx_py, _mx_py),
        mutual_sorder=_m_sorder,
    )
    _st = _st_cls.to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldxt_py = utils._EngineFloat64ArrayType.get_nd(xt, {'fun_name': fun_name, 'entity': 'xt'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldxt = utils._EngineIntScalarType(
        _ldxt_py, {'fun_name': fun_name, 'entity': '_ldxt'},
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('icomm', 'rcomm'),
    )
    if _comm_needs_init:
        _icomm = utils._EngineIntArrayType.empty_ctypes(30)
        comm['icomm'] = _icomm
    else:
        _icomm = comm['icomm']
    _licomm_py = utils._EngineIntArrayType.get_shape(comm['icomm'], {'fun_name': fun_name, 'entity': 'comm[\'icomm\']'}, exp_rank=1, validate=False)[0]
    _licomm = utils._EngineIntScalarType(
        _licomm_py, {'fun_name': fun_name, 'entity': '_licomm'},
    ).to_ctypes()
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _irevcm_py == 0:
        _xt_dim2 = max(_my_py, _n_py)
    elif _irevcm_py in (1, 2, 3) and _licomm_py >= 10:
        _xt_dim2 = max(_my_py, utils._EngineIntArrayType.get_element(comm['icomm'], {'fun_name': fun_name, 'entity': 'comm[\'icomm\']'}, (9,), validate=False))
    else:
        _xt_dim2 = 0
    exp_shape_xt = (_mx_py, _xt_dim2)
    _xt_cls = utils._EngineFloat64ArrayType(
        xt, {'fun_name': fun_name, 'entity': 'xt'},
        exp_shape=exp_shape_xt,
        mutual_sorder=_m_sorder,
        check_contents=_irevcm_py != 0,
    )
    _xt = _xt_cls.to_ctypes(can_reference_contents=_irevcm_py != 0)
    _ldfxt_py = utils._EngineFloat64ArrayType.get_nd(fxt, {'fun_name': fun_name, 'entity': 'fxt'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldfxt = utils._EngineIntScalarType(
        _ldfxt_py, {'fun_name': fun_name, 'entity': '_ldfxt'},
    ).to_ctypes()
    if _irevcm_py in (1, 2):
        _fxt_dim1 = max(_mx_py, _my_py)
    else:
        _fxt_dim1 = 0
    if _irevcm_py == 1:
        _fxt_dim2 = _n_py+_mx_py
    elif _irevcm_py == 2:
        _fxt_dim2 = _n_py+_my_py
    else:
        _fxt_dim2 = 0
    exp_shape_fxt = (_fxt_dim1, _fxt_dim2)
    _fxt_cls = utils._EngineFloat64ArrayType(
        fxt, {'fun_name': fun_name, 'entity': 'fxt'},
        exp_shape=exp_shape_fxt,
        mutual_sorder=_m_sorder,
        check_contents=_irevcm_py != 0,
    )
    _fxt = _fxt_cls.to_ctypes(can_reference_contents=_irevcm_py != 0)
    _lropt = utils._EngineIntScalarType(
        lropt, {'fun_name': fun_name, 'entity': 'lropt'},
    ).to_ctypes()
    _lropt_py = utils._EngineIntScalarType.to_py(_lropt)
    if _lropt_py != 0:
        if ropt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ropt'},
                'must not be None when lropt != 0',
            )
    _ropt = utils._EngineFloat64ArrayType(
        ropt, {'fun_name': fun_name, 'entity': 'ropt'},
        exp_shape=(_lropt_py,),
        check_contents=_lropt_py != 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_lropt_py != 0)
    if _comm_needs_init:
        _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(30+_my_py+_mx_py*_my_py+2*max(_mx_py, _my_py))
        comm['rcomm'] = _rcomm
    else:
        _rcomm = comm['rcomm']
    _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, exp_rank=1, validate=False)[0]
    _lrcomm = utils._EngineIntScalarType(
        _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _irevcm,
        _mx,
        _my,
        _y,
        _lx,
        _ldlx,
        _ly,
        _ldly,
        _x,
        _st,
        _ldst,
        _n,
        _xt,
        _ldxt,
        _fxt,
        _ldfxt,
        _ropt,
        _lropt,
        _icomm,
        _licomm,
        _rcomm,
        _lrcomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _irevcm_py != 3 or _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irevcm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irevcm = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'mx has changed between calls.',
                    'On intermediate entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'my has changed between calls.',
                    'On intermediate entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On initial entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, augmented sigma points requested,',
                    'n = {:s} and mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 122 and _ifmt == 122:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, redrawn sigma points requested,',
                    'n = {:s} and mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 123 and _ifmt == 123:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'n has changed between calls.',
                    'On intermediate entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On intermediate exit, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 171 and _ifmt == 171:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ropt[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ropt[0] = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 172 and _ifmt == 172:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ropt[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: kappa > {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 173 and _ifmt == 173:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ropt[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: alpha > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 181 and _ifmt == 181:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lropt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= lropt <= 7.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 191 and _ifmt == 191:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'icomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 202 and _ifmt == 202:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, licomm = {:s} and lrcomm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: licomm >= 30 and',
                    'lrcomm >= 30+my+mx*my+2*max(mx, my).',
                    'The minimum required values for licomm and lrcomm are returned in',
                    'comm[\'icomm\'][0] and comm[\'icomm\'][1] respectively.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 211 and _ifmt == 211:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'rcomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 301 and _ifmt == 301:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A weight was negative and it was not possible',
                    'to downdate the Cholesky factorization.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 302 and _ifmt == 302:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the Kalman gain matrix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 303 and _ifmt == 303:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the Cholesky factorization of the',
                    'updated state covariance matrix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _st_cls.refresh_py(_st)
        _n_py = utils._EngineIntScalarType.to_py(_n)
        _irevcm_py = utils._EngineIntScalarType.to_py(_irevcm)
        if _irevcm_py != 3:
            _xt_cls.refresh_py(_xt)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['irevcm', 'n'],
        )(
            _irevcm_py,
            _n_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def kalman_unscented_state(mx, my, y, lx, ly, f, h, x, st, data=None):
    u"""
    Combined time and measurement update, one iteration of the Unscented
    Kalman Filter for a nonlinear state space model, with additive
    noise.

    ``kalman_unscented_state`` applies the Unscented Kalman Filter (UKF)
    to a nonlinear state space model, with additive noise.

    ``kalman_unscented_state`` uses direct communication for evaluating
    the nonlinear functionals of the state space model.

    For full information please refer to the NAG Library document for
    g13ek

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ekf.html

    Parameters
    ----------
    mx : int
        m_x, the number of state variables.

    my : int
        m_y, the number of observed variables.

    y : float, shape (`my`)
        y_t, the observed data at the current time point.

    lx : float, shape (`mx`, `mx`)
        L_x, such that L_x L_x^T = Sigma_x, i.e., the lower triangular
        part of a Cholesky decomposition of the process noise covariance
        structure. Only the lower triangular part of `lx` is referenced.

        If Sigma_x is time dependent, the value supplied should be for
        time t.

    ly : float, shape (`my`, `my`)
        L_y, such that L_y L_y^T = Sigma_y, i.e., the lower triangular
        part of a Cholesky decomposition of the observation noise
        covariance structure. Only the lower triangular part of `ly` is
        referenced.

        If Sigma_y is time dependent, the value supplied should be for
        time t.

    f : callable f(mx, n, xt, fxt, data=None)
        The state function, F.

        Parameters
        ~~~~~~~~~~
        mx : int
            m_x, the number of state variables.

        n : int
            n, the number of sigma points.

        xt : float, ndarray, shape (`mx`, `n`)
            X_t, the sigma points generated in (a). For the jth sigma
            point, the value for the ith parameter is held in
            `xt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,m_x.

        fxt : float, ndarray, shape (`mx`, `n`), to be modified in place
            On exit: F(X_t).

            For the jth sigma point the value for the ith parameter
            should be held in `fxt`[i-1,j-1], for j = 1,2,...,n, for i =
            1,2,...,m_x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    h : callable h(mx, my, n, yt, hyt, data=None)
        The measurement function, H as described in (e).

        Parameters
        ~~~~~~~~~~
        mx : int
            m_x, the number of state variables.

        my : int
            m_y, the number of observed variables.

        n : int
            n, the number of sigma points.

        yt : float, ndarray, shape (`mx`, `n`)
            Y_t, the sigma points generated in (d). For the jth sigma
            point, the value for the ith parameter is held in
            `yt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,m_x, where
            m_x is the number of state variables and n is the number of
            sigma points.

        hyt : float, ndarray, shape (`my`, `n`), to be modified in place
            On exit: H(Y_t).

            For the jth sigma point the value for the ith parameter
            should be held in `hyt`[i-1,j-1], for j = 1,2,...,n, for i =
            1,2,...,m_y.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    x : float, shape (`mx`), modified in place
        On entry: x^_{t-1} the state vector for the previous time point.

        On exit: x^_t the updated state vector.

    st : float, shape (`mx`, `mx`), modified in place
        On entry: S_t, such that S_{t-1}S_{t-1}^T = P_{t-1}, i.e., the
        lower triangular part of a Cholesky decomposition of the state
        covariance matrix at the previous time point. Only the lower
        triangular part of `st` is referenced.

        On exit: S_t, the lower triangular part of a Cholesky
        factorization of the updated state covariance matrix.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `mx` = *<value>*.

            Constraint: `mx` >= 1.

        (`errno` 21)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 301)
            A weight was negative and it was not possible to downdate
            the Cholesky factorization.

        (`errno` 302)
            Unable to calculate the Kalman gain matrix.

        (`errno` 303)
            Unable to calculate the Cholesky factorization of the
            updated state covariance matrix.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 61)
            User requested termination in `f`.

        (`errno` 71)
            User requested termination in `h`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.kalman_unscented_state_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.kalman_unscented_state'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_f_name = fun_name + ':f'

    fh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mx.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, xt[mx, n].
        utils._EngineFloat64ArrayType.c_type, # Output, fxt[mx, n].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Input/Output, info.
    )

    def cb_fh(_f, en_data, mx, n, xt, fxt, _iuser, _ruser, info):
        """Helper for f."""
        try:
            _mx_fh = utils._EngineIntScalarType.to_py(mx.contents)
            _n_fh = utils._EngineIntScalarType.to_py(n.contents)
            _xt_fh = utils._EngineFloat64ArrayType.from_pointer(
                xt,
                (_mx_fh, _n_fh),
                sorder=_m_sorder,
            )
            _fxt_fh = utils._EngineFloat64ArrayType.from_pointer(
                fxt,
                (_mx_fh, _n_fh),
                can_reference_contents=False,
                sorder=_m_sorder,
            )
            if data is not None:
                f(
                    _mx_fh,
                    _n_fh,
                    _xt_fh,
                    _fxt_fh,
                    data,
                )
            else:
                f(
                    _mx_fh,
                    _n_fh,
                    _xt_fh,
                    _fxt_fh,
                )
            utils._unlink_pointer(xt)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _fxt_fh, {'fun_name': cb_f_name, 'returned': 'fxt'}, fxt,
            )
            utils._unlink_pointer(fxt)
        except:
            excs.append(_sys.exc_info())
            info.contents.value = -1
            if excs[-1][0] is not utils.UserCallbackTerminate:
                en_data.contents.hlperr = utils._IERR_HELPER

    cb_h_name = fun_name + ':h'

    hh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mx.
        utils._BYREF_INT, # Input, my.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, yt[mx, n].
        utils._EngineFloat64ArrayType.c_type, # Output, hyt[my, n].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Input/Output, info.
    )

    def cb_hh(_h, en_data, mx, my, n, yt, hyt, _iuser, _ruser, info):
        """Helper for h."""
        try:
            _mx_hh = utils._EngineIntScalarType.to_py(mx.contents)
            _my_hh = utils._EngineIntScalarType.to_py(my.contents)
            _n_hh = utils._EngineIntScalarType.to_py(n.contents)
            _yt_hh = utils._EngineFloat64ArrayType.from_pointer(
                yt,
                (_mx_hh, _n_hh),
                sorder=_m_sorder,
            )
            _hyt_hh = utils._EngineFloat64ArrayType.from_pointer(
                hyt,
                (_my_hh, _n_hh),
                can_reference_contents=False,
                sorder=_m_sorder,
            )
            if data is not None:
                h(
                    _mx_hh,
                    _my_hh,
                    _n_hh,
                    _yt_hh,
                    _hyt_hh,
                    data,
                )
            else:
                h(
                    _mx_hh,
                    _my_hh,
                    _n_hh,
                    _yt_hh,
                    _hyt_hh,
                )
            utils._unlink_pointer(yt)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _hyt_hh, {'fun_name': cb_h_name, 'returned': 'hyt'}, hyt,
            )
            utils._unlink_pointer(hyt)
        except:
            excs.append(_sys.exc_info())
            info.contents.value = -1
            if excs[-1][0] is not utils.UserCallbackTerminate:
                en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g13ekft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mx.
        utils._BYREF_INT, # Input, my.
        utils._EngineFloat64ArrayType.c_type, # Input, y[my].
        utils._EngineFloat64ArrayType.c_type, # Input, lx[mx, mx].
        utils._EngineFloat64ArrayType.c_type, # Input, ly[my, my].
        utils._UNUSED_FUNCTYPE,
        fh,
        utils._UNUSED_FUNCTYPE,
        hh,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[mx].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, st[mx, mx].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((lx, 'lx'), (ly, 'ly'), (st, 'st')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _lx = utils._EngineFloat64ArrayType(
        lx, {'fun_name': fun_name, 'entity': 'lx'},
        exp_shape=(_mx_py, _mx_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ly = utils._EngineFloat64ArrayType(
        ly, {'fun_name': fun_name, 'entity': 'ly'},
        exp_shape=(_my_py, _my_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = utils._UNUSED_CALLBACK
    _fh = utils._cast_callback(
        fh, cb_fh,
        {'fun_name': fun_name, 'entity': 'f'},
    )
    utils._check_callable(h, {'fun_name': fun_name, 'entity': 'h'})
    _h = utils._UNUSED_CALLBACK
    _hh = utils._cast_callback(
        hh, cb_hh,
        {'fun_name': fun_name, 'entity': 'h'},
    )
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_mx_py,),
    )
    _x = _x_cls.to_ctypes()
    _st_cls = utils._EngineFloat64ArrayType(
        st, {'fun_name': fun_name, 'entity': 'st'},
        exp_shape=(_mx_py, _mx_py),
        mutual_sorder=_m_sorder,
    )
    _st = _st_cls.to_ctypes()
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mx,
        _my,
        _y,
        _lx,
        _ly,
        _f,
        _fh,
        _h,
        _hh,
        _x,
        _st,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._UserCallbackExit(
                fun_name=fun_name,
                msg_lines=[
                    'User requested termination in f.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
                cb_excs=excs,
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._UserCallbackExit(
                fun_name=fun_name,
                msg_lines=[
                    'User requested termination in h.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
                cb_excs=excs,
            )
        elif _ierr_py == 301 and _ifmt == 301:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A weight was negative and it was not possible',
                    'to downdate the Cholesky factorization.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 302 and _ifmt == 302:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the Kalman gain matrix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 303 and _ifmt == 303:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Unable to calculate the Cholesky factorization of the',
                    'updated state covariance matrix.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _st_cls.refresh_py(_st)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_garch_asym1_estim(dist, yt, x, num, ip, iq, nreg, mn, isym, npar, theta, se, sc, covr, pht, et, ht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for either a symmetric
    GARCH process or a GARCH process with asymmetry of the form
    (epsilon_{t-1}+gamma)^2.

    ``uni_garch_asym1_estim`` estimates the parameters of either a
    standard univariate regression GARCH process, or a univariate
    regression-type I AGARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13faf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, shape (`num`)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, shape (`num`, `nreg`)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    nreg : int
        k, the number of regression coefficients.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    isym : int
        If `isym` = 1, the asymmetry term gamma will be included in the
        model.

    npar : int
        The number of parameters to be included in the model. `npar` =
        1+`iq`+`ip`+`isym`+`mn`+`nreg` when `dist` = 'N', and `npar` =
        2+`iq`+`ip`+`isym`+`mn`+`nreg` when `dist` = 'T'.

    theta : float, shape (`npar`), modified in place
        On entry: the initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        If `isym` = 1, the next element must contain the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element must contain df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts`[1] = False, the remaining `nreg` elements are taken
        as initial estimates of the linear regression coefficients b_i,
        for i = 1,2,...,k.

        On exit: the estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the coefficients beta_j, for j =
        1,2,...,p.

        If `isym` = 1, the next element contains the estimate for the
        asymmetry parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final `nreg` elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, shape (`npar`), modified in place
        On exit: the standard errors for theta^.

        The first element contains the standard error for alpha_o.

        The next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        If `isym` = 1, the next element contains the standard error for
        gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final `nreg` elements are the standard errors for b_j, for j
        = 1,2,...,k.

    sc : float, shape (`npar`), modified in place
        On exit: the scores for theta^.

        The first element contains the score for alpha_o.

        The next `iq` elements contain the score for alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the scores for beta_j, for j =
        1,2,...,p.

        If `isym` = 1, the next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final `nreg` elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, shape (`npar`, `npar`), modified in place
        On exit: the covariance matrix of the parameter estimates
        theta^, that is the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    et : float, shape (`num`), modified in place
        On exit: the estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, shape (`num`), modified in place
        On exit: the estimated conditional variances, h_t, for t =
        1,2,...,T.

    copts : bool, shape (2)
        The options to be used by ``uni_garch_asym1_estim``.

        `copts`[0] = True

          Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

          The function provides initial parameter estimates of the
          regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= `nreg`+`mn`.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` < 20.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: if `dist` = 'N' then `npar` =
            1+`iq`+`ip`+`isym`+`mn`+`nreg`, else `npar` =
            2+`iq`+`ip`+`isym`+`mn`+`nreg`.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, `isym` = *<value>*.

            Constraint: `isym` = 0 or 1.

        (`errno` 1)
            On entry, `nreg` = *<value>*.

            Constraint: `nreg` >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_garch_asym1_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_asym1_estim'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13faft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, yt[num].
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, nreg].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_INT, # Input, nreg.
        utils._BYREF_INT, # Input, mn.
        utils._BYREF_INT, # Input, isym.
        utils._BYREF_INT, # Input, npar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, se[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, sc[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, covr[ldcovr, npar].
        utils._BYREF_INT, # Input, ldcovr.
        utils._BYREF_FLOAT64, # Input/Output, pht.
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._BYREF_FLOAT64, # Output, lgf.
        utils._EngineLogicalArrayType.c_type, # Input, copts[2].
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (covr, 'covr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _yt = utils._EngineFloat64ArrayType(
        yt, {'fun_name': fun_name, 'entity': 'yt'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _nreg = utils._EngineIntScalarType(
        nreg, {'fun_name': fun_name, 'entity': 'nreg'},
    ).to_ctypes()
    _nreg_py = utils._EngineIntScalarType.to_py(_nreg)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_num_py, _nreg_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _mn = utils._EngineIntScalarType(
        mn, {'fun_name': fun_name, 'entity': 'mn'},
    ).to_ctypes()
    _isym = utils._EngineIntScalarType(
        isym, {'fun_name': fun_name, 'entity': 'isym'},
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_npar_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _sc_cls = utils._EngineFloat64ArrayType(
        sc, {'fun_name': fun_name, 'entity': 'sc'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _sc = _sc_cls.to_ctypes(can_reference_contents=False)
    _ldcovr_py = utils._EngineFloat64ArrayType.get_nd(covr, {'fun_name': fun_name, 'entity': 'covr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcovr = utils._EngineIntScalarType(
        _ldcovr_py, {'fun_name': fun_name, 'entity': '_ldcovr'},
    ).to_ctypes()
    _covr_cls = utils._EngineFloat64ArrayType(
        covr, {'fun_name': fun_name, 'entity': 'covr'},
        exp_shape=(_npar_py, _npar_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _covr = _covr_cls.to_ctypes(can_reference_contents=False)
    _pht = utils._EngineFloat64ScalarType(
        pht, {'fun_name': fun_name, 'entity': 'pht'},
    ).to_ctypes()
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _lgf = utils._EngineFloat64ScalarType.empty_ctypes()
    _copts = utils._EngineLogicalArrayType(
        copts, {'fun_name': fun_name, 'entity': 'copts'},
        exp_shape=(2,),
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _dist,
        _yt,
        _x,
        _ldx,
        _num,
        _ip,
        _iq,
        _nreg,
        _mn,
        _isym,
        _npar,
        _theta,
        _se,
        _sc,
        _covr,
        _ldcovr,
        _pht,
        _et,
        _ht,
        _lgf,
        _copts,
        _maxit,
        _tol,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 88888:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= nreg+mn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar < 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if dist = \'N\' then npar = 1+iq+ip+isym+mn+nreg,',
                    'else npar = 2+iq+ip+isym+mn+nreg.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dist = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mn = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isym = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isym = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nreg = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nreg >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the matrix X is not full rank.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The information matrix is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The maximum number of iterations has been reached.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The log-likelihood cannot be optimized any further.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No feasible model parameters could be found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _theta_cls.refresh_py(_theta)
        _se_cls.refresh_py(_se)
        _sc_cls.refresh_py(_sc)
        _covr_cls.refresh_py(_covr)
        _pht_py = utils._EngineFloat64ScalarType.to_py(_pht)
        _et_cls.refresh_py(_et)
        _ht_cls.refresh_py(_ht)
        _lgf_py = utils._EngineFloat64ScalarType.to_py(_lgf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['pht', 'lgf'],
        )(
            _pht_py,
            _lgf_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_garch_asym1_forecast(num, nt, ip, iq, theta, gamma, fht, ht, et):
    u"""
    Univariate time series, forecast function for either a symmetric
    GARCH process or a GARCH process with asymmetry of the form
    (epsilon_{t-1}+gamma)^2.

    ``uni_garch_asym1_forecast`` forecasts the conditional variances
    h_t, for t = T+1,...,T+xi, from a type I AGARCH(p,q) sequence, where
    xi is the forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13fb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fbf.html

    Parameters
    ----------
    num : int
        The number of terms in the arrays `ht` and `et` from the
        modelled sequence.

    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    fht : float, shape (`nt`), modified in place
        On exit: the forecast values of the conditional variance, h_t,
        for t = T+1,...,T+xi.

    ht : float, shape (`num`)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, shape (`num`)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_garch_asym1_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_asym1_forecast'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13fbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, nt.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[iq+ip+1].
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._EngineFloat64ArrayType.c_type, # Output, fht[nt].
        utils._EngineFloat64ArrayType.c_type, # Input, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Input, et[num].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_iq_py+_ip_py+1,),
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    _fht_cls = utils._EngineFloat64ArrayType(
        fht, {'fun_name': fun_name, 'entity': 'fht'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _fht = _fht_cls.to_ctypes(can_reference_contents=False)
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _et = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _num,
        _nt,
        _ip,
        _iq,
        _theta,
        _gamma,
        _fht,
        _ht,
        _et,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, max(ip-1, iq-1) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip-1, iq-1) <= 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _fht_cls.refresh_py(_fht)
        return_data = None
    _err_class.manage()
    return return_data

def uni_garch_asym2_estim(dist, yt, x, num, ip, iq, nreg, mn, npar, theta, se, sc, covr, pht, et, ht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for a GARCH process
    with asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``uni_garch_asym2_estim`` estimates the parameters of a univariate
    regression-type II AGARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fcf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, shape (`num`)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, shape (`num`, `nreg`)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    nreg : int
        k, the number of regression coefficients.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    npar : int
        The number of parameters to be included in the model. `npar` =
        2+`iq`+`ip`+`mn`+`nreg` when `dist` = 'N' and `npar` =
        3+`iq`+`ip`+`mn`+`nreg` when `dist` = 'T'.

    theta : float, shape (`npar`), modified in place
        On entry: the initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        The next element must contain the asymmetry parameter gamma.

        If `dist` = 'T', the next element must contain df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the mean term b_o.

        If `copts`[1] = False, the remaining `nreg` elements are taken
        as initial estimates of the linear regression coefficients b_i,
        for i = 1,2,...,k.

        On exit: the estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the coefficients beta_j, for j =
        1,2,...,p.

        The next element contains the estimate for the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final `nreg` elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, shape (`npar`), modified in place
        On exit: the standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        The next element contains the standard error for gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final `nreg` elements are the standard errors for b_j, for j
        = 1,2,...,k.

    sc : float, shape (`npar`), modified in place
        On exit: the scores for theta^.

        The first element contains the score for alpha_o and the next
        `iq` elements contain the score for alpha_i, for i = 1,2,...,q.

        The next `ip` elements are the scores for beta_j, for j =
        1,2,...,p.

        The next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final `nreg` elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, shape (`npar`, `npar`), modified in place
        On exit: the covariance matrix of the parameter estimates
        theta^, that is the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    et : float, shape (`num`), modified in place
        On exit: the estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, shape (`num`), modified in place
        On exit: the estimated conditional variances, h_t, for t =
        1,2,...,T.

    copts : bool, shape (2)
        The options to be used by ``uni_garch_asym2_estim``.

        `copts`[0] = True

          Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

          The function provides initial parameter estimates of the
          regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= `nreg`+`mn`.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` < 20.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: if `dist` = 'N' then `npar` =
            2+`iq`+`ip`+`mn`+`nreg`, else `npar` =
            3+`iq`+`ip`+`mn`+`nreg`.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, `nreg` = *<value>*.

            Constraint: `nreg` >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_asym2_estim'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13fcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, yt[num].
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, nreg].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_INT, # Input, nreg.
        utils._BYREF_INT, # Input, mn.
        utils._BYREF_INT, # Input, npar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, se[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, sc[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, covr[ldcovr, npar].
        utils._BYREF_INT, # Input, ldcovr.
        utils._BYREF_FLOAT64, # Input/Output, pht.
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._BYREF_FLOAT64, # Output, lgf.
        utils._EngineLogicalArrayType.c_type, # Input, copts[2].
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (covr, 'covr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _yt = utils._EngineFloat64ArrayType(
        yt, {'fun_name': fun_name, 'entity': 'yt'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _nreg = utils._EngineIntScalarType(
        nreg, {'fun_name': fun_name, 'entity': 'nreg'},
    ).to_ctypes()
    _nreg_py = utils._EngineIntScalarType.to_py(_nreg)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_num_py, _nreg_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _mn = utils._EngineIntScalarType(
        mn, {'fun_name': fun_name, 'entity': 'mn'},
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_npar_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _sc_cls = utils._EngineFloat64ArrayType(
        sc, {'fun_name': fun_name, 'entity': 'sc'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _sc = _sc_cls.to_ctypes(can_reference_contents=False)
    _ldcovr_py = utils._EngineFloat64ArrayType.get_nd(covr, {'fun_name': fun_name, 'entity': 'covr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcovr = utils._EngineIntScalarType(
        _ldcovr_py, {'fun_name': fun_name, 'entity': '_ldcovr'},
    ).to_ctypes()
    _covr_cls = utils._EngineFloat64ArrayType(
        covr, {'fun_name': fun_name, 'entity': 'covr'},
        exp_shape=(_npar_py, _npar_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _covr = _covr_cls.to_ctypes(can_reference_contents=False)
    _pht = utils._EngineFloat64ScalarType(
        pht, {'fun_name': fun_name, 'entity': 'pht'},
    ).to_ctypes()
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _lgf = utils._EngineFloat64ScalarType.empty_ctypes()
    _copts = utils._EngineLogicalArrayType(
        copts, {'fun_name': fun_name, 'entity': 'copts'},
        exp_shape=(2,),
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _dist,
        _yt,
        _x,
        _ldx,
        _num,
        _ip,
        _iq,
        _nreg,
        _mn,
        _npar,
        _theta,
        _se,
        _sc,
        _covr,
        _ldcovr,
        _pht,
        _et,
        _ht,
        _lgf,
        _copts,
        _maxit,
        _tol,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 88888:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= nreg+mn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar < 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if dist = \'N\' then npar = 2+iq+ip+mn+nreg,',
                    'else npar = 3+iq+ip+mn+nreg.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dist = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mn = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nreg = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nreg >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the matrix X is not full rank.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The information matrix is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The maximum number of iterations has been reached.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The log-likelihood cannot be optimized any further.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No feasible model parameters could be found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _theta_cls.refresh_py(_theta)
        _se_cls.refresh_py(_se)
        _sc_cls.refresh_py(_sc)
        _covr_cls.refresh_py(_covr)
        _pht_py = utils._EngineFloat64ScalarType.to_py(_pht)
        _et_cls.refresh_py(_et)
        _ht_cls.refresh_py(_ht)
        _lgf_py = utils._EngineFloat64ScalarType.to_py(_lgf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['pht', 'lgf'],
        )(
            _pht_py,
            _lgf_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_garch_asym2_forecast(num, nt, ip, iq, theta, gamma, fht, ht, et):
    u"""
    Univariate time series, forecast function for a GARCH process with
    asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``uni_garch_asym2_forecast`` forecasts the conditional variances,
    h_t,t = T+1,...,T+xi from a type II AGARCH(p,q) sequence, where xi
    is the forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13fd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fdf.html

    Parameters
    ----------
    num : int
        The number of terms in the arrays `ht` and `et` from the
        modelled sequence.

    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    fht : float, shape (`nt`), modified in place
        On exit: the forecast values of the conditional variance, h_t,
        for t = T+1,...,T+xi.

    ht : float, shape (`num`)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, shape (`num`)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_asym2_forecast'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13fdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, nt.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[iq+ip+1].
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._EngineFloat64ArrayType.c_type, # Output, fht[nt].
        utils._EngineFloat64ArrayType.c_type, # Input, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Input, et[num].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_iq_py+_ip_py+1,),
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    _fht_cls = utils._EngineFloat64ArrayType(
        fht, {'fun_name': fun_name, 'entity': 'fht'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _fht = _fht_cls.to_ctypes(can_reference_contents=False)
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _et = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _num,
        _nt,
        _ip,
        _iq,
        _theta,
        _gamma,
        _fht,
        _ht,
        _et,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, max(ip-1, iq-1) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip-1, iq-1) <= 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _fht_cls.refresh_py(_fht)
        return_data = None
    _err_class.manage()
    return return_data

def uni_garch_gjr_estim(dist, yt, x, num, ip, iq, nreg, mn, npar, theta, se, sc, covr, pht, et, ht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for an asymmetric
    Glosten, Jagannathan and Runkle (GJR) GARCH process.

    ``uni_garch_gjr_estim`` estimates the parameters of a univariate
    regression-GJR GARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fef.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, shape (`num`)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, shape (`num`, `nreg`)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    nreg : int
        k, the number of regression coefficients.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    npar : int
        The number of parameters to be included in the model. `npar` =
        2+`iq`+`ip`+`mn`+`nreg` when `dist` = 'N' and `npar` =
        3+`iq`+`ip`+`mn`+`nreg` when `dist` = 'T'.

    theta : float, shape (`npar`), modified in place
        On entry: the initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        The next element must contain the asymmetry parameter gamma.

        If `dist` = 'T', the next element contains df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts`[1] = False, the remaining `nreg` elements are taken
        as initial estimates of the linear regression coefficients b_i,
        for i = 1,2,...,k.

        On exit: the estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the moving average coefficients
        beta_j, for j = 1,2,...,p.

        The next element contains the estimate for the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final `nreg` elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, shape (`npar`), modified in place
        On exit: the standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        The next element contains the standard error for gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final `nreg` elements are the standard errors for b_j, for j
        = 1,2,...,k.

    sc : float, shape (`npar`), modified in place
        On exit: the scores for theta^.

        The first element contains the score for alpha_o, the next `iq`
        elements contain the scores for alpha_i, for i = 1,2,...,q.

        The next `ip` elements are the score for beta_j, for j =
        1,2,...,p.

        The next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final `nreg` elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, shape (`npar`, `npar`), modified in place
        On exit: the covariance matrix of the parameter estimates
        theta^, that is the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    et : float, shape (`num`), modified in place
        On exit: the estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, shape (`num`), modified in place
        On exit: the estimated conditional variances, h_t, for t =
        1,2,...,T.

    copts : bool, shape (2)
        The options to be used by ``uni_garch_gjr_estim``.

        `copts`[0] = True

          Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

          The function provides initial parameter estimates of the
          regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= `nreg`+`mn`.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` < 20.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: if `dist` = 'N' then `npar` =
            2+`iq`+`ip`+`mn`+`nreg`, else `npar` =
            3+`iq`+`ip`+`mn`+`nreg`.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, `nreg` = *<value>*.

            Constraint: `nreg` >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_gjr_estim'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13feft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, yt[num].
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, nreg].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_INT, # Input, nreg.
        utils._BYREF_INT, # Input, mn.
        utils._BYREF_INT, # Input, npar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, se[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, sc[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, covr[ldcovr, npar].
        utils._BYREF_INT, # Input, ldcovr.
        utils._BYREF_FLOAT64, # Input/Output, pht.
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._BYREF_FLOAT64, # Output, lgf.
        utils._EngineLogicalArrayType.c_type, # Input, copts[2].
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (covr, 'covr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _yt = utils._EngineFloat64ArrayType(
        yt, {'fun_name': fun_name, 'entity': 'yt'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _nreg = utils._EngineIntScalarType(
        nreg, {'fun_name': fun_name, 'entity': 'nreg'},
    ).to_ctypes()
    _nreg_py = utils._EngineIntScalarType.to_py(_nreg)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_num_py, _nreg_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _mn = utils._EngineIntScalarType(
        mn, {'fun_name': fun_name, 'entity': 'mn'},
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_npar_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _sc_cls = utils._EngineFloat64ArrayType(
        sc, {'fun_name': fun_name, 'entity': 'sc'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _sc = _sc_cls.to_ctypes(can_reference_contents=False)
    _ldcovr_py = utils._EngineFloat64ArrayType.get_nd(covr, {'fun_name': fun_name, 'entity': 'covr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcovr = utils._EngineIntScalarType(
        _ldcovr_py, {'fun_name': fun_name, 'entity': '_ldcovr'},
    ).to_ctypes()
    _covr_cls = utils._EngineFloat64ArrayType(
        covr, {'fun_name': fun_name, 'entity': 'covr'},
        exp_shape=(_npar_py, _npar_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _covr = _covr_cls.to_ctypes(can_reference_contents=False)
    _pht = utils._EngineFloat64ScalarType(
        pht, {'fun_name': fun_name, 'entity': 'pht'},
    ).to_ctypes()
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _lgf = utils._EngineFloat64ScalarType.empty_ctypes()
    _copts = utils._EngineLogicalArrayType(
        copts, {'fun_name': fun_name, 'entity': 'copts'},
        exp_shape=(2,),
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _dist,
        _yt,
        _x,
        _ldx,
        _num,
        _ip,
        _iq,
        _nreg,
        _mn,
        _npar,
        _theta,
        _se,
        _sc,
        _covr,
        _ldcovr,
        _pht,
        _et,
        _ht,
        _lgf,
        _copts,
        _maxit,
        _tol,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 88888:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= nreg+mn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar < 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if dist = \'N\' then npar = 2+iq+ip+mn+nreg,',
                    'else npar = 3+iq+ip+mn+nreg.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dist = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mn = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nreg = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nreg >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the matrix X is not full rank.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The information matrix is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The maximum number of iterations has been reached.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The log-likelihood cannot be optimized any further.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No feasible model parameters could be found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _theta_cls.refresh_py(_theta)
        _se_cls.refresh_py(_se)
        _sc_cls.refresh_py(_sc)
        _covr_cls.refresh_py(_covr)
        _pht_py = utils._EngineFloat64ScalarType.to_py(_pht)
        _et_cls.refresh_py(_et)
        _ht_cls.refresh_py(_ht)
        _lgf_py = utils._EngineFloat64ScalarType.to_py(_lgf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['pht', 'lgf'],
        )(
            _pht_py,
            _lgf_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_garch_gjr_forecast(num, nt, ip, iq, theta, gamma, fht, ht, et):
    u"""
    Univariate time series, forecast function for an asymmetric Glosten,
    Jagannathan and Runkle (GJR) GARCH process.

    ``uni_garch_gjr_forecast`` forecasts the conditional variances, h_t,
    for t = T+1,...,T+xi from a GJR GARCH(p,q) sequence, where xi is the
    forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13ff

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fff.html

    Parameters
    ----------
    num : int
        The number of terms in the arrays `ht` and `et` from the
        modelled sequence.

    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    fht : float, shape (`nt`), modified in place
        On exit: the forecast values of the conditional variance, h_t,
        for t = T+1,...,T+xi.

    ht : float, shape (`num`)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, shape (`num`)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_gjr_forecast'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ffft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, nt.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[iq+ip+1].
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._EngineFloat64ArrayType.c_type, # Output, fht[nt].
        utils._EngineFloat64ArrayType.c_type, # Input, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Input, et[num].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_iq_py+_ip_py+1,),
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    _fht_cls = utils._EngineFloat64ArrayType(
        fht, {'fun_name': fun_name, 'entity': 'fht'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _fht = _fht_cls.to_ctypes(can_reference_contents=False)
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _et = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _num,
        _nt,
        _ip,
        _iq,
        _theta,
        _gamma,
        _fht,
        _ht,
        _et,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, max(ip-1, iq-1) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip-1, iq-1) <= 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _fht_cls.refresh_py(_fht)
        return_data = None
    _err_class.manage()
    return return_data

def uni_garch_exp_estim(dist, yt, x, num, ip, iq, nreg, mn, npar, theta, se, sc, covr, pht, et, ht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for an exponential
    GARCH (EGARCH) process.

    ``uni_garch_exp_estim`` estimates the parameters of a univariate
    regression-exponential GARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fgf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, shape (`num`)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, shape (`num`, `nreg`)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    nreg : int
        k, the number of regression coefficients.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    npar : int
        The number of parameters to be included in the model. `npar` =
        1+2*`iq`+`ip`+`mn`+`nreg` when `dist` = 'N' and `npar` =
        2+2*`iq`+`ip`+`mn`+`nreg` when `dist` = 'T'.

    theta : float, shape (`npar`), modified in place
        On entry: the initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the autoregressive coefficients
        alpha_i, for i = 1,2,...,q.

        The next `iq` elements contain the coefficients phi_i, for i =
        1,2,...,q.

        The next `ip` elements must contain the moving average
        coefficients beta_i, for i = 1,2,...,p.

        If `dist` = 'T', the next element must contain an estimate for
        df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts` = False, the remaining `nreg` elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

        On exit: the estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o and the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `iq` elements contain the coefficients phi_i, for i =
        1,2,...,q.

        The next `ip` elements are the moving average coefficients
        beta_i, for i = 1,2,...,p.

        If `dist` = 'T', the next element contains an estimate for df
        then the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final `nreg` elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, shape (`npar`), modified in place
        On exit: the standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `iq` elements contain the standard errors for phi_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final `nreg` elements are the standard errors for b_j, for j
        = 1,2,...,k.

    sc : float, shape (`npar`), modified in place
        On exit: the scores for theta^.

        The first element contains the scores for alpha_o, the next `iq`
        elements contain the scores for alpha_i, for i = 1,2,...,q, the
        next `iq` elements contain the scores for phi_i, for i =
        1,2,...,q, the next `ip` elements are the scores for beta_j, for
        j = 1,2,...,p.

        If `dist` = 'T', the next element contains the scores for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final `nreg` elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, shape (`npar`, `npar`), modified in place
        On exit: the covariance matrix of the parameter estimates
        theta^, that is the inverse of the Fisher Information Matrix.

    pht : float
        If `copts` = False then `pht` is the value to be used for the
        pre-observed conditional variance, otherwise `pht` is not
        referenced.

    et : float, shape (`num`), modified in place
        On exit: the estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, shape (`num`), modified in place
        On exit: the estimated conditional variances, h_t, for t =
        1,2,...,T.

    copts : bool
        If `copts` = True, the function provides initial parameter
        estimates of the regression terms, otherwise these are provided
        by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    pht : float
        If `copts` = True then `pht` is the estimated value of the
        pre-observed conditional variance.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= `nreg`+`mn`.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: `npar` < 20.

        (`errno` 1)
            On entry, `npar` = *<value>*.

            Constraint: if `dist` = 'N' then `npar` =
            1+2*`iq`+`ip`+`mn`+`nreg`, else `npar` =
            2+2*`iq`+`ip`+`mn`+`nreg`.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, `nreg` = *<value>*.

            Constraint: `nreg` >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_exp_estim'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13fgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, yt[num].
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, nreg].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._BYREF_INT, # Input, nreg.
        utils._BYREF_INT, # Input, mn.
        utils._BYREF_INT, # Input, npar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, se[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, sc[npar].
        utils._EngineFloat64ArrayType.c_type, # Output, covr[ldcovr, npar].
        utils._BYREF_INT, # Input, ldcovr.
        utils._BYREF_FLOAT64, # Input/Output, pht.
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._BYREF_FLOAT64, # Output, lgf.
        utils._BYREF_LOGICAL, # Input, copts.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (covr, 'covr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _yt = utils._EngineFloat64ArrayType(
        yt, {'fun_name': fun_name, 'entity': 'yt'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _nreg = utils._EngineIntScalarType(
        nreg, {'fun_name': fun_name, 'entity': 'nreg'},
    ).to_ctypes()
    _nreg_py = utils._EngineIntScalarType.to_py(_nreg)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_num_py, _nreg_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _mn = utils._EngineIntScalarType(
        mn, {'fun_name': fun_name, 'entity': 'mn'},
    ).to_ctypes()
    _npar = utils._EngineIntScalarType(
        npar, {'fun_name': fun_name, 'entity': 'npar'},
    ).to_ctypes()
    _npar_py = utils._EngineIntScalarType.to_py(_npar)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_npar_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _sc_cls = utils._EngineFloat64ArrayType(
        sc, {'fun_name': fun_name, 'entity': 'sc'},
        exp_shape=(_npar_py,),
        check_contents=False,
    )
    _sc = _sc_cls.to_ctypes(can_reference_contents=False)
    _ldcovr_py = utils._EngineFloat64ArrayType.get_nd(covr, {'fun_name': fun_name, 'entity': 'covr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcovr = utils._EngineIntScalarType(
        _ldcovr_py, {'fun_name': fun_name, 'entity': '_ldcovr'},
    ).to_ctypes()
    _covr_cls = utils._EngineFloat64ArrayType(
        covr, {'fun_name': fun_name, 'entity': 'covr'},
        exp_shape=(_npar_py, _npar_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _covr = _covr_cls.to_ctypes(can_reference_contents=False)
    _pht = utils._EngineFloat64ScalarType(
        pht, {'fun_name': fun_name, 'entity': 'pht'},
    ).to_ctypes()
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _lgf = utils._EngineFloat64ScalarType.empty_ctypes()
    _copts = utils._EngineLogicalScalarType(
        copts, {'fun_name': fun_name, 'entity': 'copts'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _dist,
        _yt,
        _x,
        _ldx,
        _num,
        _ip,
        _iq,
        _nreg,
        _mn,
        _npar,
        _theta,
        _se,
        _sc,
        _covr,
        _ldcovr,
        _pht,
        _et,
        _ht,
        _lgf,
        _copts,
        _maxit,
        _tol,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 88888:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= nreg+mn.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npar < 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npar = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if dist = \'N\' then npar = 1+2*iq+ip+mn+nreg,',
                    'else npar = 2+2*iq+ip+mn+nreg.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dist = \'N\' or \'T\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mn = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nreg = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nreg >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the matrix X is not full rank.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The information matrix is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The maximum number of iterations has been reached.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The log-likelihood cannot be optimized any further.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No feasible model parameters could be found.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _theta_cls.refresh_py(_theta)
        _se_cls.refresh_py(_se)
        _sc_cls.refresh_py(_sc)
        _covr_cls.refresh_py(_covr)
        _pht_py = utils._EngineFloat64ScalarType.to_py(_pht)
        _et_cls.refresh_py(_et)
        _ht_cls.refresh_py(_ht)
        _lgf_py = utils._EngineFloat64ScalarType.to_py(_lgf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['pht', 'lgf'],
        )(
            _pht_py,
            _lgf_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def uni_garch_exp_forecast(num, nt, ip, iq, theta, fht, ht, et):
    u"""
    Univariate time series, forecast function for an exponential GARCH
    (EGARCH) process.

    ``uni_garch_exp_forecast`` forecasts the conditional variances,
    h_t,t = T+1,...,T+xi from an exponential GARCH(p,q) sequence, where
    xi is the forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13fh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fhf.html

    Parameters
    ----------
    num : int
        The number of terms in the arrays `ht` and `et` from the
        modelled sequence.

    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (2*`iq`+`ip`+1)
        The initial parameter estimates for the vector theta. The first
        element must contain the coefficient alpha_o and the next `iq`
        elements must contain the autoregressive coefficients alpha_i,
        for i = 1,2,...,q. The next `iq` elements must contain the
        coefficients phi_i, for i = 1,2,...,q. The next `ip` elements
        must contain the moving average coefficients beta_j, for j =
        1,2,...,p.

    fht : float, shape (`nt`), modified in place
        On exit: the forecast values of the conditional variance, h_t,
        for t = T+1,...,T+xi.

    ht : float, shape (`num`)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, shape (`num`)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: max(`ip`, `iq`) <= `num`.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.uni_garch_exp_forecast'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13fhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, nt.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[2*iq+ip+1].
        utils._EngineFloat64ArrayType.c_type, # Output, fht[nt].
        utils._EngineFloat64ArrayType.c_type, # Input, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Input, et[num].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(2*_iq_py+_ip_py+1,),
    ).to_ctypes()
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    _fht_cls = utils._EngineFloat64ArrayType(
        fht, {'fun_name': fun_name, 'entity': 'fht'},
        exp_shape=(_nt_py,),
        check_contents=False,
    )
    _fht = _fht_cls.to_ctypes(can_reference_contents=False)
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _et = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _num,
        _nt,
        _ip,
        _iq,
        _theta,
        _fht,
        _ht,
        _et,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, max(ip-1, iq-1) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip-1, iq-1) <= 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: max(ip, iq) <= num.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nt > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _fht_cls.refresh_py(_fht)
        return_data = None
    _err_class.manage()
    return return_data

def inhom_iema(nb, iema, t, tau, m, sinit, inter, pn, comm):
    u"""
    Computes the iterated exponential moving average for a univariate
    inhomogeneous time series.

    ``inhom_iema`` calculates the iterated exponential moving average
    for an inhomogeneous time series.

    For full information please refer to the NAG Library document for
    g13me

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13mef.html

    Parameters
    ----------
    nb : int
        b, the number of observations in the current block of data. The
        size of the block of data supplied in `iema` and `t` can vary;
        therefore `nb` can change between calls to ``inhom_iema``.

    iema : float, shape (`nb`), modified in place
        On entry: z_i, the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        On exit: the iterated EMA, with `iema`[i-1] = EMA[tau,m;z](t_i).

    t : float, shape (`nb`)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 31 will be returned, but
        ``inhom_iema`` will continue as if t was strictly increasing by
        using the absolute value.

    tau : float
        tau, the parameter controlling the rate of decay, which must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau
        can be calculated without overflowing, for all i.

    m : int
        m, the number of times the EMA operator is to be iterated.

    sinit : None or float, shape (`m`+2)
        If `pn` = 0, the values used to start the iterative process,
        with

        - `sinit`[0] = t_0,

        - `sinit`[1] = z_0,

        - `sinit`[j+1] = EMA[tau,j;z](t_0), for j = 1,2,...,`m`.

        If `pn` != 0, `sinit` is not referenced.

    inter : int, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

          Previous point, with nu = 1.

        `inter`[i] = 2

          Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

          Next point, nu = mu.

        Zumbach and Mueller recommend that linear interpolation is used
        in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    pn : int
        k, the number of observations processed so far. On the first
        call to ``inhom_iema``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to ``inhom_iema``.

    comm : None or dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    pn : int
        k+b, the updated number of observations processed so far.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `nb` = *<value>*.

            Constraint: `nb` >= 0.

        (`errno` 32)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 41)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 42)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 52)
            On entry, `m` = *<value>*.

            On entry at previous call, `m` = *<value>*.

            Constraint: if `pn` > 0 then `m` must be unchanged since
            previous call.

        (`errno` 71)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 72)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 73)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the previous call.

        (`errno` 81)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 82)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 91)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 101)
            On entry, `pn` = 0, lrcomm = *<value>* and `m` = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= `m`+20.

        (`errno` 102)
            On entry, `pn` != 0, lrcomm = *<value>* and `m` = *<value>*.

            Constraint: if `pn` != 0, lrcomm >= `m`+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `iema` or for `tau`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.inhom_iema_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.inhom_iema'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13meft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nb.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, iema[nb].
        utils._EngineFloat64ArrayType.c_type, # Input, t[nb].
        utils._BYREF_FLOAT64, # Input, tau.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, sinit[m+2].
        utils._EngineIntArrayType.c_type, # Input, inter[2].
        utils._BYREF_INT, # Input/Output, pn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _iema_cls = utils._EngineFloat64ArrayType(
        iema, {'fun_name': fun_name, 'entity': 'iema'},
        exp_shape=(_nb_py,),
    )
    _iema = _iema_cls.to_ctypes()
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _tau = utils._EngineFloat64ScalarType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _pn = utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = utils._EngineIntScalarType.to_py(_pn)
    if _pn_py == 0:
        if sinit is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'sinit'},
                'must not be None when pn = 0',
            )
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _sinit = utils._EngineFloat64ArrayType(
        sinit, {'fun_name': fun_name, 'entity': 'sinit'},
        exp_shape=(_m_py+2,),
        check_contents=_pn_py == 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_pn_py == 0)
    _inter = utils._EngineIntArrayType(
        inter, {'fun_name': fun_name, 'entity': 'inter'},
        exp_shape=(2,),
    ).to_ctypes()
    _vl_dummy_rcomm = utils._EngineFloat64ArrayType.empty_ctypes(0)
    if _pn_py != 0:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    if _pn_py != 0:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('rcomm',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('rcomm',),
            allow_none=True,
        )
    if comm is None:
        _rcomm = _vl_dummy_rcomm
    else:
        if _comm_needs_init:
            if comm is not None:
                _rcomm_dim1 = _m_py+20
            else:
                _rcomm_dim1 = 0
            _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_rcomm_dim1)
            comm['rcomm'] = _rcomm
        else:
            _rcomm = comm['rcomm']
    if comm is None:
        _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_rcomm, {'fun_name': fun_name, 'entity': '_vl_dummy_rcomm'}, exp_rank=1)[0]
        _lrcomm = utils._EngineIntScalarType(
            _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
        ).to_ctypes()
    else:
        _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(_rcomm, {'fun_name': fun_name, 'entity': '_rcomm'}, exp_rank=1, validate=False)[0]
        _lrcomm = utils._EngineIntScalarType(
            _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
        ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nb,
        _iema,
        _t,
        _tau,
        _m,
        _sinit,
        _inter,
        _pn,
        _rcomm,
        _lrcomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nb >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, t[i-2] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t should be strictly increasing.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, t[i-2] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i-1] != t[i-2] if linear interpolation is being used.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tau > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then tau must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then m must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inter[0] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 72 and _ifmt == 72:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inter[1] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 73 and _ifmt == 73:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[0] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'On entry at previous call, inter[0] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if pn != 0, inter must be unchanged since the previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pn >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On exit from previous call, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then pn must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'rcomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = 0, lrcomm = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn = 0, lrcomm = 0 or lrcomm >= m+20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn != 0, lrcomm = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn != 0, lrcomm >= m+20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 301 and _ifmt == 301:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Truncation occurred to avoid overflow, check for extreme values in t, iema',
                    'or for tau.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _iema_cls.refresh_py(_iema)
        _pn_py = utils._EngineIntScalarType.to_py(_pn)
        return_data = _pn_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['pn'],
            )(
                _pn_py,
            )
    _err_class.manage()
    return return_data

def inhom_iema_all(sorder, nb, z, iema, t, tau, m1, m2, sinit, inter, ftype, p, x, pn, comm):
    u"""
    Computes the iterated exponential moving average for a univariate
    inhomogeneous time series, intermediate results are also returned.

    ``inhom_iema_all`` calculates the iterated exponential moving
    average for an inhomogeneous time series, returning the intermediate
    results.

    For full information please refer to the NAG Library document for
    g13mf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13mff.html

    Parameters
    ----------
    sorder : int
        Determines the storage order of output returned in `iema`.

    nb : int
        b, the number of observations in the current block of data. At
        each call the size of the block of data supplied in `z`, `t` and
        `x` can vary; therefore `nb` can change between calls to
        ``inhom_iema_all``.

    z : float, shape (`nb`)
        z_i, the current block of observations, for i = k+1,...,k+b,
        where k is the number of observations processed so far, i.e.,
        the value supplied in `pn` on entry.

    iema : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `nb`; otherwise:
        `m2`-`m1`+1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `m2`-`m1`+1; if
        `sorder`=2: `nb`; otherwise: 0.

        On exit: the iterated exponential moving average.

        If `sorder` = 1, `iema`[i-1,j-1] = EMA[tau,j+`m1`-1;y](t_{i+k}).

        If `sorder` = 2, `iema`[j-1,i-1] = EMA[tau,j+`m1`-1;y](t_{i+k}).

        For i = 1,2,...,`nb`, j = 1,2,...,`m2`-`m1`+1 and k is the
        number of observations processed so far, i.e., the value
        supplied in `pn` on entry.

    t : float, shape (`nb`)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 61 will be returned, but
        ``inhom_iema_all`` will continue as if t was strictly increasing
        by using the absolute value.

    tau : float
        tau, the parameter controlling the rate of decay. tau must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau
        can be calculated without overflowing, for all i.

    m1 : int
        The minimum number of times the EMA operator is to be iterated.

    m2 : int
        The maximum number of times the EMA operator is to be iterated.
        Therefore ``inhom_iema_all`` returns EMA[tau,m;y], for m =
        `m1`,`m1`+1,...,`m2`.

    sinit : None or float, shape (`m2`+2)
        If `pn` = 0, the values used to start the iterative process,
        with

        - `sinit`[0] = t_0,

        - `sinit`[1] = y_0,

        - `sinit`[j+1] = EMA[tau,j;y](t_0), j = 1,2,...,`m2`.

        If `pn` != 0 then `sinit` is not referenced.

    inter : int, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

          Previous point, with nu = 1.

        `inter`[i] = 2

          Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

          Next point, nu = mu.

        Zumbach and Mueller recommend that linear interpolation is used
        in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    ftype : int
        The function type used to define the relationship between y and
        z when calculating EMA[tau,1;y]. Three functions are provided:

        `ftype` = 1

          The identity function, with y_i = z_i^[p].

        `ftype` = 2

          The absolute value, with y_i = |z_i|^p.

        `ftype` = 3

          The absolute difference, with y_i = |z_i-x_i|^p, where the
          vector x is supplied in `x`.

    p : float
        p, the power used in the transformation function.

    x : None or float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ftype`=3: `nb`; otherwise: 0.

        If `ftype` = 3, x_i, the vector used to shift the current block
        of observations, for i = k+1,...,k+b, where k is the number of
        observations processed so far, i.e., the value supplied in `pn`
        on entry.

        If `ftype` != 3 then `x` is not referenced.

    pn : int
        k, the number of observations processed so far. On the first
        call to ``inhom_iema_all``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to
        ``inhom_iema_all``.

    comm : None or dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    p : float
        If `ftype` = 1, then [p], the actual power used in the
        transformation function is returned, otherwise `p` is unchanged.

    pn : int
        k+b, the updated number of observations processed so far.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `sorder` = *<value>*.

            Constraint: `sorder` = 1 or 2.

        (`errno` 21)
            On entry, `nb` = *<value>*.

            Constraint: `nb` >= 0.

        (`errno` 51)
            On entry, `sorder` = 1, ldiema = *<value>* and `nb` =
            *<value>*.

            Constraint: ldiema >= `nb`.

        (`errno` 51)
            On entry, `sorder` = 2, ldiema = *<value>* and `m2`-`m1`+1 =
            *<value>*.

            Constraint: ldiema >= `m2`-`m1`+1.

        (`errno` 62)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 71)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 72)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 81)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 82)
            On entry, `m1` = *<value>*.

            On entry at previous call, `m1` = *<value>*.

            Constraint: if `pn` > 0 then `m1` must be unchanged since
            previous call.

        (`errno` 91)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m2` >= `m1`.

        (`errno` 92)
            On entry, `m2` = *<value>*.

            On entry at previous call, `m2` = *<value>*.

            Constraint: if `pn` > 0 then `m2` must be unchanged since
            previous call.

        (`errno` 101)
            On entry, `ftype` != 1, j = *<value>* and `sinit`[j-1] =
            *<value>*.

            Constraint: if `ftype` != 1, `sinit`[j-1] >= 0.0, for j =
            2,3,...,`m2`+2.

        (`errno` 111)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 112)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 113)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the last call.

        (`errno` 121)
            On entry, `ftype` = *<value>*.

            Constraint: `ftype` = 1, 2 or 3.

        (`errno` 122)
            On entry, `ftype` = *<value>*, On entry at previous call,
            `ftype` = *<value>*.

            Constraint: if `pn` != 0, `ftype` must be unchanged since
            the previous call.

        (`errno` 131)
            On entry, `p` = *<value>*.

            Constraint: absolute value of `p` must be representable as
            an integer.

        (`errno` 132)
            On entry, `p` = *<value>*.

            Constraint: if `ftype` != 1, `p` != 0.0. If `ftype` = 1, the
            nearest integer to `p` must not be 0.

        (`errno` 133)
            On entry, i = *<value>*, `z`[i-1] = *<value>* and `p` =
            *<value>*.

            Constraint: if `ftype` = 1 or 2 and `z`[i] = 0 for any i
            then `p` > 0.0.

        (`errno` 134)
            On entry, i = *<value>*, `z`[i-1] = *<value>*, `x`[i-1] =
            *<value>* and `p` = *<value>*.

            Constraint: if `ftype` = 3 and `z`[i] = `x`[i] for any i
            then `p` > 0.0.

        (`errno` 135)
            On entry, `p` = *<value>*.

            On exit from previous call, `p` = *<value>*.

            Constraint: if `pn` > 0 then `p` must be unchanged since
            previous call.

        (`errno` 151)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 152)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 161)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 171)
            On entry, `pn` = 0, lrcomm = *<value>* and `m2` = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= `m2`+20.

        (`errno` 172)
            On entry, `pn` != 0, lrcomm = *<value>* and `m2` =
            *<value>*.

            Constraint: if `pn` != 0 then lrcomm >= `m2`+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 61)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `z`, `x` or for `tau`.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.inhom_iema_all'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13mfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, nb.
        utils._EngineFloat64ArrayType.c_type, # Input, z[nb].
        utils._EngineFloat64ArrayType.c_type, # Output, iema[ldiema, sdiema].
        utils._BYREF_INT, # Input, ldiema.
        utils._EngineFloat64ArrayType.c_type, # Input, t[nb].
        utils._BYREF_FLOAT64, # Input, tau.
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineFloat64ArrayType.c_type, # Input, sinit[m2+2].
        utils._EngineIntArrayType.c_type, # Input, inter[2].
        utils._BYREF_INT, # Input, ftype.
        utils._BYREF_FLOAT64, # Input/Output, p.
        utils._EngineFloat64ArrayType.c_type, # Input, x[*].
        utils._BYREF_INT, # Input/Output, pn.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._BYREF_INT, # Input, sdiema.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((iema, 'iema'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _ldiema_py = utils._EngineFloat64ArrayType.get_nd(iema, {'fun_name': fun_name, 'entity': 'iema'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldiema = utils._EngineIntScalarType(
        _ldiema_py, {'fun_name': fun_name, 'entity': '_ldiema'},
    ).to_ctypes()
    _sdiema_py = utils._EngineFloat64ArrayType.get_nd(iema, {'fun_name': fun_name, 'entity': 'iema'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdiema = utils._EngineIntScalarType(
        _sdiema_py, {'fun_name': fun_name, 'entity': '_sdiema'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m1_py = utils._EngineIntScalarType.to_py(_m1)
    if _sorder_py == 1:
        _iema_dim1 = _nb_py
    else:
        _iema_dim1 = _m2_py-_m1_py+1
    if _sorder_py == 1:
        _iema_dim2 = _m2_py-_m1_py+1
    elif _sorder_py == 2:
        _iema_dim2 = _nb_py
    else:
        _iema_dim2 = 0
    exp_shape_iema = (_iema_dim1, _iema_dim2)
    _iema_cls = utils._EngineFloat64ArrayType(
        iema, {'fun_name': fun_name, 'entity': 'iema'},
        exp_shape=exp_shape_iema,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _iema = _iema_cls.to_ctypes(can_reference_contents=False)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _tau = utils._EngineFloat64ScalarType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
    ).to_ctypes()
    _pn = utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = utils._EngineIntScalarType.to_py(_pn)
    if _pn_py == 0:
        if sinit is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'sinit'},
                'must not be None when pn = 0',
            )
    _sinit = utils._EngineFloat64ArrayType(
        sinit, {'fun_name': fun_name, 'entity': 'sinit'},
        exp_shape=(_m2_py+2,),
        check_contents=_pn_py == 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_pn_py == 0)
    _inter = utils._EngineIntArrayType(
        inter, {'fun_name': fun_name, 'entity': 'inter'},
        exp_shape=(2,),
    ).to_ctypes()
    _ftype = utils._EngineIntScalarType(
        ftype, {'fun_name': fun_name, 'entity': 'ftype'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _ftype_py = utils._EngineIntScalarType.to_py(_ftype)
    if _ftype_py == 3:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when ftype = 3',
            )
    if _ftype_py == 3:
        _x_dim1 = _nb_py
    else:
        _x_dim1 = 0
    exp_shape_x = (_x_dim1,)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        check_contents=_ftype_py == 3,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_ftype_py == 3)
    _vl_dummy_rcomm = utils._EngineFloat64ArrayType.empty_ctypes(0)
    if _pn_py != 0:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    if _pn_py != 0:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('rcomm',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('rcomm',),
            allow_none=True,
        )
    if comm is None:
        _rcomm = _vl_dummy_rcomm
    else:
        if _comm_needs_init:
            if comm is not None:
                _rcomm_dim1 = _m2_py+20
            else:
                _rcomm_dim1 = 0
            _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_rcomm_dim1)
            comm['rcomm'] = _rcomm
        else:
            _rcomm = comm['rcomm']
    if comm is None:
        _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_rcomm, {'fun_name': fun_name, 'entity': '_vl_dummy_rcomm'}, exp_rank=1)[0]
        _lrcomm = utils._EngineIntScalarType(
            _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
        ).to_ctypes()
    else:
        _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(_rcomm, {'fun_name': fun_name, 'entity': '_rcomm'}, exp_rank=1, validate=False)[0]
        _lrcomm = utils._EngineIntScalarType(
            _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
        ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _nb,
        _z,
        _iema,
        _ldiema,
        _t,
        _tau,
        _m1,
        _m2,
        _sinit,
        _inter,
        _ftype,
        _p,
        _x,
        _pn,
        _rcomm,
        _lrcomm,
        _sdiema,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nb >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sorder = 1, ldiema = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldiema >= nb.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _ierr_fl = 51
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sorder = 2, ldiema = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm2-m1+1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldiema >= m2-m1+1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, t[i-2] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t should be strictly increasing.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, t[i-2] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i-1] != t[i-2] if linear interpolation is being used.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tau > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 72 and _ifmt == 72:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then tau must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then m1 must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m2 >= m1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 92 and _ifmt == 92:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then m2 must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ftype != 1, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'sinit[j-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if ftype != 1, sinit[j-1] >= 0.0, for j = 2,3,...,m2+2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inter[0] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inter[1] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 113 and _ifmt == 113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[0] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'On entry at previous call, inter[0] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if pn != 0, inter must be unchanged since the last call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ftype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ftype = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 122 and _ifmt == 122:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ftype = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, ftype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn != 0, ftype must be unchanged since the previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 131 and _ifmt == 131:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: absolute value of p must be representable as an integer.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 132 and _ifmt == 132:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if ftype != 1, p != 0.0.',
                    'If ftype = 1, the nearest integer to p must not be 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 133 and _ifmt == 133:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, z[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ftype = 1 or 2 and z[i] = 0 for any i then p > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 134 and _ifmt == 134:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, z[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[i-1] = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if ftype = 3 and z[i] = x[i] for any i then p > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 135 and _ifmt == 135:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On exit from previous call, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then p must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 151 and _ifmt == 151:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pn >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 152 and _ifmt == 152:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On exit from previous call, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then pn must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 161 and _ifmt == 161:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'rcomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 171 and _ifmt == 171:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = 0, lrcomm = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn = 0, lrcomm = 0 or lrcomm >= m2+20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 172 and _ifmt == 172:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn != 0, lrcomm = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn != 0 then lrcomm >= m2+20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 301 and _ifmt == 301:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Truncation occurred to avoid overflow, check for extreme values in t, z,',
                    'x or for tau.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _iema_cls.refresh_py(_iema)
        _p_py = utils._EngineFloat64ScalarType.to_py(_p)
        _pn_py = utils._EngineIntScalarType.to_py(_pn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['p', 'pn'],
        )(
            _p_py,
            _pn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def inhom_ma(nb, ma, t, tau, m1, m2, sinit, inter, ftype, p, pn, wma, comm):
    u"""
    Computes the exponential moving average for a univariate
    inhomogeneous time series.

    ``inhom_ma`` provides a moving average, moving norm, moving variance
    and moving standard deviation operator for an inhomogeneous time
    series.

    For full information please refer to the NAG Library document for
    g13mg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13mgf.html

    Parameters
    ----------
    nb : int
        b, the number of observations in the current block of data. At
        each call the size of the block of data supplied in `ma` and `t`
        can vary; therefore `nb` can change between calls to
        ``inhom_ma``.

    ma : float, shape (`nb`), modified in place
        On entry: z_i, the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        On exit: the moving average:

        if `ftype` = 4 or 5

          `ma`[i-1] = {MA[tau,m_1,m_2;y](t_i)}^(1/p),

        otherwise

          `ma`[i-1] = MA[tau,m_1,m_2;y](t_i).

    t : float, shape (`nb`)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 31 will be returned, but
        ``inhom_ma`` will continue as if t was strictly increasing by
        using the absolute value.

        The lagged difference, t_i-t_{i-1} must be sufficiently small
        that e^(-alpha), alpha = (t_i-t_{i-1})/tau~ can be calculated
        without overflowing, for all i.

    tau : float
        tau, the parameter controlling the rate of decay. tau must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau~
        can be calculated without overflowing, for all i, where tau~ =
        2*tau/(m_2+m_1).

    m1 : int
        m_1, the iteration of the EMA operator at which the sum is
        started.

    m2 : int
        m_2, the iteration of the EMA operator at which the sum is
        ended.

    sinit : None or float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ftype` in (3, 5): 2*`m2`+3; otherwise: `m2`+2.

        If `pn` = 0, the values used to start the iterative process,
        with

        - `sinit`[0] = t_0,

        - `sinit`[1] = y_0,

        - `sinit`[j+1] = EMA[tau,j;y](t_0), for i = 1,2,...,`m2`.

        In addition, if `ftype` = 3 or 5 then

        - `sinit`[`m2`+2] = z_0,

        - `sinit`[`m2`+j+1] = EMA[tau,j;z](t_0), for j = 1,2,...,`m2`.

        i.e., initial values based on the original data z as opposed to
        the transformed data y.

        If `pn` != 0, `sinit` is not referenced.

    inter : int, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

          Previous point, with nu = 1.

        `inter`[i] = 2

          Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

          Next point, nu = mu.

        Zumbach and Mueller recommend that linear interpolation is used
        in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    ftype : int
        The function type used to define the relationship between y and
        z when calculating EMA[tau,1;y]. Three functions are provided:

        `ftype` = 1

          The identity function, with y_i = z_i^[p].

        `ftype` = 2 or 4

          The absolute value, with y_i = |z_i|^p.

        `ftype` = 3 or 5

          The absolute difference, with y_i = |z_i-MA[tau,m;y](t_i)|^p.

        If `ftype` = 4 or 5 then the resulting vector of averages is
        scaled by p^(-1) as described in `ma`.

    p : float
        p, the power used in the transformation function.

    pn : int
        k, the number of observations processed so far. On the first
        call to ``inhom_ma``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to ``inhom_ma``.

    wma : float, shape (`nb`), modified in place
        On exit: either the moving average or exponential moving
        average, depending on the value of `ftype`.

        if `ftype` = 3 or 5

          `wma`[i-1] = MA[tau;y](t_i)

        otherwise

          `wma`[i-1] = EMA[tau~;y](t_i).

    comm : None or dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    p : float
        If `ftype` = 1, then [p], the actual power used in the
        transformation function is returned, otherwise `p` is unchanged.

    pn : int
        k+b, the updated number of observations processed so far.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `nb` = *<value>*.

            Constraint: `nb` >= 0.

        (`errno` 32)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 41)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 42)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 52)
            On entry, `m1` = *<value>*.

            On entry at previous call, `m1` = *<value>*.

            Constraint: if `pn` > 0 then `m1` must be unchanged since
            previous call.

        (`errno` 61)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m2` >= `m1`.

        (`errno` 62)
            On entry, `m2` = *<value>*.

            On entry at previous call, `m2` = *<value>*.

            Constraint: if `pn` > 0 then `m2` must be unchanged since
            previous call.

        (`errno` 71)
            On entry, `ftype` != 1, j = *<value>* and `sinit`[j-1] =
            *<value>*.

            Constraint: if `ftype` != 1, `sinit`[j-1] >= 0.0, for j =
            2,3,...,`m2`+2.

        (`errno` 81)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 82)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 83)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the last call.

        (`errno` 91)
            On entry, `ftype` = *<value>*.

            Constraint: `ftype` = 1, 2, 3, 4 or 5.

        (`errno` 92)
            On entry, `ftype` = *<value>*, On entry at previous call,
            `ftype` = *<value>*.

            Constraint: if `pn` != 0, `ftype` must be unchanged since
            the previous call.

        (`errno` 101)
            On entry, `p` = *<value>*.

            Constraint: absolute value of `p` must be representable as
            an integer.

        (`errno` 102)
            On entry, `p` = *<value>*.

            Constraint: if `ftype` != 1, `p` != 0.0. If `ftype` = 1, the
            nearest integer to `p` must not be 0.

        (`errno` 103)
            On entry, i = *<value>*, `ma`[i-1] = *<value>* and `p` =
            *<value>*.

            Constraint: if `ftype` = 1, 2 or 4 and `ma`[i-1] = 0 for any
            i then `p` > 0.0.

        (`errno` 104)
            On entry, i = *<value>*, `ma`[i-1] = *<value>*, `wma`[i-1] =
            *<value>* and `p` = *<value>*.

            Constraint: if `p` < 0.0, `ma`[i-1]-`wma`[i-1] != 0.0, for
            any i.

        (`errno` 105)
            On entry, `p` = *<value>*.

            On exit from previous call, `p` = *<value>*.

            Constraint: if `pn` > 0 then `p` must be unchanged since
            previous call.

        (`errno` 111)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 112)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 131)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 141)
            On entry, `pn` = 0, lrcomm = *<value>* and `m2` = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= 2*`m2`+20.

        (`errno` 142)
            On entry, `pn` != 0, lrcomm = *<value>* and `m2` =
            *<value>*.

            Constraint: if `pn` != 0, lrcomm >= 2*`m2`+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `ma` or for `tau`.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.inhom_ma'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13mgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nb.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ma[nb].
        utils._EngineFloat64ArrayType.c_type, # Input, t[nb].
        utils._BYREF_FLOAT64, # Input, tau.
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineFloat64ArrayType.c_type, # Input, sinit[*].
        utils._EngineIntArrayType.c_type, # Input, inter[2].
        utils._BYREF_INT, # Input, ftype.
        utils._BYREF_FLOAT64, # Input/Output, p.
        utils._BYREF_INT, # Input/Output, pn.
        utils._EngineFloat64ArrayType.c_type, # Output, wma[nb].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._BYREF_INT, # Input, lsinit.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nb = utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = utils._EngineIntScalarType.to_py(_nb)
    _ma_cls = utils._EngineFloat64ArrayType(
        ma, {'fun_name': fun_name, 'entity': 'ma'},
        exp_shape=(_nb_py,),
    )
    _ma = _ma_cls.to_ctypes()
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_nb_py,),
    ).to_ctypes()
    _tau = utils._EngineFloat64ScalarType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _ftype = utils._EngineIntScalarType(
        ftype, {'fun_name': fun_name, 'entity': 'ftype'},
    ).to_ctypes()
    _pn = utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = utils._EngineIntScalarType.to_py(_pn)
    if _pn_py == 0:
        if sinit is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'sinit'},
                'must not be None when pn = 0',
            )
    _ftype_py = utils._EngineIntScalarType.to_py(_ftype)
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    if _ftype_py in (3, 5):
        _sinit_dim1 = 2*_m2_py+3
    else:
        _sinit_dim1 = _m2_py+2
    exp_shape_sinit = (_sinit_dim1,)
    _sinit = utils._EngineFloat64ArrayType(
        sinit, {'fun_name': fun_name, 'entity': 'sinit'},
        exp_shape=exp_shape_sinit,
        check_contents=_pn_py == 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_pn_py == 0)
    _inter = utils._EngineIntArrayType(
        inter, {'fun_name': fun_name, 'entity': 'inter'},
        exp_shape=(2,),
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _wma_cls = utils._EngineFloat64ArrayType(
        wma, {'fun_name': fun_name, 'entity': 'wma'},
        exp_shape=(_nb_py,),
        check_contents=False,
    )
    _wma = _wma_cls.to_ctypes(can_reference_contents=False)
    _vl_dummy_rcomm = utils._EngineFloat64ArrayType.empty_ctypes(0)
    if _pn_py != 0:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    if _pn_py != 0:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('rcomm',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('rcomm',),
            allow_none=True,
        )
    if comm is None:
        _rcomm = _vl_dummy_rcomm
    else:
        if _comm_needs_init:
            if comm is not None:
                _rcomm_dim1 = 2*_m2_py+20
            else:
                _rcomm_dim1 = 0
            _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_rcomm_dim1)
            comm['rcomm'] = _rcomm
        else:
            _rcomm = comm['rcomm']
    if comm is None:
        _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_rcomm, {'fun_name': fun_name, 'entity': '_vl_dummy_rcomm'}, exp_rank=1)[0]
        _lrcomm = utils._EngineIntScalarType(
            _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
        ).to_ctypes()
    else:
        _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(_rcomm, {'fun_name': fun_name, 'entity': '_rcomm'}, exp_rank=1, validate=False)[0]
        _lrcomm = utils._EngineIntScalarType(
            _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
        ).to_ctypes()
    _lsinit_py = utils._EngineFloat64ArrayType.get_shape(sinit, {'fun_name': fun_name, 'entity': 'sinit'}, exp_rank=1, allow_none=True)[0]
    _lsinit = utils._EngineIntScalarType(
        _lsinit_py, {'fun_name': fun_name, 'entity': '_lsinit'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nb,
        _ma,
        _t,
        _tau,
        _m1,
        _m2,
        _sinit,
        _inter,
        _ftype,
        _p,
        _pn,
        _wma,
        _rcomm,
        _lrcomm,
        _lsinit,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nb >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, t[i-2] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t should be strictly increasing.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, t[i-2] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't[i-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: t[i-1] != t[i-2] if linear interpolation is being used.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tau > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then tau must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then m1 must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m2 >= m1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then m2 must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ftype != 1, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'sinit[j-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if ftype != 1, sinit[j-1] >= 0.0, for j = 2,3,...,m2+2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[0] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inter[0] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: inter[1] = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 83 and _ifmt == 83:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, inter[0] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'On entry at previous call, inter[0] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'inter[1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if pn != 0, inter must be unchanged since the last call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ftype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ftype = 1, 2, 3, 4 or 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 92 and _ifmt == 92:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ftype = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, ftype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn != 0, ftype must be unchanged since the previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: absolute value of p must be representable as an integer.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if ftype != 1, p != 0.0.',
                    'If ftype = 1, the nearest integer to p must not be 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 103 and _ifmt == 103:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ma[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ftype = 1, 2 or 4 and ma[i-1] = 0 for any i then p > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 104 and _ifmt == 104:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, ma[i-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'wma[i-1] = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: if p < 0.0, ma[i-1]-wma[i-1] != 0.0, for any i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 105 and _ifmt == 105:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On exit from previous call, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then p must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pn >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On exit from previous call, pn = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn > 0 then pn must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 131 and _ifmt == 131:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'rcomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 141 and _ifmt == 141:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn = 0, lrcomm = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn = 0, lrcomm = 0 or lrcomm >= 2*m2+20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 142 and _ifmt == 142:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pn != 0, lrcomm = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if pn != 0, lrcomm >= 2*m2+20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 301 and _ifmt == 301:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Truncation occurred to avoid overflow, check for extreme values in t, ma',
                    'or for tau.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ma_cls.refresh_py(_ma)
        _p_py = utils._EngineFloat64ScalarType.to_py(_p)
        _pn_py = utils._EngineIntScalarType.to_py(_pn)
        _wma_cls.refresh_py(_wma)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['p', 'pn'],
        )(
            _p_py,
            _pn_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def cp_pelt(ctype, n, y, beta, minss, param, tau, sparam):
    u"""
    Change point detection, using the PELT algorithm.

    ``cp_pelt`` detects change points in a univariate time series, that
    is, the time points at which some feature of the data, for example
    the mean, changes.
    Change points are detected using the PELT (Pruned Exact Linear Time)
    algorithm using one of a provided set of cost functions.

    For full information please refer to the NAG Library document for
    g13na

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13naf.html

    Parameters
    ----------
    ctype : int
        A flag indicating the assumed distribution of the data and the
        type of change point being looked for.

        `ctype` = 1

          Data from a Normal distribution, looking for changes in the
          mean, mu.

        `ctype` = 2

          Data from a Normal distribution, looking for changes in the
          standard deviation sigma.

        `ctype` = 3

          Data from a Normal distribution, looking for changes in the
          mean, mu and standard deviation sigma.

        `ctype` = 4

          Data from a Gamma distribution, looking for changes in the
          scale parameter b.

        `ctype` = 5

          Data from an exponential distribution, looking for changes in
          lambda.

        `ctype` = 6

          Data from a Poisson distribution, looking for changes in
          lambda.

    n : int
        n, the length of the time series.

    y : float, shape (`n`)
        y, the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n)

        AIC

          beta = 2*p

        Hannan-Quinn

          beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        This is usually set to 2 when `ctype` = 3 and 1 otherwise.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    param : None or float, shape (1)
        phi, values for the parameters that will be treated as fixed. If
        `ctype` = 4 then `param` must be supplied.

        If supplied, then when

        `ctype` = 1

          `param`[0] = sigma, the standard deviation of the normal
          distribution. If not supplied then sigma is estimated from the
          full input data,

        `ctype` = 2

          `param`[0] = mu, the mean of the normal distribution. If not
          supplied then mu is estimated from the full input data,

        `ctype` = 4

          `param`[0] must hold the shape, a, for the gamma distribution,

        otherwise

          `param` is not referenced.

    tau : int, shape (`n`), modified in place
        On exit: the first m elements of `tau` hold the location of the
        change points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

        The remainder of `tau` is used as workspace.

    sparam : float, shape (2*`n`+2), modified in place
        On exit: the estimated values of the distribution parameters in
        each segment

        `ctype` = 1, 2 or 3

          `sparam`[2*i-2] = mu_i and `sparam`[2*i-1] = sigma_i for i =
          1,2,...,m, where mu_i and sigma_i is the mean and standard
          deviation, respectively, of the values of y in the ith
          segment.

          It should be noted that sigma_i = sigma_j when `ctype` = 1 and
          mu_i = mu_j when `ctype` = 2, for all i and j.

        `ctype` = 4

          `sparam`[2*i-2] = a_i and `sparam`[2*i-1] = b_i for i =
          1,2,...,m, where a_i and b_i are the shape and scale
          parameters, respectively, for the values of y in the ith
          segment. It should be noted that a_i = `param`[0] for all i.

        `ctype` = 5 or 6

          `sparam`[i-1] = lambda_i for i = 1,2,...,m, where lambda_i is
          the mean of the values of y in the ith segment.

        The remainder of `sparam` is used as workspace.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ctype` = *<value>*.

            Constraint: `ctype` = 1, 2, 3, 4, 5 or 6.

        (`errno` 21)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `ctype` = *<value>* and `y`[*<value>*] =
            *<value>*.

            Constraint: if `ctype` = 4, 5 or 6 then `y`[i-1] >= 0.0, for
            i = 1,2,...,`n`.

        (`errno` 32)
            On entry, `y`[*<value>*] = *<value>*, is too large.

        (`errno` 51)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 71)
            On entry, `ctype` = *<value>* and `param`[0] = *<value>*.

            Constraint: if `ctype` = 1 or 4 and `param` has been
            supplied, then `param`[0] > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 200)
            To avoid overflow some truncation occurred when calculating
            the cost function, C. All output is returned as normal.

        (`errno` 201)
            To avoid overflow some truncation occurred when calculating
            the parameter estimates returned in `sparam`. All output is
            returned as normal.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.cp_pelt_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.cp_pelt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13naft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ctype.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_FLOAT64, # Input, beta.
        utils._BYREF_INT, # Input, minss.
        utils._BYREF_INT, # Input, iparam.
        utils._EngineFloat64ArrayType.c_type, # Input, param[1].
        utils._BYREF_INT, # Output, ntau.
        utils._EngineIntArrayType.c_type, # Output, tau[n].
        utils._EngineFloat64ArrayType.c_type, # Output, sparam[2*n+2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ctype = utils._EngineIntScalarType(
        ctype, {'fun_name': fun_name, 'entity': 'ctype'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _minss = utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    if param is None:
        _iparam = utils._EngineIntScalarType.c_type(0)
    else:
        _iparam = utils._EngineIntScalarType.c_type(1)
    _ctype_py = utils._EngineIntScalarType.to_py(_ctype)
    if _ctype_py == 4:
        if param is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'param'},
                'must not be None when ctype = 4',
            )
    _param = utils._EngineFloat64ArrayType(
        param, {'fun_name': fun_name, 'entity': 'param'},
        exp_shape=(1,),
        check_contents=_ctype_py == 4 or (_ctype_py in (1, 2) and param is not None),
        allow_none=True,
    ).to_ctypes(can_reference_contents=_ctype_py == 4 or (_ctype_py in (1, 2) and param is not None))
    _ntau = utils._EngineIntScalarType.empty_ctypes()
    _tau_cls = utils._EngineIntArrayType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _tau = _tau_cls.to_ctypes(can_reference_contents=False)
    _sparam_cls = utils._EngineFloat64ArrayType(
        sparam, {'fun_name': fun_name, 'entity': 'sparam'},
        exp_shape=(2*_n_py+2,),
        check_contents=False,
    )
    _sparam = _sparam_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ctype,
        _n,
        _y,
        _beta,
        _minss,
        _iparam,
        _param,
        _ntau,
        _tau,
        _sparam,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ctype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ctype = 1, 2, 3, 4, 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ctype = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'y[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ctype = 4, 5 or 6 then y[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'is too large.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, minss = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: minss >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ctype = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'param[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if ctype = 1 or 4 and param has been supplied, then param[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 200 and _ifmt == 200:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'To avoid overflow some truncation occurred when calculating',
                    'the cost function, C. All output is returned as normal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'To avoid overflow some truncation occurred when calculating the parameter',
                    'estimates returned in sparam. All output is returned as normal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _tau_cls.refresh_py(_tau)
        _sparam_cls.refresh_py(_sparam)
        _ntau_py = utils._EngineIntScalarType.to_py(_ntau)
        return_data = _ntau_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ntau'],
            )(
                _ntau_py,
            )
    _err_class.manage()
    return return_data

def cp_pelt_user(n, beta, minss, k, costfn, tau, data=None):
    u"""
    Change points detection using the PELT algorithm, user supplied cost
    function.

    ``cp_pelt_user`` detects change points in a univariate time series,
    that is, the time points at which some feature of the data, for
    example the mean, changes.
    Change points are detected using the PELT (Pruned Exact Linear Time)
    algorithm and a user-supplied cost function.

    For full information please refer to the NAG Library document for
    g13nb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13nbf.html

    Parameters
    ----------
    n : int
        n, the length of the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n)

        AIC

          beta = 2*p

        Hannan-Quinn

          beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        The value of p will depend on the cost function being used.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    k : float
        K, the constant value that satisfies equation [equation]. If K
        exists, it is unlikely to be unique in such cases, it is
        recommened that the largest value of K, that satisfies equation
        [equation], is chosen. No check is made that K is the correct
        value for the supplied cost function.

    costfn : callable costfn(ts, nr, r, c, data=None)
        The cost function, C. `costfn` must calculate a vector of costs
        for a number of segments.

        Parameters
        ~~~~~~~~~~
        ts : int
            A reference time point.

        nr : int
            Number of segments being considered.

        r : int, ndarray, shape (`nr`)
            Time points which, along with `ts`, define the segments
            being considered, 0 <= `r`[i-1] <= n for i = 1,2,...`nr`.

        c : float, ndarray, shape (`nr`), to be modified in place
            On exit, the cost function, C.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    tau : int, shape (`n`), modified in place
        On exit: the first m elements of `tau` hold the location of the
        change points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

        The remainder of `tau` is used as workspace.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 51)
            User requested termination.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.cp_pelt_user'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_costfn_name = fun_name + ':costfn'

    costfnh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ts.
        utils._BYREF_INT, # Input, nr.
        utils._EngineIntArrayType.c_type, # Input, r[nr].
        utils._EngineFloat64ArrayType.c_type, # Output, c[nr].
        utils._BYREF_INT, # Input, my.
        utils._BYREF_INT, # Input, ny.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[0].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Input/Output, info.
    )

    def cb_costfnh(_costfn, en_data, ts, nr, r, c, my, ny, y, _iuser, _ruser, info):
        """Helper for costfn."""
        try:
            _ts_costfnh = utils._EngineIntScalarType.to_py(ts.contents)
            _nr_costfnh = utils._EngineIntScalarType.to_py(nr.contents)
            _r_costfnh = utils._EngineIntArrayType.from_pointer(
                r,
                (_nr_costfnh,),
            )
            _c_costfnh = utils._EngineFloat64ArrayType.from_pointer(
                c,
                (_nr_costfnh,),
                can_reference_contents=False,
            )
            if data is not None:
                costfn(
                    _ts_costfnh,
                    _nr_costfnh,
                    _r_costfnh,
                    _c_costfnh,
                    data,
                )
            else:
                costfn(
                    _ts_costfnh,
                    _nr_costfnh,
                    _r_costfnh,
                    _c_costfnh,
                )
            utils._unlink_pointer(r)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _c_costfnh, {'fun_name': cb_costfn_name, 'returned': 'c'}, c,
            )
            utils._unlink_pointer(c)
        except:
            excs.append(_sys.exc_info())
            info.contents.value = -1
            if excs[-1][0] is not utils.UserCallbackTerminate:
                en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g13nbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._BYREF_INT, # Input, minss.
        utils._BYREF_FLOAT64, # Input, k.
        utils._UNUSED_FUNCTYPE,
        costfnh,
        utils._BYREF_INT, # Output, ntau.
        utils._EngineIntArrayType.c_type, # Output, tau[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[0].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _minss = utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    _k = utils._EngineFloat64ScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    utils._check_callable(costfn, {'fun_name': fun_name, 'entity': 'costfn'})
    _costfn = utils._UNUSED_CALLBACK
    _costfnh = utils._cast_callback(
        costfnh, cb_costfnh,
        {'fun_name': fun_name, 'entity': 'costfn'},
    )
    _ntau = utils._EngineIntScalarType.empty_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _tau_cls = utils._EngineIntArrayType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _tau = _tau_cls.to_ctypes(can_reference_contents=False)
    _y = None
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _beta,
        _minss,
        _k,
        _costfn,
        _costfnh,
        _ntau,
        _tau,
        _y,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, minss = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: minss >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._UserCallbackExit(
                fun_name=fun_name,
                msg_lines=[
                    'User requested termination.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
                cb_excs=excs,
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _tau_cls.refresh_py(_tau)
        _ntau_py = utils._EngineIntScalarType.to_py(_ntau)
        return_data = _ntau_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ntau'],
            )(
                _ntau_py,
            )
    _err_class.manage()
    return return_data

def cp_binary(ctype, n, y, beta, minss, param, mdepth, tau, sparam):
    u"""
    Change point detection, using binary segmentation.

    ``cp_binary`` detects change points in a univariate time series,
    that is, the time points at which some feature of the data, for
    example the mean, changes.
    Change points are detected using binary segmentation using one of a
    provided set of cost functions.

    For full information please refer to the NAG Library document for
    g13nd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ndf.html

    Parameters
    ----------
    ctype : int
        A flag indicating the assumed distribution of the data and the
        type of change point being looked for.

        `ctype` = 1

          Data from a Normal distribution, looking for changes in the
          mean, mu.

        `ctype` = 2

          Data from a Normal distribution, looking for changes in the
          standard deviation sigma.

        `ctype` = 3

          Data from a Normal distribution, looking for changes in the
          mean, mu and standard deviation sigma.

        `ctype` = 4

          Data from a Gamma distribution, looking for changes in the
          scale parameter b.

        `ctype` = 5

          Data from an exponential distribution, looking for changes in
          lambda.

        `ctype` = 6

          Data from a Poisson distribution, looking for changes in
          lambda.

    n : int
        n, the length of the time series.

    y : float, shape (`n`)
        y, the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n)

        AIC

          beta = 2*p

        Hannan-Quinn

          beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        This is usually set to 2 when `ctype` = 3 and 1 otherwise.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    param : None or float, shape (1)
        phi, values for the parameters that will be treated as fixed. If
        `ctype` = 4 then `param` must be supplied.

        If supplied, then when

        `ctype` = 1

          `param`[0] = sigma, the standard deviation of the normal
          distribution. If not supplied then sigma is estimated from the
          full input data,

        `ctype` = 2

          `param`[0] = mu, the mean of the normal distribution. If not
          supplied then mu is estimated from the full input data,

        `ctype` = 4

          `param`[0] must hold the shape, a, for the gamma distribution,

        otherwise

          `param` is not referenced.

    mdepth : int
        K, the maximum depth for the iterative process, which in turn
        puts an upper limit on the number of change points with m <=
        2^K.

        If K <= 0 then no limit is put on the depth of the iterative
        process and no upper limit is put on the number of change
        points, other than that inherent in the length of the series and
        the value of `minss`.

    tau : int, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `mdepth`>0: min(ceiling(`n`/`minss`), 2^`mdepth`);
        otherwise: ceiling(`n`/`minss`).

        On exit: the first m elements of `tau` hold the location of the
        change points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

        The remainder of `tau` is used as workspace.

    sparam : float, shape (2*`n`), modified in place
        On exit: the estimated values of the distribution parameters in
        each segment

        `ctype` = 1, 2 or 3

          `sparam`[2*i-2] = mu_i and `sparam`[2*i-1] = sigma_i for i =
          1,2,...,m, where mu_i and sigma_i is the mean and standard
          deviation, respectively, of the values of y in the ith
          segment.

          It should be noted that sigma_i = sigma_j when `ctype` = 1 and
          mu_i = mu_j when `ctype` = 2, for all i and j.

        `ctype` = 4

          `sparam`[2*i-2] = a_i and `sparam`[2*i-1] = b_i for i =
          1,2,...,m, where a_i and b_i are the shape and scale
          parameters, respectively, for the values of y in the ith
          segment. It should be noted that a_i = `param`[0] for all i.

        `ctype` = 5 or 6

          `sparam`[i-1] = lambda_i for i = 1,2,...,m, where lambda_i is
          the mean of the values of y in the ith segment.

        The remainder of `sparam` is used as workspace.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ctype` = *<value>*.

            Constraint: `ctype` = 1, 2, 3, 4, 5 or 6.

        (`errno` 21)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `ctype` = *<value>* and `y`[*<value>*] =
            *<value>*.

            Constraint: if `ctype` = 4, 5 or 6 then `y`[i-1] >= 0.0, for
            i = 1,2,...,`n`.

        (`errno` 32)
            On entry, `y`[*<value>*] = *<value>*, is too large.

        (`errno` 51)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 71)
            On entry, `ctype` = *<value>* and `param`[0] = *<value>*.

            Constraint: if `ctype` = 1 or 4 and `param` has been
            supplied, then `param`[0] > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 200)
            To avoid overflow some truncation occurred when calculating
            the cost function, C. All output is returned as normal.

        (`errno` 201)
            To avoid overflow some truncation occurred when calculating
            the parameter estimates returned in `sparam`. All output is
            returned as normal.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.cp_binary'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g13ndft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ctype.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_FLOAT64, # Input, beta.
        utils._BYREF_INT, # Input, minss.
        utils._BYREF_INT, # Input, iparam.
        utils._EngineFloat64ArrayType.c_type, # Input, param[1].
        utils._BYREF_INT, # Input, mdepth.
        utils._BYREF_INT, # Output, ntau.
        utils._EngineIntArrayType.c_type, # Output, tau[ltau].
        utils._BYREF_INT, # Input, ltau.
        utils._EngineFloat64ArrayType.c_type, # Output, sparam[2*n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ctype = utils._EngineIntScalarType(
        ctype, {'fun_name': fun_name, 'entity': 'ctype'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _minss = utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    if param is None:
        _iparam = utils._EngineIntScalarType.c_type(0)
    else:
        _iparam = utils._EngineIntScalarType.c_type(1)
    _ctype_py = utils._EngineIntScalarType.to_py(_ctype)
    if _ctype_py == 4:
        if param is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'param'},
                'must not be None when ctype = 4',
            )
    _param = utils._EngineFloat64ArrayType(
        param, {'fun_name': fun_name, 'entity': 'param'},
        exp_shape=(1,),
        check_contents=_ctype_py == 4 or (_ctype_py in (1, 2) and param is not None),
        allow_none=True,
    ).to_ctypes(can_reference_contents=_ctype_py == 4 or (_ctype_py in (1, 2) and param is not None))
    _mdepth = utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    _ntau = utils._EngineIntScalarType.empty_ctypes()
    _ltau_py = utils._EngineIntArrayType.get_shape(tau, {'fun_name': fun_name, 'entity': 'tau'}, exp_rank=1)[0]
    _ltau = utils._EngineIntScalarType(
        _ltau_py, {'fun_name': fun_name, 'entity': '_ltau'},
    ).to_ctypes()
    _mdepth_py = utils._EngineIntScalarType.to_py(_mdepth)
    _minss_py = utils._EngineIntScalarType.to_py(_minss)
    if _mdepth_py > 0:
        _tau_dim1 = min(int(ceil(float(_n_py)/float(_minss_py))), 2**_mdepth_py)
    else:
        _tau_dim1 = int(ceil(float(_n_py)/float(_minss_py)))
    exp_shape_tau = (_tau_dim1,)
    _tau_cls = utils._EngineIntArrayType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
        exp_shape=exp_shape_tau,
        check_contents=False,
    )
    _tau = _tau_cls.to_ctypes(can_reference_contents=False)
    _sparam_cls = utils._EngineFloat64ArrayType(
        sparam, {'fun_name': fun_name, 'entity': 'sparam'},
        exp_shape=(2*_n_py,),
        check_contents=False,
    )
    _sparam = _sparam_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ctype,
        _n,
        _y,
        _beta,
        _minss,
        _iparam,
        _param,
        _mdepth,
        _ntau,
        _tau,
        _ltau,
        _sparam,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ctype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ctype = 1, 2, 3, 4, 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ctype = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'y[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: if ctype = 4, 5 or 6 then y[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'is too large.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, minss = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: minss >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ctype = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'param[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if ctype = 1 or 4 and param has been supplied, then param[0] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 200 and _ifmt == 200:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'To avoid overflow some truncation occurred when calculating',
                    'the cost function, C. All output is returned as normal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'To avoid overflow some truncation occurred when calculating the parameter',
                    'estimates returned in sparam. All output is returned as normal.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _tau_cls.refresh_py(_tau)
        _sparam_cls.refresh_py(_sparam)
        _ntau_py = utils._EngineIntScalarType.to_py(_ntau)
        return_data = _ntau_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ntau'],
            )(
                _ntau_py,
            )
    _err_class.manage()
    return return_data

def cp_binary_user(n, beta, minss, mdepth, chgpfn, tau, data=None):
    u"""
    Change point detection, using binary segmentation, user supplied
    cost function.

    ``cp_binary_user`` detects change points in a univariate time
    series, that is, the time points at which some feature of the data,
    for example the mean, changes.
    Change points are detected using binary segmentation for a
    user-supplied cost function.

    For full information please refer to the NAG Library document for
    g13ne

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13nef.html

    Parameters
    ----------
    n : int
        n, the length of the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n).

        AIC

          beta = 2*p.

        Hannan-Quinn

          beta = 2*p*log(log(n)).

        where p is the number of parameters being treated as estimated
        in each segment.

        The value of p will depend on the cost function being used.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    mdepth : int
        K, the maximum depth for the iterative process, which in turn
        puts an upper limit on the number of change points with m <=
        2^K.

        If K <= 0 then no limit is put on the depth of the iterative
        process and no upper limit is put on the number of change
        points, other than that inherent in the length of the series and
        the value of `minss`.

    chgpfn : callable (v, info) = chgpfn(side, u, w, minss, cost, info,
        data=None)
        `chgpfn` must calculate a proposed change point, and the
        associated costs, within a specified segment.

        Parameters
        ~~~~~~~~~~
        side : int
            Flag indicating what `chgpfn` must calculate and at which
            point of the Binary Segmentation it has been called.

            `side` = -1

              only C(y_{u:w}) need be calculated and returned in
              `cost`[0], neither `v` nor the other elements of `cost`
              need be set. In this case, u = 1 and w = n.

            `side` = 0

              all elements of `cost` and `v` must be set. In this case,
              u = 1 and w = n.

            `side` = 1

              the segment, y_{u:w}, is a left-hand side subsegment from
              a previous iteration of the Binary Segmentation algorithm.
              All elements of `cost` and `v` must be set.

            `side` = 2

              the segment, y_{u:w}, is a right-hand side subsegment from
              a previous iteration of the Binary Segmentation algorithm.
              All elements of `cost` and `v` must be set.

            The distinction between `side` = 1 or 2 may allow for
            `chgpfn` to be implemented in a more efficient manner.

            See [Example] for one such example.

            The first call to `chgpfn` will always have `side` = -1 and
            the second call will always have `side` = 0.

            All subsequent calls will be made with `side` = 1 or 2.

        u : int
            u, the start of the segment of interest.

        w : int
            w, the end of the segment of interest.

        minss : int
            The minimum distance between two change points, as passed to
            ``cp_binary_user``.

        cost : float, ndarray, shape (3), to be modified in place
            On exit: costs associated with the proposed change point, v.

            If `side` = -1 then `cost`[0] = C(y_{u:w}) and the remaining
            two elements of `cost` need not be set.

            If `side` != -1 then

            - `cost`[0] = C(y_{u:v})+C(y_{v+1:w}).

            - `cost`[1] = C(y_{u:v}).

            - `cost`[2] = C(y_{v+1:w}).

        info : int
            `info` = 0.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        v : int
            If `side` = -1 then `v` need not be set.

            if `side` != -1 then v, the proposed change point.

            That is, the value which minimizes

                minimize_v C(y_{u:v})+C(y_{v+1:w})

            for v = u+`minss`-1 to w-`minss`.

        info : int
            In most circumstances `info` should remain unchanged.

            If `info` is set to a strictly positive value then
            ``cp_binary_user`` terminates with `errno` = 51.

            If `info` is set to a strictly negative value the current
            segment is skipped and ``cp_binary_user`` continues as
            normal.

            If `info` was set to a strictly negative value at any point
            and no other errors occur then ``cp_binary_user`` will
            terminate with `errno` = 52.

    tau : int, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `mdepth`>0: min(ceiling(`n`/`minss`), 2^`mdepth`);
        otherwise: ceiling(`n`/`minss`).

        On exit: the first m elements of `tau` hold the location of the
        change points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

        The remainder of `tau` is used as workspace.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 51)
            User requested termination by setting `info` = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 52)
            User requested a segment to be skipped by setting `info` =
            *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.tsa.cp_binary_user'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        from logging import getLogger
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_chgpfn_name = fun_name + ':chgpfn'

    chgpfnh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, side.
        utils._BYREF_INT, # Input, u.
        utils._BYREF_INT, # Input, w.
        utils._BYREF_INT, # Input, minss.
        utils._BYREF_INT, # Output, v.
        utils._EngineFloat64ArrayType.c_type, # Output, cost[3].
        utils._BYREF_INT, # Input, my.
        utils._BYREF_INT, # Input, ny.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[0].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, fy[1].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._BYREF_INT, # Input/Output, info.
    )

    def cb_chgpfnh(_chgpfn, en_data, side, u, w, minss, v, cost, my, ny, y, fy, _iuser, _ruser, info):
        """Helper for chgpfn."""
        try:
            _side_chgpfnh = utils._EngineIntScalarType.to_py(side.contents)
            _u_chgpfnh = utils._EngineIntScalarType.to_py(u.contents)
            _w_chgpfnh = utils._EngineIntScalarType.to_py(w.contents)
            _minss_chgpfnh = utils._EngineIntScalarType.to_py(minss.contents)
            _cost_chgpfnh = utils._EngineFloat64ArrayType.from_pointer(
                cost,
                (3,),
                can_reference_contents=False,
            )
            _info_chgpfnh = utils._EngineIntScalarType.to_py(info.contents)
            if data is not None:
                _v_chgpfnh, _info_chgpfnh = chgpfn(
                    _side_chgpfnh,
                    _u_chgpfnh,
                    _w_chgpfnh,
                    _minss_chgpfnh,
                    _cost_chgpfnh,
                    _info_chgpfnh,
                    data,
                )
            else:
                _v_chgpfnh, _info_chgpfnh = chgpfn(
                    _side_chgpfnh,
                    _u_chgpfnh,
                    _w_chgpfnh,
                    _minss_chgpfnh,
                    _cost_chgpfnh,
                    _info_chgpfnh,
                )
            utils._EngineIntScalarType(
                _v_chgpfnh, {'fun_name': cb_chgpfn_name, 'returned': 'v'},
            ).refresh_ctypes_buffer(v)
            utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _cost_chgpfnh, {'fun_name': cb_chgpfn_name, 'returned': 'cost'}, cost,
            )
            utils._unlink_pointer(cost)
            utils._EngineIntScalarType(
                _info_chgpfnh, {'fun_name': cb_chgpfn_name, 'returned': 'info'},
            ).refresh_ctypes_buffer(info)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g13neft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._BYREF_INT, # Input, minss.
        utils._BYREF_INT, # Input, mdepth.
        utils._UNUSED_FUNCTYPE,
        chgpfnh,
        utils._BYREF_INT, # Output, ntau.
        utils._EngineIntArrayType.c_type, # Output, tau[ltau].
        utils._BYREF_INT, # Input, ltau.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[0].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _minss = utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    _mdepth = utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    utils._check_callable(chgpfn, {'fun_name': fun_name, 'entity': 'chgpfn'})
    _chgpfn = utils._UNUSED_CALLBACK
    _chgpfnh = utils._cast_callback(
        chgpfnh, cb_chgpfnh,
        {'fun_name': fun_name, 'entity': 'chgpfn'},
    )
    _ntau = utils._EngineIntScalarType.empty_ctypes()
    _ltau_py = utils._EngineIntArrayType.get_shape(tau, {'fun_name': fun_name, 'entity': 'tau'}, exp_rank=1)[0]
    _ltau = utils._EngineIntScalarType(
        _ltau_py, {'fun_name': fun_name, 'entity': '_ltau'},
    ).to_ctypes()
    _mdepth_py = utils._EngineIntScalarType.to_py(_mdepth)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _minss_py = utils._EngineIntScalarType.to_py(_minss)
    if _mdepth_py > 0:
        _tau_dim1 = min(int(ceil(float(_n_py)/float(_minss_py))), 2**_mdepth_py)
    else:
        _tau_dim1 = int(ceil(float(_n_py)/float(_minss_py)))
    exp_shape_tau = (_tau_dim1,)
    _tau_cls = utils._EngineIntArrayType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
        exp_shape=exp_shape_tau,
        check_contents=False,
    )
    _tau = _tau_cls.to_ctypes(can_reference_contents=False)
    _y = None
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _beta,
        _minss,
        _mdepth,
        _chgpfn,
        _chgpfnh,
        _ntau,
        _tau,
        _ltau,
        _y,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, minss = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: minss >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'User requested termination',
                    'by setting info = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'User requested a segment to be skipped',
                    'by setting info = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _tau_cls.refresh_py(_tau)
        _ntau_py = utils._EngineIntScalarType.to_py(_ntau)
        return_data = _ntau_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ntau'],
            )(
                _ntau_py,
            )
    _err_class.manage()
    return return_data
