# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `blas` Chapter.

``blas`` - Linear Algebra Support Routines

This module is concerned with basic linear algebra functions which
perform elementary algebraic operations involving scalars, vectors and
matrices.
It includes functions which conform to the specifications of the BLAS
(Basic Linear Algebra Subprograms).

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections

from ..base import utils as _b_utils
from ..library import _utils

def drotg(a, b):
    u"""
    Generate real plane rotation.

    ``drotg`` generates a real Givens plane rotation.

    For full information please refer to the NAG Library document for
    f06aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06aaf.html

    Parameters
    ----------
    a : float
        The value a, the first element of the vector which determines
        the rotation.

    b : float
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : float
        The value d.

    b : float
        The value z, from which c and s can be reconstructed.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import drotg as b_drotg

    _return_scalars = b_drotg(
        a,
        b,
    )
    _a_py = _return_scalars.a
    _b_py = _return_scalars.b
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return _return_data

def drotgc(a, b):
    u"""
    Generate real plane rotation, storing tangent.

    ``drotgc`` generates a real Givens plane rotation and the tangent of
    that rotation.

    For full information please refer to the NAG Library document for
    f06ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06baf.html

    Parameters
    ----------
    a : float
        The value a, the first element of the vector which determines
        the rotation.

    b : float
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : float
        The value d.

    b : float
        The value t, the tangent of the rotation.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import drotgc as b_drotgc

    _return_scalars = b_drotgc(
        a,
        b,
    )
    _a_py = _return_scalars.a
    _b_py = _return_scalars.b
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return _return_data

def dcsg(t):
    u"""
    Recover cosine and sine from given real tangent.

    ``dcsg`` reconstructs the parameters c and s of a real plane
    rotation from the tangent of that rotation.

    For full information please refer to the NAG Library document for
    f06bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bcf.html

    Parameters
    ----------
    t : float
        The value t, the tangent of the rotation.

    Returns
    -------
    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dcsg as b_dcsg

    _return_scalars = b_dcsg(
        t,
    )
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's'],
    )(
        _c_py,
        _s_py,
    )
    return _return_data

def drotj(job, x, y, z):
    u"""
    Generate real Jacobi plane rotation.

    ``drotj`` generates a real Jacobi plane rotation.

    For full information please refer to the NAG Library document for
    f06be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bef.html

    Parameters
    ----------
    job : str, length 1
        Specifies the property which determines the precise form of the
        rotation.

        `job` = 'B'

          c >= 1/sqrt(2).

        `job` = 'S'

          0 <= c <= 1/sqrt(2).

        `job` = 'M'

          |a| >= |b|.

    x : float
        The value x, the (1,1) element of the input matrix.

    y : float
        The value y, the (1,2) or (2,1) element of the input matrix.

    z : float
        The value z. the (2,2) element of the input matrix.

    Returns
    -------
    x : float
        The value a.

    y : float
        The value t, the tangent of the rotation.

    z : float
        The value b.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `job`.

            Constraint: `job` = 'B', 'S' or 'M'.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import drotj as b_drotj

    _return_scalars = b_drotj(
        job,
        x,
        y,
        z,
    )
    _x_py = _return_scalars.x
    _y_py = _return_scalars.y
    _z_py = _return_scalars.z
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'z', 'c', 's'],
    )(
        _x_py,
        _y_py,
        _z_py,
        _c_py,
        _s_py,
    )
    return _return_data

def drot2(x, y, z, c, s):
    u"""
    Apply real similarity rotation to 2 by 2 symmetric matrix.

    ``drot2`` applies a real similarity rotation to a 2 by 2 real
    symmetric matrix.

    For full information please refer to the NAG Library document for
    f06bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bhf.html

    Parameters
    ----------
    x : float
        The value x, the (1,1) element of the input matrix.

    y : float
        The value y, the (1,2) or (2,1) element of the input matrix.

    z : float
        The value z, the (2,2) element of the input matrix.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Returns
    -------
    x : float
        The transformed value x.

    y : float
        The transformed value y.

    z : float
        The transformed value z.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import drot2 as b_drot2

    _return_scalars = b_drot2(
        x,
        y,
        z,
        c,
        s,
    )
    _x_py = _return_scalars.x
    _y_py = _return_scalars.y
    _z_py = _return_scalars.z
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'z'],
    )(
        _x_py,
        _y_py,
        _z_py,
    )
    return _return_data

def ddiv(a, b):
    u"""
    Compute quotient of two real scalars, with overflow flag.

    ``ddiv`` computes the quotient of two real scalars.

    For full information please refer to the NAG Library document for
    f06bl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06blf.html

    Parameters
    ----------
    a : float
        The value a.

    b : float
        The value b.

    Returns
    -------
    div : float
        The quotient of the two real scalars.

    fail : bool
        True if a/b would overflow (in which case |q| = flmax) or a = b
        = 0 (in which case q = 0); otherwise False.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import ddiv as b_ddiv

    _return_scalars = b_ddiv(
        a,
        b,
    )
    _div_py = _return_scalars.div
    _fail_py = _return_scalars.fail
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['div', 'fail'],
    )(
        _div_py,
        _fail_py,
    )
    return _return_data

def dnorm(scal, ssq):
    u"""
    Compute Euclidean norm from scaled form.

    ``dnorm`` completes the safe computation of the Euclidean length of
    a vector, following a call to ``dssq`` or ``dzssq``.

    For full information please refer to the NAG Library document for
    f06bm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bmf.html

    Parameters
    ----------
    scal : float
        The scaling factor alpha, returned by ``dssq`` or ``dzssq``.

    ssq : float
        The scaled sum of squares xi, returned by ``dssq`` or ``dzssq``.

    Returns
    -------
    nrm2 : float
        The Euclidean length of the vector.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `scal`.

            Constraint: `scal` >= 0.0.

        (`errno` 2)
            On entry, error in parameter `ssq`.

            Constraint: `ssq` >= 1.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dnorm as b_dnorm

    _return_scalar = b_dnorm(
        scal,
        ssq,
    )
    return _return_scalar

def dpyth(a, b):
    u"""
    Compute square root of (a^2+b^2), real a and b.

    ``dpyth`` returns the value sqrt(a^2+b^2).

    For full information please refer to the NAG Library document for
    f06bn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bnf.html

    Parameters
    ----------
    a : float
        The value a.

    b : float
        The value b.

    Returns
    -------
    p : float
        The value sqrt(a^2+b^2).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dpyth as b_dpyth

    _return_scalar = b_dpyth(
        a,
        b,
    )
    return _return_scalar

def deig2(a, b, c):
    u"""
    Compute eigenvalue of 2 by 2 real symmetric matrix.

    ``deig2`` returns an eigenvalue of a 2 by 2 real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06bp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06bpf.html

    Parameters
    ----------
    a : float
        The value a, the (1,1) element of the input matrix.

    b : float
        The value b, the (1,2) or (2,1) element of the input matrix.

    c : float
        The value c, the (2,2) element of the input matrix.

    Returns
    -------
    ev : float
        The eigenvalue.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import deig2 as b_deig2

    _return_scalar = b_deig2(
        a,
        b,
        c,
    )
    return _return_scalar

def zrotgc(a, b):
    u"""
    Generate complex plane rotation, storing tangent, real cosine.

    ``zrotgc`` generates a complex Givens plane rotation having real
    cosine.

    For full information please refer to the NAG Library document for
    f06ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06caf.html

    Parameters
    ----------
    a : complex
        The value a, the first element of the vector which determines
        the rotation.

    b : complex
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : complex
        The value d.

    b : complex
        The value t, the tangent of the rotation.

    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zrotgc as b_zrotgc

    _return_scalars = b_zrotgc(
        a,
        b,
    )
    _a_py = _return_scalars.a
    _b_py = _return_scalars.b
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return _return_data

def zrotgs(a, b):
    u"""
    Generate complex plane rotation, storing tangent, real sine.

    ``zrotgs`` generates a complex Givens plane rotation having real
    sine.

    For full information please refer to the NAG Library document for
    f06cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06cbf.html

    Parameters
    ----------
    a : complex
        The value a, the first element of the vector which determines
        the rotation.

    b : complex
        The value b, the second element of the vector which determines
        the rotation.

    Returns
    -------
    a : complex
        The value d.

    b : complex
        The value t, the tangent of the rotation.

    c : complex
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zrotgs as b_zrotgs

    _return_scalars = b_zrotgs(
        a,
        b,
    )
    _a_py = _return_scalars.a
    _b_py = _return_scalars.b
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
    )
    return _return_data

def zcsg(t):
    u"""
    Recover cosine and sine from given complex tangent, real cosine.

    ``zcsg`` reconstructs the parameters c (real) and s (complex) of a
    complex plane rotation from the tangent of that rotation.

    For full information please refer to the NAG Library document for
    f06cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ccf.html

    Parameters
    ----------
    t : complex
        The value t, the tangent of the rotation.

    Returns
    -------
    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zcsg as b_zcsg

    _return_scalars = b_zcsg(
        t,
    )
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's'],
    )(
        _c_py,
        _s_py,
    )
    return _return_data

def zcsgs(t):
    u"""
    Recover cosine and sine from given complex tangent, real sine.

    ``zcsgs`` reconstructs the parameters c (complex) and s (real) of a
    complex plane rotation from the tangent of that rotation.

    For full information please refer to the NAG Library document for
    f06cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06cdf.html

    Parameters
    ----------
    t : complex
        The value t, the tangent of the rotation.

    Returns
    -------
    c : complex
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zcsgs as b_zcsgs

    _return_scalars = b_zcsgs(
        t,
    )
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's'],
    )(
        _c_py,
        _s_py,
    )
    return _return_data

def zrot2(x, y, z, c, s):
    u"""
    Apply complex similarity rotation to 2 by 2 Hermitian matrix.

    ``zrot2`` applies a complex similarity rotation having real cosine
    and complex sine to a 2 by 2 complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06ch

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06chf.html

    Parameters
    ----------
    x : complex
        The value x, the (1,1) element of the input matrix.

    y : complex
        The value y, the (1,2) element of the input matrix.

    z : complex
        The value z, the (2,2) element of the input matrix.

    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Returns
    -------
    x : complex
        The transformed value x.

    y : complex
        The transformed value y.

    z : complex
        The transformed value z.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zrot2 as b_zrot2

    _return_scalars = b_zrot2(
        x,
        y,
        z,
        c,
        s,
    )
    _x_py = _return_scalars.x
    _y_py = _return_scalars.y
    _z_py = _return_scalars.z
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'z'],
    )(
        _x_py,
        _y_py,
        _z_py,
    )
    return _return_data

def zdiv(a, b):
    u"""
    Compute quotient of two complex scalars, with overflow flag.

    ``zdiv`` computes the quotient of two complex scalars.

    For full information please refer to the NAG Library document for
    f06cl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06clf.html

    Parameters
    ----------
    a : complex
        The value a.

    b : complex
        The value b.

    Returns
    -------
    sdiv : complex
        The value q.

    fail : bool
        True if a/b would overflow (in which case |Re(q)| = |Im(q)| =
        flmax) or a = b = 0 (in which case q = 0); otherwise False.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdiv as b_zdiv

    _return_scalars = b_zdiv(
        a,
        b,
    )
    _sdiv_py = _return_scalars.sdiv
    _fail_py = _return_scalars.fail
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['sdiv', 'fail'],
    )(
        _sdiv_py,
        _fail_py,
    )
    return _return_data

def iload(n, con):
    u"""
    Broadcast scalar into integer vector.

    ``iload`` broadcasts an integer scalar into an integer vector.

    For full information please refer to the NAG Library document for
    f06db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06dbf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    con : int
        The scalar alpha.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The elements x_i of the vector x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import iload as b_iload

    fun_name = 'naginterfaces.library.blas.iload'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _incx = 1
    _incx_py = _incx
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (1+(_n_py-1)*_incx_py,),
    )
    b_iload(
        _n,
        con,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def icopy(x):
    u"""
    Copy integer vector.

    ``icopy`` copies an integer vector.

    For full information please refer to the NAG Library document for
    f06df

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06dff.html

    Parameters
    ----------
    x : int, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    y : int, ndarray, shape (n)
        The elements y_i of the vector y will be stored in `y` as
        follows.

        If incy > 0, y_i will be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i will be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import icopy as b_icopy

    fun_name = 'naginterfaces.library.blas.icopy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _incy = 1
    _incy_py = _incy
    _y = _b_utils._EngineIntArrayType.empty_ndarray(
        (1+(_n_py-1)*abs(_incy_py),),
    )
    b_icopy(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def ddot(x, y):
    u"""
    Dot product of two real vectors.

    ``ddot`` computes the scalar product of two real vectors.

    For full information please refer to the NAG Library document for
    f06ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06eaf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : float, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    Returns
    -------
    d : float
        The scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import ddot as b_ddot

    fun_name = 'naginterfaces.library.blas.ddot'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    _return_scalar = b_ddot(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    return _return_scalar

def daxpy(alpha, x, y):
    u"""
    Add scalar times real vector to real vector.

    ``daxpy`` adds a scaled real vector to an unscaled real vector.

    For full information please refer to the NAG Library document for
    f06ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ecf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : float, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    Returns
    -------
    y : float, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import daxpy as b_daxpy

    fun_name = 'naginterfaces.library.blas.daxpy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    b_daxpy(
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dscal(alpha, x):
    u"""
    Multiply real vector by scalar.

    ``dscal`` multiplies a real vector by a real scalar.

    For full information please refer to the NAG Library document for
    f06ed

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06edf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, ndarray, shape (n), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*incx], for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

        On exit: the vector alpha x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dscal as b_dscal

    fun_name = 'naginterfaces.library.blas.dscal'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    b_dscal(
        _n,
        alpha,
        x,
        _incx,
    )

def dcopy(x):
    u"""
    Copy real vector.

    ``dcopy`` makes a copy of a real vector.

    For full information please refer to the NAG Library document for
    f06ef

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06eff.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    y : float, ndarray, shape (n)
        The elements y_i of the vector y will be stored in `y` as
        follows.

        If incy > 0, y_i will be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i will be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dcopy as b_dcopy

    fun_name = 'naginterfaces.library.blas.dcopy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _incy = 1
    _incy_py = _incy
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (1+(_n_py-1)*abs(_incy_py),),
    )
    b_dcopy(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dswap(x, y):
    u"""
    Swap two real vectors.

    ``dswap`` interchanges two n-element real vectors x and y.

    For full information please refer to the NAG Library document for
    f06eg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06egf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : float, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    Returns
    -------
    x : float, ndarray, shape (n)
        The original vector y stored in the array elements used to
        supply the original vector x.

    y : float, ndarray, shape (n)
        The original vector x stored in the array elements used to
        supply the original vector y.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dswap as b_dswap

    fun_name = 'naginterfaces.library.blas.dswap'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    b_dswap(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def dnrm2(x):
    u"""
    Compute Euclidean norm of real vector.

    ``dnrm2`` returns the Euclidean norm of the n-element real vector x.

    For full information please refer to the NAG Library document for
    f06ej

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ejf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

    Returns
    -------
    nrm : float
        The Euclidean norm.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dnrm2 as b_dnrm2

    fun_name = 'naginterfaces.library.blas.dnrm2'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_dnrm2(
        _n,
        _x,
        _incx,
    )
    return _return_scalar

def dasum(x):
    u"""
    Sum absolute values of real vector elements.

    ``dasum`` returns the 1-norm of the n-element real vector x.

    For full information please refer to the NAG Library document for
    f06ek

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ekf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

    Returns
    -------
    sumx : float
        The 1-norm of the n-element real vector x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dasum as b_dasum

    fun_name = 'naginterfaces.library.blas.dasum'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_dasum(
        _n,
        _x,
        _incx,
    )
    return _return_scalar

def drot(x, y, c, s):
    u"""
    Apply real plane rotation.

    ``drot`` applies a real plane rotation to two real vectors.

    For full information please refer to the NAG Library document for
    f06ep

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06epf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : float, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Returns
    -------
    x : float, ndarray, shape (n)
        The transformed vector x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    y : float, ndarray, shape (n)
        The transformed vector y stored in the array elements used to
        supply the original vector y.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import drot as b_drot

    fun_name = 'naginterfaces.library.blas.drot'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    b_drot(
        _n,
        _x,
        _incx,
        _y,
        _incy,
        c,
        s,
    )
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def ddoti(x, indx, y):
    u"""
    Dot product of a real sparse and a full vector.

    ``ddoti`` computes the scalar product of a sparse real vector,
    stored in compressed form, with a real vector.

    For full information please refer to the NAG Library document for
    f06er

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06erf.html

    Parameters
    ----------
    x : float, array-like, shape (nz)
        The nonzero elements of the sparse vector x.

    indx : int, array-like, shape (nz)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,nz.

    y : float, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    d : float
        The scalar product of the sparse real vector, x, with the real
        vector y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import ddoti as b_ddoti

    fun_name = 'naginterfaces.library.blas.ddoti'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nz_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_ddoti(
        _nz,
        _x,
        _indx,
        _y,
    )
    return _return_scalar

def daxpyi(a, x, indx, y):
    u"""
    Add scalar times real sparse vector to a full vector.

    ``daxpyi`` adds a scaled sparse real vector, stored in compressed
    form, to an unscaled real vector.

    For full information please refer to the NAG Library document for
    f06et

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06etf.html

    Parameters
    ----------
    a : float
        The scalar alpha.

    x : float, array-like, shape (nz)
        The nonzero elements of the sparse vector x.

    indx : int, array-like, shape (nz)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,nz.

    y : float, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    y : float, ndarray, shape (max(`indx`))
        The updated vector y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import daxpyi as b_daxpyi

    fun_name = 'naginterfaces.library.blas.daxpyi'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nz_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_daxpyi(
        _nz,
        a,
        _x,
        _indx,
        _y,
    )
    _y_py = _y
    return _y_py

def dgthr(y, indx):
    u"""
    Gather real sparse vector.

    ``dgthr`` gathers specified (usually nonzero) elements of a real
    vector y in full storage form into a sparse real vector x in
    compressed form.

    For full information please refer to the NAG Library document for
    f06eu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06euf.html

    Parameters
    ----------
    y : float, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    indx : int, array-like, shape (nz)
        `indx`[i-1] must contain the index `y`[i-1], for i = 1,2,...,nz,
        which is to be gathered into x.

    Returns
    -------
    x : float, ndarray, shape (nz)
        The compressed vector x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dgthr as b_dgthr

    fun_name = 'naginterfaces.library.blas.dgthr'

    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _nz_py = _b_utils._EngineIntArrayType.get_shape(_indx, {'fun_name': fun_name, 'entity': 'indx'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nz_py,),
    )
    b_dgthr(
        _nz,
        _y,
        _x,
        _indx,
    )
    _x_py = _x
    return _x_py

def dgthrz(y, indx):
    u"""
    Gather and set to zero real sparse vector.

    ``dgthrz`` gathers specified (usually nonzero) elements of a real
    vector y in full storage form into a sparse real vector x in
    compressed form.
    The specified elements of y are set to zero.

    For full information please refer to the NAG Library document for
    f06ev

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06evf.html

    Parameters
    ----------
    y : float, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    indx : int, array-like, shape (nz)
        `indx`[i-1] must contain the index `y`[i-1], for i = 1,2,...,nz,
        which is to be gathered into x.

    Returns
    -------
    y : float, ndarray, shape (max(`indx`))
        The elements of y corresponding to indices in `indx` are set to
        zero.

    x : float, ndarray, shape (nz)
        The compressed vector x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `indx`.

            Constraint: the indices must be distinct.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dgthrz as b_dgthrz

    fun_name = 'naginterfaces.library.blas.dgthrz'

    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _nz_py = _b_utils._EngineIntArrayType.get_shape(_indx, {'fun_name': fun_name, 'entity': 'indx'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nz_py,),
    )
    b_dgthrz(
        _nz,
        _y,
        _x,
        _indx,
    )
    _y_py = _y
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['y', 'x'],
    )(
        _y_py,
        _x_py,
    )
    return _return_data

def dsctr(x, indx, y):
    u"""
    Scatter real sparse vector.

    ``dsctr`` scatters the elements of a sparse real vector x stored in
    compressed form, into a real vector y in full storage form.

    For full information please refer to the NAG Library document for
    f06ew

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ewf.html

    Parameters
    ----------
    x : float, array-like, shape (nz)
        The nonzero elements of the sparse vector x.

    indx : int, array-like, shape (nz)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,nz.

    y : float, array-like, shape (max(`indx`))
        The vector y.

    Returns
    -------
    y : float, ndarray, shape (max(`indx`))
        The vector y, with the elements corresponding to indices in
        `indx` altered.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dsctr as b_dsctr

    fun_name = 'naginterfaces.library.blas.dsctr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nz_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dsctr(
        _nz,
        _x,
        _indx,
        _y,
    )
    _y_py = _y
    return _y_py

def droti(x, indx, y, c, s):
    u"""
    Apply plane rotation to a real sparse and a full vector.

    ``droti`` applies a real plane rotation to a sparse real vector and
    a real vector.

    For full information please refer to the NAG Library document for
    f06ex

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06exf.html

    Parameters
    ----------
    x : float, array-like, shape (nz)
        The nonzero elements of the sparse vector x.

    indx : int, array-like, shape (nz)
        `indx`[i-1] must contain the index of `x`[i-1] in the sparse
        vector x, for i = 1,2,...,nz.

    y : float, array-like, shape (max(`indx`))
        The vector y. Only the elements corresponding to indices in
        `indx` are referenced.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Returns
    -------
    x : float, ndarray, shape (nz)
        The transformed vector x.

    y : float, ndarray, shape (max(`indx`))
        The transformed vector y. Only elements corresponding to indices
        in `indx` are altered.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import droti as b_droti

    fun_name = 'naginterfaces.library.blas.droti'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nz_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_droti(
        _nz,
        _x,
        _indx,
        _y,
        c,
        s,
    )
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def dvcos(j, tolx, x, toly, y):
    u"""
    Compute cosine of angle between two real vectors.

    ``dvcos`` computes the cosine of the angle between two real vectors.

    For full information please refer to the NAG Library document for
    f06fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06faf.html

    Parameters
    ----------
    j : int
        If the vector y is supplied in `y`, `j` should be set to 0.
        Otherwise, `j` specifies the index j of the unit vector e_j to
        be used as y.

    tolx : float
        The value tolx, used to determine whether ||x||_2 is effectively
        zero.

        If `tolx` is negative, the value zero is used.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    toly : float
        The value toly, used to determine whether ||y||_2 is effectively
        zero.

        If `toly` is negative, the value zero is used.

    y : float, array-like, shape (n)
        If 1 <= `j` <= n, `y` is not referenced. Otherwise, `y` holds
        the vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    Returns
    -------
    vcos : float
        The cosine of the angle between x and y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dvcos as b_dvcos

    fun_name = 'naginterfaces.library.blas.dvcos'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    _return_scalar = b_dvcos(
        _n,
        j,
        tolx,
        _x,
        _incx,
        toly,
        _y,
        _incy,
    )
    return _return_scalar

def dload(n, con):
    u"""
    Broadcast scalar into real vector.

    ``dload`` broadcasts a real scalar into a real vector.

    For full information please refer to the NAG Library document for
    f06fb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fbf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    con : float
        The scalar alpha.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The elements x_i of the vector x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dload as b_dload

    fun_name = 'naginterfaces.library.blas.dload'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _incx = 1
    _incx_py = _incx
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (1+(_n_py-1)*_incx_py,),
    )
    b_dload(
        _n,
        con,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def ddscl(d, x):
    u"""
    Multiply real vector by diagonal matrix.

    ``ddscl`` multiplies a real vector by a real diagonal matrix.

    For full information please refer to the NAG Library document for
    f06fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fcf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        The vector d.

        If incd > 0, d_i must be stored in `d`[(i-1)*incd], for i =
        1,2,...,n.

        If incd < 0, d_i must be stored in `d`[ 1-(n-i)*incd -1], for i
        = 1,2,...,n.

    x : float, array-like, shape (n)
        The array `x` must contain the n-element vector x.

        If incx > 0, x_i must be stored in `x`[ 1+(i-1)*incx -1], for i
        = 1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

        Intermediate elements of `x` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import ddscl as b_ddscl

    fun_name = 'naginterfaces.library.blas.ddscl'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incd = 1
    _incd_py = _incd
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_ddscl(
        _n,
        _d,
        _incd,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def axpzy(alpha, x):
    u"""
    Multiply real vector by scalar, preserving input vector.

    ``axpzy`` multiplies a real vector by a scalar, preserving the input
    vector.

    For full information please refer to the NAG Library document for
    f06fd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fdf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    y : float, ndarray, shape (n)
        The elements y_i of the vector y will be stored in `y` as
        follows.

        If incy > 0, y_i will be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i will be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import axpzy as b_axpzy

    fun_name = 'naginterfaces.library.blas.axpzy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _incy = 1
    _incy_py = _incy
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (1+(_n_py-1)*abs(_incy_py),),
    )
    b_axpzy(
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def drscl(alpha, x):
    u"""
    Multiply real vector by reciprocal of scalar.

    ``drscl`` multiplies a real vector by the reciprocal of a scalar.

    For full information please refer to the NAG Library document for
    f06fe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fef.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x, stored in the same array elements used to
        supply the original vector.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `alpha`.

            Constraint: `alpha` != 0.0.

        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.
    """

    from ..base.blas import drscl as b_drscl

    fun_name = 'naginterfaces.library.blas.drscl'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    b_drscl(
        _n,
        alpha,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dnegv(x):
    u"""
    Negate real vector.

    ``dnegv`` negates a real vector.

    For full information please refer to the NAG Library document for
    f06fg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fgf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    Returns
    -------
    x : float, ndarray, shape (n)
        The vector -x stored in the array elements used to supply the
        original vector x.

        Intermediate elements of `x` are unchanged.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dnegv as b_dnegv

    fun_name = 'naginterfaces.library.blas.dnegv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    b_dnegv(
        _n,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dssq(x, scal, sumsq):
    u"""
    Update Euclidean norm of real vector in scaled form.

    ``dssq`` updates the Euclidean norm of real vector in scaled form.

    For full information please refer to the NAG Library document for
    f06fj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fjf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

    scal : float
        The scaling factor alpha. On the first call to ``dssq`` `scal` =
        0.0.

    sumsq : float
        The scaled sum of squares xi. On the first call to ``dssq``
        `sumsq` = 1.0.

    Returns
    -------
    scal : float
        The updated scaling factor alpha~ = max_i(alpha,|x_i|).

    sumsq : float
        The updated scaled sum of squares xi~, satisfying: 1 <= xi~ <=
        xi+n.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

        (`errno` 4)
            On entry, error in parameter `scal`.

            Constraint: `scal` >= 0.0.

        (`errno` 5)
            On entry, error in parameter `sumsq`.

            Constraint: `sumsq` >= 1.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dssq as b_dssq

    fun_name = 'naginterfaces.library.blas.dssq'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalars = b_dssq(
        _n,
        _x,
        _incx,
        scal,
        sumsq,
    )
    _scal_py = _return_scalars.scal
    _sumsq_py = _return_scalars.sumsq
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['scal', 'sumsq'],
    )(
        _scal_py,
        _sumsq_py,
    )
    return _return_data

def dnrm2w(w, x):
    u"""
    Compute weighted Euclidean norm of real vector.

    ``dnrm2w`` computes the weighted Euclidean norm of a real vector.

    For full information please refer to the NAG Library document for
    f06fk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fkf.html

    Parameters
    ----------
    w : float, array-like, shape (n)
        w, the vector of weights.

        If incw > 0, w_i must be stored in `w`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incw < 0, w_i must be stored in `w`[1-(n-i)*incw-1], for i =
        1,2,...,n.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    nrm : float
        The weighted Euclidean norm of the real vector, x.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter w.

            Constraint: the elements of the array must not be negative.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dnrm2w as b_dnrm2w

    fun_name = 'naginterfaces.library.blas.dnrm2w'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incw = 1
    _incw_py = _incw
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_dnrm2w(
        _n,
        _w,
        _incw,
        _x,
        _incx,
    )
    return _return_scalar

def darang(x):
    u"""
    Elements of real vector with largest and smallest absolute value.

    ``darang`` returns the absolutely largest and absolutely smallest
    values from a real vector.

    For full information please refer to the NAG Library document for
    f06fl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06flf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

    Returns
    -------
    xmax : float
        The value x_max = max_i|x_i|.

    xmin : float
        The value x_min = min_i|x_i|.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import darang as b_darang

    fun_name = 'naginterfaces.library.blas.darang'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalars = b_darang(
        _n,
        _x,
        _incx,
    )
    _xmax_py = _return_scalars.xmax
    _xmin_py = _return_scalars.xmin
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xmax', 'xmin'],
    )(
        _xmax_py,
        _xmin_py,
    )
    return _return_data

def drots(x, y, c, s):
    u"""
    Apply real symmetric plane rotation to two vectors.

    ``drots`` applies a real symmetric plane rotation to two real
    vectors.

    For full information please refer to the NAG Library document for
    f06fp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fpf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The original vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

        Intermediate elements of `x` are not referenced.

    y : float, array-like, shape (n)
        The original vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Returns
    -------
    x : float, ndarray, shape (n)
        The transformed vector x stored in the same elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    y : float, ndarray, shape (n)
        The transformed vector y stored in the same elements used to
        supply the original vector y.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import drots as b_drots

    fun_name = 'naginterfaces.library.blas.drots'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    b_drots(
        _n,
        _x,
        _incx,
        _y,
        _incy,
        c,
        s,
    )
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def dsrotg(pivot, direct, alpha, x):
    u"""
    Generate sequence of real plane rotations.

    ``dsrotg`` generates a sequence of real plane rotations.

    For full information please refer to the NAG Library document for
    f06fq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fqf.html

    Parameters
    ----------
    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'F' (fixed pivot)

          P_k rotates the (1,k+1) plane if `direct` = 'F', or the
          (k,n+1) plane if `direct` = 'B'.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_n...P_2 P_1.

        `direct` = 'B' (backward sequence)

          P = P_1 P_2...P_n.

    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    Returns
    -------
    alpha : float
        The scalar beta.

    x : float, ndarray, shape (n)
        The referenced elements are overwritten by details of the
        sequence of plane rotations.

    c : float, ndarray, shape (n)
        The values c_k, the cosines of the rotations.

    s : float, ndarray, shape (n)
        The values s_k, the sines of the rotations.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V' or 'F'.

        (`errno` 2)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 6)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dsrotg as b_dsrotg

    fun_name = 'naginterfaces.library.blas.dsrotg'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalar = b_dsrotg(
        pivot,
        direct,
        _n,
        alpha,
        _x,
        _incx,
        _c,
        _s,
    )
    _alpha_py = _return_scalar
    _x_py = _x
    _c_py = _c
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'x', 'c', 's'],
    )(
        _alpha_py,
        _x_py,
        _c_py,
        _s_py,
    )
    return _return_data

def dnhousg(alpha, x, tol):
    u"""
    Generate real elementary reflection, NAG style.

    ``dnhousg`` generates a real elementary reflection in the NAG (as
    opposed to LINPACK) style.

    For full information please refer to the NAG Library document for
    f06fr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06frf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    tol : float
        The value tol.

    Returns
    -------
    alpha : float
        The scalar beta.

    x : float, ndarray, shape (n)
        The referenced elements are overwritten by details of the real
        elementary reflection.

    zeta : float
        The scalar zeta.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dnhousg as b_dnhousg

    fun_name = 'naginterfaces.library.blas.dnhousg'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalars = b_dnhousg(
        _n,
        alpha,
        _x,
        _incx,
        tol,
    )
    _alpha_py = _return_scalars.alpha
    _x_py = _x
    _zeta_py = _return_scalars.zeta
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'x', 'zeta'],
    )(
        _alpha_py,
        _x_py,
        _zeta_py,
    )
    return _return_data

def dlhousg(alpha, x, tol):
    u"""
    Generate real elementary reflection, LINPACK style.

    ``dlhousg`` generates a real elementary reflection in the LINPACK
    (as opposed to NAG) style.

    For full information please refer to the NAG Library document for
    f06fs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fsf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    tol : float
        The value tol.

        If `tol` is not in the range (0,1), the value 0 is used for tol.

    Returns
    -------
    alpha : float
        The scalar beta.

    x : float, ndarray, shape (n)
        The referenced elements are overwritten by details of the real
        elementary reflection.

    z1 : float
        The scalar zeta.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlhousg as b_dlhousg

    fun_name = 'naginterfaces.library.blas.dlhousg'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalars = b_dlhousg(
        _n,
        alpha,
        _x,
        _incx,
        tol,
    )
    _alpha_py = _return_scalars.alpha
    _x_py = _x
    _z1_py = _return_scalars.z1
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'x', 'z1'],
    )(
        _alpha_py,
        _x_py,
        _z1_py,
    )
    return _return_data

def dnhous(delta, y, zeta, z):
    u"""
    Apply real elementary reflection, NAG style.

    ``dnhous`` applies a NAG (as opposed to LINPACK) style real
    elementary reflection to a real vector.

    For full information please refer to the NAG Library document for
    f06ft

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ftf.html

    Parameters
    ----------
    delta : float
        The original scalar delta.

    y : float, array-like, shape (n)
        The original vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1) *incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i) *incy-1], for i =
        1,2,...,n.

    zeta : float
        The scalar zeta, as returned by ``dnhousg``.

        If zeta = 0, P is assumed to be the unit matrix and the
        transformation is skipped.

    z : float, array-like, shape (n)
        The vector z, as returned by ``dnhousg``.

        If incz > 0, z_i must be stored in `z`[1+(i-1)*incz-1], for i =
        1,2,...,n.

        If incz < 0, z_i must be stored in `z`[1-(n-i)*incz-1], for i =
        1,2,...,n.

    Returns
    -------
    delta : float
        The transformed scalar delta.

    y : float, ndarray, shape (n)
        The transformed stored in the same array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 5)
            On entry, error in parameter `zeta`.

            Constraint: `zeta` = 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dnhous as b_dnhous

    fun_name = 'naginterfaces.library.blas.dnhous'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incy = 1
    _incy_py = _incy
    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incz = 1
    _incz_py = _incz
    _return_scalar = b_dnhous(
        _n,
        delta,
        _y,
        _incy,
        zeta,
        _z,
        _incz,
    )
    _delta_py = _return_scalar
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['delta', 'y'],
    )(
        _delta_py,
        _y_py,
    )
    return _return_data

def dlhous(z, z1, alpha, x):
    u"""
    Apply real elementary reflection, LINPACK style.

    ``dlhous`` applies a LINPACK (as opposed to NAG) style real
    elementary reflection to a real vector.

    For full information please refer to the NAG Library document for
    f06fu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06fuf.html

    Parameters
    ----------
    z : float, array-like, shape (n)
        The vector z, as returned by ``dlhousg``.

        If incz > 0, z_i must be stored in `z`[1+(i-1)*incz-1], for i =
        1,2,...,n.

        If incz < 0, z_i must be stored in `z`[1-(n-i)*incz-1], for i =
        1,2,...,n.

    z1 : float
        The scalar zeta, as returned by ``dlhousg``.

        If zeta = 0, P is assumed to be the unit matrix and the
        transformation is skipped.

    alpha : float
        The original scalar alpha.

    x : float, array-like, shape (n)
        The original vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    alpha : float
        The transformed scalar alpha.

    x : float, ndarray, shape (n)
        The transformed vector x stored in the same array elements used
        to supply the original vector x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlhous as b_dlhous

    fun_name = 'naginterfaces.library.blas.dlhous'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incz = 1
    _incz_py = _incz
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_dlhous(
        _n,
        _z,
        _incz,
        z1,
        alpha,
        _x,
        _incx,
    )
    _alpha_py = _return_scalar
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'x'],
    )(
        _alpha_py,
        _x_py,
    )
    return _return_data

def zdotu(x, y):
    u"""
    Dot product of two complex vectors, unconjugated.

    ``zdotu`` computes the scalar product of two complex vectors.

    For full information please refer to the NAG Library document for
    f06ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gaf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    Returns
    -------
    zdu : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdotu as b_zdotu

    fun_name = 'naginterfaces.library.blas.zdotu'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _incy = 1
    _incy_py = _incy
    _return_scalar = b_zdotu(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    return _return_scalar

def zdotc(x, y):
    u"""
    Dot product of two complex vectors, conjugated.

    ``zdotc`` computes the scalar product of two complex vectors.

    For full information please refer to the NAG Library document for
    f06gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gbf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    Returns
    -------
    zdc : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdotc as b_zdotc

    fun_name = 'naginterfaces.library.blas.zdotc'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _incy = 1
    _incy_py = _incy
    _return_scalar = b_zdotc(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    return _return_scalar

def zaxpy(alpha, x, y):
    u"""
    Add scalar times complex vector to complex vector.

    ``zaxpy`` adds a scaled complex vector to an unscaled complex
    vector.

    For full information please refer to the NAG Library document for
    f06gc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gcf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The updated vector y.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zaxpy as b_zaxpy

    fun_name = 'naginterfaces.library.blas.zaxpy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _incy = 1
    _incy_py = _incy
    b_zaxpy(
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zscal(alpha, x):
    u"""
    Multiply complex vector by complex scalar.

    ``zscal`` scales a complex vector by a complex scalar.

    For full information please refer to the NAG Library document for
    f06gd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gdf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, ndarray, shape (n), modified in place
        On entry: the n-element vector x. x_i must be stored in
        `x`[1+(i-1)*incx], for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

        On exit: the vector alpha x stored in the array elements used to
        supply the original vector x(y).

        Intermediate elements of `x` are unchanged.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zscal as b_zscal

    fun_name = 'naginterfaces.library.blas.zscal'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    b_zscal(
        _n,
        alpha,
        x,
        _incx,
    )

def zcopy(x):
    u"""
    Copy complex vector.

    ``zcopy`` copies a complex vector to a complex vector.

    For full information please refer to the NAG Library document for
    f06gf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gff.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The elements y_i of the vector y will be stored in `y` as
        follows.

        If incy > 0, y_i will be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i will be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zcopy as b_zcopy

    fun_name = 'naginterfaces.library.blas.zcopy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _incy = 1
    _incy_py = _incy
    _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (1+(_n_py-1)*abs(_incy_py),),
    )
    b_zcopy(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zswap(x, y):
    u"""
    Swap two complex vectors.

    ``zswap`` interchanges two n-element complex vectors x and y.

    For full information please refer to the NAG Library document for
    f06gg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ggf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The original vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx], for i =
        1,2,...,n.

        If incx <= 0, x_i must be stored in `x`[1+(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The original vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy <= 0, y_i must be stored in `y`[1+(n-i)*incy], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The original vector y stored in the array elements used to store
        the original vector x. Intermediate elements of `x` are
        unchanged.

    y : complex, ndarray, shape (n)
        The original vector x stored in the array elements used to store
        the original vector y. Intermediate elements of `y` are
        unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zswap as b_zswap

    fun_name = 'naginterfaces.library.blas.zswap'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _incy = 1
    _incy_py = _incy
    b_zswap(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def zdotui(x, indx, y):
    u"""
    Dot product of a complex sparse and a full vector, unconjugated.

    ``zdotui`` computes the scalar product of an unconjugated sparse
    complex vector with a complex vector.

    For full information please refer to the NAG Library document for
    f06gr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06grf.html

    Parameters
    ----------
    x : complex, array-like, shape (nz)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, array-like, shape (nz)
        `indx` must contain the set of indices J.

    y : complex, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    zdu : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdotui as b_zdotui

    fun_name = 'naginterfaces.library.blas.zdotui'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _nz_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _return_scalar = b_zdotui(
        _nz,
        _x,
        _indx,
        _y,
    )
    return _return_scalar

def zdotci(x, indx, y):
    u"""
    Dot product of a complex sparse and a full vector, conjugated.

    ``zdotci`` computes the scalar product of a conjugated sparse
    complex vector with a complex vector.

    For full information please refer to the NAG Library document for
    f06gs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gsf.html

    Parameters
    ----------
    x : complex, array-like, shape (nz)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, array-like, shape (nz)
        `indx` must contain the set of indices J.

    y : complex, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    zdc : complex
        The value of the scalar product.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdotci as b_zdotci

    fun_name = 'naginterfaces.library.blas.zdotci'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _nz_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _return_scalar = b_zdotci(
        _nz,
        _x,
        _indx,
        _y,
    )
    return _return_scalar

def zaxpyi(a, x, indx, y):
    u"""
    Add scalar times complex sparse vector to a full vector.

    ``zaxpyi`` adds a scaled sparse complex vector to an unscaled
    complex vector.

    For full information please refer to the NAG Library document for
    f06gt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gtf.html

    Parameters
    ----------
    a : complex
        The scalar alpha.

    x : complex, array-like, shape (nz)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, array-like, shape (nz)
        The indices of the elements in the compressed vector x.

    y : complex, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    Returns
    -------
    y : complex, ndarray, shape (max(`indx`))
        The updated vector y.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter `indx`.

            Constraint: the indices must be distinct.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zaxpyi as b_zaxpyi

    fun_name = 'naginterfaces.library.blas.zaxpyi'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _nz_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    b_zaxpyi(
        _nz,
        a,
        _x,
        _indx,
        _y,
    )
    _y_py = _y
    return _y_py

def zgthr(y, indx):
    u"""
    Gather complex sparse vector.

    ``zgthr`` gathers specified (usually nonzero) elements of a complex
    vector y in full storage form into a sparse complex vector x in
    compressed form.

    For full information please refer to the NAG Library document for
    f06gu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06guf.html

    Parameters
    ----------
    y : complex, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    indx : int, array-like, shape (nz)
        `indx` must contain the set of indices J.

    Returns
    -------
    x : complex, ndarray, shape (nz)
        The compressed vector x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zgthr as b_zgthr

    fun_name = 'naginterfaces.library.blas.zgthr'

    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _nz_py = _b_utils._EngineIntArrayType.get_shape(_indx, {'fun_name': fun_name, 'entity': 'indx'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_nz_py,),
    )
    b_zgthr(
        _nz,
        _y,
        _x,
        _indx,
    )
    _x_py = _x
    return _x_py

def zgthrz(y, indx):
    u"""
    Gather and set to zero complex sparse vector.

    ``zgthrz`` gathers specified (usually nonzero) elements of a complex
    vector y in full storage form into a sparse complex vector x in
    compressed form.
    The specified elements of y are set to zero.

    For full information please refer to the NAG Library document for
    f06gv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gvf.html

    Parameters
    ----------
    y : complex, array-like, shape (max(`indx`))
        The vector y. Only elements corresponding to indices in `indx`
        are accessed.

    indx : int, array-like, shape (nz)
        `indx` must contain the set of indices J.

    Returns
    -------
    y : complex, ndarray, shape (max(`indx`))
        The elements of y corresponding to indices in `indx` are set to
        zero.

    x : complex, ndarray, shape (nz)
        The compressed vector x.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zgthrz as b_zgthrz

    fun_name = 'naginterfaces.library.blas.zgthrz'

    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _nz_py = _b_utils._EngineIntArrayType.get_shape(_indx, {'fun_name': fun_name, 'entity': 'indx'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_nz_py,),
    )
    b_zgthrz(
        _nz,
        _y,
        _x,
        _indx,
    )
    _y_py = _y
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['y', 'x'],
    )(
        _y_py,
        _x_py,
    )
    return _return_data

def zsctr(x, indx, y):
    u"""
    Scatter complex sparse vector.

    ``zsctr`` scatters the elements of a sparse complex vector x stored
    in compressed form, into a complex vector y in full storage form.

    For full information please refer to the NAG Library document for
    f06gw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06gwf.html

    Parameters
    ----------
    x : complex, array-like, shape (nz)
        The compressed vector x. `x` contains x_i for i in J.

    indx : int, array-like, shape (nz)
        The indices of the elements in the compressed vector x.

    y : complex, array-like, shape (max(`indx`))
        The vector y.

    Returns
    -------
    y : complex, ndarray, shape (max(`indx`))
        The vector y, with the elements corresponding to indices in
        `indx` altered.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter `indx`.

            Constraint: the indices must be distinct.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zsctr as b_zsctr

    fun_name = 'naginterfaces.library.blas.zsctr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _nz_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=False,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    b_zsctr(
        _nz,
        _x,
        _indx,
        _y,
    )
    _y_py = _y
    return _y_py

def zload(n, con):
    u"""
    Broadcast scalar into complex vector.

    ``zload`` broadcasts a complex scalar into a complex vector.

    For full information please refer to the NAG Library document for
    f06hb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hbf.html

    Parameters
    ----------
    n : int
        n, the number of elements in x.

    con : complex
        The scalar alpha.

    Returns
    -------
    x : complex, ndarray, shape (`n`)
        The elements x_i of the vector x.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zload as b_zload

    fun_name = 'naginterfaces.library.blas.zload'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _incx = 1
    _incx_py = _incx
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (1+(_n_py-1)*_incx_py,),
    )
    b_zload(
        _n,
        con,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def zdscl(d, x):
    u"""
    Multiply complex vector by complex diagonal matrix.

    ``zdscl`` multiplies a complex vector by a complex diagonal matrix.

    For full information please refer to the NAG Library document for
    f06hc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hcf.html

    Parameters
    ----------
    d : complex, array-like, shape (n)
        The vector d.

        If incd > 0, d_i must be stored in `d`[(i-1)*incd], for i =
        1,2,...,n.

        If incd < 0, d_i must be stored in `d`[ 1-(n-i)*incd -1], for i
        = 1,2,...,n.

    x : complex, ndarray, shape (n), modified in place
        On entry: the array `x` must contain the n-element vector x.

        If incx > 0, x_i must be stored in `x`[ 1+(i-1)*incx -1], for i
        = 1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

        On exit: the updated vector x stored in the array elements used
        to supply the original vector x.

        Intermediate elements of `x` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdscl as b_zdscl

    fun_name = 'naginterfaces.library.blas.zdscl'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incd = 1
    _incd_py = _incd
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _incx = 1
    _incx_py = _incx
    b_zdscl(
        _n,
        _d,
        _incd,
        x,
        _incx,
    )

def zaxpzy(alpha, x):
    u"""
    Multiply complex vector by complex scalar, preserving input vector.

    ``zaxpzy`` multiplies a complex vector by a scalar, preserving the
    input vector.

    For full information please refer to the NAG Library document for
    f06hd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hdf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The elements y_i of the vector y will be stored in `y` as
        follows.

        If incy > 0, y_i will be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i will be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zaxpzy as b_zaxpzy

    fun_name = 'naginterfaces.library.blas.zaxpzy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _incy = 1
    _incy_py = _incy
    _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (1+(_n_py-1)*abs(_incy_py),),
    )
    b_zaxpzy(
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def znegv(x):
    u"""
    Negate complex vector.

    ``znegv`` negates a complex vector.

    For full information please refer to the NAG Library document for
    f06hg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hgf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The vector -x stored in the array elements used to supply the
        original vector x.

        Intermediate elements of `x` are unchanged.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import znegv as b_znegv

    fun_name = 'naginterfaces.library.blas.znegv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    b_znegv(
        _n,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def zrot(cx, cy, c, s):
    u"""
    Apply plane rotation with real cosine and complex sine.

    ``zrot`` applies a plane rotation with a real cosine and complex
    sine to two complex vectors.

    For full information please refer to the NAG Library document for
    f06hm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hmf.html

    Parameters
    ----------
    cx : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `cx`[1-(n-i)*incx], for i =
        1,2,...,n.

    cy : complex, array-like, shape (n)
        The n-element vector y.

        If incy < 0, y_i must be stored in `cy`[1-(n-i)*incy], for i =
        1,2,...,n.

    c : float
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Returns
    -------
    cx : complex, ndarray, shape (n)
        The transformed vector x stored in the array elements used to
        supply the original vector x.

    cy : complex, ndarray, shape (n)
        The transformed vector y.

        Intermediate elements of `cy` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zrot as b_zrot

    fun_name = 'naginterfaces.library.blas.zrot'

    _cx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cx'},
        cx,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_cx, {'fun_name': fun_name, 'entity': 'cx'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _cy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cy'},
        cy,
        copy=True,
    )
    _incy = 1
    _incy_py = _incy
    b_zrot(
        _n,
        _cx,
        _incx,
        _cy,
        _incy,
        c,
        s,
    )
    _cx_py = _cx
    _cy_py = _cy
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['cx', 'cy'],
    )(
        _cx_py,
        _cy_py,
    )
    return _return_data

def zcrot(x, y, c, s):
    u"""
    Apply complex plane rotation.

    ``zcrot`` applies a complex plane rotation to two complex vectors.

    For full information please refer to the NAG Library document for
    f06hp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hpf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    c : complex
        The value c, the cosine of the rotation.

    s : complex
        The value s, the sine of the rotation.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The transformed vector x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    y : complex, ndarray, shape (n)
        The transformed vector y.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zcrot as b_zcrot

    fun_name = 'naginterfaces.library.blas.zcrot'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _incy = 1
    _incy_py = _incy
    b_zcrot(
        _n,
        _x,
        _incx,
        _y,
        _incy,
        c,
        s,
    )
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def zsrotg(pivot, direct, alpha, x):
    u"""
    Generate sequence of complex plane rotations.

    ``zsrotg`` generates a sequence of complex plane rotations.

    For full information please refer to the NAG Library document for
    f06hq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hqf.html

    Parameters
    ----------
    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'F' (fixed pivot)

          P_k rotates the (1,k+1) plane if `direct` = 'F', or the
          (k,n+1) plane if `direct` = 'B'.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_n...P_2 P_1.

        `direct` = 'B' (backward sequence)

          P = P_1 P_2...P_n.

    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    Returns
    -------
    alpha : complex
        The scalar beta.

    x : complex, ndarray, shape (n)
        The referenced elements are overwritten by details of the plane
        rotations.

    c : float, ndarray, shape (n)
        The values c_k, the cosines of the rotations.

    s : complex, ndarray, shape (n)
        The values s_k, the sines of the rotations.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V' or 'F'.

        (`errno` 2)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 6)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zsrotg as b_zsrotg

    fun_name = 'naginterfaces.library.blas.zsrotg'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _s = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalar = b_zsrotg(
        pivot,
        direct,
        _n,
        alpha,
        _x,
        _incx,
        _c,
        _s,
    )
    _alpha_py = _return_scalar
    _x_py = _x
    _c_py = _c
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'x', 'c', 's'],
    )(
        _alpha_py,
        _x_py,
        _c_py,
        _s_py,
    )
    return _return_data

def zhousg(alpha, x, tol):
    u"""
    Generate complex elementary reflection.

    ``zhousg`` generates a complex elementary reflection.

    For full information please refer to the NAG Library document for
    f06hr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06hrf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    tol : float
        The value tol.

    Returns
    -------
    alpha : complex
        The scalar beta.

    x : complex, ndarray, shape (n)
        The referenced elements are overwritten by details of the
        complex elementary reflection.

    theta : complex
        The scalar theta.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zhousg as b_zhousg

    fun_name = 'naginterfaces.library.blas.zhousg'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalars = b_zhousg(
        _n,
        alpha,
        _x,
        _incx,
        tol,
    )
    _alpha_py = _return_scalars.alpha
    _x_py = _x
    _theta_py = _return_scalars.theta
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['alpha', 'x', 'theta'],
    )(
        _alpha_py,
        _x_py,
        _theta_py,
    )
    return _return_data

def zhous(delta, y, theta, z):
    u"""
    Apply complex elementary reflection.

    ``zhous`` applies a complex elementary reflection to a complex
    vector.

    For full information please refer to the NAG Library document for
    f06ht

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06htf.html

    Parameters
    ----------
    delta : complex
        The original scalar delta.

    y : complex, array-like, shape (n)
        The original vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1) *incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i) *incy-1], for i =
        1,2,...,n.

    theta : complex
        The value theta, as returned by ``zhousg``.

        If theta = 0, P is assumed to be the unit matrix and the
        transformation is skipped.

    z : complex, array-like, shape (n)
        The vector z, as returned by ``zhousg``.

        If incz > 0, z_i must be stored in `z`[1+(i-1)*incz-1], for i =
        1,2,...,n.

        If incz < 0, z_i must be stored in `z`[1-(n-i)*incz-1], for i =
        1,2,...,n.

    Returns
    -------
    delta : complex
        The transformed scalar delta.

    y : complex, ndarray, shape (n)
        The transformed stored in the same array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 5)
            On entry, error in parameter `theta`.

            Constraint: Re(`theta`) <= 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zhous as b_zhous

    fun_name = 'naginterfaces.library.blas.zhous'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incy = 1
    _incy_py = _incy
    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
    )
    _incz = 1
    _incz_py = _incz
    _return_scalar = b_zhous(
        _n,
        delta,
        _y,
        _incy,
        theta,
        _z,
        _incz,
    )
    _delta_py = _return_scalar
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['delta', 'y'],
    )(
        _delta_py,
        _y_py,
    )
    return _return_data

def zdscal(alpha, x):
    u"""
    Multiply complex vector by real scalar.

    ``zdscal`` scales a complex vector by a real scalar.

    For full information please refer to the NAG Library document for
    f06jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jdf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The vector alpha x stored in the array elements used to supply
        the original vector x.

        Intermediate elements of `x` are unchanged.

    Raises
    ------
    NagValueError
        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdscal as b_zdscal

    fun_name = 'naginterfaces.library.blas.zdscal'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    b_zdscal(
        _n,
        alpha,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dznrm2(x):
    u"""
    Compute Euclidean norm of complex vector.

    ``dznrm2`` computes the Euclidean norm of a complex vector.

    For full information please refer to the NAG Library document for
    f06jj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jjf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

    Returns
    -------
    nrm : float
        The Euclidean norm of the complex vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dznrm2 as b_dznrm2

    fun_name = 'naginterfaces.library.blas.dznrm2'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_dznrm2(
        _n,
        _x,
        _incx,
    )
    return _return_scalar

def dzasum(x):
    u"""
    Sum absolute values of complex vector elements.

    ``dzasum`` returns the sum of the absolute values of the real and
    imaginary parts of the elements in a complex vector.

    For full information please refer to the NAG Library document for
    f06jk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jkf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

    Returns
    -------
    sumx : float
        The sum of the absolute values of the real and imaginary parts
        of the elements in the complex vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dzasum as b_dzasum

    fun_name = 'naginterfaces.library.blas.dzasum'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_dzasum(
        _n,
        _x,
        _incx,
    )
    return _return_scalar

def idamax(x):
    u"""
    Index, real vector element with largest absolute value.

    ``idamax`` computes the index of the absolutely largest component of
    a real vector.

    For full information please refer to the NAG Library document for
    f06jl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jlf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

    Returns
    -------
    imax : int
        The index of the absolutely largest component of the real
        vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import idamax as b_idamax

    fun_name = 'naginterfaces.library.blas.idamax'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_idamax(
        _n,
        _x,
        _incx,
    )
    return _return_scalar

def izamax(x):
    u"""
    Index, complex vector element with largest absolute value.

    ``izamax`` computes the index of the absolutely largest component of
    a complex vector.

    For full information please refer to the NAG Library document for
    f06jm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06jmf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

    Returns
    -------
    imax : int
        The index of the absolutely largest component of a complex
        vector, x.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import izamax as b_izamax

    fun_name = 'naginterfaces.library.blas.izamax'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_izamax(
        _n,
        _x,
        _incx,
    )
    return _return_scalar

def zddscl(d, x):
    u"""
    Multiply complex vector by real diagonal matrix.

    ``zddscl`` multiplies a complex vector by a real diagonal matrix.

    For full information please refer to the NAG Library document for
    f06kc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kcf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        The vector d.

        If incd > 0, d_i must be stored in `d`[(i-1)*incd], for i =
        1,2,...,n.

        If incd < 0, d_i must be stored in `d`[ 1-(n-i)*incd -1], for i
        = 1,2,...,n.

    x : complex, array-like, shape (n)
        The array `x` must contain the n-element vector x.

        If incx > 0, x_i must be stored in `x`[ 1+(i-1)*incx -1], for i
        = 1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

        Intermediate elements of `x` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zddscl as b_zddscl

    fun_name = 'naginterfaces.library.blas.zddscl'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incd = 1
    _incd_py = _incd
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _incx = 1
    _incx_py = _incx
    b_zddscl(
        _n,
        _d,
        _incd,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def zdaxpzy(alpha, x):
    u"""
    Multiply complex vector by real scalar, preserving input vector.

    ``zdaxpzy`` multiplies a complex vector by a real scalar, preserving
    the input vector.

    For full information please refer to the NAG Library document for
    f06kd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kdf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The elements y_i of the vector y will be stored in `y` as
        follows.

        If incy > 0, y_i will be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i will be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdaxpzy as b_zdaxpzy

    fun_name = 'naginterfaces.library.blas.zdaxpzy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _incy = 1
    _incy_py = _incy
    _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (1+(_n_py-1)*abs(_incy_py),),
    )
    b_zdaxpzy(
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zdrscl(alpha, x):
    u"""
    Multiply complex vector by reciprocal of real scalar.

    ``zdrscl`` multiplies a complex vector by the reciprocal of a real
    scalar.

    For full information please refer to the NAG Library document for
    f06ke

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kef.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x, stored in the same array elements used to
        supply the original vector.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `alpha`.

            Constraint: `alpha` != 0.0.

        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.
    """

    from ..base.blas import zdrscl as b_zdrscl

    fun_name = 'naginterfaces.library.blas.zdrscl'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    b_zdrscl(
        _n,
        alpha,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def zdcopy(x):
    u"""
    Copy real vector to complex vector.

    ``zdcopy`` copies a real vector to a complex vector.

    For full information please refer to the NAG Library document for
    f06kf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kff.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The elements y_i of the vector y will be stored in `y` as
        follows.

        If incy > 0, y_i will be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i will be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdcopy as b_zdcopy

    fun_name = 'naginterfaces.library.blas.zdcopy'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _incy = 1
    _incy_py = _incy
    _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (1+(_n_py-1)*abs(_incy_py),),
    )
    b_zdcopy(
        _n,
        _x,
        _incx,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dzssq(x, scal, sumsq):
    u"""
    Update Euclidean norm of complex vector in scaled form.

    ``dzssq`` updates the Euclidean norm of complex vector in scaled
    form.

    For full information please refer to the NAG Library document for
    f06kj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kjf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

    scal : float
        The scaling factor alpha. On the first call to ``dzssq`` `scal`
        = 0.0.

    sumsq : float
        The scaled sum of squares xi. On the first call to ``dzssq``
        `sumsq` = 1.0.

    Returns
    -------
    scal : float
        The updated scaling factor alpha~ =
        max_i(alpha,|Re(x_i)|,|Im(x_i)|).

    sumsq : float
        The updated scaled sum of squares xi~, satisfying: 1 <= xi~ <=
        xi+2n.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

        (`errno` 4)
            On entry, error in parameter `scal`.

            Constraint: `scal` >= 0.

        (`errno` 5)
            On entry, error in parameter `sumsq`.

            Constraint: `sumsq` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dzssq as b_dzssq

    fun_name = 'naginterfaces.library.blas.dzssq'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalars = b_dzssq(
        _n,
        _x,
        _incx,
        scal,
        sumsq,
    )
    _scal_py = _return_scalars.scal
    _sumsq_py = _return_scalars.sumsq
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['scal', 'sumsq'],
    )(
        _scal_py,
        _sumsq_py,
    )
    return _return_data

def idrank(x, tol):
    u"""
    Last non-negligible element of real vector.

    ``idrank`` determines the index of the first negligible element of a
    real vector.

    For full information please refer to the NAG Library document for
    f06kl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06klf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The n-element vector x.

    tol : float
        The value tol.

    Returns
    -------
    k : int
        The index of the first negligible element of the real vector.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import idrank as b_idrank

    fun_name = 'naginterfaces.library.blas.idrank'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _return_scalar = b_idrank(
        _n,
        _x,
        _incx,
        tol,
    )
    return _return_scalar

def zdrot(x, y, c, s):
    u"""
    Apply real plane rotation to two complex vectors.

    ``zdrot`` applies a real plane rotation to two complex vectors.

    For full information please refer to the NAG Library document for
    f06kp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06kpf.html

    Parameters
    ----------
    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    c : float
        The value c, the cosine of the rotation.

    s : float
        The value s, the sine of the rotation.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The transformed vector x stored in the array elements used to
        supply the original vector x.

        Intermediate elements of `x` are unchanged.

    y : complex, ndarray, shape (n)
        The transformed vector y.

        Intermediate elements of `y` are unchanged.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zdrot as b_zdrot

    fun_name = 'naginterfaces.library.blas.zdrot'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _incy = 1
    _incy_py = _incy
    b_zdrot(
        _n,
        _x,
        _incx,
        _y,
        _incy,
        c,
        s,
    )
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def dgemv(trans, m, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, real rectangular matrix.

    ``dgemv`` computes the matrix-vector product for a real general
    matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06paf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T' or 'C'

          y←alpha A^Tx+beta y.

    m : int
        m, the number of rows of the matrix A.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (`m`, n)
        The m by n matrix A.

    x : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': n; if `trans` in ('T', 'C'): `m`;
        otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,n;

        - if incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i
          = 1,2,...,n.

        If `trans` = 'T' or 'C',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,`m`;

        - if incx < 0, x_i must be stored in `x`[1-(`m`-i)*incx-1], for
          i = 1,2,...,`m`.

    beta : float
        The scalar beta.

    y : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `trans`='N': `m`; if `trans` in ('T', 'C'): n;
        otherwise: 0.

        The vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,`m`;

        - if incy < 0, y_i must be stored in `y`[1-(`m`-i)*incy-1], for
          i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,n;

        - if incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i
          = 1,2,...,n.

    Returns
    -------
    y : float, ndarray, shape (:)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 11)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dgemv as b_dgemv

    fun_name = 'naginterfaces.library.blas.dgemv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineFloat64ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if _beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1 or m = 0 or n = 0',
            )
    _trans = _b_utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = _b_utils._EngineCharScalarType.to_py(_trans)
    if y is None:
        if _trans_py in ('N', 'n'):
            _y_dim1 = _m_py
        elif _trans_py in ('T', 't', 'C', 'c'):
            _y_dim1 = _n_py
        else:
            _y_dim1 = 0
        exp_shape_y = (_y_dim1,)
        _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_y,
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    b_dgemv(
        trans,
        _m,
        _n,
        alpha,
        _a,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dgbmv(trans, m, kl, ku, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, real rectangular band matrix.

    ``dgbmv`` computes the matrix-vector product for a real general band
    matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pbf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T' or 'C'

          y←alpha A^Tx+beta y.

    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (`kl`+`ku`+1, n)
        The m by n band matrix A.

    x : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': n; if `trans` in ('T', 'C'): `m`;
        otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,n;

        - if incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i
          = 1,2,...,n.

        If `trans` = 'T' or 'C',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,`m`;

        - if incx < 0, x_i must be stored in `x`[1-(`m`-i)*incx-1], for
          i = 1,2,...,`m`.

    beta : float
        The scalar beta.

    y : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `trans`='N': `m`; if `trans` in ('T', 'C'): n;
        otherwise: 0.

        The vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,`m`;

        - if incy < 0, y_i must be stored in `y`[1-(`m`-i)*incy-1], for
          i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,n;

        - if incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i
          = 1,2,...,n.

    Returns
    -------
    y : float, ndarray, shape (:)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` 10)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 13)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dgbmv as b_dgbmv

    fun_name = 'naginterfaces.library.blas.dgbmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineFloat64ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if _beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1 or m = 0 or n = 0',
            )
    _trans = _b_utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = _b_utils._EngineCharScalarType.to_py(_trans)
    if y is None:
        if _trans_py in ('N', 'n'):
            _y_dim1 = _m_py
        elif _trans_py in ('T', 't', 'C', 'c'):
            _y_dim1 = _n_py
        else:
            _y_dim1 = 0
        exp_shape_y = (_y_dim1,)
        _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_y,
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    b_dgbmv(
        trans,
        _m,
        _n,
        kl,
        ku,
        alpha,
        _a,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dsymv(uplo, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, real symmetric matrix.

    ``dsymv`` the matrix-vector product for a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06pc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pcf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : float
        The scalar beta.

    y : None or float, array-like, shape (n), optional
        The n-element vector y, if `beta` = 0.0 `y` need not be set.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : float, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 7)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 10)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dsymv as b_dsymv

    fun_name = 'naginterfaces.library.blas.dsymv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineFloat64ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    if _beta_py != 0.0 or abs(_incy_py) > 1:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1',
            )
    if y is None:
        _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (1+(_n_py-1)*abs(_incy_py),),
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    b_dsymv(
        uplo,
        _n,
        alpha,
        _a,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dsbmv(uplo, k, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, real symmetric band matrix.

    ``dsbmv`` computes the matrix-vector product for a real symmetric
    band matrix.

    For full information please refer to the NAG Library document for
    f06pd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (`k`+1, n)
        The n by n symmetric band matrix A.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : float
        The scalar beta.

    y : None or float, array-like, shape (n), optional
        The n-element vector y, if `beta` = 0.0 `y` need not be set.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : float, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 3)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 11)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dsbmv as b_dsbmv

    fun_name = 'naginterfaces.library.blas.dsbmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineFloat64ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    if _beta_py != 0.0 or abs(_incy_py) > 1:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1',
            )
    if y is None:
        _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (1+(_n_py-1)*abs(_incy_py),),
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    b_dsbmv(
        uplo,
        _n,
        k,
        alpha,
        _a,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dspmv(uplo, alpha, ap, x, beta, y=None):
    u"""
    Matrix-vector product, real symmetric packed matrix.

    ``dspmv`` computes the matrix-vector product for a real symmetric
    matrix stored in packed form.

    For full information please refer to the NAG Library document for
    f06pe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : float
        The scalar beta.

    y : None or float, array-like, shape (n), optional
        The n-element vector y, if `beta` = 0.0 `y` need not be set.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : float, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 6)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 9)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dspmv as b_dspmv

    fun_name = 'naginterfaces.library.blas.dspmv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineFloat64ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    if _beta_py != 0.0 or abs(_incy_py) > 1:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1',
            )
    if y is None:
        _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (1+(_n_py-1)*abs(_incy_py),),
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    b_dspmv(
        uplo,
        _n,
        alpha,
        _ap,
        _x,
        _incx,
        _beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def dtrmv(uplo, trans, diag, a, x):
    u"""
    Matrix-vector product, real triangular matrix.

    ``dtrmv`` computes the matrix-vector product for a real triangular
    matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pff.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T' or 'C'

          x←A^Tx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    a : float, array-like, shape (n, n)
        The n by n triangular matrix A.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtrmv as b_dtrmv

    fun_name = 'naginterfaces.library.blas.dtrmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_dtrmv(
        uplo,
        trans,
        diag,
        _n,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dtbmv(uplo, trans, diag, k, a, x):
    u"""
    Matrix-vector product, real triangular band matrix.

    ``dtbmv`` computes the matrix-vector product for a real triangular
    band matrix or its transpose.

    For full information please refer to the NAG Library document for
    f06pg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T' or 'C'

          x←A^Tx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : float, array-like, shape (`k`+1, n)
        The n by n triangular band matrix A

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtbmv as b_dtbmv

    fun_name = 'naginterfaces.library.blas.dtbmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_dtbmv(
        uplo,
        trans,
        diag,
        _n,
        k,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dtpmv(uplo, trans, diag, ap, x):
    u"""
    Matrix-vector product, real triangular packed matrix.

    ``dtpmv`` computes the matrix-vector product for a real triangular
    matrix, or its transpose, stored in packed form.

    For full information please refer to the NAG Library document for
    f06ph

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06phf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T' or 'C'

          x←A^Tx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 7)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtpmv as b_dtpmv

    fun_name = 'naginterfaces.library.blas.dtpmv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_dtpmv(
        uplo,
        trans,
        diag,
        _n,
        _ap,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dtrsv(uplo, trans, diag, a, x):
    u"""
    System of equations, real triangular matrix.

    ``dtrsv`` solves a real triangular system of equations with a single
    right-hand side.

    For full information please refer to the NAG Library document for
    f06pj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T' or 'C'

          x←A^invTx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    a : float, array-like, shape (n, n)
        The n by n triangular matrix A.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtrsv as b_dtrsv

    fun_name = 'naginterfaces.library.blas.dtrsv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_dtrsv(
        uplo,
        trans,
        diag,
        _n,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dtbsv(uplo, trans, diag, k, a, x):
    u"""
    System of equations, real triangular band matrix.

    ``dtbsv`` solves a real triangular banded system of equations with a
    single right-hand side.

    For full information please refer to the NAG Library document for
    f06pk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pkf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T' or 'C'

          x←A^invTx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : float, array-like, shape (`k`+1, n)
        The n by n triangular band matrix A

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtbsv as b_dtbsv

    fun_name = 'naginterfaces.library.blas.dtbsv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_dtbsv(
        uplo,
        trans,
        diag,
        _n,
        k,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dtpsv(uplo, trans, diag, ap, x):
    u"""
    System of equations, real triangular packed matrix.

    ``dtpsv`` solves a real triangular system of equations, stored in
    packed form, with a single right-hand side.

    For full information please refer to the NAG Library document for
    f06pl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06plf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T' or 'C'

          x←A^invTx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : float, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 7)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtpsv as b_dtpsv

    fun_name = 'naginterfaces.library.blas.dtpsv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_dtpsv(
        uplo,
        trans,
        diag,
        _n,
        _ap,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def dger(alpha, x, y, a):
    u"""
    Rank-1 update, real rectangular matrix.

    ``dger`` computes the rank-1 update of a real general matrix.

    For full information please refer to the NAG Library document for
    f06pm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pmf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (m)
        The m element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx], for i =
        1,2,...,m.

        If incx < 0, x_i must be stored in `x`[1-(m-i)*incx], for i =
        1,2,...,m.

        Intermediate elements of `x` are not referenced.

    y : float, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    a : float, array-like, shape (m, n)
        The m by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (m, n)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 7)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dger as b_dger

    fun_name = 'naginterfaces.library.blas.dger'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    b_dger(
        _m,
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
    )
    _a_py = _a
    return _a_py

def dsyr(uplo, alpha, x, a):
    u"""
    Rank-1 update, real symmetric matrix.

    ``dsyr`` computes the rank-1 update of a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06pp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ppf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dsyr as b_dsyr

    fun_name = 'naginterfaces.library.blas.dsyr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    b_dsyr(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _a,
    )
    _a_py = _a
    return _a_py

def dspr(uplo, alpha, x, ap):
    u"""
    Rank-1 update, real symmetric packed matrix.

    ``dspr`` computes the rank-1 update of a real symmetric matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06pq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06pqf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (n*(n+1)/2)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dspr as b_dspr

    fun_name = 'naginterfaces.library.blas.dspr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dspr(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def dsyr2(uplo, alpha, x, y, a):
    u"""
    Rank-2 update, real symmetric matrix.

    ``dsyr2`` computes the rank-2 update of a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06pr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06prf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : float, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 7)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dsyr2 as b_dsyr2

    fun_name = 'naginterfaces.library.blas.dsyr2'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    b_dsyr2(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
    )
    _a_py = _a
    return _a_py

def dspr2(uplo, alpha, x, y, ap):
    u"""
    Rank-2 update, real symmetric packed matrix.

    ``dspr2`` computes the rank-2 update of a real symmetric matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06ps

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06psf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : float, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (n*(n+1)/2)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 7)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dspr2 as b_dspr2

    fun_name = 'naginterfaces.library.blas.dspr2'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dspr2(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def dmcopy(matrix, m, a):
    u"""
    Matrix copy, real rectangular or trapezoidal matrix.

    ``dmcopy`` performs the matrix-copy operation

        B←A

    where A and B are m by n real general or trapezoidal matrices.

    For full information please refer to the NAG Library document for
    f06qf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qff.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrices A and B.

    a : float, array-like, shape (`m`, n)
        The m by n general or trapezoidal matrix A.

    Returns
    -------
    b : float, ndarray, shape (`m`, n)
        The m by n general or trapezoidal matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dmcopy as b_dmcopy

    fun_name = 'naginterfaces.library.blas.dmcopy'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_m_sorder,
    )
    b_dmcopy(
        matrix,
        _m,
        _n,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def dmload(matrix, m, n, con, diag):
    u"""
    Matrix initialization, real rectangular matrix.

    ``dmload`` forms the real m by n rectangular or trapezoidal matrix A
    given by

        a_{ij} = {[table omitted].

    For full information please refer to the NAG Library document for
    f06qh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qhf.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    con : float
        The value to be assigned to the off-diagonal elements of A.

    diag : float
        The value to be assigned to the diagonal elements of A.

    Returns
    -------
    a : float, ndarray, shape (`m`, `n`)
        The m by n general or trapezoidal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dmload as b_dmload

    fun_name = 'naginterfaces.library.blas.dmload'

    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_dmload(
        matrix,
        _m,
        _n,
        con,
        diag,
        _a,
    )
    _a_py = _a
    return _a_py

def dgeap(side, trans, perm, k, b):
    u"""
    Permute rows or columns, real rectangular matrix, permutations
    represented by an integer array.

    ``dgeap`` permutes the rows or columns of a real rectangular matrix
    using an int array of permutations.

    For full information please refer to the NAG Library document for
    f06qj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    perm : int, array-like, shape (n)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(n, m); if `side`='R':
        `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R': max(n,
        max(`perm`)); otherwise: 0.

        The original matrix B; B is m by k if `side` = 'L', or k by m if
        `side` = 'R'.

    Returns
    -------
    b : float, ndarray, shape (:, :)
        The permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dgeap as b_dgeap

    fun_name = 'naginterfaces.library.blas.dgeap'

    _perm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'perm'},
        perm,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_perm, {'fun_name': fun_name, 'entity': 'perm'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _vl_m_py = max(_perm[0:max(1, _n_py)-1+1])
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dgeap(
        side,
        trans,
        _n,
        _perm,
        k,
        _b,
    )
    _b_py = _b
    return _b_py

def dgeapr(side, trans, perm, k, b):
    u"""
    Permute rows or columns, real rectangular matrix, permutations
    represented by a real array.

    ``dgeapr`` permutes the rows or columns of a real rectangular matrix
    using a double array of permutations.

    For full information please refer to the NAG Library document for
    f06qk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qkf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    perm : float, array-like, shape (n)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(n, m); if `side`='R':
        `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R': max(n,
        int(max(`perm`))); otherwise: 0.

        The original matrix B; B is m by k if `side` = 'L', or k by m if
        `side` = 'R'.

    Returns
    -------
    b : float, ndarray, shape (:, :)
        The permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dgeapr as b_dgeapr

    fun_name = 'naginterfaces.library.blas.dgeapr'

    _perm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'perm'},
        perm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_perm, {'fun_name': fun_name, 'entity': 'perm'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _vl_m_py = int(max(_perm[0:max(1, _n_py)-1+1]))
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dgeapr(
        side,
        trans,
        _n,
        _perm,
        k,
        _b,
    )
    _b_py = _b
    return _b_py

def dsysrc(uplo, pivot, direct, k1, k2, c, s, a):
    u"""
    Orthogonal similarity transformation of real symmetric matrix as a
    sequence of plane rotations.

    ``dsysrc`` performs an orthogonal similarity transformation (as a
    sequence of plane rotations) of a real symmetric matrix.

    For full information please refer to the NAG Library document for
    f06qm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qmf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dsysrc as b_dsysrc

    fun_name = 'naginterfaces.library.blas.dsysrc'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dsysrc(
        uplo,
        pivot,
        direct,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _a_py = _a
    return _a_py

def dutr1(alpha, x, y, a):
    u"""
    QR factorization by sequence of plane rotations, rank-1 update of
    real upper triangular matrix.

    ``dutr1`` performs a QR factorization (as a sequence of plane
    rotations) of a real upper triangular matrix that has been modified
    by a rank-1 update.

    For full information please refer to the NAG Library document for
    f06qp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qpf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    y : float, array-like, shape (n)
        The n-element vector y. y_i must be stored in `y`[1+(i-1)*incy],
        for i = 1,2,...,n.

        Intermediate elements of `y` are not referenced.

    a : float, array-like, shape (n, n)
        The n by n upper triangular matrix U.

    Returns
    -------
    x : float, ndarray, shape (n)
        The referenced elements are overwritten by details of the
        sequence of plane rotations.

    a : float, ndarray, shape (n, n)
        The upper triangular matrix R.

    c : float, ndarray, shape (n-1)
        The cosines of the rotations Q_k, for k = 1,2,...,n-1.

    s : float, ndarray, shape (n-1)
        The sines of the rotations Q_k, for k = 1,2,...,n-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

        (`errno` 6)
            On entry, error in parameter incy.

            Constraint: incy > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dutr1 as b_dutr1

    fun_name = 'naginterfaces.library.blas.dutr1'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _incy = 1
    _incy_py = _incy
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py-1,),
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py-1,),
    )
    b_dutr1(
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _c,
        _s,
    )
    _x_py = _x
    _a_py = _a
    _c_py = _c
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'a', 'c', 's'],
    )(
        _x_py,
        _a_py,
        _c_py,
        _s_py,
    )
    return _return_data

def dutupd(alpha, x, a):
    u"""
    QR factorization by sequence of plane rotations, real upper
    triangular matrix augmented by a full row.

    ``dutupd`` performs a QR factorization (as a sequence of plane
    rotations) of a real upper triangular matrix that has been augmented
    by a full row.

    For full information please refer to the NAG Library document for
    f06qq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qqf.html

    Parameters
    ----------
    alpha : float
        The scalar alpha.

    x : float, array-like, shape (n)
        The vector x. x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

    a : float, array-like, shape (n, n)
        The n by n upper triangular matrix U.

    Returns
    -------
    x : float, ndarray, shape (n)
        The referenced elements are overwritten by the tangents of the
        rotations Q_k, for k = 1,2,...,n.

    a : float, ndarray, shape (n, n)
        The upper triangular matrix R.

    c : float, ndarray, shape (n)
        The values c_k, the cosines of the rotations Q_k, for k =
        1,2,...,n.

    s : float, ndarray, shape (n)
        The values s_k, the sines of the rotations Q_k, for k =
        1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dutupd as b_dutupd

    fun_name = 'naginterfaces.library.blas.dutupd'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dutupd(
        _n,
        alpha,
        _x,
        _incx,
        _a,
        _c,
        _s,
    )
    _x_py = _x
    _a_py = _a
    _c_py = _c
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'a', 'c', 's'],
    )(
        _x_py,
        _a_py,
        _c_py,
        _s_py,
    )
    return _return_data

def duhqr(side, k1, k2, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, real upper
    Hessenberg matrix.

    ``duhqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a real upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06qr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qrf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    s : float, array-like, shape (`k2`-1)
        The nonzero subdiagonal elements of H: `s`[k] must hold
        h_{k+1,k}, for k = k_1,...,k_2-1.

    a : float, array-like, shape (n, n)
        The upper triangular part of the n by n upper Hessenberg matrix
        H.

    Returns
    -------
    c : float, ndarray, shape (`k2`-1)
        `c`[k] holds c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, ndarray, shape (`k2`-1)
        `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, ndarray, shape (n, n)
        The upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import duhqr as b_duhqr

    fun_name = 'naginterfaces.library.blas.duhqr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k2 = _b_utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = _b_utils._EngineIntScalarType.to_py(_k2)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k2_py-1,),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_duhqr(
        side,
        _n,
        k1,
        _k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _c_py = _c
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's', 'a'],
    )(
        _c_py,
        _s_py,
        _a_py,
    )
    return _return_data

def dusqr(side, k1, k2, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, real upper
    spiked matrix.

    ``dusqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a real upper spiked matrix.

    For full information please refer to the NAG Library document for
    f06qs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qsf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    s : float, array-like, shape (`k2`-1)
        The nonzero elements of the spike of H: `s`[k] must hold
        h_{k_2,k} if `side` = 'L', and h_{k+1,k_1} if `side` = 'R', for
        k = k_1,...,k_2-1.

    a : float, array-like, shape (n, n)
        The upper triangular part of the n by n upper spiked matrix H.

    Returns
    -------
    c : float, ndarray, shape (`k2`-1)
        `c`[k] holds c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, ndarray, shape (`k2`-1)
        `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, ndarray, shape (n, n)
        The upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dusqr as b_dusqr

    fun_name = 'naginterfaces.library.blas.dusqr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k2 = _b_utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = _b_utils._EngineIntScalarType.to_py(_k2)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k2_py-1,),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dusqr(
        side,
        _n,
        k1,
        _k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _c_py = _c
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's', 'a'],
    )(
        _c_py,
        _s_py,
        _a_py,
    )
    return _return_data

def dutsqr(side, k1, k2, c, s, a):
    u"""
    QR factorization of UP or RQ factorization of PU, U real upper
    triangular, P a sequence of plane rotations.

    ``dutsqr`` performs a QR or RQ factorization of the product of a
    real upper triangular matrix and a real matrix of plane rotations.

    For full information please refer to the NAG Library document for
    f06qt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qtf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether P is applied from the left or the right in the
        transformation.

        `side` = 'L'

          P is applied from the left.

        `side` = 'R'

          P is applied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, array-like, shape (n, n)
        The n by n upper triangular matrix U.

    Returns
    -------
    c : float, ndarray, shape (`k2`-1)
        `c`[k] holds the cosine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    s : float, ndarray, shape (`k2`-1)
        `s`[k] holds the sine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    a : float, ndarray, shape (n, n)
        The upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dutsqr as b_dutsqr

    fun_name = 'naginterfaces.library.blas.dutsqr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dutsqr(
        side,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _c_py = _c
    _s_py = _s
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's', 'a'],
    )(
        _c_py,
        _s_py,
        _a_py,
    )
    return _return_data

def dutsrh(side, k1, k2, c, s, a):
    u"""
    Compute upper Hessenberg matrix by sequence of plane rotations, real
    upper triangular matrix.

    ``dutsrh`` transforms a real upper triangular matrix to an upper
    Hessenberg matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06qv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qvf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, array-like, shape (n, n)
        The n by n upper triangular matrix U.

    Returns
    -------
    s : float, ndarray, shape (`k2`-1)
        `s`[k] holds h_{k+1,k}, the subdiagonal element of H, for k =
        k_1,...,k_2-1.

    a : float, ndarray, shape (n, n)
        The upper triangular part of the upper Hessenberg matrix H.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dutsrh as b_dutsrh

    fun_name = 'naginterfaces.library.blas.dutsrh'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dutsrh(
        side,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'a'],
    )(
        _s_py,
        _a_py,
    )
    return _return_data

def dutsrs(side, k1, k2, c, s, a):
    u"""
    Compute upper spiked matrix by sequence of plane rotations, real
    upper triangular matrix.

    ``dutsrs`` transforms a real upper triangular matrix to an upper
    spiked matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06qw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qwf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, array-like, shape (n, n)
        The n by n upper triangular matrix U.

    Returns
    -------
    s : float, ndarray, shape (`k2`-1)
        `s`[k] holds a nonzero element of the spike of H: h_{k_2,k} if
        `side` = 'L', or h_{k+1,k_1} if `side` = 'R', for k =
        k_1,...,k_2-1.

    a : float, ndarray, shape (n, n)
        The upper triangular part of the upper spiked matrix H.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dutsrs as b_dutsrs

    fun_name = 'naginterfaces.library.blas.dutsrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dutsrs(
        side,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'a'],
    )(
        _s_py,
        _a_py,
    )
    return _return_data

def dgesrc(side, pivot, direct, m, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, real rectangular matrix.

    ``dgesrc`` applies a sequence of plane rotations to a real
    rectangular matrix.

    For full information please refer to the NAG Library document for
    f06qx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06qxf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

        If `m` < 1, an immediate return is effected.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : float, array-like, shape (`m`, n)
        The m by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (`m`, n)
        The transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dgesrc as b_dgesrc

    fun_name = 'naginterfaces.library.blas.dgesrc'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dgesrc(
        side,
        pivot,
        direct,
        m,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _a_py = _a
    return _a_py

def dlange(norm, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real general matrix.

    ``dlange`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real m by n matrix.

    For full information please refer to the NAG Library document for
    f06ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06raf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    a : float, array-like, shape (m, n)
        The m by n matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real m by n
        matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlange as b_dlange

    fun_name = 'naginterfaces.library.blas.dlange'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_dlange(
        norm,
        _m,
        _n,
        _a,
    )
    return _return_scalar

def dlangb(norm, kl, ku, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real band matrix.

    ``dlangb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n band matrix.

    For full information please refer to the NAG Library document for
    f06rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rbf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    ab : float, array-like, shape (`kl`+`ku`+1, n)
        The n by n band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlangb as b_dlangb

    fun_name = 'naginterfaces.library.blas.dlangb'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_dlangb(
        norm,
        _n,
        kl,
        ku,
        _ab,
    )
    return _return_scalar

def dlansy(norm, uplo, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric matrix.

    ``dlansy`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric matrix.

    For full information please refer to the NAG Library document for
    f06rc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rcf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlansy as b_dlansy

    fun_name = 'naginterfaces.library.blas.dlansy'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_dlansy(
        norm,
        uplo,
        _n,
        _a,
    )
    return _return_scalar

def dlansp(norm, uplo, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric matrix, packed storage.

    ``dlansp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06rd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rdf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlansp`` returns zero.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlansp as b_dlansp

    fun_name = 'naginterfaces.library.blas.dlansp'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_dlansp(
        norm,
        uplo,
        n,
        _ap,
    )
    return _return_scalar

def dlansb(norm, uplo, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric band matrix.

    ``dlansb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric band matrix.

    For full information please refer to the NAG Library document for
    f06re

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ref.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : float, array-like, shape (`k`+1, n)
        The n by n symmetric band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlansb as b_dlansb

    fun_name = 'naginterfaces.library.blas.dlansb'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_dlansb(
        norm,
        uplo,
        _n,
        k,
        _ab,
    )
    return _return_scalar

def dlantr(norm, uplo, diag, m, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real trapezoidal/triangular matrix.

    ``dlantr`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real m by n trapezoidal matrix (triangular if m = n).

    For full information please refer to the NAG Library document for
    f06rj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rjf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower trapezoidal.

        `uplo` = 'U'

          A is upper trapezoidal.

        `uplo` = 'L'

          A is lower trapezoidal.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix A.

        When `m` = 0, ``dlantr`` is set to zero.

    a : float, array-like, shape (`m`, n)
        The m by n trapezoidal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real m by n
        trapezoidal matrix (triangular if m = n).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlantr as b_dlantr

    fun_name = 'naginterfaces.library.blas.dlantr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_dlantr(
        norm,
        uplo,
        diag,
        m,
        _n,
        _a,
    )
    return _return_scalar

def dlantp(norm, uplo, diag, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real triangular matrix, packed storage.

    ``dlantp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n triangular matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06rk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rkf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlantp`` returns zero.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        triangular matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlantp as b_dlantp

    fun_name = 'naginterfaces.library.blas.dlantp'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_dlantp(
        norm,
        uplo,
        diag,
        n,
        _ap,
    )
    return _return_scalar

def dlantb(norm, uplo, diag, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real triangular band matrix.

    ``dlantb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n triangular band matrix.

    For full information please refer to the NAG Library document for
    f06rl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rlf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : float, array-like, shape (`k`+1, n)
        The n by n triangular band matrix A

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        triangular band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlantb as b_dlantb

    fun_name = 'naginterfaces.library.blas.dlantb'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_dlantb(
        norm,
        uplo,
        diag,
        _n,
        k,
        _ab,
    )
    return _return_scalar

def dlanhs(norm, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real upper Hessenberg matrix.

    ``dlanhs`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06rm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rmf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    a : float, array-like, shape (n, n)
        The n by n upper Hessenberg matrix A; elements of the array
        below the first subdiagonal are not referenced.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        upper Hessenberg matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlanhs as b_dlanhs

    fun_name = 'naginterfaces.library.blas.dlanhs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_dlanhs(
        norm,
        _n,
        _a,
    )
    return _return_scalar

def dlangt(norm, dl, d, du):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real tridiagonal matrix.

    ``dlangt`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06rn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rnf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    dl : float, array-like, shape (n-1)
        The (n-1) subdiagonal elements of A.

    d : float, array-like, shape (n)
        The n diagonal elements of A.

    du : float, array-like, shape (n-1)
        The (n-1) superdiagonal elements of A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlangt as b_dlangt

    fun_name = 'naginterfaces.library.blas.dlangt'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_dlangt(
        norm,
        _n,
        _dl,
        _d,
        _du,
    )
    return _return_scalar

def dlanst(norm, d, e):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric tridiagonal matrix.

    ``dlanst`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real n by n symmetric tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06rp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06rpf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : float, array-like, shape (n-1)
        The (n-1) subdiagonal or superdiagonal elements of the
        tridiagonal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the real n by n
        symmetric tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlanst as b_dlanst

    fun_name = 'naginterfaces.library.blas.dlanst'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_dlanst(
        norm,
        _n,
        _d,
        _e,
    )
    return _return_scalar

def zgemv(trans, m, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, complex rectangular matrix.

    ``zgemv`` computes the matrix-vector product for a complex general
    matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06saf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T'

          y←alpha A^Tx+beta y.

        `trans` = 'C'

          y←alpha A^Hx+beta y.

    m : int
        m, the number of rows of the matrix A.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (`m`, n)
        The m by n matrix A.

    x : complex, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': n; if `trans` in ('T', 'C'): `m`;
        otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,n;

        - if incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i
          = 1,2,...,n.

        If `trans` = 'T' or 'C',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,`m`;

        - if incx < 0, x_i must be stored in `x`[1-(`m`-i)*incx-1], for
          i = 1,2,...,`m`.

    beta : complex
        The scalar beta.

    y : None or complex, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `trans`='N': `m`; if `trans` in ('T', 'C'): n;
        otherwise: 0.

        The vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,`m`;

        - if incy < 0, y_i must be stored in `y`[1-(`m`-i)*incy-1], for
          i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,n;

        - if incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i
          = 1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (:)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 11)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zgemv as b_zgemv

    fun_name = 'naginterfaces.library.blas.zgemv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if _beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1 or m = 0 or n = 0',
            )
    _trans = _b_utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = _b_utils._EngineCharScalarType.to_py(_trans)
    if y is None:
        if _trans_py in ('N', 'n'):
            _y_dim1 = _m_py
        elif _trans_py in ('T', 't', 'C', 'c'):
            _y_dim1 = _n_py
        else:
            _y_dim1 = 0
        exp_shape_y = (_y_dim1,)
        _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            exp_shape_y,
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
        )
    b_zgemv(
        trans,
        _m,
        _n,
        alpha,
        _a,
        _x,
        _incx,
        beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zgbmv(trans, m, kl, ku, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, complex rectangular band matrix.

    ``zgbmv`` computes the matrix-vector product for a complex general
    band matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sbf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          y←alpha Ax+beta y.

        `trans` = 'T'

          y←alpha A^Tx+beta y.

        `trans` = 'C'

          y←alpha A^Hx+beta y.

    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (`kl`+`ku`+1, n)
        The m by n band matrix A.

    x : complex, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `trans`='N': n; if `trans` in ('T', 'C'): `m`;
        otherwise: 0.

        The vector x.

        If `trans` = 'N',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,n;

        - if incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i
          = 1,2,...,n.

        If `trans` = 'T' or 'C',

        - if incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i
          = 1,2,...,`m`;

        - if incx < 0, x_i must be stored in `x`[1-(`m`-i)*incx-1], for
          i = 1,2,...,`m`.

    beta : complex
        The scalar beta.

    y : None or complex, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `trans`='N': `m`; if `trans` in ('T', 'C'): n;
        otherwise: 0.

        The vector y, if `beta` = 0.0, `y` need not be set.

        If `trans` = 'N',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,`m`;

        - if incy < 0, y_i must be stored in `y`[1-(`m`-i)*incy-1], for
          i = 1,2,...,`m`.

        If `trans` = 'T' or 'C',

        - if incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i
          = 1,2,...,n;

        - if incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i
          = 1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (:)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` 10)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 13)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zgbmv as b_zgbmv

    fun_name = 'naginterfaces.library.blas.zgbmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if _beta_py != 0.0 or abs(_incy_py) > 1 or _m_py == 0 or _n_py == 0:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1 or m = 0 or n = 0',
            )
    _trans = _b_utils._EngineCharScalarType(
        trans, {'fun_name': fun_name, 'entity': 'trans'}, exp_chrlen=1,
    ).to_ctypes()
    _trans_py = _b_utils._EngineCharScalarType.to_py(_trans)
    if y is None:
        if _trans_py in ('N', 'n'):
            _y_dim1 = _m_py
        elif _trans_py in ('T', 't', 'C', 'c'):
            _y_dim1 = _n_py
        else:
            _y_dim1 = 0
        exp_shape_y = (_y_dim1,)
        _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            exp_shape_y,
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
        )
    b_zgbmv(
        trans,
        _m,
        _n,
        kl,
        ku,
        alpha,
        _a,
        _x,
        _incx,
        beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zhemv(uplo, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, complex Hermitian matrix.

    ``zhemv`` computes the matrix-vector product for a complex Hermitian
    matrix.

    For full information please refer to the NAG Library document for
    f06sc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06scf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : complex
        The scalar beta.

    y : None or complex, array-like, shape (n), optional
        The n-element vector y, if `beta` = 0.0, `y` need not be set.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 7)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 10)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zhemv as b_zhemv

    fun_name = 'naginterfaces.library.blas.zhemv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    if _beta_py != 0.0 or abs(_incy_py) > 1:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1',
            )
    if y is None:
        _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            (1+(_n_py-1)*abs(_incy_py),),
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
        )
    b_zhemv(
        uplo,
        _n,
        alpha,
        _a,
        _x,
        _incx,
        beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zhbmv(uplo, k, alpha, a, x, beta, y=None):
    u"""
    Matrix-vector product, complex Hermitian band matrix.

    ``zhbmv`` computes the matrix-vector product for a complex Hermitian
    band matrix.

    For full information please refer to the NAG Library document for
    f06sd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (`k`+1, n)
        The n by n Hermitian band matrix A.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : complex
        The scalar beta.

    y : None or complex, array-like, shape (n), optional
        The n-element vector y, if `beta` = 0.0, `y` need not be set.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 3)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 11)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zhbmv as b_zhbmv

    fun_name = 'naginterfaces.library.blas.zhbmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    if _beta_py != 0.0 or abs(_incy_py) > 1:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1',
            )
    if y is None:
        _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            (1+(_n_py-1)*abs(_incy_py),),
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
        )
    b_zhbmv(
        uplo,
        _n,
        k,
        alpha,
        _a,
        _x,
        _incx,
        beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zhpmv(uplo, alpha, ap, x, beta, y=None):
    u"""
    Matrix-vector product, complex Hermitian packed matrix.

    ``zhpmv`` computes the matrix-vector product for a complex Hermitian
    matrix stored in packed form.

    For full information please refer to the NAG Library document for
    f06se

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : complex
        The scalar beta.

    y : None or complex, array-like, shape (n), optional
        The n-element vector y, if `beta` = 0.0, `y` need not be set.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 6)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 9)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zhpmv as b_zhpmv

    fun_name = 'naginterfaces.library.blas.zhpmv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    _incy = 1
    _incy_py = _incy
    if _beta_py != 0.0 or abs(_incy_py) > 1:
        if y is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'y'},
                'must not be None when beta != 0.0 or abs(incy) > 1',
            )
    if y is None:
        _y = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            (1+(_n_py-1)*abs(_incy_py),),
        )
    else:
        _y = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'y'},
            y,
            copy=True,
        )
    b_zhpmv(
        uplo,
        _n,
        alpha,
        _ap,
        _x,
        _incx,
        beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def ztrmv(uplo, trans, diag, a, x):
    u"""
    Matrix-vector product, complex triangular matrix.

    ``ztrmv`` computes the matrix-vector product for a complex
    triangular matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sff.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T'

          x←A^Tx.

        `trans` = 'C'

          x←A^Hx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    a : complex, array-like, shape (n, n)
        The n by n triangular matrix A.

    x : complex, array-like, shape (n)
        The vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztrmv as b_ztrmv

    fun_name = 'naginterfaces.library.blas.ztrmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _incx = 1
    _incx_py = _incx
    b_ztrmv(
        uplo,
        trans,
        diag,
        _n,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def ztbmv(uplo, trans, diag, k, a, x):
    u"""
    Matrix-vector product, complex triangular band matrix.

    ``ztbmv`` computes the matrix-vector product for a complex
    triangular band matrix, its transpose or its conjugate transpose.

    For full information please refer to the NAG Library document for
    f06sg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T'

          x←A^Tx.

        `trans` = 'C'

          x←A^Hx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : complex, array-like, shape (`k`+1, n)
        The n by n triangular band matrix A

    x : complex, array-like, shape (n)
        The vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztbmv as b_ztbmv

    fun_name = 'naginterfaces.library.blas.ztbmv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _incx = 1
    _incx_py = _incx
    b_ztbmv(
        uplo,
        trans,
        diag,
        _n,
        k,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def ztpmv(uplo, trans, diag, ap, x):
    u"""
    Matrix-vector product, complex triangular packed matrix.

    ``ztpmv`` computes the matrix-vector product for a complex
    triangular matrix, its transpose or its conjugate transpose, stored
    in packed form.

    For full information please refer to the NAG Library document for
    f06sh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06shf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←Ax.

        `trans` = 'T'

          x←A^Tx.

        `trans` = 'C'

          x←A^Hx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : complex, array-like, shape (n)
        The vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 7)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztpmv as b_ztpmv

    fun_name = 'naginterfaces.library.blas.ztpmv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    b_ztpmv(
        uplo,
        trans,
        diag,
        _n,
        _ap,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def ztrsv(uplo, trans, diag, a, x):
    u"""
    System of equations, complex triangular matrix.

    ``ztrsv`` solves a complex triangular system of equations with a
    single right-hand side.

    For full information please refer to the NAG Library document for
    f06sj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T'

          x←A^invTx.

        `trans` = 'C'

          x←A^invHx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    a : complex, array-like, shape (n, n)
        The n by n triangular matrix A.

    x : complex, array-like, shape (n)
        The vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 8)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztrsv as b_ztrsv

    fun_name = 'naginterfaces.library.blas.ztrsv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _incx = 1
    _incx_py = _incx
    b_ztrsv(
        uplo,
        trans,
        diag,
        _n,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def ztbsv(uplo, trans, diag, k, a, x):
    u"""
    System of equations, complex triangular band matrix.

    ``ztbsv`` solves a complex triangular banded system of equations
    with a single right-hand side.

    For full information please refer to the NAG Library document for
    f06sk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06skf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T'

          x←A^invTx.

        `trans` = 'C'

          x←A^invHx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    a : complex, array-like, shape (`k`+1, n)
        The n by n triangular band matrix A

    x : complex, array-like, shape (n)
        The vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

        (`errno` 9)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztbsv as b_ztbsv

    fun_name = 'naginterfaces.library.blas.ztbsv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _incx = 1
    _incx_py = _incx
    b_ztbsv(
        uplo,
        trans,
        diag,
        _n,
        k,
        _a,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def ztpsv(uplo, trans, diag, ap, x):
    u"""
    System of equations, complex triangular packed matrix.

    ``ztpsv`` solves a complex triangular system of equations, stored in
    packed form, with a single right-hand side.

    For full information please refer to the NAG Library document for
    f06sl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06slf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          x←A^-1x.

        `trans` = 'T'

          x←A^invTx.

        `trans` = 'C'

          x←A^invHx.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n triangular matrix A, packed by columns.

    x : complex, array-like, shape (n)
        The vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,n.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx-1], for i =
        1,2,...,n.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The updated vector x stored in the array elements used to supply
        the original vector x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 7)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztpsv as b_ztpsv

    fun_name = 'naginterfaces.library.blas.ztpsv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    b_ztpsv(
        uplo,
        trans,
        diag,
        _n,
        _ap,
        _x,
        _incx,
    )
    _x_py = _x
    return _x_py

def zgeru(alpha, x, y, a):
    u"""
    Rank-1 update, complex rectangular matrix, unconjugated vector.

    ``zgeru`` computes the rank-1 update of a complex general matrix
    using an unconjugated vector.

    For full information please refer to the NAG Library document for
    f06sm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06smf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (m)
        The m element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,m.

        If incx < 0, x_i must be stored in `x`[1-(m-i)*incx-1], for i =
        1,2,...,m.

        Intermediate elements of X are not referenced.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    a : complex, array-like, shape (m, n)
        The m by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (m, n)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 7)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zgeru as b_zgeru

    fun_name = 'naginterfaces.library.blas.zgeru'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _incy = 1
    _incy_py = _incy
    b_zgeru(
        _m,
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
    )
    _a_py = _a
    return _a_py

def zgerc(alpha, x, y, a):
    u"""
    Rank-1 update, complex rectangular matrix, conjugated vector.

    ``zgerc`` computes the rank-1 update of a complex general matrix
    using a conjugated vector.

    For full information please refer to the NAG Library document for
    f06sn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06snf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (m)
        The m element vector x.

        If incx > 0, x_i must be stored in `x`[1+(i-1)*incx-1], for i =
        1,2,...,m.

        If incx < 0, x_i must be stored in `x`[1-(m-i)*incx-1], for i =
        1,2,...,m.

        Intermediate elements of X are not referenced.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    a : complex, array-like, shape (m, n)
        The m by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (m, n)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 7)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zgerc as b_zgerc

    fun_name = 'naginterfaces.library.blas.zgerc'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _incy = 1
    _incy_py = _incy
    b_zgerc(
        _m,
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
    )
    _a_py = _a
    return _a_py

def zher(uplo, alpha, x, a):
    u"""
    Rank-1 update, complex Hermitian matrix.

    ``zher`` computes the rank-1 update of a complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06sp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06spf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The updated matrix A. The imaginary parts of the diagonal
        elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zher as b_zher

    fun_name = 'naginterfaces.library.blas.zher'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    b_zher(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _a,
    )
    _a_py = _a
    return _a_py

def zhpr(uplo, alpha, x, ap):
    u"""
    Rank-1 update, complex Hermitian packed matrix.

    ``zhpr`` computes the rank-1 update of a complex Hermitian matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06sq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06sqf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : float
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The updated matrix A. The imaginary parts of the diagonal
        elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zhpr as b_zhpr

    fun_name = 'naginterfaces.library.blas.zhpr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    b_zhpr(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def zher2(uplo, alpha, x, y, a):
    u"""
    Rank-2 update, complex Hermitian matrix.

    ``zher2`` computes the rank-2 update of a complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06sr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06srf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The updated matrix A. The imaginary parts of the diagonal
        elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 7)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zher2 as b_zher2

    fun_name = 'naginterfaces.library.blas.zher2'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _incy = 1
    _incy_py = _incy
    b_zher2(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
    )
    _a_py = _a
    return _a_py

def zhpr2(uplo, alpha, x, y, ap):
    u"""
    Rank-2 update, complex Hermitian packed matrix.

    ``zhpr2`` computes the rank-2 update of a complex Hermitian matrix
    stored in packed form.

    For full information please refer to the NAG Library document for
    f06ss

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ssf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy], for i =
        1,2,...,n.

        Intermediate elements of `y` are not referenced.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The updated matrix A. The imaginary parts of the diagonal
        elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 7)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zhpr2 as b_zhpr2

    fun_name = 'naginterfaces.library.blas.zhpr2'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _incy = 1
    _incy_py = _incy
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    b_zhpr2(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def zsymv(uplo, alpha, a, x, beta, y):
    u"""
    Matrix-vector product, complex symmetric matrix.

    ``zsymv`` performs the matrix-vector operation

        y←alpha Ax+beta y,

    where A is an n by n complex symmetric matrix, x and y are n-element
    complex vectors, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06ta

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06taf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (n, n)
        The n by n symmetric matrix A.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : complex
        The scalar beta.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 7)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 10)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zsymv as b_zsymv

    fun_name = 'naginterfaces.library.blas.zsymv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _incy = 1
    _incy_py = _incy
    b_zsymv(
        uplo,
        _n,
        alpha,
        _a,
        _x,
        _incx,
        beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zsyr(uplo, alpha, x, a):
    u"""
    Rank-1 update, complex symmetric matrix.

    ``zsyr`` performs the symmetric rank-1 update operation

        A←alpha xx^T+A,

    where A is an n by n complex symmetric matrix, x is an n-element
    complex vector, and alpha is a complex scalar.

    For full information please refer to the NAG Library document for
    f06tb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tbf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    a : complex, array-like, shape (n, n)
        The n by n symmetric matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zsyr as b_zsyr

    fun_name = 'naginterfaces.library.blas.zsyr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    b_zsyr(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _a,
    )
    _a_py = _a
    return _a_py

def zspmv(uplo, alpha, ap, x, beta, y):
    u"""
    Matrix-vector product, complex symmetric packed matrix.

    ``zspmv`` performs the matrix-vector operation

        y←alpha Ax+beta y

    where A is an n by n complex symmetric matrix stored in packed form,
    x and y are n-element complex vectors, and alpha and beta are
    complex scalars.

    For full information please refer to the NAG Library document for
    f06tc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tcf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    beta : complex
        The scalar beta.

    y : complex, array-like, shape (n)
        The n-element vector y.

        If incy > 0, y_i must be stored in `y`[1+(i-1)*incy-1], for i =
        1,2,...,n.

        If incy < 0, y_i must be stored in `y`[1-(n-i)*incy-1], for i =
        1,2,...,n.

    Returns
    -------
    y : complex, ndarray, shape (n)
        The updated vector y stored in the array elements used to supply
        the original vector y.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 6)
            On entry, error in parameter incx.

            Constraint: incx != 0.

        (`errno` 9)
            On entry, error in parameter incy.

            Constraint: incy != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zspmv as b_zspmv

    fun_name = 'naginterfaces.library.blas.zspmv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
    )
    _incy = 1
    _incy_py = _incy
    b_zspmv(
        uplo,
        _n,
        alpha,
        _ap,
        _x,
        _incx,
        beta,
        _y,
        _incy,
    )
    _y_py = _y
    return _y_py

def zspr(uplo, alpha, x, ap):
    u"""
    Rank-1 update, complex symmetric packed matrix.

    ``zspr`` performs the symmetric rank-1 update operation

        A←alpha xx^T+A,

    where A is an n by n complex symmetric matrix, stored in packed
    form, x is an n-element complex vector, and alpha is a complex
    scalar.

    For full information please refer to the NAG Library document for
    f06td

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x.

        If incx < 0, x_i must be stored in `x`[1-(n-i)*incx], for i =
        1,2,...,n.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The updated matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter incx.

            Constraint: incx != 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zspr as b_zspr

    fun_name = 'naginterfaces.library.blas.zspr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    b_zspr(
        uplo,
        _n,
        alpha,
        _x,
        _incx,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def zmcopy(matrix, m, a):
    u"""
    Matrix copy, complex rectangular or trapezoidal matrix.

    ``zmcopy`` performs the matrix-copy operation

        B←A

    where A and B are m by n complex general or trapezoidal matrices.

    For full information please refer to the NAG Library document for
    f06tf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tff.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrices A and B.

    a : complex, array-like, shape (`m`, n)
        The m by n general or trapezoidal matrix A.

    Returns
    -------
    b : complex, ndarray, shape (`m`, n)
        The m by n general or trapezoidal matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zmcopy as b_zmcopy

    fun_name = 'naginterfaces.library.blas.zmcopy'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'),),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _b = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_m_sorder,
    )
    b_zmcopy(
        matrix,
        _m,
        _n,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def zmload(matrix, m, n, con, diag):
    u"""
    Matrix initialization, complex rectangular matrix.

    ``zmload`` forms the complex m by n rectangular or trapezoidal
    matrix A given by

        a_{ij} = {[table omitted].

    For full information please refer to the NAG Library document for
    f06th

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06thf.html

    Parameters
    ----------
    matrix : str, length 1
        The matrix type.

        `matrix` = 'G'

          General matrix.

        `matrix` = 'U'

          Upper trapezoidal matrix (upper triangular if m = n).

        `matrix` = 'L'

          Lower trapezoidal matrix (lower triangular if m = n).

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    con : complex
        The value to be assigned to the off-diagonal elements of A.

    diag : complex
        The value to be assigned to the diagonal elements of A.

    Returns
    -------
    a : complex, ndarray, shape (`m`, `n`)
        The m by n general or trapezoidal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `matrix`.

            Constraint: `matrix` = 'G', 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zmload as b_zmload

    fun_name = 'naginterfaces.library.blas.zmload'

    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _a = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_zmload(
        matrix,
        _m,
        _n,
        con,
        diag,
        _a,
    )
    _a_py = _a
    return _a_py

def zhesrc(uplo, pivot, direct, k1, k2, c, s, a):
    u"""
    Unitary similarity transformation of Hermitian matrix as a sequence
    of plane rotations.

    ``zhesrc`` performs a Unitary similarity transformation (as a
    sequence of plane rotations) of a complex Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06tm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tmf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The transformed matrix A. The imaginary parts of the diagonal
        elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zhesrc as b_zhesrc

    fun_name = 'naginterfaces.library.blas.zhesrc'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
    )
    b_zhesrc(
        uplo,
        pivot,
        direct,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _a_py = _a
    return _a_py

def zutr1(alpha, x, y, a):
    u"""
    QR factorization by sequence of plane rotations, rank-1 update of
    complex upper triangular matrix.

    ``zutr1`` performs a QR factorization (as a sequence of plane
    rotations) of a complex upper triangular matrix that has been
    modified by a rank-1 update.

    For full information please refer to the NAG Library document for
    f06tp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tpf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    y : complex, array-like, shape (n)
        The n-element vector y. y_i must be stored in `y`[1+(i-1)*incy],
        for i = 1,2,...,n.

        Intermediate elements of `y` are not referenced.

    a : complex, array-like, shape (n, n)
        The n by n upper triangular matrix U. The imaginary parts of the
        diagonal elements must be zero.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The referenced elements are overwritten by details of the
        sequence of plane rotations.

    a : complex, ndarray, shape (n, n)
        The upper triangular matrix R. The imaginary parts of the
        diagonal elements must be zero.

    c : float, ndarray, shape (n-1)
        The cosines of the rotations Q_k, for k = 1,2,...,n-1.

    s : complex, ndarray, shape (n)
        The sines of the rotations Q_k, for k = 1,2,...,n-1; `s`[n]
        holds d_n, the nth diagonal element of D.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

        (`errno` 6)
            On entry, error in parameter incy.

            Constraint: incy > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zutr1 as b_zutr1

    fun_name = 'naginterfaces.library.blas.zutr1'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
    )
    _incy = 1
    _incy_py = _incy
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py-1,),
    )
    _s = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_zutr1(
        _n,
        alpha,
        _x,
        _incx,
        _y,
        _incy,
        _a,
        _c,
        _s,
    )
    _x_py = _x
    _a_py = _a
    _c_py = _c
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'a', 'c', 's'],
    )(
        _x_py,
        _a_py,
        _c_py,
        _s_py,
    )
    return _return_data

def zutupd(alpha, x, a):
    u"""
    QR factorization by sequence of plane rotations, complex upper
    triangular matrix augmented by a full row.

    ``zutupd`` performs a QR factorization (as a sequence of plane
    rotations) of a complex upper triangular matrix that has been
    augmented by a full row.

    For full information please refer to the NAG Library document for
    f06tq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tqf.html

    Parameters
    ----------
    alpha : complex
        The scalar alpha.

    x : complex, array-like, shape (n)
        The n-element vector x. x_i must be stored in `x`[1+(i-1)*incx],
        for i = 1,2,...,n.

        Intermediate elements of `x` are not referenced.

    a : complex, array-like, shape (n, n)
        The n by n upper triangular matrix U.

    Returns
    -------
    x : complex, ndarray, shape (n)
        The referenced elements are overwritten by details of the
        sequence of plane rotations.

    a : complex, ndarray, shape (n, n)
        The upper triangular matrix R.

    c : float, ndarray, shape (n)
        The values c_k, the cosines of the rotations Q_k, for k =
        1,2,...,n.

    s : complex, ndarray, shape (n)
        The values s_k, the sines of the rotations Q_k, for k =
        1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter incx.

            Constraint: incx > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zutupd as b_zutupd

    fun_name = 'naginterfaces.library.blas.zutupd'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _incx = 1
    _incx_py = _incx
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _s = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_zutupd(
        _n,
        alpha,
        _x,
        _incx,
        _a,
        _c,
        _s,
    )
    _x_py = _x
    _a_py = _a
    _c_py = _c
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'a', 'c', 's'],
    )(
        _x_py,
        _a_py,
        _c_py,
        _s_py,
    )
    return _return_data

def zuhqr(side, k1, k2, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, complex upper
    Hessenberg matrix.

    ``zuhqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a complex upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06tr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06trf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    s : float, array-like, shape (`k2`-1)
        The nonzero subdiagonal elements of H: `s`[k] must hold
        h_{k+1,k}, for k = k_1,...,k_2-1.

    a : complex, array-like, shape (n, n)
        The upper triangular part of the n by n upper Hessenberg matrix
        H.

    Returns
    -------
    c : complex, ndarray, shape (`k2`)
        `c`[k] holds c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1; `c`[k_2] holds d_k_2, the k_2th diagonal element
        of D, if `side` = 'L', or d_k_1, the k_1th diagonal element of
        D, if `side` = 'R'.

    s : float, ndarray, shape (`k2`-1)
        `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, ndarray, shape (n, n)
        The upper triangular matrix R. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zuhqr as b_zuhqr

    fun_name = 'naginterfaces.library.blas.zuhqr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k2 = _b_utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = _b_utils._EngineIntScalarType.to_py(_k2)
    _c = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_k2_py,),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_zuhqr(
        side,
        _n,
        k1,
        _k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _c_py = _c
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's', 'a'],
    )(
        _c_py,
        _s_py,
        _a_py,
    )
    return _return_data

def zusqr(side, k1, k2, s, a):
    u"""
    QR or RQ factorization by sequence of plane rotations, complex upper
    spiked matrix.

    ``zusqr`` performs a QR or RQ factorization (as a sequence of plane
    rotations) of a complex upper spiked matrix.

    For full information please refer to the NAG Library document for
    f06ts

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tsf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether H is operated on from the left or the right.

        `side` = 'L'

          H is pre-multiplied from the left.

        `side` = 'R'

          H is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    s : complex, array-like, shape (`k2`)
        The nonzero elements of the spike of H: `s`[k] must hold
        h_{k_2,k} if `side` = 'L', and h_{k+1,k_1} if `side` = 'R', for
        k = k_1,...,k_2-1.

    a : complex, array-like, shape (n, n)
        The upper triangular part of the n by n upper spiked matrix H.
        The imaginary parts of the diagonal elements must be zero,
        except for the (k_2,k_2) element if `side` = 'L', or the
        (k_1,k_1) element if `side` = 'R'.

    Returns
    -------
    c : float, ndarray, shape (`k2`-1)
        `c`[k] holds c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, ndarray, shape (`k2`)
        `s`[k] holds s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1; `s`[k_2] holds d_k_2, the k_2th diagonal element
        of D, if `side` = 'L', or d_k_1, the k_1th diagonal element of
        D, if `side` = 'R'.

    a : complex, ndarray, shape (n, n)
        The upper triangular matrix R. The imaginary parts of the
        diagonal elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zusqr as b_zusqr

    fun_name = 'naginterfaces.library.blas.zusqr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k2 = _b_utils._EngineIntScalarType(
        k2, {'fun_name': fun_name, 'entity': 'k2'},
    ).to_ctypes()
    _k2_py = _b_utils._EngineIntScalarType.to_py(_k2)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k2_py-1,),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
    )
    b_zusqr(
        side,
        _n,
        k1,
        _k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _c_py = _c
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's', 'a'],
    )(
        _c_py,
        _s_py,
        _a_py,
    )
    return _return_data

def zutsqr(side, k1, k2, c, s, a):
    u"""
    QR factorization of UP or RQ factorization of PU, U complex upper
    triangular, P a sequence of plane rotations.

    ``zutsqr`` performs a QR or RQ factorization of the product of a
    complex upper triangular matrix and a complex matrix of plane
    rotations.

    For full information please refer to the NAG Library document for
    f06tt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ttf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether P is applied from the left or the right in the
        transformation.

        `side` = 'L'

          P is applied from the left.

        `side` = 'R'

          P is applied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, array-like, shape (`k2`-1)
        `s`[k] must hold the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, array-like, shape (n, n)
        The n by n upper triangular matrix U.

    Returns
    -------
    c : float, ndarray, shape (`k2`-1)
        `c`[k] holds the cosine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    s : complex, ndarray, shape (`k2`-1)
        `s`[k] holds the sine of the rotation Q_k, for k =
        k_1,...,k_2-1.

    a : complex, ndarray, shape (n, n)
        The upper triangular matrix R.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zutsqr as b_zutsqr

    fun_name = 'naginterfaces.library.blas.zutsqr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
    )
    b_zutsqr(
        side,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _c_py = _c
    _s_py = _s
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 's', 'a'],
    )(
        _c_py,
        _s_py,
        _a_py,
    )
    return _return_data

def zutsrh(side, k1, k2, c, s, a):
    u"""
    Compute upper Hessenberg matrix by sequence of plane rotations,
    complex upper triangular matrix.

    ``zutsrh`` transforms a complex upper triangular matrix to an upper
    Hessenberg matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06tv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tvf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : complex, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, array-like, shape (n, n)
        The n by n upper triangular matrix U. The imaginary parts of the
        diagonal elements must be zero.

    Returns
    -------
    s : float, ndarray, shape (`k2`-1)
        `s`[k] holds h_{k+1,k}, the subdiagonal element of H, for k =
        k_1,...,k_2-1.

    a : complex, ndarray, shape (n, n)
        The upper triangular part of the upper Hessenberg matrix H.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zutsrh as b_zutsrh

    fun_name = 'naginterfaces.library.blas.zutsrh'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_zutsrh(
        side,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'a'],
    )(
        _s_py,
        _a_py,
    )
    return _return_data

def zutsrs(side, k1, k2, c, s, a):
    u"""
    Compute upper spiked matrix by sequence of plane rotations, complex
    upper triangular matrix.

    ``zutsrs`` transforms a complex upper triangular matrix to an upper
    spiked matrix by applying a given sequence of plane rotations.

    For full information please refer to the NAG Library document for
    f06tw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06twf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether U is operated on from the left or the right.

        `side` = 'L'

          U is pre-multiplied from the left.

        `side` = 'R'

          U is post-multiplied from the right.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, array-like, shape (n, n)
        The n by n upper triangular matrix U. The imaginary parts of the
        diagonal elements must be zero.

    Returns
    -------
    s : complex, ndarray, shape (`k2`-1)
        `s`[k] holds a nonzero element of the spike of H: h_{k_2,k} if
        `side` = 'L', or h_{k+1,k_1} if `side` = 'R', for k =
        k_1,...,k_2-1.

    a : complex, ndarray, shape (n, n)
        The upper triangular part of the upper spiked matrix H. The
        imaginary parts of the diagonal elements are set to zero except
        for the (k_2,k_2) element if `side` = 'L', or the (k_1,k_1)
        element if `side` = 'R'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zutsrs as b_zutsrs

    fun_name = 'naginterfaces.library.blas.zutsrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
    )
    b_zutsrs(
        side,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _s_py = _s
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'a'],
    )(
        _s_py,
        _a_py,
    )
    return _return_data

def zgesrc(side, pivot, direct, m, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, complex rectangular matrix, real
    cosine and complex sine.

    ``zgesrc`` applies to a complex rectangular matrix a sequence of
    plane rotations having real cosines and complex sines.

    For full information please refer to the NAG Library document for
    f06tx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06txf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : complex, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, array-like, shape (`m`, n)
        The m by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (`m`, n)
        The transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zgesrc as b_zgesrc

    fun_name = 'naginterfaces.library.blas.zgesrc'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
    )
    b_zgesrc(
        side,
        pivot,
        direct,
        m,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _a_py = _a
    return _a_py

def zgesrs(side, pivot, direct, m, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, complex rectangular matrix,
    complex cosine and real sine.

    ``zgesrs`` applies to a complex rectangular matrix a sequence of
    plane rotations having real sines and complex cosines.

    For full information please refer to the NAG Library document for
    f06ty

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06tyf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : complex, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, array-like, shape (`m`, n)
        The m by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (`m`, n)
        The transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zgesrs as b_zgesrs

    fun_name = 'naginterfaces.library.blas.zgesrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_zgesrs(
        side,
        pivot,
        direct,
        m,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _a_py = _a
    return _a_py

def zlange(norm, m, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex general matrix.

    ``zlange`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex m by n matrix.

    For full information please refer to the NAG Library document for
    f06ua

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uaf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    m : int
        m, the number of rows of the matrix A.

        When `m` = 0, ``zlange`` is set to zero.

    a : complex, array-like, shape (`m`, n)
        The m by n matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex m
        by n matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlange as b_zlange

    fun_name = 'naginterfaces.library.blas.zlange'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlange(
        norm,
        m,
        _n,
        _a,
    )
    return _return_scalar

def zlangb(norm, kl, ku, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex band matrix.

    ``zlangb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n band matrix.

    For full information please refer to the NAG Library document for
    f06ub

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ubf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    kl : int
        k_l, the number of subdiagonals within the band of A.

    ku : int
        k_u, the number of superdiagonals within the band of A.

    ab : complex, array-like, shape (`kl`+`ku`+1, n)
        The n by n band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` 4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlangb as b_zlangb

    fun_name = 'naginterfaces.library.blas.zlangb'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlangb(
        norm,
        _n,
        kl,
        ku,
        _ab,
    )
    return _return_scalar

def zlanhe(norm, uplo, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian matrix.

    ``zlanhe`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian matrix.

    For full information please refer to the NAG Library document for
    f06uc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ucf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a Hermitian matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlanhe as b_zlanhe

    fun_name = 'naginterfaces.library.blas.zlanhe'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlanhe(
        norm,
        uplo,
        _n,
        _a,
    )
    return _return_scalar

def zlanhp(norm, uplo, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian matrix, packed storage.

    ``zlanhp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06ud

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06udf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a Hermitian matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanhp`` returns zero.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlanhp as b_zlanhp

    fun_name = 'naginterfaces.library.blas.zlanhp'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = b_zlanhp(
        norm,
        uplo,
        n,
        _ap,
    )
    return _return_scalar

def zlanhb(norm, uplo, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian band matrix.

    ``zlanhb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian band matrix.

    For full information please refer to the NAG Library document for
    f06ue

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uef.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a Hermitian matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : complex, array-like, shape (`k`+1, n)
        The n by n Hermitian band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlanhb as b_zlanhb

    fun_name = 'naginterfaces.library.blas.zlanhb'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlanhb(
        norm,
        uplo,
        _n,
        k,
        _ab,
    )
    return _return_scalar

def zlansy(norm, uplo, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex symmetric matrix.

    ``zlansy`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n symmetric matrix.

    For full information please refer to the NAG Library document for
    f06uf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uff.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    a : complex, array-like, shape (n, n)
        The n by n symmetric matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n symmetric matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlansy as b_zlansy

    fun_name = 'naginterfaces.library.blas.zlansy'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlansy(
        norm,
        uplo,
        _n,
        _a,
    )
    return _return_scalar

def zlansp(norm, uplo, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex symmetric matrix, packed storage.

    ``zlansp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n symmetric matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06ug

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ugf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlansp`` returns zero.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n symmetric matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlansp as b_zlansp

    fun_name = 'naginterfaces.library.blas.zlansp'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = b_zlansp(
        norm,
        uplo,
        n,
        _ap,
    )
    return _return_scalar

def zlansb(norm, uplo, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex symmetric band matrix.

    ``zlansb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n symmetric band matrix.

    For full information please refer to the NAG Library document for
    f06uh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06uhf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm (= the 1-norm for a symmetric matrix).

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : complex, array-like, shape (`k`+1, n)
        The n by n symmetric band matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n symmetric band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlansb as b_zlansb

    fun_name = 'naginterfaces.library.blas.zlansb'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlansb(
        norm,
        uplo,
        _n,
        k,
        _ab,
    )
    return _return_scalar

def zlantr(norm, uplo, diag, m, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex trapezoidal/triangular matrix.

    ``zlantr`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex m by n trapezoidal matrix (triangular if m = n).

    For full information please refer to the NAG Library document for
    f06uj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ujf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower trapezoidal.

        `uplo` = 'U'

          A is upper trapezoidal.

        `uplo` = 'L'

          A is lower trapezoidal.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix A.

        When `m` = 0, ``zlantr`` is set to zero.

    a : complex, array-like, shape (`m`, n)
        The m by n trapezoidal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex m
        by n trapezoidal matrix (triangular if m = n).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlantr as b_zlantr

    fun_name = 'naginterfaces.library.blas.zlantr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlantr(
        norm,
        uplo,
        diag,
        m,
        _n,
        _a,
    )
    return _return_scalar

def zlantp(norm, uplo, diag, n, ap):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex triangular matrix, packed storage.

    ``zlantp`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n triangular matrix, stored in packed form.

    For full information please refer to the NAG Library document for
    f06uk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ukf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlantp`` returns zero.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n triangular matrix, stored in packed form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlantp as b_zlantp

    fun_name = 'naginterfaces.library.blas.zlantp'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = b_zlantp(
        norm,
        uplo,
        diag,
        n,
        _ap,
    )
    return _return_scalar

def zlantb(norm, uplo, diag, k, ab):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex triangular band matrix.

    ``zlantb`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n triangular band matrix.

    For full information please refer to the NAG Library document for
    f06ul

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ulf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    k : int
        k, the number of subdiagonals or superdiagonals of the matrix A.

    ab : complex, array-like, shape (`k`+1, n)
        The n by n triangular band matrix A

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n triangular band matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlantb as b_zlantb

    fun_name = 'naginterfaces.library.blas.zlantb'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlantb(
        norm,
        uplo,
        diag,
        _n,
        k,
        _ab,
    )
    return _return_scalar

def zlanhs(norm, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hessenberg matrix.

    ``zlanhs`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n upper Hessenberg matrix.

    For full information please refer to the NAG Library document for
    f06um

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06umf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    a : complex, array-like, shape (n, n)
        The n by n upper Hessenberg matrix A; elements of the array
        below the first subdiagonal are not referenced.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n upper Hessenberg matrix.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlanhs as b_zlanhs

    fun_name = 'naginterfaces.library.blas.zlanhs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_zlanhs(
        norm,
        _n,
        _a,
    )
    return _return_scalar

def zlangt(norm, dl, d, du):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex tridiagonal matrix.

    ``zlangt`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06un

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06unf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    dl : complex, array-like, shape (n-1)
        The (n-1) subdiagonal elements of A.

    d : complex, array-like, shape (n)
        The n diagonal elements of A.

    du : complex, array-like, shape (n-1)
        The (n-1) superdiagonal elements of A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlangt as b_zlangt

    fun_name = 'naginterfaces.library.blas.zlangt'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _return_scalar = b_zlangt(
        norm,
        _n,
        _dl,
        _d,
        _du,
    )
    return _return_scalar

def zlanht(norm, d, e):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian tridiagonal matrix.

    ``zlanht`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex n by n Hermitian tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f06up

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06upf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : complex, array-like, shape (n-1)
        The (n-1) subdiagonal or superdiagonal elements of the
        tridiagonal matrix A.

    Returns
    -------
    nrm : float
        The value of the 1-norm, the infinity-norm, the Frobenius norm,
        or the maximum absolute value of the elements of the complex n
        by n Hermitian tridiagonal matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlanht as b_zlanht

    fun_name = 'naginterfaces.library.blas.zlanht'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _return_scalar = b_zlanht(
        norm,
        _n,
        _d,
        _e,
    )
    return _return_scalar

def zgeap(side, trans, perm, k, b):
    u"""
    Permute rows or columns, complex rectangular matrix, permutations
    represented by an integer array.

    ``zgeap`` permutes the rows or columns of a complex rectangular
    matrix using an int array of permutations.

    For full information please refer to the NAG Library document for
    f06vj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06vjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    perm : int, array-like, shape (n)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(n, m); if `side`='R':
        `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R': max(n,
        max(`perm`)); otherwise: 0.

        The original matrix B; B is m by k if `side` = 'L', or k by m if
        `side` = 'R'.

    Returns
    -------
    b : complex, ndarray, shape (:, :)
        The permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zgeap as b_zgeap

    fun_name = 'naginterfaces.library.blas.zgeap'

    _perm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'perm'},
        perm,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_perm, {'fun_name': fun_name, 'entity': 'perm'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _vl_m_py = max(_perm[0:max(1, _n_py)-1+1])
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    b_zgeap(
        side,
        trans,
        _n,
        _perm,
        k,
        _b,
    )
    _b_py = _b
    return _b_py

def zgeapr(side, trans, perm, k, b):
    u"""
    Permute rows or columns, complex rectangular matrix, permutations
    represented by a real array.

    ``zgeapr`` permutes the rows or columns of a complex rectangular
    matrix using a double array of permutations.

    For full information please refer to the NAG Library document for
    f06vk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06vkf.html

    Parameters
    ----------
    side : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    trans : str, length 1
        Specifies the operation to be performed.

        `side` = 'L' and `trans` = 'T'

          B←P^TB.

        `side` = 'L' and `trans` = 'N'

          B←PB.

        `side` = 'R' and `trans` = 'T'

          B←BP^T.

        `side` = 'R' and `trans` = 'N'

          B←BP.

    perm : float, array-like, shape (n)
        The n indices p_i which define the interchanges in the
        representation of P. It is usual to have p_i >= i, but this is
        not necessary.

    k : int
        k, the number of columns of B if `side` = 'L', or the number of
        rows of B if `side` = 'R'.

    b : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': max(n, m); if `side`='R':
        `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `k`; if `side`='R': max(n,
        int(max(`perm`))); otherwise: 0.

        The original matrix B; B is m by k if `side` = 'L', or k by m if
        `side` = 'R'.

    Returns
    -------
    b : complex, ndarray, shape (:, :)
        The permuted matrix B.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `perm`.

            Constraint: 1 <= `perm`[i-1] <= m.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zgeapr as b_zgeapr

    fun_name = 'naginterfaces.library.blas.zgeapr'

    _perm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'perm'},
        perm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_perm, {'fun_name': fun_name, 'entity': 'perm'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _vl_m_py = int(max(_perm[0:max(1, _n_py)-1+1]))
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    b_zgeapr(
        side,
        trans,
        _n,
        _perm,
        k,
        _b,
    )
    _b_py = _b
    return _b_py

def zsgesr(side, pivot, direct, m, k1, k2, c, s, a):
    u"""
    Apply sequence of plane rotations, complex rectangular matrix, real
    cosine and sine.

    ``zsgesr`` applies to a complex rectangular matrix a sequence of
    plane rotations having real cosines and complex sines.

    For full information please refer to the NAG Library document for
    f06vx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06vxf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether A is operated on from the left or the right.

        `side` = 'L'

          A is pre-multiplied from the left.

        `side` = 'R'

          A is post-multiplied from the right.

    pivot : str, length 1
        Specifies the plane rotated by P_k.

        `pivot` = 'V' (variable pivot)

          P_k rotates the (k,k+1) plane.

        `pivot` = 'T' (top pivot)

          P_k rotates the (k_1,k+1) plane.

        `pivot` = 'B' (bottom pivot)

          P_k rotates the (k,k_2) plane.

    direct : str, length 1
        Specifies the sequence direction.

        `direct` = 'F' (forward sequence)

          P = P_{k_2-1}...P_{k_1+1}P_k_1.

        `direct` = 'B' (backward sequence)

          P = P_k_1 P_{k_1+1}...P_{k_2-1}.

    m : int
        m, the number of rows of the matrix A.

    k1 : int
        The value k_1.

    k2 : int
        The value k_2.

    c : float, array-like, shape (`k2`-1)
        `c`[k] must hold c_k, the cosine of the rotation P_k, for k =
        k_1,...,k_2-1.

    s : float, array-like, shape (`k2`-1)
        `s`[k] must hold s_k, the sine of the rotation P_k, for k =
        k_1,...,k_2-1.

    a : complex, array-like, shape (`m`, n)
        The m by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (`m`, n)
        The transformed matrix A.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `pivot`.

            Constraint: `pivot` = 'V', 'T' or 'B'.

        (`errno` 3)
            On entry, error in parameter `direct`.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 5)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zsgesr as b_zsgesr

    fun_name = 'naginterfaces.library.blas.zsgesr'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_zsgesr(
        side,
        pivot,
        direct,
        m,
        _n,
        k1,
        k2,
        _c,
        _s,
        _a,
    )
    _a_py = _a
    return _a_py

def dlansf(norm, transr, uplo, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    real symmetric matrix, Rectangular Full Packed format.

    ``dlansf`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    real symmetric matrix A stored in Rectangular Full Packed (RFP)
    format.

    For full information please refer to the NAG Library document for
    f06wa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06waf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'T'

          The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``dlansf`` returns zero.

    a : float, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n symmetric matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    nrm : float
        The infinity-norm, the Frobenius norm, or the maximum absolute
        value of the elements of the real symmetric matrix A stored in
        Rectangular Full Packed (RFP) format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dlansf as b_dlansf

    fun_name = 'naginterfaces.library.blas.dlansf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_dlansf(
        norm,
        transr,
        uplo,
        n,
        _a,
    )
    return _return_scalar

def dtfsm(transr, side, uplo, trans, diag, alpha, a, b):
    u"""
    Solves a system of equations with multiple right-hand sides, real
    triangular coefficient matrix, Rectangular Full Packed format.

    ``dtfsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a real triangular matrix stored in Rectangular Full
    Packed (RFP) format, B is an m by n real matrix, and alpha is a real
    scalar. A^inv denotes (A^T)^-1 or equivalently (A^-1)^T.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06wb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wbf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'T'

          The matrix A is stored in transposed RFP format.

    side : str, length 1
        Specifies whether B is operated on from the left or the right,
        or similarly whether A (or its transpose) appears to the left or
        right of the solution matrix in the linear system to be solved.

        `side` = 'L'

          B is pre-multiplied from the left. The system to be solved has
          the form AX = alpha B or A^TX = alpha B.

        `side` = 'R'

          B is post-multiplied from the right. The system to be solved
          has the form XA = alpha B or XA^T = alpha B.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies whether the operation involves A^-1 or A^invT, i.e.,
        whether or not A is transposed in the linear system to be
        solved.

        `trans` = 'N'

          The operation involves A^-1, i.e., A is not transposed.

        `trans` = 'T'

          The operation involves A^invT, i.e., A is transposed.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements of A are stored explicitly.

        `diag` = 'U'

          The diagonal elements of A are assumed to be 1, the
          corresponding elements of `a` are not referenced.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `side`='L': m*(m+1)/2; if `side`='R': n*(n+1)/2;
        otherwise: 0.

        A, the m by m triangular matrix A if `side` = 'L' or the n by n
        triangular matrix A if `side` = 'R', stored in RFP format (as
        specified by `transr`). The storage format is described in
        detail in [the F07 Introduction]. If `alpha` = 0.0, `a` is not
        referenced.

    b : float, array-like, shape (m, n)
        The m by n matrix B.

        If `alpha` = 0, `b` need not be set.

    Returns
    -------
    b : float, ndarray, shape (m, n)
        The updated matrix B, or similarly the solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` 2)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 5)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 6)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` 7)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dtfsm as b_dtfsm

    fun_name = 'naginterfaces.library.blas.dtfsm'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    b_dtfsm(
        transr,
        side,
        uplo,
        trans,
        diag,
        _m,
        _n,
        alpha,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def dsfrk(transr, uplo, trans, k, alpha, a, beta, c):
    u"""
    Rank-k update of a real symmetric matrix, Rectangular Full Packed
    format.

    ``dsfrk`` performs one of the symmetric rank-k update operations

        C←alpha AA^T+beta C or C←alpha A^TA+beta C,

    where A is a real matrix, C is an n by n real symmetric matrix
    stored in Rectangular Full Packed (RFP) format, and alpha and beta
    are real scalars.

    For full information please refer to the NAG Library document for
    f06wc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wcf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of C is normal or
        transposed.

        `transr` = 'N'

          The matrix C is stored in normal RFP format.

        `transr` = 'T'

          The matrix C is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored in RFP format.

        `uplo` = 'U'

          The upper triangular part of C is stored in RFP format.

        `uplo` = 'L'

          The lower triangular part of C is stored in RFP format.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^T+beta C.

        `trans` = 'T'

          C←alpha A^TA+beta C.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'T'.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': n;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'. If `alpha` = 0.0, `a` is not referenced.

    beta : float
        The scalar beta.

    c : float, array-like, shape (n*(n+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n symmetric matrix C, stored in RFP format (as
        specified by `transr`). The storage format is described in
        detail in [the F07 Introduction].

    Returns
    -------
    c : float, ndarray, shape (n*(n+1)/2)
        The updated matrix C, that is its upper or lower triangular part
        stored in RFP format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import dsfrk as b_dsfrk

    fun_name = 'naginterfaces.library.blas.dsfrk'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dsfrk(
        transr,
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def zlanhf(norm, transr, uplo, n, a):
    u"""
    1-norm, infinity-norm, Frobenius norm, largest absolute element,
    complex Hermitian matrix, Rectangular Full Packed format.

    ``zlanhf`` returns the value of the 1-norm, the infinity-norm, the
    Frobenius norm, or the maximum absolute value of the elements of a
    complex Hermitian matrix A stored in Rectangular Full Packed (RFP)
    format.

    For full information please refer to the NAG Library document for
    f06wn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wnf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the value to be returned.

        `norm` = '1' or 'O'

          The 1-norm.

        `norm` = 'I'

          The infinity-norm.

        `norm` = 'F' or 'E'

          The Frobenius (or Euclidean) norm.

        `norm` = 'M'

          The value max_{i,j}(|a_{ij}|) (not a norm).

    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'C'

          The conjugate transpose of the RFP representation of the
          matrix A is stored.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    n : int
        n, the order of the matrix A.

        When `n` = 0, ``zlanhf`` returns zero.

    a : complex, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n Hermitian matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    nrm : float
        The infinity-norm, the Frobenius norm, or the maximum absolute
        value of the elements of the complex Hermitian matrix A stored
        in Rectangular Full Packed (RFP) format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O', 'I', 'F', 'E' or 'M'.

        (`errno` 2)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zlanhf as b_zlanhf

    fun_name = 'naginterfaces.library.blas.zlanhf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _return_scalar = b_zlanhf(
        norm,
        transr,
        uplo,
        n,
        _a,
    )
    return _return_scalar

def ztfsm(transr, side, uplo, trans, diag, alpha, a, b):
    u"""
    Solves system of equations with multiple right-hand sides, complex
    triangular coefficient matrix, Rectangular Full Packed format.

    ``ztfsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a complex triangular matrix stored in Rectangular Full
    Packed (RFP) format, B is an m by n complex matrix, and alpha is a
    complex scalar. A^invH denotes (A^H)^-1 or equivalently (A^-1)^H.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06wp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wpf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

          The matrix A is stored in normal RFP format.

        `transr` = 'C'

          The conjugate transpose of the RFP representation of the
          matrix A is stored.

    side : str, length 1
        Specifies whether B is operated on from the left or the right,
        or similarly whether A (or its transpose) appears to the left or
        right of the solution matrix in the linear system to be solved.

        `side` = 'L'

          B is pre-multiplied from the left. The system to be solved has
          the form AX = alpha B or A^HX = alpha B.

        `side` = 'R'

          B is post-multiplied from the right. The system to be solved
          has the form XA = alpha B or XA^H = alpha B.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    trans : str, length 1
        Specifies whether the operation involves A^-1 or A^invH, i.e.,
        whether or not A is transpose conjugated in the linear system to
        be solved.

        `trans` = 'N'

          The operation involves A^-1, i.e., A is not transpose
          conjugated.

        `trans` = 'C'

          The operation involves A^invH, i.e., A is transpose
          conjugated.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements of A are stored explicitly.

        `diag` = 'U'

          The diagonal elements of A are assumed to be 1, the
          corresponding elements of `a` are not referenced.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `side`='L': m*(m+1)/2; if `side`='R': n*(n+1)/2;
        otherwise: 0.

        A, the m by m triangular matrix A if `side` = 'L' or the n by n
        triangular matrix A if `side` = 'R', stored in RFP format (as
        specified by `transr`). The storage format is described in
        detail in [the F07 Introduction]. If `alpha` = 0.0, `a` is not
        referenced.

    b : complex, array-like, shape (m, n)
        The m by n matrix B.

        If `alpha` = 0, `b` need not be set.

    Returns
    -------
    b : complex, ndarray, shape (m, n)
        The updated matrix B, or similarly the solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` 2)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 3)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 4)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 5)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 6)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` 7)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import ztfsm as b_ztfsm

    fun_name = 'naginterfaces.library.blas.ztfsm'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    b_ztfsm(
        transr,
        side,
        uplo,
        trans,
        diag,
        _m,
        _n,
        alpha,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def zhfrk(transr, uplo, trans, k, alpha, a, beta, c):
    u"""
    Rank-k update of a complex Hermitian matrix, Rectangular Full Packed
    format.

    ``zhfrk`` performs one of the Hermitian rank-k update operations

        C←alpha AA^H+beta C or C←alpha A^HA+beta C,

    where A is a complex matrix, C is an n by n complex Hermitian matrix
    stored in Rectangular Full Packed (RFP) format, and alpha and beta
    are real scalars.

    For full information please refer to the NAG Library document for
    f06wq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06wqf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of C or its
        conjugate transpose is stored.

        `transr` = 'N'

          The matrix C is stored in normal RFP format.

        `transr` = 'C'

          The conjugate transpose of the RFP representation of the
          matrix C is stored.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored in RFP format.

        `uplo` = 'U'

          The upper triangular part of C is stored in RFP format.

        `uplo` = 'L'

          The lower triangular part of C is stored in RFP format.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^H+beta C.

        `trans` = 'C'

          C←alpha A^HA+beta C.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'C'.

    alpha : float
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': max(1, n); if
        `trans`='C': max(1, `k`); otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': max(1, `k`); if
        `trans`='C': max(1, n); otherwise: 1.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'. If `alpha` = 0.0, `a` is not referenced.

    beta : float
        The scalar beta.

    c : complex, array-like, shape (n*(n+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n Hermitian matrix C, stored in RFP format (as
        specified by `transr`). The storage format is described in
        detail in [the F07 Introduction].

    Returns
    -------
    c : complex, ndarray, shape (n*(n+1)/2)
        The updated matrix C, that is its upper or lower triangular part
        stored in RFP format.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.blas import zhfrk as b_zhfrk

    fun_name = 'naginterfaces.library.blas.zhfrk'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    b_zhfrk(
        transr,
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def dgemm(transa, transb, alpha, a, b, beta, c=None):
    u"""
    Matrix-matrix product, two real rectangular matrices.

    ``dgemm`` performs one of the matrix-matrix operations

        [table omitted]

    where A, B and C are real matrices, and alpha and beta are real
    scalars; C is always m by n.

    For full information please refer to the NAG Library document for
    f06ya

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yaf.html

    Parameters
    ----------
    transa : str, length 1
        Specifies whether the operation involves A or A^T.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T' or 'C'

          The operation involves A^T.

    transb : str, length 1
        Specifies whether the operation involves B or B^T.

        `transb` = 'N'

          The operation involves B.

        `transb` = 'T' or 'C'

          The operation involves B^T.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transa`='N': m; if `transa` in ('T',
        'C'): k; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transa`='N': k; if `transa` in ('T',
        'C'): m; otherwise: 0.

        The matrix A; A is m by k if `transa` = 'N', or k by m if
        `transa` = 'T' or 'C'.

    b : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transb`='N': k; if `transb` in ('T',
        'C'): n; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transb`='N': n; if `transb` in ('T',
        'C'): k; otherwise: 0.

        The matrix B; B is k by n if `transb` = 'N', or n by k if
        `transb` = 'T' or 'C'.

    beta : float
        The scalar beta.

    c : None or float, array-like, shape (m, n), optional
        The m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

    Returns
    -------
    c : float, ndarray, shape (m, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `transb`.

            Constraint: `transb` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter k.

            Constraint: k >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dgemm as b_dgemm

    fun_name = 'naginterfaces.library.blas.dgemm'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_c, 'c')),
    )
    _transa = _b_utils._EngineCharScalarType(
        transa, {'fun_name': fun_name, 'entity': 'transa'}, exp_chrlen=1,
    ).to_ctypes()
    _transa_py = _b_utils._EngineCharScalarType.to_py(_transa)
    if _transa_py in ('T', 't', 'C', 'c'):
        _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
        _m = _b_utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
    elif _transa_py in ('N', 'n'):
        _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
        _m = _b_utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
    else:
        _m = 0
        _m_py = _m
    _transb = _b_utils._EngineCharScalarType(
        transb, {'fun_name': fun_name, 'entity': 'transb'}, exp_chrlen=1,
    ).to_ctypes()
    _transb_py = _b_utils._EngineCharScalarType.to_py(_transb)
    if _transb_py in ('T', 't', 'C', 'c'):
        _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    elif _transb_py in ('N', 'n'):
        _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    else:
        _n = 0
        _n_py = _n
    if _transa_py in ('T', 't', 'C', 'c'):
        _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
        _k = _b_utils._EngineIntScalarType(
            _k_py, {'fun_name': fun_name, 'entity': '_k'},
        ).to_ctypes()
    elif _transa_py in ('N', 'n'):
        _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
        _k = _b_utils._EngineIntScalarType(
            _k_py, {'fun_name': fun_name, 'entity': '_k'},
        ).to_ctypes()
    else:
        _k = 0
        _k_py = _k
    _beta = _b_utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineFloat64ScalarType.to_py(_beta)
    if _beta_py != 0.0:
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when beta != 0.0',
            )
    if c is None:
        _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_m_py, _n_py),
            sorder=_m_sorder,
        )
    b_dgemm(
        transa,
        transb,
        _m,
        _n,
        _k,
        alpha,
        _a,
        _b,
        _beta,
        _c,
    )
    _c_py = _c
    return _c_py

def dsymm(side, uplo, m, alpha, a, b, beta, c=None):
    u"""
    Matrix-matrix product, one real symmetric matrix, one real
    rectangular matrix.

    ``dsymm`` performs one of the matrix-matrix operations

        C←alpha AB+beta C or C←alpha BA+beta C,

    where A is a real symmetric matrix, B and C are m by n real
    matrices, and alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06yc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ycf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    m : int
        m, the number of rows of the matrices B and C; the order of A if
        `side` = 'L'.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        The symmetric matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : float, array-like, shape (`m`, n)
        The m by n matrix B.

    beta : float
        The scalar beta.

    c : None or float, array-like, shape (`m`, n), optional
        The m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

    Returns
    -------
    c : float, ndarray, shape (`m`, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dsymm as b_dsymm

    fun_name = 'naginterfaces.library.blas.dsymm'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_c, 'c')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _beta = _b_utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineFloat64ScalarType.to_py(_beta)
    if _beta_py != 0.0:
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when beta != 0.0',
            )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if c is None:
        _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_m_py, _n_py),
            sorder=_m_sorder,
        )
    b_dsymm(
        side,
        uplo,
        _m,
        _n,
        alpha,
        _a,
        _b,
        _beta,
        _c,
    )
    _c_py = _c
    return _c_py

def dtrmm(side, uplo, transa, diag, m, alpha, a, b):
    u"""
    Matrix-matrix product, one real triangular matrix, one real
    rectangular matrix.

    ``dtrmm`` performs one of the matrix-matrix operations

        [table omitted]

    where B is an m by n real matrix, A is a real triangular matrix, and
    alpha is a real scalar.

    For full information please refer to the NAG Library document for
    f06yf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yff.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A or A^T.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T' or 'C'

          The operation involves A^T.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : float, array-like, shape (`m`, n)
        The m by n matrix B.

        If `alpha` = 0, `b` need not be set.

    Returns
    -------
    b : float, ndarray, shape (`m`, n)
        The updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtrmm as b_dtrmm

    fun_name = 'naginterfaces.library.blas.dtrmm'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dtrmm(
        side,
        uplo,
        transa,
        diag,
        m,
        _n,
        alpha,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def dtrsm(side, uplo, transa, diag, m, alpha, a, b):
    u"""
    Solves a system of equations with multiple right-hand sides, real
    triangular coefficient matrix.

    ``dtrsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a real triangular matrix, B is an m by n real matrix, and
    alpha is a real scalar. A^inv denotes (A^T)^-1 or equivalently
    (A^-1)^T.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06yj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A^-1 or A^invT.

        `transa` = 'N'

          The operation involves A^-1.

        `transa` = 'T' or 'C'

          The operation involves A^invT.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : float, array-like, shape (`m`, n)
        The m by n matrix B.

        If `alpha` = 0, `b` need not be set.

    Returns
    -------
    b : float, ndarray, shape (`m`, n)
        The updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dtrsm as b_dtrsm

    fun_name = 'naginterfaces.library.blas.dtrsm'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dtrsm(
        side,
        uplo,
        transa,
        diag,
        m,
        _n,
        alpha,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def dsyrk(uplo, trans, k, alpha, a, beta, c):
    u"""
    Rank-k update of a real symmetric matrix.

    ``dsyrk`` performs one of the symmetric rank-k update operations

        C←alpha AA^T+beta C or C←alpha A^TA+beta C,

    where A is a real matrix, C is an n by n real symmetric matrix, and
    alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06yp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ypf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^T+beta C.

        `trans` = 'T' or 'C'

          C←alpha A^TA+beta C.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'T' or 'C'.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans` in ('T',
        'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans` in ('T',
        'C'): n; otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T' or 'C'.

    beta : float
        The scalar beta.

    c : float, array-like, shape (n, n)
        The n by n symmetric matrix C.

    Returns
    -------
    c : float, ndarray, shape (n, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dsyrk as b_dsyrk

    fun_name = 'naginterfaces.library.blas.dsyrk'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dsyrk(
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def dsyr2k(uplo, trans, k, alpha, a, b, beta, c):
    u"""
    Rank-2k update of a real symmetric matrix.

    ``dsyr2k`` performs one of the symmetric rank-2k update operations

        C←alpha AB^T+alpha BA^T+beta C or C←alpha A^TB+alpha B^TA+beta
        C,

    where A and B are real matrices, C is an n by n real symmetric
    matrix, and alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06yr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06yrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AB^T+alpha BA^T+beta C.

        `trans` = 'T' or 'C'

          C←alpha A^TB+alpha B^TA+beta C.

    k : int
        k, the number of columns of A and B if `trans` = 'N', or the
        number of rows of A and B if `trans` = 'T' or 'C'.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans` in ('T',
        'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans` in ('T',
        'C'): n; otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T' or 'C'.

    b : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans` in ('T',
        'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans` in ('T',
        'C'): n; otherwise: 0.

        The matrix B; B is n by k if `trans` = 'N', or k by n if `trans`
        = 'T' or 'C'.

    beta : float
        The scalar beta.

    c : float, array-like, shape (n, n)
        The n by n symmetric matrix C.

    Returns
    -------
    c : float, ndarray, shape (n, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import dsyr2k as b_dsyr2k

    fun_name = 'naginterfaces.library.blas.dsyr2k'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dsyr2k(
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        _b,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def zgemm(transa, transb, m, k, alpha, a, b, beta, c=None):
    u"""
    Matrix-matrix product, two complex rectangular matrices.

    ``zgemm`` performs one of the matrix-matrix operations

        [table omitted]

    where A, B and C are complex matrices, and alpha and beta are
    complex scalars; C is always m by n.

    For full information please refer to the NAG Library document for
    f06za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zaf.html

    Parameters
    ----------
    transa : str, length 1
        Specifies whether the operation involves A, A^T or A^H.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T'

          The operation involves A^T.

        `transa` = 'C'

          The operation involves A^H.

    transb : str, length 1
        Specifies whether the operation involves B, B^T or B^H.

        `transb` = 'N'

          The operation involves B.

        `transb` = 'T'

          The operation involves B^T.

        `transb` = 'C'

          The operation involves B^H.

    m : int
        m, the number of rows of the matrix C; the number of rows of A
        if `transa` = 'N', or the number of columns of A if `transa` =
        'T' or 'C'.

    k : int
        k, the number of columns of A if `transa` = 'N', or the number
        of rows of A if `transa` = 'T' or 'C'; the number of rows of B
        if `transb` = 'N', or the number of columns of B if `transb` =
        'T' or 'C'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transa`='N': `m`; if `transa` in
        ('T', 'C'): `k`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transa`='N': `k`; if `transa` in
        ('T', 'C'): `m`; otherwise: 0.

        The matrix A; A is m by k if `transa` = 'N', or k by m if
        `transa` = 'T' or 'C'.

    b : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transb`='N': `k`; if `transb` in
        ('T', 'C'): n; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transb`='N': n; if `transb` in ('T',
        'C'): `k`; otherwise: 0.

        The matrix B; B is k by n if `transb` = 'N', or n by k if
        `transb` = 'T' or 'C'.

    beta : complex
        The scalar beta.

    c : None or complex, array-like, shape (`m`, n), optional
        The m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

    Returns
    -------
    c : complex, ndarray, shape (`m`, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, error in parameter `transb`.

            Constraint: `transb` = 'N', 'T' or 'C'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 5)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zgemm as b_zgemm

    fun_name = 'naginterfaces.library.blas.zgemm'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_c, 'c')),
    )
    _transb = _b_utils._EngineCharScalarType(
        transb, {'fun_name': fun_name, 'entity': 'transb'}, exp_chrlen=1,
    ).to_ctypes()
    _transb_py = _b_utils._EngineCharScalarType.to_py(_transb)
    if _transb_py in ('T', 't', 'C', 'c'):
        _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    elif _transb_py in ('N', 'n'):
        _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    else:
        _n = 0
        _n_py = _n
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    if _beta_py != 0.0:
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when beta != 0.0',
            )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if c is None:
        _c = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            (_m_py, _n_py),
            sorder=_m_sorder,
        )
    b_zgemm(
        transa,
        transb,
        _m,
        _n,
        k,
        alpha,
        _a,
        _b,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def zhemm(side, uplo, m, alpha, a, b, beta, c=None):
    u"""
    Matrix-matrix product, one complex Hermitian matrix, one complex
    rectangular matrix.

    ``zhemm`` performs one of the matrix-matrix operations

        C←alpha AB+beta C or C←alpha BA+beta C,

    where A is a complex Hermitian matrix, B and C are m by n complex
    matrices, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zcf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    m : int
        m, the number of rows of the matrices B and C; the order of A if
        `side` = 'L'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        The Hermitian matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, array-like, shape (`m`, n)
        The m by n matrix B.

    beta : complex
        The scalar beta.

    c : None or complex, array-like, shape (`m`, n), optional
        The m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

    Returns
    -------
    c : complex, ndarray, shape (`m`, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zhemm as b_zhemm

    fun_name = 'naginterfaces.library.blas.zhemm'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_c, 'c')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    if _beta_py != 0.0:
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when beta != 0.0',
            )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if c is None:
        _c = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            (_m_py, _n_py),
            sorder=_m_sorder,
        )
    b_zhemm(
        side,
        uplo,
        _m,
        _n,
        alpha,
        _a,
        _b,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def ztrmm(side, uplo, transa, diag, m, alpha, a, b):
    u"""
    Matrix-matrix product, one complex triangular matrix, one complex
    rectangular matrix.

    ``ztrmm`` performs one of the matrix-matrix operations

        [table omitted]

    where B is an m by n complex matrix, A is a complex triangular
    matrix, and alpha is a complex scalar.

    For full information please refer to the NAG Library document for
    f06zf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zff.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A, A^T or A^H.

        `transa` = 'N'

          The operation involves A.

        `transa` = 'T'

          The operation involves A^T.

        `transa` = 'C'

          The operation involves A^H.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, array-like, shape (`m`, n)
        The m by n matrix B.

        If `alpha` = 0, `b` need not be set.

    Returns
    -------
    b : complex, ndarray, shape (`m`, n)
        The updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztrmm as b_ztrmm

    fun_name = 'naginterfaces.library.blas.ztrmm'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    b_ztrmm(
        side,
        uplo,
        transa,
        diag,
        m,
        _n,
        alpha,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def ztrsm(side, uplo, transa, diag, m, alpha, a, b):
    u"""
    Solves system of equations with multiple right-hand sides, complex
    triangular coefficient matrix.

    ``ztrsm`` performs one of the matrix-matrix operations

        [table omitted]

    where A is a complex triangular matrix, B is an m by n complex
    matrix, and alpha is a complex scalar. A^inv denotes (A^T)^-1 or
    equivalently (A^-1)^T; A^invH denotes (A^H)^-1 or equivalently
    (A^-1)^H.

    No test for singularity or near-singularity of A is included in this
    function.
    Such tests must be performed before calling this function.

    For full information please refer to the NAG Library document for
    f06zj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zjf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

          A is upper triangular.

        `uplo` = 'L'

          A is lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A^-1, A^invT or A^invH.

        `transa` = 'N'

          The operation involves A^-1.

        `transa` = 'T'

          The operation involves A^invT.

        `transa` = 'C'

          The operation involves A^invH.

    diag : str, length 1
        Specifies whether A has nonunit or unit diagonal elements.

        `diag` = 'N'

          The diagonal elements are stored explicitly.

        `diag` = 'U'

          The diagonal elements are assumed to be 1, and are not
          referenced.

    m : int
        m, the number of rows of the matrix B; the order of A if `side`
        = 'L'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        The triangular matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, array-like, shape (`m`, n)
        The m by n matrix B.

        If `alpha` = 0, `b` need not be set.

    Returns
    -------
    b : complex, ndarray, shape (`m`, n)
        The updated matrix B.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `transa`.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` 5)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 6)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import ztrsm as b_ztrsm

    fun_name = 'naginterfaces.library.blas.ztrsm'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    b_ztrsm(
        side,
        uplo,
        transa,
        diag,
        m,
        _n,
        alpha,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def zherk(uplo, trans, k, alpha, a, beta, c):
    u"""
    Rank-k update of a complex Hermitian matrix.

    ``zherk`` performs one of the Hermitian rank-k update operations

        C←alpha AA^H+beta C or C←alpha A^HA+beta C

    where A is a complex matrix, C is an n by n complex Hermitian
    matrix, and alpha and beta are real scalars.

    For full information please refer to the NAG Library document for
    f06zp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zpf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^H+beta C.

        `trans` = 'C'

          C←alpha A^HA+beta C.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'C'.

    alpha : float
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans`='C': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='C': n;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'.

    beta : float
        The scalar beta.

    c : complex, array-like, shape (n, n)
        The n by n Hermitian matrix C.

    Returns
    -------
    c : complex, ndarray, shape (n, n)
        The updated matrix C. The imaginary parts of the diagonal
        elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zherk as b_zherk

    fun_name = 'naginterfaces.library.blas.zherk'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    b_zherk(
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def zher2k(uplo, trans, k, alpha, a, b, beta, c):
    u"""
    Rank-2k update of a complex Hermitian matrix.

    ``zher2k`` performs one of the Hermitian rank-2k update operations

        C←alpha AB^H+alpha_BA^H+beta C or C←alpha A^HB+alpha_B^HA+beta
        C,

    where A and B are complex matrices, C is an n by n complex Hermitian
    matrix, alpha is a complex scalar, and beta is a real scalar.

    For full information please refer to the NAG Library document for
    f06zr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AB^H+alpha_BA^H+beta C.

        `trans` = 'C'

          C←alpha A^HB+alpha_B^HA+beta C.

    k : int
        k, the number of columns of A if `trans` = 'N', or the number of
        rows of A if `trans` = 'C'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans`='C': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='C': n;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'.

    b : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans`='C': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='C': n;
        otherwise: 0.

        The matrix B; B is n by k if `trans` = 'N', or k by n if `trans`
        = 'C'.

    beta : float
        The scalar beta.

    c : complex, array-like, shape (n, n)
        The n by n Hermitian matrix C.

    Returns
    -------
    c : complex, ndarray, shape (n, n)
        The updated matrix C. The imaginary parts of the diagonal
        elements are set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'C'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zher2k as b_zher2k

    fun_name = 'naginterfaces.library.blas.zher2k'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    b_zher2k(
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        _b,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def zsymm(side, uplo, m, alpha, a, b, beta, c=None):
    u"""
    Matrix-matrix product, one complex symmetric matrix, one complex
    rectangular matrix.

    ``zsymm`` performs one of the matrix-matrix operations

        C←alpha AB+beta C or C←alpha BA+beta C,

    where A is a complex symmetric matrix, B and C are m by n complex
    matrices, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06ztf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

          B is pre-multiplied from the left.

        `side` = 'R'

          B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

          The upper triangular part of A is stored.

        `uplo` = 'L'

          The lower triangular part of A is stored.

    m : int
        m, the number of rows of the matrices B and C; the order of A if
        `side` = 'L'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `side`='L': `m`; if `side`='R': n;
        otherwise: 0.

        The symmetric matrix A; A is m by m if `side` = 'L', or n by n
        if `side` = 'R'.

    b : complex, array-like, shape (`m`, n)
        The m by n matrix B.

    beta : complex
        The scalar beta.

    c : None or complex, array-like, shape (`m`, n), optional
        The m by n matrix C.

        If `beta` = 0.0, `c` need not be set.

    Returns
    -------
    c : complex, ndarray, shape (`m`, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `side`.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zsymm as b_zsymm

    fun_name = 'naginterfaces.library.blas.zsymm'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_c, 'c')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _beta = _b_utils._EngineComplex128ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _beta_py = _b_utils._EngineComplex128ScalarType.to_py(_beta)
    if _beta_py != 0.0:
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when beta != 0.0',
            )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if c is None:
        _c = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            (_m_py, _n_py),
            sorder=_m_sorder,
        )
    b_zsymm(
        side,
        uplo,
        _m,
        _n,
        alpha,
        _a,
        _b,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def zsyrk(uplo, trans, k, alpha, a, beta, c):
    u"""
    Rank-k update of a complex symmetric matrix.

    ``zsyrk`` performs one of the symmetric rank-k update operations

        C←alpha AA^T+beta C or C←alpha A^TA+beta C,

    where A is a complex matrix, C is an n by n complex symmetric
    matrix, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zuf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AA^T+beta C.

        `trans` = 'T'

          C←alpha A^TA+beta C.

    k : int
        k, the number of columns of A and B if `trans` = 'N', or the
        number of rows of A and B if `trans` = 'T'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': n;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'.

    beta : complex
        The scalar beta.

    c : complex, array-like, shape (n, n)
        The n by n symmetric matrix C.

    Returns
    -------
    c : complex, ndarray, shape (n, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zsyrk as b_zsyrk

    fun_name = 'naginterfaces.library.blas.zsyrk'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    b_zsyrk(
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py

def zsyr2k(uplo, trans, k, alpha, a, b, beta, c):
    u"""
    Rank-2k update of a complex symmetric matrix.

    ``zsyr2k`` performs one of the symmetric rank-2k update operations

        C←alpha AB^T+alpha BA^T+beta C or C←alpha A^TB+alpha B^TA+beta
        C,

    where A and B are complex matrices, C is an n by n complex symmetric
    matrix, and alpha and beta are complex scalars.

    For full information please refer to the NAG Library document for
    f06zw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f06/f06zwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of C is
        stored.

        `uplo` = 'U'

          The upper triangular part of C is stored.

        `uplo` = 'L'

          The lower triangular part of C is stored.

    trans : str, length 1
        Specifies the operation to be performed.

        `trans` = 'N'

          C←alpha AB^T+alpha BA^T+beta C.

        `trans` = 'T'

          C←alpha A^TB+alpha B^TA+beta C.

    k : int
        k, the number of columns of A and B if `trans` = 'N', or the
        number of rows of A and B if `trans` = 'T'.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': n;
        otherwise: 0.

        The matrix A; A is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'.

    b : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `trans`='N': n; if `trans`='T': `k`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `trans`='N': `k`; if `trans`='T': n;
        otherwise: 0.

        The matrix B; B is n by k if `trans` = 'N', or k by n if `trans`
        = 'T'.

    beta : complex
        The scalar beta.

    c : complex, array-like, shape (n, n)
        The n by n symmetric matrix C.

    Returns
    -------
    c : complex, ndarray, shape (n, n)
        The updated matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N' or 'T'.

        (`errno` 3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` 4)
            On entry, error in parameter `k`.

            Constraint: `k` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.blas import zsyr2k as b_zsyr2k

    fun_name = 'naginterfaces.library.blas.zsyr2k'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    b_zsyr2k(
        uplo,
        trans,
        _n,
        k,
        alpha,
        _a,
        _b,
        beta,
        _c,
    )
    _c_py = _c
    return _c_py
