# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `ode` Chapter.

``ode`` - Ordinary Differential Equations

This module is concerned with the numerical solution of ordinary
differential equations.
There are two main types of problem: those in which all boundary
conditions are specified at one point (initial value problems), and
those in which the boundary conditions are distributed between two or
more points (boundary value problems and eigenvalue problems). Functions
are available for initial value problems, two-point boundary value
problems and Sturm--Liouville eigenvalue problems.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02intro.html

See Also
--------
``naginterfaces.library.examples.ode`` :
    This subpackage contains examples for the ``ode`` module.
    See also the :ref:`library_ode_ex` subsection.
"""

# NAG Copyright 2017-2019.

import collections as _collections

from ..base import utils as _b_utils
from ..library import _utils

def bvp_shoot_genpar_intern(h, e, parerr, param, m1, aux, bcaux, raaux, prsol, data=None):
    u"""
    Ordinary differential equations, boundary value problem, shooting
    and matching technique, allowing interior matching point, general
    parameters to be determined.

    ``bvp_shoot_genpar_intern`` solves a two-point boundary value
    problem for a system of ordinary differential equations, using
    initial value techniques and Newton iteration; it generalizes
    ``bvp_shoot_bval`` to include the case where parameters other than
    boundary values are to be determined.

    For full information please refer to the NAG Library document for
    d02ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02agf.html

    Parameters
    ----------
    h : float
        `h` must be set to an estimate of the step size, h, needed for
        integration.

    e : float, array-like, shape (n)
        `e`[i-1] must be set to a small quantity to control the ith
        solution component. The element `e`[i-1] is used:

        - in the bound on the local error in the ith component of the
          solution y_i during integration,

        - in the convergence test on the ith component of the solution
          y_i at the matching point in the Newton iteration.

        The elements `e`[i-1] should not be chosen too small.

        They should usually be several orders of magnitude larger than
        machine precision.

    parerr : float, array-like, shape (n1)
        `parerr`[i-1] must be set to a small quantity to control the ith
        parameter component. The element `parerr`[i-1] is used:

        - in the convergence test on the ith parameter in the Newton
          iteration,

        - in perturbing the ith parameter when approximating the
          derivatives of the components of the solution with respect to
          the ith parameter, for use in the Newton iteration.

        The elements `parerr`[i-1] should not be chosen too small.

        They should usually be several orders of magnitude larger than
        machine precision.

    param : float, array-like, shape (n1)
        `param`[i-1] must be set to an estimate for the ith parameter,
        p_i, for i = 1,2,...,n1.

    m1 : int
        Determines whether or not the final solution is computed as well
        as the parameter values.

        `m1` = 1

          The final solution is not calculated;

        `m1` > 1

          The final values of the solution at interval (length of
          range)/(`m1`-1) are calculated and stored sequentially in the
          array `c` starting with the values of y_i evaluated at the
          first end point (see `raaux`) stored in `c`[0,i-1].

    aux : callable f = aux(n, y, x, param, data=None)
        `aux` must evaluate the functions f_i (i.e., the derivatives
        y'_i) for given values of its arguments, x,y_1,...,y_n,
        p_1,...,p_n_1.

        Parameters
        ~~~~~~~~~~
        n : int
            n, the total number of differential equations.

        y : float, ndarray, shape (`n`)
            y_i, for i = 1,2,...,n, the value of the argument.

        x : float
            x, the value of the argument.

        param : float, ndarray, shape (n1)
            p_i, for i = 1,2,...,n_1, the value of the parameters.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (`n`)
            The value of f_i, for i = 1,2,...,n.

    bcaux : callable (g0, g1) = bcaux(n, param, data=None)
        `bcaux` must evaluate the values of y_i at the end points of the
        range given the values of p_1,...,p_n_1.

        Parameters
        ~~~~~~~~~~
        n : int
            n, the total number of differential equations.

        param : float, ndarray, shape (n1)
            p_i, for i = 1,2,...,n, the value of the parameters.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        g0 : float, array-like, shape (`n`)
            The values y_i, for i = 1,2,...,n, at the boundary point x_0
            (see `raaux`).

        g1 : float, array-like, shape (`n`)
            The values y_i, for i = 1,2,...,n, at the boundary point x_1
            (see `raaux`).

    raaux : callable (x0, x1, r) = raaux(param, data=None)
        `raaux` must evaluate the end points, x_0 and x_1, of the range
        and the matching point, r, given the values p_1,p_2,...,p_n_1.

        Parameters
        ~~~~~~~~~~
        param : float, ndarray, shape (n1)
            p_i, for i = 1,2,...,n_1, the value of the parameters.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        x0 : float
            Must contain the left-hand end of the range, x_0.

        x1 : float
            Must contain the right-hand end of the range x_1.

        r : float
            Must contain the matching point, r.

    prsol : callable prsol(param, res, err, data=None)
        `prsol` is called at each iteration of the Newton method and can
        be used to print the current values of the parameters p_i, for i
        = 1,2,...,n_1, their errors, e_i, and the sum of squares of the
        errors at the matching point, r.

        Parameters
        ~~~~~~~~~~
        param : float, ndarray, shape (n1)
            p_i, for i = 1,2,...,n_1, the current value of the
            parameters.

        res : float
            The sum of squares of the errors in the arguments, sum_i =
            1^n_1e_i^2.

        err : float, ndarray, shape (n1)
            The errors in the parameters, e_i, for i = 1,2,...,n_1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    h : float
        The last step length used.

    param : float, ndarray, shape (n1)
        The corrected value for the ith parameter, unless an error has
        occurred, when it contains the last calculated value of the
        parameter (possibly perturbed by
        `parerr`[i-1]*(1+|`param`[i-1]|) if the error occurred when
        calculating the approximate derivatives).

    c : float, ndarray, shape (`m1`, n)
        The solution when `m1` > 1 (see `m1`).

        If `m1` = 1, the elements of `c` are not used.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n1 = *<value>* and n = *<value>*.

            Constraint: n1 <= n.

        (`errno` 2)
            No further progress can be made when stepping to obtain a
            Jacobian update for the current parameter values. Step
            length `h` = *<value>*.

        (`errno` 3)
            Currently the matching point r does not lie in range
            [x_0,x_1].

            If x_0, x_1 or r depend on the parameters then this may
            occur when care is not taken to avoid it.

            r = *<value>*, x_0 = *<value>* and x_1 = *<value>*.

        (`errno` 4)
            No further progress can be made when stepping to a solution
            corresponding to the current parameter values.

            Step length `h` = *<value>*.

        (`errno` 5)
            The Jacobian for parameter corrections is singular.

        (`errno` 6)
            The Newton method failed to converge while updating
            parameter values.

        (`errno` 7)
            The Newton method has not converged after 12 iterations
            while updating parameter values.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import bvp_shoot_genpar_intern as b_bvp_shoot_genpar_intern

    fun_name = 'naginterfaces.library.ode.bvp_shoot_genpar_intern'

    cb_aux_name = fun_name + ':aux'

    def cb_aux(n, n1, f, y, x, param, data=None):
        """Wrapping user's aux. Called as base aux."""
        if data is not None:
            _f_aux = aux(
                n,
                y,
                x,
                param,
                data,
            )
        else:
            _f_aux = aux(
                n,
                y,
                x,
                param,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_aux_name, 'returned': 'f'},
            _f_aux,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_bcaux_name = fun_name + ':bcaux'

    def cb_bcaux(n, n1, g0, g1, param, data=None):
        """Wrapping user's bcaux. Called as base bcaux."""
        if data is not None:
            _g0_bcaux, _g1_bcaux = bcaux(
                n,
                param,
                data,
            )
        else:
            _g0_bcaux, _g1_bcaux = bcaux(
                n,
                param,
            )
        g0[:] = _utils._to_ndarray(
            {'fun_name': cb_bcaux_name, 'returned': 'g0'},
            _g0_bcaux,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        g1[:] = _utils._to_ndarray(
            {'fun_name': cb_bcaux_name, 'returned': 'g1'},
            _g1_bcaux,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_raaux(n1, param, data=None):
        """Wrapping user's raaux. Called as base raaux."""
        if data is not None:
            _x0_raaux, _x1_raaux, _r_raaux = raaux(
                param,
                data,
            )
        else:
            _x0_raaux, _x1_raaux, _r_raaux = raaux(
                param,
            )
        _x0_raaux_base = _x0_raaux
        _x1_raaux_base = _x1_raaux
        _r_raaux_base = _r_raaux
        return (
            _x0_raaux_base,
            _x1_raaux_base,
            _r_raaux_base,
        )

    def cb_prsol(param, res, n1, err, data=None):
        """Wrapping user's prsol. Called as base prsol."""
        if data is not None:
            prsol(
                param,
                res,
                err,
                data,
            )
        else:
            prsol(
                param,
                res,
                err,
            )

    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_e, {'fun_name': fun_name, 'entity': 'e'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _parerr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parerr'},
        parerr,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_parerr, {'fun_name': fun_name, 'entity': 'parerr'}, exp_rank=1)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m1 = _b_utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m1_py = _b_utils._EngineIntScalarType.to_py(_m1)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m1_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_utils._check_callable(aux, {'fun_name': fun_name, 'entity': 'aux'})
    _aux = cb_aux
    _b_utils._check_callable(bcaux, {'fun_name': fun_name, 'entity': 'bcaux'})
    _bcaux = cb_bcaux
    _b_utils._check_callable(raaux, {'fun_name': fun_name, 'entity': 'raaux'})
    _raaux = cb_raaux
    _b_utils._check_callable(prsol, {'fun_name': fun_name, 'entity': 'prsol'})
    _prsol = cb_prsol
    _return_scalar = b_bvp_shoot_genpar_intern(
        h,
        _e,
        _parerr,
        _param,
        _c,
        _n,
        _n1,
        _m1,
        _aux,
        _bcaux,
        _raaux,
        _prsol,
        data,
    )
    _h_py = _return_scalar
    _param_py = _param
    _c_py = _c
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['h', 'param', 'c'],
    )(
        _h_py,
        _param_py,
        _c_py,
    )
    return _return_data

def ivp_rkm_val_simple(x, xend, y, tol, hmax, m, val, fcn, data=None):
    u"""
    Ordinary differential equations, initial value problem,
    Runge--Kutta--Merson method, until a component attains given value
    (simple driver).

    ``ivp_rkm_val_simple`` integrates a system of first-order ordinary
    differential equations over an interval with suitable initial
    conditions, using a Runge--Kutta--Merson method, until a specified
    component attains a given value.

    For full information please refer to the NAG Library document for
    d02bg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02bgf.html

    Parameters
    ----------
    x : float
        Must be set to the initial value of the independent variable x.

    xend : float
        The final value of the independent variable x.

        If `xend` < `x` on entry integration will proceed in the
        negative direction.

    y : float, array-like, shape (n)
        The initial values of the solution y_1,y_2,...,y_n.

    tol : float
        Must be set to a positive tolerance for controlling the error in
        the integration and in the determination of the position where
        y_m = alpha.

        ``ivp_rkm_val_simple`` has been designed so that, for most
        problems, a reduction in `tol` leads to an approximately
        proportional reduction in the error in the solution obtained in
        the integration.

        The relation between changes in `tol` and the error in the
        determination of the position where y_m = alpha is less clear,
        but for `tol` small enough the error should be approximately
        proportional to `tol`.

        However, the actual relation between `tol` and the accuracy
        cannot be guaranteed.

        You are strongly recommended to call ``ivp_rkm_val_simple`` with
        more than one value for `tol` and to compare the results
        obtained to estimate their accuracy.

        In the absence of any prior knowledge you might compare results
        obtained by calling ``ivp_rkm_val_simple`` with `tol` =
        10.0^(-p) and `tol` = 10.0^(-p-1) if p correct decimal digits in
        the solution are required.

    hmax : float
        Controls how the sign of y_m-alpha is checked.

        `hmax` = 0.0

          y_m-alpha is checked at every internal integration step.

        `hmax` != 0.0

          The computed solution is checked for a change in sign of
          y_m-alpha at steps of not greater than |`hmax`|. This facility
          should be used if there is any chance of 'missing' the change
          in sign by checking too infrequently. For example, if two
          changes of sign of y_m-alpha are expected within a distance h,
          say, of each other then a suitable value for `hmax` might be
          `hmax` = h/2. If only one change of sign in y_m-alpha is
          expected on the range `x` to `xend` then `hmax` = 0.0 is most
          appropriate.

    m : int
        The index m of the component of the solution whose value is to
        be checked.

    val : float
        The value of alpha in the equation y_m = alpha to be solved for
        `x`.

    fcn : callable f = fcn(x, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i) for given values of its arguments x,y_1,...,y_n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the argument.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the argument.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            The value of f_i, for i = 1,2,...,n.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float
        The point where the component y_m attains the value alpha unless
        an error has occurred, when it contains the value of x at the
        error. In particular, if y_m != alpha anywhere on the range x =
        `x` to x = `xend`, it will contain `xend` on exit.

    y : float, ndarray, shape (n)
        The computed values of the solution at a point near the solution
        `x`, unless an error has occurred when they contain the computed
        values at the final value of `x`.

    tol : float
        Normally unchanged. However if the range from `x` to the
        position where y_m = alpha (or to the final value of `x` if an
        error occurs) is so short that a small change in `tol` is
        unlikely to make any change in the computed solution then, on
        return, `tol` has its sign changed. To check results returned
        with `tol` < 0.0, ``ivp_rkm_val_simple`` should be called again
        with a positive value of `tol` whose magnitude is considerably
        smaller than that of the previous call.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` <= n.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 2)
            The value of `tol`, *<value>*, is too small for the function
            to make any further progress across the integration range.
            Current value of `x` = *<value>*.

        (`errno` 3)
            The value of `tol`, *<value>*, is too small for the function
            to take an initial step.

        (`errno` 5)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

        (`errno` 6)
            A serious error occurred in a call to the internal
            integrator.

            The error code internally was *<value>*. Please contact NAG.

        (`errno` 7)
            Unexpected internal error in call to interpolation routine.

            The interpolation routine returned error flag *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            No change in sign of the function y_m-alpha was detected in
            the integration range.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_rkm_val_simple as b_ivp_rkm_val_simple

    fun_name = 'naginterfaces.library.ode.ivp_rkm_val_simple'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, x, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    try:
        _return_scalars = b_ivp_rkm_val_simple(
            x,
            xend,
            _n,
            _y,
            tol,
            hmax,
            m,
            val,
            _fcn,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _return_scalars.x
    _y_py = _y
    _tol_py = _return_scalars.tol
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'tol'],
    )(
        _x_py,
        _y_py,
        _tol_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_rkm_zero_simple(x, xend, y, tol, irelab, hmax, fcn, g, data=None):
    u"""
    Ordinary differential equations, initial value problem,
    Runge--Kutta--Merson method, until function of solution is zero
    (simple driver).

    ``ivp_rkm_zero_simple`` integrates a system of first-order ordinary
    differential equations over an interval with suitable initial
    conditions, using a Runge--Kutta--Merson method, until a
    user-specified function of the solution is zero.

    For full information please refer to the NAG Library document for
    d02bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02bhf.html

    Parameters
    ----------
    x : float
        Must be set to the initial value of the independent variable x.

    xend : float
        The final value of the independent variable x.

        If `xend` < `x` on entry, integration proceeds in a negative
        direction.

    y : float, array-like, shape (n)
        The initial values of the solution y_1,y_2,...,y_n.

    tol : float
        Must be set to a positive tolerance for controlling the error in
        the integration and in the determination of the position where
        g(x,y) = 0.0.

        ``ivp_rkm_zero_simple`` has been designed so that, for most
        problems, a reduction in `tol` leads to an approximately
        proportional reduction in the error in the solution obtained in
        the integration.

        The relation between changes in `tol` and the error in the
        determination of the position where g(x,y) = 0.0 is less clear,
        but for `tol` small enough the error should be approximately
        proportional to `tol`.

        However, the actual relation between `tol` and the accuracy
        cannot be guaranteed.

        You are strongly recommended to call ``ivp_rkm_zero_simple``
        with more than one value for `tol` and to compare the results
        obtained to estimate their accuracy.

        In the absence of any prior knowledge you might compare results
        obtained by calling ``ivp_rkm_zero_simple`` with `tol` =
        10.0^(-p) and `tol` = 10.0^(-p-1) if p correct decimal digits in
        the solution are required.

    irelab : int
        Determines the type of error control. At each step in the
        numerical solution an estimate of the local error, est, is made.
        For the current step to be accepted the following condition must
        be satisfied:

        `irelab` = 0

          est <= `tol`*max{1.0,|y_1|,|y_2|,...,|y_n|};

        `irelab` = 1

          est <= `tol`;

        `irelab` = 2

          est <= `tol`*max{epsilon,|y_1|,|y_2|,...,|y_n|}, where epsilon
          is machine precision.

        If the appropriate condition is not satisfied, the step size is
        reduced and the solution recomputed on the current step.

        If you wish to measure the error in the computed solution in
        terms of the number of correct decimal places, set `irelab` = 1
        on entry, whereas if the error requirement is in terms of the
        number of correct significant digits, set `irelab` = 2.

        Where there is no preference in the choice of error test,
        `irelab` = 0 will result in a mixed error test.

        It should be borne in mind that the computed solution will be
        used in evaluating g(x,y).

    hmax : float
        If `hmax` = 0.0, no special action is taken.

        If `hmax` != 0.0, a check is made for a change in sign of g(x,y)
        at steps not greater than |`hmax`|.

        This facility should be used if there is any chance of 'missing'
        the change in sign by checking too infrequently.

        For example, if two changes of sign of g(x,y) are expected
        within a distance h, say, of each other, a suitable value for
        `hmax` might be `hmax` = h/2.

        If only one change of sign in g(x,y) is expected on the range
        `x` to `xend`, the choice `hmax` = 0.0 is most appropriate.

    fcn : callable f = fcn(x, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i) for given values of its arguments x,y_1,...,y_n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the argument.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the argument.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            The value of f_i, for i = 1,2,...,n.

    g : callable retval = g(x, y, data=None)
        `g` must evaluate the function g(x,y) at a specified point.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            The value of y_i, for i = 1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(x,y) at the specified point.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float
        The point where g(x,y) = 0.0 unless an error has occurred, when
        it contains the value of x at the error. In particular, if
        g(x,y) != 0.0 anywhere on the range `x` to `xend`, it will
        contain `xend` on exit.

    y : float, ndarray, shape (n)
        The computed values of the solution at the final point x = `x`.

    tol : float
        Normally unchanged. However if the range from x = `x` to the
        position where g(x,y) = 0.0 (or to the final value of x if an
        error occurs) is so short that a small change in `tol` is
        unlikely to make any change in the computed solution, `tol` is
        returned with its sign changed. To check results returned with
        `tol` < 0.0, ``ivp_rkm_zero_simple`` should be called again with
        a positive value of `tol` whose magnitude is considerably
        smaller than that of the previous call.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irelab` = *<value>*.

            Constraint: 0 <= `irelab` <= 2.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 3)
            The value of `tol`, *<value>*, is too small for the function
            to take an initial step.

        (`errno` 5)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

        (`errno` 6)
            A serious error occurred in a call to the internal
            integrator.

            The error code internally was *<value>*. Please contact NAG.

        (`errno` 7)
            Unexpected internal error in call to interpolation routine.

            The interpolation routine returned error flag *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The value of `tol`, *<value>*, is too small for the function
            to make any further progress across the integration range.
            Current value of x = *<value>*.

        (`errno` 4)
            No change in sign of the function g(x,y) was detected in the
            integration range.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_rkm_zero_simple as b_ivp_rkm_zero_simple

    fun_name = 'naginterfaces.library.ode.ivp_rkm_zero_simple'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, x, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_g(n, x, y, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _retval_g = g(
                x,
                y,
                data,
            )
        else:
            _retval_g = g(
                x,
                y,
            )
        return _retval_g

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'})
    _g = cb_g
    try:
        _return_scalars = b_ivp_rkm_zero_simple(
            x,
            xend,
            _n,
            _y,
            tol,
            irelab,
            hmax,
            _fcn,
            _g,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _return_scalars.x
    _y_py = _y
    _tol_py = _return_scalars.tol
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'tol'],
    )(
        _x_py,
        _y_py,
        _tol_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_rk_zero_simple(x, xend, y, fcn, tol, relabs, output=None, g=None, data=None):
    u"""
    Ordinary differential equations, initial value problem, Runge--Kutta
    method, until function of solution is zero, integration over range
    with intermediate output (simple driver).

    ``ivp_rk_zero_simple`` integrates a system of first-order ordinary
    differential equations over an interval with suitable initial
    conditions, using a fixed order Runge--Kutta method, until a
    user-specified function, if supplied, of the solution is zero, and
    returns the solution at points specified by you, if desired.

    For full information please refer to the NAG Library document for
    d02bj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02bjf.html

    Parameters
    ----------
    x : float
        The initial value of the independent variable x.

    xend : float
        The final value of the independent variable. If `xend` < `x`,
        integration will proceed in the negative direction.

    y : float, array-like, shape (n)
        The initial values of the solution y_1,y_2,...,y_n at x = `x`.

    fcn : callable f = fcn(x, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i) for given values of its arguments x,y_1,...,y_n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            The value of f_i, for i = 1,2,...,n.

    tol : float
        A positive tolerance for controlling the error in the
        integration. Hence `tol` affects the determination of the
        position where g(x,y) = 0, if g is supplied.

        ``ivp_rk_zero_simple`` has been designed so that, for most
        problems, a reduction in `tol` leads to an approximately
        proportional reduction in the error in the solution.

        However, the actual relation between `tol` and the accuracy
        achieved cannot be guaranteed.

        You are strongly recommended to call ``ivp_rk_zero_simple`` with
        more than one value for `tol` and to compare the results
        obtained to estimate their accuracy.

        In the absence of any prior knowledge, you might compare the
        results obtained by calling ``ivp_rk_zero_simple`` with `relabs`
        = 'D' and with each of `tol` = 10.0^(-p) and `tol` = 10.0^(-p-1)
        where p correct significant digits are required in the solution,
        y.

        The accuracy of the value x such that g(x,y) = 0 is indirectly
        controlled by varying `tol`.

        You should experiment to determine this accuracy.

    relabs : str, length 1
        The type of error control. At each step in the numerical
        solution an estimate of the local error, est, is made. For the
        current step to be accepted the following condition must be
        satisfied:

            est = max(e_i/(tau_r*max(|y_i|,tau_a))) <= 1.0

        where tau_r and tau_a are defined by

        [table omitted]

        where epsilon_r and epsilon_a are small machine-dependent
        numbers and e_i is an estimate of the local error at y_i,
        computed internally. If the condition is not satisfied, the step
        size is reduced and the solution is recomputed on the current
        step. If you wish to measure the error in the computed solution
        in terms of the number of correct decimal places, `relabs`
        should be set to 'A' on entry, whereas if the error requirement
        is in terms of the number of correct significant digits,
        `relabs` should be set to 'R'. If you prefer a mixed error test,
        `relabs` should be set to 'M', otherwise if you have no
        preference, `relabs` should be set to the default 'D'. Note that
        in this case 'D' is taken to be 'R'.

    output : None or callable xsol = output(xsol, y, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `output` permits access to intermediate values of the computed
        solution (for example to print or plot them), at successive
        user-specified points.

        It is initially called by ``ivp_rk_zero_simple`` with `xsol` =
        `x` (the initial value of x).

        You must reset `xsol` to the next point (between the current
        `xsol` and `xend`) where `output` is to be called, and so on at
        each call to `output`.

        If, after a call to `output`, the reset point `xsol` is beyond
        `xend`, ``ivp_rk_zero_simple`` will integrate to `xend` with no
        further calls to `output`; if a call to `output` is required at
        the point `xsol` = `xend`, `xsol` must be given precisely the
        value `xend`.

        Parameters
        ~~~~~~~~~~
        xsol : float
            The output value of the independent variable x.

        y : float, ndarray, shape (n)
            The computed solution at the point `xsol`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        xsol : float
            You must set `xsol` to the next value of x at which `output`
            is to be called.

    g : None or callable retval = g(x, y, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `g` must evaluate the function g(x,y) for specified values x,y.

        It specifies the function g for which the first position x where
        g(x,y) = 0 is to be found.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(x,y) at the specified point.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float
        If g is supplied by you, it contains the point where g(x,y) = 0,
        unless g(x,y) != 0 anywhere on the range `x` to `xend`, in which
        case, `x` will contain `xend` (and the error indicator `errno` =
        6 is set); if g is not supplied by you it contains `xend`.
        However, if an error has occurred, it contains the value of x at
        which the error occurred.

    y : float, ndarray, shape (n)
        The computed values of the solution at the final point x = `x`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = `xend` = *<value>*.

            Constraint: `x` != `xend`.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: *<value>* <= `tol` < 0.01.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `relabs` = *<value>*.

            Constraint: `relabs` = 'M', 'A', 'R' or 'D'.

        (`errno` 2)
            With the given value of `tol`, no further progress can be
            made across the integration range from the current point x =
            `x`.

            The solution is returned at this point, `x` = *<value>*.

        (`errno` 3)
            With the given value of `tol`, no initial progress can be
            made from the starting point of integration provided.

        (`errno` 4)
            On the first call to `output`, `xsol` was returned as
            *<value>*, which is inconsistent with (`x`,`xend`) ---
            (*<value>*,*<value>*).

        (`errno` 4)
            On the first call to `output`, `xsol` remained unchanged.

        (`errno` 5)
            On a call to `output`, `xsol` was returned as *<value>*,
            which is inconsistent with previous `xsol` and `xend` ---
            (*<value>*,*<value>*).

        (`errno` 5)
            On a call to `output`, `xsol` remained unchanged.

        (`errno` 6)
            No change in sign of the function g(x,y) was detected in the
            integration range.

        (`errno` 7)
            Unexpected internal error in call to zero-finding routine.

            The zero-finding routine returned error flag *<value>*.

        (`errno` 7)
            Unexpected internal error in call to interpolation routine.

            The interpolation routine returned error flag *<value>*.

        (`errno` 8)
            Unexpected internal error in call to step integrator.

            The step integrator returned error flag *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_rk_zero_simple as b_ivp_rk_zero_simple

    fun_name = 'naginterfaces.library.ode.ivp_rk_zero_simple'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(x, n, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_output(xsol, n, y, data=None):
        """Wrapping user's output. Called as base output."""
        if data is not None:
            _xsol_output = output(
                xsol,
                y,
                data,
            )
        else:
            _xsol_output = output(
                xsol,
                y,
            )
        _xsol_output_base = _xsol_output
        return _xsol_output_base

    def cb_g(x, n, y, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _retval_g = g(
                x,
                y,
                data,
            )
        else:
            _retval_g = g(
                x,
                y,
            )
        return _retval_g

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _b_utils._check_callable(output, {'fun_name': fun_name, 'entity': 'output'}, allow_none=True)
    _output = None if output is None else cb_output
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'}, allow_none=True)
    _g = None if g is None else cb_g
    _return_scalar = b_ivp_rk_zero_simple(
        x,
        xend,
        _n,
        _y,
        _fcn,
        tol,
        relabs,
        _output,
        _g,
        data,
    )
    _x_py = _return_scalar
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    return _return_data

def ivp_adams_zero_simple(x, xend, y, fcn, tol, relabs, output=None, g=None, data=None):
    u"""
    Ordinary differential equations, initial value problem, Adams'
    method, until function of solution is zero, intermediate output
    (simple driver).

    ``ivp_adams_zero_simple`` integrates a system of first-order
    ordinary differential equations over a range with suitable initial
    conditions, using a variable-order, variable-step Adams' method
    until a user-specified function, if supplied, of the solution is
    zero, and returns the solution at points specified by you, if
    desired.

    For full information please refer to the NAG Library document for
    d02cj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02cjf.html

    Parameters
    ----------
    x : float
        The initial value of the independent variable x.

    xend : float
        The final value of the independent variable. If `xend` < `x`,
        integration will proceed in the negative direction.

    y : float, array-like, shape (n)
        The initial values of the solution y_1,y_2,...,y_n at x = `x`.

    fcn : callable f = fcn(x, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i) for given values of its arguments x,y_1,...,y_n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            The value of f_i, for i = 1,2,...,n.

    tol : float
        A positive tolerance for controlling the error in the
        integration. Hence `tol` affects the determination of the
        position where g(x,y) = 0.0, if g is supplied.

        ``ivp_adams_zero_simple`` has been designed so that, for most
        problems, a reduction in `tol` leads to an approximately
        proportional reduction in the error in the solution.

        However, the actual relation between `tol` and the accuracy
        achieved cannot be guaranteed.

        You are strongly recommended to call ``ivp_adams_zero_simple``
        with more than one value for `tol` and to compare the results
        obtained to estimate their accuracy.

        In the absence of any prior knowledge, you might compare the
        results obtained by calling ``ivp_adams_zero_simple`` with `tol`
        = 10.0^(-p) and `tol` = 10.0^(-p-1) where p correct decimal
        digits are required in the solution.

    relabs : str, length 1
        The type of error control. At each step in the numerical
        solution an estimate of the local error, est, is made. For the
        current step to be accepted the following condition must be
        satisfied:

            est = sqrt(sum_i = 1^n(e_i/(tau_r*|y_i|+tau_a))^2) <= 1.0

        where tau_r and tau_a are defined by

        [table omitted]

        where epsilon is a small machine-dependent number and e_i is an
        estimate of the local error at y_i, computed internally. If the
        appropriate condition is not satisfied, the step size is reduced
        and the solution is recomputed on the current step. If you wish
        to measure the error in the computed solution in terms of the
        number of correct decimal places, `relabs` should be set to 'A'
        on entry, whereas if the error requirement is in terms of the
        number of correct significant digits, `relabs` should be set to
        'R'. If you prefer a mixed error test, `relabs` should be set to
        'M', otherwise if you have no preference, `relabs` should be set
        to the default 'D'. Note that in this case 'D' is taken to be
        'M'.

    output : None or callable xsol = output(xsol, y, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `output` permits access to intermediate values of the computed
        solution (for example to print or plot them), at successive
        user-specified points.

        It is initially called by ``ivp_adams_zero_simple`` with `xsol`
        = `x` (the initial value of x).

        You must reset `xsol` to the next point (between the current
        `xsol` and `xend`) where `output` is to be called, and so on at
        each call to `output`.

        If, after a call to `output`, the reset point `xsol` is beyond
        `xend`, ``ivp_adams_zero_simple`` will integrate to `xend` with
        no further calls to `output`; if a call to `output` is required
        at the point `xsol` = `xend`, `xsol` must be given precisely the
        value `xend`.

        Parameters
        ~~~~~~~~~~
        xsol : float
            The output value of the independent variable x.

        y : float, ndarray, shape (n)
            The computed solution at the point `xsol`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        xsol : float
            You must set `xsol` to the next value of x at which `output`
            is to be called.

    g : None or callable retval = g(x, y, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `g` must evaluate the function g(x,y) for specified values x,y.

        It specifies the function g for which the first position x where
        g(x,y) = 0 is to be found.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g(x,y) at the specified point.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float
        If g is supplied by you, it contains the point where g(x,y) =
        0.0, unless g(x,y) != 0.0 anywhere on the range `x` to `xend`,
        in which case, `x` will contain `xend`. If g is not supplied by
        you it contains `xend`, unless an error has occurred, when it
        contains the value of x at the error.

    y : float, ndarray, shape (n)
        The computed values of the solution at the final point x = `x`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `relabs` = *<value>*.

            Constraint: `relabs` = 'M', 'A', 'R' or 'D'.

        (`errno` 1)
            On entry, `x` = *<value>* and `xend` = *<value>*.

            Constraint: `x` != `xend`.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            No integration steps have been taken. Progress not possible
            with the input value of `tol` = *<value>*.

        (`errno` 4)
            On the first call to `output`, `xsol` was returned as
            *<value>*, which is inconsistent with (`x`,`xend`) ---
            (*<value>*,*<value>*).

        (`errno` 4)
            On the first call to `output`, `xsol` remained unchanged.

        (`errno` 5)
            On a call to `output`, `xsol` remained unchanged.

            `xsol` = *<value>*.

        (`errno` 7)
            Impossible error --- internal variable IDID = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Integration successful as far as x = *<value>*, but further
            progress not possible with the input value of `tol` =
            *<value>*.

        (`errno` 5)
            On a call to `output`, `xsol` was returned as *<value>*,
            which is inconsistent with previous `xsol` and `xend` ---
            (*<value>*,*<value>*).

        (`errno` 6)
            No change in sign of the function g(x,y) was detected in the
            integration range.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import ivp_adams_zero_simple as b_ivp_adams_zero_simple

    fun_name = 'naginterfaces.library.ode.ivp_adams_zero_simple'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(x, n, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_output(xsol, n, y, data=None):
        """Wrapping user's output. Called as base output."""
        if data is not None:
            _xsol_output = output(
                xsol,
                y,
                data,
            )
        else:
            _xsol_output = output(
                xsol,
                y,
            )
        _xsol_output_base = _xsol_output
        return _xsol_output_base

    def cb_g(x, n, y, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _retval_g = g(
                x,
                y,
                data,
            )
        else:
            _retval_g = g(
                x,
                y,
            )
        return _retval_g

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _b_utils._check_callable(output, {'fun_name': fun_name, 'entity': 'output'}, allow_none=True)
    _output = None if output is None else cb_output
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'}, allow_none=True)
    _g = None if g is None else cb_g
    try:
        _return_scalar = b_ivp_adams_zero_simple(
            x,
            xend,
            _n,
            _y,
            _fcn,
            tol,
            relabs,
            _output,
            _g,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _return_scalar
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y'],
    )(
        _x_py,
        _y_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_bdf_zero_simple(x, xend, y, fcn, tol, relabs, pederv=None, output=None, g=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Ordinary differential equations, stiff initial value problem,
    backward differentiation formulae method, until function of solution
    is zero, intermediate output (simple driver).

    ``ivp_bdf_zero_simple`` integrates a stiff system of first-order
    ordinary differential equations over an interval with suitable
    initial conditions, using a variable-order, variable-step method
    implementing the Backward Differentiation Formulae (BDF), until a
    user-specified function, if supplied, of the solution is zero, and
    returns the solution at points specified by you, if desired.

    For full information please refer to the NAG Library document for
    d02ej

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ejf.html

    Parameters
    ----------
    x : float
        The initial value of the independent variable x.

    xend : float
        The final value of the independent variable. If `xend` < `x`,
        integration will proceed in the negative direction.

    y : float, array-like, shape (n)
        The initial values of the solution y_1,y_2,...,y_n at x = `x`.

    fcn : callable f = fcn(x, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i) for given values of its arguments x,y_1,...,y_n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            The value of f_i, for i = 1,2,...,n.

    tol : float
        Must be set to a positive tolerance for controlling the error in
        the integration. Hence `tol` affects the determination of the
        position where g(x,y) = 0.0, if `g` is supplied.

        ``ivp_bdf_zero_simple`` has been designed so that, for most
        problems, a reduction in `tol` leads to an approximately
        proportional reduction in the error in the solution.

        However, the actual relation between `tol` and the accuracy
        achieved cannot be guaranteed.

        You are strongly recommended to call ``ivp_bdf_zero_simple``
        with more than one value for `tol` and to compare the results
        obtained to estimate their accuracy.

        In the absence of any prior knowledge, you might compare the
        results obtained by calling ``ivp_bdf_zero_simple`` with `tol` =
        10^(-p) and `tol` = 10^(-p-1) if p correct decimal digits are
        required in the solution.

    relabs : str, length 1
        The type of error control. At each step in the numerical
        solution an estimate of the local error, est, is made. For the
        current step to be accepted the following condition must be
        satisfied:

            est = sqrt(1/(n)sum_i = 1^n(e_i/(tau_r*|y_i|+tau_a))^2) <=
            1.0

        where tau_r and tau_a are defined by

        [table omitted]

        where epsilon is a small machine-dependent number and e_i is an
        estimate of the local error at y_i, computed internally. If the
        appropriate condition is not satisfied, the step size is reduced
        and the solution is recomputed on the current step. If you wish
        to measure the error in the computed solution in terms of the
        number of correct decimal places, `relabs` should be set to 'A'
        on entry, whereas if the error requirement is in terms of the
        number of correct significant digits, `relabs` should be set to
        'R'. If you prefer a mixed error test, `relabs` should be set to
        'M', otherwise if you have no preference, `relabs` should be set
        to the default 'D'. Note that in this case 'D' is taken to be
        'R'.

    pederv : None or callable pw = pederv(x, y, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `pederv` must evaluate the Jacobian of the system (that is, the
        partial derivatives df_i/dy_j) for given values of the variables
        x,y_1,y_2,...,y_n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        pw : float, array-like, shape (n, n)
            `pw`[i-1,j-1] must contain the value of df_i/dy_j, for j =
            1,2,...,n, for i = 1,2,...,n.

    output : None or callable xsol = output(xsol, y, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `output` permits access to intermediate values of the computed
        solution (for example to print or plot them), at successive
        user-specified points.

        It is initially called by ``ivp_bdf_zero_simple`` with `xsol` =
        `x` (the initial value of x).

        You must reset `xsol` to the next point (between the current
        `xsol` and `xend`) where `output` is to be called, and so on at
        each call to `output`.

        If, after a call to `output`, the reset point `xsol` is beyond
        `xend`, ``ivp_bdf_zero_simple`` will integrate to `xend` with no
        further calls to `output`; if a call to `output` is required at
        the point `xsol` = `xend`, `xsol` must be given precisely the
        value `xend`.

        Parameters
        ~~~~~~~~~~
        xsol : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            The computed solution at the point `xsol`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        xsol : float
            You must set `xsol` to the next value of x at which `output`
            is to be called.

    g : None or callable retval = g(x, y, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `g` must evaluate the function g(x,y) for specified values x,y.

        It specifies the function g for which the first position x where
        g(x,y) = 0 is to be found.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of g at the specified points.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `pw` in `pederv` is spiked (i.e., has unit extent in all but
        one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float
        If `g` is supplied by you, `x` contains the point where g(x,y) =
        0.0, unless g(x,y) != 0.0 anywhere on the range `x` to `xend`,
        in which case, `x` will contain `xend`. If `g` is not supplied
        `x` contains `xend`, unless an error has occurred, when it
        contains the value of x at the error.

    y : float, ndarray, shape (n)
        The computed values of the solution at the final point x = `x`.

    tol : float
        Normally unchanged. However if the range `x` to `xend` is so
        short that a small change in `tol` is unlikely to make any
        change in the computed solution, then, on return, `tol` has its
        sign changed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `relabs` = *<value>*.

            Constraint: `relabs` = 'M', 'A', 'R' or 'D'.

        (`errno` 1)
            On entry, `x` = *<value>* and `xend` = *<value>*.

            Constraint: `x` != `xend`.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            No integration steps have been taken. Progress not possible
            with the input value of `tol` = *<value>*.

        (`errno` 4)
            No integration steps have been taken. `xsol` has been set
            illegally.

        (`errno` 7)
            Integration successful as far as x = *<value>*, but an
            internal error has occurred during rootfinding.

        (`errno` 9)
            Impossible error --- internal variable IERR = *<value>*.

        (`errno` 9)
            Impossible error --- internal variable IREVCM = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Integration successful as far as x = *<value>*, but further
            progress not possible with the input value of `tol` =
            *<value>*.

        (`errno` 5)
            Integration successful as far as x = *<value>*, but `xsol`
            has been reset illegally.

        (`errno` 6)
            No change in sign of the function g(x,y) was detected in the
            integration range.

        (`errno` 8)
            Integration successful as far as x = *<value>*, but an
            internal error has occurred during interpolation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.ode.ivp_bdf_zero_simple_ex.main`
    """

    from ..base.ode import ivp_bdf_zero_simple as b_ivp_bdf_zero_simple

    fun_name = 'naginterfaces.library.ode.ivp_bdf_zero_simple'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, x, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_pederv_name = fun_name + ':pederv'

    def cb_pederv(n, x, y, pw, data=None):
        """Wrapping user's pederv. Called as base pederv."""
        if data is not None:
            _pw_pederv = pederv(
                x,
                y,
                data,
            )
        else:
            _pw_pederv = pederv(
                x,
                y,
            )
        _pw_pederv_np = _utils._to_ndarray(
            {'fun_name': cb_pederv_name, 'returned': 'pw'},
            _pw_pederv,
            copy=False,
            exp_shape=(n, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_pederv_name},
            ((_pw_pederv_np, 'pw'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        pw[:, :] = _pw_pederv_np

    def cb_output(n, xsol, y, data=None):
        """Wrapping user's output. Called as base output."""
        if data is not None:
            _xsol_output = output(
                xsol,
                y,
                data,
            )
        else:
            _xsol_output = output(
                xsol,
                y,
            )
        _xsol_output_base = _xsol_output
        return _xsol_output_base

    def cb_g(n, x, y, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _retval_g = g(
                x,
                y,
                data,
            )
        else:
            _retval_g = g(
                x,
                y,
            )
        return _retval_g

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _b_utils._check_callable(pederv, {'fun_name': fun_name, 'entity': 'pederv'}, allow_none=True)
    _pederv = None if pederv is None else cb_pederv
    _b_utils._check_callable(output, {'fun_name': fun_name, 'entity': 'output'}, allow_none=True)
    _output = None if output is None else cb_output
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'}, allow_none=True)
    _g = None if g is None else cb_g
    try:
        _return_scalars = b_ivp_bdf_zero_simple(
            x,
            xend,
            _n,
            _y,
            _fcn,
            _pederv,
            tol,
            relabs,
            _output,
            _g,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _return_scalars.x
    _y_py = _y
    _tol_py = _return_scalars.tol
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'tol'],
    )(
        _x_py,
        _y_py,
        _tol_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bvp_fd_nonlin_fixedbc(u, v, a, b, tol, fcn, x, np, itrace, comm, data=None, io_manager=None):
    u"""
    Ordinary differential equations, boundary value problem, finite
    difference technique with deferred correction, simple nonlinear
    problem.

    ``bvp_fd_nonlin_fixedbc`` solves a two-point boundary value problem
    with assigned boundary values for a system of ordinary differential
    equations, using a deferred correction technique and a Newton
    iteration.

    For full information please refer to the NAG Library document for
    d02ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02gaf.html

    Parameters
    ----------
    u : float, array-like, shape (n, 2)
        `u`[i-1,0] must be set to the known or estimated value of y_i at
        a and `u`[i-1,1] must be set to the known or estimated value of
        y_i at b, for i = 1,2,...,n.

    v : float, array-like, shape (n, 2)
        `v`[i-1,j-1] must be set to 0.0 if `u`[i-1,j-1] is a known value
        and to 1.0 if `u`[i-1,j-1] is an estimated value, for j =
        1,2,...,2, for i = 1,2,...,n.

    a : float
        a, the left-hand boundary point.

    b : float
        b, the right-hand boundary point.

    tol : float
        A positive absolute error tolerance. If

            a = x_1 < x_2 < ... < x_`np` = b

        is the final mesh, z_j(x_i) is the jth component of the
        approximate solution at x_i, and y_j(x) is the jth component of
        the true solution of equation (1) (see [Description]) and the
        boundary conditions, then, except in extreme cases, it is
        expected that

            |z_j(x_i)-y_j(x_i)| <= `tol`, i = 1,2,...,`np` and j =
            1,2,...,n.

    fcn : callable f = fcn(n, x, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i), for i = 1,2,...,n, at a general point x.

        Parameters
        ~~~~~~~~~~
        n : int
            n, the number of equations.

        x : float
            x, the value of the argument.

        y : float, ndarray, shape (`n`)
            y_i, for i = 1,2,...,n, the value of the argument.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (`n`)
            The values of f_i, for i = 1,2,...,n.

    x : float, array-like, shape (mnp)
        If `np` >= 4 (see `np`), the first `np` elements must define an
        initial mesh. Otherwise the elements of `x` need not be set.

    np : int
        Determines whether a default or user-supplied mesh is used.

        `np` = 0

          A default value of 4 for `np` and a corresponding equispaced
          mesh `x`[0],`x`[1],...,`x`[`np`-1] are used.

        `np` >= 4

          You must define an initial mesh using the array `x` as
          described.

    itrace : int
        If `itrace` = 0 warning messages be suppressed, otherwise
        warning messages will be printed (see [Errors]).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (mnp)
        `x`[0],`x`[1],...,`x`[`np`-1] define the final mesh (with the
        returned value of `np`) satisfying the relation (3).

    y : float, ndarray, shape (n, mnp)
        The approximate solution z_j(x_i) satisfying (2), on the final
        mesh, that is

            `y`[j-1,i-1] = z_j(x_i), i = 1,2,...,`np` and j = 1,2,...,n,

        where `np` is the number of points in the final mesh.

        The remaining columns of `y` are not used.

    np : int
        The number of points in the final (returned) mesh.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The sequence `x` is not strictly increasing. For i =
            *<value>*, `x`[i-1] = *<value>* and `x`[i] = *<value>*.

        (`errno` 1)
            On entry: `b` = *<value>* and `x`[`np`-1] = *<value>*.

            Constraint: `a` = `x`[0] < `x`[1] < ... < `x`[`np`-1] = `b`,
            `np` >= 4.

        (`errno` 1)
            On entry: `a` = *<value>* and `x`[0] = *<value>*.

            Constraint: `a` = `x`[0] < `x`[1] < ... < `x`[`np`-1] = `b`,
            `np` >= 4.

        (`errno` 1)
            The sum of known left and right boundary values must equal
            the number of equations: the number of known left boundary
            values = *<value>*, the number of known right boundary
            values = *<value>*, the number of equations = *<value>*.

        (`errno` 1)
            The number of known right boundary values must be less than
            the number of equations: the number of known right boundary
            values = *<value>*, the number of equations = *<value>*.

        (`errno` 1)
            The number of known left boundary values must be less than
            the number of equations: the number of known left boundary
            values = *<value>*, the number of equations = *<value>*.

        (`errno` 1)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > `a`.

        (`errno` 1)
            On entry, mnp = *<value>*.

            Constraint: mnp >= 32.

        (`errno` 1)
            On entry, `np` = *<value>* and mnp = *<value>*.

            Constraint: `np` <= mnp.

        (`errno` 1)
            On entry, `np` = *<value>*.

            Constraint: `np` = 0 or `np` >= 4.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            The Newton iteration has failed to converge.

            This could be due to there being too few points in the
            initial mesh or to the initial approximate solution being
            too inaccurate. If this latter reason is suspected or you
            cannot make changes to prevent this error, you should use
            the routine with a continuation facility instead.

        (`errno` 5)
            A serious error occurred in a call to the internal
            integrator.

            The error code internally was *<value>*. Please contact NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Newton iteration has reached round-off level.

            If desired accuracy has not been reached, `tol` is too small
            for this problem and this machine precision.

        (`errno` 4)
            A finer mesh is required for the accuracy requested; that
            is, mnp = *<value>* is not large enough.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import bvp_fd_nonlin_fixedbc as b_bvp_fd_nonlin_fixedbc

    fun_name = 'naginterfaces.library.ode.bvp_fd_nonlin_fixedbc'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, x, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                n,
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                n,
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_u, 'u'), (_v, 'v')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mnp_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mnp = _b_utils._EngineIntScalarType(
        _mnp_py, {'fun_name': fun_name, 'entity': '_mnp'},
    ).to_ctypes()
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _mnp_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = b_bvp_fd_nonlin_fixedbc(
            _u,
            _v,
            _n,
            a,
            b,
            tol,
            _fcn,
            _mnp,
            _x,
            _y,
            np,
            itrace,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _np_py = _return_scalar
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'np'],
    )(
        _x_py,
        _y_py,
        _np_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bvp_fd_lin_gen(a, b, tol, fcnf, fcng, c, d, gam, x, np, itrace, comm, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Ordinary differential equations, boundary value problem, finite
    difference technique with deferred correction, general linear
    problem.

    ``bvp_fd_lin_gen`` solves a general linear two-point boundary value
    problem for a system of ordinary differential equations, using a
    deferred correction technique.

    For full information please refer to the NAG Library document for
    d02gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02gbf.html

    Parameters
    ----------
    a : float
        a, the left-hand boundary point.

    b : float
        b, the right-hand boundary point.

    tol : float
        A positive absolute error tolerance. If

            a = x_1 < x_2 < ... < x_`np` = b

        is the final mesh, z(x) is the approximate solution from
        ``bvp_fd_lin_gen`` and y(x) is the true solution of equations
        (1) and (2) then, except in extreme cases, it is expected that

            ||z-y|| <= `tol`

        where

            ||u|| = max_{1 <= i <= n}max_{1 <= j <= `np`}|u_i(x_j)|.

    fcnf : callable f = fcnf(n, x, data=None)
        `fcnf` must evaluate the matrix F(x) in (1) at a general point
        x.

        Parameters
        ~~~~~~~~~~
        n : int
            n, the number of equations.

        x : float
            x, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (`n`, `n`)
            `f`[j-1,i-1] must contain the (i,j)th element of the matrix
            F(x), for j = 1,2,...,n, for i = 1,2,...,n. (See [Example]
            for an example.)

    fcng : callable g = fcng(n, x, data=None)
        `fcng` must evaluate the vector g(x) in (1) at a general point
        x.

        Parameters
        ~~~~~~~~~~
        n : int
            n, the number of equations.

        x : float
            x, the value of the independent variable.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        g : float, array-like, shape (`n`)
            The ith element of the vector g(x), for i = 1,2,...,n. (See
            [Example] for an example.)

    c : float, array-like, shape (n, n)
        The arrays `c` and `d` must be set to the matrices C and D in
        (2)). `gam` must be set to the vector gamma in (2).

    d : float, array-like, shape (n, n)
        The arrays `c` and `d` must be set to the matrices C and D in
        (2)). `gam` must be set to the vector gamma in (2).

    gam : float, array-like, shape (n)
        The arrays `c` and `d` must be set to the matrices C and D in
        (2)). `gam` must be set to the vector gamma in (2).

    x : float, array-like, shape (mnp)
        If `np` >= 4 (see `np`), the first `np` elements must define an
        initial mesh. Otherwise the elements of x need not be set.

    np : int
        Determines whether a default mesh or user-supplied mesh is used.

        `np` = 0

          A default value of 4 for `np` and a corresponding equispaced
          mesh `x`[0],`x`[1],...,`x`[`np`-1] are used.

        `np` >= 4

          You must define an initial mesh `x` as in [equation].

    itrace : int
        If `itrace` = 0 warning messages be suppressed, otherwise
        warning messages will be printed (see [Errors]).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `c` and `d` are spiked (i.e., have unit extent in all but one
        dimension, or have size 1), `spiked_sorder` selects the storage
        order to associate with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    c : float, ndarray, shape (n, n)
        The rows of `c` and `d` and the components of `gam` are
        reordered so that the boundary conditions are in the order:

        - conditions on y(a) only;

        - condition involving y(a) and y(b); and

        - conditions on y(b) only.

        The function will be slightly more efficient if the arrays `c`,
        `d` and `gam` are ordered in this way before entry, and in this
        event they will be unchanged on exit.

        Note that the problems (1) and (2) must be of boundary value
        type, that is neither C nor D may be identically zero.

        Note also that the rank of the matrix [C,D] must be n for the
        problem to be properly posed.

        Any violation of these conditions will lead to an error exit.

    d : float, ndarray, shape (n, n)
        The rows of `c` and `d` and the components of `gam` are
        reordered so that the boundary conditions are in the order:

        - conditions on y(a) only;

        - condition involving y(a) and y(b); and

        - conditions on y(b) only.

        The function will be slightly more efficient if the arrays `c`,
        `d` and `gam` are ordered in this way before entry, and in this
        event they will be unchanged on exit.

        Note that the problems (1) and (2) must be of boundary value
        type, that is neither C nor D may be identically zero.

        Note also that the rank of the matrix [C,D] must be n for the
        problem to be properly posed.

        Any violation of these conditions will lead to an error exit.

    gam : float, ndarray, shape (n)
        The rows of `c` and `d` and the components of `gam` are
        reordered so that the boundary conditions are in the order:

        - conditions on y(a) only;

        - condition involving y(a) and y(b); and

        - conditions on y(b) only.

        The function will be slightly more efficient if the arrays `c`,
        `d` and `gam` are ordered in this way before entry, and in this
        event they will be unchanged on exit.

        Note that the problems (1) and (2) must be of boundary value
        type, that is neither C nor D may be identically zero.

        Note also that the rank of the matrix [C,D] must be n for the
        problem to be properly posed.

        Any violation of these conditions will lead to an error exit.

    x : float, ndarray, shape (mnp)
        `x`[0],`x`[1],...,`x`[`np`-1] define the final mesh (with the
        returned value of `np`) satisfying the relation [equation].

    y : float, ndarray, shape (n, mnp)
        The approximate solution z(x) satisfying [equation], on the
        final mesh, that is

            `y`[j-1,i-1] = z_j(x_i), i = 1,2,...,`np` and j = 1,2,...,n

        where `np` is the number of points in the final mesh.

        The remaining columns of `y` are not used.

    np : int
        The number of points in the final (returned) mesh.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The sequence `x` is not strictly increasing. For i =
            *<value>*, `x`[i-1] = *<value>* and `x`[i] = *<value>*.

        (`errno` 1)
            On entry: `b` = *<value>* and `x`[`np`-1] = *<value>*.

            Constraint: `a` = `x`[0] < `x`[1] < ... < `x`[`np`-1] = `b`,
            `np` >= 4.

        (`errno` 1)
            On entry: `a` = *<value>* and `x`[0] = *<value>*.

            Constraint: `a` = `x`[0] < `x`[1] < ... < `x`[`np`-1] = `b`,
            `np` >= 4.

        (`errno` 1)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > `a`.

        (`errno` 1)
            On entry, mnp = *<value>*.

            Constraint: mnp >= 32.

        (`errno` 1)
            On entry, `np` = *<value>* and mnp = *<value>*.

            Constraint: `np` <= mnp.

        (`errno` 1)
            On entry, `np` = *<value>*.

            Constraint: `np` = 0 or `np` >= 4.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            More than n columns of the n by 2*n matrix [C,D] are
            identically zero, i.e., the boundary conditions are rank
            deficient. The number of non-identically zero columns is
            *<value>*.

        (`errno` 2)
            Row *<value>* of the array `c` and the corresponding row of
            array `d` are identically zero, i.e., the boundary
            conditions are rank deficient.

        (`errno` 2)
            `c` is identically zero; n conditions are set in `d`.

            At least one condition must be on the left. n = *<value>*.

        (`errno` 2)
            `d` is identically zero; n conditions are set in `c`.

            At least one condition must be on the right. n = *<value>*.

        (`errno` 4)
            A serious error occurred in a call to the internal
            integrator.

            The error code internally was *<value>*. Please contact NAG.

        (`errno` 5)
            At least one row of the n by 2*n matrix [C,D] is a linear
            combination of the other rows determined up to a numerical
            tolerance, i.e., the boundary conditions are rank deficient.
            The index of first such row is *<value>*.

        (`errno` 5)
            At least one row of the n by 2*n matrix [C,D] is a linear
            combination of the other rows, i.e., the boundary conditions
            are rank deficient. The index of the first such row is
            *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Newton iteration has reached round-off level.

            If desired accuracy has not been reached, `tol` is too small
            for this problem and this machine precision.

        (`errno` 3)
            The Newton iteration has failed to converge.

            This could be due to there being too few points in the
            initial mesh or to the initial approximate solution being
            too inaccurate. If this latter reason is suspected or you
            cannot make changes to prevent this error, you should use
            the routine with a continuation facility instead.

        (`errno` 3)
            A finer mesh is required for the accuracy requested; that
            is, mnp is not large enough.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import bvp_fd_lin_gen as b_bvp_fd_lin_gen

    fun_name = 'naginterfaces.library.ode.bvp_fd_lin_gen'

    cb_fcnf_name = fun_name + ':fcnf'

    def cb_fcnf(n, x, f, data=None):
        """Wrapping user's fcnf. Called as base fcnf."""
        if data is not None:
            _f_fcnf = fcnf(
                n,
                x,
                data,
            )
        else:
            _f_fcnf = fcnf(
                n,
                x,
            )
        _f_fcnf_np = _utils._to_ndarray(
            {'fun_name': cb_fcnf_name, 'returned': 'f'},
            _f_fcnf,
            copy=False,
            exp_shape=(n, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_fcnf_name},
            ((_f_fcnf_np, 'f'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        f[:, :] = _f_fcnf_np

    cb_fcng_name = fun_name + ':fcng'

    def cb_fcng(n, x, g, data=None):
        """Wrapping user's fcng. Called as base fcng."""
        if data is not None:
            _g_fcng = fcng(
                n,
                x,
                data,
            )
        else:
            _g_fcng = fcng(
                n,
                x,
            )
        g[:] = _utils._to_ndarray(
            {'fun_name': cb_fcng_name, 'returned': 'g'},
            _g_fcng,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_c, 'c'), (_d, 'd')),
        spiked_sorder=spiked_sorder,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcnf, {'fun_name': fun_name, 'entity': 'fcnf'})
    _fcnf = cb_fcnf
    _b_utils._check_callable(fcng, {'fun_name': fun_name, 'entity': 'fcng'})
    _fcng = cb_fcng
    _gam = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'gam'},
        gam,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mnp_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mnp = _b_utils._EngineIntScalarType(
        _mnp_py, {'fun_name': fun_name, 'entity': '_mnp'},
    ).to_ctypes()
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _mnp_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = b_bvp_fd_lin_gen(
            a,
            b,
            _n,
            tol,
            _fcnf,
            _fcng,
            _c,
            _d,
            _gam,
            _mnp,
            _x,
            _y,
            np,
            itrace,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _c_py = _c
    _d_py = _d
    _gam_py = _gam
    _x_py = _x
    _np_py = _return_scalar
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 'd', 'gam', 'x', 'y', 'np'],
    )(
        _c_py,
        _d_py,
        _gam_py,
        _x_py,
        _y_py,
        _np_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bvp_shoot_bval(u, v, a, b, tol, fcn, m1, monlev, comm, data=None, io_manager=None):
    u"""
    Ordinary differential equations, boundary value problem, shooting
    and matching, boundary values to be determined.

    ``bvp_shoot_bval`` solves a two-point boundary value problem for a
    system of ordinary differential equations, using a
    Runge--Kutta--Merson method and a Newton iteration in a shooting and
    matching technique.

    For full information please refer to the NAG Library document for
    d02ha

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02haf.html

    Parameters
    ----------
    u : float, array-like, shape (n, 2)
        `u`[i-1,0] must be set to the known or estimated value of y_i at
        a and `u`[i-1,1] must be set to the known or estimated value of
        y_i at b, for i = 1,2,...,n.

    v : float, array-like, shape (n, 2)
        `v`[i-1,j-1] must be set to 0.0 if `u`[i-1,j-1] is a known value
        and to 1.0 if `u`[i-1,j-1] is an estimated value, for j =
        1,2,...,2, for i = 1,2,...,n.

    a : float
        a, the initial point of the interval of integration.

    b : float
        b, the final point of the interval of integration.

    tol : float
        Must be set to a small quantity suitable for:

        - testing the local error in y_i during integration,

        - testing for the convergence of y_i at b,

        - calculating the perturbation in estimated boundary values for
          y_i, which are used to obtain the approximate derivatives of
          the residuals for use in the Newton iteration.

        You are advised to check your results by varying `tol`.

    fcn : callable f = fcn(n, x, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i), for i = 1,2,...,n, at a general point x.

        Parameters
        ~~~~~~~~~~
        n : int
            n, the number of equations.

        x : float
            x, the value of the argument.

        y : float, ndarray, shape (`n`)
            y_i, for i = 1,2,...,n, the value of the argument.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (`n`)
            The values of f_i(x), for i = 1,2,...,n.

    m1 : int
        A value which controls output.

        `m1` = 1

          The final solution is not evaluated.

        `m1` > 1

          The final values of y_i at interval (b-a)/(`m1`-1) are
          calculated and stored in the array `soln` by columns, starting
          with values y_i at a stored in `soln`[i-1,0], for i =
          1,2,...,n.

    monlev : int
        Setting `monlev` = 0 disables monitoring of the pseudo-Newton
        iteration. Setting `monlev` = 1 enables this monitoring.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    u : float, ndarray, shape (n, 2)
        The known values unaltered, and corrected values of the
        estimates, unless an error has occurred. If an error has
        occurred, `u` contains the known values and the latest values of
        the estimates.

    soln : float, ndarray, shape (n, `m1`)
        The solution when `m1` > 1.

    w : float, ndarray, shape (n, 3*n+17+max(11, `n`))
        If `errno` = 2, 3, 4 or 5, `w`[i-1,0], for i = 1,2,...,n,
        contains the solution at the point where the integration fails
        and the point of failure is returned in `w`[0,1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `monlev` = *<value>*.

            Constraint: `monlev` = 0 or 1.

        (`errno` 1)
            On entry, incorrect number of boundary values were flagged
            as known.

            Number flagged as known: *<value>*, but number should be
            *<value>*.

        (`errno` 1)
            On entry all left-hand boundary values were flagged as
            known.

        (`errno` 1)
            On entry no left-hand boundary values were flagged as known.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            In the integration with initial or final parameters, the
            step size was reduced too far for the integration to
            proceed. Either this function is not a suitable method for
            solving the problem, or the initial choice of parameters is
            very poor.

        (`errno` 3)
            In the integration with initial or final parameters, a
            suitable initial step could not be found. Either this
            function is not suitable for solving the problem, or the
            initial choice of parameters is very poor.

        (`errno` 4)
            An initial step-length could be found for integration to
            proceed with the current parameters.

        (`errno` 5)
            The step-length required to calculate the Jacobian to
            sufficient accuracy is too small

        (`errno` 6)
            The Jacobian has an insignificant column. Make sure that the
            solution vector depends on all the parameters.

        (`errno` 7)
            An internal singular value decomposition has failed.

            This error can be avoided by changing the initial parameter
            estimates.

        (`errno` 8)
            The Newton iteration has failed to converge.

            This can indicate a poor initial choice of parameters or a
            very difficult problem.

            Consider varying elements of the parameter convergence
            control if the residuals are small; otherwise vary initial
            parameter estimates.

        (`errno` 10)
            Internal error in calculating residual. Please contact NAG.

        (`errno` 11)
            Internal error in calculating Jacobian. Please contact NAG.

        (`errno` 12)
            Internal error in Newton method. Please contact NAG.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import bvp_shoot_bval as b_bvp_shoot_bval

    fun_name = 'naginterfaces.library.ode.bvp_shoot_bval'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, x, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                n,
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                n,
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_u, 'u'), (_v, 'v')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _m1 = _b_utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m1_py = _b_utils._EngineIntScalarType.to_py(_m1)
    _soln = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m1_py),
        sorder=_m_sorder,
    )
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, 3*_n_py+17+max(11, _n_py)),
        sorder=_m_sorder,
    )
    b_bvp_shoot_bval(
        _u,
        _v,
        _n,
        a,
        b,
        tol,
        _fcn,
        _soln,
        _m1,
        monlev,
        _w,
        comm,
        data,
        io_manager,
    )
    _u_py = _u
    _soln_py = _soln
    _w_py = _w
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['u', 'soln', 'w'],
    )(
        _u_py,
        _soln_py,
        _w_py,
    )
    return _return_data

def bvp_shoot_genpar(p, pe, e, m1, monlev, fcn, bc, ebndry, comm, data=None, io_manager=None):
    u"""
    Ordinary differential equations, boundary value problem, shooting
    and matching, general parameters to be determined.

    ``bvp_shoot_genpar`` solves a two-point boundary value problem for a
    system of ordinary differential equations, using initial value
    techniques and Newton iteration; it generalizes function
    ``bvp_shoot_bval`` to include the case where parameters other than
    boundary values are to be determined.

    For full information please refer to the NAG Library document for
    d02hb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02hbf.html

    Parameters
    ----------
    p : float, array-like, shape (n1)
        An estimate for the ith argument, p_i, for i = 1,2,...,n_1.

    pe : float, array-like, shape (n1)
        The elements of `pe` must be given small positive values. The
        element `pe`[i-1] is used

        - in the convergence test on the ith argument in the Newton
          iteration, and

        - in perturbing the ith argument when approximating the
          derivatives of the components of the solution with respect to
          this argument for use in the Newton iteration.

        The elements `pe`[i-1] should not be chosen too small.

        They should usually be several orders of magnitude larger than
        machine precision.

    e : float, array-like, shape (n)
        The elements of `e` must be given positive values. The element
        `e`[i-1] is used in the bound on the local error in the ith
        component of the solution y_i during integration.

        The elements `e`[i-1] should not be chosen too small.

        They should usually be several orders of magnitude larger than
        machine precision.

    m1 : int
        A value which controls exit values.

        `m1` = 1

          The final solution is not calculated.

        `m1` > 1

          The final values of the solution at interval (length of
          range)/(`m1`-1) are calculated and stored sequentially in the
          array `soln` starting with the values of the solutions
          evaluated at the first end point (see `ebndry`) stored in the
          first column of `soln`.

    monlev : int
        Setting `monlev` = 0 disables monitoring of the pseudo-Newton
        iteration. Setting `monlev` = 1 enables this monitoring.

    fcn : callable f = fcn(n, n1, x, y, p, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i), for i = 1,2,...,n, at a general point x.

        Parameters
        ~~~~~~~~~~
        n : int
            n, the number of equations.

        n1 : int
            n_1, the number of parameters.

        x : float
            x, the value of the argument.

        y : float, ndarray, shape (`n`)
            y_i, for i = 1,2,...,n, the value of the argument.

        p : float, ndarray, shape (`n1`)
            The current estimate of the argument p_i, for i =
            1,2,...,n_1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (`n`)
            The value of f_i, for i = 1,2,...,n. The f_i may depend upon
            the parameters p_j, for j = 1,2,...,n_1. If there are any
            driving equations (see [Description]) then these must be
            numbered first in the ordering of the components of `f` in
            `fcn`.

    bc : callable (g1, g2) = bc(n, n1, p, data=None)
        `bc` must place in `g1` and `g2` the boundary conditions at a
        and b respectively (see `ebndry`).

        Parameters
        ~~~~~~~~~~
        n : int
            n, the number of equations.

        n1 : int
            n_1, the number of parameters.

        p : float, ndarray, shape (`n1`)
            An estimate of the argument p_i, for i = 1,2,...,n_1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        g1 : float, array-like, shape (`n`)
            The value of y_i(a), (where this may be a known value or a
            function of the parameters p_j, for j = 1,2,...,n_1, for i =
            1,2,...,n).

        g2 : float, array-like, shape (`n`)
            The value of y_i(b), for i = 1,2,...,n, (where these may be
            known values or functions of the parameters p_j, for j =
            1,2,...,n_1). If n > n_1, so that there are some driving
            equations, the first n-n_1 values of `g2` need not be set
            since they are never used.

    ebndry : callable (a, b) = ebndry(n1, p, data=None)
        `ebndry` must evaluate the boundary points a and b, each of
        which may depend on the arguments p_1,p_2,...,p_n_1.

        The integrations in the shooting method are always from a to b.

        Parameters
        ~~~~~~~~~~
        n1 : int
            n_1, the number of parameters.

        p : float, ndarray, shape (`n1`)
            The current estimate of the ith argument, p_i, for i =
            1,2,...,n_1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        a : float
            a, one of the boundary points.

        b : float
            The second boundary point, b. Note that `b` > `a` forces the
            direction of integration to be that of increasing x. If `a`
            and `b` are interchanged the direction of integration is
            reversed.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    p : float, ndarray, shape (n1)
        The corrected value for the ith argument, unless an error has
        occurred, when it contains the last calculated value of the
        argument.

    soln : float, ndarray, shape (n, `m1`)
        The solution when `m1` > 1.

    w : float, ndarray, shape (n, 3*n+14+max(11, `n`))
        With `errno` = 2, 3, 4 or 5 (see [Errors]), `w`[i-1,0], for i =
        1,2,...,n, contains the solution at the point x when the error
        occurred. `w`[0,1] contains x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `monlev` = *<value>*.

            Constraint: `monlev` = 0 or 1.

        (`errno` 1)
            On entry a negative or zero local error tolerance has been
            set.

        (`errno` 1)
            On entry a negative or zero convergence test tolerance has
            been set.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, n = *<value>* and n1 = *<value>*.

            Constraint: n >= n1.

        (`errno` 1)
            On entry, n1 = *<value>*.

            Constraint: n1 >= 1.

        (`errno` 2)
            The step length for the integration became too short to
            proceed when calculating the residual.

        (`errno` 2)
            In the integration with initial or final parameters, the
            step size was reduced too far for the integration to
            proceed. Either this function is not a suitable method for
            solving the problem, or the initial choice of parameters is
            very poor.

        (`errno` 3)
            An initial step-length could be found for integration to
            proceed with the current parameters.

        (`errno` 3)
            In the integration with initial or final parameters, a
            suitable initial step could not be found. Either this
            function is not suitable for solving the problem, or the
            initial choice of parameters is very poor.

        (`errno` 4)
            The step-length required to calculate the Jacobian to
            sufficient accuracy is too small

        (`errno` 5)
            An initial step-length could be found for Jacobian
            calculation to proceed with the current parameters.

        (`errno` 6)
            The Jacobian has an insignificant column. Make sure that the
            solution vector depends on all the parameters.

        (`errno` 7)
            An internal singular value decomposition has failed.

            This error can be avoided by changing the initial parameter
            estimates.

        (`errno` 8)
            The Newton iteration has failed to converge.

            This can indicate a poor initial choice of parameters or a
            very difficult problem.

            Consider varying elements of the parameter convergence
            control if the residuals are small; otherwise vary initial
            parameter estimates.

        (`errno` 9)
            Internal error in Newton method. Please contact NAG.

        (`errno` 10)
            Internal error in calculating Jacobian. Please contact NAG.

        (`errno` 11)
            Internal error in calculating residual. Please contact NAG.

        (`errno` 12)
            Internal error in calculating residual. Please contact NAG.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import bvp_shoot_genpar as b_bvp_shoot_genpar

    fun_name = 'naginterfaces.library.ode.bvp_shoot_genpar'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, n1, x, y, f, p, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                n,
                n1,
                x,
                y,
                p,
                data,
            )
        else:
            _f_fcn = fcn(
                n,
                n1,
                x,
                y,
                p,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_bc_name = fun_name + ':bc'

    def cb_bc(n, n1, g1, g2, p, data=None):
        """Wrapping user's bc. Called as base bc."""
        if data is not None:
            _g1_bc, _g2_bc = bc(
                n,
                n1,
                p,
                data,
            )
        else:
            _g1_bc, _g2_bc = bc(
                n,
                n1,
                p,
            )
        g1[:] = _utils._to_ndarray(
            {'fun_name': cb_bc_name, 'returned': 'g1'},
            _g1_bc,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        g2[:] = _utils._to_ndarray(
            {'fun_name': cb_bc_name, 'returned': 'g2'},
            _g2_bc,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_range(n1, p, data=None):
        """Wrapping user's range. Called as base range."""
        if data is not None:
            _a_range, _b_range = ebndry(
                n1,
                p,
                data,
            )
        else:
            _a_range, _b_range = ebndry(
                n1,
                p,
            )
        _a_range_base = _a_range
        _b_range_base = _b_range
        return (
            _a_range_base,
            _b_range_base,
        )

    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _pe = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'pe'},
        pe,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_e, {'fun_name': fun_name, 'entity': 'e'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m1 = _b_utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m1_py = _b_utils._EngineIntScalarType.to_py(_m1)
    _soln = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m1_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _b_utils._check_callable(bc, {'fun_name': fun_name, 'entity': 'bc'})
    _bc = cb_bc
    _b_utils._check_callable(ebndry, {'fun_name': fun_name, 'entity': 'ebndry'})
    _ebndry = cb_range
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, 3*_n_py+14+max(11, _n_py)),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_bvp_shoot_genpar(
        _p,
        _n1,
        _pe,
        _e,
        _n,
        _soln,
        _m1,
        monlev,
        _fcn,
        _bc,
        _ebndry,
        _w,
        comm,
        data,
        io_manager,
    )
    _p_py = _p
    _soln_py = _soln
    _w_py = _w
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'soln', 'w'],
    )(
        _p_py,
        _soln_py,
        _w_py,
    )
    return _return_data

def bvp_coll_nth(n, cf, bc, x0, x1, k1, kp, comm, data=None):
    u"""
    Ordinary differential equations, boundary value problem, collocation
    and least squares, single nth-order linear equation.

    ``bvp_coll_nth`` solves a regular linear two-point boundary value
    problem for a single nth-order ordinary differential equation by
    Chebyshev series using collocation and least squares.

    For full information please refer to the NAG Library document for
    d02ja

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02jaf.html

    Parameters
    ----------
    n : int
        n, the order of the differential equation.

    cf : callable retval = cf(j, x, data=None)
        `cf` defines the differential equation (see [Description]).

        It must return the value of a function f_j(x) at a given point
        x, where, for 1 <= j <= n+1, f_j(x) is the coefficient of
        y^(j-1)(x) in the equation, and f_0(x) is the right-hand side.

        Parameters
        ~~~~~~~~~~
        j : int
            The index of the function f_j to be evaluated.

        x : float
            The point at which f_j is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f_j(x) at the given point x.

    bc : callable (j, rhs) = bc(i, data=None)
        `bc` defines the boundary conditions, each of which has the form
        y^(k-1)(x_1) = s_k or y^(k-1)(x_0) = s_k.

        The boundary conditions may be specified in any order.

        Parameters
        ~~~~~~~~~~
        i : int
            The index of the boundary condition to be defined.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        j : int
            Must be set to -k if the boundary condition is y^(k-1)(x_0)
            = s_k, and to +k if it is y^(k-1)(x_1) = s_k.

            `j` must not be set to the same value k for two different
            values of `i`.

        rhs : float
            Must be set to the value s_k.

    x0 : float
        The left- and right-hand boundaries, x_0 and x_1, respectively.

    x1 : float
        The left- and right-hand boundaries, x_0 and x_1, respectively.

    k1 : int
        The number of coefficients to be returned in the Chebyshev
        series representation of the solution (hence the degree of the
        polynomial approximation is `k1`-1).

    kp : int
        The number of collocation points to be used.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    c : float, ndarray, shape (`k1`)
        The computed Chebyshev coefficients; that is, the computed
        solution is:

            ⎲_i = 1^`k1``c`[i-1]T_{i-1}(x)

        where T_i(x) is the ith Chebyshev polynomial of the first kind,
        and ⎲ denotes that the first coefficient, `c`[0], is halved.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kp` = *<value>*, `n` = *<value>* and `k1` =
            *<value>*.

            Constraint: `kp`+`n` >= `k1`.

        (`errno` 1)
            On entry, `k1` = *<value>* and `n` = *<value>*.

            Constraint: `k1` >= `n`+1.

        (`errno` 1)
            On entry, `x1` = *<value>* and `x0` = *<value>*.

            Constraint: `x1` > `x0`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            Either the boundary conditions are not linearly independent,
            or the coefficient matrix is rank deficient. Increasing the
            number of collocation points may overcome this latter
            problem.

        (`errno` 4)
            Iterative refinement in the least squares solution has
            failed to converge. The coefficient matrix is too
            ill-conditioned.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import bvp_coll_nth as b_bvp_coll_nth

    def cb_cf(j, x, data=None):
        """Wrapping user's cf. Called as base cf."""
        if data is not None:
            _retval_cf = cf(
                j,
                x,
                data,
            )
        else:
            _retval_cf = cf(
                j,
                x,
            )
        return _retval_cf

    def cb_bc(i, data=None):
        """Wrapping user's bc. Called as base bc."""
        if data is not None:
            _j_bc, _rhs_bc = bc(
                i,
                data,
            )
        else:
            _j_bc, _rhs_bc = bc(
                i,
            )
        _j_bc_base = _j_bc
        _rhs_bc_base = _rhs_bc
        return (
            _j_bc_base,
            _rhs_bc_base,
        )

    fun_name = 'naginterfaces.library.ode.bvp_coll_nth'

    _b_utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = cb_cf
    _b_utils._check_callable(bc, {'fun_name': fun_name, 'entity': 'bc'})
    _bc = cb_bc
    _k1 = _b_utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k1_py = _b_utils._EngineIntScalarType.to_py(_k1)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k1_py,),
    )
    b_bvp_coll_nth(
        n,
        _cf,
        _bc,
        x0,
        x1,
        _k1,
        kp,
        _c,
        comm,
        data,
    )
    _c_py = _c
    return _c_py

def bvp_coll_sys(n, cf, bc, x0, x1, k1, kp, comm, data=None):
    u"""
    Ordinary differential equations, boundary value problem, collocation
    and least squares, system of first-order linear equations.

    ``bvp_coll_sys`` solves a regular linear two-point boundary value
    problem for a system of ordinary differential equations by Chebyshev
    series using collocation and least squares.

    For full information please refer to the NAG Library document for
    d02jb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02jbf.html

    Parameters
    ----------
    n : int
        n, the order of the system of differential equations.

    cf : callable retval = cf(i, j, x, data=None)
        `cf` defines the system of differential equations (see
        [Description]).

        It must return the value of a coefficient function a_{i,j}(x),
        of A, at a given point x, or of a right-hand side function
        r_i(x) if `j` = 0.

        Parameters
        ~~~~~~~~~~
        i : int
            Indicate the function to be evaluated, namely a_{i,j}(x) if
            1 <= `j` <= n, or r_i(x) if `j` = 0.

            1 <= `i` <= n, 0 <= `j` <= n.

        j : int
            Indicate the function to be evaluated, namely a_{i,j}(x) if
            1 <= `j` <= n, or r_i(x) if `j` = 0.

            1 <= `i` <= n, 0 <= `j` <= n.

        x : float
            The point at which the function is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of a coefficient function a_{i,j}(x), of A, at a
            given point x, or of a right-hand side function r_i(x) if
            `j` = 0.

    bc : callable (j, rhs) = bc(i, data=None)
        `bc` defines the n boundary conditions, which have the form
        y_k(x_0) = s or y_k(x_1) = s.

        The boundary conditions may be specified in any order.

        Parameters
        ~~~~~~~~~~
        i : int
            The index of the boundary condition to be defined.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        j : int
            Must be set to -k if the ith boundary condition is y_k(x_0)
            = s, or to +k if it is y_k(x_1) = s.

            `j` must not be set to the same value k for two different
            values of `i`.

        rhs : float
            The value s.

    x0 : float
        The left- and right-hand boundaries, x_0 and x_1, respectively.

    x1 : float
        The left- and right-hand boundaries, x_0 and x_1, respectively.

    k1 : int
        The number of coefficients to be returned in the Chebyshev
        series representation of the components of the solution (hence
        the degree of the polynomial approximation is `k1`-1).

    kp : int
        The number of collocation points to be used.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    c : float, ndarray, shape (`k1`, `n`)
        The computed Chebyshev coefficients of the kth component of the
        solution, y_k; that is, the computed solution is:

            y_k = ⎲_i = 1^`k1``c`[i-1,k-1]T_{i-1}(x), 1 <= k <= n

        where T_i(x) is the ith Chebyshev polynomial of the first kind,
        and ⎲ denotes that the first coefficient, `c`[0,k-1], is halved.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kp` = *<value>* and `k1` = *<value>*.

            Constraint: `kp`+1 >= `k1`.

        (`errno` 1)
            On entry, `k1` = *<value>*.

            Constraint: `k1` >= 2.

        (`errno` 1)
            On entry, `x1` = *<value>* and `x0` = *<value>*.

            Constraint: `x1` > `x0`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            Either the boundary conditions are not linearly independent,
            or the coefficient matrix is rank deficient. Increasing the
            number of collocation points may overcome this latter
            problem.

        (`errno` 4)
            Iterative refinement in the least squares solution has
            failed to converge. The coefficient matrix is too
            ill-conditioned.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import bvp_coll_sys as b_bvp_coll_sys

    def cb_cf(i, j, x, data=None):
        """Wrapping user's cf. Called as base cf."""
        if data is not None:
            _retval_cf = cf(
                i,
                j,
                x,
                data,
            )
        else:
            _retval_cf = cf(
                i,
                j,
                x,
            )
        return _retval_cf

    def cb_bc(i, data=None):
        """Wrapping user's bc. Called as base bc."""
        if data is not None:
            _j_bc, _rhs_bc = bc(
                i,
                data,
            )
        else:
            _j_bc, _rhs_bc = bc(
                i,
            )
        _j_bc_base = _j_bc
        _rhs_bc_base = _rhs_bc
        return (
            _j_bc_base,
            _rhs_bc_base,
        )

    fun_name = 'naginterfaces.library.ode.bvp_coll_sys'

    _b_utils._check_callable(cf, {'fun_name': fun_name, 'entity': 'cf'})
    _cf = cb_cf
    _b_utils._check_callable(bc, {'fun_name': fun_name, 'entity': 'bc'})
    _bc = cb_bc
    _k1 = _b_utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k1_py = _b_utils._EngineIntScalarType.to_py(_k1)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k1_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_bvp_coll_sys(
        _n,
        _cf,
        _bc,
        x0,
        x1,
        _k1,
        kp,
        _c,
        comm,
        data,
    )
    _c_py = _c
    return _c_py

def sl2_reg_finite(xl, xr, coeffn, bcond, k, tol, elam, delam, monit=None, data=None):
    u"""
    Second-order Sturm--Liouville problem, regular system, finite range,
    eigenvalue only.

    ``sl2_reg_finite`` finds a specified eigenvalue of a regular
    second-order Sturm--Liouville system defined on a finite range,
    using a Pruefer transformation and a shooting method.

    For full information please refer to the NAG Library document for
    d02ka

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02kaf.html

    Parameters
    ----------
    xl : float
        The left- and right-hand end points a and b respectively, of the
        interval of definition of the problem.

    xr : float
        The left- and right-hand end points a and b respectively, of the
        interval of definition of the problem.

    coeffn : callable (p, q, dqdl) = coeffn(x, elam, jint, data=None)
        `coeffn` must compute the values of the coefficient functions
        p(x) and q(x;lambda) for given values of x and lambda.
        [Description] states the conditions which p and q must satisfy.

        Parameters
        ~~~~~~~~~~
        x : float
            The current value of x.

        elam : float
            The current trial value of the eigenvalue argument lambda.

        jint : int
            Need not be set.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float
            The value of p(x) for the current value of x.

        q : float
            The value of q(x;lambda) for the current value of x and the
            current trial value of lambda.

        dqdl : float
            The value of dq/d lambda(x;lambda) for the current value of
            x and the current trial value of lambda. However `dqdl` is
            only used in error estimation and, in the rare cases where
            it may be difficult to evaluate, an approximation (say to
            within 20%) will suffice.

    bcond : float, array-like, shape (3, 2)
        `bcond`[0,0] and `bcond`[1,0] must contain the numbers a_1, a_2
        specifying the left-hand boundary condition in the form

            a_2 y(a) = a_1 p(a)y'(a)

        where |a_2|+|a_1 p(a)| != 0.

        `bcond`[0,1] and `bcond`[1,1] must contain b_1, b_2 such that

            b_2 y(b) = b_1 p(b)y'(b)

        where |b_2|+|b_1 p(b)| != 0.

        Note the occurrence of p(a), p(b) in these formulae.

    k : int
        k, the index of the required eigenvalue when the eigenvalues are
        ordered

            lambda_0 < lambda_1 < lambda_2 < ... < lambda_k < ....

    tol : float
        The tolerance argument which determines the accuracy of the
        computed eigenvalue. The error estimate held in `delam` on exit
        satisfies the mixed absolute/relative error test

            `delam` <= `tol`*max(1.0, |`elam`|),

        where `elam` is the final estimate of the eigenvalue. `delam` is
        usually somewhat smaller than the right-hand side of (1) but not
        several orders of magnitude smaller.

    elam : float
        An initial estimate of the eigenvalue lambda~.

    delam : float
        An indication of the scale of the problem in the
        lambda-direction. `delam` holds the initial 'search step'
        (positive or negative). Its value is not critical, but the first
        two trial evaluations are made at `elam` and `elam`+`delam`, so
        the function will work most efficiently if the eigenvalue lies
        between these values. A reasonable choice (if a closer bound is
        not known) is about half the distance between adjacent
        eigenvalues in the neighbourhood of the one sought. In practice,
        there will often be a problem, similar to the one in hand but
        with known eigenvalues, which will help one to choose initial
        values for `elam` and `delam`.

        If `delam` = 0.0 on entry, it is given the default value of
        0.25*max(1.0, |`elam`|).

    monit : None or callable monit(nit, iflag, elam, finfo, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is called by ``sl2_reg_finite`` at the end of each
        iteration for lambda and allows you to monitor the course of the
        computation by printing out the arguments (see [Example] for an
        example).

        Parameters
        ~~~~~~~~~~
        nit : int
            15 minus the number of iterations used so far in the search
            for lambda~. (Up to 15 iterations are permitted.)

        iflag : int
            Describes what phase the computation is in.

            `iflag` < 0

              An error occurred in the computation at this iteration; an
              error exit from ``sl2_reg_finite`` will follow.

            `iflag` = 1

              The function is trying to bracket the eigenvalue lambda~.

            `iflag` = 2

              The function is converging to the eigenvalue lambda~
              (having already bracketed it).

            Normally, the iteration will terminate after a sequence of
            iterates with `iflag` = 2, but occasionally the bracket on
            lambda~ thus determined will not be sufficiently small and
            the iteration will be repeated with tighter accuracy
            control.

        elam : float
            The current trial value of lambda~.

        finfo : float, ndarray, shape (15)
            Information about the behaviour of the shooting method, and
            diagnostic information in the case of errors. It should not
            normally be printed in full if no error has occurred (that
            is, if `iflag` >= 0), though the first few components may be
            of interest to you. In case of an error (`iflag` < 0) all
            the components of `finfo` should be printed.

            The contents of `finfo` are as follows:

            `finfo`[0]

              The current value of the 'miss-distance' or 'residual'
              function f(lambda) on which the shooting method is based.
              f(lambda~) = 0 in theory. This is set to zero if `iflag` <
              0.

            `finfo`[1]

              An estimate of the quantity d lambda defined as follows.
              Consider the perturbation in the miss-distance f(lambda)
              that would result if the local error in the solution of
              the differential equation were always positive and equal
              to its maximum permitted value. Then d lambda is the
              perturbation in lambda that would have the same effect on
              f(lambda). Thus, at the zero of f(lambda),|d lambda| is an
              approximate bound on the perturbation of the zero (that is
              the eigenvalue) caused by errors in numerical solution. If
              d lambda is very large then it is possible that there has
              been a programming error in `coeffn` such that q is
              independent of lambda. If this is the case, an error exit
              with `errno` = 5 should follow. `finfo`[1] is set to zero
              if `iflag` < 0.

            `finfo`[2]

              The number of internal iterations, using the same value of
              lambda and tighter accuracy tolerances, needed to bring
              the accuracy (that is, the value of d lambda) to an
              acceptable value. Its value should normally be 1.0, and
              should almost never exceed 2.0.

            `finfo`[3]

              The number of calls to `coeffn` at this iteration.

            `finfo`[4]

              The number of successful steps taken by the internal
              differential equation solver at this iteration. A step is
              successful if it is used to advance the integration.

            `finfo`[5]

              The number of unsuccessful steps used by the internal
              integrator at this iteration.

            `finfo`[6]

              The number of successful steps at the maximum step size
              taken by the internal integrator at this iteration.

            `finfo`[7]

              Not used.

            `finfo`[8] to `finfo`[14]

              Set to zero, unless `iflag` < 0 in which case they hold
              the following values describing the point of failure:

              `finfo`[8]

                1 or 2 depending on whether integration was in a forward
                or backward direction at the time of failure.

              `finfo`[9]

                The value of the independent variable, x, the point at
                which the error occurred.

              `finfo`[10], `finfo`[11], `finfo`[12]

                The current values of the Pruefer dependent variables
                beta, phi and rho respectively. See [Description] for a
                description of these variables.

              `finfo`[13]

                The local-error tolerance being used by the internal
                integrator at the point of failure.

              `finfo`[14]

                The last integration mesh point.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    bcond : float, ndarray, shape (3, 2)
        `bcond`[2,0] and `bcond`[2,1] hold values sigma_l,sigma_r
        estimating the sensitivity of the computed eigenvalue to changes
        in the boundary conditions. These values should only be of
        interest if the boundary conditions are, in some sense, an
        approximation to some 'true' boundary conditions. For example,
        if the range [`xl`, `xr`] should really be [0, infinity ] but
        instead `xr` has been given a large value and the boundary
        conditions at infinity applied at `xr`, the sensitivity argument
        sigma_r may be of interest. Refer to [Further Comments], for the
        actual meaning of sigma_r and sigma_l.

    elam : float
        The final computed estimate, whether or not an error occurred.

    delam : float
        If no exception or warning is raised, `delam` holds an estimate
        of the absolute error in the computed eigenvalue, that is
        |lambda~-`elam`| ~= `delam`, where lambda~ is the true
        eigenvalue.

        If an exception is raised, `delam` may hold an estimate of the
        error, or its initial value, depending on the value of errno.

        See [Errors] for further details.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            The constraint on the left- or right-hand boundary condition
            has been violated.

        (`errno` 3)
            The function p(x) became zero or changed sign in the
            interval [a,b].

        (`errno` 4)
            After 15 iterations the eigenvalue had not been found to the
            required accuracy.

        (`errno` 5)
            The bracketing phase failed to bracket the eigenvalue within
            ten iterations. This may be due to an error in formulating
            the problem (for example, q is independent of lambda), or by
            very poor initial estimates for the eigenvalue and search
            step.

        (`errno` 6)
            Too small a step size was required at the start of a
            sub-interval to obtain the desired accuracy.

        (`errno` 7)
            Too small a step size was required during solution in a
            sub-interval to obtain the desired accuracy.

        (`errno` 9)
            An internal error has occurred corresponding to a pole of
            the matching function. Try solving the problem again with a
            smaller tolerance.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            The tolerance set is too small for the problem being solved
            and the machine precision being used. The local eigenvalue
            estimate returned is likely to be a very good approximation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import sl2_reg_finite as b_sl2_reg_finite

    def cb_coeffn(x, elam, jint, data=None):
        """Wrapping user's coeffn. Called as base coeffn."""
        if data is not None:
            _p_coeffn, _q_coeffn, _dqdl_coeffn = coeffn(
                x,
                elam,
                jint,
                data,
            )
        else:
            _p_coeffn, _q_coeffn, _dqdl_coeffn = coeffn(
                x,
                elam,
                jint,
            )
        _p_coeffn_base = _p_coeffn
        _q_coeffn_base = _q_coeffn
        _dqdl_coeffn_base = _dqdl_coeffn
        return (
            _p_coeffn_base,
            _q_coeffn_base,
            _dqdl_coeffn_base,
        )

    def cb_monit(nit, iflag, elam, finfo, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                nit,
                iflag,
                elam,
                finfo,
                data,
            )
        else:
            monit(
                nit,
                iflag,
                elam,
                finfo,
            )

    fun_name = 'naginterfaces.library.ode.sl2_reg_finite'

    _b_utils._check_callable(coeffn, {'fun_name': fun_name, 'entity': 'coeffn'})
    _coeffn = cb_coeffn
    _bcond = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bcond'},
        bcond,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    try:
        _return_scalars = b_sl2_reg_finite(
            xl,
            xr,
            _coeffn,
            _bcond,
            k,
            tol,
            elam,
            delam,
            _monit,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _bcond_py = _bcond
    _elam_py = _return_scalars.elam
    _delam_py = _return_scalars.delam
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bcond', 'elam', 'delam'],
    )(
        _bcond_py,
        _elam_py,
        _delam_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sl2_breaks_vals(xpoint, coeffn, bdyval, k, tol, elam, delam, hmax, maxit=0, maxfun=0, monit=None, data=None):
    u"""
    Second-order Sturm--Liouville problem, regular/singular system,
    finite/infinite range, eigenvalue only, user-specified break-points.

    ``sl2_breaks_vals`` finds a specified eigenvalue of a regular or
    singular second-order Sturm--Liouville system on a finite or
    infinite interval, using a Pruefer transformation and a shooting
    method.
    Provision is made for discontinuities in the coefficient functions
    or their derivatives.

    For full information please refer to the NAG Library document for
    d02kd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02kdf.html

    Parameters
    ----------
    xpoint : float, array-like, shape (m)
        The points where the boundary conditions computed by `bdyval`
        are to be imposed, and also any break-points, i.e., `xpoint`[0]
        to `xpoint`[m-1] must contain values x_1,...,x_m such that

            x_1 <= x_2 < x_3 < ... < x_{m-1} <= x_m

        with the following meanings:

        - x_1 and x_m are the left- and right-hand end points, a and b,
          of the domain of definition of the Sturm--Liouville system if
          these are finite. If either a or b is infinite, the
          corresponding value x_1 or x_m may be a more-or-less
          arbitrarily 'large' number of appropriate sign.

        - x_2 and x_{m-1} are the Boundary Matching Points (BMPs), that
          is the points at which the left and right boundary conditions
          computed in `bdyval` are imposed.

          If the left-hand end point is a regular point then you should
          set x_2 = x_1 (= a), while if it is a singular point you must
          set x_2 > x_1.

          Similarly x_{m-1} = x_m (= b) if the right-hand end point is
          regular, and x_{m-1} < x_m if it is singular.

        - The remaining m-4 points x_3,...,x_{m-2}, if any, define
          'break-points' which divide the interval [x_2,x_{m-1}] into
          m-3 sub-intervals

              i_1 = [x_2,x_3],...,i_{m-3} = [x_{m-2},x_{m-1}].

          Numerical integration of the differential equation is stopped
          and restarted at each break-point. In simple cases no
          break-points are needed. However, if p(x) or q(x;lambda) are
          given by different formulae in different parts of the
          interval, integration is more efficient if the range is broken
          up by break-points in the appropriate way. Similarly points
          where any jumps occur in p(x) or q(x;lambda), or in their
          derivatives up to the fifth-order, should appear as
          break-points.

          Examples are given in [Further Comments]. `xpoint` determines
          the position of the Shooting Matching Point (SMP), as
          explained in [Further Comments].

    coeffn : callable (p, q, dqdl) = coeffn(x, elam, jint, data=None)
        `coeffn` must compute the values of the coefficient functions
        p(x) and q(x;lambda) for given values of x and lambda.
        [Description] states the conditions which p and q must satisfy.

        See [Further Comments] for examples.

        Parameters
        ~~~~~~~~~~
        x : float
            The current value of x.

        elam : float
            The current trial value of the eigenvalue argument lambda.

        jint : int
            The index j of the sub-interval i_j (see specification of
            `xpoint`) in which x lies.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float
            The value of p(x) for the current value of x.

        q : float
            The value of q(x;lambda) for the current value of x and the
            current trial value of lambda.

        dqdl : float
            The value of dq/d lambda(x;lambda) for the current value of
            x and the current trial value of lambda. However `dqdl` is
            only used in error estimation and, in the rare cases where
            it may be difficult to evaluate, an approximation (say to
            within 20%) will suffice.

    bdyval : callable (yl, yr) = bdyval(xl, xr, elam, data=None)
        `bdyval` must define the boundary conditions.

        For each end point, `bdyval` must return (in `yl` or `yr`)
        values of y(x) and p(x)y'(x) which are consistent with the
        boundary conditions at the end points; only the ratio of the
        values matters.

        Here x is a given point (`xl` or `xr`) equal to, or close to,
        the end point.

        Parameters
        ~~~~~~~~~~
        xl : float
            If a is a regular end point of the system (so that a = x_1 =
            x_2), `xl` contains a. If a is a singular point (so that a
            <= x_1 < x_2), `xl` contains a point x such that x_1 < x <=
            x_2.

        xr : float
            If b is a regular end point of the system (so that x_{m-1} =
            x_m = b), `xr` contains b. If b is a singular point (so that
            x_{m-1} < x_m <= b), `xr` contains a point x such that
            x_{m-1} <= x < x_m.

        elam : float
            The current trial value of lambda.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        yl : float, array-like, shape (3)
            `yl`[0] and `yl`[1] should contain values of y(x) and
            p(x)y'(x) respectively (not both zero) which are consistent
            with the boundary condition at the left-hand end point,
            given by x = `xl`. `yl`[2] should not be set.

        yr : float, array-like, shape (3)
            `yr`[0] and `yr`[1] should contain values of y(x) and
            p(x)y'(x) respectively (not both zero) which are consistent
            with the boundary condition at the right-hand end point,
            given by x = `xr`. `yr`[2] should not be set.

    k : int
        k, the index of the required eigenvalue when the eigenvalues are
        ordered

            lambda_0 < lambda_1 < lambda_2 < ... < lambda_k < ....

    tol : float
        The tolerance argument which determines the accuracy of the
        computed eigenvalue. The error estimate held in `delam` on exit
        satisfies the mixed absolute/relative error test

            `delam` <= `tol`*max(1.0, |`elam`|),

        where `elam` is the final estimate of the eigenvalue. `delam` is
        usually somewhat smaller than the right-hand side of (1) but not
        several orders of magnitude smaller.

    elam : float
        An initial estimate of the eigenvalue lambda~.

    delam : float
        An indication of the scale of the problem in the
        lambda-direction. `delam` holds the initial 'search step'
        (positive or negative). Its value is not critical, but the first
        two trial evaluations are made at `elam` and `elam`+`delam`, so
        the function will work most efficiently if the eigenvalue lies
        between these values. A reasonable choice (if a closer bound is
        not known) is half the distance between adjacent eigenvalues in
        the neighbourhood of the one sought. In practice, there will
        often be a problem, similar to the one in hand but with known
        eigenvalues, which will help one to choose initial values for
        `elam` and `delam`.

        If `delam` = 0.0 on entry, it is given the default value of
        0.25*max(1.0, |`elam`|).

    hmax : float, array-like, shape (2, m)
        `hmax`[0,j-1] should contain a maximum step size to be used by
        the differential equation code in the jth sub-interval i_j (as
        described in the specification of argument `xpoint`), for j =
        1,2,...,m-3. If it is zero the function generates a maximum step
        size internally.

        It is recommended that `hmax`[0,j-1] be set to zero unless the
        coefficient functions p and q have features (such as a narrow
        peak) within the jth sub-interval that could be 'missed' if a
        long step were taken.

        In such a case `hmax`[0,j-1] should be set to about half the
        distance over which the feature should be observed.

        Too small a value will increase the computing time for the
        function.

        See [Further Comments] for further suggestions.

        The rest of the array is used as workspace.

    maxit : int, optional
        A bound on n_r, the number of root-finding iterations allowed,
        that is the number of trial values of lambda that are used. If
        `maxit` <= 0, no such bound is assumed. (See also `maxfun`.)

    maxfun : int, optional
        A bound on n_f, the number of calls to `coeffn` made in any one
        root-finding iteration. If `maxfun` <= 0, no such bound is
        assumed.

    monit : None or callable monit(nit, iflag, elam, finfo, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is called by ``sl2_breaks_vals`` at the end of each
        root-finding iteration and allows you to monitor the course of
        the computation by printing out the arguments (see [Example] for
        an example).

        Parameters
        ~~~~~~~~~~
        nit : int
            The current value of the argument `maxit` of
            ``sl2_breaks_vals``, this is decreased by one at each
            iteration.

        iflag : int
            Describes what phase the computation is in.

            `iflag` < 0

              An error occurred in the computation at this iteration; an
              error exit from ``sl2_breaks_vals`` with errno = -`iflag`
              will follow.

            `iflag` = 1

              The function is trying to bracket the eigenvalue lambda~.

            `iflag` = 2

              The function is converging to the eigenvalue lambda~
              (having already bracketed it).

        elam : float
            The current trial value of lambda.

        finfo : float, ndarray, shape (15)
            Information about the behaviour of the shooting method, and
            diagnostic information in the case of errors. It should not
            normally be printed in full if no error has occurred (that
            is, if `iflag` > 0), though the first few components may be
            of interest to you. In case of an error (`iflag` < 0) all
            the components of `finfo` should be printed.

            The contents of `finfo` are as follows:

            `finfo`[0]

              The current value of the 'miss-distance' or 'residual'
              function f(lambda) on which the shooting method is based.
              (See [Further Comments] for further information.)
              `finfo`[0] is set to zero if `iflag` < 0.

            `finfo`[1]

              An estimate of the quantity d lambda defined as follows.
              Consider the perturbation in the miss-distance f(lambda)
              that would result if the local error in the solution of
              the differential equation were always positive and equal
              to its maximum permitted value. Then d lambda is the
              perturbation in lambda that would have the same effect on
              f(lambda). Thus, at the zero of f(lambda),|d lambda| is an
              approximate bound on the perturbation of the zero (that is
              the eigenvalue) caused by errors in numerical solution. If
              d lambda is very large then it is possible that there has
              been a programming error in `coeffn` such that q is
              independent of lambda. If this is the case, an error exit
              with `errno` = 5 should follow. `finfo`[1] is set to zero
              if `iflag` < 0.

            `finfo`[2]

              The number of internal iterations, using the same value of
              lambda and tighter accuracy tolerances, needed to bring
              the accuracy (that is, the value of d lambda) to an
              acceptable value. Its value should normally be 1.0, and
              should almost never exceed 2.0.

            `finfo`[3]

              The number of calls to `coeffn` at this iteration.

            `finfo`[4]

              The number of successful steps taken by the internal
              differential equation solver at this iteration. A step is
              successful if it is used to advance the integration.

            `finfo`[5]

              The number of unsuccessful steps used by the internal
              integrator at this iteration.

            `finfo`[6]

              The number of successful steps at the maximum step size
              taken by the internal integrator at this iteration.

            `finfo`[7]

              Not used.

            `finfo`[8] to `finfo`[14]

              Set to zero, unless `iflag` < 0 in which case they hold
              the following values describing the point of failure:

              `finfo`[8]

                The index of the sub-interval where failure occurred, in
                the range 1 to m-3. In case of an error in `bdyval`, it
                is set to 0 or m-2 depending on whether the left or
                right boundary condition caused the error.

              `finfo`[9]

                The value of the independent variable, x, the point at
                which the error occurred. In case of an error in
                `bdyval`, it is set to the value of `xl` or `xr` as
                appropriate (see the specification of `bdyval`).

              `finfo`[10], `finfo`[11], `finfo`[12]

                The current values of the Pruefer dependent variables
                beta, phi and rho respectively. These are set to zero in
                case of an error in `bdyval`. (See ``sl2_breaks_funs``
                for a description of these variables.)

              `finfo`[13]

                The local-error tolerance being used by the internal
                integrator at the point of failure. This is set to zero
                in the case of an error in `bdyval`.

              `finfo`[14]

                The last integration mesh point. This is set to zero in
                the case of an error in `bdyval`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    elam : float
        The final computed estimate, whether or not an error occurred.

    delam : float
        If no exception or warning is raised, `delam` holds an estimate
        of the absolute error in the computed eigenvalue, that is
        |lambda~-`elam`| ~= `delam`. (In [Further Comments] we discuss
        the assumptions under which this is true.) The true error is
        rarely more than twice, or less than a tenth, of the estimated
        error.

        If an exception is raised, `delam` may hold an estimate of the
        error, or its initial value, depending on the value of errno.

        See [Errors] for further details.

    hmax : float, ndarray, shape (2, m)
        `hmax`[0,m-2] and `hmax`[0,m-1] contain the sensitivity
        coefficients sigma_l,sigma_r, described in [Further Comments].
        Other entries contain diagnostic output in the case of an error
        exit (see [Errors]).

    maxit : int
        Will have been decreased by the number of iterations actually
        performed, whether or not it was positive on entry.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, the sequence of boundary and break points is not
            monotonic.

            For i = *<value>*, point i-1 = *<value>* and point i =
            *<value>*.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 4.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            The constraint on the specification of left- or right-hand
            boundary condition has been violated.

        (`errno` 3)
            The function p(x) became zero or changed sign in the
            interval [a,b].

        (`errno` 6)
            The last iteration was terminated because more than
            *<value>* calls to evaluate p, q and its derivative were
            performed.

        (`errno` 7)
            Too small a step size was required at the start of a
            sub-interval to obtain the desired accuracy.

        (`errno` 8)
            Too small a step size was required during solution in a
            sub-interval to obtain the desired accuracy.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            After *<value>* iterations the eigenvalue had not been found
            to the required accuracy.

        (`errno` 5)
            The bracketing phase failed to bracket the eigenvalue within
            ten iterations. This may be due to an error in formulating
            the problem (for example, q is independent of lambda), or by
            very poor initial estimates for the eigenvalue and search
            step.

        (`errno` 9)
            The tolerance set is too small for the problem being solved
            and the machine precision being used. The local eigenvalue
            estimate returned is likely to be a very good approximation.

        (`errno` 10)
            An internal error has occurred corresponding to a pole of
            the matching function. Try solving the problem again with a
            smaller tolerance.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import sl2_breaks_vals as b_sl2_breaks_vals

    def cb_coeffn(x, elam, jint, data=None):
        """Wrapping user's coeffn. Called as base coeffn."""
        if data is not None:
            _p_coeffn, _q_coeffn, _dqdl_coeffn = coeffn(
                x,
                elam,
                jint,
                data,
            )
        else:
            _p_coeffn, _q_coeffn, _dqdl_coeffn = coeffn(
                x,
                elam,
                jint,
            )
        _p_coeffn_base = _p_coeffn
        _q_coeffn_base = _q_coeffn
        _dqdl_coeffn_base = _dqdl_coeffn
        return (
            _p_coeffn_base,
            _q_coeffn_base,
            _dqdl_coeffn_base,
        )

    fun_name = 'naginterfaces.library.ode.sl2_breaks_vals'

    cb_bdyval_name = fun_name + ':bdyval'

    def cb_bdyval(xl, xr, elam, yl, yr, data=None):
        """Wrapping user's bdyval. Called as base bdyval."""
        if data is not None:
            _yl_bdyval, _yr_bdyval = bdyval(
                xl,
                xr,
                elam,
                data,
            )
        else:
            _yl_bdyval, _yr_bdyval = bdyval(
                xl,
                xr,
                elam,
            )
        yl[:] = _utils._to_ndarray(
            {'fun_name': cb_bdyval_name, 'returned': 'yl'},
            _yl_bdyval,
            copy=False,
            exp_shape=(3,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        yr[:] = _utils._to_ndarray(
            {'fun_name': cb_bdyval_name, 'returned': 'yr'},
            _yr_bdyval,
            copy=False,
            exp_shape=(3,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_monit(nit, iflag, elam, finfo, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                nit,
                iflag,
                elam,
                finfo,
                data,
            )
        else:
            monit(
                nit,
                iflag,
                elam,
                finfo,
            )

    _xpoint = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xpoint'},
        xpoint,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_xpoint, {'fun_name': fun_name, 'entity': 'xpoint'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _b_utils._check_callable(coeffn, {'fun_name': fun_name, 'entity': 'coeffn'})
    _coeffn = cb_coeffn
    _b_utils._check_callable(bdyval, {'fun_name': fun_name, 'entity': 'bdyval'})
    _bdyval = cb_bdyval
    _hmax = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'hmax'},
        hmax,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    try:
        _return_scalars = b_sl2_breaks_vals(
            _xpoint,
            _m,
            _coeffn,
            _bdyval,
            k,
            tol,
            elam,
            delam,
            _hmax,
            maxit,
            maxfun,
            _monit,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _elam_py = _return_scalars.elam
    _delam_py = _return_scalars.delam
    _hmax_py = _hmax
    _maxit_py = _return_scalars.maxit
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['elam', 'delam', 'hmax', 'maxit'],
    )(
        _elam_py,
        _delam_py,
        _hmax_py,
        _maxit_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sl2_breaks_funs(xpoint, match, coeffn, bdyval, k, tol, elam, delam, hmax, report, maxit=0, maxfun=0, monit=None, data=None):
    u"""
    Second-order Sturm--Liouville problem, regular/singular system,
    finite/infinite range, eigenvalue and eigenfunction, user-specified
    break-points.

    ``sl2_breaks_funs`` finds a specified eigenvalue of a regular or
    singular second-order Sturm--Liouville system on a finite or
    infinite interval, using a Pruefer transformation and a shooting
    method.
    It also reports values of the eigenfunction and its derivatives.
    Provision is made for discontinuities in the coefficient functions
    or their derivatives.

    For full information please refer to the NAG Library document for
    d02ke

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02kef.html

    Parameters
    ----------
    xpoint : float, array-like, shape (m)
        The points where the boundary conditions computed by `bdyval`
        are to be imposed, and also any break-points, i.e., `xpoint`[0]
        to `xpoint`[m-1] must contain values x_1,...,x_m such that

            x_1 <= x_2 < x_3 < ... < x_{m-1} <= x_m

        with the following meanings:

        - x_1 and x_m are the left- and right-hand end points, a and b,
          of the domain of definition of the Sturm--Liouville system if
          these are finite. If either a or b is infinite, the
          corresponding value x_1 or x_m may be a more-or-less
          arbitrarily 'large' number of appropriate sign.

        - x_2 and x_{m-1} are the Boundary Matching Points (BMPs), that
          is the points at which the left and right boundary conditions
          computed in `bdyval` are imposed.

          If the left-hand end point is a regular point then you should
          set x_2 = x_1 (= a), while if it is a singular point you must
          set x_2 > x_1.

          Similarly x_{m-1} = x_m (= b) if the right-hand end point is
          regular, and x_{m-1} < x_m if it is singular.

        - The remaining m-4 points x_3,...,x_{m-2}, if any, define
          'break-points' which divide the interval [x_2,x_{m-1}] into
          m-3 sub-intervals

              i_1 = [x_2,x_3],...,i_{m-3} = [x_{m-2},x_{m-1}].

          Numerical integration of the differential equation is stopped
          and restarted at each break-point. In simple cases no
          break-points are needed. However, if p(x) or q(x;lambda) are
          given by different formulae in different parts of the
          interval, integration is more efficient if the range is broken
          up by break-points in the appropriate way. Similarly points
          where any jumps occur in p(x) or q(x;lambda), or in their
          derivatives up to the fifth-order, should appear as
          break-points.

          Examples are given in [Further Comments]. `xpoint` determines
          the position of the Shooting Matching Point (SMP), as
          explained in [Further Comments].

    match : int
        Must be set to the index of the 'break-point' to be used as the
        matching point (see [Further Comments]). If `match` is set to a
        value outside the range [2,m-1] then a default value is taken,
        corresponding to the break-point nearest the centre of the
        interval [`xpoint`[1],`xpoint`[m-1]].

    coeffn : callable (p, q, dqdl) = coeffn(x, elam, jint, data=None)
        `coeffn` must compute the values of the coefficient functions
        p(x) and q(x;lambda) for given values of x and lambda.
        [Description] states the conditions which p and q must satisfy.

        See [Further Comments] for examples.

        Parameters
        ~~~~~~~~~~
        x : float
            The current value of x.

        elam : float
            The current trial value of the eigenvalue argument lambda.

        jint : int
            The index j of the sub-interval i_j (see specification of
            `xpoint`) in which x lies.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float
            The value of p(x) for the current value of x.

        q : float
            The value of q(x;lambda) for the current value of x and the
            current trial value of lambda.

        dqdl : float
            The value of dq/d lambda(x;lambda) for the current value of
            x and the current trial value of lambda. However `dqdl` is
            only used in error estimation and, in the rare cases where
            it may be difficult to evaluate, an approximation (say to
            within 20%) will suffice.

    bdyval : callable (yl, yr) = bdyval(xl, xr, elam, data=None)
        `bdyval` must define the boundary conditions.

        For each end point, `bdyval` must return (in `yl` or `yr`)
        values of y(x) and p(x)y'(x) which are consistent with the
        boundary conditions at the end points; only the ratio of the
        values matters.

        Here x is a given point (`xl` or `xr`) equal to, or close to,
        the end point.

        Parameters
        ~~~~~~~~~~
        xl : float
            If a is a regular end point of the system (so that a = x_1 =
            x_2), `xl` contains a. If a is a singular point (so that a
            <= x_1 < x_2), `xl` contains a point x such that x_1 < x <=
            x_2.

        xr : float
            If b is a regular end point of the system (so that x_{m-1} =
            x_m = b), `xr` contains b. If b is a singular point (so that
            x_{m-1} < x_m <= b), `xr` contains a point x such that
            x_{m-1} <= x < x_m.

        elam : float
            The current trial value of lambda.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        yl : float, array-like, shape (3)
            `yl`[0] and `yl`[1] should contain values of y(x) and
            p(x)y'(x) respectively (not both zero) which are consistent
            with the boundary condition at the left-hand end point,
            given by x = `xl`. `yl`[2] should not be set.

        yr : float, array-like, shape (3)
            `yr`[0] and `yr`[1] should contain values of y(x) and
            p(x)y'(x) respectively (not both zero) which are consistent
            with the boundary condition at the right-hand end point,
            given by x = `xr`. `yr`[2] should not be set.

    k : int
        k, the index of the required eigenvalue when the eigenvalues are
        ordered

            lambda_0 < lambda_1 < lambda_2 < ... < lambda_k < ....

    tol : float
        The tolerance argument which determines the accuracy of the
        computed eigenvalue. The error estimate held in `delam` on exit
        satisfies the mixed absolute/relative error test

            `delam` <= `tol`*max(1.0, |`elam`|),

        where `elam` is the final estimate of the eigenvalue. `delam` is
        usually somewhat smaller than the right-hand side of (1) but not
        several orders of magnitude smaller.

    elam : float
        An initial estimate of the eigenvalue lambda~.

    delam : float
        An indication of the scale of the problem in the
        lambda-direction. `delam` holds the initial 'search step'
        (positive or negative). Its value is not critical, but the first
        two trial evaluations are made at `elam` and `elam`+`delam`, so
        the function will work most efficiently if the eigenvalue lies
        between these values. A reasonable choice (if a closer bound is
        not known) is half the distance between adjacent eigenvalues in
        the neighbourhood of the one sought. In practice, there will
        often be a problem, similar to the one in hand but with known
        eigenvalues, which will help one to choose initial values for
        `elam` and `delam`.

        If `delam` = 0.0 on entry, it is given the default value of
        0.25*max(1.0, |`elam`|).

    hmax : float, array-like, shape (2, m)
        `hmax`[0,j-1] should contain a maximum step size to be used by
        the differential equation code in the jth sub-interval i_j (as
        described in the specification of argument `xpoint`), for j =
        1,2,...,m-3. If it is zero the function generates a maximum step
        size internally.

        It is recommended that `hmax`[0,j-1] be set to zero unless the
        coefficient functions p and q have features (such as a narrow
        peak) within the jth sub-interval that could be 'missed' if a
        long step were taken.

        In such a case `hmax`[0,j-1] should be set to about half the
        distance over which the feature should be observed.

        Too small a value will increase the computing time for the
        function.

        See [Further Comments] for further suggestions.

        The rest of the array is used as workspace.

    report : callable report(x, v, jint, data=None)
        `report` provides the means by which you may compute the
        eigenfunction y(x) and its derivative at each integration mesh
        point x. (See [Further Comments] for an example.)

        Parameters
        ~~~~~~~~~~
        x : float
            The current value of the independent variable x. See
            [Further Comments] for the order in which values of x are
            supplied.

        v : float, ndarray, shape (3)
            `v`[0], `v`[1], `v`[2] hold the current values of the
            Pruefer variables beta, phi, rho respectively.

        jint : int
            Indicates the sub-interval between break-points in which `x`
            lies exactly as for `coeffn`, except that at the extreme
            left-hand end point (when x = `xpoint`[1]) `jint` is set to
            0 and at the extreme right-hand end point (when x = x_r =
            `xpoint`[m-1]) `jint` is set to m-2.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    maxit : int, optional
        A bound on n_r, the number of root-finding iterations allowed,
        that is the number of trial values of lambda that are used. If
        `maxit` <= 0, no such bound is assumed. (See also `maxfun`.)

    maxfun : int, optional
        A bound on n_f, the number of calls to `coeffn` made in any one
        root-finding iteration. If `maxfun` <= 0, no such bound is
        assumed.

    monit : None or callable monit(nit, iflag, elam, finfo, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is called by ``sl2_breaks_funs`` at the end of each
        root-finding iteration and allows you to monitor the course of
        the computation by printing out the arguments (see [Example] for
        an example).

        Parameters
        ~~~~~~~~~~
        nit : int
            The current value of the argument `maxit` of
            ``sl2_breaks_funs``, this is decreased by one at each
            iteration.

        iflag : int
            Describes what phase the computation is in.

            `iflag` < 0

              An error occurred in the computation at this iteration; an
              error exit from ``sl2_breaks_funs`` with errno = -`iflag`
              will follow.

            `iflag` = 1

              The function is trying to bracket the eigenvalue lambda~.

            `iflag` = 2

              The function is converging to the eigenvalue lambda~
              (having already bracketed it).

        elam : float
            The current trial value of lambda.

        finfo : float, ndarray, shape (15)
            Information about the behaviour of the shooting method, and
            diagnostic information in the case of errors. It should not
            normally be printed in full if no error has occurred (that
            is, if `iflag` > 0), though the first few components may be
            of interest to you. In case of an error (`iflag` < 0) all
            the components of `finfo` should be printed.

            The contents of `finfo` are as follows:

            `finfo`[0]

              The current value of the 'miss-distance' or 'residual'
              function f(lambda) on which the shooting method is based.
              (See [Further Comments] for further information.)
              `finfo`[0] is set to zero if `iflag` < 0.

            `finfo`[1]

              An estimate of the quantity d lambda defined as follows.
              Consider the perturbation in the miss-distance f(lambda)
              that would result if the local error in the solution of
              the differential equation were always positive and equal
              to its maximum permitted value. Then d lambda is the
              perturbation in lambda that would have the same effect on
              f(lambda). Thus, at the zero of f(lambda),|d lambda| is an
              approximate bound on the perturbation of the zero (that is
              the eigenvalue) caused by errors in numerical solution. If
              d lambda is very large then it is possible that there has
              been a programming error in `coeffn` such that q is
              independent of lambda. If this is the case, an error exit
              with `errno` = 5 should follow. `finfo`[1] is set to zero
              if `iflag` < 0.

            `finfo`[2]

              The number of internal iterations, using the same value of
              lambda and tighter accuracy tolerances, needed to bring
              the accuracy (that is, the value of d lambda) to an
              acceptable value. Its value should normally be 1.0, and
              should almost never exceed 2.0.

            `finfo`[3]

              The number of calls to `coeffn` at this iteration.

            `finfo`[4]

              The number of successful steps taken by the internal
              differential equation solver at this iteration. A step is
              successful if it is used to advance the integration.

            `finfo`[5]

              The number of unsuccessful steps used by the internal
              integrator at this iteration.

            `finfo`[6]

              The number of successful steps at the maximum step size
              taken by the internal integrator at this iteration.

            `finfo`[7]

              Not used.

            `finfo`[8] to `finfo`[14]

              Set to zero, unless `iflag` < 0 in which case they hold
              the following values describing the point of failure:

              `finfo`[8]

                The index of the sub-interval where failure occurred, in
                the range 1 to m-3. In case of an error in `bdyval`, it
                is set to 0 or m-2 depending on whether the left or
                right boundary condition caused the error.

              `finfo`[9]

                The value of the independent variable, x, the point at
                which the error occurred. In case of an error in
                `bdyval`, it is set to the value of `xl` or `xr` as
                appropriate (see the specification of `bdyval`).

              `finfo`[10], `finfo`[11], `finfo`[12]

                The current values of the Pruefer dependent variables
                beta, phi and rho respectively. These are set to zero in
                case of an error in `bdyval`.

              `finfo`[13]

                The local-error tolerance being used by the internal
                integrator at the point of failure. This is set to zero
                in the case of an error in `bdyval`.

              `finfo`[14]

                The last integration mesh point. This is set to zero in
                the case of an error in `bdyval`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    match : int
        The index of the break-point actually used as the matching
        point.

    elam : float
        The final computed estimate, whether or not an error occurred.

    delam : float
        If no exception or warning is raised, `delam` holds an estimate
        of the absolute error in the computed eigenvalue, that is
        |lambda~-`elam`| ~= `delam`. (In [Further Comments] we discuss
        the assumptions under which this is true.) The true error is
        rarely more than twice, or less than a tenth, of the estimated
        error.

        If an exception is raised, `delam` may hold an estimate of the
        error, or its initial value, depending on the value of errno.

        See [Errors] for further details.

    hmax : float, ndarray, shape (2, m)
        `hmax`[0,m-2] and `hmax`[0,m-1] contain the sensitivity
        coefficients sigma_l,sigma_r, described in [Further Comments].
        Other entries contain diagnostic output in the case of an error
        exit (see [Errors]).

    maxit : int
        Will have been decreased by the number of iterations actually
        performed, whether or not it was positive on entry.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 4.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            On entry, the sequence of boundary and break points is not
            monotonic.

            For i = *<value>*, point i-1 = *<value>* and point i =
            *<value>*.

        (`errno` 2)
            The constraint on the specification of left- or right-hand
            boundary condition has been violated.

        (`errno` 3)
            The function p(x) became zero or changed sign in the
            interval [a,b].

        (`errno` 4)
            After *<value>* iterations the eigenvalue had not been found
            to the required accuracy.

        (`errno` 5)
            The bracketing phase failed to bracket the eigenvalue within
            ten iterations. This may be due to an error in formulating
            the problem (for example, q is independent of lambda), or by
            very poor initial estimates for the eigenvalue and search
            step.

        (`errno` 6)
            The last iteration was terminated because more than
            *<value>* calls to evaluate p, q and its derivative were
            performed.

        (`errno` 7)
            Too small a step size was required at the start of a
            sub-interval to obtain the desired accuracy.

        (`errno` 8)
            Too small a step size was required during solution in a
            sub-interval to obtain the desired accuracy.

        (`errno` 10)
            An internal error has occurred corresponding to a pole of
            the matching function. Try solving the problem again with a
            smaller tolerance.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 9)
            The tolerance set is too small for the problem being solved
            and the machine precision being used. The local eigenvalue
            estimate returned is likely to be a very good approximation.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import sl2_breaks_funs as b_sl2_breaks_funs

    def cb_coeffn(x, elam, jint, data=None):
        """Wrapping user's coeffn. Called as base coeffn."""
        if data is not None:
            _p_coeffn, _q_coeffn, _dqdl_coeffn = coeffn(
                x,
                elam,
                jint,
                data,
            )
        else:
            _p_coeffn, _q_coeffn, _dqdl_coeffn = coeffn(
                x,
                elam,
                jint,
            )
        _p_coeffn_base = _p_coeffn
        _q_coeffn_base = _q_coeffn
        _dqdl_coeffn_base = _dqdl_coeffn
        return (
            _p_coeffn_base,
            _q_coeffn_base,
            _dqdl_coeffn_base,
        )

    fun_name = 'naginterfaces.library.ode.sl2_breaks_funs'

    cb_bdyval_name = fun_name + ':bdyval'

    def cb_bdyval(xl, xr, elam, yl, yr, data=None):
        """Wrapping user's bdyval. Called as base bdyval."""
        if data is not None:
            _yl_bdyval, _yr_bdyval = bdyval(
                xl,
                xr,
                elam,
                data,
            )
        else:
            _yl_bdyval, _yr_bdyval = bdyval(
                xl,
                xr,
                elam,
            )
        yl[:] = _utils._to_ndarray(
            {'fun_name': cb_bdyval_name, 'returned': 'yl'},
            _yl_bdyval,
            copy=False,
            exp_shape=(3,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        yr[:] = _utils._to_ndarray(
            {'fun_name': cb_bdyval_name, 'returned': 'yr'},
            _yr_bdyval,
            copy=False,
            exp_shape=(3,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_monit(nit, iflag, elam, finfo, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                nit,
                iflag,
                elam,
                finfo,
                data,
            )
        else:
            monit(
                nit,
                iflag,
                elam,
                finfo,
            )

    def cb_report(x, v, jint, data=None):
        """Wrapping user's report. Called as base report."""
        if data is not None:
            report(
                x,
                v,
                jint,
                data,
            )
        else:
            report(
                x,
                v,
                jint,
            )

    _xpoint = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xpoint'},
        xpoint,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_xpoint, {'fun_name': fun_name, 'entity': 'xpoint'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _b_utils._check_callable(coeffn, {'fun_name': fun_name, 'entity': 'coeffn'})
    _coeffn = cb_coeffn
    _b_utils._check_callable(bdyval, {'fun_name': fun_name, 'entity': 'bdyval'})
    _bdyval = cb_bdyval
    _hmax = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'hmax'},
        hmax,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _b_utils._check_callable(report, {'fun_name': fun_name, 'entity': 'report'})
    _report = cb_report
    try:
        _return_scalars = b_sl2_breaks_funs(
            _xpoint,
            _m,
            match,
            _coeffn,
            _bdyval,
            k,
            tol,
            elam,
            delam,
            _hmax,
            maxit,
            maxfun,
            _monit,
            _report,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _match_py = _return_scalars.match
    _elam_py = _return_scalars.elam
    _delam_py = _return_scalars.delam
    _hmax_py = _hmax
    _maxit_py = _return_scalars.maxit
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['match', 'elam', 'delam', 'hmax', 'maxit'],
    )(
        _match_py,
        _elam_py,
        _delam_py,
        _hmax_py,
        _maxit_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_2nd_rkn(fcn, t, tend, y, yp, ydp, comm, data=None):
    u"""
    Second-order ordinary differential equations, initial value problem,
    Runge--Kutta--Nystrom method.

    ``ivp_2nd_rkn`` is a function for integrating a non-stiff system of
    second-order ordinary differential equations using
    Runge--Kutta--Nystrom techniques.

    For full information please refer to the NAG Library document for
    d02la

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02laf.html

    Parameters
    ----------
    fcn : callable f = fcn(t, y, data=None)
        `fcn` must evaluate the functions f_i (that is the second
        derivatives y_i^'') for given values of its arguments x,
        y_1,y_2,...,y_neq.

        Parameters
        ~~~~~~~~~~
        t : float
            x, the value of the argument.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the value of the argument.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (neq)
            The value of f_i, for i = 1,2,...,neq.

    t : float
        The initial value of the independent variable x.

    tend : float
        The end point of the range of integration. If `tend` < `t` on
        initial entry, integration will proceed in the negative
        direction. `tend` may be reset, in the direction of integration,
        before any continuation call.

    y : float, array-like, shape (neq)
        The initial values of the solution y_1,y_2,...,y_neq.

    yp : float, array-like, shape (neq)
        The initial values of the derivatives y'_1,y'_2,...,y'_neq.

    ydp : float, array-like, shape (neq)
        Must be unchanged from a previous call to ``ivp_2nd_rkn``.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_2nd_rkn_setup``.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    t : float
        The value of the independent variable, which is usually `tend`,
        unless an error has occurred or the code is operating in
        one-step mode. If the integration is to be continued, possibly
        with a new value for `tend`, `t` must not be changed.

    y : float, ndarray, shape (neq)
        The computed values of the solution at the exit value of `t`. If
        the integration is to be continued, possibly with a new value
        for `tend`, these values must not be changed.

    yp : float, ndarray, shape (neq)
        The computed values of the derivatives at the exit value of `t`.
        If the integration is to be continued, possibly with a new value
        for `tend`, these values must not be changed.

    ydp : float, ndarray, shape (neq)
        The computed values of the second derivative at the exit value
        of `t`, unless illegal input is detected, in which case the
        elements of `ydp` may not have been initialized. If the
        integration is to be continued, possibly with a new value for
        `tend`, these values must not be changed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, neq = *<value>*, but in the previous call to the
            setup function neq = *<value>*.

        (`errno` 1)
            The previous call to the setup function resulted in the
            error exit *<value>*.

        (`errno` 1)
            The setup function ``ivp_2nd_rkn_setup`` has not been
            called.

        (`errno` 1)
            `tend` has been reset such that the direction of integration
            is reversed.

        (`errno` 3)
            To satisfy the accuracy requirements the step size,
            *<value>*, at `t` = *<value>*, is too small for the machine
            precision.

        (`errno` 4)
            Two successive errors detected at the current value of `t` =
            *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `t` = *<value>* and `tend` = *<value>*.

            Constraint: `t` != `tend`.

        (`errno` 2)
            The maximum number of steps, *<value>*, has been attempted.

        (`errno` 5)
            Inefficiency detected in integrating exactly to values of
            `tend`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_2nd_rkn as b_ivp_2nd_rkn

    fun_name = 'naginterfaces.library.ode.ivp_2nd_rkn'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(neq, t, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                t,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                t,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _yp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yp'},
        yp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ydp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ydp'},
        ydp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = b_ivp_2nd_rkn(
            _fcn,
            _neq,
            t,
            tend,
            _y,
            _yp,
            _ydp,
            comm,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _t_py = _return_scalar
    _y_py = _y
    _yp_py = _yp
    _ydp_py = _ydp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'y', 'yp', 'ydp'],
    )(
        _t_py,
        _y_py,
        _yp_py,
        _ydp_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_2nd_rkn_setup(h, tol, thres, thresp, maxstp, start, onestp, high, comm):
    u"""
    Second-order ordinary differential equations, initial value problem,
    setup for ``ivp_2nd_rkn``.

    ``ivp_2nd_rkn_setup`` is a setup function which must be called prior
    to the first call of the integrator ``ivp_2nd_rkn`` and may be
    called prior to any continuation call to ``ivp_2nd_rkn``.

    For full information please refer to the NAG Library document for
    d02lx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02lxf.html

    Parameters
    ----------
    h : float
        If `start` = True, `h` may specify an initial step size to be
        attempted in ``ivp_2nd_rkn``.

        If `start` = False, `h` may specify a step size to override the
        choice of next step attempted made internally to
        ``ivp_2nd_rkn``.

        The sign of `h` is not important, as the absolute value of `h`
        is chosen and the appropriate sign is selected by
        ``ivp_2nd_rkn``.

        If this option is not required then you must set `h` = 0.0.

    tol : float
        Must be set to a relative tolerance for controlling the error in
        the integration by ``ivp_2nd_rkn``. ``ivp_2nd_rkn`` has been
        designed so that, for most problems, a reduction in `tol` leads
        to an approximately proportional reduction in the error in the
        solution. However the actual relation between `tol` and the
        accuracy of the solution cannot be guaranteed. You are strongly
        recommended to repeat the integration with a smaller value of
        `tol` and compare the results. See the description of `thres`
        and `thresp` for further details of how `tol` is used.

    thres : float, array-like, shape (neq)
        `thres` and `thresp` may be set to thresholds for use in the
        error control of ``ivp_2nd_rkn``. At each step in the numerical
        integration estimates of the local errors E1(i) and E2(i) in the
        solution, y_i, and its derivative, y'_i, respectively are
        computed, for i = 1,2,...,neq. For the step to be accepted
        conditions of the following type must be satisfied:

            max_{1 <= i <= neq} (E1(i)/max(`thres`[i-1], |y_i|)) <=
            `tol`,;;;max_{1 <= i <= neq} (E2(i)/max(`thresp`[i-1],
            |y'_i|)) <= `tol`.

        If one or both of these is not satisfied then the step size is
        reduced and the solution is recomputed.

        If `thres`[0] <= 0.0 on entry, a value of 50.0*epsilon is used
        for `thres`[i-1], for i = 1,2,...,neq, where epsilon is machine
        precision.

        Similarly for `thresp`.

    thresp : float, array-like, shape (neq)
        `thresp` and `thresp` may be set to thresholds for use in the
        error control of ``ivp_2nd_rkn``. At each step in the numerical
        integration estimates of the local errors E1(i) and E2(i) in the
        solution, y_i, and its derivative, y'_i, respectively are
        computed, for i = 1,2,...,neq. For the step to be accepted
        conditions of the following type must be satisfied:

            max_{1 <= i <= neq} (E1(i)/max(`thres`[i-1], |y_i|)) <=
            `tol`,;;;max_{1 <= i <= neq} (E2(i)/max(`thresp`[i-1],
            |y'_i|)) <= `tol`.

        If one or both of these is not satisfied then the step size is
        reduced and the solution is recomputed.

        If `thres`[0] <= 0.0 on entry, a value of 50.0*epsilon is used
        for `thres`[i-1], for i = 1,2,...,neq, where epsilon is machine
        precision.

        Similarly for `thresp`.

    maxstp : int
        A bound on the number of steps attempted in any one call of
        ``ivp_2nd_rkn``.

        If `maxstp` <= 0 on entry, a value of 1000 is used.

    start : bool
        Specifies whether or not the call of ``ivp_2nd_rkn`` is for a
        new problem. `start` = True indicates that a new problem is to
        be solved. `start` = False indicates the call of
        ``ivp_2nd_rkn_setup`` is prior to a continuation call of
        ``ivp_2nd_rkn``.

    onestp : bool
        The mode of operation for ``ivp_2nd_rkn``.

        `onestp` = True

          ``ivp_2nd_rkn`` will operate in one-step mode, that is it will
          return after each successful step.

        `onestp` = False

          ``ivp_2nd_rkn`` will operate in interval mode, that is it will
          return at the end of the integration interval.

    high : bool
        If `high` = True, a high-order method will be used, whereas if
        `high` = False, a low-order method will be used. (See the
        specification of ``ivp_2nd_rkn`` for further details.)

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    start : bool
        `start` = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `thresp`[*<value>*] = *<value>* is not positive.

        (`errno` 1)
            On entry, `thres`[*<value>*] = *<value>* is not positive.

        (`errno` 2)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 2)
            On entry, `high` = False and lrwork = *<value>*.

            Constraint: if `high` = False then lrwork >= *<value>*.

        (`errno` 2)
            On entry, `high` = True and lrwork = *<value>*.

            Constraint: if `high` = True then lrwork >= *<value>*.

        (`errno` 3)
            On entry, `tol` = *<value>*.

            Constraint: *<value>* < `tol` < *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_2nd_rkn_setup as b_ivp_2nd_rkn_setup

    fun_name = 'naginterfaces.library.ode.ivp_2nd_rkn_setup'

    _thres = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'thres'},
        thres,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_thres, {'fun_name': fun_name, 'entity': 'thres'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _thresp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'thresp'},
        thresp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_ivp_2nd_rkn_setup(
        _neq,
        h,
        tol,
        _thres,
        _thresp,
        maxstp,
        start,
        onestp,
        high,
        comm,
    )
    _start_py = _return_scalar
    return _start_py

def ivp_2nd_rkn_diag(neq, comm):
    u"""
    Second-order ordinary differential equations, initial value problem,
    diagnostics for ``ivp_2nd_rkn``.

    ``ivp_2nd_rkn_diag`` is a diagnostic function which may be called
    after a call of the integrator ``ivp_2nd_rkn``.

    For full information please refer to the NAG Library document for
    d02ly

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02lyf.html

    Parameters
    ----------
    neq : int
        The number of second-order ordinary differential equations
        solved by ``ivp_2nd_rkn``. It must be the same as the argument
        `neq` supplied to ``ivp_2nd_rkn_setup`` and ``ivp_2nd_rkn``.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_2nd_rkn_setup``.

    Returns
    -------
    hnext : float
        The next step size which ``ivp_2nd_rkn``, if called, would
        attempt.

    hused : float
        The last successful step size used by ``ivp_2nd_rkn``.

    hstart : float
        The initial step size used on the current integration problem by
        ``ivp_2nd_rkn``.

    nsucc : int
        The number of steps attempted by ``ivp_2nd_rkn`` that have been
        successful since the start of the current problem.

    nfail : int
        The number of steps attempted by ``ivp_2nd_rkn`` that have
        failed since the start of the current problem.

    natt : int
        The number of steps attempted before the initial step was
        successful. Over a large number of problems the cost of an
        attempted step of this type is approximately half that of a
        normal attempted step.

    thres : float, ndarray, shape (`neq`)
        The ith solution threshold value used in the error control
        strategy. (See ``ivp_2nd_rkn_setup``.)

    thresp : float, ndarray, shape (`neq`)
        The ith derivative threshold value used in the error control
        strategy. (See ``ivp_2nd_rkn_setup``.)

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `neq` = *<value>*, but in the previous call to the
            setup function neq = *<value>*.

        (`errno` 1)
            The integrator function ``ivp_2nd_rkn`` has not been called.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_2nd_rkn_diag as b_ivp_2nd_rkn_diag

    fun_name = 'naginterfaces.library.ode.ivp_2nd_rkn_diag'

    _neq = _b_utils._EngineIntScalarType(
        neq, {'fun_name': fun_name, 'entity': 'neq'},
    ).to_ctypes()
    _neq_py = _b_utils._EngineIntScalarType.to_py(_neq)
    _thres = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neq_py,),
    )
    _thresp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neq_py,),
    )
    _return_scalars = b_ivp_2nd_rkn_diag(
        _neq,
        _thres,
        _thresp,
        comm,
    )
    _hnext_py = _return_scalars.hnext
    _hused_py = _return_scalars.hused
    _hstart_py = _return_scalars.hstart
    _nsucc_py = _return_scalars.nsucc
    _nfail_py = _return_scalars.nfail
    _natt_py = _return_scalars.natt
    _thres_py = _thres
    _thresp_py = _thresp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['hnext', 'hused', 'hstart', 'nsucc', 'nfail', 'natt', 'thres', 'thresp'],
    )(
        _hnext_py,
        _hused_py,
        _hstart_py,
        _nsucc_py,
        _nfail_py,
        _natt_py,
        _thres_py,
        _thresp_py,
    )
    return _return_data

def ivp_2nd_rkn_interp(t, y, yp, nwant, twant, comm):
    u"""
    Second-order ordinary differential equations, initial value problem,
    interpolation for ``ivp_2nd_rkn``.

    ``ivp_2nd_rkn_interp`` interpolates components of the solution of a
    non-stiff system of second-order differential equations from
    information provided by the integrator ``ivp_2nd_rkn``, when the
    low-order method has been used.

    For full information please refer to the NAG Library document for
    d02lz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02lzf.html

    Parameters
    ----------
    t : float
        t, the current value at which the solution and its derivative
        have been computed (as returned in argument `t` on output from
        ``ivp_2nd_rkn``).

    y : float, array-like, shape (neq)
        The ith component of the solution at t, for i = 1,2,...,neq, as
        returned from ``ivp_2nd_rkn``.

    yp : float, array-like, shape (neq)
        The ith component of the derivative at t, for i = 1,2,...,neq,
        as returned from ``ivp_2nd_rkn``.

    nwant : int
        The number of components of the solution and derivative whose
        values at `twant` are required. The first `nwant` components are
        evaluated.

    twant : float
        The point at which components of the solution and derivative are
        to be evaluated. `twant` should not normally be an extrapolation
        point, that is `twant` should satisfy

            told <= `twant` <= `t`,

        or if integration is proceeding in the negative direction

            told >= `twant` >= `t`,

        where told is the previous integration point which is held in an
        element of the array `comm`['rwork'] and is, to within rounding,
        `t`-hused. (hused is given by ``ivp_2nd_rkn_diag``.)
        Extrapolation is permitted but not recommended, and `errno` = 2
        is returned whenever extrapolation is attempted.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_2nd_rkn_setup``.

    Returns
    -------
    ywant : float, ndarray, shape (`nwant`)
        The calculated value of the ith component of the solution at t =
        `twant`, for i = 1,2,...,`nwant`.

    ypwant : float, ndarray, shape (`nwant`)
        The calculated value of the ith component of the derivative at t
        = `twant`, for i = 1,2,...,`nwant`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, neq = *<value>*, but in a previous call to the
            setup function neq = *<value>*.

        (`errno` 1)
            The integrator function ``ivp_2nd_rkn`` has not been called.

        (`errno` 1)
            No successful steps, interpolation impossible at `twant` =
            *<value>*.

        (`errno` 1)
            On entry, `nwant` = *<value>*.

            Constraint: `nwant` >= 1.

        (`errno` 1)
            On entry, `nwant` = *<value>* and neq = *<value>*.

            Constraint: `nwant` <= neq.

        (`errno` 3)
            Interpolation is not permitted on data from the high order
            formulas.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Extrapolation at `twant` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_2nd_rkn_interp as b_ivp_2nd_rkn_interp

    fun_name = 'naginterfaces.library.ode.ivp_2nd_rkn_interp'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _yp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yp'},
        yp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nwant = _b_utils._EngineIntScalarType(
        nwant, {'fun_name': fun_name, 'entity': 'nwant'},
    ).to_ctypes()
    _nwant_py = _b_utils._EngineIntScalarType.to_py(_nwant)
    _ywant = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nwant_py,),
    )
    _ypwant = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nwant_py,),
    )
    try:
        b_ivp_2nd_rkn_interp(
            _neq,
            t,
            _y,
            _yp,
            _nwant,
            twant,
            _ywant,
            _ypwant,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ywant_py = _ywant
    _ypwant_py = _ypwant
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ywant', 'ypwant'],
    )(
        _ywant_py,
        _ypwant_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dae_dassl_cont(comm):
    u"""
    Implicit ordinary differential equations/DAEs, initial value
    problem, DASSL method continuation for ``dae_dassl_gen``.

    ``dae_dassl_cont`` is a setup function which must be called prior to
    a continuation call to ``dae_dassl_gen``.

    For full information please refer to the NAG Library document for
    d02mc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02mcf.html

    Parameters
    ----------
    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``dae_dassl_gen``.
    """

    from ..base.ode import dae_dassl_cont as b_dae_dassl_cont

    b_dae_dassl_cont(
        comm,
    )

def ivp_stiff_dassl(neqmax, sdysav, maxord, con, tcrit, hmin, hmax, h0, maxstp, mxhnil, norm):
    u"""
    Ordinary differential equations, initial value problem, DASSL
    method, setup for d02m--n functions.

    ``ivp_stiff_dassl`` is an integration method specific setup function
    which must be called prior to linear algebra setup functions and
    integrators from the SPRINT suite of functions, if the DASSL
    implementation of Backward Differentiation Formulae (BDF) is to be
    used.
    Note that this method is also available, independent from the SPRINT
    suite, using ``dae_dassl_gen``

    For full information please refer to the NAG Library document for
    d02mv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02mvf.html

    Parameters
    ----------
    neqmax : int
        A bound on the maximum number of differential equations to be
        solved.

    sdysav : int
        The second dimension of the array ysav that will be supplied to
        the integrator, as declared in the (sub)program from which the
        integrator is called (e.g., see ``ivp_stiff_exp_fulljac``).

    maxord : int
        The maximum order to be used for the BDF method. If `maxord` = 0
        or `maxord` > 5, `maxord` = 5 is assumed.

    con : float, array-like, shape (3)
        Values to be used to control step size choice during
        integration. If any `con`[i] = 0.0 on entry, it is replaced by
        its default value described below. In most cases this is the
        recommended setting.

        `con`[0], `con`[1], and `con`[2] are factors used to bound step
        size changes.

        If the current step size h fails, the modulus of the next step
        size is bounded by `con`[0]*|h|.

        The default value of `con`[0] is 2.0.

        Note that the new step size may be used with a method of
        different order to the failed step.

        If the initial step size is h, the modulus of the step size on
        the second step is bounded by `con`[2]*|h|.

        At any other stage in the integration, if the current step size
        is h, the modulus of the next step size is bounded by
        `con`[1]*|h|.

        The default values are 10.0 for `con`[1] and 1000.0 for
        `con`[2].

    tcrit : float
        A point beyond which integration must not be attempted. The use
        of `tcrit` is described under the argument itask in the
        specification for the integrator (e.g., see
        ``ivp_stiff_exp_fulljac``). A value, 0.0 say, must be specified
        even if itask subsequently specifies that `tcrit` will not be
        used.

    hmin : float
        The minimum absolute step size to be allowed. Set `hmin` = 0.0
        if this option is not required.

    hmax : float
        The maximum absolute step size to be allowed. Set `hmax` = 0.0
        if this option is not required.

    h0 : float
        The step size to be attempted on the first step. Set `h0` = 0.0
        if the initial step size is calculated internally.

    maxstp : int
        The maximum number of steps to be attempted during one call to
        the integrator after which it will return with `errno` = 2
        (e.g., see ``ivp_stiff_exp_fulljac``). Set `maxstp` = 0 if no
        limit is to be imposed.

    mxhnil : int
        The maximum number of warnings printed (if itrace >= 0, e.g.,
        see ``ivp_stiff_exp_fulljac``) per problem when t+h = t on a
        step (h = current step size). If `mxhnil` <= 0, a default value
        of 10 is assumed.

    norm : str, length 1
        Indicates the type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L2 norm.

        `norm` = 'D'

          Is the same as 'A'.

        If vnorm denotes the norm of the vector v of length n, for the
        averaged L2 norm

            vnormB = sqrt(1/nsum_i = 1^n((v_i)/(w_i))^2),

        while for the maximum norm

            vnorm = max_{1 <= i <= n}|(v_i)/(w_i)|.

        If you wish to weight the maximum norm or the L2 norm, rtol and
        atol should be scaled appropriately on input to the integrator
        (see under itol in the specification of the integrator for the
        formulation of the weight vector w_i from rtol and atol, e.g.,
        see ``ivp_stiff_exp_fulljac``).

        Only the first character to the actual argument `norm` is passed
        to ``ivp_stiff_dassl``; hence it is permissible for the actual
        argument to be more descriptive, e.g., 'Maximum', 'Average L2'
        or 'Default' in a call to ``ivp_stiff_dassl``.

    Returns
    -------
    con : float, ndarray, shape (3)
        The values actually to be used by the integration function.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The sequence `con` is not strictly increasing: `con`[0] <
            `con`[1] < `con`[2] is required.

        (`errno` 1)
            On entry, `con`[*<value>*] = *<value>*.

            Constraint: `con`[*<value>*] >= 1.0.

        (`errno` 1)
            On entry, `con`[*<value>*] = *<value>*.

            Constraint: `con`[*<value>*] >= 0.0.

        (`errno` 1)
            On entry, `norm` was not valid: `norm` = *<value>*.

            Constraint: `norm` = 'M', 'A' or 'D'.

        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

        (`errno` 1)
            On entry, `sdysav` = *<value>* and `maxord`+3 = *<value>*.

            Constraint: `sdysav` >= `maxord`+3.

        (`errno` 1)
            On entry, `maxord` = *<value>*.

            Constraint: `maxord` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `maxord` > 5. `maxord` = 5 assumed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_dassl as b_ivp_stiff_dassl

    fun_name = 'naginterfaces.library.ode.ivp_stiff_dassl'

    _con = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'con'},
        con,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = b_ivp_stiff_dassl(
            neqmax,
            sdysav,
            maxord,
            _con,
            tcrit,
            hmin,
            hmax,
            h0,
            maxstp,
            mxhnil,
            norm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _con_py = _con
    _comm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['con', 'comm'],
    )(
        _con_py,
        _comm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dae_dassl_setup(neq, maxord, jceval, hmax, h0, itol):
    u"""
    Implicit ordinary differential equations/DAEs, initial value
    problem, setup for ``dae_dassl_gen``.

    ``dae_dassl_setup`` is a setup function which must be called prior
    to the integrator ``dae_dassl_gen``, if the DASSL implementation of
    Backward Differentiation Formulae (BDF) is to be used.

    For full information please refer to the NAG Library document for
    d02mw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02mwf.html

    Parameters
    ----------
    neq : int
        The number of differential-algebraic equations to be solved.

    maxord : int
        The maximum order to be used for the BDF method. Orders up to
        5th order are available; setting `maxord` > 5 means that the
        maximum order used will be 5.

    jceval : str, length 1
        Specifies the technique to be used to compute the Jacobian.

        `jceval` = 'N'

          The Jacobian is to be evaluated numerically by the integrator.

        `jceval` = 'A'

          You must supply a function to evaluate the Jacobian on a call
          to the integrator.

        Only the first character of the actual paramater `jceval` is
        passed to ``dae_dassl_setup``; hence it is permissible for the
        actual argument to be more descriptive, e.g., 'Numerical' or
        'Analytical', on a call to ``dae_dassl_setup``.

    hmax : float
        The maximum absolute step size to be allowed. Set `hmax` = 0.0
        if this option is not required.

    h0 : float
        The step size to be attempted on the first step. Set `h0` = 0.0
        if the initial step size is calculated internally.

    itol : int
        A value to indicate the form of the local error test.

        `itol` = 0

          rtol and atol are single element vectors.

        `itol` = 1

          rtol and atol are vectors. This should be chosen if you want
          to apply different tolerances to each equation in the system.

        See ``dae_dassl_gen``.

        Note: the tolerances must either both be single element vectors
        or both be vectors of length `neq`.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `neq` = *<value>*.

            Constraint: `neq` >= 1.

        (`errno` 2)
            On entry, `maxord` = *<value>*.

            Constraint: `maxord` >= 1.

        (`errno` 3)
            On entry, `jceval` has an illegal value: `jceval` =
            *<value>*.

            Constraint: `jceval` = 'N' or 'A'.

        (`errno` 4)
            On entry, `hmax` = *<value>*.

            Constraint: `hmax` >= 0.0.

        (`errno` 6)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 0 or 1.
    """

    from ..base.ode import dae_dassl_setup as b_dae_dassl_setup

    _return_scalar = b_dae_dassl_setup(
        neq,
        maxord,
        jceval,
        hmax,
        h0,
        itol,
    )
    _comm_py = _return_scalar
    return _comm_py

def ivp_stiff_interp(tsol, m, comm):
    u"""
    Ordinary differential equations, initial value problem,
    interpolation for d02m--n routines (all integration methods),
    natural interpolant.

    ``ivp_stiff_interp`` interpolates components of the solution of a
    system of first-order differential equations from information
    provided by those integrators in submodule ``ode`` using methods set
    up by calls to ``ivp_stiff_dassl``, ``ivp_stiff_bdf`` or
    ``ivp_stiff_blend``.

    For full information please refer to the NAG Library document for
    d02mz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02mzf.html

    Parameters
    ----------
    tsol : float
        The point at which the first `m` components of the solution are
        to be evaluated. `tsol` should not normally be an extrapolation
        point. Extrapolation is permitted but not recommended.

    m : int
        The number of components of the solution whose values are
        required.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_exp_fulljac``, ``ivp_stiff_exp_bandjac``,
        ``ivp_stiff_exp_sparjac``, ``ivp_stiff_imp_fulljac``,
        ``ivp_stiff_imp_bandjac``, ``ivp_stiff_imp_sparjac``,
        ``ivp_stiff_exp_revcom`` or ``ivp_stiff_imp_revcom``.

    Returns
    -------
    sol : float, ndarray, shape (`m`)
        The calculated value of the solution at `tsol`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, `m` = *<value>* and neq = *<value>*.

            Constraint: `m` <= neq.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            Either the integrator function has not been called prior to
            the first call of this function or the array has become
            corrupted.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, `tsol` is an extrapolation point.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_interp as b_ivp_stiff_interp

    fun_name = 'naginterfaces.library.ode.ivp_stiff_interp'

    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _sol = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('ysav', 'rwork', 'iwsav', 'rwsav'),
    )
    _neq = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (258,), validate=False), {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _neq_py = _b_utils._EngineIntScalarType.to_py(_neq)
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    try:
        b_ivp_stiff_interp(
            tsol,
            _sol,
            _m,
            _ldysav,
            _neq,
            comm,
            _sdysav,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _sol_py = _sol
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['sol'],
    )(
        _sol_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _sol_py

def ivp_stiff_exp_fulljac(t, tout, y, comm, rtol, atol, itol, fcn, itask, itrace, jac=None, monitr=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Explicit ordinary differential equations, stiff initial value
    problem, full Jacobian (comprehensive).

    ``ivp_stiff_exp_fulljac`` is a direct communication function for
    integrating stiff systems of explicit ordinary differential
    equations when the Jacobian is a full matrix.

    For full information please refer to the NAG Library document for
    d02nb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nbf.html

    Parameters
    ----------
    t : float
        t, the value of the independent variable. The input value of `t`
        is used only on the first call as the initial point of the
        integration.

    tout : float
        The next value of t at which a computed solution is desired. For
        the initial t, the input value of `tout` is used to determine
        the direction of integration. Integration is permitted in either
        direction (see also `itask`).

    y : float, array-like, shape (neq)
        The values of the dependent variables (solution). On the first
        call the first neq elements of `y` must contain the vector of
        initial values.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_stiff_fulljac_setup`` and one of ``ivp_stiff_bdf`` or
        ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``ivp_stiff_exp_fulljac`` whether to interpret
        either or both of `rtol` or `atol` as a vector or a scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    fcn : callable (f, ires) = fcn(t, y, ires, data=None)
        `fcn` must evaluate the derivative vector for the explicit
        ordinary differential equation system, defined by y' = g(t,y).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            The value of y_i, for i = 1,2,...,neq.

        ires : int
            `ires` = 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (neq)
            The value y'_i, given by y'_i = g_i(t,y), for i =
            1,2,...,neq.

        ires : int
            You may set `ires` as follows to indicate certain conditions
            in `fcn` to the integrator:

            `ires` = 1

              Indicates a normal return from `fcn`, that is `ires` has
              not been altered by you and integration continues.

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling (sub)program with the
              error indicator set to `errno` = 11.

            `ires` = 3

              Indicates to the integrator that an error condition has
              occurred in the solution vector, its time derivative or in
              the value of t. The integrator will use a smaller time
              step to try to avoid this condition. If this is not
              possible the integrator returns to the calling
              (sub)program with the error indicator set to `errno` = 7.

            `ires` = 4

              Indicates to the integrator to stop its current operation
              and to enter `monitr` immediately with argument `imon` =
              -2.

    itask : int
        The task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit (e.g., see
          ``ivp_stiff_dassl``). tcrit must be specified as an option in
          one of the integrator setup functions before the first call to
          the integrator, or specified in the optional input function
          before a continuation call. tcrit may be equal to or beyond
          `tout`, but not before it, in the direction of integration.

        `itask` = 5

          Take one step only and return, without passing tcrit (e.g.,
          see ``ivp_stiff_dassl``). tcrit must be specified as under
          `itask` = 4.

    itrace : int
        The level of output that is printed by the integrator. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    jac : None or callable p = jac(t, y, h, d, p, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jac` must evaluate the Jacobian of the system.

        If this option is not required, the actual argument for `jac`
        must be None. You must indicate to the integrator whether this
        option is to be used by setting the argument jceval
        appropriately in a call to the full linear algebra setup
        function ``ivp_stiff_fulljac_setup``.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        h : float
            The current step size.

        d : float
            The argument d which depends on the integration method.

        p : float, ndarray, shape (neq, neq)
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (neq, neq)
            `p`[i-1,j-1] must contain dr_i/dy_j, for j = 1,2,...,neq,
            for i = 1,2,...,neq.

            Only the nonzero elements of this array need be set, since
            it is preset to zero before the call to `jac`.

    monitr : None or callable (hnext, y, imon, inln, hmin, hmax) =
        monitr(t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin,
        hmax, nqu, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitr` performs tasks requested by you.

        If this option is not required, the actual argument for `monitr`
        must be None.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable.

        hlast : float
            The last step size successfully used by the integrator.

        hnext : float
            The step size that the integrator proposes to take on the
            next step.

        y : float, ndarray, shape (neq)
            y, the values of the dependent variables evaluated at t.

        ydot : float, ndarray, shape (neq)
            The time derivatives y' of the vector y.

        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument may be passed as argument comm in a call to
            ``ivp_stiff_nat_interp`` or comm in a call to
            ``ivp_stiff_c1_interp`` to enable you to carry out
            interpolation using either of those functions.

        r : float, ndarray, shape (neq)
            If `imon` = 0 and `inln` = 3, the first neq elements contain
            the residual vector, y'-g(t,y).

        acor : float, ndarray, shape (neq, 2)
            With `imon` = 1, `acor`[i-1,0] contains the weight used for
            the ith equation when the norm is evaluated, and
            `acor`[i-1,1] contains the estimated local error for the ith
            equation. The scaled local error at the end of a timestep
            may be obtained by calling ``ivp_stiff_errest``.

        imon : int
            A flag indicating under what circumstances `monitr` was
            called:

            `imon` = -2

              Entry from the integrator after `ires` = 4 (set in `fcn`)
              caused an early termination (this facility could be used
              to locate discontinuities).

            `imon` = -1

              The current step failed repeatedly.

            `imon` = 0

              Entry after a call to the internal nonlinear equation
              solver (see `inln`).

            `imon` = 1

              The current step was successful.

        hmin : float
            The minimum step size to be taken on the next step.

        hmax : float
            The maximum step size to be taken on the next step.

        nqu : int
            The order of the integrator used on the last step. This is
            supplied to enable you to carry out interpolation using
            either of the functions ``ivp_stiff_nat_interp`` or
            ``ivp_stiff_c1_interp``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hnext : float
            The next step size to be used. If this is different from the
            input value, `imon` must be set to 4.

        y : float, array-like, shape (neq)
            These values must not be changed unless `imon` is set to 2.

        imon : int
            May be reset to determine subsequent action in
            ``ivp_stiff_exp_fulljac``.

            `imon` = -2

              Integration is to be halted. A return will be made from
              the integrator to the calling (sub)program with `errno` =
              12.

            `imon` = -1

              Allow the integrator to continue with its own internal
              strategy. The integrator will try up to three restarts
              unless `imon` is set != -1 on exit.

            `imon` = 0

              Return to the internal nonlinear equation solver, where
              the action taken is determined by the value of `inln` (see
              `inln`).

            `imon` = 1

              Normal exit to the integrator to continue integration.

            `imon` = 2

              Restart the integration at the current time point. The
              integrator will restart from order 1 when this option is
              used. The solution `y`, provided by `monitr`, will be used
              for the initial conditions.

            `imon` = 3

              Try to continue with the same step size and order as was
              to be used before the call to `monitr`. `hmin` and `hmax`
              may be altered if desired.

            `imon` = 4

              Continue the integration but using a new value of `hnext`
              and possibly new values of `hmin` and `hmax`.

        inln : int
            The action to be taken by the internal nonlinear equation
            solver when `monitr` is exited with `imon` = 0. By setting
            `inln` = 3 and returning to the integrator, the residual
            vector is evaluated and placed in the array `r`, and then
            `monitr` is called again. At present this is the only option
            available: `inln` must not be set to any other value.

        hmin : float
            The minimum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4.

        hmax : float
            The maximum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4. If `hmax` is
            set to zero, no limit is assumed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `jac` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    t : float
        The value at which the computed solution y is returned (usually
        at `tout`).

    y : float, ndarray, shape (neq)
        The computed solution vector, evaluated at `t` (usually `t` =
        `tout`).

    ydot : float, ndarray, shape (neq)
        The time derivatives y' of the vector y at the last integration
        point.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, *<value>*, is too small.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. `hmin` =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            Either the integrator setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. `hmax` =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            `monitr` set `inln` = *<value>*.

            Constraint: `inln` = 3.

        (`errno` 1)
            `monitr` appears to have overwritten the solution vector.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 7)
            The derivative evaluation procedure set the error flag
            `ires` = 3 repeatedly despite attempts by the integrator to
            avoid this.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 10)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

        (`errno` 15)
            The linear algebra setup routine for full Jacobian was not
            called first.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            There were repeated convergence-test failures on an
            attempted step, before completing the requested task, but
            the integration was successful as far as `t`. This may be
            caused by an inaccurate Jacobian matrix or one which is
            incorrectly computed.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 6)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 11)
            `fcn` set `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 12)
            `monitr` set `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    See Also
    --------
    :meth:`naginterfaces.library.examples.ode.ivp_stiff_exp_fulljac_ex.main`
    """

    from ..base.ode import ivp_stiff_exp_fulljac as b_ivp_stiff_exp_fulljac

    fun_name = 'naginterfaces.library.ode.ivp_stiff_exp_fulljac'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(neq, t, y, f, ires, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn, _ires_fcn = fcn(
                t,
                y,
                ires,
                data,
            )
        else:
            _f_fcn, _ires_fcn = fcn(
                t,
                y,
                ires,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_fcn_base = _ires_fcn
        return _ires_fcn_base

    cb_jac_name = fun_name + ':jac'

    def cb_jac(neq, t, y, h, d, p, data=None):
        """Wrapping user's jac. Called as base jac."""
        if data is not None:
            _p_jac = jac(
                t,
                y,
                h,
                d,
                p,
                data,
            )
        else:
            _p_jac = jac(
                t,
                y,
                h,
                d,
                p,
            )
        _p_jac_np = _utils._to_ndarray(
            {'fun_name': cb_jac_name, 'returned': 'p'},
            _p_jac,
            copy=False,
            exp_shape=(neq, neq),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_jac_name},
            ((_p_jac_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_jac_np

    cb_monitr_name = fun_name + ':monitr'

    def cb_monitr(neq, ldysav, sdysav, t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin, hmax, nqu, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
                data,
            )
        else:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
            )
        _hnext_monitr_base = _hnext_monitr
        y[:] = _utils._to_ndarray(
            {'fun_name': cb_monitr_name, 'returned': 'y'},
            _y_monitr,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _imon_monitr_base = _imon_monitr
        _hmin_monitr_base = _hmin_monitr
        _hmax_monitr_base = _hmax_monitr
        _inln_monitr_base = _inln_monitr
        return (
            _hnext_monitr_base,
            _imon_monitr_base,
            _inln_monitr_base,
            _hmin_monitr_base,
            _hmax_monitr_base,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _ydot = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neq_py,),
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'wkjac', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _b_utils._check_callable(jac, {'fun_name': fun_name, 'entity': 'jac'}, allow_none=True)
    _jac = None if jac is None else cb_jac
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'}, allow_none=True)
    _monitr = None if monitr is None else cb_monitr
    try:
        _return_scalar = b_ivp_stiff_exp_fulljac(
            _neq,
            _ldysav,
            t,
            tout,
            _y,
            _ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _fcn,
            _sdysav,
            _jac,
            _monitr,
            itask,
            itrace,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _t_py = _return_scalar
    _y_py = _y
    _ydot_py = _ydot
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'y', 'ydot'],
    )(
        _t_py,
        _y_py,
        _ydot_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_stiff_exp_bandjac(t, tout, y, comm, rtol, atol, itol, fcn, itask, itrace, jac=None, monitr=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Explicit ordinary differential equations, stiff initial value
    problem, banded Jacobian (comprehensive).

    ``ivp_stiff_exp_bandjac`` is a direct communication function for
    integrating stiff systems of explicit ordinary differential
    equations when the Jacobian is a banded matrix.

    For full information please refer to the NAG Library document for
    d02nc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ncf.html

    Parameters
    ----------
    t : float
        t, the value of the independent variable. The input value of `t`
        is used only on the first call as the initial point of the
        integration.

    tout : float
        The next value of t at which a computed solution is desired. For
        the initial t, the input value of `tout` is used to determine
        the direction of integration. Integration is permitted in either
        direction (see also `itask`).

    y : float, array-like, shape (neq)
        The values of the dependent variables (solution). On the first
        call the first neq elements of `y` must contain the vector of
        initial values.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_stiff_bandjac_setup`` and one of ``ivp_stiff_bdf`` or
        ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``ivp_stiff_exp_bandjac`` whether to interpret
        either or both of `rtol` or `atol` as a vector or a scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    fcn : callable (f, ires) = fcn(t, y, ires, data=None)
        `fcn` must evaluate the derivative vector for the explicit
        ordinary differential equation system, defined by y' = g(t,y).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            The value of y_i, for i = 1,2,...,neq.

        ires : int
            `ires` = 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (neq)
            The value y'_i, given by y'_i = g_i(t,y), for i =
            1,2,...,neq.

        ires : int
            You may set `ires` as follows to indicate certain conditions
            in `fcn` to the integrator:

            `ires` = 1

              Indicates a normal return from `fcn`, that is `ires` has
              not been altered by you and integration continues.

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling (sub)program with the
              error indicator set to `errno` = 11.

            `ires` = 3

              Indicates to the integrator that an error condition has
              occurred in the solution vector, its time derivative or in
              the value of t. The integrator will use a smaller time
              step to try to avoid this condition. If this is not
              possible the integrator returns to the calling
              (sub)program with the error indicator set to `errno` = 7.

            `ires` = 4

              Indicates to the integrator to stop its current operation
              and to enter `monitr` immediately with argument `imon` =
              -2.

    itask : int
        The task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit. tcrit must be specified as an
          option in one of the integrator setup functions before the
          first call to the integrator, or specified in the optional
          input function before a continuation call. tcrit may be equal
          to or beyond `tout`, but not before it, in the direction of
          integration.

        `itask` = 5

          Take one step only and return, without passing tcrit. tcrit
          must be specified as under `itask` = 4.

    itrace : int
        The level of output that is printed by the integrator. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    jac : None or callable p = jac(t, y, h, d, ml, mu, p, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jac` must evaluate the Jacobian of the system.

        If this option is not required, the actual argument for `jac`
        must be None. You must indicate to the integrator whether this
        option is to be used by setting the argument jceval
        appropriately in a call to the banded linear algebra setup
        function ``ivp_stiff_bandjac_setup``.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        h : float
            The current step size.

        d : float
            The argument d which depends on the integration method.

        ml : int
            The number of subdiagonals and superdiagonals respectively
            in the band.

        mu : int
            The number of subdiagonals and superdiagonals respectively
            in the band.

        p : float, ndarray, shape (`ml`+`mu`+1, neq)
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (`ml`+`mu`+1, neq)
            Elements of the Jacobian matrix dr/dy stored as specified by
            the following pseudocode:

            ::

              for i in range(1, neq+1):
                j1 = max(i-ml, 1)
                j2 = min(i+mu, neq)
                for j in range(j1, j2+1):
                  k = min(ml+1-i, 0) + j
                  p[k-1, i-1] = J[i-1, j-1]

            See also ``lapacklin.dgbtrf``.

            Only nonzero elements of this array need be set, since it is
            preset to zero before the call to `jac`.

    monitr : None or callable (hnext, y, imon, inln, hmin, hmax) =
        monitr(t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin,
        hmax, nqu, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitr` performs tasks requested by you.

        If this option is not required, the actual argument for `monitr`
        must be None.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable.

        hlast : float
            The last step size successfully used by the integrator.

        hnext : float
            The step size that the integrator proposes to take on the
            next step.

        y : float, ndarray, shape (neq)
            y, the values of the dependent variables evaluated at t.

        ydot : float, ndarray, shape (neq)
            The time derivatives y' of the vector y.

        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument may be passed as argument comm in a call to
            ``ivp_stiff_nat_interp`` or comm in a call to
            ``ivp_stiff_c1_interp`` to enable you to carry out
            interpolation using either of those functions.

        r : float, ndarray, shape (neq)
            If `imon` = 0 and `inln` = 3, the first neq elements contain
            the residual vector, y'-g(t,y).

        acor : float, ndarray, shape (neq, 2)
            With `imon` = 1, `acor`[i-1,0] contains the weight used for
            the ith equation when the norm is evaluated, and
            `acor`[i-1,1] contains the estimated local error for the ith
            equation. The scaled local error at the end of a timestep
            may be obtained by calling ``ivp_stiff_errest``.

        imon : int
            A flag indicating under what circumstances `monitr` was
            called:

            `imon` = -2

              Entry from the integrator after `ires` = 4 (set in `fcn`)
              caused an early termination (this facility could be used
              to locate discontinuities).

            `imon` = -1

              The current step failed repeatedly.

            `imon` = 0

              Entry after a call to the internal nonlinear equation
              solver (see `inln`).

            `imon` = 1

              The current step was successful.

        hmin : float
            The minimum step size to be taken on the next step.

        hmax : float
            The maximum step size to be taken on the next step.

        nqu : int
            The order of the integrator used on the last step. This is
            supplied to enable you to carry out interpolation using
            either of the functions ``ivp_stiff_nat_interp`` or
            ``ivp_stiff_c1_interp``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hnext : float
            The next step size to be used. If this is different from the
            input value, `imon` must be set to 4.

        y : float, array-like, shape (neq)
            These values must not be changed unless `imon` is set to 2.

        imon : int
            May be reset to determine subsequent action in
            ``ivp_stiff_exp_bandjac``.

            `imon` = -2

              Integration is to be halted. A return will be made from
              the integrator to the calling (sub)program with `errno` =
              12.

            `imon` = -1

              Allow the integrator to continue with its own internal
              strategy. The integrator will try up to three restarts
              unless `imon` is set != -1 on exit.

            `imon` = 0

              Return to the internal nonlinear equation solver, where
              the action taken is determined by the value of `inln` (see
              `inln`).

            `imon` = 1

              Normal exit to the integrator to continue integration.

            `imon` = 2

              Restart the integration at the current time point. The
              integrator will restart from order 1 when this option is
              used. The solution `y`, provided by `monitr`, will be used
              for the initial conditions.

            `imon` = 3

              Try to continue with the same step size and order as was
              to be used before the call to `monitr`. `hmin` and `hmax`
              may be altered if desired.

            `imon` = 4

              Continue the integration but using a new value of `hnext`
              and possibly new values of `hmin` and `hmax`.

        inln : int
            The action to be taken by the internal nonlinear equation
            solver when `monitr` is exited with `imon` = 0. By setting
            `inln` = 3 and returning to the integrator, the residual
            vector is evaluated and placed in the array `r`, and then
            `monitr` is called again. At present this is the only option
            available: `inln` must not be set to any other value.

        hmin : float
            The minimum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4.

        hmax : float
            The maximum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4. If `hmax` is
            set to zero, no limit is assumed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `jac` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    t : float
        The value at which the computed solution y is returned (usually
        at `tout`).

    y : float, ndarray, shape (neq)
        The computed solution vector, evaluated at `t` (usually `t` =
        `tout`).

    ydot : float, ndarray, shape (neq)
        The time derivatives y' of the vector y at the last integration
        point.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `monitr` set `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, *<value>*, is too small.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. `hmin` =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            Either the integrator setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. `hmax` =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            `monitr` set `inln` = *<value>*.

            Constraint: `inln` = 3.

        (`errno` 1)
            `monitr` appears to have overwritten the solution vector.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 7)
            The derivative evaluation procedure set the error flag
            `ires` = 3 repeatedly despite attempts by the integrator to
            avoid this.

        (`errno` 10)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

        (`errno` 15)
            The linear algebra setup routine for banded Jacobian was not
            called first.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            There were repeated convergence-test failures on an
            attempted step, before completing the requested task, but
            the integration was successful as far as `t`. This may be
            caused by an inaccurate Jacobian matrix or one which is
            incorrectly computed.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 6)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 11)
            `fcn` set `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_exp_bandjac as b_ivp_stiff_exp_bandjac

    fun_name = 'naginterfaces.library.ode.ivp_stiff_exp_bandjac'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(neq, t, y, f, ires, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn, _ires_fcn = fcn(
                t,
                y,
                ires,
                data,
            )
        else:
            _f_fcn, _ires_fcn = fcn(
                t,
                y,
                ires,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_fcn_base = _ires_fcn
        return _ires_fcn_base

    cb_jac_name = fun_name + ':jac'

    def cb_jac(neq, t, y, h, d, ml, mu, p, data=None):
        """Wrapping user's jac. Called as base jac."""
        if data is not None:
            _p_jac = jac(
                t,
                y,
                h,
                d,
                ml,
                mu,
                p,
                data,
            )
        else:
            _p_jac = jac(
                t,
                y,
                h,
                d,
                ml,
                mu,
                p,
            )
        _p_jac_np = _utils._to_ndarray(
            {'fun_name': cb_jac_name, 'returned': 'p'},
            _p_jac,
            copy=False,
            exp_shape=(ml+mu+1, neq),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_jac_name},
            ((_p_jac_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_jac_np

    cb_monitr_name = fun_name + ':monitr'

    def cb_monitr(neq, ldysav, sdysav, t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin, hmax, nqu, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
                data,
            )
        else:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
            )
        _hnext_monitr_base = _hnext_monitr
        y[:] = _utils._to_ndarray(
            {'fun_name': cb_monitr_name, 'returned': 'y'},
            _y_monitr,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _imon_monitr_base = _imon_monitr
        _hmin_monitr_base = _hmin_monitr
        _hmax_monitr_base = _hmax_monitr
        _inln_monitr_base = _inln_monitr
        return (
            _hnext_monitr_base,
            _imon_monitr_base,
            _inln_monitr_base,
            _hmin_monitr_base,
            _hmax_monitr_base,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _ydot = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neq_py,),
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'wkjac', 'jacpvt', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _b_utils._check_callable(jac, {'fun_name': fun_name, 'entity': 'jac'}, allow_none=True)
    _jac = None if jac is None else cb_jac
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'}, allow_none=True)
    _monitr = None if monitr is None else cb_monitr
    try:
        _return_scalar = b_ivp_stiff_exp_bandjac(
            _neq,
            _ldysav,
            t,
            tout,
            _y,
            _ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _fcn,
            _sdysav,
            _jac,
            _monitr,
            itask,
            itrace,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _t_py = _return_scalar
    _y_py = _y
    _ydot_py = _ydot
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'y', 'ydot'],
    )(
        _t_py,
        _y_py,
        _ydot_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_stiff_exp_sparjac(t, tout, y, comm, rtol, atol, itol, fcn, itask, itrace, jac=None, monitr=None, data=None, io_manager=None):
    u"""
    Explicit ordinary differential equations, stiff initial value
    problem, sparse Jacobian (comprehensive).

    ``ivp_stiff_exp_sparjac`` is a direct communication function for
    integrating stiff systems of explicit ordinary differential
    equations when the Jacobian is a sparse matrix.

    For full information please refer to the NAG Library document for
    d02nd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ndf.html

    Parameters
    ----------
    t : float
        t, the value of the independent variable. The input value of `t`
        is used only on the first call as the initial point of the
        integration.

    tout : float
        The next value of t at which a computed solution is desired. For
        the initial t, the input value of `tout` is used to determine
        the direction of integration. Integration is permitted in either
        direction (see also `itask`).

    y : float, array-like, shape (neq)
        The values of the dependent variables (solution). On the first
        call the first neq elements of `y` must contain the vector of
        initial values.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_stiff_sparjac_setup`` and one of ``ivp_stiff_bdf`` or
        ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``ivp_stiff_exp_sparjac`` whether to interpret
        either or both of `rtol` or `atol` as a vector or a scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    fcn : callable (f, ires) = fcn(t, y, ires, data=None)
        `fcn` must evaluate the derivative vector for the explicit
        ordinary differential equation system, defined by y' = g(t,y).

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            The value of y_i, for i = 1,2,...,neq.

        ires : int
            `ires` = 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (neq)
            The value y'_i, given by y'_i = g_i(t,y), for i =
            1,2,...,neq.

        ires : int
            You may set `ires` as follows to indicate certain conditions
            in `fcn` to the integrator:

            `ires` = 1

              Indicates a normal return from `fcn`, that is `ires` has
              not been altered by you and integration continues.

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling (sub)program with the
              error indicator set to `errno` = 11.

            `ires` = 3

              Indicates to the integrator that an error condition has
              occurred in the solution vector, its time derivative or in
              the value of t. The integrator will use a smaller time
              step to try to avoid this condition. If this is not
              possible the integrator returns to the calling
              (sub)program with the error indicator set to `errno` = 7.

            `ires` = 4

              Indicates to the integrator to stop its current operation
              and to enter `monitr` immediately with argument `imon` =
              -2.

    itask : int
        The task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit (e.g., see
          ``ivp_stiff_dassl``). tcrit must be specified as an option in
          one of the integrator setup functions before the first call to
          the integrator, or specified in the optional input function
          before a continuation call. tcrit may be equal to or beyond
          `tout`, but not before it, in the direction of integration.

        `itask` = 5

          Take one step only and return, without passing tcrit (e.g.,
          see ``ivp_stiff_dassl``). tcrit must be specified as under
          `itask` = 4.

    itrace : int
        The level of output that is printed by the integrator. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    jac : None or callable pdj = jac(t, y, h, d, j, pdj, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jac` must evaluate the Jacobian of the system.

        If this option is not required, the actual argument for `jac`
        must be None. You must indicate to the integrator whether this
        option is to be used by setting the argument jceval
        appropriately in a call to the sparse linear algebra setup
        function ``ivp_stiff_sparjac_setup``.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        h : float
            The current step size.

        d : float
            The argument d which depends on the integration method.

        j : int
            The column of the Jacobian that `jac` must return in the
            array `pdj`.

        pdj : float, ndarray, shape (neq)
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        pdj : float, array-like, shape (neq)
            `pdj`[i-1] should be set to the (i,j)th element of the
            Jacobian, where j is given by `j`. Only nonzero elements of
            this array need be set, since it is preset to zero before
            the call to `jac`.

    monitr : None or callable (hnext, y, imon, inln, hmin, hmax) =
        monitr(t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin,
        hmax, nqu, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitr` performs tasks requested by you.

        If this option is not required, the actual argument for `monitr`
        must be None.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable.

        hlast : float
            The last step size successfully used by the integrator.

        hnext : float
            The step size that the integrator proposes to take on the
            next step.

        y : float, ndarray, shape (neq)
            y, the values of the dependent variables evaluated at t.

        ydot : float, ndarray, shape (neq)
            The time derivatives y' of the vector y.

        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument may be passed as argument comm in a call to
            ``ivp_stiff_nat_interp`` or comm in a call to
            ``ivp_stiff_c1_interp`` to enable you to carry out
            interpolation using either of those functions.

        r : float, ndarray, shape (neq)
            If `imon` = 0 and `inln` = 3, the first neq elements contain
            the residual vector, y'-g(t,y).

        acor : float, ndarray, shape (neq, 2)
            With `imon` = 1, `acor`[i-1,0] contains the weight used for
            the ith equation when the norm is evaluated, and
            `acor`[i-1,1] contains the estimated local error for the ith
            equation. The scaled local error at the end of a timestep
            may be obtained by calling ``ivp_stiff_errest``.

        imon : int
            A flag indicating under what circumstances `monitr` was
            called:

            `imon` = -2

              Entry from the integrator after `ires` = 4 (set in `fcn`)
              caused an early termination (this facility could be used
              to locate discontinuities).

            `imon` = -1

              The current step failed repeatedly.

            `imon` = 0

              Entry after a call to the internal nonlinear equation
              solver (see `inln`).

            `imon` = 1

              The current step was successful.

        hmin : float
            The minimum step size to be taken on the next step.

        hmax : float
            The maximum step size to be taken on the next step.

        nqu : int
            The order of the integrator used on the last step. This is
            supplied to enable you to carry out interpolation using
            either of the functions ``ivp_stiff_nat_interp`` or
            ``ivp_stiff_c1_interp``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hnext : float
            The next step size to be used. If this is different from the
            input value, `imon` must be set to 4.

        y : float, array-like, shape (neq)
            These values must not be changed unless `imon` is set to 2.

        imon : int
            May be reset to determine subsequent action in
            ``ivp_stiff_exp_sparjac``.

            `imon` = -2

              Integration is to be halted. A return will be made from
              the integrator to the calling (sub)program with `errno` =
              12.

            `imon` = -1

              Allow the integrator to continue with its own internal
              strategy. The integrator will try up to three restarts
              unless `imon` is set != -1 on exit.

            `imon` = 0

              Return to the internal nonlinear equation solver, where
              the action taken is determined by the value of `inln` (see
              `inln`).

            `imon` = 1

              Normal exit to the integrator to continue integration.

            `imon` = 2

              Restart the integration at the current time point. The
              integrator will restart from order 1 when this option is
              used. The solution `y`, provided by `monitr`, will be used
              for the initial conditions.

            `imon` = 3

              Try to continue with the same step size and order as was
              to be used before the call to `monitr`. `hmin` and `hmax`
              may be altered if desired.

            `imon` = 4

              Continue the integration but using a new value of `hnext`
              and possibly new values of `hmin` and `hmax`.

        inln : int
            The action to be taken by the internal nonlinear equation
            solver when `monitr` is exited with `imon` = 0. By setting
            `inln` = 3 and returning to the integrator, the residual
            vector is evaluated and placed in the array `r`, and then
            `monitr` is called again. At present this is the only option
            available: `inln` must not be set to any other value.

        hmin : float
            The minimum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4.

        hmax : float
            The maximum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4. If `hmax` is
            set to zero, no limit is assumed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    t : float
        The value at which the computed solution y is returned (usually
        at `tout`).

    y : float, ndarray, shape (neq)
        The computed solution vector, evaluated at `t` (usually `t` =
        `tout`).

    ydot : float, ndarray, shape (neq)
        The time derivatives y' of the vector y at the last integration
        point.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `monitr` set `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, *<value>*, is too small.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. `hmin` =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            Either the integrator setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. `hmax` =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            `monitr` set `inln` = *<value>*.

            Constraint: `inln` = 3.

        (`errno` 1)
            `monitr` appears to have overwritten the solution vector.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 7)
            The derivative evaluation procedure set the error flag
            `ires` = 3 repeatedly despite attempts by the integrator to
            avoid this.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 10)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

        (`errno` 15)
            The linear algebra setup function for sparse Jacobian was
            not called first.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            There were repeated convergence-test failures on an
            attempted step, before completing the requested task, but
            the integration was successful as far as `t`. This may be
            caused by an inaccurate Jacobian matrix or one which is
            incorrectly computed.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 6)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 11)
            `fcn` set `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_exp_sparjac as b_ivp_stiff_exp_sparjac

    fun_name = 'naginterfaces.library.ode.ivp_stiff_exp_sparjac'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(neq, t, y, f, ires, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn, _ires_fcn = fcn(
                t,
                y,
                ires,
                data,
            )
        else:
            _f_fcn, _ires_fcn = fcn(
                t,
                y,
                ires,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_fcn_base = _ires_fcn
        return _ires_fcn_base

    cb_jac_name = fun_name + ':jac'

    def cb_jac(neq, t, y, h, d, j, pdj, data=None):
        """Wrapping user's jac. Called as base jac."""
        if data is not None:
            _pdj_jac = jac(
                t,
                y,
                h,
                d,
                j,
                pdj,
                data,
            )
        else:
            _pdj_jac = jac(
                t,
                y,
                h,
                d,
                j,
                pdj,
            )
        pdj[:] = _utils._to_ndarray(
            {'fun_name': cb_jac_name, 'returned': 'pdj'},
            _pdj_jac,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_monitr_name = fun_name + ':monitr'

    def cb_monitr(neq, ldysav, sdysav, t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin, hmax, nqu, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
                data,
            )
        else:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
            )
        _hnext_monitr_base = _hnext_monitr
        y[:] = _utils._to_ndarray(
            {'fun_name': cb_monitr_name, 'returned': 'y'},
            _y_monitr,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _imon_monitr_base = _imon_monitr
        _hmin_monitr_base = _hmin_monitr
        _hmax_monitr_base = _hmax_monitr
        _inln_monitr_base = _inln_monitr
        return (
            _hnext_monitr_base,
            _imon_monitr_base,
            _inln_monitr_base,
            _hmin_monitr_base,
            _hmax_monitr_base,
        )

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _ydot = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neq_py,),
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'wkjac', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'jacpvt', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _b_utils._check_callable(jac, {'fun_name': fun_name, 'entity': 'jac'}, allow_none=True)
    _jac = None if jac is None else cb_jac
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'}, allow_none=True)
    _monitr = None if monitr is None else cb_monitr
    try:
        _return_scalar = b_ivp_stiff_exp_sparjac(
            _neq,
            _ldysav,
            t,
            tout,
            _y,
            _ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _fcn,
            _sdysav,
            _jac,
            _monitr,
            itask,
            itrace,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _t_py = _return_scalar
    _y_py = _y
    _ydot_py = _ydot
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'y', 'ydot'],
    )(
        _t_py,
        _y_py,
        _ydot_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dae_dassl_gen(t, tout, y, ydot, rtol, atol, itask, res, comm, jac=None, data=None):
    u"""
    Implicit ordinary differential equations/DAEs, initial value
    problem, DASSL method integrator.

    ``dae_dassl_gen`` is a function for integrating stiff systems of
    implicit ordinary differential equations coupled with algebraic
    equations.

    For full information please refer to the NAG Library document for
    d02ne

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nef.html

    Parameters
    ----------
    t : float
        On initial entry: the initial value of the independent variable,
        t.

    tout : float
        The next value of t at which a computed solution is desired.

        On initial entry: `tout` is used to determine the direction of
        integration. Integration is permitted in either direction (see
        also `itask`).

    y : float, array-like, shape (neq)
        On initial entry: the vector of initial values of the dependent
        variables y.

    ydot : float, array-like, shape (neq)
        On initial entry: `ydot` must contain approximations to the time
        derivatives y' of the vector y evaluated at the initial value of
        the independent variable.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `comm`['icom'][1]=1: neq; if `comm`['icom'][1]=0: 1;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `comm`['icom'][1]=1: neq; if `comm`['icom'][1]=0: 1;
        otherwise: 0.

        The absolute local error tolerance.

    itask : int
        On initial entry: need not be set.

    res : callable (r, ires) = res(t, y, ydot, ires, data=None)
        `res` must evaluate the residual

            R = F(t,y,y').

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        ydot : float, ndarray, shape (neq)
            The derivative of the solution at the current point t.

        ires : int
            Is always equal to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (neq)
            `r`[i-1] must contain the ith component of R, for i =
            1,2,...,neq where

                R = F(`t`,`y`,`ydot`).

        ires : int
            `ires` should normally be left unchanged. However, if an
            illegal value of `y` is encountered, `ires` should be set to
            -1; ``dae_dassl_gen`` will then attempt to resolve the
            problem so that illegal values of `y` are not encountered.
            `ires` should be set to -2 if you wish to return control to
            the calling function; this will cause ``dae_dassl_gen`` to
            exit with `errno` = 23.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``dae_dassl_setup``.

    jac : None or callable pd = jac(t, y, ydot, pd, cj, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        Evaluates the matrix of partial derivatives, J, where

            J_{ij} = dF_i/dy_j+`cj`*dF_i/dy'_j, i,j = 1,2,...,neq.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        ydot : float, ndarray, shape (neq)
            The derivative of the solution at the current point t.

        pd : float, ndarray, shape (:)
            `pd` is preset to zero before the call to `jac`.

        cj : float
            `cj` is a scalar constant which will be defined in
            ``dae_dassl_gen``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        pd : float, array-like, shape (:)
            If the Jacobian is full then `pd`[(j-1)*neq+i-1] = J_{ij},
            for j = 1,2,...,neq, for i = 1,2,...,neq; if the Jacobian is
            banded then `pd`[(j-1)*(2ml+mu+1)+ml+mu+i-j] = J_{ij}, for
            max(1, j-mu) <= i <= min(n, j+ml); (see also in
            ``lapacklin.dgbtrf``).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    t : float
        On intermediate exit: t, the current value of the independent
        variable.

        On final exit: the value of the independent variable at which
        the computed solution y is returned (usually at `tout`).

    y : float, ndarray, shape (neq)
        On intermediate exit: the computed solution vector, y, evaluated
        at t = T.

        On final exit: the computed solution vector, evaluated at t
        (usually t = `tout`).

    ydot : float, ndarray, shape (neq)
        The time derivatives y' of the vector y at the last integration
        point.

    rtol : float, ndarray, shape (:)
        `rtol` remains unchanged unless ``dae_dassl_gen`` exits with
        `errno` = 16 in which case the values may have been increased to
        values estimated to be appropriate for continuing the
        integration.

    atol : float, ndarray, shape (:)
        `atol` remains unchanged unless ``dae_dassl_gen`` exits with
        `errno` = 16 in which case the values may have been increased to
        values estimated to be appropriate for continuing the
        integration.

    itask : int
        The task performed by the integrator on successful completion or
        an indicator that a problem occurred during integration.

        `itask` = 2

          The integration to `tout` was successfully completed (`t` =
          `tout`) by stepping exactly to `tout`.

        `itask` = 3

          The integration to `tout` was successfully completed (`t` =
          `tout`) by stepping past `tout`. `y` and `ydot` are obtained
          by interpolation.

        `itask` < 0

          Different negative values of `itask` returned correspond to
          different failure exits. errno should always be checked in
          such cases and the corrective action taken where appropriate.

        `itask` must remain unchanged between calls to
        ``dae_dassl_gen``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 3)
            On entry, `t` = *<value>*.

            Constraint: `tout` != `t`.

        (`errno` 3)
            `tout` is too close to `t` to start integration: `tout`-`t`
            = *<value>*: hmin = *<value>*.

        (`errno` 3)
            `tout` is behind `t` in the direction of h: `tout`-`t` =
            *<value>*, h = *<value>*.

        (`errno` 6)
            Some element of `rtol` is less than zero.

        (`errno` 7)
            Some element of `atol` is less than zero.

        (`errno` 8)
            A previous call to this function returned with `itask` =
            *<value>* and no appropriate action was taken.

        (`errno` 11)
            Either the initialization function has not been called prior
            to the first call of this function or a communication array
            has become corrupted.

        (`errno` 12)
            Either the initialization function has not been called prior
            to the first call of this function or a communication array
            has become corrupted.

        (`errno` 13)
            `comm`['com'] array is of insufficient length; length
            required = *<value>*; actual length lcom = *<value>*.

        (`errno` 14)
            All elements of `rtol` and `atol` are zero.

        (`errno` 15)
            Maximum number of steps taken on this call before reaching
            `tout`: `t` = *<value>*, maximum number of steps =
            *<value>*.

        (`errno` 16)
            Too much accuracy requested for precision of machine. `rtol`
            and `atol` were increased by scale factor R. Try running
            again with these scaled tolerances. `t` = *<value>*, R =
            *<value>*.

        (`errno` 17)
            A solution component has become zero when a purely relative
            tolerance (zero absolute tolerance) was selected for that
            component. `t` = *<value>*, `y`[I-1] = *<value>* for
            component I = *<value>*.

        (`errno` 18)
            The error test failed repeatedly with abs(h) = hmin. `t` =
            *<value>*. Stepsize h = *<value>*.

        (`errno` 19)
            The corrector repeatedly failed to converge with abs(h) =
            hmin. `t` = *<value>*. Stepsize h = *<value>*.

        (`errno` 20)
            The iteration matrix is singular. `t` = *<value>*. Stepsize
            h = *<value>*.

        (`errno` 21)
            The corrector could not converge and the error test failed
            repeatedly. `t` = *<value>*. Stepsize h = *<value>*.

        (`errno` 22)
            `ires` was set to -1 during a call to `res` and could not be
            resolved. `t` = *<value>*. Stepsize h = *<value>*.

        (`errno` 23)
            `ires` was set to -2 during a call to `res`. `t` =
            *<value>*. Stepsize = *<value>*.

        (`errno` 24)
            The initial `ydot` could not be computed. `t` = *<value>*.
            Stepsize h = *<value>*.

        (`errno` 25)
            Repeated occurrences of input constraint violations have
            been detected. This could result in a potential infinite
            loop: `itask` = *<value>*. Current violation corresponds to
            exit with errno = *<value>*.
    """

    from ..base.ode import dae_dassl_gen as b_dae_dassl_gen

    fun_name = 'naginterfaces.library.ode.dae_dassl_gen'

    cb_res_name = fun_name + ':res'

    def cb_res(neq, t, y, ydot, r, ires, data=None):
        """Wrapping user's res. Called as base res."""
        if data is not None:
            _r_res, _ires_res = res(
                t,
                y,
                ydot,
                ires,
                data,
            )
        else:
            _r_res, _ires_res = res(
                t,
                y,
                ydot,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_res_name, 'returned': 'r'},
            _r_res,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_res_base = _ires_res
        return _ires_res_base

    cb_jac_name = fun_name + ':jac'

    def cb_jac(neq, t, y, ydot, pd, lenpd, cj, data=None):
        """Wrapping user's jac. Called as base jac."""
        if data is not None:
            _pd_jac = jac(
                t,
                y,
                ydot,
                pd,
                cj,
                data,
            )
        else:
            _pd_jac = jac(
                t,
                y,
                ydot,
                pd,
                cj,
            )
        pd[:] = _utils._to_ndarray(
            {'fun_name': cb_jac_name, 'returned': 'pd'},
            _pd_jac,
            copy=False,
            exp_shape=(lenpd,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ydot = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ydot'},
        ydot,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(res, {'fun_name': fun_name, 'entity': 'res'})
    _res = cb_res
    _b_utils._check_callable(jac, {'fun_name': fun_name, 'entity': 'jac'}, allow_none=True)
    _jac = None if jac is None else cb_jac
    _return_scalars = b_dae_dassl_gen(
        _neq,
        t,
        tout,
        _y,
        _ydot,
        _rtol,
        _atol,
        itask,
        _res,
        _jac,
        comm,
        data,
    )
    _t_py = _return_scalars.t
    _y_py = _y
    _ydot_py = _ydot
    _rtol_py = _rtol
    _atol_py = _atol
    _itask_py = _return_scalars.itask
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'y', 'ydot', 'rtol', 'atol', 'itask'],
    )(
        _t_py,
        _y_py,
        _ydot_py,
        _rtol_py,
        _atol_py,
        _itask_py,
    )
    return _return_data

def ivp_stiff_imp_fulljac(t, tout, y, ydot, comm, rtol, atol, itol, resid, lderiv, itask, itrace, jac=None, monitr=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Implicit/algebraic ordinary differential equations, stiff initial
    value problem, full Jacobian (comprehensive).

    ``ivp_stiff_imp_fulljac`` is a direct communication function for
    integrating stiff systems of implicit ordinary differential
    equations coupled with algebraic equations when the Jacobian is a
    full matrix.

    For full information please refer to the NAG Library document for
    d02ng

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ngf.html

    Parameters
    ----------
    t : float
        t, the value of the independent variable. The input value of `t`
        is used only on the first call as the initial point of the
        integration.

    tout : float
        The next value of t at which a computed solution is desired. For
        the initial t, the input value of `tout` is used to determine
        the direction of integration. Integration is permitted in either
        direction (see also `itask`).

    y : float, array-like, shape (neq)
        The values of the dependent variables (solution). On the first
        call the first neq elements of `y` must contain the vector of
        initial values.

    ydot : float, array-like, shape (neq)
        If `lderiv`[0] = True, `ydot` must contain approximations to the
        time derivatives y' of the vector y.

        If `lderiv`[0] = False, `ydot` need not be set on entry.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_stiff_fulljac_setup`` and one of ``ivp_stiff_dassl``,
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``ivp_stiff_imp_fulljac`` whether to interpret
        either or both of `rtol` or `atol` as a vector or a scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    resid : callable (r, ires) = resid(t, y, ydot, ires, data=None)
        `resid` must evaluate the residual

            r = g(t,y)-A(t,y)y'

        in one case and

            r = -A(t,y)y'

        in another.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            The value of y_i, for i = 1,2,...,neq.

        ydot : float, ndarray, shape (neq)
            The value of y'_i, for i = 1,2,...,neq, at t.

        ires : int
            The form of the residual that must be returned in array `r`.

            `ires` = -1

              The residual defined in equation (2) must be returned.

            `ires` = 1

              The residual defined in equation (1) must be returned.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (neq)
            `r`[i-1] must contain the ith component of r, for i =
            1,2,...,neq, where

                r = g(t,y)-A(t,y)y'

            or

                r = -A(t,y)y'

            and where the definition of r is determined by the input
            value of `ires`.

        ires : int
            Should be unchanged unless one of the following actions is
            required of the integrator, in which case `ires` should be
            set accordingly.

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling (sub)program with the
              error indicator set to `errno` = 11.

            `ires` = 3

              Indicates to the integrator that an error condition has
              occurred in the solution vector, its time derivative or in
              the value of t. The integrator will use a smaller time
              step to try to avoid this condition. If this is not
              possible, the integrator returns to the calling
              (sub)program with the error indicator set to `errno` = 7.

            `ires` = 4

              Indicates to the integrator to stop its current operation
              and to enter `monitr` immediately with argument `imon` =
              -2.

    lderiv : bool, array-like, shape (2)
        `lderiv`[0] must be set to True if you have supplied both an
        initial y and an initial y'. `lderiv`[0] must be set to False if
        only the initial y has been supplied.

        `lderiv`[1] must be set to True if the integrator is to use a
        modified Newton method to evaluate the initial y and y'.

        Note that y and y', if supplied, are used as initial estimates.

        This method involves taking a small step at the start of the
        integration, and if `itask` = 6 on entry, `t` and `tout` will be
        set to the result of taking this small step. `lderiv`[1] must be
        set to False if the integrator is to use functional iteration to
        evaluate the initial y and y', and if this fails a modified
        Newton method will then be attempted. `lderiv`[1] = True is
        recommended if there are implicit equations or the initial y and
        y' are zero.

    itask : int
        The task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit. tcrit must be specified as an
          option in one of the integrator setup functions before the
          first call to the integrator, or specified in the optional
          input function before a continuation call. tcrit may be equal
          to or beyond `tout`, but not before it, in the direction of
          integration.

        `itask` = 5

          Take one step only and return, without passing tcrit. tcrit
          must be specified as under `itask` = 4.

        `itask` = 6

          The integrator will solve for the initial values of y and y'
          only and then return to the calling (sub)program without doing
          the integration. This option can be used to check the initial
          values of y and y'. Functional iteration or a 'small' backward
          Euler method used in conjunction with a damped Newton
          iteration is used to calculate these values (see `lderiv`).
          Note that if a backward Euler step is used then the value of t
          will have been advanced a short distance from the initial
          point.

        Note: if ``ivp_stiff_imp_fulljac`` is recalled with a different
        value of `itask` (and `tout` altered), the initialization
        procedure is repeated, possibly leading to different initial
        conditions.

    itrace : int
        The level of output that is printed by the integrator. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    jac : None or callable p = jac(t, y, ydot, h, d, p, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jac` must evaluate the Jacobian of the system.

        If this option is not required, the actual argument for `jac`
        must be None. You must indicate to the integrator whether this
        option is to be used by setting the argument jceval
        appropriately in a call to the full linear algebra setup
        function ``ivp_stiff_fulljac_setup``.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        ydot : float, ndarray, shape (neq)
            The derivative of the solution at the current point t.

        h : float
            The current step size.

        d : float
            The argument d which depends on the integration method.

        p : float, ndarray, shape (neq, neq)
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (neq, neq)
            `p`[i-1,j-1] must contain dr_i/dy_j, for j = 1,2,...,neq,
            for i = 1,2,...,neq.

            Only the nonzero elements of this array need be set, since
            it is preset to zero before the call to `jac`.

    monitr : None or callable (hnext, y, imon, inln, hmin, hmax) =
        monitr(t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin,
        hmax, nqu, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitr` performs tasks requested by you.

        If this option is not required, the actual argument for `monitr`
        must be None.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable.

        hlast : float
            The last step size successfully used by the integrator.

        hnext : float
            The step size that the integrator proposes to take on the
            next step.

        y : float, ndarray, shape (neq)
            y, the values of the dependent variables evaluated at t.

        ydot : float, ndarray, shape (neq)
            The time derivatives y' of the vector y.

        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument may be passed as argument comm in a call to
            ``ivp_stiff_nat_interp`` or comm in a call to
            ``ivp_stiff_c1_interp`` to enable you to carry out
            interpolation using either of those functions.

        r : float, ndarray, shape (neq)
            If `imon` = 0 and `inln` = 3, the first neq elements contain
            the residual vector A(t,y)y'-g(t,y).

        acor : float, ndarray, shape (neq, 2)
            With `imon` = 1, `acor`[i-1,0] contains the weight used for
            the ith equation when the norm is evaluated, and
            `acor`[i-1,1] contains the estimated local error for the ith
            equation. The scaled local error at the end of a timestep
            may be obtained by calling ``ivp_stiff_errest``.

        imon : int
            A flag indicating under what circumstances `monitr` was
            called:

            `imon` = -2

              Entry from the integrator after `ires` = 4 (set in
              `resid`) caused an early termination (this facility could
              be used to locate discontinuities).

            `imon` = -1

              The current step failed repeatedly.

            `imon` = 0

              Entry after a call to the internal nonlinear equation
              solver (see `inln`).

            `imon` = 1

              The current step was successful.

        hmin : float
            The minimum step size to be taken on the next step.

        hmax : float
            The maximum step size to be taken on the next step.

        nqu : int
            The order of the integrator used on the last step. This is
            supplied to enable you to carry out interpolation using
            either of the functions ``ivp_stiff_nat_interp`` or
            ``ivp_stiff_c1_interp``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hnext : float
            The next step size to be used. If this is different from the
            input value, `imon` must be set to 4.

        y : float, array-like, shape (neq)
            These values must not be changed unless `imon` is set to 2.

        imon : int
            May be reset to determine subsequent action in
            ``ivp_stiff_imp_fulljac``.

            `imon` = -2

              Integration is to be halted. A return will be made from
              the integrator to the calling (sub)program with `errno` =
              12.

            `imon` = -1

              Allow the integrator to continue with its own internal
              strategy. The integrator will try up to three restarts
              unless `imon` != -1 on exit.

            `imon` = 0

              Return to the internal nonlinear equation solver, where
              the action taken is determined by the value of `inln` (see
              `inln`).

            `imon` = 1

              Normal exit to the integrator to continue integration.

            `imon` = 2

              Restart the integration at the current time point. The
              integrator will restart from order 1 when this option is
              used. The solution `y`, provided by `monitr`, will be used
              for the initial conditions.

            `imon` = 3

              Try to continue with the same step size and order as was
              to be used before the call to `monitr`. `hmin` and `hmax`
              may be altered if desired.

            `imon` = 4

              Continue the integration but using a new value of `hnext`
              and possibly new values of `hmin` and `hmax`.

        inln : int
            The action to be taken by the internal nonlinear equation
            solver when `monitr` is exited with `imon` = 0. By setting
            `inln` = 3 and returning to the integrator, the residual
            vector is evaluated and placed in the array `r`, and then
            `monitr` is called again. At present this is the only option
            available: `inln` must not be set to any other value.

        hmin : float
            The minimum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4.

        hmax : float
            The maximum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4. If `hmax` is
            set to zero, no limit is assumed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `jac` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    t : float
        The value at which the computed solution y is returned (usually
        at `tout`).

    tout : float
        Normally unchanged. However, when `itask` = 6, `tout` contains
        the value of `t` at which initial values have been computed
        without performing any integration. See descriptions of `itask`
        and `lderiv`.

    y : float, ndarray, shape (neq)
        The computed solution vector, evaluated at `t` (usually `t` =
        `tout`).

    ydot : float, ndarray, shape (neq)
        The time derivatives y' of the vector y at the last integration
        point.

    lderiv : bool, ndarray, shape (2)
        `lderiv`[0] is normally unchanged. However if `itask` = 6 and
        internal initialization was successful then `lderiv`[0] = True.

        `lderiv`[1] = True, if implicit equations were detected.

        Otherwise `lderiv`[1] = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ires` = *<value>* and dy/dt = 0.0 for all
            elements.

            Check the evaluation of the residual for this value of
            `ires`.

        (`errno` 1)
            `ires` was set to an illegal value during initialization.

            `ires` = *<value>* at time *<value>*.

        (`errno` 1)
            `monitr` set `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, *<value>*, is too small.

        (`errno` 1)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. `hmin` =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            Either the integrator setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. `hmax` =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            `monitr` set `inln` = *<value>*.

            Constraint: `inln` = 3.

        (`errno` 1)
            `monitr` appears to have overwritten the solution vector.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            There were repeated convergence-test failures on an
            attempted step, before completing the requested task, but
            the integration was successful as far as `t`. This may be
            caused by an inaccurate Jacobian matrix or one which is
            incorrectly computed.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 7)
            `resid` set `ires` = 3, which signals that an error
            condition has occurred in the solution vector, its time
            derivative or in the value of t. It was not possible to
            remove this condition.

            `ires` = *<value>* at t = *<value>*.

        (`errno` 8)
            Attempt was made to reduce the step size to a value less
            than the minimum step size during the calculation of initial
            values.

            Minimum stepsize: *<value>*.

        (`errno` 8)
            The user problem has one or more inconsistencies between the
            `ires` = 1 and `ires` = -1 parts. Integration will not be
            attempted.

        (`errno` 8)
            The residual function returned an error when calculating the
            initial values of the solution and its time derivative.

        (`errno` 8)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough integer store provided for internal storage
            pointers.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

        (`errno` 15)
            Either the full matrix linear algebra setup function was not
            called first or a communication array has become corrupted.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            Nonlinear solver failed to converge using a damped Newton
            method to solve for initial values.

            Damping factor: *<value>*; convergence rate: *<value>*.

        (`errno` 11)
            `resid` set `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_imp_fulljac as b_ivp_stiff_imp_fulljac

    fun_name = 'naginterfaces.library.ode.ivp_stiff_imp_fulljac'

    cb_resid_name = fun_name + ':resid'

    def cb_resid(neq, t, y, ydot, r, ires, data=None):
        """Wrapping user's resid. Called as base resid."""
        if data is not None:
            _r_resid, _ires_resid = resid(
                t,
                y,
                ydot,
                ires,
                data,
            )
        else:
            _r_resid, _ires_resid = resid(
                t,
                y,
                ydot,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_resid_name, 'returned': 'r'},
            _r_resid,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_resid_base = _ires_resid
        return _ires_resid_base

    cb_jac_name = fun_name + ':jac'

    def cb_jac(neq, t, y, ydot, h, d, p, data=None):
        """Wrapping user's jac. Called as base jac."""
        if data is not None:
            _p_jac = jac(
                t,
                y,
                ydot,
                h,
                d,
                p,
                data,
            )
        else:
            _p_jac = jac(
                t,
                y,
                ydot,
                h,
                d,
                p,
            )
        _p_jac_np = _utils._to_ndarray(
            {'fun_name': cb_jac_name, 'returned': 'p'},
            _p_jac,
            copy=False,
            exp_shape=(neq, neq),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_jac_name},
            ((_p_jac_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_jac_np

    cb_monitr_name = fun_name + ':monitr'

    def cb_monitr(neq, ldysav, sdysav, t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin, hmax, nqu, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
                data,
            )
        else:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
            )
        _hnext_monitr_base = _hnext_monitr
        y[:] = _utils._to_ndarray(
            {'fun_name': cb_monitr_name, 'returned': 'y'},
            _y_monitr,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _imon_monitr_base = _imon_monitr
        _hmin_monitr_base = _hmin_monitr
        _hmax_monitr_base = _hmax_monitr
        _inln_monitr_base = _inln_monitr
        return (
            _hnext_monitr_base,
            _imon_monitr_base,
            _inln_monitr_base,
            _hmin_monitr_base,
            _hmax_monitr_base,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _ydot = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ydot'},
        ydot,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(resid, {'fun_name': fun_name, 'entity': 'resid'})
    _resid = cb_resid
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'wkjac', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _b_utils._check_callable(jac, {'fun_name': fun_name, 'entity': 'jac'}, allow_none=True)
    _jac = None if jac is None else cb_jac
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'}, allow_none=True)
    _monitr = None if monitr is None else cb_monitr
    _lderiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lderiv'},
        lderiv,
        copy=True,
    )
    try:
        _return_scalars = b_ivp_stiff_imp_fulljac(
            _neq,
            _ldysav,
            t,
            tout,
            _y,
            _ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _resid,
            _sdysav,
            _jac,
            _monitr,
            _lderiv,
            itask,
            itrace,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _t_py = _return_scalars.t
    _tout_py = _return_scalars.tout
    _y_py = _y
    _ydot_py = _ydot
    _lderiv_py = _lderiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'tout', 'y', 'ydot', 'lderiv'],
    )(
        _t_py,
        _tout_py,
        _y_py,
        _ydot_py,
        _lderiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_stiff_imp_bandjac(t, tout, y, ydot, comm, rtol, atol, itol, resid, lderiv, itask, itrace, jac=None, monitr=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Implicit/algebraic ordinary differential equations, stiff initial
    value problem, banded Jacobian (comprehensive).

    ``ivp_stiff_imp_bandjac`` is a direct communication function for
    integrating stiff systems of implicit ordinary differential
    equations coupled with algebraic equations when the Jacobian is a
    banded matrix.

    For full information please refer to the NAG Library document for
    d02nh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nhf.html

    Parameters
    ----------
    t : float
        t, the value of the independent variable. The input value of `t`
        is used only on the first call as the initial point of the
        integration.

    tout : float
        The next value of t at which a computed solution is desired. For
        the initial t, the input value of `tout` is used to determine
        the direction of integration. Integration is permitted in either
        direction (see also `itask`).

    y : float, array-like, shape (neq)
        The values of the dependent variables (solution). On the first
        call the first neq elements of `y` must contain the vector of
        initial values.

    ydot : float, array-like, shape (neq)
        If `lderiv`[0] = True, `ydot` must contain approximations to the
        time derivatives y' of the vector y.

        If `lderiv`[0] = False, `ydot` need not be set on entry.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_stiff_bandjac_setup`` and one of ``ivp_stiff_dassl``,
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``ivp_stiff_imp_bandjac`` whether to interpret
        either or both of `rtol` or `atol` as a vector or a scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    resid : callable (r, ires) = resid(t, y, ydot, ires, data=None)
        `resid` must evaluate the residual

            r = g(t,y)-A(t,y)y'

        in one case and

            r = -A(t,y)y'

        in another.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            The value of y_i, for i = 1,2,...,neq.

        ydot : float, ndarray, shape (neq)
            The value of y'_i, for i = 1,2,...,neq, at t.

        ires : int
            The form of the residual that must be returned in array `r`.

            `ires` = -1

              The residual defined in equation (2) must be returned.

            `ires` = 1

              The residual defined in equation (1) must be returned.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (neq)
            `r`[i-1] must contain the ith component of r, for i =
            1,2,...,neq, where

                r = g(t,y)-A(t,y)y'

            or

                r = -A(t,y)y'

            and where the definition of r is determined by the input
            value of `ires`.

        ires : int
            Should be unchanged unless one of the following actions is
            required of the integrator, in which case `ires` should be
            set accordingly.

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling (sub)program with the
              error indicator set to `errno` = 11.

            `ires` = 3

              Indicates to the integrator that an error condition has
              occurred in the solution vector, its time derivative or in
              the value of t. The integrator will use a smaller time
              step to try to avoid this condition. If this is not
              possible, the integrator returns to the calling
              (sub)program with the error indicator set to `errno` = 7.

            `ires` = 4

              Indicates to the integrator to stop its current operation
              and to enter `monitr` immediately with argument `imon` =
              -2.

    lderiv : bool, array-like, shape (2)
        `lderiv`[0] must be set to True if you have supplied both an
        initial y and an initial y'. `lderiv`[0] must be set to False if
        only the initial y has been supplied.

        `lderiv`[1] must be set to True if the integrator is to use a
        modified Newton method to evaluate the initial y and y'.

        Note that y and y', if supplied, are used as initial estimates.

        This method involves taking a small step at the start of the
        integration, and if `itask` = 6 on entry, `t` and `tout` will be
        set to the result of taking this small step. `lderiv`[1] must be
        set to False if the integrator is to use functional iteration to
        evaluate the initial y and y', and if this fails a modified
        Newton method will then be attempted. `lderiv`[1] = True is
        recommended if there are implicit equations or the initial y and
        y' are zero.

    itask : int
        The task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit. tcrit must be specified as an
          option in one of the integrator setup functions before the
          first call to the integrator, or specified in the optional
          input function before a continuation call. tcrit may be equal
          to or beyond `tout`, but not before it, in the direction of
          integration.

        `itask` = 5

          Take one step only and return, without passing tcrit. tcrit
          must be specified as under `itask` = 4.

        `itask` = 6

          The integrator will solve for the initial values of y and y'
          only and then return to the calling (sub)program without doing
          the integration. This option can be used to check the initial
          values of y and y'. Functional iteration or a 'small' backward
          Euler method used in conjunction with a damped Newton
          iteration is used to calculate these values (see `lderiv`).
          Note that if a backward Euler step is used then the value of t
          will have been advanced a short distance from the initial
          point.

        Note: if ``ivp_stiff_imp_bandjac`` is recalled with a different
        value of `itask` (and `tout` altered), the initialization
        procedure is repeated, possibly leading to different initial
        conditions.

    itrace : int
        The level of output that is printed by the integrator. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    jac : None or callable p = jac(t, y, ydot, h, d, ml, mu, p,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jac` must evaluate the Jacobian of the system.

        If this option is not required, the actual argument for `jac`
        must be None. You must indicate to the integrator whether this
        option is to be used by setting the argument jceval
        appropriately in a call to the banded linear algebra setup
        function ``ivp_stiff_bandjac_setup``.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        ydot : float, ndarray, shape (neq)
            The derivative of the solution at the current point t.

        h : float
            The current step size.

        d : float
            The argument d which depends on the integration method.

        ml : int
            The number of subdiagonals and superdiagonals respectively
            in the band.

        mu : int
            The number of subdiagonals and superdiagonals respectively
            in the band.

        p : float, ndarray, shape (`ml`+`mu`+1, neq)
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (`ml`+`mu`+1, neq)
            Elements of the Jacobian matrix dr/dy stored as specified by
            the following pseudocode:

            ::

              for i in range(1, neq+1):
                j1 = max(i-ml, 1)
                j2 = min(i+mu, neq)
                for j in range(j1, j2+1):
                  k = min(ml+1-i, 0) + j
                  p[k-1, i-1] = J[i-1, j-1]

            See also ``lapacklin.dgbtrf``.

            Only nonzero elements of this array need be set, since it is
            preset to zero before the call to `jac`.

    monitr : None or callable (hnext, y, imon, inln, hmin, hmax) =
        monitr(t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin,
        hmax, nqu, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitr` performs tasks requested by you.

        If this option is not required, the actual argument for `monitr`
        must be None.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable.

        hlast : float
            The last step size successfully used by the integrator.

        hnext : float
            The step size that the integrator proposes to take on the
            next step.

        y : float, ndarray, shape (neq)
            y, the values of the dependent variables evaluated at t.

        ydot : float, ndarray, shape (neq)
            The time derivatives y' of the vector y.

        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument may be passed as argument comm in a call to
            ``ivp_stiff_nat_interp`` or comm in a call to
            ``ivp_stiff_c1_interp`` to enable you to carry out
            interpolation using either of those functions.

        r : float, ndarray, shape (neq)
            If `imon` = 0 and `inln` = 3, the first neq elements contain
            the residual vector A(t,y)y'-g(t,y).

        acor : float, ndarray, shape (neq, 2)
            With `imon` = 1, `acor`[i-1,0] contains the weight used for
            the ith equation when the norm is evaluated, and
            `acor`[i-1,1] contains the estimated local error for the ith
            equation. The scaled local error at the end of a timestep
            may be obtained by calling ``ivp_stiff_errest``.

        imon : int
            A flag indicating under what circumstances `monitr` was
            called:

            `imon` = -2

              Entry from the integrator after `ires` = 4 (set in
              `resid`) caused an early termination (this facility could
              be used to locate discontinuities).

            `imon` = -1

              The current step failed repeatedly.

            `imon` = 0

              Entry after a call to the internal nonlinear equation
              solver (see `inln`).

            `imon` = 1

              The current step was successful.

        hmin : float
            The minimum step size to be taken on the next step.

        hmax : float
            The maximum step size to be taken on the next step.

        nqu : int
            The order of the integrator used on the last step. This is
            supplied to enable you to carry out interpolation using
            either of the functions ``ivp_stiff_nat_interp`` or
            ``ivp_stiff_c1_interp``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hnext : float
            The next step size to be used. If this is different from the
            input value, `imon` must be set to 4.

        y : float, array-like, shape (neq)
            These values must not be changed unless `imon` is set to 2.

        imon : int
            May be reset to determine subsequent action in
            ``ivp_stiff_imp_bandjac``.

            `imon` = -2

              Integration is to be halted. A return will be made from
              the integrator to the calling (sub)program with `errno` =
              12.

            `imon` = -1

              Allow the integrator to continue with its own internal
              strategy. The integrator will try up to three restarts
              unless `imon` != -1 on exit.

            `imon` = 0

              Return to the internal nonlinear equation solver, where
              the action taken is determined by the value of `inln` (see
              `inln`).

            `imon` = 1

              Normal exit to the integrator to continue integration.

            `imon` = 2

              Restart the integration at the current time point. The
              integrator will restart from order 1 when this option is
              used. The solution `y`, provided by `monitr`, will be used
              for the initial conditions.

            `imon` = 3

              Try to continue with the same step size and order as was
              to be used before the call to `monitr`. `hmin` and `hmax`
              may be altered if desired.

            `imon` = 4

              Continue the integration but using a new value of `hnext`
              and possibly new values of `hmin` and `hmax`.

        inln : int
            The action to be taken by the internal nonlinear equation
            solver when `monitr` is exited with `imon` = 0. By setting
            `inln` = 3 and returning to the integrator, the residual
            vector is evaluated and placed in the array `r`, and then
            `monitr` is called again. At present this is the only option
            available: `inln` must not be set to any other value.

        hmin : float
            The minimum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4.

        hmax : float
            The maximum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4. If `hmax` is
            set to zero, no limit is assumed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `jac` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    t : float
        The value at which the computed solution y is returned (usually
        at `tout`).

    tout : float
        Normally unchanged. However, when `itask` = 6, `tout` contains
        the value of `t` at which initial values have been computed
        without performing any integration. See descriptions of `itask`
        and `lderiv`.

    y : float, ndarray, shape (neq)
        The computed solution vector, evaluated at `t` (usually `t` =
        `tout`).

    ydot : float, ndarray, shape (neq)
        The time derivatives y' of the vector y at the last integration
        point.

    lderiv : bool, ndarray, shape (2)
        `lderiv`[0] is normally unchanged. However if `itask` = 6 and
        internal initialization was successful then `lderiv`[0] = True.

        `lderiv`[1] = True, if implicit equations were detected.

        Otherwise `lderiv`[1] = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ires` = *<value>* and dy/dt = 0.0 for all
            elements.

            Check the evaluation of the residual for this value of
            `ires`.

        (`errno` 1)
            `ires` was set to an illegal value during initialization.

            `ires` = *<value>* at time *<value>*.

        (`errno` 1)
            `monitr` set `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, *<value>*, is too small.

        (`errno` 1)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. `hmin` =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. `hmax` =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            `monitr` set `inln` = *<value>*.

            Constraint: `inln` = 3.

        (`errno` 1)
            `monitr` appears to have overwritten the solution vector.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

        (`errno` 7)
            `resid` set `ires` = 3, which signals that an error
            condition has occurred in the solution vector, its time
            derivative or in the value of t. It was not possible to
            remove this condition.

            `ires` = *<value>* at t = *<value>*.

        (`errno` 8)
            Attempt was made to reduce the step size to a value less
            than the minimum step size during the calculation of initial
            values.

            Minimum stepsize: *<value>*.

        (`errno` 8)
            The residual function returned an error when calculating the
            initial values of the solution and its time derivative.

        (`errno` 8)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough integer store provided for internal storage
            pointers.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

        (`errno` 15)
            Either the banded matrix linear algebra setup function was
            not called first or a communication array has become
            corrupted.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            There were repeated convergence-test failures on an
            attempted step, before completing the requested task, but
            the integration was successful as far as `t`. This may be
            caused by an inaccurate Jacobian matrix or one which is
            incorrectly computed.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 8)
            Nonlinear solver failed to converge using a damped Newton
            method to solve for initial values.

            Damping factor: *<value>*; convergence rate: *<value>*.

        (`errno` 8)
            The user problem has one or more inconsistencies between the
            `ires` = 1 and `ires` = -1 parts. Integration will not be
            attempted.

        (`errno` 11)
            `resid` set `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_imp_bandjac as b_ivp_stiff_imp_bandjac

    fun_name = 'naginterfaces.library.ode.ivp_stiff_imp_bandjac'

    cb_resid_name = fun_name + ':resid'

    def cb_resid(neq, t, y, ydot, r, ires, data=None):
        """Wrapping user's resid. Called as base resid."""
        if data is not None:
            _r_resid, _ires_resid = resid(
                t,
                y,
                ydot,
                ires,
                data,
            )
        else:
            _r_resid, _ires_resid = resid(
                t,
                y,
                ydot,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_resid_name, 'returned': 'r'},
            _r_resid,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_resid_base = _ires_resid
        return _ires_resid_base

    cb_jac_name = fun_name + ':jac'

    def cb_jac(neq, t, y, ydot, h, d, ml, mu, p, data=None):
        """Wrapping user's jac. Called as base jac."""
        if data is not None:
            _p_jac = jac(
                t,
                y,
                ydot,
                h,
                d,
                ml,
                mu,
                p,
                data,
            )
        else:
            _p_jac = jac(
                t,
                y,
                ydot,
                h,
                d,
                ml,
                mu,
                p,
            )
        _p_jac_np = _utils._to_ndarray(
            {'fun_name': cb_jac_name, 'returned': 'p'},
            _p_jac,
            copy=False,
            exp_shape=(ml+mu+1, neq),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_jac_name},
            ((_p_jac_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_jac_np

    cb_monitr_name = fun_name + ':monitr'

    def cb_monitr(neq, ldysav, sdysav, t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin, hmax, nqu, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
                data,
            )
        else:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
            )
        _hnext_monitr_base = _hnext_monitr
        y[:] = _utils._to_ndarray(
            {'fun_name': cb_monitr_name, 'returned': 'y'},
            _y_monitr,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _imon_monitr_base = _imon_monitr
        _hmin_monitr_base = _hmin_monitr
        _hmax_monitr_base = _hmax_monitr
        _inln_monitr_base = _inln_monitr
        return (
            _hnext_monitr_base,
            _imon_monitr_base,
            _inln_monitr_base,
            _hmin_monitr_base,
            _hmax_monitr_base,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _ydot = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ydot'},
        ydot,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(resid, {'fun_name': fun_name, 'entity': 'resid'})
    _resid = cb_resid
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'wkjac', 'jacpvt', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _b_utils._check_callable(jac, {'fun_name': fun_name, 'entity': 'jac'}, allow_none=True)
    _jac = None if jac is None else cb_jac
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'}, allow_none=True)
    _monitr = None if monitr is None else cb_monitr
    _lderiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lderiv'},
        lderiv,
        copy=True,
    )
    try:
        _return_scalars = b_ivp_stiff_imp_bandjac(
            _neq,
            _ldysav,
            t,
            tout,
            _y,
            _ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _resid,
            _sdysav,
            _jac,
            _monitr,
            _lderiv,
            itask,
            itrace,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _t_py = _return_scalars.t
    _tout_py = _return_scalars.tout
    _y_py = _y
    _ydot_py = _ydot
    _lderiv_py = _lderiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'tout', 'y', 'ydot', 'lderiv'],
    )(
        _t_py,
        _tout_py,
        _y_py,
        _ydot_py,
        _lderiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_stiff_imp_sparjac(t, tout, y, ydot, comm, rtol, atol, itol, resid, lderiv, itask, itrace, jac=None, monitr=None, data=None, io_manager=None):
    u"""
    Implicit/algebraic ordinary differential equations, stiff initial
    value problem, sparse Jacobian (comprehensive).

    ``ivp_stiff_imp_sparjac`` is a direct communication function for
    integrating stiff systems of implicit ordinary differential
    equations coupled with algebraic equations when the Jacobian is a
    sparse matrix.

    For full information please refer to the NAG Library document for
    d02nj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02njf.html

    Parameters
    ----------
    t : float
        t, the value of the independent variable. The input value of `t`
        is used only on the first call as the initial point of the
        integration.

    tout : float
        The next value of t at which a computed solution is desired. For
        the initial t, the input value of `tout` is used to determine
        the direction of integration. Integration is permitted in either
        direction (see also `itask`).

    y : float, array-like, shape (neq)
        The values of the dependent variables (solution). On the first
        call the first neq elements of `y` must contain the vector of
        initial values.

    ydot : float, array-like, shape (neq)
        If `lderiv`[0] = True, `ydot` must contain approximations to the
        time derivatives y' of the vector y.

        If `lderiv`[0] = False, `ydot` need not be set on entry.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_stiff_sparjac_setup`` and one of ``ivp_stiff_dassl``,
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``ivp_stiff_imp_sparjac`` whether to interpret
        either or both of `rtol` or `atol` as a vector or a scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    resid : callable (r, ires) = resid(t, y, ydot, ires, data=None)
        `resid` must evaluate the residual

            r = g(t,y)-A(t,y)y'

        in one case and

            r = -A(t,y)y'

        in another.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            The value of y_i, for i = 1,2,...,neq.

        ydot : float, ndarray, shape (neq)
            The value of y'_i, for i = 1,2,...,neq, at t.

        ires : int
            The form of the residual that must be returned in array `r`.

            `ires` = -1

              The residual defined in equation (2) must be returned.

            `ires` = 1

              The residual defined in equation (1) must be returned.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (neq)
            `r`[i-1] must contain the ith component of r, for i =
            1,2,...,neq, where

                r = g(t,y)-A(t,y)y'

            or

                r = -A(t,y)y'

            and where the definition of r is determined by the input
            value of `ires`.

        ires : int
            Should be unchanged unless one of the following actions is
            required of the integrator, in which case `ires` should be
            set accordingly.

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling (sub)program with the
              error indicator set to `errno` = 11.

            `ires` = 3

              Indicates to the integrator that an error condition has
              occurred in the solution vector, its time derivative or in
              the value of t. The integrator will use a smaller time
              step to try to avoid this condition. If this is not
              possible, the integrator returns to the calling
              (sub)program with the error indicator set to `errno` = 7.

            `ires` = 4

              Indicates to the integrator to stop its current operation
              and to enter `monitr` immediately with argument `imon` =
              -2.

    lderiv : bool, array-like, shape (2)
        `lderiv`[0] must be set to True if you have supplied both an
        initial y and an initial y'. `lderiv`[0] must be set to False if
        only the initial y has been supplied.

        `lderiv`[1] must be set to True if the integrator is to use a
        modified Newton method to evaluate the initial y and y'.

        Note that y and y', if supplied, are used as initial estimates.

        This method involves taking a small step at the start of the
        integration, and if `itask` = 6 on entry, `t` and `tout` will be
        set to the result of taking this small step. `lderiv`[1] must be
        set to False if the integrator is to use functional iteration to
        evaluate the initial y and y', and if this fails a modified
        Newton method will then be attempted. `lderiv`[1] = True is
        recommended if there are implicit equations or the initial y and
        y' are zero.

    itask : int
        The task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit. tcrit must be specified as an
          option in one of the integrator setup functions before the
          first call to the integrator, or specified in the optional
          input function before a continuation call. tcrit may be equal
          to or beyond `tout`, but not before it, in the direction of
          integration.

        `itask` = 5

          Take one step only and return, without passing tcrit. tcrit
          must be specified as under `itask` = 4.

        `itask` = 6

          The integrator will solve for the initial values of y and y'
          only and then return to the calling (sub)program without doing
          the integration. This option can be used to check the initial
          values of y and y'. Functional iteration or a 'small' backward
          Euler method used in conjunction with a damped Newton
          iteration is used to calculate these values (see `lderiv`).
          Note that if a backward Euler step is used then the value of t
          will have been advanced a short distance from the initial
          point.

        Note: if ``ivp_stiff_imp_sparjac`` is recalled with a different
        value of `itask` (and `tout` altered), the initialization
        procedure is repeated, possibly leading to different initial
        conditions.

    itrace : int
        The level of output that is printed by the integrator. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    jac : None or callable pdj = jac(t, y, ydot, h, d, j, pdj,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jac` must evaluate the Jacobian of the system.

        If this option is not required, the actual argument for `jac`
        must be None. You must indicate to the integrator whether this
        option is to be used by setting the argument jceval
        appropriately in a call to the sparse linear algebra setup
        function ``ivp_stiff_sparjac_setup``.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (neq)
            y_i, for i = 1,2,...,neq, the current solution component.

        ydot : float, ndarray, shape (neq)
            The derivative of the solution at the current point t.

        h : float
            The current step size.

        d : float
            The argument d which depends on the integration method.

        j : int
            The column of the Jacobian that `jac` must return in the
            array `pdj`.

        pdj : float, ndarray, shape (neq)
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        pdj : float, array-like, shape (neq)
            `pdj`[i-1] should be set to the (i,j)th element of the
            Jacobian, where j is given by `j`. Only nonzero elements of
            this array need be set, since it is preset to zero before
            the call to `jac`.

    monitr : None or callable (hnext, y, imon, inln, hmin, hmax) =
        monitr(t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin,
        hmax, nqu, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitr` performs tasks requested by you.

        If this option is not required, the actual argument for `monitr`
        must be None.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable.

        hlast : float
            The last step size successfully used by the integrator.

        hnext : float
            The step size that the integrator proposes to take on the
            next step.

        y : float, ndarray, shape (neq)
            y, the values of the dependent variables evaluated at t.

        ydot : float, ndarray, shape (neq)
            The time derivatives y' of the vector y.

        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument may be passed as argument comm in a call to
            ``ivp_stiff_nat_interp`` or comm in a call to
            ``ivp_stiff_c1_interp`` to enable you to carry out
            interpolation using either of those functions.

        r : float, ndarray, shape (neq)
            If `imon` = 0 and `inln` = 3, the first neq elements contain
            the residual vector A(t,y)y'-g(t,y).

        acor : float, ndarray, shape (neq, 2)
            With `imon` = 1, `acor`[i-1,0] contains the weight used for
            the ith equation when the norm is evaluated, and
            `acor`[i-1,1] contains the estimated local error for the ith
            equation. The scaled local error at the end of a timestep
            may be obtained by calling ``ivp_stiff_errest``.

        imon : int
            A flag indicating under what circumstances `monitr` was
            called:

            `imon` = -2

              Entry from the integrator after `ires` = 4 (set in
              `resid`) caused an early termination (this facility could
              be used to locate discontinuities).

            `imon` = -1

              The current step failed repeatedly.

            `imon` = 0

              Entry after a call to the internal nonlinear equation
              solver (see `inln`).

            `imon` = 1

              The current step was successful.

        hmin : float
            The minimum step size to be taken on the next step.

        hmax : float
            The maximum step size to be taken on the next step.

        nqu : int
            The order of the integrator used on the last step. This is
            supplied to enable you to carry out interpolation using
            either of the functions ``ivp_stiff_nat_interp`` or
            ``ivp_stiff_c1_interp``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hnext : float
            The next step size to be used. If this is different from the
            input value, `imon` must be set to 4.

        y : float, array-like, shape (neq)
            These values must not be changed unless `imon` is set to 2.

        imon : int
            May be reset to determine subsequent action in
            ``ivp_stiff_imp_sparjac``.

            `imon` = -2

              Integration is to be halted. A return will be made from
              the integrator to the calling (sub)program with `errno` =
              12.

            `imon` = -1

              Allow the integrator to continue with its own internal
              strategy. The integrator will try up to three restarts
              unless `imon` != -1 on exit.

            `imon` = 0

              Return to the internal nonlinear equation solver, where
              the action taken is determined by the value of `inln` (see
              `inln`).

            `imon` = 1

              Normal exit to the integrator to continue integration.

            `imon` = 2

              Restart the integration at the current time point. The
              integrator will restart from order 1 when this option is
              used. The solution `y`, provided by `monitr`, will be used
              for the initial conditions.

            `imon` = 3

              Try to continue with the same step size and order as was
              to be used before the call to `monitr`. `hmin` and `hmax`
              may be altered if desired.

            `imon` = 4

              Continue the integration but using a new value of `hnext`
              and possibly new values of `hmin` and `hmax`.

        inln : int
            The action to be taken by the internal nonlinear equation
            solver when `monitr` is exited with `imon` = 0. By setting
            `inln` = 3 and returning to the integrator, the residual
            vector is evaluated and placed in the array `r`, and then
            `monitr` is called again. At present this is the only option
            available: `inln` must not be set to any other value.

        hmin : float
            The minimum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4.

        hmax : float
            The maximum step size to be used. If this is different from
            the input value, `imon` must be set to 3 or 4. If `hmax` is
            set to zero, no limit is assumed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    t : float
        The value at which the computed solution y is returned (usually
        at `tout`).

    tout : float
        Normally unchanged. However, when `itask` = 6, `tout` contains
        the value of `t` at which initial values have been computed
        without performing any integration. See descriptions of `itask`
        and `lderiv`.

    y : float, ndarray, shape (neq)
        The computed solution vector, evaluated at `t` (usually `t` =
        `tout`).

    ydot : float, ndarray, shape (neq)
        The time derivatives y' of the vector y at the last integration
        point.

    lderiv : bool, ndarray, shape (2)
        `lderiv`[0] is normally unchanged. However if `itask` = 6 and
        internal initialization was successful then `lderiv`[0] = True.

        `lderiv`[1] = True, if implicit equations were detected.

        Otherwise `lderiv`[1] = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ires` = *<value>* and dy/dt = 0.0 for all
            elements.

            Check the evaluation of the residual for this value of
            `ires`.

        (`errno` 1)
            `ires` was set to an illegal value during initialization.

            `ires` = *<value>* at time *<value>*.

        (`errno` 1)
            `monitr` set `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, h = *<value>*, is too small.

        (`errno` 1)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. hmin =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            Either the integrator setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. hmax =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            `monitr` set `inln` = *<value>*.

            Constraint: `inln` = 3.

        (`errno` 1)
            `monitr` appears to have overwritten the solution vector.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

            See ``ivp_stiff_dassl`` for details of maxstp.

        (`errno` 7)
            `resid` set `ires` = 3, which signals that an error
            condition has occurred in the solution vector, its time
            derivative or in the value of t. It was not possible to
            remove this condition.

            `ires` = *<value>* at t = *<value>*.

        (`errno` 8)
            Attempt was made to reduce the step size to a value less
            than the minimum step size during the calculation of initial
            values.

            Minimum stepsize: *<value>*.

        (`errno` 8)
            The residual function returned an error when calculating the
            initial values of the solution and its time derivative.

        (`errno` 8)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough integer store provided for internal storage
            pointers.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

        (`errno` 15)
            Either the sparse matrix linear algebra setup function was
            not called first or a communication array has become
            corrupted.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            There were repeated convergence-test failures on an
            attempted step, before completing the requested task, but
            the integration was successful as far as `t`. This may be
            caused by an inaccurate Jacobian matrix or one which is
            incorrectly computed.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 8)
            Nonlinear solver failed to converge using a damped Newton
            method to solve for initial values.

            Damping factor: *<value>*; convergence rate: *<value>*.

        (`errno` 8)
            The user problem has one or more inconsistencies between the
            `ires` = 1 and `ires` = -1 parts. Integration will not be
            attempted.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 11)
            `resid` set `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_imp_sparjac as b_ivp_stiff_imp_sparjac

    fun_name = 'naginterfaces.library.ode.ivp_stiff_imp_sparjac'

    cb_resid_name = fun_name + ':resid'

    def cb_resid(neq, t, y, ydot, r, ires, data=None):
        """Wrapping user's resid. Called as base resid."""
        if data is not None:
            _r_resid, _ires_resid = resid(
                t,
                y,
                ydot,
                ires,
                data,
            )
        else:
            _r_resid, _ires_resid = resid(
                t,
                y,
                ydot,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_resid_name, 'returned': 'r'},
            _r_resid,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_resid_base = _ires_resid
        return _ires_resid_base

    cb_jac_name = fun_name + ':jac'

    def cb_jac(neq, t, y, ydot, h, d, j, pdj, data=None):
        """Wrapping user's jac. Called as base jac."""
        if data is not None:
            _pdj_jac = jac(
                t,
                y,
                ydot,
                h,
                d,
                j,
                pdj,
                data,
            )
        else:
            _pdj_jac = jac(
                t,
                y,
                ydot,
                h,
                d,
                j,
                pdj,
            )
        pdj[:] = _utils._to_ndarray(
            {'fun_name': cb_jac_name, 'returned': 'pdj'},
            _pdj_jac,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_monitr_name = fun_name + ':monitr'

    def cb_monitr(neq, ldysav, sdysav, t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin, hmax, nqu, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
                data,
            )
        else:
            _hnext_monitr, _y_monitr, _imon_monitr, _inln_monitr, _hmin_monitr, _hmax_monitr = monitr(
                t,
                hlast,
                hnext,
                y,
                ydot,
                comm,
                r,
                acor,
                imon,
                hmin,
                hmax,
                nqu,
            )
        _hnext_monitr_base = _hnext_monitr
        y[:] = _utils._to_ndarray(
            {'fun_name': cb_monitr_name, 'returned': 'y'},
            _y_monitr,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _imon_monitr_base = _imon_monitr
        _hmin_monitr_base = _hmin_monitr
        _hmax_monitr_base = _hmax_monitr
        _inln_monitr_base = _inln_monitr
        return (
            _hnext_monitr_base,
            _imon_monitr_base,
            _inln_monitr_base,
            _hmin_monitr_base,
            _hmax_monitr_base,
        )

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _ydot = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ydot'},
        ydot,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(resid, {'fun_name': fun_name, 'entity': 'resid'})
    _resid = cb_resid
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'wkjac', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'jacpvt', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _b_utils._check_callable(jac, {'fun_name': fun_name, 'entity': 'jac'}, allow_none=True)
    _jac = None if jac is None else cb_jac
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'}, allow_none=True)
    _monitr = None if monitr is None else cb_monitr
    _lderiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lderiv'},
        lderiv,
        copy=True,
    )
    try:
        _return_scalars = b_ivp_stiff_imp_sparjac(
            _neq,
            _ldysav,
            t,
            tout,
            _y,
            _ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _resid,
            _sdysav,
            _jac,
            _monitr,
            _lderiv,
            itask,
            itrace,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _t_py = _return_scalars.t
    _tout_py = _return_scalars.tout
    _y_py = _y
    _ydot_py = _ydot
    _lderiv_py = _lderiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'tout', 'y', 'ydot', 'lderiv'],
    )(
        _t_py,
        _tout_py,
        _y_py,
        _ydot_py,
        _lderiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_stiff_exp_revcom(t, tout, y, ydot, comm, rtol, atol, itol, wkjac, imon, inln, ires, irevcm, itask, itrace, io_manager=None):
    u"""
    Explicit ordinary differential equations, stiff initial value
    problem (reverse communication, comprehensive).

    ``ivp_stiff_exp_revcom`` is a reverse communication function for
    integrating stiff systems of explicit ordinary differential
    equations.

    For full information please refer to the NAG Library document for
    d02nm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nmf.html

    Parameters
    ----------
    t : float
        On initial entry: t, the value of the independent variable. The
        input value of `t` is used only on the first call as the initial
        point of the integration.

    tout : float
        On initial entry: the next value of t at which a computed
        solution is desired. For the initial t, the input value of
        `tout` is used to determine the direction of integration.
        Integration is permitted in either direction (see also `itask`).

    y : float, ndarray, shape (neq), modified in place
        On initial entry: the values of the dependent variables
        (solution). On the first call the first neq elements of y must
        contain the vector of initial values.

        On final exit: the computed solution vector evaluated at `t`
        (usually t = `tout`).

    ydot : float, ndarray, shape (neq), modified in place
        On intermediate entry: must be set to the derivatives as defined
        under the description of `irevcm`.

        On final exit: the time derivatives y' of the vector y at the
        last integration point.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to one
        of ``ivp_stiff_fulljac_setup``, ``ivp_stiff_bandjac_setup`` or
        ``ivp_stiff_sparjac_setup`` and one of ``ivp_stiff_dassl``,
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        On initial entry: the relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        On initial entry: the absolute local error tolerance.

    itol : int
        On initial entry: a value to indicate the form of the local
        error test. `itol` indicates to ``ivp_stiff_exp_revcom`` whether
        to interpret either or both of `rtol` or `atol` as a vector or a
        scalar. The error test to be satisfied is ||e_i/w_i|| < 1.0,
        where w_i is defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    wkjac : float, ndarray, shape (nwkjac), modified in place
        On intermediate entry: elements of the Jacobian as defined under
        the description of `irevcm`. If a numerical Jacobian was
        requested then `wkjac` is used for workspace.

        On intermediate exit: the Jacobian is overwritten.

    imon : int
        On intermediate entry: may be reset to determine subsequent
        action in ``ivp_stiff_exp_revcom``.

        `imon` = -2

          Integration is to be halted. A return will be made from
          ``ivp_stiff_exp_revcom`` to the calling (sub)program with
          `errno` = 12.

        `imon` = -1

          Allow ``ivp_stiff_exp_revcom`` to continue with its own
          internal strategy. The integrator will try up to three
          restarts unless `imon` != -1.

        `imon` = 0

          Return to the internal nonlinear equation solver, where the
          action taken is determined by the value of `inln`.

        `imon` = 1

          Normal exit to ``ivp_stiff_exp_revcom`` to continue
          integration.

        `imon` = 2

          Restart the integration at the current time point. The
          integrator will restart from order 1 when this option is used.
          The solution `y`, provided by the `monitr` operation (see
          [Description]), will be used for the initial conditions.

        `imon` = 3

          Try to continue with the same step size and order as was to be
          used before entering the `monitr` operation (see
          [Description]). `hmin` and `hmax` may be altered if desired.

        `imon` = 4

          Continue the integration but using a new value of `hnext` and
          possibly new values of `hmin` and `hmax`.

    inln : int
        On intermediate entry: with `imon` = 0 and `irevcm` = 9, `inln`
        specifies the action to be taken by the internal nonlinear
        equation solver. By setting `inln` = 3 and returning to
        ``ivp_stiff_exp_revcom``, the residual vector is evaluated and
        placed in `comm`['rwork'][50+2*neq+i-1], for i = 1,2,...,neq and
        then the `monitr` operation (see [Description]) is invoked
        again. At present this is the only option available: `inln` must
        not be set to any other value.

    ires : int
        On intermediate entry: should be unchanged unless one of the
        following actions is required of ``ivp_stiff_exp_revcom`` in
        which case `ires` should be set accordingly.

        `ires` = 2

          Indicates to ``ivp_stiff_exp_revcom`` that control should be
          passed back immediately to the calling (sub)program with the
          error indicator set to `errno` = 11.

        `ires` = 3

          Indicates to ``ivp_stiff_exp_revcom`` that an error condition
          has occurred in the solution vector, its time derivative or in
          the value of t. The integrator will use a smaller time step to
          try to avoid this condition. If this is not possible
          ``ivp_stiff_exp_revcom`` returns to the calling (sub)program
          with the error indicator set to `errno` = 7.

        `ires` = 4

          Indicates to ``ivp_stiff_exp_revcom`` to stop its current
          operation and to enter the `monitr` operation (see
          [Description]) immediately.

    irevcm : int
        On initial entry: must contain 0.

        On intermediate entry: should remain unchanged.

    itask : int
        On initial entry: the task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit. tcrit must be specified as an
          option in one of the integrator setup functions before the
          first call to the integrator, or specified in the optional
          input function before a continuation call. tcrit (e.g., see
          ``ivp_stiff_bdf``) may be equal to or beyond `tout`, but not
          before it in the direction of integration.

        `itask` = 5

          Take one step only and return, without passing tcrit (e.g.,
          see ``ivp_stiff_bdf``). tcrit must be specified under `itask`
          = 4.

    itrace : int
        On initial entry: the level of output that is printed by the
        integrator. `itrace` may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    t : float
        On final exit: the value at which the computed solution y is
        returned (usually at `tout`).

    imon : int
        On intermediate exit: used to pass information between
        ``ivp_stiff_exp_revcom`` and the MONITR operation (see
        [Description]). With `irevcm` = 9, `imon` contains a flag
        indicating under what circumstances the return from
        ``ivp_stiff_exp_revcom`` occurred:

        `imon` = -2

          Exit from ``ivp_stiff_exp_revcom`` after `ires` = 4 caused an
          early termination (this facility could be used to locate
          discontinuities).

        `imon` = -1

          The current step failed repeatedly.

        `imon` = 0

          Exit from ``ivp_stiff_exp_revcom`` after a call to the
          internal nonlinear equation solver.

        `imon` = 1

          The current step was successful.

    inln : int
        On intermediate exit: contains a flag indicating the action to
        be taken, if any, by the internal nonlinear equation solver.

    ires : int
        On intermediate exit: with `irevcm` = 1, 2, 3, 4 or 5, `ires`
        contains the value 1.

    irevcm : int
        On intermediate exit: indicates what action you must take before
        re-entering. The possible exit values of `irevcm` are 1, 3, 4,
        5, 8, 9 or 10, which should be interpreted as follows:

        `irevcm` = 1, 3, 4 or 5

          Indicates that an FCN operation (see [Description]) is
          required: y' = g(t,y) must be supplied, where `y`[i-1] is
          located in y_i, for i = 1,2,...,neq.

          For `irevcm` = 1 or 3, y'_i should be placed in location
          `comm`['rwork'][50+2*neq+i-1], for i = 1,2,...,neq.

          For `irevcm` = 4, y'_i should be placed in location
          `comm`['rwork'][50+neq+i-1], for i = 1,2,...,neq.

          For `irevcm` = 5, y'_i should be placed in location
          `ydot`[i-1], for i = 1,2,...,neq.

        `irevcm` = 8

          Indicates that a JAC operation (see [Description]) is
          required: the Jacobian matrix must be supplied.

          If full matrix linear algebra is being used, the (i,j)th
          element of the Jacobian must be stored in
          `wkjac`[(j-1)*neq+i-1].

          If banded matrix linear algebra is being used then the (i,j)th
          element of the Jacobian must be stored in
          `wkjac`[(i-1)*m_B+k-1], where m_B = m_L+m_U+1 and k =
          min(m_L-i+1, 0)+j; here m_L and m_U are the number of
          subdiagonals and superdiagonals, respectively, in the band.

          If sparse matrix linear algebra is being used then
          ``ivp_stiff_sparjac_enq`` must be called to determine which
          column of the Jacobian is required and where it should be
          stored.

        `irevcm` = 9

          Indicates that a MONITR operation (see [Description]) can be
          performed.

        `irevcm` = 10

          Indicates that the current step was not successful, due to
          error test failure or convergence test failure. The only
          information supplied to you on this return is the current
          value of the independent variable t, located in
          `comm`['rwork'][18]. No values must be changed before
          re-entering ``ivp_stiff_exp_revcom``; this facility enables
          you to determine the number of unsuccessful steps.

        On final exit: `irevcm` = 0 indicated the user-specified task
        has been completed or an error has been encountered (see the
        descriptions for `itask` and errno).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On re-entry, `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, *<value>*, is too small.

        (`errno` 1)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of nwkjac is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            nwkjac = *<value>*, nwkjac (setup) = *<value>*.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. hmin =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            Either the integrator setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. hmax =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On re-entry, `inln` != 3 for the case `irevcm` = 9 and
            `imon` = 0.

            `inln` = *<value>*.

        (`errno` 1)
            On entry, `irevcm` is invalid: `irevcm` = *<value>*.

        (`errno` 1)
            On re-entry, the solution vector appears to have been
            overwritten.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

        (`errno` 7)
            The derivative evaluation procedure set the error flag
            `ires` = 3 repeatedly despite attempts by the integrator to
            avoid this.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 10)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            There were repeated convergence-test failures on an
            attempted step, before completing the requested task, but
            the integration was successful as far as `t`. This may be
            caused by an inaccurate Jacobian matrix or one which is
            incorrectly computed.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 11)
            On re-entry, `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_exp_revcom as b_ivp_stiff_exp_revcom

    fun_name = 'naginterfaces.library.ode.ivp_stiff_exp_revcom'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'ydot'},
        ydot,
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'jacpvt', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'wkjac'},
        wkjac,
    )
    _nwkjac_py = _b_utils._EngineFloat64ArrayType.get_shape(wkjac, {'fun_name': fun_name, 'entity': 'wkjac'}, exp_rank=1)[0]
    _nwkjac = _b_utils._EngineIntScalarType(
        _nwkjac_py, {'fun_name': fun_name, 'entity': '_nwkjac'},
    ).to_ctypes()
    try:
        _return_scalars = b_ivp_stiff_exp_revcom(
            _neq,
            _ldysav,
            t,
            tout,
            y,
            ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _sdysav,
            wkjac,
            _nwkjac,
            imon,
            inln,
            ires,
            irevcm,
            itask,
            itrace,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _t_py = _return_scalars.t
    _imon_py = _return_scalars.imon
    _inln_py = _return_scalars.inln
    _ires_py = _return_scalars.ires
    _irevcm_py = _return_scalars.irevcm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'imon', 'inln', 'ires', 'irevcm'],
    )(
        _t_py,
        _imon_py,
        _inln_py,
        _ires_py,
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_stiff_imp_revcom(t, tout, y, ydot, comm, rtol, atol, itol, wkjac, imon, inln, ires, irevcm, lderiv, itask, itrace, io_manager=None):
    u"""
    Implicit/algebraic ordinary differential equations, stiff initial
    value problem (reverse communication, comprehensive).

    ``ivp_stiff_imp_revcom`` is a reverse communication function for
    integrating stiff systems of implicit ordinary differential
    equations coupled with algebraic equations.

    For full information please refer to the NAG Library document for
    d02nn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nnf.html

    Parameters
    ----------
    t : float
        On initial entry: t, the value of the independent variable. The
        input value of `t` is used only on the first call as the initial
        point of the integration.

    tout : float
        On initial entry: the next value of t at which a computed
        solution is desired. For the initial t, the input value of
        `tout` is used to determine the direction of integration.
        Integration is permitted in either direction (see also `itask`).

    y : float, ndarray, shape (neq), modified in place
        On initial entry: the values of the dependent variables
        (solution). On the first call the first neq elements of y must
        contain the vector of initial values.

        On final exit: the computed solution vector evaluated at `t`
        (usually t = `tout`).

    ydot : float, ndarray, shape (neq), modified in place
        On initial entry: if `lderiv`[0] = True, `ydot` must contain
        approximations to the time derivatives y' of the vector y. If
        `lderiv`[0] = False, `ydot` need not be set on entry.

        On final exit: contains the time derivatives y' of the vector y
        at the last integration point.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to one
        of ``ivp_stiff_fulljac_setup``, ``ivp_stiff_bandjac_setup`` or
        ``ivp_stiff_sparjac_setup`` and one of ``ivp_stiff_dassl``,
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; otherwise: neq.

        On initial entry: the relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; otherwise: neq.

        On initial entry: the absolute local error tolerance.

    itol : int
        On initial entry: a value to indicate the form of the local
        error test. `itol` indicates to ``ivp_stiff_imp_revcom`` whether
        to interpret either or both of `rtol` or `atol` as a vector or a
        scalar. The error test to be satisfied is ||e_i/w_i|| < 1.0,
        where w_i is defined as follows:

        [table omitted]

        e_i is an estimate of the local error in y_i, computed
        internally, and the choice of norm to be used is defined by a
        previous call to an integrator setup function.

    wkjac : float, ndarray, shape (nwkjac), modified in place
        On intermediate entry: elements of the Jacobian as defined under
        the description of `irevcm`. If a numerical Jacobian was
        requested then `wkjac` is used for workspace.

        On intermediate exit: the Jacobian is overwritten.

    imon : int
        On intermediate entry: may be reset to determine subsequent
        action in ``ivp_stiff_imp_revcom``.

        `imon` = -2

          Integration is to be halted. A return will be made from
          ``ivp_stiff_imp_revcom`` to the calling (sub)program with
          `errno` = 12.

        `imon` = -1

          Allow ``ivp_stiff_imp_revcom`` to continue with its own
          internal strategy. The integrator will try up to three
          restarts unless `imon` != -1.

        `imon` = 0

          Return to the internal nonlinear equation solver, where the
          action taken is determined by the value of `inln`.

        `imon` = 1

          Normal exit to ``ivp_stiff_imp_revcom`` to continue
          integration.

        `imon` = 2

          Restart the integration at the current time point. The
          integrator will restart from order 1 when this option is used.
          The internal initialization module solves for new values of y
          and y' by using the values supplied in `y` and `ydot` by the
          `monitr` operation (see [Description]) as initial estimates.

        `imon` = 3

          Try to continue with the same step size and order as was to be
          used before entering the `monitr` operation (see
          [Description]). `hmin` and `hmax` may be altered if desired.

        `imon` = 4

          Continue the integration but using a new value of `hnext` and
          possibly new values of `hmin` and `hmax`.

    inln : int
        On intermediate entry: with `imon` = 0 and `irevcm` = 9, `inln`
        specifies the action to be taken by the internal nonlinear
        equation solver. By setting `inln` = 3 and returning to
        ``ivp_stiff_imp_revcom``, the residual vector is evaluated and
        placed in `comm`['rwork'][50+2*neq+i-1], for i = 1,2,...,neq and
        then the `monitr` operation (see [Description]) is invoked
        again. At present this is the only option available: `inln` must
        not be set to any other value.

    ires : int
        On intermediate entry: should be unchanged unless one of the
        following actions is required of ``ivp_stiff_imp_revcom`` in
        which case `ires` should be set accordingly.

        `ires` = 2

          Indicates to ``ivp_stiff_imp_revcom`` that control should be
          passed back immediately to the calling (sub)program with the
          error indicator set to `errno` = 11.

        `ires` = 3

          Indicates to ``ivp_stiff_imp_revcom`` that an error condition
          has occurred in the solution vector, its time derivative or in
          the value of t. The integrator will use a smaller time step to
          try to avoid this condition. If this is not possible
          ``ivp_stiff_imp_revcom`` returns to the calling (sub)program
          with the error indicator set to `errno` = 7.

        `ires` = 4

          Indicates to ``ivp_stiff_imp_revcom`` to stop its current
          operation and to enter the `monitr` operation (see
          [Description]) immediately.

    irevcm : int
        On initial entry: must contain 0.

        On intermediate entry: should remain unchanged.

    lderiv : bool, ndarray, shape (2), modified in place
        On initial entry: `lderiv`[0] must be set to True if you have
        supplied both an initial y and an initial y'. `lderiv`[0] must
        be set to False if only the initial y has been supplied.

        `lderiv`[1] must be set to True if the integrator is to use a
        modified Newton method to evaluate the initial y and y'.

        Note that y and y', if supplied, are used as initial estimates.

        This method involves taking a small step at the start of the
        integration, and if `itask` = 6 on entry, `t` and `tout` will be
        set to the result of taking this small step. `lderiv`[1] must be
        set to False if the integrator is to use functional iteration to
        evaluate the initial y and y', and if this fails a modified
        Newton method will then be attempted. `lderiv`[1] = True is
        recommended if there are implicit equations or the initial y and
        y' are zero.

        On final exit: `lderiv`[0] is normally unchanged. However if
        `itask` = 6 and internal initialization was successful then
        `lderiv`[0] = True.

        `lderiv`[1] = True, if implicit equations were detected.

        Otherwise `lderiv`[1] = False.

    itask : int
        On initial entry: the task to be performed by the integrator.

        `itask` = 1

          Normal computation of output values of y(t) at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step only and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout` and return.

        `itask` = 4

          Normal computation of output values of y(t) at t = `tout` but
          without overshooting t = tcrit. tcrit must be specified as an
          option in one of the integrator setup functions before the
          first call to the integrator, or specified in the optional
          input function before a continuation call. tcrit (e.g., see
          ``ivp_stiff_bdf``) may be equal to or beyond `tout`, but not
          before it in the direction of integration.

        `itask` = 5

          Take one step only and return, without passing tcrit (e.g.,
          see ``ivp_stiff_bdf``). tcrit must be specified under `itask`
          = 4.

        `itask` = 6

          The integrator will solve for the initial values of y and y'
          only and then return to the calling (sub)program without doing
          the integration. This option can be used to check the initial
          values of y and y'. Functional iteration or a 'small' backward
          Euler method used in conjunction with a damped Newton
          iteration is used to calculate these values (see `lderiv`).
          Note that if a backward Euler step is used then the value of t
          will have been advanced a short distance from the initial
          point.

        Note: if ``ivp_stiff_imp_revcom`` is recalled with a different
        value of `itask` (and `tout` altered) then the initialization
        procedure is repeated, possibly leading to different initial
        conditions.

    itrace : int
        On initial entry: the level of output that is printed by the
        integrator. `itrace` may take the value -1, 0, 1, 2 or 3.

        `itrace` < -1

          -1 is assumed and similarly if `itrace` > 3, 3 is assumed.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed on the current error message
          unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        `itrace` > 0

          Warning messages are printed as above, and on the file object
          associated with the advisory I/O unit (see
          :class:`~naginterfaces.base.utils.FileObjManager`) output is
          generated which details Jacobian entries, the nonlinear
          iteration and the time integration. The advisory messages are
          given in greater detail the larger the value of `itrace`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    t : float
        On final exit: the value at which the computed solution y is
        returned (usually at `tout`).

    tout : float
        Is unaltered unless `itask` = 6 and `lderiv`[1] = True on entry
        (see also `itask` and `lderiv`) in which case `tout` will be set
        to the result of taking a small step at the start of the
        integration.

    imon : int
        On intermediate exit: used to pass information between
        ``ivp_stiff_imp_revcom`` and the `monitr` operation (see
        [Description]). With `irevcm` = 9, `imon` contains a flag
        indicating under what circumstances the return from
        ``ivp_stiff_imp_revcom`` occurred:

        `imon` = -2

          Exit from ``ivp_stiff_imp_revcom`` after `ires` = 4 (set in
          the `resid` operation (see [Description]) caused an early
          termination (this facility could be used to locate
          discontinuities).

        `imon` = -1

          The current step failed repeatedly.

        `imon` = 0

          Exit from ``ivp_stiff_imp_revcom`` after a call to the
          internal nonlinear equation solver.

        `imon` = 1

          The current step was successful.

    inln : int
        On intermediate exit: contains a flag indicating the action to
        be taken, if any, by the internal nonlinear equation solver.

    ires : int
        On intermediate exit: with `irevcm` = 1, 2, 3, 4, 5, 6, 7 or 11,
        `ires` specifies the form of the residual to be returned by the
        `resid` operation (see [Description]).

        If `ires` = 1, -r = g(t,y)-A(t,y)y' must be returned.

        If `ires` = -1, -r^= -A(t,y)y' must be returned.

    irevcm : int
        On intermediate exit: indicates what action you must take before
        re-entering ``ivp_stiff_imp_revcom``. The possible exit values
        of `irevcm` are 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 or 11 which should
        be interpreted as follows:

        `irevcm` = 1, 2, 3, 4, 5, 6, 7 or 11

          Indicates that a `resid` operation (see [Description]) is
          required: you must supply the residual of the system. For each
          of these values of `irevcm`, y_i is located in `y`[i-1], for i
          = 1,2,...,neq.

          For `irevcm` = 1, 3, 6 or 11, y'_i is located in `ydot`[i-1]
          and r_i should be stored in `comm`['rwork'][50+2*neq+i-1], for
          i = 1,2,...,neq.

          For `irevcm` = 2, y'_i is located in
          `comm`['rwork'][50+neq+i-1] and r_i should be stored in
          `comm`['rwork'][50+2*neq+i-1], for i = 1,2,...,neq.

          For `irevcm` = 4 or 7, y'_i is located in `ydot`[i-1] and r_i
          should be stored in `comm`['rwork'][50+neq+i-1], for i =
          1,2,...,neq.

          For `irevcm` = 5, y'_i is located in
          `comm`['rwork'][50+2*neq+i-1] and r_i should be stored in
          `ydot`[i-1], for i = 1,2,...,neq.

        `irevcm` = 8

          Indicates that a `jac` operation (see [Description]) is
          required: you must supply the Jacobian matrix.

          If full matrix linear algebra is being used, the (i,j)th
          element of the Jacobian must be stored in
          `wkjac`[(j-1)*neq+i-1].

          If banded matrix linear algebra is being used, the (i,j)th
          element of the Jacobian must be stored in
          `wkjac`[(i-1)*m_B+k-1], where m_B = m_L+m_U+1 and k =
          min(m_L-i+1, 0)+j; here m_L and m_U are the number of
          subdiagonals and superdiagonals, respectively, in the band.

          If sparse matrix linear algebra is being used,
          ``ivp_stiff_sparjac_enq`` must be called to determine which
          column of the Jacobian is required and where it should be
          stored.

        `irevcm` = 9

          Indicates that a `monitr` operation (see [Description]) can be
          performed.

        `irevcm` = 10

          Indicates that the current step was not successful, due to
          error test failure or convergence test failure. The only
          information supplied to you on this return is the current
          value of the variable t, located in `comm`['rwork'][18]. No
          values must be changed before re-entering
          ``ivp_stiff_imp_revcom``; this facility enables you to
          determine the number of unsuccessful steps.

        On final exit: `irevcm` = 0 indicating that the user-specified
        task has been completed or an error has been encountered (see
        the descriptions for `itask` and errno).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ires` = *<value>* and dy/dt = 0.0 for all
            elements.

            Check the evaluation of the residual for this value of
            `ires`.

        (`errno` 1)
            On re-entry, `ires` has been set to an illegal value during
            initialization.

            `ires` = *<value>* at time *<value>*.

        (`errno` 1)
            On re-entry, `imon` = *<value>*.

            Constraint: -2 <= `imon` <= 4.

        (`errno` 1)
            Failure during internal time interpolation. `tout` and the
            current time are too close.

            `itask` = *<value>* and `tout` = *<value>* and the current
            time is *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before the current
            time in the direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and the current time
            is *<value>*.

        (`errno` 1)
            `itask` = 3 and `tout` is more than an integration step
            behind the current time.

            `tout` = *<value>*, current time minus step size: *<value>*.

        (`errno` 1)
            The initial stepsize, h = *<value>*, is too small.

        (`errno` 1)
            Weight number i = *<value>* used in the local error test is
            too small. Check the values of `rtol` and `atol`.

            `atol`[i-1] and `y`[i-1] may both be zero.

            Weight i = *<value>*.

        (`errno` 1)
            On entry, `itask` = 4 or 5 and tcrit is before `tout` in the
            direction of integration.

            `itask` = *<value>*, tcrit = *<value>* and `tout` =
            *<value>*.

        (`errno` 1)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 1)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 1)
            Either the value of nwkjac is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            nwkjac = *<value>*, nwkjac (setup) = *<value>*.

        (`errno` 1)
            Either the value of sdysav is not the same as the value
            supplied to the setup function or a communication array has
            become corrupted.

            sdysav = *<value>*, sdysav (setup) = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) minimum stepsize was
            provided in a prior call to a setup function. hmin =
            *<value>*.

        (`errno` 1)
            On entry, `tout` is less than `t` with respect to the
            direction of integration given by the sign of h0 in a prior
            call to a setup function.

            `tout` = *<value>*, `t` = *<value>* and h0 = *<value>*.

        (`errno` 1)
            On entry, an illegal (negative) maximum number of steps was
            provided in a prior call to a setup function. maxstp =
            *<value>*.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: 1 <= `itol` <= 4.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            Either the function was entered on a continuation call
            without a prior call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, `tout` is too close to `t` to start integration.

            `tout` = *<value>* and `t` = *<value>*.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: 1 <= `itask` <= 5.

        (`errno` 1)
            Either the integrator setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On entry, an illegal (negative) maximum stepsize was
            provided in a prior call to a setup function. hmax =
            *<value>*.

        (`errno` 1)
            Either the linear algebra setup function has not been called
            prior to the first call of this function, or a communication
            array has become corrupted.

        (`errno` 1)
            On re-entry, `inln` = *<value>* for the case `irevcm` = 9
            and `imon` = 0.

            Constraint: `inln` = 3.

        (`errno` 1)
            On entry, `irevcm` = *<value>*.

            Constraint: 0 <= `irevcm` <= 11.

        (`errno` 1)
            On re-entry, the solution vector appears to have been
            overwritten.

            Further integration will not be attempted.

        (`errno` 2)
            At time *<value>* the maximum number of allowed steps on
            this call was taken before reaching the next output point
            `tout` = *<value>*.

            Maximum number of steps = *<value>*.

        (`errno` 7)
            On re-entry, `ires` = 3 which signals that an error
            condition has occurred in the solution vector, its time
            derivative or in the value of t. It was not possible to
            remove this condition.

            `ires` = *<value>* at t = *<value>*.

        (`errno` 8)
            Attempt was made to reduce the step size to a value less
            than the minimum step size during the calculation of initial
            values.

            Minimum stepsize: *<value>*.

        (`errno` 8)
            The user problem has one or more inconsistencies between the
            `ires` = 1 and `ires` = -1 parts. Integration will not be
            attempted.

        (`errno` 8)
            The residual function returned an error when calculating the
            initial values of the solution and its time derivative.

        (`errno` 8)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 9)
            A singular Jacobian has been encountered. You should check
            the problem formulation and Jacobian calculation.

        (`errno` 10)
            Workspace error occurred when trying to form the Jacobian
            matrix in calculating the initial values of the solution and
            its time derivative.

        (`errno` 10)
            Not enough integer store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 10)
            Larger integer workspace required.

            Provided: *<value>*; required: *<value>*.

        (`errno` 10)
            Not enough real store provided for sparse matrix solver.

            Units of store needed: *<value>*. Amount provided:
            *<value>*.

        (`errno` 14)
            On entry, too much accuracy requested for precision of the
            machine at the start of problem. The tolerances should be
            checked; the requested accuracy should be reduced by a
            factor of at least *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Too much accuracy requested for precision of the machine at
            time *<value>*.

            The tolerances should be checked; the requested accuracy
            should be reduced by a factor of at least *<value>*.

        (`errno` 4)
            There were repeated error-test failures on an attempted
            step, before completing the requested task, but the
            integration was successful as far as `t`. The problem may
            have a singularity, or the local error requirements may be
            inappropriate.

        (`errno` 5)
            Nonlinear solver failed to converge using a damped Newton
            method to solve for initial values.

            Damping factor: *<value>*; convergence rate: *<value>*.

        (`errno` 6)
            At time *<value>*, error weight *<value>* became zero. Check
            the values of `atol`, `rtol` and `itol` supplied.

        (`errno` 11)
            On re-entry, `ires` = 2, which signals that the integration
            should terminate. `ires` = *<value>* at time *<value>*.

        (`errno` 12)
            A return was forced by setting `imon` = -2, but the
            integration was successful as far as `t`.

        (`errno` 13)
            The requested task has been completed, but it is estimated
            that a small change in `rtol` and `atol` is unlikely to
            produce any change in the computed solution. (This ONLY
            applies when you are NOT operating in one step mode; that
            is, when `itask` != 2 or 5.)

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_imp_revcom as b_ivp_stiff_imp_revcom

    fun_name = 'naginterfaces.library.ode.ivp_stiff_imp_revcom'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'ydot'},
        ydot,
    )
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('inform', 'ysav', 'jacpvt', 'lwsav', 'iwsav', 'rwsav'),
        req_keys=('rwork', 'cwsav'),
    )
    _sdysav = _b_utils._EngineIntScalarType(
        int(_b_utils._EngineFloat64ArrayType.get_element(comm['rwork'], {'fun_name': fun_name, 'entity': 'comm[\'rwork\']'}, (32,), validate=False)), {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'wkjac'},
        wkjac,
    )
    _nwkjac_py = _b_utils._EngineFloat64ArrayType.get_shape(wkjac, {'fun_name': fun_name, 'entity': 'wkjac'}, exp_rank=1)[0]
    _nwkjac = _b_utils._EngineIntScalarType(
        _nwkjac_py, {'fun_name': fun_name, 'entity': '_nwkjac'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'lderiv'},
        lderiv,
    )
    try:
        _return_scalars = b_ivp_stiff_imp_revcom(
            _neq,
            _ldysav,
            t,
            tout,
            y,
            ydot,
            comm,
            _rtol,
            _atol,
            itol,
            _sdysav,
            wkjac,
            _nwkjac,
            imon,
            inln,
            ires,
            irevcm,
            lderiv,
            itask,
            itrace,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _t_py = _return_scalars.t
    _tout_py = _return_scalars.tout
    _imon_py = _return_scalars.imon
    _inln_py = _return_scalars.inln
    _ires_py = _return_scalars.ires
    _irevcm_py = _return_scalars.irevcm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'tout', 'imon', 'inln', 'ires', 'irevcm'],
    )(
        _t_py,
        _tout_py,
        _imon_py,
        _inln_py,
        _ires_py,
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dae_dassl_linalg(neq, ml, mu, comm):
    u"""
    Implicit ordinary differential equations/DAEs, initial value problem
    linear algebra setup function for ``dae_dassl_gen``.

    ``dae_dassl_linalg`` is a setup function which you must call prior
    to ``dae_dassl_gen`` and after a call to ``dae_dassl_setup``, if the
    Jacobian is to be considered as having a banded structure.

    For full information please refer to the NAG Library document for
    d02np

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02npf.html

    Parameters
    ----------
    neq : int
        The number of differential-algebraic equations to be solved.

    ml : int
        m_L, the number of subdiagonals in the band.

    mu : int
        m_U, the number of superdiagonals in the band.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``dae_dassl_setup``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `neq` = *<value>*.

            Constraint: `neq` >= 1.

        (`errno` 2)
            On entry, `ml` = *<value>*.

            Constraint: `ml` >= 0.

        (`errno` 2)
            On entry, `ml` = *<value>* and `neq` = *<value>*.

            Constraint: `ml` <= `neq`-1.

        (`errno` 3)
            On entry, `mu` = *<value>*.

            Constraint: `mu` >= 0.

        (`errno` 3)
            On entry, `mu` = *<value>* and `neq` = *<value>*.

            Constraint: `mu` <= `neq`-1.

        (`errno` 4)
            Either the initialization function has not been called prior
            to the first call of this function or the communication
            array has become corrupted.

        (`errno` 5)
            On entry, licom is too small: licom = *<value>*.
    """

    from ..base.ode import dae_dassl_linalg as b_dae_dassl_linalg

    fun_name = 'naginterfaces.library.ode.dae_dassl_linalg'

    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('icom',),
    )
    _licom_py = _b_utils._EngineIntArrayType.get_shape(comm['icom'], {'fun_name': fun_name, 'entity': 'comm[\'icom\']'}, exp_rank=1, validate=False)[0]
    _licom = _b_utils._EngineIntScalarType(
        _licom_py, {'fun_name': fun_name, 'entity': '_licom'},
    ).to_ctypes()
    b_dae_dassl_linalg(
        neq,
        ml,
        mu,
        comm,
        _licom,
    )

def ivp_stiff_sparjac_enq(comm):
    u"""
    Ordinary differential equations, initial value problem, for use with
    d02m--n functions, sparse Jacobian, enquiry function.

    ``ivp_stiff_sparjac_enq`` is an enquiry function for communicating
    with ``ivp_stiff_exp_revcom`` or ``ivp_stiff_imp_revcom`` when
    supplying columns of a sparse Jacobian matrix.

    For full information please refer to the NAG Library document for
    d02nr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nrf.html

    Parameters
    ----------
    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_exp_revcom`` or ``ivp_stiff_imp_revcom``.

    Returns
    -------
    j : int
        The index j of the column of the Jacobian which is required.

    iplace : int
        Indicates which locations in the array rwork to fill with the
        jth column.

        If `iplace` = 1, the (i,j)th element of the Jacobian must be
        placed in rwork[50+2*ldysav+i-1], otherwise the (i,j)th element
        must be placed in rwork[50+ldysav+i-1].

        If jceval = 'F', in the previous call to
        ``ivp_stiff_sparjac_setup``, `iplace` = 2 always, hence the jth
        column of the Jacobian must be placed in rwork[50+ldysav+i-1],
        for i = 1,2,...,neq.

        rwork, neq and ldysav are arguments of ``ivp_stiff_exp_revcom``
        and ``ivp_stiff_imp_revcom``.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_sparjac_enq as b_ivp_stiff_sparjac_enq

    _return_scalars = b_ivp_stiff_sparjac_enq(
        comm,
    )
    _j_py = _return_scalars.j
    _iplace_py = _return_scalars.iplace
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['j', 'iplace'],
    )(
        _j_py,
        _iplace_py,
    )
    return _return_data

def ivp_stiff_fulljac_setup(neq, neqmax, jceval, nwkjac, comm):
    u"""
    Ordinary differential equations, initial value problem, for use with
    d02m--n functions, full Jacobian, linear algebra set up.

    ``ivp_stiff_fulljac_setup`` is a setup function which must be called
    prior to an integrator in submodule ``ode``, if full matrix linear
    algebra is required.

    For full information please refer to the NAG Library document for
    d02ns

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nsf.html

    Parameters
    ----------
    neq : int
        The number of differential equations.

    neqmax : int
        A bound on the maximum number of differential equations to be
        solved during the integration.

    jceval : str, length 1
        Specifies the technique to be used to compute the Jacobian.

        `jceval` = 'N'

          The Jacobian is to be evaluated numerically by the integrator.
          If this option is used, the actual argument corresponding to
          jac in the call to ``ivp_stiff_exp_fulljac`` or
          ``ivp_stiff_imp_fulljac`` must be specified as None.

        `jceval` = 'A'

          You must supply a (sub)program to evaluate the Jacobian on a
          call to the integrator.

        `jceval` = 'D'

          The default choice is to be made. In this case 'D' is
          interpreted as 'N'.

        Only the first character of the actual argument `jceval` is
        passed to ``ivp_stiff_fulljac_setup``; hence it is permissible
        for the actual argument to be more descriptive 'Numerical',
        'Analytical' or 'Default' on a call to
        ``ivp_stiff_fulljac_setup``.

    nwkjac : int
        The size of the workspace array wkjac, which you are supplying
        to the integrator, as declared in the (sub)program from which
        ``ivp_stiff_fulljac_setup`` is called.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `jceval` = *<value>*.

            Constraint: `jceval` = 'A', 'N' or 'D'.

        (`errno` 1)
            On entry, `nwkjac` = *<value>* and `neqmax` = *<value>*.

            Constraint: `nwkjac` >= (`neqmax`+1)*`neqmax`.

        (`errno` 1)
            On entry, `neq` = *<value>* and `neqmax` = *<value>*.

            Constraint: `neq` <= `neqmax`.

        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

        (`errno` 1)
            On entry, `neq` = *<value>*.

            Constraint: `neq` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    See Also
    --------
    :meth:`naginterfaces.library.examples.ode.ivp_stiff_exp_fulljac_ex.main`
    """

    from ..base.ode import ivp_stiff_fulljac_setup as b_ivp_stiff_fulljac_setup

    b_ivp_stiff_fulljac_setup(
        neq,
        neqmax,
        jceval,
        nwkjac,
        comm,
    )

def ivp_stiff_bandjac_setup(neq, neqmax, jceval, ml, mu, nwkjac, njcpvt, comm):
    u"""
    Ordinary differential equations, initial value problem, for use with
    d02m--n functions, banded Jacobian, linear algebra set up.

    ``ivp_stiff_bandjac_setup`` is a setup function which you must call
    prior to an integrator in submodule ``ode``, if banded matrix linear
    algebra is required.

    For full information please refer to the NAG Library document for
    d02nt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ntf.html

    Parameters
    ----------
    neq : int
        The number of differential equations.

    neqmax : int
        A bound on the maximum number of differential equations to be
        solved during the integration.

    jceval : str, length 1
        Specifies the technique to be used to compute the Jacobian as
        follows:

        `jceval` = 'N'

          The Jacobian is to be evaluated numerically by the integrator.
          If this option is used, the actual argument corresponding to
          jac in the call to ``ivp_stiff_exp_bandjac`` or
          ``ivp_stiff_imp_bandjac`` must be specified as None.

        `jceval` = 'A'

          You must supply a (sub)program to evaluate the Jacobian on a
          call to the integrator.

        `jceval` = 'D'

          The default choice is to be made. In this case 'D' is
          interpreted as 'N'.

        Only the first character of the actual argument `jceval` is
        passed to ``ivp_stiff_bandjac_setup``; hence it is permissible
        for the actual argument to be more descriptive, e.g.,
        'Numerical', 'Analytical' or 'Default', on a call to
        ``ivp_stiff_bandjac_setup``.

    ml : int
        m_L, the number of subdiagonals in the band.

    mu : int
        m_U, the number of superdiagonals in the band.

    nwkjac : int
        The size of the workspace array wkjac, which you are supplying
        to the integrator, as declared in the (sub)program from which
        ``ivp_stiff_bandjac_setup`` is called.

    njcpvt : int
        The size of the workspace array jacpvt, which you are supplying
        to the integrator, as declared in the (sub)program from which
        ``ivp_stiff_bandjac_setup`` is called.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `jceval` = *<value>*.

            Constraint: `jceval` = 'A', 'N' or 'D'.

        (`errno` 1)
            On entry, `mu` = *<value>* and `neq` = *<value>*.

            Constraint: `mu` <= `neq`-1.

        (`errno` 1)
            On entry, `mu` = *<value>*.

            Constraint: `mu` >= 0.

        (`errno` 1)
            On entry, `ml` = *<value>* and `neq` = *<value>*.

            Constraint: `ml` <= `neq`-1.

        (`errno` 1)
            On entry, `ml` = *<value>*.

            Constraint: `ml` >= 0.

        (`errno` 1)
            On entry, `nwkjac` = *<value>*, `ml` = *<value>*, `mu` =
            *<value>* and `neqmax` = *<value>*.

            Constraint: `nwkjac` >= (2*`ml`+`mu`+1)*`neqmax`.

        (`errno` 1)
            On entry, `njcpvt` = *<value>* and `neqmax` = *<value>*.

            Constraint: `njcpvt` >= `neqmax`.

        (`errno` 1)
            On entry, `neq` = *<value>* and `neqmax` = *<value>*.

            Constraint: `neq` <= `neqmax`.

        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

        (`errno` 1)
            On entry, `neq` = *<value>*.

            Constraint: `neq` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_bandjac_setup as b_ivp_stiff_bandjac_setup

    b_ivp_stiff_bandjac_setup(
        neq,
        neqmax,
        jceval,
        ml,
        mu,
        nwkjac,
        njcpvt,
        comm,
    )

def ivp_stiff_sparjac_setup(neq, neqmax, jceval, nwkjac, ia, ja, comm, njcpvt, sens, u, eta, lblock, isplit=73):
    u"""
    Ordinary differential equations, initial value problem, for use with
    d02m--n functions, sparse Jacobian, linear algebra set up.

    ``ivp_stiff_sparjac_setup`` is a setup function which must be called
    prior to an integrator in submodule ``ode``, if sparse matrix linear
    algebra is required.

    For full information please refer to the NAG Library document for
    d02nu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nuf.html

    Parameters
    ----------
    neq : int
        The number of differential equations.

    neqmax : int
        A bound on the maximum number of differential equations to be
        solved during the integration.

    jceval : str, length 1
        Specifies the technique to be used to compute the Jacobian.

        `jceval` = 'N'

          The sparsity structure and the value of the Jacobian are to be
          determined numerically by the integrator.

        `jceval` = 'S'

          The sparsity structure of the Jacobian is supplied in the
          arrays `ia` and `ja` but its value is to be determined
          numerically. This is the recommended mode of operation unless
          it is a simple matter to supply the Jacobian.

        `jceval` = 'A'

          The Jacobian will be evaluated by calls to jac. The sparsity
          structure will be estimated by calls to jac; that is, no
          explicit sparsity structure need be supplied in the arrays
          `ia` and `ja`.

        `jceval` = 'F'

          The sparsity structure of the Jacobian is supplied in `ia` and
          `ja`, and its value will be determined by calls to jac. This
          is the recommended mode of operation if the jac is simple to
          form.

        `jceval` = 'D'

          The default choice is to be made. In this case 'D' is
          interpreted as 'S'.

        If the sparsity structure is supplied in arrays `ia` and `ja`,
        any evidence from the numerical or analytical formation of the
        Jacobian that this structure is not correct, is ignored.

        Only the first character of the actual argument `jceval` is
        passed to ``ivp_stiff_sparjac_setup``; hence it is permissible
        for the actual argument to be more descriptive, e.g.,
        'Numerical', 'Structural', 'Analytical', 'Full information' or
        'Default' in a call to ``ivp_stiff_sparjac_setup``.

        If the option `jceval` = 'N', 'S' or 'D' is used then the actual
        argument corresponding to jac in the call to
        ``ivp_stiff_exp_sparjac`` or ``ivp_stiff_imp_sparjac`` must be
        specified as None.

        If integration is to be performed by reverse communication
        (``ivp_stiff_exp_revcom`` or ``ivp_stiff_imp_revcom``) then
        `jceval` should be set to either 'N' or 'A'.

        In this case `ia` and `ja` are not used and their lengths may be
        set to 1.

    nwkjac : int
        The size of the array wkjac, which you are supplying to the
        integrator, as declared in the (sub)program from which
        ``ivp_stiff_sparjac_setup`` is called.

    ia : int, array-like, shape (nia)
        If `jceval` = 'S', 'F' or 'D', `ia` must contain details of the
        sparsity pattern to be used for the Jacobian. See `ja`.

        `ia` is not used if `jceval` = 'N' or 'A'.

    ja : int, array-like, shape (nja)
        If `jceval` = 'S', 'F' or 'D', `ja` must contain details of the
        sparsity pattern to be used for the Jacobian. `ja` contains the
        row indices where nonzero elements occur, reading in column-wise
        order, and `ia` contains the starting locations in `ja` of the
        descriptions of columns 1,2,...,`neq` in that order, with
        `ia`[0] = 1. Thus for each column index j = 1,2,...,`neq`, the
        values of the row index i in column j where a nonzero element
        may occur are given by

            i = `ja`[k]

        where `ia`[j-1] <= k < `ia`[j].

        Thus the total number of nonzeros, nelement, must be
        `ia`[`neq`]-1.

        For example, for the following matrix

            [table omitted]

        where x represents nonzero elements (13 in all) the arrays `ia`
        and `ja` should be

            [table omitted]

        `ja` is not used if `jceval` = 'N' or 'A'.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    njcpvt : int
        The length of the array `comm`['jacpvt'], which you are
        supplying to the integrator, as dimensioned in the sub(program)
        from which ``ivp_stiff_sparjac_setup`` is called.

    sens : float
        A threshold argument used to determine whether or not a matrix
        element is zero; when `sens` is set to 0.0 on entry, the
        function will use `sens` = 100.0*machine precision. Otherwise
        the absolute value of `sens` is used.

    u : float
        Should have a value between 0.0 and 0.9999. Otherwise a default
        value of 0.1 is used. When the sparsity pattern has been
        evaluated, the first Jacobian computed is decomposed with `u`
        governing the choice of pivots; subsequent Jacobian
        decompositions use the same pattern of decomposition until the
        sparsity pattern is re-evaluated. When searching a row for a
        pivot, any element is excluded from the search which is less
        than `u` times the largest of those elements in the row
        available as pivots. Thus decreasing `u` biases the algorithm
        towards maintaining sparsity at the expense of numerical
        stability.

    eta : float
        A relative pivot threshold, below which on subsequent
        decompositions (as described under `u`), an internal error is
        provoked.

        `eta` > 1.0

          No check on pivot size is made.

        `eta` <= 0.0

          The default value `eta` = 1.0e-4 is used.

    lblock : bool
        Indicates if preordering is used before decomposition.

        If `lblock` = True, on entry, the Jacobian matrix is preordered
        to block lower triangular form before a decomposition is
        performed (this is the recommended mode).

        If you know the structure of the Jacobian to be irreducible,
        that is not permutable to block lower triangular form, you
        should set `lblock` = False.

        For example, a Jacobian arising from using the method of lines
        for parabolic partial differential equations would normally be
        irreducible. (See the specification of
        ``ivp_stiff_sparjac_diag`` for optional output concerning
        `lblock`.)

    isplit : int, optional
        This argument is used for splitting the integer workspace
        `comm`['jacpvt'] to effect an efficient decomposition. It must
        satisfy 1 <= `isplit` <= 99. If `isplit` lies outside this range
        on entry, a default value of 73 is used. An appropriate value
        for `isplit` for subsequent runs on similar problems is
        available via the optional output ``ivp_stiff_sparjac_diag``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nwkjac` = *<value>* and `ia`[`neq`]-1+2*`neq` =
            *<value>*.

            Constraint: `nwkjac` >= `ia`[`neq`]-1+2*`neq`.

        (`errno` 1)
            On entry, `ja` defines duplicate elements in row *<value>*
            and column *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `ja`[i] = *<value>* and `neq` =
            *<value>*.

            Constraint: 1 <= `ja`[i] <= `neq` for all i.

        (`errno` 1)
            On entry, i = *<value>*, `ia`[i] = *<value>*, i-1 =
            *<value>*, `ia`[i-1] = *<value>*, `neq` = *<value>*.

            Constraint: `ia`[*<value>*]-`ia`[*<value>*] <= `neq`.

        (`errno` 1)
            On entry, `ia`[*<value>*] = *<value>* and `ia`[*<value>*] =
            *<value>*.

            Constraint: `ia`[*<value>*] >= `ia`[*<value>*].

        (`errno` 1)
            On entry, `ia`[0] = *<value>*.

            Constraint: `ia`[0] = 1.

        (`errno` 1)
            On entry, `njcpvt` = *<value>* and
            3*(`ia`[`neq`]-1)+14*`neq`+1 = *<value>*.

            Constraint: `njcpvt` >= 3*(`ia`[`neq`]-1)+14*`neq`+1.

        (`errno` 1)
            On entry, nja = *<value>* and `ia`[`neq`]-1 = *<value>*.

            Constraint: nja >= `ia`[`neq`]-1.

        (`errno` 1)
            On entry, `ia`[`neq`]-1 = *<value>* and `neq` = *<value>*.

            Constraint: `ia`[`neq`]-1 >= `neq`.

        (`errno` 1)
            On entry, nia = *<value>* and `neq`+1 = *<value>*.

            Constraint: nia >= `neq`+1.

        (`errno` 1)
            On entry, `jceval` = *<value>*.

            Constraint: `jceval` = 'A', 'N', 'S', 'F' or 'D'.

        (`errno` 1)
            On entry, `neq` = *<value>* and `neqmax` = *<value>*.

            Constraint: `neq` <= `neqmax`.

        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

        (`errno` 1)
            On entry, `neq` = *<value>*.

            Constraint: `neq` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_sparjac_setup as b_ivp_stiff_sparjac_setup

    fun_name = 'naginterfaces.library.ode.ivp_stiff_sparjac_setup'

    _ia = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ia'},
        ia,
        copy=False,
    )
    _nia_py = _b_utils._EngineIntArrayType.get_shape(_ia, {'fun_name': fun_name, 'entity': 'ia'}, exp_rank=1)[0]
    _nia = _b_utils._EngineIntScalarType(
        _nia_py, {'fun_name': fun_name, 'entity': '_nia'},
    ).to_ctypes()
    _ja = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ja'},
        ja,
        copy=False,
    )
    _nja_py = _b_utils._EngineIntArrayType.get_shape(_ja, {'fun_name': fun_name, 'entity': 'ja'}, exp_rank=1)[0]
    _nja = _b_utils._EngineIntScalarType(
        _nja_py, {'fun_name': fun_name, 'entity': '_nja'},
    ).to_ctypes()
    b_ivp_stiff_sparjac_setup(
        neq,
        neqmax,
        jceval,
        nwkjac,
        _ia,
        _nia,
        _ja,
        _nja,
        comm,
        njcpvt,
        sens,
        u,
        eta,
        lblock,
        isplit,
    )

def ivp_stiff_bdf(neqmax, sdysav, maxord, method, petzld, con, tcrit, hmin, hmax, h0, maxstp, mxhnil, norm):
    u"""
    Ordinary differential equations, initial value problem, backward
    differentiation formulae method, setup for d02m--n functions.

    ``ivp_stiff_bdf`` is a setup function which must be called prior to
    linear algebra setup functions and integrators from the SPRINT suite
    of functions, if Backward Differentiation Formulae (BDF) are to be
    used.

    For full information please refer to the NAG Library document for
    d02nv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nvf.html

    Parameters
    ----------
    neqmax : int
        A bound on the maximum number of differential equations to be
        solved.

    sdysav : int
        The second dimension of the array ysav that will be supplied to
        the integrator, as declared in the (sub)program from which the
        integrator is called.

    maxord : int
        The maximum order to be used for the BDF method.

    method : str, length 1
        Specifies the method to be used to solve the system of nonlinear
        equations arising on each step of the BDF code.

        `method` = 'N'

          A modified Newton iteration is used.

        `method` = 'F'

          Functional iteration is used.

        `method` = 'D'

          A modified Newton iteration is used.

        Note: a linear algebra setup function must be called even when
        using functional iteration, since if difficulty is encountered a
        switch is made to a modified Newton method.

        Only the first character of the actual argument `method` is
        passed to ``ivp_stiff_bdf``; hence it is permissible for the
        actual argument to be more descriptive e.g., 'Newton',
        'Functional iteration' or 'Default' in a call to
        ``ivp_stiff_bdf``.

    petzld : bool
        Specifies whether the Petzold local error test is to be used. If
        `petzld` is set to True on entry, the Petzold local error test
        is used, otherwise a conventional test is used. The Petzold test
        results in extra overhead cost but is more stable and reliable
        for differential/algebraic equations.

    con : float, array-like, shape (6)
        Values to be used to control step size choice during
        integration. If any `con`[i-1] = 0.0 on entry, it is replaced by
        its default value described below. In most cases this is the
        recommended setting.

        `con`[0], `con`[1], and `con`[2] are factors used to bound step
        size changes.

        If the current step size h fails, the modulus of the next step
        size is bounded by `con`[0]*|h|.

        The default value of `con`[0] is 2.0.

        Note that the new step size may be used with a method of
        different order to the failed step.

        If the initial step size is h, the modulus of the step size on
        the second step is bounded by `con`[2]*|h|.

        At any other stage in the integration, if the current step size
        is h, the modulus of the next step size is bounded by
        `con`[1]*|h|.

        The default values are 10.0 for `con`[1] and 1000.0 for
        `con`[2].

        `con`[3], `con`[4] and `con`[5] are 'tuning' constants used in
        determining the next order and step size.

        They are used to scale the error estimates used in determining
        whether to keep the same order of the BDF method, decrease the
        order or increase the order respectively.

        The larger the value of `con`[i-1], for i = 4,5,...,6, the less
        likely the choice of the corresponding order.

        The default values are: `con`[3] = 1.2, `con`[4] = 1.3, `con`[5]
        = 1.4.

    tcrit : float
        A point beyond which integration must not be attempted. The use
        of `tcrit` is described under the argument itask in the
        specification for the integrator (e.g., see
        ``ivp_stiff_exp_fulljac``). A value, 0.0 say, must be specified
        even if itask subsequently specifies that `tcrit` will not be
        used.

    hmin : float
        The minimum absolute step size to be allowed. Set `hmin` = 0.0
        if this option is not required.

    hmax : float
        The maximum absolute step size to be allowed. Set `hmax` = 0.0
        if this option is not required.

    h0 : float
        The step size to be attempted on the first step. Set `h0` = 0.0
        if the initial step size is calculated internally.

    maxstp : int
        The maximum number of steps to be attempted during one call to
        the integrator after which it will return with `errno` = 2. Set
        `maxstp` = 0 if no limit is to be imposed.

    mxhnil : int
        The maximum number of warnings printed (if itrace >= 0) per
        problem when t+h = t on a step (h = current step size). If
        `mxhnil` <= 0, a default value of 10 is assumed.

    norm : str, length 1
        Indicates the type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L2 norm.

        `norm` = 'D'

          Is the same as `norm` = 'A'.

        If vnorm denotes the norm of the vector v of length n, for the
        averaged L2 norm

            vnorm = sqrt(1/nsum_i = 1^n(v_i/w_i)^2),

        while for the maximum norm

            vnorm = max_i|v_i/w_i|.

        If you wish to weight the maximum norm or the L2 norm, rtol and
        atol should be scaled appropriately on input to the integrator
        (see under itol in the specification of the integrator for the
        formulation of the weight vector w_i from rtol and atol, e.g.,
        see ``ivp_stiff_exp_fulljac``).

        Only the first character to the actual argument `norm` is passed
        to ``ivp_stiff_bdf``; hence it is permissible for the actual
        argument to be more descriptive e.g., 'Maximum', 'Average L2' or
        'Default' in a call to ``ivp_stiff_bdf``.

    Returns
    -------
    con : float, ndarray, shape (6)
        The values actually used by ``ivp_stiff_bdf``.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `con`[1] = *<value>* and `con`[2] = *<value>*.

            Constraint: `con`[1] <= `con`[2].

        (`errno` 1)
            On entry, `con`[0] = *<value>* and `con`[1] = *<value>*.

            Constraint: `con`[0] <= `con`[1].

        (`errno` 1)
            On entry, `con`[*<value>*] = *<value>* was less than 1.0.

        (`errno` 1)
            On entry, `con`[*<value>*] = *<value>* was less than 0.0.

        (`errno` 1)
            On entry, `method` = *<value>* and `maxord` = *<value>*.

            Constraint: `maxord` <= *<value>*, the maximum allowed order
            for the method define by `method`.

        (`errno` 1)
            On entry, `maxord` = *<value>*.

            Constraint: `maxord` >= 1.

        (`errno` 1)
            On entry, `sdysav` = *<value>* and `maxord` = *<value>*.

            Constraint: `sdysav` > `maxord`.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 'M', 'A' or 'D'.

        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'N', 'F' or 'D'.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    See Also
    --------
    :meth:`naginterfaces.library.examples.ode.ivp_stiff_exp_fulljac_ex.main`
    """

    from ..base.ode import ivp_stiff_bdf as b_ivp_stiff_bdf

    fun_name = 'naginterfaces.library.ode.ivp_stiff_bdf'

    _con = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'con'},
        con,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_ivp_stiff_bdf(
        neqmax,
        sdysav,
        maxord,
        method,
        petzld,
        _con,
        tcrit,
        hmin,
        hmax,
        h0,
        maxstp,
        mxhnil,
        norm,
    )
    _con_py = _con
    _comm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['con', 'comm'],
    )(
        _con_py,
        _comm_py,
    )
    return _return_data

def ivp_stiff_blend(neqmax, sdysav, maxord, con, tcrit, hmin, hmax, h0, maxstp, mxhnil, norm):
    u"""
    Ordinary differential equations, initial value problem, Blend
    method, setup for d02m--n functions.

    ``ivp_stiff_blend`` is a setup function which must be called prior
    to linear algebra setup functions and integrators from the SPRINT
    suite of functions, if the BLEND formulae are to be used.

    For full information please refer to the NAG Library document for
    d02nw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nwf.html

    Parameters
    ----------
    neqmax : int
        A bound on the maximum number of differential equations to be
        solved.

    sdysav : int
        The second dimension of the array ysav that will be supplied to
        the integrator, as declared in the (sub)program from which the
        integrator is called (e.g., see ``ivp_stiff_exp_fulljac``).

    maxord : int
        The maximum order to be used for the BLEND method.

    con : float, array-like, shape (6)
        Values to be used to control step size choice during
        integration. If any `con`[i-1] = 0.0 on entry, it is replaced by
        its default value described below. In most cases this is the
        recommended setting.

        `con`[0], `con`[1], and `con`[2] are factors used to bound step
        size changes.

        If the current step size h fails, the modulus of the next step
        size is bounded by `con`[0]*|h|.

        The default value of `con`[0] is 2.0.

        Note that the new step size may be used with a method of
        different order to the failed step.

        If the initial step size is h, the modulus of the step size on
        the second step is bounded by `con`[2]*|h|.

        At any other stage in the integration, if the current step size
        is h, the modulus of the next step size is bounded by
        `con`[1]*|h|.

        The default values are 10.0 for `con`[1] and 1000.0 for
        `con`[2].

        `con`[3], `con`[4] and `con`[5] are 'tuning' constants used in
        determining the next order and step size.

        They are used to scale the error estimates used in determining
        whether to keep the same order of the BLEND method, decrease the
        order or increase the order respectively.

        The larger the value of `con`[i-1], for i = 4,5,...,6, the less
        likely the choice of the corresponding order.

        The default values are: `con`[3] = 1.2, `con`[4] = 1.3, `con`[5]
        = 1.4.

    tcrit : float
        A point beyond which integration must not be attempted. The use
        of `tcrit` is described under the argument itask in the
        specification for the integrator (e.g., see
        ``ivp_stiff_exp_fulljac``). A value, 0.0 say, must be specified
        even if itask subsequently specifies that `tcrit` will not be
        used.

    hmin : float
        The minimum absolute step size to be allowed. Set `hmin` = 0.0
        if this option is not required.

    hmax : float
        The maximum absolute step size to be allowed. Set `hmax` = 0.0
        if this option is not required.

    h0 : float
        The step size to be attempted on the first step. Set `h0` = 0.0
        if the initial step size is calculated internally.

    maxstp : int
        The maximum number of steps to be attempted during one call to
        the integrator after which it will return with `errno` = 2. Set
        `maxstp` = 0 if no limit is to be imposed.

    mxhnil : int
        The maximum number of warnings printed (if itrace >= 0, e.g.,
        see ``ivp_stiff_exp_fulljac``) per problem when t+h = t on a
        step (h = current step size). If `mxhnil` <= 0, a default value
        of 10 is assumed.

    norm : str, length 1
        Indicates the type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L2 norm.

        `norm` = 'D'

          Is the same as 'A'.

        If vnorm denotes the norm of the vector v of length n, for the
        averaged L2 norm

            vnorm = sqrt(1/nsum_i = 1^n(v_i/w_i)^2),

        while for the maximum norm

            vnorm = max_i|v_i/w_i|.

        If you wish to weight the maximum norm or the L2 norm, rtol and
        atol should be scaled appropriately on input to the integrator
        (see under itol in the specification of the integrator for the
        formulation of the weight vector w_i from rtol and atol, e.g.,
        ``ivp_stiff_exp_fulljac``).

        Only the first character of the actual argument `norm` is passed
        to ``ivp_stiff_blend``; hence it is permissible for the actual
        argument to be more descriptive e.g., 'Maximum', 'Average L2' or
        'Default' in a call to ``ivp_stiff_blend``.

    Returns
    -------
    con : float, ndarray, shape (6)
        The values actually used by ``ivp_stiff_blend``.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `con`[1] = *<value>* and `con`[2] = *<value>*.

            Constraint: `con`[1] <= `con`[2].

        (`errno` 1)
            On entry, `maxord` = *<value>* and `sdysav` = *<value>*.

            Constraint: `sdysav` >= `maxord`+3.

        (`errno` 1)
            On entry, `maxord` = *<value>*.

            Constraint: 0 < `maxord` <= 11.

        (`errno` 1)
            On entry, `con`[0] = *<value>* and `con`[1] = *<value>*.

            Constraint: `con`[0] <= `con`[1].

        (`errno` 1)
            On entry, `con`[*<value>*] = *<value>* was less than 1.0.

        (`errno` 1)
            On entry, `con`[*<value>*] = *<value>* was less than 0.0.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 'M', 'A' or 'D'.

        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_blend as b_ivp_stiff_blend

    fun_name = 'naginterfaces.library.ode.ivp_stiff_blend'

    _con = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'con'},
        con,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_ivp_stiff_blend(
        neqmax,
        sdysav,
        maxord,
        _con,
        tcrit,
        hmin,
        hmax,
        h0,
        maxstp,
        mxhnil,
        norm,
    )
    _con_py = _con
    _comm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['con', 'comm'],
    )(
        _con_py,
        _comm_py,
    )
    return _return_data

def ivp_stiff_sparjac_diag(icall, lblock, comm):
    u"""
    Ordinary differential equations, initial value problem, sparse
    Jacobian, linear algebra diagnostics, for use with d02m--n
    functions.

    ``ivp_stiff_sparjac_diag`` is an optional output function which you
    may call, on exit from an integrator in submodule ``ode``, if sparse
    matrix linear algebra has been selected.

    For full information please refer to the NAG Library document for
    d02nx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nxf.html

    Parameters
    ----------
    icall : int
        Indicates whether or not all output arguments have been set
        during the call to the integrator. If so, that is, if the
        integrator returned with the function exits successfully or
        `errno` = 12, then `icall` must be set to 0. Otherwise `icall`
        must be set to 1, indicating that integration did not take place
        due to lack of space in arrays wkjac and jacpvt, and only
        `liwreq`, `liwusd`, `lrwreq`, `lrwusd` have been set.

    lblock : bool
        The value used for the argument `lblock` when calling
        ``ivp_stiff_sparjac_setup``.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_exp_sparjac``, ``ivp_stiff_imp_sparjac`` or
        ``ivp_stiff_imp_revcom``.

    Returns
    -------
    liwreq : int
        The length of the int workspace jacpvt reserved for the sparse
        matrix functions.

    liwusd : int
        The length of the int workspace jacpvt actually used by the
        sparse matrix functions.

    lrwreq : int
        The length of the double workspace wkjac reserved for the sparse
        matrix functions.

    lrwusd : int
        The length of the double workspace wkjac actually used by the
        sparse matrix functions.

    nlu : int
        The number of LU decompositions done during the integration.

    nnz : int
        The number of nonzeros in the Jacobian.

    ngp : int
        The number of fcn or resid calls needed to form the Jacobian.

    isplit : int
        An appropriate value for the argument `isplit` when calling
        ``ivp_stiff_sparjac_setup`` for subsequent runs of similar
        problems.

    igrow : int
        An estimate of the growth of the elements encountered during the
        last LU decomposition performed. If the actual estimate exceeds
        the largest possible integer value for the machine being used
        (see ``machine.integer_max``) `igrow` is set to the value
        returned by ``machine.integer_max``.

    nblock : int
        If `lblock` = True, `nblock` contains the number of diagonal
        blocks in the Jacobian matrix permuted to block lower triangular
        form. If `nblock` = 1 then on subsequent runs of a similar
        problem `lblock` should be set to False in the call to
        ``ivp_stiff_sparjac_setup``.

        If `lblock` = False, `nblock` = 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_sparjac_diag as b_ivp_stiff_sparjac_diag

    _return_scalars = b_ivp_stiff_sparjac_diag(
        icall,
        lblock,
        comm,
    )
    _liwreq_py = _return_scalars.liwreq
    _liwusd_py = _return_scalars.liwusd
    _lrwreq_py = _return_scalars.lrwreq
    _lrwusd_py = _return_scalars.lrwusd
    _nlu_py = _return_scalars.nlu
    _nnz_py = _return_scalars.nnz
    _ngp_py = _return_scalars.ngp
    _isplit_py = _return_scalars.isplit
    _igrow_py = _return_scalars.igrow
    _nblock_py = _return_scalars.nblock
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['liwreq', 'liwusd', 'lrwreq', 'lrwusd', 'nlu', 'nnz', 'ngp', 'isplit', 'igrow', 'nblock'],
    )(
        _liwreq_py,
        _liwusd_py,
        _lrwreq_py,
        _lrwusd_py,
        _nlu_py,
        _nnz_py,
        _ngp_py,
        _isplit_py,
        _igrow_py,
        _nblock_py,
    )
    return _return_data

def ivp_stiff_integ_diag(neq, neqmax, comm):
    u"""
    Ordinary differential equations, initial value problem, integrator
    diagnostics, for use with d02m--n functions.

    ``ivp_stiff_integ_diag`` is a diagnostic function which you may call
    either after any user-specified exit or after a mid-integration
    error exit from any of those integrators in submodule ``ode`` that
    use methods set up by calls to ``ivp_stiff_dassl``,
    ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    For full information please refer to the NAG Library document for
    d02ny

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nyf.html

    Parameters
    ----------
    neq : int
        The value used for the argument `neq` when calling the
        integrator.

    neqmax : int
        The value used for the argument `neqmax` when calling the
        integrator.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_exp_fulljac``, ``ivp_stiff_exp_bandjac``,
        ``ivp_stiff_exp_sparjac``, ``ivp_stiff_imp_fulljac``,
        ``ivp_stiff_imp_bandjac``, ``ivp_stiff_imp_sparjac``,
        ``ivp_stiff_exp_revcom`` or ``ivp_stiff_imp_revcom``.

    Returns
    -------
    hu : float
        The last successful step size.

    h : float
        The proposed next step size for continuing the integration.

    tcur : float
        t, the value of the independent variable which the integrator
        has actually reached. `tcur` will always be at least as far as
        the output value of the argument t in the direction of
        integration, but may be further (if overshooting and
        interpolation at tout was specified, e.g., see
        ``ivp_stiff_exp_fulljac``).

    tolsf : float
        A tolerance scale factor, `tolsf` >= 1.0, which is computed when
        a request for too much accuracy is detected by the integrator
        (indicated by a return with `errno` = 3 or 14). If itol is left
        unaltered but rtol and atol are uniformly scaled up by a factor
        of `tolsf` the next call to the integrator is deemed likely to
        succeed.

    nst : int
        The number of steps taken in the integration so far.

    nre : int
        The number of function or residual evaluations (fcn (e.g., see
        ``ivp_stiff_exp_fulljac``) or resid (e.g., see
        ``ivp_stiff_imp_fulljac``) calls) used in the integration so
        far.

    nje : int
        The number of Jacobian evaluations used in the integration so
        far. This equals the number of matrix LU decompositions.

    nqu : int
        The order of the method last used (successfully) in the
        integration.

    nq : int
        The proposed order of the method for continuing the integration.

    niter : int
        The number of iterations performed in the integration so far by
        the nonlinear equation solver.

    imxer : int
        The index of the component of largest magnitude in the weighted
        local error vector (e_i/w_i), for i = 1,2,...,`neq`.

    algequ : bool, ndarray, shape (`neq`)
        `algequ`[i-1] = True if the ith equation integrated was detected
        to be algebraic, otherwise `algequ`[i-1] = False. Note that when
        the integrators for explicit equations are being used, then
        `algequ`[i-1] = False, for i = 1,2,...,`neq`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `neq` = *<value>* and `neqmax` = *<value>*.

            Constraint: `neq` <= `neqmax`.

        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

        (`errno` 1)
            On entry, `neq` = *<value>*.

            Constraint: `neq` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_integ_diag as b_ivp_stiff_integ_diag

    fun_name = 'naginterfaces.library.ode.ivp_stiff_integ_diag'

    _neq = _b_utils._EngineIntScalarType(
        neq, {'fun_name': fun_name, 'entity': 'neq'},
    ).to_ctypes()
    _neq_py = _b_utils._EngineIntScalarType.to_py(_neq)
    _algequ = _b_utils._EngineLogicalArrayType.empty_ndarray(
        (_neq_py,),
    )
    _return_scalars = b_ivp_stiff_integ_diag(
        _neq,
        neqmax,
        comm,
        _algequ,
    )
    _hu_py = _return_scalars.hu
    _h_py = _return_scalars.h
    _tcur_py = _return_scalars.tcur
    _tolsf_py = _return_scalars.tolsf
    _nst_py = _return_scalars.nst
    _nre_py = _return_scalars.nre
    _nje_py = _return_scalars.nje
    _nqu_py = _return_scalars.nqu
    _nq_py = _return_scalars.nq
    _niter_py = _return_scalars.niter
    _imxer_py = _return_scalars.imxer
    _algequ_py = _algequ
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['hu', 'h', 'tcur', 'tolsf', 'nst', 'nre', 'nje', 'nqu', 'nq', 'niter', 'imxer', 'algequ'],
    )(
        _hu_py,
        _h_py,
        _tcur_py,
        _tolsf_py,
        _nst_py,
        _nre_py,
        _nje_py,
        _nqu_py,
        _nq_py,
        _niter_py,
        _imxer_py,
        _algequ_py,
    )
    return _return_data

def ivp_stiff_contin(neqmax, tcrit, h, hmin, hmax, maxstp, mxhnil, comm):
    u"""
    Ordinary differential equations, initial value problem, setup for
    continuation calls to integrator, for use with d02m--n functions.

    ``ivp_stiff_contin`` is a setup function which must be called, if
    optional inputs need resetting, prior to a continuation call to any
    of those integrators in submodule ``ode`` that use methods set up by
    calls to ``ivp_stiff_dassl``, ``ivp_stiff_bdf`` or
    ``ivp_stiff_blend``.

    For full information please refer to the NAG Library document for
    d02nz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02nzf.html

    Parameters
    ----------
    neqmax : int
        The value used for the argument `neqmax` when calling the
        integrator.

    tcrit : float
        A point beyond which integration must not be attempted. The use
        of `tcrit` is described under the argument itask in the
        specification for the integrator (e.g., see
        ``ivp_stiff_exp_fulljac``). A value, 0.0 say, must be specified
        even if itask subsequently specifies that `tcrit` will not be
        used.

    h : float
        The next step size to be attempted. Set `h` = 0.0 if the current
        value of `h` is not to be changed.

    hmin : float
        The minimum absolute step size to be allowed. Set `hmin` = 0.0
        if this option is not required. Set `hmin` < 0.0 if the current
        value of `hmin` is not to be changed.

    hmax : float
        The maximum absolute step size to be allowed. Set `hmax` = 0.0
        if this option is not required. Set `hmax` < 0.0 if the current
        value of `hmax` is not to be changed.

    maxstp : int
        The maximum number of steps to be attempted during one call to
        the integrator after which it will return with `errno` = 2 (see
        ``ivp_stiff_exp_bandjac``). Set `maxstp` = 0 if this option is
        not required. Set `maxstp` < 0 if the current value of `maxstp`
        is not to be changed.

    mxhnil : int
        The maximum number of warnings printed (if itrace >= 0, e.g.,
        see ``ivp_stiff_exp_fulljac``) per problem when t+h = t on a
        step (h = current step size). If `mxhnil` <= 0, a default value
        of 10 is assumed.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_stiff_dassl``, ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `neqmax` = *<value>*.

            Constraint: `neqmax` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_contin as b_ivp_stiff_contin

    b_ivp_stiff_contin(
        neqmax,
        tcrit,
        h,
        hmin,
        hmax,
        maxstp,
        mxhnil,
        comm,
    )

def ivp_rkts_range(f, twant, ygot, ymax, comm, data=None):
    u"""
    Ordinary differential equations, initial value problem, Runge--Kutta
    method, integration over range with output.

    ``ivp_rkts_range`` solves an initial value problem for a first-order
    system of ordinary differential equations using Runge--Kutta
    methods.

    For full information please refer to the NAG Library document for
    d02pe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02pef.html

    Parameters
    ----------
    f : callable yp = f(t, y, data=None)
        `f` must evaluate the functions f_i (that is the first
        derivatives y'_i) for given values of the arguments t, y_i.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (n)
            The current values of the dependent variables, y_i, for i =
            1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        yp : float, array-like, shape (n)
            The values of f_i, for i = 1,2,...,n.

    twant : float
        t, the next value of the independent variable where a solution
        is desired.

    ygot : float, array-like, shape (n)
        On the first call to ``ivp_rkts_range``, `ygot` need not be set.
        On all subsequent calls `ygot` must remain unchanged.

    ymax : float, array-like, shape (n)
        On the first call to ``ivp_rkts_range``, `ymax` need not be set.
        On all subsequent calls `ymax` must remain unchanged.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    tgot : float
        t, the value of the independent variable at which a solution has
        been computed. On successful exit with no exception or warning
        is raised, `tgot` will equal `twant`. On exit with `errno` > 1,
        a solution has still been computed at the value of `tgot` but in
        general `tgot` will not equal `twant`.

    ygot : float, ndarray, shape (n)
        An approximation to the true solution at the value of `tgot`. At
        each step of the integration to `tgot`, the local error has been
        controlled as specified in ``ivp_rkts_setup``. The local error
        has still been controlled even when `tgot` != `twant`, that is
        after a return with `errno` > 1.

    ypgot : float, ndarray, shape (n)
        An approximation to the first derivative of the true solution at
        `tgot`.

    ymax : float, ndarray, shape (n)
        `ymax`[i-1] contains the largest value of |y_i| computed at any
        step in the integration so far.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted.

        (`errno` 1)
            On entry, n = *<value>*, but the value passed to the setup
            function was n = *<value>*.

        (`errno` 1)
            On entry, the communication arrays have become corrupted, or
            a catastrophic error has already been detected elsewhere.
            You cannot continue integrating the problem.

        (`errno` 1)
            `twant` is too close to the last value of `tgot` (tstart on
            setup).

            When using the method of order 8 at setup, these must differ
            by at least *<value>*. Their absolute difference is
            *<value>*.

        (`errno` 1)
            `twant` lies beyond tend (setup) in the direction of
            integration.

            `twant` = *<value>* and tend = *<value>*.

        (`errno` 1)
            `twant` lies beyond tend (setup) in the direction of
            integration, but is very close to tend.

            You may have intended `twant` = tend.

            abs(`twant`-tend) = *<value>*.

        (`errno` 1)
            `twant` does not lie in the direction of integration.
            `twant` = *<value>*.

        (`errno` 1)
            tend (setup) had already been reached in a previous call.

            To start a new problem, you will need to call the setup
            function.

        (`errno` 1)
            You cannot call this function after it has returned an
            error.

            You must call the setup function to start another problem.

        (`errno` 1)
            You cannot call this function when you have specified, in
            the setup function, that the step integrator will be used.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            This function is being used inefficiently because the step
            size has been reduced drastically many times to obtain
            answers at many points. Using the order 4 and 5 pair method
            at setup is more appropriate here. You can continue
            integrating this problem.

        (`errno` 3)
            Approximately *<value>* function evaluations have been used
            to compute the solution since the integration started or
            since this message was last printed. However, you can
            continue integrating the problem.

        (`errno` 4)
            Approximately *<value>* function evaluations have been used
            to compute the solution since the integration started or
            since this message was last printed. Your problem has been
            diagnosed as stiff. If the situation persists, it will cost
            roughly *<value>* times as much to reach tend (setup) as it
            has cost to reach the current time. You should probably call
            functions intended for stiff problems. However, you can
            continue integrating the problem.

        (`errno` 4)
            Your problem has been diagnosed as stiff. If the situation
            persists, it will cost roughly *<value>* times as much to
            reach tend (setup) as it has cost to reach the current time.
            You should probably call functions intended for stiff
            problems. However, you can continue integrating the problem.

        (`errno` 5)
            In order to satisfy your error requirements the solver has
            to use a step size of *<value>* at the current time,
            *<value>*. This step size is too small for the machine
            precision, and is smaller than *<value>*.

        (`errno` 6)
            The global error assessment algorithm failed at start of
            integration.

            The integration is being terminated.

        (`errno` 6)
            The global error assessment may not be reliable for times
            beyond *<value>*.

            The integration is being terminated.
    """

    from ..base.ode import ivp_rkts_range as b_ivp_rkts_range

    fun_name = 'naginterfaces.library.ode.ivp_rkts_range'

    cb_f_name = fun_name + ':f'

    def cb_f(t, n, y, yp, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _yp_f = f(
                t,
                y,
                data,
            )
        else:
            _yp_f = f(
                t,
                y,
            )
        yp[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'yp'},
            _yp_f,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    _ygot = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ygot'},
        ygot,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ygot, {'fun_name': fun_name, 'entity': 'ygot'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ypgot = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _ymax = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ymax'},
        ymax,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = b_ivp_rkts_range(
            _f,
            _n,
            twant,
            _ygot,
            _ypgot,
            _ymax,
            comm,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ygot_py = _ygot
    _ymax_py = _ymax
    _tgot_py = _return_scalar
    _ypgot_py = _ypgot
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['tgot', 'ygot', 'ypgot', 'ymax'],
    )(
        _tgot_py,
        _ygot_py,
        _ypgot_py,
        _ymax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_rkts_onestep(f, n, comm, data=None):
    u"""
    Ordinary differential equations, initial value problem, Runge--Kutta
    method, integration over one step.

    ``ivp_rkts_onestep`` is a one-step function for solving an initial
    value problem for a first-order system of ordinary differential
    equations using Runge--Kutta methods.

    For full information please refer to the NAG Library document for
    d02pf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02pff.html

    Parameters
    ----------
    f : callable yp = f(t, y, data=None)
        `f` must evaluate the functions f_i (that is the first
        derivatives y'_i) for given values of the arguments t, y_i.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (n)
            The current values of the dependent variables, y_i, for i =
            1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        yp : float, array-like, shape (n)
            The values of f_i, for i = 1,2,...,n.

    n : int
        n, the number of ordinary differential equations in the system
        to be solved.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    tnow : float
        t, the value of the independent variable at which a solution has
        been computed.

    ynow : float, ndarray, shape (`n`)
        An approximation to the solution at `tnow`. The local error of
        the step to `tnow` was no greater than permitted by the
        specified tolerances (see ``ivp_rkts_setup``).

    ypnow : float, ndarray, shape (`n`)
        An approximation to the first derivative of the solution at
        `tnow`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            tend, as specified in the setup function, has already been
            reached. To start a new problem, you will need to call the
            setup function. To continue integration beyond tend then
            ``ivp_rkts_reset_tend`` must first be called to reset tend
            to a new end value.

        (`errno` 1)
            A call to this function cannot be made after it has returned
            an error.

            The setup function must be called to start another problem.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted.

        (`errno` 1)
            On entry, `n` = *<value>*, but the value passed to the setup
            function was n = *<value>*.

        (`errno` 1)
            On entry, the communication arrays have become corrupted, or
            a catastrophic error has already been detected elsewhere.
            You cannot continue integrating the problem.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            More than 100 output points have been obtained by
            integrating to tend (as specified in the setup function).
            They have been so clustered that it would probably be (much)
            more efficient to use the interpolation function (if
            abs(method) = 3, switch to abs(method) = 2 at setup).
            However, you can continue integrating the problem.

        (`errno` 3)
            Approximately *<value>* function evaluations have been used
            to compute the solution since the integration started or
            since this message was last printed. However, you can
            continue integrating the problem.

        (`errno` 4)
            Approximately *<value>* function evaluations have been used
            to compute the solution since the integration started or
            since this message was last printed. Your problem has been
            diagnosed as stiff. If the situation persists, it will cost
            roughly *<value>* times as much to reach tend (setup) as it
            has cost to reach the current time. You should probably call
            functions intended for stiff problems. However, you can
            continue integrating the problem.

        (`errno` 4)
            Your problem has been diagnosed as stiff. If the situation
            persists, it will cost roughly *<value>* times as much to
            reach tend (setup) as it has cost to reach the current time.
            You should probably call functions intended for stiff
            problems. However, you can continue integrating the problem.

        (`errno` 5)
            In order to satisfy your error requirements the solver has
            to use a step size of *<value>* at the current time,
            *<value>*. This step size is too small for the machine
            precision, and is smaller than *<value>*.

        (`errno` 6)
            The global error assessment may not be reliable for times
            beyond *<value>*.

            The integration is being terminated.

        (`errno` 6)
            The global error assessment algorithm failed at start of
            integration.

            The integration is being terminated.
    """

    from ..base.ode import ivp_rkts_onestep as b_ivp_rkts_onestep

    fun_name = 'naginterfaces.library.ode.ivp_rkts_onestep'

    cb_f_name = fun_name + ':f'

    def cb_f(t, n, y, yp, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _yp_f = f(
                t,
                y,
                data,
            )
        else:
            _yp_f = f(
                t,
                y,
            )
        yp[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'yp'},
            _yp_f,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ynow = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _ypnow = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = b_ivp_rkts_onestep(
            _f,
            _n,
            _ynow,
            _ypnow,
            comm,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _tnow_py = _return_scalar
    _ynow_py = _ynow
    _ypnow_py = _ypnow
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['tnow', 'ynow', 'ypnow'],
    )(
        _tnow_py,
        _ynow_py,
        _ypnow_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_rk_step_revcomm(irevcm, yp, comm):
    u"""
    Ordinary differential equations, initial value problem, Runge--Kutta
    method, integration by reverse communication.

    ``ivp_rk_step_revcomm`` is a reverse communication one-step function
    for solving an initial value problem for a first-order system of
    ordinary differential equations using Runge--Kutta methods.
    The direct communication version of this function is
    ``ivp_rkts_onestep``.
    See [Direct and Reverse Communication Routines] for the difference
    between forward and reverse communication.

    For full information please refer to the NAG Library document for
    d02pg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02pgf.html

    Parameters
    ----------
    irevcm : int
        On initial entry `irevcm` must be set to zero, otherwise should
        be unchanged from the previous call.

    yp : float, array-like, shape (n)
        On initial entry: `yp` need not be set.

        On intermediate entry: `yp` must contain the value of the
        derivatives y' = f(t,y) where t is supplied in `t` and y is
        supplied in the array `y`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    Returns
    -------
    irevcm : int
        On intermediate exit `irevcm` returns a value > 0 to indicate
        that a function evaluation is required prior to re-entry.

        On final exit returns -1 or -2 denoting success or failure,
        respectively.

    t : float
        On intermediate exit: `t` contains the value of the independent
        variable t at which the derivatives y' are to be evaluated.

        On final exit: the value of t at which a solution has been
        computed following a successful step.

    y : float, ndarray, shape (n)
        On intermediate exit: `y` contains the value of the solution y
        at which the derivatives y' are to be evaluated.

        On final exit: the approximation to the solution computed
        following a successful step.

    Raises
    ------
    NagValueError
        (`errno` 1)
            tend, as specified in the setup function, has already been
            reached. To start a new problem, you will need to call the
            setup function. To continue integration beyond tend then
            ``ivp_rkts_reset_tend`` must first be called to reset tend
            to a new end value.

        (`errno` 1)
            A call to this function cannot be made after it has returned
            an error.

            The setup function must be called to start another problem.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted.

        (`errno` 1)
            `irevcm` < 0 on entry.

        (`errno` 1)
            On entry, n = *<value>*, but the value passed to the setup
            function was n = *<value>*.

        (`errno` 1)
            On entry, the communication arrays have become corrupted, or
            a catastrophic error has already been detected elsewhere.
            You cannot continue integrating the problem.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            More than 100 output points have been obtained by
            integrating to tend (as specified in the setup function).
            They have been so clustered that it would probably be (much)
            more efficient to use the interpolation function. However,
            you can continue integrating the problem.

        (`errno` 3)
            Approximately *<value>* function evaluations have been used
            to compute the solution since the integration started or
            since this message was last printed. However, you can
            continue integrating the problem.

        (`errno` 4)
            Approximately *<value>* function evaluations have been used
            to compute the solution since the integration started or
            since this message was last printed. Your problem has been
            diagnosed as stiff. If the situation persists, it will cost
            roughly *<value>* times as much to reach tend (setup) as it
            has cost to reach the current time. You should probably call
            functions intended for stiff problems. However, you can
            continue integrating the problem.

        (`errno` 4)
            Your problem has been diagnosed as stiff. If the situation
            persists, it will cost roughly *<value>* times as much to
            reach tend (setup) as it has cost to reach the current time.
            You should probably call functions intended for stiff
            problems. However, you can continue integrating the problem.

        (`errno` 5)
            In order to satisfy your error requirements the solver has
            to use a step size of *<value>* at the current time,
            *<value>*. This step size is too small for the machine
            precision, and is smaller than *<value>*.

        (`errno` 6)
            The global error assessment may not be reliable for times
            beyond *<value>*.

            The integration is being terminated.

        (`errno` 6)
            The global error assessment algorithm failed at start of
            integration.

            The integration is being terminated.
    """

    from ..base.ode import ivp_rk_step_revcomm as b_ivp_rk_step_revcomm

    fun_name = 'naginterfaces.library.ode.ivp_rk_step_revcomm'

    _yp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yp'},
        yp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_yp, {'fun_name': fun_name, 'entity': 'yp'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = b_ivp_rk_step_revcomm(
            irevcm,
            _n,
            _y,
            _yp,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _irevcm_py = _return_scalars.irevcm
    _t_py = _return_scalars.t
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm', 't', 'y'],
    )(
        _irevcm_py,
        _t_py,
        _y_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_rk_interp_setup(irevcm, nwant, yp, comm):
    u"""
    Set up interpolant by reverse communication for solution and
    derivative evaluations at points within the range of the last
    integration step taken by ``ivp_rk_step_revcomm``.

    ``ivp_rk_interp_setup`` is a reverse communication function that
    computes the interpolant for evaluation by ``ivp_rk_interp_eval``
    anywhere on an integration step taken by ``ivp_rk_step_revcomm``.
    The direct communication version of the ``ivp_rk_interp_setup`` and
    ``ivp_rk_interp_eval`` pair is ``ivp_rkts_interp``.
    A significant difference in functionality between the forward and
    reverse communication versions is that ``ivp_rk_interp_setup`` and
    ``ivp_rk_interp_eval`` can interpolate for the high-order
    Runge--Kutta method.

    For full information please refer to the NAG Library document for
    d02ph

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02phf.html

    Parameters
    ----------
    irevcm : int
        On initial entry: `irevcm` must be set to zero to indicate that
        the interpolant for a new step is being taken.

        On intermediate entry: `irevcm` should remain unchanged.

    nwant : int
        The number of components of the solution to be computed. The
        first `nwant` components are evaluated.

    yp : float, array-like, shape (n)
        On initial entry: need not be set.

        On intermediate entry: `yp` must contain the values of the
        derivatives y'_i for the given values of the parameters t, y_i.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    Returns
    -------
    irevcm : int
        On intermediate exit: `irevcm` returns a value 1 to indicate
        that a function evaluation is required prior to re-entry; the
        value of the derivatives must be returned in `yp` where the
        value of t is supplied in `t` and the values y(t) are supplied
        in the array `y`.

        On final exit:

        `irevcm` = -1

          Successful exit; `comm`['rwsav'] and `comm`['wcomm'] contain
          details of the interpolant.

        `irevcm` = -2

          Error exit; errno should be interrogated to determine the
          nature of the error.

    t : float
        On intermediate exit: `t` contains the value of the independent
        variable t at which the derivatives y' are to be evaluated.

        On final exit: contains no useful information.

    y : float, ndarray, shape (n)
        On intermediate exit: `y` contains the value of the solution y
        at which the derivatives y' are to be evaluated.

        On final exit: contains no useful information.

    Raises
    ------
    NagValueError
        (`errno` 1)
            You cannot call this function after the integrator has
            returned an error.

        (`errno` 1)
            You cannot call this function before you have called the
            step integrator.

        (`errno` 1)
            You cannot call this function after the range integrator has
            been called.

        (`errno` 1)
            On entry, n = *<value>*, but the value passed to the setup
            function was n = *<value>*.

        (`errno` 1)
            On entry, `nwant` = *<value>* and n = *<value>*.

            Constraint: 1 <= `nwant` <= n.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted, or a
            catastrophic error has already been detected elsewhere.

            You cannot continue integrating the problem.
    """

    from ..base.ode import ivp_rk_interp_setup as b_ivp_rk_interp_setup

    fun_name = 'naginterfaces.library.ode.ivp_rk_interp_setup'

    _yp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yp'},
        yp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_yp, {'fun_name': fun_name, 'entity': 'yp'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = b_ivp_rk_interp_setup(
        irevcm,
        _n,
        nwant,
        _y,
        _yp,
        comm,
    )
    _irevcm_py = _return_scalars.irevcm
    _t_py = _return_scalars.t
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm', 't', 'y'],
    )(
        _irevcm_py,
        _t_py,
        _y_py,
    )
    return _return_data

def ivp_rk_interp_eval(icheck, n, nwant, t, ideriv, comm):
    u"""
    Evaluate interpolant, set up using ``ivp_rkts_setup``, to
    approximate solution and/or solution derivatives at a point within
    the range of the last integration step taken by
    ``ivp_rk_step_revcomm``.

    ``ivp_rk_interp_eval`` evaluates the interpolant calculated by
    ``ivp_rk_interp_setup``, following an integration step performed by
    ``ivp_rk_step_revcomm`` to solve an initial value problem.

    For full information please refer to the NAG Library document for
    d02pj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02pjf.html

    Parameters
    ----------
    icheck : int
        Indicates whether consistency checks on input arguments should
        be performed

        `icheck` != 1

          Don't perform checks on input arguments.

        `icheck` = 1

          Perform consistency checks on input arguments.

        It is recommended to use `icheck` = 1 on the first call
        following a call to ``ivp_rk_interp_setup`` and to set `icheck`
        != 1 on subsequent calls within the last step to avoid the
        overhead of argument checking.

    n : int
        n, the dimension of the system of ODEs being integrated.

    nwant : int
        Only the first `nwant` system components to be computed. This
        should be the same value as passed to ``ivp_rk_interp_setup``
        when computing the interpolant.

    t : float
        t, the value of the independent variable where a solution is
        desired. Although any value of t can be supplied, accurate
        solutions can only be obtained for values in the range of the
        last time-step taken by ``ivp_rk_step_revcomm``.

    ideriv : int
        `ideriv` = 0

          Compute approximations to the first `nwant` components of the
          solution y(t).

        `ideriv` = 1

          Compute approximations to the first `nwant` components of the
          first derivatives of the solution y'(t).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    Returns
    -------
    sol : float, ndarray, shape (`nwant`)
        `ideriv` = 0

          The first `nwant` components of the solution y(t).

        `ideriv` = 1

          The first `nwant` components of the first derivatives of the
          solution y'(t).

    Raises
    ------
    NagValueError
        (`errno` 1)
            The previous call to the interpolation setup function
            returned an error.

        (`errno` 1)
            You cannot call this function before you have called the
            interpolation setup.

        (`errno` 1)
            On entry, `n` = *<value>*, but the value passed to the setup
            routine was `n` = *<value>*.

        (`errno` 1)
            On entry, `ideriv` = *<value>*.

            Constraint: `ideriv` = 0 or 1.

        (`errno` 1)
            On entry, `nwant` = *<value>*, but on interpolation setup
            `nwant` = *<value>*.

            Constraint: `nwant` must be unchanged from setup.

        (`errno` 1)
            On entry, lwcomm = *<value>* and `nwant` = *<value>*.

            Constraint: for method = -3 or 3, lwcomm >= 8*`nwant`.

        (`errno` 1)
            On entry, lwcomm = *<value>*, `n` = *<value>* and `nwant` =
            *<value>*.

            Constraint: for method = -2 or 2, lwcomm >= `n`+max(`n`,
            5*`nwant`).

        (`errno` 1)
            On entry, lwcomm = *<value>*.

            Constraint: for method = -1 or 1, lwcomm >= 1.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted, or a
            catastrophic error has already been detected elsewhere.

            You cannot continue integrating the problem.
    """

    from ..base.ode import ivp_rk_interp_eval as b_ivp_rk_interp_eval

    fun_name = 'naginterfaces.library.ode.ivp_rk_interp_eval'

    _nwant = _b_utils._EngineIntScalarType(
        nwant, {'fun_name': fun_name, 'entity': 'nwant'},
    ).to_ctypes()
    _nwant_py = _b_utils._EngineIntScalarType.to_py(_nwant)
    _sol = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nwant_py,),
    )
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('wcomm', 'iwsav', 'rwsav'),
    )
    _lwcomm_py = _b_utils._EngineFloat64ArrayType.get_shape(comm['wcomm'], {'fun_name': fun_name, 'entity': 'comm[\'wcomm\']'}, exp_rank=1, validate=False)[0]
    _lwcomm = _b_utils._EngineIntScalarType(
        _lwcomm_py, {'fun_name': fun_name, 'entity': '_lwcomm'},
    ).to_ctypes()
    b_ivp_rk_interp_eval(
        icheck,
        n,
        _nwant,
        t,
        ideriv,
        _sol,
        comm,
        _lwcomm,
    )
    _sol_py = _sol
    return _sol_py

def ivp_rkts_setup(tstart, tend, yinit, tol, thresh, method, hstart=0.0):
    u"""
    Ordinary differential equations, initial value problem, setup for
    ``ivp_rkts_range`` and ``ivp_rkts_onestep``.

    ``ivp_rkts_setup`` is a setup function which must be called prior to
    the first call of either of the integration functions
    ``ivp_rkts_range``, ``ivp_rkts_onestep`` and
    ``ivp_rk_step_revcomm``.

    For full information please refer to the NAG Library document for
    d02pq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02pqf.html

    Parameters
    ----------
    tstart : float
        The initial value of the independent variable, t_0.

    tend : float
        The final value of the independent variable, t_f, at which the
        solution is required. `tstart` and `tend` together determine the
        direction of integration.

    yinit : float, array-like, shape (n)
        y_0, the initial values of the solution, y_i, for i = 1,2,...,n.

    tol : float
        A relative error tolerance. The actual tolerance used is
        max(10*machine precision, min(`tol`, 0.01)); that is, the
        minimum tolerance is set at 10 times machine precision and the
        maximum tolerance is set at 0.01.

    thresh : float, array-like, shape (n)
        A vector of thresholds. For the ith component, the actual
        threshold used is max(sqrt(safe range), `thresh`[i-1]), where
        safe range is the value returned by ``machine.real_safe``.

    method : int
        The Runge--Kutta method to be used.

        `method` = 1 or -1

          A 2(3) pair is used.

        `method` = 2 or -2

          A 4(5) pair is used.

        `method` = 3 or -3

          A 7(8) pair is used.

    hstart : float, optional
        A value for the size of the first step in the integration to be
        attempted. The absolute value of `hstart` is used with the
        direction being determined by `tstart` and `tend`. The actual
        first step taken by the integrator may be different to `hstart`
        if the underlying algorithm determines that `hstart` is
        unsuitable. If `hstart` = 0.0 then the size of the first step is
        computed automatically.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `tstart` = *<value>*.

            Constraint: `tstart` != `tend`.

        (`errno` 1)
            On entry, `tstart` is too close to `tend`.

            abs(`tstart`-`tend`) = *<value>*, but this quantity should
            be at least *<value>*.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = -3, -2, -1, 1, 2 or 3.

        (`errno` 1)
            On entry, too much workspace required.

            Workspace provided was *<value>*, workspace required is
            *<value>*.
    """

    from ..base.ode import ivp_rkts_setup as b_ivp_rkts_setup

    fun_name = 'naginterfaces.library.ode.ivp_rkts_setup'

    _yinit = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yinit'},
        yinit,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_yinit, {'fun_name': fun_name, 'entity': 'yinit'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _thresh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'thresh'},
        thresh,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_ivp_rkts_setup(
        _n,
        tstart,
        tend,
        _yinit,
        tol,
        _thresh,
        method,
        hstart,
    )
    _comm_py = _return_scalar
    return _comm_py

def ivp_rkts_reset_tend(tendnu, comm):
    u"""
    Ordinary differential equations, initial value problem, resets end
    of range for ``ivp_rkts_onestep``.

    ``ivp_rkts_reset_tend`` resets the end point in an integration
    performed by ``ivp_rkts_onestep`` and ``ivp_rk_step_revcomm``.

    For full information please refer to the NAG Library document for
    d02pr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02prf.html

    Parameters
    ----------
    tendnu : float
        The new value for t_f.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            You cannot call this function before you have called the
            setup function.

        (`errno` 1)
            You cannot call this function after the integrator has
            returned an error.

        (`errno` 1)
            You cannot call this function before you have called the
            step integrator.

        (`errno` 1)
            You cannot call this function when the range integrator has
            been used.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted, or a
            catastrophic error has already been detected elsewhere. You
            cannot continue integrating the problem.

        (`errno` 1)
            On entry, `tendnu` is not beyond tnow (step integrator) in
            the direction of integration.

            The direction is positive, `tendnu` = *<value>* and tnow =
            *<value>*.

        (`errno` 1)
            On entry, `tendnu` is not beyond tnow (step integrator) in
            the direction of integration.

            The direction is negative, `tendnu` = *<value>* and tnow =
            *<value>*.

        (`errno` 1)
            On entry, `tendnu` is too close to tnow (step integrator).
            Their difference is *<value>*, but this quantity must be at
            least *<value>*.
    """

    from ..base.ode import ivp_rkts_reset_tend as b_ivp_rkts_reset_tend

    b_ivp_rkts_reset_tend(
        tendnu,
        comm,
    )

def ivp_rkts_interp(n, twant, ideriv, nwant, f, comm, data=None):
    u"""
    Ordinary differential equations, initial value problem,
    interpolation for ``ivp_rkts_onestep``.

    ``ivp_rkts_interp`` computes the solution of a system of ordinary
    differential equations using interpolation anywhere on an
    integration step taken by ``ivp_rkts_onestep``.

    For full information please refer to the NAG Library document for
    d02ps

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02psf.html

    Parameters
    ----------
    n : int
        n, the number of ordinary differential equations in the system
        to be solved by the integration function.

    twant : float
        t, the value of the independent variable where a solution is
        desired.

    ideriv : int
        Determines whether the solution and/or its first derivative are
        to be computed

        `ideriv` = 0

          compute approximate solution.

        `ideriv` = 1

          compute approximate first derivative.

        `ideriv` = 2

          compute approximate solution and first derivative.

    nwant : int
        The number of components of the solution to be computed. The
        first `nwant` components are evaluated.

    f : callable yp = f(t, y, data=None)
        `f` must evaluate the functions f_i (that is the first
        derivatives y'_i) for given values of the arguments t,y_i.

        It must be the same procedure as supplied to
        ``ivp_rkts_onestep``.

        Parameters
        ~~~~~~~~~~
        t : float
            t, the current value of the independent variable.

        y : float, ndarray, shape (n)
            The current values of the dependent variables, y_i, for i =
            1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        yp : float, array-like, shape (n)
            The values of f_i, for i = 1,2,...,n.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    ywant : float, ndarray, shape (`nwant`)
        An approximation to the first `nwant` components of the solution
        at `twant` if `ideriv` = 0 or 2. Otherwise `ywant` is not
        defined.

    ypwant : float, ndarray, shape (`nwant`)
        An approximation to the first `nwant` components of the first
        derivative at `twant` if `ideriv` = 1 or 2. Otherwise `ypwant`
        is not defined.

    Raises
    ------
    NagValueError
        (`errno` 1)
            You cannot call this function after the integrator has
            returned an error.

        (`errno` 1)
            You cannot call this function before you have called the
            step integrator.

        (`errno` 1)
            You cannot call this function when you have specified, in
            the setup function, that the range integrator will be used.

        (`errno` 1)
            method = -3 or 3 in setup, but interpolation is not
            available for this method. Either use method = -2 or 2 in
            setup or use reset function to force the integrator to step
            to particular points.

        (`errno` 1)
            On entry, `n` = *<value>*, but the value passed to the setup
            function was n = *<value>*.

        (`errno` 1)
            On entry, `ideriv` = *<value>*.

            Constraint: `ideriv` = 0, 1 or 2.

        (`errno` 1)
            On entry, `nwant` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nwant` <= `n`.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted, or a
            catastrophic error has already been detected elsewhere.

            You cannot continue integrating the problem.
    """

    from ..base.ode import ivp_rkts_interp as b_ivp_rkts_interp

    fun_name = 'naginterfaces.library.ode.ivp_rkts_interp'

    cb_f_name = fun_name + ':f'

    def cb_f(t, n, y, yp, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _yp_f = f(
                t,
                y,
                data,
            )
        else:
            _yp_f = f(
                t,
                y,
            )
        yp[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'yp'},
            _yp_f,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _nwant = _b_utils._EngineIntScalarType(
        nwant, {'fun_name': fun_name, 'entity': 'nwant'},
    ).to_ctypes()
    _nwant_py = _b_utils._EngineIntScalarType.to_py(_nwant)
    _ywant = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nwant_py,),
    )
    _ypwant = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nwant_py,),
    )
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    b_ivp_rkts_interp(
        n,
        twant,
        ideriv,
        _nwant,
        _ywant,
        _ypwant,
        _f,
        comm,
        data,
    )
    _ywant_py = _ywant
    _ypwant_py = _ypwant
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ywant', 'ypwant'],
    )(
        _ywant_py,
        _ypwant_py,
    )
    return _return_data

def ivp_rkts_diag(comm):
    u"""
    Ordinary differential equations, initial value problem, integration
    diagnostics for ``ivp_rkts_range`` and ``ivp_rkts_onestep``.

    ``ivp_rkts_diag`` provides details about an integration performed by
    either ``ivp_rkts_range``, ``ivp_rkts_onestep`` or
    ``ivp_rk_step_revcomm``.

    For full information please refer to the NAG Library document for
    d02pt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ptf.html

    Parameters
    ----------
    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_rkts_setup`` and one of ``ivp_rkts_range``,
        ``ivp_rkts_onestep`` or ``ivp_rk_step_revcomm``.

    Returns
    -------
    fevals : int
        The total number of evaluations of f used in the integration so
        far; this includes evaluations of f required for the secondary
        integration necessary if ``ivp_rkts_setup`` had previously been
        called with method > 0.

    stepcost : int
        The cost in terms of number of evaluations of f of a typical
        step with the method being used for the integration. The method
        is specified by the argument method in a prior call to
        ``ivp_rkts_setup``.

    waste : float
        The number of attempted steps that failed to meet the local
        error requirement divided by the total number of steps attempted
        so far in the integration. A 'large' fraction indicates that the
        integrator is having trouble with the problem being solved. This
        can happen when the problem is 'stiff' and also when the
        solution has discontinuities in a low-order derivative.

    stepsok : int
        The number of accepted steps.

    hnext : float
        The step size the integrator will attempt to use for the next
        step.

    Raises
    ------
    NagValueError
        (`errno` 1)
            You cannot call this function before you have called the
            integrator.

        (`errno` 1)
            You have already made one call to this function after the
            integrator could not achieve specified accuracy.

            You cannot call this function again.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted, or a
            catastrophic error has already been detected elsewhere.

            You cannot continue integrating the problem.
    """

    from ..base.ode import ivp_rkts_diag as b_ivp_rkts_diag

    _return_scalars = b_ivp_rkts_diag(
        comm,
    )
    _fevals_py = _return_scalars.fevals
    _stepcost_py = _return_scalars.stepcost
    _waste_py = _return_scalars.waste
    _stepsok_py = _return_scalars.stepsok
    _hnext_py = _return_scalars.hnext
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['fevals', 'stepcost', 'waste', 'stepsok', 'hnext'],
    )(
        _fevals_py,
        _stepcost_py,
        _waste_py,
        _stepsok_py,
        _hnext_py,
    )
    return _return_data

def ivp_rkts_errass(n, comm):
    u"""
    Ordinary differential equations, initial value problem, error
    assessment diagnostics for ``ivp_rkts_range`` and
    ``ivp_rkts_onestep``.

    ``ivp_rkts_errass`` provides details about global error assessment
    computed during an integration with either ``ivp_rkts_range``,
    ``ivp_rkts_onestep`` or ``ivp_rk_step_revcomm``.

    For full information please refer to the NAG Library document for
    d02pu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02puf.html

    Parameters
    ----------
    n : int
        n, the number of ordinary differential equations in the system
        to be solved by the integration function.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_rkts_setup``.

    Returns
    -------
    rmserr : float, ndarray, shape (`n`)
        `rmserr`[i-1] approximates the RMS average of the true error of
        the numerical solution for the ith solution component, for i =
        1,2,...,n. The average is taken over all steps from the
        beginning of the integration to the current integration point.

    errmax : float
        The maximum weighted approximate true error taken over all
        solution components and all steps.

    terrmx : float
        The first value of the independent variable where an approximate
        true error attains the maximum value, `errmax`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*, but the value passed to the setup
            function was n = *<value>*.

        (`errno` 1)
            No error assessment is available since the integrator has
            not actually taken any successful steps.

        (`errno` 1)
            No error assessment is available since you did not ask for
            it in your call to the setup function.

        (`errno` 1)
            You cannot call this function before you have called the
            integrator.

        (`errno` 1)
            You have already made one call to this function after the
            integrator could not achieve specified accuracy.

            You cannot call this function again.

        (`errno` 1)
            On entry, a previous call to the setup function has not been
            made or the communication arrays have become corrupted, or a
            catastrophic error has already been detected elsewhere.

            You cannot continue integrating the problem.
    """

    from ..base.ode import ivp_rkts_errass as b_ivp_rkts_errass

    fun_name = 'naginterfaces.library.ode.ivp_rkts_errass'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _rmserr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = b_ivp_rkts_errass(
        _n,
        _rmserr,
        comm,
    )
    _rmserr_py = _rmserr
    _errmax_py = _return_scalars.errmax
    _terrmx_py = _return_scalars.terrmx
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['rmserr', 'errmax', 'terrmx'],
    )(
        _rmserr_py,
        _errmax_py,
        _terrmx_py,
    )
    return _return_data

def ivp_adams_roots(fcn, t, y, tout, neqg, comm, g=None, data=None):
    u"""
    Ordinary differential equations, initial value problem, Adams'
    method with root-finding (direct communication, comprehensive).

    ``ivp_adams_roots`` is a function for integrating a non-stiff system
    of first-order ordinary differential equations using a
    variable-order variable-step Adams' method.
    A root-finding facility is provided.

    For full information please refer to the NAG Library document for
    d02qf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02qff.html

    Parameters
    ----------
    fcn : callable f = fcn(x, y, data=None)
        `fcn` must evaluate the functions f_i (that is the first
        derivatives y'_i) for given values of its arguments x,
        y_1,y_2,...,y_neqf.

        Parameters
        ~~~~~~~~~~
        x : float
            The current value of the argument x.

        y : float, ndarray, shape (neqf)
            y_i, for i = 1,2,...,neqf, the current value of the
            argument.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (neqf)
            The value of f_i, for i = 1,2,...,neqf.

    t : float
        After a call to ``ivp_adams_setup`` with statef = 'S', `t` must
        be set to the initial value of the independent variable x.

    y : float, array-like, shape (neqf)
        The initial values of the solution y_1,y_2,...,y_neqf.

    tout : float
        The next value of x at which a computed solution is required.
        For the initial `t`, the input value of `tout` is used to
        determine the direction of integration. Integration is permitted
        in either direction. If `tout` = `t` on exit, `tout` must be
        reset beyond `t` in the direction of integration, before any
        continuation call.

    neqg : int
        The number of event functions which you are defining for
        root-finding. If root-finding is not required the value for
        `neqg` must be <= 0. Otherwise it must be the same argument
        `neqg` used in the prior call to ``ivp_adams_setup``.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_adams_setup``.

    g : None or callable retval = g(x, y, yp, k, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `g` must evaluate a given component of g(x,y,y') at a specified
        point.

        Parameters
        ~~~~~~~~~~
        x : float
            The current value of the independent variable.

        y : float, ndarray, shape (neqf)
            The current values of the dependent variables.

        yp : float, ndarray, shape (neqf)
            The current values of the derivatives of the dependent
            variables.

        k : int
            The component of g which must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the component of g(x,y,y') at the specified
            point.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    t : float
        The value of x at which y has been computed. This may be an
        intermediate output point, a root, `tout` or a point at which an
        error has occurred. If the integration is to be continued,
        possibly with a new value for `tout`, `t` must not be changed.

    y : float, ndarray, shape (neqf)
        The computed values of the solution at the exit value of `t`. If
        the integration is to be continued, possibly with a new value
        for `tout`, these values must not be changed.

    root : bool
        If root-finding was required (`neqg` > 0 on entry), `root`
        specifies whether or not the output value of the argument `t` is
        a root of one of the event functions. If `root` = False, no root
        was detected, whereas `root` = True indicates a root and you
        should make a call to ``ivp_adams_rootdiag`` for further
        information.

        If root-finding was not required (`neqg` = 0 on entry), then on
        exit `root` = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The value of `tout`, *<value>*, indicates a change in the
            integration direction. This is not permitted on a
            continuation call.

        (`errno` 1)
            The value of `t` has been changed from *<value>* to
            *<value>*.

            This is not permitted on a continuation call.

        (`errno` 1)
            On entry, `tout` = *<value>* but crit = True in
            ``ivp_adams_setup``.

            Integration cannot be attempted beyond tcrit = *<value>*.

        (`errno` 1)
            On entry, `tout` = `t` = *<value>*.

        (`errno` 1)
            On entry, `neqg` = *<value>* and neqg = 0 in
            ``ivp_adams_setup``.

            Constraint: if `neqg` <= 0 then `neqg` = neqg in
            ``ivp_adams_setup``.

        (`errno` 1)
            On entry, `neqg` = *<value>* and neqg = 0 in
            ``ivp_adams_setup``.

            Constraint: if `neqg` > 0 then neqg > 0 in
            ``ivp_adams_setup``.

        (`errno` 1)
            On entry, `neqg` = *<value>* and neqg = *<value>* in
            ``ivp_adams_setup``.

            Constraint: if `neqg` <= 0 then neqg <= 0 in
            ``ivp_adams_setup``.

        (`errno` 1)
            On entry, neqf = *<value>* and neqf = *<value>* in
            ``ivp_adams_setup``.

            Constraint: neqf = neqf in ``ivp_adams_setup``.

        (`errno` 1)
            The call to setup function ``ivp_adams_setup`` produced an
            error.

        (`errno` 1)
            The setup function ``ivp_adams_setup`` has not been called.

        (`errno` 3)
            The error tolerances are too stringent.

        (`errno` 7)
            Two successive errors detected at the current value of `t`,
            *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The maximum number of steps has been attempted.

            If integration is to be continued then the function may be
            called again and a further maxstp in ``ivp_adams_setup``
            steps will be attempted.

        (`errno` 4)
            Error weight i has become zero during the integration.

            atol[i] = 0.0 in ``ivp_adams_setup`` but `y`[i] is now 0.0.
            Integration successful as far as `t`.

        (`errno` 5)
            The problem appears to be stiff.

        (`errno` 6)
            A change in sign of an event function has been detected but
            the root-finding process appears to have converged to a
            singular point of `t` rather than a root. Integration may be
            continued by calling the function again.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import ivp_adams_roots as b_ivp_adams_roots

    fun_name = 'naginterfaces.library.ode.ivp_adams_roots'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(neqf, x, y, f, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(neqf,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_g(neqf, x, y, yp, k, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _retval_g = g(
                x,
                y,
                yp,
                k,
                data,
            )
        else:
            _retval_g = g(
                x,
                y,
                yp,
                k,
            )
        return _retval_g

    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqf_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neqf = _b_utils._EngineIntScalarType(
        _neqf_py, {'fun_name': fun_name, 'entity': '_neqf'},
    ).to_ctypes()
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'}, allow_none=True)
    _g = None if g is None else cb_g
    try:
        _return_scalars = b_ivp_adams_roots(
            _fcn,
            _neqf,
            t,
            _y,
            tout,
            _g,
            neqg,
            comm,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _t_py = _return_scalars.t
    _y_py = _y
    _root_py = _return_scalars.root
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'y', 'root'],
    )(
        _t_py,
        _y_py,
        _root_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_adams_roots_revcom(t, y, tout, neqg, irevcm, grvcm, kgrvcm, comm):
    u"""
    Ordinary differential equations, initial value problem, Adams'
    method with root-finding (reverse communication, comprehensive).

    ``ivp_adams_roots_revcom`` is a reverse communication function for
    integrating a non-stiff system of first-order ordinary differential
    equations using a variable-order variable-step Adams' method.
    A root-finding facility is provided.

    For full information please refer to the NAG Library document for
    d02qg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02qgf.html

    Parameters
    ----------
    t : float
        On initial entry: that is after a call to ``ivp_adams_setup``
        with statef = 'S', `t` must be set to the initial value of the
        independent variable x.

    y : float, ndarray, shape (neqf), modified in place
        On initial entry: the initial values of the solution
        y_1,y_2,...,y_neqf.

        On final exit: the computed values of the solution at the exit
        value of `t`. If the integration is to be continued, possibly
        with a new value for `tout`, these values must not be changed.

    tout : float
        On initial entry: the next value of x at which a computed
        solution is required. For the initial `t`, the input value of
        `tout` is used to determine the direction of integration.
        Integration is permitted in either direction. If `tout` = `t` on
        exit, `tout` must be reset beyond `t` in the direction of
        integration, before any continuation call.

    neqg : int
        On initial entry: the number of event functions which you are
        defining for root-finding. If root-finding is not required the
        value for `neqg` must be <= 0. Otherwise it must be the same
        value as the argument `neqg` used in the prior call to
        ``ivp_adams_setup``.

    irevcm : int
        On initial entry: must have the value 0.

    grvcm : float
        On initial entry: need not be set.

        On intermediate entry: with `irevcm` = 9, 10, 11 or 12, `grvcm`
        must contain the value of g_k(x,y,y'), where k is given by
        `kgrvcm`.

    kgrvcm : int
        On intermediate entry: with `irevcm` = 9, 10, 11 or 12, `kgrvcm`
        must remain unchanged from a previous call to
        ``ivp_adams_roots_revcom``.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``ivp_adams_setup``.

    Returns
    -------
    t : float
        On final exit: the value of x at which y has been computed. This
        may be an intermediate output point, a root, `tout` or a point
        at which an error has occurred. If the integration is to be
        continued, possibly with a new value for `tout`, `t` must not be
        changed.

    root : bool
        On final exit: if root-finding was required (`neqg` > 0 on
        entry), `root` specifies whether or not the output value of the
        argument `t` is a root of one of the event functions. If `root`
        = False, no root was detected, whereas `root` = True indicates a
        root and you should make a call to ``ivp_adams_rootdiag`` for
        further information.

        If root-finding was not required (`neqg` = 0 on entry), `root` =
        False.

    irevcm : int
        On intermediate exit: specifies what action you must take before
        re-entering ``ivp_adams_roots_revcom`` with `irevcm` unchanged.

        `irevcm` = 1, 2, 3, 4, 5, 6 or 7

          Indicates that you must supply y' = f(x,y), where x is given
          by `trvcm` and y_i is returned in `y`[i-1], for i =
          1,2,...,neqf when `yrvcm` = 0 and
          `comm`['rwork'][`yrvcm`+i-2], for i = 1,2,...,neqf when
          `yrvcm` != 0. y'_i should be placed in location
          `comm`['rwork'][`yprvcm`+i-2], for i = 1,2,...,neqf.

        `irevcm` = 8

          Indicates that the current step was not successful due to
          error test failure. The only information supplied to you on
          this return is the current value of the independent variable
          `t`, as given by `trvcm`. No values must be changed before
          re-entering ``ivp_adams_roots_revcom``. This facility enables
          you to determine the number of unsuccessful steps.

        `irevcm` = 9, 10, 11 or 12

          Indicates that you must supply g_k(x,y,y'), where k is given
          by `kgrvcm`, x is given by `trvcm`, y_i is given by `y`[i-1]
          and y'_i is given by `comm`['rwork'][`yprvcm`-1+i-1]. The
          result g_k should be placed in the variable `grvcm`.

        On final exit: has the value 0, which indicates that an output
        point or root has been reached or an error has occurred (see
        errno).

    trvcm : float
        On intermediate exit: the current value of the independent
        variable.

    yrvcm : int
        On intermediate exit: with `irevcm` = 1, 2, 3, 4, 5, 6, 7, 9,
        10, 11 or 12, `yrvcm` specifies the locations of the dependent
        variables y for use in evaluating the differential system or the
        event functions.

        `yrvcm` = 0

          y_i is given by `y`[i], for i = 1,2,...,neqf.

        `yrvcm` != 0

          y_i is given by `comm`['rwork'][`yrvcm`+i-2], for i =
          1,2,...,neqf.

    yprvcm : int
        On intermediate exit: with `irevcm` = 1, 2, 3, 4, 5, 6 or 7,
        `yprvcm` specifies the positions in `comm`['rwork'] at which you
        should place the derivatives y'. y'_i should be placed in
        location `comm`['rwork'][`yprvcm`+i-2], for i = 1,2,...,neqf.

        With `irevcm` = 9, 10, 11 or 12, `yprvcm` specifies the
        locations of the derivatives y' for use in evaluating the event
        functions. y'_i is given by `comm`['rwork'][`yprvcm`+i-2], for i
        = 1,2,...,neqf.

    kgrvcm : int
        On intermediate exit: with `irevcm` = 9, 10, 11 or 12, `kgrvcm`
        specifies which event function g_k(x,y,y') you must evaluate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irevcm` had an illegal value, *<value>*.

        (`errno` 1)
            The value of `tout`, *<value>*, indicates a change in the
            integration direction. This is not permitted on a
            continuation call.

        (`errno` 1)
            The value of `t` has been changed from *<value>* to
            *<value>*.

            This is not permitted on a continuation call.

        (`errno` 1)
            On entry, `tout` = *<value>* but crit = True in
            ``ivp_adams_setup``.

            Integration cannot be attempted beyond tcrit = *<value>*.

        (`errno` 1)
            On entry, `tout` = `t` = *<value>*.

        (`errno` 1)
            On entry, `neqg` = *<value>* and neqg = 0 in
            ``ivp_adams_setup``.

            Constraint: if `neqg` <= 0 then `neqg` = neqg in
            ``ivp_adams_setup``.

        (`errno` 1)
            On entry, `neqg` = *<value>* and neqg = 0 in
            ``ivp_adams_setup``.

            Constraint: if `neqg` > 0 then neqg > 0 in
            ``ivp_adams_setup``.

        (`errno` 1)
            On entry, `neqg` = *<value>* and neqg = *<value>* in
            ``ivp_adams_setup``.

            Constraint: if `neqg` <= 0 then neqg <= 0 in
            ``ivp_adams_setup``.

        (`errno` 1)
            On entry, neqf = *<value>* and neqf = *<value>* in
            ``ivp_adams_setup``.

            Constraint: neqf = neqf in ``ivp_adams_setup``.

        (`errno` 1)
            The call to setup function ``ivp_adams_setup`` produced an
            error.

        (`errno` 1)
            The setup function ``ivp_adams_setup`` has not been called.

        (`errno` 3)
            The error tolerances are too stringent.

        (`errno` 7)
            Two successive errors detected at the current value of `t`,
            *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The maximum number of steps has been attempted.

            If integration is to be continued then the function may be
            called again and a further maxstp in ``ivp_adams_setup``
            steps will be attempted.

        (`errno` 4)
            Error weight i has become zero during the integration.

            atol[i] = 0.0 in ``ivp_adams_setup`` but `y`[i] is now 0.0.
            Integration successful as far as `t`.

        (`errno` 5)
            The problem appears to be stiff.

        (`errno` 6)
            A change in sign of an event function has been detected but
            the root-finding process appears to have converged to a
            singular point of `t` rather than a root. Integration may be
            continued by calling the function again.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_adams_roots_revcom as b_ivp_adams_roots_revcom

    fun_name = 'naginterfaces.library.ode.ivp_adams_roots_revcom'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
    )
    _neqf_py = _b_utils._EngineFloat64ArrayType.get_shape(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _neqf = _b_utils._EngineIntScalarType(
        _neqf_py, {'fun_name': fun_name, 'entity': '_neqf'},
    ).to_ctypes()
    try:
        _return_scalars = b_ivp_adams_roots_revcom(
            _neqf,
            t,
            y,
            tout,
            neqg,
            irevcm,
            grvcm,
            kgrvcm,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _t_py = _return_scalars.t
    _irevcm_py = _return_scalars.irevcm
    _kgrvcm_py = _return_scalars.kgrvcm
    _root_py = _return_scalars.root
    _trvcm_py = _return_scalars.trvcm
    _yrvcm_py = _return_scalars.yrvcm
    _yprvcm_py = _return_scalars.yprvcm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'root', 'irevcm', 'trvcm', 'yrvcm', 'yprvcm', 'kgrvcm'],
    )(
        _t_py,
        _root_py,
        _irevcm_py,
        _trvcm_py,
        _yrvcm_py,
        _yprvcm_py,
        _kgrvcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def ivp_adams_setup(statef, neqf, vectol, atol, rtol, onestp, crit, tcrit, hmax, maxstp, neqg, alterg, sophst, comm):
    u"""
    Ordinary differential equations, initial value problem, setup for
    ``ivp_adams_roots`` and ``ivp_adams_roots_revcom``.

    ``ivp_adams_setup`` is a setup function which must be called prior
    to the first call of either of the integrators ``ivp_adams_roots``
    and ``ivp_adams_roots_revcom`` and may be called prior to any
    continuation call to these functions.

    For full information please refer to the NAG Library document for
    d02qw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02qwf.html

    Parameters
    ----------
    statef : str, length 1
        Specifies whether the integration function (``ivp_adams_roots``
        or ``ivp_adams_roots_revcom``) is to start a new system of
        ordinary differential equations, restart a system or continue
        with a system.

        `statef` = 'S'

          Start integration with a new differential system.

        `statef` = 'R'

          Restart integration with the current differential system.

        `statef` = 'C'

          Continue integration with the current differential system.

    neqf : int
        The number of ordinary differential equations to be solved by
        the integration function. `neqf` must remain unchanged on
        subsequent calls to ``ivp_adams_setup`` with `statef` = 'C' or
        'R'.

    vectol : bool
        Specifies whether vector or scalar error control is to be
        employed for the local error test in the integration.

        If `vectol` = True, vector error control will be used and you
        must specify values of `rtol`[i-1] and `atol`[i-1], for i =
        1,2,...,`neqf`.

        Otherwise scalar error control will be used and you must specify
        values of just `rtol`[0] and `atol`[0].

        The error test to be satisfied is of the form

            sqrt(sum_i = 1^`neqf`((e_i)/(w_i))^2) <= 1.0.

        where w_i is defined as follows:

        [table omitted]

        and e_i is an estimate of the local error in y_i, computed
        internally. `vectol` must remain unchanged on subsequent calls
        to ``ivp_adams_setup`` with `statef` = 'C' or 'R'.

    atol : float, array-like, shape (latol)
        The absolute local error tolerance (see `vectol`).

    rtol : float, array-like, shape (lrtol)
        The relative local error tolerance (see `vectol`).

    onestp : bool
        The mode of operation of the integration function. If `onestp` =
        True, the integration function will operate in one-step mode,
        that is it will return after each successful step. Otherwise the
        integration function will operate in interval mode, that is it
        will return at the end of the integration interval.

    crit : bool
        Specifies whether or not there is a value for the independent
        variable beyond which integration is not to be attempted.
        Setting `crit` = True indicates that there is such a point,
        whereas `crit` = False indicates that there is no such
        restriction.

    tcrit : float
        With `crit` = True, `tcrit` must be set to a value of the
        independent variable beyond which integration is not to be
        attempted. Otherwise `tcrit` is not referenced.

    hmax : float
        If `hmax` != 0.0, a bound on the absolute step size during the
        integration is taken to be |`hmax`|.

        If `hmax` = 0.0, no bound is assumed on the step size during the
        integration.

        A bound may be required if there are features of the solution on
        very short ranges of integration which may be missed.

        You should try `hmax` = 0.0 first.

        Note: this argument only affects the step size if the option
        `crit` = True is being used.

    maxstp : int
        A bound on the number of attempted steps in any one call to the
        integration function. If `maxstp` <= 0 on entry, a value of 1000
        is used.

    neqg : int
        Specifies whether or not root-finding is required in
        ``ivp_adams_roots`` or ``ivp_adams_roots_revcom``.

        `neqg` <= 0

          No root-finding is attempted.

        `neqg` > 0

          Root-finding is required and `neqg` event functions will be
          specified for the integration function.

    alterg : bool
        Specifies whether or not the event functions have been
        redefined. `alterg` need not be set if `statef` = 'S'. On
        subsequent calls to ``ivp_adams_setup``, if `neqg` has been set
        positive, `alterg` = False specifies that the event functions
        remain unchanged, whereas `alterg` = True specifies that the
        event functions have changed. Because of the expense in
        reinitializing the root searching procedure, `alterg` should be
        set to True only if the event functions really have been
        altered. `alterg` need not be set if the root-finding option is
        not used.

    sophst : bool
        The type of search technique to be used in the root-finding. If
        `sophst` = True then a sophisticated and reliable but expensive
        technique will be used, whereas for `sophst` = False a simple
        but less reliable technique will be used. If `neqg` <= 0,
        `sophst` is not referenced.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    statef : str, length 1
        Is set to 'C', except that if an error is detected, `statef` is
        unchanged.

    alterg : bool
        Is set to False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `statef` = *<value>*.

            Constraint: `statef` = 'S', 'R' or 'C'.

        (`errno` 1)
            On entry, for i = *<value>*, `atol`[i] = 0.0, `rtol`[i] =
            *<value>* and 4*eps = *<value>*.

            Constraint: if `atol`[i] = 0.0 then `rtol`[i] >= 4*eps for
            all i.

        (`errno` 1)
            On entry, `rtol`[*<value>*] = *<value>*.

            Constraint: `rtol`[i] >= 0.0 for all i.

        (`errno` 1)
            On entry, `atol`[*<value>*] = *<value>*.

            Constraint: `atol`[i] >= 0.0 for all i.

        (`errno` 1)
            On entry, `statef` = 'R' or 'C', `alterg` = False and `neqg`
            = *<value>*.

            Constraint: if `statef` = 'R' or 'C' and `alterg` = False
            then `neqg` must retain its value from the prior call with
            `statef` = 'S'.

        (`errno` 1)
            On entry, `statef` = 'R' or 'C' and `vectol` = False.

            Constraint: if `statef` = 'R' or 'C' then `vectol` must
            retain its value from the prior call with `statef` = 'S'.

        (`errno` 1)
            On entry, `statef` = 'R' or 'C' and `vectol` = True.

            Constraint: if `statef` = 'R' or 'C' then `vectol` must
            retain its value from the prior call with `statef` = 'S'.

        (`errno` 1)
            On entry, `statef` = 'R' or 'C' and `neqf` = *<value>*.

            Constraint: if `statef` = 'R' or 'C' then `neqf` must retain
            its value, *<value>*, from the prior call with `statef` =
            'S'.

        (`errno` 1)
            On entry, latol = *<value>*.

            Constraint: latol >= 1.

        (`errno` 1)
            On entry, lrtol = *<value>*.

            Constraint: lrtol >= 1.

        (`errno` 1)
            On entry, `vectol` = True, latol = *<value>* and `neqf` =
            *<value>*.

            Constraint: if `vectol` = True then latol >= `neqf`.

        (`errno` 1)
            On entry, `vectol` = True, lrtol = *<value>* and `neqf` =
            *<value>*.

            Constraint: if `vectol` = True then lrtol >= `neqf`.

        (`errno` 1)
            On entry, `neqf` = *<value>*.

            Constraint: `neqf` >= 1.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import ivp_adams_setup as b_ivp_adams_setup

    fun_name = 'naginterfaces.library.ode.ivp_adams_setup'

    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _latol_py = _b_utils._EngineFloat64ArrayType.get_shape(_atol, {'fun_name': fun_name, 'entity': 'atol'}, exp_rank=1)[0]
    _latol = _b_utils._EngineIntScalarType(
        _latol_py, {'fun_name': fun_name, 'entity': '_latol'},
    ).to_ctypes()
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lrtol_py = _b_utils._EngineFloat64ArrayType.get_shape(_rtol, {'fun_name': fun_name, 'entity': 'rtol'}, exp_rank=1)[0]
    _lrtol = _b_utils._EngineIntScalarType(
        _lrtol_py, {'fun_name': fun_name, 'entity': '_lrtol'},
    ).to_ctypes()
    _return_scalars = b_ivp_adams_setup(
        statef,
        neqf,
        vectol,
        _atol,
        _latol,
        _rtol,
        _lrtol,
        onestp,
        crit,
        tcrit,
        hmax,
        maxstp,
        neqg,
        alterg,
        sophst,
        comm,
    )
    _statef_py = _return_scalars.statef
    _alterg_py = _return_scalars.alterg
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['statef', 'alterg'],
    )(
        _statef_py,
        _alterg_py,
    )
    return _return_data

def ivp_adams_diag(neqf, comm):
    u"""
    Ordinary differential equations, initial value problem, diagnostics
    for ``ivp_adams_roots`` and ``ivp_adams_roots_revcom``.

    ``ivp_adams_diag`` is a diagnostic function which may be called
    after a call to either of the integration functions
    ``ivp_adams_roots`` and ``ivp_adams_roots_revcom``.

    For full information please refer to the NAG Library document for
    d02qx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02qxf.html

    Parameters
    ----------
    neqf : int
        The number of first-order ordinary differential equations solved
        by the integration function. It must be the same argument `neqf`
        supplied to the setup function ``ivp_adams_setup`` and the
        integration functions ``ivp_adams_roots`` or
        ``ivp_adams_roots_revcom``.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_adams_setup`` and one of ``ivp_adams_roots`` or
        ``ivp_adams_roots_revcom``.

    Returns
    -------
    yp : float, ndarray, shape (`neqf`)
        The approximate derivative of the solution component y_i, as
        supplied in y_i on output from the integration function at the
        output value of t. These values are obtained by the evaluation
        of y' = f(x,y) except when the output value of the argument t in
        the call to the integration function is tout and `tcurr` !=
        tout, in which case they are obtained by interpolation.

    tcurr : float
        The value of the independent variable which the integrator has
        actually reached. `tcurr` will always be at least as far as the
        output value of the argument t (from the integration function)
        in the direction of integration, but may be further.

    hlast : float
        The last successful step size used by the integrator.

    hnext : float
        The next step size which the integration function would attempt.

    odlast : int
        The order of the method last used (successfully) by the
        integration function.

    odnext : int
        The order of the method which the integration function would
        attempt on the next step.

    nsucc : int
        The number of steps attempted by the integration function that
        have been successful since the start of the current problem.

    nfail : int
        The number of steps attempted by the integration function that
        have failed since the start of the current problem.

    tolfac : float
        A tolerance scale factor, `tolfac` >= 1.0, returned when the
        integration function exits with `errno` = 3. If rtol and atol
        are uniformly scaled up by a factor of `tolfac` and
        ``ivp_adams_setup`` is called, the next call to the integration
        function is deemed likely to succeed.

    badcmp : int
        If the integration function returned with `errno` = 4, `badcmp`
        specifies the index of the component which forced the error
        exit. Otherwise `badcmp` is 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `neqf` = *<value>* and neqf = *<value>* in
            ``ivp_adams_setup``.

            Constraint: `neqf` = neqf in ``ivp_adams_setup``.

        (`errno` 1)
            Neither of the appropriate two integrator functions has been
            called.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_adams_diag as b_ivp_adams_diag

    fun_name = 'naginterfaces.library.ode.ivp_adams_diag'

    _neqf = _b_utils._EngineIntScalarType(
        neqf, {'fun_name': fun_name, 'entity': 'neqf'},
    ).to_ctypes()
    _neqf_py = _b_utils._EngineIntScalarType.to_py(_neqf)
    _yp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neqf_py,),
    )
    _return_scalars = b_ivp_adams_diag(
        _neqf,
        _yp,
        comm,
    )
    _yp_py = _yp
    _tcurr_py = _return_scalars.tcurr
    _hlast_py = _return_scalars.hlast
    _hnext_py = _return_scalars.hnext
    _odlast_py = _return_scalars.odlast
    _odnext_py = _return_scalars.odnext
    _nsucc_py = _return_scalars.nsucc
    _nfail_py = _return_scalars.nfail
    _tolfac_py = _return_scalars.tolfac
    _badcmp_py = _return_scalars.badcmp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['yp', 'tcurr', 'hlast', 'hnext', 'odlast', 'odnext', 'nsucc', 'nfail', 'tolfac', 'badcmp'],
    )(
        _yp_py,
        _tcurr_py,
        _hlast_py,
        _hnext_py,
        _odlast_py,
        _odnext_py,
        _nsucc_py,
        _nfail_py,
        _tolfac_py,
        _badcmp_py,
    )
    return _return_data

def ivp_adams_rootdiag(neqg, comm):
    u"""
    Ordinary differential equations, initial value problem, root-finding
    diagnostics for ``ivp_adams_roots`` and ``ivp_adams_roots_revcom``.

    ``ivp_adams_rootdiag`` is a diagnostic function which may be called
    after a call to the integrator functions ``ivp_adams_roots`` or
    ``ivp_adams_roots_revcom``.

    For full information please refer to the NAG Library document for
    d02qy

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02qyf.html

    Parameters
    ----------
    neqg : int
        The number of event functions defined for the integration
        function. It must be the same argument `neqg` supplied to the
        setup function ``ivp_adams_setup`` and to the integration
        function (``ivp_adams_roots`` or ``ivp_adams_roots_revcom``).

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_adams_setup`` and one of ``ivp_adams_roots`` or
        ``ivp_adams_roots_revcom``.

    Returns
    -------
    index : int
        The index k of the event equation g_k(x,y,y') = 0 for which the
        root has been detected.

    itype : int
        Information about the root detected for the event equation
        defined by `index`. The possible values of `itype` with their
        interpretations are as follows:

        `itype` = 1

          A simple root, or lack of distinguishing information
          available.

        `itype` = 2

          A root of even multiplicity is believed to have been detected,
          that is no change in sign of the event function was found.

        `itype` = 3

          A high-order root of odd multiplicity.

        `itype` = 4

          A possible root, but due to high multiplicity or a clustering
          of roots accurate evaluation of the event function was
          prohibited by round-off error and/or cancellation.

        In general, the accuracy of the root is less reliable for values
        of `itype` > 1.

    events : int, ndarray, shape (`neqg`)
        Information about the kth event function on a very small
        interval containing the root, t (see ``ivp_adams_roots`` and
        ``ivp_adams_roots_revcom``), as output from the integration
        function. All roots lying in this interval are considered
        indistinguishable numerically and therefore should be regarded
        as defining a root at t. The possible values of `events`[k-1]
        with their interpretations are as follows:

        `events`[k-1] = 0

          The kth event function did not have a root.

        `events`[k-1] = -1

          The kth event function changed sign from positive to negative
          about a root, in the direction of integration.

        `events`[k-1] = 1

          The kth event function changed sign from negative to positive
          about a root, in the direction of integration.

        `events`[k-1] = 2

          A root was identified, but no change in sign was observed.

    resids : float, ndarray, shape (`neqg`)
        The value of the kth event function computed at the root, t (see
        ``ivp_adams_roots`` and ``ivp_adams_roots_revcom``).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `neqg` = *<value>* and neqg = *<value>* in
            ``ivp_adams_setup``.

            Constraint: `neqg` = neqg in ``ivp_adams_setup``.

        (`errno` 1)
            The integrator did not end at an event.

        (`errno` 1)
            Neither of the appropriate two integrator functions has been
            called.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import ivp_adams_rootdiag as b_ivp_adams_rootdiag

    fun_name = 'naginterfaces.library.ode.ivp_adams_rootdiag'

    _neqg = _b_utils._EngineIntScalarType(
        neqg, {'fun_name': fun_name, 'entity': 'neqg'},
    ).to_ctypes()
    _neqg_py = _b_utils._EngineIntScalarType.to_py(_neqg)
    _events = _b_utils._EngineIntArrayType.empty_ndarray(
        (_neqg_py,),
    )
    _resids = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neqg_py,),
    )
    _return_scalars = b_ivp_adams_rootdiag(
        _neqg,
        _events,
        _resids,
        comm,
    )
    _index_py = _return_scalars.index
    _itype_py = _return_scalars.itype
    _events_py = _events
    _resids_py = _resids
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['index', 'itype', 'events', 'resids'],
    )(
        _index_py,
        _itype_py,
        _events_py,
        _resids_py,
    )
    return _return_data

def ivp_adams_interp(neqf, twant, nwant, comm):
    u"""
    Ordinary differential equations, initial value problem,
    interpolation for ``ivp_adams_roots`` or ``ivp_adams_roots_revcom``.

    ``ivp_adams_interp`` interpolates components of the solution of a
    non-stiff system of first-order differential equations from
    information provided by the integrator functions ``ivp_adams_roots``
    or ``ivp_adams_roots_revcom``.

    For full information please refer to the NAG Library document for
    d02qz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02qzf.html

    Parameters
    ----------
    neqf : int
        The number of first-order ordinary differential equations being
        solved by the integration function. It must contain the same
        value as the argument `neqf` in a prior call to the setup
        function ``ivp_adams_setup``.

    twant : float
        The point at which components of the solution and derivative are
        to be evaluated. `twant` should not normally be an extrapolation
        point, that is `twant` should satisfy

        - told <= `twant` <= T,

        or if integration is proceeding in the negative direction

        - told >= `twant` >= T,

        where told is the previous integration point and is, to within
        rounding, tcurr -- hlast (see ``ivp_adams_diag``). Extrapolation
        is permitted but not recommended and `errno` = 2 is returned
        whenever extrapolation is attempted.

    nwant : int
        The number of components of the solution and derivative whose
        values at `twant` are required. The first `nwant` components are
        evaluated.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_adams_setup`` and one of ``ivp_adams_roots`` or
        ``ivp_adams_roots_revcom``.

    Returns
    -------
    ywant : float, ndarray, shape (`nwant`)
        The calculated value of the ith component of the solution at
        `twant`, for i = 1,2,...,`nwant`.

    ypwant : float, ndarray, shape (`nwant`)
        The calculated value of the ith component of the derivative at
        `twant`, for i = 1,2,...,`nwant`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nwant` = *<value>*.

            Constraint: `nwant` >= 1.

        (`errno` 1)
            On entry, `nwant` = *<value>* and `neqf` = *<value>*.

            Constraint: `nwant` <= `neqf`.

        (`errno` 1)
            On entry, `neqf` = *<value>* and neqf = *<value>* in
            ``ivp_adams_setup``.

            Constraint: `neqf` = neqf in ``ivp_adams_setup``.

        (`errno` 1)
            Neither of the appropriate two integrator functions has been
            called.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            No successful steps had been taken, so interpolation is
            impossible at `twant` = *<value>*.

        (`errno` 2)
            Extrapolation performed at `twant` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import ivp_adams_interp as b_ivp_adams_interp

    fun_name = 'naginterfaces.library.ode.ivp_adams_interp'

    _nwant = _b_utils._EngineIntScalarType(
        nwant, {'fun_name': fun_name, 'entity': 'nwant'},
    ).to_ctypes()
    _nwant_py = _b_utils._EngineIntScalarType.to_py(_nwant)
    _ywant = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nwant_py,),
    )
    _ypwant = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nwant_py,),
    )
    try:
        b_ivp_adams_interp(
            neqf,
            twant,
            _nwant,
            _ywant,
            _ypwant,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ywant_py = _ywant
    _ypwant_py = _ypwant
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ywant', 'ypwant'],
    )(
        _ywant_py,
        _ypwant_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bvp_fd_nonlin_gen(np, numbeg, nummix, tol, init, x, y, fcn, g, ijac, deleps, itrace, comm, jacobf=None, jacobg=None, jaceps=None, jacgep=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Ordinary differential equations, general nonlinear boundary value
    problem, finite difference technique with deferred correction,
    continuation facility.

    ``bvp_fd_nonlin_gen`` solves a two-point boundary value problem with
    general boundary conditions for a system of ordinary differential
    equations, using a deferred correction technique and Newton
    iteration.

    For full information please refer to the NAG Library document for
    d02ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02raf.html

    Parameters
    ----------
    np : int
        Must be set to the number of points to be used in the initial
        mesh.

    numbeg : int
        The number of left-hand boundary conditions (that is the number
        involving y(a) only).

    nummix : int
        The number of coupled boundary conditions (that is the number
        involving both y(a) and y(b)).

    tol : float
        A positive absolute error tolerance. If

            a = x_1 < x_2 < ... < x_`np` = b

        is the final mesh, z_j(x_i) is the jth component of the
        approximate solution at x_i, and y_j(x) is the jth component of
        the true solution of (1) and (2), then, except in extreme
        circumstances, it is expected that

            |z_j(x_i)-y_j(x_i)| <= `tol`, i = 1,2,...,`np` and j =
            1,2,...,n.

    init : int
        Indicates whether you wish to supply an initial mesh and
        approximate solution (`init` = 1) or whether default values are
        to be used, (`init` = 0).

    x : float, array-like, shape (mnp)
        You must set `x`[0] = a and `x`[`np`-1] = b. If `init` = 0 on
        entry a default equispaced mesh will be used, otherwise you must
        specify a mesh by setting `x`[i-1] = x_i, for i =
        2,3,...,`np`-1.

    y : float, array-like, shape (n, mnp)
        If `init` = 0, `y` need not be set.

        If `init` = 1, the array `y` must contain an initial
        approximation to the solution such that `y`[j-1,i-1] contains an
        approximation to

            y_j(x_i), i = 1,2,...,`np` and j = 1,2,...,n.

    fcn : callable f = fcn(x, eps, y, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i) at a general point x for a given value of epsilon, the
        continuation parameter (see [Description]).

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        eps : float
            epsilon, the value of the continuation parameter. This is 1
            if continuation is not being used.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the values of the dependent
            variables at x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            The values of the derivatives f_i evaluated at x given
            epsilon, for i = 1,2,...,n.

    g : callable bc = g(eps, ya, yb, data=None)
        `g` must evaluate the boundary conditions in equation (3) and
        place them in the array `bc`.

        Parameters
        ~~~~~~~~~~
        eps : float
            epsilon, the value of the continuation parameter. This is 1
            if continuation is not being used.

        ya : float, ndarray, shape (n)
            The value y_i(a), for i = 1,2,...,n.

        yb : float, ndarray, shape (n)
            The value y_i(b), for i = 1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        bc : float, array-like, shape (n)
            The values g_i(y(a),y(b),epsilon), for i = 1,2,...,n. These
            must be ordered as follows:

            - first, the conditions involving only y(a) (see `numbeg`);

            - next, the `nummix` coupled conditions involving both y(a)
              and y(b) (see `nummix`); and,

            - finally, the conditions involving only y(b)
              (n-`numbeg`-`nummix`).

    ijac : int
        Indicates whether or not you are supplying Jacobian evaluation
        functions.

        `ijac` != 0

          You must supply `jacobf` and `jacobg` and also, when
          continuation is used, `jaceps` and `jacgep`.

        `ijac` = 0

          Numerical differentiation is used to calculate the Jacobian
          and None may be used for the respective callbacks.

    deleps : float
        Must be given a value which specifies whether continuation is
        required. If `deleps` <= 0.0 or `deleps` >= 1.0 then it is
        assumed that continuation is not required. If 0.0 < `deleps` <
        1.0 then it is assumed that continuation is required unless
        `deleps` < sqrt(machine precision) when an error exit is taken.
        `deleps` is used as the increment epsilon_2-epsilon_1 (see (4))
        and the choice `deleps` = 0.1 is recommended.

    itrace : int
        If `itrace` = 0 warning messages be suppressed, otherwise
        warning messages will be printed (see [Errors]).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    jacobf : None or callable f = jacobf(x, eps, y, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jacobf` evaluates the Jacobian (df_i/dy_j), for j = 1,2,...,n,
        for i = 1,2,...,n, given x and y_j, for j = 1,2,...,n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        eps : float
            epsilon, the value of the continuation parameter. This is 1
            if continuation is not being used.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the values of the dependent
            variables at x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n, n)
            `f`[j-1,i-1] must be set to the value of df_i/dy_j,
            evaluated at the point (x,y), for j = 1,2,...,n, for i =
            1,2,...,n.

    jacobg : None or callable (aj, bj) = jacobg(eps, ya, yb, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jacobg` evaluates the Jacobians (dg_i/(dy_j(a))) and
        (dg_i/(dy_j(b))).

        The ordering of the rows of `aj` and `bj` must correspond to the
        ordering of the boundary conditions described in the
        specification of `g`.

        Parameters
        ~~~~~~~~~~
        eps : float
            epsilon, the value of the continuation parameter. This is 1
            if continuation is not being used.

        ya : float, ndarray, shape (n)
            The value y_i(a), for i = 1,2,...,n.

        yb : float, ndarray, shape (n)
            The value y_i(b), for i = 1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        aj : float, array-like, shape (n, n)
            `aj`[i-1,j-1] must be set to the value dg_i/(dy_j(a)), for j
            = 1,2,...,n, for i = 1,2,...,n.

        bj : float, array-like, shape (n, n)
            `bj`[i-1,j-1] must be set to the value dg_i/(dy_j(b)), for j
            = 1,2,...,n, for i = 1,2,...,n.

    jaceps : None or callable f = jaceps(x, eps, y, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jaceps` evaluates the derivative df_i/d epsilon given x and y
        if continuation is being used.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the independent variable.

        eps : float
            epsilon, the value of the continuation parameter.

        y : float, ndarray, shape (n)
            The solution values y_i, for i = 1,2,...,n, at the point x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            `f`[i-1] must contain the value df_i/d epsilon at the point
            (x,y), for i = 1,2,...,n.

    jacgep : None or callable bcep = jacgep(eps, ya, yb, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `jacgep` evaluates the derivatives dg_i/d epsilon if
        continuation is being used.

        Parameters
        ~~~~~~~~~~
        eps : float
            epsilon, the value of the continuation parameter.

        ya : float, ndarray, shape (n)
            The value of y_i(a), for i = 1,2,...,n.

        yb : float, ndarray, shape (n)
            The value of y_i(b), for i = 1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        bcep : float, array-like, shape (n)
            `bcep`[i-1] must contain the value of dg_i/d epsilon, for i
            = 1,2,...,n.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `y` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    np : int
        The number of points in the final mesh.

    x : float, ndarray, shape (mnp)
        `x`[0],`x`[1],...,`x`[`np`-1] define the final mesh (with the
        returned value of `np`) and `x`[0] = a and `x`[`np`-1] = b.

    y : float, ndarray, shape (n, mnp)
        The approximate solution z_j(x_i) satisfying (5) on the final
        mesh, that is

            `y`[j-1,i-1] = z_j(x_i), i = 1,2,...,`np` and j = 1,2,...,n,

        where `np` is the number of points in the final mesh. If an
        error has occurred then `y` contains the latest approximation to
        the solution. The remaining columns of `y` are not used.

    abt : float, ndarray, shape (n)
        `abt`[i-1], for i = 1,2,...,n, holds the largest estimated error
        (in magnitude) of the ith component of the solution over all
        mesh points.

    deleps : float
        An overestimate of the increment epsilon_p-epsilon_{p-1} (in
        fact the value of the increment which would have been tried if
        the restriction epsilon_p = 1 had not been imposed). If
        continuation was not requested then `deleps` = 0.0.

        If continuation is not requested then `jaceps` and `jacgep` may
        each be replaced by None.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry the mesh points are not in strictly ascending
            order.

            For i = *<value>*, mesh point i = *<value>*, but mesh point
            i+1 = *<value>*.

        (`errno` 1)
            On entry, `numbeg` = *<value>*, `nummix` = *<value>* and n =
            *<value>*.

            Constraint: `numbeg`+`nummix` <= n.

        (`errno` 1)
            On entry, `nummix` = *<value>*.

            Constraint: `nummix` >= 0.

        (`errno` 1)
            On entry, `numbeg` = *<value>* and n = *<value>*.

            Constraint: `numbeg` < n.

        (`errno` 1)
            On entry, `numbeg` = *<value>*.

            Constraint: `numbeg` >= 0.

        (`errno` 1)
            On entry, `x`[0] = *<value>* and `x`[`np`-1] = *<value>*.

            Constraint: `x`[0] < `x`[`np`-1].

        (`errno` 1)
            On entry, mnp = *<value>*.

            Constraint: mnp >= 32.

        (`errno` 1)
            On entry, `np` = *<value>* and mnp = *<value>*.

            Constraint: `np` <= mnp.

        (`errno` 1)
            On entry, `np` = *<value>*.

            Constraint: `np` >= 4.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            A finer mesh is required for the accuracy requested; that
            is, mnp = *<value>* is not large enough.

        (`errno` 3)
            The Newton iteration has failed to converge.

            This could be due to there being too few points in the
            initial mesh or to the initial approximate solution being
            too inaccurate.

            If this latter reason is suspected or you cannot make
            changes to prevent this error, you should use the function
            with a continuation facility instead.

        (`errno` 5)
            The Jacobian for the boundary conditions is singular.

            This may occur due to faulty coding of the Jacobian or, in
            some circumstances, to a zero initial choice of approximate
            solution.

        (`errno` 6)
            There is no dependence on the continuation parameter when
            continuation is being used. This can be due to faulty coding
            of derivatives with respect to the continuation parameter or
            to a zero initial choice of approximate solution.

        (`errno` 7)
            The continuation step is required to be less than machine
            precision for continuation to proceed. It is likely that
            either the problem has no solution for some value of the
            continuation parameter near the current value or that the
            problem is so difficult that even with continuation it is
            unlikely to be solved using this function. In the latter
            case using more mesh points initially may help.

        (`errno` 8)
            A serious error occurred in a call to the internal
            integrator.

            The error code internally was *<value>*.

            Please contact NAG.

        (`errno` 9)
            A continuation error occurred, but continuation is not being
            used.

            Please contact NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Newton iteration has reached round-off level.

            If desired accuracy has not been reached, `tol` is too small
            for this problem and this machine precision.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.ode import bvp_fd_nonlin_gen as b_bvp_fd_nonlin_gen

    fun_name = 'naginterfaces.library.ode.bvp_fd_nonlin_gen'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(x, eps, y, f, n, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                eps,
                y,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                eps,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_g_name = fun_name + ':g'

    def cb_g(eps, ya, yb, bc, n, data=None):
        """Wrapping user's g. Called as base g."""
        if data is not None:
            _bc_g = g(
                eps,
                ya,
                yb,
                data,
            )
        else:
            _bc_g = g(
                eps,
                ya,
                yb,
            )
        bc[:] = _utils._to_ndarray(
            {'fun_name': cb_g_name, 'returned': 'bc'},
            _bc_g,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_jacobf_name = fun_name + ':jacobf'

    def cb_jacobf(x, eps, y, f, n, data=None):
        """Wrapping user's jacobf. Called as base jacobf."""
        if data is not None:
            _f_jacobf = jacobf(
                x,
                eps,
                y,
                data,
            )
        else:
            _f_jacobf = jacobf(
                x,
                eps,
                y,
            )
        _f_jacobf_np = _utils._to_ndarray(
            {'fun_name': cb_jacobf_name, 'returned': 'f'},
            _f_jacobf,
            copy=False,
            exp_shape=(n, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_jacobf_name},
            ((_f_jacobf_np, 'f'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        f[:, :] = _f_jacobf_np

    cb_jacobg_name = fun_name + ':jacobg'

    def cb_jacobg(eps, ya, yb, aj, bj, n, data=None):
        """Wrapping user's jacobg. Called as base jacobg."""
        if data is not None:
            _aj_jacobg, _bj_jacobg = jacobg(
                eps,
                ya,
                yb,
                data,
            )
        else:
            _aj_jacobg, _bj_jacobg = jacobg(
                eps,
                ya,
                yb,
            )
        _aj_jacobg_np = _utils._to_ndarray(
            {'fun_name': cb_jacobg_name, 'returned': 'aj'},
            _aj_jacobg,
            copy=False,
            exp_shape=(n, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _bj_jacobg_np = _utils._to_ndarray(
            {'fun_name': cb_jacobg_name, 'returned': 'bj'},
            _bj_jacobg,
            copy=False,
            exp_shape=(n, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_jacobg_name},
            ((_aj_jacobg_np, 'aj'), (_bj_jacobg_np, 'bj')),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        aj[:, :] = _aj_jacobg_np
        bj[:, :] = _bj_jacobg_np

    cb_jaceps_name = fun_name + ':jaceps'

    def cb_jaceps(x, eps, y, f, n, data=None):
        """Wrapping user's jaceps. Called as base jaceps."""
        if data is not None:
            _f_jaceps = jaceps(
                x,
                eps,
                y,
                data,
            )
        else:
            _f_jaceps = jaceps(
                x,
                eps,
                y,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_jaceps_name, 'returned': 'f'},
            _f_jaceps,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_jacgep_name = fun_name + ':jacgep'

    def cb_jacgep(eps, ya, yb, bcep, n, data=None):
        """Wrapping user's jacgep. Called as base jacgep."""
        if data is not None:
            _bcep_jacgep = jacgep(
                eps,
                ya,
                yb,
                data,
            )
        else:
            _bcep_jacgep = jacgep(
                eps,
                ya,
                yb,
            )
        bcep[:] = _utils._to_ndarray(
            {'fun_name': cb_jacgep_name, 'returned': 'bcep'},
            _bcep_jacgep,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_y, 'y'),),
        spiked_sorder=spiked_sorder,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mnp_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mnp = _b_utils._EngineIntScalarType(
        _mnp_py, {'fun_name': fun_name, 'entity': '_mnp'},
    ).to_ctypes()
    _abt = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _b_utils._check_callable(g, {'fun_name': fun_name, 'entity': 'g'})
    _g = cb_g
    _b_utils._check_callable(jacobf, {'fun_name': fun_name, 'entity': 'jacobf'}, allow_none=True)
    _jacobf = None if jacobf is None else cb_jacobf
    _b_utils._check_callable(jacobg, {'fun_name': fun_name, 'entity': 'jacobg'}, allow_none=True)
    _jacobg = None if jacobg is None else cb_jacobg
    _b_utils._check_callable(jaceps, {'fun_name': fun_name, 'entity': 'jaceps'}, allow_none=True)
    _jaceps = None if jaceps is None else cb_jaceps
    _b_utils._check_callable(jacgep, {'fun_name': fun_name, 'entity': 'jacgep'}, allow_none=True)
    _jacgep = None if jacgep is None else cb_jacgep
    try:
        _return_scalars = b_bvp_fd_nonlin_gen(
            _n,
            _mnp,
            np,
            numbeg,
            nummix,
            tol,
            init,
            _x,
            _y,
            _abt,
            _fcn,
            _g,
            ijac,
            _jacobf,
            _jacobg,
            deleps,
            _jaceps,
            _jacgep,
            itrace,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _np_py = _return_scalars.np
    _x_py = _x
    _y_py = _y
    _deleps_py = _return_scalars.deleps
    _abt_py = _abt
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['np', 'x', 'y', 'abt', 'deleps'],
    )(
        _np_py,
        _x_py,
        _y_py,
        _abt_py,
        _deleps_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bvp_shoot_genpar_algeq(p, n, n1, pe, pf, dp, swp, icount, brkpts, bc, fcn, ymax, monlev, comm, e=None, eqn=None, constr=None, monit=None, prsol=None, data=None, io_manager=None):
    u"""
    Ordinary differential equations, boundary value problem, shooting
    and matching technique, subject to extra algebraic equations,
    general parameters to be determined.

    ``bvp_shoot_genpar_algeq`` solves a two-point boundary value problem
    for a system of first-order ordinary differential equations with
    boundary conditions, combined with additional algebraic equations.
    It uses initial value techniques and a modified Newton iteration in
    a shooting and matching method.

    For full information please refer to the NAG Library document for
    d02sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02saf.html

    Parameters
    ----------
    p : float, array-like, shape (m)
        `p`[i] must be set to an estimate of the ith argument, p_i, for
        i = 1,2,...,m.

    n : int
        n, the total number of differential equations.

    n1 : int
        n_1, the number of differential equations active in the matching
        process. The active equations must be placed last in the
        numbering in `fcn` and `bc`. The first `n`-`n1` equations are
        used as the driving equations.

    pe : float, array-like, shape (m)
        `pe`[i-1], for i = 1,2,...,m, must be set to a positive value
        for use in the convergence test in the ith argument p_i. See the
        description of `pf` for further details.

    pf : float, array-like, shape (m)
        `pf`[i-1], for i = 1,2,...,m, should be set to a 'floor' value
        in the convergence test on the ith argument p_i. If `pf`[i-1] <=
        0.0 on entry then it is set to the small positive value
        sqrt(epsilon) (where epsilon may in most cases be considered to
        be machine precision); otherwise it is used unchanged.

        The Newton iteration is presumed to have converged if a full
        Newton step is taken (`istate` = 1 in the specification of
        `monit`), the singular values of the Jacobian are not being
        significantly perturbed (also see `monit`) and if the Newton
        correction C_i satisfies

            |C_i| <= `pe`[i-1]*max(|p_i|, `pf`[i-1]), i = 1,2,...,m,

        where p_i is the current value of the ith argument.

        The values `pf`[i-1] are also used in determining the Newton
        iterates as discussed in [Description], see equation (6).

    dp : float, array-like, shape (m)
        A value to be used in perturbing the argument p_i in the
        numerical differentiation to estimate the Jacobian used in
        Newton's method. If `dp`[i] = 0.0 on entry, an estimate is made
        internally by setting

            `dp`[i] = sqrt(epsilon)*max(`pf`[i], |p_i|),

        where p_i is the initial value of the argument supplied by you
        and epsilon may in most cases be considered to be machine
        precision. The estimate of the Jacobian, J, is made using
        forward differences, that is for each i, for i = 1,2,...,m, p_i
        is perturbed to p_i+`dp`[i] and the ith column of J is estimated
        as

            (r(p_i+`dp`[i])-r(p_i))/`dp`[i]

        where the other components of p are unchanged (see (3) for the
        notation used). If this fails to produce a Jacobian with
        significant columns, backward differences are tried by
        perturbing p_i to p_i-`dp`[i] and if this also fails then
        central differences are used with p_i perturbed to
        p_i+10.0*`dp`[i]. If this also fails then the calculation of the
        Jacobian is abandoned. If the Jacobian has not previously been
        calculated then an error exit is taken. If an earlier estimate
        of the Jacobian is available then the current argument set, p_i,
        for i = 1,2,...,m, is abandoned in favour of the last argument
        set from which useful progress was made and the singular values
        of the Jacobian used at the point are modified before proceeding
        with the Newton iteration. You are recommended to use the
        default value `dp`[i] = 0.0 unless you have prior knowledge of a
        better choice. If any of the perturbations described are likely
        to lead to an unfortunate set of argument values then you should
        use `constr` to prevent such perturbations (all changes of
        arguments are checked by a call to `constr`).

    swp : float, array-like, shape (npoint, 6)
        `swp`[i-1,0] must contain an estimate for an initial step size
        for integration across the ith sub-interval [`x`[i-1],`x`[i]],
        for i = 1,2,...,npoint-1, (see `brkpts`). `swp`[i-1,0] should
        have the same sign as `x`[i]-`x`[i-1] if it is nonzero. If
        `swp`[i-1,0] = 0.0, on entry, a default value for the initial
        step size is calculated internally. This is the recommended mode
        of entry.

        `swp`[i-1,2] must contain a lower bound for the modulus of the
        step size on the ith sub-interval [`x`[ i -1],`x`[ i + 1 -1]],
        for i = 1,2,...,npoint-1.

        If `swp`[i-1,2] = 0.0 on entry, a very small default value is
        used.

        By setting `swp`[i-1,2] > 0.0 but smaller than the expected step
        sizes (assuming you have some insight into the likely step
        sizes) expensive integrations with arguments p far from the
        solution can be avoided.

        `swp`[i-1,1] must contain an upper bound on the modulus of the
        step size to be used in the integration on [`x`[ i -1],`x`[ i +
        1 -1]], for i = 1,2,...,npoint-1.

        If `swp`[i-1,1] = 0.0 on entry no bound is assumed.

        This is the recommended mode of entry unless the solution is
        expected to have important features which might be 'missed' in
        the integration if the step size were permitted to be chosen
        freely.

    icount : int
        An upper bound on the number of Newton iterations. If `icount` =
        0 on entry, no check on the number of iterations is made (this
        is the recommended mode of entry).

    brkpts : callable x = brkpts(npoint, p, data=None)
        `brkpts` must specify the break-points x_i, for i =
        1,2,...,`npoint`, which may depend on the arguments p_j, for j =
        1,2,...,m.

        Parameters
        ~~~~~~~~~~
        npoint : int
            2 plus the number of break-points in (a,b).

        p : float, ndarray, shape (m)
            The current estimate of the ith argument, for i = 1,2,...,m.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        x : float, array-like, shape (`npoint`)
            The ith break-point, for i = 1,2,...,`npoint`. The sequence
            (`x`[i]) must be strictly monotonic, that is either

                a = `x`[0] < `x`[1] < ... < `x`[`npoint`-1] = b

            or

                a = `x`[0] > `x`[1] > ... > `x`[`npoint`-1] = b.

    bc : callable (g1, g2) = bc(p, n, data=None)
        `bc` must place in `g1` and `g2` the boundary conditions at a
        and b respectively.

        Parameters
        ~~~~~~~~~~
        p : float, ndarray, shape (m)
            An estimate of the ith argument, p_i, for i = 1,2,...,m.

        n : int
            n, the number of differential equations.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        g1 : float, array-like, shape (`n`)
            The value of y_i(a), for i = 1,2,...,n, (where this may be a
            known value or a function of the parameters p_j, for j =
            1,2,...,m).

        g2 : float, array-like, shape (`n`)
            The value of y_i(b), for i = 1,2,...,n, (where these may be
            known values or functions of the parameters p_j, for j =
            1,2,...,m). If n > n_1, so that there are some driving
            equations, the first n-n_1 values of `g2` need not be set
            since they are never used.

    fcn : callable f = fcn(x, y, p, i, data=None)
        `fcn` must evaluate the functions f_i (i.e., the derivatives
        y'_i), for i = 1,2,...,n.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the value of the argument.

        y : float, ndarray, shape (n)
            y_i, for i = 1,2,...,n, the value of the argument.

        p : float, ndarray, shape (m)
            The current estimate of the ith argument p_i, for i =
            1,2,...,m.

        i : int
            Specifies the sub-interval [x_i,x_{i+1}] on which the
            derivatives are to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (n)
            The derivative of y_i, for i = 1,2,...,n, evaluated at x.
            `f`[i] may depend upon the parameters p_j, for j =
            1,2,...,m. If there are any driving equations (see
            [Description]) then these must be numbered first in the
            ordering of the components of `f`.

    ymax : float
        A non-negative value which is used as a bound on all values
        ||y(x)||_infinity where y(x) is the solution at any point x
        between `x`[0] and `x`[npoint-1] for the current arguments
        p_1,p_2,...,p_m. If this bound is exceeded the integration is
        terminated and the current arguments are rejected. Such a
        rejection will result in an error exit if it prevents the
        initial residual or Jacobian, or the final solution, being
        calculated. If `ymax` = 0 on entry, no bound on the solution y
        is used; that is the integrations proceed without any checking
        on the size of ||y||_infinity.

    monlev : int
        Setting `monlev` = 0 disables monitoring of the pseudo-Newton
        iteration. Setting `monlev` = 1 enables this monitoring.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    e : None or float, array-like, shape (`n`), optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 10^-5.

        Values for use in controlling the local error in the integration
        of the differential equations. If err_i is an estimate of the
        local error in y_i, for i = 1,2,...,n,

            |err_i| <= `e`[i-1]*max{sqrt(epsilon),|y_i|},

        where epsilon may in most cases be considered to be machine
        precision.

    eqn : None or callable e = eqn(q, p, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `eqn` is used to describe the additional algebraic equations to
        be solved in the determination of the parameters, p_i, for i =
        1,2,...,m.

        If there are no additional algebraic equations (i.e., m = n_1)
        then `eqn` is never called and may be None.

        Parameters
        ~~~~~~~~~~
        q : int
            The number of algebraic equations, m-n_1.

        p : float, ndarray, shape (m)
            The current estimate of the ith argument p_i, for i =
            1,2,...,m.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        e : float, array-like, shape (`q`)
            The vector of residuals, r_2(p), that is the amount by which
            the current estimates of the arguments fail to satisfy the
            algebraic equations.

    constr : None or callable retval = constr(p, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `constr` is used to prevent the pseudo-Newton iteration running
        into difficulty. `constr` should return the value True if the
        constraints are satisfied by the parameters p_1,p_2,...,p_m.

        Otherwise `constr` should return the value False. If None is
        supplied the function will act as if `constr` returns the value
        True at all times, and in the first instance this is recommended
        as the actual argument.

        Parameters
        ~~~~~~~~~~
        p : float, ndarray, shape (m)
            An estimate of the ith argument, p_i, for i = 1,2,...,m.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : bool
            Must be True if the constraints are satisfied by the
            parameters `p`.

    monit : None or callable monit(istate, iflag, ifail1, p, f, pnorm,
        pnorm1, eps, d, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` enables you to monitor the values of various quantities
        during the calculation. If `monlev` = 0 on entry no monitoring
        is performed and this argument is never referenced. Otherwise,
        if `monlev` = 1 on entry, you may supply None, to use
        NAG-defined default monitoring, or you may supply your own
        custom monitoring function. The function is called by
        ``bvp_shoot_genpar_algeq`` after every calculation of the norm
        ||`d`^(-1)J~^+r||_2 which determines the strategy of the Newton
        method, every time there is an internal error exit leading to a
        change of strategy, and before an error exit when calculating
        the initial Jacobian.

        Parameters
        ~~~~~~~~~~
        istate : int
            The state of the Newton iteration.

            `istate` = 0

              The calculation of the residual, Jacobian and
              ||`d`^(-1)J~^+r||_2 are taking place.

            `istate` = 1 to 5

              During the Newton iteration a factor of 2^(-`istate`+1) of
              the Newton step is being used to try to reduce the norm.

            `istate` = 6

              The current Newton step has been rejected and the Jacobian
              is being re-calculated.

            `istate` = -6 to -1

              An internal error exit has caused the rejection of the
              current set of argument values, p. -`istate` is the value
              which `istate` would have taken if the error had not
              occurred.

            `istate` = -7

              An internal error exit has occurred when calculating the
              initial Jacobian.

        iflag : int
            Whether or not the Jacobian being used has been calculated
            at the beginning of the current iteration. If the Jacobian
            has been updated then `iflag` = 1; otherwise `iflag` = 2.
            The Jacobian is only calculated when convergence to the
            current argument values has been slow.

        ifail1 : int
            If -6 <= `istate` <= -1, `ifail1` specifies the errno error
            number that would be produced were control returned to you.
            `ifail1` is unspecified for values of `istate` outside this
            range.

        p : float, ndarray, shape (m)
            The current estimate of the ith argument p_i, for i =
            1,2,...,m.

        f : float, ndarray, shape (m)
            r, the residual corresponding to the current argument
            values, provided 1 <= `istate` <= 5 or `istate` = -7. `f` is
            unspecified for other values of `istate`.

        pnorm : float
            A quantity against which all reductions in norm are
            currently measured.

        pnorm1 : float
            p, the norm of the current arguments. It is set for 1 <=
            `istate` <= 5 and is undefined for other values of `istate`.

        eps : float
            Gives some indication of the convergence rate. It is the
            current singular value modification factor. It is zero
            initially and whenever convergence is proceeding steadily.
            `eps` is epsilon^(3/8) or greater (where epsilon may in most
            cases be considered machine precision) when the singular
            values of J are approximately zero or when convergence is
            not being achieved. The larger the value of `eps` the worse
            the convergence rate. When `eps` becomes too large the
            Newton iteration is terminated.

        d : float, ndarray, shape (m)
            J, the singular values of the current modified Jacobian
            matrix. If `d`[m-1] is small relative to `d`[0] for a number
            of Jacobians corresponding to different argument values then
            the computed results should be viewed with suspicion. It
            could be that the matching equations do not depend
            significantly on some argument (which could be due to a
            programming error in `fcn`, `bc`, `brkpts` or `eqn`).
            Alternatively, the system of differential equations may be
            very ill-conditioned when viewed as an initial value
            problem, in which case ``bvp_shoot_genpar_algeq`` is
            unsuitable. This may also be indicated by some singular
            values being very large. These values of `d`[i-1], for i =
            1,2,...,m, should not be changed.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    prsol : None or callable z = prsol(z, y, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `prsol` can be used to obtain values of the solution y at a
        selected point z by integration across the final range
        [`x`[0],`x`[ `npoint` -1]].

        If no output is required None can be used as the actual
        argument.

        Parameters
        ~~~~~~~~~~
        z : float
            Contains x_1 on the first call. On subsequent calls `z`
            contains its previous output value.

        y : float, ndarray, shape (n)
            The solution value y_i, for i = 1,2,...,n, at z.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        z : float
            The next point at which output is required. The new point
            must be nearer `x`[`npoint`-1] than the old.

            If `z` is set to a point outside [`x`[0],`x`[ `npoint` -1]]
            the process stops and control returns from
            ``bvp_shoot_genpar_algeq`` to the (sub)program from which
            ``bvp_shoot_genpar_algeq`` is called.

            Otherwise the next call to `prsol` is made by
            ``bvp_shoot_genpar_algeq`` at the point `z`, with solution
            values y_1,y_2,...,y_n at `z` contained in `y`.

            If `z` is set to `x`[`npoint`-1] exactly, the final call to
            `prsol` is made with y_1,y_2,...,y_n as values of the
            solution at `x`[`npoint`-1] produced by the integration.

            In general the solution values obtained at `x`[`npoint`-1]
            from `prsol` will differ from the values obtained at this
            point by a call to `bc`.

            The difference between the two solutions is the residual r.

            You are reminded that the points
            `x`[0],`x`[1],...,`x`[`npoint`-1] are available in the
            locations `swp`[0,3],`swp`[1,3],...,`swp`[`npoint`-1,3] at
            all times.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    p : float, ndarray, shape (m)
        The corrected value for the ith argument, unless an error has
        occurred, when it contains the last calculated value of the
        argument.

    pf : float, ndarray, shape (m)
        The values actually used.

    dp : float, ndarray, shape (m)
        The values actually used.

    swp : float, ndarray, shape (npoint, 6)
        `swp`[i-1,0] contains the initial step size used on the last
        integration on [`x`[ i -1],`x`[ i + 1 -1]], for i =
        1,2,...,npoint-1, (excluding integrations during the calculation
        of the Jacobian).

        `swp`[i-1,1], for i = 1,2,...,npoint-1, is usually unchanged.

        If the maximum step size `swp`[i-1,1] is so small or the length
        of the range [`x`[ i -1],`x`[ i + 1 -1]] is so short that on the
        last integration the step size was not controlled in the main by
        the size of the error tolerances `e`[i-1] but by these other
        factors, `swp`[npoint-1,1] is set to the floating-point value of
        i if the problem last occurred in [`x`[ i -1],`x`[ i + 1 -1]].

        Any results obtained when this value is returned as nonzero
        should be viewed with caution.

        `swp`[i-1,2], for i = 1,2,...,npoint-1, are unchanged.

        If an error exit with `errno` = 4, 5 or 6 (see [Errors]) occurs
        on the integration made from `x`[i-1] to `x`[i] the
        floating-point value of i is returned in `swp`[npoint-1,0].

        The actual point x in [`x`[ i -1],`x`[ i + 1 -1]] where the
        error occurred is returned in `swp`[0,4] (see also the
        specification of `w`).

        The floating-point value of npoint is returned in
        `swp`[npoint-1,0] if the error exit is caused by a call to `bc`.

        If an error exit occurs when estimating the Jacobian matrix
        (`errno` = 7, 8, 9, 10, 11 or 12, see [Errors]) and if argument
        p_i was the cause of the failure then on exit `swp`[npoint-1,0]
        contains the floating-point value of i.

        `swp`[i-1,3] contains the point `x`[i-1], for i =
        1,2,...,npoint, used at the solution p or at the final values of
        p if an error occurred.

        `swp` is also partly used as workspace.

    w : float, ndarray, shape (max(`n`, m), 3*m+12+max(11, m))
        In the case of an error exit of the type where the point of
        failure is returned in `swp`[0,4], the solution at this point of
        failure is returned in `w`[i-1,0], for i = 1,2,...,n.

        Otherwise `w` is used for workspace.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `monlev` = *<value>*.

            Constraint: `monlev` = 0 or 1.

        (`errno` 1)
            On entry, an element of the parameter convergence control
            array is zero or negative.

        (`errno` 1)
            On entry an element of the local error control array is zero
            or negative.

        (`errno` 1)
            On entry, `ymax` = *<value>*.

            Constraint: `ymax` >= 0.0.

        (`errno` 1)
            On entry, npoint = *<value>*.

            Constraint: npoint >= 2.

        (`errno` 1)
            On entry, `icount` = *<value>*.

            Constraint: `icount` >= 0.

        (`errno` 1)
            On entry, m = *<value>* and `n1` = *<value>*.

            Constraint: m >= `n1`.

        (`errno` 1)
            On entry, `n` = *<value>* and `n1` = *<value>*.

            Constraint: `n` >= `n1`.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 2)
            The constraints have been violated by the initial
            parameters.

        (`errno` 3)
            The sequence of break-points is not strictly monotonic in
            their initial specification.

        (`errno` 4)
            In the integration from first to last break-point with
            initial or final parameters, the step size was reduced too
            far for the integration to proceed. Consider reversing the
            order of break-points or relaxing the local error control.
            If this error exit still results, either this function is
            not a suitable method for solving the problem, or the
            initial choice of parameters is very poor.

        (`errno` 5)
            In the integration from first to last break-point with
            initial or final parameters, a suitable initial step could
            not be found on one of the intervals. Consider reversing the
            order of break-points or relaxing the local error control.
            If this error exit still results, either this function is
            not suitable for solving the problem, or the initial choice
            of parameters is very poor.

        (`errno` 6)
            During integration the solution exceeded `ymax` in
            magnitude, where, on entry, `ymax` = *<value>*.

        (`errno` 7)
            On calculating the initial approximation to the Jacobian,
            the constraints were violated.

        (`errno` 8)
            On perturbing the parameters when calculating the initial
            approximation to the Jacobian, the monotonicity condition on
            break-points is violated.

        (`errno` 9)
            An initial step-length could be found for integration to
            proceed with the current parameters.

        (`errno` 10)
            The step-length required to calculate the Jacobian to
            sufficient accuracy is too small.

        (`errno` 11)
            On calculating the initial approximation to the Jacobian,
            the solution to the system of differential equations
            exceeded `ymax` in magnitude, where, on entry, `ymax` =
            *<value>*.

        (`errno` 12)
            The Jacobian has an insignificant column. Make sure that the
            solution vector depends on all the parameters.

        (`errno` 13)
            An internal singular value decomposition has failed. This
            error can be avoided by changing the initial parameter
            estimates.

        (`errno` 14)
            The Newton iteration has failed to converge. This can
            indicate a poor initial choice of parameters or a very
            difficult problem. Consider varying elements of the
            parameter convergence control if the residuals are small;
            otherwise vary initial parameter estimates.

        (`errno` 15)
            The number of iterations permitted by `icount` has been
            exceeded where this was set to a positive value on entry.
            `icount` = *<value>*.

        (`errno` 16)
            Internal error in calculating Jacobian. Please contact NAG.

        (`errno` 17)
            Internal error in calculating residual. Please contact NAG.

        (`errno` 18)
            Internal error in integration. Please contact NAG.

        (`errno` 19)
            Internal error in Newton method. Please contact NAG.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import bvp_shoot_genpar_algeq as b_bvp_shoot_genpar_algeq

    fun_name = 'naginterfaces.library.ode.bvp_shoot_genpar_algeq'

    cb_range_name = fun_name + ':range'

    def cb_range(x, npoint, p, m, data=None):
        """Wrapping user's range. Called as base range."""
        if data is not None:
            _x_range = brkpts(
                npoint,
                p,
                data,
            )
        else:
            _x_range = brkpts(
                npoint,
                p,
            )
        x[:] = _utils._to_ndarray(
            {'fun_name': cb_range_name, 'returned': 'x'},
            _x_range,
            copy=False,
            exp_shape=(npoint,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_bc_name = fun_name + ':bc'

    def cb_bc(g1, g2, p, m, n, data=None):
        """Wrapping user's bc. Called as base bc."""
        if data is not None:
            _g1_bc, _g2_bc = bc(
                p,
                n,
                data,
            )
        else:
            _g1_bc, _g2_bc = bc(
                p,
                n,
            )
        g1[:] = _utils._to_ndarray(
            {'fun_name': cb_bc_name, 'returned': 'g1'},
            _g1_bc,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        g2[:] = _utils._to_ndarray(
            {'fun_name': cb_bc_name, 'returned': 'g2'},
            _g2_bc,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(x, y, f, n, p, m, i, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _f_fcn = fcn(
                x,
                y,
                p,
                i,
                data,
            )
        else:
            _f_fcn = fcn(
                x,
                y,
                p,
                i,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'f'},
            _f_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_eqn_name = fun_name + ':eqn'

    def cb_eqn(e, q, p, m, data=None):
        """Wrapping user's eqn. Called as base eqn."""
        if data is not None:
            _e_eqn = eqn(
                q,
                p,
                data,
            )
        else:
            _e_eqn = eqn(
                q,
                p,
            )
        e[:] = _utils._to_ndarray(
            {'fun_name': cb_eqn_name, 'returned': 'e'},
            _e_eqn,
            copy=False,
            exp_shape=(q,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_constr(p, m, data=None):
        """Wrapping user's constr. Called as base constr."""
        if data is not None:
            _retval_constr = constr(
                p,
                data,
            )
        else:
            _retval_constr = constr(
                p,
            )
        return _retval_constr

    def cb_monit(istate, iflag, ifail1, p, m, f, pnorm, pnorm1, eps, d, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                istate,
                iflag,
                ifail1,
                p,
                f,
                pnorm,
                pnorm1,
                eps,
                d,
                data,
            )
        else:
            monit(
                istate,
                iflag,
                ifail1,
                p,
                f,
                pnorm,
                pnorm1,
                eps,
                d,
            )

    def cb_prsol(z, y, n, data=None):
        """Wrapping user's prsol. Called as base prsol."""
        if data is not None:
            _z_prsol = prsol(
                z,
                y,
                data,
            )
        else:
            _z_prsol = prsol(
                z,
                y,
            )
        _z_prsol_base = _z_prsol
        return _z_prsol_base

    _swp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'swp'},
        swp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_swp, 'swp'),),
    )
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _pe = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'pe'},
        pe,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'pf'},
        pf,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _dp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dp'},
        dp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npoint_py = _b_utils._EngineFloat64ArrayType.get_shape(_swp, {'fun_name': fun_name, 'entity': 'swp'}, exp_rank=2)[0]
    _npoint = _b_utils._EngineIntScalarType(
        _npoint_py, {'fun_name': fun_name, 'entity': '_npoint'},
    ).to_ctypes()
    _b_utils._check_callable(brkpts, {'fun_name': fun_name, 'entity': 'brkpts'})
    _brkpts = cb_range
    _b_utils._check_callable(bc, {'fun_name': fun_name, 'entity': 'bc'})
    _bc = cb_bc
    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _b_utils._check_callable(eqn, {'fun_name': fun_name, 'entity': 'eqn'}, allow_none=True)
    _eqn = None if eqn is None else cb_eqn
    _b_utils._check_callable(constr, {'fun_name': fun_name, 'entity': 'constr'}, allow_none=True)
    _constr = None if constr is None else cb_constr
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _b_utils._check_callable(prsol, {'fun_name': fun_name, 'entity': 'prsol'}, allow_none=True)
    _prsol = None if prsol is None else cb_prsol
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_n_py, _m_py), 3*_m_py+12+max(11, _m_py)),
        sorder=_m_sorder,
    )
    b_bvp_shoot_genpar_algeq(
        _p,
        _m,
        _n,
        n1,
        _pe,
        _pf,
        _e,
        _dp,
        _npoint,
        _swp,
        icount,
        _brkpts,
        _bc,
        _fcn,
        _eqn,
        _constr,
        ymax,
        monlev,
        _monit,
        _prsol,
        _w,
        comm,
        data,
        io_manager,
    )
    _p_py = _p
    _pf_py = _pf
    _dp_py = _dp
    _swp_py = _swp
    _w_py = _w
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'pf', 'dp', 'swp', 'w'],
    )(
        _p_py,
        _pf_py,
        _dp_py,
        _swp_py,
        _w_py,
    )
    return _return_data

def bvp_coll_nth_comp(m, l, x0, x1, k1, kp, coeff, bdyc, data=None, spiked_sorder='C'):
    u"""
    nth-order linear ordinary differential equations, boundary value
    problem, collocation and least squares.

    ``bvp_coll_nth_comp`` solves a system of linear ordinary
    differential equations by least squares fitting of a series of
    Chebyshev polynomials using collocation.

    For full information please refer to the NAG Library document for
    d02tg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02tgf.html

    Parameters
    ----------
    m : int, array-like, shape (n)
        `m`[i] must be set to the highest order derivative occurring in
        the ith equation, for i = 1,2,...,n.

    l : int, array-like, shape (n)
        `l`[i] must be set to the number of boundary conditions
        associated with the ith equation, for i = 1,2,...,n.

    x0 : float
        The left-hand boundary, x_0.

    x1 : float
        The right-hand boundary, x_1.

    k1 : int
        The number of coefficients, k_1, to be returned in the Chebyshev
        series representation of the solution (hence, the degree of the
        polynomial approximation is `k1`-1).

    kp : int
        The number of collocation points to be used, k_p.

    coeff : callable (a, rhs) = coeff(x, i, a, rhs, data=None)
        `coeff` defines the system of differential equations (see
        [Description]).

        It must evaluate the coefficient functions f_kj^i(x) and the
        right-hand side function r^i(x) of the ith equation at a given
        point.

        Only nonzero entries of the array `a` and `rhs` need be
        specifically assigned, since all elements are set to zero by
        ``bvp_coll_nth_comp`` before calling `coeff`.

        Parameters
        ~~~~~~~~~~
        x : float
            x, the point at which the functions must be evaluated.

        i : int
            The equation for which the coefficients and right-hand side
            are to be evaluated.

        a : float, ndarray, shape (ia, ia1)
            All elements of `a` are set to zero.

        rhs : float
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        a : float, array-like, shape (ia, ia1)
            `a`[k-1,j-1] must contain the value f_kj^i(x), for 1 <= k <=
            n, 1 <= j <= m_i+1.

        rhs : float
            It must contain the value r^i(x).

    bdyc : callable (x, a, rhs) = bdyc(i, j, a, rhs, data=None)
        `bdyc` defines the boundary conditions (see [Description]).

        It must evaluate the coefficient functions f_kj^ij and
        right-hand side function r^(ij) in the jth boundary condition
        associated with the ith equation, at the point x^(ij) at which
        the boundary condition is applied.

        Only nonzero entries of the array `a` and `rhs` need be
        specifically assigned, since all elements are set to zero by
        ``bvp_coll_nth_comp`` before calling `bdyc`.

        Parameters
        ~~~~~~~~~~
        i : int
            The differential equation with which the condition is
            associated.

        j : int
            The boundary condition for which the coefficients and
            right-hand side are to be evaluated.

        a : float, ndarray, shape (ia, ia1)
            All elements of `a` are set to zero.

        rhs : float
            Is set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        x : float
            x^(ij), the value at which the boundary condition is
            applied.

        a : float, array-like, shape (ia, ia1)
            The value f_kj^ij(x^(ij)), for 1 <= k <= n, 1 <= j <= m_i+1.

        rhs : float
            The value r^(ij)(x^(ij)).

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `a` in `bdyc` or `a` in `coeff` are spiked (i.e., have unit
        extent in all but one dimension, or have size 1),
        `spiked_sorder` selects the storage order to associate with them
        in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    c : float, ndarray, shape (`k1`, n)
        The kth column of `c` contains the computed Chebyshev
        coefficients of the kth component of the solution, y_k; that is,
        the computed solution is:

            y_k = ⎲_i = 1^k_1`c`[i-1,k-1]T_{i-1}(x), 1 <= k <= n,

        where T_i(x) is the Chebyshev polynomial of the first kind and ⎲
        denotes that the first coefficient, `c`[0,k-1], is halved.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kp` = *<value>* and `k1` = *<value>*.

            Constraint: `kp`+1 >= `k1`.

        (`errno` 1)
            On entry, `k1` = *<value>* and max(`m`[i-1]) = *<value>*.

            Constraint: `k1` >= 1+max(`m`[i-1]).

        (`errno` 1)
            On entry, one of the equations is of order less than 1.

        (`errno` 1)
            On entry, `x1` = *<value>* and `x0` = *<value>*.

            Constraint: `x1` > `x0`.

        (`errno` 1)
            On entry, a negative number of boundary conditions was set
            for one of the system equations.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            Either the boundary conditions are not linearly independent,
            or the coefficient matrix is rank deficient. Increasing the
            number of collocation points may overcome this latter
            problem.

        (`errno` 4)
            Iterative refinement in the least squares solution has
            failed to converge. The coefficient matrix is too
            ill-conditioned.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import bvp_coll_nth_comp as b_bvp_coll_nth_comp

    fun_name = 'naginterfaces.library.ode.bvp_coll_nth_comp'

    cb_coeff_name = fun_name + ':coeff'

    def cb_coeff(x, i, a, ia, ia1, rhs, data=None):
        """Wrapping user's coeff. Called as base coeff."""
        if data is not None:
            _a_coeff, _rhs_coeff = coeff(
                x,
                i,
                a,
                rhs,
                data,
            )
        else:
            _a_coeff, _rhs_coeff = coeff(
                x,
                i,
                a,
                rhs,
            )
        _a_coeff_np = _utils._to_ndarray(
            {'fun_name': cb_coeff_name, 'returned': 'a'},
            _a_coeff,
            copy=False,
            exp_shape=(ia, ia1),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _rhs_coeff_base = _rhs_coeff
        _utils._check_callback_sorder(
            {'fun_name': cb_coeff_name},
            ((_a_coeff_np, 'a'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        a[:, :] = _a_coeff_np
        return _rhs_coeff_base

    cb_bdyc_name = fun_name + ':bdyc'

    def cb_bdyc(i, j, a, ia, ia1, rhs, data=None):
        """Wrapping user's bdyc. Called as base bdyc."""
        if data is not None:
            _x_bdyc, _a_bdyc, _rhs_bdyc = bdyc(
                i,
                j,
                a,
                rhs,
                data,
            )
        else:
            _x_bdyc, _a_bdyc, _rhs_bdyc = bdyc(
                i,
                j,
                a,
                rhs,
            )
        _a_bdyc_np = _utils._to_ndarray(
            {'fun_name': cb_bdyc_name, 'returned': 'a'},
            _a_bdyc,
            copy=False,
            exp_shape=(ia, ia1),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _rhs_bdyc_base = _rhs_bdyc
        _x_bdyc_base = _x_bdyc
        _utils._check_callback_sorder(
            {'fun_name': cb_bdyc_name},
            ((_a_bdyc_np, 'a'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        a[:, :] = _a_bdyc_np
        return (
            _x_bdyc_base,
            _rhs_bdyc_base,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _m = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'm'},
        m,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_m, {'fun_name': fun_name, 'entity': 'm'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _l = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'l'},
        l,
        copy=False,
    )
    _k1 = _b_utils._EngineIntScalarType(
        k1, {'fun_name': fun_name, 'entity': 'k1'},
    ).to_ctypes()
    _k1_py = _b_utils._EngineIntScalarType.to_py(_k1)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k1_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_utils._check_callable(coeff, {'fun_name': fun_name, 'entity': 'coeff'})
    _coeff = cb_coeff
    _b_utils._check_callable(bdyc, {'fun_name': fun_name, 'entity': 'bdyc'})
    _bdyc = cb_bdyc
    b_bvp_coll_nth_comp(
        _n,
        _m,
        _l,
        x0,
        x1,
        _k1,
        kp,
        _c,
        _coeff,
        _bdyc,
        data,
    )
    _c_py = _c
    return _c_py

def bvp_coll_nlin_solve(ffun, fjac, gafun, gbfun, gajac, gbjac, guess, comm, data=None, spiked_sorder='C'):
    u"""
    Ordinary differential equations, general nonlinear boundary value
    problem, collocation technique (thread safe).

    ``bvp_coll_nlin_solve`` solves a general two-point boundary value
    problem for a nonlinear mixed order system of ordinary differential
    equations.

    For full information please refer to the NAG Library document for
    d02tl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02tlf.html

    Parameters
    ----------
    ffun : callable f = ffun(x, y, m, data=None)
        `ffun` must evaluate the functions f_i for given values
        x,z(y(x)).

        Parameters
        ~~~~~~~~~~
        x : float
            x, the independent variable.

        y : float, ndarray, shape (neq, :)
            `y`[i-1,j] contains y_i^(j)(x), for j = 0,1,...,`m`[i-1]-1,
            for i = 1,2,...,neq.

            Note:  y_i^(0)(x) = y_i(x).

        m : int, ndarray, shape (neq)
            `m`[i-1] contains m_i, the order of the ith differential
            equation, for i = 1,2,...,neq.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (neq)
            `f`[i-1] must contain f_i, for i = 1,2,...,neq.

    fjac : callable dfdy = fjac(x, y, m, dfdy, data=None)
        `fjac` must evaluate the partial derivatives of f_i with respect
        to the elements of

        z(y(x)) =
        (y_1(x),y_1^1(x),...,y_1^(m_1-1)(x),y_2(x),...,y_n^(m_n-1)(x)).

        Parameters
        ~~~~~~~~~~
        x : float
            x, the independent variable.

        y : float, ndarray, shape (neq, :)
            `y`[i-1,j] contains y_i^(j)(x), for j = 0,1,...,`m`[i-1]-1,
            for i = 1,2,...,neq.

            Note:  y_i^(0)(x) = y_i(x).

        m : int, ndarray, shape (neq)
            `m`[i-1] contains m_i, the order of the ith differential
            equation, for i = 1,2,...,neq.

        dfdy : float, ndarray, shape (neq, neq, :)
            Set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        dfdy : float, array-like, shape (neq, neq, :)
            `dfdy`[i-1,j-1,k] must contain the partial derivative of f_i
            with respect to y_j^(k), for k = 0,1,...,`m`[j-1]-1, for j =
            1,2,...,neq, for i = 1,2,...,neq. Only nonzero partial
            derivatives need be set.

    gafun : callable ga = gafun(ya, m, nlbc, data=None)
        `gafun` must evaluate the boundary conditions at the left-hand
        end of the range, that is functions g_i(z(y(a))) for given
        values of z(y(a)).

        Parameters
        ~~~~~~~~~~
        ya : float, ndarray, shape (neq, :)
            `ya`[i-1,j] contains y_i^(j)(a), for j = 0,1,...,`m`[i-1]-1,
            for i = 1,2,...,neq.

            Note:  y_i^(0)(a) = y_i(a).

        m : int, ndarray, shape (neq)
            `m`[i-1] contains m_i, the order of the ith differential
            equation, for i = 1,2,...,neq.

        nlbc : int
            The number of boundary conditions at a.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        ga : float, array-like, shape (`nlbc`)
            `ga`[i-1] must contain g_i(z(y(a))), for i = 1,2,...,`nlbc`.

    gbfun : callable gb = gbfun(yb, m, nrbc, data=None)
        `gbfun` must evaluate the boundary conditions at the right-hand
        end of the range, that is functions g__i(z(y(b))) for given
        values of z(y(b)).

        Parameters
        ~~~~~~~~~~
        yb : float, ndarray, shape (neq, :)
            `yb`[i-1,j] contains y_i^(j)(b), for j = 0,1,...,`m`[i-1]-1,
            for i = 1,2,...,neq.

            Note:  y_i^(0)(b) = y_i(b).

        m : int, ndarray, shape (neq)
            `m`[i-1] contains m_i, the order of the ith differential
            equation, for i = 1,2,...,neq.

        nrbc : int
            The number of boundary conditions at b.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        gb : float, array-like, shape (`nrbc`)
            `gb`[i-1] must contain g__i(z(y(b))), for i =
            1,2,...,`nrbc`.

    gajac : callable dgady = gajac(ya, m, dgady, data=None)
        `gajac` must evaluate the partial derivatives of g_i(z(y(a)))
        with respect to the elements of z(y(a)) =
        (y_1(a),y_1^1(a),...,y_1^(m_1-1)(a),y_2(a),...,y_n^(m_n-1)(a)).

        Parameters
        ~~~~~~~~~~
        ya : float, ndarray, shape (neq, :)
            `ya`[i-1,j] contains y_i^(j)(a), for j = 0,1,...,`m`[i-1]-1,
            for i = 1,2,...,neq.

            Note:  y_i^(0)(a) = y_i(a).

        m : int, ndarray, shape (neq)
            `m`[i-1] contains m_i, the order of the ith differential
            equation, for i = 1,2,...,neq.

        dgady : float, ndarray, shape (nlbc, neq, :)
            Set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        dgady : float, array-like, shape (nlbc, neq, :)
            `dgady`[i-1,j-1,k] must contain the partial derivative of
            g_i(z(y(a))) with respect to y_j^(k)(a), for k =
            0,1,...,`m`[j-1]-1, for j = 1,2,...,neq, for i =
            1,2,...,`nlbc`. Only nonzero partial derivatives need be
            set.

    gbjac : callable dgbdy = gbjac(yb, m, dgbdy, data=None)
        `gbjac` must evaluate the partial derivatives of g__i(z(y(b)))
        with respect to the elements of z(y(b)) =
        (y_1(b),y_1^1(b),...,y_1^(m_1-1)(b),y_2(b),...,y_n^(m_n-1)(b)).

        Parameters
        ~~~~~~~~~~
        yb : float, ndarray, shape (neq, :)
            `yb`[i-1,j] contains y_i^(j)(b), for j = 0,1,...,`m`[i-1]-1,
            for i = 1,2,...,neq.

            Note:  y_i^(0)(b) = y_i(b).

        m : int, ndarray, shape (neq)
            `m`[i-1] contains m_i, the order of the ith differential
            equation, for i = 1,2,...,neq.

        dgbdy : float, ndarray, shape (nrbc, neq, :)
            Set to zero.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        dgbdy : float, array-like, shape (nrbc, neq, :)
            `dgbdy`[i-1,j-1,k] must contain the partial derivative of
            g__i(z(y(b))) with respect to y_j^(k)(b), for k =
            0,1,...,`m`[j-1]-1, for j = 1,2,...,neq, for i =
            1,2,...,`nrbc`. Only nonzero partial derivatives need be
            set.

    guess : callable (y, dym) = guess(x, m, mmax, data=None)
        `guess` must return initial approximations for the solution
        components y_i^(j) and the derivatives y_i^(m_i), for j =
        0,1,...,`m`[i-1]-1, for i = 1,2,...,neq.

        Try to compute each derivative y_i^(m_i) such that it
        corresponds to your approximations to y_i^(j), for j =
        0,1,...,`m`[i-1]-1.

        You should not call `ffun` to compute y_i^(m_i).

        Parameters
        ~~~~~~~~~~
        x : float
            x, the independent variable; x in [a,b].

        m : int, ndarray, shape (neq)
            `m`[i-1] contains m_i, the order of the ith differential
            equation, for i = 1,2,...,neq.

        mmax : int
            max(`m`[i]).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        y : float, array-like, shape (neq, mmax)
            `y`[i-1,j] must contain y_i^(j)(x), for j =
            0,1,...,`m`[i-1]-1, for i = 1,2,...,neq.

            Note:  y_i^(0)(x) = y_i(x).

        dym : float, array-like, shape (neq)
            `dym`[i-1] must contain y_i^(m_i)(x), for i = 1,2,...,neq.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bvp_coll_nlin_setup``.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `dfdy` in `fjac`, `dgady` in `gajac` or `y` in `guess` are
        spiked (i.e., have unit extent in all but one dimension, or have
        size 1), `spiked_sorder` selects the storage order to associate
        with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Either the setup function has not been called or the
            communication arrays have become corrupted. No solution will
            be computed.

        (`errno` 2)
            Numerical singularity has been detected in the Jacobian used
            in the Newton iteration.

            No results have been generated. Check the coding of the
            functions for calculating the Jacobians of system and
            boundary conditions.

        (`errno` 3)
            All Newton iterations that have been attempted have failed
            to converge.

            No results have been generated. Check the coding of the
            functions for calculating the Jacobians of system and
            boundary conditions.

            Try to provide a better initial solution approximation.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            A Newton iteration has failed to converge. The computation
            has not succeeded but results have been returned for an
            intermediate mesh on which convergence was achieved.

            These results should be treated with extreme caution.

        (`errno` 5)
            The expected number of sub-intervals required to continue
            the computation exceeds the maximum specified: *<value>*.

            Results have been generated which may be useful.

            Try increasing this number or relaxing the error
            requirements.
    """

    from ..base.ode import bvp_coll_nlin_solve as b_bvp_coll_nlin_solve

    fun_name = 'naginterfaces.library.ode.bvp_coll_nlin_solve'

    cb_ffun_name = fun_name + ':ffun'

    def cb_ffun(x, y, neq, m, mmax, f, data=None):
        """Wrapping user's ffun. Called as base ffun."""
        if data is not None:
            _f_ffun = ffun(
                x,
                y,
                m,
                data,
            )
        else:
            _f_ffun = ffun(
                x,
                y,
                m,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_ffun_name, 'returned': 'f'},
            _f_ffun,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_fjac_name = fun_name + ':fjac'

    def cb_fjac(x, y, neq, m, mmax, dfdy, data=None):
        """Wrapping user's fjac. Called as base fjac."""
        if data is not None:
            _dfdy_fjac = fjac(
                x,
                y,
                m,
                dfdy,
                data,
            )
        else:
            _dfdy_fjac = fjac(
                x,
                y,
                m,
                dfdy,
            )
        _dfdy_fjac_np = _utils._to_ndarray(
            {'fun_name': cb_fjac_name, 'returned': 'dfdy'},
            _dfdy_fjac,
            copy=False,
            exp_shape=(neq, neq, mmax),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_fjac_name},
            ((_dfdy_fjac_np, 'dfdy'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        dfdy[:, :, :] = _dfdy_fjac_np

    cb_gafun_name = fun_name + ':gafun'

    def cb_gafun(ya, neq, m, mmax, nlbc, ga, data=None):
        """Wrapping user's gafun. Called as base gafun."""
        if data is not None:
            _ga_gafun = gafun(
                ya,
                m,
                nlbc,
                data,
            )
        else:
            _ga_gafun = gafun(
                ya,
                m,
                nlbc,
            )
        ga[:] = _utils._to_ndarray(
            {'fun_name': cb_gafun_name, 'returned': 'ga'},
            _ga_gafun,
            copy=False,
            exp_shape=(nlbc,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_gbfun_name = fun_name + ':gbfun'

    def cb_gbfun(yb, neq, m, mmax, nrbc, gb, data=None):
        """Wrapping user's gbfun. Called as base gbfun."""
        if data is not None:
            _gb_gbfun = gbfun(
                yb,
                m,
                nrbc,
                data,
            )
        else:
            _gb_gbfun = gbfun(
                yb,
                m,
                nrbc,
            )
        gb[:] = _utils._to_ndarray(
            {'fun_name': cb_gbfun_name, 'returned': 'gb'},
            _gb_gbfun,
            copy=False,
            exp_shape=(nrbc,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_gajac_name = fun_name + ':gajac'

    def cb_gajac(ya, neq, m, mmax, nlbc, dgady, data=None):
        """Wrapping user's gajac. Called as base gajac."""
        if data is not None:
            _dgady_gajac = gajac(
                ya,
                m,
                dgady,
                data,
            )
        else:
            _dgady_gajac = gajac(
                ya,
                m,
                dgady,
            )
        _dgady_gajac_np = _utils._to_ndarray(
            {'fun_name': cb_gajac_name, 'returned': 'dgady'},
            _dgady_gajac,
            copy=False,
            exp_shape=(nlbc, neq, mmax),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_gajac_name},
            ((_dgady_gajac_np, 'dgady'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        dgady[:, :, :] = _dgady_gajac_np

    cb_gbjac_name = fun_name + ':gbjac'

    def cb_gbjac(yb, neq, m, mmax, nrbc, dgbdy, data=None):
        """Wrapping user's gbjac. Called as base gbjac."""
        if data is not None:
            _dgbdy_gbjac = gbjac(
                yb,
                m,
                dgbdy,
                data,
            )
        else:
            _dgbdy_gbjac = gbjac(
                yb,
                m,
                dgbdy,
            )
        _dgbdy_gbjac_np = _utils._to_ndarray(
            {'fun_name': cb_gbjac_name, 'returned': 'dgbdy'},
            _dgbdy_gbjac,
            copy=False,
            exp_shape=(nrbc, neq, mmax),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_gbjac_name},
            ((_dgbdy_gbjac_np, 'dgbdy'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        dgbdy[:, :, :] = _dgbdy_gbjac_np

    cb_guess_name = fun_name + ':guess'

    def cb_guess(x, neq, m, mmax, y, dym, data=None):
        """Wrapping user's guess. Called as base guess."""
        if data is not None:
            _y_guess, _dym_guess = guess(
                x,
                m,
                mmax,
                data,
            )
        else:
            _y_guess, _dym_guess = guess(
                x,
                m,
                mmax,
            )
        _y_guess_np = _utils._to_ndarray(
            {'fun_name': cb_guess_name, 'returned': 'y'},
            _y_guess,
            copy=False,
            exp_shape=(neq, mmax),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        dym[:] = _utils._to_ndarray(
            {'fun_name': cb_guess_name, 'returned': 'dym'},
            _dym_guess,
            copy=False,
            exp_shape=(neq,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_guess_name},
            ((_y_guess_np, 'y'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        y[:, :] = _y_guess_np

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(ffun, {'fun_name': fun_name, 'entity': 'ffun'})
    _ffun = cb_ffun
    _b_utils._check_callable(fjac, {'fun_name': fun_name, 'entity': 'fjac'})
    _fjac = cb_fjac
    _b_utils._check_callable(gafun, {'fun_name': fun_name, 'entity': 'gafun'})
    _gafun = cb_gafun
    _b_utils._check_callable(gbfun, {'fun_name': fun_name, 'entity': 'gbfun'})
    _gbfun = cb_gbfun
    _b_utils._check_callable(gajac, {'fun_name': fun_name, 'entity': 'gajac'})
    _gajac = cb_gajac
    _b_utils._check_callable(gbjac, {'fun_name': fun_name, 'entity': 'gbjac'})
    _gbjac = cb_gbjac
    _b_utils._check_callable(guess, {'fun_name': fun_name, 'entity': 'guess'})
    _guess = cb_guess
    b_bvp_coll_nlin_solve(
        _ffun,
        _fjac,
        _gafun,
        _gbfun,
        _gajac,
        _gbjac,
        _guess,
        comm,
        data,
    )

def bvp_coll_nlin_setup(m, nlbc, nrbc, ncol, tols, nmesh, mesh, ipmesh):
    u"""
    Ordinary differential equations, general nonlinear boundary value
    problem, setup for ``bvp_coll_nlin_solve``.

    ``bvp_coll_nlin_setup`` is a setup function which must be called
    prior to the first call of the nonlinear two-point boundary value
    solver ``bvp_coll_nlin_solve``.

    For full information please refer to the NAG Library document for
    d02tv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02tvf.html

    Parameters
    ----------
    m : int, array-like, shape (neq)
        `m`[i-1] must contain m_i, the order of the ith differential
        equation, for i = 1,2,...,n.

    nlbc : int
        p, the number of left boundary conditions defined at the
        left-hand end, a (= `mesh`[0]).

    nrbc : int
        q, the number of right boundary conditions defined at the
        right-hand end, b (= `mesh`[`nmesh`-1]).

    ncol : int
        The number of collocation points to be used in each mesh
        sub-interval.

    tols : float, array-like, shape (neq)
        `tols`[i-1] must contain the error requirement for the ith
        solution component.

    nmesh : int
        The number of points to be used in the initial mesh of the
        solution process.

    mesh : float, array-like, shape (mxmesh)
        The positions of the initial `nmesh` mesh points. The remaining
        elements of `mesh` need not be set. You should try to place the
        mesh points in areas where you expect the solution to vary most
        rapidly. In the absence of any other information the points
        should be equally distributed on [a,b].

        `mesh`[0] must contain the left boundary point, a, and
        `mesh`[`nmesh`-1] must contain the right boundary point, b.

    ipmesh : int, array-like, shape (mxmesh)
        `ipmesh`[i-1] specifies whether or not the initial mesh point
        defined in `mesh`[i-1], for i = 1,2,...,`nmesh`, should be a
        fixed point in all meshes computed during the solution process.
        The remaining elements of `ipmesh` need not be set.

        `ipmesh`[i-1] = 1

          Indicates that `mesh`[i-1] should be a fixed point in all
          meshes.

        `ipmesh`[i-1] = 2

          Indicates that `mesh`[i-1] is not a fixed point.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ipmesh`[i] != 1 or 2 for some i.

        (`errno` 1)
            On entry, `ipmesh`[0] or `ipmesh`[`nmesh`-1] does not equal
            1.

        (`errno` 1)
            On entry, the elements of `mesh` are not strictly
            increasing.

        (`errno` 1)
            On entry, `nlbc` = *<value>*, `nrbc` = *<value>* and
            sum(`m`[i]) = *<value>*.

            Constraint: `nlbc`+`nrbc` = sum(`m`[i]).

        (`errno` 1)
            On entry, `nlbc` = *<value>* and `nrbc` = *<value>*.

            Constraint: `nlbc` >= 1 and `nrbc` >= 1.

        (`errno` 1)
            On entry, `ncol` = *<value>* and max(`m`[i]) = *<value>*.

            Constraint: max(`m`[i]) <= `ncol` <= 7.

        (`errno` 1)
            On entry, `tols`[*<value>*] = *<value>*.

            Constraint: `tols`[i-1] > *<value>* for all i.

        (`errno` 1)
            On entry, `m`[*<value>*] = *<value>*.

            Constraint: 1 <= `m`[i-1] <= 4 for all i.

        (`errno` 1)
            On entry, mxmesh = *<value>* and `nmesh` = *<value>*.

            Constraint: mxmesh >= 2*`nmesh`-1.

        (`errno` 1)
            On entry, `nmesh` = *<value>*.

            Constraint: `nmesh` >= 6.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.
    """

    from ..base.ode import bvp_coll_nlin_setup as b_bvp_coll_nlin_setup

    fun_name = 'naginterfaces.library.ode.bvp_coll_nlin_setup'

    _m = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'm'},
        m,
        copy=False,
    )
    _neq_py = _b_utils._EngineIntArrayType.get_shape(_m, {'fun_name': fun_name, 'entity': 'm'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _tols = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tols'},
        tols,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mesh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mesh'},
        mesh,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mxmesh_py = _b_utils._EngineFloat64ArrayType.get_shape(_mesh, {'fun_name': fun_name, 'entity': 'mesh'}, exp_rank=1)[0]
    _mxmesh = _b_utils._EngineIntScalarType(
        _mxmesh_py, {'fun_name': fun_name, 'entity': '_mxmesh'},
    ).to_ctypes()
    _ipmesh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipmesh'},
        ipmesh,
        copy=False,
    )
    _return_scalar = b_bvp_coll_nlin_setup(
        _neq,
        _m,
        nlbc,
        nrbc,
        ncol,
        _tols,
        _mxmesh,
        nmesh,
        _mesh,
        _ipmesh,
    )
    _comm_py = _return_scalar
    return _comm_py

def bvp_coll_nlin_contin(nmesh, mesh, ipmesh, comm):
    u"""
    Ordinary differential equations, general nonlinear boundary value
    problem, continuation facility for ``bvp_coll_nlin_solve``.

    ``bvp_coll_nlin_contin`` allows a solution to a nonlinear two-point
    boundary value problem computed by ``bvp_coll_nlin_solve`` to be
    used as an initial approximation in the solution of a related
    nonlinear two-point boundary value problem in a continuation call to
    ``bvp_coll_nlin_solve``.

    For full information please refer to the NAG Library document for
    d02tx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02txf.html

    Parameters
    ----------
    nmesh : int
        The number of points to be used in the new initial mesh. It is
        strongly recommended that if this function is called that the
        suggested value for `nmesh` is used. In this case the arrays
        mesh and ipmesh returned by ``bvp_coll_nlin_diag`` can be passed
        to this function without any modification.

    mesh : float, array-like, shape (mxmesh)
        The `nmesh` points to be used in the new initial mesh as
        specified by `ipmesh`.

    ipmesh : int, array-like, shape (mxmesh)
        Specifies the points in `mesh` to be used as the new initial
        mesh. Let {i_j:j = 1,2,...,`nmesh`} be the set of array indices
        of `ipmesh` such that `ipmesh`[i_j-1] = 1 or 2 and 1 = i_1 < i_2
        < ... < i_`nmesh`. Then `mesh`[i_j-1] will be included in the
        new initial mesh.

        If `ipmesh`[i_j-1] = 1, `mesh`[i_j-1] will be a fixed point in
        the new initial mesh.

        If `ipmesh`[k-1] = 3 for any k, `mesh`[k-1] will not be included
        in the new mesh.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bvp_coll_nlin_setup``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The entries in `mesh` are not strictly increasing.

        (`errno` 1)
            The last point of the new mesh does not coincide with the
            right hand end of the range previously specified.

            Last point of the new mesh: *<value>*; right-hand end of the
            range: *<value>*.

        (`errno` 1)
            You have set the element of `ipmesh` corresponding to the
            last element of `mesh` to be included in the new mesh as
            *<value>*, which is not 1.

        (`errno` 1)
            Expected *<value>* elements of `ipmesh` to be 1 or 2, but
            *<value>* such elements found.

        (`errno` 1)
            An element of `ipmesh` was set to -1 before `nmesh` elements
            containing 1 or 2 were detected.

        (`errno` 1)
            `ipmesh`[i] != -1, 1, 2 or 3 for some i.

        (`errno` 1)
            On entry, `ipmesh`[0] = *<value>*.

            Constraint: `ipmesh`[0] = 1.

        (`errno` 1)
            The first element of array `mesh` does not coincide with the
            left-hand end of the range previously specified.

            First element of `mesh`: *<value>*; left-hand of the range:
            *<value>*.

        (`errno` 1)
            On entry, `nmesh` = *<value>* and mxmesh = *<value>*.

            Constraint: `nmesh` <= (mxmesh+1)/2.

        (`errno` 1)
            On entry, `nmesh` = *<value>*.

            Constraint: `nmesh` >= 6.

        (`errno` 1)
            On entry, mxmesh = *<value>* and mxmesh = *<value>* in
            ``bvp_coll_nlin_setup``.

            Constraint: mxmesh = mxmesh in ``bvp_coll_nlin_setup``.

        (`errno` 1)
            The solver function did not produce any results suitable for
            remeshing.

        (`errno` 1)
            The solver function does not appear to have been called.
    """

    from ..base.ode import bvp_coll_nlin_contin as b_bvp_coll_nlin_contin

    fun_name = 'naginterfaces.library.ode.bvp_coll_nlin_contin'

    _mesh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mesh'},
        mesh,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mxmesh_py = _b_utils._EngineFloat64ArrayType.get_shape(_mesh, {'fun_name': fun_name, 'entity': 'mesh'}, exp_rank=1)[0]
    _mxmesh = _b_utils._EngineIntScalarType(
        _mxmesh_py, {'fun_name': fun_name, 'entity': '_mxmesh'},
    ).to_ctypes()
    _ipmesh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipmesh'},
        ipmesh,
        copy=False,
    )
    b_bvp_coll_nlin_contin(
        _mxmesh,
        nmesh,
        _mesh,
        _ipmesh,
        comm,
    )

def bvp_coll_nlin_interp(x, neq, mmax, comm):
    u"""
    Ordinary differential equations, general nonlinear boundary value
    problem, interpolation for ``bvp_coll_nlin_solve``.

    ``bvp_coll_nlin_interp`` interpolates on the solution of a general
    two-point boundary value problem computed by
    ``bvp_coll_nlin_solve``.

    For full information please refer to the NAG Library document for
    d02ty

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02tyf.html

    Parameters
    ----------
    x : float
        x, the independent variable.

    neq : int
        The number of differential equations.

    mmax : int
        The maximal order of the differential equations, max(m_i), for i
        = 1,2,...,`neq`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bvp_coll_nlin_setup``.

    Returns
    -------
    y : float, ndarray, shape (`neq`, `mmax`)
        `y`[i-1,j] contains an approximation to y_i^(j)(x), for j =
        0,1,...,m_i-1, for i = 1,2,...,`neq`. The remaining elements of
        `y` (where m_i < `mmax`) are initialized to 0.0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= *<value>*.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= *<value>*.

        (`errno` 1)
            On entry, `mmax` = *<value>* and max(m[i]) = *<value>* in
            ``bvp_coll_nlin_setup``.

            Constraint: `mmax` = max(m[i]) in ``bvp_coll_nlin_setup``.

        (`errno` 1)
            The solver function did not produce any results suitable for
            interpolation.

        (`errno` 1)
            The solver function does not appear to have been called.

        (`errno` 1)
            On entry, `neq` = *<value>* and neq = *<value>* in
            ``bvp_coll_nlin_setup``.

            Constraint: `neq` = neq in ``bvp_coll_nlin_setup``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The solver function did not satisfy the error requirements.

            Interpolated values should be treated with caution.

        (`errno` 2)
            The solver function did not converge to a suitable solution.

            A converged intermediate solution has been used.

            Interpolated values should be treated with caution.
    """

    from ..base.ode import bvp_coll_nlin_interp as b_bvp_coll_nlin_interp

    fun_name = 'naginterfaces.library.ode.bvp_coll_nlin_interp'

    _neq = _b_utils._EngineIntScalarType(
        neq, {'fun_name': fun_name, 'entity': 'neq'},
    ).to_ctypes()
    _neq_py = _b_utils._EngineIntScalarType.to_py(_neq)
    _mmax = _b_utils._EngineIntScalarType(
        mmax, {'fun_name': fun_name, 'entity': 'mmax'},
    ).to_ctypes()
    _mmax_py = _b_utils._EngineIntScalarType.to_py(_mmax)
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_neq_py, _mmax_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        b_bvp_coll_nlin_interp(
            x,
            _y,
            _neq,
            _mmax,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['y'],
    )(
        _y_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _y_py

def bvp_coll_nlin_diag(mxmesh, comm):
    u"""
    Ordinary differential equations, general nonlinear boundary value
    problem, diagnostics for ``bvp_coll_nlin_solve``.

    ``bvp_coll_nlin_diag`` returns information about the solution of a
    general two-point boundary value problem computed by
    ``bvp_coll_nlin_solve``.

    For full information please refer to the NAG Library document for
    d02tz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02tzf.html

    Parameters
    ----------
    mxmesh : int
        The maximum number of points allowed in the mesh.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bvp_coll_nlin_setup``.

    Returns
    -------
    nmesh : int
        The number of points in the mesh last used by
        ``bvp_coll_nlin_solve``.

    mesh : float, ndarray, shape (`mxmesh`)
        `mesh`[i-1] contains the ith point of the mesh last used by
        ``bvp_coll_nlin_solve``, for i = 1,2,...,`nmesh`. `mesh`[0] will
        contain a and `mesh`[`nmesh`-1] will contain b. The remaining
        elements of `mesh` are not initialized.

    ipmesh : int, ndarray, shape (`mxmesh`)
        `ipmesh`[i-1] specifies the nature of the point `mesh`[i-1], for
        i = 1,2,...,`nmesh`, in the final mesh computed by
        ``bvp_coll_nlin_solve``.

        `ipmesh`[i-1] = 1

          Indicates that the ith point is a fixed point and was used by
          the solver before an extrapolation-like error test.

        `ipmesh`[i-1] = 2

          Indicates that the ith point was used by the solver before an
          extrapolation-like error test.

        `ipmesh`[i-1] = 3

          Indicates that the ith point was used by the solver only as
          part of an extrapolation-like error test.

        The remaining elements of `ipmesh` are initialized to -1.

        See [Further Comments] for advice on how these values may be
        used in conjunction with a continuation process.

    ermx : float
        An estimate of the maximum error in the solution computed by
        ``bvp_coll_nlin_solve``, that is

            `ermx` = max(||y_i-v_i||/((1.0+||v_i||)))

        where v_i is the approximate solution for the ith solution
        component. If ``bvp_coll_nlin_solve`` returned successfully with
        no exception or warning is raised, `ermx` will be less than
        tols[`ijermx`-1] in ``bvp_coll_nlin_setup`` where tols contains
        the error requirements as specified in [Description] and
        [Parameters].

        If ``bvp_coll_nlin_solve`` returned with `errno` = 5, `ermx`
        will be greater than tols[`ijermx`-1] in
        ``bvp_coll_nlin_setup``.

        If ``bvp_coll_nlin_solve`` returned any other value for errno
        then an error estimate is not available and `ermx` is
        initialized to 0.0.

    iermx : int
        Indicates the mesh sub-interval where the value of `ermx` has
        been computed, that is [`mesh`[ `iermx` -1],`mesh`[ `iermx` + 1
        -1]].

        If an estimate of the error is not available then `iermx` is
        initialized to 0.

    ijermx : int
        Indicates the component i (= `ijermx`) of the solution for which
        `ermx` has been computed, that is the approximation of y_i on
        [`mesh`[`iermx`-1],`mesh`[`iermx`]] is estimated to have the
        largest error of all components y_i over mesh sub-intervals
        defined by `mesh`.

        If an estimate of the error is not available then `ijermx` is
        initialized to 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The solver function did not produce any results suitable for
            interpolation.

        (`errno` 1)
            The solver function does not appear to have been called.

        (`errno` 1)
            On entry, `mxmesh` = *<value>* and mxmesh = *<value>* in
            ``bvp_coll_nlin_setup``.

            Constraint: `mxmesh` = mxmesh in ``bvp_coll_nlin_setup``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The solver function did not satisfy the error requirements.

            Information has been supplied on the last mesh used.

        (`errno` 2)
            The solver function did not converge to a suitable solution.

            A converged intermediate solution has been used.

            Error estimate information is not available.
    """

    from ..base.ode import bvp_coll_nlin_diag as b_bvp_coll_nlin_diag

    fun_name = 'naginterfaces.library.ode.bvp_coll_nlin_diag'

    _mxmesh = _b_utils._EngineIntScalarType(
        mxmesh, {'fun_name': fun_name, 'entity': 'mxmesh'},
    ).to_ctypes()
    _mxmesh_py = _b_utils._EngineIntScalarType.to_py(_mxmesh)
    _mesh = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mxmesh_py,),
    )
    _ipmesh = _b_utils._EngineIntArrayType.empty_ndarray(
        (_mxmesh_py,),
    )
    try:
        _return_scalars = b_bvp_coll_nlin_diag(
            _mxmesh,
            _mesh,
            _ipmesh,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _nmesh_py = _return_scalars.nmesh
    _mesh_py = _mesh
    _ipmesh_py = _ipmesh
    _ermx_py = _return_scalars.ermx
    _iermx_py = _return_scalars.iermx
    _ijermx_py = _return_scalars.ijermx
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nmesh', 'mesh', 'ipmesh', 'ermx', 'iermx', 'ijermx'],
    )(
        _nmesh_py,
        _mesh_py,
        _ipmesh_py,
        _ermx_py,
        _iermx_py,
        _ijermx_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bvp_ps_lin_coeffs(f):
    u"""
    Coefficients of Chebyshev interpolating polynomial from function
    values on Chebyshev grid.

    ``bvp_ps_lin_coeffs`` obtains the Chebyshev coefficients of a
    function discretized on Chebyshev Gauss--Lobatto points.
    The set of discretization points on which the function is evaluated
    is usually obtained by a previous call to ``bvp_ps_lin_cgl_grid``.

    For full information please refer to the NAG Library document for
    d02ua

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02uaf.html

    Parameters
    ----------
    f : float, array-like, shape (n+1)
        The function values f(x_r), for r = 1,2,...,n+1.

    Returns
    -------
    c : float, ndarray, shape (n+1)
        The Chebyshev coefficients, c_j, for j = 1,2,...,n+1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n is even.
    """

    from ..base.ode import bvp_ps_lin_coeffs as b_bvp_ps_lin_coeffs

    fun_name = 'naginterfaces.library.ode.bvp_ps_lin_coeffs'

    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_f, {'fun_name': fun_name, 'entity': 'f'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+1,),
    )
    b_bvp_ps_lin_coeffs(
        _n,
        _f,
        _c,
    )
    _c_py = _c
    return _c_py

def bvp_ps_lin_cgl_vals(a, b, q, c):
    u"""
    Function or low-order-derivative values on Chebyshev grid from
    coefficients of Chebyshev interpolating polynomial.

    ``bvp_ps_lin_cgl_vals`` evaluates a function, or one of its lower
    order derivatives, from its Chebyshev series representation at
    Chebyshev Gauss--Lobatto points on [a,b].
    The coefficients of the Chebyshev series representation required are
    usually derived from those returned by ``bvp_ps_lin_coeffs`` or
    ``bvp_ps_lin_solve``.

    For full information please refer to the NAG Library document for
    d02ub

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ubf.html

    Parameters
    ----------
    a : float
        a, the lower bound of domain [a,b].

    b : float
        b, the upper bound of domain [a,b].

    q : int
        The order, q, of the derivative to evaluate.

    c : float, array-like, shape (n+1)
        The Chebyshev coefficients, c_i, for i = 1,2,...,n+1.

    Returns
    -------
    f : float, ndarray, shape (n+1)
        The derivatives S^(q)x_i, for i = 1,2,...,n+1, of the Chebyshev
        series, S.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n is even.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` < `b`.

        (`errno` 3)
            On entry, `q` = *<value>*.

            Constraint: 0 <= `q` <= 4.
    """

    from ..base.ode import bvp_ps_lin_cgl_vals as b_bvp_ps_lin_cgl_vals

    fun_name = 'naginterfaces.library.ode.bvp_ps_lin_cgl_vals'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+1,),
    )
    b_bvp_ps_lin_cgl_vals(
        _n,
        a,
        b,
        q,
        _c,
        _f,
    )
    _f_py = _f
    return _f_py

def bvp_ps_lin_cgl_grid(n, a, b):
    u"""
    Chebyshev Gauss--Lobatto grid generation.

    ``bvp_ps_lin_cgl_grid`` returns the Chebyshev Gauss--Lobatto grid
    points on [a,b].

    For full information please refer to the NAG Library document for
    d02uc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02ucf.html

    Parameters
    ----------
    n : int
        n, where the number of grid points is n+1. This is also the
        largest order of Chebyshev polynomial in the Chebyshev series to
        be computed.

    a : float
        a, the lower bound of domain [a,b].

    b : float
        b, the upper bound of domain [a,b].

    Returns
    -------
    x : float, ndarray, shape (`n`+1)
        The Chebyshev Gauss--Lobatto grid points, x_i, for i =
        1,2,...,n+1, on [a,b].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` is even.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` < `b`.
    """

    from ..base.ode import bvp_ps_lin_cgl_grid as b_bvp_ps_lin_cgl_grid

    fun_name = 'naginterfaces.library.ode.bvp_ps_lin_cgl_grid'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+1,),
    )
    b_bvp_ps_lin_cgl_grid(
        _n,
        a,
        b,
        _x,
    )
    _x_py = _x
    return _x_py

def bvp_ps_lin_cgl_deriv(f):
    u"""
    Differentiate a function by the FFT using function values on
    Chebyshev grid.

    ``bvp_ps_lin_cgl_deriv`` differentiates a function discretized on
    Chebyshev Gauss--Lobatto points.
    The grid points on which the function values are to be provided are
    normally returned by a previous call to ``bvp_ps_lin_cgl_grid``.

    For full information please refer to the NAG Library document for
    d02ud

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02udf.html

    Parameters
    ----------
    f : float, array-like, shape (n+1)
        The function values f(x_i), for i = 1,2,...,n+1

    Returns
    -------
    fd : float, ndarray, shape (n+1)
        The approximations to the derivatives of the function evaluated
        at the Chebyshev Gauss--Lobatto points. For functions defined on
        [a,b], the returned derivative values (corresponding to the
        domain [-1,1]) must be multiplied by the factor 2/(b-a) to
        obtain the correct values on [a,b].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n is even.
    """

    from ..base.ode import bvp_ps_lin_cgl_deriv as b_bvp_ps_lin_cgl_deriv

    fun_name = 'naginterfaces.library.ode.bvp_ps_lin_cgl_deriv'

    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_f, {'fun_name': fun_name, 'entity': 'f'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _fd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+1,),
    )
    b_bvp_ps_lin_cgl_deriv(
        _n,
        _f,
        _fd,
    )
    _fd_py = _fd
    return _fd_py

def bvp_ps_lin_solve(a, b, c, bmat, y, bvec, f):
    u"""
    Solve linear constant coefficient boundary value problem on
    Chebyshev grid, Integral formulation.

    ``bvp_ps_lin_solve`` finds the solution of a linear constant
    coefficient boundary value problem by using the Chebyshev
    integration formulation on a Chebyshev Gauss--Lobatto grid.

    For full information please refer to the NAG Library document for
    d02ue

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02uef.html

    Parameters
    ----------
    a : float
        a, the lower bound of domain [a,b].

    b : float
        b, the upper bound of domain [a,b].

    c : float, array-like, shape (n+1)
        The Chebyshev coefficients c_j, j = 0,1,...,n, for the
        right-hand side of the boundary value problem. Usually these are
        obtained by a previous call of ``bvp_ps_lin_coeffs``.

    bmat : float, array-like, shape (m, m+1)
        `bmat`[i-1,j] must contain the coefficients B_{i,j+1}, for j =
        0,1,...,m, for i = 1,2,...,m, in the problem formulation of
        [Description].

    y : float, array-like, shape (m)
        The points, y_i, for i = 1,2,...,m, where the boundary
        conditions are discretized.

    bvec : float, array-like, shape (m)
        The values, beta_i, for i = 1,2,...,m, in the formulation of the
        boundary conditions given in [Description].

    f : float, array-like, shape (m+1)
        The coefficients, f_j, for j = 1,2,...,m+1, in the formulation
        of the linear boundary value problem given in [Description]. The
        highest order term, `f`[m], needs to be nonzero to have a well
        posed problem.

    Returns
    -------
    bmat : float, ndarray, shape (m, m+1)
        The coefficients have been scaled to form an equivalent problem
        defined on the domain [-1,1].

    f : float, ndarray, shape (m+1)
        The coefficients have been scaled to form an equivalent problem
        defined on the domain [-1,1].

    uc : float, ndarray, shape (n+1, m+1)
        The Chebyshev coefficients in the Chebyshev series
        representations of the solution and derivatives of the solution
        to the boundary value problem. The n+1 elements `uc`[0:n+1,0]
        contain the coefficients representing the solution U(x_i), for i
        = 0,1,...,n. `uc`[0:n+1,j] contains the coefficients
        representing the jth derivative of U, for j = 1,2,...,m.

    resid : float
        The maximum residual resulting from substituting the solution
        vectors returned in `uc` into the equations representing the
        linear boundary value problem and associated boundary
        conditions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 8.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n is even.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` < `b`.

        (`errno` 3)
            On entry, `f`[m] = 0.0.

        (`errno` 6)
            On entry, m = *<value>*.

            Constraint: 1 <= m <= 4.

        (`errno` 7)
            Internal error while unpacking matrix during iterative
            refinement.

            Please contact NAG.

        (`errno` 8)
            Singular matrix encountered during iterative refinement.

            Please check that your system is well posed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 9)
            During iterative refinement, the maximum number of
            iterations was reached.

            Number of iterations = *<value>* and residual achieved =
            *<value>*.

        (`errno` 10)
            During iterative refinement, convergence was achieved, but
            the residual is more than 100*machine precision. Residual
            achieved on convergence = *<value>*.
    """

    from ..base.ode import bvp_ps_lin_solve as b_bvp_ps_lin_solve

    fun_name = 'naginterfaces.library.ode.bvp_ps_lin_solve'

    _bmat = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bmat'},
        bmat,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_bmat, 'bmat'),),
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_bmat, {'fun_name': fun_name, 'entity': 'bmat'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bvec = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bvec'},
        bvec,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _uc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+1, _m_py+1),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = b_bvp_ps_lin_solve(
            _n,
            a,
            b,
            _m,
            _c,
            _bmat,
            _y,
            _bvec,
            _f,
            _uc,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bmat_py = _bmat
    _f_py = _f
    _uc_py = _uc
    _resid_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bmat', 'f', 'uc', 'resid'],
    )(
        _bmat_py,
        _f_py,
        _uc_py,
        _resid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bvp_ps_lin_grid_vals(nip, x, f):
    u"""
    Interpolate a function from Chebyshev grid to uniform grid using
    barycentric Lagrange interpolation.

    ``bvp_ps_lin_grid_vals`` interpolates from a set of function values
    on a supplied grid onto a set of values for a uniform grid on the
    same range.
    The interpolation is performed using barycentric Lagrange
    interpolation. ``bvp_ps_lin_grid_vals`` is primarily a utility
    function to map a set of function values specified on a Chebyshev
    Gauss--Lobatto grid onto a uniform grid.

    For full information please refer to the NAG Library document for
    d02uw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02uwf.html

    Parameters
    ----------
    nip : int
        The number, m, of grid points in the uniform mesh x^ onto which
        function values are interpolated. If `nip` = 1 then on
        successful exit from ``bvp_ps_lin_grid_vals``, `fip`[0] will
        contain the value f(x_n).

    x : float, array-like, shape (n+1)
        The grid points, x_i, for i = 0,1,...,n, at which the function
        is specified.

        Usually this should be the array of Chebyshev Gauss--Lobatto
        points returned in ``bvp_ps_lin_cgl_grid``.

    f : float, array-like, shape (n+1)
        The function values, f(x_i), for i = 0,1,...,n.

    Returns
    -------
    xip : float, ndarray, shape (`nip`)
        The evenly-spaced grid points, x^_j, for j = 1,2,...,m.

    fip : float, ndarray, shape (`nip`)
        The set of interpolated values f^(x^_j), for j = 1,2,...,m. Here
        f^(x^_j)≈f(x = x^_j).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n is even.

        (`errno` 2)
            On entry, `nip` = *<value>*.

            Constraint: `nip` > 0.
    """

    from ..base.ode import bvp_ps_lin_grid_vals as b_bvp_ps_lin_grid_vals

    fun_name = 'naginterfaces.library.ode.bvp_ps_lin_grid_vals'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nip = _b_utils._EngineIntScalarType(
        nip, {'fun_name': fun_name, 'entity': 'nip'},
    ).to_ctypes()
    _nip_py = _b_utils._EngineIntScalarType.to_py(_nip)
    _xip = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nip_py,),
    )
    _fip = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nip_py,),
    )
    b_bvp_ps_lin_grid_vals(
        _n,
        _nip,
        _x,
        _f,
        _xip,
        _fip,
    )
    _xip_py = _xip
    _fip_py = _fip
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xip', 'fip'],
    )(
        _xip_py,
        _fip_py,
    )
    return _return_data

def bvp_ps_lin_quad_weights(n):
    u"""
    Clenshaw--Curtis quadrature weights for integration using computed
    Chebyshev coefficients.

    ``bvp_ps_lin_quad_weights`` obtains the weights for Clenshaw--Curtis
    quadrature at Chebyshev points.
    This allows for fast approximations of integrals for functions
    specified on Chebyshev Gauss--Lobatto points on [-1,1].

    For full information please refer to the NAG Library document for
    d02uy

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02uyf.html

    Parameters
    ----------
    n : int
        n, where the number of grid points is n+1.

    Returns
    -------
    w : float, ndarray, shape (`n`+1)
        The Clenshaw--Curtis quadrature weights, w_i, for i = 0,1,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` is even.
    """

    from ..base.ode import bvp_ps_lin_quad_weights as b_bvp_ps_lin_quad_weights

    fun_name = 'naginterfaces.library.ode.bvp_ps_lin_quad_weights'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+1,),
    )
    b_bvp_ps_lin_quad_weights(
        _n,
        _w,
    )
    _w_py = _w
    return _w_py

def bvp_ps_lin_cheb_eval(k, x):
    u"""
    Chebyshev polynomial evaluation, T_k(x).

    ``bvp_ps_lin_cheb_eval`` returns the value of the kth Chebyshev
    polynomial evaluated at a point x in [-1,1].
    ``bvp_ps_lin_cheb_eval`` is primarily a utility function for use by
    the Chebyshev boundary value problem solvers.

    For full information please refer to the NAG Library document for
    d02uz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02uzf.html

    Parameters
    ----------
    k : int
        The order of the Chebyshev polynomial.

    x : float
        The point at which to evaluate the polynomial.

    Returns
    -------
    t : float
        The value, T, of the Chebyshev polynomial order k evaluated at
        x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: -1.0 <= `x` <= 1.0.
    """

    from ..base.ode import bvp_ps_lin_cheb_eval as b_bvp_ps_lin_cheb_eval

    _return_scalar = b_bvp_ps_lin_cheb_eval(
        k,
        x,
    )
    _t_py = _return_scalar
    return _t_py

def ivp_stiff_nat_interp(xsol, m, comm, neq, x, nqu, hu, h):
    u"""
    Ordinary differential equations, initial value problem,
    interpolation for d02m--n routines (BLEND and BDF methods only),
    natural interpolant.

    ``ivp_stiff_nat_interp`` interpolates components of the solution of
    a system of first-order ordinary differential equations from
    information provided by the integrators in submodule ``ode``.

    For full information please refer to the NAG Library document for
    d02xj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02xjf.html

    Parameters
    ----------
    xsol : float
        The point at which the first m components of the solution are to
        be evaluated. `xsol` should not be an extrapolation point, that
        is `xsol` should satisfy (`xsol`-`x`)*`hu` <= 0.0. Extrapolation
        is permitted but not recommended.

    m : int
        m, the number of components of the solution whose values at
        `xsol` are required. The first `m` components are evaluated.

    comm : dict, communication object
        Communication structure populated by a previous call to the
        integrator.

    neq : int
        The value used for the argument `neq` when calling the
        integrator.

    x : float
        The latest value at which the solution has been computed, as
        provided in the argument tcur on return from the optional output
        ``ivp_stiff_integ_diag``.

    nqu : int
        The order of the method used up to the latest value at which the
        solution has been computed, as provided in the argument nqu on
        return from the optional output ``ivp_stiff_integ_diag``.

    hu : float
        The last successful step used, that is the step used in the
        integration to get to `x`, as provided in the argument hu on
        return from the optional output ``ivp_stiff_integ_diag``.

    h : float
        The next step size to be attempted in the integration, as
        provided in the argument h on return from the optional output
        ``ivp_stiff_integ_diag``.

    Returns
    -------
    sol : float, ndarray, shape (`m`)
        The calculated value of the ith component of the solution at
        `xsol`, for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, sdysav = *<value>* and `nqu` = *<value>*.

            Constraint: sdysav >= `nqu`+1.

        (`errno` 1)
            On entry, `nqu` = *<value>*.

            Constraint: `nqu` >= 1.

        (`errno` 1)
            On entry, `neq` = *<value>* and ldysav = *<value>*.

            Constraint: `neq` <= ldysav.

        (`errno` 1)
            On entry, `m` = *<value>* and `neq` = *<value>*.

            Constraint: `m` <= `neq`.

        (`errno` 1)
            On entry, `neq` = *<value>*.

            Constraint: `neq` >= 1.

        (`errno` 1)
            On entry, ldysav = *<value>*.

            Constraint: ldysav >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `hu` = 0.0.

        (`errno` 2)
            On entry, `h` = 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, `xsol` = *<value>*, `x` = *<value>* and `h` =
            *<value>*.

            Constraint: `xsol` should not be an extrapolation point.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_nat_interp as b_ivp_stiff_nat_interp

    fun_name = 'naginterfaces.library.ode.ivp_stiff_nat_interp'

    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _sol = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _neq = _b_utils._EngineIntScalarType(
        neq, {'fun_name': fun_name, 'entity': 'neq'},
    ).to_ctypes()
    _neq_py = _b_utils._EngineIntScalarType.to_py(_neq)
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _nqu = _b_utils._EngineIntScalarType(
        nqu, {'fun_name': fun_name, 'entity': 'nqu'},
    ).to_ctypes()
    _nqu_py = _b_utils._EngineIntScalarType.to_py(_nqu)
    _sdysav = _b_utils._EngineIntScalarType(
        _nqu_py+1, {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    try:
        b_ivp_stiff_nat_interp(
            xsol,
            _sol,
            _m,
            comm,
            _ldysav,
            _sdysav,
            _neq,
            x,
            _nqu,
            hu,
            h,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _sol_py = _sol
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['sol'],
    )(
        _sol_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _sol_py

def ivp_stiff_c1_interp(xsol, m, comm, acor, x, nqu, hu, h):
    u"""
    Ordinary differential equations, initial value problem,
    interpolation for d02m--n functions, C^1 interpolant.

    ``ivp_stiff_c1_interp`` interpolates components of the solution of a
    system of first-order ordinary differential equations from
    information provided by the integrators in submodule ``ode``.
    It provides C^1 interpolation suitable for general use.

    For full information please refer to the NAG Library document for
    d02xk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02xkf.html

    Parameters
    ----------
    xsol : float
        The point at which the first m components of the solution are to
        be evaluated. `xsol` should not be an extrapolation point, that
        is `xsol` should satisfy (`xsol`-`x`)*`hu` <= 0.0. Extrapolation
        is permitted but not recommended.

    m : int
        The number of components of the solution whose values at `xsol`
        are required. The first m components are evaluated.

    comm : dict, communication object
        Communication structure populated by a previous call to the
        integrator.

    acor : float, array-like, shape (neq)
        The value returned in position (ldysav+50+i), for i =
        1,2,...,neq, of the argument rwork returned by the integrator.
        If one of the direct communication D02N functions is being
        employed and ``ivp_stiff_c1_interp`` is to be used in monitr,
        acor[i-1] must contain the value given in position (i,2) of the
        monitr argument acor, for i = 1,2,...,neq (e.g., see
        ``ivp_stiff_exp_fulljac``).

    x : float
        The latest value at which the solution has been computed, as
        provided in the argument tcur on return from the optional output
        ``ivp_stiff_integ_diag``.

    nqu : int
        The order of the method used up to the latest value at which the
        solution has been computed, as provided in the argument nqu on
        return from the optional output ``ivp_stiff_integ_diag``.

    hu : float
        The last successful step used, that is the step used in the
        integration to get to `x`, as provided in the argument hu on
        return from the optional output ``ivp_stiff_integ_diag``.

    h : float
        The next step size to be attempted in the integration, as
        provided in the argument h on return from the optional output
        ``ivp_stiff_integ_diag``.

    Returns
    -------
    sol : float, ndarray, shape (`m`)
        The calculated value of the ith component of the solution at
        `xsol`, for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The Petzold error test was selected.

        (`errno` 1)
            The BDF integration method is not being used.

        (`errno` 1)
            On entry, sdysav = *<value>* and `nqu` = *<value>*.

            Constraint: sdysav >= `nqu`+1.

        (`errno` 1)
            On entry, `nqu` = *<value>*.

            Constraint: `nqu` >= 1.

        (`errno` 1)
            On entry, neq = *<value>* and ldysav = *<value>*.

            Constraint: neq <= ldysav.

        (`errno` 1)
            On entry, `m` = *<value>* and neq = *<value>*.

            Constraint: `m` <= neq.

        (`errno` 1)
            On entry, neq = *<value>*.

            Constraint: neq >= 1.

        (`errno` 1)
            On entry, ldysav = *<value>*.

            Constraint: ldysav >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `hu` = 0.0.

        (`errno` 2)
            On entry, `h` = 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Extrapolation has been performed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_c1_interp as b_ivp_stiff_c1_interp

    fun_name = 'naginterfaces.library.ode.ivp_stiff_c1_interp'

    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _sol = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _acor = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'acor'},
        acor,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_acor, {'fun_name': fun_name, 'entity': 'acor'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _ldysav = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_ldysav'},
    ).to_ctypes()
    _nqu = _b_utils._EngineIntScalarType(
        nqu, {'fun_name': fun_name, 'entity': 'nqu'},
    ).to_ctypes()
    _nqu_py = _b_utils._EngineIntScalarType.to_py(_nqu)
    _sdysav = _b_utils._EngineIntScalarType(
        _nqu_py+1, {'fun_name': fun_name, 'entity': '_sdysav'},
    ).to_ctypes()
    try:
        b_ivp_stiff_c1_interp(
            xsol,
            _sol,
            _m,
            comm,
            _ldysav,
            _sdysav,
            _acor,
            _neq,
            x,
            _nqu,
            hu,
            h,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _sol_py = _sol
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['sol'],
    )(
        _sol_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _sol_py

def ivp_stiff_errest(v, w, comm):
    u"""
    Ordinary differential equations, initial value problem, weighted
    norm of local error estimate for d02m--n functions.

    ``ivp_stiff_errest`` calculates the weighted norm of the local error
    estimate from inside a monitr called from an integrator in submodule
    ``ode`` (e.g., see ``ivp_stiff_exp_fulljac``).

    For full information please refer to the NAG Library document for
    d02za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d02/d02zaf.html

    Parameters
    ----------
    v : float, array-like, shape (neq)
        The vector, the weighted norm of which is to be evaluated by
        ``ivp_stiff_errest``. `v` is calculated internally by the
        integrator being used.

    w : float, array-like, shape (neq)
        The weights, calculated internally by the integrator, to be used
        in the norm evaluation.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by prior calls to
        ``ivp_stiff_fulljac_setup`` and one of ``ivp_stiff_dassl``,
        ``ivp_stiff_bdf`` or ``ivp_stiff_blend``.

    Returns
    -------
    errnrm : float
        ||v||, the norm of the scaled local error estimate.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The value of the norm would either overflow or is close to
            overflowing.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.ode import ivp_stiff_errest as b_ivp_stiff_errest

    fun_name = 'naginterfaces.library.ode.ivp_stiff_errest'

    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neq_py = _b_utils._EngineFloat64ArrayType.get_shape(_v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=1)[0]
    _neq = _b_utils._EngineIntScalarType(
        _neq_py, {'fun_name': fun_name, 'entity': '_neq'},
    ).to_ctypes()
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = b_ivp_stiff_errest(
            _neq,
            _v,
            _w,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['errnrm'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar
