# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `pde` Chapter.

``pde`` - Partial Differential Equations

This module is concerned with the numerical solution of partial
differential equations.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03intro.html
"""

# NAG Copyright 2017-2019.

import collections as _collections

from ..base import utils as _b_utils
from ..library import _utils

def dim2_laplace(stage1, ext, dorm, p, q, x, y, phi, phid, alpha):
    u"""
    Elliptic PDE, Laplace's equation, two-dimensional arbitrary domain.

    ``dim2_laplace`` solves Laplace's equation in two dimensions for an
    arbitrary domain bounded internally or externally by one or more
    closed contours, given the value of either the unknown function or
    its normal derivative (into the domain) at each point of the
    boundary.

    For full information please refer to the NAG Library document for
    d03ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03eaf.html

    Parameters
    ----------
    stage1 : bool
        Indicates whether the function call is for stage 1 of the
        computation as defined in [Description].

        `stage1` = True

          The call is for stage 1.

        `stage1` = False

          The call is for stage 2.

    ext : bool
        The form of the domain. If `ext` = True, the domain is
        unbounded. Otherwise the domain is an interior one.

    dorm : bool
        The form of the boundary conditions. If `dorm` = True, the
        problem is a Dirichlet or mixed boundary value problem.
        Otherwise it is a Neumann problem.

    p : float
        To stage 2, `p` and `q` must specify the x and y coordinates
        respectively of a point at which the solution is required.

        When `stage1` is True, `p` and `q` are ignored.

    q : float
        To stage 2, `p` and `q` must specify the x and y coordinates
        respectively of a point at which the solution is required.

        When `stage1` is True, `p` and `q` are ignored.

    x : float, array-like, shape (n1p1)
        The x and y coordinates respectively of points on the one or
        more closed contours which define the domain of the problem.

        Note: each contour is described in such a manner that the
        subscripts of the coordinates increase when the domain is kept
        on the left. The final point on each contour coincides with the
        first and, if a further contour is to be described, the
        coordinates of this point are repeated in the arrays. In this
        way each interval is defined by three points, the second of
        which (the nodal point) always has an even subscript. In the
        case of the interior Neumann problem, the outermost boundary
        contour must be given first, if there is more than one.

    y : float, array-like, shape (n1p1)
        The x and y coordinates respectively of points on the one or
        more closed contours which define the domain of the problem.

        Note: each contour is described in such a manner that the
        subscripts of the coordinates increase when the domain is kept
        on the left. The final point on each contour coincides with the
        first and, if a further contour is to be described, the
        coordinates of this point are repeated in the arrays. In this
        way each interval is defined by three points, the second of
        which (the nodal point) always has an even subscript. In the
        case of the interior Neumann problem, the outermost boundary
        contour must be given first, if there is more than one.

    phi : float, array-like, shape (n)
        For stage 1, `phi` must contain the nodal values of phi or its
        normal derivative (into the domain) as prescribed in each
        interval. For stage 2 it must retain its output values from
        stage 1.

    phid : float, array-like, shape (n)
        For stage 1, `phid`[i-1] must hold the value 0.0 or 1.0
        accordingly as `phi`[i-1] contains a value of phi or its normal
        derivative, for i = 1,2,...,n. For stage 2 it must retain its
        output values from stage 1.

    alpha : float
        For stage 1, the use of `alpha` depends on the nature of the
        problem:

        - if `dorm` = True, `alpha` need not be set;

        - if `dorm` = False and `ext` = True, `alpha` must contain the
          prescribed constant c (see [Description]);

        - if `dorm` = False and `ext` = False, `alpha` must contain an
          appropriate value (often zero) for the integral of phi around
          the outermost boundary.

        For stage 2, on every call `alpha` must contain the value
        returned at stage 1.

    Returns
    -------
    phi : float, ndarray, shape (n)
        From stage 1, it contains the constants which approximate phi in
        each interval. It remains unchanged on exit from stage 2.

    phid : float, ndarray, shape (n)
        From stage 1, `phid` contains the constants which approximate
        the normal derivative of phi in each interval. It remains
        unchanged on exit from stage 2.

    alpha : float
        From stage 1:

        - if `ext` = False, `alpha` contains 0.0;

        - if `ext` = True and `dorm` = False `alpha` is unchanged;

        - if `ext` = True and `dorm` = True `alpha` contains a computed
          estimate for c.

        From stage 2:

        - `alpha` contains the computed value of phi at the point
          (`p`,`q`).

    Raises
    ------
    NagValueError
        (`errno` 1)
            Tolerance is too small: icint[0] = *<value>*.

        (`errno` 1)
            Tolerance is too large: icint[0] = 0.

        (`errno` 2)
            Incorrect rank: icint[0] = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_laplace as b_dim2_laplace

    fun_name = 'naginterfaces.library.pde.dim2_laplace'

    _phi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'phi'},
        phi,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_phi, {'fun_name': fun_name, 'entity': 'phi'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1p1_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n1p1 = _b_utils._EngineIntScalarType(
        _n1p1_py, {'fun_name': fun_name, 'entity': '_n1p1'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _phid = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'phid'},
        phid,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_dim2_laplace(
        stage1,
        ext,
        dorm,
        _n,
        p,
        q,
        _x,
        _y,
        _n1p1,
        _phi,
        _phid,
        alpha,
    )
    _phi_py = _phi
    _phid_py = _phid
    _alpha_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['phi', 'phid', 'alpha'],
    )(
        _phi_py,
        _phid_py,
        _alpha_py,
    )
    return _return_data

def dim2_ellip_fd(a, b, c, d, e, q, t, aparam, itmax, itcoun, ndir, ixn, iyn, conres, conchn):
    u"""
    Elliptic PDE, solution of finite difference equations by SIP,
    five-point two-dimensional molecule, iterate to convergence.

    ``dim2_ellip_fd`` uses the Strongly Implicit Procedure to calculate
    the solution to a system of simultaneous algebraic equations of
    five-point molecule form on a two-dimensional
    topologically-rectangular mesh. ('Topological' means that a polar
    grid, for example (r,theta), can be used, being equivalent to a
    rectangular box.)

    For full information please refer to the NAG Library document for
    d03eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03ebf.html

    Parameters
    ----------
    a : float, array-like, shape (n1, n2)
        `a`[i-1,j-1] must contain the coefficient of the 'southerly'
        term involving t_{i,j-1} in the (i,j)th equation of the system
        (1), for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of
        `a`, for j = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        rectangle.

    b : float, array-like, shape (n1, n2)
        `b`[i-1,j-1] must contain the coefficient of the 'westerly' term
        involving t_{i-1,j} in the (i,j)th equation of the system (1),
        for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of `b`, for
        i = 1, must be zero after incorporating the boundary conditions,
        since they involve nodal values from outside the rectangle.

    c : float, array-like, shape (n1, n2)
        `c`[i-1,j-1] must contain the coefficient of the 'central' term
        involving t_{ij} in the (i,j)th equation of the system (1), for
        j = 1,2,...,n2, for i = 1,2,...,n1. The elements of `c` are
        checked to ensure that they are nonzero. If any element is found
        to be zero, the corresponding algebraic equation is assumed to
        be t_{ij} = q_{ij}. This feature can be used to define the
        equations for nodes at which, for example, Dirichlet boundary
        conditions are applied, or for nodes external to the problem of
        interest, by setting `c`[i-1,j-1] = 0.0 at appropriate points,
        and the corresponding value of `q`[i-1,j-1] to the appropriate
        value, namely the prescribed value of `t`[i-1,j-1] in the
        Dirichlet case or zero at an external point.

    d : float, array-like, shape (n1, n2)
        `d`[i-1,j-1] must contain the coefficient of the 'easterly' term
        involving t_{i+1,j} in the (i,j)th equation of the system (1),
        for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of `d`, for
        i = n1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        rectangle.

    e : float, array-like, shape (n1, n2)
        `e`[i-1,j-1] must contain the coefficient of the 'northerly'
        term involving t_{i,j+1} in the (i,j)th equation of the system
        (1), for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of
        `e`, for j = n2 must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        rectangle.

    q : float, array-like, shape (n1, n2)
        `q`[i-1,j-1] must contain q_{ij}, for j = 1,2,...,n2, for i =
        1,2,...,n1, i.e., the source term values at the nodal points for
        the system (1).

    t : float, array-like, shape (n1, n2)
        `t`[i-1,j-1] must contain the element t_{ij} of the approximate
        solution to the equations supplied by the calling program as an
        initial starting value, for j = 1,2,...,n2, for i = 1,2,...,n1.

        If no better approximation is known, an array of zeros can be
        used.

    aparam : float
        The iteration acceleration factor. A value of 1.0 is adequate
        for most typical problems. However, if convergence is slow, the
        value can be reduced, typically to 0.2 or 0.1. If divergence is
        obtained, the value can be increased, typically to 2.0, 5.0 or
        10.0.

    itmax : int
        The maximum number of iterations to be used by the function in
        seeking the solution. A reasonable value might be 30 if n1 = n2
        = 10 or 100 if n1 = n2 = 50.

    itcoun : int
        On the first call of ``dim2_ellip_fd``, `itcoun` must be set to
        0. On subsequent entries, its value must be unchanged from the
        previous call.

    ndir : int
        Indicates whether or not the system of equations has a unique
        solution. For systems which have a unique solution, `ndir` must
        be set to any nonzero value. For systems derived from, for
        example, Laplace's equation with all Neumann boundary
        conditions, i.e., problems in which an arbitrary constant can be
        added to the solution, `ndir` should be set to 0 and the values
        of the next two arguments must be specified. For such problems
        the function subtracts the value of the function derived at the
        node (`ixn`, `iyn`) from the whole solution after each iteration
        to reduce the possibility of large rounding errors. You must
        also ensure that for such problems the appropriate consistency
        condition on the source terms `q` is satisfied.

    ixn : int
        Is ignored unless `ndir` is equal to zero, in which case it must
        specify the first index of the nodal point at which the solution
        is to be set to zero. The node should not correspond to a corner
        node, or to a node external to the region of interest.

    iyn : int
        Is ignored unless `ndir` is equal to zero, in which case it must
        specify the second index of the nodal point at which the
        solution is to be set to zero. The node should not correspond to
        a corner node, or to a node external to the region of interest.

    conres : float
        The convergence criterion to be used on the maximum absolute
        value of the normalized residual vector components. The latter
        is defined as the residual of the algebraic equation divided by
        the central coefficient when the latter is not equal to 0.0, and
        defined as the residual when the central coefficient is zero.

        Clearly `conres` should not be less than a reasonable multiple
        of the machine precision.

    conchn : float
        The convergence criterion to be used on the maximum absolute
        value of the change made at each iteration to the elements of
        the array `t`, namely the dependent variable. Clearly `conchn`
        should not be less than a reasonable multiple of the machine
        precision multiplied by the maximum value of `t` attained.

        Convergence is achieved when both the convergence criteria are
        satisfied.

        You can therefore set convergence on either the residual or on
        the change, or (as is recommended) on a requirement that both
        are below prescribed limits.

    Returns
    -------
    t : float, ndarray, shape (n1, n2)
        The solution derived by the function.

    itcoun : int
        Its value is increased by the number of iterations used on this
        call (namely `itused`). It therefore stores the accumulated
        number of iterations actually used. For subsequent calls for the
        same problem, i.e., with the same n1 and n2 but possibly
        different coefficients and/or source terms, as occur with
        nonlinear systems or with time-dependent systems, `itcoun` is
        the accumulated number of iterations. This applies to the second
        and subsequent calls to ``dim2_ellip_fd``. In this way a
        suitable cycling of the sequence of iteration arguments is
        obtained in the calls to ``dim2_ellip_fd_iter``.

    itused : int
        The number of iterations actually used on that call.

    resids : float, ndarray, shape (`itmax`)
        The maximum absolute value of the residuals calculated at the
        ith iteration, for i = 1,2,...,`itused`. If you want to know the
        maximum absolute residual of the solution which is returned you
        must calculate this in the calling program. The sequence of
        values `resids` indicates the rate of convergence.

    chngs : float, ndarray, shape (`itmax`)
        The maximum absolute value of the changes made to the components
        of the dependent variable `t` at the ith iteration, for i =
        1,2,...,`itused`. The sequence of values `chngs` indicates the
        rate of convergence.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 > 1.

        (`errno` 1)
            On entry, n1 = *<value>*.

            Constraint: n1 > 1.

        (`errno` 3)
            On entry, `aparam` = *<value>*.

            Constraint: `aparam` > 0.0.

        (`errno` 4)
            On entry, `aparam` = *<value>*, n1 = *<value>* and n2 =
            *<value>*.

            Constraint: `aparam` <= ((n1-1)^2+(n2-1)^2)/2.

        (`errno` 5)
            Convergence not achieved in `itmax` iterations: `itmax` =
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_ellip_fd as b_dim2_ellip_fd

    fun_name = 'naginterfaces.library.pde.dim2_ellip_fd'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _n2_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _itmax = _b_utils._EngineIntScalarType(
        itmax, {'fun_name': fun_name, 'entity': 'itmax'},
    ).to_ctypes()
    _itmax_py = _b_utils._EngineIntScalarType.to_py(_itmax)
    _resids = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_itmax_py,),
    )
    _chngs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_itmax_py,),
    )
    _return_scalars = b_dim2_ellip_fd(
        _n1,
        _n2,
        _a,
        _b,
        _c,
        _d,
        _e,
        _q,
        _t,
        aparam,
        _itmax,
        itcoun,
        ndir,
        ixn,
        iyn,
        conres,
        conchn,
        _resids,
        _chngs,
    )
    _t_py = _t
    _itcoun_py = _return_scalars.itcoun
    _itused_py = _return_scalars.itused
    _resids_py = _resids
    _chngs_py = _chngs
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'itcoun', 'itused', 'resids', 'chngs'],
    )(
        _t_py,
        _itcoun_py,
        _itused_py,
        _resids_py,
        _chngs_py,
    )
    return _return_data

def dim3_ellip_fd(n2, a, b, c, d, e, f, g, q, t, aparam, itmax, itcoun, ndir, ixn, iyn, izn, conres, conchn):
    u"""
    Elliptic PDE, solution of finite difference equations by SIP for
    seven-point three-dimensional molecule, iterate to convergence.

    ``dim3_ellip_fd`` uses the Strongly Implicit Procedure to calculate
    the solution to a system of simultaneous algebraic equations of
    seven-point molecule form on a three-dimensional
    topologically-rectangular mesh. ('Topological' means that a polar
    grid, for example, can be used if it is equivalent to a rectangular
    box.)

    For full information please refer to the NAG Library document for
    d03ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03ecf.html

    Parameters
    ----------
    n2 : int
        The number of nodes in the second coordinate direction, n_2.

    a : float, array-like, shape (n1, `n2`, n3)
        `a`[i-1,j-1,k-1] must contain the coefficient of t_{ij,k-1} in
        the (i,j,k)th equation of the system (1), for k = 1,2,...,n3,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `a`,
        for k = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    b : float, array-like, shape (n1, `n2`, n3)
        `b`[i-1,j-1,k-1] must contain the coefficient of t_{i,j-1,k} in
        the (i,j,k)th equation of the system (1), for k = 1,2,...,n3,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `b`,
        for j = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    c : float, array-like, shape (n1, `n2`, n3)
        `c`[i-1,j-1,k-1] must contain the coefficient of t_{i-1,jk} in
        the (i,j,k)th equation of the system (1), for k = 1,2,...,n3,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `c`,
        for i = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    d : float, array-like, shape (n1, `n2`, n3)
        `d`[i-1,j-1,k-1] must contain the coefficient of t_{ijk} (the
        'central' term) in the (i,j,k)th equation of the system (1), for
        k = 1,2,...,n3, for j = 1,2,...,`n2`, for i = 1,2,...,n1. The
        elements of `d` are checked to ensure that they are nonzero. If
        any element is found to be zero, the corresponding algebraic
        equation is assumed to be t_{ijk} = q_{ijk}. This feature can be
        used to define the equations for nodes at which, for example,
        Dirichlet boundary conditions are applied, or for nodes external
        to the problem of interest. Setting `d`[i-1,j-1,k-1] = 0.0 at
        appropriate points, and the corresponding value of
        `q`[i-1,j-1,k-1] to the appropriate value, namely the prescribed
        value of `t`[i-1,j-1,k-1] in the Dirichlet case, or to zero at
        an external point.

    e : float, array-like, shape (n1, `n2`, n3)
        `e`[i-1,j-1,k-1] must contain the coefficient of t_{i+1,jk} in
        the (i,j,k)th equation of the system (1), for k = 1,2,...,n3,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `e`,
        for i = n1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    f : float, array-like, shape (n1, `n2`, n3)
        `f`[i-1,j-1,k-1] must contain the coefficient of t_{i,j+1,k} in
        the (i,j,k)th equation of the system (1), for k = 1,2,...,n3,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `f`,
        for j = `n2`, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    g : float, array-like, shape (n1, `n2`, n3)
        `g`[i-1,j-1,k-1] must contain the coefficient of t_{ij,k+1} in
        the (i,j,k)th equation of the system (1), for k = 1,2,...,n3,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `g`,
        for k = n3, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    q : float, array-like, shape (n1, `n2`, n3)
        `q`[i-1,j-1,k-1] must contain q_{ijk}, for k = 1,2,...,n3, for j
        = 1,2,...,`n2`, for i = 1,2,...,n1, i.e., the source-term values
        at the nodal points of the system (1).

    t : float, array-like, shape (n1, `n2`, n3)
        `t`[i-1,j-1,k-1] must contain the element t_{ijk} of an
        approximate solution to the equations, for k = 1,2,...,n3, for j
        = 1,2,...,`n2`, for i = 1,2,...,n1.

        If no better approximation is known, an array of zeros can be
        used.

    aparam : float
        The iteration acceleration factor. A value of 1.0 is adequate
        for most typical problems. However, if convergence is slow, the
        value can be reduced, typically to 0.2 or 0.1. If divergence is
        obtained, the value can be increased, typically to 2.0, 5.0 or
        10.0.

    itmax : int
        The maximum number of iterations to be used by the function in
        seeking the solution. A reasonable value might be 20 for a
        problem with 3000 nodes and convergence criteria of about
        10^(-3) of the original residual and change.

    itcoun : int
        On the first call of ``dim3_ellip_fd``, `itcoun` must be set to
        0. On subsequent entries, its value must be unchanged from the
        previous call.

    ndir : int
        Indicates whether or not the system of equations has a unique
        solution. For systems which have a unique solution, `ndir` must
        be set to any nonzero value. For systems derived from problems
        to which an arbitrary constant can be added to the solution, for
        example Poisson's equation with all Neumann boundary conditions,
        `ndir` should be set to 0 and the values of the next three
        arguments must be specified. For such problems the function
        subtracts the value of the function derived at the node (`ixn`,
        `iyn`, `izn`) from the whole solution after each iteration to
        reduce the possibility of large rounding errors. You must also
        ensure for such problems that the appropriate compatibility
        condition on the source terms `q` is satisfied. See the comments
        at the end of [Description].

    ixn : int
        Is ignored unless `ndir` is equal to zero, in which case it must
        specify the first index of the nodal point at which the solution
        is to be set to zero. The node should not correspond to a corner
        node, or to a node external to the region of interest.

    iyn : int
        Is ignored unless `ndir` is equal to zero, in which case it must
        specify the second index of the nodal point at which the
        solution is to be set to zero. The node should not correspond to
        a corner node, or to a node external to the region of interest.

    izn : int
        Is ignored unless `ndir` is equal to zero, in which case it must
        specify the third index of the nodal point at which the solution
        is to be set to zero. The node should not correspond to a corner
        node, or to a node external to the region of interest.

    conres : float
        The convergence criterion to be used on the maximum absolute
        value of the normalized residual vector components. The latter
        is defined as the residual of the algebraic equation divided by
        the central coefficient when the latter is not equal to 0.0, and
        defined as the residual when the central coefficient is zero.

        `conres` should not be less than a reasonable multiple of the
        machine precision.

    conchn : float
        The convergence criterion to be used on the maximum absolute
        value of the change made at each iteration to the elements of
        the array `t`, namely the dependent variable. `conchn` should
        not be less than a reasonable multiple of the machine accuracy
        multiplied by the maximum value of `t` attained.

        Convergence is achieved when both the convergence criteria are
        satisfied.

        You can therefore set convergence on either the residual or on
        the change, or (as is recommended) on a requirement that both
        are below prescribed limits.

    Returns
    -------
    t : float, ndarray, shape (n1, `n2`, n3)
        The solution derived by the function.

    itcoun : int
        Its value is increased by the number of iterations used on this
        call (namely `itused`). It therefore stores the accumulated
        number of iterations actually used.

        For subsequent calls for the same problem, i.e., with the same
        n1, `n2` and n3 but possibly different coefficients and/or
        source terms, as occur with nonlinear systems or with
        time-dependent systems, `itcoun` should not be reset, i.e., it
        must contain the accumulated number of iterations.

        In this way a suitable cycling of the sequence of iteration
        arguments is obtained in the calls to ``dim3_ellip_fd_iter``.

    itused : int
        The number of iterations actually used on that call.

    resids : float, ndarray, shape (`itmax`)
        The maximum absolute value of the residuals calculated at the
        ith iteration, for i = 1,2,...,`itused`. If the residual of the
        solution is sought you must calculate this in the function from
        which ``dim3_ellip_fd`` is called. The sequence of values
        `resids` indicates the rate of convergence.

    chngs : float, ndarray, shape (`itmax`)
        The maximum absolute value of the changes made to the components
        of the dependent variable `t` at the ith iteration, for i =
        1,2,...,`itused`. The sequence of values `chngs` indicates the
        rate of convergence.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n3 = *<value>*.

            Constraint: n3 > 1.

        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` > 1.

        (`errno` 1)
            On entry, n1 = *<value>*.

            Constraint: n1 > 1.

        (`errno` 2)
            On entry, sda = *<value>* and `n2` = *<value>*.

            Constraint: sda >= `n2`.

        (`errno` 3)
            On entry, `aparam` = *<value>*.

            Constraint: `aparam` > 0.0.

        (`errno` 4)
            On entry, `aparam` = *<value>*, n1 = *<value>*, `n2` =
            *<value>* and n3 = *<value>*.

            Constraint: `aparam` <= ((n1-1)^2+(`n2`-1)^2+(n3-1)^2)/3.

        (`errno` 5)
            Convergence not achieved in `itmax` iterations: `itmax` =
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim3_ellip_fd as b_dim3_ellip_fd

    fun_name = 'naginterfaces.library.pde.dim3_ellip_fd'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=3)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _n3_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=3)[2]
    _n3 = _b_utils._EngineIntScalarType(
        _n3_py, {'fun_name': fun_name, 'entity': '_n3'},
    ).to_ctypes()
    _sda_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=3)[1]
    _sda = _b_utils._EngineIntScalarType(
        _sda_py, {'fun_name': fun_name, 'entity': '_sda'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _g = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'g'},
        g,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _itmax = _b_utils._EngineIntScalarType(
        itmax, {'fun_name': fun_name, 'entity': 'itmax'},
    ).to_ctypes()
    _itmax_py = _b_utils._EngineIntScalarType.to_py(_itmax)
    _resids = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_itmax_py,),
    )
    _chngs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_itmax_py,),
    )
    _return_scalars = b_dim3_ellip_fd(
        _n1,
        n2,
        _n3,
        _sda,
        _a,
        _b,
        _c,
        _d,
        _e,
        _f,
        _g,
        _q,
        _t,
        aparam,
        _itmax,
        itcoun,
        ndir,
        ixn,
        iyn,
        izn,
        conres,
        conchn,
        _resids,
        _chngs,
    )
    _t_py = _t
    _itcoun_py = _return_scalars.itcoun
    _itused_py = _return_scalars.itused
    _resids_py = _resids
    _chngs_py = _chngs
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'itcoun', 'itused', 'resids', 'chngs'],
    )(
        _t_py,
        _itcoun_py,
        _itused_py,
        _resids_py,
        _chngs_py,
    )
    return _return_data

def dim2_ellip_mgrid(ngx, ngy, a, rhs, ub, maxit, acc, iout, io_manager=None):
    u"""
    Elliptic PDE, solution of finite difference equations by a multigrid
    technique.

    ``dim2_ellip_mgrid`` solves seven-diagonal systems of linear
    equations which arise from the discretization of an elliptic partial
    differential equation on a rectangular region.
    This function uses a multigrid technique.

    For full information please refer to the NAG Library document for
    d03ed

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03edf.html

    Parameters
    ----------
    ngx : int
        The number of interior grid points in the x-direction, n_x.
        `ngx`-1 should preferably be divisible by as high a power of 2
        as possible.

    ngy : int
        The number of interior grid points in the y-direction, n_y.
        `ngy`-1 should preferably be divisible by as high a power of 2
        as possible.

    a : float, array-like, shape ((4*(`ngx`+1)*(`ngy`+1))/3, 7)
        `a`[i+(j-1)*`ngx`-1,k-1] must be set to `a`_ij^k, for k =
        1,2,...,7, for j = 1,2,...,`ngy`, for i = 1,2,...,`ngx`.

    rhs : float, array-like, shape ((4*(`ngx`+1)*(`ngy`+1))/3)
        `rhs`[i+(j-1)*`ngx`-1] must be set to f_{ij}, for j =
        1,2,...,`ngy`, for i = 1,2,...,`ngx`.

    ub : float, array-like, shape (`ngx`*`ngy`)
        `ub`[i+(j-1)*`ngx`-1] must be set to the initial estimate for
        the solution u_{ij}.

    maxit : int
        The maximum permitted number of multigrid iterations. If `maxit`
        = 0, no multigrid iterations are performed, but the coarse-grid
        approximations and incomplete Crout decompositions are computed,
        and may be output if `iout` is set accordingly.

    acc : float
        The required tolerance for convergence of the residual 2-norm:

            ||r||_2 = sqrt(sum_k = 1^`ngx`*`ngy`(r_k)^2)

        where r = f-Au and u is the computed solution. Note that the
        norm is not scaled by the number of equations. The function will
        stop after fewer than `maxit` iterations if the residual 2-norm
        is less than the specified tolerance. (If `maxit` > 0, at least
        one iteration is always performed.)

        If on entry `acc` = 0.0, the machine precision is used as a
        default value for the tolerance; if `acc` > 0.0, but `acc` is
        less than the machine precision, the function will stop when the
        residual 2-norm is less than the machine precision and errno
        will be set to 4.

    iout : int
        Controls the output of printed information to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        `iout` = 0

          No output.

        `iout` = 1

          The solution u_{ij}, for j = 1,2,...,`ngy`, for i =
          1,2,...,`ngx`.

        `iout` = 2

          The residual 2-norm after each iteration, with the reduction
          factor over the previous iteration.

        `iout` = 3

          As for `iout` = 1 and `iout` = 2.

        `iout` = 4

          As for `iout` = 3, plus the final residual (as returned in
          `ub`).

        `iout` = 5

          As for `iout` = 4, plus the initial elements of `a` and `rhs`.

        `iout` = 6

          As for `iout` = 5, plus the Galerkin coarse grid
          approximations.

        `iout` = 7

          As for `iout` = 6, plus the incomplete Crout decompositions.

        `iout` = 8

          As for `iout` = 7, plus the residual after each iteration.

        The elements `a`[p-1,k-1], the Galerkin coarse grid
        approximations and the incomplete Crout decompositions are
        output in the format:

        - Y-index = j

        - X-index =
          i`a`[p-1,0]`a`[p-1,1]`a`[p-1,2]`a`[p-1,3]`a`[p-1,4]`a`[p-1,5]`a`[p-1,6]

        - where p = i+(j-1)*`ngx`, for j = 1,2,...,`ngy`, for i =
          1,2,...,`ngx`.

        The vectors `u`[p-1], `ub`[p-1], `rhs`[p-1] are output in matrix
        form with `ngy` rows and `ngx` columns.

        Where `ngx` > 10, the `ngx` values for a given j value are
        produced in rows of 10.

        Values of `iout` > 4 may yield considerable amounts of output.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    a : float, ndarray, shape ((4*(`ngx`+1)*(`ngy`+1))/3, 7)
        Is overwritten.

    rhs : float, ndarray, shape ((4*(`ngx`+1)*(`ngy`+1))/3)
        The first `ngx`*`ngy` elements are unchanged and the rest of the
        array is used as workspace.

    ub : float, ndarray, shape (`ngx`*`ngy`)
        The corresponding component of the residual r = f-`a`u.

    us : float, ndarray, shape ((4*(`ngx`+1)*(`ngy`+1))/3)
        The residual 2-norm, stored in element `us`[0].

    u : float, ndarray, shape ((4*(`ngx`+1)*(`ngy`+1))/3)
        The computed solution u_{ij} is returned in
        `u`[i+(j-1)*`ngx`-1], for j = 1,2,...,`ngy`, for i =
        1,2,...,`ngx`.

    numit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `acc` = *<value>*.

            Constraint: `acc` >= 0.0.

        (`errno` 1)
            On entry, `iout` = *<value>*.

            Constraint: 0 <= `iout` <= 8.

        (`errno` 1)
            On entry, `ngy` = *<value>*.

            Constraint: `ngy` >= 3.

        (`errno` 1)
            On entry, `ngx` = *<value>*.

            Constraint: `ngx` >= 3.

        (`errno` 3)
            After `maxit` iterations the residual norm is not less than
            the tolerance `maxit` = *<value>*, residual norm =
            *<value>*, tolerance = *<value>*. The residual norm
            increased at one or more iterations after the first.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            After `maxit` iterations the residual norm is not less than
            the tolerance `maxit` = *<value>*, residual norm =
            *<value>*, tolerance = *<value>*. The residual norm has
            decreased at each iteration after the first.

        (`errno` 4)
            On entry, `acc` is less than machine precision. The function
            terminated because the residual norm is less than machine
            precision. residual norm = *<value>*, machine precision =
            *<value>* and `acc` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_ellip_mgrid as b_dim2_ellip_mgrid

    fun_name = 'naginterfaces.library.pde.dim2_ellip_mgrid'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rhs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rhs'},
        rhs,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ub = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ub'},
        ub,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ngx = _b_utils._EngineIntScalarType(
        ngx, {'fun_name': fun_name, 'entity': 'ngx'},
    ).to_ctypes()
    _ngx_py = _b_utils._EngineIntScalarType.to_py(_ngx)
    _ngy = _b_utils._EngineIntScalarType(
        ngy, {'fun_name': fun_name, 'entity': 'ngy'},
    ).to_ctypes()
    _ngy_py = _b_utils._EngineIntScalarType.to_py(_ngy)
    _us = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((4*(_ngx_py+1)*(_ngy_py+1))//3,),
    )
    _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((4*(_ngx_py+1)*(_ngy_py+1))//3,),
    )
    try:
        _return_scalar = b_dim2_ellip_mgrid(
            _ngx,
            _ngy,
            _a,
            _rhs,
            _ub,
            maxit,
            acc,
            _us,
            _u,
            iout,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _rhs_py = _rhs
    _ub_py = _ub
    _us_py = _us
    _u_py = _u
    _numit_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'rhs', 'ub', 'us', 'u', 'numit'],
    )(
        _a_py,
        _rhs_py,
        _ub_py,
        _us_py,
        _u_py,
        _numit_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim2_ellip_discret(xmin, xmax, ymin, ymax, pdef, bndy, ngx, ngy, scheme, data=None):
    u"""
    Discretize a second-order elliptic PDE on a rectangle.

    ``dim2_ellip_discret`` discretizes a second-order elliptic partial
    differential equation (PDE) on a rectangular region.

    For full information please refer to the NAG Library document for
    d03ee

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03eef.html

    Parameters
    ----------
    xmin : float
        The lower and upper x coordinates of the rectangular region
        respectively, x_A and x_B.

    xmax : float
        The lower and upper x coordinates of the rectangular region
        respectively, x_A and x_B.

    ymin : float
        The lower and upper y coordinates of the rectangular region
        respectively, y_A and y_B.

    ymax : float
        The lower and upper y coordinates of the rectangular region
        respectively, y_A and y_B.

    pdef : callable (alpha, beta, gamma, delta, epslon, phi, psi) =
        pdef(x, y, data=None)
        `pdef` must evaluate the functions alpha(x,y), beta(x,y),
        gamma(x,y), delta(x,y), epsilon(x,y), phi(x,y) and psi(x,y)
        which define the equation at a general point (x,y).

        Parameters
        ~~~~~~~~~~
        x : float
            The x and y coordinates of the point at which the
            coefficients of the partial differential equation are to be
            evaluated.

        y : float
            The x and y coordinates of the point at which the
            coefficients of the partial differential equation are to be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        alpha : float
            `alpha` must be set to the value of alpha(`x`,`y`) at the
            point specified by `x` and `y`.

        beta : float
            `beta` must be set to the value of beta(`x`,`y`) at the
            point specified by `x` and `y`.

        gamma : float
            `gamma` must be set to the value of gamma(`x`,`y`) at the
            point specified by `x` and `y`.

        delta : float
            `delta` must be set to the value of delta(`x`,`y`) at the
            point specified by `x` and `y`.

        epslon : float
            `epslon` must be set to the value of epislon(`x`,`y`) at the
            point specified by `x` and `y`.

        phi : float
            `phi` must be set to the value of phi(`x`,`y`) at the point
            specified by `x` and `y`.

        psi : float
            `psi` must be set to the value of psi(`x`,`y`) at the point
            specified by `x` and `y`.

    bndy : callable (a, b, c) = bndy(x, y, ibnd, data=None)
        `bndy` must evaluate the functions a(x,y), b(x,y), and c(x,y)
        involved in the boundary conditions.

        Parameters
        ~~~~~~~~~~
        x : float
            The x and y coordinates of the point at which the boundary
            conditions are to be evaluated.

        y : float
            The x and y coordinates of the point at which the boundary
            conditions are to be evaluated.

        ibnd : int
            Specifies on which boundary the point (`x`,`y`) lies. `ibnd`
            = 0, 1, 2 or 3 according as the point lies on the bottom,
            right, top or left boundary.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        a : float
            `a`, `b` and `c` must be set to the values of the functions
            appearing in the boundary conditions.

        b : float
            `a`, `b` and `c` must be set to the values of the functions
            appearing in the boundary conditions.

        c : float
            `a`, `b` and `c` must be set to the values of the functions
            appearing in the boundary conditions.

    ngx : int
        The number of interior grid points in the x- and y-directions
        respectively, n_x and n_y. If the seven-diagonal equations are
        to be solved by ``dim2_ellip_mgrid``, `ngx`-1 and `ngy`-1 should
        preferably be divisible by as high a power of 2 as possible.

    ngy : int
        The number of interior grid points in the x- and y-directions
        respectively, n_x and n_y. If the seven-diagonal equations are
        to be solved by ``dim2_ellip_mgrid``, `ngx`-1 and `ngy`-1 should
        preferably be divisible by as high a power of 2 as possible.

    scheme : str, length 1
        The type of approximation to be used for the first derivatives
        which occur in the partial differential equation.

        `scheme` = 'C'

          Central differences are used.

        `scheme` = 'U'

          Upwind differences are used.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : float, ndarray, shape ((4*(`ngx`+1)*(`ngy`+1))/3, 7)
        `a`[i-1,j-1], for j = 1,2,...,7, for i = 1,2,...,`ngx`*`ngy`,
        contains the seven-diagonal linear equations produced by the
        discretization described above. If lda > `ngx`*`ngy`, the
        remaining elements are not referenced by the function, but if
        lda >= (4*(`ngx`+1)*(`ngy`+1))/3 then the array `a` can be
        passed directly to ``dim2_ellip_mgrid``, where these elements
        are used as workspace.

    rhs : float, ndarray, shape ((4*(`ngx`+1)*(`ngy`+1))/3)
        The first `ngx`*`ngy` elements contain the right-hand sides of
        the seven-diagonal linear equations produced by the
        discretization described above. If lda > `ngx`*`ngy`, the
        remaining elements are not referenced by the function, but if
        lda >= (4*(`ngy`+1)*(`ngy`+1))/3 then the array `rhs` can be
        passed directly to ``dim2_ellip_mgrid``, where these elements
        are used as workspace.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `scheme` = *<value>*.

            Constraint: `scheme` = 'C' or 'U'.

        (`errno` 1)
            On entry, `ngy` = *<value>*.

            Constraint: `ngy` >= 3.

        (`errno` 1)
            On entry, `ngx` = *<value>*.

            Constraint: `ngx` >= 3.

        (`errno` 1)
            On entry, `ngx`*`ngy` = *<value>* and lda = *<value>*.

            Constraint: lda must be at least `ngx`*`ngy`.

        (`errno` 1)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymin` < `ymax`.

        (`errno` 1)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at top left boundary. (x,y) = (*<value>*,
            *<value>*).

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at top right boundary. (x,y) = (*<value>*,
            *<value>*).

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at bottom left boundary. (x,y) = (*<value>*,
            *<value>*).

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at bottom right boundary. (x,y) = (*<value>*,
            *<value>*).

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at right boundary. (x,y) = (*<value>*, *<value>*).

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at left boundary. (x,y) = (*<value>*, *<value>*).

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at top boundary. (x,y) = (*<value>*, *<value>*).

        (`errno` 2)
            Mixed derivative in equation and derivative in boundary
            condition at bottom boundary. (x,y) = (*<value>*,
            *<value>*).

        (`errno` 3)
            Null boundary condition at left boundary, top end. (x,y) =
            (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at top boundary, left end. (x,y) =
            (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at right boundary, top end. (x,y) =
            (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at top boundary, right end. (x,y) =
            (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at left boundary, bottom end. (x,y)
            = (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at bottom boundary, left end. (x,y)
            = (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at right boundary, bottom end. (x,y)
            = (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at bottom boundary, right end. (x,y)
            = (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at right boundary. (x,y) =
            (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at left boundary. (x,y) =
            (*<value>*, *<value>*).

        (`errno` 3)
            Null boundary condition at top boundary. (x,y) = (*<value>*,
            *<value>*).

        (`errno` 3)
            Null boundary condition at bottom boundary. (x,y) =
            (*<value>*, *<value>*).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Equation not elliptic at some point.

        (`errno` 5)
            There is no unique solution with Neumann Boundary
            conditions.

        (`errno` 6)
            The linear equations were not diagonally dominant.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_ellip_discret as b_dim2_ellip_discret

    def cb_pdef(x, y, data=None):
        """Wrapping user's pdef. Called as base pdef."""
        if data is not None:
            _alpha_pdef, _beta_pdef, _gamma_pdef, _delta_pdef, _epslon_pdef, _phi_pdef, _psi_pdef = pdef(
                x,
                y,
                data,
            )
        else:
            _alpha_pdef, _beta_pdef, _gamma_pdef, _delta_pdef, _epslon_pdef, _phi_pdef, _psi_pdef = pdef(
                x,
                y,
            )
        _alpha_pdef_base = _alpha_pdef
        _beta_pdef_base = _beta_pdef
        _gamma_pdef_base = _gamma_pdef
        _delta_pdef_base = _delta_pdef
        _epslon_pdef_base = _epslon_pdef
        _phi_pdef_base = _phi_pdef
        _psi_pdef_base = _psi_pdef
        return (
            _alpha_pdef_base,
            _beta_pdef_base,
            _gamma_pdef_base,
            _delta_pdef_base,
            _epslon_pdef_base,
            _phi_pdef_base,
            _psi_pdef_base,
        )

    def cb_bndy(x, y, ibnd, data=None):
        """Wrapping user's bndy. Called as base bndy."""
        if data is not None:
            _a_bndy, _b_bndy, _c_bndy = bndy(
                x,
                y,
                ibnd,
                data,
            )
        else:
            _a_bndy, _b_bndy, _c_bndy = bndy(
                x,
                y,
                ibnd,
            )
        _a_bndy_base = _a_bndy
        _b_bndy_base = _b_bndy
        _c_bndy_base = _c_bndy
        return (
            _a_bndy_base,
            _b_bndy_base,
            _c_bndy_base,
        )

    fun_name = 'naginterfaces.library.pde.dim2_ellip_discret'

    _b_utils._check_callable(pdef, {'fun_name': fun_name, 'entity': 'pdef'})
    _pdef = cb_pdef
    _b_utils._check_callable(bndy, {'fun_name': fun_name, 'entity': 'bndy'})
    _bndy = cb_bndy
    _ngx = _b_utils._EngineIntScalarType(
        ngx, {'fun_name': fun_name, 'entity': 'ngx'},
    ).to_ctypes()
    _ngx_py = _b_utils._EngineIntScalarType.to_py(_ngx)
    _ngy = _b_utils._EngineIntScalarType(
        ngy, {'fun_name': fun_name, 'entity': 'ngy'},
    ).to_ctypes()
    _ngy_py = _b_utils._EngineIntScalarType.to_py(_ngy)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((4*(_ngx_py+1)*(_ngy_py+1))//3, 7),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _rhs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((4*(_ngx_py+1)*(_ngy_py+1))//3,),
    )
    try:
        b_dim2_ellip_discret(
            xmin,
            xmax,
            ymin,
            ymax,
            _pdef,
            _bndy,
            _ngx,
            _ngy,
            _a,
            _rhs,
            scheme,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _rhs_py = _rhs
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'rhs'],
    )(
        _a_py,
        _rhs_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim3_ellip_helmholtz(xs, xf, l, lbdcnd, bdxs, bdxf, ys, yf, mbdcnd, bdys, bdyf, zs, zf, nbdcnd, bdzs, bdzf, lamda, f):
    u"""
    Elliptic PDE, Helmholtz equation, three-dimensional Cartesian
    coordinates.

    ``dim3_ellip_helmholtz`` solves the Helmholtz equation in Cartesian
    coordinates in three dimensions using the standard seven-point
    finite difference approximation.
    This function is designed to be particularly efficient on vector
    processors.

    For full information please refer to the NAG Library document for
    d03fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03faf.html

    Parameters
    ----------
    xs : float
        The lower bound of the range of x, i.e., `xs` <= x <= `xf`.

    xf : float
        The upper bound of the range of x, i.e., `xs` <= x <= `xf`.

    l : int
        The number of panels into which the interval (`xs`,`xf`) is
        subdivided. Hence, there will be `l`+1 grid points in the
        x-direction given by x_i = `xs`+(i-1)*delta x, for i =
        1,2,...,`l`+1, where delta x = (`xf`-`xs`)/`l` is the panel
        width.

    lbdcnd : int
        Indicates the type of boundary conditions at x = `xs` and x =
        `xf`.

        `lbdcnd` = 0

          If the solution is periodic in x, i.e., u(`xs`,y,z) =
          u(`xf`,y,z).

        `lbdcnd` = 1

          If the solution is specified at x = `xs` and x = `xf`.

        `lbdcnd` = 2

          If the solution is specified at x = `xs` and the derivative of
          the solution with respect to x is specified at x = `xf`.

        `lbdcnd` = 3

          If the derivative of the solution with respect to x is
          specified at x = `xs` and x = `xf`.

        `lbdcnd` = 4

          If the derivative of the solution with respect to x is
          specified at x = `xs` and the solution is specified at x =
          `xf`.

    bdxs : float, array-like, shape (m+1, n+1)
        The values of the derivative of the solution with respect to x
        at x = `xs`. When `lbdcnd` = 3 or 4, `bdxs`[j-1,k-1] =
        u_x(`xs`,y_j,z_k), for k = 1,2,...,n+1, for j = 1,2,...,m+1.

        When `lbdcnd` has any other value, `bdxs` is not referenced.

    bdxf : float, array-like, shape (m+1, n+1)
        The values of the derivative of the solution with respect to x
        at x = `xf`. When `lbdcnd` = 2 or 3, `bdxf`[j-1,k-1] =
        u_x(`xf`,y_j,z_k), for k = 1,2,...,n+1, for j = 1,2,...,m+1.

        When `lbdcnd` has any other value, `bdxf` is not referenced.

    ys : float
        The lower bound of the range of y, i.e., `ys` <= y <= `yf`.

    yf : float
        The upper bound of the range of y, i.e., `ys` <= y <= `yf`.

    mbdcnd : int
        Indicates the type of boundary conditions at y = `ys` and y =
        `yf`.

        `mbdcnd` = 0

          If the solution is periodic in y, i.e., u(x,`yf`,z) =
          u(x,`ys`,z).

        `mbdcnd` = 1

          If the solution is specified at y = `ys` and y = `yf`.

        `mbdcnd` = 2

          If the solution is specified at y = `ys` and the derivative of
          the solution with respect to y is specified at y = `yf`.

        `mbdcnd` = 3

          If the derivative of the solution with respect to y is
          specified at y = `ys` and y = `yf`.

        `mbdcnd` = 4

          If the derivative of the solution with respect to y is
          specified at y = `ys` and the solution is specified at y =
          `yf`.

    bdys : float, array-like, shape (`l`+1, n+1)
        The values of the derivative of the solution with respect to y
        at y = `ys`. When `mbdcnd` = 3 or 4, `bdys`[i-1,k-1] =
        u_y(x_i,`ys`,z_k), for k = 1,2,...,n+1, for i = 1,2,...,`l`+1.

        When `mbdcnd` has any other value, `bdys` is not referenced.

    bdyf : float, array-like, shape (`l`+1, n+1)
        The values of the derivative of the solution with respect to y
        at y = `yf`. When `mbdcnd` = 2 or 3, `bdyf`[i-1,k-1] =
        u_y(x_i,`yf`,z_k), for k = 1,2,...,n+1, for i = 1,2,...,`l`+1.

        When `mbdcnd` has any other value, `bdyf` is not referenced.

    zs : float
        The lower bound of the range of z, i.e., `zs` <= z <= `zf`.

    zf : float
        The upper bound of the range of z, i.e., `zs` <= z <= `zf`.

    nbdcnd : int
        Specifies the type of boundary conditions at z = `zs` and z =
        `zf`.

        `nbdcnd` = 0

          if the solution is periodic in z, i.e., u(x,y,`zf`) =
          u(x,y,`zs`).

        `nbdcnd` = 1

          if the solution is specified at z = `zs` and z = `zf`.

        `nbdcnd` = 2

          if the solution is specified at z = `zs` and the derivative of
          the solution with respect to z is specified at z = `zf`.

        `nbdcnd` = 3

          if the derivative of the solution with respect to z is
          specified at z = `zs` and z = `zf`.

        `nbdcnd` = 4

          if the derivative of the solution with respect to z is
          specified at z = `zs` and the solution is specified at z =
          `zf`.

    bdzs : float, array-like, shape (`l`+1, m+1)
        The values of the derivative of the solution with respect to z
        at z = `zs`. When `nbdcnd` = 3 or 4, `bdzs`[i-1,j-1] =
        u_z(x_i,y_j,`zs`), for j = 1,2,...,m+1, for i = 1,2,...,`l`+1.

        When `nbdcnd` has any other value, `bdzs` is not referenced.

    bdzf : float, array-like, shape (`l`+1, m+1)
        The values of the derivative of the solution with respect to z
        at z = `zf`. When `nbdcnd` = 2 or 3, `bdzf`[i-1,j-1] =
        u_z(x_i,y_j,`zf`), for j = 1,2,...,m+1, for i = 1,2,...,`l`+1.

        When `nbdcnd` has any other value, `bdzf` is not referenced.

    lamda : float
        The constant lambda in the Helmholtz equation. For certain
        positive values of lambda a solution to the differential
        equation may not exist, and close to these values the solution
        of the discretized problem will be extremely ill-conditioned. If
        lambda > 0, ``dim3_ellip_helmholtz`` will set `errno` = 3, but
        will still attempt to find a solution. However, since in general
        the values of lambda for which no solution exists cannot be
        predicted a priori, you are advised to treat any results
        computed with lambda > 0 with great caution.

    f : float, array-like, shape (`l`+1, m+1, n+1)
        The values of the right-side of the Helmholtz equation and
        boundary values (if any).

            `f`[i-1,j-1,k-1] = f(x_i,y_j,z_k) i = 2,3,...,`l`,j =
            2,3,...,m and k = 2,3,...,n.

        On the boundaries `f` is defined by

        [table omitted]

        Note: if the table calls for both the solution u and the
        right-hand side f on a boundary, the solution must be specified.

    Returns
    -------
    f : float, ndarray, shape (`l`+1, m+1, n+1)
        Contains the solution u(i,j,k) of the finite difference
        approximation for the grid point (x_i,y_j,z_k), for k =
        1,2,...,n+1, for j = 1,2,...,m+1, for i = 1,2,...,`l`+1.

    pertrb : float
        `pertrb` = 0, unless a solution to Poisson's equation (lambda =
        0) is required with a combination of periodic or derivative
        boundary conditions (`lbdcnd`, `mbdcnd` and `nbdcnd` = 0 or 3).
        In this case a solution may not exist. `pertrb` is a constant,
        calculated and subtracted from the array `f`, which ensures that
        a solution exists. ``dim3_ellip_helmholtz`` then computes this
        solution, which is a least squares solution to the original
        approximation. This solution is not unique and is unnormalized.
        The value of `pertrb` should be small compared to the right-hand
        side `f`, otherwise a solution has been obtained to an
        essentially different problem. This comparison should always be
        made to ensure that a meaningful solution has been obtained.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nbdcnd` = *<value>*.

            Constraint: `nbdcnd` >= 0 and `nbdcnd` <= 4.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 5.

        (`errno` 1)
            On entry, `zs` = *<value>* and `zf` = *<value>*.

            Constraint: `zs` < `zf`.

        (`errno` 1)
            On entry, `mbdcnd` = *<value>*.

            Constraint: `mbdcnd` >= 0 and `mbdcnd` <= 4.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 5.

        (`errno` 1)
            On entry, `ys` = *<value>* and `yf` = *<value>*.

            Constraint: `ys` < `yf`

        (`errno` 1)
            On entry, `lbdcnd` = *<value>*.

            Constraint: `lbdcnd` >= 0 and `lbdcnd` <= 4.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 5.

        (`errno` 1)
            On entry, `xs` = *<value>* and `xf` = *<value>*.

            Constraint: `xs` < `xf`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            lambda > 0.0 in Helmholtz's equation -- a solution may not
            exist.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim3_ellip_helmholtz as b_dim3_ellip_helmholtz

    fun_name = 'naginterfaces.library.pde.dim3_ellip_helmholtz'

    _bdzs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bdzs'},
        bdzs,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_bdzs, {'fun_name': fun_name, 'entity': 'bdzs'}, exp_rank=2)[1]-1, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _bdxs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bdxs'},
        bdxs,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_bdxs, {'fun_name': fun_name, 'entity': 'bdxs'}, exp_rank=2)[1]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _bdxf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bdxf'},
        bdxf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bdys = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bdys'},
        bdys,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bdyf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bdyf'},
        bdyf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bdzf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bdzf'},
        bdzf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = b_dim3_ellip_helmholtz(
            xs,
            xf,
            l,
            lbdcnd,
            _bdxs,
            _bdxf,
            ys,
            yf,
            _m,
            mbdcnd,
            _bdys,
            _bdyf,
            zs,
            zf,
            _n,
            nbdcnd,
            _bdzs,
            _bdzf,
            lamda,
            _f,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _f_py = _f
    _pertrb_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['f', 'pertrb'],
    )(
        _f_py,
        _pertrb_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim2_triangulate(h, m, n, nb, sdindx, isin, data=None):
    u"""
    Triangulation of plane region.

    ``dim2_triangulate`` places a triangular mesh over a given
    two-dimensional region.
    The region may have any shape, including one with holes.

    For full information please refer to the NAG Library document for
    d03ma

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03maf.html

    Parameters
    ----------
    h : float
        h, the required length for the sides of the triangles of the
        uniform mesh.

    m : int
        The value m such that all points inside the region satisfy the
        documented inequalities.

    n : int
        The value n such that all points inside the region satisfy the
        documented inequalities.

    nb : int
        The number of times a triangle side is bisected to find a point
        on the boundary. A value of 10 is adequate for most purposes
        (see [Accuracy]).

    sdindx : int
        The second dimension of the arrays `places` and `indx`.

    isin : callable retval = isin(x, y, data=None)
        `isin` must return the value 1 if the given point (`x`,`y`) lies
        inside the region, and 0 if it lies outside.

        Parameters
        ~~~~~~~~~~
        x : float
            The coordinates of the given point.

        y : float
            The coordinates of the given point.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : int
            The value 1 if the given point (`x`,`y`) lies inside the
            region, and 0 if it lies outside.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    npts : int
        The number of points in the triangulation.

    places : float, ndarray, shape (2, `npts`)
        The x and y coordinates respectively of the ith point of the
        triangulation.

    indx : int, ndarray, shape (4, `npts`)
        `indx`[0,i-1] contains i if point i is inside the region and -i
        if it is on the boundary. For each triangle side between points
        i and j with j > i, `indx`[k-1,i-1], k > 1, contains j or -j
        according to whether point j is internal or on the boundary.
        There can never be more than three such points. If there are
        less, some values `indx`[k-1,i-1], k > 1, are zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `sdindx` is too small: `sdindx` = *<value>*.

        (`errno` 2)
            A point inside the region violates one of the constraints.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` > 2.

        (`errno` 5)
            On entry, `n` = *<value>*.

            Constraint: `n` > 2.

        (`errno` 6)
            On entry, `nb` = *<value>*.

            Constraint: `nb` > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_triangulate as b_dim2_triangulate

    def cb_isin(x, y, data=None):
        """Wrapping user's isin. Called as base isin."""
        if data is not None:
            _retval_isin = isin(
                x,
                y,
                data,
            )
        else:
            _retval_isin = isin(
                x,
                y,
            )
        return _retval_isin

    fun_name = 'naginterfaces.library.pde.dim2_triangulate'

    _sdindx = _b_utils._EngineIntScalarType(
        sdindx, {'fun_name': fun_name, 'entity': 'sdindx'},
    ).to_ctypes()
    _sdindx_py = _b_utils._EngineIntScalarType.to_py(_sdindx)
    _places = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2, _sdindx_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _indx = _b_utils._EngineIntArrayType.empty_ndarray(
        (4, _sdindx_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_utils._check_callable(isin, {'fun_name': fun_name, 'entity': 'isin'})
    _isin = cb_isin
    _return_scalar = b_dim2_triangulate(
        h,
        m,
        n,
        nb,
        _places,
        _indx,
        _sdindx,
        _isin,
        data,
    )
    _npts_py = _return_scalar
    _places_py = _places[:, :_npts_py]
    _indx_py = _indx[:, :_npts_py]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['npts', 'places', 'indx'],
    )(
        _npts_py,
        _places_py,
        _indx_py,
    )
    return _return_data

def dim1_blackscholes_fd(kopt, x, mesh, s, t, tdpar, r, q, sigma, ntkeep, alpha=0.55):
    u"""
    Finite difference solution of the Black--Scholes equations.

    ``dim1_blackscholes_fd`` solves the Black--Scholes equation for
    financial option pricing using a finite difference scheme.

    For full information please refer to the NAG Library document for
    d03nc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03ncf.html

    Parameters
    ----------
    kopt : int
        Specifies the kind of option to be valued.

        `kopt` = 1

          A European call option.

        `kopt` = 2

          An American call option.

        `kopt` = 3

          A European put option.

        `kopt` = 4

          An American put option.

    x : float
        The exercise price X.

    mesh : str, length 1
        Indicates the type of finite difference mesh to be used:

        `mesh` = 'U'

          Uniform mesh.

        `mesh` = 'C'

          Custom mesh supplied by you.

    s : float, array-like, shape (ns)
        If `mesh` = 'C', `s`[i-1] must contain the ith stock price in
        the mesh, for i = 1,2,...,ns. These values should be in
        increasing order, with `s`[0] = S_min and `s`[ns-1] = S_max.

        If `mesh` = 'U', `s`[0] must be set to S_min and `s`[ns-1] to
        S_max, but `s`[1],`s`[2],...,`s`[ns-2] need not be initialized,
        as they will be set internally by the function in order to
        define a uniform mesh.

    t : float, array-like, shape (nt)
        If `mesh` = 'C' then `t`[j-1] must contain the jth time in the
        mesh, for j = 1,2,...,nt. These values should be in increasing
        order, with `t`[0] = t_min and `t`[nt-1] = t_max.

        If `mesh` = 'U' then `t`[0] must be set to t_min and `t`[nt-1]
        to t_max, but `t`[1],`t`[2],...,`t`[nt-2] need not be
        initialized, as they will be set internally by the function in
        order to define a uniform mesh.

    tdpar : bool, array-like, shape (3)
        Specifies whether or not various arguments are time-dependent.
        More precisely, r is time-dependent if `tdpar`[0] = True and
        constant otherwise. Similarly, `tdpar`[1] specifies whether q is
        time-dependent and `tdpar`[2] specifies whether sigma is
        time-dependent.

    r : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `tdpar`[0]= True: nt; otherwise: 1.

        If `tdpar`[0] = True then `r`[j-1] must contain the value of the
        risk-free interest rate r(t) at the jth time in the mesh, for j
        = 1,2,...,nt.

        If `tdpar`[0] = False then `r`[0] must contain the constant
        value of the risk-free interest rate r.

        The remaining elements need not be set.

    q : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `tdpar`[1]= True: nt; otherwise: 1.

        If `tdpar`[1] = True then `q`[j-1] must contain the value of the
        continuous dividend q(t) at the jth time in the mesh, for j =
        1,2,...,nt.

        If `tdpar`[1] = False then `q`[0] must contain the constant
        value of the continuous dividend q.

        The remaining elements need not be set.

    sigma : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `tdpar`[2]= True: nt; otherwise: 1.

        If `tdpar`[2] = True then `sigma`[j-1] must contain the value of
        the volatility sigma(t) at the jth time in the mesh, for j =
        1,2,...,nt.

        If `tdpar`[2] = False then `sigma`[0] must contain the constant
        value of the volatility sigma.

        The remaining elements need not be set.

    ntkeep : int
        The number of solutions to be stored in the time direction. The
        function calculates the solution backwards from `t`[nt-1] to
        `t`[0] at all times in the mesh. These time solutions and the
        corresponding Greeks will be stored at times `t`[i-1], for i =
        1,2,...,`ntkeep`, in the arrays `f`, `theta`, `delta`, `gamma`,
        `lamda` and `rho`. Other time solutions will be discarded. To
        store all time solutions set `ntkeep` = nt.

    alpha : float, optional
        The value of lambda to be used in the time-stepping scheme.
        Typical values include:

        `alpha` = 0.0

          Explicit forward Euler scheme.

        `alpha` = 0.5

          Implicit Crank--Nicolson scheme.

        `alpha` = 1.0

          Implicit backward Euler scheme.

        The value 0.5 gives second-order accuracy in time.

        Values greater than 0.5 give unconditional stability.

        Since 0.5 is at the limit of unconditional stability this value
        does not damp oscillations.

    Returns
    -------
    s : float, ndarray, shape (ns)
        If `mesh` = 'U', the elements of `s` define a uniform mesh over
        [S_min,S_max].

        If `mesh` = 'C', the elements of `s` are unchanged.

    t : float, ndarray, shape (nt)
        If `mesh` = 'U', the elements of `t` define a uniform mesh over
        [t_min,t_max].

        If `mesh` = 'C', the elements of `t` are unchanged.

    f : float, ndarray, shape (ns, `ntkeep`)
        `f`[i-1,j-1], for j = 1,2,...,`ntkeep`, for i = 1,2,...,ns,
        contains the value f of the option at the ith mesh point
        `s`[i-1] at time `t`[j-1].

    theta : float, ndarray, shape (ns, `ntkeep`)
        The values of various Greeks at the ith mesh point `s`[i-1] at
        time `t`[j-1], as follows:

            [table omitted]

    delta : float, ndarray, shape (ns, `ntkeep`)
        The values of various Greeks at the ith mesh point `s`[i-1] at
        time `t`[j-1], as follows:

            [table omitted]

    gamma : float, ndarray, shape (ns, `ntkeep`)
        The values of various Greeks at the ith mesh point `s`[i-1] at
        time `t`[j-1], as follows:

            [table omitted]

    lamda : float, ndarray, shape (ns, `ntkeep`)
        The values of various Greeks at the ith mesh point `s`[i-1] at
        time `t`[j-1], as follows:

            [table omitted]

    rho : float, ndarray, shape (ns, `ntkeep`)
        The values of various Greeks at the ith mesh point `s`[i-1] at
        time `t`[j-1], as follows:

            [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ntkeep` = *<value>* and nt = *<value>*.

            Constraint: `ntkeep` <= nt.

        (`errno` 1)
            On entry, `ntkeep` = *<value>*.

            Constraint: `ntkeep` >= 1.

        (`errno` 1)
            On entry, `alpha` = *<value>*.

            Constraint: `alpha` <= 1.0.

        (`errno` 1)
            On entry, `alpha` = *<value>*.

            Constraint: `alpha` >= 0.0.

        (`errno` 1)
            On entry, `mesh` = *<value>*.

            Constraint: `mesh` = 'U' or 'C'.

        (`errno` 1)
            On entry, `t`[0] = *<value>*.

            Constraint: `t`[0] >= 0.0.

        (`errno` 1)
            On entry, `s`[0] = *<value>*.

            Constraint: `s`[0] >= 0.0.

        (`errno` 1)
            On entry, nt = *<value>*.

            Constraint: nt >= 2.

        (`errno` 1)
            On entry, ns = *<value>*.

            Constraint: ns >= 2.

        (`errno` 1)
            On entry, `kopt` = *<value>*.

            Constraint: `kopt` = 1, 2, 3 or 4.

        (`errno` 2)
            On entry, `t`[nt-1] = *<value>* and `t`[0] = *<value>*.

            Constraint: `t`[nt-1] > `t`[0].

        (`errno` 2)
            On entry, `s`[ns-1] = *<value>* and `s`[0] = *<value>*.

            Constraint: `s`[ns-1] > `s`[0].

        (`errno` 3)
            On entry, `t`[0] <= `t`[*<value>*].

            Constraint: when `mesh` = 'C', `t`[i-1] < `t`[i], for i =
            1,2,...,nt-1.

        (`errno` 3)
            On entry, `s`[0] <= `s`[*<value>*].

            Constraint: when `mesh` = 'C', `s`[i-1] < `s`[i], for i =
            1,2,...,ns-1.
    """

    from ..base.pde import dim1_blackscholes_fd as b_dim1_blackscholes_fd

    fun_name = 'naginterfaces.library.pde.dim1_blackscholes_fd'

    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ns_py = _b_utils._EngineFloat64ArrayType.get_shape(_s, {'fun_name': fun_name, 'entity': 's'}, exp_rank=1)[0]
    _ns = _b_utils._EngineIntScalarType(
        _ns_py, {'fun_name': fun_name, 'entity': '_ns'},
    ).to_ctypes()
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt_py = _b_utils._EngineFloat64ArrayType.get_shape(_t, {'fun_name': fun_name, 'entity': 't'}, exp_rank=1)[0]
    _nt = _b_utils._EngineIntScalarType(
        _nt_py, {'fun_name': fun_name, 'entity': '_nt'},
    ).to_ctypes()
    _tdpar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tdpar'},
        tdpar,
        copy=False,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sigma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sigma'},
        sigma,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ntkeep = _b_utils._EngineIntScalarType(
        ntkeep, {'fun_name': fun_name, 'entity': 'ntkeep'},
    ).to_ctypes()
    _ntkeep_py = _b_utils._EngineIntScalarType.to_py(_ntkeep)
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ntkeep_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _theta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ntkeep_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _delta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ntkeep_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _gamma = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ntkeep_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _lamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ntkeep_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _rho = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ntkeep_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_dim1_blackscholes_fd(
        kopt,
        x,
        mesh,
        _ns,
        _s,
        _nt,
        _t,
        _tdpar,
        _r,
        _q,
        _sigma,
        alpha,
        _ntkeep,
        _f,
        _theta,
        _delta,
        _gamma,
        _lamda,
        _rho,
    )
    _s_py = _s
    _t_py = _t
    _f_py = _f
    _theta_py = _theta
    _delta_py = _delta
    _gamma_py = _gamma
    _lamda_py = _lamda
    _rho_py = _rho
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 't', 'f', 'theta', 'delta', 'gamma', 'lamda', 'rho'],
    )(
        _s_py,
        _t_py,
        _f_py,
        _theta_py,
        _delta_py,
        _gamma_py,
        _lamda_py,
        _rho_py,
    )
    return _return_data

def dim1_blackscholes_closed(kopt, x, s, t, tmat, tdpar, r, q, sigma):
    u"""
    Analytic solution of the Black--Scholes equations.

    ``dim1_blackscholes_closed`` computes an analytic solution to the
    Black--Scholes equation for a certain set of option types.

    For full information please refer to the NAG Library document for
    d03nd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03ndf.html

    Parameters
    ----------
    kopt : int
        Specifies the kind of option to be valued:

        `kopt` = 1

          A European call option.

        `kopt` = 2

          An American call option.

        `kopt` = 3

          A European put option.

    x : float
        The exercise price X.

    s : float
        The stock price at which the option value and the Greeks should
        be evaluated.

    t : float
        The time at which the option value and the Greeks should be
        evaluated.

    tmat : float
        The maturity time of the option.

    tdpar : bool, array-like, shape (3)
        Specifies whether or not various arguments are time-dependent.
        More precisely, r is time-dependent if `tdpar`[0] = True and
        constant otherwise. Similarly, `tdpar`[1] specifies whether q is
        time-dependent and `tdpar`[2] specifies whether sigma is
        time-dependent.

    r : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `tdpar`[0]= True: 3; otherwise: 1.

        If `tdpar`[0] = False then `r`[0] must contain the constant
        value of r. The remaining elements need not be set.

        If `tdpar`[0] = True then `r`[0] must contain the value of r at
        time `t` and `r`[1] must contain its average instantaneous value
        over the remaining life of the option:

            r^= int_`t`^`tmat`r(zeta)d zeta.

        The auxiliary function ``dim1_blackscholes_means`` may be used
        to construct `r` from a set of values of r at discrete times.

    q : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `tdpar`[1]= True: 3; otherwise: 1.

        If `tdpar`[1] = False then `q`[0] must contain the constant
        value of q. The remaining elements need not be set.

        If `tdpar`[1] = True then `q`[0] must contain the constant value
        of q and `q`[1] must contain its average instantaneous value
        over the remaining life of the option:

            q^= int_`t`^`tmat`q(zeta)d zeta.

        The auxiliary function ``dim1_blackscholes_means`` may be used
        to construct `q` from a set of values of q at discrete times.

    sigma : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `tdpar`[2]= True: 3; otherwise: 1.

        If `tdpar`[2] = False then `sigma`[0] must contain the constant
        value of sigma. The remaining elements need not be set.

        If `tdpar`[2] = True then `sigma`[0] must contain the value of
        sigma at time `t`, `sigma`[1] the average instantaneous value
        sigma^, and `sigma`[2] the second-order average sigma_, where:

            sigma^= int_`t`^`tmat`sigma(zeta)d zeta,

            sigma_= (int_`t`^`tmat`sigma^2(zeta)d zeta)^(1/2).

        The auxiliary function ``dim1_blackscholes_means`` may be used
        to compute `sigma` from a set of values at discrete times.

    Returns
    -------
    f : float
        The value f of the option at the stock price `s` and time `t`.

    theta : float
        The values of various Greeks at the stock price `s` and time
        `t`, as follows:

            [table omitted]

    delta : float
        The values of various Greeks at the stock price `s` and time
        `t`, as follows:

            [table omitted]

    gamma : float
        The values of various Greeks at the stock price `s` and time
        `t`, as follows:

            [table omitted]

    lamda : float
        The values of various Greeks at the stock price `s` and time
        `t`, as follows:

            [table omitted]

    rho : float
        The values of various Greeks at the stock price `s` and time
        `t`, as follows:

            [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `sigma`[*<value>*] = *<value>*.

            Constraint: `sigma`[i-1] > 0.0.

        (`errno` 1)
            On entry, `tmat` = *<value>* and `t` = *<value>*.

            Constraint: `tmat` >= `t`.

        (`errno` 1)
            On entry, `t` = *<value>*.

            Constraint: `t` >= 0.0.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 0.0.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 1)
            On entry, `q`[0] is not equal to 0.0 with American call
            option. `q`[0] = *<value>*.

        (`errno` 1)
            On entry, `kopt` = *<value>*.

            Constraint: `kopt` = 1, 2 or 3.
    """

    from ..base.pde import dim1_blackscholes_closed as b_dim1_blackscholes_closed

    fun_name = 'naginterfaces.library.pde.dim1_blackscholes_closed'

    _tdpar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tdpar'},
        tdpar,
        copy=False,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sigma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sigma'},
        sigma,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalars = b_dim1_blackscholes_closed(
        kopt,
        x,
        s,
        t,
        tmat,
        _tdpar,
        _r,
        _q,
        _sigma,
    )
    _f_py = _return_scalars.f
    _theta_py = _return_scalars.theta
    _delta_py = _return_scalars.delta
    _gamma_py = _return_scalars.gamma
    _lamda_py = _return_scalars.lamda
    _rho_py = _return_scalars.rho
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['f', 'theta', 'delta', 'gamma', 'lamda', 'rho'],
    )(
        _f_py,
        _theta_py,
        _delta_py,
        _gamma_py,
        _lamda_py,
        _rho_py,
    )
    return _return_data

def dim1_blackscholes_means(t0, tmat, td, phid):
    u"""
    Compute average values for ``dim1_blackscholes_closed``.

    ``dim1_blackscholes_means`` computes average values of a continuous
    function of time over the remaining life of an option.
    It is used together with ``dim1_blackscholes_closed`` to value
    options with time-dependent arguments.

    For full information please refer to the NAG Library document for
    d03ne

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03nef.html

    Parameters
    ----------
    t0 : float
        The current time t_0.

    tmat : float
        The maturity time T.

    td : float, array-like, shape (ntd)
        The discrete times at which phi is specified.

    phid : float, array-like, shape (ntd)
        `phid`[i-1] must contain the value of phi at time `td`[i-1], for
        i = 1,2,...,ntd.

    Returns
    -------
    phiav : float, ndarray, shape (3)
        `phiav`[0] contains the value of phi interpolated to t_0,
        `phiav`[1] contains the first-order average phi^ and `phiav`[2]
        contains the second-order average phi_, where:

            phi^= 1/(T-t_0)int_t_0^T phi(zeta)d zeta, phi_=
            (1/(T-t_0)int_t_0^T phi^2(zeta)d zeta)^(1/2).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `td`[0] <= `td`[*<value>*].

            Constraint: `td`[0] < `td`[1] < ... < `td`[ntd-1].

        (`errno` 1)
            On entry, ntd = *<value>*.

            Constraint: ntd >= 2.

        (`errno` 1)
            On entry, `tmat` = *<value>*, `td`[0] = *<value>* and
            `td`[ntd-1] = *<value>*.

            Constraint: `td`[0] <= `tmat` <= `td`[ntd-1].

        (`errno` 1)
            On entry, `t0` = *<value>*, `td`[0] = *<value>* and
            `td`[ntd-1] = *<value>*.

            Constraint: `td`[0] <= `t0` <= `td`[ntd-1].

        (`errno` 2)
            Unexpected failure in internal call to spline function.
    """

    from ..base.pde import dim1_blackscholes_means as b_dim1_blackscholes_means

    fun_name = 'naginterfaces.library.pde.dim1_blackscholes_means'

    _td = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'td'},
        td,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ntd_py = _b_utils._EngineFloat64ArrayType.get_shape(_td, {'fun_name': fun_name, 'entity': 'td'}, exp_rank=1)[0]
    _ntd = _b_utils._EngineIntScalarType(
        _ntd_py, {'fun_name': fun_name, 'entity': '_ntd'},
    ).to_ctypes()
    _phid = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'phid'},
        phid,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _phiav = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    b_dim1_blackscholes_means(
        t0,
        tmat,
        _ntd,
        _td,
        _phid,
        _phiav,
    )
    _phiav_py = _phiav
    return _phiav_py

def dim1_parab_fd(m, ts, tout, pdedef, bndary, u, x, acc, comm, itask, itrace, ind, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of parabolic PDEs, method of lines, finite
    differences, one space variable.

    ``dim1_parab_fd`` integrates a system of linear or nonlinear
    parabolic partial differential equations (PDEs) in one space
    variable.
    The spatial discretization is performed using finite differences,
    and the method of lines is employed to reduce the PDEs to a system
    of ordinary differential equations (ODEs).
    The resulting system is solved using a backward differentiation
    formula method.

    For full information please refer to the NAG Library document for
    d03pc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pcf.html

    Parameters
    ----------
    m : int
        The coordinate system used:

        `m` = 0

          Indicates Cartesian coordinates.

        `m` = 1

          Indicates cylindrical polar coordinates.

        `m` = 2

          Indicates spherical polar coordinates.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (p, q, r, ires) = pdedef(t, x, u, ux, ires,
        data=None)
        `pdedef` must compute the functions P_{i,j}, Q_i and R_i which
        define the system of PDEs. `pdedef` is called approximately
        midway between each pair of mesh points in turn by
        ``dim1_parab_fd``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,npde.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,npde.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde)
            `p`[i-1,j-1] must be set to the value of P_{i,j}(x,t,U,U_x),
            for j = 1,2,...,npde, for i = 1,2,...,npde.

        q : float, array-like, shape (npde)
            `q`[i-1] must be set to the value of Q_i(x,t,U,U_x), for i =
            1,2,...,npde.

        r : float, array-like, shape (npde)
            `r`[i-1] must be set to the value of R_i(x,t,U,U_x), for i =
            1,2,...,npde.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    bndary : callable (beta, gamma, ires) = bndary(t, u, ux, ibnd, ires,
        data=None)
        `bndary` must compute the functions beta_i and gamma_i which
        define the boundary conditions as in equation (3).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,npde.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx
            at the boundary specified by `ibnd`, for i = 1,2,...,npde.

        ibnd : int
            Determines the position of the boundary conditions.

            `ibnd` = 0

              `bndary` must set up the coefficients of the left-hand
              boundary, x = a.

            `ibnd` != 0

              Indicates that `bndary` must set up the coefficients of
              the right-hand boundary, x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        beta : float, array-like, shape (npde)
            `beta`[i-1] must be set to the value of beta_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,npde.

        gamma : float, array-like, shape (npde)
            `gamma`[i-1] must be set to the value of gamma_i(x,t,U,U_x)
            at the boundary specified by `ibnd`, for i = 1,2,...,npde.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    u : float, array-like, shape (npde, npts)
        The initial values of U(x,t) at t = `ts` and the mesh points
        `x`[j-1], for j = 1,2,...,npts.

    x : float, array-like, shape (npts)
        The mesh points in the spatial direction. `x`[0] must specify
        the left-hand boundary, a, and `x`[npts-1] must specify the
        right-hand boundary, b.

    acc : float
        A positive quantity for controlling the local error estimate in
        the time integration. If E(i,j) is the estimated error for U_i
        at the jth mesh point, the error test is:

            |E(i,j)| = `acc`*(1.0+|`u`[i-1,j-1]|).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    itask : int
        Specifies the task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout`.

        `itask` = 2

          One step and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

    itrace : int
        The level of trace information required from ``dim1_parab_fd``
        and the underlying ODE solver. `itrace` may take the value -1,
        0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` > 0

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases. You are advised to set `itrace` = 0, unless you are
        experienced with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_fd``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `u` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (npde, npts)
        `u`[i-1,j-1] will contain the computed solution at t = `ts`.

    ind : int
        `ind` = 1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, `acc` = *<value>*.

            Constraint: `acc` > 0.0.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `m` = *<value>* and `x`[0] = *<value>*.

            Constraint: `m` <= 0 or `x`[0] >= 0.0

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` = 0, 1 or 2.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2 or 3.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `acc` was too small to start integration: `acc` = *<value>*.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef` or
            `bndary`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 14)
            Flux function appears to depend on time derivatives.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied value of `acc`. `ts` =
            *<value>*, `acc` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef` or `bndary`. Integration is successful as
            far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but a small change in `acc` is
            unlikely to result in a changed solution. `acc` = *<value>*.
    """

    from ..base.pde import dim1_parab_fd as b_dim1_parab_fd

    fun_name = 'naginterfaces.library.pde.dim1_parab_fd'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ux, p, q, r, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                ires,
                data,
            )
        else:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        q[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'q'},
            _q_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'r'},
            _r_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_pdedef_np
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, u, ux, ibnd, beta, gamma, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                ibnd,
                ires,
                data,
            )
        else:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                ibnd,
                ires,
            )
        beta[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'beta'},
            _beta_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        gamma[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'gamma'},
            _gamma_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_u, 'u'),),
        spiked_sorder=spiked_sorder,
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[1]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_dim1_parab_fd(
            _npde,
            m,
            ts,
            tout,
            _pdedef,
            _bndary,
            _u,
            _npts,
            _x,
            acc,
            comm,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'ind'],
    )(
        _ts_py,
        _u_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_coll(m, ts, tout, pdedef, bndary, u, xbkpts, npoly, uinit, acc, comm, itask, itrace, ind, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of parabolic PDEs, method of lines, Chebyshev C^0
    collocation, one space variable.

    ``dim1_parab_coll`` integrates a system of linear or nonlinear
    parabolic partial differential equations (PDEs) in one space
    variable.
    The spatial discretization is performed using a Chebyshev C^0
    collocation method, and the method of lines is employed to reduce
    the PDEs to a system of ordinary differential equations (ODEs).
    The resulting system is solved using a backward differentiation
    formula method.

    For full information please refer to the NAG Library document for
    d03pd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pdf.html

    Parameters
    ----------
    m : int
        The coordinate system used:

        `m` = 0

          Indicates Cartesian coordinates.

        `m` = 1

          Indicates cylindrical polar coordinates.

        `m` = 2

          Indicates spherical polar coordinates.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (p, q, r, ires) = pdedef(t, x, u, ux, ires,
        data=None)
        `pdedef` must compute the values of the functions P_{i,j}, Q_i
        and R_i which define the system of PDEs.

        The functions may depend on x, t, U and U_x and must be
        evaluated at a set of points.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (nptl)
            Contains a set of mesh points at which P_{i,j}, Q_i and R_i
            are to be evaluated. `x`[0] and `x`[nptl-1] contain
            successive user-supplied break-points and the elements of
            the array will satisfy `x`[0] < `x`[1] < ... < `x`[nptl-1].

        u : float, ndarray, shape (npde, nptl)
            `u`[i-1,j-1] contains the value of the component U_i(x,t)
            where x = `x`[j-1], for j = 1,2,...,nptl, for i =
            1,2,...,`npde`.

        ux : float, ndarray, shape (npde, nptl)
            `ux`[i-1,j-1] contains the value of the component
            dU_i(x,t)/dx where x = `x`[j-1], for j = 1,2,...,nptl, for i
            = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde, nptl)
            `p`[ `npde`*`npde`*(k-1)+ `npde`*(j-1)+ (i-1)] must be set
            to the value of P_{i,j}(x,t,U,U_x) where x = `x`[k-1], for k
            = 1,2,...,nptl, for j = 1,2,...,`npde`, for i =
            1,2,...,`npde`.

        q : float, array-like, shape (npde, nptl)
            `q`[i-1,j-1] must be set to the value of Q_i(x,t,U,U_x)
            where x = `x`[j-1], for j = 1,2,...,nptl, for i =
            1,2,...,`npde`.

        r : float, array-like, shape (npde, nptl)
            `r`[i-1,j-1] must be set to the value of R_i(x,t,U,U_x)
            where x = `x`[j-1], for j = 1,2,...,nptl, for i =
            1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_coll``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    bndary : callable (beta, gamma, ires) = bndary(t, u, ux, ibnd, ires,
        data=None)
        `bndary` must compute the functions beta_i and gamma_i which
        define the boundary conditions as in equation (3).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx
            at the boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must set up the coefficients of the left-hand
              boundary, x = a.

            `ibnd` != 0

              `bndary` must set up the coefficients of the right-hand
              boundary, x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        beta : float, array-like, shape (npde)
            `beta`[i-1] must be set to the value of beta_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        gamma : float, array-like, shape (npde)
            `gamma`[i-1] must be set to the value of gamma_i(x,t,U,U_x)
            at the boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_coll``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    u : float, array-like, shape (npde, (nbkpts-1)*`npoly`+1)
        If `ind` = 1 the value of `u` must be unchanged from the
        previous call.

    xbkpts : float, array-like, shape (nbkpts)
        The values of the break-points in the space direction.
        `xbkpts`[0] must specify the left-hand boundary, a, and
        `xbkpts`[nbkpts-1] must specify the right-hand boundary, b.

    npoly : int
        The degree of the Chebyshev polynomial to be used in
        approximating the PDE solution between each pair of
        break-points.

    uinit : callable u = uinit(npde, x, data=None)
        `uinit` must compute the initial values of the PDE components
        U_i(x_j,t_0), for j = 1,2,...,npts, for i = 1,2,...,`npde`.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        x : float, ndarray, shape (npts)
            `x`[j-1], contains the values of the jth mesh point, for j =
            1,2,...,npts.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float, array-like, shape (`npde`, npts)
            `u`[i-1,j-1] must be set to the initial value U_i(x_j,t_0),
            for j = 1,2,...,npts, for i = 1,2,...,`npde`.

    acc : float
        A positive quantity for controlling the local error estimate in
        the time integration. If E(i,j) is the estimated error for U_i
        at the jth mesh point, the error test is:

            |E(i,j)| = `acc`*(1.0+|`u`[i-1,j-1]|).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    itask : int
        Specifies the task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout`.

        `itask` = 2

          One step and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

    itrace : int
        The level of trace information required from ``dim1_parab_coll``
        and the underlying ODE solver. `itrace` may take the value -1,
        0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` > 0

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases. You are advised to set `itrace` = 0, unless you are
        experienced with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_coll``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `u` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (npde, (nbkpts-1)*`npoly`+1)
        `u`[i-1,j-1] will contain the computed solution at t = `ts`.

    x : float, ndarray, shape ((nbkpts-1)*`npoly`+1)
        The mesh points chosen by ``dim1_parab_coll`` in the spatial
        direction. The values of `x` will satisfy `x`[0] < `x`[1] < ...
        < `x`[npts-1].

    ind : int
        `ind` = 1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, i = *<value>*, `xbkpts`[i-1] = *<value>*, j =
            *<value>* and `xbkpts`[j-1] = *<value>*.

            Constraint: `xbkpts`[0] < `xbkpts`[1] < ... <
            `xbkpts`[nbkpts-1].

        (`errno` 1)
            On entry, `acc` = *<value>*.

            Constraint: `acc` > 0.0.

        (`errno` 1)
            On entry, npts = *<value>*, nbkpts = *<value>* and `npoly` =
            *<value>*.

            Constraint: npts = (nbkpts-1)*`npoly`+1.

        (`errno` 1)
            On entry, `npoly` = *<value>*.

            Constraint: `npoly` <= 49.

        (`errno` 1)
            On entry, `npoly` = *<value>*.

            Constraint: `npoly` >= 1.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde >= 1.

        (`errno` 1)
            On entry, nbkpts = *<value>*.

            Constraint: nbkpts >= 2.

        (`errno` 1)
            On entry, `m` = *<value>* and `xbkpts`[0] = *<value>*.

            Constraint: `m` <= 0 or `xbkpts`[0] >= 0.0

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` = 0, 1 or 2.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2 or 3.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `acc` was too small to start integration: `acc` = *<value>*.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef` or
            `bndary`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 14)
            Flux function appears to depend on time derivatives.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied value of `acc`. `ts` =
            *<value>*, `acc` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef` or `bndary`. Integration is successful as
            far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but a small change in `acc` is
            unlikely to result in a changed solution. `acc` = *<value>*.
    """

    from ..base.pde import dim1_parab_coll as b_dim1_parab_coll

    fun_name = 'naginterfaces.library.pde.dim1_parab_coll'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, nptl, u, ux, p, q, r, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                ires,
                data,
            )
        else:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde, nptl),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _q_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'q'},
            _q_pdedef,
            copy=False,
            exp_shape=(npde, nptl),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _r_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'r'},
            _r_pdedef,
            copy=False,
            exp_shape=(npde, nptl),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'), (_q_pdedef_np, 'q'), (_r_pdedef_np, 'r')),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :, :] = _p_pdedef_np
        q[:, :] = _q_pdedef_np
        r[:, :] = _r_pdedef_np
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, u, ux, ibnd, beta, gamma, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                ibnd,
                ires,
                data,
            )
        else:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                ibnd,
                ires,
            )
        beta[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'beta'},
            _beta_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        gamma[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'gamma'},
            _gamma_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_uinit_name = fun_name + ':uinit'

    def cb_uinit(npde, npts, x, u, data=None):
        """Wrapping user's uinit. Called as base uinit."""
        if data is not None:
            _u_uinit = uinit(
                npde,
                x,
                data,
            )
        else:
            _u_uinit = uinit(
                npde,
                x,
            )
        _u_uinit_np = _utils._to_ndarray(
            {'fun_name': cb_uinit_name, 'returned': 'u'},
            _u_uinit,
            copy=False,
            exp_shape=(npde, npts),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_uinit_name},
            ((_u_uinit_np, 'u'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        u[:, :] = _u_uinit_np

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_u, 'u'),),
        spiked_sorder=spiked_sorder,
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[1]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _xbkpts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xbkpts'},
        xbkpts,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nbkpts_py = _b_utils._EngineFloat64ArrayType.get_shape(_xbkpts, {'fun_name': fun_name, 'entity': 'xbkpts'}, exp_rank=1)[0]
    _nbkpts = _b_utils._EngineIntScalarType(
        _nbkpts_py, {'fun_name': fun_name, 'entity': '_nbkpts'},
    ).to_ctypes()
    _npoly = _b_utils._EngineIntScalarType(
        npoly, {'fun_name': fun_name, 'entity': 'npoly'},
    ).to_ctypes()
    _npoly_py = _b_utils._EngineIntScalarType.to_py(_npoly)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_nbkpts_py-1)*_npoly_py+1,),
    )
    _b_utils._check_callable(uinit, {'fun_name': fun_name, 'entity': 'uinit'})
    _uinit = cb_uinit
    try:
        _return_scalars = b_dim1_parab_coll(
            _npde,
            m,
            ts,
            tout,
            _pdedef,
            _bndary,
            _u,
            _nbkpts,
            _xbkpts,
            _npoly,
            _npts,
            _x,
            _uinit,
            acc,
            comm,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'x', 'ind'],
    )(
        _ts_py,
        _u_py,
        _x_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_keller(ts, tout, pdedef, bndary, u, x, nleft, acc, comm, itask, itrace, ind, data=None, io_manager=None):
    u"""
    General system of first-order PDEs, method of lines, Keller box
    discretization, one space variable.

    ``dim1_parab_keller`` integrates a system of linear or nonlinear,
    first-order, time-dependent partial differential equations (PDEs) in
    one space variable.
    The spatial discretization is performed using the Keller box scheme
    and the method of lines is employed to reduce the PDEs to a system
    of ordinary differential equations (ODEs).
    The resulting system is solved using a Backward Differentiation
    Formula (BDF) method.

    For full information please refer to the NAG Library document for
    d03pe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pef.html

    Parameters
    ----------
    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (res, ires) = pdedef(t, x, u, ut, ux, ires,
        data=None)
        `pdedef` must compute the functions G_i which define the system
        of PDEs. `pdedef` is called approximately midway between each
        pair of mesh points in turn by ``dim1_parab_keller``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,npde.

        ut : float, ndarray, shape (npde)
            `ut`[i-1] contains the value of the component dU_i(x,t)/dt,
            for i = 1,2,...,npde.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,npde.

        ires : int
            The form of G_i that must be returned in the array `res`.

            `ires` = -1

              Equation (8) must be used.

            `ires` = 1

              Equation (9) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (npde)
            `res`[i-1] must contain the ith component of G, for i =
            1,2,...,npde, where G is defined as

                G_i = sum_j = 1^npdeP_{i,j}dU_j/dt,

            i.e., only terms depending explicitly on time derivatives,
            or

                G_i = sum_j = 1^npdeP_{i,j}dU_j/dt+Q_i,

            i.e., all terms in equation (2).

            The definition of G is determined by the input value of
            `ires`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions,
            as described below:

    bndary : callable (res, ires) = bndary(t, ibnd, nobc, u, ut, ires,
        data=None)
        `bndary` must compute the functions G_i^L and G_i^R which define
        the boundary conditions as in equations (4) and (5).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        ibnd : int
            Determines the position of the boundary conditions.

            `ibnd` = 0

              `bndary` must compute the left-hand boundary condition at
              x = a.

            `ibnd` != 0

              Indicates that `bndary` must compute the right-hand
              boundary condition at x = b.

        nobc : int
            Specifies the number of boundary conditions at the boundary
            specified by `ibnd`.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,npde.

        ut : float, ndarray, shape (npde)
            `ut`[i-1] contains the value of the component dU_i(x,t)/dt
            at the boundary specified by `ibnd`, for i = 1,2,...,npde.

        ires : int
            The form G_i^L (or G_i^R) that must be returned in the array
            `res`.

            `ires` = -1

              Equation (0) must be used.

            `ires` = 1

              Equation (1) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (`nobc`)
            `res`[i-1] must contain the ith component of G^L or G^R,
            depending on the value of `ibnd`, for i = 1,2,...,`nobc`,
            where G^L is defined as

                G_i^L = sum_j = 1^npdeE_i,j^LdU_j/dt,

            i.e., only terms depending explicitly on time derivatives,
            or

                G_i^L = sum_j = 1^npdeE_i,j^LdU_j/dt+S_i^L,

            i.e., all terms in equation (6), and similarly for G_i^R.

            The definitions of G^L and G^R are determined by the input
            value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions,
            as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_keller``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    u : float, array-like, shape (npde, npts)
        The initial values of U(x,t) at t = `ts` and the mesh points
        `x`[j-1], for j = 1,2,...,npts.

    x : float, array-like, shape (npts)
        The mesh points in the spatial direction. `x`[0] must specify
        the left-hand boundary, a, and `x`[npts-1] must specify the
        right-hand boundary, b.

    nleft : int
        The number n_a of boundary conditions at the left-hand mesh
        point `x`[0].

    acc : float
        A positive quantity for controlling the local error estimate in
        the time integration. If E(i,j) is the estimated error for U_i
        at the jth mesh point, the error test is:

            |E(i,j)| = `acc`*(1.0+|`u`[i-1,j-1]|).

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    itask : int
        Specifies the task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout`.

        `itask` = 2

          Take one step and return.

        `itask` = 3

          Stop at the first internal integration point at or beyond t =
          `tout`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_keller`` and the underlying ODE solver as follows:

        `itrace` <= -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` = 1

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        `itrace` = 2

          Output from the underlying ODE solver is similar to that
          produced when `itrace` = 1, except that the advisory messages
          are given in greater detail.

        `itrace` >= 3

          Output from the underlying ODE solver is similar to that
          produced when `itrace` = 2, except that the advisory messages
          are given in greater detail.

        You are advised to set `itrace` = 0.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_keller``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (npde, npts)
        `u`[i-1,j-1] will contain the computed solution at t = `ts`.

    ind : int
        `ind` = 1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, `acc` = *<value>*.

            Constraint: `acc` > 0.0.

        (`errno` 1)
            On entry, `nleft` = *<value>*, npde = *<value>*.

            Constraint: `nleft` <= npde.

        (`errno` 1)
            On entry, `nleft` = *<value>*.

            Constraint: `nleft` >= 0.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2 or 3.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `acc` was too small to start integration: `acc` = *<value>*.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef` or
            `bndary`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied value of `acc`. `ts` =
            *<value>*, `acc` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef` or `bndary`. Integration is successful as
            far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but a small change in `acc` is
            unlikely to result in a changed solution. `acc` = *<value>*.
    """

    from ..base.pde import dim1_parab_keller as b_dim1_parab_keller

    fun_name = 'naginterfaces.library.pde.dim1_parab_keller'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ut, ux, res, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _res_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ut,
                ux,
                ires,
                data,
            )
        else:
            _res_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ut,
                ux,
                ires,
            )
        res[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'res'},
            _res_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, ibnd, nobc, u, ut, res, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _res_bndary, _ires_bndary = bndary(
                t,
                ibnd,
                nobc,
                u,
                ut,
                ires,
                data,
            )
        else:
            _res_bndary, _ires_bndary = bndary(
                t,
                ibnd,
                nobc,
                u,
                ut,
                ires,
            )
        res[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'res'},
            _res_bndary,
            copy=False,
            exp_shape=(nobc,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[1]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_dim1_parab_keller(
            _npde,
            ts,
            tout,
            _pdedef,
            _bndary,
            _u,
            _npts,
            _x,
            nleft,
            acc,
            comm,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'ind'],
    )(
        _ts_py,
        _u_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_convdiff(ts, tout, numflx, bndary, u, x, acc, tsmax, comm, itask, itrace, ind, pdedef=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of convection-diffusion PDEs with source terms in
    conservative form, method of lines, upwind scheme using numerical
    flux function based on Riemann solver, one space variable.

    ``dim1_parab_convdiff`` integrates a system of linear or nonlinear
    convection-diffusion equations in one space dimension, with optional
    source terms.
    The system must be posed in conservative form.
    Convection terms are discretized using a sophisticated upwind scheme
    involving a user-supplied numerical flux function based on the
    solution of a Riemann problem at each mesh point.
    The method of lines is employed to reduce the PDEs to a system of
    ordinary differential equations (ODEs), and the resulting system is
    solved using a backward differentiation formula (BDF) method.

    For full information please refer to the NAG Library document for
    d03pf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pff.html

    Parameters
    ----------
    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    numflx : callable (flux, ires) = numflx(t, x, uleft, uright, ires,
        data=None)
        `numflx` must supply the numerical flux for each PDE given the
        left and right values of the solution vector `u`. `numflx` is
        called approximately midway between each pair of mesh points in
        turn by ``dim1_parab_convdiff``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        uleft : float, ndarray, shape (npde)
            `uleft`[i-1] contains the left value of the component
            U_i(x), for i = 1,2,...,npde.

        uright : float, ndarray, shape (npde)
            `uright`[i-1] contains the right value of the component
            U_i(x), for i = 1,2,...,npde.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        flux : float, array-like, shape (npde)
            `flux`[i-1] must be set to the numerical flux F^_i, for i =
            1,2,...,npde.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_convdiff``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    bndary : callable (g, ires) = bndary(t, x, u, ibnd, ires, data=None)
        `bndary` must evaluate the functions G_i^L and G_i^R which
        describe the physical and numerical boundary conditions, as
        given by (7) and (8).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            The mesh points in the spatial direction. `x`[0] corresponds
            to the left-hand boundary, a, and `x`[npts-1] corresponds to
            the right-hand boundary, b.

        u : float, ndarray, shape (npde, 3)
            Contains the value of solution components in the boundary
            region.

            If `ibnd` = 0, `u`[i-1,j-1] contains the value of the
            component U_i(x,t) at x = `x`[j-1], for j = 1,2,...,3, for i
            = 1,2,...,npde.

            If `ibnd` != 0, `u`[i-1,j-1] contains the value of the
            component U_i(x,t) at x = `x`[npts-j], for j = 1,2,...,3,
            for i = 1,2,...,npde.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must evaluate the left-hand boundary condition at
              x = a.

            `ibnd` != 0

              `bndary` must evaluate the right-hand boundary condition
              at x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        g : float, array-like, shape (npde)
            `g`[i-1] must contain the ith component of either `g`^L or
            `g`^R in (7) and (8), depending on the value of `ibnd`, for
            i = 1,2,...,npde.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_convdiff``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    u : float, array-like, shape (npde, npts)
        `u`[i-1,j-1] must contain the initial value of U_i(x,t) at x =
        `x`[j-1] and t = `ts`, for j = 1,2,...,npts, for i =
        1,2,...,npde.

    x : float, array-like, shape (npts)
        The mesh points in the space direction. `x`[0] must specify the
        left-hand boundary, a, and `x`[npts-1] must specify the
        right-hand boundary, b.

    acc : float, array-like, shape (2)
        The components of `acc` contain the relative and absolute error
        tolerances used in the local error test in the time integration.

        If E(i,j) is the estimated error for U_i at the jth mesh point,
        the error test is

            E(i,j) = `acc`[0]*`u`[i-1,j-1]+`acc`[1].

    tsmax : float
        The maximum absolute step size to be allowed in the time
        integration. If `tsmax` = 0.0 then no maximum is imposed.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    itask : int
        The task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step in the time direction and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_convdiff`` and the underlying ODE solver. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` > 0

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases. You are advised to set `itrace` = 0, unless you are
        experienced with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_convdiff``.

    pdedef : None or callable (p, c, d, s, ires) = pdedef(t, x, u, ux,
        ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `pdedef` must evaluate the functions P_{i,j}, C_i, D_i and S_i
        which partially define the system of PDEs. P_{i,j}, C_i and S_i
        may depend on x, t and U; D_i may depend on x, t, U and U_x.
        `pdedef` is called approximately midway between each pair of
        mesh points in turn by ``dim1_parab_convdiff``. None may be used
        for `pdedef` for problems in the form (2).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,npde.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,npde.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde)
            `p`[i-1,j-1] must be set to the value of P_{i,j}(x,t,U), for
            j = 1,2,...,npde, for i = 1,2,...,npde.

        c : float, array-like, shape (npde)
            `c`[i-1] must be set to the value of C_i(x,t,U), for i =
            1,2,...,npde.

        d : float, array-like, shape (npde)
            `d`[i-1] must be set to the value of D_i(x,t,U,U_x), for i =
            1,2,...,npde.

        s : float, array-like, shape (npde)
            `s`[i-1] must be set to the value of S_i(x,t,U), for i =
            1,2,...,npde.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_convdiff``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `u` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (npde, npts)
        `u`[i-1,j-1] will contain the computed solution U_i(x,t) at x =
        `x`[j-1] and t = `ts`, for j = 1,2,...,npts, for i =
        1,2,...,npde.

    ind : int
        `ind` = 1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, `tsmax` = *<value>*.

            Constraint: `tsmax` >= 0.0.

        (`errno` 1)
            On entry, `acc`[0] and `acc`[1] are both zero.

        (`errno` 1)
            On entry, `acc`[1] = *<value>*.

            Constraint: `acc`[1] >= 0.0.

        (`errno` 1)
            On entry, `acc`[0] = *<value>*.

            Constraint: `acc`[0] >= 0.0.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2 or 3.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef`, `numflx`, or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            Values in `acc` are too small to start integration: `acc`[0]
            = *<value>*, `acc`[1] = *<value>*.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef`,
            `numflx`, or `bndary`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 14)
            The functions P, D, or C appear to depend on time
            derivatives.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `acc`. `ts` =
            *<value>*, `acc`[0] = *<value>*, `acc`[1] = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef`, `numflx`, or `bndary`. Integration is
            successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `acc` are
            unlikely to result in a changed solution. `acc`[0] =
            *<value>*, `acc`[1] = *<value>*.
    """

    from ..base.pde import dim1_parab_convdiff as b_dim1_parab_convdiff

    fun_name = 'naginterfaces.library.pde.dim1_parab_convdiff'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ux, p, c, d, s, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _c_pdedef, _d_pdedef, _s_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                ires,
                data,
            )
        else:
            _p_pdedef, _c_pdedef, _d_pdedef, _s_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'c'},
            _c_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        d[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'd'},
            _d_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        s[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 's'},
            _s_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_pdedef_np
        return _ires_pdedef_base

    cb_numflx_name = fun_name + ':numflx'

    def cb_numflx(npde, t, x, uleft, uright, flux, ires, data=None):
        """Wrapping user's numflx. Called as base numflx."""
        if data is not None:
            _flux_numflx, _ires_numflx = numflx(
                t,
                x,
                uleft,
                uright,
                ires,
                data,
            )
        else:
            _flux_numflx, _ires_numflx = numflx(
                t,
                x,
                uleft,
                uright,
                ires,
            )
        flux[:] = _utils._to_ndarray(
            {'fun_name': cb_numflx_name, 'returned': 'flux'},
            _flux_numflx,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_numflx_base = _ires_numflx
        return _ires_numflx_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, npts, t, x, u, ibnd, g, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _g_bndary, _ires_bndary = bndary(
                t,
                x,
                u,
                ibnd,
                ires,
                data,
            )
        else:
            _g_bndary, _ires_bndary = bndary(
                t,
                x,
                u,
                ibnd,
                ires,
            )
        g[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'g'},
            _g_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_u, 'u'),),
        spiked_sorder=spiked_sorder,
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'}, allow_none=True)
    _pdedef = None if pdedef is None else cb_pdedef
    _b_utils._check_callable(numflx, {'fun_name': fun_name, 'entity': 'numflx'})
    _numflx = cb_numflx
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[1]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _acc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'acc'},
        acc,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_dim1_parab_convdiff(
            _npde,
            ts,
            tout,
            _pdedef,
            _numflx,
            _bndary,
            _u,
            _npts,
            _x,
            _acc,
            tsmax,
            comm,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'ind'],
    )(
        _ts_py,
        _u_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_dae_fd(npde, m, ts, tout, pdedef, bndary, u, x, nv, xi, rtol, atol, itol, norm, laopt, algopt, comm, itask, itrace, ind, odedef=None, lrsave=None, lisave=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of parabolic PDEs, coupled DAEs, method of lines,
    finite differences, one space variable.

    ``dim1_parab_dae_fd`` integrates a system of linear or nonlinear
    parabolic partial differential equations (PDEs) in one space
    variable, with scope for coupled ordinary differential equations
    (ODEs).
    The spatial discretization is performed using finite differences,
    and the method of lines is employed to reduce the PDEs to a system
    of ODEs.
    The resulting system is solved using a backward differentiation
    formula method or a Theta method (switching between Newton's method
    and functional iteration).

    For full information please refer to the NAG Library document for
    d03ph

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03phf.html

    Parameters
    ----------
    npde : int
        The number of PDEs to be solved.

    m : int
        The coordinate system used:

        `m` = 0

          Indicates Cartesian coordinates.

        `m` = 1

          Indicates cylindrical polar coordinates.

        `m` = 2

          Indicates spherical polar coordinates.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (p, q, r, ires) = pdedef(t, x, u, ux, v, vdot,
        ires, data=None)
        `pdedef` must evaluate the functions P_{i,j}, Q_i and R_i which
        define the system of PDEs.

        The functions may depend on x, t, U, U_x and V. Q_i may depend
        linearly on V^.. `pdedef` is called approximately midway between
        each pair of mesh points in turn by ``dim1_parab_dae_fd``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in Q_j, for j = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde)
            `p`[i-1,j-1] must be set to the value of
            P_{i,j}(x,t,U,U_x,V), for j = 1,2,...,`npde`, for i =
            1,2,...,`npde`.

        q : float, array-like, shape (npde)
            `q`[i-1] must be set to the value of Q_i(x,t,U,U_x,V,V^.),
            for i = 1,2,...,`npde`.

        r : float, array-like, shape (npde)
            `r`[i-1] must be set to the value of R_i(x,t,U,U_x,V), for i
            = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_dae_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    bndary : callable (beta, gamma, ires) = bndary(t, u, ux, v, vdot,
        ibnd, ires, data=None)
        `bndary` must evaluate the functions beta_i and gamma_i which
        describe the boundary conditions, as given in [equation].

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx
            at the boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in Q_j, for j = 1,2,...,`npde`.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must set up the coefficients of the left-hand
              boundary, x = a.

            `ibnd` != 0

              `bndary` must set up the coefficients of the right-hand
              boundary, x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        beta : float, array-like, shape (npde)
            `beta`[i-1] must be set to the value of beta_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        gamma : float, array-like, shape (npde)
            `gamma`[i-1] must be set to the value of
            gamma_i(x,t,U,U_x,V,V^.) at the boundary specified by
            `ibnd`, for i = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_dae_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    u : float, array-like, shape (neqn)
        The initial values of the dependent variables defined as
        follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t_0), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t_0), for i = 1,2,...,`nv`.

    x : float, array-like, shape (npts)
        The mesh points in the space direction. `x`[0] must specify the
        left-hand boundary, a, and `x`[npts-1] must specify the
        right-hand boundary, b.

    nv : int
        The number of coupled ODE components.

    xi : float, array-like, shape (nxi)
        If nxi > 0, `xi`[i-1], for i = 1,2,...,nxi, must be set to the
        ODE/PDE coupling points.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; if `itol` in (3, 4): neqn;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; if `itol` in (2, 4): neqn;
        otherwise: 0.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``dim1_parab_dae_fd`` whether to interpret either
        or both of `rtol` or `atol` as a vector or scalar. The error
        test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is defined
        as follows:

        [table omitted]

        In the above, e_i denotes the estimated local error for the ith
        component of the coupled PDE/ODE system in time, `u`[i-1], for i
        = 1,2,...,neqn.

        The choice of norm used is defined by the argument `norm`.

    norm : str, length 1
        The type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L_2 norm.

        If `u`_norm denotes the norm of the vector `u` of length neqn,
        then for the averaged L_2 norm

            `u`_norm = sqrt(1/neqnsum_i = 1^neqn(`u`[i-1]/w_i)^2),

        while for the maximum norm

            `u`_norm = max_i(|`u`[i-1]/w_i|).

        See the description of `itol` for the formulation of the weight
        vector w.

    laopt : str, length 1
        The type of matrix algebra required.

        `laopt` = 'F'

          Full matrix methods to be used.

        `laopt` = 'B'

          Banded matrix methods to be used.

        `laopt` = 'S'

          Sparse matrix methods to be used.

    algopt : float, array-like, shape (30)
        May be set to control various options available in the
        integrator. If you wish to employ all the default options,
        `algopt`[0] should be set to 0.0. Default values will also be
        used for any other elements of `algopt` set to zero. The
        permissible values, default values, and meanings are as follows:

        `algopt`[0]

          Selects the ODE integration method to be used. If `algopt`[0]
          = 1.0, a BDF method is used and if `algopt`[0] = 2.0, a Theta
          method is used. The default value is `algopt`[0] = 1.0.

        If `algopt`[0] = 2.0, `algopt`[i-1], for i = 2,3,...,4 are not
        used.

        `algopt`[1]

          Specifies the maximum order of the BDF integration formula to
          be used. `algopt`[1] may be 1.0, 2.0, 3.0, 4.0 or 5.0. The
          default value is `algopt`[1] = 5.0.

        `algopt`[2]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the BDF method. If
          `algopt`[2] = 1.0 a modified Newton iteration is used and if
          `algopt`[2] = 2.0 a functional iteration method is used. If
          functional iteration is selected and the integrator encounters
          difficulty, there is an automatic switch to the modified
          Newton iteration. The default value is `algopt`[2] = 1.0.

        `algopt`[3]

          Specifies whether or not the Petzold error test is to be
          employed. The Petzold error test results in extra overhead but
          is more suitable when algebraic equations are present, such as
          P_{i,j} = 0.0, for j = 1,2,...,`npde`, for some i or when
          there is no V^._i(t) dependence in the coupled ODE system. If
          `algopt`[3] = 1.0, the Petzold test is used. If `algopt`[3] =
          2.0, the Petzold test is not used. The default value is
          `algopt`[3] = 1.0.

        If `algopt`[0] = 1.0, `algopt`[i-1], for i = 5,6,...,7, are not
        used.

        `algopt`[4]

          Specifies the value of Theta to be used in the Theta
          integration method. 0.51 <= `algopt`[4] <= 0.99. The default
          value is `algopt`[4] = 0.55.

        `algopt`[5]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the Theta method.
          If `algopt`[5] = 1.0, a modified Newton iteration is used and
          if `algopt`[5] = 2.0, a functional iteration method is used.
          The default value is `algopt`[5] = 1.0.

        `algopt`[6]

          Specifies whether or not the integrator is allowed to switch
          automatically between modified Newton and functional iteration
          methods in order to be more efficient. If `algopt`[6] = 1.0,
          switching is allowed and if `algopt`[6] = 2.0, switching is
          not allowed. The default value is `algopt`[6] = 1.0.

        `algopt`[10]

          Specifies a point in the time direction, t_crit, beyond which
          integration must not be attempted. The use of t_crit is
          described under the argument `itask`. If `algopt`[0] != 0.0, a
          value of 0.0 for `algopt`[10], say, should be specified even
          if `itask` subsequently specifies that t_crit will not be
          used.

        `algopt`[11]

          Specifies the minimum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[11]
          should be set to 0.0.

        `algopt`[12]

          Specifies the maximum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[12]
          should be set to 0.0.

        `algopt`[13]

          Specifies the initial step size to be attempted by the
          integrator. If `algopt`[13] = 0.0, the initial step size is
          calculated internally.

        `algopt`[14]

          Specifies the maximum number of steps to be attempted by the
          integrator in any one call. If `algopt`[14] = 0.0, no limit is
          imposed.

        `algopt`[22]

          Specifies what method is to be used to solve the nonlinear
          equations at the initial point to initialize the values of U,
          U_t, V and V^.. If `algopt`[22] = 1.0, a modified Newton
          iteration is used and if `algopt`[22] = 2.0, functional
          iteration is used. The default value is `algopt`[22] = 1.0.

        `algopt`[28] and `algopt`[29] are used only for the sparse
        matrix algebra option, `laopt` = 'S'.

        `algopt`[28]

          Governs the choice of pivots during the decomposition of the
          first Jacobian matrix. It should lie in the range 0.0 <
          `algopt`[28] < 1.0, with smaller values biasing the algorithm
          towards maintaining sparsity at the expense of numerical
          stability. If `algopt`[28] lies outside this range then the
          default value is used. If the functions regard the Jacobian
          matrix as numerically singular then increasing `algopt`[28]
          towards 1.0 may help, but at the cost of increased fill-in.
          The default value is `algopt`[28] = 0.1.

        `algopt`[29]

          Is used as a relative pivot threshold during subsequent
          Jacobian decompositions (see `algopt`[28]) below which an
          internal error is invoked. If `algopt`[29] is greater than 1.0
          no check is made on the pivot size, and this may be a
          necessary option if the Jacobian is found to be numerically
          singular (see `algopt`[28]). The default value is `algopt`[29]
          = 0.0001.

    comm : dict, communication object, modified in place
        Note: this argument will be (re-)initialized when it is an empty
        ``dict`` or under the following condition: `ind` = 0.

        Communication structure.

        On initial entry: need not be set.

    itask : int
        Specifies the task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout`.

        `itask` = 2

          One step and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

        `itask` = 4

          Normal computation of output values `u` at t = `tout` but
          without overshooting t = t_crit where t_crit is described
          under the argument `algopt`.

        `itask` = 5

          Take one step in the time direction and return, without
          passing t_crit, where t_crit is described under the argument
          `algopt`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_dae_fd`` and the underlying ODE solver. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` > 0

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases. You are advised to set `itrace` = 0, unless you are
        experienced with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_dae_fd``.

    odedef : None or callable (f, ires) = odedef(t, v, vdot, xi, ucp,
        ucpx, rcp, ucpt, ucptx, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `odedef` must evaluate the functions F, which define the system
        of ODEs, as given in (3).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling points,
            xi_i, for i = 1,2,...,nxi.

        ucp : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucp`[i-1,j-1] contains the value of U_i(x,t) at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpx : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpx`[i-1,j-1] contains the value of
            dU_i(x,t)/dx at the coupling point x = xi_j, for j =
            1,2,...,nxi, for i = 1,2,...,`npde`.

        rcp : float, ndarray, shape (npde, nxi)
            `rcp`[i-1,j-1] contains the value of the flux R_i at the
            coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpt : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpt`[i-1,j-1] contains the value of dU_i/dt at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucptx : float, ndarray, shape (npde, nxi)
            `ucptx`[i-1,j-1] contains the value of d^2U_i/dxdt at the
            coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ires : int
            The form of F that must be returned in the array `f`.

            `ires` = 1

              Equation (5) must be used.

            `ires` = -1

              Equation (6) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (nv)
            On exit, `f`[i-1] must contain the ith component of F as
            specified by the input value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may reset
            `ires` to force the integration function to take certain
            actions as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_dae_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    lrsave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['rsave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lrsave` is computed by the
        function.

        If your supplied `lrsave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lrsave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lrsave` set to at least the lower-bound value returned in
        `lrsave_min`.

    lisave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['isave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lisave` is computed by the
        function.

        If your supplied `lisave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lisave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lisave` set to at least the lower-bound value returned in
        `lisave_min`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `pdedef` is spiked (i.e., has unit extent in all but
        one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (neqn)
        The computed solution U_i(x_j,t), for j = 1,2,...,npts, for i =
        1,2,...,`npde`, and V_k(t), for k = 1,2,...,`nv`, evaluated at t
        = `ts`, as follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t), for i = 1,2,...,`nv`.

    ind : int
        `ind` = 1.

    lrsave_min : int
        A lower bound on the sufficient size for `comm`['rsave'].

    lisave_min : int
        A lower bound on the sufficient size for `comm`['isave'].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `lrsave` = *<value>*.

            Constraint: `lrsave` >= *<value>*.

        (`errno` 1)
            On entry, `lisave` = *<value>*.

            Constraint: `lisave` >= *<value>*.

        (`errno` 1)
            On entry, at least one point in `xi` lies outside
            [`x`[0],`x`[npts-1]]: `x`[0] = *<value>* and `x`[npts-1] =
            *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xi`[i] = *<value>* and `xi`[i-1] =
            *<value>*.

            Constraint: `xi`[i] > `xi`[i-1].

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: corresponding elements `atol`[i-1] and
            `rtol`[j-1] cannot both be 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `rtol`[i-1] = *<value>*.

            Constraint: `rtol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `atol`[i-1] = *<value>*.

            Constraint: `atol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, neqn = *<value>*, `npde` = *<value>*, npts =
            *<value>* and `nv` = *<value>*.

            Constraint: neqn = `npde`*npts+`nv`.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi = 0 when `nv` = 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi >= 0 when `nv` > 0.

        (`errno` 1)
            On entry, `npde` = *<value>*.

            Constraint: `npde` >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `m` = *<value>* and `x`[0] = *<value>*.

            Constraint: `m` <= 0 or `x`[0] >= 0.0

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` = 0, 1 or 2.

        (`errno` 1)
            On entry, `laopt` = *<value>*.

            Constraint: `laopt` = 'F', 'B' or 'S'.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 'A' or 'M'.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2, 3, 4 or 5.

        (`errno` 1)
            On entry, `nv` = *<value>*.

            Constraint: `nv` >= 0.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `atol` and `rtol` were too small to start integration.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef`,
            `bndary`, or `odedef`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 12)
            In solving ODE system, the maximum number of steps
            `algopt`[14] has been exceeded. `algopt`[14] = *<value>*.

        (`errno` 14)
            Flux function appears to depend on time derivatives.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `atol` and `rtol`.
            `ts` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef`, `bndary`, or `odedef`. Integration is
            successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `atol` or `rtol`
            are unlikely to result in a changed solution.

        (`errno` 13)
            Zero error weights encountered during time integration.

        (`errno` 15)
            When using the sparse option `lisave` or `lrsave` is too
            small: `lisave` = *<value>*, `lrsave` = *<value>*.
    """

    from ..base.pde import dim1_parab_dae_fd as b_dim1_parab_dae_fd

    fun_name = 'naginterfaces.library.pde.dim1_parab_dae_fd'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ux, nv, v, vdot, p, q, r, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        q[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'q'},
            _q_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'r'},
            _r_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_pdedef_np
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, u, ux, nv, v, vdot, ibnd, beta, gamma, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                v,
                vdot,
                ibnd,
                ires,
                data,
            )
        else:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                v,
                vdot,
                ibnd,
                ires,
            )
        beta[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'beta'},
            _beta_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        gamma[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'gamma'},
            _gamma_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_odedef_name = fun_name + ':odedef'

    def cb_odedef(npde, t, nv, v, vdot, nxi, xi, ucp, ucpx, rcp, ucpt, ucptx, f, ires, data=None):
        """Wrapping user's odedef. Called as base odedef."""
        if data is not None:
            _f_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                rcp,
                ucpt,
                ucptx,
                ires,
                data,
            )
        else:
            _f_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                rcp,
                ucpt,
                ucptx,
                ires,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_odedef_name, 'returned': 'f'},
            _f_odedef,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_odedef_base = _ires_odedef
        return _ires_odedef_base

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqn_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _neqn = _b_utils._EngineIntScalarType(
        _neqn_py, {'fun_name': fun_name, 'entity': '_neqn'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _b_utils._check_callable(odedef, {'fun_name': fun_name, 'entity': 'odedef'}, allow_none=True)
    _odedef = None if odedef is None else cb_odedef
    _xi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xi'},
        xi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxi_py = _b_utils._EngineFloat64ArrayType.get_shape(_xi, {'fun_name': fun_name, 'entity': 'xi'}, exp_rank=1)[0]
    _nxi = _b_utils._EngineIntScalarType(
        _nxi_py, {'fun_name': fun_name, 'entity': '_nxi'},
    ).to_ctypes()
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _algopt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'algopt'},
        algopt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_dim1_parab_dae_fd(
            npde,
            m,
            ts,
            tout,
            _pdedef,
            _bndary,
            _u,
            _npts,
            _x,
            nv,
            _odedef,
            _nxi,
            _xi,
            _neqn,
            _rtol,
            _atol,
            itol,
            norm,
            laopt,
            _algopt,
            comm,
            lrsave,
            lisave,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _lrsave_min_py = _return_scalars.lrsave_min
    _lisave_min_py = _return_scalars.lisave_min
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'ind', 'lrsave_min', 'lisave_min'],
    )(
        _ts_py,
        _u_py,
        _ind_py,
        _lrsave_min_py,
        _lisave_min_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_dae_coll(npde, m, ts, tout, pdedef, bndary, u, xbkpts, npoly, npts, nv, xi, uvinit, rtol, atol, itol, norm, laopt, algopt, comm, itask, itrace, ind, odedef=None, lrsave=None, lisave=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of parabolic PDEs, coupled DAEs, method of lines,
    Chebyshev C^0 collocation, one space variable.

    ``dim1_parab_dae_coll`` integrates a system of linear or nonlinear
    parabolic partial differential equations (PDEs), in one space
    variable with scope for coupled ordinary differential equations
    (ODEs).
    The spatial discretization is performed using a Chebyshev C^0
    collocation method, and the method of lines is employed to reduce
    the PDEs to a system of ODEs.
    The resulting system is solved using a backward differentiation
    formula (BDF) method or a Theta method (switching between Newton's
    method and functional iteration).

    For full information please refer to the NAG Library document for
    d03pj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pjf.html

    Parameters
    ----------
    npde : int
        The number of PDEs to be solved.

    m : int
        The coordinate system used:

        `m` = 0

          Indicates Cartesian coordinates.

        `m` = 1

          Indicates cylindrical polar coordinates.

        `m` = 2

          Indicates spherical polar coordinates.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (p, q, r, ires) = pdedef(t, x, u, ux, v, vdot,
        ires, data=None)
        `pdedef` must compute the functions P_{i,j}, Q_i and R_i which
        define the system of PDEs.

        The functions may depend on x, t, U, U_x and V; Q_i may depend
        linearly on V^..

        The functions must be evaluated at a set of points.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (nptl)
            Contains a set of mesh points at which P_{i,j}, Q_i and R_i
            are to be evaluated. `x`[0] and `x`[nptl-1] contain
            successive user-supplied break-points and the elements of
            the array will satisfy `x`[0] < `x`[1] < ... < `x`[nptl-1].

        u : float, ndarray, shape (npde, nptl)
            `u`[i-1,j-1] contains the value of the component U_i(x,t)
            where x = `x`[j-1], for j = 1,2,...,nptl, for i =
            1,2,...,`npde`.

        ux : float, ndarray, shape (npde, nptl)
            `ux`[i-1,j-1] contains the value of the component
            dU_i(x,t)/dx where x = `x`[j-1], for j = 1,2,...,nptl, for i
            = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in Q_j, for j = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde, nptl)
            `p`[ `npde`*`npde`*(k-1)+ `npde`*(j-1)+ (i-1)] must be set
            to the value of P_{i,j}(x,t,U,U_x,V) where x = `x`[k-1], for
            k = 1,2,...,nptl, for j = 1,2,...,`npde`, for i =
            1,2,...,`npde`.

        q : float, array-like, shape (npde, nptl)
            `q`[i-1,j-1] must be set to the value of
            Q_i(x,t,U,U_x,V,V^.) where x = `x`[j-1], for j =
            1,2,...,nptl, for i = 1,2,...,`npde`.

        r : float, array-like, shape (npde, nptl)
            `r`[i-1,j-1] must be set to the value of R_i(x,t,U,U_x,V)
            where x = `x`[i-1], for j = 1,2,...,nptl, for i =
            1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_dae_coll``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    bndary : callable (beta, gamma, ires) = bndary(t, u, ux, v, vdot,
        ibnd, ires, data=None)
        `bndary` must compute the functions beta_i and gamma_i which
        define the boundary conditions as in equation (4).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx
            at the boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in Q_j, for j = 1,2,...,`npde`.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must set up the coefficients of the left-hand
              boundary, x = a.

            `ibnd` != 0

              `bndary` must set up the coefficients of the right-hand
              boundary, x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        beta : float, array-like, shape (npde)
            `beta`[i-1] must be set to the value of beta_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        gamma : float, array-like, shape (npde)
            `gamma`[i-1] must be set to the value of
            gamma_i(x,t,U,U_x,V,V^.) at the boundary specified by
            `ibnd`, for i = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_dae_coll``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    u : float, array-like, shape (neqn)
        If `ind` = 1 the value of `u` must be unchanged from the
        previous call.

    xbkpts : float, array-like, shape (nbkpts)
        The values of the break-points in the space direction.
        `xbkpts`[0] must specify the left-hand boundary, a, and
        `xbkpts`[nbkpts-1] must specify the right-hand boundary, b.

    npoly : int
        The degree of the Chebyshev polynomial to be used in
        approximating the PDE solution between each pair of
        break-points.

    npts : int
        The number of mesh points in the interval [a,b].

    nv : int
        The number of coupled ODE components.

    xi : float, array-like, shape (nxi)
        `xi`[i-1], for i = 1,2,...,nxi, must be set to the ODE/PDE
        coupling points.

    uvinit : callable (u, v) = uvinit(npde, x, nv, data=None)
        `uvinit` must compute the initial values of the PDE and the ODE
        components U_i(x_j,t_0), for j = 1,2,...,`npts`, for i =
        1,2,...,`npde`, and V_k(t_0), for k = 1,2,...,`nv`.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        x : float, ndarray, shape (npts)
            `x`[i-1], for i = 1,2,...,`npts`, contains the current
            values of the space variable x_i.

        nv : int
            The number of coupled ODEs in the system.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float, array-like, shape (`npde`, npts)
            `u`[i-1,j-1] contains the value of the component
            U_i(x_j,t_0), for j = 1,2,...,`npts`, for i =
            1,2,...,`npde`.

        v : float, array-like, shape (`nv`)
            `v`[i-1] contains the value of component V_i(t_0), for i =
            1,2,...,`nv`.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; if `itol` in (3, 4): neqn;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; if `itol` in (2, 4): neqn;
        otherwise: 0.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``dim1_parab_dae_coll`` whether to interpret either
        or both of `rtol` or `atol` as a vector or scalar. The error
        test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is defined
        as follows:

        [table omitted]

        In the above, e_i denotes the estimated local error for the ith
        component of the coupled PDE/ODE system in time, `u`[i-1], for i
        = 1,2,...,neqn.

        The choice of norm used is defined by the argument `norm`.

    norm : str, length 1
        The type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L_2 norm.

        If `u`_norm denotes the norm of the vector `u` of length neqn,
        then for the averaged L_2 norm

            `u`_norm = sqrt(1/neqnsum_i = 1^neqn(`u`[i-1]/w_i)^2),

        while for the maximum norm

            `u`_norm = max_i(|`u`[i-1]/w_i|).

        See the description of `itol` for the formulation of the weight
        vector w.

    laopt : str, length 1
        The type of matrix algebra required.

        `laopt` = 'F'

          Full matrix methods to be used.

        `laopt` = 'B'

          Banded matrix methods to be used.

        `laopt` = 'S'

          Sparse matrix methods to be used.

    algopt : float, array-like, shape (30)
        May be set to control various options available in the
        integrator. If you wish to employ all the default options,
        `algopt`[0] should be set to 0.0. Default values will also be
        used for any other elements of `algopt` set to zero. The
        permissible values, default values, and meanings are as follows:

        `algopt`[0]

          Selects the ODE integration method to be used. If `algopt`[0]
          = 1.0, a BDF method is used and if `algopt`[0] = 2.0, a Theta
          method is used. The default value is `algopt`[0] = 1.0.

        If `algopt`[0] = 2.0, `algopt`[i-1], for i = 2,3,...,4 are not
        used.

        `algopt`[1]

          Specifies the maximum order of the BDF integration formula to
          be used. `algopt`[1] may be 1.0, 2.0, 3.0, 4.0 or 5.0. The
          default value is `algopt`[1] = 5.0.

        `algopt`[2]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the BDF method. If
          `algopt`[2] = 1.0 a modified Newton iteration is used and if
          `algopt`[2] = 2.0 a functional iteration method is used. If
          functional iteration is selected and the integrator encounters
          difficulty, there is an automatic switch to the modified
          Newton iteration. The default value is `algopt`[2] = 1.0.

        `algopt`[3]

          Specifies whether or not the Petzold error test is to be
          employed. The Petzold error test results in extra overhead but
          is more suitable when algebraic equations are present, such as
          P_{i,j} = 0.0, for j = 1,2,...,`npde`, for some i or when
          there is no V^._i(t) dependence in the coupled ODE system. If
          `algopt`[3] = 1.0, the Petzold test is used. If `algopt`[3] =
          2.0, the Petzold test is not used. The default value is
          `algopt`[3] = 1.0.

        If `algopt`[0] = 1.0, `algopt`[i-1], for i = 5,6,...,7, are not
        used.

        `algopt`[4]

          Specifies the value of Theta to be used in the Theta
          integration method. 0.51 <= `algopt`[4] <= 0.99. The default
          value is `algopt`[4] = 0.55.

        `algopt`[5]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the Theta method.
          If `algopt`[5] = 1.0, a modified Newton iteration is used and
          if `algopt`[5] = 2.0, a functional iteration method is used.
          The default value is `algopt`[5] = 1.0.

        `algopt`[6]

          Specifies whether or not the integrator is allowed to switch
          automatically between modified Newton and functional iteration
          methods in order to be more efficient. If `algopt`[6] = 1.0,
          switching is allowed and if `algopt`[6] = 2.0, switching is
          not allowed. The default value is `algopt`[6] = 1.0.

        `algopt`[10]

          Specifies a point in the time direction, t_crit, beyond which
          integration must not be attempted. The use of t_crit is
          described under the argument `itask`. If `algopt`[0] != 0.0, a
          value of 0.0 for `algopt`[10], say, should be specified even
          if `itask` subsequently specifies that t_crit will not be
          used.

        `algopt`[11]

          Specifies the minimum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[11]
          should be set to 0.0.

        `algopt`[12]

          Specifies the maximum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[12]
          should be set to 0.0.

        `algopt`[13]

          Specifies the initial step size to be attempted by the
          integrator. If `algopt`[13] = 0.0, the initial step size is
          calculated internally.

        `algopt`[14]

          Specifies the maximum number of steps to be attempted by the
          integrator in any one call. If `algopt`[14] = 0.0, no limit is
          imposed.

        `algopt`[22]

          Specifies what method is to be used to solve the nonlinear
          equations at the initial point to initialize the values of U,
          U_t, V and V^.. If `algopt`[22] = 1.0, a modified Newton
          iteration is used and if `algopt`[22] = 2.0, functional
          iteration is used. The default value is `algopt`[22] = 1.0.

        `algopt`[28] and `algopt`[29] are used only for the sparse
        matrix algebra option, `laopt` = 'S'.

        `algopt`[28]

          Governs the choice of pivots during the decomposition of the
          first Jacobian matrix. It should lie in the range 0.0 <
          `algopt`[28] < 1.0, with smaller values biasing the algorithm
          towards maintaining sparsity at the expense of numerical
          stability. If `algopt`[28] lies outside this range then the
          default value is used. If the functions regard the Jacobian
          matrix as numerically singular then increasing `algopt`[28]
          towards 1.0 may help, but at the cost of increased fill-in.
          The default value is `algopt`[28] = 0.1.

        `algopt`[29]

          Is used as a relative pivot threshold during subsequent
          Jacobian decompositions (see `algopt`[28]) below which an
          internal error is invoked. If `algopt`[29] is greater than 1.0
          no check is made on the pivot size, and this may be a
          necessary option if the Jacobian is found to be numerically
          singular (see `algopt`[28]). The default value is `algopt`[29]
          = 0.0001.

    comm : dict, communication object, modified in place
        Note: this argument will be (re-)initialized when it is an empty
        ``dict`` or under the following condition: `ind` = 0.

        Communication structure.

        On initial entry: need not be set.

    itask : int
        Specifies the task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout`.

        `itask` = 2

          One step and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

        `itask` = 4

          Normal computation of output values `u` at t = `tout` but
          without overshooting t = t_crit where t_crit is described
          under the argument `algopt`.

        `itask` = 5

          Take one step in the time direction and return, without
          passing t_crit, where t_crit is described under the argument
          `algopt`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_dae_coll`` and the underlying ODE solver. `itrace`
        may take the value -1, 0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` > 0

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases. You are advised to set `itrace` = 0, unless you are
        experienced with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_dae_coll``.

    odedef : None or callable (f, ires) = odedef(t, v, vdot, xi, ucp,
        ucpx, rcp, ucpt, ucptx, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `odedef` must evaluate the functions F, which define the system
        of ODEs, as given in (3).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling points,
            xi_i, for i = 1,2,...,nxi.

        ucp : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucp`[i-1,j-1] contains the value of U_i(x,t) at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpx : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpx`[i-1,j-1] contains the value of
            dU_i(x,t)/dx at the coupling point x = xi_j, for j =
            1,2,...,nxi, for i = 1,2,...,`npde`.

        rcp : float, ndarray, shape (npde, nxi)
            `rcp`[i-1,j-1] contains the value of the flux R_i at the
            coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpt : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpt`[i-1,j-1] contains the value of dU_i/dt at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucptx : float, ndarray, shape (npde, nxi)
            `ucptx`[i-1,j-1] contains the value of d^2U_i/dxdt at the
            coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ires : int
            The form of F that must be returned in the array `f`.

            `ires` = 1

              Equation (5) must be used.

            `ires` = -1

              Equation (6) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (nv)
            On exit, `f`[i-1] must contain the ith component of F as
            specified by the input value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may reset
            `ires` to force the integration function to take certain
            actions as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_dae_coll``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    lrsave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['rsave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lrsave` is computed by the
        function.

        If your supplied `lrsave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lrsave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lrsave` set to at least the lower-bound value returned in
        `lrsave_min`.

    lisave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['isave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lisave` is computed by the
        function.

        If your supplied `lisave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lisave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lisave` set to at least the lower-bound value returned in
        `lisave_min`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `pdedef` and `r` in `pdedef` or `u` in `uvinit` are
        spiked (i.e., have unit extent in all but one dimension, or have
        size 1), `spiked_sorder` selects the storage order to associate
        with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (neqn)
        The computed solution U_i(x_j,t), for j = 1,2,...,`npts`, for i
        = 1,2,...,`npde`, and V_k(t), for k = 1,2,...,`nv`, evaluated at
        t = `ts`, as follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t), for j =
          1,2,...,`npts`, for i = 1,2,...,`npde`, and

        - `u`[`npts`*`npde`+i-1] contain V_i(t), for i = 1,2,...,`nv`.

    x : float, ndarray, shape (`npts`)
        The mesh points chosen by ``dim1_parab_dae_coll`` in the spatial
        direction. The values of `x` will satisfy `x`[0] < `x`[1] < ...
        < `x`[`npts`-1].

    ind : int
        `ind` = 1.

    lrsave_min : int
        A lower bound on the sufficient size for `comm`['rsave'].

    lisave_min : int
        A lower bound on the sufficient size for `comm`['isave'].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `lrsave` = *<value>*.

            Constraint: `lrsave` >= *<value>*.

        (`errno` 1)
            On entry, `lisave` = *<value>*.

            Constraint: `lisave` >= *<value>*.

        (`errno` 1)
            On entry, at least one point in `xi` lies outside
            [`xbkpts`[0],`xbkpts`[nbkpts-1]]: `xbkpts`[0] = *<value>*
            and `xbkpts`[nbkpts-1] = *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xi`[i] = *<value>* and `xi`[i-1] =
            *<value>*.

            Constraint: `xi`[i] > `xi`[i-1].

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: corresponding elements `atol`[i-1] and
            `rtol`[j-1] cannot both be 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `rtol`[i-1] = *<value>*.

            Constraint: `rtol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `atol`[i-1] = *<value>*.

            Constraint: `atol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, i = *<value>*, `xbkpts`[i-1] = *<value>*, j =
            *<value>* and `xbkpts`[j-1] = *<value>*.

            Constraint: `xbkpts`[0] < `xbkpts`[1] < ... <
            `xbkpts`[nbkpts-1].

        (`errno` 1)
            On entry, neqn = *<value>*, `npde` = *<value>*, `npts` =
            *<value>* and `nv` = *<value>*.

            Constraint: neqn = `npde`*`npts`+`nv`.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi = 0 when `nv` = 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi >= 0 when `nv` > 0.

        (`errno` 1)
            On entry, `npts` = *<value>*, nbkpts = *<value>* and `npoly`
            = *<value>*.

            Constraint: `npts` = (nbkpts-1)*`npoly`+1.

        (`errno` 1)
            On entry, `npde` = *<value>*.

            Constraint: `npde` >= 1.

        (`errno` 1)
            On entry, nbkpts = *<value>*.

            Constraint: nbkpts >= 2.

        (`errno` 1)
            On entry, `m` = *<value>* and `xbkpts`[0] = *<value>*.

            Constraint: `m` <= 0 or `xbkpts`[0] >= 0.0

        (`errno` 1)
            On entry, `npoly` = *<value>*.

            Constraint: `npoly` <= 49.

        (`errno` 1)
            On entry, `npoly` = *<value>*.

            Constraint: `npoly` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` = 0, 1 or 2.

        (`errno` 1)
            On entry, `laopt` = *<value>*.

            Constraint: `laopt` = 'F', 'B' or 'S'.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 'A' or 'M'.

        (`errno` 1)
            On entry, `algopt`[0] = *<value>*.

            Constraint: `algopt`[0] = 0.0, 1.0 or 2.0.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2, 3, 4 or 5.

        (`errno` 1)
            On entry, `nv` = *<value>*.

            Constraint: `nv` >= 0.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `atol` and `rtol` were too small to start integration.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef`,
            `bndary`, or `odedef`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 12)
            In solving ODE system, the maximum number of steps
            `algopt`[14] has been exceeded. `algopt`[14] = *<value>*.

        (`errno` 14)
            Flux function appears to depend on time derivatives.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `atol` and `rtol`.
            `ts` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef`, `bndary`, or `odedef`. Integration is
            successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `atol` or `rtol`
            are unlikely to result in a changed solution.

        (`errno` 13)
            Zero error weights encountered during time integration.

        (`errno` 15)
            When using the sparse option `lisave` or `lrsave` is too
            small: `lisave` = *<value>*, `lrsave` = *<value>*.
    """

    from ..base.pde import dim1_parab_dae_coll as b_dim1_parab_dae_coll

    fun_name = 'naginterfaces.library.pde.dim1_parab_dae_coll'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, nptl, u, ux, nv, v, vdot, p, q, r, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde, nptl),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _q_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'q'},
            _q_pdedef,
            copy=False,
            exp_shape=(npde, nptl),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _r_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'r'},
            _r_pdedef,
            copy=False,
            exp_shape=(npde, nptl),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'), (_q_pdedef_np, 'q'), (_r_pdedef_np, 'r')),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :, :] = _p_pdedef_np
        q[:, :] = _q_pdedef_np
        r[:, :] = _r_pdedef_np
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, u, ux, nv, v, vdot, ibnd, beta, gamma, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                v,
                vdot,
                ibnd,
                ires,
                data,
            )
        else:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                v,
                vdot,
                ibnd,
                ires,
            )
        beta[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'beta'},
            _beta_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        gamma[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'gamma'},
            _gamma_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_odedef_name = fun_name + ':odedef'

    def cb_odedef(npde, t, nv, v, vdot, nxi, xi, ucp, ucpx, rcp, ucpt, ucptx, f, ires, data=None):
        """Wrapping user's odedef. Called as base odedef."""
        if data is not None:
            _f_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                rcp,
                ucpt,
                ucptx,
                ires,
                data,
            )
        else:
            _f_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                rcp,
                ucpt,
                ucptx,
                ires,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_odedef_name, 'returned': 'f'},
            _f_odedef,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_odedef_base = _ires_odedef
        return _ires_odedef_base

    cb_uvinit_name = fun_name + ':uvinit'

    def cb_uvinit(npde, npts, x, u, nv, v, data=None):
        """Wrapping user's uvinit. Called as base uvinit."""
        if data is not None:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                nv,
                data,
            )
        else:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                nv,
            )
        _u_uvinit_np = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'u'},
            _u_uvinit,
            copy=False,
            exp_shape=(npde, npts),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        v[:] = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'v'},
            _v_uvinit,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_uvinit_name},
            ((_u_uvinit_np, 'u'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        u[:, :] = _u_uvinit_np

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqn_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _neqn = _b_utils._EngineIntScalarType(
        _neqn_py, {'fun_name': fun_name, 'entity': '_neqn'},
    ).to_ctypes()
    _xbkpts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xbkpts'},
        xbkpts,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nbkpts_py = _b_utils._EngineFloat64ArrayType.get_shape(_xbkpts, {'fun_name': fun_name, 'entity': 'xbkpts'}, exp_rank=1)[0]
    _nbkpts = _b_utils._EngineIntScalarType(
        _nbkpts_py, {'fun_name': fun_name, 'entity': '_nbkpts'},
    ).to_ctypes()
    _npts = _b_utils._EngineIntScalarType(
        npts, {'fun_name': fun_name, 'entity': 'npts'},
    ).to_ctypes()
    _npts_py = _b_utils._EngineIntScalarType.to_py(_npts)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npts_py,),
    )
    _b_utils._check_callable(odedef, {'fun_name': fun_name, 'entity': 'odedef'}, allow_none=True)
    _odedef = None if odedef is None else cb_odedef
    _xi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xi'},
        xi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxi_py = _b_utils._EngineFloat64ArrayType.get_shape(_xi, {'fun_name': fun_name, 'entity': 'xi'}, exp_rank=1)[0]
    _nxi = _b_utils._EngineIntScalarType(
        _nxi_py, {'fun_name': fun_name, 'entity': '_nxi'},
    ).to_ctypes()
    _b_utils._check_callable(uvinit, {'fun_name': fun_name, 'entity': 'uvinit'})
    _uvinit = cb_uvinit
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _algopt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'algopt'},
        algopt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_dim1_parab_dae_coll(
            npde,
            m,
            ts,
            tout,
            _pdedef,
            _bndary,
            _u,
            _nbkpts,
            _xbkpts,
            npoly,
            _npts,
            _x,
            nv,
            _odedef,
            _nxi,
            _xi,
            _neqn,
            _uvinit,
            _rtol,
            _atol,
            itol,
            norm,
            laopt,
            _algopt,
            comm,
            lrsave,
            lisave,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _x_py = _x
    _lrsave_min_py = _return_scalars.lrsave_min
    _lisave_min_py = _return_scalars.lisave_min
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'x', 'ind', 'lrsave_min', 'lisave_min'],
    )(
        _ts_py,
        _u_py,
        _x_py,
        _ind_py,
        _lrsave_min_py,
        _lisave_min_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_dae_keller(npde, ts, tout, pdedef, bndary, u, x, nleft, nv, xi, rtol, atol, itol, norm, laopt, algopt, comm, itask, itrace, ind, odedef=None, lrsave=None, lisave=None, data=None, io_manager=None):
    u"""
    General system of first-order PDEs, coupled DAEs, method of lines,
    Keller box discretization, one space variable.

    ``dim1_parab_dae_keller`` integrates a system of linear or
    nonlinear, first-order, time-dependent partial differential
    equations (PDEs) in one space variable, with scope for coupled
    ordinary differential equations (ODEs).
    The spatial discretization is performed using the Keller box scheme
    and the method of lines is employed to reduce the PDEs to a system
    of ODEs.
    The resulting system is solved using a Backward Differentiation
    Formula (BDF) method or a Theta method (switching between Newton's
    method and functional iteration).

    For full information please refer to the NAG Library document for
    d03pk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pkf.html

    Parameters
    ----------
    npde : int
        The number of PDEs to be solved.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (res, ires) = pdedef(t, x, u, ut, ux, v, vdot,
        ires, data=None)
        `pdedef` must evaluate the functions G_i which define the system
        of PDEs. `pdedef` is called approximately midway between each
        pair of mesh points in turn by ``dim1_parab_dae_keller``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,`npde`.

        ut : float, ndarray, shape (npde)
            `ut`[i-1] contains the value of the component dU_i(x,t)/dt,
            for i = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        ires : int
            The form of G_i that must be returned in the array `res`.

            `ires` = -1

              Equation (9) must be used.

            `ires` = 1

              Equation (0) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (npde)
            `res`[i-1] must contain the ith component of G, for i =
            1,2,...,`npde`, where G is defined as

                G_i = sum_j = 1^`npde`P_{i,j}dU_j/dt+sum_j =
                1^`nv`Q_{i,j}V^._j,

            i.e., only terms depending explicitly on time derivatives,
            or

                G_i = sum_j = 1^`npde`P_{i,j}dU_j/dt+sum_j =
                1^`nv`Q_{i,j}V^._j+S_i,

            i.e., all terms in equation (3).

            The definition of G is determined by the input value of
            `ires`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions,
            as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_dae_keller`` returns to the calling function
              with the error indicator set to `errno` = 4.

    bndary : callable (res, ires) = bndary(t, ibnd, nobc, u, ut, v,
        vdot, ires, data=None)
        `bndary` must evaluate the functions G_i^L and G_i^R which
        describe the boundary conditions, as given in (5) and (6).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must compute the left-hand boundary condition at
              x = a.

            `ibnd` != 0

              `bndary` must compute the right-hand boundary condition at
              x = b.

        nobc : int
            Specifies the number of boundary conditions at the boundary
            specified by `ibnd`.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ut : float, ndarray, shape (npde)
            `ut`[i-1] contains the value of the component dU_i(x,t)/dt
            at the boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: `vdot`[i-1], for i = 1,2,...,`nv`, may only appear
            linearly as in (7) and (8).

        ires : int
            The form of G_i^L (or G_i^R) that must be returned in the
            array `res`.

            `ires` = -1

              Equation (1) must be used.

            `ires` = 1

              Equation (2) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (`nobc`)
            `res`[i-1] must contain the ith component of G^L or G^R,
            depending on the value of `ibnd`, for i = 1,2,...,`nobc`,
            where G^L is defined as

                G_i^L = sum_j = 1^`npde`E_i,j^LdU_j/dt+sum_j =
                1^`nv`H_i,j^LV^._j,

            i.e., only terms depending explicitly on time derivatives,
            or

                G_i^L = sum_j = 1^`npde`E_i,j^LdU_j/dt+sum_j =
                1^`nv`H_i,j^LV^._j+K_i^L,

            i.e., all terms in equation (7), and similarly for G_i^R.

            The definitions of G^L and G^R are determined by the input
            value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_dae_keller`` returns to the calling function
              with the error indicator set to `errno` = 4.

    u : float, array-like, shape (neqn)
        The initial values of the dependent variables defined as
        follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t_0), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t_0), for i = 1,2,...,`nv`.

    x : float, array-like, shape (npts)
        The mesh points in the space direction. `x`[0] must specify the
        left-hand boundary, a, and `x`[npts-1] must specify the
        right-hand boundary, b.

    nleft : int
        The number n_a of boundary conditions at the left-hand mesh
        point `x`[0].

    nv : int
        The number of coupled ODE components.

    xi : float, array-like, shape (nxi)
        `xi`[i-1], for i = 1,2,...,nxi, must be set to the ODE/PDE
        coupling points, xi_i.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; if `itol` in (3, 4): neqn;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; if `itol` in (2, 4): neqn;
        otherwise: 0.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``dim1_parab_dae_keller`` whether to interpret
        either or both of `rtol` or `atol` as a vector or scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        In the above, e_i denotes the estimated local error for the ith
        component of the coupled PDE/ODE system in time, `u`[i-1], for i
        = 1,2,...,neqn.

        The choice of norm used is defined by the argument `norm`.

    norm : str, length 1
        The type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L_2 norm.

        If `u`_norm denotes the norm of the vector `u` of length neqn,
        then for the averaged L_2 norm

            `u`_norm = sqrt(1/neqnsum_i = 1^neqn(`u`[i-1]/w_i)^2),

        while for the maximum norm

            `u`_norm = max_i(|`u`[i-1]/w_i|).

        See the description of `itol` for the formulation of the weight
        vector w.

    laopt : str, length 1
        The type of matrix algebra required.

        `laopt` = 'F'

          Full matrix methods to be used.

        `laopt` = 'B'

          Banded matrix methods to be used.

        `laopt` = 'S'

          Sparse matrix methods to be used.

    algopt : float, array-like, shape (30)
        May be set to control various options available in the
        integrator. If you wish to employ all the default options,
        `algopt`[0] should be set to 0.0. Default values will also be
        used for any other elements of `algopt` set to zero. The
        permissible values, default values, and meanings are as follows:

        `algopt`[0]

          Selects the ODE integration method to be used. If `algopt`[0]
          = 1.0, a BDF method is used and if `algopt`[0] = 2.0, a Theta
          method is used. The default value is `algopt`[0] = 1.0.

        If `algopt`[0] = 2.0, then `algopt`[i-1], for i = 2,3,...,4, are
        not used.

        `algopt`[1]

          Specifies the maximum order of the BDF integration formula to
          be used. `algopt`[1] may be 1.0, 2.0, 3.0, 4.0 or 5.0. The
          default value is `algopt`[1] = 5.0.

        `algopt`[2]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the BDF method. If
          `algopt`[2] = 1.0 a modified Newton iteration is used and if
          `algopt`[2] = 2.0 a functional iteration method is used. If
          functional iteration is selected and the integrator encounters
          difficulty, there is an automatic switch to the modified
          Newton iteration. The default value is `algopt`[2] = 1.0.

        `algopt`[3]

          Specifies whether or not the Petzold error test is to be
          employed. The Petzold error test results in extra overhead but
          is more suitable when algebraic equations are present, such as
          P_{i,j} = 0.0, for j = 1,2,...,`npde`, for some i or when
          there is no V^._i(t) dependence in the coupled ODE system. If
          `algopt`[3] = 1.0, the Petzold test is used. If `algopt`[3] =
          2.0, the Petzold test is not used. The default value is
          `algopt`[3] = 1.0.

        If `algopt`[0] = 1.0, `algopt`[i-1], for i = 5,6,...,7, are not
        used.

        `algopt`[4]

          Specifies the value of Theta to be used in the Theta
          integration method. 0.51 <= `algopt`[4] <= 0.99. The default
          value is `algopt`[4] = 0.55.

        `algopt`[5]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the Theta method.
          If `algopt`[5] = 1.0, a modified Newton iteration is used and
          if `algopt`[5] = 2.0, a functional iteration method is used.
          The default value is `algopt`[5] = 1.0.

        `algopt`[6]

          Specifies whether or not the integrator is allowed to switch
          automatically between modified Newton and functional iteration
          methods in order to be more efficient. If `algopt`[6] = 1.0,
          switching is allowed and if `algopt`[6] = 2.0, switching is
          not allowed. The default value is `algopt`[6] = 1.0.

        `algopt`[10]

          Specifies a point in the time direction, t_crit, beyond which
          integration must not be attempted. The use of t_crit is
          described under the argument `itask`. If `algopt`[0] != 0.0, a
          value of 0.0, for `algopt`[10], say, should be specified even
          if `itask` subsequently specifies that t_crit will not be
          used.

        `algopt`[11]

          Specifies the minimum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[11]
          should be set to 0.0.

        `algopt`[12]

          Specifies the maximum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[12]
          should be set to 0.0.

        `algopt`[13]

          Specifies the initial step size to be attempted by the
          integrator. If `algopt`[13] = 0.0, the initial step size is
          calculated internally.

        `algopt`[14]

          Specifies the maximum number of steps to be attempted by the
          integrator in any one call. If `algopt`[14] = 0.0, no limit is
          imposed.

        `algopt`[22]

          Specifies what method is to be used to solve the nonlinear
          equations at the initial point to initialize the values of U,
          U_t, V and V^.. If `algopt`[22] = 1.0, a modified Newton
          iteration is used and if `algopt`[22] = 2.0, functional
          iteration is used. The default value is `algopt`[22] = 1.0.

        `algopt`[28] and `algopt`[29] are used only for the sparse
        matrix algebra option, i.e., `laopt` = 'S'.

        `algopt`[28]

          Governs the choice of pivots during the decomposition of the
          first Jacobian matrix. It should lie in the range 0.0 <
          `algopt`[28] < 1.0, with smaller values biasing the algorithm
          towards maintaining sparsity at the expense of numerical
          stability. If `algopt`[28] lies outside this range then the
          default value is used. If the functions regard the Jacobian
          matrix as numerically singular then increasing `algopt`[28]
          towards 1.0 may help, but at the cost of increased fill-in.
          The default value is `algopt`[28] = 0.1.

        `algopt`[29]

          Used as a relative pivot threshold during subsequent Jacobian
          decompositions (see `algopt`[28]) below which an internal
          error is invoked. `algopt`[29] must be greater than zero,
          otherwise the default value is used. If `algopt`[29] is
          greater than 1.0 no check is made on the pivot size, and this
          may be a necessary option if the Jacobian is found to be
          numerically singular (see `algopt`[28]). The default value is
          `algopt`[29] = 0.0001.

    comm : dict, communication object, modified in place
        Note: this argument will be (re-)initialized when it is an empty
        ``dict`` or under the following condition: `ind` = 0.

        Communication structure.

        On initial entry: need not be set.

    itask : int
        The task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step in the time direction and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

        `itask` = 4

          Normal computation of output values `u` at t = `tout` but
          without overshooting t = t_crit where t_crit is described
          under the argument `algopt`.

        `itask` = 5

          Take one step in the time direction and return, without
          passing t_crit, where t_crit is described under the argument
          `algopt`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_dae_keller`` and the underlying ODE solver as
        follows:

        `itrace` <= -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` = 1

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        `itrace` = 2

          Output from the underlying ODE solver is similar to that
          produced when `itrace` = 1, except that the advisory messages
          are given in greater detail.

        `itrace` >= 3

          Output from the underlying ODE solver is similar to that
          produced when `itrace` = 2, except that the advisory messages
          are given in greater detail.

        You advised to set `itrace` = 0, unless you are experienced with
        submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_dae_keller``.

    odedef : None or callable (r, ires) = odedef(t, v, vdot, xi, ucp,
        ucpx, ucpt, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `odedef` must evaluate the functions R, which define the system
        of ODEs, as given in (4).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling points,
            xi_i, for i = 1,2,...,nxi.

        ucp : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucp`[i-1,j-1] contains the value of U_i(x,t) at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpx : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpx`[i-1,j-1] contains the value of
            dU_i(x,t)/dx at the coupling point x = xi_j, for j =
            1,2,...,nxi, for i = 1,2,...,`npde`.

        ucpt : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpt`[i-1,j-1] contains the value of dU_i/dt at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ires : int
            The form of R that must be returned in the array `r`.

            `ires` = -1

              Equation (3) must be used.

            `ires` = 1

              Equation (4) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (nv)
            If `nv` > 0, `r`[i-1] must contain the ith component of R,
            for i = 1,2,...,`nv`, where R is defined as

                R = -BV^.-CU_t^*,

            i.e., only terms depending explicitly on time derivatives,
            or

                R = A-BV^.-CU_t^*,

            i.e., all terms in equation (4). The definition of R is
            determined by the input value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may reset
            `ires` to force the integration function to take certain
            actions, as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_dae_keller`` returns to the calling function
              with the error indicator set to `errno` = 4.

    lrsave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['rsave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lrsave` is computed by the
        function.

        If your supplied `lrsave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lrsave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lrsave` set to at least the lower-bound value returned in
        `lrsave_min`.

    lisave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['isave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lisave` is computed by the
        function.

        If your supplied `lisave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lisave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lisave` set to at least the lower-bound value returned in
        `lisave_min`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution in `u`. Normally
        `ts` = `tout`.

    u : float, ndarray, shape (neqn)
        The computed solution U_i(x_j,t), for j = 1,2,...,npts, for i =
        1,2,...,`npde`, and V_k(t), for k = 1,2,...,`nv`, evaluated at t
        = `ts`, as follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t), for i = 1,2,...,`nv`.

    ind : int
        `ind` = 1.

    lrsave_min : int
        A lower bound on the sufficient size for `comm`['rsave'].

    lisave_min : int
        A lower bound on the sufficient size for `comm`['isave'].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `lrsave` = *<value>*.

            Constraint: `lrsave` >= *<value>*.

        (`errno` 1)
            On entry, `lisave` = *<value>*.

            Constraint: `lisave` >= *<value>*.

        (`errno` 1)
            On entry, at least one point in `xi` lies outside
            [`x`[0],`x`[npts-1]]: `x`[0] = *<value>* and `x`[npts-1] =
            *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xi`[i] = *<value>* and `xi`[i-1] =
            *<value>*.

            Constraint: `xi`[i] > `xi`[i-1].

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: corresponding elements `atol`[i-1] and
            `rtol`[j-1] cannot both be 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `rtol`[i-1] = *<value>*.

            Constraint: `rtol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `atol`[i-1] = *<value>*.

            Constraint: `atol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, neqn = *<value>*, `npde` = *<value>*, npts =
            *<value>* and `nv` = *<value>*.

            Constraint: neqn = `npde`*npts+`nv`.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi = 0 when `nv` = 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi >= 0 when `nv` > 0.

        (`errno` 1)
            On entry, `nleft` = *<value>*, `npde` = *<value>*.

            Constraint: `nleft` <= `npde`.

        (`errno` 1)
            On entry, `nleft` = *<value>*.

            Constraint: `nleft` >= 0.

        (`errno` 1)
            On entry, `npde` = *<value>*.

            Constraint: `npde` >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `laopt` = *<value>*.

            Constraint: `laopt` = 'F', 'B' or 'S'.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 'A' or 'M'.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2, 3, 4 or 5.

        (`errno` 1)
            On entry, `nv` = *<value>*.

            Constraint: `nv` >= 0.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `atol` and `rtol` were too small to start integration.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef`,
            `bndary`, or `odedef`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 12)
            In solving ODE system, the maximum number of steps
            `algopt`[14] has been exceeded. `algopt`[14] = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `atol` and `rtol`.
            `ts` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef`, `bndary`, or `odedef`. Integration is
            successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `atol` or `rtol`
            are unlikely to result in a changed solution.

        (`errno` 13)
            Zero error weights encountered during time integration.

        (`errno` 15)
            When using the sparse option `lisave` or `lrsave` is too
            small: `lisave` = *<value>*, `lrsave` = *<value>*.
    """

    from ..base.pde import dim1_parab_dae_keller as b_dim1_parab_dae_keller

    fun_name = 'naginterfaces.library.pde.dim1_parab_dae_keller'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ut, ux, nv, v, vdot, res, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _res_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ut,
                ux,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _res_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ut,
                ux,
                v,
                vdot,
                ires,
            )
        res[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'res'},
            _res_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, ibnd, nobc, u, ut, nv, v, vdot, res, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _res_bndary, _ires_bndary = bndary(
                t,
                ibnd,
                nobc,
                u,
                ut,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _res_bndary, _ires_bndary = bndary(
                t,
                ibnd,
                nobc,
                u,
                ut,
                v,
                vdot,
                ires,
            )
        res[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'res'},
            _res_bndary,
            copy=False,
            exp_shape=(nobc,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_odedef_name = fun_name + ':odedef'

    def cb_odedef(npde, t, nv, v, vdot, nxi, xi, ucp, ucpx, ucpt, r, ires, data=None):
        """Wrapping user's odedef. Called as base odedef."""
        if data is not None:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
                data,
            )
        else:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_odedef_name, 'returned': 'r'},
            _r_odedef,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_odedef_base = _ires_odedef
        return _ires_odedef_base

    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqn_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _neqn = _b_utils._EngineIntScalarType(
        _neqn_py, {'fun_name': fun_name, 'entity': '_neqn'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _b_utils._check_callable(odedef, {'fun_name': fun_name, 'entity': 'odedef'}, allow_none=True)
    _odedef = None if odedef is None else cb_odedef
    _xi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xi'},
        xi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxi_py = _b_utils._EngineFloat64ArrayType.get_shape(_xi, {'fun_name': fun_name, 'entity': 'xi'}, exp_rank=1)[0]
    _nxi = _b_utils._EngineIntScalarType(
        _nxi_py, {'fun_name': fun_name, 'entity': '_nxi'},
    ).to_ctypes()
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _algopt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'algopt'},
        algopt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_dim1_parab_dae_keller(
            npde,
            ts,
            tout,
            _pdedef,
            _bndary,
            _u,
            _npts,
            _x,
            nleft,
            nv,
            _odedef,
            _nxi,
            _xi,
            _neqn,
            _rtol,
            _atol,
            itol,
            norm,
            laopt,
            _algopt,
            comm,
            lrsave,
            lisave,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _lrsave_min_py = _return_scalars.lrsave_min
    _lisave_min_py = _return_scalars.lisave_min
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'ind', 'lrsave_min', 'lisave_min'],
    )(
        _ts_py,
        _u_py,
        _ind_py,
        _lrsave_min_py,
        _lisave_min_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_convdiff_dae(npde, ts, tout, numflx, bndary, u, x, nv, xi, rtol, atol, itol, norm, laopt, algopt, comm, itask, itrace, ind, pdedef=None, odedef=None, lrsave=None, lisave=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of convection-diffusion PDEs with source terms in
    conservative form, coupled DAEs, method of lines, upwind scheme
    using numerical flux function based on Riemann solver, one space
    variable.

    ``dim1_parab_convdiff_dae`` integrates a system of linear or
    nonlinear convection-diffusion equations in one space dimension,
    with optional source terms and scope for coupled ordinary
    differential equations (ODEs).
    The system must be posed in conservative form.
    Convection terms are discretized using a sophisticated upwind scheme
    involving a user-supplied numerical flux function based on the
    solution of a Riemann problem at each mesh point.
    The method of lines is employed to reduce the partial differential
    equations (PDEs) to a system of ODEs, and the resulting system is
    solved using a backward differentiation formula (BDF) method or a
    Theta method.

    For full information please refer to the NAG Library document for
    d03pl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03plf.html

    Parameters
    ----------
    npde : int
        The number of PDEs to be solved.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    numflx : callable (flux, ires) = numflx(t, x, v, uleft, uright,
        ires, data=None)
        `numflx` must supply the numerical flux for each PDE given the
        left and right values of the solution vector `u`. `numflx` is
        called approximately midway between each pair of mesh points in
        turn by ``dim1_parab_convdiff_dae``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        uleft : float, ndarray, shape (npde)
            `uleft`[i-1] contains the left value of the component
            U_i(x), for i = 1,2,...,`npde`.

        uright : float, ndarray, shape (npde)
            `uright`[i-1] contains the right value of the component
            U_i(x), for i = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        flux : float, array-like, shape (npde)
            `flux`[i-1] must be set to the numerical flux F^_i, for i =
            1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_dae`` returns to the calling
              function with the error indicator set to `errno` = 4.

    bndary : callable (g, ires) = bndary(t, x, u, v, vdot, ibnd, ires,
        data=None)
        `bndary` must evaluate the functions G_i^L and G_i^R which
        describe the physical and numerical boundary conditions, as
        given by (9) and (0).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            The mesh points in the spatial direction. `x`[0] corresponds
            to the left-hand boundary, a, and `x`[npts-1] corresponds to
            the right-hand boundary, b.

        u : float, ndarray, shape (npde, npts)
            `u`[i-1,j-1] contains the value of the component U_i(x,t) at
            x = `x`[j-1], for j = 1,2,...,npts, for i = 1,2,...,`npde`.

            Note: if banded matrix algebra is to be used then the
            functions G_i^L and G_i^R may depend on the value of
            U_i(x,t) at the boundary point and the two adjacent points
            only.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in G_j^L and G_j^R, for j = 1,2,...,`npde`.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must evaluate the left-hand boundary condition at
              x = a.

            `ibnd` != 0

              `bndary` must evaluate the right-hand boundary condition
              at x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        g : float, array-like, shape (npde)
            `g`[i-1] must contain the ith component of either G_i^L or
            G_i^R in (9) and (0), depending on the value of `ibnd`, for
            i = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_dae`` returns to the calling
              function with the error indicator set to `errno` = 4.

    u : float, array-like, shape (neqn)
        The initial values of the dependent variables defined as
        follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t_0), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t_0), for i = 1,2,...,`nv`.

    x : float, array-like, shape (npts)
        The mesh points in the space direction. `x`[0] must specify the
        left-hand boundary, a, and `x`[npts-1] must specify the
        right-hand boundary, b.

    nv : int
        The number of coupled ODE components.

    xi : float, array-like, shape (nxi)
        `xi`[i-1], for i = 1,2,...,nxi, must be set to the ODE/PDE
        coupling points.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; if `itol` in (3, 4): neqn;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; if `itol` in (2, 4): neqn;
        otherwise: 0.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. If e_i is
        the estimated local error for `u`[i-1], for i = 1,2,...,neqn,
        and || || denotes the norm, the error test to be satisfied is
        ||e_i|| < 1.0. `itol` indicates to ``dim1_parab_convdiff_dae``
        whether to interpret either or both of `rtol` and `atol` as a
        vector or scalar in the formation of the weights w_i used in the
        calculation of the norm (see the description of `norm`):

        [table omitted]

    norm : str, length 1
        The type of norm to be used.

        `norm` = '1'

          Averaged L_1 norm.

        `norm` = '2'

          Averaged L_2 norm.

        If U_norm denotes the norm of the vector `u` of length neqn,
        then for the averaged L_1 norm

            U_norm = 1/neqnsum_i = 1^neqn`u`[i-1]/w_i,

        and for the averaged L_2 norm

            U_norm = sqrt(1/neqnsum_i = 1^neqn(`u`[i-1]/w_i)^2).

        See the description of `itol` for the formulation of the weight
        vector w.

    laopt : str, length 1
        The type of matrix algebra required.

        `laopt` = 'F'

          Full matrix methods to be used.

        `laopt` = 'B'

          Banded matrix methods to be used.

        `laopt` = 'S'

          Sparse matrix methods to be used.

    algopt : float, array-like, shape (30)
        May be set to control various options available in the
        integrator. If you wish to employ all the default options,
        `algopt`[0] should be set to 0.0. Default values will also be
        used for any other elements of `algopt` set to zero. The
        permissible values, default values, and meanings are as follows:

        `algopt`[0]

          Selects the ODE integration method to be used. If `algopt`[0]
          = 1.0, a BDF method is used and if `algopt`[0] = 2.0, a Theta
          method is used. The default is `algopt`[0] = 1.0.

        If `algopt`[0] = 2.0, then `algopt`[i-1], for i = 2,3,...,4, are
        not used.

        `algopt`[1]

          Specifies the maximum order of the BDF integration formula to
          be used. `algopt`[1] may be 1.0, 2.0, 3.0, 4.0 or 5.0. The
          default value is `algopt`[1] = 5.0.

        `algopt`[2]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the BDF method. If
          `algopt`[2] = 1.0 a modified Newton iteration is used and if
          `algopt`[2] = 2.0 a functional iteration method is used. If
          functional iteration is selected and the integrator encounters
          difficulty, there is an automatic switch to the modified
          Newton iteration. The default value is `algopt`[2] = 1.0.

        `algopt`[3]

          Specifies whether or not the Petzold error test is to be
          employed. The Petzold error test results in extra overhead but
          is more suitable when algebraic equations are present, such as
          P_{i,j} = 0.0, for j = 1,2,...,`npde`, for some i or when
          there is no V^._i(t) dependence in the coupled ODE system. If
          `algopt`[3] = 1.0, the Petzold test is used. If `algopt`[3] =
          2.0, the Petzold test is not used. The default value is
          `algopt`[3] = 1.0.

        If `algopt`[0] = 1.0, `algopt`[i-1], for i = 5,6,...,7, are not
        used.

        `algopt`[4]

          Specifies the value of Theta to be used in the Theta
          integration method. 0.51 <= `algopt`[4] <= 0.99. The default
          value is `algopt`[4] = 0.55.

        `algopt`[5]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the Theta method.
          If `algopt`[5] = 1.0, a modified Newton iteration is used and
          if `algopt`[5] = 2.0, a functional iteration method is used.
          The default value is `algopt`[5] = 1.0.

        `algopt`[6]

          Specifies whether or not the integrator is allowed to switch
          automatically between modified Newton and functional iteration
          methods in order to be more efficient. If `algopt`[6] = 1.0,
          switching is allowed and if `algopt`[6] = 2.0, switching is
          not allowed. The default value is `algopt`[6] = 1.0.

        `algopt`[10]

          Specifies a point in the time direction, t_crit, beyond which
          integration must not be attempted. The use of t_crit is
          described under the argument `itask`. If `algopt`[0] != 0.0, a
          value of 0.0 for `algopt`[10], say, should be specified even
          if `itask` subsequently specifies that t_crit will not be
          used.

        `algopt`[11]

          Specifies the minimum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[11]
          should be set to 0.0.

        `algopt`[12]

          Specifies the maximum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[12]
          should be set to 0.0.

        `algopt`[13]

          Specifies the initial step size to be attempted by the
          integrator. If `algopt`[13] = 0.0, the initial step size is
          calculated internally.

        `algopt`[14]

          Specifies the maximum number of steps to be attempted by the
          integrator in any one call. If `algopt`[14] = 0.0, no limit is
          imposed.

        `algopt`[22]

          Specifies what method is to be used to solve the nonlinear
          equations at the initial point to initialize the values of U,
          U_t, V and V^.. If `algopt`[22] = 1.0, a modified Newton
          iteration is used and if `algopt`[22] = 2.0, functional
          iteration is used. The default value is `algopt`[22] = 1.0.

        `algopt`[28] and `algopt`[29] are used only for the sparse
        matrix algebra option, i.e., `laopt` = 'S'.

        `algopt`[28]

          Governs the choice of pivots during the decomposition of the
          first Jacobian matrix. It should lie in the range 0.0 <
          `algopt`[28] < 1.0, with smaller values biasing the algorithm
          towards maintaining sparsity at the expense of numerical
          stability. If `algopt`[28] lies outside the range then the
          default value is used. If the functions regard the Jacobian
          matrix as numerically singular, increasing `algopt`[28]
          towards 1.0 may help, but at the cost of increased fill-in.
          The default value is `algopt`[28] = 0.1.

        `algopt`[29]

          Used as the relative pivot threshold during subsequent
          Jacobian decompositions (see `algopt`[28]) below which an
          internal error is invoked. `algopt`[29] must be greater than
          zero, otherwise the default value is used. If `algopt`[29] is
          greater than 1.0 no check is made on the pivot size, and this
          may be a necessary option if the Jacobian matrix is found to
          be numerically singular (see `algopt`[28]). The default value
          is `algopt`[29] = 0.0001.

    comm : dict, communication object, modified in place
        Note: this argument will be (re-)initialized when it is an empty
        ``dict`` or under the following condition: `ind` = 0.

        Communication structure.

        On initial entry: need not be set.

    itask : int
        The task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step in the time direction and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

        `itask` = 4

          Normal computation of output values `u` at t = `tout` but
          without overshooting t = t_crit where t_crit is described
          under the argument `algopt`.

        `itask` = 5

          Take one step in the time direction and return, without
          passing t_crit, where t_crit is described under the argument
          `algopt`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_convdiff_dae`` and the underlying ODE solver.
        `itrace` may take the value -1, 0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` > 0

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases. You are advised to set `itrace` = 0, unless you are
        experienced with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` should be
          reset between calls to ``dim1_parab_convdiff_dae``.

    pdedef : None or callable (p, c, d, s, ires) = pdedef(t, x, u, ux,
        v, vdot, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `pdedef` must evaluate the functions P_{i,j}, C_i, D_i and S_i
        which partially define the system of PDEs. P_{i,j} and C_i may
        depend on x, t, U and V; D_i may depend on x, t, U, U_x and V;
        and S_i may depend on x, t, U, V and linearly on V^.. `pdedef`
        is called approximately midway between each pair of mesh points
        in turn by ``dim1_parab_convdiff_dae``. None may be used for
        `pdedef` for problems in the form (2).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in S_j, for j = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde)
            `p`[i-1,j-1] must be set to the value of P_{i,j}(x,t,U,V),
            for j = 1,2,...,`npde`, for i = 1,2,...,`npde`.

        c : float, array-like, shape (npde)
            `c`[i-1] must be set to the value of C_i(x,t,U,V), for i =
            1,2,...,`npde`.

        d : float, array-like, shape (npde)
            `d`[i-1] must be set to the value of D_i(x,t,U,U_x,V), for i
            = 1,2,...,`npde`.

        s : float, array-like, shape (npde)
            `s`[i-1] must be set to the value of S_i(x,t,U,V,V^.), for i
            = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_dae`` returns to the calling
              function with the error indicator set to `errno` = 4.

    odedef : None or callable (r, ires) = odedef(t, v, vdot, xi, ucp,
        ucpx, ucpt, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `odedef` must evaluate the functions R, which define the system
        of ODEs, as given in (4).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling point,
            xi_i, for i = 1,2,...,nxi.

        ucp : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucp`[i-1,j-1] contains the value of U_i(x,t) at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpx : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpx`[i-1,j-1] contains the value of
            dU_i(x,t)/dx at the coupling point x = xi_j, for j =
            1,2,...,nxi, for i = 1,2,...,`npde`.

        ucpt : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpt`[i-1,j-1] contains the value of dU_i/dt at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ires : int
            The form of R that must be returned in the array `r`.

            `ires` = 1

              Equation (1) must be used.

            `ires` = -1

              Equation (2) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (nv)
            `r`[i-1] must contain the ith component of R, for i =
            1,2,...,`nv`, where R is defined as

                R = L-MV^.-NU_t^*,

            or

                R = -MV^.-NU_t^*.

            The definition of R is determined by the input value of
            `ires`.

        ires : int
            Should usually remain unchanged. However, you may reset
            `ires` to force the integration function to take certain
            actions, as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_dae`` returns to the calling
              function with the error indicator set to `errno` = 4.

    lrsave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['rsave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lrsave` is computed by the
        function.

        If your supplied `lrsave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lrsave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lrsave` set to at least the lower-bound value returned in
        `lrsave_min`.

    lisave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['isave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lisave` is computed by the
        function.

        If your supplied `lisave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lisave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lisave` set to at least the lower-bound value returned in
        `lisave_min`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `pdedef` is spiked (i.e., has unit extent in all but
        one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (neqn)
        The computed solution U_i(x_j,t), for j = 1,2,...,npts, for i =
        1,2,...,`npde`, and V_k(t), for k = 1,2,...,`nv`, evaluated at t
        = `ts`, as follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t), for i = 1,2,...,`nv`.

    ind : int
        `ind` = 1.

    lrsave_min : int
        A lower bound on the sufficient size for `comm`['rsave'].

    lisave_min : int
        A lower bound on the sufficient size for `comm`['isave'].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `lrsave` = *<value>*.

            Constraint: `lrsave` >= *<value>*.

        (`errno` 1)
            On entry, `lisave` = *<value>*.

            Constraint: `lisave` >= *<value>*.

        (`errno` 1)
            On entry, at least one point in `xi` lies outside
            [`x`[0],`x`[npts-1]]: `x`[0] = *<value>* and `x`[npts-1] =
            *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xi`[i] = *<value>* and `xi`[i-1] =
            *<value>*.

            Constraint: `xi`[i] > `xi`[i-1].

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: corresponding elements `atol`[i-1] and
            `rtol`[j-1] cannot both be 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `rtol`[i-1] = *<value>*.

            Constraint: `rtol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `atol`[i-1] = *<value>*.

            Constraint: `atol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, neqn = *<value>*, `npde` = *<value>*, npts =
            *<value>* and `nv` = *<value>*.

            Constraint: neqn = `npde`*npts+`nv`.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi = 0 when `nv` = 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi >= 0 when `nv` > 0.

        (`errno` 1)
            On entry, `npde` = *<value>*.

            Constraint: `npde` >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `laopt` = *<value>*.

            Constraint: `laopt` = 'F', 'B' or 'S'.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1' or '2'.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2, 3, 4 or 5.

        (`errno` 1)
            On entry, `nv` = *<value>*.

            Constraint: `nv` >= 0.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef`, `numflx`, `bndary` or
            `odedef`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `atol` and `rtol` were too small to start integration.

        (`errno` 8)
            `ires` set to an invalid value in a call to functions
            `pdedef`, `numflx`, `bndary` or `odedef`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 12)
            In solving ODE system, the maximum number of steps
            `algopt`[14] has been exceeded. `algopt`[14] = *<value>*.

        (`errno` 14)
            The functions P, D, or C appear to depend on time
            derivatives.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `atol` and `rtol`.
            `ts` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in functions `pdedef`, `numflx`, `bndary` or `odedef`.
            Integration is successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `atol` or `rtol`
            are unlikely to result in a changed solution.

        (`errno` 13)
            Zero error weights encountered during time integration.

        (`errno` 15)
            When using the sparse option `lisave` or `lrsave` is too
            small: `lisave` = *<value>*, `lrsave` = *<value>*.
    """

    from ..base.pde import dim1_parab_convdiff_dae as b_dim1_parab_convdiff_dae

    fun_name = 'naginterfaces.library.pde.dim1_parab_convdiff_dae'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ux, nv, v, vdot, p, c, d, s, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _c_pdedef, _d_pdedef, _s_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _p_pdedef, _c_pdedef, _d_pdedef, _s_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'c'},
            _c_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        d[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'd'},
            _d_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        s[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 's'},
            _s_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_pdedef_np
        return _ires_pdedef_base

    cb_numflx_name = fun_name + ':numflx'

    def cb_numflx(npde, t, x, nv, v, uleft, uright, flux, ires, data=None):
        """Wrapping user's numflx. Called as base numflx."""
        if data is not None:
            _flux_numflx, _ires_numflx = numflx(
                t,
                x,
                v,
                uleft,
                uright,
                ires,
                data,
            )
        else:
            _flux_numflx, _ires_numflx = numflx(
                t,
                x,
                v,
                uleft,
                uright,
                ires,
            )
        flux[:] = _utils._to_ndarray(
            {'fun_name': cb_numflx_name, 'returned': 'flux'},
            _flux_numflx,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_numflx_base = _ires_numflx
        return _ires_numflx_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, npts, t, x, u, nv, v, vdot, ibnd, g, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _g_bndary, _ires_bndary = bndary(
                t,
                x,
                u,
                v,
                vdot,
                ibnd,
                ires,
                data,
            )
        else:
            _g_bndary, _ires_bndary = bndary(
                t,
                x,
                u,
                v,
                vdot,
                ibnd,
                ires,
            )
        g[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'g'},
            _g_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_odedef_name = fun_name + ':odedef'

    def cb_odedef(npde, t, nv, v, vdot, nxi, xi, ucp, ucpx, ucpt, r, ires, data=None):
        """Wrapping user's odedef. Called as base odedef."""
        if data is not None:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
                data,
            )
        else:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_odedef_name, 'returned': 'r'},
            _r_odedef,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_odedef_base = _ires_odedef
        return _ires_odedef_base

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'}, allow_none=True)
    _pdedef = None if pdedef is None else cb_pdedef
    _b_utils._check_callable(numflx, {'fun_name': fun_name, 'entity': 'numflx'})
    _numflx = cb_numflx
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqn_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _neqn = _b_utils._EngineIntScalarType(
        _neqn_py, {'fun_name': fun_name, 'entity': '_neqn'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _b_utils._check_callable(odedef, {'fun_name': fun_name, 'entity': 'odedef'}, allow_none=True)
    _odedef = None if odedef is None else cb_odedef
    _xi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xi'},
        xi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxi_py = _b_utils._EngineFloat64ArrayType.get_shape(_xi, {'fun_name': fun_name, 'entity': 'xi'}, exp_rank=1)[0]
    _nxi = _b_utils._EngineIntScalarType(
        _nxi_py, {'fun_name': fun_name, 'entity': '_nxi'},
    ).to_ctypes()
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _algopt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'algopt'},
        algopt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_dim1_parab_convdiff_dae(
            npde,
            ts,
            tout,
            _pdedef,
            _numflx,
            _bndary,
            _u,
            _npts,
            _x,
            nv,
            _odedef,
            _nxi,
            _xi,
            _neqn,
            _rtol,
            _atol,
            itol,
            norm,
            laopt,
            _algopt,
            comm,
            lrsave,
            lisave,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _ind_py = _return_scalars.ind
    _lrsave_min_py = _return_scalars.lrsave_min
    _lisave_min_py = _return_scalars.lisave_min
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'ind', 'lrsave_min', 'lisave_min'],
    )(
        _ts_py,
        _u_py,
        _ind_py,
        _lrsave_min_py,
        _lisave_min_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_remesh_fd(npde, m, ts, tout, pdedef, bndary, uvinit, u, x, nv, xi, rtol, atol, itol, norm, laopt, algopt, remesh, xfix, nrmesh, dxmesh, trmesh, ipminf, comm, itask, itrace, ind, odedef=None, xratio=1.5, con=None, monitf=None, lrsave=None, lisave=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of parabolic PDEs, coupled DAEs, method of lines,
    finite differences, remeshing, one space variable.

    ``dim1_parab_remesh_fd`` integrates a system of linear or nonlinear
    parabolic partial differential equations (PDEs) in one space
    variable, with scope for coupled ordinary differential equations
    (ODEs), and automatic adaptive spatial remeshing.
    The spatial discretization is performed using finite differences,
    and the method of lines is employed to reduce the PDEs to a system
    of ODEs.
    The resulting system is solved using a Backward Differentiation
    Formula (BDF) method or a Theta method (switching between Newton's
    method and functional iteration).

    For full information please refer to the NAG Library document for
    d03pp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03ppf.html

    Parameters
    ----------
    npde : int
        The number of PDEs to be solved.

    m : int
        The coordinate system used:

        `m` = 0

          Indicates Cartesian coordinates.

        `m` = 1

          Indicates cylindrical polar coordinates.

        `m` = 2

          Indicates spherical polar coordinates.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (p, q, r, ires) = pdedef(t, x, u, ux, v, vdot,
        ires, data=None)
        `pdedef` must evaluate the functions P_{i,j}, Q_i and R_i which
        define the system of PDEs.

        The functions may depend on x, t, U, U_x and V. Q_i may depend
        linearly on V^.. `pdedef` is called approximately midway between
        each pair of mesh points in turn by ``dim1_parab_remesh_fd``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in Q_j, for j = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde)
            `p`[i-1,j-1] must be set to the value of
            P_{i,j}(x,t,U,U_x,V), for j = 1,2,...,`npde`, for i =
            1,2,...,`npde`.

        q : float, array-like, shape (npde)
            `q`[i-1] must be set to the value of Q_i(x,t,U,U_x,V,V^.),
            for i = 1,2,...,`npde`.

        r : float, array-like, shape (npde)
            `r`[i-1] must be set to the value of R_i(x,t,U,U_x,V), for i
            = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_remesh_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    bndary : callable (beta, gamma, ires) = bndary(t, u, ux, v, vdot,
        ibnd, ires, data=None)
        `bndary` must evaluate the functions beta_i and gamma_i which
        describe the boundary conditions, as given in [equation].

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx
            at the boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            `vdot`[i-1] contains the value of component V^._i(t), for i
            = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in gamma_j, for j = 1,2,...,`npde`.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must set up the coefficients of the left-hand
              boundary, x = a.

            `ibnd` != 0

              `bndary` must set up the coefficients of the right-hand
              boundary, x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        beta : float, array-like, shape (npde)
            `beta`[i-1] must be set to the value of beta_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        gamma : float, array-like, shape (npde)
            `gamma`[i-1] must be set to the value of
            gamma_i(x,t,U,U_x,V,V^.) at the boundary specified by
            `ibnd`, for i = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_remesh_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    uvinit : callable (u, v) = uvinit(npde, x, xi, nv, data=None)
        `uvinit` must supply the initial (t = t_0) values of U(x,t) and
        V(t) for all values of x in the interval a <= x <= b.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        x : float, ndarray, shape (npts)
            The current mesh. `x`[i-1] contains the value of x_i, for i
            = 1,2,...,npts.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the value of the ODE/PDE
            coupling point, xi_i, for i = 1,2,...,nxi.

        nv : int
            The number of coupled ODEs in the system.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float, array-like, shape (`npde`, npts)
            `u`[i-1,j-1] contains the value of the component
            U_i(x_j,t_0), for j = 1,2,...,npts, for i = 1,2,...,`npde`.

        v : float, array-like, shape (`nv`)
            `v`[i-1] contains the value of component V_i(t_0), for i =
            1,2,...,`nv`.

    u : float, array-like, shape (neqn)
        If `ind` = 1 the value of `u` must be unchanged from the
        previous call.

    x : float, array-like, shape (npts)
        The initial mesh points in the space direction. `x`[0] must
        specify the left-hand boundary, a, and `x`[npts-1] must specify
        the right-hand boundary, b.

    nv : int
        The number of coupled ODE in the system.

    xi : float, array-like, shape (nxi)
        If nxi > 0, `xi`[i-1], for i = 1,2,...,nxi, must be set to the
        ODE/PDE coupling points.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; if `itol` in (3, 4): neqn;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; if `itol` in (2, 4): neqn;
        otherwise: 0.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``dim1_parab_remesh_fd`` whether to interpret
        either or both of `rtol` or `atol` as a vector or scalar. The
        error test to be satisfied is ||e_i/w_i|| < 1.0, where w_i is
        defined as follows:

        [table omitted]

        In the above, e_i denotes the estimated local error for the ith
        component of the coupled PDE/ODE system in time, `u`[i-1], for i
        = 1,2,...,neqn.

        The choice of norm used is defined by the argument `norm`.

    norm : str, length 1
        The type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L_2 norm.

        If `u`_norm denotes the norm of the vector `u` of length neqn,
        then for the averaged L_2 norm

            `u`_norm = sqrt(1/neqnsum_i = 1^neqn(`u`[i-1]/w_i)^2),

        while for the maximum norm

            `u`_norm = max_i(|`u`[i-1]/w_i|).

        See the description of `itol` for the formulation of the weight
        vector w.

    laopt : str, length 1
        The type of matrix algebra required.

        `laopt` = 'F'

          Full matrix methods to be used.

        `laopt` = 'B'

          Banded matrix methods to be used.

        `laopt` = 'S'

          Sparse matrix methods to be used.

    algopt : float, array-like, shape (30)
        May be set to control various options available in the
        integrator. If you wish to employ all the default options,
        `algopt`[0] should be set to 0.0. Default values will also be
        used for any other elements of `algopt` set to zero. The
        permissible values, default values, and meanings are as follows:

        `algopt`[0]

          Selects the ODE integration method to be used. If `algopt`[0]
          = 1.0, a BDF method is used and if `algopt`[0] = 2.0, a Theta
          method is used. The default value is `algopt`[0] = 1.0.

        If `algopt`[0] = 2.0, `algopt`[i-1], for i = 2,3,...,4 are not
        used.

        `algopt`[1]

          Specifies the maximum order of the BDF integration formula to
          be used. `algopt`[1] may be 1.0, 2.0, 3.0, 4.0 or 5.0. The
          default value is `algopt`[1] = 5.0.

        `algopt`[2]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the BDF method. If
          `algopt`[2] = 1.0 a modified Newton iteration is used and if
          `algopt`[2] = 2.0 a functional iteration method is used. If
          functional iteration is selected and the integrator encounters
          difficulty, there is an automatic switch to the modified
          Newton iteration. The default value is `algopt`[2] = 1.0.

        `algopt`[3]

          Specifies whether or not the Petzold error test is to be
          employed. The Petzold error test results in extra overhead but
          is more suitable when algebraic equations are present, such as
          P_{i,j} = 0.0, for j = 1,2,...,`npde`, for some i or when
          there is no V^._i(t) dependence in the coupled ODE system. If
          `algopt`[3] = 1.0, the Petzold test is used. If `algopt`[3] =
          2.0, the Petzold test is not used. The default value is
          `algopt`[3] = 1.0.

        If `algopt`[0] = 1.0, `algopt`[i-1], for i = 5,6,...,7, are not
        used.

        `algopt`[4]

          Specifies the value of Theta to be used in the Theta
          integration method. 0.51 <= `algopt`[4] <= 0.99. The default
          value is `algopt`[4] = 0.55.

        `algopt`[5]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the Theta method.
          If `algopt`[5] = 1.0, a modified Newton iteration is used and
          if `algopt`[5] = 2.0, a functional iteration method is used.
          The default value is `algopt`[5] = 1.0.

        `algopt`[6]

          Specifies whether or not the integrator is allowed to switch
          automatically between modified Newton and functional iteration
          methods in order to be more efficient. If `algopt`[6] = 1.0,
          switching is allowed and if `algopt`[6] = 2.0, switching is
          not allowed. The default value is `algopt`[6] = 1.0.

        `algopt`[10]

          Specifies a point in the time direction, t_crit, beyond which
          integration must not be attempted. The use of t_crit is
          described under the argument `itask`. If `algopt`[0] != 0.0, a
          value of 0.0 for `algopt`[10], say, should be specified even
          if `itask` subsequently specifies that t_crit will not be
          used.

        `algopt`[11]

          Specifies the minimum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[11]
          should be set to 0.0.

        `algopt`[12]

          Specifies the maximum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[12]
          should be set to 0.0.

        `algopt`[13]

          Specifies the initial step size to be attempted by the
          integrator. If `algopt`[13] = 0.0, the initial step size is
          calculated internally.

        `algopt`[14]

          Specifies the maximum number of steps to be attempted by the
          integrator in any one call. If `algopt`[14] = 0.0, no limit is
          imposed.

        `algopt`[22]

          Specifies what method is to be used to solve the nonlinear
          equations at the initial point to initialize the values of U,
          U_t, V and V^.. If `algopt`[22] = 1.0, a modified Newton
          iteration is used and if `algopt`[22] = 2.0, functional
          iteration is used. The default value is `algopt`[22] = 1.0.

        `algopt`[28] and `algopt`[29] are used only for the sparse
        matrix algebra option, `laopt` = 'S'.

        `algopt`[28]

          Governs the choice of pivots during the decomposition of the
          first Jacobian matrix. It should lie in the range 0.0 <
          `algopt`[28] < 1.0, with smaller values biasing the algorithm
          towards maintaining sparsity at the expense of numerical
          stability. If `algopt`[28] lies outside this range then the
          default value is used. If the functions regard the Jacobian
          matrix as numerically singular then increasing `algopt`[28]
          towards 1.0 may help, but at the cost of increased fill-in.
          The default value is `algopt`[28] = 0.1.

        `algopt`[29]

          Is used as a relative pivot threshold during subsequent
          Jacobian decompositions (see `algopt`[28]) below which an
          internal error is invoked. If `algopt`[29] is greater than 1.0
          no check is made on the pivot size, and this may be a
          necessary option if the Jacobian is found to be numerically
          singular (see `algopt`[28]). The default value is `algopt`[29]
          = 0.0001.

    remesh : bool
        Indicates whether or not spatial remeshing should be performed.

        `remesh` = True

          Indicates that spatial remeshing should be performed as
          specified.

        `remesh` = False

          Indicates that spatial remeshing should be suppressed.

        Note: `remesh` should not be changed between consecutive calls
        to ``dim1_parab_remesh_fd``. Remeshing can be switched off or on
        at specified times by using appropriate values for the arguments
        `nrmesh` and `trmesh` at each call.

    xfix : float, array-like, shape (nxfix)
        `xfix`[i-1], for i = 1,2,...,nxfix, must contain the value of
        the x coordinate at the ith fixed mesh point.

    nrmesh : int
        Specifies the spatial remeshing frequency and criteria for the
        calculation and adoption of a new mesh.

        `nrmesh` < 0

          Indicates that a new mesh is adopted according to the argument
          `dxmesh`. The mesh is tested every |`nrmesh`| timesteps.

        `nrmesh` = 0

          Indicates that remeshing should take place just once at the
          end of the first time step reached when t > `trmesh`.

        `nrmesh` > 0

          Indicates that remeshing will take place every `nrmesh` time
          steps, with no testing using `dxmesh`.

        Note: `nrmesh` may be changed between consecutive calls to
        ``dim1_parab_remesh_fd`` to give greater flexibility over the
        times of remeshing.

    dxmesh : float
        Determines whether a new mesh is adopted when `nrmesh` is set
        less than zero. A possible new mesh is calculated at the end of
        every |`nrmesh`| time steps, but is adopted only if

            x_i^(new) > x_i^(old)+`dxmesh`*(x_i+1^(old)-x_i^(old))

        or

            x_i^(new) < x_i^(old)-`dxmesh`*(x_i^(old)-x_i-1^(old))

        `dxmesh` thus imposes a lower limit on the difference between
        one mesh and the next.

    trmesh : float
        Specifies when remeshing will take place when `nrmesh` is set to
        zero. Remeshing will occur just once at the end of the first
        time step reached when t is greater than `trmesh`.

        Note: `trmesh` may be changed between consecutive calls to
        ``dim1_parab_remesh_fd`` to force remeshing at several specified
        times.

    ipminf : int
        The level of trace information regarding the adaptive remeshing.
        Details are directed to the file object associated with the
        advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

        `ipminf` = 0

          No trace information.

        `ipminf` = 1

          Brief summary of mesh characteristics.

        `ipminf` = 2

          More detailed information, including old and new mesh points,
          mesh sizes and monitor function values.

    comm : dict, communication object, modified in place
        Note: this argument will be (re-)initialized when it is an empty
        ``dict`` or under the following condition: `ind` = 0.

        Communication structure.

        On initial entry: need not be set.

    itask : int
        Specifies the task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout`.

        `itask` = 2

          One step and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

        `itask` = 4

          Normal computation of output values `u` at t = `tout` but
          without overshooting t = t_crit where t_crit is described
          under the argument `algopt`.

        `itask` = 5

          Take one step in the time direction and return, without
          passing t_crit, where t_crit is described under the argument
          `algopt`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_remesh_fd`` and the underlying ODE solver:

        `itrace` <= -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` = 1

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        `itrace` = 2

          Output from the underlying ODE solver is similar to that
          produced when `itrace` = 1, except that the advisory messages
          are given in greater detail.

        `itrace` >= 3

          Output from the underlying ODE solver is similar to that
          produced when `itrace` = 2, except that the advisory messages
          are given in greater detail.

    ind : int
        Must be set to 0 or 1.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` and the
          remeshing arguments `nrmesh`, `dxmesh`, `trmesh`, `xratio` and
          `con` may be reset between calls to ``dim1_parab_remesh_fd``.

    odedef : None or callable (f, ires) = odedef(t, v, vdot, xi, ucp,
        ucpx, rcp, ucpt, ucptx, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `odedef` must evaluate the functions F, which define the system
        of ODEs, as given in (3).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling points,
            xi_i, for i = 1,2,...,nxi.

        ucp : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucp`[i-1,j-1] contains the value of U_i(x,t) at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpx : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpx`[i-1,j-1] contains the value of
            dU_i(x,t)/dx at the coupling point x = xi_j, for j =
            1,2,...,nxi, for i = 1,2,...,`npde`.

        rcp : float, ndarray, shape (npde, nxi)
            `rcp`[i-1,j-1] contains the value of the flux R_i at the
            coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpt : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpt`[i-1,j-1] contains the value of dU_i/dt at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucptx : float, ndarray, shape (npde, nxi)
            `ucptx`[i-1,j-1] contains the value of d^2U_i/dxdt at the
            coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ires : int
            The form of F that must be returned in the array `f`.

            `ires` = 1

              Equation (5) must be used.

            `ires` = -1

              Equation (6) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (nv)
            On exit, `f`[i-1] must contain the ith component of F as
            specified by the input value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may reset
            `ires` to force the integration function to take certain
            actions as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3, ``dim1_parab_remesh_fd``
              returns to the calling function with the error indicator
              set to `errno` = 4.

    xratio : float, optional
        An input bound on the adjacent mesh ratio (greater than 1.0 and
        typically in the range 1.5 to 3.0). The remeshing functions will
        attempt to ensure that

            (x_i-x_{i-1})/`xratio` < x_{i+1}-x_i <
            `xratio`*(x_i-x_{i-1}).

    con : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 2.0/(npts-1).

        An input bound on the sub-integral of the monitor function
        F^mon(x) over each space step. The remeshing functions will
        attempt to ensure that

            int_x_i^x_{i+1}F^mon(x)dx <= `con`int_x_1^x_nptsF^mon(x)dx,

        . `con` gives you more control over the mesh distribution e.g.,
        decreasing `con` allows more clustering. A typical value is
        2/(npts-1), but you are encouraged to experiment with different
        values. Its value is not critical and the mesh should be
        qualitatively correct for all values in the range given below.

    monitf : None or callable fmon = monitf(t, x, u, r, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitf` must supply and evaluate a remesh monitor function to
        indicate the solution behaviour of interest.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            The current mesh. `x`[i-1] contains the value of x_i, for i
            = 1,2,...,npts.

        u : float, ndarray, shape (npde, npts)
            `u`[i-1,j-1] contains the value of U_i(x,t) at x = `x`[j-1]
            and time t, for j = 1,2,...,npts, for i = 1,2,...,`npde`.

        r : float, ndarray, shape (npde, npts)
            `r`[i-1,j-1] contains the value of R_i(x,t,U,U_x,V) at x =
            `x`[j-1] and time t, for j = 1,2,...,npts, for i =
            1,2,...,`npde`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fmon : float, array-like, shape (npts)
            `fmon`[i-1] must contain the value of the monitor function
            F^mon(x) at mesh point x = `x`[i-1].

    lrsave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['rsave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lrsave` is computed by the
        function.

        If your supplied `lrsave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lrsave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lrsave` set to at least the lower-bound value returned in
        `lrsave_min`.

    lisave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['isave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lisave` is computed by the
        function.

        If your supplied `lisave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lisave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lisave` set to at least the lower-bound value returned in
        `lisave_min`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `pdedef` or `u` in `uvinit` are spiked (i.e., have
        unit extent in all but one dimension, or have size 1),
        `spiked_sorder` selects the storage order to associate with them
        in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (neqn)
        The computed solution U_i(x_j,t), for j = 1,2,...,npts, for i =
        1,2,...,`npde`, and V_k(t), for k = 1,2,...,`nv`, evaluated at t
        = `ts`, as follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t), for i = 1,2,...,`nv`.

    x : float, ndarray, shape (npts)
        The final values of the mesh points.

    ind : int
        `ind` = 1.

    lrsave_min : int
        A lower bound on the sufficient size for `comm`['rsave'].

    lisave_min : int
        A lower bound on the sufficient size for `comm`['isave'].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `lrsave` = *<value>*.

            Constraint: `lrsave` >= *<value>*.

        (`errno` 1)
            On entry, `lisave` = *<value>*.

            Constraint: `lisave` >= *<value>*.

        (`errno` 1)
            On entry, the point `xfix`[i-1] does not coincide with any
            `x`[j-1]: i = *<value>* and `xfix`[i-1] = *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xfix`[i] = *<value>* and
            `xfix`[i-1] = *<value>*.

            Constraint: `xfix`[i] > `xfix`[i-1].

        (`errno` 1)
            On entry, at least one point in `xi` lies outside
            [`x`[0],`x`[npts-1]]: `x`[0] = *<value>* and `x`[npts-1] =
            *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xi`[i] = *<value>* and `xi`[i-1] =
            *<value>*.

            Constraint: `xi`[i] > `xi`[i-1].

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: corresponding elements `atol`[i-1] and
            `rtol`[j-1] cannot both be 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `rtol`[i-1] = *<value>*.

            Constraint: `rtol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `atol`[i-1] = *<value>*.

            Constraint: `atol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, neqn = *<value>*, `npde` = *<value>*, npts =
            *<value>* and `nv` = *<value>*.

            Constraint: neqn = `npde`*npts+`nv`.

        (`errno` 1)
            On entry, `con` = *<value>*, npts = *<value>*.

            Constraint: `con` <= 10.0/(npts-1).

        (`errno` 1)
            On entry, `con` = *<value>*, npts = *<value>*.

            Constraint: `con` >= 0.1/(npts-1).

        (`errno` 1)
            On entry, `xratio` = *<value>*.

            Constraint: `xratio` > 1.0.

        (`errno` 1)
            On entry, `ipminf` = *<value>*.

            Constraint: `ipminf` = 0, 1 or 2.

        (`errno` 1)
            On entry, `dxmesh` = *<value>*.

            Constraint: `dxmesh` >= 0.0.

        (`errno` 1)
            On entry, nxfix = *<value>*, npts = *<value>*.

            Constraint: nxfix <= npts-2.

        (`errno` 1)
            On entry, nxfix = *<value>*.

            Constraint: nxfix >= 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi = 0 when `nv` = 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi >= 0 when `nv` > 0.

        (`errno` 1)
            On entry, `npde` = *<value>*.

            Constraint: `npde` >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `m` = *<value>* and `x`[0] = *<value>*.

            Constraint: `m` <= 0 or `x`[0] >= 0.0

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` = 0, 1 or 2.

        (`errno` 1)
            On entry, `laopt` = *<value>*.

            Constraint: `laopt` = 'F', 'B' or 'S'.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 'A' or 'M'.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2, 3, 4 or 5.

        (`errno` 1)
            On entry, `nv` = *<value>*.

            Constraint: `nv` >= 0.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `atol` and `rtol` were too small to start integration.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef`,
            `bndary`, or `odedef`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 12)
            In solving ODE system, the maximum number of steps
            `algopt`[14] has been exceeded. `algopt`[14] = *<value>*.

        (`errno` 14)
            Flux function appears to depend on time derivatives.

        (`errno` 16)
            `remesh` has been changed between calls to
            ``dim1_parab_remesh_fd``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `atol` and `rtol`.
            `ts` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef`, `bndary`, or `odedef`. Integration is
            successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `atol` or `rtol`
            are unlikely to result in a changed solution.

        (`errno` 13)
            Zero error weights encountered during time integration.

        (`errno` 15)
            When using the sparse option `lisave` or `lrsave` is too
            small: `lisave` = *<value>*, `lrsave` = *<value>*.
    """

    from ..base.pde import dim1_parab_remesh_fd as b_dim1_parab_remesh_fd

    fun_name = 'naginterfaces.library.pde.dim1_parab_remesh_fd'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ux, nv, v, vdot, p, q, r, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _p_pdedef, _q_pdedef, _r_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        q[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'q'},
            _q_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'r'},
            _r_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_pdedef_np
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, u, ux, nv, v, vdot, ibnd, beta, gamma, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                v,
                vdot,
                ibnd,
                ires,
                data,
            )
        else:
            _beta_bndary, _gamma_bndary, _ires_bndary = bndary(
                t,
                u,
                ux,
                v,
                vdot,
                ibnd,
                ires,
            )
        beta[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'beta'},
            _beta_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        gamma[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'gamma'},
            _gamma_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_uvinit_name = fun_name + ':uvinit'

    def cb_uvinit(npde, npts, nxi, x, xi, u, nv, v, data=None):
        """Wrapping user's uvinit. Called as base uvinit."""
        if data is not None:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                xi,
                nv,
                data,
            )
        else:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                xi,
                nv,
            )
        _u_uvinit_np = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'u'},
            _u_uvinit,
            copy=False,
            exp_shape=(npde, npts),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        v[:] = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'v'},
            _v_uvinit,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_uvinit_name},
            ((_u_uvinit_np, 'u'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        u[:, :] = _u_uvinit_np

    cb_odedef_name = fun_name + ':odedef'

    def cb_odedef(npde, t, nv, v, vdot, nxi, xi, ucp, ucpx, rcp, ucpt, ucptx, f, ires, data=None):
        """Wrapping user's odedef. Called as base odedef."""
        if data is not None:
            _f_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                rcp,
                ucpt,
                ucptx,
                ires,
                data,
            )
        else:
            _f_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                rcp,
                ucpt,
                ucptx,
                ires,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_odedef_name, 'returned': 'f'},
            _f_odedef,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_odedef_base = _ires_odedef
        return _ires_odedef_base

    cb_monitf_name = fun_name + ':monitf'

    def cb_monitf(t, npts, npde, x, u, r, fmon, data=None):
        """Wrapping user's monitf. Called as base monitf."""
        if data is not None:
            _fmon_monitf = monitf(
                t,
                x,
                u,
                r,
                data,
            )
        else:
            _fmon_monitf = monitf(
                t,
                x,
                u,
                r,
            )
        fmon[:] = _utils._to_ndarray(
            {'fun_name': cb_monitf_name, 'returned': 'fmon'},
            _fmon_monitf,
            copy=False,
            exp_shape=(npts,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _b_utils._check_callable(uvinit, {'fun_name': fun_name, 'entity': 'uvinit'})
    _uvinit = cb_uvinit
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqn_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _neqn = _b_utils._EngineIntScalarType(
        _neqn_py, {'fun_name': fun_name, 'entity': '_neqn'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _b_utils._check_callable(odedef, {'fun_name': fun_name, 'entity': 'odedef'}, allow_none=True)
    _odedef = None if odedef is None else cb_odedef
    _xi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xi'},
        xi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxi_py = _b_utils._EngineFloat64ArrayType.get_shape(_xi, {'fun_name': fun_name, 'entity': 'xi'}, exp_rank=1)[0]
    _nxi = _b_utils._EngineIntScalarType(
        _nxi_py, {'fun_name': fun_name, 'entity': '_nxi'},
    ).to_ctypes()
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _algopt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'algopt'},
        algopt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xfix = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xfix'},
        xfix,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxfix_py = _b_utils._EngineFloat64ArrayType.get_shape(_xfix, {'fun_name': fun_name, 'entity': 'xfix'}, exp_rank=1)[0]
    _nxfix = _b_utils._EngineIntScalarType(
        _nxfix_py, {'fun_name': fun_name, 'entity': '_nxfix'},
    ).to_ctypes()
    _con = 2.0/(_npts_py-1) if con is None else con
    _b_utils._check_callable(monitf, {'fun_name': fun_name, 'entity': 'monitf'}, allow_none=True)
    _monitf = None if monitf is None else cb_monitf
    try:
        _return_scalars = b_dim1_parab_remesh_fd(
            npde,
            m,
            ts,
            tout,
            _pdedef,
            _bndary,
            _uvinit,
            _u,
            _npts,
            _x,
            nv,
            _odedef,
            _nxi,
            _xi,
            _neqn,
            _rtol,
            _atol,
            itol,
            norm,
            laopt,
            _algopt,
            remesh,
            _nxfix,
            _xfix,
            nrmesh,
            dxmesh,
            trmesh,
            ipminf,
            xratio,
            _con,
            _monitf,
            comm,
            lrsave,
            lisave,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _x_py = _x
    _ind_py = _return_scalars.ind
    _lrsave_min_py = _return_scalars.lrsave_min
    _lisave_min_py = _return_scalars.lisave_min
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'x', 'ind', 'lrsave_min', 'lisave_min'],
    )(
        _ts_py,
        _u_py,
        _x_py,
        _ind_py,
        _lrsave_min_py,
        _lisave_min_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_remesh_keller(npde, ts, tout, pdedef, bndary, uvinit, u, x, nleft, nv, xi, rtol, atol, itol, norm, laopt, algopt, remesh, xfix, nrmesh, dxmesh, trmesh, ipminf, comm, itask, itrace, ind, odedef=None, xratio=1.5, con=None, monitf=None, lrsave=None, lisave=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of first-order PDEs, coupled DAEs, method of lines,
    Keller box discretization, remeshing, one space variable.

    ``dim1_parab_remesh_keller`` integrates a system of linear or
    nonlinear, first-order, time-dependent partial differential
    equations (PDEs) in one space variable, with scope for coupled
    ordinary differential equations (ODEs), and automatic adaptive
    spatial remeshing.
    The spatial discretization is performed using the Keller box scheme
    and the method of lines is employed to reduce the PDEs to a system
    of ODEs.
    The resulting system is solved using a Backward Differentiation
    Formula (BDF) method or a Theta method (switching between Newton's
    method and functional iteration).

    For full information please refer to the NAG Library document for
    d03pr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03prf.html

    Parameters
    ----------
    npde : int
        The number of PDEs to be solved.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    pdedef : callable (res, ires) = pdedef(t, x, u, ut, ux, v, vdot,
        ires, data=None)
        `pdedef` must evaluate the functions G_i which define the system
        of PDEs. `pdedef` is called approximately midway between each
        pair of mesh points in turn by ``dim1_parab_remesh_keller``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,`npde`.

        ut : float, ndarray, shape (npde)
            `ut`[i-1] contains the value of the component dU_i(x,t)/dt,
            for i = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        ires : int
            The form of G_i that must be returned in the array `res`.

            `ires` = -1

              Equation (9) must be used.

            `ires` = 1

              Equation (0) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (npde)
            `res`[i-1] must contain the ith component of G, for i =
            1,2,...,`npde`, where G is defined as

                G_i = sum_j = 1^`npde`P_{i,j}dU_j/dt+sum_j =
                1^`nv`Q_{i,j}V^._j,

            i.e., only terms depending explicitly on time derivatives,
            or

                G_i = sum_j = 1^`npde`P_{i,j}dU_j/dt+sum_j =
                1^`nv`Q_{i,j}V^._j+S_i,

            i.e., all terms in equation (3).

            The definition of G is determined by the input value of
            `ires`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions,
            as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_remesh_keller`` returns to the calling
              function with the error indicator set to `errno` = 4.

    bndary : callable (res, ires) = bndary(t, ibnd, nobc, u, ut, v,
        vdot, ires, data=None)
        `bndary` must evaluate the functions G_i^L and G_i^R which
        describe the boundary conditions, as given in (5) and (6).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must compute the left-hand boundary condition at
              x = a.

            `ibnd` != 0

              `bndary` must compute of the right-hand boundary condition
              at x = b.

        nobc : int
            Specifies the number n_a of boundary conditions at the
            boundary specified by `ibnd`.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t) at the
            boundary specified by `ibnd`, for i = 1,2,...,`npde`.

        ut : float, ndarray, shape (npde)
            `ut`[i-1] contains the value of the component dU_i(x,t)/dt,
            for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: `vdot`[i-1], for i = 1,2,...,`nv`, may only appear
            linearly as in (1) and (2).

        ires : int
            The form of G_i^L (or G_i^R) that must be returned in the
            array `res`.

            `ires` = -1

              Equation (1) must be used.

            `ires` = 1

              Equation (2) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (`nobc`)
            `res`[i-1] must contain the ith component of G^L or G^R,
            depending on the value of `ibnd`, for i = 1,2,...,`nobc`,
            where G^L is defined as

                G_i^L = sum_j = 1^`npde`E_i,j^LdU_j/dt+sum_j =
                1^`nv`H_i,j^LV^._j,

            i.e., only terms depending explicitly on time derivatives,
            or

                G_i^L = sum_j = 1^`npde`E_i,j^LdU_j/dt+sum_j =
                1^`nv`H_i,j^LV^._j+K_i^L,

            i.e., all terms in equation (7), and similarly for G_i^R.

            The definitions of G^L and G^R are determined by the input
            value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_remesh_keller`` returns to the calling
              function with the error indicator set to `errno` = 4.

    uvinit : callable (u, v) = uvinit(npde, x, xi, nv, data=None)
        `uvinit` must supply the initial (t = t_0) values of U(x,t) and
        V(t) for all values of x in the interval [a,b].

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        x : float, ndarray, shape (npts)
            The current mesh. `x`[i-1] contains the value of x_i, for i
            = 1,2,...,npts.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling point,
            xi_i, for i = 1,2,...,nxi.

        nv : int
            The number of coupled ODEs in the system.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float, array-like, shape (`npde`, npts)
            `u`[i-1,j-1] contains the value of the component
            U_i(x_j,t_0), for j = 1,2,...,npts, for i = 1,2,...,`npde`.

        v : float, array-like, shape (`nv`)
            If `nv` > 0, `v`[i-1] must contain the value of component
            V_i(t_0), for i = 1,2,...,`nv`.

    u : float, array-like, shape (neqn)
        If `ind` = 1 the value of `u` must be unchanged from the
        previous call.

    x : float, array-like, shape (npts)
        The initial mesh points in the space direction. `x`[0] must
        specify the left-hand boundary, a, and `x`[npts-1] must specify
        the right-hand boundary, b.

    nleft : int
        The number n_a of boundary conditions at the left-hand mesh
        point `x`[0].

    nv : int
        The number of coupled ODE components.

    xi : float, array-like, shape (nxi)
        `xi`[i-1], for i = 1,2,...,nxi, must be set to the ODE/PDE
        coupling points, xi_i.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; if `itol` in (3, 4): neqn;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; if `itol` in (2, 4): neqn;
        otherwise: 0.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. `itol`
        indicates to ``dim1_parab_remesh_keller`` whether to interpret
        either or both of `rtol` or `atol` as a vector or scalar.

    norm : str, length 1
        The type of norm to be used.

        `norm` = 'M'

          Maximum norm.

        `norm` = 'A'

          Averaged L_2 norm.

        If U_norm denotes the norm of the vector `u` of length neqn,
        then for the averaged L_2 norm

            U_norm = sqrt(1/neqnsum_i = 1^neqn(U(i)/w_i)^2),

        while for the maximum norm

            U_norm = max_i|`u`[i-1]/w_i|.

        See the description of `itol` for the formulation of the weight
        vector w.

    laopt : str, length 1
        The type of matrix algebra required.

        `laopt` = 'F'

          Full matrix methods to be used.

        `laopt` = 'B'

          Banded matrix methods to be used.

        `laopt` = 'S'

          Sparse matrix methods to be used.

    algopt : float, array-like, shape (30)
        May be set to control various options available in the
        integrator. If you wish to employ all the default options,
        `algopt`[0] should be set to 0.0. Default values will also be
        used for any other elements of `algopt` set to zero. The
        permissible values, default values, and meanings are as follows:

        `algopt`[0]

          Selects the ODE integration method to be used. If `algopt`[0]
          = 1.0, a BDF method is used and if `algopt`[0] = 2.0, a Theta
          method is used. The default value is `algopt`[0] = 1.0.

        If `algopt`[0] = 2.0, then `algopt`[i-1], for i = 2,3,...,4, are
        not used.

        `algopt`[1]

          Specifies the maximum order of the BDF integration formula to
          be used. `algopt`[1] may be 1.0, 2.0, 3.0, 4.0 or 5.0. The
          default value is `algopt`[1] = 5.0.

        `algopt`[2]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the BDF method. If
          `algopt`[2] = 1.0 a modified Newton iteration is used and if
          `algopt`[2] = 2.0 a functional iteration method is used. If
          functional iteration is selected and the integrator encounters
          difficulty, there is an automatic switch to the modified
          Newton iteration. The default value is `algopt`[2] = 1.0.

        `algopt`[3]

          Specifies whether or not the Petzold error test is to be
          employed. The Petzold error test results in extra overhead but
          is more suitable when algebraic equations are present, such as
          P_{i,j} = 0.0, for j = 1,2,...,`npde`, for some i or when
          there is no V^._i(t) dependence in the coupled ODE system. If
          `algopt`[3] = 1.0, the Petzold test is used. If `algopt`[3] =
          2.0, the Petzold test is not used. The default value is
          `algopt`[3] = 1.0.

        If `algopt`[0] = 1.0, `algopt`[i-1], for i = 5,6,...,7, are not
        used.

        `algopt`[4]

          Specifies the value of Theta to be used in the Theta
          integration method. 0.51 <= `algopt`[4] <= 0.99. The default
          value is `algopt`[4] = 0.55.

        `algopt`[5]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the Theta method.
          If `algopt`[5] = 1.0, a modified Newton iteration is used and
          if `algopt`[5] = 2.0, a functional iteration method is used.
          The default value is `algopt`[5] = 1.0.

        `algopt`[6]

          Specifies whether or not the integrator is allowed to switch
          automatically between modified Newton and functional iteration
          methods in order to be more efficient. If `algopt`[6] = 1.0,
          switching is allowed and if `algopt`[6] = 2.0, switching is
          not allowed. The default value is `algopt`[6] = 1.0.

        `algopt`[10]

          Specifies a point in the time direction, t_crit, beyond which
          integration must not be attempted. The use of t_crit is
          described under the argument `itask`. If `algopt`[0] != 0.0, a
          value of 0.0, for `algopt`[10], say, should be specified even
          if `itask` subsequently specifies that t_crit will not be
          used.

        `algopt`[11]

          Specifies the minimum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[11]
          should be set to 0.0.

        `algopt`[12]

          Specifies the maximum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[12]
          should be set to 0.0.

        `algopt`[13]

          Specifies the initial step size to be attempted by the
          integrator. If `algopt`[13] = 0.0, the initial step size is
          calculated internally.

        `algopt`[14]

          Specifies the maximum number of steps to be attempted by the
          integrator in any one call. If `algopt`[14] = 0.0, no limit is
          imposed.

        `algopt`[22]

          Specifies what method is to be used to solve the nonlinear
          equations at the initial point to initialize the values of U,
          U_t, V and V^.. If `algopt`[22] = 1.0, a modified Newton
          iteration is used and if `algopt`[22] = 2.0, functional
          iteration is used. The default value is `algopt`[22] = 1.0.

        `algopt`[28] and `algopt`[29] are used only for the sparse
        matrix algebra option, i.e., `laopt` = 'S'.

        `algopt`[28]

          Governs the choice of pivots during the decomposition of the
          first Jacobian matrix. It should lie in the range 0.0 <
          `algopt`[28] < 1.0, with smaller values biasing the algorithm
          towards maintaining sparsity at the expense of numerical
          stability. If `algopt`[28] lies outside this range then the
          default value is used. If the functions regard the Jacobian
          matrix as numerically singular then increasing `algopt`[28]
          towards 1.0 may help, but at the cost of increased fill-in.
          The default value is `algopt`[28] = 0.1.

        `algopt`[29]

          Used as a relative pivot threshold during subsequent Jacobian
          decompositions (see `algopt`[28]) below which an internal
          error is invoked. `algopt`[29] must be greater than zero,
          otherwise the default value is used. If `algopt`[29] is
          greater than 1.0 no check is made on the pivot size, and this
          may be a necessary option if the Jacobian is found to be
          numerically singular (see `algopt`[28]). The default value is
          `algopt`[29] = 0.0001.

    remesh : bool
        Indicates whether or not spatial remeshing should be performed.

        `remesh` = True

          Indicates that spatial remeshing should be performed as
          specified.

        `remesh` = False

          Indicates that spatial remeshing should be suppressed.

        Note: `remesh` should not be changed between consecutive calls
        to ``dim1_parab_remesh_keller``. Remeshing can be switched off
        or on at specified times by using appropriate values for the
        arguments `nrmesh` and `trmesh` at each call.

    xfix : float, array-like, shape (nxfix)
        `xfix`[i-1], for i = 1,2,...,nxfix, must contain the value of
        the x coordinate at the ith fixed mesh point.

    nrmesh : int
        Indicates the form of meshing to be performed.

        `nrmesh` < 0

          Indicates that a new mesh is adopted according to the argument
          `dxmesh`. The mesh is tested every |`nrmesh`| timesteps.

        `nrmesh` = 0

          Indicates that remeshing should take place just once at the
          end of the first time step reached when t > `trmesh`.

        `nrmesh` > 0

          Indicates that remeshing will take place every `nrmesh` time
          steps, with no testing using `dxmesh`.

        Note: `nrmesh` may be changed between consecutive calls to
        ``dim1_parab_remesh_keller`` to give greater flexibility over
        the times of remeshing.

    dxmesh : float
        Determines whether a new mesh is adopted when `nrmesh` is set
        less than zero. A possible new mesh is calculated at the end of
        every |`nrmesh`| time steps, but is adopted only if

            x_i^new > x_i^old+`dxmesh`*(x_i+1^old-x_i^old),

        or

            x_i^new < x_i^old-`dxmesh`*(x_i^old-x_i-1^old).

        `dxmesh` thus imposes a lower limit on the difference between
        one mesh and the next.

    trmesh : float
        Specifies when remeshing will take place when `nrmesh` is set to
        zero. Remeshing will occur just once at the end of the first
        time step reached when t is greater than `trmesh`.

        Note: `trmesh` may be changed between consecutive calls to
        ``dim1_parab_remesh_keller`` to force remeshing at several
        specified times.

    ipminf : int
        The level of trace information regarding the adaptive remeshing.
        Details are directed to the file object associated with the
        advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

        `ipminf` = 0

          No trace information.

        `ipminf` = 1

          Brief summary of mesh characteristics.

        `ipminf` = 2

          More detailed information, including old and new mesh points,
          mesh sizes and monitor function values.

    comm : dict, communication object, modified in place
        Note: this argument will be (re-)initialized when it is an empty
        ``dict`` or under the following condition: `ind` = 0.

        Communication structure.

        On initial entry: need not be set.

    itask : int
        The task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step in the time direction and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

        `itask` = 4

          Normal computation of output values `u` at t = `tout` but
          without overshooting t = t_crit where t_crit is described
          under the argument `algopt`.

        `itask` = 5

          Take one step in the time direction and return, without
          passing t_crit, where t_crit is described under the argument
          `algopt`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_remesh_keller`` and the underlying ODE solver as
        follows:

        `itrace` <= -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` = 1

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        `itrace` = 2

          Output from the underlying ODE solver is similar to that
          produced when `itrace` = 1, except that the advisory messages
          are given in greater detail.

        `itrace` >= 3

          The output from the underlying ODE solver is similar to that
          produced when `itrace` = 2, except that the advisory messages
          are given in greater detail.

        You are advised to set `itrace` = 0, unless you are experienced
        with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the argument `tout` and the
          remeshing arguments `nrmesh`, `dxmesh`, `trmesh`, `xratio` and
          `con` may be reset between calls to
          ``dim1_parab_remesh_keller``.

    odedef : None or callable (r, ires) = odedef(t, v, vdot, xi, ucp,
        ucpx, ucpt, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `odedef` must evaluate the functions R, which define the system
        of ODEs, as given in (4).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling point,
            xi_i, for i = 1,2,...,nxi.

        ucp : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucp`[i-1,j-1] contains the value of U_i(x,t) at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpx : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpx`[i-1,j-1] contains the value of
            dU_i(x,t)/dx at the coupling point x = xi_j, for j =
            1,2,...,nxi, for i = 1,2,...,`npde`.

        ucpt : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpt`[i-1,j-1] contains the value of dU_i/dt at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ires : int
            The form of R that must be returned in the array `r`.

            `ires` = -1

              Equation (3) must be used.

            `ires` = 1

              Equation (4) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (nv)
            If `nv` > 0, `r`[i-1] must contain the ith component of R,
            for i = 1,2,...,`nv`, where R is defined as

                R = -BV^.-CU_t^*,

            i.e., only terms depending explicitly on time derivatives,
            or

                R = A-BV^.-CU_t^*,

            i.e., all terms in equation (4). The definition of R is
            determined by the input value of `ires`.

        ires : int
            Should usually remain unchanged. However, you may reset
            `ires` to force the integration function to take certain
            actions, as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_remesh_keller`` returns to the calling
              function with the error indicator set to `errno` = 4.

    xratio : float, optional
        An input bound on the adjacent mesh ratio (greater than 1.0 and
        typically in the range 1.5 to 3.0). The remeshing functions will
        attempt to ensure that

            (x_i-x_{i-1})/`xratio` < x_{i+1}-x_i <
            `xratio`*(x_i-x_{i-1}).

    con : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 2.0/(npts-1).

        An input bound on the sub-integral of the monitor function
        F^mon(x) over each space step. The remeshing functions will
        attempt to ensure that

            int_x_1^x_{i+1}F^mon(x)dx <= `con`int_x_1^x_nptsF^mon(x)dx,

        . `con` gives you more control over the mesh distribution e.g.,
        decreasing `con` allows more clustering. A typical value is
        2/(npts-1), but you are encouraged to experiment with different
        values. Its value is not critical and the mesh should be
        qualitatively correct for all values in the range given below.

    monitf : None or callable fmon = monitf(t, x, u, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitf` must supply and evaluate a remesh monitor function to
        indicate the solution behaviour of interest.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            The current mesh. `x`[i-1] contains the value of x_i, for i
            = 1,2,...,npts.

        u : float, ndarray, shape (npde, npts)
            `u`[i-1,j-1] contains the value of U_i(x,t) at x = `x`[j-1]
            and time t, for j = 1,2,...,npts, for i = 1,2,...,`npde`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fmon : float, array-like, shape (npts)
            `fmon`[i-1] must contain the value of the monitor function
            F^mon(x) at mesh point x = `x`[i-1].

    lrsave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['rsave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lrsave` is computed by the
        function.

        If your supplied `lrsave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lrsave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lrsave` set to at least the lower-bound value returned in
        `lrsave_min`.

    lisave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['isave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lisave` is computed by the
        function.

        If your supplied `lisave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lisave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lisave` set to at least the lower-bound value returned in
        `lisave_min`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `u` in `uvinit` is spiked (i.e., has unit extent in all but
        one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (neqn)
        The computed solution U_i(x_j,t), for j = 1,2,...,npts, for i =
        1,2,...,`npde`, and V_k(t), for k = 1,2,...,`nv`, evaluated at t
        = `ts`, as follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t), for i = 1,2,...,`nv`.

    x : float, ndarray, shape (npts)
        The final values of the mesh points.

    ind : int
        `ind` = 1.

    lrsave_min : int
        A lower bound on the sufficient size for `comm`['rsave'].

    lisave_min : int
        A lower bound on the sufficient size for `comm`['isave'].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `lrsave` = *<value>*.

            Constraint: `lrsave` >= *<value>*.

        (`errno` 1)
            On entry, `lisave` = *<value>*.

            Constraint: `lisave` >= *<value>*.

        (`errno` 1)
            On entry, the point `xfix`[i-1] does not coincide with any
            `x`[j-1]: i = *<value>* and `xfix`[i-1] = *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xfix`[i] = *<value>* and
            `xfix`[i-1] = *<value>*.

            Constraint: `xfix`[i] > `xfix`[i-1].

        (`errno` 1)
            On entry, at least one point in `xi` lies outside
            [`x`[0],`x`[npts-1]]: `x`[0] = *<value>* and `x`[npts-1] =
            *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xi`[i] = *<value>* and `xi`[i-1] =
            *<value>*.

            Constraint: `xi`[i] > `xi`[i-1].

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: corresponding elements `atol`[i-1] and
            `rtol`[j-1] cannot both be 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `rtol`[i-1] = *<value>*.

            Constraint: `rtol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `atol`[i-1] = *<value>*.

            Constraint: `atol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, neqn = *<value>*, `npde` = *<value>*, npts =
            *<value>* and `nv` = *<value>*.

            Constraint: neqn = `npde`*npts+`nv`.

        (`errno` 1)
            On entry, `con` = *<value>*, npts = *<value>*.

            Constraint: `con` <= 10.0/(npts-1).

        (`errno` 1)
            On entry, `con` = *<value>*, npts = *<value>*.

            Constraint: `con` >= 0.1/(npts-1).

        (`errno` 1)
            On entry, `xratio` = *<value>*.

            Constraint: `xratio` > 1.0.

        (`errno` 1)
            On entry, `ipminf` = *<value>*.

            Constraint: `ipminf` = 0, 1 or 2.

        (`errno` 1)
            On entry, `dxmesh` = *<value>*.

            Constraint: `dxmesh` >= 0.0.

        (`errno` 1)
            On entry, nxfix = *<value>*, npts = *<value>*.

            Constraint: nxfix <= npts-2.

        (`errno` 1)
            On entry, nxfix = *<value>*.

            Constraint: nxfix >= 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi = 0 when `nv` = 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi >= 0 when `nv` > 0.

        (`errno` 1)
            On entry, `nleft` = *<value>*, `npde` = *<value>*.

            Constraint: `nleft` <= `npde`.

        (`errno` 1)
            On entry, `nleft` = *<value>*.

            Constraint: `nleft` >= 0.

        (`errno` 1)
            On entry, `npde` = *<value>*.

            Constraint: `npde` >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `laopt` = *<value>*.

            Constraint: `laopt` = 'F', 'B' or 'S'.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 'A' or 'M'.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2, 3, 4 or 5.

        (`errno` 1)
            On entry, `nv` = *<value>*.

            Constraint: `nv` >= 0.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef` or `bndary`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `atol` and `rtol` were too small to start integration.

        (`errno` 8)
            `ires` set to an invalid value in call to `pdedef`,
            `bndary`, or `odedef`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 12)
            In solving ODE system, the maximum number of steps
            `algopt`[14] has been exceeded. `algopt`[14] = *<value>*.

        (`errno` 16)
            `remesh` has been changed between calls to
            ``dim1_parab_remesh_keller``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `atol` and `rtol`.
            `ts` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in `pdedef`, `bndary`, or `odedef`. Integration is
            successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `atol` or `rtol`
            are unlikely to result in a changed solution.

        (`errno` 13)
            Zero error weights encountered during time integration.

        (`errno` 15)
            When using the sparse option `lisave` or `lrsave` is too
            small: `lisave` = *<value>*, `lrsave` = *<value>*.
    """

    from ..base.pde import dim1_parab_remesh_keller as b_dim1_parab_remesh_keller

    fun_name = 'naginterfaces.library.pde.dim1_parab_remesh_keller'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ut, ux, nv, v, vdot, res, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _res_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ut,
                ux,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _res_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ut,
                ux,
                v,
                vdot,
                ires,
            )
        res[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'res'},
            _res_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        return _ires_pdedef_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, t, ibnd, nobc, u, ut, nv, v, vdot, res, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _res_bndary, _ires_bndary = bndary(
                t,
                ibnd,
                nobc,
                u,
                ut,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _res_bndary, _ires_bndary = bndary(
                t,
                ibnd,
                nobc,
                u,
                ut,
                v,
                vdot,
                ires,
            )
        res[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'res'},
            _res_bndary,
            copy=False,
            exp_shape=(nobc,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_uvinit_name = fun_name + ':uvinit'

    def cb_uvinit(npde, npts, nxi, x, xi, u, nv, v, data=None):
        """Wrapping user's uvinit. Called as base uvinit."""
        if data is not None:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                xi,
                nv,
                data,
            )
        else:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                xi,
                nv,
            )
        _u_uvinit_np = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'u'},
            _u_uvinit,
            copy=False,
            exp_shape=(npde, npts),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        v[:] = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'v'},
            _v_uvinit,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_uvinit_name},
            ((_u_uvinit_np, 'u'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        u[:, :] = _u_uvinit_np

    cb_odedef_name = fun_name + ':odedef'

    def cb_odedef(npde, t, nv, v, vdot, nxi, xi, ucp, ucpx, ucpt, r, ires, data=None):
        """Wrapping user's odedef. Called as base odedef."""
        if data is not None:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
                data,
            )
        else:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_odedef_name, 'returned': 'r'},
            _r_odedef,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_odedef_base = _ires_odedef
        return _ires_odedef_base

    cb_monitf_name = fun_name + ':monitf'

    def cb_monitf(t, npts, npde, x, u, fmon, data=None):
        """Wrapping user's monitf. Called as base monitf."""
        if data is not None:
            _fmon_monitf = monitf(
                t,
                x,
                u,
                data,
            )
        else:
            _fmon_monitf = monitf(
                t,
                x,
                u,
            )
        fmon[:] = _utils._to_ndarray(
            {'fun_name': cb_monitf_name, 'returned': 'fmon'},
            _fmon_monitf,
            copy=False,
            exp_shape=(npts,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _b_utils._check_callable(uvinit, {'fun_name': fun_name, 'entity': 'uvinit'})
    _uvinit = cb_uvinit
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqn_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _neqn = _b_utils._EngineIntScalarType(
        _neqn_py, {'fun_name': fun_name, 'entity': '_neqn'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _b_utils._check_callable(odedef, {'fun_name': fun_name, 'entity': 'odedef'}, allow_none=True)
    _odedef = None if odedef is None else cb_odedef
    _xi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xi'},
        xi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxi_py = _b_utils._EngineFloat64ArrayType.get_shape(_xi, {'fun_name': fun_name, 'entity': 'xi'}, exp_rank=1)[0]
    _nxi = _b_utils._EngineIntScalarType(
        _nxi_py, {'fun_name': fun_name, 'entity': '_nxi'},
    ).to_ctypes()
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _algopt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'algopt'},
        algopt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xfix = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xfix'},
        xfix,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxfix_py = _b_utils._EngineFloat64ArrayType.get_shape(_xfix, {'fun_name': fun_name, 'entity': 'xfix'}, exp_rank=1)[0]
    _nxfix = _b_utils._EngineIntScalarType(
        _nxfix_py, {'fun_name': fun_name, 'entity': '_nxfix'},
    ).to_ctypes()
    _con = 2.0/(_npts_py-1) if con is None else con
    _b_utils._check_callable(monitf, {'fun_name': fun_name, 'entity': 'monitf'}, allow_none=True)
    _monitf = None if monitf is None else cb_monitf
    try:
        _return_scalars = b_dim1_parab_remesh_keller(
            npde,
            ts,
            tout,
            _pdedef,
            _bndary,
            _uvinit,
            _u,
            _npts,
            _x,
            nleft,
            nv,
            _odedef,
            _nxi,
            _xi,
            _neqn,
            _rtol,
            _atol,
            itol,
            norm,
            laopt,
            _algopt,
            remesh,
            _nxfix,
            _xfix,
            nrmesh,
            dxmesh,
            trmesh,
            ipminf,
            xratio,
            _con,
            _monitf,
            comm,
            lrsave,
            lisave,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _x_py = _x
    _ind_py = _return_scalars.ind
    _lrsave_min_py = _return_scalars.lrsave_min
    _lisave_min_py = _return_scalars.lisave_min
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'x', 'ind', 'lrsave_min', 'lisave_min'],
    )(
        _ts_py,
        _u_py,
        _x_py,
        _ind_py,
        _lrsave_min_py,
        _lisave_min_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_convdiff_remesh(npde, ts, tout, numflx, bndary, uvinit, u, x, nv, xi, rtol, atol, itol, norm, laopt, algopt, remesh, xfix, nrmesh, dxmesh, trmesh, ipminf, comm, itask, itrace, ind, pdedef=None, odedef=None, xratio=1.5, con=None, monitf=None, lrsave=None, lisave=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of convection-diffusion PDEs, coupled DAEs, method of
    lines, upwind scheme, remeshing, one space variable.

    ``dim1_parab_convdiff_remesh`` integrates a system of linear or
    nonlinear convection-diffusion equations in one space dimension,
    with optional source terms and scope for coupled ordinary
    differential equations (ODEs).
    The system must be posed in conservative form.
    This function also includes the option of automatic adaptive spatial
    remeshing.
    Convection terms are discretized using a sophisticated upwind scheme
    involving a user-supplied numerical flux function based on the
    solution of a Riemann problem at each mesh point.
    The method of lines is employed to reduce the partial differential
    equations (PDEs) to a system of ODEs, and the resulting system is
    solved using a backward differentiation formula (BDF) method or a
    Theta method.

    For full information please refer to the NAG Library document for
    d03ps

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03psf.html

    Parameters
    ----------
    npde : int
        The number of PDEs to be solved.

    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    numflx : callable (flux, ires) = numflx(t, x, v, uleft, uright,
        ires, data=None)
        `numflx` must supply the numerical flux for each PDE given the
        left and right values of the solution vector `u`. `numflx` is
        called approximately midway between each pair of mesh points in
        turn by ``dim1_parab_convdiff_remesh``.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        uleft : float, ndarray, shape (npde)
            `uleft`[i-1] contains the left value of the component
            U_i(x), for i = 1,2,...,`npde`.

        uright : float, ndarray, shape (npde)
            `uright`[i-1] contains the right value of the component
            U_i(x), for i = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        flux : float, array-like, shape (npde)
            `flux`[i-1] must be set to the numerical flux F^_i, for i =
            1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_remesh`` returns to the calling
              function with the error indicator set to `errno` = 4.

    bndary : callable (g, ires) = bndary(t, x, u, v, vdot, ibnd, ires,
        data=None)
        `bndary` must evaluate the functions G_i^L and G_i^R which
        describe the physical and numerical boundary conditions, as
        given by (9) and (0).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            The mesh points in the spatial direction. `x`[0] corresponds
            to the left-hand boundary, a, and `x`[npts-1] corresponds to
            the right-hand boundary, b.

        u : float, ndarray, shape (npde, npts)
            `u`[i-1,j-1] contains the value of the component U_i(x,t) at
            x = `x`[j-1], for j = 1,2,...,npts, for i = 1,2,...,`npde`.

            Note: if banded matrix algebra is to be used then the
            functions G_i^L and G_i^R may depend on the value of
            U_i(x,t) at the boundary point and the two adjacent points
            only.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in G_j^L and G_j^R, for j = 1,2,...,`npde`.

        ibnd : int
            Specifies which boundary conditions are to be evaluated.

            `ibnd` = 0

              `bndary` must evaluate the left-hand boundary condition at
              x = a.

            `ibnd` != 0

              `bndary` must evaluate the right-hand boundary condition
              at x = b.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        g : float, array-like, shape (npde)
            `g`[i-1] must contain the ith component of either G_i^L or
            G_i^R in (9) and (0), depending on the value of `ibnd`, for
            i = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_remesh`` returns to the calling
              function with the error indicator set to `errno` = 4.

    uvinit : callable (u, v) = uvinit(npde, x, xi, nv, data=None)
        `uvinit` must supply the initial (t = t_0) values of U(x,t) and
        V(t) for all values of x in the interval a <= x <= b.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        x : float, ndarray, shape (npts)
            The current mesh. `x`[i-1] contains the value of x_i, for i
            = 1,2,...,npts.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling point,
            xi_i, for i = 1,2,...,nxi.

        nv : int
            The number of coupled ODEs in the system.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float, array-like, shape (`npde`, npts)
            `u`[i-1,j-1] contains the value of the component
            U_i(x_j,t_0), for j = 1,2,...,npts, for i = 1,2,...,`npde`.

        v : float, array-like, shape (`nv`)
            If `nv` > 0, `v`[i-1] must contain the value of component
            V_i(t_0), for i = 1,2,...,`nv`.

    u : float, array-like, shape (neqn)
        If `ind` = 1 the value of `u` must be unchanged from the
        previous call.

    x : float, array-like, shape (npts)
        The mesh points in the space direction. `x`[0] must specify the
        left-hand boundary, a, and `x`[npts-1] must specify the
        right-hand boundary, b.

    nv : int
        The number of coupled ODE components.

    xi : float, array-like, shape (nxi)
        If nxi > 0, `xi`[i-1], for i = 1,2,...,nxi, must be set to the
        ODE/PDE coupling points.

    rtol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 2): 1; if `itol` in (3, 4): neqn;
        otherwise: 0.

        The relative local error tolerance.

    atol : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itol` in (1, 3): 1; if `itol` in (2, 4): neqn;
        otherwise: 0.

        The absolute local error tolerance.

    itol : int
        A value to indicate the form of the local error test. If e_i is
        the estimated local error for `u`[i-1], for i = 1,2,...,neqn,
        and || ||, denotes the norm, the error test to be satisfied is
        ||e_i|| < 1.0. `itol` indicates to
        ``dim1_parab_convdiff_remesh`` whether to interpret either or
        both of `rtol` and `atol` as a vector or scalar in the formation
        of the weights w_i used in the calculation of the norm (see the
        description of `norm`):

        [table omitted]

    norm : str, length 1
        The type of norm to be used.

        `norm` = '1'

          Averaged L_1 norm.

        `norm` = '2'

          Averaged L_2 norm.

        If U_norm denotes the norm of the vector `u` of length neqn,
        then for the averaged L_1 norm

            U_norm = 1/neqnsum_i = 1^neqn`u`[i-1]/w_i,

        and for the averaged L_2 norm

            U_norm = sqrt(1/neqnsum_i = 1^neqn(`u`[i-1]/w_i)^2),

        See the description of `itol` for the formulation of the weight
        vector w.

    laopt : str, length 1
        The type of matrix algebra required.

        `laopt` = 'F'

          Full matrix methods to be used.

        `laopt` = 'B'

          Banded matrix methods to be used.

        `laopt` = 'S'

          Sparse matrix methods to be used.

    algopt : float, array-like, shape (30)
        May be set to control various options available in the
        integrator. If you wish to employ all the default options,
        `algopt`[0] should be set to 0.0. Default values will also be
        used for any other elements of `algopt` set to zero. The
        permissible values, default values, and meanings are as follows:

        `algopt`[0]

          Selects the ODE integration method to be used. If `algopt`[0]
          = 1.0, a BDF method is used and if `algopt`[0] = 2.0, a Theta
          method is used. The default is `algopt`[0] = 1.0.

        If `algopt`[0] = 2.0, then `algopt`[i-1], for i = 2,3,...,4, are
        not used.

        `algopt`[1]

          Specifies the maximum order of the BDF integration formula to
          be used. `algopt`[1] may be 1.0, 2.0, 3.0, 4.0 or 5.0. The
          default value is `algopt`[1] = 5.0.

        `algopt`[2]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the BDF method. If
          `algopt`[2] = 1.0 a modified Newton iteration is used and if
          `algopt`[2] = 2.0 a functional iteration method is used. If
          functional iteration is selected and the integrator encounters
          difficulty, there is an automatic switch to the modified
          Newton iteration. The default value is `algopt`[2] = 1.0.

        `algopt`[3]

          Specifies whether or not the Petzold error test is to be
          employed. The Petzold error test results in extra overhead but
          is more suitable when algebraic equations are present, such as
          P_{i,j} = 0.0, for j = 1,2,...,`npde`, for some i or when
          there is no V^._i(t) dependence in the coupled ODE system. If
          `algopt`[3] = 1.0, the Petzold test is used. If `algopt`[3] =
          2.0, the Petzold test is not used. The default value is
          `algopt`[3] = 1.0.

        If `algopt`[0] = 1.0, `algopt`[i-1], for i = 5,6,...,7, are not
        used.

        `algopt`[4]

          Specifies the value of Theta to be used in the Theta
          integration method. 0.51 <= `algopt`[4] <= 0.99. The default
          value is `algopt`[4] = 0.55.

        `algopt`[5]

          Specifies what method is to be used to solve the system of
          nonlinear equations arising on each step of the Theta method.
          If `algopt`[5] = 1.0, a modified Newton iteration is used and
          if `algopt`[5] = 2.0, a functional iteration method is used.
          The default value is `algopt`[5] = 1.0.

        `algopt`[6]

          Specifies whether or not the integrator is allowed to switch
          automatically between modified Newton and functional iteration
          methods in order to be more efficient. If `algopt`[6] = 1.0,
          switching is allowed and if `algopt`[6] = 2.0, switching is
          not allowed. The default value is `algopt`[6] = 1.0.

        `algopt`[10]

          Specifies a point in the time direction, t_crit, beyond which
          integration must not be attempted. The use of t_crit is
          described under the argument `itask`. If `algopt`[0] != 0.0, a
          value of 0.0 for `algopt`[10], say, should be specified even
          if `itask` subsequently specifies that t_crit will not be
          used.

        `algopt`[11]

          Specifies the minimum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[11]
          should be set to 0.0.

        `algopt`[12]

          Specifies the maximum absolute step size to be allowed in the
          time integration. If this option is not required, `algopt`[12]
          should be set to 0.0.

        `algopt`[13]

          Specifies the initial step size to be attempted by the
          integrator. If `algopt`[13] = 0.0, the initial step size is
          calculated internally.

        `algopt`[14]

          Specifies the maximum number of steps to be attempted by the
          integrator in any one call. If `algopt`[14] = 0.0, no limit is
          imposed.

        `algopt`[22]

          Specifies what method is to be used to solve the nonlinear
          equations at the initial point to initialize the values of U,
          U_t, V and V^.. If `algopt`[22] = 1.0, a modified Newton
          iteration is used and if `algopt`[22] = 2.0, functional
          iteration is used. The default value is `algopt`[22] = 1.0.

        `algopt`[28] and `algopt`[29] are used only for the sparse
        matrix algebra option, i.e., `laopt` = 'S'.

        `algopt`[28]

          Governs the choice of pivots during the decomposition of the
          first Jacobian matrix. It should lie in the range 0.0 <
          `algopt`[28] < 1.0, with smaller values biasing the algorithm
          towards maintaining sparsity at the expense of numerical
          stability. If `algopt`[28] lies outside the range then the
          default value is used. If the functions regard the Jacobian
          matrix as numerically singular, increasing `algopt`[28]
          towards 1.0 may help, but at the cost of increased fill-in.
          The default value is `algopt`[28] = 0.1.

        `algopt`[29]

          Used as the relative pivot threshold during subsequent
          Jacobian decompositions (see `algopt`[28]) below which an
          internal error is invoked. `algopt`[29] must be greater than
          zero, otherwise the default value is used. If `algopt`[29] is
          greater than 1.0 no check is made on the pivot size, and this
          may be a necessary option if the Jacobian matrix is found to
          be numerically singular (see `algopt`[28]). The default value
          is `algopt`[29] = 0.0001.

    remesh : bool
        Indicates whether or not spatial remeshing should be performed.

        `remesh` = True

          Indicates that spatial remeshing should be performed as
          specified.

        `remesh` = False

          Indicates that spatial remeshing should be suppressed.

        Note: `remesh` should not be changed between consecutive calls
        to ``dim1_parab_convdiff_remesh``. Remeshing can be switched off
        or on at specified times by using appropriate values for the
        arguments `nrmesh` and `trmesh` at each call.

    xfix : float, array-like, shape (nxfix)
        `xfix`[i-1], for i = 1,2,...,nxfix, must contain the value of
        the x coordinate at the ith fixed mesh point.

    nrmesh : int
        Specifies the spatial remeshing frequency and criteria for the
        calculation and adoption of a new mesh.

        `nrmesh` < 0

          Indicates that a new mesh is adopted according to the argument
          `dxmesh`. The mesh is tested every |`nrmesh`| timesteps.

        `nrmesh` = 0

          Indicates that remeshing should take place just once at the
          end of the first time step reached when t > `trmesh`.

        `nrmesh` > 0

          Indicates that remeshing will take place every `nrmesh` time
          steps, with no testing using `dxmesh`.

        Note: `nrmesh` may be changed between consecutive calls to
        ``dim1_parab_convdiff_remesh`` to give greater flexibility over
        the times of remeshing.

    dxmesh : float
        Determines whether a new mesh is adopted when `nrmesh` is set
        less than zero. A possible new mesh is calculated at the end of
        every |`nrmesh`| time steps, but is adopted only if

            x_i^new > x_i^old+`dxmesh`*(x_i+1^old-x_i^old)

        or

            x_i^new < x_i^old-`dxmesh`*(x_i^old-x_i-1^old)

        `dxmesh` thus imposes a lower limit on the difference between
        one mesh and the next.

    trmesh : float
        Specifies when remeshing will take place when `nrmesh` is set to
        zero. Remeshing will occur just once at the end of the first
        time step reached when t is greater than `trmesh`.

        Note: `trmesh` may be changed between consecutive calls to
        ``dim1_parab_convdiff_remesh`` to force remeshing at several
        specified times.

    ipminf : int
        The level of trace information regarding the adaptive remeshing.
        Details are directed to the file object associated with the
        advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

        `ipminf` = 0

          No trace information.

        `ipminf` = 1

          Brief summary of mesh characteristics.

        `ipminf` = 2

          More detailed information, including old and new mesh points,
          mesh sizes and monitor function values.

    comm : dict, communication object, modified in place
        Note: this argument will be (re-)initialized when it is an empty
        ``dict`` or under the following condition: `ind` = 0.

        Communication structure.

        On initial entry: need not be set.

    itask : int
        The task to be performed by the ODE integrator.

        `itask` = 1

          Normal computation of output values `u` at t = `tout` (by
          overshooting and interpolating).

        `itask` = 2

          Take one step in the time direction and return.

        `itask` = 3

          Stop at first internal integration point at or beyond t =
          `tout`.

        `itask` = 4

          Normal computation of output values `u` at t = `tout` but
          without overshooting t = t_crit where t_crit is described
          under the argument `algopt`.

        `itask` = 5

          Take one step in the time direction and return, without
          passing t_crit, where t_crit is described under the argument
          `algopt`.

    itrace : int
        The level of trace information required from
        ``dim1_parab_convdiff_remesh`` and the underlying ODE solver.
        `itrace` may take the value -1, 0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages from the PDE solver are printed.

        `itrace` > 0

          Output from the underlying ODE solver is printed. This output
          contains details of Jacobian entries, the nonlinear iteration
          and the time integration during the computation of the ODE
          system.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases. You are advised to set `itrace` = 0, unless you are
        experienced with submodule ``ode``.

    ind : int
        Indicates whether this is a continuation call or a new
        integration.

        `ind` = 0

          Starts or restarts the integration in time.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the arguments `tout`, `nrmesh`
          and `trmesh` may be reset between calls to
          ``dim1_parab_convdiff_remesh``.

    pdedef : None or callable (p, c, d, s, ires) = pdedef(t, x, u, ux,
        v, vdot, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `pdedef` must evaluate the functions P_{i,j}, C_i, D_i and S_i
        which partially define the system of PDEs. P_{i,j} and C_i may
        depend on x, t, U and V; D_i may depend on x, t, U, U_x and V;
        and S_i may depend on x, t, U, V and linearly on V^.. `pdedef`
        is called approximately midway between each pair of mesh points
        in turn by ``dim1_parab_convdiff_remesh``. None may be used for
        `pdedef` for problems in the form (2).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float
            The current value of the space variable x.

        u : float, ndarray, shape (npde)
            `u`[i-1] contains the value of the component U_i(x,t), for i
            = 1,2,...,`npde`.

        ux : float, ndarray, shape (npde)
            `ux`[i-1] contains the value of the component dU_i(x,t)/dx,
            for i = 1,2,...,`npde`.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

            Note: V^._i(t), for i = 1,2,...,`nv`, may only appear
            linearly in S_j, for j = 1,2,...,`npde`.

        ires : int
            Set to -1 or 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        p : float, array-like, shape (npde, npde)
            `p`[i-1,j-1] must be set to the value of P_{i,j}(x,t,U,V),
            for j = 1,2,...,`npde`, for i = 1,2,...,`npde`.

        c : float, array-like, shape (npde)
            `c`[i-1] must be set to the value of C_i(x,t,U,V), for i =
            1,2,...,`npde`.

        d : float, array-like, shape (npde)
            `d`[i-1] must be set to the value of D_i(x,t,U,U_x,V), for i
            = 1,2,...,`npde`.

        s : float, array-like, shape (npde)
            `s`[i-1] must be set to the value of S_i(x,t,U,V,V^.), for i
            = 1,2,...,`npde`.

        ires : int
            Should usually remain unchanged. However, you may set `ires`
            to force the integration function to take certain actions as
            described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_remesh`` returns to the calling
              function with the error indicator set to `errno` = 4.

    odedef : None or callable (r, ires) = odedef(t, v, vdot, xi, ucp,
        ucpx, ucpt, ires, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `odedef` must evaluate the functions R, which define the system
        of ODEs, as given in (4).

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        v : float, ndarray, shape (nv)
            If `nv` > 0, `v`[i-1] contains the value of the component
            V_i(t), for i = 1,2,...,`nv`.

        vdot : float, ndarray, shape (nv)
            If `nv` > 0, `vdot`[i-1] contains the value of component
            V^._i(t), for i = 1,2,...,`nv`.

        xi : float, ndarray, shape (nxi)
            If nxi > 0, `xi`[i-1] contains the ODE/PDE coupling point,
            xi_i, for i = 1,2,...,nxi.

        ucp : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucp`[i-1,j-1] contains the value of U_i(x,t) at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ucpx : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpx`[i-1,j-1] contains the value of
            dU_i(x,t)/dx at the coupling point x = xi_j, for j =
            1,2,...,nxi, for i = 1,2,...,`npde`.

        ucpt : float, ndarray, shape (npde, nxi)
            If nxi > 0, `ucpt`[i-1,j-1] contains the value of dU_i/dt at
            the coupling point x = xi_j, for j = 1,2,...,nxi, for i =
            1,2,...,`npde`.

        ires : int
            The form of R that must be returned in the array `r`.

            `ires` = 1

              Equation (1) must be used.

            `ires` = -1

              Equation (2) must be used.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        r : float, array-like, shape (nv)
            `r`[i-1] must contain the ith component of R, for i =
            1,2,...,`nv`, where R is defined as

                R = L-MV^.-NU_t^*,

            or

                R = -MV^.-NU_t^*.

            The definition of R is determined by the input value of
            `ires`.

        ires : int
            Should usually remain unchanged. However, you may reset
            `ires` to force the integration function to take certain
            actions, as described below:

            `ires` = 2

              Indicates to the integrator that control should be passed
              back immediately to the calling function with the error
              indicator set to `errno` = 6.

            `ires` = 3

              Indicates to the integrator that the current time step
              should be abandoned and a smaller time step used instead.
              You may wish to set `ires` = 3 when a physically
              meaningless input or output value has been generated. If
              you consecutively set `ires` = 3,
              ``dim1_parab_convdiff_remesh`` returns to the calling
              function with the error indicator set to `errno` = 4.

    xratio : float, optional
        An input bound on the adjacent mesh ratio (greater than 1.0 and
        typically in the range 1.5 to 3.0). The remeshing functions will
        attempt to ensure that

            (x_i-x_{i-1})/`xratio` < x_{i+1}-x_i <
            `xratio`*(x_i-x_{i-1}).

    con : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 2.0/(npts-1).

        An input bound on the sub-integral of the monitor function
        F^mon(x) over each space step. The remeshing functions will
        attempt to ensure that

            int_x_i^x_{i+1}F^mon(x)dx <= `con`int_x_1^x_nptsF^mon(x)dx,

        . `con` gives you more control over the mesh distribution, e.g.,
        decreasing `con` allows more clustering. A typical value is
        2.0/(npts-1), but you are encouraged to experiment with
        different values. Its value is not critical and the mesh should
        be qualitatively correct for all values in the range given
        below.

    monitf : None or callable fmon = monitf(t, x, u, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monitf` must supply and evaluate a remesh monitor function to
        indicate the solution behaviour of interest.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            The current mesh. `x`[i-1] contains the value of x_i, for i
            = 1,2,...,npts.

        u : float, ndarray, shape (npde, npts)
            `u`[i-1,j-1] contains the value of U_i(x,t) at x = `x`[j-1]
            and time t, for j = 1,2,...,npts, for i = 1,2,...,`npde`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fmon : float, array-like, shape (npts)
            `fmon`[i-1] must contain the value of the monitor function
            F^mon(x) at mesh point x = `x`[i-1].

    lrsave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['rsave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lrsave` is computed by the
        function.

        If your supplied `lrsave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lrsave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lrsave` set to at least the lower-bound value returned in
        `lrsave_min`.

    lisave : None or int, optional
        When performing a new integration, the size to use for the
        communication array `comm`['isave'].

        Otherwise, the value is not referenced.

        An initial estimate for an adequate `lisave` is computed by the
        function.

        If your supplied `lisave` is None or too small, the estimated
        value will be used instead.

        In some cases the estimated value will be sufficient for
        continuation calls to the function.

        When `laopt` = 'S', even the function's initial estimated value
        of `lisave` may be too small.

        If so, the function returns with `errno` = 15.

        You are advised to call the function again with `ind` = 0 and
        `lisave` set to at least the lower-bound value returned in
        `lisave_min`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `p` in `pdedef` or `u` in `uvinit` are spiked (i.e., have
        unit extent in all but one dimension, or have size 1),
        `spiked_sorder` selects the storage order to associate with them
        in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    ts : float
        The value of t corresponding to the solution values in `u`.
        Normally `ts` = `tout`.

    u : float, ndarray, shape (neqn)
        The computed solution U_i(x_j,t), for j = 1,2,...,npts, for i =
        1,2,...,`npde`, and V_k(t), for k = 1,2,...,`nv`, evaluated at t
        = `ts`, as follows:

        - `u`[`npde`*(j-1)+i-1] contain U_i(x_j,t), for j =
          1,2,...,npts, for i = 1,2,...,`npde`, and

        - `u`[npts*`npde`+i-1] contain V_i(t), for i = 1,2,...,`nv`.

    x : float, ndarray, shape (npts)
        The final values of the mesh points.

    ind : int
        `ind` = 1.

    lrsave_min : int
        A lower bound on the sufficient size for `comm`['rsave'].

    lisave_min : int
        A lower bound on the sufficient size for `comm`['isave'].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, on initial entry `ind` = 1.

            Constraint: on initial entry `ind` = 0.

        (`errno` 1)
            On entry, `lrsave` = *<value>*.

            Constraint: `lrsave` >= *<value>*.

        (`errno` 1)
            On entry, `lisave` = *<value>*.

            Constraint: `lisave` >= *<value>*.

        (`errno` 1)
            On entry, the point `xfix`[i-1] does not coincide with any
            `x`[j-1]: i = *<value>* and `xfix`[i-1] = *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xfix`[i] = *<value>* and
            `xfix`[i-1] = *<value>*.

            Constraint: `xfix`[i] > `xfix`[i-1].

        (`errno` 1)
            On entry, at least one point in `xi` lies outside
            [`x`[0],`x`[npts-1]]: `x`[0] = *<value>* and `x`[npts-1] =
            *<value>*.

        (`errno` 1)
            On entry, i = *<value>*, `xi`[i] = *<value>* and `xi`[i-1] =
            *<value>*.

            Constraint: `xi`[i] > `xi`[i-1].

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: corresponding elements `atol`[i-1] and
            `rtol`[j-1] cannot both be 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `rtol`[i-1] = *<value>*.

            Constraint: `rtol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `atol`[i-1] = *<value>*.

            Constraint: `atol`[i-1] >= 0.0.

        (`errno` 1)
            On entry, `itol` = *<value>*.

            Constraint: `itol` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, neqn = *<value>*, `npde` = *<value>*, npts =
            *<value>* and `nv` = *<value>*.

            Constraint: neqn = `npde`*npts+`nv`.

        (`errno` 1)
            On entry, `con` = *<value>*, npts = *<value>*.

            Constraint: `con` <= 10.0/(npts-1).

        (`errno` 1)
            On entry, `con` = *<value>*, npts = *<value>*.

            Constraint: `con` >= 0.1/(npts-1).

        (`errno` 1)
            On entry, `xratio` = *<value>*.

            Constraint: `xratio` > 1.0.

        (`errno` 1)
            On entry, `ipminf` = *<value>*.

            Constraint: `ipminf` = 0, 1 or 2.

        (`errno` 1)
            On entry, `dxmesh` = *<value>*.

            Constraint: `dxmesh` >= 0.0.

        (`errno` 1)
            On entry, nxfix = *<value>*, npts = *<value>*.

            Constraint: nxfix <= npts-2.

        (`errno` 1)
            On entry, nxfix = *<value>*.

            Constraint: nxfix >= 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi = 0 when `nv` = 0.

        (`errno` 1)
            On entry, `nv` = *<value>* and nxi = *<value>*.

            Constraint: nxi >= 0 when `nv` > 0.

        (`errno` 1)
            On entry, `npde` = *<value>*.

            Constraint: `npde` >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts >= 3.

        (`errno` 1)
            On entry, `laopt` = *<value>*.

            Constraint: `laopt` = 'F', 'B' or 'S'.

        (`errno` 1)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = '1' or '2'.

        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0 or 1.

        (`errno` 1)
            On entry, `itask` = *<value>*.

            Constraint: `itask` = 1, 2, 3, 4 or 5.

        (`errno` 1)
            On entry, `nv` = *<value>*.

            Constraint: `nv` >= 0.

        (`errno` 1)
            On entry, `tout`-`ts` is too small: `tout` = *<value>* and
            `ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 4)
            In setting up the ODE system an internal auxiliary was
            unable to initialize the derivative. This could be due to
            your setting `ires` = 3 in `pdedef`, `numflx`, `bndary` or
            `odedef`.

        (`errno` 5)
            Singular Jacobian of ODE system. Check problem formulation.

        (`errno` 7)
            `atol` and `rtol` were too small to start integration.

        (`errno` 8)
            `ires` set to an invalid value in a call to functions
            `pdedef`, `numflx`, `bndary` or `odedef`.

        (`errno` 9)
            Serious error in internal call to an auxiliary. Increase
            `itrace` for further details.

        (`errno` 11)
            Error during Jacobian formulation for ODE system. Increase
            `itrace` for further details.

        (`errno` 12)
            In solving ODE system, the maximum number of steps
            `algopt`[14] has been exceeded. `algopt`[14] = *<value>*.

        (`errno` 14)
            The functions P, D, or C appear to depend on time
            derivatives.

        (`errno` 16)
            `remesh` has been changed between calls to
            ``dim1_parab_convdiff_remesh``.

        (`errno` 17)
            `fmon` is negative at one or more mesh points, or zero mesh
            spacing has been obtained due to a poor monitor function.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Underlying ODE solver cannot make further progress from the
            point `ts` with the supplied values of `atol` and `rtol`.
            `ts` = *<value>*.

        (`errno` 3)
            Repeated errors in an attempted step of underlying ODE
            solver. Integration was successful as far as `ts`: `ts` =
            *<value>*.

        (`errno` 6)
            In evaluating residual of ODE system, `ires` = 2 has been
            set in functions `pdedef`, `numflx`, `bndary` or `odedef`.
            Integration is successful as far as `ts`: `ts` = *<value>*.

        (`errno` 10)
            Integration completed, but small changes in `atol` or `rtol`
            are unlikely to result in a changed solution.

        (`errno` 13)
            Zero error weights encountered during time integration.

        (`errno` 15)
            When using the sparse option `lisave` or `lrsave` is too
            small: `lisave` = *<value>*, `lrsave` = *<value>*.
    """

    from ..base.pde import dim1_parab_convdiff_remesh as b_dim1_parab_convdiff_remesh

    fun_name = 'naginterfaces.library.pde.dim1_parab_convdiff_remesh'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npde, t, x, u, ux, nv, v, vdot, p, c, d, s, ires, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _p_pdedef, _c_pdedef, _d_pdedef, _s_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
                data,
            )
        else:
            _p_pdedef, _c_pdedef, _d_pdedef, _s_pdedef, _ires_pdedef = pdedef(
                t,
                x,
                u,
                ux,
                v,
                vdot,
                ires,
            )
        _p_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'p'},
            _p_pdedef,
            copy=False,
            exp_shape=(npde, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'c'},
            _c_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        d[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'd'},
            _d_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        s[:] = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 's'},
            _s_pdedef,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_pdedef_base = _ires_pdedef
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_p_pdedef_np, 'p'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        p[:, :] = _p_pdedef_np
        return _ires_pdedef_base

    cb_numflx_name = fun_name + ':numflx'

    def cb_numflx(npde, t, x, nv, v, uleft, uright, flux, ires, data=None):
        """Wrapping user's numflx. Called as base numflx."""
        if data is not None:
            _flux_numflx, _ires_numflx = numflx(
                t,
                x,
                v,
                uleft,
                uright,
                ires,
                data,
            )
        else:
            _flux_numflx, _ires_numflx = numflx(
                t,
                x,
                v,
                uleft,
                uright,
                ires,
            )
        flux[:] = _utils._to_ndarray(
            {'fun_name': cb_numflx_name, 'returned': 'flux'},
            _flux_numflx,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_numflx_base = _ires_numflx
        return _ires_numflx_base

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npde, npts, t, x, u, nv, v, vdot, ibnd, g, ires, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _g_bndary, _ires_bndary = bndary(
                t,
                x,
                u,
                v,
                vdot,
                ibnd,
                ires,
                data,
            )
        else:
            _g_bndary, _ires_bndary = bndary(
                t,
                x,
                u,
                v,
                vdot,
                ibnd,
                ires,
            )
        g[:] = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'g'},
            _g_bndary,
            copy=False,
            exp_shape=(npde,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_bndary_base = _ires_bndary
        return _ires_bndary_base

    cb_uvinit_name = fun_name + ':uvinit'

    def cb_uvinit(npde, npts, nxi, x, xi, u, nv, v, data=None):
        """Wrapping user's uvinit. Called as base uvinit."""
        if data is not None:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                xi,
                nv,
                data,
            )
        else:
            _u_uvinit, _v_uvinit = uvinit(
                npde,
                x,
                xi,
                nv,
            )
        _u_uvinit_np = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'u'},
            _u_uvinit,
            copy=False,
            exp_shape=(npde, npts),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        v[:] = _utils._to_ndarray(
            {'fun_name': cb_uvinit_name, 'returned': 'v'},
            _v_uvinit,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_uvinit_name},
            ((_u_uvinit_np, 'u'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        u[:, :] = _u_uvinit_np

    cb_odedef_name = fun_name + ':odedef'

    def cb_odedef(npde, t, nv, v, vdot, nxi, xi, ucp, ucpx, ucpt, r, ires, data=None):
        """Wrapping user's odedef. Called as base odedef."""
        if data is not None:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
                data,
            )
        else:
            _r_odedef, _ires_odedef = odedef(
                t,
                v,
                vdot,
                xi,
                ucp,
                ucpx,
                ucpt,
                ires,
            )
        r[:] = _utils._to_ndarray(
            {'fun_name': cb_odedef_name, 'returned': 'r'},
            _r_odedef,
            copy=False,
            exp_shape=(nv,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _ires_odedef_base = _ires_odedef
        return _ires_odedef_base

    cb_monitf_name = fun_name + ':monitf'

    def cb_monitf(t, npts, npde, x, u, fmon, data=None):
        """Wrapping user's monitf. Called as base monitf."""
        if data is not None:
            _fmon_monitf = monitf(
                t,
                x,
                u,
                data,
            )
        else:
            _fmon_monitf = monitf(
                t,
                x,
                u,
            )
        fmon[:] = _utils._to_ndarray(
            {'fun_name': cb_monitf_name, 'returned': 'fmon'},
            _fmon_monitf,
            copy=False,
            exp_shape=(npts,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'}, allow_none=True)
    _pdedef = None if pdedef is None else cb_pdedef
    _b_utils._check_callable(numflx, {'fun_name': fun_name, 'entity': 'numflx'})
    _numflx = cb_numflx
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _b_utils._check_callable(uvinit, {'fun_name': fun_name, 'entity': 'uvinit'})
    _uvinit = cb_uvinit
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _neqn_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _neqn = _b_utils._EngineIntScalarType(
        _neqn_py, {'fun_name': fun_name, 'entity': '_neqn'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _b_utils._check_callable(odedef, {'fun_name': fun_name, 'entity': 'odedef'}, allow_none=True)
    _odedef = None if odedef is None else cb_odedef
    _xi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xi'},
        xi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxi_py = _b_utils._EngineFloat64ArrayType.get_shape(_xi, {'fun_name': fun_name, 'entity': 'xi'}, exp_rank=1)[0]
    _nxi = _b_utils._EngineIntScalarType(
        _nxi_py, {'fun_name': fun_name, 'entity': '_nxi'},
    ).to_ctypes()
    _rtol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rtol'},
        rtol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _atol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'atol'},
        atol,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _algopt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'algopt'},
        algopt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xfix = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xfix'},
        xfix,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxfix_py = _b_utils._EngineFloat64ArrayType.get_shape(_xfix, {'fun_name': fun_name, 'entity': 'xfix'}, exp_rank=1)[0]
    _nxfix = _b_utils._EngineIntScalarType(
        _nxfix_py, {'fun_name': fun_name, 'entity': '_nxfix'},
    ).to_ctypes()
    _con = 2.0/(_npts_py-1) if con is None else con
    _b_utils._check_callable(monitf, {'fun_name': fun_name, 'entity': 'monitf'}, allow_none=True)
    _monitf = None if monitf is None else cb_monitf
    try:
        _return_scalars = b_dim1_parab_convdiff_remesh(
            npde,
            ts,
            tout,
            _pdedef,
            _numflx,
            _bndary,
            _uvinit,
            _u,
            _npts,
            _x,
            nv,
            _odedef,
            _nxi,
            _xi,
            _neqn,
            _rtol,
            _atol,
            itol,
            norm,
            laopt,
            _algopt,
            remesh,
            _nxfix,
            _xfix,
            nrmesh,
            dxmesh,
            trmesh,
            ipminf,
            xratio,
            _con,
            _monitf,
            comm,
            lrsave,
            lisave,
            itask,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _u_py = _u
    _x_py = _x
    _ind_py = _return_scalars.ind
    _lrsave_min_py = _return_scalars.lrsave_min
    _lisave_min_py = _return_scalars.lisave_min
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'u', 'x', 'ind', 'lrsave_min', 'lisave_min'],
    )(
        _ts_py,
        _u_py,
        _x_py,
        _ind_py,
        _lrsave_min_py,
        _lisave_min_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim1_parab_euler_roe(uleft, uright, gamma, comm):
    u"""
    Roe's approximate Riemann solver for Euler equations in conservative
    form, for use with ``dim1_parab_convdiff``,
    ``dim1_parab_convdiff_dae`` and ``dim1_parab_convdiff_remesh``.

    ``dim1_parab_euler_roe`` calculates a numerical flux function using
    Roe's Approximate Riemann Solver for the Euler equations in
    conservative form.
    It is designed primarily for use with the upwind discretization
    schemes ``dim1_parab_convdiff``, ``dim1_parab_convdiff_dae`` or
    ``dim1_parab_convdiff_remesh``, but may also be applicable to other
    conservative upwind schemes requiring numerical flux functions.

    For full information please refer to the NAG Library document for
    d03pu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03puf.html

    Parameters
    ----------
    uleft : float, array-like, shape (3)
        `uleft`[i-1] must contain the left value of the component U_i,
        for i = 1,2,...,3. That is, `uleft`[0] must contain the left
        value of rho, `uleft`[1] must contain the left value of m and
        `uleft`[2] must contain the left value of e.

    uright : float, array-like, shape (3)
        `uright`[i-1] must contain the right value of the component U_i,
        for i = 1,2,...,3. That is, `uright`[0] must contain the right
        value of rho, `uright`[1] must contain the right value of m and
        `uright`[2] must contain the right value of e.

    gamma : float
        The ratio of specific heats, gamma.

    comm : dict, communication object
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    flux : float, ndarray, shape (3)
        `flux`[i-1] contains the numerical flux component F^_i, for i =
        1,2,...,3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `gamma` = *<value>*.

            Constraint: `gamma` > 0.0.

        (`errno` 2)
            Right pressure value pr < 0.0: pr = *<value>*.

        (`errno` 2)
            Left pressure value pl < 0.0: pl = *<value>*.

        (`errno` 2)
            On entry, `uright`[0] = *<value>*.

            Constraint: `uright`[0] >= 0.0.

        (`errno` 2)
            On entry, `uleft`[0] = *<value>*.

            Constraint: `uleft`[0] >= 0.0.
    """

    from ..base.pde import dim1_parab_euler_roe as b_dim1_parab_euler_roe

    fun_name = 'naginterfaces.library.pde.dim1_parab_euler_roe'

    _uleft = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uleft'},
        uleft,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _uright = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uright'},
        uright,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _flux = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    b_dim1_parab_euler_roe(
        _uleft,
        _uright,
        gamma,
        _flux,
        comm,
    )
    _flux_py = _flux
    return _flux_py

def dim1_parab_euler_osher(uleft, uright, gamma, path, comm):
    u"""
    Osher's approximate Riemann solver for Euler equations in
    conservative form, for use with ``dim1_parab_convdiff``,
    ``dim1_parab_convdiff_dae`` and ``dim1_parab_convdiff_remesh``.

    ``dim1_parab_euler_osher`` calculates a numerical flux function
    using Osher's Approximate Riemann Solver for the Euler equations in
    conservative form.
    It is designed primarily for use with the upwind discretization
    schemes ``dim1_parab_convdiff``, ``dim1_parab_convdiff_dae`` or
    ``dim1_parab_convdiff_remesh``, but may also be applicable to other
    conservative upwind schemes requiring numerical flux functions.

    For full information please refer to the NAG Library document for
    d03pv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pvf.html

    Parameters
    ----------
    uleft : float, array-like, shape (3)
        `uleft`[i-1] must contain the left value of the component U_i,
        for i = 1,2,...,3. That is, `uleft`[0] must contain the left
        value of rho, `uleft`[1] must contain the left value of m and
        `uleft`[2] must contain the left value of e.

    uright : float, array-like, shape (3)
        `uright`[i-1] must contain the right value of the component U_i,
        for i = 1,2,...,3. That is, `uright`[0] must contain the right
        value of rho, `uright`[1] must contain the right value of m and
        `uright`[2] must contain the right value of e.

    gamma : float
        The ratio of specific heats, gamma.

    path : str, length 1
        The variant of the Osher scheme.

        `path` = 'O'

          Original.

        `path` = 'P'

          Physical.

    comm : dict, communication object
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    flux : float, ndarray, shape (3)
        `flux`[i-1] contains the numerical flux component F^_i, for i =
        1,2,...,3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `gamma` = *<value>*.

            Constraint: `gamma` > 0.0.

        (`errno` 1)
            On entry, `path` = *<value>*.

            Constraint: `path` = 'O' or 'P'.

        (`errno` 2)
            Right pressure value pr < 0.0: pr = *<value>*.

        (`errno` 2)
            Left pressure value pl < 0.0: pl = *<value>*.

        (`errno` 2)
            On entry, `uright`[0] = *<value>*.

            Constraint: `uright`[0] >= 0.0.

        (`errno` 2)
            On entry, `uleft`[0] = *<value>*.

            Constraint: `uleft`[0] >= 0.0.
    """

    from ..base.pde import dim1_parab_euler_osher as b_dim1_parab_euler_osher

    fun_name = 'naginterfaces.library.pde.dim1_parab_euler_osher'

    _uleft = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uleft'},
        uleft,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _uright = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uright'},
        uright,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _flux = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    b_dim1_parab_euler_osher(
        _uleft,
        _uright,
        gamma,
        path,
        _flux,
        comm,
    )
    _flux_py = _flux
    return _flux_py

def dim1_parab_euler_hll(uleft, uright, gamma, comm):
    u"""
    Modified HLL Riemann solver for Euler equations in conservative
    form, for use with ``dim1_parab_convdiff``,
    ``dim1_parab_convdiff_dae`` and ``dim1_parab_convdiff_remesh``.

    ``dim1_parab_euler_hll`` calculates a numerical flux function using
    a modified HLL (Harten--Lax--van Leer) Approximate Riemann Solver
    for the Euler equations in conservative form.
    It is designed primarily for use with the upwind discretization
    schemes ``dim1_parab_convdiff``, ``dim1_parab_convdiff_dae`` or
    ``dim1_parab_convdiff_remesh``, but may also be applicable to other
    conservative upwind schemes requiring numerical flux functions.

    For full information please refer to the NAG Library document for
    d03pw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pwf.html

    Parameters
    ----------
    uleft : float, array-like, shape (3)
        `uleft`[i-1] must contain the left value of the component U_i,
        for i = 1,2,...,3. That is, `uleft`[0] must contain the left
        value of rho, `uleft`[1] must contain the left value of m and
        `uleft`[2] must contain the left value of e.

    uright : float, array-like, shape (3)
        `uright`[i-1] must contain the right value of the component U_i,
        for i = 1,2,...,3. That is, `uright`[0] must contain the right
        value of rho, `uright`[1] must contain the right value of m and
        `uright`[2] must contain the right value of e.

    gamma : float
        The ratio of specific heats, gamma.

    comm : dict, communication object
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    flux : float, ndarray, shape (3)
        `flux`[i-1] contains the numerical flux component F^_i, for i =
        1,2,...,3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `gamma` = *<value>*.

            Constraint: `gamma` > 0.0.

        (`errno` 2)
            Right pressure value pr < 0.0: pr = *<value>*.

        (`errno` 2)
            Left pressure value pl < 0.0: pl = *<value>*.

        (`errno` 2)
            On entry, `uright`[0] = *<value>*.

            Constraint: `uright`[0] >= 0.0.

        (`errno` 2)
            On entry, `uleft`[0] = *<value>*.

            Constraint: `uleft`[0] >= 0.0.
    """

    from ..base.pde import dim1_parab_euler_hll as b_dim1_parab_euler_hll

    fun_name = 'naginterfaces.library.pde.dim1_parab_euler_hll'

    _uleft = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uleft'},
        uleft,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _uright = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uright'},
        uright,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _flux = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    b_dim1_parab_euler_hll(
        _uleft,
        _uright,
        gamma,
        _flux,
        comm,
    )
    _flux_py = _flux
    return _flux_py

def dim1_parab_euler_exact(uleft, uright, gamma, tol, niter, comm):
    u"""
    Exact Riemann solver for Euler equations in conservative form, for
    use with ``dim1_parab_convdiff``, ``dim1_parab_convdiff_dae`` and
    ``dim1_parab_convdiff_remesh``.

    ``dim1_parab_euler_exact`` calculates a numerical flux function
    using an Exact Riemann Solver for the Euler equations in
    conservative form.
    It is designed primarily for use with the upwind discretization
    schemes ``dim1_parab_convdiff``, ``dim1_parab_convdiff_dae`` or
    ``dim1_parab_convdiff_remesh``, but may also be applicable to other
    conservative upwind schemes requiring numerical flux functions.

    For full information please refer to the NAG Library document for
    d03px

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pxf.html

    Parameters
    ----------
    uleft : float, array-like, shape (3)
        `uleft`[i-1] must contain the left value of the component U_i,
        for i = 1,2,...,3. That is, `uleft`[0] must contain the left
        value of rho, `uleft`[1] must contain the left value of m and
        `uleft`[2] must contain the left value of e.

    uright : float, array-like, shape (3)
        `uright`[i-1] must contain the right value of the component U_i,
        for i = 1,2,...,3. That is, `uright`[0] must contain the right
        value of rho, `uright`[1] must contain the right value of m and
        `uright`[2] must contain the right value of e.

    gamma : float
        The ratio of specific heats, gamma.

    tol : float
        The tolerance to be used in the Newton--Raphson procedure to
        calculate the pressure. If `tol` is set to zero then the default
        value of 1.0*10^(-6) is used.

    niter : int
        The maximum number of Newton--Raphson iterations allowed. If
        `niter` is set to zero then the default value of 20 is used.

    comm : dict, communication object
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    flux : float, ndarray, shape (3)
        `flux`[i-1] contains the numerical flux component F^_i, for i =
        1,2,...,3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `niter` = *<value>*.

            Constraint: `niter` >= 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `gamma` = *<value>*.

            Constraint: `gamma` > 0.0.

        (`errno` 2)
            Right pressure value pr < 0.0: pr = *<value>*.

        (`errno` 2)
            Left pressure value pl < 0.0: pl = *<value>*.

        (`errno` 2)
            On entry, `uright`[0] = *<value>*.

            Constraint: `uright`[0] >= 0.0.

        (`errno` 2)
            On entry, `uleft`[0] = *<value>*.

            Constraint: `uleft`[0] >= 0.0.

        (`errno` 3)
            A vacuum condition has been detected.

        (`errno` 4)
            Newton--Raphson iteration failed to converge.
    """

    from ..base.pde import dim1_parab_euler_exact as b_dim1_parab_euler_exact

    fun_name = 'naginterfaces.library.pde.dim1_parab_euler_exact'

    _uleft = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uleft'},
        uleft,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _uright = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'uright'},
        uright,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _flux = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    b_dim1_parab_euler_exact(
        _uleft,
        _uright,
        gamma,
        tol,
        niter,
        _flux,
        comm,
    )
    _flux_py = _flux
    return _flux_py

def dim1_parab_coll_interp(u, xbkpts, npoly, xp, itype, comm):
    u"""
    PDEs, spatial interpolation with ``dim1_parab_coll`` or
    ``dim1_parab_dae_coll``.

    ``dim1_parab_coll_interp`` may be used in conjunction with either
    ``dim1_parab_coll`` or ``dim1_parab_dae_coll``.
    It computes the solution and its first derivative at user-specified
    points in the spatial coordinate.

    For full information please refer to the NAG Library document for
    d03py

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pyf.html

    Parameters
    ----------
    u : float, array-like, shape (npde, npts)
        The PDE part of the original solution returned in the argument
        `u` by the function ``dim1_parab_coll`` or
        ``dim1_parab_dae_coll``.

    xbkpts : float, array-like, shape (nbkpts)
        `xbkpts`[i-1], for i = 1,2,...,nbkpts, must contain the
        break-points as used by ``dim1_parab_coll`` or
        ``dim1_parab_dae_coll``.

    npoly : int
        The degree of the Chebyshev polynomial used for approximation as
        used by ``dim1_parab_coll`` or ``dim1_parab_dae_coll``.

    xp : float, array-like, shape (intpts)
        `xp`[i-1], for i = 1,2,...,intpts, must contain the spatial
        interpolation points.

    itype : int
        Specifies the interpolation to be performed.

        `itype` = 1

          The solution at the interpolation points are computed.

        `itype` = 2

          Both the solution and the first derivative at the
          interpolation points are computed.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by prior calls to
        ``dim1_parab_coll`` or ``dim1_parab_dae_coll``.

    Returns
    -------
    up : float, ndarray, shape (npde, intpts, `itype`)
        If `itype` = 1, `up`[i-1,j-1,0], contains the value of the
        solution U_i(x_j,t_out), at the interpolation points x_j =
        `xp`[j-1], for i = 1,2,...,npde, for j = 1,2,...,intpts.

        If `itype` = 2, `up`[i-1,j-1,0] contains U_i(x_j,t_out) and
        `up`[i-1,j-1,1] contains dU_i/dx at these points.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1 or 2.

        (`errno` 1)
            On entry, i = *<value>*, `xbkpts`[i-1] = *<value>*, j =
            *<value>* and `xbkpts`[j-1] = *<value>*.

            Constraint: `xbkpts`[0] < `xbkpts`[1] < ... <
            `xbkpts`[nbkpts-1].

        (`errno` 1)
            On entry, npts = *<value>*, nbkpts = *<value>* and `npoly` =
            *<value>*.

            Constraint: npts = (nbkpts-1)*`npoly`+1.

        (`errno` 1)
            On entry, intpts = *<value>*.

            Constraint: intpts >= 1.

        (`errno` 1)
            On entry, `npoly` = *<value>*.

            Constraint: `npoly` > 0.

        (`errno` 1)
            On entry, nbkpts = *<value>*.

            Constraint: nbkpts >= 2.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde > 0.

        (`errno` 2)
            On entry, `itype` = 2 and at least one interpolation point
            coincides with a break-point, i.e., interpolation point no
            *<value>* with value *<value>* is close to break-point
            *<value>* with value *<value>*.

        (`errno` 2)
            On entry, i = *<value>*, `xp`[i-1] = *<value>*, j =
            *<value>* and `xp`[j-1] = *<value>*.

            Constraint: `xbkpts`[0] <= `xp`[0] < `xp`[1] < ... <
            `xp`[intpts-1] <= `xbkpts`[nbkpts-1].

        (`errno` 3)
            Extrapolation is not allowed.
    """

    from ..base.pde import dim1_parab_coll_interp as b_dim1_parab_coll_interp

    fun_name = 'naginterfaces.library.pde.dim1_parab_coll_interp'

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_u, 'u'),),
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[1]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _xbkpts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xbkpts'},
        xbkpts,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nbkpts_py = _b_utils._EngineFloat64ArrayType.get_shape(_xbkpts, {'fun_name': fun_name, 'entity': 'xbkpts'}, exp_rank=1)[0]
    _nbkpts = _b_utils._EngineIntScalarType(
        _nbkpts_py, {'fun_name': fun_name, 'entity': '_nbkpts'},
    ).to_ctypes()
    _xp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xp'},
        xp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _intpts_py = _b_utils._EngineFloat64ArrayType.get_shape(_xp, {'fun_name': fun_name, 'entity': 'xp'}, exp_rank=1)[0]
    _intpts = _b_utils._EngineIntScalarType(
        _intpts_py, {'fun_name': fun_name, 'entity': '_intpts'},
    ).to_ctypes()
    _itype = _b_utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _itype_py = _b_utils._EngineIntScalarType.to_py(_itype)
    _up = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npde_py, _intpts_py, _itype_py),
        sorder=_m_sorder,
    )
    b_dim1_parab_coll_interp(
        _npde,
        _u,
        _nbkpts,
        _xbkpts,
        npoly,
        _npts,
        _xp,
        _intpts,
        _itype,
        _up,
        comm,
    )
    _up_py = _up
    return _up_py

def dim1_parab_fd_interp(m, u, x, xp, itype):
    u"""
    PDEs, spatial interpolation with ``dim1_parab_fd``,
    ``dim1_parab_keller``, ``dim1_parab_convdiff``,
    ``dim1_parab_dae_fd``, ``dim1_parab_dae_keller``,
    ``dim1_parab_convdiff_dae``, ``dim1_parab_remesh_fd``,
    ``dim1_parab_remesh_keller`` or ``dim1_parab_convdiff_remesh``.

    ``dim1_parab_fd_interp`` interpolates in the spatial coordinate the
    solution and derivative of a system of partial differential
    equations (PDEs).
    The solution must first be computed using one of the finite
    difference schemes ``dim1_parab_fd``, ``dim1_parab_dae_fd`` or
    ``dim1_parab_remesh_fd``, or one of the Keller box schemes
    ``dim1_parab_keller``, ``dim1_parab_dae_keller`` or
    ``dim1_parab_remesh_keller``.

    For full information please refer to the NAG Library document for
    d03pz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03pzf.html

    Parameters
    ----------
    m : int
        The coordinate system used. If the call to
        ``dim1_parab_fd_interp`` follows one of the finite difference
        functions then `m` must be the same argument `m` as used in that
        call. For the Keller box scheme only Cartesian coordinate
        systems are valid and so `m` must be set to zero. No check will
        be made by ``dim1_parab_fd_interp`` in this case.

        `m` = 0

          Indicates Cartesian coordinates.

        `m` = 1

          Indicates cylindrical polar coordinates.

        `m` = 2

          Indicates spherical polar coordinates.

    u : float, array-like, shape (npde, npts)
        The PDE part of the original solution returned in the argument
        `u` by the PDE function.

    x : float, array-like, shape (npts)
        `x`[i-1], for i = 1,2,...,npts, must contain the mesh points as
        used by the PDE function.

    xp : float, array-like, shape (intpts)
        `xp`[i-1], for i = 1,2,...,intpts, must contain the spatial
        interpolation points.

    itype : int
        Specifies the interpolation to be performed.

        `itype` = 1

          The solutions at the interpolation points are computed.

        `itype` = 2

          Both the solutions and their first derivatives at the
          interpolation points are computed.

    Returns
    -------
    up : float, ndarray, shape (npde, intpts, `itype`)
        If `itype` = 1, `up`[i-1,j-1,0], contains the value of the
        solution U_i(x_j,t_out), at the interpolation points x_j =
        `xp`[j-1], for i = 1,2,...,npde, for j = 1,2,...,intpts.

        If `itype` = 2, `up`[i-1,j-1,0] contains U_i(x_j,t_out) and
        `up`[i-1,j-1,1] contains dU_i/dx at these points.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1 or 2.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` = 0, 1 or 2.

        (`errno` 1)
            On entry, i = *<value>*, `x`[i-1] = *<value>*, j = *<value>*
            and `x`[j-1] = *<value>*.

            Constraint: `x`[0] < `x`[1] < ... < `x`[npts-1].

        (`errno` 1)
            On entry, intpts = *<value>*.

            Constraint: intpts >= 1.

        (`errno` 1)
            On entry, npts = *<value>*.

            Constraint: npts > 2.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde > 0.

        (`errno` 2)
            On entry, i = *<value>*, `xp`[i-1] = *<value>*, j =
            *<value>* and `xp`[j-1] = *<value>*.

            Constraint: `x`[0] <= `xp`[0] < `xp`[1] < ... <
            `xp`[intpts-1] <= `x`[npts-1].

        (`errno` 3)
            On entry, interpolating point *<value>* with the value
            *<value>* is outside the `x` range.
    """

    from ..base.pde import dim1_parab_fd_interp as b_dim1_parab_fd_interp

    fun_name = 'naginterfaces.library.pde.dim1_parab_fd_interp'

    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_u, 'u'),),
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[0]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _npts_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2)[1]
    _npts = _b_utils._EngineIntScalarType(
        _npts_py, {'fun_name': fun_name, 'entity': '_npts'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xp'},
        xp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _intpts_py = _b_utils._EngineFloat64ArrayType.get_shape(_xp, {'fun_name': fun_name, 'entity': 'xp'}, exp_rank=1)[0]
    _intpts = _b_utils._EngineIntScalarType(
        _intpts_py, {'fun_name': fun_name, 'entity': '_intpts'},
    ).to_ctypes()
    _itype = _b_utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _itype_py = _b_utils._EngineIntScalarType.to_py(_itype)
    _up = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npde_py, _intpts_py, _itype_py),
        sorder=_m_sorder,
    )
    b_dim1_parab_fd_interp(
        _npde,
        m,
        _u,
        _npts,
        _x,
        _xp,
        _intpts,
        _itype,
        _up,
    )
    _up_py = _up
    return _up_py

def dim2_gen_order2_rectangle(ts, tout, dt, xmin, xmax, ymin, ymax, nx, ny, tols, tolt, pdedef, bndary, pdeiv, monitr, opti, optr, comm, itrace, ind, lenrwk=None, leniwk=None, lenlwk=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of second-order PDEs, method of lines, finite
    differences, remeshing, two space variables, rectangular region.

    ``dim2_gen_order2_rectangle`` integrates a system of linear or
    nonlinear, time-dependent partial differential equations (PDEs) in
    two space dimensions on a rectangular domain.
    The method of lines is employed to reduce the PDEs to a system of
    ordinary differential equations (ODEs) which are solved using a
    backward differentiation formula (BDF) method.
    The resulting system of nonlinear equations is solved using a
    modified Newton method and a BI-CGSTAB iterative linear solver with
    ILU preconditioning.
    Local uniform grid refinement is used to improve the accuracy of the
    solution. ``dim2_gen_order2_rectangle`` originates from the VLUGR2
    package.

    For full information please refer to the NAG Library document for
    d03ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03raf.html

    Parameters
    ----------
    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    dt : float, array-like, shape (3)
        The initial, minimum and maximum time step sizes respectively.

        `dt`[0]

          Specifies the initial time step size to be used on the first
          entry, i.e., when `ind` = 0. If `dt`[0] = 0.0 then the default
          value `dt`[0] = 0.01*(`tout`-`ts`) is used. On subsequent
          entries (`ind` = 1), the value of `dt`[0] is not referenced.

        `dt`[1]

          Specifies the minimum time step size to be attempted by the
          integrator. If `dt`[1] = 0.0 the default value `dt`[1] =
          10.0*machine precision is used.

        `dt`[2]

          Specifies the maximum time step size to be attempted by the
          integrator. If `dt`[2] = 0.0 the default value `dt`[2] =
          `tout`-`ts` is used.

    xmin : float
        The extents of the rectangular domain in the x-direction, i.e.,
        the x coordinates of the left and right boundaries respectively.

    xmax : float
        The extents of the rectangular domain in the x-direction, i.e.,
        the x coordinates of the left and right boundaries respectively.

    ymin : float
        The extents of the rectangular domain in the y-direction, i.e.,
        the y coordinates of the lower and upper boundaries
        respectively.

    ymax : float
        The extents of the rectangular domain in the y-direction, i.e.,
        the y coordinates of the lower and upper boundaries
        respectively.

    nx : int
        The number of grid points in the x-direction (including the
        boundary points).

    ny : int
        The number of grid points in the y-direction (including the
        boundary points).

    tols : float
        The space tolerance used in the grid refinement strategy (sigma
        in equation (4)). See [Further Comments].

    tolt : float
        The time tolerance used to determine the time step size (tau in
        equation (7)). See [Further Comments].

    pdedef : callable res = pdedef(t, x, y, u, ut, ux, uy, uxx, uxy,
        uyy, data=None)
        `pdedef` must evaluate the functions F_j, for j =
        1,2,...,`npde`, in equation (1) which define the system of PDEs
        at all interior points of the domain.

        Values at points on the boundaries of the domain are ignored and
        will be overwritten by `bndary`. `pdedef` is called for each
        subgrid in turn.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            `x`[i-1] contains the x coordinate of the ith grid point,
            for i = 1,2,...,npts.

        y : float, ndarray, shape (npts)
            `y`[i-1] contains the y coordinate of the ith grid point,
            for i = 1,2,...,npts.

        u : float, ndarray, shape (npts, npde)
            `u`[i-1,j-1] contains the value of the jth PDE component at
            the ith grid point, for j = 1,2,...,`npde`, for i =
            1,2,...,npts.

        ut : float, ndarray, shape (npts, npde)
            `ut`[i-1,j-1] contains the value of du/dt for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,npts.

        ux : float, ndarray, shape (npts, npde)
            `ux`[i-1,j-1] contains the value of du/dx for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,npts.

        uy : float, ndarray, shape (npts, npde)
            `uy`[i-1,j-1] contains the value of du/dy for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,npts.

        uxx : float, ndarray, shape (npts, npde)
            `uxx`[i-1,j-1] contains the value of d^2u/dx^2 for the jth
            PDE component at the ith grid point, for j = 1,2,...,`npde`,
            for i = 1,2,...,npts.

        uxy : float, ndarray, shape (npts, npde)
            `uxy`[i-1,j-1] contains the value of d^2u/(dxdy) for the jth
            PDE component at the ith grid point, for j = 1,2,...,`npde`,
            for i = 1,2,...,npts.

        uyy : float, ndarray, shape (npts, npde)
            `uyy`[i-1,j-1] contains the value of d^2u/dy^2 for the jth
            PDE component at the ith grid point, for j = 1,2,...,`npde`,
            for i = 1,2,...,npts.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (npts, npde)
            `res`[i-1,j-1] must contain the value of F_j, for j =
            1,2,...,`npde`, at the ith grid point, for i = 1,2,...,npts,
            although the residuals at boundary points will be ignored
            (and overwritten later on) and so they need not be specified
            here.

    bndary : callable res = bndary(t, x, y, u, ut, ux, uy, lbnd, res,
        data=None)
        `bndary` must evaluate the functions G_j, for j =
        1,2,...,`npde`, in equation (2) which define the boundary
        conditions at all boundary points of the domain.

        Residuals at interior points must not be altered by this
        function.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            `x`[i-1] contains the x coordinate of the ith grid point,
            for i = 1,2,...,npts.

        y : float, ndarray, shape (npts)
            `y`[i-1] contains the y coordinate of the ith grid point,
            for i = 1,2,...,npts.

        u : float, ndarray, shape (npts, npde)
            `u`[i-1,j-1] contains the value of the jth PDE component at
            the ith grid point, for j = 1,2,...,`npde`, for i =
            1,2,...,npts.

        ut : float, ndarray, shape (npts, npde)
            `ut`[i-1,j-1] contains the value of du/dt for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,npts.

        ux : float, ndarray, shape (npts, npde)
            `ux`[i-1,j-1] contains the value of du/dx for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,npts.

        uy : float, ndarray, shape (npts, npde)
            `uy`[i-1,j-1] contains the value of du/dy for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,npts.

        lbnd : int, ndarray, shape (nbpts)
            `lbnd`[i-1] contains the grid index for the ith boundary
            point, for i = 1,2,...,nbpts. Hence the ith boundary point
            has coordinates `x`[`lbnd`[i-1]-1] and `y`[`lbnd`[i-1]-1],
            and the corresponding solution values are
            `u`[`lbnd`[i-1]-1,`npde`-1], etc.

        res : float, ndarray, shape (npts, npde)
            `res`[i-1,j-1] contains the value of F_j, for i =
            1,2,...,`npde`, at the ith grid point, for i = 1,2,...,npts,
            as returned by `pdedef`. The residuals at the boundary
            points will be overwritten and so need not have been set by
            `pdedef`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (npts, npde)
            `res`[`lbnd`[i-1]-1,j-1] must contain the value of G_j, for
            j = 1,2,...,`npde`, at the ith boundary point, for i =
            1,2,...,nbpts.

            Note: elements of `res` corresponding to interior points
            must not be altered.

    pdeiv : callable u = pdeiv(npde, t, x, y, data=None)
        `pdeiv` must specify the initial values of the PDE components u
        at all points in the grid. `pdeiv` is not referenced if, on
        entry, `ind` = 1.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        t : float
            The (initial) value of the independent variable t.

        x : float, ndarray, shape (npts)
            `x`[i-1] contains the x coordinate of the ith grid point,
            for i = 1,2,...,npts.

        y : float, ndarray, shape (npts)
            `y`[i-1] contains the y coordinate of the ith grid point,
            for i = 1,2,...,npts.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float, array-like, shape (npts, `npde`)
            `u`[i-1,j-1] must contain the value of the jth PDE component
            at the ith grid point, for j = 1,2,...,`npde`, for i =
            1,2,...,npts.

    monitr : callable ierr = monitr(npde, t, dt, dtnew, tlast, ngpts,
        xpts, ypts, lsol, sol, ierr, data=None)
        `monitr` is called by ``dim2_gen_order2_rectangle`` at the end
        of every successful time step, and may be used to examine or
        print the solution or perform other tasks such as error
        calculations, particularly at the final time step, indicated by
        the argument `tlast`.

        The input arguments contain information about the grid and
        solution at all grid levels used.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        t : float
            The current value of the independent variable t, i.e., the
            time at the end of the integration step just completed.

        dt : float
            The current time step size Delta t, i.e., the time step size
            used for the integration step just completed.

        dtnew : float
            The step size that will be used for the next time step.

        tlast : bool
            Indicates if intermediate or final time step. `tlast` =
            False for an intermediate step, `tlast` = True for the last
            call to `monitr` before returning to your program.

        ngpts : int, ndarray, shape (nlev)
            `ngpts`[l-1] contains the number of grid points at level l,
            for l = 1,2,...,nlev.

        xpts : float, ndarray, shape (sum(`ngpts`))
            Contains the x coordinates of the grid points in each level
            in turn, i.e., `x`[i-1], for l = 1,2,...,nlev, for i =
            1,2,...,`ngpts`[l-1].

            So for level l, `x`[i-1] = `xpts`[k+i-1], where k =
            `ngpts`[0]+`ngpts`[1]+...+`ngpts`[l-2], for l =
            1,2,...,nlev, for i = 1,2,...,`ngpts`[l-1].

        ypts : float, ndarray, shape (sum(`ngpts`))
            Contains the y coordinates of the grid points in each level
            in turn, i.e., `y`[i-1], for l = 1,2,...,nlev, for i =
            1,2,...,`ngpts`[l-1].

            So for level l, `y`[i-1] = `ypts`[k+i-1], where k =
            `ngpts`[0]+`ngpts`[1]+...+`ngpts`[l-2], for l =
            1,2,...,nlev, for i = 1,2,...,`ngpts`[l-1].

        lsol : int, ndarray, shape (nlev)
            `lsol`[l-1] contains the pointer to the solution in `sol` at
            grid level l and time `t`. (`lsol`[l-1] actually contains
            the array index immediately preceding the start of the
            solution in `sol`.)

        sol : float, ndarray, shape (:)
            Contains the solution `u`[`ngpts`[l-1]-1,`npde`-1] at time
            `t` for each grid level l in turn, positioned according to
            `lsol`, i.e., for level l, `u`[i-1,j-1] =
            `sol`[`lsol`[l-1]+(j-1)*`ngpts`[l-1]+i-1], for l =
            1,2,...,nlev, for j = 1,2,...,`npde`, for i =
            1,2,...,`ngpts`[l-1].

        ierr : int
            Will be set to 0.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        ierr : int
            Should be set to 1 to force a tidy termination and an
            immediate return to the calling program with `errno` = 4.
            `ierr` should remain unchanged otherwise.

    opti : int, array-like, shape (4)
        May be set to control various options available in the
        integrator.

        `opti`[0] = 0

          All the default options are employed.

        `opti`[0] > 0

          The default value of `opti`[i-1], for i = 2,3,...,4, can be
          obtained by setting `opti`[i-1] = 0.

        `opti`[0]

          Specifies the maximum number of grid levels allowed (including
          the base grid). `opti`[0] >= 0. The default value is `opti`[0]
          = 3.

        `opti`[1]

          Specifies the maximum number of Jacobian evaluations allowed
          during each nonlinear equations solution. `opti`[1] >= 0. The
          default value is `opti`[1] = 2.

        `opti`[2]

          Specifies the maximum number of Newton iterations in each
          nonlinear equations solution. `opti`[2] >= 0. The default
          value is `opti`[2] = 10.

        `opti`[3]

          Specifies the maximum number of iterations in each linear
          equations solution. `opti`[3] >= 0. The default value is
          `opti`[3] = 100.

    optr : float, array-like, shape (3, npde)
        May be used to specify the optional vectors u^max, w^s and w^t
        in the space and time monitors (see [Further Comments]).

        If an optional vector is not required then all its components
        should be set to 1.0.

        `optr`[0,j-1], for j = 1,2,...,npde, specifies u_j^max, the
        approximate maximum absolute value of the jth component of u, as
        used in (4) and (7). `optr`[0,j-1] > 0.0, for j = 1,2,...,npde.

        `optr`[1,j-1], for j = 1,2,...,npde, specifies w_j^s, the
        weighting factors used in the space monitor (see (4)) to
        indicate the relative importance of the jth component of u on
        the space monitor. `optr`[1,j-1] >= 0.0, for j = 1,2,...,npde.

        `optr`[2,j-1], for j = 1,2,...,npde, specifies w_j^t, the
        weighting factors used in the time monitor (see (6)) to indicate
        the relative importance of the jth component of u on the time
        monitor. `optr`[2,j-1] >= 0.0, for j = 1,2,...,npde.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    itrace : int
        The level of trace information required from
        ``dim2_gen_order2_rectangle``. `itrace` may take the value -1,
        0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed.

        `itrace` > 0

          Output from the underlying solver is printed. This output
          contains details of the time integration, the nonlinear
          iteration and the linear solver.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases.

        Setting `itrace` = 1 allows you to monitor the progress of the
        integration without possibly excessive information.

    ind : int
        Must be set to 0 or 1, alternatively 10 or 11.

        `ind` = 0

          Starts the integration in time. `pdedef` is assumed to be
          serial.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the following parameters may be
          reset between calls to ``dim2_gen_order2_rectangle``: tout,
          dt, tols, tolt, opti, optr, itrace and errno. `pdedef` is
          assumed to be serial.

        `ind` = 10

          Starts the integration in time. `pdedef` is assumed to have
          been parallelized by you, as described in [Parallelism and
          Performance]. In all other respects, this is equivalent to
          `ind` = 0.

        `ind` = 11

          Continues the integration after an earlier exit from the
          function. In this case, only the following parameters may be
          reset between calls to ``dim2_gen_order2_rectangle``: tout,
          dt, tols, tolt, opti, optr, itrace and errno. `pdedef` is
          assumed to have been parallelized by you, as described in
          [Parallelism and Performance]. In all other respects, this is
          equivalent to `ind` = 1.

    lenrwk : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows:
        maxpts*`npde`*(5*maxlev+18*`npde`+9)+2*maxpts.

        The required value of `lenrwk` cannot be determined exactly in
        advance, but a suggested value is

            `lenrwk` = maxpts*npde*(5*l+18*npde+9)+2*maxpts,

        where l = `opti`[0] if `opti`[0] != 0 and l = 3 otherwise, and
        maxpts is the expected maximum number of grid points at any one
        level.

        If during the execution the supplied value is found to be too
        small then the function returns with `errno` = 3 and an
        estimated required size is printed.

    leniwk : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: maxpts*(14+5*maxlev)+7+2.

        The required value of `leniwk` cannot be determined exactly in
        advance, but a suggested value is

            `leniwk` = maxpts*(14+5*m)+7*m+2,

        where maxpts is the expected maximum number of grid points at
        any one level and m = `opti`[0] if `opti`[0] > 0 and m = 3
        otherwise.

        If during the execution the supplied value is found to be too
        small then the function returns with `errno` = 3 and an
        estimated required size is printed.

    lenlwk : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: maxpts+1.

        The required value of `lenlwk` cannot be determined exactly in
        advanced, but a suggested value is

            `lenlwk` = maxpts+1,

        where maxpts is the expected maximum number of grid points at
        any one level.

        If during the execution the supplied value is found to be too
        small then the function returns with `errno` = 3 and an
        estimated required size is printed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `optr` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    ts : float
        The value of t which has been reached. Normally `ts` = `tout`.

    dt : float, ndarray, shape (3)
        `dt`[0] contains the time step size for the next time step.
        `dt`[1] and `dt`[2] are unchanged or set to their default values
        if zero on entry.

    ind : int
        `ind` = 1, if `ind` on input was 0 or 1, or `ind` = 11, if `ind`
        on input was 10 or 11.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ind` is not equal to 0 or 1: `ind` = *<value>*.

        (`errno` 1)
            On entry, `lenlwk` too small for initial grid level:
            `lenlwk` = *<value>*, minimum value = *<value>*. Note that
            subsequent levels will require more. Consult document.

        (`errno` 1)
            On entry, `leniwk` too small for initial grid level:
            `leniwk` = *<value>*, minimum value = *<value>*. Note that
            subsequent levels will require more. Consult document.

        (`errno` 1)
            On entry, `lenrwk` too small for initial grid level:
            `lenrwk` = *<value>*, minimum value = *<value>*. Note that
            subsequent levels will require more. Consult document.

        (`errno` 1)
            On entry, `ymax` too close to `ymin`: `ymax` = *<value>* and
            `ymin` = *<value>*.

        (`errno` 1)
            On entry, `xmax` too close to `xmin`: `xmax` = *<value>* and
            `xmin` = *<value>*.

        (`errno` 1)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymax` > `ymin`.

        (`errno` 1)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmax` > `xmin`.

        (`errno` 1)
            On entry, `ny` = *<value>*.

            Constraint: `ny` >= 4.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 4.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde >= 1.

        (`errno` 1)
            On entry, `dt`[0] = *<value>* and `dt`[2] = *<value>*. Note
            that `dt`[2] was reset to default if zero on entry.

            Constraint: if `ind` = 0, `dt`[0] <= `dt`[2].

        (`errno` 1)
            On entry, `dt`[0] = *<value>* and `dt`[1] = *<value>*. Note
            that `dt`[1] was reset to default if zero on entry.

            Constraint: if `ind` = 0, `dt`[0] >= `dt`[1].

        (`errno` 1)
            On entry, `ind` = 0 and `dt`[0] too large: `dt`[0] =
            *<value>* and maximum value = *<value>*.

        (`errno` 1)
            On entry, `ind` = 0 and `dt`[0] too small: `dt`[0] =
            *<value>* and minimum value = *<value>*.

        (`errno` 1)
            On entry, `dt`[0] = *<value>*.

            Constraint: if `ind` = 0, `dt`[0] >= 0.0.

        (`errno` 1)
            On entry, `dt`[1] = *<value>* and `dt`[2] = *<value>*.

            Constraint: `dt`[1] <= `dt`[2].

        (`errno` 1)
            On entry, `dt`[2] = *<value>*.

            Constraint: `dt`[2] >= 0.0.

        (`errno` 1)
            On entry, `dt`[1] = *<value>*.

            Constraint: `dt`[1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>*, j = *<value>* and `optr`[i-1,j-1] =
            *<value>*.

            Constraint: `optr`[i-1,j-1] >= 0.0.

        (`errno` 1)
            On entry, j = *<value>* and `optr`[0,j-1] = *<value>*.

            Constraint: `optr`[0,j-1] > 0.0.

        (`errno` 1)
            On entry, `opti`[0] = *<value>*, i = *<value>* and
            `opti`[i-1] = *<value>*.

            Constraint: if `opti`[0] > 0, `opti`[i-1] >= 0.

        (`errno` 1)
            On entry, `opti`[0] = *<value>*.

            Constraint: `opti`[0] >= 0.

        (`errno` 1)
            On entry, `tolt` = *<value>*.

            Constraint: `tolt` > 0.0.

        (`errno` 1)
            On entry, `tols` = *<value>*.

            Constraint: `tols` > 0.0.

        (`errno` 1)
            On entry, `tout`-`ts` too small: `tout`-`ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 2)
            Attempted time-step smaller than specified minimum. Check
            problem formulation in `pdedef`, `bndary` and `pdeiv`. Try
            increasing `itrace` for more information.

        (`errno` 3)
            One or more of the workspace arrays are too small. Try
            increasing `itrace` for more information.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            IERR set to 1 in `monitr`. Integration completed as far as
            `ts`: `ts` = *<value>*.

        (`errno` 5)
            Integration completed, but maximum number of levels too
            small for required accuracy.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_gen_order2_rectangle as b_dim2_gen_order2_rectangle

    fun_name = 'naginterfaces.library.pde.dim2_gen_order2_rectangle'

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npts, npde, t, x, y, u, ut, ux, uy, uxx, uxy, uyy, res, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _res_pdedef = pdedef(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                uxx,
                uxy,
                uyy,
                data,
            )
        else:
            _res_pdedef = pdedef(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                uxx,
                uxy,
                uyy,
            )
        _res_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'res'},
            _res_pdedef,
            copy=False,
            exp_shape=(npts, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_res_pdedef_np, 'res'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        res[:, :] = _res_pdedef_np

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npts, npde, t, x, y, u, ut, ux, uy, nbpts, lbnd, res, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _res_bndary = bndary(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                lbnd,
                res,
                data,
            )
        else:
            _res_bndary = bndary(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                lbnd,
                res,
            )
        _res_bndary_np = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'res'},
            _res_bndary,
            copy=False,
            exp_shape=(npts, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_bndary_name},
            ((_res_bndary_np, 'res'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        res[:, :] = _res_bndary_np

    cb_pdeiv_name = fun_name + ':pdeiv'

    def cb_pdeiv(npts, npde, t, x, y, u, data=None):
        """Wrapping user's pdeiv. Called as base pdeiv."""
        if data is not None:
            _u_pdeiv = pdeiv(
                npde,
                t,
                x,
                y,
                data,
            )
        else:
            _u_pdeiv = pdeiv(
                npde,
                t,
                x,
                y,
            )
        _u_pdeiv_np = _utils._to_ndarray(
            {'fun_name': cb_pdeiv_name, 'returned': 'u'},
            _u_pdeiv,
            copy=False,
            exp_shape=(npts, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_pdeiv_name},
            ((_u_pdeiv_np, 'u'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        u[:, :] = _u_pdeiv_np

    def cb_monitr(npde, t, dt, dtnew, tlast, nlev, ngpts, xpts, ypts, lsol, sol, ierr, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _ierr_monitr = monitr(
                npde,
                t,
                dt,
                dtnew,
                tlast,
                ngpts,
                xpts,
                ypts,
                lsol,
                sol,
                ierr,
                data,
            )
        else:
            _ierr_monitr = monitr(
                npde,
                t,
                dt,
                dtnew,
                tlast,
                ngpts,
                xpts,
                ypts,
                lsol,
                sol,
                ierr,
            )
        _ierr_monitr_base = _ierr_monitr
        return _ierr_monitr_base

    _optr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'optr'},
        optr,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_optr, 'optr'),),
        spiked_sorder=spiked_sorder,
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_optr, {'fun_name': fun_name, 'entity': 'optr'}, exp_rank=2)[1]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _dt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dt'},
        dt,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _b_utils._check_callable(pdeiv, {'fun_name': fun_name, 'entity': 'pdeiv'})
    _pdeiv = cb_pdeiv
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'})
    _monitr = cb_monitr
    _opti = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'opti'},
        opti,
        copy=False,
    )
    _nx = _b_utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = _b_utils._EngineIntScalarType.to_py(_nx)
    _ny = _b_utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _ny_py = _b_utils._EngineIntScalarType.to_py(_ny)
    if _b_utils._EngineIntArrayType.get_element(_opti, {'fun_name': fun_name, 'entity': 'opti'}, (0,)) > 0:
        _vl_maxlev_py = _b_utils._EngineIntArrayType.get_element(_opti, {'fun_name': fun_name, 'entity': 'opti'}, (0,))
    else:
        _vl_maxlev_py = 3
    _vl_maxpts_py = 4**(_vl_maxlev_py-1)*_nx_py*_ny_py
    _lenrwk = _vl_maxpts_py*_npde_py*(5*_vl_maxlev_py+18*_npde_py+9)+2*_vl_maxpts_py if lenrwk is None else lenrwk
    _leniwk = _vl_maxpts_py*(14+5*_vl_maxlev_py)+7+2 if leniwk is None else leniwk
    _lenlwk = _vl_maxpts_py+1 if lenlwk is None else lenlwk
    try:
        _return_scalars = b_dim2_gen_order2_rectangle(
            _npde,
            ts,
            tout,
            _dt,
            xmin,
            xmax,
            ymin,
            ymax,
            _nx,
            _ny,
            tols,
            tolt,
            _pdedef,
            _bndary,
            _pdeiv,
            _monitr,
            _opti,
            _optr,
            comm,
            _lenrwk,
            _leniwk,
            _lenlwk,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _dt_py = _dt
    _ind_py = _return_scalars.ind
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'dt', 'ind'],
    )(
        _ts_py,
        _dt_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim2_gen_order2_rectilinear(ts, tout, dt, tols, tolt, inidom, pdedef, bndary, pdeiv, monitr, opti, optr, comm, itrace, ind, lenrwk=None, leniwk=None, lenlwk=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    General system of second-order PDEs, method of lines, finite
    differences, remeshing, two space variables, rectilinear region.

    ``dim2_gen_order2_rectilinear`` integrates a system of linear or
    nonlinear, time-dependent partial differential equations (PDEs) in
    two space dimensions on a rectilinear domain.
    The method of lines is employed to reduce the PDEs to a system of
    ordinary differential equations (ODEs) which are solved using a
    backward differentiation formula (BDF) method.
    The resulting system of nonlinear equations is solved using a
    modified Newton method and a BI-CGSTAB iterative linear solver with
    ILU preconditioning.
    Local uniform grid refinement is used to improve the accuracy of the
    solution. ``dim2_gen_order2_rectilinear`` originates from the VLUGR2
    package.

    For full information please refer to the NAG Library document for
    d03rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03rbf.html

    Parameters
    ----------
    ts : float
        The initial value of the independent variable t.

    tout : float
        The final value of t to which the integration is to be carried
        out.

    dt : float, array-like, shape (3)
        The initial, minimum and maximum time step sizes respectively.

        `dt`[0]

          Specifies the initial time step size to be used on the first
          entry, i.e., when `ind` = 0. If `dt`[0] = 0.0 then the default
          value `dt`[0] = 0.01*(`tout`-`ts`) is used. On subsequent
          entries (`ind` = 1), the value of `dt`[0] is not referenced.

        `dt`[1]

          Specifies the minimum time step size to be attempted by the
          integrator. If `dt`[1] = 0.0 the default value `dt`[1] =
          10.0*machine precision is used.

        `dt`[2]

          Specifies the maximum time step size to be attempted by the
          integrator. If `dt`[2] = 0.0 the default value `dt`[2] =
          `tout`-`ts` is used.

    tols : float
        The space tolerance used in the grid refinement strategy (sigma
        in equation (4)). See [Further Comments].

    tolt : float
        The time tolerance used to determine the time step size (tau in
        equation (7)). See [Further Comments].

    inidom : callable (xmin, xmax, ymin, ymax, nx, ny, npts, nrows,
        nbnds, nbpts, lrow, irow, icol, llbnd, ilbnd, lbnd, ierr) =
        inidom(maxpts, ierr, data=None)
        `inidom` must specify the base grid in terms of the data
        structure described in [Description]. `inidom` is not referenced
        if, on entry, `ind` = 1.

        Parameters
        ~~~~~~~~~~
        maxpts : int
            The maximum number of base grid points allowed by the
            available workspace.

        ierr : int
            Will be initialized by ``dim2_gen_order2_rectilinear`` to
            some value prior to internal calls to `inidom`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        xmin : float
            The extents of the virtual grid in the x-direction, i.e.,
            the x coordinates of the left and right boundaries
            respectively.

        xmax : float
            The extents of the virtual grid in the x-direction, i.e.,
            the x coordinates of the left and right boundaries
            respectively.

        ymin : float
            The extents of the virtual grid in the y-direction, i.e.,
            the y coordinates of the left and right boundaries
            respectively.

        ymax : float
            The extents of the virtual grid in the y-direction, i.e.,
            the y coordinates of the left and right boundaries
            respectively.

        nx : int
            The number of virtual grid points in the x- and y-direction
            respectively (including the boundary points).

        ny : int
            The number of virtual grid points in the x- and y-direction
            respectively (including the boundary points).

        npts : int
            The total number of points in the base grid. If the required
            number of points is greater than `maxpts` then `inidom` must
            be exited immediately with `ierr` set to -1 to avoid
            overwriting memory.

        nrows : int
            The total number of rows of the virtual grid that contain
            base grid points. This is the maximum base row index.

        nbnds : int
            The total number of physical boundaries and corners in the
            base grid.

        nbpts : int
            The total number of boundary points in the base grid.

        lrow : int, array-like, shape (`maxpts`)
            `lrow`[i-1], for i = 1,2,...,`nrows`, must contain the base
            grid index of the first grid point in base grid row i.

        irow : int, array-like, shape (`maxpts`)
            `irow`[i-1], for i = 1,2,...,`nrows`, must contain the
            virtual row number v_y that corresponds to base grid row i.

        icol : int, array-like, shape (`maxpts`)
            `icol`[i-1], for i = 1,2,...,`npts`, must contain the
            virtual column number v_x that contains base grid point i.

        llbnd : int, array-like, shape (`maxpts`)
            `llbnd`[i-1], for i = 1,2,...,`nbnds`, must contain the
            element of `lbnd` corresponding to the start of the ith
            boundary or corner.

            Note: the order of the boundaries and corners in `llbnd`
            must be first all the boundaries and then all the corners.
            The end points of a boundary must not be included in the
            list of points on that boundary. Also, if a corner is shared
            by two pairs of physical boundaries then it has two types
            and must therefore be treated as two corners.

        ilbnd : int, array-like, shape (`maxpts`)
            `ilbnd`[i-1], for i = 1,2,...,`nbnds`, must contain the type
            of the ith boundary (or corner), as given in [Description].

        lbnd : int, array-like, shape (`maxpts`)
            `lbnd`[i-1], for i = 1,2,...,`nbpts`, must contain the grid
            index of the ith boundary point. The order of the boundaries
            is as specified in `llbnd`, but within this restriction the
            order of the points in `lbnd` is arbitrary.

        ierr : int
            If the required number of grid points is larger than
            `maxpts`, `ierr` must be set to -1 to force a termination of
            the integration and an immediate return to the calling
            program with `errno` = 3. Otherwise, `ierr` should remain
            unchanged.

    pdedef : callable res = pdedef(t, x, y, u, ut, ux, uy, uxx, uxy,
        uyy, data=None)
        `pdedef` must evaluate the functions F_j, for j =
        1,2,...,`npde`, in equation (1) which define the system of PDEs
        at all interior points of the domain.

        Values at points on the boundaries of the domain are ignored and
        will be overwritten by `bndary`. `pdedef` is called for each
        subgrid in turn.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            `x`[i-1] contains the x coordinate of the ith grid point,
            for i = 1,2,...,`npts`.

        y : float, ndarray, shape (npts)
            `y`[i-1] contains the y coordinate of the ith grid point,
            for i = 1,2,...,`npts`.

        u : float, ndarray, shape (npts, npde)
            `u`[i-1,j-1] contains the value of the jth PDE component at
            the ith grid point, for j = 1,2,...,`npde`, for i =
            1,2,...,`npts`.

        ut : float, ndarray, shape (npts, npde)
            `ut`[i-1,j-1] contains the value of du/dt for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,`npts`.

        ux : float, ndarray, shape (npts, npde)
            `ux`[i-1,j-1] contains the value of du/dx for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,`npts`.

        uy : float, ndarray, shape (npts, npde)
            `uy`[i-1,j-1] contains the value of du/dy for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,`npts`.

        uxx : float, ndarray, shape (npts, npde)
            `uxx`[i-1,j-1] contains the value of d^2u/dx^2 for the jth
            PDE component at the ith grid point, for j = 1,2,...,`npde`,
            for i = 1,2,...,`npts`.

        uxy : float, ndarray, shape (npts, npde)
            `uxy`[i-1,j-1] contains the value of d^2u/(dxdy) for the jth
            PDE component at the ith grid point, for j = 1,2,...,`npde`,
            for i = 1,2,...,`npts`.

        uyy : float, ndarray, shape (npts, npde)
            `uyy`[i-1,j-1] contains the value of d^2u/dy^2 for the jth
            PDE component at the ith grid point, for j = 1,2,...,`npde`,
            for i = 1,2,...,`npts`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (npts, npde)
            `res`[i-1,j-1] must contain the value of F_j, for j =
            1,2,...,`npde`, at the ith grid point, for i =
            1,2,...,`npts`, although the residuals at boundary points
            will be ignored (and overwritten later on) and so they need
            not be specified here.

    bndary : callable res = bndary(t, x, y, u, ut, ux, uy, llbnd, ilbnd,
        lbnd, res, data=None)
        `bndary` must evaluate the functions G_j, for j =
        1,2,...,`npde`, in equation (2) which define the boundary
        conditions at all boundary points of the domain.

        Residuals at interior points must not be altered by this
        function.

        Parameters
        ~~~~~~~~~~
        t : float
            The current value of the independent variable t.

        x : float, ndarray, shape (npts)
            `x`[i-1] contains the x coordinate of the ith grid point,
            for i = 1,2,...,`npts`.

        y : float, ndarray, shape (npts)
            `y`[i-1] contains the y coordinate of the ith grid point,
            for i = 1,2,...,`npts`.

        u : float, ndarray, shape (npts, npde)
            `u`[i-1,j-1] contains the value of the jth PDE component at
            the ith grid point, for j = 1,2,...,`npde`, for i =
            1,2,...,`npts`.

        ut : float, ndarray, shape (npts, npde)
            `ut`[i-1,j-1] contains the value of du/dt for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,`npts`.

        ux : float, ndarray, shape (npts, npde)
            `ux`[i-1,j-1] contains the value of du/dx for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,`npts`.

        uy : float, ndarray, shape (npts, npde)
            `uy`[i-1,j-1] contains the value of du/dy for the jth PDE
            component at the ith grid point, for j = 1,2,...,`npde`, for
            i = 1,2,...,`npts`.

        llbnd : int, ndarray, shape (nbnds)
            `llbnd`[i-1], for i = 1,2,...,`nbnds`, contains the element
            of `lbnd` corresponding to the start of the ith boundary (or
            corner).

        ilbnd : int, ndarray, shape (nbnds)
            `ilbnd`[i-1], for i = 1,2,...,`nbnds`, contains the type of
            the ith boundary, as given in [Description].

        lbnd : int, ndarray, shape (nbpts)
            `lbnd`[i-1], contains the grid index of the ith boundary
            point, where the order of the boundaries is as specified in
            `llbnd`. Hence the ith boundary point has coordinates
            `x`[`lbnd`[i-1]-1] and `y`[`lbnd`[i-1]-1], and the
            corresponding solution values are `u`[`lbnd`[i-1]-1,j-1],
            for j = 1,2,...,`npde`, for i = 1,2,...,`nbpts`.

        res : float, ndarray, shape (npts, npde)
            Contains function values returned by `pdedef`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        res : float, array-like, shape (npts, npde)
            `res`[`lbnd`[i-1]-1,j-1] must contain the value of G_j, for
            j = 1,2,...,`npde`, at the ith boundary point, for i =
            1,2,...,`nbpts`.

            Note: elements of `res` corresponding to interior points,
            i.e., points not included in `lbnd`, must not be altered.

    pdeiv : callable u = pdeiv(npde, t, x, y, data=None)
        `pdeiv` must specify the initial values of the PDE components u
        at all points in the base grid. `pdeiv` is not referenced if, on
        entry, `ind` = 1.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        t : float
            The (initial) value of the independent variable t.

        x : float, ndarray, shape (npts)
            `x`[i-1] contains the x coordinate of the ith grid point,
            for i = 1,2,...,`npts`.

        y : float, ndarray, shape (npts)
            `y`[i-1] contains the y coordinate of the ith grid point,
            for i = 1,2,...,`npts`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float, array-like, shape (npts, `npde`)
            `u`[i-1,j-1] must contain the value of the jth PDE component
            at the ith grid point, for j = 1,2,...,`npde`, for i =
            1,2,...,`npts`.

    monitr : callable ierr = monitr(npde, t, dt, dtnew, tlast, xmin,
        ymin, dxb, dyb, lgrid, istruc, lsol, sol, ierr, data=None)
        `monitr` is called by ``dim2_gen_order2_rectilinear`` at the end
        of every successful time step, and may be used to examine or
        print the solution or perform other tasks such as error
        calculations, particularly at the final time step, indicated by
        the argument `tlast`.

        Parameters
        ~~~~~~~~~~
        npde : int
            The number of PDEs in the system.

        t : float
            The current value of the independent variable t, i.e., the
            time at the end of the integration step just completed.

        dt : float
            The current time step size Delta t, i.e., the time step size
            used for the integration step just completed.

        dtnew : float
            The time step size that will be used for the next time step.

        tlast : bool
            Indicates if intermediate or final time step. `tlast` =
            False for an intermediate step, `tlast` = True for the last
            call to `monitr` before returning to your program.

        xmin : float
            The (x,y) coordinates of the lower-left corner of the
            virtual grid.

        ymin : float
            The (x,y) coordinates of the lower-left corner of the
            virtual grid.

        dxb : float
            The sizes of the base grid spacing in the x- and y-direction
            respectively.

        dyb : float
            The sizes of the base grid spacing in the x- and y-direction
            respectively.

        lgrid : int, ndarray, shape (nlev+1)
            Contains pointers to the start of the grid structures in
            `istruc`, and must be passed unchanged to
            ``dim2_gen_order2_rectilinear_extractgrid`` in order to
            extract the grid information.

        istruc : int, ndarray, shape (nistruc)
            Contains the grid structures for each grid level and must be
            passed unchanged to
            ``dim2_gen_order2_rectilinear_extractgrid`` in order to
            extract the grid information.

        lsol : int, ndarray, shape (nlev)
            `lsol`[l-1] contains the pointer to the solution in `sol` at
            grid level l and time `t`. (`lsol`[l-1] actually contains
            the array index immediately preceding the start of the
            solution in `sol`.)

        sol : float, ndarray, shape (nsol)
            Contains the solution u at time `t` for each grid level l in
            turn, positioned according to `lsol`. More precisely

                `u`[i-1,j-1] = `sol`[ `lsol`[l-1] + (j-1) * n_l+i -1]

            represents the jth component of the solution at the ith grid
            point in the lth level, for l = 1,2,...,nlev, for j =
            1,2,...,`npde`, for i = 1,2,...,n_l, where n_l is the number
            of grid points at level l (obtainable by a call to
            ``dim2_gen_order2_rectilinear_extractgrid``).

        ierr : int
            Will be initialized by ``dim2_gen_order2_rectilinear`` to
            some value prior to internal calls to `ierr`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        ierr : int
            Should be set to 1 to force a termination of the integration
            and an immediate return to the calling program with `errno`
            = 4. `ierr` should remain unchanged otherwise.

    opti : int, array-like, shape (4)
        May be set to control various options available in the
        integrator.

        `opti`[0] = 0

          All the default options are employed.

        `opti`[0] > 0

          The default value of `opti`[i-1], for i = 2,3,...,4, can be
          obtained by setting `opti`[i-1] = 0.

        `opti`[0]

          Specifies the maximum number of grid levels allowed (including
          the base grid). `opti`[0] >= 0. The default value is `opti`[0]
          = 3.

        `opti`[1]

          Specifies the maximum number of Jacobian evaluations allowed
          during each nonlinear equations solution. `opti`[1] >= 0. The
          default value is `opti`[1] = 2.

        `opti`[2]

          Specifies the maximum number of Newton iterations in each
          nonlinear equations solution. `opti`[2] >= 0. The default
          value is `opti`[2] = 10.

        `opti`[3]

          Specifies the maximum number of iterations in each linear
          equations solution. `opti`[3] >= 0. The default value is
          `opti`[3] = 100.

    optr : float, array-like, shape (3, npde)
        May be used to specify the optional vectors u^max, w^s and w^t
        in the space and time monitors (see [Further Comments]).

        If an optional vector is not required then all its components
        should be set to 1.0.

        `optr`[0,j-1], for j = 1,2,...,npde, specifies u_j^max, the
        approximate maximum absolute value of the jth component of u, as
        used in (4) and (7). `optr`[0,j-1] > 0.0, for j = 1,2,...,npde.

        `optr`[1,j-1], for j = 1,2,...,npde, specifies w_j^s, the
        weighting factors used in the space monitor (see (4)) to
        indicate the relative importance of the jth component of u on
        the space monitor. `optr`[1,j-1] >= 0.0, for j = 1,2,...,npde.

        `optr`[2,j-1], for j = 1,2,...,npde, specifies w_j^t, the
        weighting factors used in the time monitor (see (6)) to indicate
        the relative importance of the jth component of u on the time
        monitor. `optr`[2,j-1] >= 0.0, for j = 1,2,...,npde.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    itrace : int
        The level of trace information required from
        ``dim2_gen_order2_rectilinear``. `itrace` may take the value -1,
        0, 1, 2 or 3.

        `itrace` = -1

          No output is generated.

        `itrace` = 0

          Only warning messages are printed.

        `itrace` > 0

          Output from the underlying solver is printed. This output
          contains details of the time integration, the nonlinear
          iteration and the linear solver.

        If `itrace` < -1, -1 is assumed and similarly if `itrace` > 3, 3
        is assumed.

        The advisory messages are given in greater detail as `itrace`
        increases.

        Setting `itrace` = 1 allows you to monitor the progress of the
        integration without possibly excessive information.

    ind : int
        Must be set to 0 or 1, alternatively 10 or 11.

        `ind` = 0

          Starts the integration in time. `pdedef` is assumed to be
          serial.

        `ind` = 1

          Continues the integration after an earlier exit from the
          function. In this case, only the following parameters may be
          reset between calls to ``dim2_gen_order2_rectilinear``: tout,
          dt, tols, tolt, opti, optr, itrace and errno. `pdedef` is
          assumed to be serial.

        `ind` = 10

          Starts the integration in time. `pdedef` is assumed to have
          been parallelized by you, as described in [Parallelism and
          Performance]. In all other respects, this is equivalent to
          `ind` = 0.

        `ind` = 11

          Continues the integration after an earlier exit from the
          function. In this case, only the following parameters may be
          reset between calls to ``dim2_gen_order2_rectangle``: tout,
          dt, tols, tolt, opti, optr, itrace and errno. `pdedef` is
          assumed to have been parallelized by you, as described in
          [Parallelism and Performance]. In all other respects, this is
          equivalent to `ind` = 1.

    lenrwk : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows:
        maxpts*`npde`*(5*maxlev+18*`npde`+9)+2*maxpts.

        The required value of `lenrwk` cannot be determined exactly in
        advance, but a suggested value is

            `lenrwk` = maxpts*npde*(5*l+18*npde+9)+2*maxpts,

        where l = `opti`[0] if `opti`[0] != 0 and l = 3 otherwise, and
        maxpts is the expected maximum number of grid points at any one
        level.

        If during the execution the supplied value is found to be too
        small then the function returns with `errno` = 3 and an
        estimated required size is printed on the current error message
        unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        Note: the size of `lenrwk` cannot be checked upon initial entry
        to ``dim2_gen_order2_rectilinear`` since the number of grid
        points on the base grid is not known.

    leniwk : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: maxpts*(14+5*maxlev)+7+2.

        The required value of `leniwk` cannot be determined exactly in
        advance, but a suggested value is

            `leniwk` = maxpts*(14+5*m)+7*m+2,

        where maxpts is the expected maximum number of grid points at
        any one level and m = `opti`[0] if `opti`[0] > 0 and m = 3
        otherwise.

        If during the execution the supplied value is found to be too
        small then the function returns with `errno` = 3 and an
        estimated required size is printed on the current error message
        unit (see :class:`~naginterfaces.base.utils.FileObjManager`).

        Note: the size of `leniwk` cannot be checked upon initial entry
        to ``dim2_gen_order2_rectilinear`` since the number of grid
        points on the base grid is not known.

    lenlwk : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: maxpts+1.

        The required value of `lenlwk` cannot be determined exactly in
        advance, but a suggested value is

            `lenlwk` = maxpts+1,

        where maxpts is the expected maximum number of grid points at
        any one level.

        If during the execution the supplied value is found to be too
        small then the function returns with `errno` = 3 and an
        estimated required size is printed.

        Note: the size of `lenlwk` cannot be checked upon initial entry
        to ``dim2_gen_order2_rectilinear`` since the number of grid
        points on the base grid is not known.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `optr` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    ts : float
        The value of t which has been reached. Normally `ts` = `tout`.

    dt : float, ndarray, shape (3)
        `dt`[0] contains the time step size for the next time step.
        `dt`[1] and `dt`[2] are unchanged or set to their default values
        if zero on entry.

    ind : int
        `ind` = 1, if `ind` on input was 0 or 1, or `ind` = 11, if `ind`
        on input was 10 or 11.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ind` is not equal to 0 or 1: `ind` = *<value>*.

        (`errno` 1)
            On entry, npde = *<value>*.

            Constraint: npde >= 1.

        (`errno` 1)
            On entry, `dt`[0] = *<value>* and `dt`[2] = *<value>*. Note
            that `dt`[2] was reset to default if zero on entry.

            Constraint: if `ind` = 0, `dt`[0] <= `dt`[2].

        (`errno` 1)
            On entry, `dt`[0] = *<value>* and `dt`[1] = *<value>*. Note
            that `dt`[1] was reset to default if zero on entry.

            Constraint: if `ind` = 0, `dt`[0] >= `dt`[1].

        (`errno` 1)
            On entry, `ind` = 0 and `dt`[0] too large: `dt`[0] =
            *<value>* and maximum value = *<value>*.

        (`errno` 1)
            On entry, `ind` = 0 and `dt`[0] too small: `dt`[0] =
            *<value>* and minimum value = *<value>*.

        (`errno` 1)
            On entry, `dt`[0] = *<value>*.

            Constraint: if `ind` = 0, `dt`[0] >= 0.0.

        (`errno` 1)
            On entry, `dt`[1] = *<value>* and `dt`[2] = *<value>*.

            Constraint: `dt`[1] <= `dt`[2].

        (`errno` 1)
            On entry, `dt`[2] = *<value>*.

            Constraint: `dt`[2] >= 0.0.

        (`errno` 1)
            On entry, `dt`[1] = *<value>*.

            Constraint: `dt`[1] >= 0.0.

        (`errno` 1)
            On entry, i = *<value>*, j = *<value>* and `optr`[i-1,j-1] =
            *<value>*.

            Constraint: `optr`[i-1,j-1] >= 0.0.

        (`errno` 1)
            On entry, j = *<value>* and `optr`[0,j-1] = *<value>*.

            Constraint: `optr`[0,j-1] > 0.0.

        (`errno` 1)
            On entry, `opti`[0] = *<value>*, i = *<value>* and
            `opti`[i-1] = *<value>*.

            Constraint: if `opti`[0] > 0, `opti`[i-1] >= 0.

        (`errno` 1)
            On entry, `opti`[0] = *<value>*.

            Constraint: `opti`[0] >= 0.

        (`errno` 1)
            On entry, `tolt` = *<value>*.

            Constraint: `tolt` > 0.0.

        (`errno` 1)
            On entry, `tols` = *<value>*.

            Constraint: `tols` > 0.0.

        (`errno` 1)
            On entry, `tout`-`ts` too small: `tout`-`ts` = *<value>*.

        (`errno` 1)
            On entry, `tout` = *<value>* and `ts` = *<value>*.

            Constraint: `tout` > `ts`.

        (`errno` 2)
            Attempted time-step smaller than specified minimum. Check
            problem formulation in `pdedef`, `bndary` and `pdeiv`. Try
            increasing `itrace` for more information.

        (`errno` 3)
            One or more of the workspace arrays are too small. Try
            increasing `itrace` for more information.

        (`errno` 6)
            Invalid output argument from `inidom`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            IERR set to 1 in `monitr`. Integration completed as far as
            `ts`: `ts` = *<value>*.

        (`errno` 5)
            Integration completed, but maximum number of levels too
            small for required accuracy.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_gen_order2_rectilinear as b_dim2_gen_order2_rectilinear

    fun_name = 'naginterfaces.library.pde.dim2_gen_order2_rectilinear'

    cb_inidom_name = fun_name + ':inidom'

    def cb_inidom(maxpts, lrow, irow, icol, llbnd, ilbnd, lbnd, ierr, data=None):
        """Wrapping user's inidom. Called as base inidom."""
        if data is not None:
            _xmin_inidom, _xmax_inidom, _ymin_inidom, _ymax_inidom, _nx_inidom, _ny_inidom, _npts_inidom, _nrows_inidom, _nbnds_inidom, _nbpts_inidom, _lrow_inidom, _irow_inidom, _icol_inidom, _llbnd_inidom, _ilbnd_inidom, _lbnd_inidom, _ierr_inidom = inidom(
                maxpts,
                ierr,
                data,
            )
        else:
            _xmin_inidom, _xmax_inidom, _ymin_inidom, _ymax_inidom, _nx_inidom, _ny_inidom, _npts_inidom, _nrows_inidom, _nbnds_inidom, _nbpts_inidom, _lrow_inidom, _irow_inidom, _icol_inidom, _llbnd_inidom, _ilbnd_inidom, _lbnd_inidom, _ierr_inidom = inidom(
                maxpts,
                ierr,
            )
        lrow[:] = _utils._to_ndarray(
            {'fun_name': cb_inidom_name, 'returned': 'lrow'},
            _lrow_inidom,
            copy=False,
            exp_shape=(maxpts,),
            exp_class=_b_utils._EngineIntArrayType,
        )
        irow[:] = _utils._to_ndarray(
            {'fun_name': cb_inidom_name, 'returned': 'irow'},
            _irow_inidom,
            copy=False,
            exp_shape=(maxpts,),
            exp_class=_b_utils._EngineIntArrayType,
        )
        icol[:] = _utils._to_ndarray(
            {'fun_name': cb_inidom_name, 'returned': 'icol'},
            _icol_inidom,
            copy=False,
            exp_shape=(maxpts,),
            exp_class=_b_utils._EngineIntArrayType,
        )
        llbnd[:] = _utils._to_ndarray(
            {'fun_name': cb_inidom_name, 'returned': 'llbnd'},
            _llbnd_inidom,
            copy=False,
            exp_shape=(maxpts,),
            exp_class=_b_utils._EngineIntArrayType,
        )
        ilbnd[:] = _utils._to_ndarray(
            {'fun_name': cb_inidom_name, 'returned': 'ilbnd'},
            _ilbnd_inidom,
            copy=False,
            exp_shape=(maxpts,),
            exp_class=_b_utils._EngineIntArrayType,
        )
        lbnd[:] = _utils._to_ndarray(
            {'fun_name': cb_inidom_name, 'returned': 'lbnd'},
            _lbnd_inidom,
            copy=False,
            exp_shape=(maxpts,),
            exp_class=_b_utils._EngineIntArrayType,
        )
        _ierr_inidom_base = _ierr_inidom
        _xmin_inidom_base = _xmin_inidom
        _xmax_inidom_base = _xmax_inidom
        _ymin_inidom_base = _ymin_inidom
        _ymax_inidom_base = _ymax_inidom
        _nx_inidom_base = _nx_inidom
        _ny_inidom_base = _ny_inidom
        _npts_inidom_base = _npts_inidom
        _nrows_inidom_base = _nrows_inidom
        _nbnds_inidom_base = _nbnds_inidom
        _nbpts_inidom_base = _nbpts_inidom
        return (
            _xmin_inidom_base,
            _xmax_inidom_base,
            _ymin_inidom_base,
            _ymax_inidom_base,
            _nx_inidom_base,
            _ny_inidom_base,
            _npts_inidom_base,
            _nrows_inidom_base,
            _nbnds_inidom_base,
            _nbpts_inidom_base,
            _ierr_inidom_base,
        )

    cb_pdedef_name = fun_name + ':pdedef'

    def cb_pdedef(npts, npde, t, x, y, u, ut, ux, uy, uxx, uxy, uyy, res, data=None):
        """Wrapping user's pdedef. Called as base pdedef."""
        if data is not None:
            _res_pdedef = pdedef(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                uxx,
                uxy,
                uyy,
                data,
            )
        else:
            _res_pdedef = pdedef(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                uxx,
                uxy,
                uyy,
            )
        _res_pdedef_np = _utils._to_ndarray(
            {'fun_name': cb_pdedef_name, 'returned': 'res'},
            _res_pdedef,
            copy=False,
            exp_shape=(npts, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_pdedef_name},
            ((_res_pdedef_np, 'res'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        res[:, :] = _res_pdedef_np

    cb_bndary_name = fun_name + ':bndary'

    def cb_bndary(npts, npde, t, x, y, u, ut, ux, uy, nbnds, nbpts, llbnd, ilbnd, lbnd, res, data=None):
        """Wrapping user's bndary. Called as base bndary."""
        if data is not None:
            _res_bndary = bndary(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                llbnd,
                ilbnd,
                lbnd,
                res,
                data,
            )
        else:
            _res_bndary = bndary(
                t,
                x,
                y,
                u,
                ut,
                ux,
                uy,
                llbnd,
                ilbnd,
                lbnd,
                res,
            )
        _res_bndary_np = _utils._to_ndarray(
            {'fun_name': cb_bndary_name, 'returned': 'res'},
            _res_bndary,
            copy=False,
            exp_shape=(npts, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_bndary_name},
            ((_res_bndary_np, 'res'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        res[:, :] = _res_bndary_np

    cb_pdeiv_name = fun_name + ':pdeiv'

    def cb_pdeiv(npts, npde, t, x, y, u, data=None):
        """Wrapping user's pdeiv. Called as base pdeiv."""
        if data is not None:
            _u_pdeiv = pdeiv(
                npde,
                t,
                x,
                y,
                data,
            )
        else:
            _u_pdeiv = pdeiv(
                npde,
                t,
                x,
                y,
            )
        _u_pdeiv_np = _utils._to_ndarray(
            {'fun_name': cb_pdeiv_name, 'returned': 'u'},
            _u_pdeiv,
            copy=False,
            exp_shape=(npts, npde),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_pdeiv_name},
            ((_u_pdeiv_np, 'u'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        u[:, :] = _u_pdeiv_np

    def cb_monitr(npde, t, dt, dtnew, tlast, nlev, xmin, ymin, dxb, dyb, lgrid, istruc, nistruc, lsol, sol, nsol, ierr, data=None):
        """Wrapping user's monitr. Called as base monitr."""
        if data is not None:
            _ierr_monitr = monitr(
                npde,
                t,
                dt,
                dtnew,
                tlast,
                xmin,
                ymin,
                dxb,
                dyb,
                lgrid,
                istruc,
                lsol,
                sol,
                ierr,
                data,
            )
        else:
            _ierr_monitr = monitr(
                npde,
                t,
                dt,
                dtnew,
                tlast,
                xmin,
                ymin,
                dxb,
                dyb,
                lgrid,
                istruc,
                lsol,
                sol,
                ierr,
            )
        _ierr_monitr_base = _ierr_monitr
        return _ierr_monitr_base

    _optr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'optr'},
        optr,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_optr, 'optr'),),
        spiked_sorder=spiked_sorder,
    )
    _npde_py = _b_utils._EngineFloat64ArrayType.get_shape(_optr, {'fun_name': fun_name, 'entity': 'optr'}, exp_rank=2)[1]
    _npde = _b_utils._EngineIntScalarType(
        _npde_py, {'fun_name': fun_name, 'entity': '_npde'},
    ).to_ctypes()
    _dt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dt'},
        dt,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(inidom, {'fun_name': fun_name, 'entity': 'inidom'})
    _inidom = cb_inidom
    _b_utils._check_callable(pdedef, {'fun_name': fun_name, 'entity': 'pdedef'})
    _pdedef = cb_pdedef
    _b_utils._check_callable(bndary, {'fun_name': fun_name, 'entity': 'bndary'})
    _bndary = cb_bndary
    _b_utils._check_callable(pdeiv, {'fun_name': fun_name, 'entity': 'pdeiv'})
    _pdeiv = cb_pdeiv
    _b_utils._check_callable(monitr, {'fun_name': fun_name, 'entity': 'monitr'})
    _monitr = cb_monitr
    _opti = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'opti'},
        opti,
        copy=False,
    )
    if _b_utils._EngineIntArrayType.get_element(_opti, {'fun_name': fun_name, 'entity': 'opti'}, (0,)) > 0:
        _vl_maxlev_py = _b_utils._EngineIntArrayType.get_element(_opti, {'fun_name': fun_name, 'entity': 'opti'}, (0,))
    else:
        _vl_maxlev_py = 3
    _vl_nx_est_py = 100
    _vl_ny_est_py = 100
    _vl_maxpts_py = 4**(_vl_maxlev_py-1)*_vl_nx_est_py*_vl_ny_est_py
    _lenrwk = _vl_maxpts_py*_npde_py*(5*_vl_maxlev_py+18*_npde_py+9)+2*_vl_maxpts_py if lenrwk is None else lenrwk
    _leniwk = _vl_maxpts_py*(14+5*_vl_maxlev_py)+7+2 if leniwk is None else leniwk
    _lenlwk = _vl_maxpts_py+1 if lenlwk is None else lenlwk
    try:
        _return_scalars = b_dim2_gen_order2_rectilinear(
            _npde,
            ts,
            tout,
            _dt,
            tols,
            tolt,
            _inidom,
            _pdedef,
            _bndary,
            _pdeiv,
            _monitr,
            _opti,
            _optr,
            comm,
            _lenrwk,
            _leniwk,
            _lenlwk,
            itrace,
            ind,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ts_py = _return_scalars.ts
    _dt_py = _dt
    _ind_py = _return_scalars.ind
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts', 'dt', 'ind'],
    )(
        _ts_py,
        _dt_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim2_gen_order2_rectilinear_extractgrid(level, xmin, ymin, dxb, dyb, lgrid, istruc, lenxy):
    u"""
    Extract grid data from ``dim2_gen_order2_rectilinear``.

    ``dim2_gen_order2_rectilinear_extractgrid`` is designed to be used
    in conjunction with ``dim2_gen_order2_rectilinear``.
    It can be called from the monitr to obtain the number of grid points
    and their (x,y) coordinates on a solution grid.

    For full information please refer to the NAG Library document for
    d03rz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03rzf.html

    Parameters
    ----------
    level : int
        The grid level at which the coordinates are required.

    xmin : float
        `xmin` as supplied to monitr must be passed unchanged to
        ``dim2_gen_order2_rectilinear_extractgrid``

    ymin : float
        `ymin` as supplied to monitr must be passed unchanged to
        ``dim2_gen_order2_rectilinear_extractgrid``

    dxb : float
        `dxb` as supplied to monitr must be passed unchanged to
        ``dim2_gen_order2_rectilinear_extractgrid``

    dyb : float
        `dyb` as supplied to monitr must be passed unchanged to
        ``dim2_gen_order2_rectilinear_extractgrid``

    lgrid : int, array-like, shape (nlev)
        `lgrid` as supplied to monitr must be passed unchanged to
        ``dim2_gen_order2_rectilinear_extractgrid``.

    istruc : int, array-like, shape (`lgrid`[nlev-1]+2*nrows+npts+1)
        `istruc` as supplied to monitr must be passed unchanged to
        ``dim2_gen_order2_rectilinear_extractgrid``.

    lenxy : int
        The dimension of the arrays `x` and `y`.

    Returns
    -------
    npts : int
        The number of grid points in the grid level `level`.

    x : float, ndarray, shape (`lenxy`)
        `x`[i-1] and `y`[i-1] contain the (x,y) coordinates respectively
        of the ith grid point, for i = 1,2,...,`npts`.

    y : float, ndarray, shape (`lenxy`)
        `x`[i-1] and `y`[i-1] contain the (x,y) coordinates respectively
        of the ith grid point, for i = 1,2,...,`npts`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `level` = *<value>* and nlev = *<value>*.

            Constraint: `level` <= nlev.

        (`errno` 1)
            On entry, `level` = *<value>*.

            Constraint: `level` >= 1.

        (`errno` 2)
            On entry, `lenxy` = *<value>*.

            Constraint: `lenxy` >= *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_gen_order2_rectilinear_extractgrid as b_dim2_gen_order2_rectilinear_extractgrid

    fun_name = 'naginterfaces.library.pde.dim2_gen_order2_rectilinear_extractgrid'

    _lgrid = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lgrid'},
        lgrid,
        copy=False,
    )
    _nlev_py = _b_utils._EngineIntArrayType.get_shape(_lgrid, {'fun_name': fun_name, 'entity': 'lgrid'}, exp_rank=1)[0]
    _nlev = _b_utils._EngineIntScalarType(
        _nlev_py, {'fun_name': fun_name, 'entity': '_nlev'},
    ).to_ctypes()
    _istruc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'istruc'},
        istruc,
        copy=False,
    )
    _lenxy = _b_utils._EngineIntScalarType(
        lenxy, {'fun_name': fun_name, 'entity': 'lenxy'},
    ).to_ctypes()
    _lenxy_py = _b_utils._EngineIntScalarType.to_py(_lenxy)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lenxy_py,),
    )
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lenxy_py,),
    )
    _return_scalar = b_dim2_gen_order2_rectilinear_extractgrid(
        level,
        _nlev,
        xmin,
        ymin,
        dxb,
        dyb,
        _lgrid,
        _istruc,
        _x,
        _y,
        _lenxy,
    )
    _npts_py = _return_scalar
    _x_py = _x
    _y_py = _y
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['npts', 'x', 'y'],
    )(
        _npts_py,
        _x_py,
        _y_py,
    )
    return _return_data

def dim2_ellip_fd_iter(a, b, c, d, e, aparam, it, r):
    u"""
    Elliptic PDE, solution of finite difference equations by SIP,
    five-point two-dimensional molecule, one iteration.

    ``dim2_ellip_fd_iter`` performs at each call one iteration of the
    Strongly Implicit Procedure.
    It is used to calculate on successive calls a sequence of
    approximate corrections to the current estimate of the solution when
    solving a system of simultaneous algebraic equations for which the
    iterative update matrix is of five-point molecule form on a
    two-dimensional topologically-rectangular mesh. ('Topological' means
    that a polar grid (r,theta), for example, can be used as it is
    equivalent to a rectangular box.)

    For full information please refer to the NAG Library document for
    d03ua

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03uaf.html

    Parameters
    ----------
    a : float, array-like, shape (n1, n2)
        `a`[i-1,j-1] must contain the coefficient of the 'southerly'
        term involving s_{i,j-1} in the (i,j)th equation of the system
        (2), for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of
        `a`, for j = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        rectangle.

    b : float, array-like, shape (n1, n2)
        `b`[i-1,j-1] must contain the coefficient of the 'westerly' term
        involving s_{i-1,j} in the (i,j)th equation of the system (2),
        for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of `b`, for
        i = 1, must be zero after incorporating the boundary conditions,
        since they involve nodal values from outside the rectangle.

    c : float, array-like, shape (n1, n2)
        `c`[i-1,j-1] must contain the coefficient of the 'central' term
        involving s_{ij} in the (i,j)th equation of the system (2), for
        j = 1,2,...,n2, for i = 1,2,...,n1. The elements of `c` are
        checked to ensure that they are nonzero. If any element is found
        to be zero, the corresponding algebraic equation is assumed to
        be s_{ij} = r_{ij}. This feature can be used to define the
        equations for nodes at which, for example, Dirichlet boundary
        conditions are applied, or for nodes external to the problem of
        interest, by setting `c`[i-1,j-1] = 0.0 at appropriate points.
        The corresponding value of `r`[i-1,j-1] is set equal to the
        appropriate value, namely the difference between the prescribed
        value of t_{ij} and the current value of t_{ij} in the Dirichlet
        case, or zero at an external point.

    d : float, array-like, shape (n1, n2)
        `d`[i-1,j-1] must contain the coefficient of the 'easterly' term
        involving s_{i+1,j} in the (i,j)th equation of the system (2),
        for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of `d`, for
        i = n1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        rectangle.

    e : float, array-like, shape (n1, n2)
        `e`[i-1,j-1] must contain the coefficient of the 'northerly'
        term involving s_{i,j+1} in the (i,j)th equation of the system
        (2), for j = 1,2,...,n2, for i = 1,2,...,n1. The elements of
        `e`, for j = n2, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        rectangle.

    aparam : float
        The iteration acceleration factor. A value of 1.0 is adequate
        for most typical problems. However, if convergence is slow, the
        value can be reduced, typically to 0.2 or 0.1. If divergence is
        obtained, the value can be increased, typically to 2.0, 5.0 or
        10.0.

    it : int
        The iteration number. It must be initialized, but not
        necessarily to 1, before the first call, and must be incremented
        by one in the calling program for each subsequent call.
        ``dim2_ellip_fd_iter`` uses the counter to select the
        appropriate acceleration argument from a sequence of nine, each
        one being used twice in succession. (Note that the acceleration
        argument depends on the value of `aparam`.)

    r : float, array-like, shape (n1, n2)
        `r`[i-1,j-1] must contain the current residual r_{ij} on the
        right-hand side of the (i,j)th equation of the system (2), for j
        = 1,2,...,n2, for i = 1,2,...,n1.

    Returns
    -------
    r : float, ndarray, shape (n1, n2)
        These residuals are overwritten by the corresponding components
        of solution s to the system (2), i.e., the changes to be made to
        the vector t to reduce the residuals supplied.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 > 1.

        (`errno` 1)
            On entry, n1 = *<value>*.

            Constraint: n1 > 1.

        (`errno` 3)
            On entry, `aparam` = *<value>*.

            Constraint: `aparam` > 0.0.

        (`errno` 4)
            On entry, `aparam` = *<value>*, n1 = *<value>* and n2 =
            *<value>*.

            Constraint: `aparam` <= ((n1-1)^2+(n2-1)^2)/2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim2_ellip_fd_iter as b_dim2_ellip_fd_iter

    fun_name = 'naginterfaces.library.pde.dim2_ellip_fd_iter'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _n2_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dim2_ellip_fd_iter(
        _n1,
        _n2,
        _a,
        _b,
        _c,
        _d,
        _e,
        aparam,
        it,
        _r,
    )
    _r_py = _r
    return _r_py

def dim3_ellip_fd_iter(n2, n3, a, b, c, d, e, f, g, aparam, it, r):
    u"""
    Elliptic PDE, solution of finite difference equations by SIP,
    seven-point three-dimensional molecule, one iteration.

    ``dim3_ellip_fd_iter`` performs at each call one iteration of the
    Strongly Implicit Procedure.
    It is used to calculate on successive calls a sequence of
    approximate corrections to the current estimate of the solution when
    solving a system of simultaneous algebraic equations for which the
    iterative update matrix is of seven-point molecule form on a
    three-dimensional topologically-rectangular mesh. ('Topological'
    means that a polar grid (r,theta), for example, can be used as it is
    equivalent to a rectangular box.)

    For full information please refer to the NAG Library document for
    d03ub

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/d03/d03ubf.html

    Parameters
    ----------
    n2 : int
        The number of nodes in the second coordinate direction, n_2.

    n3 : int
        The number of nodes in the third coordinate direction, n_3.

    a : float, array-like, shape (n1, `n2`, `n3`)
        `a`[i-1,j-1,k-1] must contain the coefficient of s_{ij,k-1} in
        the (i,j,k)th equation of the system (2), for k = 1,2,...,`n3`,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `a`,
        for k = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    b : float, array-like, shape (n1, `n2`, `n3`)
        `b`[i-1,j-1,k-1] must contain the coefficient of s_{i,j-1,k} in
        the (i,j,k)th equation of the system (2), for k = 1,2,...,`n3`,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `b`,
        for j = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    c : float, array-like, shape (n1, `n2`, `n3`)
        `c`[i-1,j-1,k-1] must contain the coefficient of s_{i-1,j,k} in
        the (i,j,k)th equation of the system (2), for k = 1,2,...,`n3`,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `c`,
        for i = 1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    d : float, array-like, shape (n1, `n2`, `n3`)
        `d`[i-1,j-1,k-1] must contain the coefficient of s_{ijk}, the
        'central' term, in the (i,j,k)th equation of the system (2), for
        k = 1,2,...,`n3`, for j = 1,2,...,`n2`, for i = 1,2,...,n1. The
        elements of `d` are checked to ensure that they are nonzero. If
        any element is found to be zero, the corresponding algebraic
        equation is assumed to be s_{ijk} = r_{ijk}. This feature can be
        used to define the equations for nodes at which, for example,
        Dirichlet boundary conditions are applied, or for nodes external
        to the problem of interest, by setting `d`[i-1,j-1,k-1] = 0.0 at
        appropriate points. The corresponding value of r_{ijk} is set
        equal to the appropriate value, namely the difference between
        the prescribed value of t_{ijk} and the current value in the
        Dirichlet case, or zero at an external point.

    e : float, array-like, shape (n1, `n2`, `n3`)
        `e`[i-1,j-1,k-1] must contain the coefficient of s_{i+1,j,k} in
        the (i,j,k)th equation of the system (2), for k = 1,2,...,`n3`,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `e`,
        for i = n1, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    f : float, array-like, shape (n1, `n2`, `n3`)
        `f`[i-1,j-1,k-1] must contain the coefficient of s_{i,j+1,k} in
        the (i,j,k)th equation of the system (2), for k = 1,2,...,`n3`,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `f`,
        for j = `n2`, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    g : float, array-like, shape (n1, `n2`, `n3`)
        `g`[i-1,j-1,k-1] must contain the coefficient of s_{i,j,k+1} in
        the (i,j,k)th equation of the system (2), for k = 1,2,...,`n3`,
        for j = 1,2,...,`n2`, for i = 1,2,...,n1. The elements of `g`,
        for k = `n3`, must be zero after incorporating the boundary
        conditions, since they involve nodal values from outside the
        box.

    aparam : float
        The iteration acceleration factor. A value of 1.0 is adequate
        for most typical problems. However, if convergence is slow, the
        value can be reduced, typically to 0.2 or 0.1. If divergence is
        obtained, the value can be increased, typically to 2.0, 5.0 or
        10.0.

    it : int
        The iteration number. It must be initialized, but not
        necessarily to 1, before the first call, and should be
        incremented by one in the calling program for each subsequent
        call. The function uses this counter to select the appropriate
        acceleration argument from a sequence of nine, each one being
        used twice in succession. (Note that the acceleration argument
        depends on the value of `aparam`.)

    r : float, array-like, shape (n1, `n2`, `n3`)
        The current residual r_{ijk} on the right-hand side of the
        (i,j,k)th equation of the system (2), for k = 1,2,...,`n3`, for
        j = 1,2,...,`n2`, for i = 1,2,...,n1.

    Returns
    -------
    r : float, ndarray, shape (n1, `n2`, `n3`)
        These residuals are overwritten by the corresponding components
        of the solution s of the system (2), i.e., the changes to be
        made to the vector t to reduce the residuals supplied.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n3` = *<value>*.

            Constraint: `n3` > 1.

        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` > 1.

        (`errno` 1)
            On entry, n1 = *<value>*.

            Constraint: n1 > 1.

        (`errno` 2)
            On entry, sda = *<value>* and `n2` = *<value>*.

            Constraint: sda >= `n2`.

        (`errno` 3)
            On entry, `aparam` = *<value>*.

            Constraint: `aparam` > 0.0.

        (`errno` 4)
            On entry, `aparam` = *<value>*, n1 = *<value>*, `n2` =
            *<value>* and `n3` = *<value>*.

            Constraint: `aparam` <= ((n1-1)^2+(`n2`-1)^2+(`n3`-1)^2)/3.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.pde import dim3_ellip_fd_iter as b_dim3_ellip_fd_iter

    fun_name = 'naginterfaces.library.pde.dim3_ellip_fd_iter'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=3)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _sda_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=3)[1]
    _sda = _b_utils._EngineIntScalarType(
        _sda_py, {'fun_name': fun_name, 'entity': '_sda'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _g = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'g'},
        g,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_dim3_ellip_fd_iter(
        _n1,
        n2,
        n3,
        _sda,
        _a,
        _b,
        _c,
        _d,
        _e,
        _f,
        _g,
        aparam,
        it,
        _r,
    )
    _r_py = _r
    return _r_py
