# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `rand` Chapter.

``rand`` - Random Number Generators

This module is concerned with the generation of sequences of independent
pseudorandom and quasi-random numbers from various distributions, and
models.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05intro.html

See Also
--------
``naginterfaces.library.examples.rand`` :
    This subpackage contains examples for the ``rand`` module.
    See also the :ref:`library_rand_ex` subsection.
"""

# NAG Copyright 2017-2019.

from __future__ import absolute_import

import collections as _collections

from math import ceil
from math import log
from math import sqrt

from ..base import utils as _b_utils
from ..library import _utils

def init_repeat(genid, seed, subid=0):
    u"""
    Initializes a pseudorandom number generator to give a repeatable
    sequence.

    ``init_repeat`` initializes the selected base generator, as used by
    the group of pseudorandom number functions (see
    ``init_leapfrog``--``init_skipahead``, ``permute``, ``sample``,
    ``times_garch_asym1``--``times_mv_varma``,
    ``matrix_orthog``--``matrix_2waytable``, ``copula_students_t``,
    ``copula_normal``, ``multivar_students_t``, ``multivar_normal`` and
    ``dist_uniform01``--``int_uniform``) and the quasi-random scrambled
    sequence initialization function, ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05kf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05kff.html

    Parameters
    ----------
    genid : int
        Must contain the type of base generator to use.

        `genid` = 1

          NAG basic generator.

        `genid` = 2

          Wichmann Hill I generator.

        `genid` = 3

          Mersenne Twister.

        `genid` = 4

          Wichmann Hill II generator.

        `genid` = 5

          ACORN generator.

        `genid` = 6

          L'Ecuyer MRG32k3a generator.

        See [the G05 Introduction] for details of each of the base
        generators.

    seed : int, array-like, shape (lseed)
        The initial (seed) values for the selected base generator. The
        number of initial values required varies with each of the base
        generators.

        If `genid` = 1, one seed is required.

        If `genid` = 2, four seeds are required.

        If `genid` = 3, 624 seeds are required.

        If `genid` = 4, four seeds are required.

        If `genid` = 5, (k+1)p seeds are required, where k and p are
        defined by `subid`.

        For the ACORN generator it is recommended that an odd value is
        used for `seed`[0].

        If `genid` = 6, six seeds are required.

        If insufficient seeds are provided then the first lseed-1 values
        supplied in `seed` are used and the remaining values are
        randomly generated using the NAG basic generator.

        In such cases the NAG basic generator is initialized using the
        value supplied in `seed`[lseed-1].

    subid : int, optional
        If `genid` = 2, `subid` indicates which of the 273
        sub-generators to use. In this case, the ((|`subid`|+272) mod
        273)+1 sub-generator is used.

        If `genid` = 5, `subid` indicates the values of k and p to use,
        where k is the order of the generator, and p controls the size
        of the modulus, M, with M = 2^(p*30).

        If `subid` < 1, the default values of k = 10 and p = 2 are used,
        otherwise values for k and p are calculated from the formula,
        `subid` = k+1000*(p-1).

        If `genid` = 6 and mod(`subid`, 2) = 0 the range of the
        generator is set to (0,1], otherwise the range is set to (0,1);
        in this case the sequence is identical to the implementation of
        MRG32k3a in TestU01 for identical seeds.

        For all other values of `genid`, `subid` is not referenced.

    Returns
    -------
    statecomm : dict, RNG communication object
        RNG communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: `genid` = 1, 2, 3, 4, 5 or 6.

        (`errno` 3)
            On entry, invalid `seed`.

        (`errno` 4)
            On entry, lseed = *<value>*.

            Constraint: lseed >= 1.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    :meth:`naginterfaces.library.examples.rand.copula_students_t_ex.main`
    """

    from ..base.rand import init_repeat as b_init_repeat

    fun_name = 'naginterfaces.library.rand.init_repeat'

    _seed = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'seed'},
        seed,
        copy=False,
    )
    _lseed_py = _b_utils._EngineIntArrayType.get_shape(_seed, {'fun_name': fun_name, 'entity': 'seed'}, exp_rank=1)[0]
    _lseed = _b_utils._EngineIntScalarType(
        _lseed_py, {'fun_name': fun_name, 'entity': '_lseed'},
    ).to_ctypes()
    _return_scalar = b_init_repeat(
        genid,
        subid,
        _seed,
        _lseed,
    )
    _statecomm_py = _return_scalar
    return _statecomm_py

def init_nonrepeat(genid, subid=0):
    u"""
    Initializes a pseudorandom number generator to give a non-repeatable
    sequence.

    ``init_nonrepeat`` initializes the selected base generator to
    generate a non-repeatable sequence of variates.
    The base generator can then be used by the group of pseudorandom
    number functions (see ``init_leapfrog``--``init_skipahead``,
    ``permute``, ``sample``, ``times_garch_asym1``--``times_mv_varma``,
    ``matrix_orthog``--``matrix_2waytable``, ``copula_students_t``,
    ``copula_normal``, ``multivar_students_t``, ``multivar_normal`` and
    ``dist_uniform01``--``int_uniform``) and the quasi-random scrambled
    sequence initialization function, ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05kg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05kgf.html

    Parameters
    ----------
    genid : int
        Must contain the type of base generator to use.

        `genid` = 1

          NAG basic generator.

        `genid` = 2

          Wichmann Hill I generator.

        `genid` = 3

          Mersenne Twister.

        `genid` = 4

          Wichmann Hill II generator.

        `genid` = 5

          ACORN generator.

        `genid` = 6

          L'Ecuyer MRG32k3a generator.

        See [the G05 Introduction] for details of each of the base
        generators.

    subid : int, optional
        If `genid` = 2, `subid` indicates which of the 273
        sub-generators to use. In this case, the ((|`subid`|+272) mod
        273)+1 sub-generator is used.

        If `genid` = 5, `subid` indicates the values of k and p to use,
        where k is the order of the generator, and p controls the size
        of the modulus, M, with M = 2^(p*30).

        If `subid` < 1, the default values of k = 10 and p = 2 are used,
        otherwise values for k and p are calculated from the formula,
        `subid` = k+1000*(p-1).

        If `genid` = 6 and mod(`subid`, 2) = 0 the range of the
        generator is set to (0,1], otherwise the range is set to (0,1);
        in this case the sequence is identical to the implementation of
        MRG32k3a in TestU01 for identical seeds.

        For all other values of `genid`, `subid` is not referenced.

    Returns
    -------
    statecomm : dict, RNG communication object
        RNG communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: `genid` = 1, 2, 3, 4, 5 or 6.
    """

    from ..base.rand import init_nonrepeat as b_init_nonrepeat

    _return_scalar = b_init_nonrepeat(
        genid,
        subid,
    )
    _statecomm_py = _return_scalar
    return _statecomm_py

def init_leapfrog(n, k, statecomm):
    u"""
    Primes a pseudorandom number generator for generating multiple
    streams using leap-frog.

    ``init_leapfrog`` allows for the generation of multiple,
    independent, sequences of pseudorandom numbers using the leap-frog
    method.

    For full information please refer to the NAG Library document for
    g05kh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05khf.html

    Parameters
    ----------
    n : int
        n, the total number of sequences required.

    k : int
        k, the number of the current sequence.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: 0 < `k` <= `n`.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 4)
            On entry, cannot use leap-frog with the base generator
            defined by `statecomm`['state'].
    """

    from ..base.rand import init_leapfrog as b_init_leapfrog

    b_init_leapfrog(
        n,
        k,
        statecomm,
    )

def init_skipahead(n, statecomm):
    u"""
    Primes a pseudorandom number generator for generating multiple
    streams using skip-ahead.

    ``init_skipahead`` allows for the generation of multiple,
    independent, sequences of pseudorandom numbers using the skip-ahead
    method.

    The base pseudorandom number sequence defined by
    `statecomm`['state'] is advanced n places.

    For full information please refer to the NAG Library document for
    g05kj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05kjf.html

    Parameters
    ----------
    n : int
        n, the number of places to skip ahead.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 3)
            On entry, cannot use skip-ahead with the base generator
            defined by `statecomm`['state'].

        (`errno` 4)
            On entry, the base generator is Mersenne Twister, but the
            `statecomm`['state'] vector defined on initialization is not
            large enough to perform a skip ahead. See the initialization
            function ``init_repeat`` or ``init_nonrepeat``.
    """

    from ..base.rand import init_skipahead as b_init_skipahead

    b_init_skipahead(
        n,
        statecomm,
    )

def init_skipahead_power2(n, statecomm):
    u"""
    Primes a pseudorandom number generator for generating multiple
    streams using skip-ahead, skipping ahead a power of 2.

    ``init_skipahead_power2`` allows for the generation of multiple,
    independent, sequences of pseudorandom numbers using the skip-ahead
    method.
    The base pseudorandom number sequence defined by
    `statecomm`['state'] is advanced 2^n places.

    For full information please refer to the NAG Library document for
    g05kk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05kkf.html

    Parameters
    ----------
    n : int
        n, where the number of places to skip-ahead is defined as 2^n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 3)
            On entry, cannot use skip-ahead with the base generator
            defined by `statecomm`['state'].

        (`errno` 4)
            On entry, the `statecomm`['state'] vector defined on
            initialization is not large enough to perform a skip-ahead
            (applies to Mersenne Twister base generator). See the
            initialization function ``init_repeat`` or
            ``init_nonrepeat``.
    """

    from ..base.rand import init_skipahead_power2 as b_init_skipahead_power2

    b_init_skipahead_power2(
        n,
        statecomm,
    )

def permute(indx, statecomm):
    u"""
    Pseudorandom permutation of an integer vector.

    ``permute`` performs a pseudorandom permutation of a vector of
    integers.

    For full information please refer to the NAG Library document for
    g05nc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ncf.html

    Parameters
    ----------
    indx : int, array-like, shape (n)
        The n integer values to be permuted.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    indx : int, ndarray, shape (n)
        The n permuted integer values.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import permute as b_permute

    fun_name = 'naginterfaces.library.rand.permute'

    _indx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'indx'},
        indx,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_indx, {'fun_name': fun_name, 'entity': 'indx'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    b_permute(
        _indx,
        _n,
        statecomm,
    )
    _indx_py = _indx
    return _indx_py

def sample(m, statecomm, ipop=None, n=None):
    u"""
    Pseudorandom sample from an integer vector.

    ``sample`` selects a pseudorandom sample without replacement from an
    integer vector.

    For full information please refer to the NAG Library document for
    g05nd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ndf.html

    Parameters
    ----------
    m : int
        The sample size.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    ipop : None or int, array-like, shape (`n`), optional
        Note: when `n` is None this argument must not be None.

        The population to be sampled.

        If `ipop` is None, then the population is assumed to be the set
        of values (1,2,...,`n`) and the array `ipop` is not referenced.

    n : None or int, optional
        Note: when `ipop` is None this argument must not be None.

        Note: if this argument is None then a default value will be
        used, determined as follows: `ipop`.shape[0].

        The number of elements in the population to be sampled.

    Returns
    -------
    isampl : int, ndarray, shape (`m`)
        The selected sample.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `m` <= `n`.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import sample as b_sample

    fun_name = 'naginterfaces.library.rand.sample'

    _ipop = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipop'},
        ipop,
        copy=False,
        allow_none=True,
    )
    _lipop_py = _b_utils._EngineIntArrayType.get_shape(_ipop, {'fun_name': fun_name, 'entity': 'ipop'}, exp_rank=1, allow_none=True)[0]
    _lipop = _b_utils._EngineIntScalarType(
        _lipop_py, {'fun_name': fun_name, 'entity': '_lipop'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _isampl = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m_py,),
    )
    b_sample(
        _ipop,
        n,
        _isampl,
        _m,
        statecomm,
        _lipop,
    )
    _isampl_py = _isampl
    return _isampl_py

def sample_wgt(order, wt, m, statecomm, ipop=None):
    u"""
    Pseudorandom sample, without replacement, unequal weights.

    ``sample_wgt`` selects a pseudorandom sample, without replacement
    and allowing for unequal probabilities.

    For full information please refer to the NAG Library document for
    g05ne

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05nef.html

    Parameters
    ----------
    order : str, length 1
        A flag indicating the sorted status of the `wt` vector.

        `order` = 'A'

          `wt` is sorted in ascending order,

        `order` = 'D'

          `wt` is sorted in descending order,

        `order` = 'U'

          `wt` is unsorted and ``sample_wgt`` will sort the weights
          prior to using them.

        Irrespective of the value of `order`, no checks are made on the
        sorted status of `wt`, e.g., it is possible to supply `order` =
        'A', even when `wt` is not sorted.

        In such cases the `wt` array will not be sorted internally, but
        ``sample_wgt`` will still work correctly except, possibly, in
        cases of extreme weight values.

        It is usually more efficient to specify a value of `order` that
        is consistent with the status of `wt`.

    wt : float, array-like, shape (n)
        w_i, the relative probability weights. These weights need not
        sum to 1.0.

    m : int
        m, the size of the sample required.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    ipop : None or int, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ipop` is not None: n; otherwise: 0.

        The population to be sampled. If `ipop` is None then the
        population is assumed to be the set of values (1,2,...,n) and
        the array `ipop` is not referenced. Elements of `ipop` with the
        same value are not combined, therefore if `wt`[i-1] !=
        0,`wt`[j-1] != 0 and i != j then there is a nonzero probability
        that the sample will contain both `ipop`[i-1] and `ipop`[j-1].
        If `ipop`[i-1] = `ipop`[j-1] then that value can appear in
        `isampl` more than once.

    Returns
    -------
    isampl : int, ndarray, shape (`m`)
        The selected sample.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `order` had an illegal value.

        (`errno` 2)
            On entry, at least one weight was less than zero.

        (`errno` 5)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 7)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: 0 <= `m` <= n.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 21)
            On entry, `m` = *<value>*, number of nonzero weights =
            *<value>*.

            Constraint: must be at least `m` nonzero weights.
    """

    from ..base.rand import sample_wgt as b_sample_wgt

    fun_name = 'naginterfaces.library.rand.sample_wgt'

    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_wt, {'fun_name': fun_name, 'entity': 'wt'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipop = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipop'},
        ipop,
        copy=False,
        allow_none=True,
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _isampl = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m_py,),
    )
    b_sample_wgt(
        order,
        _wt,
        _ipop,
        _n,
        _isampl,
        _m,
        statecomm,
    )
    _isampl_py = _isampl
    return _isampl_py

def times_garch_asym1(dist, num, ip, iq, theta, gamma, df, fcall, r, statecomm):
    u"""
    Generates a realization of a time series from a GARCH process with
    asymmetry of the form (epsilon_{t-1}+gamma)^2.

    ``times_garch_asym1`` generates a given number of terms of a type I
    AGARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g05pd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pdf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o, the next
        `iq` elements must contain the coefficients alpha_i, for i =
        1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `r`.

    r : float, array-like, shape (lr)
        The array contains information required to continue a sequence
        if `fcall` = False.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    ht : float, ndarray, shape (`num`)
        The conditional variances h_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    et : float, ndarray, shape (`num`)
        The observations epsilon_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    r : float, ndarray, shape (lr)
        Contains information that can be used in a subsequent call of
        ``times_garch_asym1``, with `fcall` = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 5)
            On entry, `theta`[*<value>*] = *<value>*.

            Constraint: `theta`[i-1] >= 0.0.

        (`errno` 7)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 11)
            `ip` or `iq` is not the same as when `r` was set up in a
            previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 12)
            On entry, lr is not large enough, lr = *<value>*: minimum
            length required = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 51)
            On entry, sum of `theta`[i-1], for i = 2,3,...,`ip`+`iq`+1
            is >= 1.0: sum = *<value>*.
    """

    from ..base.rand import times_garch_asym1 as b_times_garch_asym1

    fun_name = 'naginterfaces.library.rand.times_garch_asym1'

    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num = _b_utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = _b_utils._EngineIntScalarType.to_py(_num)
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lr_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _lr = _b_utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    b_times_garch_asym1(
        dist,
        _num,
        ip,
        iq,
        _theta,
        gamma,
        df,
        _ht,
        _et,
        fcall,
        _r,
        _lr,
        statecomm,
    )
    _r_py = _r
    _ht_py = _ht
    _et_py = _et
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ht', 'et', 'r'],
    )(
        _ht_py,
        _et_py,
        _r_py,
    )
    return _return_data

def times_garch_asym2(dist, num, ip, iq, theta, gamma, df, fcall, r, statecomm):
    u"""
    Generates a realization of a time series from a GARCH process with
    asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``times_garch_asym2`` generates a given number of terms of a type II
    AGARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g05pe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pef.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o, the next
        `iq` elements must contain the coefficients alpha_i, for i =
        1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `r`.

    r : float, array-like, shape (lr)
        The array contains information required to continue a sequence
        if `fcall` = False.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    ht : float, ndarray, shape (`num`)
        The conditional variances h_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    et : float, ndarray, shape (`num`)
        The observations epsilon_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    r : float, ndarray, shape (lr)
        Contains information that can be used in a subsequent call of
        ``times_garch_asym2``, with `fcall` = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 5)
            On entry, `theta`[*<value>*] = *<value>*.

            Constraint: `theta`[i-1] >= 0.0.

        (`errno` 7)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 11)
            `ip` or `iq` is not the same as when `r` was set up in a
            previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 12)
            On entry, lr is not large enough, lr = *<value>*: minimum
            length required = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 51)
            On entry, sum of `theta`[i-1], for i = 2,3,...,`ip`+`iq`+1
            is >= 1.0: sum = *<value>*.
    """

    from ..base.rand import times_garch_asym2 as b_times_garch_asym2

    fun_name = 'naginterfaces.library.rand.times_garch_asym2'

    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num = _b_utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = _b_utils._EngineIntScalarType.to_py(_num)
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lr_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _lr = _b_utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    b_times_garch_asym2(
        dist,
        _num,
        ip,
        iq,
        _theta,
        gamma,
        df,
        _ht,
        _et,
        fcall,
        _r,
        _lr,
        statecomm,
    )
    _r_py = _r
    _ht_py = _ht
    _et_py = _et
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ht', 'et', 'r'],
    )(
        _ht_py,
        _et_py,
        _r_py,
    )
    return _return_data

def times_garch_gjr(dist, num, ip, iq, theta, gamma, df, fcall, r, statecomm):
    u"""
    Generates a realization of a time series from an asymmetric Glosten,
    Jagannathan and Runkle (GJR) GARCH process.

    ``times_garch_gjr`` generates a given number of terms of a GJR
    GARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g05pf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pff.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o, the next
        `iq` elements must contain the coefficients alpha_i, for i =
        1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `r`.

    r : float, array-like, shape (lr)
        The array contains information required to continue a sequence
        if `fcall` = False.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    ht : float, ndarray, shape (`num`)
        The conditional variances h_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    et : float, ndarray, shape (`num`)
        The observations epsilon_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    r : float, ndarray, shape (lr)
        Contains information that can be used in a subsequent call of
        ``times_garch_gjr``, with `fcall` = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 5)
            On entry, `theta`[*<value>*] = *<value>* and gamma =
            *<value>*.

            Constraint: alpha_i+gamma >= 0.

        (`errno` 7)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 11)
            `ip` or `iq` is not the same as when `r` was set up in a
            previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 12)
            On entry, lr is not large enough, lr = *<value>*: minimum
            length required = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 51)
            On entry, `theta`[*<value>*] = *<value>*.

            Constraint: `theta`[i] >= 0.0.

        (`errno` 52)
            On entry, sum of `theta`[i] = *<value>*.

            Constraint: sum of `theta`[i], for i = 1,2,...,`ip`+`iq` is
            < 1.0.
    """

    from ..base.rand import times_garch_gjr as b_times_garch_gjr

    fun_name = 'naginterfaces.library.rand.times_garch_gjr'

    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num = _b_utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = _b_utils._EngineIntScalarType.to_py(_num)
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lr_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _lr = _b_utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    b_times_garch_gjr(
        dist,
        _num,
        ip,
        iq,
        _theta,
        gamma,
        df,
        _ht,
        _et,
        fcall,
        _r,
        _lr,
        statecomm,
    )
    _r_py = _r
    _ht_py = _ht
    _et_py = _et
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ht', 'et', 'r'],
    )(
        _ht_py,
        _et_py,
        _r_py,
    )
    return _return_data

def times_garch_exp(dist, num, ip, iq, theta, df, fcall, r, statecomm):
    u"""
    Generates a realization of a time series from an exponential GARCH
    (EGARCH) process.

    ``times_garch_exp`` generates a given number of terms of an
    exponential GARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g05pg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pgf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (2*`iq`+`ip`+1)
        The initial parameter estimates for the vector theta. The first
        element must contain the coefficient alpha_o and the next `iq`
        elements must contain the autoregressive coefficients alpha_i,
        for i = 1,2,...,q. The next `iq` elements must contain the
        coefficients phi_i, for i = 1,2,...,q. The next `ip` elements
        must contain the moving average coefficients beta_j, for j =
        1,2,...,p.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `r`.

    r : float, array-like, shape (lr)
        The array contains information required to continue a sequence
        if `fcall` = False.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    ht : float, ndarray, shape (`num`)
        The conditional variances h_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    et : float, ndarray, shape (`num`)
        The observations epsilon_t, for t = 1,2,...,T, for the
        GARCH(p,q) sequence.

    r : float, ndarray, shape (lr)
        Contains information that can be used in a subsequent call of
        ``times_garch_exp``, with `fcall` = False.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 6)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 10)
            `ip` or `iq` is not the same as when `r` was set up in a
            previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 11)
            On entry, lr is not large enough, lr = *<value>*: minimum
            length required = *<value>*.

        (`errno` 12)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 20)
            Invalid sequence generated, use different parameters.
    """

    from ..base.rand import times_garch_exp as b_times_garch_exp

    fun_name = 'naginterfaces.library.rand.times_garch_exp'

    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num = _b_utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _num_py = _b_utils._EngineIntScalarType.to_py(_num)
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lr_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _lr = _b_utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    b_times_garch_exp(
        dist,
        _num,
        ip,
        iq,
        _theta,
        df,
        _ht,
        _et,
        fcall,
        _r,
        _lr,
        statecomm,
    )
    _r_py = _r
    _ht_py = _ht
    _et_py = _et
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ht', 'et', 'r'],
    )(
        _ht_py,
        _et_py,
        _r_py,
    )
    return _return_data

def times_arma(mode, n, xmean, phi, theta, avar, comm, statecomm):
    u"""
    Generates a realization of a time series from an ARMA model.

    ``times_arma`` generates a realization of a univariate time series
    from an autoregressive moving average (ARMA) model.
    The realization may be continued or a new realization generated at
    subsequent calls to ``times_arma``.

    For full information please refer to the NAG Library document for
    g05ph

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05phf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate terms in the time series using reference vector set
          up in a prior call to ``times_arma``.

        `mode` = 2

          Set up reference vector and generate terms in the time series.

    n : int
        n, the number of observations to be generated.

    xmean : float
        The mean of the time series.

    phi : float, array-like, shape (ip)
        The autoregressive coefficients of the model,
        phi_1,phi_2,...,phi_p.

    theta : float, array-like, shape (iq)
        The moving average coefficients of the model,
        theta_1,theta_2,...,theta_q.

    avar : float
        sigma^2, the variance of the Normal perturbations.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``times_arma``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    var : float
        The proportion of the variance of a term in the series that is
        due to the moving-average (error) terms in the model. The
        smaller this is, the nearer is the model to non-stationarity.

    x : float, ndarray, shape (`n`)
        Contains the next n observations from the time series.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, ip = *<value>*.

            Constraint: ip >= 0.

        (`errno` 5)
            On entry, the AR parameters are outside the stationarity
            region.

        (`errno` 6)
            On entry, iq = *<value>*.

            Constraint: iq >= 0.

        (`errno` 8)
            On entry, `avar` = *<value>*.

            Constraint: `avar` >= 0.0.

        (`errno` 9)
            Reference vector `comm`['r'] has been corrupted or not
            initialized correctly.

        (`errno` 9)
            ip or iq is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of ip = *<value>* and ip = *<value>*.

            Previous value of iq = *<value>* and iq = *<value>*.

        (`errno` 11)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import times_arma as b_times_arma

    fun_name = 'naginterfaces.library.rand.times_arma'

    _phi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'phi'},
        phi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ip_py = _b_utils._EngineFloat64ArrayType.get_shape(_phi, {'fun_name': fun_name, 'entity': 'phi'}, exp_rank=1)[0]
    _ip = _b_utils._EngineIntScalarType(
        _ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _iq = _b_utils._EngineIntScalarType(
        _iq_py, {'fun_name': fun_name, 'entity': '_iq'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalar = b_times_arma(
        mode,
        _n,
        xmean,
        _ip,
        _phi,
        _iq,
        _theta,
        avar,
        comm,
        statecomm,
        _x,
    )
    _var_py = _return_scalar
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['var', 'x'],
    )(
        _var_py,
        _x_py,
    )
    return _return_data

def times_mv_varma(mode, n, xmean, ip, phi, iq, theta, var, comm, statecomm):
    u"""
    Generates a realization of a multivariate time series from a VARMA
    model.

    ``times_mv_varma`` generates a realization of a multivariate time
    series from a vector autoregressive moving average (VARMA) model.
    The realization may be continued or a new realization generated at
    subsequent calls to ``times_mv_varma``.

    For full information please refer to the NAG Library document for
    g05pj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pjf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector and compute a realization of the
          recent history.

        `mode` = 1

          Generate terms in the time series using reference vector set
          up in a prior call to ``times_mv_varma``.

        `mode` = 2

          Combine the operations of `mode` = 0 or 1.

        `mode` = 3

          A new realization of the recent history is computed using
          information stored in the reference vector, and the following
          sequence of time series values are generated.

        If `mode` = 1 or 3, you must ensure that the reference vector
        `comm`['r'] and the values of k, `ip`, `iq`, `xmean`, `phi`,
        `theta`, `var` and ldvar have not been changed between calls to
        ``times_mv_varma``.

    n : int
        n, the number of observations to be generated.

    xmean : float, array-like, shape (k)
        mu, the vector of means of the multivariate time series.

    ip : int
        p, the number of autoregressive parameter matrices.

    phi : float, array-like, shape (k*k*`ip`)
        Must contain the elements of the `ip`*k*k autoregressive
        parameter matrices of the model, phi_1,phi_2,...,phi_p. If `phi`
        is considered as a three-dimensional array, dimensioned as
        `phi`[k-1,k-1,`ip`-1], the (i,j)th element of phi_l would be
        stored in `phi`[i-1,j-1,l-1]; that is,
        `phi`[(l-1)*k*k+(j-1)*k+i-1] must be set equal to the (i,j)th
        element of phi_l, for l = 1,2,...,p, i = 1,2,...,k and j =
        1,2,...,k.

    iq : int
        q, the number of moving average parameter matrices.

    theta : float, array-like, shape (k*k*`iq`)
        Must contain the elements of the `iq`*k*k moving average
        parameter matrices of the model, theta_1,theta_2,...,theta_q. If
        `theta` is considered as a three-dimensional array, dimensioned
        as `theta`(k,k,`iq`), the (i,j)th element of theta_l would be
        stored in `theta`[i-1,j-1,l-1]; that is,
        `theta`[(l-1)*k*k+(j-1)*k+i-1] must be set equal to the (i,j)th
        element of theta_l, for j = 1,2,...,k, for i = 1,2,...,k, for l
        = 1,2,...,q.

    var : float, array-like, shape (k, k)
        `var`[i-1,j-1] must contain the (i,j)th element of Sigma, for j
        = 1,2,...,k, for i = 1,2,...,k. Only the lower triangle is
        required.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1 or 3, this argument must have been initialized by
        a prior call to ``times_mv_varma``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (k, `n`)
        `x`[i-1,t-1] will contain a realization of the ith component of
        X_t, for t = 1,2,...,n, for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 5)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 6)
            On entry, the AR parameters are outside the stationarity
            region.

        (`errno` 7)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 8)
            On entry, the moving average parameter matrices are such
            that the model is non-invertible.

        (`errno` 9)
            On entry, the covariance matrix `var` is not positive
            semidefinite to machine precision.

        (`errno` 11)
            k is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of k = *<value>* and k = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 20)
            An excessive number of iterations were required by the NAG
            function used to evaluate the eigenvalues of the matrices
            used to test for stationarity or invertibility.

        (`errno` 21)
            The reference vector cannot be computed because the AR
            parameters are too close to the boundary of the stationarity
            region.

        (`errno` 22)
            An excessive number of iterations were required by the NAG
            function used to evaluate the eigenvalues of the covariance
            matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 23)
            An excessive number of iterations were required by the NAG
            function used to evaluate the eigenvalues stored in the
            reference vector.
    """

    from ..base.rand import times_mv_varma as b_times_mv_varma

    fun_name = 'naginterfaces.library.rand.times_mv_varma'

    _var = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'var'},
        var,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_var, 'var'),),
    )
    _xmean = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmean'},
        xmean,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmean, {'fun_name': fun_name, 'entity': 'xmean'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _phi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'phi'},
        phi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _n_py),
        sorder=_m_sorder,
    )
    try:
        b_times_mv_varma(
            mode,
            _n,
            _k,
            _xmean,
            ip,
            _phi,
            iq,
            _theta,
            _var,
            comm,
            statecomm,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _x_py

def times_smooth_exp(mode, n, itype, p, param, init, var, r, statecomm, e):
    u"""
    Generates a realization of a time series from an exponential
    smoothing model.

    ``times_smooth_exp`` simulates from an exponential smoothing model,
    where the model uses either single exponential, double exponential
    or a Holt--Winters method.

    For full information please refer to the NAG Library document for
    g05pm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pmf.html

    Parameters
    ----------
    mode : int
        Indicates if ``times_smooth_exp`` is continuing from a previous
        call or, if not, how the initial values are computed.

        `mode` = 0

          Values for m_0, r_0 and s_{-j}, for j = 0,1,...,p-1, are
          supplied in `init`.

        `mode` = 1

          ``times_smooth_exp`` continues from a previous call using
          values that are supplied in `r`. `r` is not updated.

        `mode` = 2

          ``times_smooth_exp`` continues from a previous call using
          values that are supplied in `r`. `r` is updated.

    n : int
        The number of terms of the time series being generated.

    itype : int
        The smoothing function.

        `itype` = 1

          Single exponential.

        `itype` = 2

          Brown's double exponential.

        `itype` = 3

          Linear Holt.

        `itype` = 4

          Additive Holt--Winters.

        `itype` = 5

          Multiplicative Holt--Winters.

    p : int
        If `itype` = 4 or 5, the seasonal order, p, otherwise `p` is not
        referenced.

    param : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2): 1; if `itype`=3: 3; if `itype` in
        (4, 5): 4; otherwise: 0.

        The smoothing parameters.

        If `itype` = 1 or 2, `param`[0] = alpha and any remaining
        elements of `param` are not referenced.

        If `itype` = 3, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = phi and any remaining elements of `param` are not
        referenced.

        If `itype` = 4 or 5, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = beta and `param`[3] = phi and any remaining
        elements of `param` are not referenced.

    init : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype`=1: 1; if `itype` in (2, 3): 2; if `itype` in
        (4, 5): 2+`p`; otherwise: 0.

        If `mode` = 0, the initial values for m_0, r_0 and s_{-j}, for j
        = 0,1,...,p-1, used to initialize the smoothing.

        If `itype` = 1, `init`[0] = m_0 and any remaining elements of
        `init` are not referenced.

        If `itype` = 2 or 3, `init`[0] = m_0 and `init`[1] = r_0 and any
        remaining elements of `init` are not referenced.

        If `itype` = 4 or 5, `init`[0] = m_0, `init`[1] = r_0 and
        `init`[2] to `init`[2+p-1] hold the values for s_{-j}, for j =
        0,1,...,p-1.

        Any remaining elements of `init` are not referenced.

    var : float
        The variance, sigma^2 of the Normal distribution used to
        generate the errors epsilon_i. If `var` <= 0.0 then Normally
        distributed errors are not used.

    r : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2, 3): 13; if `itype` in (4, 5):
        13+`p`; otherwise: 0.

        If `mode` = 1 or 2, `r` must contain the values as returned by a
        previous call to ``times_smooth_exp``, `r` need not be set
        otherwise.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    e : float, array-like, shape (en)
        If en > 0 and `var` <= 0.0, a vector from which the errors,
        epsilon_t are randomly drawn, with replacement.

        If en <= 0, `e` is not referenced.

    Returns
    -------
    r : float, ndarray, shape (:)
        If `mode` = 1, `r` is unchanged. Otherwise, `r` contains the
        information on the current state of smoothing.

    x : float, ndarray, shape (`n`)
        The generated time series, x_t, for t = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1, 2, 3, 4 or 5.

        (`errno` 4)
            On entry, `p` = *<value>*.

            Constraint: if `itype` = 4 or 5, `p` >= 2.

        (`errno` 5)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: 0 <= `param`[i] <= 1.

        (`errno` 5)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: if `itype` = 2, 0 < `param`[i] <= 1.

        (`errno` 5)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: `param`[i] >= 0.

        (`errno` 8)
            On entry, some of the elements of the array `r` have been
            corrupted or have not been initialized.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 12)
            Model unsuitable for multiplicative Holt--Winter, try a
            different set of parameters.
    """

    from ..base.rand import times_smooth_exp as b_times_smooth_exp

    fun_name = 'naginterfaces.library.rand.times_smooth_exp'

    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _init = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'init'},
        init,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _en_py = _b_utils._EngineFloat64ArrayType.get_shape(_e, {'fun_name': fun_name, 'entity': 'e'}, exp_rank=1)[0]
    _en = _b_utils._EngineIntScalarType(
        _en_py, {'fun_name': fun_name, 'entity': '_en'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_times_smooth_exp(
        mode,
        _n,
        itype,
        p,
        _param,
        _init,
        var,
        _r,
        statecomm,
        _e,
        _en,
        _x,
    )
    _r_py = _r
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'x'],
    )(
        _r_py,
        _x_py,
    )
    return _return_data

def kfold_xyw(k, fold, x, statecomm, sordx=1, y=None, w=None):
    u"""
    Permutes a matrix, vector, vector triplet into a form suitable for
    K-fold cross validation.

    ``kfold_xyw`` generates training and validation datasets suitable
    for use in cross-validation or jack-knifing.

    For full information please refer to the NAG Library document for
    g05pv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pvf.html

    Parameters
    ----------
    k : int
        K, the number of folds.

    fold : int
        The number of the fold to return as the validation dataset.

        On the first call to ``kfold_xyw`` `fold` should be set to 1 and
        then incremented by one at each subsequent call until all K sets
        of training and validation datasets have been produced.

        See [Further Comments] for more details on how a different
        calling sequence can be used.

    x : float, ndarray, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sordx`=2: m; otherwise: n.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sordx`=1: m; if `sordx`=2: n;
        otherwise: 0.

        On entry: if `fold` = 1, `x` must hold X_o, the values of X for
        the original dataset, otherwise, `x` must not be changed since
        the last call to ``kfold_xyw``.

        On exit: values of X for the training and validation datasets,
        with X_t held in observations 1 to `nt` and X_v in observations
        `nt`+1 to n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    sordx : int, optional
        Determines how variables are stored in `x`.

    y : None or float, ndarray, shape (:), optional, modified in place
        Note: the required length for this argument is determined as
        follows: if `y` is not None: n; otherwise: 0.

        Optionally, on entry: y_o, the values of y for the original
        dataset. If `fold` != 1, `y` must hold the vector returned in sy
        by the last call to ``kfold_xyw``.

        On exit, if not None on entry: values of y for the training and
        validation datasets, with y_t held in elements 1 to `nt` and y_v
        in elements `nt`+1 to n.

    w : None or float, ndarray, shape (:), optional, modified in place
        Note: the required length for this argument is determined as
        follows: if `w` is not None: n; otherwise: 0.

        Optionally, on entry: if `fold` != 1, `w` must hold the vector
        returned in sw by the last call to ``kfold_xyw``.

        On exit, if not None on entry: values of w for the training and
        validation datasets, with w_t held in elements 1 to `nt` and w_v
        in elements `nt`+1 to n.

    Returns
    -------
    nt : int
        n_t, the number of observations in the training dataset.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `k` = *<value>* and n = *<value>*.

            Constraint: 2 <= `k` <= n.

        (`errno` 21)
            On entry, `fold` = *<value>* and `k` = *<value>*.

            Constraint: 1 <= `fold` <= `k`.

        (`errno` 31)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 41)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 51)
            On entry, `sordx` = *<value>*.

            Constraint: `sordx` = 1 or 2.

        (`errno` 71)
            On entry, ldx = *<value>* and n = *<value>*.

            Constraint: if `sordx` = 1, ldx >= n.

        (`errno` 72)
            On entry, ldx = *<value>* and m = *<value>*.

            Constraint: if `sordx` = 2, ldx >= m.

        (`errno` 131)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 61)
            More than 50% of the data did not move when the data was
            shuffled. *<value>* of the *<value>* observations stayed
            put.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    """

    from ..base.rand import kfold_xyw as b_kfold_xyw

    fun_name = 'naginterfaces.library.rand.kfold_xyw'

    _sordx = _b_utils._EngineIntScalarType(
        sordx, {'fun_name': fun_name, 'entity': 'sordx'},
    ).to_ctypes()
    _sordx_py = _b_utils._EngineIntScalarType.to_py(_sordx)
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    if _sordx_py == 2:
        _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    else:
        _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    if _sordx_py == 2:
        _m_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
        _m = _b_utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
    else:
        _m_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
        _m = _b_utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        allow_none=True,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        allow_none=True,
    )
    try:
        _return_scalar = b_kfold_xyw(
            k,
            fold,
            _n,
            _m,
            _sordx,
            x,
            y,
            w,
            statecomm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _nt_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nt'],
    )(
        _nt_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _nt_py

def subsamp_xyw(nt, x, statecomm, sordx=1, y=None, w=None):
    u"""
    Permutes a matrix, vector, vector triplet into a form suitable for
    random sub-sampling validation.

    ``subsamp_xyw`` generates a dataset suitable for use with repeated
    random sub-sampling validation.

    For full information please refer to the NAG Library document for
    g05pw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pwf.html

    Parameters
    ----------
    nt : int
        n_t, the number of observations in the training dataset.

    x : float, ndarray, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sordx`=2: m; otherwise: n.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sordx`=1: m; if `sordx`=2: n;
        otherwise: 0.

        On entry: `x` must hold X_o, the values of X for the original
        dataset. This may be the same `x` as updated by a previous call
        to ``subsamp_xyw``.

        On exit: values of X for the training and validation datasets,
        with X_t held in observations 1 to `nt` and X_v in observations
        `nt`+1 to n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    sordx : int, optional
        Determines how variables are stored in `x`.

    y : None or float, ndarray, shape (:), optional, modified in place
        Note: the required length for this argument is determined as
        follows: if `y` is not None: n; otherwise: 0.

        Optionally, on entry: `y` must hold y_o, the values of y for the
        original dataset. This may be the same `y` as updated by a
        previous call to ``subsamp_xyw``.

        On exit, if not None on entry: values of y for the training and
        validation datasets, with y_t held in elements 1 to `nt` and y_v
        in elements `nt`+1 to n.

    w : None or float, ndarray, shape (:), optional, modified in place
        Note: the required length for this argument is determined as
        follows: if `w` is not None: n; otherwise: 0.

        Optionally, on entry: `w` must hold w_o, the values of w for the
        original dataset. This may be the same `w` as updated by a
        previous call to ``subsamp_xyw``.

        On exit, if not None on entry: values of w for the training and
        validation datasets, with w_t held in elements 1 to `nt` and w_v
        in elements `nt`+1 to n.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `nt` = *<value>* and n = *<value>*.

            Constraint: 1 <= `nt` <= n.

        (`errno` 21)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 31)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 41)
            On entry, `sordx` = *<value>*.

            Constraint: `sordx` = 1 or 2.

        (`errno` 61)
            On entry, ldx = *<value>* and n = *<value>*.

            Constraint: if `sordx` = 1, ldx >= n.

        (`errno` 62)
            On entry, ldx = *<value>* and m = *<value>*.

            Constraint: if `sordx` = 2, ldx >= m.

        (`errno` 111)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import subsamp_xyw as b_subsamp_xyw

    fun_name = 'naginterfaces.library.rand.subsamp_xyw'

    _sordx = _b_utils._EngineIntScalarType(
        sordx, {'fun_name': fun_name, 'entity': 'sordx'},
    ).to_ctypes()
    _sordx_py = _b_utils._EngineIntScalarType.to_py(_sordx)
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    if _sordx_py == 2:
        _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    else:
        _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
        _n = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_n'},
        ).to_ctypes()
    if _sordx_py == 2:
        _m_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
        _m = _b_utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
    else:
        _m_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
        _m = _b_utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        allow_none=True,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        allow_none=True,
    )
    b_subsamp_xyw(
        nt,
        _n,
        _m,
        _sordx,
        x,
        y,
        w,
        statecomm,
    )

def matrix_orthog(side, init, statecomm, a):
    u"""
    Generates a random orthogonal matrix.

    ``matrix_orthog`` generates a random orthogonal matrix.

    For full information please refer to the NAG Library document for
    g05px

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pxf.html

    Parameters
    ----------
    side : str, length 1
        Indicates whether the matrix A is multiplied on the left or
        right by the random orthogonal matrix U.

        `side` = 'L'

          The matrix A is multiplied on the left, i.e., premultiplied.

        `side` = 'R'

          The matrix A is multiplied on the right, i.e.,
          post-multiplied.

    init : str, length 1
        Indicates whether or not `a` should be initialized to the
        identity matrix.

        `init` = 'I'

          `a` is initialized to the identity matrix.

        `init` = 'N'

          `a` is not initialized and the matrix A must be supplied in
          `a`.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    a : float, array-like, shape (m, n)
        If `init` = 'N', `a` must contain the matrix A.

    Returns
    -------
    a : float, ndarray, shape (m, n)
        The matrix UA when `side` = 'L' or the matrix AU when `side` =
        'R'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `side` is not valid: `side` = *<value>*.

        (`errno` 2)
            On entry, `init` is not valid: `init` = *<value>*.

        (`errno` 3)
            On entry, `side` = *<value>*, m = *<value>*.

            Constraint: if `side` = 'L', m > 1; otherwise m >= 1.

        (`errno` 4)
            On entry, `side` = *<value>*, n = *<value>*.

            Constraint: if `side` = 'R', n > 1; otherwise n >= 1.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 8)
            On entry, `side` = *<value>*, m = *<value>*.

            Constraint: if `side` = 'L', m > 1; otherwise m >= 1.

        (`errno` 8)
            On entry, `side` = *<value>*, n = *<value>*.

            Constraint: if `side` = 'R', n > 1; otherwise n >= 1.
    """

    from ..base.rand import matrix_orthog as b_matrix_orthog

    fun_name = 'naginterfaces.library.rand.matrix_orthog'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    b_matrix_orthog(
        side,
        init,
        _m,
        _n,
        statecomm,
        _a,
    )
    _a_py = _a
    return _a_py

def matrix_corr(d, statecomm, eps=0.00001):
    u"""
    Generates a random correlation matrix.

    ``matrix_corr`` generates a random correlation matrix with given
    eigenvalues.

    For full information please refer to the NAG Library document for
    g05py

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pyf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        The n eigenvalues, lambda_i, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    eps : float, optional
        The maximum acceptable error in the diagonal elements.

    Returns
    -------
    c : float, ndarray, shape (n, n)
        A random correlation matrix, C, of dimension n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, an eigenvalue is negative.

        (`errno` 2)
            On entry, the eigenvalues do not sum to n.

        (`errno` 3)
            On entry, `eps` = *<value>*.

            Constraint: `eps` >= n*machine precision.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 5)
            The diagonals of the returned matrix are not unity, try
            increasing the value of `eps`, or rerun the code using a
            different seed.
    """

    from ..base.rand import matrix_corr as b_matrix_corr

    fun_name = 'naginterfaces.library.rand.matrix_corr'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_matrix_corr(
        _n,
        _d,
        eps,
        statecomm,
        _c,
    )
    _c_py = _c
    return _c_py

def matrix_2waytable(mode, totr, totc, comm, statecomm):
    u"""
    Generates a random two-way table.

    ``matrix_2waytable`` generates a random two-way table.

    For full information please refer to the NAG Library document for
    g05pz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05pzf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate two-way table using reference vector set up in a
          prior call to ``matrix_2waytable``.

        `mode` = 2

          Set up reference vector and generate two-way table.

    totr : int, array-like, shape (nrow)
        The m row totals, R_i, for i = 1,2,...,m.

    totc : int, array-like, shape (ncol)
        The n column totals, C_j, for j = 1,2,...,n.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``matrix_2waytable``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : int, ndarray, shape (nrow, ncol)
        If `mode` = 1 or 2, a pseudorandom two-way m by n table, X, with
        element `x`[i-1,j-1] containing the (i,j)th entry in the table
        such that sum(`x`[i-1,j-1]) = `totc`[j-1] and sum(`x`[i-1,j-1])
        = `totr`[i-1]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, nrow = *<value>*.

            Constraint: nrow >= 2.

        (`errno` 3)
            On entry, ncol = *<value>*.

            Constraint: ncol >= 2.

        (`errno` 4)
            On entry, at least one element of `totr` is negative or
            `totr` sums to zero.

        (`errno` 5)
            On entry, `totc` has at least one negative element.

        (`errno` 6)
            nrow or ncol is not the same as when `comm`['r'] was set up
            in a previous call.

            Previous value of nrow = *<value>* and nrow = *<value>*.

            Previous value of ncol = *<value>* and ncol = *<value>*.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 15)
            On entry, the arrays `totr` and `totc` do not sum to the
            same total: `totr` array total is *<value>*, `totc` array
            total is *<value>*.
    """

    from ..base.rand import matrix_2waytable as b_matrix_2waytable

    fun_name = 'naginterfaces.library.rand.matrix_2waytable'

    _totr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'totr'},
        totr,
        copy=False,
    )
    _nrow_py = _b_utils._EngineIntArrayType.get_shape(_totr, {'fun_name': fun_name, 'entity': 'totr'}, exp_rank=1)[0]
    _nrow = _b_utils._EngineIntScalarType(
        _nrow_py, {'fun_name': fun_name, 'entity': '_nrow'},
    ).to_ctypes()
    _totc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'totc'},
        totc,
        copy=False,
    )
    _ncol_py = _b_utils._EngineIntArrayType.get_shape(_totc, {'fun_name': fun_name, 'entity': 'totc'}, exp_rank=1)[0]
    _ncol = _b_utils._EngineIntScalarType(
        _ncol_py, {'fun_name': fun_name, 'entity': '_ncol'},
    ).to_ctypes()
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nrow_py, _ncol_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_matrix_2waytable(
        mode,
        _nrow,
        _ncol,
        _totr,
        _totc,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_students_t(mode, n, df, c, comm, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a Student's
    t-copula.

    ``copula_students_t`` sets up a reference vector and generates an
    array of pseudorandom numbers from a Student's t copula with nu
    degrees of freedom and covariance matrix nu/(nu-2)C.

    For full information please refer to the NAG Library document for
    g05rc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rcf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``copula_students_t``.

        `mode` = 2

          Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    df : int
        nu, the number of degrees of freedom of the distribution.

    c : float, array-like, shape (m, m)
        Matrix which, along with `df`, defines the covariance of the
        distribution. Only the upper triangle need be set.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``copula_students_t``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`, m)
        The array of values from a multivariate Student's t copula, with
        `x`[i-1,j-1] holding the jth dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 4)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 5)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 7)
            m is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of m = *<value>* and m = *<value>*.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.copula_students_t_ex.main`
    """

    from ..base.rand import copula_students_t as b_copula_students_t

    fun_name = 'naginterfaces.library.rand.copula_students_t'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_c, 'c'),),
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    b_copula_students_t(
        mode,
        _n,
        df,
        _m,
        _c,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_normal(mode, n, c, comm, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a Gaussian copula.

    ``copula_normal`` sets up a reference vector and generates an array
    of pseudorandom numbers from a Normal (Gaussian) copula with
    covariance matrix C.

    For full information please refer to the NAG Library document for
    g05rd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rdf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``copula_normal``.

        `mode` = 2

          Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    c : float, array-like, shape (m, m)
        The covariance matrix of the distribution. Only the upper
        triangle need be set.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``copula_normal``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`, :)
        The array of values from a multivariate Gaussian copula, with
        `x`[i-1,j-1] holding the jth dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 4)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 6)
            m is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of m = *<value>* and m = *<value>*.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import copula_normal as b_copula_normal

    fun_name = 'naginterfaces.library.rand.copula_normal'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_c, 'c'),),
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 0:
        _x_dim2 = 0
    else:
        _x_dim2 = _m_py
    exp_shape_x = (_n_py, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_m_sorder,
    )
    b_copula_normal(
        _mode,
        _n,
        _m,
        _c,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_clayton_bivar(n, theta, sorder, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a bivariate
    Clayton/Cook--Johnson copula.

    ``copula_clayton_bivar`` generates pseudorandom uniform bivariates
    with joint distribution of a Clayton/Cook--Johnson Archimedean
    copula.

    For full information please refer to the NAG Library document for
    g05re

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ref.html

    Parameters
    ----------
    n : int
        n, the number of bivariates to generate.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,2, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (:, :)
        The n bivariate uniforms with joint distribution described by
        C_theta, with `x`[i-1,j-1] holding the ith value for the jth
        dimension if `sorder` = 1 and the jth value for the ith
        dimension if `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= -1.0.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.
    """

    from ..base.rand import copula_clayton_bivar as b_copula_clayton_bivar

    fun_name = 'naginterfaces.library.rand.copula_clayton_bivar'

    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = 2
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = 2
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_copula_clayton_bivar(
        _n,
        theta,
        _sorder,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_frank_bivar(n, theta, sorder, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a bivariate Frank
    copula.

    ``copula_frank_bivar`` generates pseudorandom uniform bivariates
    with joint distribution of a Frank Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rff.html

    Parameters
    ----------
    n : int
        n, the number of bivariates to generate.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,2, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (:, :)
        The n bivariate uniforms with joint distribution described by
        C_theta, with `x`[i-1,j-1] holding the ith value for the jth
        dimension if `sorder` = 1 and the jth value for the ith
        dimension if `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.
    """

    from ..base.rand import copula_frank_bivar as b_copula_frank_bivar

    fun_name = 'naginterfaces.library.rand.copula_frank_bivar'

    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = 2
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = 2
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_copula_frank_bivar(
        _n,
        theta,
        _sorder,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_plackett_bivar(n, theta, sorder, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a bivariate Plackett
    copula.

    ``copula_plackett_bivar`` generates pseudorandom uniform bivariates
    with joint distribution of a Plackett copula.

    For full information please refer to the NAG Library document for
    g05rg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rgf.html

    Parameters
    ----------
    n : int
        n, the number of bivariates to generate.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,2, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (:, :)
        The n bivariate uniforms with joint distribution described by
        C_theta, with `x`[i-1,j-1] holding the ith value for the jth
        dimension if `sorder` = 1 and the jth value for the ith
        dimension if `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 0.0.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.
    """

    from ..base.rand import copula_plackett_bivar as b_copula_plackett_bivar

    fun_name = 'naginterfaces.library.rand.copula_plackett_bivar'

    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = 2
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = 2
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_copula_plackett_bivar(
        _n,
        theta,
        _sorder,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_clayton(n, m, theta, sorder, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate
    Clayton/Cook--Johnson copula.

    ``copula_clayton`` generates pseudorandom uniform variates with
    joint distribution of a Clayton/Cook--Johnson Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rhf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom uniform variates to generate.

    m : int
        m, the number of dimensions.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (:, :)
        The pseudorandom uniform variates with joint distribution
        described by C_theta, with `x`[i-1,j-1] holding the ith value
        for the jth dimension if `sorder` = 1 and the jth value for the
        ith dimension of `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 1.0*10^(-6).

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 5)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.
    """

    from ..base.rand import copula_clayton as b_copula_clayton

    fun_name = 'naginterfaces.library.rand.copula_clayton'

    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_copula_clayton(
        _n,
        _m,
        theta,
        _sorder,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_frank(n, m, theta, sorder, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate Frank
    copula.

    ``copula_frank`` generates pseudorandom uniform variates with joint
    distribution of a Frank Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rjf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom uniform variates to generate.

    m : int
        m, the number of dimensions.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (:, :)
        The pseudorandom uniform variates with joint distribution
        described by C_theta, with `x`[i-1,j-1] holding the ith value
        for the jth dimension if `sorder` = 1 and the jth value for the
        ith dimension of `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 1.0*10^(-6).

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 5)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.
    """

    from ..base.rand import copula_frank as b_copula_frank

    fun_name = 'naginterfaces.library.rand.copula_frank'

    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_copula_frank(
        _n,
        _m,
        theta,
        _sorder,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def copula_gumbel(n, m, theta, sorder, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a Gumbel--Hougaard
    copula.

    ``copula_gumbel`` generates pseudorandom uniform variates with joint
    distribution of a Gumbel--Hougaard Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rkf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom uniform variates to generate.

    m : int
        m, the number of dimensions.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (:, :)
        The pseudorandom uniform variates with joint distribution
        described by C_theta, with `x`[i-1,j-1] holding the ith value
        for the jth dimension if `sorder` = 1 and the jth value for the
        ith dimension of `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 1.0.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 5)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.
    """

    from ..base.rand import copula_gumbel as b_copula_gumbel

    fun_name = 'naginterfaces.library.rand.copula_gumbel'

    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_copula_gumbel(
        _n,
        _m,
        theta,
        _sorder,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def multivar_students_t(mode, n, df, xmu, c, r, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate
    Student's t-distribution.

    ``multivar_students_t`` sets up a reference vector and generates an
    array of pseudorandom numbers from a multivariate Student's t
    distribution with nu degrees of freedom, mean vector a and
    covariance matrix nu/(nu-2)C.

    For full information please refer to the NAG Library document for
    g05ry

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ryf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``multivar_students_t``.

        `mode` = 2

          Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    df : int
        nu, the number of degrees of freedom of the distribution.

    xmu : float, array-like, shape (m)
        a, the vector of means of the distribution.

    c : float, array-like, shape (m, m)
        Matrix which, along with `df`, defines the covariance of the
        distribution. Only the upper triangle need be set.

    r : float, array-like, shape (lr)
        If `mode` = 1, the reference vector as set up by
        ``multivar_students_t`` in a previous call with `mode` = 0 or 2.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    r : float, ndarray, shape (lr)
        If `mode` = 0 or 2, the reference vector that can be used in
        subsequent calls to ``multivar_students_t`` with `mode` = 1.

    x : float, ndarray, shape (`n`, :)
        The array of pseudorandom multivariate Student's t vectors
        generated by the function, with `x`[i-1,j-1] holding the jth
        dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 4)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 6)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 8)
            m is not the same as when `r` was set up in a previous call.

            Previous value of m = *<value>* and m = *<value>*.

        (`errno` 9)
            On entry, lr is not large enough, lr = *<value>*: minimum
            length required = *<value>*.

        (`errno` 10)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import multivar_students_t as b_multivar_students_t

    fun_name = 'naginterfaces.library.rand.multivar_students_t'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_c, 'c'),),
    )
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lr_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _lr = _b_utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 0:
        _x_dim2 = 0
    else:
        _x_dim2 = _m_py
    exp_shape_x = (_n_py, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_m_sorder,
    )
    b_multivar_students_t(
        _mode,
        _n,
        df,
        _m,
        _xmu,
        _c,
        _r,
        _lr,
        statecomm,
        _x,
    )
    _r_py = _r
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'x'],
    )(
        _r_py,
        _x_py,
    )
    return _return_data

def multivar_normal(sorder, mode12, n, xmu, c, r, statecomm):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate
    Normal distribution.

    ``multivar_normal`` sets up a reference vector and generates an
    array of pseudorandom numbers from a multivariate Normal
    distribution with mean vector a and covariance matrix C.

    For full information please refer to the NAG Library document for
    g05rz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05rzf.html

    Parameters
    ----------
    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    mode12 : int
        A code for selecting the operation to be performed by the
        function.

        `mode12` = 0

          Set up reference vector only.

        `mode12` = 1

          Generate variates using reference vector set up in a prior
          call to ``multivar_normal``.

        `mode12` = 2

          Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    xmu : float, array-like, shape (m)
        a, the vector of means of the distribution.

    c : float, array-like, shape (m, m)
        The covariance matrix of the distribution. Only the upper
        triangle need be set.

    r : float, array-like, shape (lr)
        If `mode12` = 1, the reference vector as set up by
        ``multivar_normal`` in a previous call with `mode12` = 0 or 2.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    r : float, ndarray, shape (lr)
        If `mode12` = 0 or 2, the reference vector that can be used in
        subsequent calls to ``multivar_normal`` with `mode12` = 1.

    x : float, ndarray, shape (:, :)
        The array of pseudorandom multivariate Normal vectors generated
        by the function.

        Two possible storage orders are available.

        If `sorder` = 1 then `x`[i-1,j-1] holds the jth dimension for
        the ith variate.

        If `sorder` = 2 this ordering is reversed and `x`[j-1,i-1] holds
        the jth dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode12` = *<value>*.

            Constraint: `mode12` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 5)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 7)
            m is not the same as when `r` was set up in a previous call.

            Previous value of m = *<value>* and m = *<value>*.

        (`errno` 8)
            On entry, lr is not large enough, lr = *<value>*: minimum
            length required = *<value>*.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import multivar_normal as b_multivar_normal

    fun_name = 'naginterfaces.library.rand.multivar_normal'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_c, 'c'),),
    )
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lr_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _lr = _b_utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    _mode12 = _b_utils._EngineIntScalarType(
        mode12, {'fun_name': fun_name, 'entity': 'mode12'},
    ).to_ctypes()
    _mode12_py = _b_utils._EngineIntScalarType.to_py(_mode12)
    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if _mode12_py == 0:
        _x_dim1 = 0
    elif _sorder_py == 1:
        _x_dim1 = _n_py
    else:
        _x_dim1 = _m_py
    if _mode12_py == 0:
        _x_dim2 = 0
    elif _sorder_py == 1:
        _x_dim2 = _m_py
    else:
        _x_dim2 = _n_py
    exp_shape_x = (_x_dim1, _x_dim2)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_x,
        sorder=_m_sorder,
    )
    b_multivar_normal(
        _sorder,
        _mode12,
        _n,
        _m,
        _xmu,
        _c,
        _r,
        _lr,
        statecomm,
        _x,
    )
    _r_py = _r
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'x'],
    )(
        _r_py,
        _x_py,
    )
    return _return_data

def dist_uniform01(n, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a uniform
    distribution over (0,1].

    ``dist_uniform01`` generates a vector of pseudorandom numbers taken
    from a uniform distribution between 0 and 1.

    For full information please refer to the NAG Library document for
    g05sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05saf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from a uniform distribution over the
        half closed interval (0,1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    from ..base.rand import dist_uniform01 as b_dist_uniform01

    fun_name = 'naginterfaces.library.rand.dist_uniform01'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_uniform01(
        _n,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_beta(n, a, b, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a beta distribution.

    ``dist_beta`` generates a vector of pseudorandom numbers taken from
    a beta distribution with parameters a and b.

    For full information please refer to the NAG Library document for
    g05sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05sbf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the parameter of the beta distribution.

    b : float
        b, the parameter of the beta distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified beta distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_beta as b_dist_beta

    fun_name = 'naginterfaces.library.rand.dist_beta'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_beta(
        _n,
        a,
        b,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_cauchy(n, xmed, semiqr, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a Cauchy
    distribution.

    ``dist_cauchy`` generates a vector of pseudorandom numbers from a
    Cauchy distribution with median a and semi-interquartile range b.

    For full information please refer to the NAG Library document for
    g05sc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05scf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmed : float
        a, the median of the distribution.

    semiqr : float
        b, the semi-interquartile range of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified Cauchy
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `semiqr` = *<value>*.

            Constraint: `semiqr` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_cauchy as b_dist_cauchy

    fun_name = 'naginterfaces.library.rand.dist_cauchy'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_cauchy(
        _n,
        xmed,
        semiqr,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_chisq(n, df, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a chi^2
    distribution.

    ``dist_chisq`` generates a vector of pseudorandom numbers taken from
    a chi^2-distribution with nu degrees of freedom.

    For full information please refer to the NAG Library document for
    g05sd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05sdf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    df : int
        nu, the number of degrees of freedom of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified
        chi^2-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_chisq as b_dist_chisq

    fun_name = 'naginterfaces.library.rand.dist_chisq'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_chisq(
        _n,
        df,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_dirichlet(n, a, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a Dirichlet
    distribution.

    ``dist_dirichlet`` generates a vector of pseudorandom numbers taken
    from a Dirichlet distribution.

    For full information please refer to the NAG Library document for
    g05se

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05sef.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float, array-like, shape (m)
        The parameter vector for the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`, m)
        The n pseudorandom numbers from the specified Dirichlet
        distribution, with `x`[i-1,j-1] holding the jth dimension for
        the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 3)
            On entry, at least one `a`[i] <= 0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_dirichlet as b_dist_dirichlet

    fun_name = 'naginterfaces.library.rand.dist_dirichlet'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_dist_dirichlet(
        _n,
        _m,
        _a,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_exp(n, a, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from an exponential
    distribution.

    ``dist_exp`` generates a vector of pseudorandom numbers from a
    (negative) exponential distribution with mean a.

    For full information please refer to the NAG Library document for
    g05sf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05sff.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the mean of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified exponential
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_exp as b_dist_exp

    fun_name = 'naginterfaces.library.rand.dist_exp'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_exp(
        _n,
        a,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_expmix(n, a, wgt, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from an exponential mix
    distribution.

    ``dist_expmix`` generates a vector of pseudorandom numbers from an
    exponential mix distribution composed of m exponential distributions
    each having a mean a_i and weight w_i.

    For full information please refer to the NAG Library document for
    g05sg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05sgf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float, array-like, shape (nmix)
        The m parameters a_i for the m exponential distributions in the
        mix.

    wgt : float, array-like, shape (nmix)
        The m weights w_i for the m exponential distributions in the
        mix.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified exponential mix
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, nmix = *<value>*.

            Constraint: nmix >= 1.

        (`errno` 3)
            On entry, `a`[*<value>*] = *<value>*.

            Constraint: `a`[i-1] > 0.0.

        (`errno` 4)
            On entry, `wgt`[*<value>*] = *<value>*.

            Constraint: `wgt`[i-1] >= 0.0.

        (`errno` 4)
            On entry, sum of `wgt` = *<value>*.

            Constraint: sum of `wgt` = 1.0.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_expmix as b_dist_expmix

    fun_name = 'naginterfaces.library.rand.dist_expmix'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nmix_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _nmix = _b_utils._EngineIntScalarType(
        _nmix_py, {'fun_name': fun_name, 'entity': '_nmix'},
    ).to_ctypes()
    _wgt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wgt'},
        wgt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_expmix(
        _n,
        _nmix,
        _a,
        _wgt,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_f(n, df1, df2, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from an F-distribution.

    ``dist_f`` generates a vector of pseudorandom numbers taken from an
    F (or Fisher's variance ratio) distribution with mu and nu degrees
    of freedom.

    For full information please refer to the NAG Library document for
    g05sh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05shf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    df1 : int
        mu, the number of degrees of freedom of the distribution.

    df2 : int
        nu, the number of degrees of freedom of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified F-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `df1` = *<value>*.

            Constraint: `df1` >= 1.

        (`errno` 3)
            On entry, `df2` = *<value>*.

            Constraint: `df2` >= 1.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_f as b_dist_f

    fun_name = 'naginterfaces.library.rand.dist_f'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_f(
        _n,
        df1,
        df2,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_gamma(n, a, b, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a gamma
    distribution.

    ``dist_gamma`` generates a vector of pseudorandom numbers taken from
    a gamma distribution with parameters a and b.

    For full information please refer to the NAG Library document for
    g05sj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05sjf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the parameter of the gamma distribution.

    b : float
        b, the parameter of the gamma distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified gamma
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_gamma as b_dist_gamma

    fun_name = 'naginterfaces.library.rand.dist_gamma'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_gamma(
        _n,
        a,
        b,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_normal(n, xmu, var, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a Normal
    distribution.

    ``dist_normal`` generates a vector of pseudorandom numbers taken
    from a Normal (Gaussian) distribution with mean mu and variance
    sigma^2.

    For full information please refer to the NAG Library document for
    g05sk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05skf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmu : float
        mu, the mean of the distribution.

    var : float
        sigma^2, the variance of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified Normal
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_normal as b_dist_normal

    fun_name = 'naginterfaces.library.rand.dist_normal'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_normal(
        _n,
        xmu,
        var,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_logistic(n, a, b, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a logistic
    distribution.

    ``dist_logistic`` generates a vector of pseudorandom numbers from a
    logistic distribution with mean a and spread b.

    For full information please refer to the NAG Library document for
    g05sl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05slf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the mean of the distribution.

    b : float
        b, the spread of the distribution, where 'spread' is
        (sqrt(3))/pi*standard deviation.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified logistic
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_logistic as b_dist_logistic

    fun_name = 'naginterfaces.library.rand.dist_logistic'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_logistic(
        _n,
        a,
        b,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_lognormal(n, xmu, var, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a log-normal
    distribution.

    ``dist_lognormal`` generates a vector of pseudorandom numbers from a
    log-normal distribution with parameters mu and sigma^2.

    For full information please refer to the NAG Library document for
    g05sm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05smf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmu : float
        mu, the mean of the distribution of ln(x).

    var : float
        sigma^2, the variance of the distribution of ln(x).

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified log-normal
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `xmu` is too large to take the exponential of
            `xmu` = *<value>*.

        (`errno` 3)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_lognormal as b_dist_lognormal

    fun_name = 'naginterfaces.library.rand.dist_lognormal'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_lognormal(
        _n,
        xmu,
        var,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_students_t(n, df, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a Student's
    t-distribution.

    ``dist_students_t`` generates a vector of pseudorandom numbers taken
    from a Student's t-distribution with nu degrees of freedom.

    For full information please refer to the NAG Library document for
    g05sn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05snf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    df : int
        nu, the number of degrees of freedom of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified Student's
        t-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_students_t as b_dist_students_t

    fun_name = 'naginterfaces.library.rand.dist_students_t'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_students_t(
        _n,
        df,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_triangular(n, xmin, xmed, xmax, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a triangular
    distribution.

    ``dist_triangular`` generates a vector of pseudorandom numbers from
    a triangular distribution with parameters x_min, x_med and x_max.

    For full information please refer to the NAG Library document for
    g05sp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05spf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmin : float
        The end point x_min of the triangular distribution.

    xmed : float
        The median of the distribution x_med (also the location of the
        vertex of the triangular distribution at which the PDF reaches a
        maximum).

    xmax : float
        The end point x_max of the triangular distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified triangular
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `xmed` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmed` >= `xmin`.

        (`errno` 4)
            On entry, `xmax` = *<value>* and `xmed` = *<value>*.

            Constraint: `xmax` >= `xmed`.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_triangular as b_dist_triangular

    fun_name = 'naginterfaces.library.rand.dist_triangular'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_triangular(
        _n,
        xmin,
        xmed,
        xmax,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_uniform(n, a, b, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a uniform
    distribution over [a,b].

    ``dist_uniform`` generates a vector of pseudorandom numbers
    uniformly distributed over the interval [a,b].

    For full information please refer to the NAG Library document for
    g05sq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05sqf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        The end points a and b of the uniform distribution.

    b : float
        The end points a and b of the uniform distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified uniform
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` >= `a`.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_uniform as b_dist_uniform

    fun_name = 'naginterfaces.library.rand.dist_uniform'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_uniform(
        _n,
        a,
        b,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_vonmises(n, vk, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a von Mises
    distribution.

    ``dist_vonmises`` generates a vector of pseudorandom numbers from a
    von Mises distribution with concentration parameter kappa.

    For full information please refer to the NAG Library document for
    g05sr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05srf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    vk : float
        kappa, the concentration parameter of the required von Mises
        distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified von Mises
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `vk` <= 0.0 or `vk` too large: `vk` = *<value>*.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_vonmises as b_dist_vonmises

    fun_name = 'naginterfaces.library.rand.dist_vonmises'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_vonmises(
        _n,
        vk,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def dist_weibull(n, a, b, statecomm):
    u"""
    Generates a vector of pseudorandom numbers from a Weibull
    distribution.

    ``dist_weibull`` generates a vector of pseudorandom numbers from a
    two parameter Weibull distribution with shape parameter a and scale
    parameter b.

    For full information please refer to the NAG Library document for
    g05ss

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ssf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the shape parameter of the distribution.

    b : float
        b, the scale parameter of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : float, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified Weibull
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import dist_weibull as b_dist_weibull

    fun_name = 'naginterfaces.library.rand.dist_weibull'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_dist_weibull(
        _n,
        a,
        b,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_binomial(mode, n, m, p, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a binomial
    distribution.

    ``int_binomial`` generates a vector of pseudorandom integers from
    the discrete binomial distribution with parameters m and p.

    For full information please refer to the NAG Library document for
    g05ta

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05taf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_binomial``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    m : int
        m, the number of trials of the distribution.

    p : float
        p, the probability of success of the binomial distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_binomial``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified binomial
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, `p` = *<value>*.

            Constraint: 0.0 <= `p` <= 1.0.

        (`errno` 5)
            `p` or `m` is not the same as when `comm`['r'] was set up in
            a previous call.

            Previous value of `p` = *<value>* and `p` = *<value>*.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 5)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 7)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_binomial as b_int_binomial

    fun_name = 'naginterfaces.library.rand.int_binomial'

    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_binomial(
        _mode,
        _n,
        m,
        p,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def logical(n, p, statecomm):
    u"""
    Generates a vector of pseudorandom logical values.

    ``logical`` generates a vector of pseudorandom logical values --
    True with probability p and False with probability (1-p).

    For full information please refer to the NAG Library document for
    g05tb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tbf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom logical values to be generated.

    p : float
        Must contain the probability of ``logical`` returning True.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : bool, ndarray, shape (`n`)
        The n logical values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: 0.0 <= `p` <= 1.0.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import logical as b_logical

    fun_name = 'naginterfaces.library.rand.logical'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineLogicalArrayType.empty_ndarray(
        (_n_py,),
    )
    b_logical(
        _n,
        p,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_geom(mode, n, p, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a geometric
    distribution.

    ``int_geom`` generates a vector of pseudorandom integers from the
    discrete geometric distribution with probability p of success at a
    trial.

    For full information please refer to the NAG Library document for
    g05tc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tcf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_geom``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    p : float
        The parameter p of the geometric distribution representing the
        probability of success at a single trial.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_geom``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified geometric
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `p` = *<value>*.

            Constraint: machine precision <= `p` <= 1.0.

        (`errno` 3)
            `p` is so small that lr would have to be larger than the
            largest representable integer. Use `mode` = 3 instead. `p` =
            *<value>*

        (`errno` 4)
            `p` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `p` = *<value>* and `p` = *<value>*.

        (`errno` 4)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_geom as b_int_geom

    fun_name = 'naginterfaces.library.rand.int_geom'

    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_geom(
        _mode,
        _n,
        p,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_general(mode, n, p, ip1, itype, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a general discrete
    distribution.

    ``int_general`` generates a vector of pseudorandom integers from a
    discrete distribution with a given PDF (probability density
    function) or CDF (cumulative distribution function) p.

    For full information please refer to the NAG Library document for
    g05td

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tdf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_general``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    p : float, array-like, shape (np)
        The PDF or CDF of the distribution.

    ip1 : int
        The value of the variate, a whole number, to which the
        probability in `p`[0] corresponds.

    itype : int
        Indicates the type of information contained in `p`.

        `itype` = 1

          `p` contains a probability distribution function (PDF).

        `itype` = 2

          `p` contains a cumulative distribution function (CDF).

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_general``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        Contains n pseudorandom numbers from the specified discrete
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, at least one element of the vector `p` is less
            than 0.0 or greater than 1.0.

        (`errno` 3)
            On entry, `itype` = 1 and the sum of the elements of `p` do
            not equal one.

        (`errno` 3)
            On entry, `itype` = 2 and the values of `p` are not all in
            stricly ascending order.

        (`errno` 3)
            On entry, `p`[np-1] = *<value>*.

            Constraint: if `itype` = 2, `p`[np-1] = 1.0.

        (`errno` 4)
            On entry, np = *<value>*.

            Constraint: np > 0.

        (`errno` 6)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1 or 2.

        (`errno` 7)
            The value of np or `ip1` is not the same as when `comm`['r']
            was set up in a previous call.

            Previous value of np = *<value>* and np = *<value>*.

            Previous value of `ip1` = *<value>* and `ip1` = *<value>*.

        (`errno` 7)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_general as b_int_general

    fun_name = 'naginterfaces.library.rand.int_general'

    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _np_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _np = _b_utils._EngineIntScalarType(
        _np_py, {'fun_name': fun_name, 'entity': '_np'},
    ).to_ctypes()
    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_general(
        _mode,
        _n,
        _p,
        _np,
        ip1,
        itype,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_hypergeom(mode, n, ns, np, m, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a hypergeometric
    distribution.

    ``int_hypergeom`` generates a vector of pseudorandom integers from
    the discrete hypergeometric distribution of the number of specified
    items in a sample of size l, taken from a population of size k with
    m specified items in it.

    For full information please refer to the NAG Library document for
    g05te

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tef.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_hypergeom``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    ns : int
        l, the sample size of the hypergeometric distribution.

    np : int
        k, the population size of the hypergeometric distribution.

    m : int
        m, the number of specified items of the hypergeometric
        distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_hypergeom``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The pseudorandom numbers from the specified hypergeometric
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `ns` = *<value>* and `np` = *<value>*.

            Constraint: 0 <= `ns` <= `np`.

        (`errno` 4)
            On entry, `np` = *<value>*.

            Constraint: `np` >= 0.

        (`errno` 5)
            On entry, `m` = *<value>* and `np` = *<value>*.

            Constraint: 0 <= `m` <= `np`.

        (`errno` 6)
            The value of `ns`, `np` or `m` is not the same as when
            `comm`['r'] was set up in a previous call with `mode` = 0 or
            2.

        (`errno` 6)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_hypergeom as b_int_hypergeom

    fun_name = 'naginterfaces.library.rand.int_hypergeom'

    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_hypergeom(
        _mode,
        _n,
        ns,
        np,
        m,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_log(mode, n, a, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a logarithmic
    distribution.

    ``int_log`` generates a vector of pseudorandom integers from the
    discrete logarithmic distribution with parameter a.

    For full information please refer to the NAG Library document for
    g05tf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tff.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_log``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the parameter of the logarithmic distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_log``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified logarithmic
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `a` = *<value>*.

            Constraint: 0.0 < `a` < 1.0.

        (`errno` 4)
            The value of `a` is not the same as when `comm`['r'] was set
            up in a previous call.

            Previous value of `a` = *<value>* and `a` = *<value>*.

        (`errno` 4)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_log as b_int_log

    fun_name = 'naginterfaces.library.rand.int_log'

    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_log(
        _mode,
        _n,
        a,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_multinomial(mode, n, m, p, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a multinomial
    distribution.

    ``int_multinomial`` generates a sequence of n variates, each
    consisting of k pseudorandom integers, from the discrete multinomial
    distribution with k outcomes and m trials, where the outcomes have
    probabilities p_1,p_2,...,p_k respectively.

    For full information please refer to the NAG Library document for
    g05tg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tgf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_multinomial``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    m : int
        m, the number of trials of the multinomial distribution.

    p : float, array-like, shape (k)
        Contains the probabilities p_j, for j = 1,2,...,k, of the k
        possible outcomes of the multinomial distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_multinomial``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`, k)
        The first n rows of `x`[i-1,j-1] each contain k pseudorandom
        numbers representing a k-dimensional variate from the specified
        multinomial distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, k = *<value>*.

            Constraint: k >= 2.

        (`errno` 5)
            On entry, at least one element of the vector `p` is less
            than 0.0 or greater than 1.0.

        (`errno` 5)
            On entry, the sum of the elements of `p` do not equal one.

        (`errno` 6)
            The value of `m` or k is not the same as when `comm`['r']
            was set up in a previous call.

            Previous value of `m` = *<value>* and `m` = *<value>*.

            Previous value of k = *<value>* and k = *<value>*.

        (`errno` 6)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_multinomial as b_int_multinomial

    fun_name = 'naginterfaces.library.rand.int_multinomial'

    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py, _k_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_int_multinomial(
        _mode,
        _n,
        m,
        _k,
        _p,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_negbin(mode, n, m, p, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a negative binomial
    distribution.

    ``int_negbin`` generates a vector of pseudorandom integers from the
    discrete negative binomial distribution with parameter m and
    probability p of success at a trial.

    For full information please refer to the NAG Library document for
    g05th

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05thf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_negbin``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    m : int
        m, the number of failures of the distribution.

    p : float
        p, the parameter of the negative binomial distribution
        representing the probability of success at a single trial.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_negbin``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified negative binomial
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, `p` = *<value>*.

            Constraint: 0.0 <= `p` < 1.0.

        (`errno` 5)
            `p` or `m` is not the same as when `comm`['r'] was set up in
            a previous call.

            Previous value of `p` = *<value>* and `p` = *<value>*.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 5)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 7)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_negbin as b_int_negbin

    fun_name = 'naginterfaces.library.rand.int_negbin'

    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_negbin(
        _mode,
        _n,
        m,
        p,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_poisson(mode, n, lamda, statecomm, comm=None):
    u"""
    Generates a vector of pseudorandom integers from a Poisson
    distribution.

    ``int_poisson`` generates a vector of pseudorandom integers from the
    discrete Poisson distribution with mean lambda.

    For full information please refer to the NAG Library document for
    g05tj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tjf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

          Set up reference vector only.

        `mode` = 1

          Generate variates using reference vector set up in a prior
          call to ``int_poisson``.

        `mode` = 2

          Set up reference vector and generate variates.

        `mode` = 3

          Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    lamda : float
        lambda, the mean of the Poisson distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_poisson``.

        If `mode` = 3, `comm` is not referenced and may be None.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified Poisson
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `lamda` = *<value>*.

            Constraint: `lamda` >= 0.0.

        (`errno` 3)
            `lamda` is such that lr would have to be larger than the
            largest representable integer. Use `mode` = 3 instead.
            `lamda` = *<value>*.

        (`errno` 4)
            `lamda` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `lamda` = *<value>* and `lamda` =
            *<value>*.

        (`errno` 4)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_poisson as b_int_poisson

    fun_name = 'naginterfaces.library.rand.int_poisson'

    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_poisson(
        _mode,
        _n,
        lamda,
        comm,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_poisson_varmean(vlamda, statecomm):
    u"""
    Generates a vector of pseudorandom integers from a Poisson
    distribution with varying mean.

    ``int_poisson_varmean`` generates a vector of pseudorandom integers,
    each from a discrete Poisson distribution with differing parameter.

    For full information please refer to the NAG Library document for
    g05tk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tkf.html

    Parameters
    ----------
    vlamda : float, array-like, shape (m)
        The means, lambda_j, for j = 1,2,...,m, of the Poisson
        distributions.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : int, ndarray, shape (m)
        The m pseudorandom numbers from the specified Poisson
        distributions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 2)
            On entry, at least one element of `vlamda` is less than
            zero.

        (`errno` 2)
            On entry, at least one element of `vlamda` is too large.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_poisson_varmean as b_int_poisson_varmean

    fun_name = 'naginterfaces.library.rand.int_poisson_varmean'

    _vlamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'vlamda'},
        vlamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_vlamda, {'fun_name': fun_name, 'entity': 'vlamda'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m_py,),
    )
    b_int_poisson_varmean(
        _m,
        _vlamda,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def int_uniform(n, a, b, statecomm):
    u"""
    Generates a vector of pseudorandom integers from a uniform
    distribution.

    ``int_uniform`` generates a vector of pseudorandom integers
    uniformly distributed over the interval [a,b].

    For full information please refer to the NAG Library document for
    g05tl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05tlf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : int
        The end points a and b of the uniform distribution.

    b : int
        The end points a and b of the uniform distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    x : int, ndarray, shape (`n`)
        The n pseudorandom numbers from the specified uniform
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` >= `a`.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import int_uniform as b_int_uniform

    fun_name = 'naginterfaces.library.rand.int_uniform'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    b_int_uniform(
        _n,
        a,
        b,
        statecomm,
        _x,
    )
    _x_py = _x
    return _x_py

def bb_init(t0, tend, times):
    u"""
    Initializes the Brownian bridge generator.

    ``bb_init`` initializes the Brownian bridge generator ``bb``.
    It must be called before any calls to ``bb``.

    For full information please refer to the NAG Library document for
    g05xa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05xaf.html

    Parameters
    ----------
    t0 : float
        The starting value t_0 of the time interval.

    tend : float
        The end value T of the time interval.

    times : float, array-like, shape (ntimes)
        The points in the time interval (t_0,T) at which the Wiener
        process is to be constructed. The order in which points are
        listed in `times` determines the bridge construction order. The
        function ``bb_make_bridge_order`` can be used to create
        predefined bridge construction orders from a set of input times.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of
            ``bb_init``. Please contact NAG with the following error
            message: error in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `tend` = *<value>* and `t0` = *<value>*.

            Constraint: `tend` > `t0`.

        (`errno` 2)
            On entry, ntimes = *<value>*.

            Constraint: ntimes >= 1.

        (`errno` 3)
            On entry, `times`[*<value>*] = *<value>*, `t0` = *<value>*
            and `tend` = *<value>*.

            Constraint: `t0` < `times`[i-1] < `tend` for all i.

        (`errno` 4)
            On entry, `times`[*<value>*] and `times`[*<value>*] both
            equal *<value>*.

            Constraint: all elements of `times` must be unique.
    """

    from ..base.rand import bb_init as b_bb_init

    fun_name = 'naginterfaces.library.rand.bb_init'

    _times = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'times'},
        times,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ntimes_py = _b_utils._EngineFloat64ArrayType.get_shape(_times, {'fun_name': fun_name, 'entity': 'times'}, exp_rank=1)[0]
    _ntimes = _b_utils._EngineIntScalarType(
        _ntimes_py, {'fun_name': fun_name, 'entity': '_ntimes'},
    ).to_ctypes()
    _return_scalar = b_bb_init(
        t0,
        tend,
        _times,
        _ntimes,
    )
    _comm_py = _return_scalar
    return _comm_py

def bb(npaths, start, term, z, c, comm, rcord=2, a=1):
    u"""
    Generate paths for a free or non-free Wiener process using the
    Brownian bridge algorithm.

    ``bb`` uses a Brownian bridge algorithm to construct sample paths
    for a free or non-free Wiener process.
    The initialization function ``bb_init`` must be called prior to the
    first call to ``bb``.

    For full information please refer to the NAG Library document for
    g05xb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05xbf.html

    Parameters
    ----------
    npaths : int
        The number of Wiener sample paths to create.

    start : float, array-like, shape (d)
        The starting value of the Wiener process.

    term : float, array-like, shape (d)
        The terminal value at which the non-free Wiener process should
        end. If `a` = 0, `term` is ignored.

    z : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rcord`=1: d*(N+1-`a`); if `rcord`=2:
        `npaths`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rcord`=1: `npaths`; if `rcord`=2:
        d*(N+1-`a`); otherwise: 0.

        The Normal random numbers used to construct the sample paths.

        If quasi-random numbers are used, the d*(N+1-`a`)-dimensional
        quasi-random points should be stored in successive rows of Z.

    c : float, array-like, shape (d, d)
        The lower triangular Cholesky factorization C such that CC^T
        gives the covariance matrix of the Wiener process. Elements of C
        above the diagonal are not referenced.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bb_init``.

    rcord : int, optional
        The order in which Normal random numbers are stored in `z` and
        in which the generated values are returned in `b`.

    a : int, optional
        If `a` = 0, a free Wiener process is created beginning at
        `start` and `term` is ignored.

        If `a` = 1, a non-free Wiener process is created beginning at
        `start` and ending at `term`.

    Returns
    -------
    z : float, ndarray, shape (:, :)
        The Normal random numbers premultiplied by C.

    b : float, ndarray, shape (:, :)
        The values of the Wiener sample paths.

        Let X_p,i^k denote the kth dimension of the ith point of the pth
        sample path where 1 <= k <= d, 1 <= i <= N+1 and 1 <= p <=
        `npaths`.

        The point X_p,i^k is stored at B(p,k+(i-1)*d).

        The starting value `start` is never stored, whereas the terminal
        value is always stored.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of ``bb``.
            Please contact NAG with the following error message: error
            in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `comm`['rcomm'] was not initialized or has been
            corrupted.

        (`errno` 2)
            On entry, `npaths` = *<value>*.

            Constraint: `npaths` >= 1.

        (`errno` 3)
            On entry, the value of `rcord` is invalid.

        (`errno` 4)
            On entry, d = *<value>*.

            Constraint: d >= 1.

        (`errno` 5)
            On entry, `a` = *<value>*.

            Constraint: `a` = 0 or 1.

        (`errno` 6)
            On entry, ldz = *<value>* and d*(ntimes+1-`a`) = *<value>*.

            Constraint: ldz >= d*(ntimes+1-`a`).
    """

    from ..base.rand import bb as b_bb

    fun_name = 'naginterfaces.library.rand.bb'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_c, 'c')),
    )
    _start = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'start'},
        start,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d_py = _b_utils._EngineFloat64ArrayType.get_shape(_start, {'fun_name': fun_name, 'entity': 'start'}, exp_rank=1)[0]
    _d = _b_utils._EngineIntScalarType(
        _d_py, {'fun_name': fun_name, 'entity': '_d'},
    ).to_ctypes()
    _term = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'term'},
        term,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('rcomm',),
    )
    _vu_n_py = (int(_b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)+0.5) if _b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False) > 0 else int(_b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)-0.5))
    _rcord = _b_utils._EngineIntScalarType(
        rcord, {'fun_name': fun_name, 'entity': 'rcord'},
    ).to_ctypes()
    _rcord_py = _b_utils._EngineIntScalarType.to_py(_rcord)
    _npaths = _b_utils._EngineIntScalarType(
        npaths, {'fun_name': fun_name, 'entity': 'npaths'},
    ).to_ctypes()
    _npaths_py = _b_utils._EngineIntScalarType.to_py(_npaths)
    if _rcord_py == 1:
        _b_dim1 = _d_py*(_vu_n_py+1)
    elif _rcord_py == 2:
        _b_dim1 = _npaths_py
    else:
        _b_dim1 = 0
    if _rcord_py == 1:
        _b_dim2 = _npaths_py
    elif _rcord_py == 2:
        _b_dim2 = _d_py*(_vu_n_py+1)
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_b,
        sorder=_m_sorder,
    )
    b_bb(
        _npaths,
        _rcord,
        _d,
        _start,
        a,
        _term,
        _z,
        _c,
        _b,
        comm,
    )
    _z_py = _z
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['z', 'b'],
    )(
        _z_py,
        _b_py,
    )
    return _return_data

def bb_inc_init(t0, tend, times):
    u"""
    Initializes the generator which backs out the increments of sample
    paths generated by a Brownian bridge algorithm.

    ``bb_inc_init`` initializes the Brownian bridge increments generator
    ``bb_inc``.
    It must be called before any calls to ``bb_inc``.

    For full information please refer to the NAG Library document for
    g05xc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05xcf.html

    Parameters
    ----------
    t0 : float
        The starting value t_0 of the time interval.

    tend : float
        The end value T of the time interval.

    times : float, array-like, shape (ntimes)
        The points in the time interval (t_0,T) at which the Wiener
        process is to be constructed. The order in which points are
        listed in `times` determines the bridge construction order. The
        function ``bb_make_bridge_order`` can be used to create
        predefined bridge construction orders from a set of input times.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of
            ``bb_inc_init``. Please contact NAG with the following error
            message: error in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `tend` = *<value>* and `t0` = *<value>*.

            Constraint: `tend` > `t0`.

        (`errno` 2)
            On entry, ntimes = *<value>*.

            Constraint: ntimes >= 1.

        (`errno` 3)
            On entry, `times`[*<value>*] = *<value>*, `t0` = *<value>*
            and `tend` = *<value>*.

            Constraint: `t0` < `times`[i-1] < `tend` for all i.

        (`errno` 4)
            On entry, `times`[i-1] = `times`[j-1] = *<value>*, for i =
            *<value>* and j = *<value>*.

            Constraint: all elements of `times` must be unique.

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    from ..base.rand import bb_inc_init as b_bb_inc_init

    fun_name = 'naginterfaces.library.rand.bb_inc_init'

    _times = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'times'},
        times,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ntimes_py = _b_utils._EngineFloat64ArrayType.get_shape(_times, {'fun_name': fun_name, 'entity': 'times'}, exp_rank=1)[0]
    _ntimes = _b_utils._EngineIntScalarType(
        _ntimes_py, {'fun_name': fun_name, 'entity': '_ntimes'},
    ).to_ctypes()
    _return_scalar = b_bb_inc_init(
        t0,
        tend,
        _times,
        _ntimes,
    )
    _comm_py = _return_scalar
    return _comm_py

def bb_inc(npaths, diff, z, c, comm, rcord=2, a=1):
    u"""
    Backs out the increments from sample paths generated by a Brownian
    bridge algorithm.

    ``bb_inc`` computes scaled increments of sample paths of a free or
    non-free Wiener process, where the sample paths are constructed by a
    Brownian bridge algorithm.
    The initialization function ``bb_inc_init`` must be called prior to
    the first call to ``bb_inc``.

    For full information please refer to the NAG Library document for
    g05xd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05xdf.html

    Parameters
    ----------
    npaths : int
        The number of Wiener sample paths.

    diff : float, array-like, shape (d)
        The difference between the terminal value and starting value of
        the Wiener process. If `a` = 0, `diff` is ignored.

    z : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rcord`=1: d*(N+1-`a`); if `rcord`=2:
        `npaths`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rcord`=1: `npaths`; if `rcord`=2:
        d*(N+1-`a`); otherwise: 0.

        The Normal random numbers used to construct the sample paths.

        If quasi-random numbers are used, the d*(N+1-`a`)-dimensional
        quasi-random points should be stored in successive rows of Z.

    c : float, array-like, shape (d, d)
        The lower triangular Cholesky factorization C such that CC^T
        gives the covariance matrix of the Wiener process. Elements of C
        above the diagonal are not referenced.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bb_inc_init``.

    rcord : int, optional
        The order in which Normal random numbers are stored in `z` and
        in which the generated values are returned in `b`.

    a : int, optional
        If `a` = 0, a free Wiener process is created and `diff` is
        ignored.

        If `a` = 1, a non-free Wiener process is created where `diff` is
        the difference between the terminal value and the starting value
        of the process.

    Returns
    -------
    z : float, ndarray, shape (:, :)
        The Normal random numbers premultiplied by `c`.

    b : float, ndarray, shape (:, :)
        The scaled Wiener increments.

        Let X_p,i^k denote the kth dimension of the ith point of the pth
        sample path where 1 <= k <= d, 1 <= i <= N+1 and 1 <= p <=
        `npaths`.

        The increment ((X_p,i^k-X_p,i-1^k))/((t_i-t_{i-1})) is stored at
        B(p,k+(i-1)*d).

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of ``bb_inc``.
            Please contact NAG with the following error message: error
            in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `comm`['rcomm'] was not initialized or has been
            corrupted.

        (`errno` 2)
            On entry, `npaths` = *<value>*.

            Constraint: `npaths` >= 1.

        (`errno` 3)
            On entry, the value of `rcord` is invalid.

        (`errno` 4)
            On entry, d = *<value>*.

            Constraint: d >= 1.

        (`errno` 5)
            On entry, `a` = *<value>*.

            Constraint: `a` = 0 or 1.

        (`errno` 6)
            On entry, ldz = *<value>* and d*(ntimes+1-`a`) = *<value>*.

            Constraint: ldz >= d*(ntimes+1-`a`).

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    from ..base.rand import bb_inc as b_bb_inc

    fun_name = 'naginterfaces.library.rand.bb_inc'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_c, 'c')),
    )
    _diff = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'diff'},
        diff,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d_py = _b_utils._EngineFloat64ArrayType.get_shape(_diff, {'fun_name': fun_name, 'entity': 'diff'}, exp_rank=1)[0]
    _d = _b_utils._EngineIntScalarType(
        _d_py, {'fun_name': fun_name, 'entity': '_d'},
    ).to_ctypes()
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('rcomm',),
    )
    _vu_n_py = (int(_b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)+0.5) if _b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False) > 0 else int(_b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)-0.5))
    _rcord = _b_utils._EngineIntScalarType(
        rcord, {'fun_name': fun_name, 'entity': 'rcord'},
    ).to_ctypes()
    _rcord_py = _b_utils._EngineIntScalarType.to_py(_rcord)
    _npaths = _b_utils._EngineIntScalarType(
        npaths, {'fun_name': fun_name, 'entity': 'npaths'},
    ).to_ctypes()
    _npaths_py = _b_utils._EngineIntScalarType.to_py(_npaths)
    if _rcord_py == 1:
        _b_dim1 = _d_py*(_vu_n_py+1)
    elif _rcord_py == 2:
        _b_dim1 = _npaths_py
    else:
        _b_dim1 = 0
    if _rcord_py == 1:
        _b_dim2 = _npaths_py
    elif _rcord_py == 2:
        _b_dim2 = _d_py*(_vu_n_py+1)
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_b,
        sorder=_m_sorder,
    )
    b_bb_inc(
        _npaths,
        _rcord,
        _d,
        a,
        _diff,
        _z,
        _c,
        _b,
        comm,
    )
    _z_py = _z
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['z', 'b'],
    )(
        _z_py,
        _b_py,
    )
    return _return_data

def bb_make_bridge_order(t0, tend, intime, bgord=1, move=None):
    u"""
    Creates a Brownian bridge construction order out of a set of input
    times.

    ``bb_make_bridge_order`` takes a set of input times and permutes
    them to specify one of several predefined Brownian bridge
    construction orders.
    The permuted times can be passed to ``bb_init`` or ``bb_inc_init``
    to initialize the Brownian bridge generators with the chosen bridge
    construction order.

    For full information please refer to the NAG Library document for
    g05xe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05xef.html

    Parameters
    ----------
    t0 : float
        t_0, the start value of the time interval on which the Wiener
        process is to be constructed.

    tend : float
        T, the largest time at which the Wiener process is to be
        constructed.

    intime : float, array-like, shape (ntimes)
        The time points, t_1,t_2,...,t_N, at which the Wiener process is
        to be constructed. Note that the final time T is not included in
        this array.

    bgord : int, optional
        The bridge construction order to use.

    move : None or int, array-like, shape (nmove), optional
        The indices of the entries in `intime` which should be moved to
        the front of the `times` array, with `move`[j-1] = i setting the
        jth element of `times` to t_i. Note that i ranges from 1 to
        ntimes. When nmove = 0, `move` is not referenced.

    Returns
    -------
    times : float, ndarray, shape (ntimes)
        The output bridge construction order. This should be passed to
        ``bb_init`` or ``bb_inc_init``.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of
            ``bb_make_bridge_order``. Please contact NAG with the
            following error message: error in *<value>*, ASSERT =
            *<value>*.

        (`errno` 1)
            On entry, `bgord` = *<value>*.

            Constraint: `bgord` = 1, 2, 3 or 4

        (`errno` 2)
            On entry, ntimes = *<value>*.

            Constraint: ntimes >= 1.

        (`errno` 3)
            On entry, nmove = *<value>* and ntimes = *<value>*.

            Constraint: 0 <= nmove <= ntimes.

        (`errno` 4)
            On entry, `intime`[0] = *<value>* and `t0` = *<value>*.

            Constraint: `intime`[0] > `t0`.

        (`errno` 4)
            On entry, `intime`[*<value>*] = *<value>* and
            `intime`[*<value>*] = *<value>*.

            Constraint: the elements in `intime` must be in increasing
            order.

        (`errno` 4)
            On entry, ntimes = *<value>*, `intime`[ntimes-1] = *<value>*
            and `tend` = *<value>*.

            Constraint: `intime`[ntimes-1] < `tend`.

        (`errno` 5)
            On entry, `move`[*<value>*] = *<value>*.

            Constraint: `move`[i] >= 1 for all i.

        (`errno` 5)
            On entry, `move`[*<value>*] = *<value>* and ntimes =
            *<value>*.

            Constraint: `move`[i] <= ntimes for all i.

        (`errno` 6)
            On entry, `move`[*<value>*] and `move`[*<value>*] both equal
            *<value>*.

            Constraint: all elements in `move` must be unique.

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    from ..base.rand import bb_make_bridge_order as b_bb_make_bridge_order

    fun_name = 'naginterfaces.library.rand.bb_make_bridge_order'

    _intime = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'intime'},
        intime,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ntimes_py = _b_utils._EngineFloat64ArrayType.get_shape(_intime, {'fun_name': fun_name, 'entity': 'intime'}, exp_rank=1)[0]
    _ntimes = _b_utils._EngineIntScalarType(
        _ntimes_py, {'fun_name': fun_name, 'entity': '_ntimes'},
    ).to_ctypes()
    _move = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'move'},
        move,
        copy=False,
        allow_none=True,
    )
    if move is None:
        _nmove = 0
        _nmove_py = _nmove
    else:
        _nmove_py = _b_utils._EngineIntArrayType.get_shape(_move, {'fun_name': fun_name, 'entity': 'move'}, exp_rank=1, allow_none=True)[0]
        _nmove = _b_utils._EngineIntScalarType(
            _nmove_py, {'fun_name': fun_name, 'entity': '_nmove'},
        ).to_ctypes()
    _times = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ntimes_py,),
    )
    b_bb_make_bridge_order(
        bgord,
        t0,
        tend,
        _ntimes,
        _intime,
        _nmove,
        _move,
        _times,
    )
    _times_py = _times
    return _times_py

def quasi_normal(xmean, std, n, comm):
    u"""
    Generates a Normal quasi-random number sequence.

    ``quasi_normal`` generates a quasi-random sequence from a Normal
    (Gaussian) distribution.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05yj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05yjf.html

    Parameters
    ----------
    xmean : float, array-like, shape (idim)
        Specifies, for each dimension, the mean of the Normal
        distribution.

    std : float, array-like, shape (idim)
        Specifies, for each dimension, the standard deviation of the
        Normal distribution.

    n : int
        The number of quasi-random numbers required.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Returns
    -------
    quas : float, ndarray, shape (`n`, idim)
        Contains the `n` quasi-random numbers of dimension idim

        If sorder = 2, `quas`[i-1,j-1] holds the jth value for the ith
        dimension.

        If sorder = 1, `quas`[i-1,j-1] holds the ith value for the jth
        dimension.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, i = *<value>* and `std`[i-1] = *<value>*.

            Constraint: `std`[i-1] >= 0.0.

        (`errno` 4)
            There have been too many calls to the generator.

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    from ..base.rand import quasi_normal as b_quasi_normal

    fun_name = 'naginterfaces.library.rand.quasi_normal'

    _xmean = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmean'},
        xmean,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _std = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'std'},
        std,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = _b_utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    _quas = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _vl_idim_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_quasi_normal(
        _xmean,
        _std,
        _n,
        _quas,
        comm,
    )
    _quas_py = _quas
    return _quas_py

def quasi_lognormal(xmean, std, n, comm):
    u"""
    Generates a log-normal quasi-random number sequence.

    ``quasi_lognormal`` generates a quasi-random sequence from a
    log-normal distribution.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05yk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ykf.html

    Parameters
    ----------
    xmean : float, array-like, shape (idim)
        Specifies, for each dimension, the mean of the underlying Normal
        distribution.

    std : float, array-like, shape (idim)
        Specifies, for each dimension, the standard deviation of the
        underlying Normal distribution.

    n : int
        The number of quasi-random numbers required.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Returns
    -------
    quas : float, ndarray, shape (`n`, idim)
        Contains the `n` quasi-random numbers of dimension idim.

        If sorder = 2, `quas`[i-1,j-1] holds the jth value for the ith
        dimension.

        If sorder = 1, `quas`[i-1,j-1] holds the ith value for the jth
        dimension.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, i = *<value>* and `std`[i-1] = *<value>*.

            Constraint: `std`[i-1] >= 0.

        (`errno` 4)
            There have been too many calls to the generator.

        (`errno` 5)
            On entry, i = *<value>* and `xmean`[i-1] = *<value>*.

            Constraint: abs(`xmean`[i-1]) <= *<value>*.
    """

    from ..base.rand import quasi_lognormal as b_quasi_lognormal

    fun_name = 'naginterfaces.library.rand.quasi_lognormal'

    _xmean = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmean'},
        xmean,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _std = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'std'},
        std,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = _b_utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    _quas = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _vl_idim_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_quasi_lognormal(
        _xmean,
        _std,
        _n,
        _quas,
        comm,
    )
    _quas_py = _quas
    return _quas_py

def quasi_init(genid, idim, iskip):
    u"""
    Initializes a quasi-random number generator.

    ``quasi_init`` initializes a quasi-random generator prior to calling
    ``quasi_uniform``, ``quasi_normal`` or ``quasi_lognormal``.

    For full information please refer to the NAG Library document for
    g05yl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ylf.html

    Parameters
    ----------
    genid : int
        Must identify the quasi-random generator to use.

        `genid` = 1

          Sobol generator.

        `genid` = 2

          Sobol (A659) generator.

        `genid` = 3

          Niederreiter generator.

        `genid` = 4

          Faure generator.

    idim : int
        The number of dimensions required.

    iskip : int
        The number of terms of the sequence to skip on initialization
        for the Sobol and Niederreiter generators. If `genid` = 4,
        `iskip` is ignored.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: `genid` = 1, 2, 3 or 4.

        (`errno` 2)
            On entry, `idim` = *<value>*.

            Constraint: 1 <= `idim` <= *<value>*.

        (`errno` 5)
            On entry, `iskip` < 0 or `iskip` is too large: `iskip` =
            *<value>*, maximum value is *<value>*.
    """

    from ..base.rand import quasi_init as b_quasi_init

    _return_scalar = b_quasi_init(
        genid,
        idim,
        iskip,
    )
    _comm_py = _return_scalar
    return _comm_py

def quasi_uniform(n, comm, rcord=1):
    u"""
    Generates a uniform quasi-random number sequence.

    ``quasi_uniform`` generates a uniformly distributed low-discrepancy
    sequence as proposed by Sobol, Faure or Niederreiter.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05ym

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ymf.html

    Parameters
    ----------
    n : int
        The number of quasi-random numbers required.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    rcord : int, optional
        The order in which the generated values are returned.

    Returns
    -------
    quas : float, ndarray, shape (:, :)
        Contains the `n` quasi-random numbers of dimension idim.

        If `rcord` = 1, `quas`[i-1,j-1] holds the jth value for the ith
        dimension.

        If `rcord` = 2, `quas`[i-1,j-1] holds the ith value for the jth
        dimension.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, value of `n` would result in too many calls to the
            generator: `n` = *<value>*, generator has previously been
            called *<value>* times.

        (`errno` 2)
            On entry, `rcord` = *<value>*.

            Constraint: `rcord` = 1 or 2.

        (`errno` 4)
            On entry, ldquas = *<value>* and `n` = *<value>*.

            Constraint: if `rcord` = 2, ldquas >= `n`.

        (`errno` 5)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.
    """

    from ..base.rand import quasi_uniform as b_quasi_uniform

    fun_name = 'naginterfaces.library.rand.quasi_uniform'

    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = _b_utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    _rcord = _b_utils._EngineIntScalarType(
        rcord, {'fun_name': fun_name, 'entity': 'rcord'},
    ).to_ctypes()
    _rcord_py = _b_utils._EngineIntScalarType.to_py(_rcord)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if _rcord_py == 1:
        _quas_dim1 = _vl_idim_py
    elif _rcord_py == 2:
        _quas_dim1 = _n_py
    else:
        _quas_dim1 = 0
    if _rcord_py == 1:
        _quas_dim2 = _n_py
    elif _rcord_py == 2:
        _quas_dim2 = _vl_idim_py
    else:
        _quas_dim2 = 0
    exp_shape_quas = (_quas_dim1, _quas_dim2)
    _quas = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_quas,
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_quasi_uniform(
        _n,
        _rcord,
        _quas,
        comm,
    )
    _quas_py = _quas
    return _quas_py

def quasi_init_scrambled(genid, stype, idim, iskip, nsdigi, statecomm):
    u"""
    Initializes a scrambled quasi-random number generator.

    ``quasi_init_scrambled`` initializes a scrambled quasi-random
    generator prior to calling ``quasi_uniform``, ``quasi_normal`` or
    ``quasi_lognormal``.
    It must be preceded by a call to one of the pseudorandom
    initialization functions ``init_repeat`` or ``init_nonrepeat``.

    For full information please refer to the NAG Library document for
    g05yn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ynf.html

    Parameters
    ----------
    genid : int
        Must identify the quasi-random generator to use.

        `genid` = 1

          Sobol generator.

        `genid` = 2

          Sobol (A659) generator.

        `genid` = 3

          Niederreiter generator.

    stype : int
        Must identify the scrambling method to use.

        `stype` = 0

          No scrambling. This is equivalent to calling ``quasi_init``.

        `stype` = 1

          Owen like scrambling.

        `stype` = 2

          Faure--Tezuka scrambling.

        `stype` = 3

          Owen and Faure--Tezuka scrambling.

    idim : int
        The number of dimensions required.

    iskip : int
        The number of terms of the sequence to skip on initialization
        for the Sobol and Niederreiter generators.

    nsdigi : int
        Controls the number of digits (bits) to scramble when `genid` =
        1 or 2, otherwise `nsdigi` is ignored. If `nsdigi` < 1 or
        `nsdigi` > 30 then all the digits are scrambled.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: 1 <= `genid` <= *<value>*.

        (`errno` 2)
            On entry, `stype` = *<value>*.

            Constraint: 0 <= `stype` <= 3.

        (`errno` 3)
            On entry, `idim` = *<value>*.

            Constraint: 1 <= `idim` <= *<value>*.

        (`errno` 6)
            On entry, `iskip` = *<value>*.

            Constraint: 0 <= `iskip` <= 2^30.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    from ..base.rand import quasi_init_scrambled as b_quasi_init_scrambled

    _return_scalar = b_quasi_init_scrambled(
        genid,
        stype,
        idim,
        iskip,
        nsdigi,
        statecomm,
    )
    _comm_py = _return_scalar
    return _comm_py

def field_1d_user_setup(ns, xmin, xmax, var, cov1, maxm=None, pad=1, icorr=0, data=None):
    u"""
    Setup for simulating one-dimensional random fields, user-defined
    variogram.

    ``field_1d_user_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in one dimension, for a
    user-defined variogram, using the circulant embedding method.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_1d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05zmf.html

    Parameters
    ----------
    ns : int
        The number of sample points to be generated in realizations of
        the random field.

    xmin : float
        The lower bound for the interval over which the random field is
        to be simulated.

    xmax : float
        The upper bound for the interval over which the random field is
        to be simulated.

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    cov1 : callable gamma = cov1(x, data=None)
        `cov1` must evaluate the variogram gamma(x), without the
        multiplicative factor sigma^2, for all x >= 0.

        The value returned in `gamma` is multiplied internally by `var`.

        Parameters
        ~~~~~~~~~~
        x : float
            The value x at which the variogram gamma(x) is to be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        gamma : float
            The value of the variogram gamma(x)/sigma^2.

    maxm : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 2^(3+ceiling(log2((`ns`-1)))).

        The maximum size of the circulant matrix to use. For example, if
        the embedding matrix is to be allowed to double in size three
        times before the approximation procedure is used, then choose
        `maxm` = 2^(k+2) where k = 1+⌈log2(`ns`-1)⌉.

    pad : int, optional
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

          The embedding matrix is padded with zeros.

        `pad` = 1

          The embedding matrix is padded with values of the variogram.

    icorr : int, optional
        Determines which approximation to implement if required, as
        described in [Description].

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    lam : float, ndarray, shape (`maxm`)
        Contains the square roots of the eigenvalues of the embedding
        matrix.

    xx : float, ndarray, shape (`ns`)
        The points at which values of the random field will be output.

    m : int
        The size of the embedding matrix.

    approx : int
        Indicates whether approximation was used.

        `approx` = 0

          No approximation was used.

        `approx` = 1

          Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    eig : float, ndarray, shape (3)
        Indicates information about the negative eigenvalues in the
        embedding matrix which have had to be set to zero. `eig`[0]
        contains the smallest eigenvalue, `eig`[1] contains the sum of
        the squares of the negative eigenvalues, and `eig`[2] contains
        the sum of the absolute values of the negative eigenvalues.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `maxm` = *<value>*.

            Constraint: the minimum calculated value for `maxm` is
            *<value>*.

        (`errno` 5)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 7)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 8)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.
    """

    from ..base.rand import field_1d_user_setup as b_field_1d_user_setup

    def cb_cov1(x, data=None):
        """Wrapping user's cov1. Called as base cov1."""
        if data is not None:
            _gamma_cov1 = cov1(
                x,
                data,
            )
        else:
            _gamma_cov1 = cov1(
                x,
            )
        _gamma_cov1_base = _gamma_cov1
        return _gamma_cov1_base

    fun_name = 'naginterfaces.library.rand.field_1d_user_setup'

    _ns = _b_utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ns_py = _b_utils._EngineIntScalarType.to_py(_ns)
    _maxm = 2**3+int(ceil(log(float((_ns_py-1))) / log(2.))) if maxm is None else maxm
    _b_utils._check_callable(cov1, {'fun_name': fun_name, 'entity': 'cov1'})
    _cov1 = cb_cov1
    _maxm = _b_utils._EngineIntScalarType(
        _maxm, {'fun_name': fun_name, 'entity': '_maxm'},
    ).to_ctypes()
    _maxm_py = _b_utils._EngineIntScalarType.to_py(_maxm)
    _lam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxm_py,),
    )
    _xx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py,),
    )
    _eig = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _return_scalars = b_field_1d_user_setup(
        _ns,
        xmin,
        xmax,
        _maxm,
        var,
        _cov1,
        pad,
        icorr,
        _lam,
        _xx,
        _eig,
        data,
    )
    _lam_py = _lam
    _xx_py = _xx
    _m_py = _return_scalars.m
    _approx_py = _return_scalars.approx
    _rho_py = _return_scalars.rho
    _icount_py = _return_scalars.icount
    _eig_py = _eig
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lam', 'xx', 'm', 'approx', 'rho', 'icount', 'eig'],
    )(
        _lam_py,
        _xx_py,
        _m_py,
        _approx_py,
        _rho_py,
        _icount_py,
        _eig_py,
    )
    return _return_data

def field_1d_predef_setup(ns, xmin, xmax, var, icov1, params, maxm=None, pad=1, icorr=0):
    u"""
    Setup for simulating one-dimensional random fields.

    ``field_1d_predef_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in one dimension, for a
    preset variogram, using the circulant embedding method.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_1d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05znf.html

    Parameters
    ----------
    ns : int
        The number of sample points to be generated in realizations of
        the random field.

    xmin : float
        The lower bound for the interval over which the random field is
        to be simulated. Note that if `icov1` = 14 (for simulating
        fractional Brownian motion), `xmin` is not referenced and the
        lower bound for the interval is set to zero.

    xmax : float
        The upper bound for the interval over which the random field is
        to be simulated. Note that if `icov1` = 14 (for simulating
        fractional Brownian motion), the lower bound for the interval is
        set to zero and so `xmax` is required to be greater than zero.

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    icov1 : int
        Determines which of the preset variograms to use. The choices
        are given below. Note that x' = (|x|)/l, where l is the
        correlation length and is a parameter for most of the
        variograms, and sigma^2 is the variance specified by `var`.

        `icov1` = 1

          Symmetric stable variogram

              gamma(x) = sigma^2exp(-(x')^nu),

          where

          - l = `params`[0], l > 0,

          - nu = `params`[1], 0 <= nu <= 2.

        `icov1` = 2

          Cauchy variogram

              gamma(x) = sigma^2(1+(x')^2)^(-nu),

          where

          - l = `params`[0], l > 0,

          - nu = `params`[1], nu > 0.

        `icov1` = 3

          Differential variogram with compact support

              gamma(x) = {[table omitted]

          where

          - l = `params`[0], l > 0.

        `icov1` = 4

          Exponential variogram

              gamma(x) = sigma^2exp(-x'),

          where

          - l = `params`[0], l > 0.

        `icov1` = 5

          Gaussian variogram

              gamma(x) = sigma^2exp(-(x')^2),

          where

          - l = `params`[0], l > 0.

        `icov1` = 6

          Nugget variogram

              gamma(x) = {[table omitted]

          No parameters need be set for this value of `icov1`.

        `icov1` = 7

          Spherical variogram

              gamma(x) = {[table omitted]

          where

          - l = `params`[0], l > 0.

        `icov1` = 8

          Bessel variogram

              gamma(x) = sigma^22^nuGamma(nu+1)J_nu(x')/((x')^nu),

          where

          - J_nu(·) is the Bessel function of the first kind,

          - l = `params`[0], l > 0,

          - nu = `params`[1], nu >= -0.5.

        `icov1` = 9

          Hole effect variogram

              gamma(x) = sigma^2(sin(x'))/(x'),

          where

          - l = `params`[0], l > 0.

        `icov1` = 10

          Whittle-Matérn variogram

              gamma(x) = sigma^22^(1-nu)(x')^nu K_nu(x')/Gamma(nu),

          where

          - K_nu(·) is the modified Bessel function of the second kind,

          - l = `params`[0], l > 0,

          - nu = `params`[1], nu > 0.

        `icov1` = 11

          Continuously parameterised variogram with compact support

              gamma(x) = {[table omitted]

          where

          - x^('') = (x')/s,

          - K_nu(·) is the modified Bessel function of the second kind,

          - l = `params`[0], l > 0,

          - s = `params`[1], s > 0 (second correlation length),

          - nu = `params`[2], nu > 0.

        `icov1` = 12

          Generalized hyperbolic distribution variogram

              gamma(x) =
              sigma^2((delta^2+(x')^2)^(lambda/2))/(delta^lambda
              K_lambda(kappadelta))K_lambda(kappa(delta^2+(x')^2)^(1/2)),

          where

          - K_lambda(·) is the modified Bessel function of the second
            kind,

          - l = `params`[0], l > 0,

          - lambda = `params`[1], no constraint on lambda

          - delta = `params`[2], delta > 0,

          - kappa = `params`[3], kappa > 0.

        `icov1` = 13

          Cosine variogram

              gamma(x) = sigma^2cos(x'),

          where

          - l = `params`[0], l > 0.

        `icov1` = 14

          Used for simulating fractional Brownian motion B^H(t).
          Fractional Brownian motion itself is not a stationary Gaussian
          random field, but its increments X~(i) = B^H(t_i)-B^H(t_{i-1})
          can be simulated in the same way as a stationary random field.
          The variogram for the so-called 'increment process' is

              C(X~(t_i),X~(t_j)) = gamma~(x) =
              (delta^(2H))/2(|x/delta-1|^(2H)+|x/delta+1|^(2H)-2|x/delta|^(2H)),

          where

          - x = t_j-t_i,

          - H = `params`[0], 0 < H < 1, H is the Hurst parameter,

          - delta = `params`[1], delta > 0, normally delta = t_i-t_{i-1}
            is the (fixed) step size.

          We scale the increments to set gamma(0) = 1; let X(i) =
          X~(i)/(delta^(-H)), then

              C(X(t_i),X(t_j)) = gamma(x) =
              1/2(|x/delta-1|^(2H)+|x/delta+1|^(2H)-2|x/delta|^(2H)).

          The increments X(i) can then be simulated using
          ``field_1d_generate``, then multiplied by delta^H to obtain
          the original increments X~(i) for the fractional Brownian
          motion.

    params : float, array-like, shape (np)
        The parameters set for the variogram.

    maxm : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 2^(3+ceiling(log2((`ns`-1)))).

        The maximum size of the circulant matrix to use. For example, if
        the embedding matrix is to be allowed to double in size three
        times before the approximation procedure is used, then choose
        `maxm` = 2^(k+2) where k = 1+⌈log2(`ns`-1)⌉.

    pad : int, optional
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

          The embedding matrix is padded with zeros.

        `pad` = 1

          The embedding matrix is padded with values of the variogram.

    icorr : int, optional
        Determines which approximation to implement if required, as
        described in [Description].

    Returns
    -------
    lam : float, ndarray, shape (`maxm`)
        Contains the square roots of the eigenvalues of the embedding
        matrix.

    xx : float, ndarray, shape (`ns`)
        The points at which values of the random field will be output.

    m : int
        The size of the embedding matrix.

    approx : int
        Indicates whether approximation was used.

        `approx` = 0

          No approximation was used.

        `approx` = 1

          Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    eig : float, ndarray, shape (3)
        Indicates information about the negative eigenvalues in the
        embedding matrix which have had to be set to zero. `eig`[0]
        contains the smallest eigenvalue, `eig`[1] contains the sum of
        the squares of the negative eigenvalues, and `eig`[2] contains
        the sum of the absolute values of the negative eigenvalues.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `icov1` != 14, `xmin` = *<value>* and `xmax` =
            *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 3)
            On entry, `icov1` = 14 and `xmax` = *<value>*.

            Constraint: `xmax` > 0.0.

        (`errno` 4)
            On entry, `maxm` = *<value>*.

            Constraint: the minimum calculated value for `maxm` is
            *<value>*.

        (`errno` 5)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 6)
            On entry, `icov1` = *<value>*.

            Constraint: `icov1` >= 1 and `icov1` <= 14.

        (`errno` 7)
            On entry, np = *<value>*.

            Constraint: for `icov1` = *<value>*, np = *<value>*.

        (`errno` 8)
            On entry, `params`[*<value>*] = *<value>*.

            Constraint: dependent on `icov1`.

        (`errno` 9)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 10)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.
    """

    from ..base.rand import field_1d_predef_setup as b_field_1d_predef_setup

    fun_name = 'naginterfaces.library.rand.field_1d_predef_setup'

    _ns = _b_utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ns_py = _b_utils._EngineIntScalarType.to_py(_ns)
    _maxm = 2**3+int(ceil(log(float((_ns_py-1))) / log(2.))) if maxm is None else maxm
    _params = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'params'},
        params,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _np_py = _b_utils._EngineFloat64ArrayType.get_shape(_params, {'fun_name': fun_name, 'entity': 'params'}, exp_rank=1)[0]
    _np = _b_utils._EngineIntScalarType(
        _np_py, {'fun_name': fun_name, 'entity': '_np'},
    ).to_ctypes()
    _maxm = _b_utils._EngineIntScalarType(
        _maxm, {'fun_name': fun_name, 'entity': '_maxm'},
    ).to_ctypes()
    _maxm_py = _b_utils._EngineIntScalarType.to_py(_maxm)
    _lam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxm_py,),
    )
    _xx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py,),
    )
    _eig = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _return_scalars = b_field_1d_predef_setup(
        _ns,
        xmin,
        xmax,
        _maxm,
        var,
        icov1,
        _np,
        _params,
        pad,
        icorr,
        _lam,
        _xx,
        _eig,
    )
    _lam_py = _lam
    _xx_py = _xx
    _m_py = _return_scalars.m
    _approx_py = _return_scalars.approx
    _rho_py = _return_scalars.rho
    _icount_py = _return_scalars.icount
    _eig_py = _eig
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lam', 'xx', 'm', 'approx', 'rho', 'icount', 'eig'],
    )(
        _lam_py,
        _xx_py,
        _m_py,
        _approx_py,
        _rho_py,
        _icount_py,
        _eig_py,
    )
    return _return_data

def field_1d_generate(ns, s, lam, rho, statecomm):
    u"""
    Generates realizations of a one-dimensional random field.

    ``field_1d_generate`` produces realizations of a stationary Gaussian
    random field in one dimension, using the circulant embedding method.
    The square roots of the eigenvalues of the extended covariance
    matrix (or embedding matrix) need to be input, and can be calculated
    using ``field_1d_user_setup`` or ``field_1d_predef_setup``.

    For full information please refer to the NAG Library document for
    g05zp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05zpf.html

    Parameters
    ----------
    ns : int
        The number of sample points to be generated in realizations of
        the random field. This must be the same value as supplied to
        ``field_1d_predef_setup`` or ``field_1d_user_setup`` when
        calculating the eigenvalues of the embedding matrix.

    s : int
        S, the number of realizations of the random field to simulate.

    lam : float, array-like, shape (m)
        Must contain the square roots of the eigenvalues of the
        embedding matrix, as returned by ``field_1d_predef_setup`` or
        ``field_1d_user_setup``.

    rho : float
        Indicates the scaling of the covariance matrix, as returned by
        ``field_1d_predef_setup`` or ``field_1d_user_setup``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    z : float, ndarray, shape (`ns`, `s`)
        Contains the realizations of the random field. The jth
        realization, for the `ns` sample points, is stored in
        `z`[i-1,j-1], for i = 1,2,...,`ns`. The sample points are as
        returned in xx by ``field_1d_predef_setup`` or
        ``field_1d_user_setup``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 1.

        (`errno` 3)
            On entry, m = *<value>* and `ns` = *<value>*.

            Constraint: m >= max(1,2*(`ns`-1)).

        (`errno` 4)
            On entry, at least one element of `lam` was negative.

            Constraint: all elements of `lam` must be non-negative.

        (`errno` 5)
            On entry, `rho` = *<value>*.

            Constraint: 0.0 <= `rho` <= 1.0.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import field_1d_generate as b_field_1d_generate

    fun_name = 'naginterfaces.library.rand.field_1d_generate'

    _lam = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lam'},
        lam,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_lam, {'fun_name': fun_name, 'entity': 'lam'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _ns = _b_utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ns_py = _b_utils._EngineIntScalarType.to_py(_ns)
    _s = _b_utils._EngineIntScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _s_py = _b_utils._EngineIntScalarType.to_py(_s)
    _z = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _s_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_field_1d_generate(
        _ns,
        _s,
        _m,
        _lam,
        rho,
        statecomm,
        _z,
    )
    _z_py = _z
    return _z_py

def field_2d_user_setup(ns, xmin, xmax, ymin, ymax, maxm, var, cov2, even, pad=1, icorr=0, data=None):
    u"""
    Setup for simulating two-dimensional random fields, user-defined
    variogram.

    ``field_2d_user_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in two dimensions, for a
    user-defined variogram, using the circulant embedding method.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_2d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05zqf.html

    Parameters
    ----------
    ns : int, array-like, shape (2)
        The number of sample points to use in each direction, with
        `ns`[0] sample points in the x-direction, N_1 and `ns`[1] sample
        points in the y-direction, N_2. The total number of sample
        points on the grid is therefore `ns`[0]*`ns`[1].

    xmin : float
        The lower bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    xmax : float
        The upper bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    ymin : float
        The lower bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    ymax : float
        The upper bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    maxm : int, array-like, shape (2)
        Determines the maximum size of the circulant matrix to use -- a
        maximum of `maxm`[0] elements in the x-direction, and a maximum
        of `maxm`[1] elements in the y-direction. The maximum size of
        the circulant matrix is thus `maxm`[0]*`maxm`[1].

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    cov2 : callable gamma = cov2(x, y, data=None)
        `cov2` must evaluate the variogram gamma(x) for all x if `even`
        = 0, and for all x with non-negative entries if `even` = 1.

        The value returned in `gamma` is multiplied internally by `var`.

        Parameters
        ~~~~~~~~~~
        x : float
            The coordinate x at which the variogram gamma(x) is to be
            evaluated.

        y : float
            The coordinate y at which the variogram gamma(x) is to be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        gamma : float
            The value of the variogram gamma(x).

    even : int
        Indicates whether the covariance function supplied is even or
        uneven.

        `even` = 0

          The covariance function is uneven.

        `even` = 1

          The covariance function is even.

    pad : int, optional
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

          The embedding matrix is padded with zeros.

        `pad` = 1

          The embedding matrix is padded with values of the variogram.

    icorr : int, optional
        Determines which approximation to implement if required, as
        described in [Description].

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    lam : float, ndarray, shape (:)
        Contains the square roots of the eigenvalues of the embedding
        matrix.

    xx : float, ndarray, shape (:)
        The points of the x-coordinates at which values of the random
        field will be output.

    yy : float, ndarray, shape (:)
        The points of the y-coordinates at which values of the random
        field will be output.

    m : int, ndarray, shape (2)
        `m`[0] contains M_1, the size of the circulant blocks and `m`[1]
        contains M_2, the number of blocks, resulting in a final square
        matrix of size M_1*M_2.

    approx : int
        Indicates whether approximation was used.

        `approx` = 0

          No approximation was used.

        `approx` = 1

          Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    eig : float, ndarray, shape (3)
        Indicates information about the negative eigenvalues in the
        embedding matrix which have had to be set to zero. `eig`[0]
        contains the smallest eigenvalue, `eig`[1] contains the sum of
        the squares of the negative eigenvalues, and `eig`[2] contains
        the sum of the absolute values of the negative eigenvalues.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = [*<value>*,*<value>*].

            Constraint: `ns`[0] >= 1, `ns`[1] >= 1.

        (`errno` 2)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymin` < `ymax`.

        (`errno` 6)
            On entry, `maxm` = [*<value>*,*<value>*].

            Constraint: the minima for `maxm` are [*<value>*,*<value>*].

        (`errno` 7)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 9)
            On entry, `even` = *<value>*.

            Constraint: `even` = 0 or 1.

        (`errno` 10)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 11)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.
    """

    from ..base.rand import field_2d_user_setup as b_field_2d_user_setup

    def cb_cov2(x, y, data=None):
        """Wrapping user's cov2. Called as base cov2."""
        if data is not None:
            _gamma_cov2 = cov2(
                x,
                y,
                data,
            )
        else:
            _gamma_cov2 = cov2(
                x,
                y,
            )
        _gamma_cov2_base = _gamma_cov2
        return _gamma_cov2_base

    fun_name = 'naginterfaces.library.rand.field_2d_user_setup'

    _ns = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ns'},
        ns,
        copy=False,
    )
    _maxm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'maxm'},
        maxm,
        copy=False,
    )
    _b_utils._check_callable(cov2, {'fun_name': fun_name, 'entity': 'cov2'})
    _cov2 = cb_cov2
    _lam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (0,))*_b_utils._EngineIntArrayType.get_element(_maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (1,)),),
    )
    _xx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_ns, {'fun_name': fun_name, 'entity': 'ns'}, (0,)),),
    )
    _yy = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_ns, {'fun_name': fun_name, 'entity': 'ns'}, (1,)),),
    )
    _m = _b_utils._EngineIntArrayType.empty_ndarray(
        (2,),
    )
    _eig = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _return_scalars = b_field_2d_user_setup(
        _ns,
        xmin,
        xmax,
        ymin,
        ymax,
        _maxm,
        var,
        _cov2,
        even,
        pad,
        icorr,
        _lam,
        _xx,
        _yy,
        _m,
        _eig,
        data,
    )
    _lam_py = _lam
    _xx_py = _xx
    _yy_py = _yy
    _m_py = _m
    _approx_py = _return_scalars.approx
    _rho_py = _return_scalars.rho
    _icount_py = _return_scalars.icount
    _eig_py = _eig
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lam', 'xx', 'yy', 'm', 'approx', 'rho', 'icount', 'eig'],
    )(
        _lam_py,
        _xx_py,
        _yy_py,
        _m_py,
        _approx_py,
        _rho_py,
        _icount_py,
        _eig_py,
    )
    return _return_data

def field_2d_predef_setup(ns, xmin, xmax, ymin, ymax, maxm, var, icov2, params, norm=2, pad=1, icorr=0):
    u"""
    Setup for simulating two-dimensional random fields, preset
    variogram.

    ``field_2d_predef_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in two dimensions, for a
    preset variogram, using the circulant embedding method.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_2d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05zrf.html

    Parameters
    ----------
    ns : int, array-like, shape (2)
        The number of sample points to use in each direction, with
        `ns`[0] sample points in the x-direction, N_1 and `ns`[1] sample
        points in the y-direction, N_2. The total number of sample
        points on the grid is therefore `ns`[0]*`ns`[1].

    xmin : float
        The lower bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    xmax : float
        The upper bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    ymin : float
        The lower bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    ymax : float
        The upper bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    maxm : int, array-like, shape (2)
        Determines the maximum size of the circulant matrix to use -- a
        maximum of `maxm`[0] elements in the x-direction, and a maximum
        of `maxm`[1] elements in the y-direction. The maximum size of
        the circulant matrix is thus `maxm`[0]*`maxm`[1].

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    icov2 : int
        Determines which of the preset variograms to use. The choices
        are given below. Note that x' = ||x/(l_1),y/(l_2)||, where l_1
        and l_2 are correlation lengths in the x and y directions
        respectively and are parameters for most of the variograms, and
        sigma^2 is the variance specified by `var`.

        `icov2` = 1

          Symmetric stable variogram

              gamma(x) = sigma^2exp(-(x')^nu),

          where

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0,

          - nu = `params`[2], 0 < nu <= 2.

        `icov2` = 2

          Cauchy variogram

              gamma(x) = sigma^2(1+(x')^2)^(-nu),

          where

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0,

          - nu = `params`[2], nu > 0.

        `icov2` = 3

          Differential variogram with compact support

              gamma(x) = {[table omitted]

          where

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0.

        `icov2` = 4

          Exponential variogram

              gamma(x) = sigma^2exp(-x'),

          where

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0.

        `icov2` = 5

          Gaussian variogram

              gamma(x) = sigma^2exp(-(x')^2),

          where

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0.

        `icov2` = 6

          Nugget variogram

              gamma(x) = {[table omitted]

          No parameters need be set for this value of `icov2`.

        `icov2` = 7

          Spherical variogram

              gamma(x) = {[table omitted]

          where

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0.

        `icov2` = 8

          Bessel variogram

              gamma(x) = sigma^22^nuGamma(nu+1)J_nu(x')/((x')^nu),

          where

          - J_nu(·) is the Bessel function of the first kind,

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0,

          - nu = `params`[2], nu >= 0.

        `icov2` = 9

          Hole effect variogram

              gamma(x) = sigma^2(sin(x'))/(x'),

          where

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0.

        `icov2` = 10

          Whittle-Matérn variogram

              gamma(x) = sigma^22^(1-nu)(x')^nu K_nu(x')/Gamma(nu),

          where

          - K_nu(·) is the modified Bessel function of the second kind,

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0,

          - nu = `params`[2], nu > 0.

        `icov2` = 11

          Continuously parameterised variogram with compact support

              gamma(x) = {[table omitted]

          where

          - x^′′ = ||(x')/l_1 s_1,(y')/l_2 s_2||,

          - K_nu(·) is the modified Bessel function of the second kind,

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0,

          - s_1 = `params`[2], s_1 > 0,

          - s_2 = `params`[3], s_2 > 0,

          - nu = `params`[4], nu > 0.

        `icov2` = 12

          Generalized hyperbolic distribution variogram

              gamma(x) =
              sigma^2((delta^2+(x')^2)^(lambda/2))/(delta^lambda
              K_lambda(kappadelta))K_lambda(kappa(delta^2+(x')^2)^(1/2)),

          where

          - K_lambda(·) is the modified Bessel function of the second
            kind,

          - l_1 = `params`[0], l_1 > 0,

          - l_2 = `params`[1], l_2 > 0,

          - lambda = `params`[2], no constraint on lambda,

          - delta = `params`[3], delta > 0,

          - kappa = `params`[4], kappa > 0.

    params : float, array-like, shape (np)
        The parameters for the variogram as detailed in the description
        of `icov2`.

    norm : int, optional
        Determines which norm to use when calculating the variogram.

        `norm` = 1

          The 1-norm is used, i.e., ||x,y|| = |x|+|y|.

        `norm` = 2

          The 2-norm (Euclidean norm) is used, i.e., ||x,y|| =
          sqrt(x^2+y^2).

    pad : int, optional
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

          The embedding matrix is padded with zeros.

        `pad` = 1

          The embedding matrix is padded with values of the variogram.

    icorr : int, optional
        Determines which approximation to implement if required, as
        described in [Description].

    Returns
    -------
    lam : float, ndarray, shape (:)
        Contains the square roots of the eigenvalues of the embedding
        matrix.

    xx : float, ndarray, shape (:)
        The points of the x-coordinates at which values of the random
        field will be output.

    yy : float, ndarray, shape (:)
        The points of the y-coordinates at which values of the random
        field will be output.

    m : int, ndarray, shape (2)
        `m`[0] contains M_1, the size of the circulant blocks and `m`[1]
        contains M_2, the number of blocks, resulting in a final square
        matrix of size M_1*M_2.

    approx : int
        Indicates whether approximation was used.

        `approx` = 0

          No approximation was used.

        `approx` = 1

          Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    eig : float, ndarray, shape (3)
        Indicates information about the negative eigenvalues in the
        embedding matrix which have had to be set to zero. `eig`[0]
        contains the smallest eigenvalue, `eig`[1] contains the sum of
        the squares of the negative eigenvalues, and `eig`[2] contains
        the sum of the absolute values of the negative eigenvalues.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = [*<value>*,*<value>*].

            Constraint: `ns`[0] >= 1, `ns`[1] >= 1.

        (`errno` 2)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymin` < `ymax`.

        (`errno` 6)
            On entry, `maxm` = [*<value>*,*<value>*].

            Constraint: the minimum calculated value for `maxm` are
            [*<value>*,*<value>*].

        (`errno` 7)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 8)
            On entry, `icov2` = *<value>*.

            Constraint: `icov2` >= 1 and `icov2` <= 12.

        (`errno` 9)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 1 or 2.

        (`errno` 10)
            On entry, np = *<value>*.

            Constraint: for `icov2` = *<value>*, np = *<value>*.

        (`errno` 11)
            On entry, `params`[*<value>*] = *<value>*.

            Constraint: dependent on `icov2`, see documentation.

        (`errno` 12)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 13)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.
    """

    from ..base.rand import field_2d_predef_setup as b_field_2d_predef_setup

    fun_name = 'naginterfaces.library.rand.field_2d_predef_setup'

    _ns = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ns'},
        ns,
        copy=False,
    )
    _maxm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'maxm'},
        maxm,
        copy=False,
    )
    _params = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'params'},
        params,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _np_py = _b_utils._EngineFloat64ArrayType.get_shape(_params, {'fun_name': fun_name, 'entity': 'params'}, exp_rank=1)[0]
    _np = _b_utils._EngineIntScalarType(
        _np_py, {'fun_name': fun_name, 'entity': '_np'},
    ).to_ctypes()
    _lam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (0,))*_b_utils._EngineIntArrayType.get_element(_maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (1,)),),
    )
    _xx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_ns, {'fun_name': fun_name, 'entity': 'ns'}, (0,)),),
    )
    _yy = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_ns, {'fun_name': fun_name, 'entity': 'ns'}, (1,)),),
    )
    _m = _b_utils._EngineIntArrayType.empty_ndarray(
        (2,),
    )
    _eig = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _return_scalars = b_field_2d_predef_setup(
        _ns,
        xmin,
        xmax,
        ymin,
        ymax,
        _maxm,
        var,
        icov2,
        norm,
        _np,
        _params,
        pad,
        icorr,
        _lam,
        _xx,
        _yy,
        _m,
        _eig,
    )
    _lam_py = _lam
    _xx_py = _xx
    _yy_py = _yy
    _m_py = _m
    _approx_py = _return_scalars.approx
    _rho_py = _return_scalars.rho
    _icount_py = _return_scalars.icount
    _eig_py = _eig
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lam', 'xx', 'yy', 'm', 'approx', 'rho', 'icount', 'eig'],
    )(
        _lam_py,
        _xx_py,
        _yy_py,
        _m_py,
        _approx_py,
        _rho_py,
        _icount_py,
        _eig_py,
    )
    return _return_data

def field_2d_generate(ns, s, m, lam, rho, statecomm):
    u"""
    Generates realizations of a two-dimensional random field.

    ``field_2d_generate`` produces realizations of a stationary Gaussian
    random field in two dimensions, using the circulant embedding
    method.
    The square roots of the eigenvalues of the extended covariance
    matrix (or embedding matrix) need to be input, and can be calculated
    using ``field_2d_user_setup`` or ``field_2d_predef_setup``.

    For full information please refer to the NAG Library document for
    g05zs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05zsf.html

    Parameters
    ----------
    ns : int, array-like, shape (2)
        The number of sample points to use in each direction, with
        `ns`[0] sample points in the x-direction and `ns`[1] sample
        points in the y-direction. The total number of sample points on
        the grid is therefore `ns`[0]*`ns`[1]. This must be the same
        value as supplied to ``field_2d_user_setup`` or
        ``field_2d_predef_setup`` when calculating the eigenvalues of
        the embedding matrix.

    s : int
        S, the number of realizations of the random field to simulate.

    m : int, array-like, shape (2)
        Indicates the size, M, of the embedding matrix as returned by
        ``field_2d_user_setup`` or ``field_2d_predef_setup``. The
        embedding matrix is a block circulant matrix with circulant
        blocks. `m`[0] is the size of each block, and `m`[1] is the
        number of blocks.

    lam : float, array-like, shape (`m`[0]*`m`[1])
        Contains the square roots of the eigenvalues of the embedding
        matrix, as returned by ``field_2d_user_setup`` or
        ``field_2d_predef_setup``.

    rho : float
        Indicates the scaling of the covariance matrix, as returned by
        ``field_2d_user_setup`` or ``field_2d_predef_setup``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    z : float, ndarray, shape (:, `s`)
        Contains the realizations of the random field. The kth
        realization (where k = 1,2,...,`s`) of the random field on the
        two-dimensional grid (x_i,y_j) is stored in `z`[ (j-1)* `ns`[0]
        +i -1,k-1], for i = 1,2,...,`ns`[0] and for j = 1,2,...,`ns`[1].
        The points are returned in xx and yy by ``field_2d_user_setup``
        or ``field_2d_predef_setup``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = [*<value>*,*<value>*].

            Constraint: `ns`[0] >= 1, `ns`[1] >= 1.

        (`errno` 2)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 1.

        (`errno` 3)
            On entry, `m` = [*<value>*,*<value>*], and `ns` =
            [*<value>*,*<value>*].

            Constraints: `m`[i-1] >= max(1, 2*(`ns`[i-1])-1), for i =
            1,2.

        (`errno` 4)
            On entry, at least one element of `lam` was negative.

            Constraint: all elements of `lam` must be non-negative.

        (`errno` 5)
            On entry, `rho` = *<value>*.

            Constraint: 0.0 < `rho` <= 1.0.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import field_2d_generate as b_field_2d_generate

    fun_name = 'naginterfaces.library.rand.field_2d_generate'

    _ns = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ns'},
        ns,
        copy=False,
    )
    _m = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'm'},
        m,
        copy=False,
    )
    _lam = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lam'},
        lam,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _b_utils._EngineIntScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _s_py = _b_utils._EngineIntScalarType.to_py(_s)
    _z = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_ns, {'fun_name': fun_name, 'entity': 'ns'}, (0,))*_b_utils._EngineIntArrayType.get_element(_ns, {'fun_name': fun_name, 'entity': 'ns'}, (1,)), _s_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    b_field_2d_generate(
        _ns,
        _s,
        _m,
        _lam,
        rho,
        statecomm,
        _z,
    )
    _z_py = _z
    return _z_py

def field_fracbm_generate(ns, s, xmax, h, lam, rho, statecomm):
    u"""
    Generates realizations of fractional Brownian motion.

    ``field_fracbm_generate`` produces realizations of a fractional
    Brownian motion, using the circulant embedding method.
    The square roots of the eigenvalues of the extended covariance
    matrix (or embedding matrix) need to be input, and can be calculated
    using ``field_1d_predef_setup``.

    For full information please refer to the NAG Library document for
    g05zt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g05/g05ztf.html

    Parameters
    ----------
    ns : int
        The number of steps (points) to be generated in realizations of
        the increments of the fractional Brownian motion. This must be
        the same value as supplied to ``field_1d_predef_setup`` when
        calculating the eigenvalues of the embedding matrix.

        Note: in the context of fractional Brownian motion, `ns`
        represents the number of steps from a zero starting state.
        Realizations returned in `z` include this starting state and so
        `ns`+1 values are returned for each realization.

    s : int
        S, the number of realizations of the fractional Brownian motion
        to simulate.

    xmax : float
        The upper bound for the interval over which the fractional
        Brownian motion is to be simulated, as input to
        ``field_1d_user_setup`` or ``field_1d_predef_setup``.

    h : float
        The Hurst parameter, H, for the fractional Brownian motion. This
        must be the same value as supplied to ``field_1d_predef_setup``
        in params[0], when the eigenvalues of the embedding matrix were
        calculated.

    lam : float, array-like, shape (m)
        Contains the square roots of the eigenvalues of the embedding
        matrix, as returned by ``field_1d_user_setup`` or
        ``field_1d_predef_setup``.

    rho : float
        Indicates the scaling of the covariance matrix, as returned by
        ``field_1d_user_setup`` or ``field_1d_predef_setup``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    z : float, ndarray, shape (`ns`+1, `s`)
        Contains the realizations of the fractional Brownian motion, Z.
        The jth realization, for the ith point `xx`[i-1], is stored in
        `z`[i-1,j-1], for i = 1,2,...,`ns`+1, for j = 1,2,...,`s`.

    xx : float, ndarray, shape (`ns`+1)
        The points at which values of the fractional Brownian motion are
        output. The first point is always zero, and the subsequent `ns`
        points represent the equispaced steps towards the last point,
        `xmax`. Note that in ``field_1d_user_setup`` and
        ``field_1d_predef_setup``, the returned ns sample points are the
        mid-points of the grid returned in `xx` here.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 1.

        (`errno` 3)
            On entry, m = *<value>*, and `ns` = *<value>*.

            Constraint: m >= max(1, 2*(`ns`-1)).

        (`errno` 4)
            On entry, `xmax` = *<value>*.

            Constraint: `xmax` > 0.0.

        (`errno` 5)
            On entry, `h` = *<value>*.

            Constraint: 0.0 < `h` < 1.0.

        (`errno` 6)
            On entry, at least one element of `lam` was negative.

            Constraint: all elements of `lam` must be non-negative.

        (`errno` 7)
            On entry, `rho` = *<value>*.

            Constraint: 0.0 < `rho` <= 1.0.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.
    """

    from ..base.rand import field_fracbm_generate as b_field_fracbm_generate

    fun_name = 'naginterfaces.library.rand.field_fracbm_generate'

    _lam = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lam'},
        lam,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_lam, {'fun_name': fun_name, 'entity': 'lam'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _ns = _b_utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ns_py = _b_utils._EngineIntScalarType.to_py(_ns)
    _s = _b_utils._EngineIntScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _s_py = _b_utils._EngineIntScalarType.to_py(_s)
    _z = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py+1, _s_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _xx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py+1,),
    )
    b_field_fracbm_generate(
        _ns,
        _s,
        _m,
        xmax,
        h,
        _lam,
        rho,
        statecomm,
        _z,
        _xx,
    )
    _z_py = _z
    _xx_py = _xx
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['z', 'xx'],
    )(
        _z_py,
        _xx_py,
    )
    return _return_data
