# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
# pylint: disable=simplifiable-if-statement
u"""
Interfaces for the NAG Mark 27.0 `stat` Chapter.

``stat`` - Simple Calculations on Statistical Data

This module covers the following topics:

- descriptive statistics and exploratory data analysis;

- statistical distribution functions and their inverses;

- testing for Normality and other distributions.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01intro.html

See Also
--------
``naginterfaces.library.examples.stat`` :
    This subpackage contains examples for the ``stat`` module.
    See also the :ref:`library_stat_ex` subsection.
"""

# NAG Copyright 2017-2019.

import collections as _collections

from ..base import utils as _b_utils
from ..library import _utils

def summary_2var(x1, x2, wt=None):
    u"""
    Means, corrected sums of squares and cross-products, etc., two
    variables, from raw data.

    ``summary_2var`` computes the means, standard deviations, corrected
    sums of squares and products, maximum and minimum values, and the
    product-moment correlation coefficient for two variables.
    Unequal weighting may be given.

    For full information please refer to the NAG Library document for
    g01ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01abf.html

    Parameters
    ----------
    x1 : float, array-like, shape (n)
        The observations from the first sample, x_i, for i = 1,2,...,n.

    x2 : float, array-like, shape (n)
        The observations from the second sample, y_i, for i = 1,2,...,n.

    wt : None or float, array-like, shape (n), optional
        If weights are being supplied then the elements of `wt` must
        contain the weights associated with the observations, w_i, for i
        = 1,2,...,n.

        If weights are not supplied then `wt` may be None.

    Returns
    -------
    nval : int
        Is used to indicate the number of valid observations, m; see
        [Description] (g).

    res : float, ndarray, shape (13)
        The elements of `res` contain the following results:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            The number of valid cases, m, is 0.

        (`errno` 3)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The number of valid cases, m, is 1. In this case standard
            deviation and product-moment correlation coefficient cannot
            be calculated.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.stat import summary_2var as b_summary_2var

    fun_name = 'naginterfaces.library.stat.summary_2var'

    _x1 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x1'},
        x1,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x1, {'fun_name': fun_name, 'entity': 'x1'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x2'},
        x2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (13,),
    )
    try:
        _return_scalar = b_summary_2var(
            _n,
            _x1,
            _x2,
            _wt,
            _res,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _nval_py = _return_scalar
    _res_py = _res
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nval', 'res'],
    )(
        _nval_py,
        _res_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def summary_freq(x, ifreq):
    u"""
    Mean, variance, skewness, kurtosis, etc., one variable, from
    frequency table.

    ``summary_freq`` calculates the mean, standard deviation and
    coefficients of skewness and kurtosis for data grouped in a
    frequency distribution.

    For full information please refer to the NAG Library document for
    g01ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01adf.html

    Parameters
    ----------
    x : float, array-like, shape (k)
        The elements of `x` must contain the boundary values of the
        classes in ascending order, so that class i is bounded by the
        values in `x`[i-1] and `x`[i], for i = 1,2,...,k-1.

    ifreq : int, array-like, shape (k)
        The ith element of `ifreq` must contain the frequency associated
        with the ith class, for i = 1,2,...,k-1. `ifreq`[k-1] is not
        used by the function.

    Returns
    -------
    xmean : float
        The mean value, y_.

    s2 : float
        The standard deviation, s_2.

    s3 : float
        The coefficient of skewness, s_3.

    s4 : float
        The coefficient of kurtosis, s_4.

    n : int
        The total frequency, n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k > 1.

        (`errno` 2)
            On entry, I = *<value>*, `x`[I-2] = *<value>* and `x`[I-1] =
            *<value>*.

            Constraint: `x`[I-2] <= `x`[I-1].

        (`errno` 3)
            Either `ifreq`[i] < 0 for some i, or the sum of frequencies
            is zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The total frequency, n, is less than 2, hence the quantities
            s_2, s_3 and s_4 cannot be calculated.
    """

    from ..base.stat import summary_freq as b_summary_freq

    fun_name = 'naginterfaces.library.stat.summary_freq'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _ifreq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ifreq'},
        ifreq,
        copy=False,
    )
    try:
        _return_scalars = b_summary_freq(
            _k,
            _x,
            _ifreq,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _xmean_py = _return_scalars.xmean
    _s2_py = _return_scalars.s2
    _s3_py = _return_scalars.s3
    _s4_py = _return_scalars.s4
    _n_py = _return_scalars.n
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xmean', 's2', 's3', 's4', 'n'],
    )(
        _xmean_py,
        _s2_py,
        _s3_py,
        _s4_py,
        _n_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def frequency_table(x, cb=None):
    u"""
    Frequency table from raw data.

    ``frequency_table`` constructs a frequency distribution of a
    variable, according to either user-supplied, or function-calculated
    class boundary values.

    For full information please refer to the NAG Library document for
    g01ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01aef.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample of observations of the variable for which the
        frequency distribution is required, x_i, for i = 1,2,...,n. The
        values may be in any order.

    cb : None or float, array-like, shape (k), optional
        If iclass = 0, the elements of `cb` need not be assigned values,
        as ``frequency_table`` calculates k-1 class boundary values.

        If iclass = 1, the first k-1 elements of `cb` must contain the
        class boundary values you supplied, in ascending order.

        In both cases, the element `cb`[k-1] need not be assigned, as it
        is not used in the function.

    Returns
    -------
    cb : float, ndarray, shape (k)
        The first k-1 elements of `cb` contain the class boundary values
        in ascending order.

    ifreq : int, ndarray, shape (k)
        The elements of `ifreq` contain the frequencies in each class,
        f_i, for i = 1,2,...,k. In particular `ifreq`[0] contains the
        frequency of the class up to `cb`[0], f_1, and `ifreq`[k-1]
        contains the frequency of the class greater than `cb`[k-2], f_k.

    xmin : float
        The smallest value in the sample, a.

    xmax : float
        The largest value in the sample, b.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 2.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            On entry, `cb`[*<value>*] = *<value>* and `cb`[*<value>*] =
            *<value>*.

            Constraint: `cb`[*<value>*] < `cb`[*<value>*].
    """

    from ..base.stat import frequency_table as b_frequency_table

    fun_name = 'naginterfaces.library.stat.frequency_table'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _cb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cb'},
        cb,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k = _b_utils._EngineIntScalarType(
        2 if cb is None else _b_utils._EngineFloat64ArrayType.get_shape(_cb, {'fun_name': fun_name, 'entity': 'cb'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    if cb is None:
        _iclass = 0
        _iclass_py = _iclass
    else:
        _iclass = 1
        _iclass_py = _iclass
    _k_py = _b_utils._EngineIntScalarType.to_py(_k)
    if cb is None:
        _cb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_k_py,),
        )
    _ifreq = _b_utils._EngineIntArrayType.empty_ndarray(
        (_k_py,),
    )
    _return_scalars = b_frequency_table(
        _n,
        _k,
        _x,
        _iclass,
        _cb,
        _ifreq,
    )
    _cb_py = _cb
    _ifreq_py = _ifreq
    _xmin_py = _return_scalars.xmin
    _xmax_py = _return_scalars.xmax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['cb', 'ifreq', 'xmin', 'xmax'],
    )(
        _cb_py,
        _ifreq_py,
        _xmin_py,
        _xmax_py,
    )
    return _return_data

def contingency_table(nobs, num=0):
    u"""
    Two-way contingency table analysis, with chi^2/Fisher's exact test.

    ``contingency_table`` performs the analysis of a two-way r*c
    contingency table or classification.
    If r = c = 2, and the total number of objects classified is 40 or
    fewer, then the probabilities for Fisher's exact test are computed.
    Otherwise, a test statistic is computed (with Yates' correction when
    r = c = 2), which under the assumption of no association between the
    classifications has approximately a chi-square distribution with
    (r-1)*(c-1) degrees of freedom.

    For full information please refer to the NAG Library document for
    g01af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01aff.html

    Parameters
    ----------
    nobs : int, array-like, shape (m, n)
        The elements `nobs`[i-1,j-1], for j = 1,2,...,n, for i =
        1,2,...,m, must contain the frequencies for the two-way
        classification. The (m+1)th row and the (n+1)th column of `nobs`
        need not be set.

    num : int, optional
        The value assigned to `num` must determine whether automatic
        'shrinkage' is required when any r_{ij} < 1, as outlined in
        [Description](1).

        If `num` = 1, shrinkage is required, otherwise shrinkage is not
        required.

    Returns
    -------
    nobs : int, ndarray, shape (m, n)
        Contains the following information:

        - `nobs`[i-1,j-1], for j = 1,2,...,n_1, for i = 1,2,...,m_1,
          contain the frequencies for the two-way classification after
          'shrinkage' has taken place (see [Description]).

        - `nobs`[i-1,n], for i = 1,2,...,m_1, contain the total
          frequencies in the remaining rows, R_i.

        - `nobs`[m,j-1], for j = 1,2,...,n_1, contain the total
          frequencies in the remaining columns, C_j.

        - `nobs`[m,n], contains the total frequency, T.

        If any 'shrinkage' has occurred, all other cells contain no
        useful information.

    num : int
        When Fisher's exact test for a 2*2 classification is used then
        `num` contains the number of elements used in the array `p`,
        otherwise `num` is set to zero.

    pred : float, ndarray, shape (m, n)
        The elements `pred`[i-1,j-1], where i = 1,2,...,`m1` and j =
        1,2,...,`n1` contain the expected frequencies, r_{ij}
        corresponding to the observed frequencies `nobs`[i-1,j-1],
        except in the case when Fisher's exact test for a 2*2
        classification is to be used, when `pred` is not used. No other
        elements are utilized.

    chis : float
        The value of the test statistic, chi^2, except when Fisher's
        exact test for a 2*2 classification is used in which case it is
        unspecified.

    p : float, ndarray, shape (21)
        The first `num` elements contain the probabilities associated
        with the various possible frequency tables, P_r, for r =
        0,1,...,R_1, the remainder are unspecified.

    npos : int
        `p`[`npos`-1] holds the probability associated with the given
        table of frequencies.

    ndf : int
        The value of `ndf` gives the number of degrees of freedom for
        the chi-square distribution, (m_1-1)*(n_1-1); when Fisher's
        exact test is used `ndf` = 1.

    m1 : int
        The number of rows of the two-way classification, after any
        'shrinkage', m_1.

    n1 : int
        The number of columns of the two-way classification, after any
        'shrinkage', n_1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The number of rows or columns of `nobs` is less than 2.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 2.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m > 2.

        (`errno` 2)
            At least one frequency is negative, or all frequencies are
            zero.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.stat import contingency_table as b_contingency_table

    fun_name = 'naginterfaces.library.stat.contingency_table'

    _nobs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nobs'},
        nobs,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_nobs, 'nobs'),),
    )
    _m_py = _b_utils._EngineIntArrayType.get_shape(_nobs, {'fun_name': fun_name, 'entity': 'nobs'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntArrayType.get_shape(_nobs, {'fun_name': fun_name, 'entity': 'nobs'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _pred = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_m_sorder,
    )
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (21,),
    )
    _return_scalars = b_contingency_table(
        _m,
        _n,
        _nobs,
        num,
        _pred,
        _p,
    )
    _nobs_py = _nobs
    _num_py = _return_scalars.num
    _pred_py = _pred
    _chis_py = _return_scalars.chis
    _p_py = _p
    _npos_py = _return_scalars.npos
    _ndf_py = _return_scalars.ndf
    _m1_py = _return_scalars.m1
    _n1_py = _return_scalars.n1
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nobs', 'num', 'pred', 'chis', 'p', 'npos', 'ndf', 'm1', 'n1'],
    )(
        _nobs_py,
        _num_py,
        _pred_py,
        _chis_py,
        _p_py,
        _npos_py,
        _ndf_py,
        _m1_py,
        _n1_py,
    )
    return _return_data

def five_point_summary(x):
    u"""
    Computes a five-point summary (median, hinges and extremes).

    ``five_point_summary`` calculates a five-point summary for a single
    sample.

    For full information please refer to the NAG Library document for
    g01al

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01alf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample observations, x_1,x_2,...,x_n.

    Returns
    -------
    res : float, ndarray, shape (5)
        `res` contains the five-point summary.

        `res`[0]

          The minimum.

        `res`[1]

          The lower hinge.

        `res`[2]

          The median.

        `res`[3]

          The upper hinge.

        `res`[4]

          The maximum.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 5.
    """

    from ..base.stat import five_point_summary as b_five_point_summary

    fun_name = 'naginterfaces.library.stat.five_point_summary'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (5,),
    )
    b_five_point_summary(
        _n,
        _x,
        _res,
    )
    _res_py = _res
    return _res_py

def quantiles(rv, q):
    u"""
    Find quantiles of an unordered vector, real numbers.

    ``quantiles`` finds specified quantiles from a vector of unsorted
    data.

    For full information please refer to the NAG Library document for
    g01am

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01amf.html

    Parameters
    ----------
    rv : float, array-like, shape (n)
        The vector whose quantiles are to be determined.

    q : float, array-like, shape (nq)
        The quantiles to be calculated, in ascending order. Note that
        these must be between 0.0 and 1.0, with 0.0 returning the
        smallest element and 1.0 the largest.

    Returns
    -------
    qv : float, ndarray, shape (nq)
        `qv`[i-1] contains the quantile specified by the value provided
        in `q`[i-1], or an interpolated value if the quantile falls
        between two data values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 2)
            On entry, nq = *<value>*.

            Constraint: nq > 0.

        (`errno` 3)
            On entry, an element of `q` was less than 0.0 or greater
            than 1.0.

        (`errno` 4)
            On entry, `q` was not in ascending order.
    """

    from ..base.stat import quantiles as b_quantiles

    fun_name = 'naginterfaces.library.stat.quantiles'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_rv, {'fun_name': fun_name, 'entity': 'rv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nq_py = _b_utils._EngineFloat64ArrayType.get_shape(_q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=1)[0]
    _nq = _b_utils._EngineIntScalarType(
        _nq_py, {'fun_name': fun_name, 'entity': '_nq'},
    ).to_ctypes()
    _qv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nq_py,),
    )
    b_quantiles(
        _n,
        _rv,
        _nq,
        _q,
        _qv,
    )
    _qv_py = _qv
    return _qv_py

def quantiles_stream_fixed(ind, n, rv, nb, eps, q, nq, rcomm, icomm):
    u"""
    Calculates approximate quantiles from a data stream of known size.

    ``quantiles_stream_fixed`` finds approximate quantiles from a data
    stream of known size using an out-of-core algorithm.

    For full information please refer to the NAG Library document for
    g01an

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01anf.html

    Parameters
    ----------
    ind : int
        Indicates the action required in the current call to
        ``quantiles_stream_fixed``.

        `ind` = 0

          Return the required length of `rcomm` and `icomm` in
          `icomm`[0] and `icomm`[1] respectively. `n` and `eps` must be
          set and licomm must be at least 2.

        `ind` = 1

          Initialise the communication arrays and process the first `nb`
          values from the data stream as supplied in `rv`.

        `ind` = 2

          Process the next block of `nb` values from the data stream.
          The calling program must update `rv` and (if required) `nb`,
          and re-enter ``quantiles_stream_fixed`` with all other
          parameters unchanged.

        `ind` = 3

          Calculate the `nq` epsilon-approximate quantiles specified in
          `q`. The calling program must set `q` and `nq` and re-enter
          ``quantiles_stream_fixed`` with all other parameters
          unchanged. This option can be chosen only when `np` >=
          ⌈exp(1.0)/`eps`⌉.

    n : int
        n, the total number of values in the data stream.

    rv : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind` in (1, 2): `nb`; otherwise: 0.

        If `ind` = 1 or 2, the vector containing the current block of
        data, otherwise `rv` is not referenced.

    nb : int
        If `ind` = 1 or 2, the size of the current block of data. The
        size of blocks of data in array `rv` can vary; therefore `nb`
        can change between calls to ``quantiles_stream_fixed``.

    eps : float
        Approximation factor epsilon.

    q : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind`=3: `nq`; otherwise: 0.

        If `ind` = 3, the quantiles to be calculated, otherwise `q` is
        not referenced. Note that `q`[i] = 0.0, corresponds to the
        minimum value and `q`[i] = 1.0 to the maximum value.

    nq : int
        If `ind` = 3, the number of quantiles requested, otherwise `nq`
        is not referenced.

    rcomm : float, ndarray, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `ind` != 0: the value returned in `icomm`[0] by a
        call to ``quantiles_stream_fixed`` with `ind` = 0; otherwise: 0.

        Communication array.

    icomm : int, ndarray, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `ind` != 0: the value returned in `icomm`[1] by a
        call to ``quantiles_stream_fixed`` with `ind` = 0; otherwise: 2.

        Communication array.

    Returns
    -------
    ind : int
        Indicates output from a successful call.

        `ind` = 1

          Lengths of `rcomm` and `icomm` have been returned in
          `icomm`[0] and `icomm`[1] respectively.

        `ind` = 2

          ``quantiles_stream_fixed`` has processed `np` data points and
          expects to be called again with additional data (i.e., `np` <
          `n`).

        `ind` = 3

          ``quantiles_stream_fixed`` has returned the requested
          epsilon-approximate quantiles in `qv`. These quantiles are
          based on `np` data points.

        `ind` = 4

          Routine has processed all `n` data points (i.e., `np` = `n`).

    np : int
        The number of elements processed so far.

    qv : float, ndarray, shape (:)
        If `ind` = 3, `qv`[i] contains the epsilon-approximate quantiles
        specified by the value provided in `q`[i].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 3)
            On entry, `eps` = *<value>*.

            Constraint: exp(1.0)/`n` <= `eps` <= 1.0.

        (`errno` 4)
            On entry, `ind` = 1 or 2 and `nb` = *<value>*.

            Constraint: if `ind` = 1 or 2 then `nb` > 0.

        (`errno` 7)
            Number of data elements streamed, *<value>* is not
            sufficient for a quantile query when `eps` = *<value>*.

            Supply more data or reprocess the data with a higher `eps`
            value.

        (`errno` 8)
            On entry, `ind` = 3 and `nq` = *<value>*.

            Constraint: if `ind` = 3 then `nq` > 0.

        (`errno` 9)
            On entry, `ind` = 3 and `q`[*<value>*] = *<value>*.

            Constraint: if `ind` = 3 then 0.0 <= `q`[i] <= 1.0 for all
            i.
    """

    from ..base.stat import quantiles_stream_fixed as b_quantiles_stream_fixed

    fun_name = 'naginterfaces.library.stat.quantiles_stream_fixed'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ind = _b_utils._EngineIntScalarType(
        ind, {'fun_name': fun_name, 'entity': 'ind'},
    ).to_ctypes()
    _ind_py = _b_utils._EngineIntScalarType.to_py(_ind)
    _nq = _b_utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _nq_py = _b_utils._EngineIntScalarType.to_py(_nq)
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    exp_shape_qv = (_qv_dim1,)
    _qv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_qv,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'rcomm'},
        rcomm,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'icomm'},
        icomm,
    )
    _return_scalars = b_quantiles_stream_fixed(
        _ind,
        n,
        _rv,
        nb,
        eps,
        _q,
        _qv,
        _nq,
        rcomm,
        icomm,
    )
    _ind_py = _return_scalars.ind
    _np_py = _return_scalars.np
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    _qv_py = _qv[:_qv_dim1]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ind', 'np', 'qv'],
    )(
        _ind_py,
        _np_py,
        _qv_py,
    )
    return _return_data

def quantiles_stream_arbitrary(ind, rv, nb, eps, q, nq, rcomm, icomm):
    u"""
    Calculates approximate quantiles from a data stream of unknown size.

    ``quantiles_stream_arbitrary`` finds approximate quantiles from a
    large arbitrary-sized data stream using an out-of-core algorithm.

    For full information please refer to the NAG Library document for
    g01ap

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01apf.html

    Parameters
    ----------
    ind : int
        On initial entry: must be set to 0.Indicates the action required
        in the current call to ``quantiles_stream_arbitrary``.

        `ind` = 0

          Initialize the communication arrays and attempt to process the
          first `nb` values from the data stream. `eps`, `rv` and `nb`
          must be set and licomm must be at least 10.

        `ind` = 1

          Attempt to process the next block of `nb` values from the data
          stream. The calling program must update `rv` and (if required)
          `nb`, and re-enter ``quantiles_stream_arbitrary`` with all
          other parameters unchanged.

        `ind` = 2

          Continue calculation following the reallocation of either or
          both of the communication arrays `rcomm` and `icomm`.

        `ind` = 3

          Calculate the `nq` epsilon-approximate quantiles specified in
          `q`. The calling program must set `q` and `nq` and re-enter
          ``quantiles_stream_arbitrary`` with all other parameters
          unchanged. This option can be chosen only when `np` >=
          ⌈exp(1.0)/`eps`⌉.

    rv : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind` in (0, 1, 2): `nb`; otherwise: 0.

        If `ind` = 0, 1 or 2, the vector containing the current block of
        data, otherwise `rv` is not referenced.

    nb : int
        If `ind` = 0, 1 or 2, the size of the current block of data. The
        size of blocks of data in array `rv` can vary; therefore `nb`
        can change between calls to ``quantiles_stream_arbitrary``.

    eps : float
        Approximation factor epsilon.

    q : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind`=3: `nq`; otherwise: 0.

        If `ind` = 3, the quantiles to be calculated, otherwise `q` is
        not referenced. Note that `q`[i] = 0.0, corresponds to the
        minimum value and `q`[i] = 1.0 to the maximum value.

    nq : int
        If `ind` = 3, the number of quantiles requested, otherwise `nq`
        is not referenced.

    rcomm : float, ndarray, shape (lrcomm), modified in place
        On entry: if `ind` = 1 or 2 then the first l elements of `rcomm`
        as supplied to ``quantiles_stream_arbitrary`` must be identical
        to the first l elements of `rcomm` returned from the last call
        to ``quantiles_stream_arbitrary``, where l is the value of
        lrcomm used in the last call. In other words, the contents of
        `rcomm` must not be altered between calls to this function. If
        `rcomm` needs to be reallocated then its contents must be
        preserved. If `ind` = 0 then `rcomm` need not be set.

        On exit: `rcomm` holds information required by subsequent calls
        to ``quantiles_stream_arbitrary``.

    icomm : int, ndarray, shape (licomm), modified in place
        On entry: if `ind` = 1 or 2 then the first l elements of `icomm`
        as supplied to ``quantiles_stream_arbitrary`` must be identical
        to the first l elements of `icomm` returned from the last call
        to ``quantiles_stream_arbitrary``, where l is the value of
        licomm used in the last call. In other words, the contents of
        `icomm` must not be altered between calls to this function. If
        `icomm` needs to be reallocated then its contents must be
        preserved. If `ind` = 0 then `icomm` need not be set.

        On exit: `icomm`[0] holds the minimum required length for
        `rcomm` and `icomm`[1] holds the minimum required length for
        `icomm`. The remaining elements of `icomm` are used for
        communication between subsequent calls to
        ``quantiles_stream_arbitrary``.

    Returns
    -------
    ind : int
        Indicates output from the call.

        `ind` = 1

          ``quantiles_stream_arbitrary`` has processed `np` data points
          and expects to be called again with additional data.

        `ind` = 2

          Either one or more of the communication arrays `rcomm` and
          `icomm` is too small. The new minimum lengths of `rcomm` and
          `icomm` have been returned in `icomm`[0] and `icomm`[1]
          respectively. If the new minimum length is greater than the
          current length then the corresponding communication array
          needs to be reallocated, its contents preserved and
          ``quantiles_stream_arbitrary`` called again with all other
          parameters unchanged.

          If there is more data to be processed, it is recommended that
          lrcomm and licomm are made significantly bigger than the
          minimum to limit the number of reallocations.

        `ind` = 3

          ``quantiles_stream_arbitrary`` has returned the requested
          epsilon-approximate quantiles in `qv`. These quantiles are
          based on `np` data points.

    np : int
        m, the number of elements processed so far.

    qv : float, ndarray, shape (:)
        If `ind` = 3, `qv`[i] contains the epsilon-approximate quantiles
        specified by the value provided in `q`[i].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `eps` = *<value>*.

            Constraint: 0.0 < `eps` <= 1.0.

        (`errno` 3)
            On entry, `ind` = 0, 1 or 2 and `nb` = *<value>*.

            Constraint: if `ind` = 0, 1 or 2 then `nb` > 0.

        (`errno` 6)
            The contents of `icomm` have been altered between calls to
            this function.

        (`errno` 7)
            The contents of `rcomm` have been altered between calls to
            this function.

        (`errno` 8)
            Number of data elements streamed, *<value>* is not
            sufficient for a quantile query when `eps` = *<value>*.

            Supply more data or reprocess the data with a higher `eps`
            value.

        (`errno` 9)
            On entry, `ind` = 3 and `nq` = *<value>*.

            Constraint: if `ind` = 3 then `nq` > 0.

        (`errno` 10)
            On entry, `ind` = 3 and `q`[*<value>*] = *<value>*.

            Constraint: if `ind` = 3 then 0.0 <= `q`[i] <= 1.0 for all
            i.
    """

    from ..base.stat import quantiles_stream_arbitrary as b_quantiles_stream_arbitrary

    fun_name = 'naginterfaces.library.stat.quantiles_stream_arbitrary'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ind = _b_utils._EngineIntScalarType(
        ind, {'fun_name': fun_name, 'entity': 'ind'},
    ).to_ctypes()
    _ind_py = _b_utils._EngineIntScalarType.to_py(_ind)
    _nq = _b_utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _nq_py = _b_utils._EngineIntScalarType.to_py(_nq)
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    exp_shape_qv = (_qv_dim1,)
    _qv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_qv,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'rcomm'},
        rcomm,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'icomm'},
        icomm,
    )
    _return_scalars = b_quantiles_stream_arbitrary(
        _ind,
        _rv,
        nb,
        eps,
        _q,
        _qv,
        _nq,
        rcomm,
        icomm,
    )
    _ind_py = _return_scalars.ind
    _np_py = _return_scalars.np
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    _qv_py = _qv[:_qv_dim1]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ind', 'np', 'qv'],
    )(
        _ind_py,
        _np_py,
        _qv_py,
    )
    return _return_data

def summary_onevar(x, wt=None, pn=0, rcomm=None):
    u"""
    Computes univariate summary information: mean, variance, skewness,
    kurtosis.

    ``summary_onevar`` calculates the mean, standard deviation,
    coefficients of skewness and kurtosis, and the maximum and minimum
    values for a set of (optionally weighted) data.
    The input data can be split into arbitrary sized blocks, allowing
    large datasets to be summarised.

    For full information please refer to the NAG Library document for
    g01at

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01atf.html

    Parameters
    ----------
    x : float, array-like, shape (nb)
        The current block of observations, corresponding to x_i, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far and b is the size of the current block of data.

    wt : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `wt` is not None: nb; otherwise: 0.

        If `wt` is not None, `wt` must contain the user-supplied weights
        corresponding to the block of data supplied in `x`, that is w_i,
        for i = k+1,...,k+b.

        If `wt` is None, w_i = 1 for all i.

    pn : int, optional
        The number of valid observations processed so far, that is the
        number of observations with w_i > 0, for i = 1,2,...,k. On the
        first call to ``summary_onevar``, or when starting to summarise
        a new dataset, `pn` must be set to 0.

        If `pn` != 0, it must be the same value as returned by the last
        call to ``summary_onevar``.

    rcomm : None or float, ndarray, shape (:), optional, modified in
    place
        Note: the required length for this argument is determined as
        follows: if `rcomm` is not None: 20; otherwise: 0.

        Optionally, on entry: communication array, used to store
        information between calls to ``summary_onevar``. If `pn` = 0,
        `rcomm` need not be initialized, otherwise it must be unchanged
        since the last call to this function.

        If `rcomm` is None, `rcomm` is not referenced and all the data
        must be supplied in one go.

        On exit, if not None on entry: the updated communication array.
        The first five elements of `rcomm` hold information that may be
        of interest with

            [table omitted]

        the remaining elements of `rcomm` are used for workspace and so
        are undefined.

    Returns
    -------
    pn : int
        The updated number of valid observations processed, that is the
        number of observations with w_i > 0, for i = 1,2,...,k+b.

    xmean : float
        x_, the mean of the first k+b observations.

    xsd : float
        s_2, the standard deviation of the first k+b observations.

    xskew : float
        s_3, the coefficient of skewness for the first k+b observations.

    xkurt : float
        s_4, the coefficient of kurtosis for the first k+b observations.

    xmin : float
        The smallest value in the first k+b observations.

    xmax : float
        The largest value in the first k+b observations.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 41)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: if `wt` is not None then `wt`[i-1] >= 0, for i =
            1,2,...,nb.

        (`errno` 51)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 52)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0, `pn` must be unchanged since
            previous call.

        (`errno` 54)
            On entry, `pn` = *<value>*.

            Constraint: if `rcomm` is None then `pn` = 0.

        (`errno` 121)
            `rcomm` has been corrupted between calls.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 53)
            On entry, the number of valid observations is zero.

        (`errno` 71)
            On exit we were unable to calculate `xskew` or `xkurt`. A
            value of 0 has been returned.

        (`errno` 72)
            On exit we were unable to calculate `xsd`, `xskew` or
            `xkurt`. A value of 0 has been returned.
    """

    from ..base.stat import summary_onevar as b_summary_onevar

    fun_name = 'naginterfaces.library.stat.summary_onevar'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'rcomm'},
        rcomm,
        allow_none=True,
    )
    try:
        _return_scalars = b_summary_onevar(
            _nb,
            _x,
            _wt,
            pn,
            rcomm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _pn_py = _return_scalars.pn
    _xmean_py = _return_scalars.xmean
    _xsd_py = _return_scalars.xsd
    _xskew_py = _return_scalars.xskew
    _xkurt_py = _return_scalars.xkurt
    _xmin_py = _return_scalars.xmin
    _xmax_py = _return_scalars.xmax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pn', 'xmean', 'xsd', 'xskew', 'xkurt', 'xmin', 'xmax'],
    )(
        _pn_py,
        _xmean_py,
        _xsd_py,
        _xskew_py,
        _xkurt_py,
        _xmin_py,
        _xmax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def summary_onevar_combine(mrcomm):
    u"""
    Combines multiple sets of summary information, for use after
    ``summary_onevar``.

    ``summary_onevar_combine`` combines sets of summaries produced by
    ``summary_onevar``.

    For full information please refer to the NAG Library document for
    g01au

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01auf.html

    Parameters
    ----------
    mrcomm : float, array-like, shape (20, b)
        The jth column of `mrcomm` must contain the information returned
        in rcomm from one of the runs of ``summary_onevar``.

    Returns
    -------
    pn : int
        The number of valid observations, that is the number of
        observations with w_i > 0, for i = 1,2,...,n.

    xmean : float
        x_, the mean.

    xsd : float
        s_2, the standard deviation.

    xskew : float
        s_3, the coefficient of skewness.

    xkurt : float
        s_4, the coefficient of kurtosis.

    xmin : float
        The smallest value.

    xmax : float
        The largest value.

    rcomm : float, ndarray, shape (20)
        An amalgamation of the information held in `mrcomm`. This is in
        the same format as rcomm from ``summary_onevar``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, b = *<value>*.

            Constraint: b >= 1.

        (`errno` 21)
            On entry, `mrcomm` is not in the expected format.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, the number of valid observations is zero.

        (`errno` 51)
            On exit we were unable to calculate `xskew` or `xkurt`. A
            value of 0 has been returned.

        (`errno` 52)
            On exit we were unable to calculate `xsd`, `xskew` or
            `xkurt`. A value of 0 has been returned.
    """

    from ..base.stat import summary_onevar_combine as b_summary_onevar_combine

    fun_name = 'naginterfaces.library.stat.summary_onevar_combine'

    _mrcomm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mrcomm'},
        mrcomm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_py = _b_utils._EngineFloat64ArrayType.get_shape(_mrcomm, {'fun_name': fun_name, 'entity': 'mrcomm'}, exp_rank=2)[1]
    _b = _b_utils._EngineIntScalarType(
        _b_py, {'fun_name': fun_name, 'entity': '_b'},
    ).to_ctypes()
    _rcomm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (20,),
    )
    try:
        _return_scalars = b_summary_onevar_combine(
            _b,
            _mrcomm,
            _rcomm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _pn_py = _return_scalars.pn
    _xmean_py = _return_scalars.xmean
    _xsd_py = _return_scalars.xsd
    _xskew_py = _return_scalars.xskew
    _xkurt_py = _return_scalars.xkurt
    _xmin_py = _return_scalars.xmin
    _xmax_py = _return_scalars.xmax
    _rcomm_py = _rcomm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pn', 'xmean', 'xsd', 'xskew', 'xkurt', 'xmin', 'xmax', 'rcomm'],
    )(
        _pn_py,
        _xmean_py,
        _xsd_py,
        _xskew_py,
        _xkurt_py,
        _xmin_py,
        _xmax_py,
        _rcomm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_binomial(n, p, k):
    u"""
    Binomial distribution function.

    ``prob_binomial`` returns the lower tail, upper tail and point
    probabilities associated with a binomial distribution.

    For full information please refer to the NAG Library document for
    g01bj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01bjf.html

    Parameters
    ----------
    n : int
        The parameter n of the binomial distribution.

    p : float
        The parameter p of the binomial distribution.

    k : int
        The integer k which defines the required probabilities.

    Returns
    -------
    plek : float
        The lower tail probability, Prob{X <= k}.

    pgtk : float
        The upper tail probability, Prob{X > k}.

    peqk : float
        The point probability, Prob{X = k}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.0.

        (`errno` 3)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 3)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: `k` <= `n`.

        (`errno` 4)
            On entry, `n` is too large to be represented exactly as a
            double precision number.

        (`errno` 5)
            On entry, the variance (= np(1-p)) exceeds 10^6.
    """

    from ..base.stat import prob_binomial as b_prob_binomial

    _return_scalars = b_prob_binomial(
        n,
        p,
        k,
    )
    _plek_py = _return_scalars.plek
    _pgtk_py = _return_scalars.pgtk
    _peqk_py = _return_scalars.peqk
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['plek', 'pgtk', 'peqk'],
    )(
        _plek_py,
        _pgtk_py,
        _peqk_py,
    )
    return _return_data

def prob_poisson(rlamda, k):
    u"""
    Poisson distribution function.

    ``prob_poisson`` returns the lower tail, upper tail and point
    probabilities associated with a Poisson distribution.

    For full information please refer to the NAG Library document for
    g01bk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01bkf.html

    Parameters
    ----------
    rlamda : float
        The parameter lambda of the Poisson distribution.

    k : int
        The integer k which defines the required probabilities.

    Returns
    -------
    plek : float
        The lower tail probability, Prob{X <= k}.

    pgtk : float
        The upper tail probability, Prob{X > k}.

    peqk : float
        The point probability, Prob{X = k}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: `rlamda` > 0.0.

        (`errno` 2)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 3)
            On entry, `rlamda` = *<value>*.

            Constraint: `rlamda` <= 10^6.
    """

    from ..base.stat import prob_poisson as b_prob_poisson

    _return_scalars = b_prob_poisson(
        rlamda,
        k,
    )
    _plek_py = _return_scalars.plek
    _pgtk_py = _return_scalars.pgtk
    _peqk_py = _return_scalars.peqk
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['plek', 'pgtk', 'peqk'],
    )(
        _plek_py,
        _pgtk_py,
        _peqk_py,
    )
    return _return_data

def prob_hypergeom(n, l, m, k):
    u"""
    Hypergeometric distribution function.

    ``prob_hypergeom`` returns the lower tail, upper tail and point
    probabilities associated with a hypergeometric distribution.

    For full information please refer to the NAG Library document for
    g01bl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01blf.html

    Parameters
    ----------
    n : int
        The parameter n of the hypergeometric distribution.

    l : int
        The parameter l of the hypergeometric distribution.

    m : int
        The parameter m of the hypergeometric distribution.

    k : int
        The integer k which defines the required probabilities.

    Returns
    -------
    plek : float
        The lower tail probability, Prob{X <= k}.

    pgtk : float
        The upper tail probability, Prob{X > k}.

    peqk : float
        The point probability, Prob{X = k}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `l` = *<value>* and `n` = *<value>*.

            Constraint: `l` <= `n`.

        (`errno` 2)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` <= `n`.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, `k` = *<value>*, `l` = *<value>*, `m` = *<value>*
            and `l`+`m`-`n` = *<value>*.

            Constraint: `k` >= `l`+`m`-`n`.

        (`errno` 4)
            On entry, `k` = *<value>* and `m` = *<value>*.

            Constraint: `k` <= `m`.

        (`errno` 4)
            On entry, `k` = *<value>* and `l` = *<value>*.

            Constraint: `k` <= `l`.

        (`errno` 4)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 5)
            On entry, `n` is too large to be represented exactly as a
            double precision number.

        (`errno` 6)
            On entry, the variance = l*m*(n-l)*(n-m)/(n^2*(n-1)) exceeds
            10^6.
    """

    from ..base.stat import prob_hypergeom as b_prob_hypergeom

    _return_scalars = b_prob_hypergeom(
        n,
        l,
        m,
        k,
    )
    _plek_py = _return_scalars.plek
    _pgtk_py = _return_scalars.pgtk
    _peqk_py = _return_scalars.peqk
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['plek', 'pgtk', 'peqk'],
    )(
        _plek_py,
        _pgtk_py,
        _peqk_py,
    )
    return _return_data

def normal_scores_exact(n, etol):
    u"""
    Normal scores, accurate values.

    ``normal_scores_exact`` computes a set of Normal scores, i.e., the
    expected values of an ordered set of independent observations from a
    Normal distribution with mean 0.0 and standard deviation 1.0.

    For full information please refer to the NAG Library document for
    g01da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01daf.html

    Parameters
    ----------
    n : int
        n, the size of the set.

    etol : float
        The maximum value for the estimated absolute error in the
        computed scores.

    Returns
    -------
    pp : float, ndarray, shape (`n`)
        The Normal scores. `pp`[i-1] contains the value E(x_(i)), for i
        = 1,2,...,n.

    errest : float
        A computed estimate of the maximum error in the computed scores
        (see [Accuracy]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            On entry, `etol` = *<value>*.

            Constraint: `etol` > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The function was unable to estimate the scores with
            estimated error less than `etol`. The best result obtained
            is returned together with the associated value of `errest`.
    """

    from ..base.stat import normal_scores_exact as b_normal_scores_exact

    fun_name = 'naginterfaces.library.stat.normal_scores_exact'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _pp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = b_normal_scores_exact(
            _n,
            _pp,
            etol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _pp_py = _pp
    _errest_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pp', 'errest'],
    )(
        _pp_py,
        _errest_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def normal_scores_approx(n):
    u"""
    Normal scores, approximate values.

    ``normal_scores_approx`` calculates an approximation to the set of
    Normal Scores, i.e., the expected values of an ordered set of
    independent observations from a Normal distribution with mean 0.0
    and standard deviation 1.0.

    For full information please refer to the NAG Library document for
    g01db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01dbf.html

    Parameters
    ----------
    n : int
        n, the size of the sample.

    Returns
    -------
    pp : float, ndarray, shape (`n`)
        The Normal scores. `pp`[i-1] contains the value E(x_(i)), for i
        = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.stat import normal_scores_approx as b_normal_scores_approx

    fun_name = 'naginterfaces.library.stat.normal_scores_approx'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _pp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_normal_scores_approx(
        _n,
        _pp,
    )
    _pp_py = _pp
    return _pp_py

def normal_scores_var(n, exp1, exp2, sumssq):
    u"""
    Normal scores, approximate variance-covariance matrix.

    ``normal_scores_var`` computes an approximation to the
    variance-covariance matrix of an ordered set of independent
    observations from a Normal distribution with mean 0.0 and standard
    deviation 1.0.

    For full information please refer to the NAG Library document for
    g01dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01dcf.html

    Parameters
    ----------
    n : int
        n, the sample size.

    exp1 : float
        The expected value of the largest Normal order statistic, m_n,
        from a sample of size n.

    exp2 : float
        The expected value of the second largest Normal order statistic,
        m_{n-1}, from a sample of size n.

    sumssq : float
        The sum of squares of the expected values of the Normal order
        statistics from a sample of size n.

    Returns
    -------
    vec : float, ndarray, shape (`n`*(`n`+1)/2)
        The upper triangle of the n by n variance-covariance matrix
        packed by column. Thus element V_{ij} is stored in
        `vec`[i+j*(j-1)/2-1], for 1 <= i <= j <= n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.
    """

    from ..base.stat import normal_scores_var as b_normal_scores_var

    fun_name = 'naginterfaces.library.stat.normal_scores_var'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _vec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    b_normal_scores_var(
        _n,
        exp1,
        exp2,
        sumssq,
        _vec,
    )
    _vec_py = _vec
    return _vec_py

def test_shapiro_wilk(x, a=None):
    u"""
    Shapiro and Wilk's W test for Normality.

    ``test_shapiro_wilk`` calculates Shapiro and Wilk's W statistic and
    its significance level for testing Normality.

    For full information please refer to the NAG Library document for
    g01dd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01ddf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The ordered sample values, x_i, for i = 1,2,...,n.

    a : None or float, array-like, shape (n), optional
        If calwts has been set to False then before entry `a` must
        contain the n weights as calculated in a previous call to
        ``test_shapiro_wilk``, otherwise `a` need not be set.

    Returns
    -------
    a : float, ndarray, shape (n)
        The n weights required to calculate `w`.

    w : float
        The value of the statistic, `w`.

    pw : float
        The significance level of `w`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n <= 5000.

        (`errno` 3)
            On entry, elements of `x` not in order. `x`[*<value>*] =
            *<value>*, `x`[*<value>*] = *<value>*, `x`[*<value>*] =
            *<value>*.

        (`errno` 3)
            On entry, all elements of `x` are equal.
    """

    from ..base.stat import test_shapiro_wilk as b_test_shapiro_wilk

    fun_name = 'naginterfaces.library.stat.test_shapiro_wilk'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    if a is None:
        _calwts = True
        _calwts_py = _calwts
    else:
        _calwts = False
        _calwts_py = _calwts
    if a is None:
        _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py,),
        )
    else:
        _a = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'a'},
            a,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _return_scalars = b_test_shapiro_wilk(
        _x,
        _n,
        _calwts,
        _a,
    )
    _a_py = _a
    _w_py = _return_scalars.w
    _pw_py = _return_scalars.pw
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'w', 'pw'],
    )(
        _a_py,
        _w_py,
        _pw_py,
    )
    return _return_data

def ranks_and_scores(scores, ties, x):
    u"""
    Ranks, Normal scores, approximate Normal scores or exponential
    (Savage) scores.

    ``ranks_and_scores`` computes the ranks, Normal scores, an
    approximation to the Normal scores or the exponential scores as
    requested by you.

    For full information please refer to the NAG Library document for
    g01dh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01dhf.html

    Parameters
    ----------
    scores : str, length 1
        Indicates which of the following scores are required.

        `scores` = 'R'

          The ranks.

        `scores` = 'N'

          The Normal scores, that is the expected value of the Normal
          order statistics.

        `scores` = 'B'

          The Blom version of the Normal scores.

        `scores` = 'T'

          The Tukey version of the Normal scores.

        `scores` = 'V'

          The van der Waerden version of the Normal scores.

        `scores` = 'S'

          The Savage scores, that is the expected value of the
          exponential order statistics.

    ties : str, length 1
        Indicates which of the following methods is to be used to assign
        scores to tied observations.

        `ties` = 'A'

          The average of the scores for tied observations is used.

        `ties` = 'L'

          The lowest score in the group of ties is used.

        `ties` = 'H'

          The highest score in the group of ties is used.

        `ties` = 'N'

          The nonrepeatable random number generator is used to randomly
          untie any group of tied observations.

        `ties` = 'R'

          The repeatable random number generator is used to randomly
          untie any group of tied observations.

        `ties` = 'I'

          Any ties are ignored, that is the scores are assigned to tied
          observations in the order that they appear in the data.

    x : float, array-like, shape (n)
        The sample of observations, x_i, for i = 1,2,...,n.

    Returns
    -------
    r : float, ndarray, shape (n)
        Contains the scores, s_i, for i = 1,2,...,n, as specified by
        `scores`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `scores` = *<value>*.

            Constraint: `scores` = 'R', 'N', 'B', 'T', 'V' or 'S'.

        (`errno` 1)
            On entry, `ties` = *<value>*.

            Constraint: `ties` = 'A', 'L', 'H', 'R' or 'I'.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.
    """

    from ..base.stat import ranks_and_scores as b_ranks_and_scores

    fun_name = 'naginterfaces.library.stat.ranks_and_scores'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    b_ranks_and_scores(
        scores,
        ties,
        _n,
        _x,
        _r,
    )
    _r_py = _r
    return _r_py

def prob_normal(x, tail='L'):
    u"""
    Computes probabilities for the standard Normal distribution.

    ``prob_normal`` returns a one or two tail probability for the
    standard Normal distribution.

    For full information please refer to the NAG Library document for
    g01ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01eaf.html

    Parameters
    ----------
    x : float
        x, the value of the standard Normal variate.

    tail : str, length 1, optional
        Indicates which tail the returned probability should represent.

        `tail` = 'L'

          The lower tail probability is returned, i.e., P(X <= x).

        `tail` = 'U'

          The upper tail probability is returned, i.e., P(X >= x).

        `tail` = 'S'

          The two tail (significance level) probability is returned,
          i.e., P(X >= |x|)+P(X <= -|x|).

        `tail` = 'C'

          The two tail (confidence interval) probability is returned,
          i.e., P(X <= |x|)-P(X <= -|x|).

    Returns
    -------
    p : float
        A one or two tail probability for the standard Normal
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.
    """

    from ..base.stat import prob_normal as b_prob_normal

    _return_scalar = b_prob_normal(
        tail,
        x,
    )
    return _return_scalar

def prob_students_t(t, df, tail='L'):
    u"""
    Computes probabilities for Student's t-distribution.

    ``prob_students_t`` returns the lower tail, upper tail or two tail
    probability for the Student's t-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01ebf.html

    Parameters
    ----------
    t : float
        t, the value of the Student's t variate.

    df : float
        nu, the degrees of freedom of the Student's t-distribution.

    tail : str, length 1, optional
        Indicates which tail the returned probability should represent.

        `tail` = 'U'

          The upper tail probability is returned, i.e., P(T >= t:nu).

        `tail` = 'S'

          The two tail (significance level) probability is returned,
          i.e., P(T >= |t|:nu)+P(T <= -|t|:nu).

        `tail` = 'C'

          The two tail (confidence interval) probability is returned,
          i.e., P(T <= |t|:nu)-P(T <= -|t|:nu).

        `tail` = 'L'

          The lower tail probability is returned, i.e., P(T <= t:nu).

    Returns
    -------
    p : float
        Either the lower tail, upper tail or two tail probability for
        the Student's t-distribution, depending on the value of `tail`
        supplied.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.0.
    """

    from ..base.stat import prob_students_t as b_prob_students_t

    _return_scalar = b_prob_students_t(
        tail,
        t,
        df,
    )
    return _return_scalar

def prob_chisq(x, df, tail='L'):
    u"""
    Computes probabilities for chi^2 distribution.

    ``prob_chisq`` returns the lower or upper tail probability for the
    chi^2-distribution with real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01ecf.html

    Parameters
    ----------
    x : float
        x, the value of the chi^2 variate with nu degrees of freedom.

    df : float
        nu, the degrees of freedom of the chi^2-distribution.

    tail : str, length 1, optional
        Indicates whether the upper or lower tail probability is
        required.

        `tail` = 'L'

          The lower tail probability is returned, i.e., P(X <= x:nu).

        `tail` = 'U'

          The upper tail probability is returned, i.e., P(X >= x:nu).

    Returns
    -------
    p : float
        The lower or upper tail probability for the chi^2-distribution,
        depending on the value of `tail` supplied.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L' or 'U'.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` > 0.0.

        (`errno` 4)
            The series used to calculate the gamma probabilities has
            failed to converge. The result returned should represent an
            approximation to the solution.
    """

    from ..base.stat import prob_chisq as b_prob_chisq

    _return_scalar = b_prob_chisq(
        tail,
        x,
        df,
    )
    return _return_scalar

def prob_f(f, df1, df2, tail='L'):
    u"""
    Computes probabilities for F-distribution.

    ``prob_f`` returns the probability for the lower or upper tail of
    the F or variance-ratio distribution with real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01ed

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01edf.html

    Parameters
    ----------
    f : float
        f, the value of the F variate.

    df1 : float
        The degrees of freedom of the numerator variance, nu_1.

    df2 : float
        The degrees of freedom of the denominator variance, nu_2.

    tail : str, length 1, optional
        Indicates whether an upper or lower tail probability is
        required.

        `tail` = 'L'

          The lower tail probability is returned, i.e., P(F <=
          f:nu_1,nu_2).

        `tail` = 'U'

          The upper tail probability is returned, i.e., P(F >=
          f:nu_1,nu_2).

    Returns
    -------
    p : float
        The probability for the lower or upper tail of the F or
        variance-ratio distribution with real degrees of freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L' or 'U'.

        (`errno` 2)
            On entry, `f` = *<value>*.

            Constraint: `f` >= 0.0.

        (`errno` 3)
            On entry, `df1` = *<value>* and `df2` = *<value>*.

            Constraint: `df1` > 0.0 and `df2` > 0.0.

        (`errno` 4)
            The probability is too close to 0.0 or 1.0. `f` is too far
            out into the tails for the probability to be evaluated
            exactly. The result tends to approach 1.0 if f is large, or
            0.0 if f is small. The result returned is a good
            approximation to the required solution.
    """

    from ..base.stat import prob_f as b_prob_f

    _return_scalar = b_prob_f(
        tail,
        f,
        df1,
        df2,
    )
    return _return_scalar

def prob_beta(x, a, b):
    u"""
    Computes upper and lower tail probabilities and probability density
    function for the beta distribution.

    ``prob_beta`` computes the upper and lower tail probabilities and
    the probability density function of the beta distribution with
    parameters a and b.

    For full information please refer to the NAG Library document for
    g01ee

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01eef.html

    Parameters
    ----------
    x : float
        beta, the value of the beta variate.

    a : float
        a, the first parameter of the required beta distribution.

    b : float
        b, the second parameter of the required beta distribution.

    Returns
    -------
    p : float
        The lower tail probability, P(B <= beta:a,b).

    q : float
        The upper tail probability, P(B >= beta:a,b).

    pdf : float
        The probability density function, f(B:a,b).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= 1.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > 0.0.
    """

    from ..base.stat import prob_beta as b_prob_beta

    _return_scalars = b_prob_beta(
        x,
        a,
        b,
    )
    _p_py = _return_scalars.p
    _q_py = _return_scalars.q
    _pdf_py = _return_scalars.pdf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'q', 'pdf'],
    )(
        _p_py,
        _q_py,
        _pdf_py,
    )
    return _return_data

def prob_gamma(g, a, b, tail='L'):
    u"""
    Computes probabilities for the gamma distribution.

    ``prob_gamma`` returns the lower or upper tail probability of the
    gamma distribution, with parameters alpha and beta.

    For full information please refer to the NAG Library document for
    g01ef

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01eff.html

    Parameters
    ----------
    g : float
        g, the value of the gamma variate.

    a : float
        The parameter alpha of the gamma distribution.

    b : float
        The parameter beta of the gamma distribution.

    tail : str, length 1, optional
        Indicates whether an upper or lower tail probability is
        required.

        `tail` = 'L'

          The lower tail probability is returned, that is P(G <=
          g:alpha,beta).

        `tail` = 'U'

          The upper tail probability is returned, that is P(G >=
          g:alpha,beta).

    Returns
    -------
    p : float
        The lower or upper tail probability of the gamma distribution,
        with parameters alpha and beta.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L' or 'U'.

        (`errno` 2)
            On entry, `g` = *<value>*.

            Constraint: `g` >= 0.0.

        (`errno` 3)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` > 0.0 and `b` > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The algorithm has failed to converge in *<value>*
            iterations. The probability returned should be a reasonable
            approximation to the solution.
    """

    from ..base.stat import prob_gamma as b_prob_gamma

    try:
        _return_scalar = b_prob_gamma(
            tail,
            g,
            a,
            b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_studentized_range(q, v, ir):
    u"""
    Computes probability for the Studentized range statistic.

    ``prob_studentized_range`` returns the probability associated with
    the lower tail of the distribution of the Studentized range
    statistic.

    For full information please refer to the NAG Library document for
    g01em

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01emf.html

    Parameters
    ----------
    q : float
        q, the Studentized range statistic.

    v : float
        v, the number of degrees of freedom for the experimental error.

    ir : int
        r, the number of groups.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the
        distribution of the Studentized range statistic.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `q` = *<value>*.

            Constraint: `q` > 0.0.

        (`errno` 1)
            On entry, `ir` = *<value>*.

            Constraint: `ir` >= 2.

        (`errno` 1)
            On entry, `v` = *<value>*.

            Constraint: `v` >= 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There is some doubt as to whether full accuracy has been
            achieved. The returned value should be a reasonable estimate
            of the true value.
    """

    from ..base.stat import prob_studentized_range as b_prob_studentized_range

    try:
        _return_scalar = b_prob_studentized_range(
            q,
            v,
            ir,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_durbin_watson(n, ip, d):
    u"""
    Computes bounds for the significance of a Durbin--Watson statistic.

    ``prob_durbin_watson`` calculates upper and lower bounds for the
    significance of a Durbin--Watson statistic.

    For full information please refer to the NAG Library document for
    g01ep

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01epf.html

    Parameters
    ----------
    n : int
        n, the number of observations used in calculating the
        Durbin--Watson statistic.

    ip : int
        p, the number of independent variables in the regression model,
        including the mean.

    d : float
        d, the Durbin--Watson statistic.

    Returns
    -------
    pdl : float
        Lower bound for the significance of the Durbin--Watson
        statistic, p_l.

    pdu : float
        Upper bound for the significance of the Durbin--Watson
        statistic, p_u.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>* and `ip` = *<value>*.

            Constraint: `n` > `ip`.

        (`errno` 2)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 0.0.
    """

    from ..base.stat import prob_durbin_watson as b_prob_durbin_watson

    _return_scalars = b_prob_durbin_watson(
        n,
        ip,
        d,
    )
    _pdl_py = _return_scalars.pdl
    _pdu_py = _return_scalars.pdu
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pdl', 'pdu'],
    )(
        _pdl_py,
        _pdu_py,
    )
    return _return_data

def prob_vonmises(t, vk):
    u"""
    Computes probability for von Mises distribution.

    ``prob_vonmises`` returns the probability associated with the lower
    tail of the von Mises distribution between -pi and pi through the
    function name.

    For full information please refer to the NAG Library document for
    g01er

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01erf.html

    Parameters
    ----------
    t : float
        theta, the observed von Mises statistic measured in radians.

    vk : float
        The concentration parameter kappa, of the von Mises
        distribution.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the von Mises
        distribution between -pi and pi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `vk` = *<value>*.

            Constraint: `vk` >= 0.0.
    """

    from ..base.stat import prob_vonmises as b_prob_vonmises

    _return_scalar = b_prob_vonmises(
        t,
        vk,
    )
    return _return_scalar

def prob_landau(x):
    u"""
    Landau distribution function.

    ``prob_landau`` returns the value of the Landau distribution
    function Phi(lambda).

    For full information please refer to the NAG Library document for
    g01et

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01etf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    Returns
    -------
    p : float
        The value of the Landau distribution function Phi(lambda).
    """

    from ..base.stat import prob_landau as b_prob_landau

    _return_scalar = b_prob_landau(
        x,
    )
    return _return_scalar

def prob_vavilov(x, comm):
    u"""
    Vavilov distribution function.

    ``prob_vavilov`` returns the value of the Vavilov distribution
    function Phi_V(lambda;kappa,beta^2).

    It is intended to be used after a call to ``init_vavilov``.

    For full information please refer to the NAG Library document for
    g01eu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01euf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``init_vavilov``.

    Returns
    -------
    p : float
        The value of the Vavilov distribution function
        Phi_V(lambda;kappa,beta^2).

    Raises
    ------
    NagValueError
        (`errno` 1)
            Either the initialization function has not been called prior
            to the first call of this function or a communication array
            has become corrupted.
    """

    from ..base.stat import prob_vavilov as b_prob_vavilov

    _return_scalar = b_prob_vavilov(
        x,
        comm,
    )
    return _return_scalar

def prob_dickey_fuller_unit(ts_type, n, ts, method=1, nsamp=100000, statecomm=None):
    u"""
    Computes probabilities for the Dickey--Fuller unit root test.

    ``prob_dickey_fuller_unit`` returns the probability associated with
    the lower tail of the distribution for the Dickey--Fuller unit root
    test statistic.

    For full information please refer to the NAG Library document for
    g01ew

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01ewf.html

    Parameters
    ----------
    ts_type : int
        The type of test statistic, supplied in `ts`.

    n : int
        n, the length of the time series used to calculate the test
        statistic.

    ts : float
        The Dickey--Fuller test statistic for which the probability is
        required. If

        `ts_type` = 1

          `ts` must contain tau.

        `ts_type` = 2

          `ts` must contain tau_mu.

        `ts_type` = 3

          `ts` must contain tau_tau.

        If the test statistic was calculated using
        ``tsa.uni_dickey_fuller_unit`` the value of `ts_type` and `n`
        must not change between calls to ``tsa.uni_dickey_fuller_unit``
        and ``prob_dickey_fuller_unit``.

    method : int, optional
        The method used to calculate the probability.

        `method` = 1

          The probability is interpolated from a look-up table, whose
          values were obtained via simulation.

        `method` = 2

          The probability is interpolated from a look-up table, whose
          values were obtained from Dickey.

        `method` = 3

          The probability is obtained via simulation.

        The probability calculated from the look-up table should give
        sufficient accuracy for most applications.

    nsamp : int, optional
        If `method` = 3, the number of samples used in the simulation;
        otherwise `nsamp` is not referenced and need not be set.

    statecomm : None or dict, RNG communication object, optional,
    modified in place
        RNG communication structure.

        When `method` = 3, this argument must have been initialized by a
        prior call to ``rand.init_repeat`` or ``rand.init_nonrepeat``.

    Returns
    -------
    pint : float
        The probability associated with the lower tail of the
        distribution for the (augmented) Dickey--Fuller unit root test
        statistic supplied in `ts`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `method` = *<value>*.

            Constraint: `method` = 1, 2 or 3.

        (`errno` 21)
            On entry, `ts_type` = *<value>*.

            Constraint: `ts_type` = 1, 2 or 3.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` != 3, `n` > 0.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` = 3 and `ts_type` = 1, `n` > 2.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` = 3 and `ts_type` = 2, `n` > 3.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` = 3 and `ts_type` = 3, `n` > 4.

        (`errno` 51)
            On entry, `nsamp` = *<value>*.

            Constraint: if `method` = 3, `nsamp` > 0.

        (`errno` 61)
            On entry, `method` = 3 and the `statecomm`['state'] vector
            has been corrupted or not initialized.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 201)
            The supplied input values were outside the range of at least
            one look-up table, therefore extrapolation was used.
    """

    from ..base.stat import prob_dickey_fuller_unit as b_prob_dickey_fuller_unit

    fun_name = 'naginterfaces.library.stat.prob_dickey_fuller_unit'

    _method = _b_utils._EngineIntScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'},
    ).to_ctypes()
    _method_py = _b_utils._EngineIntScalarType.to_py(_method)
    if _method_py == 3:
        if statecomm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'statecomm'},
                'must not be None when method = 3',
            )
    try:
        _return_scalar = b_prob_dickey_fuller_unit(
            _method,
            ts_type,
            n,
            ts,
            nsamp,
            statecomm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pint'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_kolmogorov1(n, d):
    u"""
    Computes probabilities for the one-sample Kolmogorov--Smirnov
    distribution.

    ``prob_kolmogorov1`` returns the upper tail probability associated
    with the one sample Kolmogorov--Smirnov distribution.

    For full information please refer to the NAG Library document for
    g01ey

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01eyf.html

    Parameters
    ----------
    n : int
        n, the number of observations in the sample.

    d : float
        Contains the test statistic, D_n^+ or D_n^-.

    Returns
    -------
    p : float
        The upper tail probability associated with the one sample
        Kolmogorov--Smirnov distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `d` < 0.0 or `d` > 1.0: `d` = *<value>*.
    """

    from ..base.stat import prob_kolmogorov1 as b_prob_kolmogorov1

    _return_scalar = b_prob_kolmogorov1(
        n,
        d,
    )
    return _return_scalar

def prob_kolmogorov2(n1, n2, d):
    u"""
    Computes probabilities for the two-sample Kolmogorov--Smirnov
    distribution.

    ``prob_kolmogorov2`` returns the probability associated with the
    upper tail of the Kolmogorov--Smirnov two sample distribution.

    For full information please refer to the NAG Library document for
    g01ez

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01ezf.html

    Parameters
    ----------
    n1 : int
        The number of observations in the first sample, n_1.

    n2 : int
        The number of observations in the second sample, n_2.

    d : float
        The test statistic D_{n_1,n_2}, for the two sample
        Kolmogorov--Smirnov goodness-of-fit test, that is the maximum
        difference between the empirical cumulative distribution
        functions (CDFs) of the two samples.

    Returns
    -------
    p : float
        The probability associated with the upper tail of the
        Kolmogorov--Smirnov two sample distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>* and `n2` = *<value>*.

            Constraint: `n1` >= 1 and `n2` >= 1.

        (`errno` 2)
            On entry, `d` < 0.0 or `d` > 1.0: `d` = *<value>*.

        (`errno` 3)
            The Smirnov approximation used for large samples did not
            converge in 200 iterations. The probability is set to 1.0.
    """

    from ..base.stat import prob_kolmogorov2 as b_prob_kolmogorov2

    _return_scalar = b_prob_kolmogorov2(
        n1,
        n2,
        d,
    )
    return _return_scalar

def inv_cdf_normal(p, tail='L'):
    u"""
    Computes deviates for the standard Normal distribution.

    ``inv_cdf_normal`` returns the deviate associated with the given
    probability of the standard Normal distribution.

    For full information please refer to the NAG Library document for
    g01fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01faf.html

    Parameters
    ----------
    p : float
        p, the probability from the standard Normal distribution as
        defined by `tail`.

    tail : str, length 1, optional
        Indicates which tail the supplied probability represents.

        `tail` = 'L'

          The lower probability, i.e., P(X <= x_p).

        `tail` = 'U'

          The upper probability, i.e., P(X >= x_p).

        `tail` = 'S'

          The two tail (significance level) probability, i.e., P(X >=
          |x_p|)+P(X <= -|x_p|).

        `tail` = 'C'

          The two tail (confidence interval) probability, i.e., P(X <=
          |x_p|)-P(X <= -|x_p|).

    Returns
    -------
    x : float
        The deviate associated with the given probability of the
        standard Normal distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.0.
    """

    from ..base.stat import inv_cdf_normal as b_inv_cdf_normal

    _return_scalar = b_inv_cdf_normal(
        tail,
        p,
    )
    return _return_scalar

def inv_cdf_students_t(p, df, tail='L'):
    u"""
    Computes deviates for Student's t-distribution.

    ``inv_cdf_students_t`` returns the deviate associated with the given
    tail probability of Student's t-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01fb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01fbf.html

    Parameters
    ----------
    p : float
        p, the probability from the required Student's t-distribution as
        defined by `tail`.

    df : float
        nu, the degrees of freedom of the Student's t-distribution.

    tail : str, length 1, optional
        Indicates which tail the supplied probability represents.

        `tail` = 'U'

          The upper tail probability, i.e., P(T >= t_p:nu).

        `tail` = 'L'

          The lower tail probability, i.e., P(T <= t_p:nu).

        `tail` = 'S'

          The two tail (significance level) probability, i.e., P(T >=
          |t_p|:nu)+P(T <= -|t_p|:nu).

        `tail` = 'C'

          The two tail (confidence interval) probability, i.e., P(T <=
          |t_p|:nu)-P(T <= -|t_p|:nu).

    Returns
    -------
    x : float
        The deviate associated with the given tail probability of
        Student's t-distribution with real degrees of freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The solution has failed to converge. However, the result
            should be a reasonable approximation.
    """

    from ..base.stat import inv_cdf_students_t as b_inv_cdf_students_t

    try:
        _return_scalar = b_inv_cdf_students_t(
            tail,
            p,
            df,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def inv_cdf_chisq(p, df):
    u"""
    Computes deviates for the chi^2 distribution.

    ``inv_cdf_chisq`` returns the deviate associated with the given
    lower tail probability of the chi^2-distribution with real degrees
    of freedom.

    For full information please refer to the NAG Library document for
    g01fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01fcf.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required
        chi^2-distribution.

    df : float
        nu, the degrees of freedom of the chi^2-distribution.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the chi^2-distribution with real degrees of freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` > 0.0.

        (`errno` 3)
            The probability is too close to 0.0 or 1.0.

        (`errno` 5)
            The series used to calculate the gamma function has failed
            to converge. This is an unlikely error exit.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The algorithm has failed to converge in *<value>*
            iterations. The result should be a reasonable approximation.
    """

    from ..base.stat import inv_cdf_chisq as b_inv_cdf_chisq

    try:
        _return_scalar = b_inv_cdf_chisq(
            p,
            df,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def inv_cdf_f(p, df1, df2):
    u"""
    Computes deviates for the F-distribution.

    ``inv_cdf_f`` returns the deviate associated with the given lower
    tail probability of the F or variance-ratio distribution with real
    degrees of freedom.

    For full information please refer to the NAG Library document for
    g01fd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01fdf.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required F-distribution.

    df1 : float
        The degrees of freedom of the numerator variance, nu_1.

    df2 : float
        The degrees of freedom of the denominator variance, nu_2.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the F or variance-ratio distribution with real degrees of
        freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 2)
            On entry, `df1` = *<value>* and `df2` = *<value>*.

            Constraint: `df1` > 0.0 and `df2` > 0.0.

        (`errno` 4)
            The probability is too close to 0.0 or 1.0. The value of f_p
            cannot be computed. This will only occur when the large
            sample approximations are used.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The solution has failed to converge. However, the result
            should be a reasonable approximation. Alternatively,
            ``inv_cdf_beta`` can be used with a suitable setting of the
            argument tol.
    """

    from ..base.stat import inv_cdf_f as b_inv_cdf_f

    try:
        _return_scalar = b_inv_cdf_f(
            p,
            df1,
            df2,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def inv_cdf_beta(p, a, b, tol=0.0):
    u"""
    Computes deviates for the beta distribution.

    ``inv_cdf_beta`` returns the deviate associated with the given lower
    tail probability of the beta distribution.

    For full information please refer to the NAG Library document for
    g01fe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01fef.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required beta
        distribution.

    a : float
        a, the first parameter of the required beta distribution.

    b : float
        b, the second parameter of the required beta distribution.

    tol : float, optional
        The relative accuracy required by you in the result. If
        ``inv_cdf_beta`` is entered with `tol` greater than or equal to
        1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the beta distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` <= 1.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` <= 10^6.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` <= 10^6.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The solution has failed to converge. However, the result
            should be a reasonable approximation. Requested accuracy not
            achieved when calculating beta probability. You should try
            setting `tol` larger.

        (`errno` 4)
            The requested accuracy has not been achieved. Use a larger
            value of `tol`. There is doubt concerning the accuracy of
            the computed result. 100 iterations of the Newton--Raphson
            method have been performed without satisfying the accuracy
            criterion (see [Further Comments]). The result should be a
            reasonable approximation of the solution.
    """

    from ..base.stat import inv_cdf_beta as b_inv_cdf_beta

    try:
        _return_scalar = b_inv_cdf_beta(
            p,
            a,
            b,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def inv_cdf_gamma(p, a, b, tol=0.0):
    u"""
    Computes deviates for the gamma distribution.

    ``inv_cdf_gamma`` returns the deviate associated with the given
    lower tail probability of the gamma distribution.

    For full information please refer to the NAG Library document for
    g01ff

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01fff.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required gamma
        distribution.

    a : float
        alpha, the shape parameter of the gamma distribution.

    b : float
        beta, the scale parameter of the gamma distribution.

    tol : float, optional
        The relative accuracy required by you in the results. The
        smallest recommended value is 50*delta, where delta =
        max(10^(-18), machine precision). If ``inv_cdf_gamma`` is
        entered with `tol` less than 50*delta or greater or equal to
        1.0, then 50*delta is used instead.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the gamma distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` <= 10^6.

        (`errno` 2)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            The probability is too close to 0.0 for the given `a` to
            enable the result to be calculated.

        (`errno` 5)
            The series used to calculate the gamma function has failed
            to converge. This is an unlikely error exit.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The algorithm has failed to converge in 100 iterations. A
            larger value of `tol` should be tried. The result may be a
            reasonable approximation.
    """

    from ..base.stat import inv_cdf_gamma as b_inv_cdf_gamma

    try:
        _return_scalar = b_inv_cdf_gamma(
            p,
            a,
            b,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def inv_cdf_studentized_range(p, v, ir):
    u"""
    Computes deviates for the Studentized range statistic.

    ``inv_cdf_studentized_range`` returns the deviate associated with
    the lower tail probability of the distribution of the Studentized
    range statistic.

    For full information please refer to the NAG Library document for
    g01fm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01fmf.html

    Parameters
    ----------
    p : float
        The lower tail probability for the Studentized range statistic,
        p_0.

    v : float
        v, the number of degrees of freedom.

    ir : int
        r, the number of groups.

    Returns
    -------
    x : float
        The deviate associated with the lower tail probability of the
        distribution of the Studentized range statistic.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: 0.0 < `p` < 1.0.

        (`errno` 1)
            On entry, `ir` = *<value>*.

            Constraint: `ir` >= 2.

        (`errno` 1)
            On entry, `v` = *<value>*.

            Constraint: `v` >= 1.0.

        (`errno` 2)
            The function was unable to find an upper bound for the value
            of q_0. This will be caused by p_0 being too close to 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            There is some doubt as to whether full accuracy has been
            achieved. The returned value should be a reasonable estimate
            of the true value.
    """

    from ..base.stat import inv_cdf_studentized_range as b_inv_cdf_studentized_range

    try:
        _return_scalar = b_inv_cdf_studentized_range(
            p,
            v,
            ir,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def inv_cdf_landau(x):
    u"""
    Landau inverse function Psi(x).

    ``inv_cdf_landau`` returns the value of the inverse Phi^-1(x) of the
    Landau distribution function.

    For full information please refer to the NAG Library document for
    g01ft

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01ftf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    invlan : float
        The value of the inverse Phi^-1(x) of the Landau distribution
        function.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` < 1.0.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.
    """

    from ..base.stat import inv_cdf_landau as b_inv_cdf_landau

    _return_scalar = b_inv_cdf_landau(
        x,
    )
    return _return_scalar

def prob_students_t_noncentral(t, df, delta, tol=0.0, maxit=100):
    u"""
    Computes probabilities for the non-central Student's t-distribution.

    ``prob_students_t_noncentral`` returns the lower tail probability
    for the noncentral Student's t-distribution.

    For full information please refer to the NAG Library document for
    g01gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01gbf.html

    Parameters
    ----------
    t : float
        t, the deviate from the Student's t-distribution with nu degrees
        of freedom.

    df : float
        nu, the degrees of freedom of the Student's t-distribution.

    delta : float
        delta, the noncentrality parameter of the Students
        t-distribution.

    tol : float, optional
        The absolute accuracy required by you in the results. If
        ``prob_students_t_noncentral`` is entered with `tol` greater
        than or equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of terms that are used in each of the
        summations.

    Returns
    -------
    p : float
        The lower tail probability for the noncentral Student's
        t-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.0.

        (`errno` 2)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 4)
            Unable to calculate the probability as it is too close to
            zero or one.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            One of the series has failed to converge with `maxit` =
            *<value>* and `tol` = *<value>*. Reconsider the requested
            tolerance and/or the maximum number of iterations.

        (`errno` 4)
            The probability is too close to 0 or 1. The returned value
            should be a reasonable estimate of the true value.
    """

    from ..base.stat import prob_students_t_noncentral as b_prob_students_t_noncentral

    try:
        _return_scalar = b_prob_students_t_noncentral(
            t,
            df,
            delta,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_chisq_noncentral(x, df, rlamda, tol=0.0, maxit=100):
    u"""
    Computes probabilities for the non-central chi^2 distribution.

    ``prob_chisq_noncentral`` returns the probability associated with
    the lower tail of the noncentral chi^2-distribution.

    For full information please refer to the NAG Library document for
    g01gc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01gcf.html

    Parameters
    ----------
    x : float
        The deviate from the noncentral chi^2-distribution with nu
        degrees of freedom and noncentrality parameter lambda.

    df : float
        nu, the degrees of freedom of the noncentral chi^2-distribution.

    rlamda : float
        lambda, the noncentrality parameter of the noncentral
        chi^2-distribution.

    tol : float, optional
        The required accuracy of the solution. If
        ``prob_chisq_noncentral`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of iterations to be performed.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the noncentral
        chi^2-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `df` = 0.0 and `rlamda` = 0.0.

            Constraint: `rlamda` > 0.0 if `df` = 0.0.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: `rlamda` >= 0.0.

        (`errno` 1)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 0.0.

        (`errno` 2)
            The initial value of the Poisson weight used in the
            summation of Equation (1) (see [Description]) was too small
            to be calculated. The computed probability is likely to be
            zero.

        (`errno` 3)
            The solution has failed to converge in *<value>* iterations.
            Consider increasing `maxit` or `tol`.

        (`errno` 4)
            The value of a term required in Equation (2) (see
            [Description]) is too large to be evaluated accurately. The
            most likely cause of this error is both `x` and `rlamda` are
            too large.

        (`errno` 5)
            The calculations for the central chi-square probability has
            failed to converge. A larger value of `tol` should be used.
    """

    from ..base.stat import prob_chisq_noncentral as b_prob_chisq_noncentral

    _return_scalar = b_prob_chisq_noncentral(
        x,
        df,
        rlamda,
        tol,
        maxit,
    )
    return _return_scalar

def prob_f_noncentral(f, df1, df2, rlamda, tol=0.0, maxit=500):
    u"""
    Computes probabilities for the non-central F-distribution.

    ``prob_f_noncentral`` returns the probability associated with the
    lower tail of the noncentral F or variance-ratio distribution.

    For full information please refer to the NAG Library document for
    g01gd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01gdf.html

    Parameters
    ----------
    f : float
        f, the deviate from the noncentral F-distribution.

    df1 : float
        The degrees of freedom of the numerator variance, nu_1.

    df2 : float
        The degrees of freedom of the denominator variance, nu_2.

    rlamda : float
        lambda, the noncentrality parameter.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``prob_f_noncentral`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of iterations to be used.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the noncentral
        F or variance-ratio distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `df2` = *<value>*.

            Constraint: `df2` > 0.0.

        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: 0.0 <= `rlamda` <= -2.0*log(U), where U is the
            safe range parameter as defined by ``machine.real_safe``.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `df1` = *<value>*.

            Constraint: 0.0 < `df1` <= 10^6.

        (`errno` 1)
            On entry, `f` = *<value>*.

            Constraint: `f` > 0.0.

        (`errno` 1)
            On entry, `df1` = *<value>*.

            Constraint: `df1` > 0.0.

        (`errno` 2)
            The solution has failed to converge in *<value>* iterations.
            Consider increasing `maxit` or `tol`.

        (`errno` 3)
            The required probability cannot be computed accurately. This
            may happen if the result would be very close to zero or one.
            Alternatively the values of `df1` and `f` may be too large.
            In the latter case you could try using a normal
            approximation, see Abramowitz and Stegun.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy was not achieved when calculating the
            initial value of the central F or chi^2 probability. You
            should try a larger value of `tol`. If the chi^2
            approximation is being used then ``prob_f_noncentral``
            returns zero otherwise the value returned should be an
            approximation to the correct value.
    """

    from ..base.stat import prob_f_noncentral as b_prob_f_noncentral

    try:
        _return_scalar = b_prob_f_noncentral(
            f,
            df1,
            df2,
            rlamda,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_beta_noncentral(x, a, b, rlamda, tol=0.0, maxit=500):
    u"""
    Computes probabilities for the non-central beta distribution.

    ``prob_beta_noncentral`` returns the probability associated with the
    lower tail of the noncentral beta distribution.

    For full information please refer to the NAG Library document for
    g01ge

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01gef.html

    Parameters
    ----------
    x : float
        beta, the deviate from the beta distribution, for which the
        probability P(B <= beta:a,b;lambda) is to be found.

    a : float
        a, the first parameter of the required beta distribution.

    b : float
        b, the second parameter of the required beta distribution.

    rlamda : float
        lambda, the noncentrality parameter of the required beta
        distribution.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``prob_beta_noncentral`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

        See [Accuracy] for the relationship between `tol` and `maxit`.

    maxit : int, optional
        The maximum number of iterations that the algorithm should use.

        See [Accuracy] for suggestions as to suitable values for `maxit`
        for different values of the arguments.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the noncentral
        beta distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `b` = *<value>*.

            Constraint: 0.0 < `b` <= 10^6.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: 0.0 < `a` <= 10^6.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: 0.0 <= `x` <= 1.0.

        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: 0.0 <= `rlamda` <= -2.0log(U), where U is the
            safe range parameter as defined by ``machine.real_safe``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The solution has failed to converge in *<value>* iterations.
            Consider increasing `maxit` or `tol`. The returned value
            will be an approximation to the correct value.

        (`errno` 3)
            The probability is too close to 0.0 or 1.0 for the algorithm
            to be able to calculate the required probability.
            ``prob_beta_noncentral`` will return 0.0 or 1.0 as
            appropriate. This should be a reasonable approximation.

        (`errno` 4)
            The required accuracy was not achieved when calculating the
            initial value of the beta distribution. You should try a
            larger value of `tol`. The returned value will be an
            approximation to the correct value.
    """

    from ..base.stat import prob_beta_noncentral as b_prob_beta_noncentral

    try:
        _return_scalar = b_prob_beta_noncentral(
            x,
            a,
            b,
            rlamda,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_bivariate_normal(x, y, rho):
    u"""
    Computes probability for the bivariate Normal distribution.

    ``prob_bivariate_normal`` returns the lower tail probability for the
    bivariate Normal distribution.

    For full information please refer to the NAG Library document for
    g01ha

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01haf.html

    Parameters
    ----------
    x : float
        x, the first argument for which the bivariate Normal
        distribution function is to be evaluated.

    y : float
        y, the second argument for which the bivariate Normal
        distribution function is to be evaluated.

    rho : float
        rho, the correlation coefficient.

    Returns
    -------
    p : float
        The lower tail probability for the bivariate Normal
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rho` = *<value>*.

            Constraint: `rho` <= 1.0.

        (`errno` 1)
            On entry, `rho` = *<value>*.

            Constraint: `rho` >= -1.0.
    """

    from ..base.stat import prob_bivariate_normal as b_prob_bivariate_normal

    _return_scalar = b_prob_bivariate_normal(
        x,
        y,
        rho,
    )
    return _return_scalar

def prob_multi_normal(xmu, sig, a=None, b=None, tol=0.0001):
    u"""
    Computes probabilities for the multivariate Normal distribution.

    ``prob_multi_normal`` returns the upper tail, lower tail or central
    probability associated with a multivariate Normal distribution of up
    to ten dimensions.

    For full information please refer to the NAG Library document for
    g01hb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01hbf.html

    Parameters
    ----------
    xmu : float, array-like, shape (n)
        mu, the mean vector of the multivariate Normal distribution.

    sig : float, array-like, shape (n, n)
        Sigma, the variance-covariance matrix of the multivariate Normal
        distribution. Only the lower triangle is referenced.

    a : None or float, array-like, shape (n), optional
        If upper tail or central probablilities are to be returned, `a`
        should supply the lower bounds, a_i, for i = 1,2,...,n.

    b : None or float, array-like, shape (n), optional
        If lower tail or central probablilities are to be returned, `b`
        should supply the upper bounds, b_i, for i = 1,2,...,n.

    tol : float, optional
        If n > 2 the relative accuracy required for the probability, and
        if the upper or the lower tail probability is requested then
        `tol` is also used to determine the cut-off points, see
        [Accuracy].

        If n = 1, `tol` is not referenced.

    Returns
    -------
    p : float
        The upper tail, lower tail or central probability associated
        with then multivariate Normal distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: 1 <= n <= 10.

        (`errno` 1)
            On entry, tail = *<value>*.

            Constraint: tail = 'L', 'U' or 'C'.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 2)
            On entry, the *<value>**<value>* value in `b` is less than
            or equal to the corresponding value in `a`.

        (`errno` 3)
            On entry, `sig` is not positive definite.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Full accuracy not achieved, relative accuracy = *<value>*.

        (`errno` 5)
            Accuracy requested by `tol` is too strict: `tol` =
            *<value>*.
    """

    from ..base.stat import prob_multi_normal as b_prob_multi_normal

    fun_name = 'naginterfaces.library.stat.prob_multi_normal'

    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sig'},
        sig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = b_prob_multi_normal(
            _n,
            _a,
            _b,
            _xmu,
            _sig,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_bivariate_students_t(df, rho, a=None, b=None):
    u"""
    Computes probabilities for the bivariate Student's t-distribution.

    ``prob_bivariate_students_t`` returns probabilities for the
    bivariate Student's t-distribution.

    For full information please refer to the NAG Library document for
    g01hc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01hcf.html

    Parameters
    ----------
    df : int
        nu, the degrees of freedom of the bivariate Student's
        t-distribution.

    rho : float
        rho, the correlation of the bivariate Student's t-distribution.

    a : None or float, array-like, shape (2), optional
        If upper tail or central probablilities are to be returned, `a`
        should supply the lower bounds, a_i, for i = 1,2,...,2.

    b : None or float, array-like, shape (2), optional
        If lower tail or central probablilities are to be returned, `b`
        should supply the upper bounds, b_i, for i = 1,2,...,2.

    Returns
    -------
    p : float
        The probabilities for the bivariate Student's t-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, tail = *<value>*.

            Constraint: tail = 'L', 'U' or 'C'.

        (`errno` 3)
            On entry, `b`[i-1] <= `a`[i-1] for central probability, for
            some i = 1,2.

        (`errno` 4)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.

        (`errno` 5)
            On entry, `rho` = *<value>*.

            Constraint: -1.0 <= `rho` <= 1.0.
    """

    from ..base.stat import prob_bivariate_students_t as b_prob_bivariate_students_t

    fun_name = 'naginterfaces.library.stat.prob_bivariate_students_t'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_prob_bivariate_students_t(
        _a,
        _b,
        df,
        rho,
    )
    return _return_scalar

def prob_multi_students_t(tail, a, b, nu, delta, iscov, rc, epsabs=0.0, epsrel=0.001, numsub=350, nsampl=8, fmax=None):
    u"""
    Computes the probability for the multivariate Student's
    t-distribution.

    ``prob_multi_students_t`` returns a probability associated with a
    multivariate Student's t-distribution.

    For full information please refer to the NAG Library document for
    g01hd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01hdf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (n)
        Defines the calculated probability, set `tail`[i-1] to:

        `tail`[i-1] = 'L'

          If the ith lower limit a_i is negative infinity.

        `tail`[i-1] = 'U'

          If the ith upper limit b_i is infinity.

        `tail`[i-1] = 'C'

          If both a_i and b_i are finite.

    a : float, array-like, shape (n)
        a_i, for i = 1,2,...,n, the lower integral limits of the
        calculation.

        If `tail`[i-1] = 'L', `a`[i-1] is not referenced and the ith
        lower limit of integration is - infinity.

    b : float, array-like, shape (n)
        b_i, for i = 1,2,...,n, the upper integral limits of the
        calculation.

        If `tail`[i-1] = 'U', `b`[i-1] is not referenced and the ith
        upper limit of integration is infinity.

    nu : float
        nu, the degrees of freedom.

    delta : float, array-like, shape (n)
        `delta`[i-1] the noncentrality parameter for the ith dimension,
        for i = 1,2,...,n; set `delta`[i-1] = 0 for the central
        probability.

    iscov : int
        Set `iscov` = 1 if the covariance matrix is supplied and `iscov`
        = 2 if the correlation matrix is supplied.

    rc : float, array-like, shape (n, n)
        The lower triangle of either the covariance matrix (if `iscov` =
        1) or the correlation matrix (if `iscov` = 2). In either case
        the array elements corresponding to the upper triangle of the
        matrix need not be set.

    epsabs : float, optional
        epsilon_a, the absolute accuracy requested in the approximation.
        If `epsabs` is negative, the absolute value is used.

    epsrel : float, optional
        epsilon_r, the relative accuracy requested in the approximation.
        If `epsrel` is negative, the absolute value is used.

    numsub : int, optional
        If quadrature is used, the number of sub-intervals used by the
        quadrature algorithm; otherwise `numsub` is not referenced.

    nsampl : int, optional
        If quadrature is used, `nsampl` is not referenced; otherwise
        `nsampl` is the number of samples used to estimate the error in
        the approximation.

    fmax : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 1000*n.

        If a number theoretic approach is used, the maximum number of
        evaluations for each integrand function.

    Returns
    -------
    p : float
        The probability associated with the multivariate Student's
        t-distribution.

    rc : float, ndarray, shape (n, n)
        The strict upper triangle of `rc` contains the correlation
        matrix used in the calculations.

    errest : float
        An estimate of the error in the calculated probability.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: 1 < n <= 1000.

        (`errno` 2)
            On entry, `tail`[k-1] = *<value>*.

            Constraint: `tail`[k-1] = 'L', 'U' or 'C'.

        (`errno` 4)
            On entry, k = *<value>*.

            Constraint: `b`[k-1] > `a`[k-1] for a central probability.

        (`errno` 5)
            On entry, `nu` = *<value>*.

            Constraint: degrees of freedom `nu` > 0.0.

        (`errno` 8)
            On entry, `iscov` = *<value>*.

            Constraint: `iscov` = 1 or 2.

        (`errno` 9)
            On entry, the information supplied in `rc` is invalid.

        (`errno` 12)
            On entry, `numsub` = *<value>*.

            Constraint: `numsub` >= 1.

        (`errno` 13)
            On entry, `nsampl` = *<value>*.

            Constraint: `nsampl` >= 1.

        (`errno` 14)
            On entry, `fmax` = *<value>*.

            Constraint: `fmax` >= 1.
    """

    from ..base.stat import prob_multi_students_t as b_prob_multi_students_t

    fun_name = 'naginterfaces.library.stat.prob_multi_students_t'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _n_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _delta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'delta'},
        delta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rc'},
        rc,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fmax = 1000*_n_py if fmax is None else fmax
    _return_scalars = b_prob_multi_students_t(
        _n,
        _tail,
        _a,
        _b,
        nu,
        _delta,
        iscov,
        _rc,
        epsabs,
        epsrel,
        numsub,
        nsampl,
        _fmax,
    )
    _rc_py = _rc
    _p_py = _return_scalars.p
    _errest_py = _return_scalars.errest
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'rc', 'errest'],
    )(
        _p_py,
        _rc_py,
        _errest_py,
    )
    return _return_data

def prob_chisq_noncentral_lincomb(a, mult, rlamda, c, tol=0.0, maxit=500):
    u"""
    Computes probability for a positive linear combination of chi^2
    variables.

    ``prob_chisq_noncentral_lincomb`` returns the lower tail probability
    of a distribution of a positive linear combination of chi^2 random
    variables.

    For full information please refer to the NAG Library document for
    g01jc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01jcf.html

    Parameters
    ----------
    a : float, array-like, shape (n)
        The weights, a_1,a_2,...,a_n.

    mult : int, array-like, shape (n)
        The degrees of freedom, m_1,m_2,...,m_n.

    rlamda : float, array-like, shape (n)
        The noncentrality parameters, lambda_1,lambda_2,...,lambda_n.

    c : float
        c, the point for which the lower tail probability is to be
        evaluated.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``prob_chisq_noncentral_lincomb`` is entered with `tol` greater
        than or equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of terms that should be used during the
        summation.

    Returns
    -------
    p : float
        The lower tail probability associated with the linear
        combination of n chi^2 random variables with m_j degrees of
        freedom, and noncentrality parameters lambda_j, for j =
        1,2,...,n.

    pdf : float
        The value of the probability density function of the linear
        combination of chi^2 variables.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `c` = *<value>*.

            Constraint: `c` >= 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `rlamda`[*<value>*] = *<value>*.

            Constraint: `rlamda`[i] >= 0.0, for i = 0,...,n-1.

        (`errno` 2)
            On entry, `mult`[*<value>*] = *<value>*.

            Constraint: `mult`[i] >= 1, for i = 0,...,n-1.

        (`errno` 2)
            On entry, `a`[*<value>*] = *<value>*.

            Constraint: `a`[i] > 0.0, for i = 0,...,n-1.

        (`errno` 3)
            The central chi^2 calculation has failed to converge. This
            is an unlikely exit. A larger value of `tol` should be
            tried.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The solution has failed to converge within `maxit`
            iterations. A larger value of `maxit` or `tol` should be
            used. The returned value should be a reasonable
            approximation to the correct value.

        (`errno` 5)
            The solution appears to be too close to 0 or 1 for accurate
            calculation. The value returned is 0 or 1 as appropriate.
    """

    from ..base.stat import prob_chisq_noncentral_lincomb as b_prob_chisq_noncentral_lincomb

    fun_name = 'naginterfaces.library.stat.prob_chisq_noncentral_lincomb'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _mult = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mult'},
        mult,
        copy=False,
    )
    _rlamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rlamda'},
        rlamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = b_prob_chisq_noncentral_lincomb(
            _a,
            _mult,
            _rlamda,
            _n,
            c,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _p_py = _return_scalars.p
    _pdf_py = _return_scalars.pdf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'pdf'],
    )(
        _p_py,
        _pdf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_chisq_lincomb(rlam, d, c, method='D'):
    u"""
    Computes lower tail probability for a linear combination of
    (central) chi^2 variables.

    ``prob_chisq_lincomb`` calculates the lower tail probability for a
    linear combination of (central) chi^2 variables.

    For full information please refer to the NAG Library document for
    g01jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01jdf.html

    Parameters
    ----------
    rlam : float, array-like, shape (n)
        The weights, lambda_i, for i = 1,2,...,n, of the central chi^2
        variables.

    d : float
        d, the multiplier of the central chi^2 variables.

    c : float
        c, the value of the constant.

    method : str, length 1, optional
        Indicates whether Pan's, Imhof's or an appropriately selected
        procedure is to be used.

        `method` = 'P'

          Pan's method is used.

        `method` = 'I'

          Imhof's method is used.

        `method` = 'D'

          Pan's method is used if lambda_i^*, for i = 1,2,...,n are at
          least 1% distinct and n <= 60; otherwise Imhof's method is
          used.

    Returns
    -------
    prob : float
        The lower tail probability for the linear combination of central
        chi^2 variables.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'P', 'I' or 'D'.

        (`errno` 1)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `rlam`[i-1] = `d` for all values of i, for i =
            1,2,...,n.
    """

    from ..base.stat import prob_chisq_lincomb as b_prob_chisq_lincomb

    fun_name = 'naginterfaces.library.stat.prob_chisq_lincomb'

    _rlam = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rlam'},
        rlam,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_rlam, {'fun_name': fun_name, 'entity': 'rlam'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = b_prob_chisq_lincomb(
        method,
        _n,
        _rlam,
        d,
        c,
    )
    _prob_py = _return_scalar
    return _prob_py

def pdf_normal(x, xmean, xstd):
    u"""
    Calculates the value for the probability density function of the
    Normal distribution at a chosen point.

    ``pdf_normal`` returns the value of the probability density function
    (PDF) for the Normal (Gaussian) distribution with mean mu and
    variance sigma^2 at a point x.

    For full information please refer to the NAG Library document for
    g01ka

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01kaf.html

    Parameters
    ----------
    x : float
        x, the value at which the PDF is to be evaluated.

    xmean : float
        mu, the mean of the Normal distribution.

    xstd : float
        sigma, the standard deviation of the Normal distribution.

    Returns
    -------
    pdf : float
        The value of the PDF.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `xstd` = *<value>*.

            Constraint: `xstd`*sqrt(2.0*pi) > U, where U is the safe
            range parameter as defined by ``machine.real_safe``.

        (`errno` 2)
            Computation abandoned owing to underflow of
            1/(sigma*sqrt(2*pi)).

        (`errno` 3)
            Computation abandoned owing to an internal calculation
            overflowing.
    """

    from ..base.stat import pdf_normal as b_pdf_normal

    _return_scalar = b_pdf_normal(
        x,
        xmean,
        xstd,
    )
    return _return_scalar

def pdf_gamma(x, a, b):
    u"""
    Calculates the value for the probability density function of the
    gamma distribution at a chosen point.

    ``pdf_gamma`` returns the value of the probability density function
    (PDF) for the gamma distribution with shape parameter alpha and
    scale parameter beta at a point x.

    For full information please refer to the NAG Library document for
    g01kf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01kff.html

    Parameters
    ----------
    x : float
        x, the value at which the PDF is to be evaluated.

    a : float
        alpha, the shape parameter of the gamma distribution.

    b : float
        beta, the scale parameter of the gamma distribution.

    Returns
    -------
    pdf : float
        The value of the PDF.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 2)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 3)
            Computation abandoned owing to overflow due to extreme
            parameter values.
    """

    from ..base.stat import pdf_gamma as b_pdf_gamma

    _return_scalar = b_pdf_gamma(
        x,
        a,
        b,
    )
    return _return_scalar

def pdf_gamma_vector(ilog, x, a, b):
    u"""
    Computes a vector of values for the probability density function of
    the gamma distribution.

    ``pdf_gamma_vector`` returns a number of values of the probability
    density function (PDF), or its logarithm, for the gamma
    distribution.

    For full information please refer to the NAG Library document for
    g01kk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01kkf.html

    Parameters
    ----------
    ilog : int
        The value of `ilog` determines whether the logarithmic value is
        returned in `pdf`.

        `ilog` = 0

          f(x_i, alpha_i, beta_i), the probability density function is
          returned.

        `ilog` = 1

          log(f(x_i, alpha_i, beta_i)), the logarithm of the probability
          density function is returned.

    x : float, array-like, shape (lx)
        x_i, the values at which the PDF is to be evaluated.

    a : float, array-like, shape (la)
        alpha_i, the shape parameter.

    b : float, array-like, shape (lb)
        beta_i, the scale parameter.

    Returns
    -------
    pdf : float, ndarray, shape (max(lx, la, lb))
        f(x_i, alpha_i, beta_i) or log(f(x_i, alpha_i, beta_i)).

    ivalid : int, ndarray, shape (max(lx, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          alpha_i <= 0.0.

        `ivalid`[i-1] = 2

          beta_i <= 0.0.

        `ivalid`[i-1] = 3

          (x_i)/(beta_i) overflows, the value returned should be a
          reasonable approximation.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `ilog` = *<value>*.

            Constraint: `ilog` = 0 or 1.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x`, `a` or `b` was invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import pdf_gamma_vector as b_pdf_gamma_vector

    fun_name = 'naginterfaces.library.stat.pdf_gamma_vector'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _pdf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lx_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lx_py, _la_py, _lb_py),),
    )
    try:
        b_pdf_gamma_vector(
            ilog,
            _lx,
            _x,
            _la,
            _a,
            _lb,
            _b,
            _pdf,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _pdf_py = _pdf
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pdf', 'ivalid'],
    )(
        _pdf_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def pdf_normal_vector(ilog, x, xmu, xstd):
    u"""
    Computes a vector of values for the probability density function of
    the Normal distribution.

    ``pdf_normal_vector`` returns a number of values of the probability
    density function (PDF), or its logarithm, for the Normal (Gaussian)
    distributions.

    For full information please refer to the NAG Library document for
    g01kq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01kqf.html

    Parameters
    ----------
    ilog : int
        The value of `ilog` determines whether the logarithmic value is
        returned in PDF.

        `ilog` = 0

          f(x_i, mu_i, sigma_i), the probability density function is
          returned.

        `ilog` = 1

          log(f(x_i, mu_i, sigma_i)), the logarithm of the probability
          density function is returned.

    x : float, array-like, shape (lx)
        x_i, the values at which the PDF is to be evaluated.

    xmu : float, array-like, shape (lxmu)
        mu_i, the means.

    xstd : float, array-like, shape (lxstd)
        sigma_i, the standard deviations.

    Returns
    -------
    pdf : float, ndarray, shape (max(lx, lxstd, lxmu))
        f(x_i, mu_i, sigma_i) or log(f(x_i, mu_i, sigma_i)).

    ivalid : int, ndarray, shape (max(lx, lxstd, lxmu))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          sigma_i < 0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `ilog` = *<value>*.

            Constraint: `ilog` = 0 or 1.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lxmu > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lxstd > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `xstd` was invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import pdf_normal_vector as b_pdf_normal_vector

    fun_name = 'naginterfaces.library.stat.pdf_normal_vector'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxmu_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _lxmu = _b_utils._EngineIntScalarType(
        _lxmu_py, {'fun_name': fun_name, 'entity': '_lxmu'},
    ).to_ctypes()
    _xstd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xstd'},
        xstd,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxstd_py = _b_utils._EngineFloat64ArrayType.get_shape(_xstd, {'fun_name': fun_name, 'entity': 'xstd'}, exp_rank=1)[0]
    _lxstd = _b_utils._EngineIntScalarType(
        _lxstd_py, {'fun_name': fun_name, 'entity': '_lxstd'},
    ).to_ctypes()
    _pdf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lx_py, _lxstd_py, _lxmu_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lx_py, _lxstd_py, _lxmu_py),),
    )
    try:
        b_pdf_normal_vector(
            ilog,
            _lx,
            _x,
            _lxmu,
            _xmu,
            _lxstd,
            _xstd,
            _pdf,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _pdf_py = _pdf
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pdf', 'ivalid'],
    )(
        _pdf_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def pdf_multi_normal_vector(ilog, x, xmu, iuld, sig):
    u"""
    Computes a vector of values for the probability density function of
    the multivariate Normal distribution.

    ``pdf_multi_normal_vector`` returns a number of values of the
    probability density function (PDF), or its logarithm, for the
    multivariate Normal (Gaussian) distribution.

    For full information please refer to the NAG Library document for
    g01lb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01lbf.html

    Parameters
    ----------
    ilog : int
        The value of `ilog` determines whether the logarithmic value is
        returned in PDF.

        `ilog` = 0

          f(X:mu, Sigma), the probability density function is returned.

        `ilog` = 1

          log(f(X:mu, Sigma)), the logarithm of the probability density
          function is returned.

    x : float, array-like, shape (n, k)
        X, the matrix of k points at which to evaluate the probability
        density function, with the ith dimension for the jth point held
        in `x`[i-1,j-1].

    xmu : float, array-like, shape (n)
        mu, the mean vector of the multivariate Normal distribution.

    iuld : int
        Indicates the form of Sigma and how it is stored in `sig`.

        `iuld` = 1

          `sig` holds the lower triangular portion of Sigma.

        `iuld` = 2

          `sig` holds the upper triangular portion of Sigma.

        `iuld` = 3

          Sigma is a diagonal matrix and `sig` only holds the diagonal
          elements.

        `iuld` = 4

          `sig` holds the lower Cholesky decomposition, L such that
          L*L^T = Sigma.

        `iuld` = 5

          `sig` holds the upper Cholesky decomposition, U such that
          U^T*U = Sigma.

    sig : float, array-like, shape (:, n)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `iuld`=3: 1; otherwise: n.

        Information defining the variance-covariance matrix, Sigma.

        `iuld` = 1 or 2

          `sig` must hold the lower or upper portion of Sigma, with
          Sigma_{ij} held in `sig`[i-1,j-1]. The supplied
          variance-covariance matrix must be positive semidefinite.

        `iuld` = 3

          Sigma is a diagonal matrix and the ith diagonal element,
          Sigma_{ii}, must be held in `sig`[0,i-1]

        `iuld` = 4 or 5

          `sig` must hold L or U, the lower or upper Cholesky
          decomposition of Sigma, with L_{ij} or U_{ij} held in
          `sig`[i-1,j-1], depending on the value of `iuld`. No check is
          made that L*L^T or U^T*U is a valid variance-covariance
          matrix. The diagonal elements of the supplied L or U must be
          greater than zero

    Returns
    -------
    pdf : float, ndarray, shape (k)
        f(X:mu, Sigma) or log(f(X:mu, Sigma)) depending on the value of
        `ilog`.

    rank : int
        r, rank of Sigma.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ilog` = *<value>*.

            Constraint: `ilog` = 0 or 1.

        (`errno` 21)
            On entry, k = *<value>*.

            Constraint: k >= 0.

        (`errno` 31)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 71)
            On entry, `iuld` = *<value>*.

            Constraint: `iuld` = 1, 2, 3, 4 or 5.

        (`errno` 81)
            On entry, Sigma is not positive semidefinite.

        (`errno` 82)
            On entry, at least one diagonal element of Sigma is less
            than or equal to 0.

        (`errno` 83)
            On entry, Sigma is not positive definite and eigenvalue
            decomposition failed.

        (`errno` 92)
            On entry, ldsig = *<value>*.

            Constraint: if `iuld` != 3, ldsig >= n.
    """

    from ..base.stat import pdf_multi_normal_vector as b_pdf_multi_normal_vector

    fun_name = 'naginterfaces.library.stat.pdf_multi_normal_vector'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sig'},
        sig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pdf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    _return_scalar = b_pdf_multi_normal_vector(
        ilog,
        _k,
        _n,
        _x,
        _xmu,
        iuld,
        _sig,
        _pdf,
    )
    _pdf_py = _pdf
    _rank_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pdf', 'rank'],
    )(
        _pdf_py,
        _rank_py,
    )
    return _return_data

def mills_ratio(x):
    u"""
    Computes reciprocal of Mills' Ratio.

    ``mills_ratio`` returns the reciprocal of Mills' Ratio.

    For full information please refer to the NAG Library document for
    g01mb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01mbf.html

    Parameters
    ----------
    x : float
        x, the argument of the reciprocal of Mills' Ratio.

    Returns
    -------
    rmr : float
        The reciprocal of Mills' Ratio.
    """

    from ..base.stat import mills_ratio as b_mills_ratio

    _return_scalar = b_mills_ratio(
        x,
    )
    return _return_scalar

def pdf_landau(x):
    u"""
    Landau density function phi(lambda).

    ``pdf_landau`` returns the value of the Landau density function
    phi(lambda).

    For full information please refer to the NAG Library document for
    g01mt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01mtf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    Returns
    -------
    pdf : float
        The value of the Landau density function phi(lambda).
    """

    from ..base.stat import pdf_landau as b_pdf_landau

    _return_scalar = b_pdf_landau(
        x,
    )
    return _return_scalar

def pdf_vavilov(x, comm):
    u"""
    Vavilov density function phi_V(lambda;kappa,beta^2).

    ``pdf_vavilov`` returns the value of the Vavilov density function
    phi_V(lambda;kappa,beta^2).

    It is intended to be used after a call to ``init_vavilov``.

    For full information please refer to the NAG Library document for
    g01mu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01muf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``init_vavilov``.

    Returns
    -------
    pdf : float
        The value of the Vavilov density function
        phi_V(lambda;kappa,beta^2).

    Raises
    ------
    NagValueError
        (`errno` 1)
            Either the initialization function has not been called prior
            to the first call of this function or a communication array
            has become corrupted.
    """

    from ..base.stat import pdf_vavilov as b_pdf_vavilov

    _return_scalar = b_pdf_vavilov(
        x,
        comm,
    )
    return _return_scalar

def moments_quad_form(a, sigma, l, emu=None):
    u"""
    Cumulants and moments of quadratic forms in Normal variables.

    ``moments_quad_form`` computes the cumulants and moments of
    quadratic forms in Normal variates.

    For full information please refer to the NAG Library document for
    g01na

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01naf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A. Only the lower triangle is
        referenced.

    sigma : float, array-like, shape (n, n)
        The n by n variance-covariance matrix Sigma. Only the lower
        triangle is referenced.

    l : int
        The required number of cumulants, and moments if specified.

    emu : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if mean='M': n; otherwise: 1.

        If mean = 'M', `emu` must contain the n elements of the vector
        mu.

        If mean = 'Z', `emu` is not referenced.

    Returns
    -------
    rkum : float, ndarray, shape (`l`)
        The `l` cumulants of the quadratic form.

    rmom : float, ndarray, shape (:)
        If mom = 'M', the `l` moments of the quadratic form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` <= 12.

        (`errno` 1)
            On entry, mean = *<value>*.

            Constraint: mean = 'Z' or 'M'.

        (`errno` 1)
            On entry, mom = *<value>*.

            Constraint: mom = 'C' or 'M'.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 2)
            On entry, `sigma` is not positive definite.
    """

    from ..base.stat import moments_quad_form as b_moments_quad_form

    fun_name = 'naginterfaces.library.stat.moments_quad_form'

    _mom = 'M'
    _mom_py = _mom
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _emu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'emu'},
        emu,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sigma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sigma'},
        sigma,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _l = _b_utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = _b_utils._EngineIntScalarType.to_py(_l)
    _rkum = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_l_py,),
    )
    if _mom_py in ('M', 'm'):
        _rmom_dim1 = _l_py
    else:
        _rmom_dim1 = 1
    exp_shape_rmom = (_rmom_dim1,)
    _rmom = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_rmom,
    )
    b_moments_quad_form(
        _mom,
        _n,
        _a,
        _emu,
        _sigma,
        _l,
        _rkum,
        _rmom,
    )
    _rkum_py = _rkum
    _rmom_py = _rmom
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['rkum', 'rmom'],
    )(
        _rkum_py,
        _rmom_py,
    )
    return _return_data

def moments_ratio_quad_forms(a, b, sigma, l1, l2, eps, c=None, ela=None, emu=None):
    u"""
    Moments of ratios of quadratic forms in Normal variables, and
    related statistics.

    ``moments_ratio_quad_forms`` computes the moments of ratios of
    quadratic forms in Normal variables and related statistics.

    For full information please refer to the NAG Library document for
    g01nb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01nbf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A. Only the lower triangle is
        referenced.

    b : float, array-like, shape (n, n)
        The n by n positive semidefinite symmetric matrix B. Only the
        lower triangle is referenced.

    sigma : float, array-like, shape (n, n)
        The n by n variance-covariance matrix Sigma. Only the lower
        triangle is referenced.

    l1 : int
        The first moment to be computed, l_1.

    l2 : int
        The last moment to be computed, l_2.

    eps : float
        The relative accuracy required for the moments, this value is
        also used in the checks for the existence of the moments.

        If `eps` = 0.0, a value of sqrt(epsilon) where epsilon is the
        machine precision used.

    c : None or float, array-like, shape (:, n), optional
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `c` is not None: n; otherwise: 0.

        If `c` is not None, `c` must contain the n by n symmetric matrix
        C; only the lower triangle is referenced.

    ela : None or float, array-like, shape (n), optional
        If `ela` is not None, `ela` must contain the vector a of length
        n, otherwise `ela` is not referenced.

    emu : None or float, array-like, shape (n), optional
        If `emu` is not None, `emu` must contain the n elements of the
        vector mu.

    Returns
    -------
    lmax : int
        The highest moment computed, l_MAX. This will be l_2 if no
        exception or warning is raised on exit.

    rmom : float, ndarray, shape (`l2`-`l1`+1)
        The l_1 to l_MAX moments.

    abserr : float
        The estimated maximum absolute error in any computed moment.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l2` = *<value>*.

            Constraint: `l2` <= 12.

        (`errno` 1)
            On entry, `l1` = *<value>* and `l2` = *<value>*.

            Constraint: `l2` >= `l1`.

        (`errno` 1)
            On entry, `l1` = *<value>*.

            Constraint: `l1` >= 1.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: if `eps` != 0.0, `eps` >= machine precision.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 2)
            On entry, `sigma` is not positive definite.

        (`errno` 2)
            On entry, `b` is not positive semidefinite or is null.

        (`errno` 3)
            Only *<value>* moments exist, less than `l1` = *<value>*,
            therefore none of the required moments can be computed.

        (`errno` 4)
            The matrix L^TBL is not positive semidefinite or is null.

        (`errno` 5)
            The computation to compute the eigenvalues required in the
            calculation of moments has failed to converge: this is an
            unlikely error exit.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            Only some of the required moments have been computed, the
            highest is given by `lmax`.

        (`errno` 7)
            The required accuracy has not been achieved in the
            integration. An estimate of the accuracy is returned in
            `abserr`.
    """

    from ..base.stat import moments_ratio_quad_forms as b_moments_ratio_quad_forms

    fun_name = 'naginterfaces.library.stat.moments_ratio_quad_forms'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ela = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ela'},
        ela,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _emu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'emu'},
        emu,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sigma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sigma'},
        sigma,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _l2 = _b_utils._EngineIntScalarType(
        l2, {'fun_name': fun_name, 'entity': 'l2'},
    ).to_ctypes()
    _l2_py = _b_utils._EngineIntScalarType.to_py(_l2)
    _l1 = _b_utils._EngineIntScalarType(
        l1, {'fun_name': fun_name, 'entity': 'l1'},
    ).to_ctypes()
    _l1_py = _b_utils._EngineIntScalarType.to_py(_l1)
    _rmom = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_l2_py-_l1_py+1,),
    )
    try:
        _return_scalars = b_moments_ratio_quad_forms(
            _n,
            _a,
            _b,
            _c,
            _ela,
            _emu,
            _sigma,
            _l1,
            _l2,
            _rmom,
            eps,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _lmax_py = _return_scalars.lmax
    _rmom_py = _rmom
    _abserr_py = _return_scalars.abserr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lmax', 'rmom', 'abserr'],
    )(
        _lmax_py,
        _rmom_py,
        _abserr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def pdf_landau_moment1(x):
    u"""
    Landau first moment function Phi_1(x) .

    ``pdf_landau_moment1`` returns the value of the first moment
    Phi_1(x) of the Landau density function.

    For full information please refer to the NAG Library document for
    g01pt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01ptf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    mom1 : float
        The value of the first moment Phi_1(x) of the Landau density
        function.
    """

    from ..base.stat import pdf_landau_moment1 as b_pdf_landau_moment1

    _return_scalar = b_pdf_landau_moment1(
        x,
    )
    return _return_scalar

def pdf_landau_moment2(x):
    u"""
    Landau second moment function Phi_2(x) .

    ``pdf_landau_moment2`` returns the value of the second moment
    Phi_2(x) of the Landau density function.

    For full information please refer to the NAG Library document for
    g01qt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01qtf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    mom2 : float
        The value of the second moment Phi_2(x) of the Landau density
        function.
    """

    from ..base.stat import pdf_landau_moment2 as b_pdf_landau_moment2

    _return_scalar = b_pdf_landau_moment2(
        x,
    )
    return _return_scalar

def pdf_landau_deriv(x):
    u"""
    Landau derivative function phi'(lambda).

    ``pdf_landau_deriv`` returns the value of the derivative
    phi'(lambda) of the Landau density function.

    For full information please refer to the NAG Library document for
    g01rt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01rtf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    Returns
    -------
    deriv : float
        The value of the derivative phi'(lambda) of the Landau density
        function.
    """

    from ..base.stat import pdf_landau_deriv as b_pdf_landau_deriv

    _return_scalar = b_pdf_landau_deriv(
        x,
    )
    return _return_scalar

def prob_normal_vector(tail, x, xmu, xstd):
    u"""
    Computes a vector of probabilities for the standard Normal
    distribution.

    ``prob_normal_vector`` returns a number of one or two tail
    probabilities for the Normal distribution.

    For full information please refer to the NAG Library document for
    g01sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01saf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the returned probabilities should
        represent. Letting Z denote a variate from a standard Normal
        distribution, and z_i = (x_i-mu_i)/(sigma_i), then for j =
        (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(lx, ltail, lxmu,
        lxstd):

        `tail`[j] = 'L'

          The lower tail probability is returned, i.e., p_i = P(Z <=
          z_i).

        `tail`[j] = 'U'

          The upper tail probability is returned, i.e., p_i = P(Z >=
          z_i).

        `tail`[j] = 'C'

          The two tail (confidence interval) probability is returned,
          i.e., p_i = P(Z <= |z_i|)-P(Z <= -|z_i|).

        `tail`[j] = 'S'

          The two tail (significance level) probability is returned,
          i.e., p_i = P(Z >= |z_i|)+P(Z <= -|z_i|).

    x : float, array-like, shape (lx)
        x_i, the Normal variate values.

    xmu : float, array-like, shape (lxmu)
        mu_i, the means.

    xstd : float, array-like, shape (lxstd)
        sigma_i, the standard deviations.

    Returns
    -------
    p : float, ndarray, shape (max(lx, ltail, lxmu, lxstd))
        p_i, the probabilities for the Normal distribution.

    ivalid : int, ndarray, shape (max(lx, ltail, lxmu, lxstd))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          p_i.

        `ivalid`[i-1] = 2

          On entry, sigma_i <= 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, ltail = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, lx = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, lxmu = *<value>*.

            Constraint: lxmu > 0.

        (`errno` 5)
            On entry, lxstd = *<value>*.

            Constraint: lxstd > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail` or `xstd` was
            invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_normal_vector as b_prob_normal_vector

    fun_name = 'naginterfaces.library.stat.prob_normal_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxmu_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _lxmu = _b_utils._EngineIntScalarType(
        _lxmu_py, {'fun_name': fun_name, 'entity': '_lxmu'},
    ).to_ctypes()
    _xstd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xstd'},
        xstd,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxstd_py = _b_utils._EngineFloat64ArrayType.get_shape(_xstd, {'fun_name': fun_name, 'entity': 'xstd'}, exp_rank=1)[0]
    _lxstd = _b_utils._EngineIntScalarType(
        _lxstd_py, {'fun_name': fun_name, 'entity': '_lxstd'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lx_py, _ltail_py, _lxmu_py, _lxstd_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lx_py, _ltail_py, _lxmu_py, _lxstd_py),),
    )
    try:
        b_prob_normal_vector(
            _ltail,
            _tail,
            _lx,
            _x,
            _lxmu,
            _xmu,
            _lxstd,
            _xstd,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'ivalid'],
    )(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_students_t_vector(tail, t, df):
    u"""
    Computes a vector of probabilities for the Student's t-distribution.

    ``prob_students_t_vector`` returns a number of one or two tail
    probabilities for the Student's t-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01sbf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the returned probabilities should
        represent. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lt, ldf):

        `tail`[j] = 'L'

          The lower tail probability is returned, i.e., p_i = (T_i <=
          t_i:nu_i).

        `tail`[j] = 'U'

          The upper tail probability is returned, i.e., p_i = (T_i >=
          t_i:nu_i).

        `tail`[j] = 'C'

          The two tail (confidence interval) probability is returned,
          i.e., p_i = (T_i <= |t_i|:nu_i)-(T_i <= -|t_i|:nu_i).

        `tail`[j] = 'S'

          The two tail (significance level) probability is returned,
          i.e., p_i = (T_i >= |t_i|:nu_i)+(T_i <= -|t_i|:nu_i).

    t : float, array-like, shape (lt)
        t_i, the values of the Student's t variates.

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the Student's t-distribution.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, lt, ldf))
        p_i, the probabilities for the Student's t distribution.

    ivalid : int, ndarray, shape (max(ltail, lt, ldf))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          p_i.

        `ivalid`[i-1] = 2

          On entry, nu_i < 1.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lt > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail` or `df` was invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_students_t_vector as b_prob_students_t_vector

    fun_name = 'naginterfaces.library.stat.prob_students_t_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lt_py = _b_utils._EngineFloat64ArrayType.get_shape(_t, {'fun_name': fun_name, 'entity': 't'}, exp_rank=1)[0]
    _lt = _b_utils._EngineIntScalarType(
        _lt_py, {'fun_name': fun_name, 'entity': '_lt'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lt_py, _ldf_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lt_py, _ldf_py),),
    )
    try:
        b_prob_students_t_vector(
            _ltail,
            _tail,
            _lt,
            _t,
            _ldf,
            _df,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'ivalid'],
    )(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_chisq_vector(tail, x, df):
    u"""
    Computes a vector of probabilities for chi^2 distribution.

    ``prob_chisq_vector`` returns a number of lower or upper tail
    probabilities for the chi^2-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01sc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01scf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether the lower or upper tail probabilities are
        required. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lx, ldf):

        `tail`[j] = 'L'

          The lower tail probability is returned, i.e., p_i = (X_i <=
          x_i:nu_i).

        `tail`[j] = 'U'

          The upper tail probability is returned, i.e., p_i = (X_i >=
          x_i:nu_i).

    x : float, array-like, shape (lx)
        x_i, the values of the chi^2 variates with nu_i degrees of
        freedom.

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the chi^2-distribution.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, ldf, lx))
        p_i, the probabilities for the chi^2 distribution.

    ivalid : int, ndarray, shape (max(ltail, ldf, lx))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          p_i.

        `ivalid`[i-1] = 2

          On entry, x_i < 0.0.

        `ivalid`[i-1] = 3

          On entry, nu_i <= 0.0.

        `ivalid`[i-1] = 4

          The solution has failed to converge while calculating the
          gamma variate. The result returned should represent an
          approximation to the solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x`, `df` or `tail` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_chisq_vector as b_prob_chisq_vector

    fun_name = 'naginterfaces.library.stat.prob_chisq_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _ldf_py, _lx_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _ldf_py, _lx_py),),
    )
    try:
        b_prob_chisq_vector(
            _ltail,
            _tail,
            _lx,
            _x,
            _ldf,
            _df,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'ivalid'],
    )(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_f_vector(tail, f, df1, df2):
    u"""
    Computes a vector of probabilities for F-distribution.

    ``prob_f_vector`` returns a number of lower or upper tail
    probabilities for the F or variance-ratio distribution with real
    degrees of freedom.

    For full information please refer to the NAG Library document for
    g01sd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01sdf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether the lower or upper tail probabilities are
        required. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lf, ldf1, ldf2):

        `tail`[j] = 'L'

          The lower tail probability is returned, i.e., p_i = (F_i <=
          f_i:u_i,v_i).

        `tail`[j] = 'U'

          The upper tail probability is returned, i.e., p_i = (F_i >=
          f_i:u_i,v_i).

    f : float, array-like, shape (lf)
        f_i, the value of the F variate.

    df1 : float, array-like, shape (ldf1)
        u_i, the degrees of freedom of the numerator variance.

    df2 : float, array-like, shape (ldf2)
        v_i, the degrees of freedom of the denominator variance.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, lf, ldf1, ldf2))
        p_i, the probabilities for the F-distribution.

    ivalid : int, ndarray, shape (max(ltail, lf, ldf1, ldf2))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          p_i.

        `ivalid`[i-1] = 2

          On entry, f_i < 0.0.

        `ivalid`[i-1] = 3

          On entry, u_i <= 0.0, or, v_i <= 0.0.

        `ivalid`[i-1] = 4

          The solution has failed to converge. The result returned
          should represent an approximation to the solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lf > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf1 > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: ldf2 > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `f`, `df1`, `df2` or `tail`
            was invalid, or the solution failed to converge.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_f_vector as b_prob_f_vector

    fun_name = 'naginterfaces.library.stat.prob_f_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lf_py = _b_utils._EngineFloat64ArrayType.get_shape(_f, {'fun_name': fun_name, 'entity': 'f'}, exp_rank=1)[0]
    _lf = _b_utils._EngineIntScalarType(
        _lf_py, {'fun_name': fun_name, 'entity': '_lf'},
    ).to_ctypes()
    _df1 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df1'},
        df1,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf1_py = _b_utils._EngineFloat64ArrayType.get_shape(_df1, {'fun_name': fun_name, 'entity': 'df1'}, exp_rank=1)[0]
    _ldf1 = _b_utils._EngineIntScalarType(
        _ldf1_py, {'fun_name': fun_name, 'entity': '_ldf1'},
    ).to_ctypes()
    _df2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df2'},
        df2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf2_py = _b_utils._EngineFloat64ArrayType.get_shape(_df2, {'fun_name': fun_name, 'entity': 'df2'}, exp_rank=1)[0]
    _ldf2 = _b_utils._EngineIntScalarType(
        _ldf2_py, {'fun_name': fun_name, 'entity': '_ldf2'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lf_py, _ldf1_py, _ldf2_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lf_py, _ldf1_py, _ldf2_py),),
    )
    try:
        b_prob_f_vector(
            _ltail,
            _tail,
            _lf,
            _f,
            _ldf1,
            _df1,
            _ldf2,
            _df2,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'ivalid'],
    )(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_beta_vector(tail, beta, a, b):
    u"""
    Computes a vector of probabilities for the beta distribution.

    ``prob_beta_vector`` computes a number of lower or upper tail
    probabilities for the beta distribution.

    For full information please refer to the NAG Library document for
    g01se

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01sef.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether a lower or upper tail probabilities are
        required. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lbeta, la, lb):

        `tail`[j] = 'L'

          The lower tail probability is returned, i.e., p_i = (B_i <=
          beta_i:a_i,b_i).

        `tail`[j] = 'U'

          The upper tail probability is returned, i.e., p_i = (B_i >=
          beta_i:a_i,b_i).

    beta : float, array-like, shape (lbeta)
        beta_i, the value of the beta variate.

    a : float, array-like, shape (la)
        a_i, the first parameter of the required beta distribution.

    b : float, array-like, shape (lb)
        b_i, the second parameter of the required beta distribution.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, lbeta, la, lb))
        p_i, the probabilities for the beta distribution.

    ivalid : int, ndarray, shape (max(ltail, lbeta, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          p_i.

        `ivalid`[i-1] = 2

          On entry, beta_i < 0.0, or, beta_i > 1.0.

        `ivalid`[i-1] = 3

          On entry, a_i <= 0.0, or, b_i <= 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lbeta > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `beta`, `a`, `b` or `tail`
            was invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_beta_vector as b_prob_beta_vector

    fun_name = 'naginterfaces.library.stat.prob_beta_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _beta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'beta'},
        beta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lbeta_py = _b_utils._EngineFloat64ArrayType.get_shape(_beta, {'fun_name': fun_name, 'entity': 'beta'}, exp_rank=1)[0]
    _lbeta = _b_utils._EngineIntScalarType(
        _lbeta_py, {'fun_name': fun_name, 'entity': '_lbeta'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lbeta_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lbeta_py, _la_py, _lb_py),),
    )
    try:
        b_prob_beta_vector(
            _ltail,
            _tail,
            _lbeta,
            _beta,
            _la,
            _a,
            _lb,
            _b,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'ivalid'],
    )(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_gamma_vector(tail, g, a, b):
    u"""
    Computes a vector of probabilities for the gamma distribution.

    ``prob_gamma_vector`` returns a number of lower or upper tail
    probabilities for the gamma distribution.

    For full information please refer to the NAG Library document for
    g01sf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01sff.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether a lower or upper tail probability is required.
        For j = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lg,
        la, lb):

        `tail`[j] = 'L'

          The lower tail probability is returned, i.e., p_i = (G_i <=
          g_i:alpha_i,beta_i).

        `tail`[j] = 'U'

          The upper tail probability is returned, i.e., p_i = (G_i >=
          g_i:alpha_i,beta_i).

    g : float, array-like, shape (lg)
        g_i, the value of the gamma variate.

    a : float, array-like, shape (la)
        The parameter alpha_i of the gamma distribution.

    b : float, array-like, shape (lb)
        The parameter beta_i of the gamma distribution.

    Returns
    -------
    p : float, ndarray, shape (max(lg, la, lb, ltail))
        p_i, the probabilities of the beta distribution.

    ivalid : int, ndarray, shape (max(lg, la, lb, ltail))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          p_i.

        `ivalid`[i-1] = 2

          On entry, g_i < 0.0.

        `ivalid`[i-1] = 3

          On entry, alpha_i <= 0.0, or, beta_i <= 0.0.

        `ivalid`[i-1] = 4

          The solution did not converge in 600 iterations, see
          ``specfun.gamma_incomplete`` for details. The probability
          returned should be a reasonable approximation to the solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lg > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `g`, `a`, `b` or `tail` was
            invalid, or the solution did not converge.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_gamma_vector as b_prob_gamma_vector

    fun_name = 'naginterfaces.library.stat.prob_gamma_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _g = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'g'},
        g,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lg_py = _b_utils._EngineFloat64ArrayType.get_shape(_g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=1)[0]
    _lg = _b_utils._EngineIntScalarType(
        _lg_py, {'fun_name': fun_name, 'entity': '_lg'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lg_py, _la_py, _lb_py, _ltail_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lg_py, _la_py, _lb_py, _ltail_py),),
    )
    try:
        b_prob_gamma_vector(
            _ltail,
            _tail,
            _lg,
            _g,
            _la,
            _a,
            _lb,
            _b,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'ivalid'],
    )(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_binomial_vector(n, p, k):
    u"""
    Computes a vector of probabilities for the binomial distribution.

    ``prob_binomial_vector`` returns a number of the lower tail, upper
    tail and point probabilities for the binomial distribution.

    For full information please refer to the NAG Library document for
    g01sj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01sjf.html

    Parameters
    ----------
    n : int, array-like, shape (ln)
        n_i, the first parameter of the binomial distribution.

    p : float, array-like, shape (lp)
        p_i, the second parameter of the binomial distribution.

    k : int, array-like, shape (lk)
        k_i, the integer which defines the required probabilities.

    Returns
    -------
    plek : float, ndarray, shape (max(ln, lp, lk))
        Prob{X_i <= k_i}, the lower tail probabilities.

    pgtk : float, ndarray, shape (max(ln, lp, lk))
        Prob{X_i > k_i}, the upper tail probabilities.

    peqk : float, ndarray, shape (max(ln, lp, lk))
        Prob{X_i = k_i}, the point probabilities.

    ivalid : int, ndarray, shape (max(ln, lp, lk))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, n_i < 0.

        `ivalid`[i-1] = 2

          On entry, p_i <= 0.0, or, p_i >= 1.0.

        `ivalid`[i-1] = 3

          On entry, k_i < 0, or, k_i > n_i.

        `ivalid`[i-1] = 4

          On entry, n_i is too large to be represented exactly as a real
          number.

        `ivalid`[i-1] = 5

          On entry, the variance (= n_i p_i(1-p_i)) exceeds 10^6.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ln > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lk > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `n`, `p` or `k` was invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_binomial_vector as b_prob_binomial_vector

    fun_name = 'naginterfaces.library.stat.prob_binomial_vector'

    _n = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'n'},
        n,
        copy=False,
    )
    _ln_py = _b_utils._EngineIntArrayType.get_shape(_n, {'fun_name': fun_name, 'entity': 'n'}, exp_rank=1)[0]
    _ln = _b_utils._EngineIntScalarType(
        _ln_py, {'fun_name': fun_name, 'entity': '_ln'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _k = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'k'},
        k,
        copy=False,
    )
    _lk_py = _b_utils._EngineIntArrayType.get_shape(_k, {'fun_name': fun_name, 'entity': 'k'}, exp_rank=1)[0]
    _lk = _b_utils._EngineIntScalarType(
        _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
    ).to_ctypes()
    _plek = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    _pgtk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    _peqk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    try:
        b_prob_binomial_vector(
            _ln,
            _n,
            _lp,
            _p,
            _lk,
            _k,
            _plek,
            _pgtk,
            _peqk,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _plek_py = _plek
    _pgtk_py = _pgtk
    _peqk_py = _peqk
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['plek', 'pgtk', 'peqk', 'ivalid'],
    )(
        _plek_py,
        _pgtk_py,
        _peqk_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_poisson_vector(l, k):
    u"""
    Computes a vector of probabilities for the Poisson distribution.

    ``prob_poisson_vector`` returns a number of the lower tail, upper
    tail and point probabilities for the Poisson distribution.

    For full information please refer to the NAG Library document for
    g01sk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01skf.html

    Parameters
    ----------
    l : float, array-like, shape (ll)
        lambda_i, the parameter of the Poisson distribution.

    k : int, array-like, shape (lk)
        k_i, the integer which defines the required probabilities.

    Returns
    -------
    plek : float, ndarray, shape (max(ll, lk))
        Prob{X_i <= k_i}, the lower tail probabilities.

    pgtk : float, ndarray, shape (max(ll, lk))
        Prob{X_i > k_i}, the upper tail probabilities.

    peqk : float, ndarray, shape (max(ll, lk))
        Prob{X_i = k_i}, the point probabilities.

    ivalid : int, ndarray, shape (max(ll, lk))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, lambda_i <= 0.0.

        `ivalid`[i-1] = 2

          On entry, k_i < 0.

        `ivalid`[i-1] = 3

          On entry, lambda_i > 10^6.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ll > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lk > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `l` or `k` was invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_poisson_vector as b_prob_poisson_vector

    fun_name = 'naginterfaces.library.stat.prob_poisson_vector'

    _l = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'l'},
        l,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ll_py = _b_utils._EngineFloat64ArrayType.get_shape(_l, {'fun_name': fun_name, 'entity': 'l'}, exp_rank=1)[0]
    _ll = _b_utils._EngineIntScalarType(
        _ll_py, {'fun_name': fun_name, 'entity': '_ll'},
    ).to_ctypes()
    _k = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'k'},
        k,
        copy=False,
    )
    _lk_py = _b_utils._EngineIntArrayType.get_shape(_k, {'fun_name': fun_name, 'entity': 'k'}, exp_rank=1)[0]
    _lk = _b_utils._EngineIntScalarType(
        _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
    ).to_ctypes()
    _plek = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    _pgtk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    _peqk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    try:
        b_prob_poisson_vector(
            _ll,
            _l,
            _lk,
            _k,
            _plek,
            _pgtk,
            _peqk,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _plek_py = _plek
    _pgtk_py = _pgtk
    _peqk_py = _peqk
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['plek', 'pgtk', 'peqk', 'ivalid'],
    )(
        _plek_py,
        _pgtk_py,
        _peqk_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_hypergeom_vector(n, l, m, k):
    u"""
    Computes a vector of probabilities for the hypergeometric
    distribution.

    ``prob_hypergeom_vector`` returns a number of the lower tail, upper
    tail and point probabilities for the hypergeometric distribution.

    For full information please refer to the NAG Library document for
    g01sl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01slf.html

    Parameters
    ----------
    n : int, array-like, shape (ln)
        n_i, the parameter of the hypergeometric distribution.

    l : int, array-like, shape (ll)
        l_i, the parameter of the hypergeometric distribution.

    m : int, array-like, shape (lm)
        m_i, the parameter of the hypergeometric distribution.

    k : int, array-like, shape (lk)
        k_i, the integer which defines the required probabilities.

    Returns
    -------
    plek : float, ndarray, shape (max(ln, ll, lm, lk))
        Prob{X_i <= k_i}, the lower tail probabilities.

    pgtk : float, ndarray, shape (max(ln, ll, lm, lk))
        Prob{X_i > k_i}, the upper tail probabilities.

    peqk : float, ndarray, shape (max(ln, ll, lm, lk))
        Prob{X_i = k_i}, the point probabilities.

    ivalid : int, ndarray, shape (max(ln, ll, lm, lk))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, n_i < 0.

        `ivalid`[i-1] = 2

          On entry, l_i < 0, or, l_i > n_i.

        `ivalid`[i-1] = 3

          On entry, m_i < 0, or, m_i > n_i.

        `ivalid`[i-1] = 4

          On entry, k_i < 0, or, k_i > l_i, or, k_i > m_i, or, k_i <
          l_i+m_i-n_i.

        `ivalid`[i-1] = 5

          On entry, n_i is too large to be represented exactly as a real
          number.

        `ivalid`[i-1] = 6

          On entry, the variance (see [Description]) exceeds 10^6.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ln > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: ll > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lm > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lk > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `n`, `l`, `m` or `k` was
            invalid, or the variance was too large.

            Check `ivalid` for more information.
    """

    from ..base.stat import prob_hypergeom_vector as b_prob_hypergeom_vector

    fun_name = 'naginterfaces.library.stat.prob_hypergeom_vector'

    _n = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'n'},
        n,
        copy=False,
    )
    _ln_py = _b_utils._EngineIntArrayType.get_shape(_n, {'fun_name': fun_name, 'entity': 'n'}, exp_rank=1)[0]
    _ln = _b_utils._EngineIntScalarType(
        _ln_py, {'fun_name': fun_name, 'entity': '_ln'},
    ).to_ctypes()
    _l = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'l'},
        l,
        copy=False,
    )
    _ll_py = _b_utils._EngineIntArrayType.get_shape(_l, {'fun_name': fun_name, 'entity': 'l'}, exp_rank=1)[0]
    _ll = _b_utils._EngineIntScalarType(
        _ll_py, {'fun_name': fun_name, 'entity': '_ll'},
    ).to_ctypes()
    _m = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'm'},
        m,
        copy=False,
    )
    _lm_py = _b_utils._EngineIntArrayType.get_shape(_m, {'fun_name': fun_name, 'entity': 'm'}, exp_rank=1)[0]
    _lm = _b_utils._EngineIntScalarType(
        _lm_py, {'fun_name': fun_name, 'entity': '_lm'},
    ).to_ctypes()
    _k = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'k'},
        k,
        copy=False,
    )
    _lk_py = _b_utils._EngineIntArrayType.get_shape(_k, {'fun_name': fun_name, 'entity': 'k'}, exp_rank=1)[0]
    _lk = _b_utils._EngineIntScalarType(
        _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
    ).to_ctypes()
    _plek = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    _pgtk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    _peqk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    try:
        b_prob_hypergeom_vector(
            _ln,
            _n,
            _ll,
            _l,
            _lm,
            _m,
            _lk,
            _k,
            _plek,
            _pgtk,
            _peqk,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _plek_py = _plek
    _pgtk_py = _pgtk
    _peqk_py = _peqk
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['plek', 'pgtk', 'peqk', 'ivalid'],
    )(
        _plek_py,
        _pgtk_py,
        _peqk_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inv_cdf_normal_vector(tail, p, xmu, xstd):
    u"""
    Computes a vector of deviates for the standard Normal distribution.

    ``inv_cdf_normal_vector`` returns a number of deviates associated
    with given probabilities of the Normal distribution.

    For full information please refer to the NAG Library document for
    g01ta

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01taf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent.
        Letting Z denote a variate from a standard Normal distribution,
        and z_i = (x_p_i-mu_i)/(sigma_i), then for j = (mod(i-1, ltail))
        +1 -1, for i = 1,2,...,max(ltail, lp, lxmu, lxstd):

        `tail`[j] = 'L'

          The lower tail probability, i.e., p_i = P(Z <= z_i).

        `tail`[j] = 'U'

          The upper tail probability, i.e., p_i = P(Z >= z_i).

        `tail`[j] = 'C'

          The two tail (confidence interval) probability, i.e., p_i =
          P(Z <= |z_i|)-P(Z <= -|z_i|).

        `tail`[j] = 'S'

          The two tail (significance level) probability, i.e., p_i = P(Z
          >= |z_i|)+P(Z <= -|z_i|).

    p : float, array-like, shape (lp)
        p_i, the probabilities for the Normal distribution as defined by
        `tail` with p_i = `p`[j], j = mod(i-1, lp).

    xmu : float, array-like, shape (lxmu)
        mu_i, the means.

    xstd : float, array-like, shape (lxstd)
        sigma_i, the standard deviations.

    Returns
    -------
    x : float, ndarray, shape (max(ltail, lxmu, lxstd, lp))
        x_p_i, the deviates for the Normal distribution.

    ivalid : int, ndarray, shape (max(ltail, lxmu, lxstd, lp))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          x_p_i.

        `ivalid`[i-1] = 2

          On entry, p_i <= 0.0, or, p_i >= 1.0.

        `ivalid`[i-1] = 3

          On entry, sigma_i <= 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lxmu > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lxstd > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `xstd` or `p` was
            invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import inv_cdf_normal_vector as b_inv_cdf_normal_vector

    fun_name = 'naginterfaces.library.stat.inv_cdf_normal_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxmu_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _lxmu = _b_utils._EngineIntScalarType(
        _lxmu_py, {'fun_name': fun_name, 'entity': '_lxmu'},
    ).to_ctypes()
    _xstd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xstd'},
        xstd,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxstd_py = _b_utils._EngineFloat64ArrayType.get_shape(_xstd, {'fun_name': fun_name, 'entity': 'xstd'}, exp_rank=1)[0]
    _lxstd = _b_utils._EngineIntScalarType(
        _lxstd_py, {'fun_name': fun_name, 'entity': '_lxstd'},
    ).to_ctypes()
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lxmu_py, _lxstd_py, _lp_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lxmu_py, _lxstd_py, _lp_py),),
    )
    try:
        b_inv_cdf_normal_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _lxmu,
            _xmu,
            _lxstd,
            _xstd,
            _x,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ivalid'],
    )(
        _x_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inv_cdf_students_t_vector(tail, p, df):
    u"""
    Computes a vector of deviates for Student's t-distribution.

    ``inv_cdf_students_t_vector`` returns a number of deviates
    associated with given probabilities of Student's t-distribution with
    real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01tb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01tbf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, ldf):

        `tail`[j] = 'L'

          The lower tail probability, i.e., p_i = (T_i <= t_p_i:nu_i).

        `tail`[j] = 'U'

          The upper tail probability, i.e., p_i = (T_i >= t_p_i:nu_i).

        `tail`[j] = 'C'

          The two tail (confidence interval) probability, i.e., p_i =
          (T_i <= |t_p_i|:nu_i)-(T_i <= -|t_p_i|:nu_i).

        `tail`[j] = 'S'

          The two tail (significance level) probability, i.e., p_i =
          (T_i >= |t_p_i|:nu_i)+(T_i <= -|t_p_i|:nu_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required Student's t-distribution as
        defined by `tail` .

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the Student's t-distribution.

    Returns
    -------
    t : float, ndarray, shape (max(ltail, lp, ldf))
        t_p_i, the deviates for the Student's t-distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, ldf))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          t_p_i.

        `ivalid`[i-1] = 2

          On entry, p_i <= 0.0, or, p_i >= 1.0.

        `ivalid`[i-1] = 3

          On entry, nu_i < 1.0.

        `ivalid`[i-1] = 4

          The solution has failed to converge. The result returned
          should represent an approximation to the solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p` or `df` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.
    """

    from ..base.stat import inv_cdf_students_t_vector as b_inv_cdf_students_t_vector

    fun_name = 'naginterfaces.library.stat.inv_cdf_students_t_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _t = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    try:
        b_inv_cdf_students_t_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _ldf,
            _df,
            _t,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _t_py = _t
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['t', 'ivalid'],
    )(
        _t_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inv_cdf_chisq_vector(tail, p, df):
    u"""
    Computes a vector of deviates for chi^2 distribution.

    ``inv_cdf_chisq_vector`` returns a number of deviates associated
    with the given probabilities of the chi^2-distribution with real
    degrees of freedom.

    For full information please refer to the NAG Library document for
    g01tc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01tcf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, ldf):

        `tail`[j] = 'L'

          The lower tail probability, i.e., p_i = (X_i <= x_p_i:nu_i).

        `tail`[j] = 'U'

          The upper tail probability, i.e., p_i = (X_i >= x_p_i:nu_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required chi^2-distribution as
        defined by `tail` .

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the chi^2-distribution.

    Returns
    -------
    x : float, ndarray, shape (max(ltail, lp, ldf))
        x_p_i, the deviates for the chi^2-distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, ldf))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          x_p_i.

        `ivalid`[i-1] = 2

          On entry, invalid value for p_i.

        `ivalid`[i-1] = 3

          On entry, nu_i <= 0.0.

        `ivalid`[i-1] = 4

          p_i is too close to 0.0 or 1.0 for the result to be
          calculated.

        `ivalid`[i-1] = 5

          The solution has failed to converge. The result should be a
          reasonable approximation.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p` or `df` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.
    """

    from ..base.stat import inv_cdf_chisq_vector as b_inv_cdf_chisq_vector

    fun_name = 'naginterfaces.library.stat.inv_cdf_chisq_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    try:
        b_inv_cdf_chisq_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _ldf,
            _df,
            _x,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ivalid'],
    )(
        _x_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inv_cdf_f_vector(tail, p, df1, df2):
    u"""
    Computes a vector of deviates for F-distribution.

    ``inv_cdf_f_vector`` returns a number of deviates associated with
    given probabilities of the F or variance-ratio distribution with
    real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01td

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01tdf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, ldf1,
        ldf2):

        `tail`[j] = 'L'

          The lower tail probability, i.e., p_i = (F_i <=
          f_p_i:u_i,v_i).

        `tail`[j] = 'U'

          The upper tail probability, i.e., p_i = (F_i >=
          f_p_i:u_i,v_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required F-distribution as defined
        by `tail` .

    df1 : float, array-like, shape (ldf1)
        u_i, the degrees of freedom of the numerator variance.

    df2 : float, array-like, shape (ldf2)
        v_i, the degrees of freedom of the denominator variance.

    Returns
    -------
    f : float, ndarray, shape (max(ltail, lp, ldf1, ldf2))
        f_p_i, the deviates for the F-distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, ldf1, ldf2))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          f_p_i.

        `ivalid`[i-1] = 2

          On entry, invalid value for p_i.

        `ivalid`[i-1] = 3

          On entry, u_i <= 0.0, or, v_i <= 0.0.

        `ivalid`[i-1] = 4

          The solution has not converged. The result should still be a
          reasonable approximation to the solution.

        `ivalid`[i-1] = 5

          The value of p_i is too close to 0.0 or 1.0 for the result to
          be computed. This will only occur when the large sample
          approximations are used.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf1 > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: ldf2 > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p`, `df1`, `df2`
            was invalid, or the solution failed to converge.

            Check `ivalid` for more information.
    """

    from ..base.stat import inv_cdf_f_vector as b_inv_cdf_f_vector

    fun_name = 'naginterfaces.library.stat.inv_cdf_f_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _df1 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df1'},
        df1,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf1_py = _b_utils._EngineFloat64ArrayType.get_shape(_df1, {'fun_name': fun_name, 'entity': 'df1'}, exp_rank=1)[0]
    _ldf1 = _b_utils._EngineIntScalarType(
        _ldf1_py, {'fun_name': fun_name, 'entity': '_ldf1'},
    ).to_ctypes()
    _df2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df2'},
        df2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf2_py = _b_utils._EngineFloat64ArrayType.get_shape(_df2, {'fun_name': fun_name, 'entity': 'df2'}, exp_rank=1)[0]
    _ldf2 = _b_utils._EngineIntScalarType(
        _ldf2_py, {'fun_name': fun_name, 'entity': '_ldf2'},
    ).to_ctypes()
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf1_py, _ldf2_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf1_py, _ldf2_py),),
    )
    try:
        b_inv_cdf_f_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _ldf1,
            _df1,
            _ldf2,
            _df2,
            _f,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _f_py = _f
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['f', 'ivalid'],
    )(
        _f_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inv_cdf_beta_vector(tail, p, a, b, tol=0.0):
    u"""
    Computes a vector of deviates for the beta distribution.

    ``inv_cdf_beta_vector`` returns a number of deviates associated with
    given probabilities of the beta distribution.

    For full information please refer to the NAG Library document for
    g01te

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01tef.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, la,
        lb):

        `tail`[j] = 'L'

          The lower tail probability, i.e., p_i = (B_i <=
          beta_p_i:a_i,b_i).

        `tail`[j] = 'U'

          The upper tail probability, i.e., p_i = (B_i >=
          beta_p_i:a_i,b_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required beta distribution as
        defined by `tail` .

    a : float, array-like, shape (la)
        a_i, the first parameter of the required beta distribution.

    b : float, array-like, shape (lb)
        b_i, the second parameter of the required beta distribution.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``inv_cdf_beta_vector`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    Returns
    -------
    beta : float, ndarray, shape (max(ltail, lp, la, lb))
        beta_p_i, the deviates for the beta distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          beta_p_i.

        `ivalid`[i-1] = 2

          On entry, p_i < 0.0, or, p_i > 1.0.

        `ivalid`[i-1] = 3

          On entry, a_i <= 0.0, or, a_i > 10^6, or, b_i <= 0.0, or, b_i
          > 10^6.

        `ivalid`[i-1] = 4

          The solution has not converged but the result should be a
          reasonable approximation to the solution.

        `ivalid`[i-1] = 5

          Requested accuracy not achieved when calculating the beta
          probability. The result should be a reasonable approximation
          to the correct solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p`, `a`, or `b` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.
    """

    from ..base.stat import inv_cdf_beta_vector as b_inv_cdf_beta_vector

    fun_name = 'naginterfaces.library.stat.inv_cdf_beta_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _beta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    try:
        b_inv_cdf_beta_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _la,
            _a,
            _lb,
            _b,
            tol,
            _beta,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _beta_py = _beta
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['beta', 'ivalid'],
    )(
        _beta_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inv_cdf_gamma_vector(tail, p, a, b, tol=0.0):
    u"""
    Computes a vector of deviates for the gamma distribution.

    ``inv_cdf_gamma_vector`` returns a number of deviates associated
    with given probabilities of the gamma distribution.

    For full information please refer to the NAG Library document for
    g01tf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01tff.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, la,
        lb):

        `tail`[j] = 'L'

          The lower tail probability, i.e., p_i = (G_i <=
          g_p_i:alpha_i,beta_i).

        `tail`[j] = 'U'

          The upper tail probability, i.e., p_i = (G_i >=
          g_p_i:alpha_i,beta_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required gamma distribution as
        defined by `tail` .

    a : float, array-like, shape (la)
        alpha_i, the first parameter of the required gamma distribution.

    b : float, array-like, shape (lb)
        beta_i, the second parameter of the required gamma distribution.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``inv_cdf_gamma_vector`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    Returns
    -------
    g : float, ndarray, shape (max(ltail, lp, la, lb))
        g_p_i, the deviates for the gamma distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

          No error.

        `ivalid`[i-1] = 1

          On entry, invalid value supplied in `tail` when calculating
          g_p_i.

        `ivalid`[i-1] = 2

          On entry, invalid value for p_i.

        `ivalid`[i-1] = 3

          On entry, alpha_i <= 0.0, or, alpha_i > 10^6, or, beta_i <=
          0.0.

        `ivalid`[i-1] = 4

          p_i is too close to 0.0 or 1.0 to enable the result to be
          calculated.

        `ivalid`[i-1] = 5

          The solution has failed to converge. The result may be a
          reasonable approximation.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p`, `a`, or `b` was
            invalid.

            Check `ivalid` for more information.
    """

    from ..base.stat import inv_cdf_gamma_vector as b_inv_cdf_gamma_vector

    fun_name = 'naginterfaces.library.stat.inv_cdf_gamma_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    try:
        b_inv_cdf_gamma_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _la,
            _a,
            _lb,
            _b,
            tol,
            _g,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _g_py = _g
    _ivalid_py = _ivalid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['g', 'ivalid'],
    )(
        _g_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def moving_average(m, x, iwt=0, wt=None, pn=0, wantsd=False, comm=None):
    u"""
    Computes the mean and standard deviation using a rolling window.

    ``moving_average`` calculates the mean and, optionally, the standard
    deviation using a rolling window for an arbitrary sized data stream.

    For full information please refer to the NAG Library document for
    g01wa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01waf.html

    Parameters
    ----------
    m : int
        m, the length of the rolling window.

        If `pn` != 0, `m` must be unchanged since the last call to
        ``moving_average``.

    x : float, array-like, shape (nb)
        The current block of observations, corresponding to x_i, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far and b is the size of the current block of data.

    iwt : int, optional
        The type of weighting to use.

        `iwt` = 0

          No weights are used.

        `iwt` = 1

          Each observation has its own weight.

        `iwt` = 2

          Each position in the window has its own weight.

        `iwt` = 3

          Each position in the window has a weight equal to its position
          number.

        If `pn` != 0, `iwt` must be unchanged since the last call to
        ``moving_average``.

    wt : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `iwt`=1: nb; if `iwt`=2: `m`; otherwise: 0.

        The user-supplied weights.

        If `iwt` = 1, `wt`[i-1] = nu_{i+k}, for i = 1,2,...,b.

        If `iwt` = 2, `wt`[j-1] = w_j, for j = 1,2,...,m.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``moving_average``, or when starting to summarise a new
        dataset, `pn` must be set to 0.

        If `pn` != 0, it must be the same value as returned by the last
        call to ``moving_average``.

    wantsd : bool, optional
        If the standard deviations are required then `wantsd` should be
        set to True.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        If None all the data must be supplied in one go, otherwise need
        not be set.

    Returns
    -------
    pn : int
        k+b, the updated number of observations processed so far.

    rmean : float, ndarray, shape (max(0, nb+min(0, `pn`-`m`+1)))
        mu_l, the (weighted) moving averages, for l = 1,2,...,b+min(0,
        k-m+1). Therefore, mu_l is the mean of the data in the window
        that ends on `x`[l+m-min(k, m-1)-2].

        If, on entry, `pn` >= `m`-1, i.e., at least one windows worth of
        data has been previously processed, then `rmean`[l-1] is the
        summary corresponding to the window that ends on `x`[l-1].

        On the other hand, if, on entry, `pn` = 0, i.e., no data has
        been previously processed, then `rmean`[l-1] is the summary
        corresponding to the window that ends on `x`[`m`+l-2] (or,
        equivalently, starts on `x`[l-1]).

    rsd : None or float, ndarray, shape (:)
        If `wantsd` = True then sigma_l, the (weighted) standard
        deviation. The ordering of `rsd` is the same as the ordering of
        `rmean`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 12)
            On entry, `m` = *<value>*.

            On entry at previous call, `m` = *<value>*.

            Constraint: if `pn` > 0, `m` must be unchanged since
            previous call.

        (`errno` 21)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 22)
            On entry, nb = *<value>*, `m` = *<value>*.

            Constraint: if `comm` is None, nb >= `m`.

        (`errno` 41)
            On entry, `iwt` = *<value>*.

            Constraint: `iwt` = 0, 1, 2 or 3.

        (`errno` 42)
            On entry, `iwt` = *<value>*.

            On entry at previous call, `iwt` = *<value>*.

            Constraint: if `pn` > 0, `iwt` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: `wt`[i-1] >= 0.

        (`errno` 52)
            On entry, `wt`[0] = *<value>*.

            Constraint: if `iwt` = 2, `wt`[0] > 0.

        (`errno` 55)
            On entry, sum of weights supplied in `wt` is *<value>*.

            Constraint: if `iwt` = 2, the sum of the weights > 0.

        (`errno` 61)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 62)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0, `pn` must be unchanged since
            previous call.

        (`errno` 101)
            `comm`['rcomm'] has been corrupted between calls.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 53)
            On entry, at least one window had all zero weights.

        (`errno` 54)
            On entry, unable to calculate at least one standard
            deviation due to the weights supplied.

    See Also
    --------
    :meth:`naginterfaces.library.examples.stat.moving_average_ex.main`
    """

    from ..base.stat import moving_average as b_moving_average

    fun_name = 'naginterfaces.library.stat.moving_average'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _rmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(0, _nb_py+min(0, _pn_py-_m_py+1)),),
    )
    _wantsd = _b_utils._EngineLogicalScalarType(
        wantsd, {'fun_name': fun_name, 'entity': 'wantsd'},
    ).to_ctypes()
    _wantsd_py = _b_utils._EngineLogicalScalarType.to_py(_wantsd)
    if _wantsd_py:
        if _wantsd_py:
            _rsd_dim1 = max(0, _nb_py+min(0, _pn_py-_m_py+1))
        else:
            _rsd_dim1 = 0
        exp_shape_rsd = (_rsd_dim1,)
        _rsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_rsd,
        )
    else:
        _rsd = None
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    try:
        _return_scalar = b_moving_average(
            _m,
            _nb,
            _x,
            iwt,
            _wt,
            _pn,
            _rmean,
            _rsd,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _pn_py = _return_scalar
    _rmean_py = _rmean[:max(0, _nb_py+min(0, _pn_py-_m_py+1))]
    if _wantsd_py:
        _rsd_dim1 = max(0, _nb_py+min(0, _pn_py-_m_py+1))
    else:
        _rsd_dim1 = 0
    if _wantsd_py:
        _rsd_py = _rsd[:_rsd_dim1]
    else:
        _rsd_py = None
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['pn', 'rmean', 'rsd'],
    )(
        _pn_py,
        _rmean_py,
        _rsd_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def init_vavilov(rkappa, beta2, mode):
    u"""
    Initialization function for ``pdf_vavilov`` and ``prob_vavilov``.

    ``init_vavilov`` is used to initialize functions ``pdf_vavilov`` and
    ``prob_vavilov``.

    It is intended to be used before a call to ``pdf_vavilov`` or
    ``prob_vavilov``.

    For full information please refer to the NAG Library document for
    g01zu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g01/g01zuf.html

    Parameters
    ----------
    rkappa : float
        The argument kappa of the function.

    beta2 : float
        The argument beta^2 of the function.

    mode : int
        If `mode` = 0, ``pdf_vavilov`` is to be called after the call to
        ``init_vavilov``. Otherwise, ``prob_vavilov`` is to be called.

    Returns
    -------
    xl : float
        x_l, a threshold value below which phi_V(lambda;kappa,beta^2)
        will be set to zero by ``pdf_vavilov`` and
        Phi_V(lambda;kappa,beta^2) will be set to zero by
        ``prob_vavilov`` if lambda < x_l.

    xu : float
        x_u, a threshold value above which phi_V(lambda;kappa,beta^2)
        will be set to zero by ``pdf_vavilov`` and
        Phi_V(lambda;kappa,beta^2) will be set to unity by
        ``prob_vavilov`` if lambda > x_u.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `beta2` = *<value>*.

            Constraint: `beta2` <= 1.0.

        (`errno` 1)
            On entry, `beta2` = *<value>*.

            Constraint: `beta2` >= 0.0.

        (`errno` 1)
            On entry, `rkappa` = *<value>*.

            Constraint: `rkappa` <= 10.0.

        (`errno` 1)
            On entry, `rkappa` = *<value>*.

            Constraint: `rkappa` >= 0.01.

        (`errno` 2)
            The initialization has been abandoned due to an internal
            error. This error exit is unlikely to occur, but if it does
            change the values of `rkappa` and/or `beta2` and rerun
            ``init_vavilov``.
    """

    from ..base.stat import init_vavilov as b_init_vavilov

    _return_scalars = b_init_vavilov(
        rkappa,
        beta2,
        mode,
    )
    _xl_py = _return_scalars.xl
    _xu_py = _return_scalars.xu
    _comm_py = _return_scalars.comm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xl', 'xu', 'comm'],
    )(
        _xl_py,
        _xu_py,
        _comm_py,
    )
    return _return_data
