# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `tsa` Chapter.

``tsa`` - Time Series Analysis

This module provides facilities for investigating and modelling the
statistical structure of series of observations collected at points in
time.
The models may then be used to forecast the series.

The module covers the following models and approaches:

- univariate time series analysis, including autocorrelation functions
  and autoregressive moving average (ARMA) models;

- univariate spectral analysis;

- transfer function (multi-input) modelling, in which one time series is
  dependent on other time series;

- bivariate spectral methods including coherency, gain and input
  response functions;

- vector ARMA models for multivariate time series;

- Kalman filter models (linear and nonlinear);

- GARCH models for volatility;

- inhomogeneous time series.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13intro.html

See Also
--------
``naginterfaces.library.examples.tsa`` :
    This subpackage contains examples for the ``tsa`` module.
    See also the :ref:`library_tsa_ex` subsection.
"""

# NAG Copyright 2017-2019.

from __future__ import absolute_import

import collections as _collections

from math import ceil
from math import log

from ..base import utils as _b_utils
from ..library import _utils

def uni_diff(x, nd, nds, ns):
    u"""
    Univariate time series, seasonal and non-seasonal differencing.

    ``uni_diff`` carries out non-seasonal and seasonal differencing on a
    time series.
    Information which allows the original series to be reconstituted
    from the differenced series is also produced.
    This information is required in time series forecasting.

    For full information please refer to the NAG Library document for
    g13aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13aaf.html

    Parameters
    ----------
    x : float, array-like, shape (nx)
        The undifferenced time series, x_i, for i = 1,2,...,n.

    nd : int
        d, the order of non-seasonal differencing.

    nds : int
        D, the order of seasonal differencing.

    ns : int
        s, the seasonality.

    Returns
    -------
    xd : float, ndarray, shape (nx)
        The differenced values in elements 0 to `nxd`-1, and
        reconstitution data in the remainder of the array.

    nxd : int
        The number of differenced values in the array `xd`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nds` = *<value>*.

            Constraint: if `ns` = 0, `nds` <= 0.

        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 0.

        (`errno` 1)
            On entry, `nds` = *<value>*.

            Constraint: `nds` >= 0.

        (`errno` 1)
            On entry, `nd` = *<value>*.

            Constraint: `nd` >= 0.

        (`errno` 2)
            On entry, nx = *<value>*, `nd` = *<value>*, `nds` =
            *<value>* and `ns` = *<value>*.

            Constraint: nx > `nd`+(`nds`*`ns`).
    """

    from ..base.tsa import uni_diff as b_uni_diff

    fun_name = 'naginterfaces.library.tsa.uni_diff'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _xd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nx_py,),
    )
    _return_scalar = b_uni_diff(
        _x,
        _nx,
        nd,
        nds,
        ns,
        _xd,
    )
    _xd_py = _xd
    _nxd_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xd', 'nxd'],
    )(
        _xd_py,
        _nxd_py,
    )
    return _return_data

def uni_autocorr(x, nk):
    u"""
    Univariate time series, sample autocorrelation function.

    ``uni_autocorr`` computes the sample autocorrelation function of a
    time series.
    It also computes the sample mean, the sample variance and a
    statistic which may be used to test the hypothesis that the true
    autocorrelation function is zero.

    For full information please refer to the NAG Library document for
    g13ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13abf.html

    Parameters
    ----------
    x : float, array-like, shape (nx)
        The time series, x_i, for i = 1,2,...,n.

    nk : int
        K, the number of lags for which the autocorrelations are
        required. The lags range from 1 to K and do not include zero.

    Returns
    -------
    xm : float
        The sample mean of the input time series.

    xv : float
        The sample variance of the input time series.

    r : float, ndarray, shape (`nk`)
        The sample autocorrelation coefficient relating to lag k, for k
        = 1,2,...,K.

    stat : float
        The statistic used to test the hypothesis that the true
        autocorrelation function of the time series is identically zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nx = *<value>* and `nk` = *<value>*.

            Constraint: nx > `nk`.

        (`errno` 1)
            On entry, nx = *<value>*.

            Constraint: nx > 1.

        (`errno` 1)
            On entry, `nk` = *<value>*.

            Constraint: `nk` > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, all values of `x` are practically identical.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_autocorr as b_uni_autocorr

    fun_name = 'naginterfaces.library.tsa.uni_autocorr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _nk = _b_utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _nk_py = _b_utils._EngineIntScalarType.to_py(_nk)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nk_py,),
    )
    try:
        _return_scalars = b_uni_autocorr(
            _x,
            _nx,
            _nk,
            _r,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _xm_py = _return_scalars.xm
    _xv_py = _return_scalars.xv
    _r_py = _r
    _stat_py = _return_scalars.stat
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xm', 'xv', 'r', 'stat'],
    )(
        _xm_py,
        _xv_py,
        _r_py,
        _stat_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_autocorr_part(r, nl):
    u"""
    Univariate time series, partial autocorrelations from
    autocorrelations.

    ``uni_autocorr_part`` calculates partial autocorrelation
    coefficients given a set of autocorrelation coefficients.
    It also calculates the predictor error variance ratios for
    increasing order of finite lag autoregressive predictor, and the
    autoregressive parameters associated with the predictor of maximum
    order.

    For full information please refer to the NAG Library document for
    g13ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13acf.html

    Parameters
    ----------
    r : float, array-like, shape (nk)
        The autocorrelation coefficient relating to lag k, for k =
        1,2,...,K.

    nl : int
        L, the number of partial autocorrelation coefficients required.

    Returns
    -------
    p : float, ndarray, shape (`nl`)
        `p`[l-1] contains the partial autocorrelation coefficient at lag
        l, p_{l,l}, for l = 1,2,...,`nvl`.

    v : float, ndarray, shape (`nl`)
        `v`[l-1] contains the predictor error variance ratio v_l, for l
        = 1,2,...,`nvl`.

    ar : float, ndarray, shape (`nl`)
        The autoregressive parameters of maximum order, i.e., p_{Lj} if
        no exception or warning is raised, or p_{l_0-1,j} if `errno` =
        3, for j = 1,2,...,`nvl`.

    nvl : int
        The number of valid values in each of `p`, `v` and `ar`. Thus in
        the case of premature termination at iteration l_0 (see
        [Description]), `nvl` is returned as l_0-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nk = *<value>* and `nl` = *<value>*.

            Constraint: nk >= `nl`.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` > 0.

        (`errno` 1)
            On entry, nk = *<value>*.

            Constraint: nk > 0.

        (`errno` 2)
            On entry, the autocorrelation coefficient of lag 1 has
            absolute value greater than or equal to 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The autocorrelation coefficients do not form a positive
            definite sequence.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_autocorr_part as b_uni_autocorr_part

    fun_name = 'naginterfaces.library.tsa.uni_autocorr_part'

    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nk_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _nk = _b_utils._EngineIntScalarType(
        _nk_py, {'fun_name': fun_name, 'entity': '_nk'},
    ).to_ctypes()
    _nl = _b_utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = _b_utils._EngineIntScalarType.to_py(_nl)
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    _ar = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    try:
        _return_scalar = b_uni_autocorr_part(
            _r,
            _nk,
            _nl,
            _p,
            _v,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _p_py = _p
    _v_py = _v
    _ar_py = _ar
    _nvl_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'v', 'ar', 'nvl'],
    )(
        _p_py,
        _v_py,
        _ar_py,
        _nvl_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_arima_prelim(mr, r, xv):
    u"""
    Univariate time series, preliminary estimation, seasonal ARIMA
    model.

    ``uni_arima_prelim`` calculates preliminary estimates of the
    parameters of an autoregressive integrated moving average (ARIMA)
    model from the autocorrelation function of the appropriately
    differenced times series.

    For full information please refer to the NAG Library document for
    g13ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13adf.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    r : float, array-like, shape (nk)
        The autocorrelations (starting at lag 1), which must have been
        calculated after the time series has been appropriately
        differenced.

    xv : float
        The series sample variance, calculated after appropriate
        differencing has been applied to the series.

    Returns
    -------
    par : float, ndarray, shape (:)
        The first npar elements of `par` contain the preliminary
        estimates of the ARIMA model parameters, in standard order.

    rv : float
        An estimate of the residual variance of the preliminarily
        estimated model.

    isf : int, ndarray, shape (4)
        Contains success/failure indicators, one for each of the four
        types of parameter (autoregressive, moving average, seasonal
        autoregressive, seasonal moving average).

        The indicator has the interpretation:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 2)
            On entry, nk = *<value>*.

            Constraint: nk >= max(p+q, s*(P+Q)).

        (`errno` 3)
            On entry, `r`[*<value>*] = *<value>*.

            Constraint: -1.0 <= `r`[i] <= 1.0, for i = 1,2,...nk.

        (`errno` 4)
            On entry, `xv` = *<value>*.

            Constraint: `xv` > 0.0.

        (`errno` 5)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_arima_prelim as b_uni_arima_prelim

    fun_name = 'naginterfaces.library.tsa.uni_arima_prelim'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nk_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _nk = _b_utils._EngineIntScalarType(
        _nk_py, {'fun_name': fun_name, 'entity': '_nk'},
    ).to_ctypes()
    _npar = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,)), {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _npar_py = _b_utils._EngineIntScalarType.to_py(_npar)
    _par = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalar = b_uni_arima_prelim(
            _mr,
            _r,
            _nk,
            xv,
            _npar,
            _par,
            _isf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _par_py = _par
    _rv_py = _return_scalar
    _isf_py = _isf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['par', 'rv', 'isf'],
    )(
        _par_py,
        _rv_py,
        _isf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_arima_estim(mr, par, c, x, iex, igh, ist, kpiv, zsp, kzsp, kfc=1, piv=None, nit=100, data=None, io_manager=None):
    u"""
    Univariate time series, estimation, seasonal ARIMA model
    (comprehensive).

    ``uni_arima_estim`` fits a seasonal autoregressive integrated moving
    average (ARIMA) model to an observed time series, using a nonlinear
    least squares procedure incorporating backforecasting. Parameter
    estimates are obtained, together with appropriate standard errors.
    The residual series is returned, and information for use in
    forecasting the time series is produced for use by the functions
    ``uni_arima_update`` and ``uni_arima_forecast_state``.

    The estimation procedure is iterative, starting with initial
    parameter values such as may be obtained using ``uni_arima_prelim``.
    It continues until a specified convergence criterion is satisfied,
    or until a specified number of iterations has been carried out.
    The progress of the procedure can be monitored by means of a
    user-supplied function.

    For full information please refer to the NAG Library document for
    g13ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13aef.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    par : float, array-like, shape (npar)
        The initial estimates of the p values of the phi parameters, the
        q values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters, in that
        order.

    c : float
        If `kfc` = 0, `c` must contain the expected value, c, of the
        differenced series.

        If `kfc` = 1, `c` must contain an initial estimate of c.

    x : float, array-like, shape (nx)
        The n values of the original undifferenced time series.

    iex : int
        The dimension of the arrays `ex`, `exr` and `al`.

    igh : int
        The dimension of the arrays `g` and `sd`.

        The second dimension of the arrays `h` and `hc`.

    ist : int
        The dimension of the array `st`.

    kpiv : int
        Must be nonzero if the progress of the optimization is to be
        monitored using `piv`. Otherwise `kpiv` must contain 0.

    zsp : float, array-like, shape (4)
        When `kzsp` = 1, the first four elements of `zsp` must contain
        the four values used to guide the search procedure. These are as
        follows.

        `zsp`[0] contains alpha, the value used to constrain the
        magnitude of the search procedure steps.

        `zsp`[1] contains beta, the multiplier which regulates the value
        alpha.

        `zsp`[2] contains delta, the value of the stationarity and
        invertibility test tolerance factor.

        `zsp`[3] contains gamma, the value of the convergence criterion.

        If `kzsp` != 1 on entry, default values for `zsp` are supplied
        by the function.

        These are 0.001, 10.0, 1000.0 and max(100*machine
        precision,0.0000001) respectively.

    kzsp : int
        The value 1 if the function is to use the input values of `zsp`,
        and any other value if the default values of `zsp` are to be
        used.

    kfc : int, optional
        Must be set to 1 if the constant, c, is to be estimated and 0 if
        it is to be held fixed at its initial value.

    piv : None or callable piv(mr, par, c, kfc, icount, s, g, h, itc,
        zsp, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `piv` is used to monitor the progress of the optimization.

        Parameters
        ~~~~~~~~~~
        mr : int, ndarray, shape (7)
            `mr`, defined as for ``uni_arima_estim``.

        par : float, ndarray, shape (npar)
            `par`, defined as for ``uni_arima_estim``.

        c : float
            `c`, defined as for ``uni_arima_estim``.

        kfc : int
            `kfc`, defined as for ``uni_arima_estim``.

        icount : int, ndarray, shape (6)
            `icount`, defined as for ``uni_arima_estim``.

        s : float
            `s`, defined as for ``uni_arima_estim``.

        g : float, ndarray, shape (igh)
            `g`, defined as for ``uni_arima_estim``.

        h : float, ndarray, shape (:, igh)
            `h`, defined as for ``uni_arima_estim``.

        itc : int
            `itc`, defined as for ``uni_arima_estim``.

        zsp : float, ndarray, shape (4)
            `zsp`, defined as for ``uni_arima_estim``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    nit : int, optional
        The maximum number of iterations to be performed.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    par : float, ndarray, shape (npar)
        The latest values of the estimates of these parameters.

    c : float
        If `kfc` = 0, `c` is unchanged.

        If `kfc` = 1, `c` contains the latest estimate of c.

        Therefore, if `c` and `kfc` are both zero on entry, there is no
        constant correction.

    icount : int, ndarray, shape (6)
        Size of various output arrays.

        `icount`[0]

          Contains q+(Q*s), the number of backforecasts.

        `icount`[1]

          Contains n-d-(D*s), the number of differenced values.

        `icount`[2]

          Contains d+(D*s), the number of values of reconstitution
          information.

        `icount`[3]

          Contains n+q+(Q*s), the number of values held in each of the
          series `ex`, `exr` and `al`.

        `icount`[4]

          Contains n-d-(D*s)-p-q-P-Q-`kfc`, the number of degrees of
          freedom associated with S.

        `icount`[5]

          Contains `icount`[0]+npar+`kfc`, the number of parameters
          being estimated.

        These values are always computed regardless of the exit value of
        errno.

    ex : float, ndarray, shape (`iex`)
        The extended differenced series which is made up of:

        `icount`[0] backforecast values of the differenced series.

        `icount`[1] actual values of the differenced series.

        `icount`[2] values of reconstitution information.

        The total number of these values held in `ex` is `icount`[3].

        If the function exits because of a faulty input parameter, the
        contents of `ex` will be indeterminate.

    exr : float, ndarray, shape (`iex`)
        The values of the model residuals which is made up of:

        `icount`[0] residuals corresponding to the backforecasts in the
        differenced series.

        `icount`[1] residuals corresponding to the actual values in the
        differenced series.

        The remaining `icount`[2] values contain zeros.

        If the function exits with errno holding a value other than 0 or
        9, the contents of `exr` will be indeterminate.

    al : float, ndarray, shape (`iex`)
        The intermediate series which is made up of:

        `icount`[0] intermediate series values corresponding to the
        backforecasts in the differenced series.

        `icount`[1] intermediate series values corresponding to the
        actual values in the differenced series.

        The remaining `icount`[2] values contain zeros.

        If an exception is raised, the contents of `al` will be
        indeterminate.

    s : float
        The residual sum of squares after the latest series of parameter
        estimates has been incorporated into the model. If the function
        exits with a faulty input parameter, `s` contains zero.

    g : float, ndarray, shape (`igh`)
        The latest value of the derivatives of S with respect to each of
        the parameters being estimated (backforecasts, `par` parameters,
        and where relevant the constant -- in that order). The contents
        of `g` will be indeterminate if the function exits with a faulty
        input parameter.

    sd : float, ndarray, shape (`igh`)
        The standard deviations corresponding to each of the parameters
        being estimated (backforecasts, `par` parameters, and where
        relevant the constant, in that order).

        If the function exits with errno containing a value other than 0
        or 9, or if the required number of iterations is zero, the
        contents of `sd` will be indeterminate.

    h : float, ndarray, shape (1+q+(Q*s)+npar+`kfc`, `igh`)
        The second derivative of S and correlation coefficients.

        - the latest values of an approximation to the second derivative
          of S with respect to each of the (q+Q*s+npar+`kfc`) parameters
          being estimated (backforecasts, `par` parameters, and where
          relevant the constant -- in that order), and

        - the correlation coefficients relating to each pair of these
          parameters.

        These are held in a matrix defined by the first
        (q+Q*s+npar+`kfc`) rows and the first (q+Q*s+npar+`kfc`) columns
        of `h`. (Note that `icount`[5] contains the value of this
        expression.) The values of (a) are contained in the upper
        triangle, and the values of (b) in the strictly lower triangle.

        These correlation coefficients are zero during intermediate
        printout using `piv`, and indeterminate if errno contains on
        exit a value other than 0 or 9.

        All the contents of `h` are indeterminate if the required number
        of iterations are zero.

        The (q+(Q*s)+npar+`kfc`+1)th row of `h` is used internally as
        workspace.

    st : float, ndarray, shape (`ist`)
        The `nst` values of the state set array. If the function exits
        with errno containing a value other than 0 or 9, the contents of
        `st` will be indeterminate.

    nst : int
        The number of values in the state set array `st`.

    itc : int
        The number of iterations performed.

    zsp : float, ndarray, shape (4)
        `zsp` contains the values, default or otherwise, used by the
        function.

    isf : int, ndarray, shape (4)
        Contains success/failure indicators, one for each of the four
        types of parameter in the model (autoregressive, moving average,
        seasonal autoregressive, seasonal moving average), in that
        order.

        Each indicator has the interpretation:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 3)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 3)
            On entry, `zsp`[3] = *<value>*.

            Constraint: 0.0 <= `zsp`[3] < 1.0.

        (`errno` 3)
            On entry, `zsp`[2] = *<value>*.

            Constraint: `zsp`[2] >= 1.0.

        (`errno` 3)
            On entry, `zsp`[1] = *<value>*.

            Constraint: `zsp`[1] > 1.0.

        (`errno` 3)
            On entry, `zsp`[0] = *<value>*.

            Constraint: `zsp`[0] > 0.0.

        (`errno` 4)
            On entry, `ist` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ist` >= (P*s)+d+(D*s)+q+max(p, Q*s).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            On entry, ldh = *<value>*.

            Constraint: ldh >= q*(Q*s)+npar+`kfc`.

        (`errno` 6)
            On entry, `igh` = *<value>*.

            Constraint: `igh` >= q*(Q*s)+npar+`kfc`.

        (`errno` 6)
            On entry, `iex` = *<value>*.

            Constraint: `iex` >= q*(Q*s)+nx.

        (`errno` 7)
            A failure in the search procedure has occurred.

        (`errno` 8)
            Failure to invert H.

        (`errno` 9)
            Unable to calculate the latest estimates of the
            backforecasts.

        (`errno` 10)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_arima_estim as b_uni_arima_estim

    def cb_piv(mr, par, npar, c, kfc, icount, s, g, h, igh, itc, zsp, data=None):
        """Wrapping user's piv. Called as base piv."""
        if data is not None:
            piv(
                mr,
                par,
                c,
                kfc,
                icount,
                s,
                g,
                h,
                itc,
                zsp,
                data,
            )
        else:
            piv(
                mr,
                par,
                c,
                kfc,
                icount,
                s,
                g,
                h,
                itc,
                zsp,
            )

    fun_name = 'naginterfaces.library.tsa.uni_arima_estim'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _icount = _b_utils._EngineIntArrayType.empty_ndarray(
        (6,),
    )
    _iex = _b_utils._EngineIntScalarType(
        iex, {'fun_name': fun_name, 'entity': 'iex'},
    ).to_ctypes()
    _iex_py = _b_utils._EngineIntScalarType.to_py(_iex)
    _ex = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_iex_py,),
    )
    _exr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_iex_py,),
    )
    _al = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_iex_py,),
    )
    _igh = _b_utils._EngineIntScalarType(
        igh, {'fun_name': fun_name, 'entity': 'igh'},
    ).to_ctypes()
    _igh_py = _b_utils._EngineIntScalarType.to_py(_igh)
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_igh_py,),
    )
    _sd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_igh_py,),
    )
    _vl_s_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _vl_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vu_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _kfc = _b_utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _kfc_py = _b_utils._EngineIntScalarType.to_py(_kfc)
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (1+_vl_q_py+(_vu_q_py*_vl_s_py)+_npar_py+_kfc_py, _igh_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _ist = _b_utils._EngineIntScalarType(
        ist, {'fun_name': fun_name, 'entity': 'ist'},
    ).to_ctypes()
    _ist_py = _b_utils._EngineIntScalarType.to_py(_ist)
    _st = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ist_py,),
    )
    _b_utils._check_callable(piv, {'fun_name': fun_name, 'entity': 'piv'}, allow_none=True)
    _piv = None if piv is None else cb_piv
    _zsp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'zsp'},
        zsp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalars = b_uni_arima_estim(
            _mr,
            _par,
            _npar,
            c,
            _kfc,
            _x,
            _nx,
            _icount,
            _ex,
            _exr,
            _al,
            _iex,
            _g,
            _igh,
            _sd,
            _h,
            _st,
            _ist,
            _piv,
            kpiv,
            nit,
            _zsp,
            kzsp,
            _isf,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _par_py = _par
    _c_py = _return_scalars.c
    _zsp_py = _zsp
    _icount_py = _icount
    _ex_py = _ex
    _exr_py = _exr
    _al_py = _al
    _s_py = _return_scalars.s
    _g_py = _g
    _sd_py = _sd
    _h_py = _h
    _st_py = _st
    _nst_py = _return_scalars.nst
    _itc_py = _return_scalars.itc
    _isf_py = _isf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['par', 'c', 'icount', 'ex', 'exr', 'al', 's', 'g', 'sd', 'h', 'st', 'nst', 'itc', 'zsp', 'isf'],
    )(
        _par_py,
        _c_py,
        _icount_py,
        _ex_py,
        _exr_py,
        _al_py,
        _s_py,
        _g_py,
        _sd_py,
        _h_py,
        _st_py,
        _nst_py,
        _itc_py,
        _zsp_py,
        _isf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_arima_estim_easy(mr, par, c, x, nppc, kfc=1, kpiv=0, nit=100, ires=None, io_manager=None):
    u"""
    Univariate time series, estimation, seasonal ARIMA model
    (easy-to-use).

    ``uni_arima_estim_easy`` is an easy-to-use version of
    ``uni_arima_estim``.
    It fits a seasonal autoregressive integrated moving average (ARIMA)
    model to an observed time series, using a nonlinear least squares
    procedure incorporating backforecasting. Parameter estimates are
    obtained, together with appropriate standard errors.
    The residual series is returned, and information for use in
    forecasting the time series is produced for use in
    ``uni_arima_update`` and ``uni_arima_forecast_state``.

    The estimation procedure is iterative, starting with initial
    parameter values such as may be obtained using ``uni_arima_prelim``.
    It continues until a specified convergence criterion is satisfied or
    until a specified number of iterations have been carried out.
    The progress of the iteration can be monitored by means of an
    optional printing facility.

    For full information please refer to the NAG Library document for
    g13af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13aff.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model whose
        parameters are to be estimated. p, q, P and Q refer respectively
        to the number of autoregressive (phi), moving average (theta),
        seasonal autoregressive (Phi) and seasonal moving average
        (Theta) parameters. d, D and s refer respectively to the order
        of non-seasonal differencing, the order of seasonal differencing
        and the seasonal period.

    par : float, array-like, shape (npar)
        The initial estimates of the p values of the phi parameters, the
        q values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters, in that
        order.

    c : float
        If `kfc` = 0, `c` must contain the expected value, c, of the
        differenced series.

        If `kfc` = 1, `c` must contain an initial estimate of c.

        Therefore, if `c` and `kfc` are both zero on entry, there is no
        constant correction.

    x : float, array-like, shape (nx)
        The n values of the original undifferenced time series.

    nppc : int
        The number of phi, theta, Phi, Theta and c parameters to be
        estimated. `nppc` = p+q+P+Q+1 if the constant is being estimated
        and `nppc` = p+q+P+Q if not.

    kfc : int, optional
        Must be set to 1 if the constant, c, is to be estimated and 0 if
        it is to be held fixed at its initial value.

    kpiv : int, optional
        Must be nonzero if the progress of the optimization is to be
        monitored using the built-in printing facility. Otherwise `kpiv`
        must contain zero. If selected, monitoring output will be sent
        to the file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`). For each
        iteration, the heading

        ::

          G13AFZ MONITORING OUTPUT - ITERATION n

        followed by the argument values, and residual sum of squares,
        are printed.

    nit : int, optional
        The maximum number of iterations to be performed.

    ires : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows:
        15*(q+Q*s)+11*nx+13*(npar+`kfc`)+8*(p+P*s)+12+2*(q+Q*s+npar+`kfc`)^2.

        The dimension of the array `res`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    par : float, ndarray, shape (npar)
        Contains the latest values of the estimates of these parameters.

    c : float
        If `kfc` = 0, `c` is unchanged.

        If `kfc` = 1, `c` contains the latest estimate of c.

    s : float
        The residual sum of squares after the latest series of parameter
        estimates has been incorporated into the model. If the function
        exits with a faulty input parameter, `s` contains zero.

    ndf : int
        The number of degrees of freedom associated with `s`,

        `ndf` = n-d-D*s-p-q-P-Q-`kfc`.

    sd : float, ndarray, shape (`nppc`)
        The standard deviations corresponding to the parameters in the
        model (p autoregressive, q moving average, P seasonal
        autoregressive, Q seasonal moving average and c, if estimated,
        in that order). If the function exits with errno containing a
        value other than 0 or 9, or if the required number of iterations
        is zero, the contents of `sd` will be indeterminate.

    cm : float, ndarray, shape (`nppc`, `nppc`)
        The correlation coefficients associated with each pair of the
        `nppc` parameters. These are held in the first `nppc` rows and
        the first `nppc` columns of `cm`. These correlation coefficients
        are indeterminate if errno contains on exit a value other than 0
        or 9, or if the required number of iterations is zero.

    st : float, ndarray, shape (nx)
        The value of the state set in its first `nst` elements. If the
        function exits with errno containing a value other than 0 or 9,
        the contents of `st` will be indeterminate.

    nst : int
        The size of the state set. `nst` = P*s+D*s+d+q+max(p, Q*s).

        `nst` should be used subsequently in ``uni_arima_update`` and
        ``uni_arima_forecast_state`` as the dimension of `st`.

    itc : int
        The number of iterations performed.

    isf : int, ndarray, shape (4)
        The first four elements of `isf` contain success/failure
        indicators, one for each of the four types of parameter in the
        model (autoregressive, moving average, seasonal autoregressive,
        seasonal moving average), in that order.

        Each indicator has the interpretation:

        [table omitted]

    res : float, ndarray, shape (`ires`)
        The first `nres` elements of `res` contain the model residuals
        derived from the differenced series. If the function exits with
        errno holding a value other than 0 or 9, these elements of `res`
        will be indeterminate. The rest of the array `res` is used as
        workspace.

    nres : int
        The number of model residuals returned in `res`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 3)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 4)
            On entry, nx = *<value>* and minimum required size for the
            state set array = *<value>*.

        (`errno` 5)
            On entry, `ires` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ires` >=
            15*Q'+11n+13*`nppc`+8*P'+12+2*(Q'+`nppc`)^2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            A failure in the search procedure has occurred.

        (`errno` 8)
            The inversion of the Hessian matrix in the calculation of
            the covariance matrix of the parameter estimates has failed.

        (`errno` 9)
            Failure whilst calculating backforecasts.

        (`errno` 10)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_arima_estim_easy as b_uni_arima_estim_easy

    fun_name = 'naginterfaces.library.tsa.uni_arima_estim_easy'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _nppc = _b_utils._EngineIntScalarType(
        nppc, {'fun_name': fun_name, 'entity': 'nppc'},
    ).to_ctypes()
    _nppc_py = _b_utils._EngineIntScalarType.to_py(_nppc)
    _sd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nppc_py,),
    )
    _cm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nppc_py, _nppc_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _st = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nx_py,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    _kfc = _b_utils._EngineIntScalarType(
        kfc, {'fun_name': fun_name, 'entity': 'kfc'},
    ).to_ctypes()
    _kfc_py = _b_utils._EngineIntScalarType.to_py(_kfc)
    _vl_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vu_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _vl_s_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _vl_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))
    _vu_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))
    _ires = 15*(_vl_q_py+_vu_q_py*_vl_s_py)+11*_nx_py+13*(_npar_py+_kfc_py)+8*(_vl_p_py+_vu_p_py*_vl_s_py)+12+2*(_vl_q_py+_vu_q_py*_vl_s_py+_npar_py+_kfc_py)**2 if ires is None else ires
    _ires = _b_utils._EngineIntScalarType(
        _ires, {'fun_name': fun_name, 'entity': '_ires'},
    ).to_ctypes()
    _ires_py = _b_utils._EngineIntScalarType.to_py(_ires)
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ires_py,),
    )
    try:
        _return_scalars = b_uni_arima_estim_easy(
            _mr,
            _par,
            _npar,
            c,
            _kfc,
            _x,
            _nx,
            _sd,
            _nppc,
            _cm,
            _st,
            kpiv,
            nit,
            _isf,
            _res,
            _ires,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _par_py = _par
    _c_py = _return_scalars.c
    _s_py = _return_scalars.s
    _ndf_py = _return_scalars.ndf
    _sd_py = _sd
    _cm_py = _cm
    _st_py = _st
    _nst_py = _return_scalars.nst
    _itc_py = _return_scalars.itc
    _isf_py = _isf
    _res_py = _res
    _nres_py = _return_scalars.nres
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['par', 'c', 's', 'ndf', 'sd', 'cm', 'st', 'nst', 'itc', 'isf', 'res', 'nres'],
    )(
        _par_py,
        _c_py,
        _s_py,
        _ndf_py,
        _sd_py,
        _cm_py,
        _st_py,
        _nst_py,
        _itc_py,
        _isf_py,
        _res_py,
        _nres_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_arima_update(st, mr, par, c, anx):
    u"""
    Univariate time series, update state set for forecasting.

    ``uni_arima_update`` accepts a series of new observations of a time
    series, the model of which is already fully specified, and updates
    the 'state set' information for use in constructing further
    forecasts.
    The previous specifications of the time series model should have
    been obtained by using ``uni_arima_estim`` or
    ``uni_arima_estim_easy`` to estimate the relevant parameters.
    The supplied state set will originally have been produced by
    ``uni_arima_estim`` or ``uni_arima_estim_easy``, but may since have
    been updated by earlier calls to ``uni_arima_update``.

    A set of residuals corresponding to the new observations is
    returned.
    These may be of use in checking that the new observations conform to
    the previously fitted model.

    For full information please refer to the NAG Library document for
    g13ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13agf.html

    Parameters
    ----------
    st : float, array-like, shape (nst)
        The state set derived from ``uni_arima_estim`` or
        ``uni_arima_estim_easy``, or as modified using earlier calls of
        ``uni_arima_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, array-like, shape (npar)
        The estimates of the p values of the phi parameters, the q
        values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters in the model
        -- in that order, using the usual notation.

    c : float
        The constant to be subtracted from the differenced data.

    anx : float, array-like, shape (nuv)
        The new undifferenced observations which are to be used to
        update `st`.

    Returns
    -------
    st : float, ndarray, shape (nst)
        The updated values of the state set.

    anexr : float, ndarray, shape (nuv)
        The residuals corresponding to the new observations in `anx`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            On entry, nst = *<value>*.

            Constraint: nst = P*s+D*s+d+q+max(Q*s, p).

        (`errno` 3)
            On entry, nuv = *<value>*.

            Constraint: nuv > 0.

        (`errno` 4)
            On entry, nwa = *<value>* and the minimum size required =
            *<value>*.

            Constraint: nwa >= 4*npar+3*nst.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_arima_update as b_uni_arima_update

    fun_name = 'naginterfaces.library.tsa.uni_arima_update'

    _st = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nst_py = _b_utils._EngineFloat64ArrayType.get_shape(_st, {'fun_name': fun_name, 'entity': 'st'}, exp_rank=1)[0]
    _nst = _b_utils._EngineIntScalarType(
        _nst_py, {'fun_name': fun_name, 'entity': '_nst'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _anx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'anx'},
        anx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nuv_py = _b_utils._EngineFloat64ArrayType.get_shape(_anx, {'fun_name': fun_name, 'entity': 'anx'}, exp_rank=1)[0]
    _nuv = _b_utils._EngineIntScalarType(
        _nuv_py, {'fun_name': fun_name, 'entity': '_nuv'},
    ).to_ctypes()
    _anexr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nuv_py,),
    )
    b_uni_arima_update(
        _st,
        _nst,
        _mr,
        _par,
        _npar,
        c,
        _anx,
        _nuv,
        _anexr,
    )
    _st_py = _st
    _anexr_py = _anexr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['st', 'anexr'],
    )(
        _st_py,
        _anexr_py,
    )
    return _return_data

def uni_arima_forecast_state(st, mr, par, c, rms, nfv):
    u"""
    Univariate time series, forecasting from state set.

    ``uni_arima_forecast_state`` produces forecasts of a time series,
    given a time series model which has already been fitted to the time
    series using ``uni_arima_estim`` or ``uni_arima_estim_easy``.
    The original observations are not required, since
    ``uni_arima_forecast_state`` uses as input either the original state
    set produced by ``uni_arima_estim`` or ``uni_arima_estim_easy`` or
    the state set updated by a series of new observations using
    ``uni_arima_update``.
    Standard errors of the forecasts are also provided.

    For full information please refer to the NAG Library document for
    g13ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ahf.html

    Parameters
    ----------
    st : float, array-like, shape (nst)
        The state set derived from ``uni_arima_estim`` or
        ``uni_arima_estim_easy`` originally, or as modified using
        earlier calls of ``uni_arima_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, array-like, shape (npar)
        The estimates of the p values of the phi parameters, the q
        values of the theta parameters, the P values of the Phi
        parameters and the Q values of the Theta parameters which
        specify the model and which were output originally by
        ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    c : float
        c, the value of the model constant. This will have been output
        by ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    rms : float
        V, the residual variance associated with the model.

        If ``uni_arima_estim_easy`` was used to estimate the model,
        `rms` should be set to s/ndf, where s and ndf were output by
        ``uni_arima_estim_easy``.

        If ``uni_arima_estim`` was used to estimate the model, `rms`
        should be set to s/icount[4], where s and icount[4] were output
        by ``uni_arima_estim``.

    nfv : int
        L, the required number of forecasts.

    Returns
    -------
    fva : float, ndarray, shape (`nfv`)
        `nfv` forecast values relating to the original undifferenced
        series.

    fsd : float, ndarray, shape (`nfv`)
        The standard errors associated with each of the `nfv` forecast
        values in `fva`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            On entry, nst = *<value>*.

            Constraint: nst = P*s+D*s+d+q+max(Q*s, p).

        (`errno` 3)
            On entry, `nfv` = *<value>*.

            Constraint: `nfv` > 0.

        (`errno` 4)
            On entry, nwa = *<value>* and the minimum size required =
            *<value>*.

            Constraint: nwa >= 4*npar+3*nst.

        (`errno` 5)
            On entry, `rms` = *<value>*.

            Constraint: `rms` >= 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_arima_forecast_state as b_uni_arima_forecast_state

    fun_name = 'naginterfaces.library.tsa.uni_arima_forecast_state'

    _st = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nst_py = _b_utils._EngineFloat64ArrayType.get_shape(_st, {'fun_name': fun_name, 'entity': 'st'}, exp_rank=1)[0]
    _nst = _b_utils._EngineIntScalarType(
        _nst_py, {'fun_name': fun_name, 'entity': '_nst'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _nfv = _b_utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _nfv_py = _b_utils._EngineIntScalarType.to_py(_nfv)
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    b_uni_arima_forecast_state(
        _st,
        _nst,
        _mr,
        _par,
        _npar,
        c,
        rms,
        _nfv,
        _fva,
        _fsd,
    )
    _fva_py = _fva
    _fsd_py = _fsd
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['fva', 'fsd'],
    )(
        _fva_py,
        _fsd_py,
    )
    return _return_data

def uni_arima_forcecast(mr, par, c, kfc, x, ist, nfv, ifv):
    u"""
    Univariate time series, state set and forecasts, from fully
    specified seasonal ARIMA model.

    ``uni_arima_forcecast`` applies a fully specified seasonal ARIMA
    model to an observed time series, generates the state set for
    forecasting and (optionally) derives a specified number of forecasts
    together with their standard deviations.

    For full information please refer to the NAG Library document for
    g13aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ajf.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model, in the
        usual notation.

    par : float, array-like, shape (npar)
        The p values of the phi parameters, the q values of the theta
        parameters, the P values of the Phi parameters, and the Q values
        of the Theta parameters, in that order.

    c : float
        c, the expected value of the differenced series (i.e., c is the
        constant correction). Where there is no constant term, `c` must
        be set to 0.0.

    kfc : int
        Must be set to 0 if `c` was not estimated, and 1 if `c` was
        estimated. This is irrespective of whether or not `c` = 0.0. The
        only effect is that the residual degrees of freedom are one
        greater when `kfc` = 0. Assuming the supplied time series to be
        the same as that to which the model was originally fitted, this
        ensures an unbiased estimate of the residual mean-square.

    x : float, array-like, shape (nx)
        The n values of the original undifferenced time series.

    ist : int
        The dimension of the array `st`.

    nfv : int
        The required number of forecasts. If `nfv` <= 0, no forecasts
        will be computed.

    ifv : int
        The dimension of the arrays `fva` and `fsd`.

    Returns
    -------
    rms : float
        The residual variance (mean square) associated with the model.

    st : float, ndarray, shape (`ist`)
        The `nst` values of the state set.

    nst : int
        The number of values in the state set array `st`.

    fva : float, ndarray, shape (`ifv`)
        If `nfv` > 0, `fva` contains the `nfv` forecast values relating
        to the original undifferenced time series.

    fsd : float, ndarray, shape (`ifv`)
        If `nfv` > 0, `fsd` contains the estimated standard errors of
        the `nfv` forecast values.

    isf : int, ndarray, shape (4)
        Contains validity indicators, one for each of the four possible
        parameter types in the model (autoregressive, moving average,
        seasonal autoregressive, seasonal moving average), in that
        order.

        Each indicator has the interpretation:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 1)
            The orders vector `mr` is invalid.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = p+q+P+Q.

        (`errno` 2)
            The model is over-parameterised.

        (`errno` 4)
            On entry, `ist` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `ist` >= (P*s)+d+(D*s)+q+max(p, Q*s).

        (`errno` 5)
            Unable to calculate the latest estimates of the
            backforecasts.

        (`errno` 6)
            Satisfactory parameter estimates could not be obtained for
            all parameter types in the model.

        (`errno` 7)
            On entry, `ifv` = *<value>* and `nfv` = *<value>*.

            Constraint: `ifv` >= max(1, `nfv`).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_arima_forcecast as b_uni_arima_forcecast

    fun_name = 'naginterfaces.library.tsa.uni_arima_forcecast'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _ist = _b_utils._EngineIntScalarType(
        ist, {'fun_name': fun_name, 'entity': 'ist'},
    ).to_ctypes()
    _ist_py = _b_utils._EngineIntScalarType.to_py(_ist)
    _st = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ist_py,),
    )
    _ifv = _b_utils._EngineIntScalarType(
        ifv, {'fun_name': fun_name, 'entity': 'ifv'},
    ).to_ctypes()
    _ifv_py = _b_utils._EngineIntScalarType.to_py(_ifv)
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ifv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ifv_py,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (4,),
    )
    _return_scalars = b_uni_arima_forcecast(
        _mr,
        _par,
        _npar,
        c,
        kfc,
        _x,
        _nx,
        _st,
        _ist,
        nfv,
        _fva,
        _fsd,
        _ifv,
        _isf,
    )
    _rms_py = _return_scalars.rms
    _st_py = _st
    _nst_py = _return_scalars.nst
    _fva_py = _fva
    _fsd_py = _fsd
    _isf_py = _isf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['rms', 'st', 'nst', 'fva', 'fsd', 'isf'],
    )(
        _rms_py,
        _st_py,
        _nst_py,
        _fva_py,
        _fsd_py,
        _isf_py,
    )
    return _return_data

def uni_smooth_exp(mode, itype, p, param, y, k, init, nf, r):
    u"""
    Univariate time series, exponential smoothing.

    ``uni_smooth_exp`` performs exponential smoothing using either
    single exponential, double exponential or a Holt--Winters method.

    For full information please refer to the NAG Library document for
    g13am

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13amf.html

    Parameters
    ----------
    mode : int
        Indicates if ``uni_smooth_exp`` is continuing from a previous
        call or, if not, how the initial values are computed.

        `mode` = 0

          Required values for m_0, r_0 and s_{-j}, for j = 0,1,...,p-1,
          are supplied in `init`.

        `mode` = 1

          ``uni_smooth_exp`` continues from a previous call using values
          that are supplied in `r`.

        `mode` = 2

          Required values for m_0, r_0 and s_{-j}, for j = 0,1,...,p-1,
          are estimated using the first k observations.

    itype : int
        The smoothing function.

        `itype` = 1

          Single exponential.

        `itype` = 2

          Brown double exponential.

        `itype` = 3

          Linear Holt.

        `itype` = 4

          Additive Holt--Winters.

        `itype` = 5

          Multiplicative Holt--Winters.

    p : int
        If `itype` = 4 or 5, the seasonal order, p, otherwise `p` is not
        referenced.

    param : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2): 1; if `itype`=3: 3; if `itype` in
        (4, 5): 4; otherwise: 0.

        The smoothing parameters.

        If `itype` = 1 or 2, `param`[0] = alpha and any remaining
        elements of `param` are not referenced.

        If `itype` = 3, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = phi and any remaining elements of `param` are not
        referenced.

        If `itype` = 4 or 5, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = beta and `param`[3] = phi.

    y : float, array-like, shape (n)
        The time series.

    k : int
        If `mode` = 2, the number of observations used to initialize the
        smoothing.

        If `mode` != 2, `k` is not referenced.

    init : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype`=1: 1; if `itype` in (2, 3): 2; if `itype` in
        (4, 5): 2+`p`; otherwise: 0.

        If `mode` = 0, the initial values for m_0, r_0 and s_{-j}, for j
        = 0,1,...,p-1, used to initialize the smoothing.

        If `itype` = 1, `init`[0] = m_0 and the remaining elements of
        `init` are not referenced.

        If `itype` = 2 or 3, `init`[0] = m_0 and `init`[1] = r_0 and the
        remaining elements of `init` are not referenced.

        If `itype` = 4 or 5, `init`[0] = m_0, `init`[1] = r_0 and
        `init`[2] to `init`[p+1] hold the values for s_{-j}, for j =
        0,1,...,p-1.

        The remaining elements of `init` are not referenced.

    nf : int
        The number of forecasts required beyond the end of the series.
        Note, the one step ahead forecast is always produced.

    r : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2, 3): 13; if `itype` in (4, 5):
        13+`p`; otherwise: 0.

        If `mode` = 1, `r` must contain the values as returned by a
        previous call to ``rand.times_smooth_exp`` or
        ``uni_smooth_exp``, `r` need not be set otherwise.

        If `itype` = 1, 2 or 3, only the first 13 elements of `r` are
        referenced, otherwise the first 13+p elements are referenced.

    Returns
    -------
    init : float, ndarray, shape (:)
        If `mode` != 1, the values used to initialize the smoothing.
        These are in the same order as described above.

    fv : float, ndarray, shape (`nf`)
        y^_{t+f}, for f = 1,2,...,`nf`, the next `nf` step forecasts.
        Where t = n, if `mode` != 1, else t is the total number of
        smoothed and forecast values already produced.

    fse : float, ndarray, shape (`nf`)
        The forecast standard errors for the values given in `fv`.

    yhat : float, ndarray, shape (n)
        y^_{t+1}, for t = 1,2,...,n, the one step ahead forecast values,
        with `yhat`[i-1] being the one step ahead forecast of `y`[i-2].

    res : float, ndarray, shape (n)
        The residuals, (y_{t+1}-y^_{t+1}), for t = 1,2,...,n.

    dv : float
        The square root of the mean deviation.

    ad : float
        The mean absolute deviation.

    r : float, ndarray, shape (:)
        The information on the current state of the smoothing.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1, 2, 3, 4 or 5.

        (`errno` 3)
            On entry, `p` = *<value>*.

            Constraint: if `itype` = 4 or 5, `p` > 1.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: 0.0 <= `param`[*<value>*] <= 1.0.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: if `itype` = 2, 0.0 < `param`[*<value>*] <= 1.0.

        (`errno` 4)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: `param`[*<value>*] >= 0.0.

        (`errno` 5)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 6)
            A multiplicative Holt--Winters model cannot be used with the
            supplied data.

        (`errno` 7)
            On entry, `k` = *<value>* and n = *<value>*.

            Constraint: if `mode` = 2 and `itype` = 4 or 5, 1 <= `k` <=
            n.

        (`errno` 7)
            On entry, `k` = *<value>*, 2*`p` = *<value>*.

            Constraint: if `mode` = 2 and `itype` = 4 or 5, 2*`p` <=
            `k`.

        (`errno` 9)
            On entry, `nf` = *<value>*.

            Constraint: `nf` >= 0.

        (`errno` 16)
            On entry, the array `r` has not been initialized correctly.
    """

    from ..base.tsa import uni_smooth_exp as b_uni_smooth_exp

    fun_name = 'naginterfaces.library.tsa.uni_smooth_exp'

    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _init = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'init'},
        init,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nf = _b_utils._EngineIntScalarType(
        nf, {'fun_name': fun_name, 'entity': 'nf'},
    ).to_ctypes()
    _nf_py = _b_utils._EngineIntScalarType.to_py(_nf)
    _fv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nf_py,),
    )
    _fse = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nf_py,),
    )
    _yhat = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalars = b_uni_smooth_exp(
        mode,
        itype,
        p,
        _param,
        _n,
        _y,
        k,
        _init,
        _nf,
        _fv,
        _fse,
        _yhat,
        _res,
        _r,
    )
    _init_py = _init
    _r_py = _r
    _fv_py = _fv
    _fse_py = _fse
    _yhat_py = _yhat
    _res_py = _res
    _dv_py = _return_scalars.dv
    _ad_py = _return_scalars.ad
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['init', 'fv', 'fse', 'yhat', 'res', 'dv', 'ad', 'r'],
    )(
        _init_py,
        _fv_py,
        _fse_py,
        _yhat_py,
        _res_py,
        _dv_py,
        _ad_py,
        _r_py,
    )
    return _return_data

def uni_arima_resid(v, mr, m, par, ishow, io_manager=None):
    u"""
    Univariate time series, diagnostic checking of residuals, following
    ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    ``uni_arima_resid`` is a diagnostic checking function suitable for
    use after fitting a Box--Jenkins ARMA model to a univariate time
    series using ``uni_arima_estim`` or ``uni_arima_estim_easy``. The
    residual autocorrelation function is returned along with an estimate
    of its asymptotic standard errors and correlations.
    Also, ``uni_arima_resid`` calculates the Box--Ljung portmanteau
    statistic and its significance level for testing model adequacy.

    For full information please refer to the NAG Library document for
    g13as

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13asf.html

    Parameters
    ----------
    v : float, array-like, shape (n)
        `v`[t-1] must contain an estimate of epsilon_t, for t =
        1,2,...,n.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim`` then the actual argument `v` must be
        exr[icount[0]] as returned by ``uni_arima_estim``.

        If ``uni_arima_resid`` is used following a call to
        ``uni_arima_estim_easy`` then the actual argument `v` must be
        res as returned by ``uni_arima_estim_easy``.

    mr : int, array-like, shape (7)
        The orders vector (p, d, q, P, D, Q, s) as supplied to
        ``uni_arima_estim`` or ``uni_arima_estim_easy``.

    m : int
        The value of m, the number of residual autocorrelations to be
        computed. See [Further Comments] for advice on the value of `m`.

    par : float, array-like, shape (npar)
        The parameter estimates in the order phi_1,phi_2,...,phi_p,
        theta_1,theta_2,...,theta_q, Phi_1,Phi_2,...,Phi_P,
        Theta_1,Theta_2,...,Theta_Q only.

    ishow : int
        Must be nonzero if the residual autocorrelations, their standard
        errors and the portmanteau statistics are to be printed and zero
        otherwise.

        These quantities are available also as output variables in `r`,
        `rcm`, `chi`, `idf` and `siglev`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    r : float, ndarray, shape (`m`)
        An estimate of the residual autocorrelation coefficient at lag
        l, for l = 1,2,...,m. If `errno` = 3 on exit then all elements
        of `r` are set to zero.

    rcm : float, ndarray, shape (`m`, `m`)
        The estimated standard errors and correlations of the elements
        in the array `r`. The correlation between `r`[i-1] and `r`[j-1]
        is returned as `rcm`[i-1,j-1] except that if i = j then
        `rcm`[i-1,j-1] contains the standard error of `r`[i-1]. If on
        exit, `errno` >= 5, then all off-diagonal elements of `rcm` are
        set to zero and all diagonal elements are set to 1/sqrt(n).

    chi : float
        The value of the portmanteau statistic, Q_(m). If `errno` = 3 on
        exit then `chi` is returned as zero.

    idf : int
        The number of degrees of freedom of `chi`.

    siglev : float
        The significance level of `chi` based on `idf` degrees of
        freedom. If `errno` = 3 on exit, `siglev` is returned as one.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` < n.

        (`errno` 1)
            On entry, `m` = *<value>* and npar = *<value>*.

            Constraint: `m` > npar.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar != 0 and npar = p+q+P+Q.

        (`errno` 1)
            On entry, `mr`[3] = *<value>* and `mr`[5] = *<value>*.

            Constraint: `mr`[3] = 0 and `mr`[5] = 0.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*.

            Constraint: `mr`[6] >= 0.

        (`errno` 1)
            On entry, `mr`[5] = *<value>*.

            Constraint: `mr`[5] >= 0.

        (`errno` 1)
            On entry, `mr`[3] = *<value>*.

            Constraint: `mr`[3] >= 0.

        (`errno` 1)
            On entry, `mr`[2] = *<value>*.

            Constraint: `mr`[2] >= 0.

        (`errno` 1)
            On entry, `mr`[0] = *<value>*.

            Constraint: `mr`[0] >= 0.

        (`errno` 2)
            On entry, the seasonal moving average operator is
            non-invertible.

        (`errno` 2)
            On entry, the seasonal autoregressive operator is
            nonstationary.

        (`errno` 2)
            On entry, the non-seasonal moving average operator is
            non-invertible.

        (`errno` 2)
            On entry, the non-seasonal autoregressive operator is
            nonstationary.

        (`errno` 4)
            Excessive iterations needed to find zeros of determinental
            polynomials.

        (`errno` 5)
            On entry, one or more of the AR operators has a factor in
            common with one or more of the MA operators.

        (`errno` 6)
            The matrix `rcm` could not be computed because one of its
            diagonal elements was found to be non-positive.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, the elements of `v` are nearly identical giving
            near-zero variance.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_arima_resid as b_uni_arima_resid

    fun_name = 'naginterfaces.library.tsa.uni_arima_resid'

    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _rcm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _return_scalars = b_uni_arima_resid(
            _n,
            _v,
            _mr,
            _m,
            _par,
            _npar,
            ishow,
            _r,
            _rcm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _rcm_py = _rcm
    _chi_py = _return_scalars.chi
    _idf_py = _return_scalars.idf
    _siglev_py = _return_scalars.siglev
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'rcm', 'chi', 'idf', 'siglev'],
    )(
        _r_py,
        _rcm_py,
        _chi_py,
        _idf_py,
        _siglev_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_means(z, m, rs):
    u"""
    Computes quantities needed for range-mean or standard deviation-mean
    plot.

    ``uni_means`` calculates the range (or standard deviation) and the
    mean for groups of successive time series values.
    It is intended for use in the construction of range-mean plots.

    For full information please refer to the NAG Library document for
    g13au

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13auf.html

    Parameters
    ----------
    z : float, array-like, shape (n)
        `z`[t-1] must contain the tth observation Z_t, for t =
        1,2,...,n.

    m : int
        m, the group size.

    rs : str, length 1
        Indicates whether ranges or standard deviations are to be
        calculated.

        `rs` = 'R'

          Ranges are calculated.

        `rs` = 'S'

          Standard deviations are calculated.

    Returns
    -------
    y : float, ndarray, shape (int((n/`m`)))
        `y`[i-1] contains the range or standard deviation, as determined
        by `rs`, of the ith group of observations, for i = 1,2,...,k.

    mean : float, ndarray, shape (int((n/`m`)))
        `mean`[i-1] contains the mean of the ith group of observations,
        for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*, `m` = *<value>* and ngrps =
            *<value>*.

            Constraint: ngrps = int(n/`m`).

        (`errno` 1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 2)
            On entry, `rs` = *<value>*.

            Constraint: `rs` = 'R' or 'S'.
    """

    from ..base.tsa import uni_means as b_uni_means

    fun_name = 'naginterfaces.library.tsa.uni_means'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _ngrps = _b_utils._EngineIntScalarType(
        int((_n_py//_m_py)), {'fun_name': fun_name, 'entity': '_ngrps'},
    ).to_ctypes()
    _y = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (int((_n_py//_m_py)),),
    )
    _mean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (int((_n_py//_m_py)),),
    )
    b_uni_means(
        _n,
        _z,
        _m,
        _ngrps,
        rs,
        _y,
        _mean,
    )
    _y_py = _y
    _mean_py = _mean
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['y', 'mean'],
    )(
        _y_py,
        _mean_py,
    )
    return _return_data

def uni_dickey_fuller_unit(ut_type, p, y):
    u"""
    Computes (augmented) Dickey--Fuller unit root test statistic.

    ``uni_dickey_fuller_unit`` returns the (augmented) Dickey--Fuller
    unit root test.

    For full information please refer to the NAG Library document for
    g13aw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13awf.html

    Parameters
    ----------
    ut_type : int
        The type of unit test for which the probability is required.

        `ut_type` = 1

          A unit root test will be performed and tau returned.

        `ut_type` = 2

          A unit root test with drift will be performed and tau_mu
          returned.

        `ut_type` = 3

          A unit root test with drift and deterministic time trend will
          be performed and tau_tau returned.

    p : int
        p, the degree of the autoregressive (AR) component of the
        Dickey--Fuller test statistic. When p > 1 the test is usually
        referred to as the augmented Dickey--Fuller test.

    y : float, array-like, shape (n)
        y, the time series.

    Returns
    -------
    ts : float
        The (augmented) Dickey--Fuller test statistic: tau when
        `ut_type` = 1; tau_mu when `ut_type` = 2 or tau_tau when
        `ut_type` = 3.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ut_type` = *<value>*.

            Constraint: `ut_type` = 1, 2 or 3.

        (`errno` 21)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.

        (`errno` 31)
            On entry, n = *<value>*.

            Constraint: if `ut_type` = 1, n > 2*`p` if `ut_type` = 2, n
            > 2*`p`+1 if `ut_type` = 3, n > 2*`p`+2

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 41)
            On entry, the design matrix used in the estimation of beta_1
            is not of full rank, this is usually due to all elements of
            the series being virtually identical. The returned statistic
            is therefore not unique and likely to be meaningless.

        (`errno` 42)
            standard error of the parameter estimate is zero, therefore
            depending on the sign of the parameter estimate, a large
            positive or negative value has been returned.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_dickey_fuller_unit_ex.main`
    """

    from ..base.tsa import uni_dickey_fuller_unit as b_uni_dickey_fuller_unit

    fun_name = 'naginterfaces.library.tsa.uni_dickey_fuller_unit'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = b_uni_dickey_fuller_unit(
            ut_type,
            p,
            _n,
            _y,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ts'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def multi_filter_arima(y, mr, par, cy=None):
    u"""
    Multivariate time series, filtering (pre-whitening) by an ARIMA
    model.

    ``multi_filter_arima`` filters a time series by an ARIMA model.

    For full information please refer to the NAG Library document for
    g13ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13baf.html

    Parameters
    ----------
    y : float, array-like, shape (ny)
        The Q'_y backforecasts, starting with backforecast at time
        1-Q'_y to backforecast at time 0, followed by the time series
        starting at time 1, where Q'_y = `mr`[9]+`mr`[12]*`mr`[13]. If
        there are no backforecasts, either because the ARIMA model for
        the time series is not known, or because it is known but has no
        moving average terms, then the time series starts at the
        beginning of `y`.

    mr : int, array-like, shape (nmr)
        The orders vector for the filtering model, followed by the
        orders vector for the ARIMA model for the time series if the
        latter is known. The orders appear in the standard sequence
        (p,d,q,P,D,Q,s) as given in [the G13 Introduction]. If the ARIMA
        model for the time series is supplied, the function will assume
        that the first Q'_y values of the array `y` are backforecasts.

    par : float, array-like, shape (npar)
        The parameters of the filtering model, followed by the
        parameters of the ARIMA model for the time series, if supplied.
        Within each model the parameters are in the standard order of
        non-seasonal AR and MA followed by seasonal AR and MA.

    cy : None or float, optional
        If the ARIMA model is known (i.e., nmr = 14), `cy` must specify
        the constant term of the ARIMA model for the time series. If
        this model is not known (i.e., nmr = 7), `cy` is not used.

    Returns
    -------
    b : float, ndarray, shape (:)
        The filtered output series. If the ARIMA model for the time
        series was known, and hence Q'_y backforecasts were supplied in
        `y`, then `b` contains Q'_y 'filtered' backforecasts followed by
        the filtered series. Otherwise, the filtered series begins at
        the start of `b` just as the original series began at the start
        of `y`. In either case, if the value of the series at time t is
        held in `y`[t-1], then the filtered value at time t is held in
        `b`[t-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nmr = *<value>*.

            Constraint: nmr = 7 or 14.

        (`errno` 2)
            On entry, the orders vector `mr` is invalid.

        (`errno` 3)
            On entry, npar = *<value>*.

            Constraint: npar must be inconsistent with `mr`.

        (`errno` 4)
            On entry, ny = *<value>* and the minimum size required =
            *<value>*.

            Constraint: ny >= max(1+Q'_y, npar).

        (`errno` 7)
            The orders vector for the filtering model is invalid.

        (`errno` 8)
            The orders vector for the ARIMA model is invalid.

        (`errno` 9)
            The initial values of the filtered series are indeterminate
            for the given models.
    """

    from ..base.tsa import multi_filter_arima as b_multi_filter_arima

    fun_name = 'naginterfaces.library.tsa.multi_filter_arima'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ny_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _ny = _b_utils._EngineIntScalarType(
        _ny_py, {'fun_name': fun_name, 'entity': '_ny'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _nmr_py = _b_utils._EngineIntArrayType.get_shape(_mr, {'fun_name': fun_name, 'entity': 'mr'}, exp_rank=1)[0]
    _nmr = _b_utils._EngineIntScalarType(
        _nmr_py, {'fun_name': fun_name, 'entity': '_nmr'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    if _nmr_py == 14:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))), {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    elif _nmr_py == 7:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py, {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    else:
        _nb = _b_utils._EngineIntScalarType.c_type(0) # Added by code generator.
        _nb_py = 0 # Added by code generator.
    if _nmr_py == 14:
        _b_dim1 = _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))*_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,)))
    elif _nmr_py == 7:
        _b_dim1 = _ny_py
    else:
        _b_dim1 = 0
    exp_shape_b = (_b_dim1,)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_b,
    )
    b_multi_filter_arima(
        _y,
        _ny,
        _mr,
        _nmr,
        _par,
        _npar,
        cy,
        _b,
        _nb,
    )
    _b_py = _b
    return _b_py

def multi_filter_transf(y, mr, par, cy=None):
    u"""
    Multivariate time series, filtering by a transfer function model.

    ``multi_filter_transf`` filters a time series by a transfer function
    model.

    For full information please refer to the NAG Library document for
    g13bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bbf.html

    Parameters
    ----------
    y : float, array-like, shape (ny)
        The Q'_y backforecasts starting with backforecast at time 1-Q'_y
        to backforecast at time 0 followed by the time series starting
        at time 1, where Q'_y = `mr`[5]+`mr`[8]*`mr`[9]. If there are no
        backforecasts either because the ARIMA model for the time series
        is not known or because it is known but has no moving average
        terms, then the time series starts at the beginning of `y`.

    mr : int, array-like, shape (nmr)
        The orders vector for the filtering transfer function model
        followed by the orders vector for the ARIMA model for the time
        series if the latter is known. The transfer function model
        orders appear in the standard form (b,q,p) as given in [the G13
        Introduction]. Note that if the ARIMA model for the time series
        is supplied then the function will assume that the first Q'_y
        values of the array `y` are backforecasts.

    par : float, array-like, shape (npar)
        The parameters of the filtering transfer function model followed
        by the parameters of the ARIMA model for the time series. In the
        transfer function model the parameters are in the standard order
        of MA-like followed by AR-like operator parameters. In the ARIMA
        model the parameters are in the standard order of non-seasonal
        AR and MA followed by seasonal AR and MA.

    cy : None or float, optional
        If the ARIMA model is known (i.e., nmr = 10), `cy` must specify
        the constant term of the ARIMA model for the time series. If
        this model is not known (i.e., nmr = 3) then `cy` is not used.

    Returns
    -------
    b : float, ndarray, shape (:)
        The filtered output series. If the ARIMA model for the time
        series was known, and hence Q'_y backforecasts were supplied in
        `y`, then `b` contains Q'_y 'filtered' backforecasts followed by
        the filtered series. Otherwise, the filtered series begins at
        the start of `b` just as the original series began at the start
        of `y`. In either case, if the value of the series at time t is
        held in `y`[t-1], then the filtered value at time t is held in
        `b`[t-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nb = *<value>* and the minimum size required =
            *<value>*.

            Constraint: if arimas is not None then nb >= max(b, p, q),
            otherwise nb >= ny.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar must be inconsistent with `mr`.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*, `mr`[7] = *<value>* and
            `mr`[8] = *<value>*.

            Constraint: `mr`[6]+`mr`[7]+`mr`[8] != 0.

        (`errno` 1)
            On entry, `mr`[6] = *<value>*, `mr`[7] = *<value>* and
            `mr`[8] = *<value>*.

            Constraint: `mr`[6]+`mr`[7]+`mr`[8] = 0.

        (`errno` 1)
            On entry, `mr`[9] = 1.

            Constraint: `mr`[9] != 1.

        (`errno` 1)
            On entry, i = *<value>*, `mr`[i-1] = *<value>*.

            Constraint: `mr`[i-1] >= 0.

        (`errno` 1)
            On entry, nmr = *<value>*.

            Constraint: nmr = 3 or 10.

        (`errno` 2)
            A supplied model has invalid parameters.

        (`errno` 3)
            The supplied time series is too short.

        (`errno` 4)
            The matrix used to solve for starting values for MA is
            singular.
    """

    from ..base.tsa import multi_filter_transf as b_multi_filter_transf

    fun_name = 'naginterfaces.library.tsa.multi_filter_transf'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ny_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _ny = _b_utils._EngineIntScalarType(
        _ny_py, {'fun_name': fun_name, 'entity': '_ny'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _nmr_py = _b_utils._EngineIntArrayType.get_shape(_mr, {'fun_name': fun_name, 'entity': 'mr'}, exp_rank=1)[0]
    _nmr = _b_utils._EngineIntScalarType(
        _nmr_py, {'fun_name': fun_name, 'entity': '_nmr'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    if _nmr_py == 3:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py, {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    elif _nmr_py == 10:
        _nb = _b_utils._EngineIntScalarType(
            _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))), {'fun_name': fun_name, 'entity': '_nb'},
        ).to_ctypes()
        _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    else:
        _nb = _b_utils._EngineIntScalarType.c_type(0) # Added by code generator.
        _nb_py = 0 # Added by code generator.
    if _nmr_py == 3:
        _b_dim1 = _ny_py
    elif _nmr_py == 10:
        _b_dim1 = _ny_py+max(_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))+_b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,)), _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,)))
    else:
        _b_dim1 = 0
    exp_shape_b = (_b_dim1,)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_b,
    )
    b_multi_filter_transf(
        _y,
        _ny,
        _mr,
        _nmr,
        _par,
        _npar,
        cy,
        _b,
        _nb,
    )
    _b_py = _b
    return _b_py

def multi_xcorr(x, y, nl):
    u"""
    Multivariate time series, cross-correlations.

    ``multi_xcorr`` calculates cross-correlations between two time
    series.

    For full information please refer to the NAG Library document for
    g13bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bcf.html

    Parameters
    ----------
    x : float, array-like, shape (nxy)
        The n values of the x series.

    y : float, array-like, shape (nxy)
        The n values of the y series.

    nl : int
        L, the maximum lag for calculating cross-correlations.

    Returns
    -------
    s : float
        The ratio of the standard deviation of the y series to the
        standard deviation of the x series, s_y/s_x.

    r0 : float
        The cross-correlation between the x and y series at lag zero.

    r : float, ndarray, shape (`nl`)
        `r`[l-1] contains the cross-correlations between the x and y
        series at lags L, r_{xy}(l), for l = 1,2,...,L.

    stat : float
        The statistic for testing for absence of cross-correlation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nxy = *<value>*.

            Constraint: nxy > 1.

        (`errno` 1)
            On entry, `nl` = *<value>* and nxy = *<value>*.

            Constraint: 1 <= `nl` < nxy.

        (`errno` 1)
            On entry, `nl` = *<value>*.

            Constraint: `nl` >= 1.

        (`errno` 2)
            One or both of the x and y series have zero variance.
    """

    from ..base.tsa import multi_xcorr as b_multi_xcorr

    fun_name = 'naginterfaces.library.tsa.multi_xcorr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxy_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nxy = _b_utils._EngineIntScalarType(
        _nxy_py, {'fun_name': fun_name, 'entity': '_nxy'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nl = _b_utils._EngineIntScalarType(
        nl, {'fun_name': fun_name, 'entity': 'nl'},
    ).to_ctypes()
    _nl_py = _b_utils._EngineIntScalarType.to_py(_nl)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nl_py,),
    )
    _return_scalars = b_multi_xcorr(
        _x,
        _y,
        _nxy,
        _nl,
        _r,
    )
    _s_py = _return_scalars.s
    _r0_py = _return_scalars.r0
    _r_py = _r
    _stat_py = _return_scalars.stat
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'r0', 'r', 'stat'],
    )(
        _s_py,
        _r0_py,
        _r_py,
        _stat_py,
    )
    return _return_data

def multi_transf_prelim(r0, r, nna, s):
    u"""
    Multivariate time series, preliminary estimation of transfer
    function model.

    ``multi_transf_prelim`` calculates preliminary estimates of the
    parameters of a transfer function model.

    For full information please refer to the NAG Library document for
    g13bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bdf.html

    Parameters
    ----------
    r0 : float
        The cross-correlation between the two series at lag 0,
        r_{xy}(0).

    r : float, array-like, shape (nl)
        The cross-correlations between the two series at lags 1 to L,
        r_{xy}(l), for l = 1,2,...,L.

    nna : int, array-like, shape (3)
        The transfer function model orders in the standard form b,q,p.

    s : float
        The ratio of the standard deviation of the y series to that of
        the x series, s_y/s_x.

    Returns
    -------
    wds : float, ndarray, shape (:)
        The preliminary estimates of the parameters of the transfer
        function model in the order of q+1 MA-like parameters followed
        by the p AR-like parameters. If the estimation of either type of
        parameter fails then these arguments are set to 0.0.

    isf : int, ndarray, shape (2)
        Indicators of the success of the estimation of MA-like and
        AR-like parameters respectively. A value 0 indicates that there
        are no parameters of that type to be estimated. A value of 1 or
        -1 indicates that there are parameters of that type in the model
        and the estimation of that type has been successful or
        unsuccessful respectively. Note that there is always at least
        one MA-like parameter in the model.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nwds = *<value>*, `nna`[1] = *<value>* and
            `nna`[2] = *<value>*.

            Constraint: nwds = `nna`[1]+`nna`[2]+1.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` > 0.0.

        (`errno` 1)
            On entry, i = *<value>* and `r`[i-1] = *<value>*.

            Constraint: -1.0 <= `r`[i] <= 1.0.

        (`errno` 1)
            On entry, `r0` = *<value>*.

            Constraint: -1.0 <= `r0` <= 1.0.

        (`errno` 1)
            On entry, nl = *<value>*, `nna`[0] = *<value>*, `nna`[1] =
            *<value>* and `nna`[2] = *<value>*.

            Constraint: nl >= max(`nna`[0]+`nna`[1]+`nna`[2], 1).

        (`errno` 1)
            On entry, i = *<value>* and `nna`[i-1] = *<value>*.

            Constraint: `nna`[i-1] >= 0.
    """

    from ..base.tsa import multi_transf_prelim as b_multi_transf_prelim

    fun_name = 'naginterfaces.library.tsa.multi_transf_prelim'

    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nl_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=1)[0]
    _nl = _b_utils._EngineIntScalarType(
        _nl_py, {'fun_name': fun_name, 'entity': '_nl'},
    ).to_ctypes()
    _nna = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nna'},
        nna,
        copy=False,
    )
    _nwds = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (1,))+_b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (2,))+1, {'fun_name': fun_name, 'entity': '_nwds'},
    ).to_ctypes()
    _wds = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (1,))+_b_utils._EngineIntArrayType.get_element(_nna, {'fun_name': fun_name, 'entity': 'nna'}, (2,))+1,),
    )
    _isf = _b_utils._EngineIntArrayType.empty_ndarray(
        (2,),
    )
    b_multi_transf_prelim(
        r0,
        _r,
        _nl,
        _nna,
        s,
        _nwds,
        _wds,
        _isf,
    )
    _wds_py = _wds
    _isf_py = _isf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['wds', 'isf'],
    )(
        _wds_py,
        _isf_py,
    )
    return _return_data

def multi_inputmod_estim(mr, mt, para, xxy, kfc=1, kef=2, nit=1000, zsp=None, kpriv=0, io_manager=None):
    u"""
    Multivariate time series, estimation of multi-input model.

    ``multi_inputmod_estim`` fits a multi-input model relating one
    output series to the input series with a choice of three different
    estimation criteria: nonlinear least squares, exact likelihood and
    marginal likelihood.
    When no input series are present, ``multi_inputmod_estim`` fits a
    univariate ARIMA model.

    For full information please refer to the NAG Library document for
    g13be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bef.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function model orders b, p and q of each of the
        input series. The order parameters for input series i are held
        in column i. Row 1 holds the value b_i, row 2 holds the value
        q_i and row 3 holds the value p_i. For a simple input, b_i = q_i
        = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, r_i
        = 2 for a transfer function input for which no allowance is to
        be made for pre-observation period effects, and r_i = 3 for a
        transfer function input for which pre-observation period effects
        will be treated by estimation of appropriate nuisance
        parameters.

        When r_i = 1, any nonzero contents of rows 1, 2, and 3 of column
        i are ignored.

    para : float, array-like, shape (npara)
        Initial values of the multi-input model parameters. These are in
        order, firstly the ARIMA model parameters: p values of phi
        parameters, q values of theta parameters, P values of Phi
        parameters and Q values of Theta parameters. These are followed
        by initial values of the transfer function model parameters
        omega_0,omega_1,...,omega_q_1, delta_1,delta_2,...,delta_p_1 for
        the first of any input series and similarly for each subsequent
        input series. The final component of `para` is the initial value
        of the constant c, whether it is fixed or is to be estimated.

    xxy : float, array-like, shape (nxxy, nser)
        The columns of `xxy` must contain the nxxy original,
        undifferenced values of each of the input series and the output
        series x_t in that order.

    kfc : int, optional
        Must be set to 0 if the constant c is to remain fixed at its
        initial value, and 1 if it is to be estimated.

    kef : int, optional
        Indicates the likelihood option.

        `kef` = 1

          Gives least squares.

        `kef` = 2

          Gives exact likelihood.

        `kef` = 3

          Gives marginal likelihood.

    nit : int, optional
        The maximum required number of iterations.

        `nit` = 0

          No change is made to any of the model parameters in array
          `para` except that the constant c (if `kfc` = 1) and any omega
          relating to simple input series are estimated. (Apart from
          these, estimates are always derived for the nuisance
          parameters relating to any backforecasts and any
          pre-observation period effects for transfer function inputs.)

    zsp : None or float, array-like, shape (4), optional
        If `zsp` is not None, then `zsp` must contain the four values
        used to control the strategy of the search procedure.

        `zsp`[0]

          Contains alpha, the value used to constrain the magnitude of
          the search procedure steps.

        `zsp`[1]

          Contains beta, the multiplier which regulates the value of
          alpha.

        `zsp`[2]

          Contains delta, the value of the stationarity and
          invertibility test tolerance factor.

        `zsp`[3]

          Contains gamma, the value of the convergence criterion.

        If `zsp` is None before entry, default values of `zsp` are
        supplied by the function.

        These are 0.01, 10.0, 1000.0 and max(100*machine
        precision,0.0000001), respectively.

    kpriv : int, optional
        Must not be set to 0, if it is required to monitor the course of
        the optimization. The course of the optimization is monitored by
        printing out at each iteration the iteration count (`itc`), the
        residual sum of squares (`s`), the objective function (`d`) and
        a description and value for each of the parameters in the `para`
        array. The descriptions are PHI for phi, THETA for theta, SPHI
        for Phi, STHETA for Theta, OMEGA/SI for omega in a simple input,
        OMEGA for omega in a transfer function input, DELTA for delta
        and CONSTANT for c. In addition SERIES 1, SERIES 2, etc.
        indicate the input series relevant to the OMEGA and DELTA
        parameters.

        `kpriv` must be set to 0 if the print-out of the above
        information is not required.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    para : float, ndarray, shape (npara)
        The latest values of the estimates of these parameters.

    xxy : float, ndarray, shape (nxxy, nser)
        If kzef = 0, `xxy` remains unchanged on exit.

        If kzef != 0, the columns of `xxy` hold the corresponding values
        of the input component series z_t in place of x_t and the output
        noise component n_t in place of y_t, in that order.

    zsp : float, ndarray, shape (4)
        Contains the values, default or otherwise, used by the function.

    itc : int
        The number of iterations carried out.

        `itc` = -1

          Indicates that the only estimates obtained up to this point
          have been for the nuisance parameters relating to
          backforecasts, unless the marginal likelihood option is used,
          in which case estimates have also been obtained for simple
          input coefficients omega and for the constant c (if `kfc` =
          1). This value of `itc` usually indicates a failure in a
          consequent step of estimating transfer function input
          pre-observation period nuisance parameters.

        `itc` = 0

          Indicates that estimates have been obtained up to this point
          for the constant c (if `kfc` = 1), for simple input
          coefficients omega and for the nuisance parameters relating to
          the backforecasts and to transfer function input
          pre-observation period effects.

    sd : float, ndarray, shape (npara)
        The npara values of the standard deviations corresponding to
        each of the parameters in `para`. When the constant is fixed its
        standard deviation is returned as zero. When the values of
        `para` are valid, the values of `sd` are usually also valid.
        However, if an exit value of `errno` = 3, 8 or 10, then the
        contents of `sd` will be indeterminate.

    cm : float, ndarray, shape (npara, npara)
        The first npara rows and columns of `cm` contain the correlation
        coefficients relating to each pair of parameters in `para`. All
        coefficients relating to the constant will be zero if the
        constant is fixed. The contents of `cm` will be indeterminate
        under the same conditions as `sd`.

    s : float
        The residual sum of squares, S, at the latest set of valid
        parameter estimates.

    d : float
        The objective function, D, at the latest set of valid parameter
        estimates.

    ndf : int
        The number of degrees of freedom associated with S.

    res : float, ndarray, shape (nxxy)
        The values of the residuals relating to the differenced values
        of the output series. The remainder of the first nxxy terms in
        the array will be zero.

    sttf : float, ndarray, shape (`nsttf`)
        The `nsttf` values of the state set array.

    nsttf : int
        The number of values in the state set array `sttf`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nser = 1 and there are no parameters in the model.

        (`errno` 1)
            On entry, nser = *<value>*.

            Constraint: nser >= 1.

        (`errno` 1)
            On entry, `nit` = *<value>*.

            Constraint: `nit` >= 0.

        (`errno` 1)
            On entry, `kef` = *<value>*.

            Constraint: `kef` = 1, 2 or 3.

        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 2)
            On entry, `ndf` = *<value>*.

            Constraint: `ndf` > 0.

        (`errno` 2)
            On entry, npara = *<value>* and `kfc` = *<value>*.

            Constraint: npara, `kfc`, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

        (`errno` 2)
            The orders vector `mr` is invalid.

        (`errno` 3)
            On entry, or during execution, one or more sets of delta
            parameters do not satisfy the stationarity or invertibility
            conditions.

        (`errno` 4)
            On entry, `zsp`[3] = *<value>*.

            Constraint: 0.0 <= `zsp`[3] < 1.0.

        (`errno` 4)
            On entry, `zsp`[2] = *<value>*.

            Constraint: `zsp`[2] >= 1.0.

        (`errno` 4)
            On entry, `zsp`[1] = *<value>*.

            Constraint: `zsp`[1] > 1.0.

        (`errno` 4)
            On entry, `zsp`[0] = *<value>*.

            Constraint: `zsp`[0] > 0.0.

        (`errno` 8)
            Unable to calculate the latest parameter estimates.

        (`errno` 9)
            Failure in inversion of second derivative matrix.

        (`errno` 10)
            On entry, or during execution, one or more sets of ARIMA
            parameters do not satisfy the stationarity or invertibility
            conditions.

        (`errno` 11)
            On entry, isttf too small.

            Constraint: isttf >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncg.

        (`errno` 12)
            The routine has failed to converge after *<value>*
            iterations.

        (`errno` 13)
            On entry, `nit` = *<value>*.

            isttf is too small and the process failed to converge after
            `nit` iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.multi_inputmod_ex.main`
    """

    from ..base.tsa import multi_inputmod_estim as b_multi_inputmod_estim

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_estim'

    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _xxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxy'},
        xxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_mt, 'mt'), (_xxy, 'xxy')),
    )
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _nxxy_py = _b_utils._EngineFloat64ArrayType.get_shape(_xxy, {'fun_name': fun_name, 'entity': 'xxy'}, exp_rank=2)[0]
    _nxxy = _b_utils._EngineIntScalarType(
        _nxxy_py, {'fun_name': fun_name, 'entity': '_nxxy'},
    ).to_ctypes()
    if zsp is None:
        _kzsp = 0
        _kzsp_py = _kzsp
    else:
        _kzsp = 1
        _kzsp_py = _kzsp
    if zsp is None:
        _zsp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (4,),
        )
    else:
        _zsp = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'zsp'},
            zsp,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _sd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npara_py,),
    )
    _cm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npara_py, _npara_py),
        sorder=_m_sorder,
    )
    _kzef = 1
    _kzef_py = _kzef
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nxxy_py,),
    )
    _vl_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (0,))
    _vl_d_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (1,))
    _vu_p_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (3,))
    _vl_s_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (6,))
    _vl_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (2,))
    _vu_q_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (5,))
    _vu_d_py = _b_utils._EngineIntArrayType.get_element(_mr, {'fun_name': fun_name, 'entity': 'mr'}, (4,))
    _vl_ncg_py = 0
    for _vl_i_py in range(1, _nser_py+1):
        if _b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (3, _vl_i_py-1)) > 1:
            _vl_ncg_py = _vl_ncg_py+_b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (0, _vl_i_py-1))+_b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (1, _vl_i_py-1))+_b_utils._EngineIntArrayType.get_element(_mt, {'fun_name': fun_name, 'entity': 'mt'}, (2, _vl_i_py-1))
    _vl_ncg = _b_utils._EngineIntScalarType(
        _vl_ncg_py, {'fun_name': fun_name, 'entity': '_vl_ncg_py'},
    ).to_ctypes()
    _isttf = _b_utils._EngineIntScalarType(
        (_vu_p_py*_vl_s_py)+_vl_d_py+(_vu_d_py*_vl_s_py)+_vl_q_py+max(_vl_p_py, _vu_q_py*_vl_s_py)+_vl_ncg_py, {'fun_name': fun_name, 'entity': '_isttf'},
    ).to_ctypes()
    _sttf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_vu_p_py*_vl_s_py)+_vl_d_py+(_vu_d_py*_vl_s_py)+_vl_q_py+max(_vl_p_py, _vu_q_py*_vl_s_py)+_vl_ncg_py,),
    )
    _return_scalars = b_multi_inputmod_estim(
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        kfc,
        _nxxy,
        _xxy,
        kef,
        nit,
        _kzsp,
        _zsp,
        _sd,
        _cm,
        _kzef,
        _res,
        _sttf,
        _isttf,
        kpriv,
        io_manager,
    )
    _para_py = _para
    _xxy_py = _xxy
    _zsp_py = _zsp
    _itc_py = _return_scalars.itc
    _sd_py = _sd
    _cm_py = _cm
    _s_py = _return_scalars.s
    _d_py = _return_scalars.d
    _ndf_py = _return_scalars.ndf
    _res_py = _res
    _nsttf_py = _return_scalars.nsttf
    _sttf_py = _sttf[:_nsttf_py]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['para', 'xxy', 'zsp', 'itc', 'sd', 'cm', 's', 'd', 'ndf', 'res', 'sttf', 'nsttf'],
    )(
        _para_py,
        _xxy_py,
        _zsp_py,
        _itc_py,
        _sd_py,
        _cm_py,
        _s_py,
        _d_py,
        _ndf_py,
        _res_py,
        _sttf_py,
        _nsttf_py,
    )
    return _return_data

def multi_inputmod_update(sttf, mr, mt, para, xxyn, kzef):
    u"""
    Multivariate time series, update state set for forecasting from
    multi-input model.

    ``multi_inputmod_update`` accepts a series of new observations of an
    output time series and any associated input time series, for which a
    multi-input model is already fully specified, and updates the 'state
    set' information for use in constructing further forecasts.

    The previous specification of the multi-input model will normally
    have been obtained by using ``multi_inputmod_estim`` to estimate the
    relevant transfer function and ARIMA parameters.
    The supplied state set will originally have been produced by
    ``multi_inputmod_estim`` (or possibly ``multi_inputmod_forecast``),
    but may since have been updated by ``multi_inputmod_update``.

    For full information please refer to the NAG Library document for
    g13bg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bgf.html

    Parameters
    ----------
    sttf : float, array-like, shape (nsttf)
        The nsttf values in the state set before updating as returned by
        ``multi_inputmod_estim`` or ``multi_inputmod_forecast``, or a
        previous call to ``multi_inputmod_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing, and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function model orders b, p and q of each of the
        input series. The data for input series i are held in column i.
        Row 1 holds the value b_i, row 2 holds the value q_i and row 3
        holds the value p_i. For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input and
        r_i = 2 or 3 for a transfer function input.

        When r_i = 1 any nonzero contents of rows 1, 2 and 3 of column i
        are ignored.

        The choice of r_i = 2 or r_i = 3 is an option for use in model
        estimation and does not affect the operation of
        ``multi_inputmod_update``.

    para : float, array-like, shape (npara)
        Estimates of the multi-input model parameters as returned by
        ``multi_inputmod_estim``. These are in order, firstly the ARIMA
        model parameters: p values of phi parameters, q values of theta
        parameters, P values of Phi parameters and Q values of Theta
        parameters. These are followed by the transfer function model
        parameter values omega_0,omega_1,...,omega_q_1,
        delta_1,delta_2,...,delta_p_1 for the first of any input series
        and similarly for each subsequent input series. The final
        component of `para` is the value of the constant c.

    xxyn : float, array-like, shape (nnv, nser)
        The nnv new observation sets being used to update the state set.
        Column i-1 contains the values of input series i, for i =
        1,2,...,nser-1. Column nser-1 contains the values of the output
        series. Consecutive rows correspond to increasing time sequence.

    kzef : int
        Must not be set to 0, if the values of the input component
        series z_t and the values of the output noise component n_t are
        to overwrite the contents of `xxyn` on exit, and must be set to
        0 if `xxyn` is to remain unchanged on exit.

    Returns
    -------
    sttf : float, ndarray, shape (nsttf)
        The state set values after updating.

    xxyn : float, ndarray, shape (nnv, nser)
        If `kzef` = 0, `xxyn` remains unchanged.

        If `kzef` != 0, the columns of `xxyn` hold the corresponding
        values of the input component series z_t and the output noise
        component n_t in that order.

    res : float, ndarray, shape (nnv)
        The values of the residual series a_t corresponding to the new
        observations of the output series.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nsttf = *<value>*.

            Constraint: nsttf, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, npara = *<value>*.

            Constraint: npara, `mr` and `mt` must be consistent.

        (`errno` 5)
            On entry, the orders vector `mr` is invalid.

        (`errno` 5)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.
    """

    from ..base.tsa import multi_inputmod_update as b_multi_inputmod_update

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_update'

    _sttf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sttf'},
        sttf,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nsttf_py = _b_utils._EngineFloat64ArrayType.get_shape(_sttf, {'fun_name': fun_name, 'entity': 'sttf'}, exp_rank=1)[0]
    _nsttf = _b_utils._EngineIntScalarType(
        _nsttf_py, {'fun_name': fun_name, 'entity': '_nsttf'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=False,
    )
    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _xxyn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxyn'},
        xxyn,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnv_py = _b_utils._EngineFloat64ArrayType.get_shape(_xxyn, {'fun_name': fun_name, 'entity': 'xxyn'}, exp_rank=2)[0]
    _nnv = _b_utils._EngineIntScalarType(
        _nnv_py, {'fun_name': fun_name, 'entity': '_nnv'},
    ).to_ctypes()
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nnv_py,),
    )
    b_multi_inputmod_update(
        _sttf,
        _nsttf,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _nnv,
        _xxyn,
        kzef,
        _res,
    )
    _sttf_py = _sttf
    _xxyn_py = _xxyn
    _res_py = _res
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['sttf', 'xxyn', 'res'],
    )(
        _sttf_py,
        _xxyn_py,
        _res_py,
    )
    return _return_data

def multi_inputmod_forecast_state(sttf, mr, mt, para, xxyn, mrx, parx, rmsxy, kzef):
    u"""
    Multivariate time series, forecasting from state set of multi-input
    model.

    ``multi_inputmod_forecast_state`` produces forecasts of a time
    series (the output series) which depends on one or more other
    (input) series via a multi-input model which will usually have been
    fitted using ``multi_inputmod_estim``.
    The future values of the input series must be supplied.
    The original observations are not required.
    ``multi_inputmod_forecast_state`` uses as input either the original
    state set obtained from ``multi_inputmod_estim``, or the state set
    updated by a series of new observations from
    ``multi_inputmod_update``.
    Standard errors of the forecasts are produced.
    If future values of some of the input series have been obtained as
    forecasts using ARIMA models for those series, this may be allowed
    for in the calculation of the standard errors.

    For full information please refer to the NAG Library document for
    g13bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bhf.html

    Parameters
    ----------
    sttf : float, array-like, shape (nsttf)
        The nsttf values in the state set as returned by
        ``multi_inputmod_estim`` or ``multi_inputmod_update``.

    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q give respectively the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing, and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function orders b, p and q of each of the input
        series. The data for input series i are held in column i. Row 1
        holds the value b_i, row 2 holds the value q_i and row 3 holds
        the value p_i. For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, r_i
        = 2 or 3 for a transfer function input.

        When r_i = 1, any nonzero contents of rows 1, 2 and 3 of column
        i are ignored.

        The choice of r_i = 2 or r_i = 3 is an option for use in model
        estimation and does not affect the operation of
        ``multi_inputmod_forecast_state``.

    para : float, array-like, shape (npara)
        Estimates of the multi-input model parameters as returned by
        ``multi_inputmod_estim``. These are in order, firstly the ARIMA
        model parameters: p values of phi parameters, q values of theta
        parameters, P values of Phi parameters and Q values of Theta
        parameters. These are followed by the transfer function model
        parameter values omega_0,omega_1,...,omega_q_1,
        delta_1,delta_2,...,delta_p_1 for the first of any input series
        and similar sets of values for any subsequent input series. The
        final component of `para` is the constant c.

    xxyn : float, array-like, shape (nfv, nser)
        The supplied nfv values for each of the input series required to
        produce the nfv output series forecasts. Column i contains the
        values for input series i. Column nser need not be supplied.

    mrx : int, array-like, shape (7, nser)
        The orders array for each of the input series ARIMA models.
        Thus, column i contains values of p, d, q, P, D, Q, s for input
        series i. In the case of those inputs for which no ARIMA model
        is available, the corresponding orders should be set to 0. (The
        model for any input series only affects the standard errors of
        the forecast values.)

    parx : float, array-like, shape (:, nser)
        Values of the parameters (phi, theta, Phi and Theta) for each of
        the input series ARIMA models. Thus column i contains
        `mrx`[0,i-1] values of phi parameters, `mrx`[2,i-1] values of
        theta parameters, `mrx`[3,i-1] values of Phi parameters and
        `mrx`[5,i-1] values of Theta parameters -- in that order.

        Values in the columns relating to those input series for which
        no ARIMA model is available are ignored. (The model for any
        input series only affects the standard errors of the forecast
        values.)

    rmsxy : float, array-like, shape (nser)
        The estimated residual variances for each input series ARIMA
        model followed by that for the output noise ARIMA model. In the
        case of those inputs for which no ARIMA model is available, or
        when its effects are to be excluded in the calculation of
        forecast standard errors, the corresponding entry of `rmsxy`
        should be set to 0.

    kzef : int
        Must not be set to 0, if the values of the input component
        series z_t and the values of the output noise component n_t are
        to overwrite the contents of `xxyn` on exit, and must be set to
        0 if `xxyn` is to remain unchanged on exit, apart from the
        appearance of the forecast values in column nser.

    Returns
    -------
    xxyn : float, ndarray, shape (nfv, nser)
        If `kzef` = 0, then column nser of `xxyn` contains the output
        series forecast values (as does `fva`), but `xxyn` is otherwise
        unchanged.

        If `kzef` != 0, then the columns of `xxyn` hold the
        corresponding values of the forecast components z_t for each of
        the input series and the values of the output noise component
        n_t in that order.

    mrx : int, ndarray, shape (7, nser)
        Unchanged, apart from column nser which is used for workspace.

    fva : float, ndarray, shape (nfv)
        The required forecast values for the output series.

    fsd : float, ndarray, shape (nfv)
        The standard errors for each of the forecast values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nsttf = *<value>* and the expected value =
            *<value>*.

            Constraint: nsttf, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, npara = *<value>* and the expected value =
            *<value>*.

            Constraint: npara, `mr` and `mt` must be consistent.

        (`errno` 6)
            The orders vector `mr` is invalid.

        (`errno` 6)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.multi_inputmod_ex.main`
    """

    from ..base.tsa import multi_inputmod_forecast_state as b_multi_inputmod_forecast_state

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_forecast_state'

    _sttf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sttf'},
        sttf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nsttf_py = _b_utils._EngineFloat64ArrayType.get_shape(_sttf, {'fun_name': fun_name, 'entity': 'sttf'}, exp_rank=1)[0]
    _nsttf = _b_utils._EngineIntScalarType(
        _nsttf_py, {'fun_name': fun_name, 'entity': '_nsttf'},
    ).to_ctypes()
    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=True,
    )
    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _xxyn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxyn'},
        xxyn,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nfv_py = _b_utils._EngineFloat64ArrayType.get_shape(_xxyn, {'fun_name': fun_name, 'entity': 'xxyn'}, exp_rank=2)[0]
    _nfv = _b_utils._EngineIntScalarType(
        _nfv_py, {'fun_name': fun_name, 'entity': '_nfv'},
    ).to_ctypes()
    _mrx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mrx'},
        mrx,
        copy=True,
    )
    _parx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parx'},
        parx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rmsxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rmsxy'},
        rmsxy,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    b_multi_inputmod_forecast_state(
        _sttf,
        _nsttf,
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        _nfv,
        _xxyn,
        _mrx,
        _parx,
        _rmsxy,
        kzef,
        _fva,
        _fsd,
    )
    _xxyn_py = _xxyn
    _mrx_py = _mrx
    _fva_py = _fva
    _fsd_py = _fsd
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xxyn', 'mrx', 'fva', 'fsd'],
    )(
        _xxyn_py,
        _mrx_py,
        _fva_py,
        _fsd_py,
    )
    return _return_data

def multi_inputmod_forecast(mr, mt, para, kfc, nev, nfv, xxy, kzef, rmsxy, mrx, parx, isttf, io_manager=None):
    u"""
    Multivariate time series, state set and forecasts from fully
    specified multi-input model.

    ``multi_inputmod_forecast`` produces forecasts of a time series (the
    output series) which depends on one or more other (input) series via
    a previously estimated multi-input model for which the state set
    information is not available.
    The future values of the input series must be supplied.
    In contrast with ``multi_inputmod_forecast_state`` the original past
    values of the input and output series are required.
    Standard errors of the forecasts are produced.
    If future values of some of the input series have been obtained as
    forecasts using ARIMA models for those series, this may be allowed
    for in the calculation of the standard errors.

    For full information please refer to the NAG Library document for
    g13bj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13bjf.html

    Parameters
    ----------
    mr : int, array-like, shape (7)
        The orders vector (p,d,q,P,D,Q,s) of the ARIMA model for the
        output noise component.

        p, q, P and Q refer respectively to the number of autoregressive
        (phi), moving average (theta), seasonal autoregressive (Phi) and
        seasonal moving average (Theta) parameters.

        d, D and s refer respectively to the order of non-seasonal
        differencing, the order of seasonal differencing and the
        seasonal period.

    mt : int, array-like, shape (4, nser)
        The transfer function model orders b, p and q of each of the
        input series. The data for input series i is held in column i.
        Row 1 holds the value b_i, row 2 holds the value q_i and row 3
        holds the value p_i.

        For a simple input, b_i = q_i = p_i = 0.

        Row 4 holds the value r_i, where r_i = 1 for a simple input, and
        r_i = 2 or 3 for a transfer function input.

        The choice r_i = 3 leads to estimation of the pre-period input
        effects as nuisance parameters, and r_i = 2 suppresses this
        estimation.

        This choice may affect the returned forecasts and the state set.

        When r_i = 1, any nonzero contents of rows 1, 2 and 3 of column
        i are ignored.

    para : float, array-like, shape (npara)
        Estimates of the multi-input model parameters. These are in
        order, firstly the ARIMA model parameters: p values of phi
        parameters, q values of theta parameters, P values of Phi
        parameters, Q values of Theta parameters.

        These are followed by the transfer function model parameter
        values omega_0,omega_1,...,omega_q_1, delta_1,...,delta_p_1 for
        the first of any input series and similarly for each subsequent
        input series.

        The final component of `para` is the value of the constant c.

    kfc : int
        Must be set to 1 if the constant was estimated when the model
        was fitted, and 0 if it was held at a fixed value. This only
        affects the degrees of freedom used in calculating the estimated
        residual variance.

    nev : int
        The number of original (undifferenced) values in each of the
        input and output time series.

    nfv : int
        The number of forecast values of the output series required.

    xxy : float, array-like, shape ((`nev`+`nfv`), nser)
        The columns of `xxy` must contain in the first `nev` places, the
        past values of each of the input and output series, in that
        order. In the next `nfv` places, the columns relating to the
        input series (i.e., columns 1 to nser-1) contain the future
        values of the input series which are necessary for construction
        of the forecasts of the output series y.

    kzef : int
        Must be set to 0 if the relevant `nfv` values of the forecasts
        (`fva`) are to be held in the output series column (nser) of
        `xxy` (which is otherwise unchanged) on exit, and must not be
        set to 0 if the values of the input component series z_t and the
        values of the output noise component n_t are to overwrite the
        contents of `xxy` on exit.

    rmsxy : float, array-like, shape (nser)
        The first (nser-1) elements of `rmsxy` must contain the
        estimated residual variance of the input series ARIMA models. In
        the case of those inputs for which no ARIMA model is available
        or its effects are to be excluded in the calculation of forecast
        standard errors, the corresponding entry of `rmsxy` should be
        set to 0.

    mrx : int, array-like, shape (7, nser)
        The orders array for each of the input series ARIMA models.
        Thus, column i contains values of p, d, q, P, D, Q, s for input
        series i. In the case of those inputs for which no ARIMA model
        is available, the corresponding orders should be set to 0.

    parx : float, array-like, shape (:, nser)
        Values of the parameters (phi, theta, Phi, and Theta) for each
        of the input series ARIMA models.

        Thus column i contains `mrx`[0,i-1] values of phi, `mrx`[2,i-1]
        values of theta, `mrx`[3,i-1] values of Phi and `mrx`[5,i-1]
        values of Theta, in that order.

        Values in the columns relating to those input series for which
        no ARIMA model is available are ignored.

    isttf : int
        The dimension of the array `sttf`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    para : float, ndarray, shape (npara)
        The parameter values may be updated using an additional
        iteration in the estimation process.

    xxy : float, ndarray, shape ((`nev`+`nfv`), nser)
        If `kzef` = 0 then `xxy` is unchanged except that the relevant
        `nfv` values in the column relating to the output series (column
        nser) contain the forecast values (`fva`), but if `kzef` != 0
        then the columns of `xxy` contain the corresponding values of
        the input component series z_t and the values of the output
        noise component n_t, in that order.

    rmsxy : float, ndarray, shape (nser)
        `rmsxy`[nser-1] contains the estimated residual variance of the
        output noise ARIMA model which is calculated from the supplied
        series. Otherwise `rmsxy` is unchanged.

    mrx : int, ndarray, shape (7, nser)
        Unchanged, except for column nser which is used as workspace.

    fva : float, ndarray, shape (`nfv`)
        The required forecast values for the output series. (Note that
        these are also output in column nser of `xxy` if `kzef` = 0.)

    fsd : float, ndarray, shape (`nfv`)
        The standard errors for each of the forecast values.

    sttf : float, ndarray, shape (`isttf`)
        The `nsttf` values of the state set based on the first `nev`
        sets of (past) values of the input and output series.

    nsttf : int
        The number of values in the state set array `sttf`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nfv` = *<value>*.

            Constraint: `nfv` > 0.

        (`errno` 1)
            On entry, ldxxy = *<value>*, `nev` = *<value>* and `nfv` =
            *<value>*.

            Constraint: ldxxy >= `nev`+`nfv`.

        (`errno` 1)
            On entry, `kfc` = *<value>*.

            Constraint: `kfc` = 0 or 1.

        (`errno` 2)
            The orders vector `mr` is invalid.

        (`errno` 2)
            npara is inconsistent with `mr` and `mt`.

        (`errno` 2)
            Insufficient degrees of freedom to solve the problem.

        (`errno` 2)
            On entry, npara = *<value>* and the expected value =
            *<value>*.

            Constraint: npara, `mr` and `mt` must be consistent.

        (`errno` 2)
            On entry, i = *<value>* and `mt`[3,i-1] = *<value>*.

            Constraint: `mt`[3,i-1] = 1, 2 or 3.

        (`errno` 3)
            One or more sets of delta parameters do not satisfy the
            stationarity or invertibility conditions.

        (`errno` 8)
            Unable to calculate the latest parameter estimates.

        (`errno` 9)
            Failure in inversion of second derivative matrix.

        (`errno` 10)
            One or more sets of ARIMA parameters do not satisfy the
            stationarity or invertibility conditions.

        (`errno` 11)
            On entry, `isttf` = *<value>*.

            Constraint: `isttf` >= (P*s)+d+(D*s)+q+max(p, Q*s)+ncf.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import multi_inputmod_forecast as b_multi_inputmod_forecast

    fun_name = 'naginterfaces.library.tsa.multi_inputmod_forecast'

    _mr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mr'},
        mr,
        copy=True,
    )
    _mt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mt'},
        mt,
        copy=False,
    )
    _nser_py = _b_utils._EngineIntArrayType.get_shape(_mt, {'fun_name': fun_name, 'entity': 'mt'}, exp_rank=2)[1]
    _nser = _b_utils._EngineIntScalarType(
        _nser_py, {'fun_name': fun_name, 'entity': '_nser'},
    ).to_ctypes()
    _para = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'para'},
        para,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npara_py = _b_utils._EngineFloat64ArrayType.get_shape(_para, {'fun_name': fun_name, 'entity': 'para'}, exp_rank=1)[0]
    _npara = _b_utils._EngineIntScalarType(
        _npara_py, {'fun_name': fun_name, 'entity': '_npara'},
    ).to_ctypes()
    _xxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xxy'},
        xxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rmsxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rmsxy'},
        rmsxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mrx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mrx'},
        mrx,
        copy=True,
    )
    _parx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parx'},
        parx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nfv = _b_utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _nfv_py = _b_utils._EngineIntScalarType.to_py(_nfv)
    _fva = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _fsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nfv_py,),
    )
    _isttf = _b_utils._EngineIntScalarType(
        isttf, {'fun_name': fun_name, 'entity': 'isttf'},
    ).to_ctypes()
    _isttf_py = _b_utils._EngineIntScalarType.to_py(_isttf)
    _sttf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_isttf_py,),
    )
    _return_scalar = b_multi_inputmod_forecast(
        _mr,
        _nser,
        _mt,
        _para,
        _npara,
        kfc,
        nev,
        _nfv,
        _xxy,
        kzef,
        _rmsxy,
        _mrx,
        _parx,
        _fva,
        _fsd,
        _sttf,
        _isttf,
        io_manager,
    )
    _para_py = _para
    _xxy_py = _xxy
    _rmsxy_py = _rmsxy
    _mrx_py = _mrx
    _fva_py = _fva
    _fsd_py = _fsd
    _sttf_py = _sttf
    _nsttf_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['para', 'xxy', 'rmsxy', 'mrx', 'fva', 'fsd', 'sttf', 'nsttf'],
    )(
        _para_py,
        _xxy_py,
        _rmsxy_py,
        _mrx_py,
        _fva_py,
        _fsd_py,
        _sttf_py,
        _nsttf_py,
    )
    return _return_data

def uni_spectrum_lag(nx, mtx, px, iw, mw, ic, c, kc, l, lg, xg):
    u"""
    Univariate time series, smoothed sample spectrum using rectangular,
    Bartlett, Tukey or Parzen lag window.

    ``uni_spectrum_lag`` calculates the smoothed sample spectrum of a
    univariate time series using one of four lag windows -- rectangular,
    Bartlett, Tukey or Parzen window.

    For full information please refer to the NAG Library document for
    g13ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13caf.html

    Parameters
    ----------
    nx : int
        n, the length of the time series.

    mtx : int
        If covariances are to be calculated by the function (`ic` = 0),
        `mtx` must specify whether the data are to be initially mean or
        trend corrected.

        `mtx` = 0

          For no correction.

        `mtx` = 1

          For mean correction.

        `mtx` = 2

          For trend correction.

    px : float
        If covariances are to be calculated by the function (`ic` = 0),
        `px` must specify the proportion of the data (totalled over both
        ends) to be initially tapered by the split cosine bell taper.

        If covariances are supplied (`ic` != 0), `px` must specify the
        proportion of data tapered before the supplied covariances were
        calculated and after any mean or trend correction. `px` is
        required for the calculation of output statistics.

        A value of 0.0 implies no tapering.

    iw : int
        The choice of lag window.

        `iw` = 1

          Rectangular.

        `iw` = 2

          Bartlett.

        `iw` = 3

          Tukey.

        `iw` = 4

          Parzen.

    mw : int
        M, the 'cut-off' point of the lag window. Windowed covariances
        at lag M or greater are zero.

    ic : int
        Indicates whether covariances are to be calculated in the
        function or supplied in the call to the function.

        `ic` = 0

          Covariances are to be calculated.

        `ic` != 0

          Covariances are to be supplied.

    c : float, array-like, shape (nc)
        If `ic` != 0, `c` must contain the nc covariances for lags from
        0 to (nc-1), otherwise `c` need not be set.

    kc : int
        If `ic` = 0, `kc` must specify the order of the fast Fourier
        transform (FFT) used to calculate the covariances.

        If `ic` != 0, that is covariances are supplied, `kc` is not
        used.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        Therefore it is also the order of the FFT used to construct the
        sample spectrum from the covariances.

    lg : int
        Indicates whether unlogged or logged spectral estimates and
        confidence limits are required.

        `lg` = 0

          Unlogged.

        `lg` != 0

          Logged.

    xg : float, array-like, shape (nxg)
        If the covariances are to be calculated, then `xg` must contain
        the `nx` data points. If covariances are supplied, `xg` may
        contain any values.

    Returns
    -------
    c : float, ndarray, shape (nc)
        If `ic` = 0, `c` will contain the nc calculated covariances.

        If `ic` != 0, the contents of `c` will be unchanged.

    xg : float, ndarray, shape (nxg)
        Contains the `ng` spectral estimates, f^(omega_i), for i =
        0,1,...,[L/2] in `xg`[0] to `xg`[`ng`-1] respectively (logged if
        `lg` = 1). The elements `xg`[i-1], for i = `ng`+1,...,nxg
        contain 0.0.

    ng : int
        The number of spectral estimates, [L/2]+1, in `xg`.

    stats : float, ndarray, shape (4)
        Four associated statistics. These are the degrees of freedom in
        `stats`[0], the lower and upper 95% confidence limit factors in
        `stats`[1] and `stats`[2] respectively (logged if `lg` = 1), and
        the bandwidth in `stats`[3].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nc = *<value>* and `nx` = *<value>*.

            Constraint: nc <= `nx`.

        (`errno` 1)
            On entry, nc = *<value>* and `mw` = *<value>*.

            Constraint: nc >= `mw`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nx` = *<value>*.

            Constraint: `mw` <= `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `iw` = *<value>*.

            Constraint: `iw` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` <= 1.0.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 0.0.

        (`errno` 1)
            On entry, nxg = *<value>* and `l` = *<value>*.

            Constraint: if `ic` != 0, nxg >= `l`.

        (`errno` 1)
            On entry, nxg = *<value>*, `kc` = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` = 0, nxg >= max(`kc`, `l`).

        (`errno` 1)
            On entry, `ic` = 0 and `mtx` > 2: `mtx` = *<value>*.

        (`errno` 1)
            On entry, `ic` = 0 and `mtx` < 0: `mtx` = *<value>*.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>*, `nx` = *<value>* and nc =
            *<value>*.

            Constraint: if `ic` = 0, `kc` >= (`nx`+nc).

        (`errno` 3)
            On entry, `l` = *<value>* and `mw` = *<value>*.

            Constraint: `l` >= 2*`mw`-1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            One or more spectral estimates are negative.

            Unlogged spectral estimates are returned in `xg`, and the
            degrees of freedom, unloged confidence limit factors and
            bandwidth in `stats`.

        (`errno` 5)
            The calculation of confidence limit factors has failed.

            Spectral estimates (logged if requested) are returned in
            `xg`, and degrees of freedom and bandwidth in `stats`.
    """

    from ..base.tsa import uni_spectrum_lag as b_uni_spectrum_lag

    fun_name = 'naginterfaces.library.tsa.uni_spectrum_lag'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nc_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]
    _nc = _b_utils._EngineIntScalarType(
        _nc_py, {'fun_name': fun_name, 'entity': '_nc'},
    ).to_ctypes()
    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxg_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _nxg = _b_utils._EngineIntScalarType(
        _nxg_py, {'fun_name': fun_name, 'entity': '_nxg'},
    ).to_ctypes()
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalar = b_uni_spectrum_lag(
            nx,
            mtx,
            px,
            iw,
            mw,
            ic,
            _nc,
            _c,
            kc,
            l,
            lg,
            _nxg,
            _xg,
            _stats,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _c_py = _c
    _xg_py = _xg
    _ng_py = _return_scalar
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['c', 'xg', 'ng', 'stats'],
    )(
        _c_py,
        _xg_py,
        _ng_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_spectrum_daniell(nx, mtx, px, mw, pw, l, lg, xg):
    u"""
    Univariate time series, smoothed sample spectrum using spectral
    smoothing by the trapezium frequency (Daniell) window.

    ``uni_spectrum_daniell`` calculates the smoothed sample spectrum of
    a univariate time series using spectral smoothing by the trapezium
    frequency (Daniell) window.

    For full information please refer to the NAG Library document for
    g13cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cbf.html

    Parameters
    ----------
    nx : int
        n, the length of the time series.

    mtx : int
        Whether the data are to be initially mean or trend corrected.

        `mtx` = 0

          For no correction.

        `mtx` = 1

          For mean correction.

        `mtx` = 2

          For trend correction.

    px : float
        The proportion of the data (totalled over both ends) to be
        initially tapered by the split cosine bell taper. (A value of
        0.0 implies no tapering.)

    mw : int
        The value of M which determines the frequency width of the
        smoothing window as 2*pi/M. A value of n implies no smoothing is
        to be carried out.

    pw : float
        p, the shape parameter of the trapezium frequency window.

        A value of 0.0 gives a triangular window, and a value of 1.0 a
        rectangular window.

        If `mw` = `nx`, `pw` is not used.

    l : int
        L, the frequency division of smoothed spectral estimates as
        2*pi/L.

    lg : int
        Indicates whether unlogged or logged spectral estimates and
        confidence limits are required.

        `lg` = 0

          For unlogged.

        `lg` != 0

          For logged.

    xg : float, array-like, shape (kc)
        The n data points.

    Returns
    -------
    xg : float, ndarray, shape (kc)
        Contains the `ng` spectral estimates f^(omega_i), for i =
        0,1,...,[L/2], in `xg`[0] to `xg`[`ng`-1] (logged if `lg` != 0).
        The elements `xg`[i-1], for i = `ng`+1,...,kc, contain 0.0.

    ng : int
        The number of spectral estimates, [L/2]+1, in `xg`.

    stats : float, ndarray, shape (4)
        Four associated statistics. These are the degrees of freedom in
        `stats`[0], the lower and upper 95% confidence limit factors in
        `stats`[1] and `stats`[2] respectively (logged if `lg` != 0),
        and the bandwidth in `stats`[3].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `px` = *<value>*, `mw` = *<value>* and `nx` =
            *<value>*.

            Constraint: if `pw` > 1.0, `mw` = `nx`.

        (`errno` 1)
            On entry, `px` = *<value>*, `mw` = *<value>* and `nx` =
            *<value>*.

            Constraint: if `pw` < 0.0, `mw` = `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nx` = *<value>*.

            Constraint: `mw` <= `nx`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` <= 1.0.

        (`errno` 1)
            On entry, `px` = *<value>*.

            Constraint: `px` >= 0.0.

        (`errno` 1)
            On entry, `mtx` = *<value>*.

            Constraint: `mtx` <= 2.

        (`errno` 1)
            On entry, `mtx` = *<value>*.

            Constraint: `mtx` >= 0.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 2)
            On entry, kc = *<value>* and `l` = *<value>*.

            Constraint: kc must be a multiple of `l`.

        (`errno` 2)
            On entry, kc = *<value>* and `nx` = *<value>*.

            Constraint: kc >= 2*`nx`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            One or more spectral estimates are negative.

            Unlogged spectral estimates are returned in `xg`, and the
            degrees of freedom, unloged confidence limit factors and
            bandwidth in `stats`.

        (`errno` 5)
            The calculation of confidence limit factors has failed.

            Spectral estimates (logged if requested) are returned in
            `xg`, and degrees of freedom and bandwidth in `stats`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_spectrum_daniell as b_uni_spectrum_daniell

    fun_name = 'naginterfaces.library.tsa.uni_spectrum_daniell'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _kc_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _kc = _b_utils._EngineIntScalarType(
        _kc_py, {'fun_name': fun_name, 'entity': '_kc'},
    ).to_ctypes()
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    try:
        _return_scalar = b_uni_spectrum_daniell(
            nx,
            mtx,
            px,
            mw,
            pw,
            l,
            _kc,
            lg,
            _xg,
            _stats,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _xg_py = _xg
    _ng_py = _return_scalar
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xg', 'ng', 'stats'],
    )(
        _xg_py,
        _ng_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_spectrum_lag(nxy, mtxy, pxy, iw, mw, ish, ic, cxy, cyx, kc, l, xg=None, yg=None):
    u"""
    Multivariate time series, smoothed sample cross spectrum using
    rectangular, Bartlett, Tukey or Parzen lag window.

    ``multi_spectrum_lag`` calculates the smoothed sample cross spectrum
    of a bivariate time series using one of four lag windows:
    rectangular, Bartlett, Tukey or Parzen.

    For full information please refer to the NAG Library document for
    g13cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ccf.html

    Parameters
    ----------
    nxy : int
        n, the length of the time series x and y.

    mtxy : int
        If cross-covariances are to be calculated by the function (`ic`
        = 0), `mtxy` must specify whether the data is to be initially
        mean or trend corrected.

        `mtxy` = 0

          For no correction.

        `mtxy` = 1

          For mean correction.

        `mtxy` = 2

          For trend correction.

        If cross-covariances are supplied (`ic` != 0), `mtxy` is not
        used.

    pxy : float
        If cross-covariances are to be calculated by the function (`ic`
        = 0), `pxy` must specify the proportion of the data (totalled
        over both ends) to be initially tapered by the split cosine bell
        taper. A value of 0.0 implies no tapering.

        If cross-covariances are supplied (`ic` != 0), `pxy` is not
        used.

    iw : int
        The choice of lag window.

        `iw` = 1

          Rectangular.

        `iw` = 2

          Bartlett.

        `iw` = 3

          Tukey.

        `iw` = 4

          Parzen.

    mw : int
        M, the 'cut-off' point of the lag window, relative to any
        alignment shift that has been applied. Windowed
        cross-covariances at lags (-`mw`+`ish`) or less, and at lags
        (`mw`+`ish`) or greater are zero.

    ish : int
        S, the alignment shift between the x and y series. If x leads y,
        the shift is positive.

    ic : int
        Indicates whether cross-covariances are to be calculated in the
        function or supplied in the call to the function.

        `ic` = 0

          Cross-covariances are to be calculated.

        `ic` != 0

          Cross-covariances are to be supplied.

    cxy : float, array-like, shape (nc)
        If `ic` != 0, `cxy` must contain the nc cross-covariances
        between values in the y series and earlier values in time in the
        x series, for lags from 0 to (nc-1).

        If `ic` = 0, `cxy` need not be set.

    cyx : float, array-like, shape (nc)
        If `ic` != 0, `cyx` must contain the nc cross-covariances
        between values in the y series and later values in time in the x
        series, for lags from 0 to (nc-1).

        If `ic` = 0, `cyx` need not be set.

    kc : int
        If `ic` = 0, `kc` must specify the order of the fast Fourier
        transform (FFT) used to calculate the cross-covariances.

        If `ic` != 0, that is if covariances are supplied, `kc` is not
        used.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        Therefore it is also the order of the FFT used to construct the
        sample spectrum from the cross-covariances.

    xg : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ic`=0: max(`kc`, `l`); otherwise: `l`.

        If the cross-covariances are to be calculated, then `xg` must
        contain the `nxy` data points of the x series. If covariances
        are supplied, `xg` need not be set.

    yg : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ic`=0: max(`kc`, `l`); otherwise: `l`.

        If cross-covariances are to be calculated, `yg` must contain the
        `nxy` data points of the y series. If covariances are supplied,
        `yg` need not be set.

    Returns
    -------
    cxy : float, ndarray, shape (nc)
        If `ic` = 0, `cxy` will contain the nc calculated
        cross-covariances.

        If `ic` != 0, the contents of `cxy` will be unchanged.

    cyx : float, ndarray, shape (nc)
        If `ic` = 0, `cyx` will contain the nc calculated
        cross-covariances.

        If `ic` != 0, the contents of `cyx` will be unchanged.

    xg : float, ndarray, shape (:)
        Contains the real parts of the `ng` complex spectral estimates
        in elements `xg`[0] to `xg`[`ng`-1], and `xg`[`ng`] to
        `xg`[nxyg-1] contain 0.0. The y series leads the x series.

    yg : float, ndarray, shape (:)
        Contains the imaginary parts of the `ng` complex spectral
        estimates in elements `yg`[0] to `yg`[`ng`-1], and `yg`[`ng`] to
        `yg`[nxyg-1] contain 0.0. The y series leads the x series.

    ng : int
        The number, [`l`/2]+1, of complex spectral estimates, whose
        separate parts are held in `xg` and `yg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ic` = *<value>*, nxyg = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` != 0, nxyg >= `l`.

        (`errno` 1)
            On entry, nxyg = *<value>*, `kc` = *<value>* and `l` =
            *<value>*.

            Constraint: if `ic` = 0, nxyg >= max(`kc`, `l`).

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: if `ic` = 0, `pxy` <= 1.0.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: if `ic` = 0, `pxy` >= 0.0.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: if `ic` = 0 then `mtxy` != 0, 1 or 2.

        (`errno` 1)
            On entry, nc = *<value>* and `nxy` = *<value>*.

            Constraint: nc <= `nxy`.

        (`errno` 1)
            On entry, nc = *<value>*, `mw` = *<value>* and `ish` =
            *<value>*.

            Constraint: nc >= `mw`+abs(`ish`).

        (`errno` 1)
            On entry, `mw` = *<value>*, `ish` = *<value>* and `nxy` =
            *<value>*.

            Constraint: `mw`+abs(`ish`) <= `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `ish` = *<value>* and `mw` = *<value>*.

            Constraint: abs(`ish`) <= `mw`.

        (`errno` 1)
            On entry, `iw` = *<value>*.

            Constraint: `iw` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `nxy` = *<value>*.

            Constraint: `nxy` >= 1.

        (`errno` 2)
            On entry, `kc` = *<value>*, `nxy` = *<value>* and nc =
            *<value>*.

            Constraint: if `ic` = 0, `kc` >= `nxy`+nc.

        (`errno` 3)
            On entry, `l` = *<value>* and `mw` = *<value>*.

            Constraint: `l` >= 2*`mw`-1.
    """

    from ..base.tsa import multi_spectrum_lag as b_multi_spectrum_lag

    fun_name = 'naginterfaces.library.tsa.multi_spectrum_lag'

    _cxy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cxy'},
        cxy,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nc_py = _b_utils._EngineFloat64ArrayType.get_shape(_cxy, {'fun_name': fun_name, 'entity': 'cxy'}, exp_rank=1)[0]
    _nc = _b_utils._EngineIntScalarType(
        _nc_py, {'fun_name': fun_name, 'entity': '_nc'},
    ).to_ctypes()
    _cyx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cyx'},
        cyx,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxyg_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1, allow_none=True)[0]
    _nxyg = _b_utils._EngineIntScalarType(
        _nxyg_py, {'fun_name': fun_name, 'entity': '_nxyg'},
    ).to_ctypes()
    _ic = _b_utils._EngineIntScalarType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
    ).to_ctypes()
    _ic_py = _b_utils._EngineIntScalarType.to_py(_ic)
    _kc = _b_utils._EngineIntScalarType(
        kc, {'fun_name': fun_name, 'entity': 'kc'},
    ).to_ctypes()
    _kc_py = _b_utils._EngineIntScalarType.to_py(_kc)
    _l = _b_utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = _b_utils._EngineIntScalarType.to_py(_l)
    if xg is None:
        if _ic_py == 0:
            _xg_dim1 = max(_kc_py, _l_py)
        else:
            _xg_dim1 = _l_py
        exp_shape_xg = (_xg_dim1,)
        _xg = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_xg,
        )
    if yg is None:
        if _ic_py == 0:
            _yg_dim1 = max(_kc_py, _l_py)
        else:
            _yg_dim1 = _l_py
        exp_shape_yg = (_yg_dim1,)
        _yg = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_yg,
        )
    else:
        _yg = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'yg'},
            yg,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _return_scalar = b_multi_spectrum_lag(
        nxy,
        mtxy,
        pxy,
        iw,
        mw,
        ish,
        _ic,
        _nc,
        _cxy,
        _cyx,
        _kc,
        _l,
        _nxyg,
        _xg,
        _yg,
    )
    _cxy_py = _cxy
    _cyx_py = _cyx
    _xg_py = _xg
    _yg_py = _yg
    _ng_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['cxy', 'cyx', 'xg', 'yg', 'ng'],
    )(
        _cxy_py,
        _cyx_py,
        _xg_py,
        _yg_py,
        _ng_py,
    )
    return _return_data

def multi_spectrum_daniell(nxy, mtxy, pxy, mw, ish, pw, l, xg, yg):
    u"""
    Multivariate time series, smoothed sample cross spectrum using
    spectral smoothing by the trapezium frequency (Daniell) window.

    ``multi_spectrum_daniell`` calculates the smoothed sample cross
    spectrum of a bivariate time series using spectral smoothing by the
    trapezium frequency (Daniell) window.

    For full information please refer to the NAG Library document for
    g13cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cdf.html

    Parameters
    ----------
    nxy : int
        n, the length of the time series x and y.

    mtxy : int
        Whether the data is to be initially mean or trend corrected.

        `mtxy` = 0

          For no correction.

        `mtxy` = 1

          For mean correction.

        `mtxy` = 2

          For trend correction.

    pxy : float
        The proportion of the data (totalled over both ends) to be
        initially tapered by the split cosine bell taper.

        A value of 0.0 implies no tapering.

    mw : int
        M, the frequency width of the smoothing window as 2*pi/M.

        A value of n implies that no smoothing is to be carried out.

    ish : int
        S, the alignment shift between the x and y series. If x leads y,
        the shift is positive.

    pw : float
        p, the shape parameter of the trapezium frequency window.

        A value of 0.0 gives a triangular window, and a value of 1.0 a
        rectangular window.

        If `mw` = `nxy` then `pw` is not used.

    l : int
        L, the frequency division of smoothed cross spectral estimates
        as 2*pi/L.

    xg : float, array-like, shape (kc)
        The `nxy` data points of the x series.

    yg : float, array-like, shape (kc)
        The `nxy` data points of the y series.

    Returns
    -------
    xg : float, ndarray, shape (kc)
        The real parts of the `ng` cross spectral estimates in elements
        `xg`[0] to `xg`[`ng`-1], and `xg`[`ng`] to `xg`[kc-1] contain
        0.0. The y series leads the x series.

    yg : float, ndarray, shape (kc)
        The imaginary parts of the `ng` cross spectral estimates in
        elements `yg`[0] to `yg`[`ng`-1], and `yg`[`ng`] to `yg`[kc-1]
        contain 0.0. The y series leads the x series.

    ng : int
        The number of spectral estimates, [L/2]+1, whose separate parts
        are held in `xg` and `yg`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ish` = *<value>* and `l` = *<value>*.

            Constraint: abs(`ish`) < `l`.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, `pxy` = *<value>*, `mw` = *<value>* and `nxy` =
            *<value>*.

            Constraint: if `pw` > 1.0, `mw` = `nxy`.

        (`errno` 1)
            On entry, `pxy` = *<value>*, `mw` = *<value>* and `nxy` =
            *<value>*.

            Constraint: if `pw` < 0.0, `mw` = `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>* and `nxy` = *<value>*.

            Constraint: `mw` <= `nxy`.

        (`errno` 1)
            On entry, `mw` = *<value>*.

            Constraint: `mw` >= 1.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: `pxy` <= 1.0.

        (`errno` 1)
            On entry, `pxy` = *<value>*.

            Constraint: `pxy` >= 0.0.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: `mtxy` <= 2.

        (`errno` 1)
            On entry, `mtxy` = *<value>*.

            Constraint: `mtxy` >= 0.

        (`errno` 1)
            On entry, `nxy` = *<value>*.

            Constraint: `nxy` >= 1.

        (`errno` 2)
            On entry, kc = *<value>* and `l` = *<value>*.

            Constraint: kc must be a multiple of `l`.

        (`errno` 2)
            On entry, kc = *<value>* and `nxy` = *<value>*.

            Constraint: kc >= 2*`nxy`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import multi_spectrum_daniell as b_multi_spectrum_daniell

    fun_name = 'naginterfaces.library.tsa.multi_spectrum_daniell'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _kc_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _kc = _b_utils._EngineIntScalarType(
        _kc_py, {'fun_name': fun_name, 'entity': '_kc'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = b_multi_spectrum_daniell(
        nxy,
        mtxy,
        pxy,
        mw,
        ish,
        pw,
        l,
        _kc,
        _xg,
        _yg,
    )
    _xg_py = _xg
    _yg_py = _yg
    _ng_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xg', 'yg', 'ng'],
    )(
        _xg_py,
        _yg_py,
        _ng_py,
    )
    return _return_data

def multi_spectrum_bivar(xg, yg, xyrg, xyig, stats):
    u"""
    Multivariate time series, cross amplitude spectrum, squared
    coherency, bounds, univariate and bivariate (cross) spectra.

    For a bivariate time series, ``multi_spectrum_bivar`` calculates the
    cross amplitude spectrum and squared coherency, together with lower
    and upper bounds from the univariate and bivariate (cross) spectra.

    For full information please refer to the NAG Library document for
    g13ce

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cef.html

    Parameters
    ----------
    xg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, array-like, shape (ng)
        The real parts, cf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, array-like, shape (ng)
        The imaginary parts, qf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        Note: the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    stats : float, array-like, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the
        degrees of freedom, `stats`[1] and `stats`[2] contain the lower
        and upper bound multiplying factors respectively and `stats`[3]
        contains the bandwidth.

    Returns
    -------
    ca : float, ndarray, shape (ng)
        The ng cross amplitude spectral estimates A^(omega) at each
        frequency of omega.

    calw : float, ndarray, shape (ng)
        The ng lower bounds for the ng cross amplitude spectral
        estimates.

    caup : float, ndarray, shape (ng)
        The ng upper bounds for the ng cross amplitude spectral
        estimates.

    t : float
        The critical value for the significance of the squared
        coherency, T.

    sc : float, ndarray, shape (ng)
        The ng squared coherency estimates, W^(omega) at each frequency
        omega.

    sclw : float, ndarray, shape (ng)
        The ng lower bounds for the ng squared coherency estimates.

    scup : float, ndarray, shape (ng)
        The ng upper bounds for the ng squared coherency estimates.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stats`[2] = *<value>*.

            Constraint: `stats`[2] >= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] <= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] > 0.0.

        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

        (`errno` 4)
            A univariate spectral estimate is zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import multi_spectrum_bivar as b_multi_spectrum_bivar

    fun_name = 'naginterfaces.library.tsa.multi_spectrum_bivar'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyrg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyrg'},
        xyrg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyig'},
        xyig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _stats = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'stats'},
        stats,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ca = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _calw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _caup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _sclw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _scup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    try:
        _return_scalar = b_multi_spectrum_bivar(
            _xg,
            _yg,
            _xyrg,
            _xyig,
            _ng,
            _stats,
            _ca,
            _calw,
            _caup,
            _sc,
            _sclw,
            _scup,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ca_py = _ca
    _calw_py = _calw
    _caup_py = _caup
    _t_py = _return_scalar
    _sc_py = _sc
    _sclw_py = _sclw
    _scup_py = _scup
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ca', 'calw', 'caup', 't', 'sc', 'sclw', 'scup'],
    )(
        _ca_py,
        _calw_py,
        _caup_py,
        _t_py,
        _sc_py,
        _sclw_py,
        _scup_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_gain_bivar(xg, yg, xyrg, xyig, stats):
    u"""
    Multivariate time series, gain, phase, bounds, univariate and
    bivariate (cross) spectra.

    For a bivariate time series, ``multi_gain_bivar`` calculates the
    gain and phase together with lower and upper bounds from the
    univariate and bivariate spectra.

    For full information please refer to the NAG Library document for
    g13cf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cff.html

    Parameters
    ----------
    xg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, array-like, shape (ng)
        The real parts, cf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, array-like, shape (ng)
        The imaginary parts, qf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        Note: the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    stats : float, array-like, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the
        degrees of freedom, `stats`[1] and `stats`[2] contain the lower
        and upper bound multiplying factors respectively and `stats`[3]
        holds the bandwidth.

    Returns
    -------
    gn : float, ndarray, shape (ng)
        The ng gain estimates, G^(omega), at each frequency omega.

    gnlw : float, ndarray, shape (ng)
        The ng lower bounds for the ng gain estimates.

    gnup : float, ndarray, shape (ng)
        The ng upper bounds for the ng gain estimates.

    ph : float, ndarray, shape (ng)
        The ng phase estimates, phi^(omega), at each frequency omega.

    phlw : float, ndarray, shape (ng)
        The ng lower bounds for the ng phase estimates.

    phup : float, ndarray, shape (ng)
        The ng upper bounds for the ng phase estimates.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

        (`errno` 4)
            A univariate spectral estimate is zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import multi_gain_bivar as b_multi_gain_bivar

    fun_name = 'naginterfaces.library.tsa.multi_gain_bivar'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyrg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyrg'},
        xyrg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyig'},
        xyig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _stats = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'stats'},
        stats,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _gn = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _gnlw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _gnup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _ph = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _phlw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _phup = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    try:
        b_multi_gain_bivar(
            _xg,
            _yg,
            _xyrg,
            _xyig,
            _ng,
            _stats,
            _gn,
            _gnlw,
            _gnup,
            _ph,
            _phlw,
            _phup,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _gn_py = _gn
    _gnlw_py = _gnlw
    _gnup_py = _gnup
    _ph_py = _ph
    _phlw_py = _phlw
    _phup_py = _phup
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['gn', 'gnlw', 'gnup', 'ph', 'phlw', 'phup'],
    )(
        _gn_py,
        _gnlw_py,
        _gnup_py,
        _ph_py,
        _phlw_py,
        _phup_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_noise_bivar(xg, yg, xyrg, xyig, stats, l, n):
    u"""
    Multivariate time series, noise spectrum, bounds, impulse response
    function and its standard error.

    For a bivariate time series, ``multi_noise_bivar`` calculates the
    noise spectrum together with multiplying factors for the bounds and
    the impulse response function and its standard error, from the
    univariate and bivariate spectra.

    For full information please refer to the NAG Library document for
    g13cg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13cgf.html

    Parameters
    ----------
    xg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{xx}(omega), for the x
        series.

    yg : float, array-like, shape (ng)
        The ng univariate spectral estimates, f_{yy}(omega), for the y
        series.

    xyrg : float, array-like, shape (ng)
        The real parts, cf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

    xyig : float, array-like, shape (ng)
        The imaginary parts, qf(omega), of the ng bivariate spectral
        estimates for the x and y series. The x series leads the y
        series.

        Note: the two univariate and the bivariate spectra must each
        have been calculated using the same method of smoothing.

        For rectangular, Bartlett, Tukey or Parzen smoothing windows,
        the same cut-off point of lag window and the same frequency
        division of the spectral estimates must be used.

        For the trapezium frequency smoothing window, the frequency
        width and the shape of the window and the frequency division of
        the spectral estimates must be the same.

        The spectral estimates and statistics must also be unlogged.

    stats : float, array-like, shape (4)
        The four associated statistics for the univariate spectral
        estimates for the x and y series. `stats`[0] contains the degree
        of freedom, `stats`[1] and `stats`[2] contain the lower and
        upper bound multiplying factors respectively and `stats`[3]
        contains the bandwidth.

    l : int
        L, the frequency division of the spectral estimates as 2*pi/L.
        It is also the order of the FFT used to calculate the impulse
        response function. `l` must relate to the parameter ng by the
        relationship.

    n : int
        The number of points in each of the time series x and y. `n`
        should have the same value as nxy in the call of
        ``multi_spectrum_lag`` or ``multi_spectrum_daniell`` which
        calculated the smoothed sample cross spectrum. `n` is used in
        calculating the impulse response function standard error
        (`rfse`).

    Returns
    -------
    er : float, ndarray, shape (ng)
        The ng estimates of the noise spectrum, f^_{y | x}(omega) at
        each frequency.

    erlw : float
        The noise spectrum lower limit multiplying factor.

    erup : float
        The noise spectrum upper limit multiplying factor.

    rf : float, ndarray, shape (`l`)
        The impulse response function. Causal responses are stored in
        ascending frequency in `rf`[0] to `rf`[ng-1] and anticipatory
        responses are stored in descending frequency in `rf`[ng] to
        `rf`[`l`-1].

    rfse : float
        The impulse response function standard error.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `stats`[2] = *<value>*.

            Constraint: `stats`[2] >= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] <= 1.0.

        (`errno` 1)
            On entry, `stats`[1] = *<value>*.

            Constraint: `stats`[1] > 0.0.

        (`errno` 1)
            On entry, `stats`[0] = *<value>*.

            Constraint: `stats`[0] >= 3.0.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 1.

        (`errno` 6)
            On entry, ng = *<value>* and `l` = *<value>*.

            Constraint: ng = `l`/2+1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A bivariate spectral estimate is zero.

            For this frequency the noise spectrum is set to zero, and
            the contribution to the impulse response function and its
            standard error is set to zero.

        (`errno` 3)
            A univariate spectral estimate is negative.

            For this frequency the noise spectrum is set to zero, and
            the contributions to the impulse response function and its
            standard error are set to zero.

        (`errno` 4)
            A univariate spectral estimate is zero.

            For this frequency the noise spectrum is set to zero and the
            contributions to the impulse response function and its
            standard error are set to zero.

        (`errno` 5)
            A calculated value of the squared coherency exceeds 1.0.

            For this frequency the squared coherency is reset to 1.0
            with the consequence that the noise spectrum is zero and the
            contribution to the impulse response function at this
            frequency is zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import multi_noise_bivar as b_multi_noise_bivar

    fun_name = 'naginterfaces.library.tsa.multi_noise_bivar'

    _xg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xg'},
        xg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng_py = _b_utils._EngineFloat64ArrayType.get_shape(_xg, {'fun_name': fun_name, 'entity': 'xg'}, exp_rank=1)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _yg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yg'},
        yg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyrg = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyrg'},
        xyrg,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xyig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xyig'},
        xyig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _stats = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'stats'},
        stats,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _er = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _l = _b_utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = _b_utils._EngineIntScalarType.to_py(_l)
    _rf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_l_py,),
    )
    try:
        _return_scalars = b_multi_noise_bivar(
            _xg,
            _yg,
            _xyrg,
            _xyig,
            _ng,
            _stats,
            _l,
            n,
            _er,
            _rf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _er_py = _er
    _erlw_py = _return_scalars.erlw
    _erup_py = _return_scalars.erup
    _rf_py = _rf
    _rfse_py = _return_scalars.rfse
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['er', 'erlw', 'erup', 'rf', 'rfse'],
    )(
        _er_py,
        _erlw_py,
        _erup_py,
        _rf_py,
        _rfse_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_autocorr_part(c0, c, nk):
    u"""
    Multivariate time series, multiple squared partial autocorrelations.

    ``multi_autocorr_part`` calculates the multivariate partial
    autocorrelation function of a multivariate time series.

    For full information please refer to the NAG Library document for
    g13db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dbf.html

    Parameters
    ----------
    c0 : float, array-like, shape (ns, ns)
        Contains the zero lag cross-covariances between the ns series as
        returned by ``multi_corrmat_cross``. (`c0` is assumed to be
        symmetric, upper triangle only is used.)

    c : float, array-like, shape (ns, ns, nl)
        Contains the cross-covariances at lags 1 to nl. `c`[i-1,j-1,k-1]
        must contain the cross-covariance, c_{ijk}, of series i and
        series j at lag k. Series j leads series i.

    nk : int
        The number of lags to which partial auto-correlations are to be
        calculated.

    Returns
    -------
    p : float, ndarray, shape (`nvp`)
        The multiple squared partial autocorrelations from lags 1 to
        `nvp`; that is, `p`[l-1] contains p_l^2, for l = 1,2,...,`nvp`.
        For lags `nvp`+1 to `nk` the elements of `p` are set to zero.

    v0 : float
        The lag zero prediction error variance (equal to the determinant
        of `c0`).

    v : float, ndarray, shape (`nvp`)
        The prediction error variance ratios from lags 1 to `nvp`; that
        is, `v`[l-1] contains v_l, for l = 1,2,...,`nvp`. For lags
        `nvp`+1 to `nk` the elements of `v` are set to zero.

    d : float, ndarray, shape (ns, ns, `nk`)
        The prediction error variance matrices at lags 1 to `nvp`.

        Element (i,j,k) of `d` contains the prediction error covariance
        of series i and series j at lag k, for k = 1,2,...,`nvp`.

        Series j leads series i; that is, the (i,j)th element of D_k.

        For lags `nvp`+1 to `nk` the elements of `d` are set to zero.

    db : float, ndarray, shape (ns, ns)
        The backward prediction error variance matrix at lag `nvp`.

        `db`[i-1,j-1] contains the backward prediction error covariance
        of series i and series j; that is, the (i,j)th element of the
        G_k, where k = `nvp`.

    w : float, ndarray, shape (ns, ns, `nk`)
        The prediction coefficient matrices at lags 1 to `nvp`.

        `w`[i-1,j-1,l-1] contains the jth prediction coefficient of
        series i at lag l; that is, the (i,j)th element of Phi_{kl},
        where k = `nvp`, for l = 1,2,...,`nvp`.

        For lags `nvp`+1 to `nk` the elements of `w` are set to zero.

    wb : float, ndarray, shape (ns, ns, `nk`)
        The backward prediction coefficient matrices at lags 1 to `nvp`.

        `wb`[i-1,j-1,l-1] contains the jth backward prediction
        coefficient of series i at lag l; that is, the (i,j)th element
        of Psi_{kl}, where k = `nvp`, for l = 1,2,...,`nvp`.

        For lags `nvp`+1 to `nk` the elements of `wb` are set to zero.

    nvp : int
        The maximum lag, L, for which calculation of `p`, `v`, `d`,
        `db`, `w` and `wb` was successful. If the function completes
        successfully `nvp` will equal `nk`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nk` = *<value>* and nl = *<value>*.

            Constraint: `nk` <= nl.

        (`errno` 1)
            On entry, `nk` = *<value>*.

            Constraint: `nk` >= 1.

        (`errno` 1)
            On entry, nl = *<value>*.

            Constraint: nl >= 1.

        (`errno` 1)
            On entry, ns = *<value>* and ldc0 = *<value>*.

            Constraint: ns <= ldc0.

        (`errno` 1)
            On entry, ns = *<value>*.

            Constraint: ns >= 1.

        (`errno` 2)
            `c0` is not positive definite.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            For `nvp` = *<value>*, at lag k = `nvp`+1, D_k was found not
            to be positive definite.
    """

    from ..base.tsa import multi_autocorr_part as b_multi_autocorr_part

    fun_name = 'naginterfaces.library.tsa.multi_autocorr_part'

    _c0 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c0'},
        c0,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_c0, 'c0'), (_c, 'c')),
    )
    _ns_py = _b_utils._EngineFloat64ArrayType.get_shape(_c0, {'fun_name': fun_name, 'entity': 'c0'}, exp_rank=2)[0]
    _ns = _b_utils._EngineIntScalarType(
        _ns_py, {'fun_name': fun_name, 'entity': '_ns'},
    ).to_ctypes()
    _nl_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=3)[2]
    _nl = _b_utils._EngineIntScalarType(
        _nl_py, {'fun_name': fun_name, 'entity': '_nl'},
    ).to_ctypes()
    _nk = _b_utils._EngineIntScalarType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
    ).to_ctypes()
    _nk_py = _b_utils._EngineIntScalarType.to_py(_nk)
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nk_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nk_py,),
    )
    _d = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py, _nk_py),
        sorder=_m_sorder,
    )
    _db = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py),
        sorder=_m_sorder,
    )
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py, _nk_py),
        sorder=_m_sorder,
    )
    _wb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py, _ns_py, _nk_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalars = b_multi_autocorr_part(
            _c0,
            _c,
            _ns,
            _nl,
            _nk,
            _p,
            _v,
            _d,
            _db,
            _w,
            _wb,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _nvp_py = _return_scalars.nvp
    _p_py = _p[:_nvp_py]
    _v0_py = _return_scalars.v0
    _v_py = _v[:_nvp_py]
    _d_py = _d[:, :, :_nvp_py]
    _db_py = _db
    _w_py = _w[:, :, :_nvp_py]
    _wb_py = _wb[:, :, :_nvp_py]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['p', 'v0', 'v', 'd', 'db', 'w', 'wb', 'nvp'],
    )(
        _p_py,
        _v0_py,
        _v_py,
        _d_py,
        _db_py,
        _w_py,
        _wb_py,
        _nvp_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_varma_estimate(ip, iq, mean, par, qq, w, parhld, exact, iprint, cgetol, ishow, maxcal=None, io_manager=None):
    u"""
    Multivariate time series, estimation of VARMA model.

    ``multi_varma_estimate`` fits a vector autoregressive moving average
    (VARMA) model to an observed vector of time series using the method
    of Maximum Likelihood (ML).
    Standard errors of parameter estimates are computed along with their
    appropriate correlation matrix.
    The function also calculates estimates of the residual series.

    For full information please refer to the NAG Library document for
    g13dd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ddf.html

    Parameters
    ----------
    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    mean : bool
        `mean` = True, if components of mu have been estimated and
        `mean` = False, if all elements of mu are to be taken as zero.

    par : float, array-like, shape (npar)
        Initial parameter estimates read in row by row in the order
        phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q,mu.

        Thus,

        - if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to
          an initial estimate of the (i,j)th element of phi_l, for j =
          1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,p;

        - if `iq` > 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
          equal to an initial estimate of the (i,j)th element of
          theta_l, l = 1,2,...,q and i,j = 1,2,...,k;

        - if `mean` = True, `par`[(p+q)*k*k+i-1] should be set equal to
          an initial estimate of the ith component of mu (mu(i)). (If
          you set `par`[(p+q)*k*k+i-1] to 0.0 then
          ``multi_varma_estimate`` will calculate the mean of the ith
          series and use this as an initial estimate of mu(i).)

        The first p*k*k elements of `par` must satisfy the stationarity
        condition and the next q*k*k elements of `par` must satisfy the
        invertibility condition.

        If in doubt set all elements of `par` to 0.0.

    qq : float, array-like, shape (k, k)
        `qq`[i-1,j-1] must be set equal to an initial estimate of the
        (i,j)th element of Sigma. The lower triangle only is needed.
        `qq` must be positive definite. It is strongly recommended that
        on entry the elements of `qq` are of the same order of magnitude
        as at the solution point. If you set `qq`[i-1,j-1] = 0.0, for j
        = 1,2,...,i, for i = 1,2,...,k, then ``multi_varma_estimate``
        will calculate the covariance matrix between the k time series
        and use this as an initial estimate of Sigma.

    w : float, array-like, shape (k, n)
        `w`[i-1,t-1] must be set equal to the ith component of W_t, for
        t = 1,2,...,n, for i = 1,2,...,k.

    parhld : bool, array-like, shape (npar)
        `parhld`[i-1] must be set to True if `par`[i-1] is to be held
        constant at its input value and False if `par`[i-1] is a free
        parameter, for i = 1,2,...,npar.

        If in doubt try setting all elements of `parhld` to False.

    exact : bool
        Must be set equal to True if you wish ``multi_varma_estimate``
        to compute exact maximum likelihood estimates. `exact` must be
        set equal to False if only conditional likelihood estimates are
        required.

    iprint : int
        The frequency with which the automatic monitoring function is to
        be called.

        `iprint` > 0

          The ML search procedure is monitored once every `iprint`
          iterations and just before exit from the search function.

        `iprint` = 0

          The search function is monitored once at the final point.

        `iprint` < 0

          The search function is not monitored at all.

    cgetol : float
        The accuracy to which the solution in `par` and `qq` is
        required.

        If `cgetol` is set to 10^(-l) and on exit no exception or
        warning is raised or `errno` >= 6, then all the elements in
        `par` and `qq` should be accurate to approximately l decimal
        places.

        For most practical purposes the value 10^(-4) should suffice.

        You should be wary of setting `cgetol` too small since the
        convergence criteria may then have become too strict for the
        machine to handle.

        If `cgetol` has been set to a value which is less than the
        machine precision, epsilon, then ``multi_varma_estimate`` will
        use the value 10.0*sqrt(epsilon) instead.

    ishow : int
        Specifies which of the following two quantities are to be
        printed.

        - table of maximum likelihood estimates and their standard
          errors (as returned in the output arrays `par`, `qq` and
          `cm`);

        - table of residual series (as returned in the output array
          `v`).

        `ishow` = 0

          None of the above are printed.

        `ishow` = 1

          (i) only is printed.

        `ishow` = 2

          are printed.

    maxcal : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 40*npar*(npar+5).

        The maximum number of likelihood evaluations to be permitted by
        the search procedure.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    par : float, ndarray, shape (npar)
        If no exception or warning is raised or `errno` >= 4 then all
        the elements of `par` will be overwritten by the latest
        estimates of the corresponding ARMA parameters.

    qq : float, ndarray, shape (k, k)
        If no exception or warning is raised or `errno` >= 4 then
        `qq`[i-1,j-1] will contain the latest estimate of the (i,j)th
        element of Sigma. The lower triangle only is returned.

    niter : int
        If no exception or warning is raised or `errno` >= 4 then
        `niter` contains the number of iterations performed by the
        search function.

    rlogl : float
        If no exception or warning is raised or `errno` >= 4 then
        `rlogl` contains the value of the log-likelihood function
        corresponding to the final point held in `par` and `qq`.

    v : float, ndarray, shape (k, n)
        If no exception or warning is raised or `errno` >= 4 then
        `v`[i-1,t-1] will contain an estimate of the ith component of
        epsilon_t, for t = 1,2,...,n, for i = 1,2,...,k, corresponding
        to the final point held in `par` and `qq`.

    g : float, ndarray, shape (npar)
        If no exception or warning is raised or `errno` >= 4 then
        `g`[i-1] will contain the estimated first derivative of the
        log-likelihood function with respect to the ith element in the
        array `par`. If the gradient cannot be computed then all the
        elements of `g` are returned as zero.

    cm : float, ndarray, shape (npar, npar)
        If no exception or warning is raised or `errno` >= 4 then
        `cm`[i-1,j-1] will contain an estimate of the correlation
        coefficient between the ith and jth elements in the `par` array
        for 1 <= i <= npar, 1 <= j <= npar. If i = j, then `cm`[i-1,j-1]
        will contain the estimated standard error of `par`[i-1]. If the
        lth component of `par` has been held constant, i.e.,
        `parhld`[l-1] was set to True, then the lth row and column of
        `cm` will be set to zero. If the second derivative matrix cannot
        be computed then all the elements of `cm` are returned as zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar >= 0.

        (`errno` 1)
            On entry, n = *<value>*, k = *<value>* and npar = *<value>*.

            Constraint: n*k > npar+k*(k+1)/2.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = 0 and `iq` = 0.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar = *<value>*.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `ishow` = *<value>*.

            Constraint: 0 <= `ishow` <= 2.

        (`errno` 2)
            The initial estimate of Sigma is not positive definite.

        (`errno` 2)
            The initial AR parameter estimates are outside the
            stationarity region.

        (`errno` 2)
            The initial MA parameter estimates are outside the
            invertibility region.

        (`errno` 2)
            The starting point is too close to the boundary of the
            admissibility region.

        (`errno` 3)
            The function cannot compute a sufficiently accurate estimate
            of the gradient vector at the user-supplied starting point.

        (`errno` 4)
            There have been `maxcal` log-likelihood evaluations made in
            the function.

            See the function document for further details.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The conditions for a solution have not all been met, but a
            point at which the log-likelihood took a larger value could
            not be found.

            See the function document for further details.

        (`errno` 6)
            The ML solution is so close to the boundary of either the
            stationarity region or the invertibility region that
            ``multi_varma_estimate`` cannot evaluate the Hessian matrix.
            The elements of `cm` are set to zero, as are the elements of
            `g`. All other output quantities are correct.

        (`errno` 7)
            An estimate of the second derivative matrix and the gradient
            vector at the solution point was computed. Either the
            Hessian matrix was found to be too ill-conditioned to be
            evaluated accurately or the gradient vector could not be
            computed to an acceptable degree of accuracy. The elements
            of `cm` are set to zero, as are the elements of `g`. All
            other output quantities are correct.

        (`errno` 8)
            The second-derivative matrix at the solution point is not
            positive definite. The elements of `cm` are set to zero. All
            other output quantities are correct.
    """

    from ..base.tsa import multi_varma_estimate as b_multi_varma_estimate

    fun_name = 'naginterfaces.library.tsa.multi_varma_estimate'

    _qq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qq'},
        qq,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_qq, 'qq'), (_w, 'w')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_qq, {'fun_name': fun_name, 'entity': 'qq'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _parhld = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parhld'},
        parhld,
        copy=False,
    )
    _maxcal = 40*_npar_py*(_npar_py+5) if maxcal is None else maxcal
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _n_py),
        sorder=_m_sorder,
    )
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _cm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalars = b_multi_varma_estimate(
            _k,
            _n,
            ip,
            iq,
            mean,
            _par,
            _npar,
            _qq,
            _w,
            _parhld,
            exact,
            iprint,
            cgetol,
            _maxcal,
            ishow,
            _v,
            _g,
            _cm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _par_py = _par
    _qq_py = _qq
    _niter_py = _return_scalars.niter
    _rlogl_py = _return_scalars.rlogl
    _v_py = _v
    _g_py = _g
    _cm_py = _cm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['par', 'qq', 'niter', 'rlogl', 'v', 'g', 'cm'],
    )(
        _par_py,
        _qq_py,
        _niter_py,
        _rlogl_py,
        _v_py,
        _g_py,
        _cm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_varma_forecast(z, tr, dord, delta, ip, iq, mean, par, qq, v, lmax, lref):
    u"""
    Multivariate time series, forecasts and their standard errors.

    ``multi_varma_forecast`` computes forecasts of a multivariate time
    series.
    It is assumed that a vector ARMA model has already been fitted to
    the appropriately differenced/transformed time series using
    ``multi_varma_estimate``.
    The standard deviations of the forecast errors are also returned.
    A reference vector is set up so that, should future series values
    become available, the forecasts and their standard errors may be
    updated by calling ``multi_varma_update``.

    For full information please refer to the NAG Library document for
    g13dj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13djf.html

    Parameters
    ----------
    z : float, array-like, shape (k, n)
        `z`[i,t] must contain, z_{it}, the ith component of Z_t, for t =
        1,2,...,n, for i = 1,2,...,k.

    tr : str, length 1, array-like, shape (k)
        `tr`[i-1] indicates whether the ith time series is to be
        transformed, for i = 1,2,...,k.

        `tr`[i-1] = 'N'

          No transformation is used.

        `tr`[i-1] = 'L'

          A log transformation is used.

        `tr`[i-1] = 'S'

          A square root transformation is used.

    dord : int, array-like, shape (k)
        `dord`[i-1] must specify, d_i, the order of differencing
        required for the ith series.

    delta : float, array-like, shape (k, d)
        If `dord`[i-1] > 0, then `delta`[i-1,j-1] must be set equal to
        delta_{ij}, for i = 1,2,...,k, for j = 1,2,...,d_i.

        If d = 0, `delta` is not referenced.

    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    mean : str, length 1
        `mean` = 'M', if components of mu have been estimated and `mean`
        = 'Z', if all elements of mu are to be taken as zero.

    par : float, array-like, shape (lpar)
        Must contain the parameter estimates read in row by row in the
        order phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q, mu.

        Thus,

        - if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to
          an estimate of the (i,j)th element of phi_l, for j =
          1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,p;

        - if `iq` > 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
          equal to an estimate of the (i,j)th element of theta_l, for j
          = 1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,q;

        - if `mean` = 'M', `par`[(p+q)*k*k+i-1] must be set equal to an
          estimate of the ith component of mu, for i = 1,2,...,k.

    qq : float, array-like, shape (k, k)
        `qq`[i-1,j-1] must contain an estimate of the (i,j)th element of
        Sigma. The lower triangle only is needed.

    v : float, array-like, shape (k, n-d)
        `v`[i-1,t-1] must contain an estimate of the ith component of
        epsilon_{t+d}, for t = 1,2,...,n-d, for i = 1,2,...,k.

        If q = 0, `v` is not used.

    lmax : int
        The number, l_max, of forecasts required.

    lref : int
        The dimension of the array `ref`.

    Returns
    -------
    qq : float, ndarray, shape (k, k)
        If `errno` != 1, then the upper triangle is set equal to the
        lower triangle.

    predz : float, ndarray, shape (k, `lmax`)
        `predz`[i-1,l-1] contains the forecast of z_{i,n+l}, for l =
        1,2,...,l_max, for i = 1,2,...,k.

    sefz : float, ndarray, shape (k, `lmax`)
        `sefz`[i-1,l-1] contains an estimate of the standard error of
        the forecast of z_{i,n+l}, for l = 1,2,...,l_max, for i =
        1,2,...,k.

    ref : float, ndarray, shape (`lref`)
        The reference vector which may be used to update forecasts using
        ``multi_varma_update``. The first (`lmax`-1)*k*k elements
        contain the psi weight matrices, psi_1,psi_2,...,psi_{l_max-1}.
        The next k*`lmax` elements contain the forecasts of the
        transformed series Z^_n+1^*,Z^_n+2^*,...,Z^_n+l_max^* and the
        next k*`lmax` contain the variances of the forecasts of the
        transformed variables. The last k elements are used to store the
        transformations for the series.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lref` = *<value>* and the minimum size required =
            *<value>*.

            Constraint: `lref` >= (`lmax`-1)*k*k+2*k*`lmax`+k.

        (`errno` 1)
            On entry, i = *<value>*, `dord`[i-1] = *<value>* and
            n-max(`ip`, `iq`) = *<value>*.

            Constraint: 0 <= `dord`[i-1] < n-max(`ip`, `iq`).

        (`errno` 1)
            On entry, `lmax` = *<value>*.

            Constraint: `lmax` >= 1.

        (`errno` 1)
            On entry, number of observations = *<value>* and number of
            parameters = *<value>*.

            Constraint: number of observations >= number of parameters.

        (`errno` 1)
            On entry, lpar is too small: lpar = *<value>* and the
            minimum size required = *<value>*.

        (`errno` 1)
            On entry, `mean` is an invalid character.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `tr`[i-1] is invalid.

            Constraint: `tr`[i-1] = 'N', 'L' or 'S'.

        (`errno` 3)
            On entry, one (or more) of the transformations requested is
            invalid.

        (`errno` 4)
            On entry, the MA parameter matrices are outside the
            invertibility region.

        (`errno` 4)
            On entry, the AR parameter matrices are outside the
            stationarity region.

        (`errno` 4)
            On entry, the covariance matrix `qq` is not positive
            definite.

        (`errno` 5)
            An excessive number of iterations were needed to evaluate
            the eigenvalues of the matrices used to test for
            stationarity and invertibility.

        (`errno` 6)
            The covariance matrix may be nearly non-positive definite.

        (`errno` 7)
            The forecasts will overflow if computed.
    """

    from ..base.tsa import multi_varma_forecast as b_multi_varma_forecast

    fun_name = 'naginterfaces.library.tsa.multi_varma_forecast'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _delta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'delta'},
        delta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _qq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qq'},
        qq,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_delta, 'delta'), (_qq, 'qq'), (_v, 'v')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _tr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tr'},
        tr,
        copy=False,
    )
    _dord = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dord'},
        dord,
        copy=False,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lpar_py = _b_utils._EngineFloat64ArrayType.get_shape(_par, {'fun_name': fun_name, 'entity': 'par'}, exp_rank=1)[0]
    _lpar = _b_utils._EngineIntScalarType(
        _lpar_py, {'fun_name': fun_name, 'entity': '_lpar'},
    ).to_ctypes()
    _lmax = _b_utils._EngineIntScalarType(
        lmax, {'fun_name': fun_name, 'entity': 'lmax'},
    ).to_ctypes()
    _lmax_py = _b_utils._EngineIntScalarType.to_py(_lmax)
    _predz = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _lmax_py),
        sorder=_m_sorder,
    )
    _sefz = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _lmax_py),
        sorder=_m_sorder,
    )
    _lref = _b_utils._EngineIntScalarType(
        lref, {'fun_name': fun_name, 'entity': 'lref'},
    ).to_ctypes()
    _lref_py = _b_utils._EngineIntScalarType.to_py(_lref)
    _ref = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lref_py,),
    )
    b_multi_varma_forecast(
        _k,
        _n,
        _z,
        _tr,
        _dord,
        _delta,
        ip,
        iq,
        mean,
        _par,
        _lpar,
        _qq,
        _v,
        _lmax,
        _predz,
        _sefz,
        _ref,
        _lref,
    )
    _qq_py = _qq
    _predz_py = _predz
    _sefz_py = _sefz
    _ref_py = _ref
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['qq', 'predz', 'sefz', 'ref'],
    )(
        _qq_py,
        _predz_py,
        _sefz_py,
        _ref_py,
    )
    return _return_data

def multi_varma_update(mlast, z, ref, predz, sefz):
    u"""
    Multivariate time series, updates forecasts and their standard
    errors.

    ``multi_varma_update`` accepts a sequence of new observations in a
    multivariate time series and updates both the forecasts and the
    standard deviations of the forecast errors.
    A call to ``multi_varma_forecast`` must be made prior to calling
    this function in order to calculate the elements of a reference
    vector together with a set of forecasts and their standard errors.
    On a successful exit from ``multi_varma_update`` the reference
    vector is updated so that should future series values become
    available these forecasts may be updated by recalling
    ``multi_varma_update``.

    For full information please refer to the NAG Library document for
    g13dk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dkf.html

    Parameters
    ----------
    mlast : int
        On the first call to ``multi_varma_update``, since calling
        ``multi_varma_forecast``, `mlast` must be set to 0 to indicate
        that no new observations have yet been used to update the
        forecasts; on subsequent calls `mlast` must contain the value of
        `mlast` as output on the previous call to
        ``multi_varma_update``.

    z : float, array-like, shape (k, m)
        `z`[i-1,j-1] must contain the value of z_{i,n+`mlast`+j}, for j
        = 1,2,...,m, for i = 1,2,...,k, and where n is the number of
        observations in the time series in the last call made to
        ``multi_varma_forecast``.

    ref : float, array-like, shape (lref)
        Must contain the first (lmax-1)*k*k+2*k*lmax+k elements of the
        reference vector as returned on a successful exit from
        ``multi_varma_forecast`` (or a previous call to
        ``multi_varma_update``).

    predz : float, array-like, shape (k, lmax)
        Nonupdated values are kept intact.

    sefz : float, array-like, shape (k, lmax)
        Nonupdated values are kept intact.

    Returns
    -------
    mlast : int
        Is incremented by m to indicate that `mlast`+m observations have
        now been used to update the forecasts since the last call to
        ``multi_varma_forecast``.

        `mlast` must not be changed between calls to
        ``multi_varma_update``, unless a call to
        ``multi_varma_forecast`` has been made between the calls in
        which case `mlast` should be reset to 0.

    ref : float, ndarray, shape (lref)
        The elements of `ref` are updated. The first (lmax-1)*k*k
        elements store the psi weights psi_1,psi_2,...,psi_{l_max-1}.
        The next k*lmax elements contain the forecasts of the
        transformed series and the next k*lmax elements contain the
        variances of the forecasts of the transformed variables; see
        ``multi_varma_forecast``. The last k elements are not updated.

    v : float, ndarray, shape (k, m)
        `v`[i-1,j-1] contains an estimate of the ith component of
        epsilon_{n+`mlast`+j}, for j = 1,2,...,m, for i = 1,2,...,k.

    predz : float, ndarray, shape (k, lmax)
        `predz`[i-1,j-1] contains the updated forecast of z_{i,n+j}, for
        j = `mlast`+m+1,...,l_max, for i = 1,2,...,k.

        The columns of `predz` corresponding to the new observations
        since the last call to either ``multi_varma_forecast`` or
        ``multi_varma_update`` are set equal to the corresponding
        columns of `z`.

    sefz : float, ndarray, shape (k, lmax)
        `sefz`[i-1,j-1] contains an estimate of the standard error of
        the corresponding element of `predz`, for j =
        `mlast`+m+1,...,l_max, for i = 1,2,...,k.

        The columns of `sefz` corresponding to the new observations
        since the last call to either ``multi_varma_forecast`` or
        ``multi_varma_update`` are set equal to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lref = *<value>* and the minimum size required =
            *<value>*.

            Constraint: lref >= (lmax-1)*k*k+2*k*lmax+k.

        (`errno` 1)
            On entry, `mlast` = *<value>*.

            Constraint: `mlast` >= 0.

        (`errno` 1)
            On entry, m = *<value>*, lmax = *<value>* and `mlast` =
            *<value>*.

            Constraint: m < lmax-`mlast`.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 1)
            On entry, lmax = *<value>*.

            Constraint: lmax >= 2.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, some of the elements of the array `ref` have been
            corrupted.

        (`errno` 3)
            On entry, one (or more) of the transformations requested is
            invalid.

        (`errno` 4)
            The updated forecasts will overflow if computed.
    """

    from ..base.tsa import multi_varma_update as b_multi_varma_update

    fun_name = 'naginterfaces.library.tsa.multi_varma_update'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _predz = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'predz'},
        predz,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sefz = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sefz'},
        sefz,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_predz, 'predz'), (_sefz, 'sefz')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _lmax_py = _b_utils._EngineFloat64ArrayType.get_shape(_predz, {'fun_name': fun_name, 'entity': 'predz'}, exp_rank=2)[1]
    _lmax = _b_utils._EngineIntScalarType(
        _lmax_py, {'fun_name': fun_name, 'entity': '_lmax'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _ref = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ref'},
        ref,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lref_py = _b_utils._EngineFloat64ArrayType.get_shape(_ref, {'fun_name': fun_name, 'entity': 'ref'}, exp_rank=1)[0]
    _lref = _b_utils._EngineIntScalarType(
        _lref_py, {'fun_name': fun_name, 'entity': '_lref'},
    ).to_ctypes()
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _m_py),
        sorder=_m_sorder,
    )
    _return_scalar = b_multi_varma_update(
        _k,
        _lmax,
        _m,
        mlast,
        _z,
        _ref,
        _lref,
        _v,
        _predz,
        _sefz,
    )
    _mlast_py = _return_scalar
    _ref_py = _ref
    _predz_py = _predz
    _sefz_py = _sefz
    _v_py = _v
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['mlast', 'ref', 'v', 'predz', 'sefz'],
    )(
        _mlast_py,
        _ref_py,
        _v_py,
        _predz_py,
        _sefz_py,
    )
    return _return_data

def multi_diff(z, tr, dord, delta):
    u"""
    Multivariate time series, differences and/or transforms.

    ``multi_diff`` differences and/or transforms a multivariate time
    series. It is intended to be used prior to ``multi_varma_estimate``
    to fit a vector autoregressive moving average (VARMA) model to the
    differenced/transformed series.

    For full information please refer to the NAG Library document for
    g13dl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dlf.html

    Parameters
    ----------
    z : float, array-like, shape (k, n)
        `z`[i,t] must contain, z_{it}, the ith component of Z_t, for t =
        1,2,...,n, for i = 1,2,...,k.

    tr : str, length 1, array-like, shape (k)
        `tr`[i-1] indicates whether the ith time series is to be
        transformed, for i = 1,2,...,k.

        `tr`[i-1] = 'N'

          No transformation is used.

        `tr`[i-1] = 'L'

          A log transformation is used.

        `tr`[i-1] = 'S'

          A square root transformation is used.

    dord : int, array-like, shape (k)
        The order of differencing for each series, d_1,d_2,...,d_k.

    delta : float, array-like, shape (k, d)
        If `dord`[i-1] > 0, then `delta`[i-1,j-1] must be set equal to
        delta_{ij}, for i = 1,2,...,k, for j = 1,2,...,d_i.

        If d = 0, `delta` is not referenced.

    Returns
    -------
    w : float, ndarray, shape (k, `nd`)
        `w`[i-1,t-1] contains the value of w_{i,t+d}, for t =
        1,2,...,n-d, for i = 1,2,...,k.

    nd : int
        The number of differenced values, n-d, in the series, where d =
        max(`dord`[i-1]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, i = *<value>*, `dord`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: `dord` < n.

        (`errno` 2)
            On entry, i = *<value>* and `dord`[i-1] = *<value>*.

            Constraint: `dord`[i-1] >= 0.

        (`errno` 3)
            On entry, i = *<value>* and `tr`[i-1] is invalid.

            Constraint: `tr`[i-1] = 'N', 'L' or 'S'.

        (`errno` 4)
            On entry, one (or more) of the transformations requested is
            invalid.
    """

    from ..base.tsa import multi_diff as b_multi_diff

    fun_name = 'naginterfaces.library.tsa.multi_diff'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _delta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'delta'},
        delta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'), (_delta, 'delta')),
    )
    _tr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tr'},
        tr,
        copy=False,
    )
    _k_py = _b_utils._EngineCharArrayType.get_shape(_tr, {'fun_name': fun_name, 'entity': 'tr'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _dord = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dord'},
        dord,
        copy=False,
    )
    _vl_d_py = max(_dord[0:_k_py-1+1])
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _n_py-_vl_d_py),
        sorder=_m_sorder,
    )
    _return_scalar = b_multi_diff(
        _k,
        _n,
        _z,
        _tr,
        _dord,
        _delta,
        _w,
    )
    _nd_py = _return_scalar
    _w_py = _w[:, :_nd_py]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['w', 'nd'],
    )(
        _w_py,
        _nd_py,
    )
    return _return_data

def multi_corrmat_cross(matrix, m, w):
    u"""
    Multivariate time series, sample cross-correlation or
    cross-covariance matrices.

    ``multi_corrmat_cross`` calculates the sample cross-correlation (or
    cross-covariance) matrices of a multivariate time series.

    For full information please refer to the NAG Library document for
    g13dm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dmf.html

    Parameters
    ----------
    matrix : str, length 1
        Indicates whether the cross-covariance or cross-correlation
        matrices are to be computed.

        `matrix` = 'V'

          The cross-covariance matrices are computed.

        `matrix` = 'R'

          The cross-correlation matrices are computed.

    m : int
        m, the number of cross-correlation (or cross-covariance)
        matrices to be computed. If in doubt set `m` = 10. However it
        should be noted that `m` is usually taken to be at most n/4.

    w : float, array-like, shape (k, n)
        `w`[i-1,t-1] must contain the observation w_{it}, for t =
        1,2,...,n, for i = 1,2,...,k.

    Returns
    -------
    wmean : float, ndarray, shape (k)
        The means, w__i, for i = 1,2,...,k.

    r0 : float, ndarray, shape (k, k)
        If i != j, then `r0`[i-1,j-1] contains an estimate of the
        (i,j)th element of the cross-correlation (or cross-covariance)
        matrix at lag zero, R^_{ij}(0); if i = j, then if `matrix` =
        'V', `r0`[i-1,i-1] contains the variance of the ith series,
        C^_{ii}(0), and if `matrix` = 'R', `r0`[i-1,i-1] contains the
        standard deviation of the ith series, sqrt(C^_{ii}(0)).

        If `errno` = 2 and `matrix` = 'R', then on exit all the elements
        in `r0` whose computation involves the zero variance are set to
        zero.

    r : float, ndarray, shape (k, k, `m`)
        `r`[i-1,j-1,l-1] contains an estimate of the (i,j)th element of
        the cross-correlation (or cross-covariance) at lag l,
        R^_{ij}(l), for j = 1,2,...,k, for i = 1,2,...,k, for l =
        1,2,...,m.

        If `errno` = 2 and `matrix` = 'R', then on exit all the elements
        in `r` whose computation involves the zero variance are set to
        zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= 1 and `m` < n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 1)
            On entry, `matrix` = *<value>*.

            Constraint: `matrix` = 'V' or 'R'.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, at least one of the series is such that all its
            elements are practically identical giving zero (or near
            zero) variance.
    """

    from ..base.tsa import multi_corrmat_cross as b_multi_corrmat_cross

    fun_name = 'naginterfaces.library.tsa.multi_corrmat_cross'

    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_w, 'w'),),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _wmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    _r0 = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py),
        sorder=_m_sorder,
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    try:
        b_multi_corrmat_cross(
            matrix,
            _k,
            _n,
            _m,
            _w,
            _wmean,
            _r0,
            _r,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _wmean_py = _wmean
    _r0_py = _r0
    _r_py = _r
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['wmean', 'r0', 'r'],
    )(
        _wmean_py,
        _r0_py,
        _r_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_corrmat_partlag(n, r0, r):
    u"""
    Multivariate time series, sample partial lag correlation matrices,
    chi^2 statistics and significance levels.

    ``multi_corrmat_partlag`` calculates the sample partial lag
    correlation matrices of a multivariate time series.
    A set of chi^2-statistics and their significance levels are also
    returned.
    A call to ``multi_corrmat_cross`` is usually made prior to calling
    this function in order to calculate the sample cross-correlation
    matrices.

    For full information please refer to the NAG Library document for
    g13dn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dnf.html

    Parameters
    ----------
    n : int
        n, the number of observations in each series.

    r0 : float, array-like, shape (k, k)
        If i != j, then `r0`[i-1,j-1] must contain the (i,j)th element
        of the sample cross-correlation matrix at lag zero, R^_{ij}(0).
        If i = j, then `r0`[i-1,i-1] must contain the standard deviation
        of the ith series.

    r : float, array-like, shape (k, k, m)
        `r`[i-1,j-1,l-1] must contain the (i,j)th element of the sample
        cross-correlation at lag l, R^_{ij}(l), for j = 1,2,...,k, for i
        = 1,2,...,k, for l = 1,2,...,m, where series j leads series i
        (see [Further Comments]).

    Returns
    -------
    maxlag : int
        The maximum lag up to which partial lag correlation matrices
        (along with chi^2-statistics and their significance levels) have
        been successfully computed. On a successful exit `maxlag` will
        equal m. If `errno` = 2 on exit, `maxlag` will be less than m.

    parlag : float, ndarray, shape (k, k, m)
        `parlag`[i-1,j-1,l-1] contains the (i,j)th element of the sample
        partial lag correlation matrix at lag l, P^_{ij}(l), for j =
        1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,`maxlag`.

    x : float, ndarray, shape (m)
        `x`[l-1] contains the chi^2-statistic at lag l, for l =
        1,2,...,`maxlag`.

    pvalue : float, ndarray, shape (m)
        `pvalue`[l-1] contains the significance level of the
        corresponding chi^2-statistic in `x`, for l = 1,2,...,`maxlag`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>* and `n` = *<value>*.

            Constraint: m < `n`.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The recursive equations used to compute the partial lag
            correlation matrices are ill-conditioned (they have been
            computed up to lag *<value>*).
    """

    from ..base.tsa import multi_corrmat_partlag as b_multi_corrmat_partlag

    fun_name = 'naginterfaces.library.tsa.multi_corrmat_partlag'

    _r0 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r0'},
        r0,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_r0, 'r0'), (_r, 'r')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_r0, {'fun_name': fun_name, 'entity': 'r0'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=3)[2]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _parlag = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _pvalue = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    try:
        _return_scalar = b_multi_corrmat_partlag(
            _k,
            n,
            _m,
            _r0,
            _r,
            _parlag,
            _x,
            _pvalue,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _maxlag_py = _return_scalar
    _parlag_py = _parlag
    _x_py = _x
    _pvalue_py = _pvalue
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['maxlag', 'parlag', 'x', 'pvalue'],
    )(
        _maxlag_py,
        _parlag_py,
        _x_py,
        _pvalue_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_regmat_partial(z, m):
    u"""
    Multivariate time series, partial autoregression matrices.

    ``multi_regmat_partial`` calculates the sample partial
    autoregression matrices of a multivariate time series.
    A set of likelihood ratio statistics and their significance levels
    are also returned.
    These quantities are useful for determining whether the series
    follows an autoregressive model and, if so, of what order.

    For full information please refer to the NAG Library document for
    g13dp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dpf.html

    Parameters
    ----------
    z : float, array-like, shape (k, n)
        `z`[i-1,t-1] must contain the observation w_{it}, for t =
        1,2,...,n, for i = 1,2,...,k.

    m : int
        m, the number of partial autoregression matrices to be computed.
        If in doubt set `m` = 10.

    Returns
    -------
    maxlag : int
        The maximum lag up to which partial autoregression matrices
        (along with their likelihood ratio statistics and their
        significance levels) have been successfully computed. On a
        successful exit `maxlag` will equal `m`. If `errno` = 2 on exit
        then `maxlag` will be less than `m`.

    parlag : float, ndarray, shape (k, k, `m`)
        `parlag`[i-1,j-1,l-1] contains an estimate of the (i,j)th
        element of the partial autoregression matrix at lag l, P^_l(ij),
        for l = 1,2,...,`maxlag`, i = 1,2,...,k and j = 1,2,...,k.

    se : float, ndarray, shape (k, k, `m`)
        `se`[i-1,j-1,l-1] contains an estimate of the standard error of
        the corresponding element in the array `parlag`.

    qq : float, ndarray, shape (k, k, `m`)
        `qq`[i-1,j-1,l-1] contains an estimate of the (i,j)th element of
        the corresponding variance-covariance matrix Sigma^_l, for j =
        1,2,...,k, for i = 1,2,...,k, for l = 1,2,...,`maxlag`.

    x : float, ndarray, shape (`m`)
        `x`[l-1] contains X_l, the likelihood ratio statistic at lag l,
        for l = 1,2,...,`maxlag`.

    pvalue : float, ndarray, shape (`m`)
        `pvalue`[l-1] contains the significance level of the statistic
        in the corresponding element of `x`.

    loglhd : float, ndarray, shape (`m`)
        `loglhd`[l-1] contains an estimate of the maximum of the
        log-likelihood function when an AR(l-1) model has been fitted to
        the series, for l = 1,2,...,`maxlag`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*, `m` = *<value>* and n = *<value>*.

            Constraint: n-`m`-(k*`m`+1) >= k.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 4.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The recursive equations used to compute the partial
            autoregression matrices are ill-conditioned. They have been
            computed up to lag *<value>*.
    """

    from ..base.tsa import multi_regmat_partial as b_multi_regmat_partial

    fun_name = 'naginterfaces.library.tsa.multi_regmat_partial'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'),),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _parlag = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _qq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _pvalue = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _loglhd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    try:
        _return_scalar = b_multi_regmat_partial(
            _k,
            _n,
            _z,
            _m,
            _parlag,
            _se,
            _qq,
            _x,
            _pvalue,
            _loglhd,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _maxlag_py = _return_scalar
    _parlag_py = _parlag
    _se_py = _se
    _qq_py = _qq
    _x_py = _x
    _pvalue_py = _pvalue
    _loglhd_py = _loglhd
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['maxlag', 'parlag', 'se', 'qq', 'x', 'pvalue', 'loglhd'],
    )(
        _maxlag_py,
        _parlag_py,
        _se_py,
        _qq_py,
        _x_py,
        _pvalue_py,
        _loglhd_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multi_varma_diag(v, ip, iq, m, par, parhld, qq, ishow, io_manager=None):
    u"""
    Multivariate time series, diagnostic checking of residuals,
    following ``multi_varma_estimate``.

    ``multi_varma_diag`` is a diagnostic checking function suitable for
    use after fitting a vector ARMA model to a multivariate time series
    using ``multi_varma_estimate``.
    The residual cross-correlation matrices are returned along with an
    estimate of their asymptotic standard errors and correlations.
    Also, ``multi_varma_diag`` calculates the modified Li--McLeod
    portmanteau statistic and its significance level for testing model
    adequacy.

    For full information please refer to the NAG Library document for
    g13ds

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dsf.html

    Parameters
    ----------
    v : float, array-like, shape (k, n)
        `v`[i-1,t-1] must contain an estimate of the ith component of
        epsilon_t, for t = 1,2,...,n, for i = 1,2,...,k.

    ip : int
        p, the number of AR parameter matrices.

    iq : int
        q, the number of MA parameter matrices.

    m : int
        The value of m, the number of residual cross-correlation
        matrices to be computed. See [Further Comments] for advice on
        the choice of `m`.

    par : float, array-like, shape ((`ip`+`iq`)*k*k)
        The parameter estimates read in row by row in the order
        phi_1,phi_2,...,phi_p, theta_1,theta_2,...,theta_q.

        Thus,

        - if `ip` > 0, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to
          an estimate of the (i,j)th element of phi_l, for i =
          1,2,...,k, for l = 1,2,...,p;

        - if `iq` >= 0, `par`[p*k*k+(l-1)*k*k+(i-1)*k+j-1] must be set
          equal to an estimate of the (i,j)th element of theta_l, for i
          = 1,2,...,k, for l = 1,2,...,q.

        The first p*k*k elements of `par` must satisfy the stationarity
        condition and the next q*k*k elements of `par` must satisfy the
        invertibility condition.

    parhld : bool, array-like, shape ((`ip`+`iq`)*k*k)
        `parhld`[i-1] must be set to True if `par`[i-1] has been held
        constant at a pre-specified value and False if `par`[i-1] is a
        free parameter, for i = 1,2,...,(p+q)*k*k.

    qq : float, array-like, shape (k, k)
        `qq`[i-1,j-1] is an efficient estimate of the (i,j)th element of
        Sigma. The lower triangle only is needed.

    ishow : int
        Must be nonzero if the residual cross-correlation matrices
        {r^_{ij}(l)} and their standard errors {se(r^_{ij}(l))}, the
        modified portmanteau statistic with its significance and a
        summary table are to be printed. The summary table indicates
        which elements of the residual correlation matrices are
        significant at the 5% level in either a positive or negative
        direction; i.e., if r^_{ij}(l) > 1.96*se(r^_{ij}(l)) then a '+'
        is printed, if r^_{ij}(l) < -1.96*se(r^_{ij}(l)) then a '-' is
        printed, otherwise a fullstop (.) is printed. The summary table
        is only printed if k <= 6 on entry.

        The residual cross-correlation matrices, their standard errors
        and the modified portmanteau statistic with its significance are
        available also as output variables in `r`, `rcm`, `chi`, `idf`
        and `siglev`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    qq : float, ndarray, shape (k, k)
        If `errno` != 1, then the upper triangle is set equal to the
        lower triangle.

    r0 : float, ndarray, shape (k, k)
        If i != j, then `r0`[i-1,j-1] contains an estimate of the
        (i,j)th element of the residual cross-correlation matrix at lag
        zero, R^_0. When i = j, `r0`[i-1,j-1] contains the standard
        deviation of the ith residual series. If `errno` = 3 on exit
        then the first k rows and columns of `r0` are set to zero.

    r : float, ndarray, shape (k, k, `m`)
        `r`[l-1,i-1,j-1] is an estimate of the (i,j)th element of the
        residual cross-correlation matrix at lag l, for l = 1,2,...,m,
        for j = 1,2,...,k, for i = 1,2,...,k. If `errno` = 3 on exit
        then all elements of `r` are set to zero.

    rcm : float, ndarray, shape (`m`*k*k, `m`*k*k)
        The estimated standard errors and correlations of the elements
        in the array `r`. The correlation between `r`[l-1,i-1,j-1] and
        `r`[l_2-1,i_2-1,j_2-1] is returned as `rcm`[s,t] where s =
        (l-1)*k*k+(j-1)*k+i and t = (l_2-1)*k*k+(j_2-1)*k+i_2 except
        that if s = t, then `rcm`[s,t] contains the standard error of
        `r`[l-1,i-1,j-1]. If on exit, `errno` >= 5, then all
        off-diagonal elements of `rcm` are set to zero and all diagonal
        elements are set to 1/sqrt(n).

    chi : float
        The value of the modified portmanteau statistic, Q_(m)^*. If
        `errno` = 3 on exit then `chi` is returned as zero.

    idf : int
        The number of degrees of freedom of `chi`.

    siglev : float
        The significance level of `chi` based on `idf` degrees of
        freedom. If `errno` = 3 on exit, `siglev` is returned as one.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ldrcm = *<value>*, `m` = *<value>* and k =
            *<value>*.

            Constraint: ldrcm >= `m`*k*k.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` < n.

        (`errno` 1)
            On entry, `m` = *<value>*, `ip` = *<value>* and `iq` =
            *<value>*.

            Constraint: `m` > `ip`+`iq`.

        (`errno` 1)
            On entry, `ip` = 0 and `iq` = 0.

            Constraint: `ip` = `iq` = 0 must not hold.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 1.

        (`errno` 2)
            On entry, the MA parameter matrices are outside the
            invertibility region.

        (`errno` 2)
            On entry, the AR parameter matrices are outside the
            stationarity region.

        (`errno` 2)
            On entry, the covariance matrix `qq` is not positive
            definite.

        (`errno` 4)
            Excessive iterations needed to find zeros of determinental
            polynomials.

        (`errno` 5)
            On entry, the AR operator has a factor in common with the MA
            operator.

        (`errno` 6)
            The matrix `rcm` could not be computed because one of its
            diagonal elements was found to be non-positive.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one of the residual series in the array
            `v` has near-zero variance.

        (`errno` 3)
            On entry, at least two of the residual series are identical.
    """

    from ..base.tsa import multi_varma_diag as b_multi_varma_diag

    fun_name = 'naginterfaces.library.tsa.multi_varma_diag'

    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _qq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qq'},
        qq,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_v, 'v'), (_qq, 'qq')),
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _parhld = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'parhld'},
        parhld,
        copy=False,
    )
    _r0 = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py),
        sorder=_m_sorder,
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py, _m_py),
        sorder=_m_sorder,
    )
    _rcm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py*_k_py*_k_py, _m_py*_k_py*_k_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalars = b_multi_varma_diag(
            _k,
            _n,
            _v,
            ip,
            iq,
            _m,
            _par,
            _parhld,
            _qq,
            ishow,
            _r0,
            _r,
            _rcm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _qq_py = _qq
    _r0_py = _r0
    _r_py = _r
    _rcm_py = _rcm
    _chi_py = _return_scalars.chi
    _idf_py = _return_scalars.idf
    _siglev_py = _return_scalars.siglev
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['qq', 'r0', 'r', 'rcm', 'chi', 'idf', 'siglev'],
    )(
        _qq_py,
        _r0_py,
        _r_py,
        _rcm_py,
        _chi_py,
        _idf_py,
        _siglev_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_arma_roots(k, ip, par):
    u"""
    Calculates the zeros of a vector autoregressive (or moving average)
    operator.

    ``uni_arma_roots`` calculates the zeros of a vector autoregressive
    (or moving average) operator. This function is likely to be used in
    conjunction with ``rand.times_mv_varma``, ``uni_arima_resid``,
    ``multi_varma_estimate`` or ``multi_varma_diag``.

    For full information please refer to the NAG Library document for
    g13dx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13dxf.html

    Parameters
    ----------
    k : int
        k, the dimension of the multivariate time series.

    ip : int
        The number of AR (or MA) parameter matrices, p (or q).

    par : float, array-like, shape (`ip`*`k`*`k`)
        The AR (or MA) parameter matrices read in row by row in the
        order phi_1,phi_2,...,phi_p (or theta_1,theta_2,...,theta_q).
        That is, `par`[(l-1)*k*k+(i-1)*k+j-1] must be set equal to the
        (i,j)th element of phi_l, for l = 1,2,...,p (or the (i,j)th
        element of theta_l, for l = 1,2,...,q).

    Returns
    -------
    rr : float, ndarray, shape (`k`*`ip`)
        The real parts of the eigenvalues.

    ri : float, ndarray, shape (`k`*`ip`)
        The imaginary parts of the eigenvalues.

    rmod : float, ndarray, shape (`k`*`ip`)
        The moduli of the eigenvalues.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 2)
            An excessive number of iterations have been required to
            calculate the eigenvalues.
    """

    from ..base.tsa import uni_arma_roots as b_uni_arma_roots

    fun_name = 'naginterfaces.library.tsa.uni_arma_roots'

    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k = _b_utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = _b_utils._EngineIntScalarType.to_py(_k)
    _ip = _b_utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ip_py = _b_utils._EngineIntScalarType.to_py(_ip)
    _rr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py*_ip_py,),
    )
    _ri = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py*_ip_py,),
    )
    _rmod = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py*_ip_py,),
    )
    b_uni_arma_roots(
        _k,
        _ip,
        _par,
        _rr,
        _ri,
        _rmod,
    )
    _rr_py = _rr
    _ri_py = _ri
    _rmod_py = _rmod
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['rr', 'ri', 'rmod'],
    )(
        _rr_py,
        _ri_py,
        _rmod_py,
    )
    return _return_data

def multi_kalman_sqrt_var(a, b, stq, q, c, r, s, tol=0.0):
    u"""
    Combined measurement and time update, one iteration of Kalman
    filter, time-varying, square root covariance filter.

    ``multi_kalman_sqrt_var`` performs a combined measurement and time
    update of one iteration of the time-varying Kalman filter using a
    square root covariance filter.

    For full information please refer to the NAG Library document for
    g13ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13eaf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The state transition matrix, A_i.

    b : float, array-like, shape (n, l)
        The noise coefficient matrix B_i.

    stq : bool
        If `stq` = True, the state noise covariance matrix Q_i is
        assumed to be the identity matrix. Otherwise the lower
        triangular Cholesky factor, Q_i^1/2, must be provided in `q`.

    q : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if not((`stq`): l; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if not((`stq`): l; otherwise: 0.

        If `stq` = False, `q` must contain the lower triangular Cholesky
        factor of the state noise covariance matrix, Q_i^1/2. Otherwise
        `q` is not referenced.

    c : float, array-like, shape (m, n)
        The measurement coefficient matrix, C_i.

    r : float, array-like, shape (m, m)
        The lower triangular Cholesky factor of the measurement noise
        covariance matrix R_i^1/2.

    s : float, array-like, shape (n, n)
        The lower triangular Cholesky factor of the state covariance
        matrix, S_i.

    tol : float, optional
        The tolerance used to test for the singularity of H_i^1/2. If
        0.0 <= `tol` < m^2*machine precision, then m^2*machine precision
        is used instead. The inverse of the condition number of H^(1/2)
        is estimated by a call to ``lapacklin.dtrcon``. If this estimate
        is less than `tol` then H^(1/2) is assumed to be singular.

    Returns
    -------
    s : float, ndarray, shape (n, n)
        The lower triangular Cholesky factor of the state covariance
        matrix, S_{i+1}.

    k : float, ndarray, shape (n, m)
        The Kalman gain matrix, K_i, premultiplied by the state
        transition matrix, A_i, A_i K_i.

    h : float, ndarray, shape (m, m)
        The lower triangular matrix H_i^1/2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, l = *<value>*.

            Constraint: l >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The matrix H_i^1/2 is singular.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import multi_kalman_sqrt_var as b_multi_kalman_sqrt_var

    fun_name = 'naginterfaces.library.tsa.multi_kalman_sqrt_var'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_q, 'q'), (_c, 'c'), (_r, 'r'), (_s, 's')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _l_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _l = _b_utils._EngineIntScalarType(
        _l_py, {'fun_name': fun_name, 'entity': '_l'},
    ).to_ctypes()
    _k = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_m_sorder,
    )
    b_multi_kalman_sqrt_var(
        _n,
        _m,
        _l,
        _a,
        _b,
        stq,
        _q,
        _c,
        _r,
        _s,
        _k,
        _h,
        tol,
    )
    _s_py = _s
    _k_py = _k
    _h_py = _h
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'k', 'h'],
    )(
        _s_py,
        _k_py,
        _h_py,
    )
    return _return_data

def multi_kalman_sqrt_invar(transf, l, a, b, stq, q, c, r, s, tol=0.0):
    u"""
    Combined measurement and time update, one iteration of Kalman
    filter, time-invariant, square root covariance filter.

    ``multi_kalman_sqrt_invar`` performs a combined measurement and time
    update of one iteration of the time-invariant Kalman filter using a
    square root covariance filter.

    For full information please refer to the NAG Library document for
    g13eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ebf.html

    Parameters
    ----------
    transf : str, length 1
        Indicates whether to transform the input matrix pair (A,C) to
        lower observer Hessenberg form. The transformation will only be
        required on the first call to ``multi_kalman_sqrt_invar``.

        `transf` = 'T'

          The matrices in arrays `a` and `c` are transformed to lower
          observer Hessenberg form and the matrices in `b` and `s` are
          transformed as described in [Description].

        `transf` = 'H'

          The matrices in arrays `a`, `c` and `b` should be as returned
          from a previous call to ``multi_kalman_sqrt_invar`` with
          `transf` = 'T'.

    l : int
        l, the dimension of the state noise.

    a : float, array-like, shape (n, n)
        If `transf` = 'T', the state transition matrix, A.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

    b : float, array-like, shape (n, `l`)
        If `transf` = 'T', the noise coefficient matrix B.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

    stq : bool
        If `stq` = True, the state noise covariance matrix Q_i is
        assumed to be the identity matrix. Otherwise the lower
        triangular Cholesky factor, Q_i^1/2, must be provided in `q`.

    q : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if not((`stq`): `l`; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if not((`stq`): `l`; otherwise: 0.

        If `stq` = False, `q` must contain the lower triangular Cholesky
        factor of the state noise covariance matrix, Q_i^1/2. Otherwise
        `q` is not referenced.

    c : float, array-like, shape (m, n)
        If `transf` = 'T', the measurement coefficient matrix, C.

        If `transf` = 'H', the transformed matrix as returned by a
        previous call to ``multi_kalman_sqrt_invar`` with `transf` =
        'T'.

    r : float, array-like, shape (m, m)
        The lower triangular Cholesky factor of the measurement noise
        covariance matrix R_i^1/2.

    s : float, array-like, shape (n, n)
        If `transf` = 'T' the lower triangular Cholesky factor of the
        state covariance matrix, S_i.

        If `transf` = 'H' the lower triangular Cholesky factor of the
        covariance matrix of the transformed state vector S_i^* as
        returned from a previous call to ``multi_kalman_sqrt_invar``
        with `transf` = 'T'.

    tol : float, optional
        The tolerance used to test for the singularity of H_i^1/2. If
        0.0 <= `tol` < m^2*machine precision, then m^2*machine precision
        is used instead. The inverse of the condition number of H^(1/2)
        is estimated by a call to ``lapacklin.dtrcon``. If this estimate
        is less than `tol` then H^(1/2) is assumed to be singular.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If `transf` = 'T', the transformed matrix, U*AU^(*T), otherwise
        `a` is unchanged.

    b : float, ndarray, shape (n, `l`)
        If `transf` = 'T', the transformed matrix, U*B, otherwise `b` is
        unchanged.

    c : float, ndarray, shape (m, n)
        If `transf` = 'T', the transformed matrix, CU^(*T), otherwise
        `c` is unchanged.

    s : float, ndarray, shape (n, n)
        The lower triangular Cholesky factor of the transformed state
        covariance matrix, S_i+1^*.

    k : float, ndarray, shape (n, m)
        The Kalman gain matrix for the transformed state vector
        premultiplied by the state transformed transition matrix,
        U*AK_i.

    h : float, ndarray, shape (m, m)
        The lower triangular matrix H_i^1/2.

    u : float, ndarray, shape (n, :)
        If `transf` = 'T' the n by n transformation matrix U*, otherwise
        `u` is not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `transf` = *<value>*.

            Constraint: `transf` = 'T' or 'H'.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The matrix H_i^1/2 is singular.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import multi_kalman_sqrt_invar as b_multi_kalman_sqrt_invar

    fun_name = 'naginterfaces.library.tsa.multi_kalman_sqrt_invar'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_q, 'q'), (_c, 'c'), (_r, 'r'), (_s, 's')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _k = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_m_sorder,
    )
    _transf = _b_utils._EngineCharScalarType(
        transf, {'fun_name': fun_name, 'entity': 'transf'}, exp_chrlen=1,
    ).to_ctypes()
    _transf_py = _b_utils._EngineCharScalarType.to_py(_transf)
    if _transf_py in ('T', 't'):
        _u_dim2 = _n_py
    else:
        _u_dim2 = 0
    exp_shape_u = (_n_py, _u_dim2)
    _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_u,
        sorder=_m_sorder,
    )
    b_multi_kalman_sqrt_invar(
        transf,
        _n,
        _m,
        l,
        _a,
        _b,
        stq,
        _q,
        _c,
        _r,
        _s,
        _k,
        _h,
        _u,
        tol,
    )
    _a_py = _a
    _b_py = _b
    _c_py = _c
    _s_py = _s
    _k_py = _k
    _h_py = _h
    _u_py = _u
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 's', 'k', 'h', 'u'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _s_py,
        _k_py,
        _h_py,
        _u_py,
    )
    return _return_data

def kalman_unscented_state_revcom(irevcm, y, ly, x, st, xt, fxt, comm, lx=None, ropt=None):
    u"""
    Combined time and measurement update, one iteration of the Unscented
    Kalman Filter for a nonlinear state space model, with additive noise
    (reverse communication).

    ``kalman_unscented_state_revcom`` applies the Unscented Kalman
    Filter to a nonlinear state space model, with additive noise.

    ``kalman_unscented_state_revcom`` uses reverse communication for
    evaluating the nonlinear functionals of the state space model.

    For full information please refer to the NAG Library document for
    g13ej

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ejf.html

    Parameters
    ----------
    irevcm : int
        On initial entry: must be set to 0 or 3.

        If `irevcm` = 0, it is assumed that t = 0, otherwise it is
        assumed that t != 0 and that ``kalman_unscented_state_revcom``
        has been called at least once before at an earlier time step.

        On intermediate entry: `irevcm` must remain unchanged.

    y : float, array-like, shape (my)
        y_t, the observed data at the current time point.

    ly : float, array-like, shape (my, my)
        L_y, such that L_y L_y^T = Sigma_y, i.e., the lower triangular
        part of a Cholesky decomposition of the observation noise
        covariance structure. Only the lower triangular part of `ly` is
        referenced.

        If Sigma_y is time dependent, the value supplied should be for
        time t.

    x : float, ndarray, shape (mx), modified in place
        On initial entry: x^_{t-1} the state vector for the previous
        time point.

        On intermediate exit: when

        `irevcm` = 1

          `x` is unchanged.

        `irevcm` = 2

          x^_t_.

        On intermediate entry: `x` must remain unchanged.

        On final exit: x^_t the updated state vector.

    st : float, ndarray, shape (mx, mx), modified in place
        On initial entry: S_t, such that S_{t-1}S_{t-1}^T = P_{t-1},
        i.e., the lower triangular part of a Cholesky decomposition of
        the state covariance matrix at the previous time point. Only the
        lower triangular part of `st` is referenced.

        On intermediate exit: when

        `irevcm` = 1

          `st` is unchanged.

        `irevcm` = 2

          S_t_, the lower triangular part of a Cholesky factorization of
          P_t_.

        On intermediate entry: `st` must remain unchanged.

        On final exit: S_t, the lower triangular part of a Cholesky
        factorization of the updated state covariance matrix.

    xt : float, ndarray, shape (mx, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `irevcm`=0: max(my, n); if `irevcm` in
        (1, 2, 3) and licomm>=10: max(my, `comm`['icomm'][9]);
        otherwise: 0.

        On initial entry: need not be set.

        On intermediate exit: X_t when `irevcm` = 1, otherwise Y_t.

        For the jth sigma point, the value for the ith parameter is held
        in `xt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,mx.

        On intermediate entry: `xt` must remain unchanged.

        On final exit: the contents of `xt` are undefined.

    fxt : float, ndarray, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `irevcm` in (1, 2): max(mx, my);
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `irevcm`=1: n+mx; if `irevcm`=2: n+my;
        otherwise: 0.

        On initial entry: need not be set.

        On intermediate exit: the contents of `fxt` are undefined.

        On intermediate entry: F(X_t) when `irevcm` = 1, otherwise
        H(Y_t) for the values of X_t and Y_t held in `xt`.

        For the jth sigma point the value for the ith parameter should
        be held in `fxt`[i-1,j-1], for j = 1,2,...,n.

        When `irevcm` = 1, i = 1,2,...,mx and when `irevcm` = 2, i =
        1,2,...,my.

        On final exit: the contents of `fxt` are undefined.

    comm : dict, communication object, modified in place
        Communication structure.

        On initial entry: need not be set.

    lx : None or float, array-like, shape (:, mx), optional
        L_x, such that L_x L_x^T = Sigma_x, i.e., the lower triangular
        part of a Cholesky decomposition of the process noise covariance
        structure. Only the lower triangular part of `lx` is referenced.

        If ldlx = 0, there is no process noise (v_t = 0 for all t) and
        `lx` is not referenced.

        If Sigma_x is time dependent, the value supplied should be for
        time t.

    ropt : None or float, array-like, shape (lropt), optional
        Options. The default value will be used for `ropt`[i-1] if lropt
        < i. Setting lropt = 0 will use the default values for all
        options and `ropt` need not be set.

        `ropt`[0]

          If set to 1 then the second set of sigma points are redrawn,
          as given by equation [equation]. If set to 2 then the second
          set of sigma points are generated via augmentation, as given
          by equation [equation].

          Default is for the sigma points to be redrawn (i.e., `ropt`[0]
          = 1)

        `ropt`[1]

          kappa_x, value of kappa used when constructing the first set
          of sigma points, X_t.

          Defaults to 3-mx.

        `ropt`[2]

          alpha_x, value of alpha used when constructing the first set
          of sigma points, X_t.

          Defaults to 1.

        `ropt`[3]

          beta_x, value of beta used when constructing the first set of
          sigma points, X_t.

          Defaults to 2.

        `ropt`[4]

          Value of kappa used when constructing the second set of sigma
          points, Y_t.

          Defaults to 3-2*mx when ldlx != 0 and the second set of sigma
          points are augmented and kappa_x otherwise.

        `ropt`[5]

          Value of alpha used when constructing the second set of sigma
          points, Y_t.

          Defaults to alpha_x.

        `ropt`[6]

          Value of beta used when constructing the second set of sigma
          points, Y_t.

          Defaults to beta_x.

    Returns
    -------
    irevcm : int
        On intermediate exit: `irevcm` = 1 or 2. The value of `irevcm`
        specifies what intermediate values are returned by this function
        and what values the calling program must assign to arguments of
        ``kalman_unscented_state_revcom`` before re-entering the
        routine. Details of the output and required input are given in
        the individual argument descriptions.

        On final exit: `irevcm` = 3

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0, 1, 2 or 3.

        (`errno` 21)
            On entry, mx = *<value>*.

            Constraint: mx >= 1.

        (`errno` 22)
            mx has changed between calls.

            On intermediate entry, mx = *<value>*.

            On initial entry, mx = *<value>*.

        (`errno` 31)
            On entry, my = *<value>*.

            Constraint: my >= 1.

        (`errno` 32)
            my has changed between calls.

            On intermediate entry, my = *<value>*.

            On initial entry, my = *<value>*.

        (`errno` 121)
            On entry, augmented sigma points requested, n = *<value>*
            and mx = *<value>*.

            Constraint: n >= *<value>*.

        (`errno` 122)
            On entry, redrawn sigma points requested, n = *<value>* and
            mx = *<value>*.

            Constraint: n >= *<value>*.

        (`errno` 123)
            n has changed between calls.

            On intermediate entry, n = *<value>*.

            On intermediate exit, n = *<value>*.

        (`errno` 171)
            On entry, `ropt`[0] = *<value>*.

            Constraint: `ropt`[0] = 1 or 2.

        (`errno` 172)
            On entry, `ropt`[*<value>*] = *<value>*.

            Constraint: kappa > *<value>*.

        (`errno` 173)
            On entry, `ropt`[*<value>*] = *<value>*.

            Constraint: alpha > 0.

        (`errno` 181)
            On entry, lropt = *<value>*.

            Constraint: 0 <= lropt <= 7.

        (`errno` 191)
            `comm`['icomm'] has been corrupted between calls.

        (`errno` 211)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 301)
            A weight was negative and it was not possible to downdate
            the Cholesky factorization.

        (`errno` 302)
            Unable to calculate the Kalman gain matrix.

        (`errno` 303)
            Unable to calculate the Cholesky factorization of the
            updated state covariance matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 202)
            On entry, licomm = *<value>* and lrcomm = *<value>*.

            Constraint: licomm >= 30 and lrcomm >= 30+my+mx*my+2*max(mx,
            my).

            The minimum required values for licomm and lrcomm are
            returned in `comm`['icomm'][0] and `comm`['icomm'][1]
            respectively.
    """

    from ..base.tsa import kalman_unscented_state_revcom as b_kalman_unscented_state_revcom

    fun_name = 'naginterfaces.library.tsa.kalman_unscented_state_revcom'

    _lx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lx'},
        lx,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mx = _b_utils._EngineIntScalarType(
        1 if lx is None else _b_utils._EngineFloat64ArrayType.get_shape(_lx, {'fun_name': fun_name, 'entity': 'lx'}, exp_rank=2, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _ly = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ly'},
        ly,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'xt'},
        xt,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(xt, {'fun_name': fun_name, 'entity': 'xt'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'fxt'},
        fxt,
    )
    _ropt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ropt'},
        ropt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lropt_py = _b_utils._EngineFloat64ArrayType.get_shape(_ropt, {'fun_name': fun_name, 'entity': 'ropt'}, exp_rank=1, allow_none=True)[0]
    _lropt = _b_utils._EngineIntScalarType(
        _lropt_py, {'fun_name': fun_name, 'entity': '_lropt'},
    ).to_ctypes()
    try:
        _return_scalars = b_kalman_unscented_state_revcom(
            irevcm,
            _mx,
            _my,
            _y,
            _lx,
            _ly,
            x,
            st,
            _n,
            xt,
            fxt,
            _ropt,
            _lropt,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _irevcm_py = _return_scalars.irevcm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm'],
    )(
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _irevcm_py

def kalman_unscented_state(y, lx, ly, f, h, x, st, data=None, spiked_sorder='C'):
    u"""
    Combined time and measurement update, one iteration of the Unscented
    Kalman Filter for a nonlinear state space model, with additive
    noise.

    ``kalman_unscented_state`` applies the Unscented Kalman Filter (UKF)
    to a nonlinear state space model, with additive noise.

    ``kalman_unscented_state`` uses direct communication for evaluating
    the nonlinear functionals of the state space model.

    For full information please refer to the NAG Library document for
    g13ek

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ekf.html

    Parameters
    ----------
    y : float, array-like, shape (my)
        y_t, the observed data at the current time point.

    lx : float, array-like, shape (mx, mx)
        L_x, such that L_x L_x^T = Sigma_x, i.e., the lower triangular
        part of a Cholesky decomposition of the process noise covariance
        structure. Only the lower triangular part of `lx` is referenced.

        If Sigma_x is time dependent, the value supplied should be for
        time t.

    ly : float, array-like, shape (my, my)
        L_y, such that L_y L_y^T = Sigma_y, i.e., the lower triangular
        part of a Cholesky decomposition of the observation noise
        covariance structure. Only the lower triangular part of `ly` is
        referenced.

        If Sigma_y is time dependent, the value supplied should be for
        time t.

    f : callable fxt = f(xt, data=None)
        The state function, F.

        Parameters
        ~~~~~~~~~~
        xt : float, ndarray, shape (mx, n)
            X_t, the sigma points generated in (a). For the jth sigma
            point, the value for the ith parameter is held in
            `xt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,m_x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fxt : float, array-like, shape (mx, n)
            F(X_t).

            For the jth sigma point the value for the ith parameter
            should be held in `fxt`[i-1,j-1], for j = 1,2,...,n, for i =
            1,2,...,m_x.

    h : callable hyt = h(my, yt, data=None)
        The measurement function, H as described in (e).

        Parameters
        ~~~~~~~~~~
        my : int
            m_y, the number of observed variables.

        yt : float, ndarray, shape (mx, n)
            Y_t, the sigma points generated in (d). For the jth sigma
            point, the value for the ith parameter is held in
            `yt`[i-1,j-1], for j = 1,2,...,n, for i = 1,2,...,m_x, where
            m_x is the number of state variables and n is the number of
            sigma points.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hyt : float, array-like, shape (`my`, n)
            H(Y_t).

            For the jth sigma point the value for the ith parameter
            should be held in `hyt`[i-1,j-1], for j = 1,2,...,n, for i =
            1,2,...,m_y.

    x : float, array-like, shape (mx)
        x^_{t-1} the state vector for the previous time point.

    st : float, array-like, shape (mx, mx)
        S_t, such that S_{t-1}S_{t-1}^T = P_{t-1}, i.e., the lower
        triangular part of a Cholesky decomposition of the state
        covariance matrix at the previous time point. Only the lower
        triangular part of `st` is referenced.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `lx` and `st` are spiked (i.e., have unit extent in all but
        one dimension, or have size 1), `spiked_sorder` selects the
        storage order to associate with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    x : float, ndarray, shape (mx)
        x^_t the updated state vector.

    st : float, ndarray, shape (mx, mx)
        S_t, the lower triangular part of a Cholesky factorization of
        the updated state covariance matrix.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, mx = *<value>*.

            Constraint: mx >= 1.

        (`errno` 21)
            On entry, my = *<value>*.

            Constraint: my >= 1.

        (`errno` 301)
            A weight was negative and it was not possible to downdate
            the Cholesky factorization.

        (`errno` 302)
            Unable to calculate the Kalman gain matrix.

        (`errno` 303)
            Unable to calculate the Cholesky factorization of the
            updated state covariance matrix.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 61)
            User requested termination in `f`.

        (`errno` 71)
            User requested termination in `h`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.kalman_unscented_state_ex.main`
    """

    from ..base.tsa import kalman_unscented_state as b_kalman_unscented_state

    fun_name = 'naginterfaces.library.tsa.kalman_unscented_state'

    cb_f_name = fun_name + ':f'

    def cb_f(mx, n, xt, fxt, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fxt_f = f(
                xt,
                data,
            )
        else:
            _fxt_f = f(
                xt,
            )
        _fxt_f_np = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fxt'},
            _fxt_f,
            copy=False,
            exp_shape=(mx, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_f_name},
            ((_fxt_f_np, 'fxt'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fxt[:, :] = _fxt_f_np

    cb_h_name = fun_name + ':h'

    def cb_h(mx, my, n, yt, hyt, data=None):
        """Wrapping user's h. Called as base h."""
        if data is not None:
            _hyt_h = h(
                my,
                yt,
                data,
            )
        else:
            _hyt_h = h(
                my,
                yt,
            )
        _hyt_h_np = _utils._to_ndarray(
            {'fun_name': cb_h_name, 'returned': 'hyt'},
            _hyt_h,
            copy=False,
            exp_shape=(my, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_h_name},
            ((_hyt_h_np, 'hyt'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        hyt[:, :] = _hyt_h_np

    _lx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lx'},
        lx,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ly = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ly'},
        ly,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _st = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'st'},
        st,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_lx, 'lx'), (_ly, 'ly'), (_st, 'st')),
        spiked_sorder=spiked_sorder,
    )
    _mx_py = _b_utils._EngineFloat64ArrayType.get_shape(_lx, {'fun_name': fun_name, 'entity': 'lx'}, exp_rank=2)[0]
    _mx = _b_utils._EngineIntScalarType(
        _mx_py, {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    _b_utils._check_callable(h, {'fun_name': fun_name, 'entity': 'h'})
    _h = cb_h
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        b_kalman_unscented_state(
            _mx,
            _my,
            _y,
            _lx,
            _ly,
            _f,
            _h,
            _x,
            _st,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _st_py = _st
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'st'],
    )(
        _x_py,
        _st_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_asym1_estim(dist, yt, x, ip, iq, mn, isym, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for either a symmetric
    GARCH process or a GARCH process with asymmetry of the form
    (epsilon_{t-1}+gamma)^2.

    ``uni_garch_asym1_estim`` estimates the parameters of either a
    standard univariate regression GARCH process, or a univariate
    regression-type I AGARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13faf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    isym : int
        If `isym` = 1, the asymmetry term gamma will be included in the
        model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        If `isym` = 1, the next element must contain the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element must contain df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts`[1] = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    copts : bool, array-like, shape (2)
        The options to be used by ``uni_garch_asym1_estim``.

        `copts`[0] = True

          Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

          The function provides initial parameter estimates of the
          regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the coefficients beta_j, for j =
        1,2,...,p.

        If `isym` = 1, the next element contains the estimate for the
        asymmetry parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o.

        The next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        If `isym` = 1, the next element contains the standard error for
        gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the score for alpha_o.

        The next `iq` elements contain the score for alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the scores for beta_j, for j =
        1,2,...,p.

        If `isym` = 1, the next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            1+`iq`+`ip`+`isym`+`mn`+nreg, else npar =
            2+`iq`+`ip`+`isym`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, `isym` = *<value>*.

            Constraint: `isym` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_garch_asym1_ex.main`
    """

    from ..base.tsa import uni_garch_asym1_estim as b_uni_garch_asym1_estim

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym1_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _copts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'copts'},
        copts,
        copy=False,
    )
    try:
        _return_scalars = b_uni_garch_asym1_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            isym,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            _copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
    )(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_asym1_forecast(nt, ip, iq, theta, gamma, ht, et):
    u"""
    Univariate time series, forecast function for either a symmetric
    GARCH process or a GARCH process with asymmetry of the form
    (epsilon_{t-1}+gamma)^2.

    ``uni_garch_asym1_forecast`` forecasts the conditional variances
    h_t, for t = T+1,...,T+xi, from a type I AGARCH(p,q) sequence, where
    xi is the forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13fb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fbf.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.uni_garch_asym1_ex.main`
    """

    from ..base.tsa import uni_garch_asym1_forecast as b_uni_garch_asym1_forecast

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym1_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_uni_garch_asym1_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        gamma,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

def uni_garch_asym2_estim(dist, yt, x, ip, iq, mn, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for a GARCH process
    with asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``uni_garch_asym2_estim`` estimates the parameters of a univariate
    regression-type II AGARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fcf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        The next element must contain the asymmetry parameter gamma.

        If `dist` = 'T', the next element must contain df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the mean term b_o.

        If `copts`[1] = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    copts : bool, array-like, shape (2)
        The options to be used by ``uni_garch_asym2_estim``.

        `copts`[0] = True

          Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

          The function provides initial parameter estimates of the
          regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the coefficients beta_j, for j =
        1,2,...,p.

        The next element contains the estimate for the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        The next element contains the standard error for gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the score for alpha_o and the next
        `iq` elements contain the score for alpha_i, for i = 1,2,...,q.

        The next `ip` elements are the scores for beta_j, for j =
        1,2,...,p.

        The next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            2+`iq`+`ip`+`mn`+nreg, else npar = 3+`iq`+`ip`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_garch_asym2_estim as b_uni_garch_asym2_estim

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym2_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _copts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'copts'},
        copts,
        copy=False,
    )
    try:
        _return_scalars = b_uni_garch_asym2_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            _copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
    )(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_asym2_forecast(nt, ip, iq, theta, gamma, ht, et):
    u"""
    Univariate time series, forecast function for a GARCH process with
    asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``uni_garch_asym2_forecast`` forecasts the conditional variances,
    h_t,t = T+1,...,T+xi from a type II AGARCH(p,q) sequence, where xi
    is the forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13fd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fdf.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_garch_asym2_forecast as b_uni_garch_asym2_forecast

    fun_name = 'naginterfaces.library.tsa.uni_garch_asym2_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_uni_garch_asym2_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        gamma,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

def uni_garch_gjr_estim(dist, yt, x, ip, iq, mn, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for an asymmetric
    Glosten, Jagannathan and Runkle (GJR) GARCH process.

    ``uni_garch_gjr_estim`` estimates the parameters of a univariate
    regression-GJR GARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fef.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements must contain the coefficients beta_j, for
        j = 1,2,...,p.

        The next element must contain the asymmetry parameter gamma.

        If `dist` = 'T', the next element contains df, the number of
        degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts`[1] = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts`[1] = False, `pht` is the value to be used for the
        pre-observed conditional variance; otherwise `pht` is not
        referenced.

    copts : bool, array-like, shape (2)
        The options to be used by ``uni_garch_gjr_estim``.

        `copts`[0] = True

          Stationary conditions are enforced, otherwise they are not.

        `copts`[1] = True

          The function provides initial parameter estimates of the
          regression terms, otherwise these are to be provided by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters. If `maxit`
        is set to 0, the standard errors, score vector and
        variance-covariance are calculated for the input value of theta
        in `theta` when `dist` = 'N'; however the value of theta is not
        updated.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o, the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `ip` elements are the moving average coefficients
        beta_j, for j = 1,2,...,p.

        The next element contains the estimate for the asymmetry
        parameter gamma.

        If `dist` = 'T', the next element contains an estimate for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        The next element contains the standard error for gamma.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the score for alpha_o, the next `iq`
        elements contain the scores for alpha_i, for i = 1,2,...,q.

        The next `ip` elements are the score for beta_j, for j =
        1,2,...,p.

        The next element contains the score for gamma.

        If `dist` = 'T', the next element contains the score for df, the
        number of degrees of freedom of the Student's t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts`[1] = True, `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            2+`iq`+`ip`+`mn`+nreg, else npar = 3+`iq`+`ip`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_garch_gjr_estim as b_uni_garch_gjr_estim

    fun_name = 'naginterfaces.library.tsa.uni_garch_gjr_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _copts = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'copts'},
        copts,
        copy=False,
    )
    try:
        _return_scalars = b_uni_garch_gjr_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            _copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
    )(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_gjr_forecast(nt, ip, iq, theta, gamma, ht, et):
    u"""
    Univariate time series, forecast function for an asymmetric Glosten,
    Jagannathan and Runkle (GJR) GARCH process.

    ``uni_garch_gjr_forecast`` forecasts the conditional variances, h_t,
    for t = T+1,...,T+xi from a GJR GARCH(p,q) sequence, where xi is the
    forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13ff

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fff.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the coefficients alpha_i, for i
        = 1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    from ..base.tsa import uni_garch_gjr_forecast as b_uni_garch_gjr_forecast

    fun_name = 'naginterfaces.library.tsa.uni_garch_gjr_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_uni_garch_gjr_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        gamma,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

def uni_garch_exp_estim(dist, yt, x, ip, iq, mn, theta, pht, copts, maxit, tol, io_manager=None):
    u"""
    Univariate time series, parameter estimation for an exponential
    GARCH (EGARCH) process.

    ``uni_garch_exp_estim`` estimates the parameters of a univariate
    regression-exponential GARCH(p,q) process.

    For full information please refer to the NAG Library document for
    g13fg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fgf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for e_t.

        `dist` = 'N'

          A Normal distribution is used.

        `dist` = 'T'

          A Student's t-distribution is used.

    yt : float, array-like, shape (num)
        The sequence of observations, y_t, for t = 1,2,...,T.

    x : float, array-like, shape (num, nreg)
        Row t of `x` must contain the time dependent exogenous vector
        x_t, where x_t^T = (x_t^1,...,x_t^k), for t = 1,2,...,T.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    mn : int
        If `mn` = 1, the mean term b_0 will be included in the model.

    theta : float, array-like, shape (npar)
        The initial parameter estimates for the vector theta.

        The first element must contain the coefficient alpha_o and the
        next `iq` elements must contain the autoregressive coefficients
        alpha_i, for i = 1,2,...,q.

        The next `iq` elements contain the coefficients phi_i, for i =
        1,2,...,q.

        The next `ip` elements must contain the moving average
        coefficients beta_i, for i = 1,2,...,p.

        If `dist` = 'T', the next element must contain an estimate for
        df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element must contain the mean term b_o.

        If `copts` = False, the remaining nreg elements are taken as
        initial estimates of the linear regression coefficients b_i, for
        i = 1,2,...,k.

    pht : float
        If `copts` = False then `pht` is the value to be used for the
        pre-observed conditional variance, otherwise `pht` is not
        referenced.

    copts : bool
        If `copts` = True, the function provides initial parameter
        estimates of the regression terms, otherwise these are provided
        by you.

    maxit : int
        The maximum number of iterations to be used by the optimization
        function when estimating the GARCH(p,q) parameters.

    tol : float
        The tolerance to be used by the optimization function when
        estimating the GARCH(p,q) parameters.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    theta : float, ndarray, shape (npar)
        The estimated values theta^ for the vector theta.

        The first element contains the coefficient alpha_o and the next
        `iq` elements contain the coefficients alpha_i, for i =
        1,2,...,q.

        The next `iq` elements contain the coefficients phi_i, for i =
        1,2,...,q.

        The next `ip` elements are the moving average coefficients
        beta_i, for i = 1,2,...,p.

        If `dist` = 'T', the next element contains an estimate for df
        then the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains an estimate for the mean
        term b_o.

        The final nreg elements are the estimated linear regression
        coefficients b_i, for i = 1,2,...,k.

    se : float, ndarray, shape (npar)
        The standard errors for theta^.

        The first element contains the standard error for alpha_o and
        the next `iq` elements contain the standard errors for alpha_i,
        for i = 1,2,...,q.

        The next `iq` elements contain the standard errors for phi_i,
        for i = 1,2,...,q.

        The next `ip` elements are the standard errors for beta_j, for j
        = 1,2,...,p.

        If `dist` = 'T', the next element contains the standard error
        for df, the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the standard error for
        b_o.

        The final nreg elements are the standard errors for b_j, for j =
        1,2,...,k.

    sc : float, ndarray, shape (npar)
        The scores for theta^.

        The first element contains the scores for alpha_o, the next `iq`
        elements contain the scores for alpha_i, for i = 1,2,...,q, the
        next `iq` elements contain the scores for phi_i, for i =
        1,2,...,q, the next `ip` elements are the scores for beta_j, for
        j = 1,2,...,p.

        If `dist` = 'T', the next element contains the scores for df,
        the number of degrees of freedom of the Student's
        t-distribution.

        If `mn` = 1, the next element contains the score for b_o.

        The final nreg elements are the scores for b_j, for j =
        1,2,...,k.

    covr : float, ndarray, shape (npar, npar)
        The covariance matrix of the parameter estimates theta^, that is
        the inverse of the Fisher Information Matrix.

    pht : float
        If `copts` = True then `pht` is the estimated value of the
        pre-observed conditional variance.

    et : float, ndarray, shape (num)
        The estimated residuals, epsilon_t, for t = 1,2,...,T.

    ht : float, ndarray, shape (num)
        The estimated conditional variances, h_t, for t = 1,2,...,T.

    lgf : float
        The value of the log-likelihood function at theta^.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: num >= nreg+`mn`.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: npar < 20.

        (`errno` 1)
            On entry, npar = *<value>*.

            Constraint: if `dist` = 'N' then npar =
            1+2*`iq`+`ip`+`mn`+nreg, else npar =
            2+2*`iq`+`ip`+`mn`+nreg.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N' or 'T'.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `mn` = *<value>*.

            Constraint: `mn` = 0 or 1.

        (`errno` 1)
            On entry, nreg = *<value>*.

            Constraint: nreg >= 0.

        (`errno` 3)
            On entry, the matrix X is not full rank.

        (`errno` 4)
            The information matrix is not positive definite.

        (`errno` 7)
            No feasible model parameters could be found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The maximum number of iterations has been reached.

        (`errno` 6)
            The log-likelihood cannot be optimized any further.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_garch_exp_estim as b_uni_garch_exp_estim

    fun_name = 'naginterfaces.library.tsa.uni_garch_exp_estim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _yt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yt'},
        yt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_yt, {'fun_name': fun_name, 'entity': 'yt'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _nreg_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _nreg = _b_utils._EngineIntScalarType(
        _nreg_py, {'fun_name': fun_name, 'entity': '_nreg'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_theta, {'fun_name': fun_name, 'entity': 'theta'}, exp_rank=1)[0]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _se = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _sc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py,),
    )
    _covr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_npar_py, _npar_py),
        sorder=_m_sorder,
    )
    _et = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    _ht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_num_py,),
    )
    try:
        _return_scalars = b_uni_garch_exp_estim(
            dist,
            _yt,
            _x,
            _num,
            ip,
            iq,
            _nreg,
            mn,
            _npar,
            _theta,
            _se,
            _sc,
            _covr,
            pht,
            _et,
            _ht,
            copts,
            maxit,
            tol,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _theta_py = _theta
    _pht_py = _return_scalars.pht
    _se_py = _se
    _sc_py = _sc
    _covr_py = _covr
    _et_py = _et
    _ht_py = _ht
    _lgf_py = _return_scalars.lgf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['theta', 'se', 'sc', 'covr', 'pht', 'et', 'ht', 'lgf'],
    )(
        _theta_py,
        _se_py,
        _sc_py,
        _covr_py,
        _pht_py,
        _et_py,
        _ht_py,
        _lgf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uni_garch_exp_forecast(nt, ip, iq, theta, ht, et):
    u"""
    Univariate time series, forecast function for an exponential GARCH
    (EGARCH) process.

    ``uni_garch_exp_forecast`` forecasts the conditional variances,
    h_t,t = T+1,...,T+xi from an exponential GARCH(p,q) sequence, where
    xi is the forecast horizon and T is the current time.

    For full information please refer to the NAG Library document for
    g13fh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13fhf.html

    Parameters
    ----------
    nt : int
        xi, the forecast horizon.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, array-like, shape (2*`iq`+`ip`+1)
        The initial parameter estimates for the vector theta. The first
        element must contain the coefficient alpha_o and the next `iq`
        elements must contain the autoregressive coefficients alpha_i,
        for i = 1,2,...,q. The next `iq` elements must contain the
        coefficients phi_i, for i = 1,2,...,q. The next `ip` elements
        must contain the moving average coefficients beta_j, for j =
        1,2,...,p.

    ht : float, array-like, shape (num)
        The sequence of past conditional variances for the GARCH(p,q)
        process, h_t, for t = 1,2,...,T.

    et : float, array-like, shape (num)
        The sequence of past residuals for the GARCH(p,q) process,
        epsilon_t, for t = 1,2,...,T.

    Returns
    -------
    fht : float, ndarray, shape (`nt`)
        The forecast values of the conditional variance, h_t, for t =
        T+1,...,T+xi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max(`ip`-1, `iq`-1) = *<value>*.

            Constraint: max(`ip`-1, `iq`-1) <= 20.

        (`errno` 1)
            On entry, num = *<value>*.

            Constraint: max(`ip`, `iq`) <= num.

        (`errno` 1)
            On entry, `nt` = *<value>*.

            Constraint: `nt` > 0.

        (`errno` 1)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    from ..base.tsa import uni_garch_exp_forecast as b_uni_garch_exp_forecast

    fun_name = 'naginterfaces.library.tsa.uni_garch_exp_forecast'

    _ht = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _num_py = _b_utils._EngineFloat64ArrayType.get_shape(_ht, {'fun_name': fun_name, 'entity': 'ht'}, exp_rank=1)[0]
    _num = _b_utils._EngineIntScalarType(
        _num_py, {'fun_name': fun_name, 'entity': '_num'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nt = _b_utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _nt_py = _b_utils._EngineIntScalarType.to_py(_nt)
    _fht = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nt_py,),
    )
    _et = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'et'},
        et,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    b_uni_garch_exp_forecast(
        _num,
        _nt,
        ip,
        iq,
        _theta,
        _fht,
        _ht,
        _et,
    )
    _fht_py = _fht
    return _fht_py

def inhom_iema(iema, t, tau, inter, sinit=None, pn=0, comm=None):
    u"""
    Computes the iterated exponential moving average for a univariate
    inhomogeneous time series.

    ``inhom_iema`` calculates the iterated exponential moving average
    for an inhomogeneous time series.

    For full information please refer to the NAG Library document for
    g13me

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13mef.html

    Parameters
    ----------
    iema : float, array-like, shape (nb)
        z_i, the current block of observations, for i = k+1,...,k+b,
        where k is the number of observations processed so far, i.e.,
        the value supplied in `pn` on entry.

    t : float, array-like, shape (nb)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 31 will be returned, but
        ``inhom_iema`` will continue as if t was strictly increasing by
        using the absolute value.

    tau : float
        tau, the parameter controlling the rate of decay, which must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau
        can be calculated without overflowing, for all i.

    inter : int, array-like, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

          Previous point, with nu = 1.

        `inter`[i] = 2

          Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

          Next point, nu = mu.

        Zumbach and Mueller recommend that linear interpolation is used
        in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    sinit : None or float, array-like, shape (m+2), optional
        If `pn` = 0, the values used to start the iterative process,
        with

        - `sinit`[0] = t_0,

        - `sinit`[1] = z_0,

        - `sinit`[j+1] = EMA[tau,j;z](t_0), for j = 1,2,...,m.

        If `pn` != 0, `sinit` is not referenced.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``inhom_iema``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to ``inhom_iema``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    iema : float, ndarray, shape (nb)
        The iterated EMA, with `iema`[i-1] = EMA[tau,m;z](t_i).

    pn : int
        k+b, the updated number of observations processed so far.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 32)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 41)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 42)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 52)
            On entry, m = *<value>*.

            On entry at previous call, m = *<value>*.

            Constraint: if `pn` > 0 then m must be unchanged since
            previous call.

        (`errno` 71)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 72)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 73)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the previous call.

        (`errno` 81)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 82)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 91)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 101)
            On entry, `pn` = 0, lrcomm = *<value>* and m = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= m+20.

        (`errno` 102)
            On entry, `pn` != 0, lrcomm = *<value>* and m = *<value>*.

            Constraint: if `pn` != 0, lrcomm >= m+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `iema` or for `tau`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.inhom_iema_ex.main`
    """

    from ..base.tsa import inhom_iema as b_inhom_iema

    fun_name = 'naginterfaces.library.tsa.inhom_iema'

    _iema = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iema'},
        iema,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_iema, {'fun_name': fun_name, 'entity': 'iema'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    if _pn_py != 0:
        _comm_needs_init = _b_utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('rcomm',),
        )
    else:
        _comm_needs_init = _b_utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('rcomm',),
            allow_none=True,
        )
    _sinit = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sinit'},
        sinit,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if sinit is None:
        _m = _b_utils._EngineIntScalarType(
            int(_b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)), {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
        _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    else:
        _m = _b_utils._EngineIntScalarType(
            _b_utils._EngineFloat64ArrayType.get_shape(_sinit, {'fun_name': fun_name, 'entity': 'sinit'}, exp_rank=1, allow_none=True)[0]-2, {'fun_name': fun_name, 'entity': '_m'},
        ).to_ctypes()
        _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _inter = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inter'},
        inter,
        copy=False,
    )
    try:
        _return_scalar = b_inhom_iema(
            _nb,
            _iema,
            _t,
            tau,
            _m,
            _sinit,
            _inter,
            _pn,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _iema_py = _iema
    _pn_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['iema', 'pn'],
    )(
        _iema_py,
        _pn_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inhom_iema_all(z, t, tau, m1, inter, ftype, p, sorder=1, sinit=None, x=None, pn=0, comm=None):
    u"""
    Computes the iterated exponential moving average for a univariate
    inhomogeneous time series, intermediate results are also returned.

    ``inhom_iema_all`` calculates the iterated exponential moving
    average for an inhomogeneous time series, returning the intermediate
    results.

    For full information please refer to the NAG Library document for
    g13mf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13mff.html

    Parameters
    ----------
    z : float, array-like, shape (nb)
        z_i, the current block of observations, for i = k+1,...,k+b,
        where k is the number of observations processed so far, i.e.,
        the value supplied in `pn` on entry.

    t : float, array-like, shape (nb)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 61 will be returned, but
        ``inhom_iema_all`` will continue as if t was strictly increasing
        by using the absolute value.

    tau : float
        tau, the parameter controlling the rate of decay. tau must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau
        can be calculated without overflowing, for all i.

    m1 : int
        The minimum number of times the EMA operator is to be iterated.

    inter : int, array-like, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

          Previous point, with nu = 1.

        `inter`[i] = 2

          Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

          Next point, nu = mu.

        Zumbach and Mueller recommend that linear interpolation is used
        in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    ftype : int
        The function type used to define the relationship between y and
        z when calculating EMA[tau,1;y]. Three functions are provided:

        `ftype` = 1

          The identity function, with y_i = z_i^[p].

        `ftype` = 2

          The absolute value, with y_i = |z_i|^p.

        `ftype` = 3

          The absolute difference, with y_i = |z_i-x_i|^p, where the
          vector x is supplied in `x`.

    p : float
        p, the power used in the transformation function.

    sorder : int, optional
        Determines the storage order of output returned in `iema`.

    sinit : None or float, array-like, shape (m2+2), optional
        If `pn` = 0, the values used to start the iterative process,
        with

        - `sinit`[0] = t_0,

        - `sinit`[1] = y_0,

        - `sinit`[j+1] = EMA[tau,j;y](t_0), j = 1,2,...,m2.

        If `pn` != 0 then `sinit` is not referenced.

    x : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ftype`=3: nb; otherwise: 0.

        If `ftype` = 3, x_i, the vector used to shift the current block
        of observations, for i = k+1,...,k+b, where k is the number of
        observations processed so far, i.e., the value supplied in `pn`
        on entry.

        If `ftype` != 3 then `x` is not referenced.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``inhom_iema_all``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to
        ``inhom_iema_all``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    iema : float, ndarray, shape (:, :)
        The iterated exponential moving average.

        If `sorder` = 1, `iema`[i-1,j-1] = EMA[tau,j+`m1`-1;y](t_{i+k}).

        If `sorder` = 2, `iema`[j-1,i-1] = EMA[tau,j+`m1`-1;y](t_{i+k}).

        For i = 1,2,...,nb, j = 1,2,...,m2-`m1`+1 and k is the number of
        observations processed so far, i.e., the value supplied in `pn`
        on entry.

    p : float
        If `ftype` = 1, then [p], the actual power used in the
        transformation function is returned, otherwise `p` is unchanged.

    pn : int
        k+b, the updated number of observations processed so far.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `sorder` = *<value>*.

            Constraint: `sorder` = 1 or 2.

        (`errno` 21)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 51)
            On entry, `sorder` = 1, ldiema = *<value>* and nb =
            *<value>*.

            Constraint: ldiema >= nb.

        (`errno` 51)
            On entry, `sorder` = 2, ldiema = *<value>* and m2-`m1`+1 =
            *<value>*.

            Constraint: ldiema >= m2-`m1`+1.

        (`errno` 62)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 71)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 72)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 81)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 82)
            On entry, `m1` = *<value>*.

            On entry at previous call, `m1` = *<value>*.

            Constraint: if `pn` > 0 then `m1` must be unchanged since
            previous call.

        (`errno` 91)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: m2 >= `m1`.

        (`errno` 92)
            On entry, m2 = *<value>*.

            On entry at previous call, m2 = *<value>*.

            Constraint: if `pn` > 0 then m2 must be unchanged since
            previous call.

        (`errno` 101)
            On entry, `ftype` != 1, j = *<value>* and `sinit`[j-1] =
            *<value>*.

            Constraint: if `ftype` != 1, `sinit`[j-1] >= 0.0, for j =
            2,3,...,m2+2.

        (`errno` 111)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 112)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 113)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the last call.

        (`errno` 121)
            On entry, `ftype` = *<value>*.

            Constraint: `ftype` = 1, 2 or 3.

        (`errno` 122)
            On entry, `ftype` = *<value>*, On entry at previous call,
            `ftype` = *<value>*.

            Constraint: if `pn` != 0, `ftype` must be unchanged since
            the previous call.

        (`errno` 131)
            On entry, `p` = *<value>*.

            Constraint: absolute value of `p` must be representable as
            an integer.

        (`errno` 132)
            On entry, `p` = *<value>*.

            Constraint: if `ftype` != 1, `p` != 0.0. If `ftype` = 1, the
            nearest integer to `p` must not be 0.

        (`errno` 133)
            On entry, i = *<value>*, `z`[i-1] = *<value>* and `p` =
            *<value>*.

            Constraint: if `ftype` = 1 or 2 and `z`[i] = 0 for any i
            then `p` > 0.0.

        (`errno` 134)
            On entry, i = *<value>*, `z`[i-1] = *<value>*, `x`[i-1] =
            *<value>* and `p` = *<value>*.

            Constraint: if `ftype` = 3 and `z`[i] = `x`[i] for any i
            then `p` > 0.0.

        (`errno` 135)
            On entry, `p` = *<value>*.

            On exit from previous call, `p` = *<value>*.

            Constraint: if `pn` > 0 then `p` must be unchanged since
            previous call.

        (`errno` 151)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 152)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 161)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 171)
            On entry, `pn` = 0, lrcomm = *<value>* and m2 = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= m2+20.

        (`errno` 172)
            On entry, `pn` != 0, lrcomm = *<value>* and m2 = *<value>*.

            Constraint: if `pn` != 0 then lrcomm >= m2+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 61)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `z`, `x` or for `tau`.
    """

    from ..base.tsa import inhom_iema_all as b_inhom_iema_all

    fun_name = 'naginterfaces.library.tsa.inhom_iema_all'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _sinit = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sinit'},
        sinit,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if sinit is None:
        _m2 = 1
        _m2_py = _m2
    else:
        _m2 = _b_utils._EngineIntScalarType(
            _b_utils._EngineFloat64ArrayType.get_shape(_sinit, {'fun_name': fun_name, 'entity': 'sinit'}, exp_rank=1, allow_none=True)[0]-2, {'fun_name': fun_name, 'entity': '_m2'},
        ).to_ctypes()
        _m2_py = _b_utils._EngineIntScalarType.to_py(_m2)
    _sorder = _b_utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _sorder_py = _b_utils._EngineIntScalarType.to_py(_sorder)
    _m1 = _b_utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m1_py = _b_utils._EngineIntScalarType.to_py(_m1)
    if _sorder_py == 1:
        _iema_dim1 = _nb_py
    else:
        _iema_dim1 = _m2_py-_m1_py+1
    if _sorder_py == 1:
        _iema_dim2 = _m2_py-_m1_py+1
    elif _sorder_py == 2:
        _iema_dim2 = _nb_py
    else:
        _iema_dim2 = 0
    exp_shape_iema = (_iema_dim1, _iema_dim2)
    _iema = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_iema,
        sorder=_b_utils._EngineData.default_sorder,
    )
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _inter = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inter'},
        inter,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    try:
        _return_scalars = b_inhom_iema_all(
            _sorder,
            _nb,
            _z,
            _iema,
            _t,
            tau,
            _m1,
            _m2,
            _sinit,
            _inter,
            ftype,
            p,
            _x,
            _pn,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _p_py = _return_scalars.p
    _pn_py = _return_scalars.pn
    _iema_py = _iema
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['iema', 'p', 'pn'],
    )(
        _iema_py,
        _p_py,
        _pn_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def inhom_ma(ma, t, tau, m1, m2, inter, ftype, p, sinit=None, pn=0, comm=None):
    u"""
    Computes the exponential moving average for a univariate
    inhomogeneous time series.

    ``inhom_ma`` provides a moving average, moving norm, moving variance
    and moving standard deviation operator for an inhomogeneous time
    series.

    For full information please refer to the NAG Library document for
    g13mg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13mgf.html

    Parameters
    ----------
    ma : float, array-like, shape (nb)
        z_i, the current block of observations, for i = k+1,...,k+b,
        where k is the number of observations processed so far, i.e.,
        the value supplied in `pn` on entry.

    t : float, array-like, shape (nb)
        t_i, the times for the current block of observations, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far, i.e., the value supplied in `pn` on entry.

        If t_i <= t_{i-1}, `errno` = 31 will be returned, but
        ``inhom_ma`` will continue as if t was strictly increasing by
        using the absolute value.

        The lagged difference, t_i-t_{i-1} must be sufficiently small
        that e^(-alpha), alpha = (t_i-t_{i-1})/tau~ can be calculated
        without overflowing, for all i.

    tau : float
        tau, the parameter controlling the rate of decay. tau must be
        sufficiently large that e^(-alpha), alpha = (t_i-t_{i-1})/tau~
        can be calculated without overflowing, for all i, where tau~ =
        2*tau/(m_2+m_1).

    m1 : int
        m_1, the iteration of the EMA operator at which the sum is
        started.

    m2 : int
        m_2, the iteration of the EMA operator at which the sum is
        ended.

    inter : int, array-like, shape (2)
        The type of interpolation used with `inter`[0] indicating the
        interpolation method to use when calculating EMA[tau,1;z] and
        `inter`[1] the interpolation method to use when calculating
        EMA[tau,j;z], j > 1.

        Three types of interpolation are possible:

        `inter`[i] = 1

          Previous point, with nu = 1.

        `inter`[i] = 2

          Linear, with nu = (1-mu)/alpha.

        `inter`[i] = 3

          Next point, nu = mu.

        Zumbach and Mueller recommend that linear interpolation is used
        in second and subsequent iterations, i.e., `inter`[1] = 2,
        irrespective of the interpolation method used at the first
        iteration, i.e., the value of `inter`[0].

    ftype : int
        The function type used to define the relationship between y and
        z when calculating EMA[tau,1;y]. Three functions are provided:

        `ftype` = 1

          The identity function, with y_i = z_i^[p].

        `ftype` = 2 or 4

          The absolute value, with y_i = |z_i|^p.

        `ftype` = 3 or 5

          The absolute difference, with y_i = |z_i-MA[tau,m;y](t_i)|^p.

        If `ftype` = 4 or 5 then the resulting vector of averages is
        scaled by p^(-1) as described in `ma`.

    p : float
        p, the power used in the transformation function.

    sinit : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ftype` in (3, 5): 2*`m2`+3; otherwise: `m2`+2.

        If `pn` = 0, the values used to start the iterative process,
        with

        - `sinit`[0] = t_0,

        - `sinit`[1] = y_0,

        - `sinit`[j+1] = EMA[tau,j;y](t_0), for i = 1,2,...,`m2`.

        In addition, if `ftype` = 3 or 5 then

        - `sinit`[`m2`+2] = z_0,

        - `sinit`[`m2`+j+1] = EMA[tau,j;z](t_0), for j = 1,2,...,`m2`.

        i.e., initial values based on the original data z as opposed to
        the transformed data y.

        If `pn` != 0, `sinit` is not referenced.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``inhom_ma``, or when starting to summarise a new
        dataset, `pn` must be set to 0. On subsequent calls it must be
        the same value as returned by the last call to ``inhom_ma``.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        On initial entry: need not be set.

    Returns
    -------
    ma : float, ndarray, shape (nb)
        The moving average:

        if `ftype` = 4 or 5

          `ma`[i-1] = {MA[tau,m_1,m_2;y](t_i)}^(1/p),

        otherwise

          `ma`[i-1] = MA[tau,m_1,m_2;y](t_i).

    p : float
        If `ftype` = 1, then [p], the actual power used in the
        transformation function is returned, otherwise `p` is unchanged.

    pn : int
        k+b, the updated number of observations processed so far.

    wma : float, ndarray, shape (nb)
        Either the moving average or exponential moving average,
        depending on the value of `ftype`.

        if `ftype` = 3 or 5

          `wma`[i-1] = MA[tau;y](t_i)

        otherwise

          `wma`[i-1] = EMA[tau~;y](t_i).

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 32)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t`[i-1] != `t`[i-2] if linear interpolation is
            being used.

        (`errno` 41)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 42)
            On entry, `tau` = *<value>*.

            On entry at previous call, `tau` = *<value>*.

            Constraint: if `pn` > 0 then `tau` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 52)
            On entry, `m1` = *<value>*.

            On entry at previous call, `m1` = *<value>*.

            Constraint: if `pn` > 0 then `m1` must be unchanged since
            previous call.

        (`errno` 61)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m2` >= `m1`.

        (`errno` 62)
            On entry, `m2` = *<value>*.

            On entry at previous call, `m2` = *<value>*.

            Constraint: if `pn` > 0 then `m2` must be unchanged since
            previous call.

        (`errno` 71)
            On entry, `ftype` != 1, j = *<value>* and `sinit`[j-1] =
            *<value>*.

            Constraint: if `ftype` != 1, `sinit`[j-1] >= 0.0, for j =
            2,3,...,`m2`+2.

        (`errno` 81)
            On entry, `inter`[0] = *<value>*.

            Constraint: `inter`[0] = 1, 2 or 3.

        (`errno` 82)
            On entry, `inter`[1] = *<value>*.

            Constraint: `inter`[1] = 1, 2 or 3.

        (`errno` 83)
            On entry, `inter`[0] = *<value>* and `inter`[1] = *<value>*.

            On entry at previous call, `inter`[0] = *<value>*,
            `inter`[1] = *<value>*.

            Constraint: if `pn` != 0, `inter` must be unchanged since
            the last call.

        (`errno` 91)
            On entry, `ftype` = *<value>*.

            Constraint: `ftype` = 1, 2, 3, 4 or 5.

        (`errno` 92)
            On entry, `ftype` = *<value>*, On entry at previous call,
            `ftype` = *<value>*.

            Constraint: if `pn` != 0, `ftype` must be unchanged since
            the previous call.

        (`errno` 101)
            On entry, `p` = *<value>*.

            Constraint: absolute value of `p` must be representable as
            an integer.

        (`errno` 102)
            On entry, `p` = *<value>*.

            Constraint: if `ftype` != 1, `p` != 0.0. If `ftype` = 1, the
            nearest integer to `p` must not be 0.

        (`errno` 103)
            On entry, i = *<value>*, `ma`[i-1] = *<value>* and `p` =
            *<value>*.

            Constraint: if `ftype` = 1, 2 or 4 and `ma`[i-1] = 0 for any
            i then `p` > 0.0.

        (`errno` 104)
            On entry, i = *<value>*, `ma`[i-1] = *<value>*, `wma`[i-1] =
            *<value>* and `p` = *<value>*.

            Constraint: if `p` < 0.0, `ma`[i-1]-`wma`[i-1] != 0.0, for
            any i.

        (`errno` 105)
            On entry, `p` = *<value>*.

            On exit from previous call, `p` = *<value>*.

            Constraint: if `pn` > 0 then `p` must be unchanged since
            previous call.

        (`errno` 111)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 112)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0 then `pn` must be unchanged since
            previous call.

        (`errno` 131)
            `comm`['rcomm'] has been corrupted between calls.

        (`errno` 141)
            On entry, `pn` = 0, lrcomm = *<value>* and `m2` = *<value>*.

            Constraint: if `pn` = 0, lrcomm = 0 or lrcomm >= 2*`m2`+20.

        (`errno` 142)
            On entry, `pn` != 0, lrcomm = *<value>* and `m2` =
            *<value>*.

            Constraint: if `pn` != 0, lrcomm >= 2*`m2`+20.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, i = *<value>*, `t`[i-2] = *<value>* and `t`[i-1] =
            *<value>*.

            Constraint: `t` should be strictly increasing.

        (`errno` 301)
            Truncation occurred to avoid overflow, check for extreme
            values in `t`, `ma` or for `tau`.
    """

    from ..base.tsa import inhom_ma as b_inhom_ma

    fun_name = 'naginterfaces.library.tsa.inhom_ma'

    _ma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ma'},
        ma,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_ma, {'fun_name': fun_name, 'entity': 'ma'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sinit = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sinit'},
        sinit,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _inter = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inter'},
        inter,
        copy=False,
    )
    _wma = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nb_py,),
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    try:
        _return_scalars = b_inhom_ma(
            _nb,
            _ma,
            _t,
            tau,
            m1,
            m2,
            _sinit,
            _inter,
            ftype,
            p,
            _pn,
            _wma,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ma_py = _ma
    _p_py = _return_scalars.p
    _pn_py = _return_scalars.pn
    _wma_py = _wma
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ma', 'p', 'pn', 'wma'],
    )(
        _ma_py,
        _p_py,
        _pn_py,
        _wma_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def cp_pelt(ctype, y, beta=None, minss=2, param=None):
    u"""
    Change point detection, using the PELT algorithm.

    ``cp_pelt`` detects change points in a univariate time series, that
    is, the time points at which some feature of the data, for example
    the mean, changes.
    Change points are detected using the PELT (Pruned Exact Linear Time)
    algorithm using one of a provided set of cost functions.

    For full information please refer to the NAG Library document for
    g13na

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13naf.html

    Parameters
    ----------
    ctype : int
        A flag indicating the assumed distribution of the data and the
        type of change point being looked for.

        `ctype` = 1

          Data from a Normal distribution, looking for changes in the
          mean, mu.

        `ctype` = 2

          Data from a Normal distribution, looking for changes in the
          standard deviation sigma.

        `ctype` = 3

          Data from a Normal distribution, looking for changes in the
          mean, mu and standard deviation sigma.

        `ctype` = 4

          Data from a Gamma distribution, looking for changes in the
          scale parameter b.

        `ctype` = 5

          Data from an exponential distribution, looking for changes in
          lambda.

        `ctype` = 6

          Data from a Poisson distribution, looking for changes in
          lambda.

    y : float, array-like, shape (n)
        y, the time series.

    beta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if `ctype` = 3: 2*log(`n`);
        otherwise: log(`n`).

        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n)

        AIC

          beta = 2*p

        Hannan-Quinn

          beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        This is usually set to 2 when `ctype` = 3 and 1 otherwise.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    param : None or float, array-like, shape (1), optional
        phi, values for the parameters that will be treated as fixed. If
        `ctype` = 4 then `param` must be supplied.

        If supplied, then when

        `ctype` = 1

          `param`[0] = sigma, the standard deviation of the normal
          distribution. If not supplied then sigma is estimated from the
          full input data,

        `ctype` = 2

          `param`[0] = mu, the mean of the normal distribution. If not
          supplied then mu is estimated from the full input data,

        `ctype` = 4

          `param`[0] must hold the shape, a, for the gamma distribution,

        otherwise

          `param` is not referenced.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    sparam : float, ndarray, shape (:) or (:, :)
        The estimated values of the distribution parameters in each
        segment

        `ctype` = 1, 2 or 3

          `sparam`[0,i-1] = mu_i and `sparam`[1,i-1] = sigma_i for i =
          1,2,...,m, where mu_i and sigma_i is the mean and standard
          deviation, respectively, of the values of y in the ith
          segment.

          It should be noted that sigma_i = sigma_j when `ctype` = 1 and
          mu_i = mu_j when `ctype` = 2, for all i and j.

        `ctype` = 4

          `sparam`[0,i-1] = a_i and `sparam`[1,i-1] = b_i for i =
          1,2,...,m, where a_i and b_i are the shape and scale
          parameters, respectively, for the values of y in the ith
          segment. It should be noted that a_i = `param`[0] for all i.

        `ctype` = 5 or 6

          `sparam`[i-1] = lambda_i for i = 1,2,...,m, where lambda_i is
          the mean of the values of y in the ith segment.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ctype` = *<value>*.

            Constraint: `ctype` = 1, 2, 3, 4, 5 or 6.

        (`errno` 21)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 31)
            On entry, `ctype` = *<value>* and `y`[*<value>*] =
            *<value>*.

            Constraint: if `ctype` = 4, 5 or 6 then `y`[i-1] >= 0.0, for
            i = 1,2,...,n.

        (`errno` 32)
            On entry, `y`[*<value>*] = *<value>*, is too large.

        (`errno` 51)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 71)
            On entry, `ctype` = *<value>* and `param`[0] = *<value>*.

            Constraint: if `ctype` = 1 or 4 and `param` has been
            supplied, then `param`[0] > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 200)
            To avoid overflow some truncation occurred when calculating
            the cost function, C. All output is returned as normal.

        (`errno` 201)
            To avoid overflow some truncation occurred when calculating
            the parameter estimates returned in `sparam`. All output is
            returned as normal.

    See Also
    --------
    :meth:`naginterfaces.library.examples.tsa.cp_pelt_ex.main`
    """

    from ..base.tsa import cp_pelt as b_cp_pelt

    fun_name = 'naginterfaces.library.tsa.cp_pelt'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ctype = _b_utils._EngineIntScalarType(
        ctype, {'fun_name': fun_name, 'entity': 'ctype'},
    ).to_ctypes()
    _ctype_py = _b_utils._EngineIntScalarType.to_py(_ctype)
    if _ctype_py == 3:
        _beta = 2*log(_n_py)
    else:
        _beta = log(_n_py)
    _beta = _beta if beta is None else beta
    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _sparam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py+2,),
    )
    try:
        _return_scalar = b_cp_pelt(
            _ctype,
            _n,
            _y,
            _beta,
            minss,
            _param,
            _tau,
            _sparam,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    if _ctype_py in (5, 6):
        _sparam_dim1 = _ntau_py
    else:
        _sparam_dim1 = 2
    if _ctype_py in (5, 6):
        _sparam_dim2 = None
    else:
        _sparam_dim2 = _ntau_py
    if _sparam_dim2 is None:
        _sparam_py = _sparam[:_sparam_dim1]
    else:
        from numpy import ascontiguousarray
        _sparam_py = ascontiguousarray(_sparam[:_sparam_dim1*_sparam_dim2].reshape((_sparam_dim1, _sparam_dim2), order='F'))
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ntau', 'tau', 'sparam'],
    )(
        _ntau_py,
        _tau_py,
        _sparam_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def cp_pelt_user(n, beta, k, costfn, minss=2, data=None):
    u"""
    Change points detection using the PELT algorithm, user supplied cost
    function.

    ``cp_pelt_user`` detects change points in a univariate time series,
    that is, the time points at which some feature of the data, for
    example the mean, changes.
    Change points are detected using the PELT (Pruned Exact Linear Time)
    algorithm and a user-supplied cost function.

    For full information please refer to the NAG Library document for
    g13nb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13nbf.html

    Parameters
    ----------
    n : int
        n, the length of the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n)

        AIC

          beta = 2*p

        Hannan-Quinn

          beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        The value of p will depend on the cost function being used.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    k : float
        K, the constant value that satisfies equation [equation]. If K
        exists, it is unlikely to be unique in such cases, it is
        recommened that the largest value of K, that satisfies equation
        [equation], is chosen. No check is made that K is the correct
        value for the supplied cost function.

    costfn : callable c = costfn(ts, r, data=None)
        The cost function, C. `costfn` must calculate a vector of costs
        for a number of segments.

        Parameters
        ~~~~~~~~~~
        ts : int
            A reference time point.

        r : int, ndarray, shape (nr)
            Time points which, along with `ts`, define the segments
            being considered, 0 <= `r`[i-1] <= n for i = 1,2,...nr.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        c : float, array-like, shape (nr)
            On exit, the cost function, C.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 51)
            User requested termination.
    """

    from ..base.tsa import cp_pelt_user as b_cp_pelt_user

    fun_name = 'naginterfaces.library.tsa.cp_pelt_user'

    cb_costfn_name = fun_name + ':costfn'

    def cb_costfn(ts, nr, r, c, data=None):
        """Wrapping user's costfn. Called as base costfn."""
        if data is not None:
            _c_costfn = costfn(
                ts,
                r,
                data,
            )
        else:
            _c_costfn = costfn(
                ts,
                r,
            )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_costfn_name, 'returned': 'c'},
            _c_costfn,
            copy=False,
            exp_shape=(nr,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(costfn, {'fun_name': fun_name, 'entity': 'costfn'})
    _costfn = cb_costfn
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = b_cp_pelt_user(
            _n,
            beta,
            minss,
            k,
            _costfn,
            _tau,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ntau', 'tau'],
    )(
        _ntau_py,
        _tau_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def cp_binary(ctype, y, beta=None, minss=2, param=None, mdepth=0):
    u"""
    Change point detection, using binary segmentation.

    ``cp_binary`` detects change points in a univariate time series,
    that is, the time points at which some feature of the data, for
    example the mean, changes.
    Change points are detected using binary segmentation using one of a
    provided set of cost functions.

    For full information please refer to the NAG Library document for
    g13nd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13ndf.html

    Parameters
    ----------
    ctype : int
        A flag indicating the assumed distribution of the data and the
        type of change point being looked for.

        `ctype` = 1

          Data from a Normal distribution, looking for changes in the
          mean, mu.

        `ctype` = 2

          Data from a Normal distribution, looking for changes in the
          standard deviation sigma.

        `ctype` = 3

          Data from a Normal distribution, looking for changes in the
          mean, mu and standard deviation sigma.

        `ctype` = 4

          Data from a Gamma distribution, looking for changes in the
          scale parameter b.

        `ctype` = 5

          Data from an exponential distribution, looking for changes in
          lambda.

        `ctype` = 6

          Data from a Poisson distribution, looking for changes in
          lambda.

    y : float, array-like, shape (n)
        y, the time series.

    beta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if `ctype` = 3: 2*log(`n`);
        otherwise: log(`n`).

        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n)

        AIC

          beta = 2*p

        Hannan-Quinn

          beta = 2*p*log(log(n))

        where p is the number of parameters being treated as estimated
        in each segment.

        This is usually set to 2 when `ctype` = 3 and 1 otherwise.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    param : None or float, array-like, shape (1), optional
        phi, values for the parameters that will be treated as fixed. If
        `ctype` = 4 then `param` must be supplied.

        If supplied, then when

        `ctype` = 1

          `param`[0] = sigma, the standard deviation of the normal
          distribution. If not supplied then sigma is estimated from the
          full input data,

        `ctype` = 2

          `param`[0] = mu, the mean of the normal distribution. If not
          supplied then mu is estimated from the full input data,

        `ctype` = 4

          `param`[0] must hold the shape, a, for the gamma distribution,

        otherwise

          `param` is not referenced.

    mdepth : int, optional
        K, the maximum depth for the iterative process, which in turn
        puts an upper limit on the number of change points with m <=
        2^K.

        If K <= 0 then no limit is put on the depth of the iterative
        process and no upper limit is put on the number of change
        points, other than that inherent in the length of the series and
        the value of `minss`.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    sparam : float, ndarray, shape (:) or (:, :)
        The estimated values of the distribution parameters in each
        segment

        `ctype` = 1, 2 or 3

          `sparam`[0,i-1] = mu_i and `sparam`[1,i-1] = sigma_i for i =
          1,2,...,m, where mu_i and sigma_i is the mean and standard
          deviation, respectively, of the values of y in the ith
          segment.

          It should be noted that sigma_i = sigma_j when `ctype` = 1 and
          mu_i = mu_j when `ctype` = 2, for all i and j.

        `ctype` = 4

          `sparam`[0,i-1] = a_i and `sparam`[1,i-1] = b_i for i =
          1,2,...,m, where a_i and b_i are the shape and scale
          parameters, respectively, for the values of y in the ith
          segment. It should be noted that a_i = `param`[0] for all i.

        `ctype` = 5 or 6

          `sparam`[i-1] = lambda_i for i = 1,2,...,m, where lambda_i is
          the mean of the values of y in the ith segment.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ctype` = *<value>*.

            Constraint: `ctype` = 1, 2, 3, 4, 5 or 6.

        (`errno` 21)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 31)
            On entry, `ctype` = *<value>* and `y`[*<value>*] =
            *<value>*.

            Constraint: if `ctype` = 4, 5 or 6 then `y`[i-1] >= 0.0, for
            i = 1,2,...,n.

        (`errno` 32)
            On entry, `y`[*<value>*] = *<value>*, is too large.

        (`errno` 51)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 71)
            On entry, `ctype` = *<value>* and `param`[0] = *<value>*.

            Constraint: if `ctype` = 1 or 4 and `param` has been
            supplied, then `param`[0] > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 200)
            To avoid overflow some truncation occurred when calculating
            the cost function, C. All output is returned as normal.

        (`errno` 201)
            To avoid overflow some truncation occurred when calculating
            the parameter estimates returned in `sparam`. All output is
            returned as normal.
    """

    from ..base.tsa import cp_binary as b_cp_binary

    fun_name = 'naginterfaces.library.tsa.cp_binary'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ctype = _b_utils._EngineIntScalarType(
        ctype, {'fun_name': fun_name, 'entity': 'ctype'},
    ).to_ctypes()
    _ctype_py = _b_utils._EngineIntScalarType.to_py(_ctype)
    if _ctype_py == 3:
        _beta = 2*log(_n_py)
    else:
        _beta = log(_n_py)
    _beta = _beta if beta is None else beta
    _param = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'param'},
        param,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mdepth = _b_utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    _mdepth_py = _b_utils._EngineIntScalarType.to_py(_mdepth)
    _minss = _b_utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    _minss_py = _b_utils._EngineIntScalarType.to_py(_minss)
    if _mdepth_py > 0:
        _tau_dim1 = min(int(ceil(float(_n_py)/float(_minss_py))), 2**_mdepth_py)
    else:
        _tau_dim1 = int(ceil(float(_n_py)/float(_minss_py)))
    exp_shape_tau = (_tau_dim1,)
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        exp_shape_tau,
    )
    _sparam = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py,),
    )
    try:
        _return_scalar = b_cp_binary(
            _ctype,
            _n,
            _y,
            _beta,
            _minss,
            _param,
            _mdepth,
            _tau,
            _sparam,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    if _ctype_py in (5, 6):
        _sparam_dim1 = _ntau_py
    else:
        _sparam_dim1 = 2
    if _ctype_py in (5, 6):
        _sparam_dim2 = None
    else:
        _sparam_dim2 = _ntau_py
    if _sparam_dim2 is None:
        _sparam_py = _sparam[:_sparam_dim1]
    else:
        from numpy import ascontiguousarray
        _sparam_py = ascontiguousarray(_sparam[:_sparam_dim1*_sparam_dim2].reshape((_sparam_dim1, _sparam_dim2), order='F'))
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ntau', 'tau', 'sparam'],
    )(
        _ntau_py,
        _tau_py,
        _sparam_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def cp_binary_user(n, beta, chgpfn, minss=2, mdepth=0, data=None):
    u"""
    Change point detection, using binary segmentation, user supplied
    cost function.

    ``cp_binary_user`` detects change points in a univariate time
    series, that is, the time points at which some feature of the data,
    for example the mean, changes.
    Change points are detected using binary segmentation for a
    user-supplied cost function.

    For full information please refer to the NAG Library document for
    g13ne

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g13/g13nef.html

    Parameters
    ----------
    n : int
        n, the length of the time series.

    beta : float
        beta, the penalty term.

        There are a number of standard ways of setting beta, including:

        SIC or BIC

          beta = p*log(n).

        AIC

          beta = 2*p.

        Hannan-Quinn

          beta = 2*p*log(log(n)).

        where p is the number of parameters being treated as estimated
        in each segment.

        The value of p will depend on the cost function being used.

        If no penalty is required then set beta = 0.

        Generally, the smaller the value of beta the larger the number
        of suggested change points.

    chgpfn : callable (v, cost, info) = chgpfn(side, u, w, minss, info,
        data=None)
        `chgpfn` must calculate a proposed change point, and the
        associated costs, within a specified segment.

        Parameters
        ~~~~~~~~~~
        side : int
            Flag indicating what `chgpfn` must calculate and at which
            point of the Binary Segmentation it has been called.

            `side` = -1

              only C(y_{u:w}) need be calculated and returned in
              `cost`[0], neither `v` nor the other elements of `cost`
              need be set. In this case, u = 1 and w = n.

            `side` = 0

              all elements of `cost` and `v` must be set. In this case,
              u = 1 and w = n.

            `side` = 1

              the segment, y_{u:w}, is a left-hand side subsegment from
              a previous iteration of the Binary Segmentation algorithm.
              All elements of `cost` and `v` must be set.

            `side` = 2

              the segment, y_{u:w}, is a right-hand side subsegment from
              a previous iteration of the Binary Segmentation algorithm.
              All elements of `cost` and `v` must be set.

            The distinction between `side` = 1 or 2 may allow for
            `chgpfn` to be implemented in a more efficient manner.

            See [Example] for one such example.

            The first call to `chgpfn` will always have `side` = -1 and
            the second call will always have `side` = 0.

            All subsequent calls will be made with `side` = 1 or 2.

        u : int
            u, the start of the segment of interest.

        w : int
            w, the end of the segment of interest.

        minss : int
            The minimum distance between two change points, as passed to
            ``cp_binary_user``.

        info : int
            `info` = 0.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        v : int
            If `side` = -1 then `v` need not be set.

            if `side` != -1 then v, the proposed change point.

            That is, the value which minimizes

                minimize_v C(y_{u:v})+C(y_{v+1:w})

            for v = u+`minss`-1 to w-`minss`.

        cost : float, array-like, shape (3)
            Costs associated with the proposed change point, v.

            If `side` = -1 then `cost`[0] = C(y_{u:w}) and the remaining
            two elements of `cost` need not be set.

            If `side` != -1 then

            - `cost`[0] = C(y_{u:v})+C(y_{v+1:w}).

            - `cost`[1] = C(y_{u:v}).

            - `cost`[2] = C(y_{v+1:w}).

        info : int
            In most circumstances `info` should remain unchanged.

            If `info` is set to a strictly positive value then
            ``cp_binary_user`` terminates with `errno` = 51.

            If `info` is set to a strictly negative value the current
            segment is skipped and ``cp_binary_user`` continues as
            normal.

            If `info` was set to a strictly negative value at any point
            and no other errors occur then ``cp_binary_user`` will
            terminate with `errno` = 52.

    minss : int, optional
        The minimum distance between two change points, that is
        tau_i-tau_{i-1} >= `minss`.

    mdepth : int, optional
        K, the maximum depth for the iterative process, which in turn
        puts an upper limit on the number of change points with m <=
        2^K.

        If K <= 0 then no limit is put on the depth of the iterative
        process and no upper limit is put on the number of change
        points, other than that inherent in the length of the series and
        the value of `minss`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    ntau : int
        m, the number of change points detected.

    tau : int, ndarray, shape (`ntau`)
        The first m elements of `tau` hold the location of the change
        points. The ith segment is defined by y_(tau_{i-1}+1) to
        y_tau_i, where tau_0 = 0 and tau_i = `tau`[i-1],1 <= i <= m.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 31)
            On entry, `minss` = *<value>*.

            Constraint: `minss` >= 2.

        (`errno` 51)
            User requested termination by setting `info` = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 52)
            User requested a segment to be skipped by setting `info` =
            *<value>*.
    """

    from ..base.tsa import cp_binary_user as b_cp_binary_user

    fun_name = 'naginterfaces.library.tsa.cp_binary_user'

    cb_chgpfn_name = fun_name + ':chgpfn'

    def cb_chgpfn(side, u, w, minss, cost, info, data=None):
        """Wrapping user's chgpfn. Called as base chgpfn."""
        if data is not None:
            _v_chgpfn, _cost_chgpfn, _info_chgpfn = chgpfn(
                side,
                u,
                w,
                minss,
                info,
                data,
            )
        else:
            _v_chgpfn, _cost_chgpfn, _info_chgpfn = chgpfn(
                side,
                u,
                w,
                minss,
                info,
            )
        cost[:] = _utils._to_ndarray(
            {'fun_name': cb_chgpfn_name, 'returned': 'cost'},
            _cost_chgpfn,
            copy=False,
            exp_shape=(3,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _info_chgpfn_base = _info_chgpfn
        _v_chgpfn_base = _v_chgpfn
        return (
            _v_chgpfn_base,
            _info_chgpfn_base,
        )

    _b_utils._check_callable(chgpfn, {'fun_name': fun_name, 'entity': 'chgpfn'})
    _chgpfn = cb_chgpfn
    _mdepth = _b_utils._EngineIntScalarType(
        mdepth, {'fun_name': fun_name, 'entity': 'mdepth'},
    ).to_ctypes()
    _mdepth_py = _b_utils._EngineIntScalarType.to_py(_mdepth)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _minss = _b_utils._EngineIntScalarType(
        minss, {'fun_name': fun_name, 'entity': 'minss'},
    ).to_ctypes()
    _minss_py = _b_utils._EngineIntScalarType.to_py(_minss)
    if _mdepth_py > 0:
        _tau_dim1 = min(int(ceil(float(_n_py)/float(_minss_py))), 2**_mdepth_py)
    else:
        _tau_dim1 = int(ceil(float(_n_py)/float(_minss_py)))
    exp_shape_tau = (_tau_dim1,)
    _tau = _b_utils._EngineIntArrayType.empty_ndarray(
        exp_shape_tau,
    )
    try:
        _return_scalar = b_cp_binary_user(
            _n,
            beta,
            _minss,
            _mdepth,
            _chgpfn,
            _tau,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ntau_py = _return_scalar
    _tau_py = _tau[:_ntau_py]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ntau', 'tau'],
    )(
        _ntau_py,
        _tau_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data
