"""
Deprecated wrappers for the NAG C Library (Mark 26.0) f07 Chapter.

f07 - Linear Equations (LAPACK).

https://www.nag.com/numeric/CL/nagdoc_cl26/html/F07/f07conts.html

NAG Copyright 2017.
"""
from ctypes import c_double
from ctypes import POINTER

from .util import get_library
from .util import Nag_RFP_Store

from .util import Nag_OrderType
from .util import Integer
from .util import NagError
from .util import Nag_FactoredFormType
from .util import Nag_TransType
from .util import Nag_EquilibrationType
from .util import Nag_NormType
from .util import Complex
from .util import Nag_UploType
from .util import Nag_DiagType

_library_cl = get_library()

f07aac_ctypes = _library_cl.f07aac
f07aac_ctypes.restype = None
f07aac_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07aac(order, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Computes the solution to a real system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07aac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07aac', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07aac', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07aac', b, 'b')
    f07aac_ctypes(order, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07abc_ctypes = _library_cl.f07abc
f07abc_ctypes.restype = None
f07abc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_TransType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07abc(order, fact, trans, n, nrhs, a, pda, af, pdaf, ipiv, equed, r, c, b, pdb, x, pdx, rcond, ferr, berr, recipgrowthfactor, fail):
    """
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07abc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, r, c, b, x, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, af, r, c, b, x, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07abc', a, 'a')
    _af = nag_double_type_check_and_cast('f07abc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07abc', ipiv, 'ipiv')
    _r = nag_double_type_check_and_cast('f07abc', r, 'r')
    _c = nag_double_type_check_and_cast('f07abc', c, 'c')
    _b = nag_double_type_check_and_cast('f07abc', b, 'b')
    _x = nag_double_type_check_and_cast('f07abc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07abc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07abc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07abc', berr, 'berr')
    _recipgrowthfactor = nag_double_type_check_and_cast('f07abc', recipgrowthfactor, 'recipgrowthfactor')
    f07abc_ctypes(order, fact, trans, n, nrhs, _a, pda, _af, pdaf, _ipiv, equed, _r, _c, _b, pdb, _x, pdx, _rcond, _ferr, _berr, _recipgrowthfactor, fail)

f07acc_ctypes = _library_cl.f07acc
f07acc_ctypes.restype = None
f07acc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(NagError)]


def f07acc(order, n, nrhs, a, pda, ipiv, b, pdb, x, pdx, _iter, fail):
    """
    Computes the solution to a real system of linear equations using mixed
    precision arithmetic.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07acc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b, x, _iter
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b, x
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv, _iter
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07acc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07acc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07acc', b, 'b')
    _x = nag_double_type_check_and_cast('f07acc', x, 'x')
    __iter = nag_integer_type_check_and_cast('f07acc', _iter, '_iter')
    f07acc_ctypes(order, n, nrhs, _a, pda, _ipiv, _b, pdb, _x, pdx, __iter, fail)

f07adc_ctypes = _library_cl.f07adc
f07adc_ctypes.restype = None
f07adc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(NagError)]


def f07adc(order, m, n, a, pda, ipiv, fail):
    """
    LU factorization of real m by n matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07adc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07adc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07adc', ipiv, 'ipiv')
    f07adc_ctypes(order, m, n, _a, pda, _ipiv, fail)

f07aec_ctypes = _library_cl.f07aec
f07aec_ctypes.restype = None
f07aec_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07aec(order, trans, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Solution of real system of linear equations, multiple right-hand sides,
    matrix already factorized by f07adc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07aec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07aec', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07aec', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07aec', b, 'b')
    f07aec_ctypes(order, trans, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07afc_ctypes = _library_cl.f07afc
f07afc_ctypes.restype = None
f07afc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07afc(order, m, n, a, pda, r, c, rowcnd, colcnd, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a general real
    matrix and reduce its condition number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07afc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, r, c, rowcnd, colcnd, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, r, c, rowcnd, colcnd, amax
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07afc', a, 'a')
    _r = nag_double_type_check_and_cast('f07afc', r, 'r')
    _c = nag_double_type_check_and_cast('f07afc', c, 'c')
    _rowcnd = nag_double_type_check_and_cast('f07afc', rowcnd, 'rowcnd')
    _colcnd = nag_double_type_check_and_cast('f07afc', colcnd, 'colcnd')
    _amax = nag_double_type_check_and_cast('f07afc', amax, 'amax')
    f07afc_ctypes(order, m, n, _a, pda, _r, _c, _rowcnd, _colcnd, _amax, fail)

f07agc_ctypes = _library_cl.f07agc
f07agc_ctypes.restype = None
f07agc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Integer, POINTER(c_double), Integer, c_double, POINTER(c_double), POINTER(NagError)]


def f07agc(order, norm, n, a, pda, anorm, rcond, fail):
    """
    Estimate condition number of real matrix, matrix already factorized by
    f07adc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07agc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, rcond
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07agc', a, 'a')
    _rcond = nag_double_type_check_and_cast('f07agc', rcond, 'rcond')
    f07agc_ctypes(order, norm, n, _a, pda, anorm, _rcond, fail)

f07ahc_ctypes = _library_cl.f07ahc
f07ahc_ctypes.restype = None
f07ahc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07ahc(order, trans, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real system of linear equations,
    multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ahc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, af, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07ahc', a, 'a')
    _af = nag_double_type_check_and_cast('f07ahc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07ahc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07ahc', b, 'b')
    _x = nag_double_type_check_and_cast('f07ahc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07ahc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07ahc', berr, 'berr')
    f07ahc_ctypes(order, trans, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07ajc_ctypes = _library_cl.f07ajc
f07ajc_ctypes.restype = None
f07ajc_ctypes.argtypes = [Nag_OrderType, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(NagError)]


def f07ajc(order, n, a, pda, ipiv, fail):
    """
    Inverse of real matrix, matrix already factorized by f07adc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ajc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07ajc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07ajc', ipiv, 'ipiv')
    f07ajc_ctypes(order, n, _a, pda, _ipiv, fail)

f07anc_ctypes = _library_cl.f07anc
f07anc_ctypes.restype = None
f07anc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07anc(order, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Computes the solution to a complex system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07anc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07anc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07anc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07anc', b, 'b')
    f07anc_ctypes(order, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07apc_ctypes = _library_cl.f07apc
f07apc_ctypes.restype = None
f07apc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_TransType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(c_double), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07apc(order, fact, trans, n, nrhs, a, pda, af, pdaf, ipiv, equed, r, c, b, pdb, x, pdx, rcond, ferr, berr, recipgrowthfactor, fail):
    """
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07apc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, r, c, b, x, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        r, c, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07apc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07apc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07apc', ipiv, 'ipiv')
    _r = nag_double_type_check_and_cast('f07apc', r, 'r')
    _c = nag_double_type_check_and_cast('f07apc', c, 'c')
    _b = nag_complex_type_check_and_cast('f07apc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07apc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07apc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07apc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07apc', berr, 'berr')
    _recipgrowthfactor = nag_double_type_check_and_cast('f07apc', recipgrowthfactor, 'recipgrowthfactor')
    f07apc_ctypes(order, fact, trans, n, nrhs, _a, pda, _af, pdaf, _ipiv, equed, _r, _c, _b, pdb, _x, pdx, _rcond, _ferr, _berr, _recipgrowthfactor, fail)

f07aqc_ctypes = _library_cl.f07aqc
f07aqc_ctypes.restype = None
f07aqc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07aqc(order, n, nrhs, a, pda, ipiv, b, pdb, x, pdx, _iter, fail):
    """
    Computes the solution to a complex system of linear equations using
    mixed precision arithmetic.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07aqc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b, x, _iter
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv, _iter
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07aqc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07aqc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07aqc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07aqc', x, 'x')
    __iter = nag_integer_type_check_and_cast('f07aqc', _iter, '_iter')
    f07aqc_ctypes(order, n, nrhs, _a, pda, _ipiv, _b, pdb, _x, pdx, __iter, fail)

f07arc_ctypes = _library_cl.f07arc
f07arc_ctypes.restype = None
f07arc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07arc(order, m, n, a, pda, ipiv, fail):
    """
    LU factorization of complex m by n matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07arc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07arc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07arc', ipiv, 'ipiv')
    f07arc_ctypes(order, m, n, _a, pda, _ipiv, fail)

f07asc_ctypes = _library_cl.f07asc
f07asc_ctypes.restype = None
f07asc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07asc(order, trans, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Solution of complex system of linear equations, multiple right-hand
    sides, matrix already factorized by f07arc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07asc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07asc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07asc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07asc', b, 'b')
    f07asc_ctypes(order, trans, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07atc_ctypes = _library_cl.f07atc
f07atc_ctypes.restype = None
f07atc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07atc(order, m, n, a, pda, r, c, rowcnd, colcnd, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a general
    complex matrix and reduce its condition number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07atc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, r, c, rowcnd, colcnd, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        r, c, rowcnd, colcnd, amax
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07atc', a, 'a')
    _r = nag_double_type_check_and_cast('f07atc', r, 'r')
    _c = nag_double_type_check_and_cast('f07atc', c, 'c')
    _rowcnd = nag_double_type_check_and_cast('f07atc', rowcnd, 'rowcnd')
    _colcnd = nag_double_type_check_and_cast('f07atc', colcnd, 'colcnd')
    _amax = nag_double_type_check_and_cast('f07atc', amax, 'amax')
    f07atc_ctypes(order, m, n, _a, pda, _r, _c, _rowcnd, _colcnd, _amax, fail)

f07auc_ctypes = _library_cl.f07auc
f07auc_ctypes.restype = None
f07auc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Integer, POINTER(Complex), Integer, c_double, POINTER(c_double), POINTER(NagError)]


def f07auc(order, norm, n, a, pda, anorm, rcond, fail):
    """
    Estimate condition number of complex matrix, matrix already factorized
    by f07arc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07auc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07auc', a, 'a')
    _rcond = nag_double_type_check_and_cast('f07auc', rcond, 'rcond')
    f07auc_ctypes(order, norm, n, _a, pda, anorm, _rcond, fail)

f07avc_ctypes = _library_cl.f07avc
f07avc_ctypes.restype = None
f07avc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07avc(order, trans, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07avc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07avc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07avc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07avc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07avc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07avc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07avc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07avc', berr, 'berr')
    f07avc_ctypes(order, trans, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07awc_ctypes = _library_cl.f07awc
f07awc_ctypes.restype = None
f07awc_ctypes.argtypes = [Nag_OrderType, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07awc(order, n, a, pda, ipiv, fail):
    """
    Inverse of complex matrix, matrix already factorized by f07arc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07awc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07awc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07awc', ipiv, 'ipiv')
    f07awc_ctypes(order, n, _a, pda, _ipiv, fail)

f07bac_ctypes = _library_cl.f07bac
f07bac_ctypes.restype = None
f07bac_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07bac(order, n, kl, ku, nrhs, ab, pdab, ipiv, b, pdb, fail):
    """
    Computes the solution to a real banded system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07bac', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07bac', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07bac', b, 'b')
    f07bac_ctypes(order, n, kl, ku, nrhs, _ab, pdab, _ipiv, _b, pdb, fail)

f07bbc_ctypes = _library_cl.f07bbc
f07bbc_ctypes.restype = None
f07bbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_TransType, Integer, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07bbc(order, fact, trans, n, kl, ku, nrhs, ab, pdab, afb, pdafb, ipiv, equed, r, c, b, pdb, x, pdx, rcond, ferr, berr, recipgrowthfactor, fail):
    """
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real banded system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, ipiv, r, c, b, x, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, afb, r, c, b, x, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07bbc', ab, 'ab')
    _afb = nag_double_type_check_and_cast('f07bbc', afb, 'afb')
    _ipiv = nag_integer_type_check_and_cast('f07bbc', ipiv, 'ipiv')
    _r = nag_double_type_check_and_cast('f07bbc', r, 'r')
    _c = nag_double_type_check_and_cast('f07bbc', c, 'c')
    _b = nag_double_type_check_and_cast('f07bbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07bbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07bbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07bbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07bbc', berr, 'berr')
    _recipgrowthfactor = nag_double_type_check_and_cast('f07bbc', recipgrowthfactor, 'recipgrowthfactor')
    f07bbc_ctypes(order, fact, trans, n, kl, ku, nrhs, _ab, pdab, _afb, pdafb, _ipiv, equed, _r, _c, _b, pdb, _x, pdx, _rcond, _ferr, _berr, _recipgrowthfactor, fail)

f07bdc_ctypes = _library_cl.f07bdc
f07bdc_ctypes.restype = None
f07bdc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(NagError)]


def f07bdc(order, m, n, kl, ku, ab, pdab, ipiv, fail):
    """
    LU factorization of real m by n band matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07bdc', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07bdc', ipiv, 'ipiv')
    f07bdc_ctypes(order, m, n, kl, ku, _ab, pdab, _ipiv, fail)

f07bec_ctypes = _library_cl.f07bec
f07bec_ctypes.restype = None
f07bec_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07bec(order, trans, n, kl, ku, nrhs, ab, pdab, ipiv, b, pdb, fail):
    """
    Solution of real band system of linear equations, multiple right-hand
    sides, matrix already factorized by f07bdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07bec', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07bec', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07bec', b, 'b')
    f07bec_ctypes(order, trans, n, kl, ku, nrhs, _ab, pdab, _ipiv, _b, pdb, fail)

f07bfc_ctypes = _library_cl.f07bfc
f07bfc_ctypes.restype = None
f07bfc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07bfc(order, m, n, kl, ku, ab, pdab, r, c, rowcnd, colcnd, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a real banded
    matrix and reduce its condition number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bfc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, r, c, rowcnd, colcnd, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, r, c, rowcnd, colcnd, amax
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07bfc', ab, 'ab')
    _r = nag_double_type_check_and_cast('f07bfc', r, 'r')
    _c = nag_double_type_check_and_cast('f07bfc', c, 'c')
    _rowcnd = nag_double_type_check_and_cast('f07bfc', rowcnd, 'rowcnd')
    _colcnd = nag_double_type_check_and_cast('f07bfc', colcnd, 'colcnd')
    _amax = nag_double_type_check_and_cast('f07bfc', amax, 'amax')
    f07bfc_ctypes(order, m, n, kl, ku, _ab, pdab, _r, _c, _rowcnd, _colcnd, _amax, fail)

f07bgc_ctypes = _library_cl.f07bgc
f07bgc_ctypes.restype = None
f07bgc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07bgc(order, norm, n, kl, ku, ab, pdab, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of real band matrix, matrix already factorized
    by f07bdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07bgc', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07bgc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07bgc', rcond, 'rcond')
    f07bgc_ctypes(order, norm, n, kl, ku, _ab, pdab, _ipiv, anorm, _rcond, fail)

f07bhc_ctypes = _library_cl.f07bhc
f07bhc_ctypes.restype = None
f07bhc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07bhc(order, trans, n, kl, ku, nrhs, ab, pdab, afb, pdafb, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real band system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, afb, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07bhc', ab, 'ab')
    _afb = nag_double_type_check_and_cast('f07bhc', afb, 'afb')
    _ipiv = nag_integer_type_check_and_cast('f07bhc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07bhc', b, 'b')
    _x = nag_double_type_check_and_cast('f07bhc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07bhc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07bhc', berr, 'berr')
    f07bhc_ctypes(order, trans, n, kl, ku, nrhs, _ab, pdab, _afb, pdafb, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07bnc_ctypes = _library_cl.f07bnc
f07bnc_ctypes.restype = None
f07bnc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07bnc(order, n, kl, ku, nrhs, ab, pdab, ipiv, b, pdb, fail):
    """
    Computes the solution to a complex banded system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07bnc', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07bnc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07bnc', b, 'b')
    f07bnc_ctypes(order, n, kl, ku, nrhs, _ab, pdab, _ipiv, _b, pdb, fail)

f07bpc_ctypes = _library_cl.f07bpc
f07bpc_ctypes.restype = None
f07bpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_TransType, Integer, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(c_double), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07bpc(order, fact, trans, n, kl, ku, nrhs, ab, pdab, afb, pdafb, ipiv, equed, r, c, b, pdb, x, pdx, rcond, ferr, berr, recipgrowthfactor, fail):
    """
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex banded system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, ipiv, r, c, b, x, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        r, c, rcond, ferr, berr, recipgrowthfactor
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, afb, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07bpc', ab, 'ab')
    _afb = nag_complex_type_check_and_cast('f07bpc', afb, 'afb')
    _ipiv = nag_integer_type_check_and_cast('f07bpc', ipiv, 'ipiv')
    _r = nag_double_type_check_and_cast('f07bpc', r, 'r')
    _c = nag_double_type_check_and_cast('f07bpc', c, 'c')
    _b = nag_complex_type_check_and_cast('f07bpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07bpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07bpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07bpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07bpc', berr, 'berr')
    _recipgrowthfactor = nag_double_type_check_and_cast('f07bpc', recipgrowthfactor, 'recipgrowthfactor')
    f07bpc_ctypes(order, fact, trans, n, kl, ku, nrhs, _ab, pdab, _afb, pdafb, _ipiv, equed, _r, _c, _b, pdb, _x, pdx, _rcond, _ferr, _berr, _recipgrowthfactor, fail)

f07brc_ctypes = _library_cl.f07brc
f07brc_ctypes.restype = None
f07brc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07brc(order, m, n, kl, ku, ab, pdab, ipiv, fail):
    """
    LU factorization of complex m by n band matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07brc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07brc', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07brc', ipiv, 'ipiv')
    f07brc_ctypes(order, m, n, kl, ku, _ab, pdab, _ipiv, fail)

f07bsc_ctypes = _library_cl.f07bsc
f07bsc_ctypes.restype = None
f07bsc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07bsc(order, trans, n, kl, ku, nrhs, ab, pdab, ipiv, b, pdb, fail):
    """
    Solution of complex band system of linear equations, multiple right-hand
    sides, matrix already factorized by f07brc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07bsc', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07bsc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07bsc', b, 'b')
    f07bsc_ctypes(order, trans, n, kl, ku, nrhs, _ab, pdab, _ipiv, _b, pdb, fail)

f07btc_ctypes = _library_cl.f07btc
f07btc_ctypes.restype = None
f07btc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07btc(order, m, n, kl, ku, ab, pdab, r, c, rowcnd, colcnd, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a complex
    banded matrix and reduce its condition number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07btc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, r, c, rowcnd, colcnd, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        r, c, rowcnd, colcnd, amax
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07btc', ab, 'ab')
    _r = nag_double_type_check_and_cast('f07btc', r, 'r')
    _c = nag_double_type_check_and_cast('f07btc', c, 'c')
    _rowcnd = nag_double_type_check_and_cast('f07btc', rowcnd, 'rowcnd')
    _colcnd = nag_double_type_check_and_cast('f07btc', colcnd, 'colcnd')
    _amax = nag_double_type_check_and_cast('f07btc', amax, 'amax')
    f07btc_ctypes(order, m, n, kl, ku, _ab, pdab, _r, _c, _rowcnd, _colcnd, _amax, fail)

f07buc_ctypes = _library_cl.f07buc
f07buc_ctypes.restype = None
f07buc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07buc(order, norm, n, kl, ku, ab, pdab, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of complex band matrix, matrix already
    factorized by f07brc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07buc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07buc', ab, 'ab')
    _ipiv = nag_integer_type_check_and_cast('f07buc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07buc', rcond, 'rcond')
    f07buc_ctypes(order, norm, n, kl, ku, _ab, pdab, _ipiv, anorm, _rcond, fail)

f07bvc_ctypes = _library_cl.f07bvc
f07bvc_ctypes.restype = None
f07bvc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07bvc(order, trans, n, kl, ku, nrhs, ab, pdab, afb, pdafb, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex band system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07bvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, afb, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07bvc', ab, 'ab')
    _afb = nag_complex_type_check_and_cast('f07bvc', afb, 'afb')
    _ipiv = nag_integer_type_check_and_cast('f07bvc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07bvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07bvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07bvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07bvc', berr, 'berr')
    f07bvc_ctypes(order, trans, n, kl, ku, nrhs, _ab, pdab, _afb, pdafb, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07cac_ctypes = _library_cl.f07cac
f07cac_ctypes.restype = None
f07cac_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(NagError)]


def f07cac(order, n, nrhs, dl, d, du, b, pdb, fail):
    """
    Computes the solution to a real tridiagonal system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        dl, d, du, b
    """
    from .util import nag_double_type_check_and_cast
    _dl = nag_double_type_check_and_cast('f07cac', dl, 'dl')
    _d = nag_double_type_check_and_cast('f07cac', d, 'd')
    _du = nag_double_type_check_and_cast('f07cac', du, 'du')
    _b = nag_double_type_check_and_cast('f07cac', b, 'b')
    f07cac_ctypes(order, n, nrhs, _dl, _d, _du, _b, pdb, fail)

f07cbc_ctypes = _library_cl.f07cbc
f07cbc_ctypes.restype = None
f07cbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_TransType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07cbc(order, fact, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real tridiagonal system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, dlf, df, duf, du2, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        dl, d, du, dlf, df, duf, du2, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_double_type_check_and_cast('f07cbc', dl, 'dl')
    _d = nag_double_type_check_and_cast('f07cbc', d, 'd')
    _du = nag_double_type_check_and_cast('f07cbc', du, 'du')
    _dlf = nag_double_type_check_and_cast('f07cbc', dlf, 'dlf')
    _df = nag_double_type_check_and_cast('f07cbc', df, 'df')
    _duf = nag_double_type_check_and_cast('f07cbc', duf, 'duf')
    _du2 = nag_double_type_check_and_cast('f07cbc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07cbc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07cbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07cbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07cbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07cbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07cbc', berr, 'berr')
    f07cbc_ctypes(order, fact, trans, n, nrhs, _dl, _d, _du, _dlf, _df, _duf, _du2, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07cdc_ctypes = _library_cl.f07cdc
f07cdc_ctypes.restype = None
f07cdc_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f07cdc(n, dl, d, du, du2, ipiv, fail):
    """
    LU factorization of real tridiagonal matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, du2, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        dl, d, du, du2
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_double_type_check_and_cast('f07cdc', dl, 'dl')
    _d = nag_double_type_check_and_cast('f07cdc', d, 'd')
    _du = nag_double_type_check_and_cast('f07cdc', du, 'du')
    _du2 = nag_double_type_check_and_cast('f07cdc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07cdc', ipiv, 'ipiv')
    f07cdc_ctypes(n, _dl, _d, _du, _du2, _ipiv, fail)

f07cec_ctypes = _library_cl.f07cec
f07cec_ctypes.restype = None
f07cec_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07cec(order, trans, n, nrhs, dl, d, du, du2, ipiv, b, pdb, fail):
    """
    Solves a real tridiagonal system of linear equations using the LU
    factorization computed by f07cdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, du2, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        dl, d, du, du2, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_double_type_check_and_cast('f07cec', dl, 'dl')
    _d = nag_double_type_check_and_cast('f07cec', d, 'd')
    _du = nag_double_type_check_and_cast('f07cec', du, 'du')
    _du2 = nag_double_type_check_and_cast('f07cec', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07cec', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07cec', b, 'b')
    f07cec_ctypes(order, trans, n, nrhs, _dl, _d, _du, _du2, _ipiv, _b, pdb, fail)

f07cgc_ctypes = _library_cl.f07cgc
f07cgc_ctypes.restype = None
f07cgc_ctypes.argtypes = [Nag_NormType, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07cgc(norm, n, dl, d, du, du2, ipiv, anorm, rcond, fail):
    """
    Estimates the reciprocal of the condition number of a real tridiagonal
    matrix using the LU factorization computed by f07cdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, du2, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        dl, d, du, du2, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_double_type_check_and_cast('f07cgc', dl, 'dl')
    _d = nag_double_type_check_and_cast('f07cgc', d, 'd')
    _du = nag_double_type_check_and_cast('f07cgc', du, 'du')
    _du2 = nag_double_type_check_and_cast('f07cgc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07cgc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07cgc', rcond, 'rcond')
    f07cgc_ctypes(norm, n, _dl, _d, _du, _du2, _ipiv, anorm, _rcond, fail)

f07chc_ctypes = _library_cl.f07chc
f07chc_ctypes.restype = None
f07chc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07chc(order, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real tridiagonal system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07chc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, dlf, df, duf, du2, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        dl, d, du, dlf, df, duf, du2, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_double_type_check_and_cast('f07chc', dl, 'dl')
    _d = nag_double_type_check_and_cast('f07chc', d, 'd')
    _du = nag_double_type_check_and_cast('f07chc', du, 'du')
    _dlf = nag_double_type_check_and_cast('f07chc', dlf, 'dlf')
    _df = nag_double_type_check_and_cast('f07chc', df, 'df')
    _duf = nag_double_type_check_and_cast('f07chc', duf, 'duf')
    _du2 = nag_double_type_check_and_cast('f07chc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07chc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07chc', b, 'b')
    _x = nag_double_type_check_and_cast('f07chc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07chc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07chc', berr, 'berr')
    f07chc_ctypes(order, trans, n, nrhs, _dl, _d, _du, _dlf, _df, _duf, _du2, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07cnc_ctypes = _library_cl.f07cnc
f07cnc_ctypes.restype = None
f07cnc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), Integer, POINTER(NagError)]


def f07cnc(order, n, nrhs, dl, d, du, b, pdb, fail):
    """
    Computes the solution to a complex tridiagonal system of linear
    equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        dl, d, du, b
    """
    from .util import nag_complex_type_check_and_cast
    _dl = nag_complex_type_check_and_cast('f07cnc', dl, 'dl')
    _d = nag_complex_type_check_and_cast('f07cnc', d, 'd')
    _du = nag_complex_type_check_and_cast('f07cnc', du, 'du')
    _b = nag_complex_type_check_and_cast('f07cnc', b, 'b')
    f07cnc_ctypes(order, n, nrhs, _dl, _d, _du, _b, pdb, fail)

f07cpc_ctypes = _library_cl.f07cpc
f07cpc_ctypes.restype = None
f07cpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_TransType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07cpc(order, fact, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex tridiagonal system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, dlf, df, duf, du2, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        dl, d, du, dlf, df, duf, du2, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_complex_type_check_and_cast('f07cpc', dl, 'dl')
    _d = nag_complex_type_check_and_cast('f07cpc', d, 'd')
    _du = nag_complex_type_check_and_cast('f07cpc', du, 'du')
    _dlf = nag_complex_type_check_and_cast('f07cpc', dlf, 'dlf')
    _df = nag_complex_type_check_and_cast('f07cpc', df, 'df')
    _duf = nag_complex_type_check_and_cast('f07cpc', duf, 'duf')
    _du2 = nag_complex_type_check_and_cast('f07cpc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07cpc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07cpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07cpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07cpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07cpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07cpc', berr, 'berr')
    f07cpc_ctypes(order, fact, trans, n, nrhs, _dl, _d, _du, _dlf, _df, _duf, _du2, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07crc_ctypes = _library_cl.f07crc
f07crc_ctypes.restype = None
f07crc_ctypes.argtypes = [Integer, POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f07crc(n, dl, d, du, du2, ipiv, fail):
    """
    LU factorization of complex tridiagonal matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07crc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, du2, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        dl, d, du, du2
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_complex_type_check_and_cast('f07crc', dl, 'dl')
    _d = nag_complex_type_check_and_cast('f07crc', d, 'd')
    _du = nag_complex_type_check_and_cast('f07crc', du, 'du')
    _du2 = nag_complex_type_check_and_cast('f07crc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07crc', ipiv, 'ipiv')
    f07crc_ctypes(n, _dl, _d, _du, _du2, _ipiv, fail)

f07csc_ctypes = _library_cl.f07csc
f07csc_ctypes.restype = None
f07csc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07csc(order, trans, n, nrhs, dl, d, du, du2, ipiv, b, pdb, fail):
    """
    Solves a complex tridiagonal system of linear equations using the LU
    factorization computed by f07cdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07csc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, du2, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        dl, d, du, du2, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_complex_type_check_and_cast('f07csc', dl, 'dl')
    _d = nag_complex_type_check_and_cast('f07csc', d, 'd')
    _du = nag_complex_type_check_and_cast('f07csc', du, 'du')
    _du2 = nag_complex_type_check_and_cast('f07csc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07csc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07csc', b, 'b')
    f07csc_ctypes(order, trans, n, nrhs, _dl, _d, _du, _du2, _ipiv, _b, pdb, fail)

f07cuc_ctypes = _library_cl.f07cuc
f07cuc_ctypes.restype = None
f07cuc_ctypes.argtypes = [Nag_NormType, Integer, POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07cuc(norm, n, dl, d, du, du2, ipiv, anorm, rcond, fail):
    """
    Estimates the reciprocal of the condition number of a complex
    tridiagonal matrix using the LU factorization computed by f07cdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cuc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, du2, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        dl, d, du, du2
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_complex_type_check_and_cast('f07cuc', dl, 'dl')
    _d = nag_complex_type_check_and_cast('f07cuc', d, 'd')
    _du = nag_complex_type_check_and_cast('f07cuc', du, 'du')
    _du2 = nag_complex_type_check_and_cast('f07cuc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07cuc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07cuc', rcond, 'rcond')
    f07cuc_ctypes(norm, n, _dl, _d, _du, _du2, _ipiv, anorm, _rcond, fail)

f07cvc_ctypes = _library_cl.f07cvc
f07cvc_ctypes.restype = None
f07cvc_ctypes.argtypes = [Nag_OrderType, Nag_TransType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07cvc(order, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex tridiagonal system of
    linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07cvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        dl, d, du, dlf, df, duf, du2, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        dl, d, du, dlf, df, duf, du2, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _dl = nag_complex_type_check_and_cast('f07cvc', dl, 'dl')
    _d = nag_complex_type_check_and_cast('f07cvc', d, 'd')
    _du = nag_complex_type_check_and_cast('f07cvc', du, 'du')
    _dlf = nag_complex_type_check_and_cast('f07cvc', dlf, 'dlf')
    _df = nag_complex_type_check_and_cast('f07cvc', df, 'df')
    _duf = nag_complex_type_check_and_cast('f07cvc', duf, 'duf')
    _du2 = nag_complex_type_check_and_cast('f07cvc', du2, 'du2')
    _ipiv = nag_integer_type_check_and_cast('f07cvc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07cvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07cvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07cvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07cvc', berr, 'berr')
    f07cvc_ctypes(order, trans, n, nrhs, _dl, _d, _du, _dlf, _df, _duf, _du2, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07fac_ctypes = _library_cl.f07fac
f07fac_ctypes.restype = None
f07fac_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07fac(order, uplo, n, nrhs, a, pda, b, pdb, fail):
    """
    Computes the solution to a real symmetric positive definite system of
    linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07fac', a, 'a')
    _b = nag_double_type_check_and_cast('f07fac', b, 'b')
    f07fac_ctypes(order, uplo, n, nrhs, _a, pda, _b, pdb, fail)

f07fbc_ctypes = _library_cl.f07fbc
f07fbc_ctypes.restype = None
f07fbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07fbc(order, fact, uplo, n, nrhs, a, pda, af, pdaf, equed, s, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the Cholesky factorization to compute the solution, error-bound and
    condition estimate for a real symmetric positive definite system of
    linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, s, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, af, s, b, x, rcond, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07fbc', a, 'a')
    _af = nag_double_type_check_and_cast('f07fbc', af, 'af')
    _s = nag_double_type_check_and_cast('f07fbc', s, 's')
    _b = nag_double_type_check_and_cast('f07fbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07fbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07fbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07fbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07fbc', berr, 'berr')
    f07fbc_ctypes(order, fact, uplo, n, nrhs, _a, pda, _af, pdaf, equed, _s, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07fdc_ctypes = _library_cl.f07fdc
f07fdc_ctypes.restype = None
f07fdc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07fdc(order, uplo, n, a, pda, fail):
    """
    Cholesky factorization of real symmetric positive definite matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07fdc', a, 'a')
    f07fdc_ctypes(order, uplo, n, _a, pda, fail)

f07fec_ctypes = _library_cl.f07fec
f07fec_ctypes.restype = None
f07fec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07fec(order, uplo, n, nrhs, a, pda, b, pdb, fail):
    """
    Solution of real symmetric positive definite system of linear equations,
    multiple right-hand sides, matrix already factorized by f07fdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07fec', a, 'a')
    _b = nag_double_type_check_and_cast('f07fec', b, 'b')
    f07fec_ctypes(order, uplo, n, nrhs, _a, pda, _b, pdb, fail)

f07ffc_ctypes = _library_cl.f07ffc
f07ffc_ctypes.restype = None
f07ffc_ctypes.argtypes = [Nag_OrderType, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07ffc(order, n, a, pda, s, scond, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite matrix and reduce its condition number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ffc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, s, scond, amax
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07ffc', a, 'a')
    _s = nag_double_type_check_and_cast('f07ffc', s, 's')
    _scond = nag_double_type_check_and_cast('f07ffc', scond, 'scond')
    _amax = nag_double_type_check_and_cast('f07ffc', amax, 'amax')
    f07ffc_ctypes(order, n, _a, pda, _s, _scond, _amax, fail)

f07fgc_ctypes = _library_cl.f07fgc
f07fgc_ctypes.restype = None
f07fgc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), Integer, c_double, POINTER(c_double), POINTER(NagError)]


def f07fgc(order, uplo, n, a, pda, anorm, rcond, fail):
    """
    Estimate condition number of real symmetric positive definite matrix,
    matrix already factorized by f07fdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, rcond
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07fgc', a, 'a')
    _rcond = nag_double_type_check_and_cast('f07fgc', rcond, 'rcond')
    f07fgc_ctypes(order, uplo, n, _a, pda, anorm, _rcond, fail)

f07fhc_ctypes = _library_cl.f07fhc
f07fhc_ctypes.restype = None
f07fhc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07fhc(order, uplo, n, nrhs, a, pda, af, pdaf, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real symmetric positive definite
    system of linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, af, b, x, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07fhc', a, 'a')
    _af = nag_double_type_check_and_cast('f07fhc', af, 'af')
    _b = nag_double_type_check_and_cast('f07fhc', b, 'b')
    _x = nag_double_type_check_and_cast('f07fhc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07fhc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07fhc', berr, 'berr')
    f07fhc_ctypes(order, uplo, n, nrhs, _a, pda, _af, pdaf, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07fjc_ctypes = _library_cl.f07fjc
f07fjc_ctypes.restype = None
f07fjc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07fjc(order, uplo, n, a, pda, fail):
    """
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by f07fdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fjc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07fjc', a, 'a')
    f07fjc_ctypes(order, uplo, n, _a, pda, fail)

f07fnc_ctypes = _library_cl.f07fnc
f07fnc_ctypes.restype = None
f07fnc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07fnc(order, uplo, n, nrhs, a, pda, b, pdb, fail):
    """
    Computes the solution to a complex Hermitian positive definite system of
    linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07fnc', a, 'a')
    _b = nag_complex_type_check_and_cast('f07fnc', b, 'b')
    f07fnc_ctypes(order, uplo, n, nrhs, _a, pda, _b, pdb, fail)

f07fpc_ctypes = _library_cl.f07fpc
f07fpc_ctypes.restype = None
f07fpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07fpc(order, fact, uplo, n, nrhs, a, pda, af, pdaf, equed, s, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the Cholesky factorization to compute the solution, error-bound and
    condition estimate for a complex Hermitian positive definite system of
    linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, s, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        s, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07fpc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07fpc', af, 'af')
    _s = nag_double_type_check_and_cast('f07fpc', s, 's')
    _b = nag_complex_type_check_and_cast('f07fpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07fpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07fpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07fpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07fpc', berr, 'berr')
    f07fpc_ctypes(order, fact, uplo, n, nrhs, _a, pda, _af, pdaf, equed, _s, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07frc_ctypes = _library_cl.f07frc
f07frc_ctypes.restype = None
f07frc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07frc(order, uplo, n, a, pda, fail):
    """
    Cholesky factorization of complex Hermitian positive definite matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07frc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07frc', a, 'a')
    f07frc_ctypes(order, uplo, n, _a, pda, fail)

f07fsc_ctypes = _library_cl.f07fsc
f07fsc_ctypes.restype = None
f07fsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07fsc(order, uplo, n, nrhs, a, pda, b, pdb, fail):
    """
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    f07frc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07fsc', a, 'a')
    _b = nag_complex_type_check_and_cast('f07fsc', b, 'b')
    f07fsc_ctypes(order, uplo, n, nrhs, _a, pda, _b, pdb, fail)

f07ftc_ctypes = _library_cl.f07ftc
f07ftc_ctypes.restype = None
f07ftc_ctypes.argtypes = [Nag_OrderType, Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07ftc(order, n, a, pda, s, scond, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite matrix and reduce its condition number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ftc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07ftc', a, 'a')
    _s = nag_double_type_check_and_cast('f07ftc', s, 's')
    _scond = nag_double_type_check_and_cast('f07ftc', scond, 'scond')
    _amax = nag_double_type_check_and_cast('f07ftc', amax, 'amax')
    f07ftc_ctypes(order, n, _a, pda, _s, _scond, _amax, fail)

f07fuc_ctypes = _library_cl.f07fuc
f07fuc_ctypes.restype = None
f07fuc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, c_double, POINTER(c_double), POINTER(NagError)]


def f07fuc(order, uplo, n, a, pda, anorm, rcond, fail):
    """
    Estimate condition number of complex Hermitian positive definite matrix,
    matrix already factorized by f07frc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fuc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07fuc', a, 'a')
    _rcond = nag_double_type_check_and_cast('f07fuc', rcond, 'rcond')
    f07fuc_ctypes(order, uplo, n, _a, pda, anorm, _rcond, fail)

f07fvc_ctypes = _library_cl.f07fvc
f07fvc_ctypes.restype = None
f07fvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07fvc(order, uplo, n, nrhs, a, pda, af, pdaf, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex Hermitian positive
    definite system of linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07fvc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07fvc', af, 'af')
    _b = nag_complex_type_check_and_cast('f07fvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07fvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07fvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07fvc', berr, 'berr')
    f07fvc_ctypes(order, uplo, n, nrhs, _a, pda, _af, pdaf, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07fwc_ctypes = _library_cl.f07fwc
f07fwc_ctypes.restype = None
f07fwc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07fwc(order, uplo, n, a, pda, fail):
    """
    Inverse of complex Hermitian positive definite matrix, matrix already
    factorized by f07frc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07fwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07fwc', a, 'a')
    f07fwc_ctypes(order, uplo, n, _a, pda, fail)

f07gac_ctypes = _library_cl.f07gac
f07gac_ctypes.restype = None
f07gac_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(NagError)]


def f07gac(order, uplo, n, nrhs, ap, b, pdb, fail):
    """
    Computes the solution to a real symmetric positive definite system of
    linear equations, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, b
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07gac', ap, 'ap')
    _b = nag_double_type_check_and_cast('f07gac', b, 'b')
    f07gac_ctypes(order, uplo, n, nrhs, _ap, _b, pdb, fail)

f07gbc_ctypes = _library_cl.f07gbc
f07gbc_ctypes.restype = None
f07gbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07gbc(order, fact, uplo, n, nrhs, ap, afp, equed, s, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the Cholesky factorization to compute the solution, error-bound and
    condition estimate for a real symmetric positive definite system of
    linear equations, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, s, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, afp, s, b, x, rcond, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07gbc', ap, 'ap')
    _afp = nag_double_type_check_and_cast('f07gbc', afp, 'afp')
    _s = nag_double_type_check_and_cast('f07gbc', s, 's')
    _b = nag_double_type_check_and_cast('f07gbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07gbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07gbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07gbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07gbc', berr, 'berr')
    f07gbc_ctypes(order, fact, uplo, n, nrhs, _ap, _afp, equed, _s, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07gdc_ctypes = _library_cl.f07gdc
f07gdc_ctypes.restype = None
f07gdc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), POINTER(NagError)]


def f07gdc(order, uplo, n, ap, fail):
    """
    Cholesky factorization of real symmetric positive definite matrix,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07gdc', ap, 'ap')
    f07gdc_ctypes(order, uplo, n, _ap, fail)

f07gec_ctypes = _library_cl.f07gec
f07gec_ctypes.restype = None
f07gec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(NagError)]


def f07gec(order, uplo, n, nrhs, ap, b, pdb, fail):
    """
    Solution of real symmetric positive definite system of linear equations,
    multiple right-hand sides, matrix already factorized by f07gdc, packed
    storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, b
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07gec', ap, 'ap')
    _b = nag_double_type_check_and_cast('f07gec', b, 'b')
    f07gec_ctypes(order, uplo, n, nrhs, _ap, _b, pdb, fail)

f07gfc_ctypes = _library_cl.f07gfc
f07gfc_ctypes.restype = None
f07gfc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07gfc(order, uplo, n, ap, s, scond, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite matrix and reduce its condition number,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gfc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, s, scond, amax
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07gfc', ap, 'ap')
    _s = nag_double_type_check_and_cast('f07gfc', s, 's')
    _scond = nag_double_type_check_and_cast('f07gfc', scond, 'scond')
    _amax = nag_double_type_check_and_cast('f07gfc', amax, 'amax')
    f07gfc_ctypes(order, uplo, n, _ap, _s, _scond, _amax, fail)

f07ggc_ctypes = _library_cl.f07ggc
f07ggc_ctypes.restype = None
f07ggc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), c_double, POINTER(c_double), POINTER(NagError)]


def f07ggc(order, uplo, n, ap, anorm, rcond, fail):
    """
    Estimate condition number of real symmetric positive definite matrix,
    matrix already factorized by f07gdc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ggc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, rcond
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07ggc', ap, 'ap')
    _rcond = nag_double_type_check_and_cast('f07ggc', rcond, 'rcond')
    f07ggc_ctypes(order, uplo, n, _ap, anorm, _rcond, fail)

f07ghc_ctypes = _library_cl.f07ghc
f07ghc_ctypes.restype = None
f07ghc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07ghc(order, uplo, n, nrhs, ap, afp, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real symmetric positive definite
    system of linear equations, multiple right-hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ghc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, afp, b, x, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07ghc', ap, 'ap')
    _afp = nag_double_type_check_and_cast('f07ghc', afp, 'afp')
    _b = nag_double_type_check_and_cast('f07ghc', b, 'b')
    _x = nag_double_type_check_and_cast('f07ghc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07ghc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07ghc', berr, 'berr')
    f07ghc_ctypes(order, uplo, n, nrhs, _ap, _afp, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07gjc_ctypes = _library_cl.f07gjc
f07gjc_ctypes.restype = None
f07gjc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), POINTER(NagError)]


def f07gjc(order, uplo, n, ap, fail):
    """
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by f07gdc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gjc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07gjc', ap, 'ap')
    f07gjc_ctypes(order, uplo, n, _ap, fail)

f07gnc_ctypes = _library_cl.f07gnc
f07gnc_ctypes.restype = None
f07gnc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), Integer, POINTER(NagError)]


def f07gnc(order, uplo, n, nrhs, ap, b, pdb, fail):
    """
    Computes the solution to a complex Hermitian positive definite system of
    linear equations, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b
    """
    from .util import nag_complex_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07gnc', ap, 'ap')
    _b = nag_complex_type_check_and_cast('f07gnc', b, 'b')
    f07gnc_ctypes(order, uplo, n, nrhs, _ap, _b, pdb, fail)

f07gpc_ctypes = _library_cl.f07gpc
f07gpc_ctypes.restype = None
f07gpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07gpc(order, fact, uplo, n, nrhs, ap, afp, equed, s, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the Cholesky factorization to compute the solution, error-bound and
    condition estimate for a complex Hermitian positive definite system of
    linear equations, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, s, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        s, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, afp, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07gpc', ap, 'ap')
    _afp = nag_complex_type_check_and_cast('f07gpc', afp, 'afp')
    _s = nag_double_type_check_and_cast('f07gpc', s, 's')
    _b = nag_complex_type_check_and_cast('f07gpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07gpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07gpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07gpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07gpc', berr, 'berr')
    f07gpc_ctypes(order, fact, uplo, n, nrhs, _ap, _afp, equed, _s, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07grc_ctypes = _library_cl.f07grc
f07grc_ctypes.restype = None
f07grc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(NagError)]


def f07grc(order, uplo, n, ap, fail):
    """
    Cholesky factorization of complex Hermitian positive definite matrix,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07grc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07grc', ap, 'ap')
    f07grc_ctypes(order, uplo, n, _ap, fail)

f07gsc_ctypes = _library_cl.f07gsc
f07gsc_ctypes.restype = None
f07gsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), Integer, POINTER(NagError)]


def f07gsc(order, uplo, n, nrhs, ap, b, pdb, fail):
    """
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    f07grc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b
    """
    from .util import nag_complex_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07gsc', ap, 'ap')
    _b = nag_complex_type_check_and_cast('f07gsc', b, 'b')
    f07gsc_ctypes(order, uplo, n, nrhs, _ap, _b, pdb, fail)

f07gtc_ctypes = _library_cl.f07gtc
f07gtc_ctypes.restype = None
f07gtc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07gtc(order, uplo, n, ap, s, scond, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite matrix and reduce its condition number,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gtc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07gtc', ap, 'ap')
    _s = nag_double_type_check_and_cast('f07gtc', s, 's')
    _scond = nag_double_type_check_and_cast('f07gtc', scond, 'scond')
    _amax = nag_double_type_check_and_cast('f07gtc', amax, 'amax')
    f07gtc_ctypes(order, uplo, n, _ap, _s, _scond, _amax, fail)

f07guc_ctypes = _library_cl.f07guc
f07guc_ctypes.restype = None
f07guc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), c_double, POINTER(c_double), POINTER(NagError)]


def f07guc(order, uplo, n, ap, anorm, rcond, fail):
    """
    Estimate condition number of complex Hermitian positive definite matrix,
    matrix already factorized by f07grc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07guc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07guc', ap, 'ap')
    _rcond = nag_double_type_check_and_cast('f07guc', rcond, 'rcond')
    f07guc_ctypes(order, uplo, n, _ap, anorm, _rcond, fail)

f07gvc_ctypes = _library_cl.f07gvc
f07gvc_ctypes.restype = None
f07gvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07gvc(order, uplo, n, nrhs, ap, afp, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex Hermitian positive
    definite system of linear equations, multiple right-hand sides, packed
    storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, afp, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07gvc', ap, 'ap')
    _afp = nag_complex_type_check_and_cast('f07gvc', afp, 'afp')
    _b = nag_complex_type_check_and_cast('f07gvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07gvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07gvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07gvc', berr, 'berr')
    f07gvc_ctypes(order, uplo, n, nrhs, _ap, _afp, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07gwc_ctypes = _library_cl.f07gwc
f07gwc_ctypes.restype = None
f07gwc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(NagError)]


def f07gwc(order, uplo, n, ap, fail):
    """
    Inverse of complex Hermitian positive definite matrix, matrix already
    factorized by f07grc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07gwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07gwc', ap, 'ap')
    f07gwc_ctypes(order, uplo, n, _ap, fail)

f07hac_ctypes = _library_cl.f07hac
f07hac_ctypes.restype = None
f07hac_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07hac(order, uplo, n, kd, nrhs, ab, pdab, b, pdb, fail):
    """
    Computes the solution to a real symmetric positive definite banded
    system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, b
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07hac', ab, 'ab')
    _b = nag_double_type_check_and_cast('f07hac', b, 'b')
    f07hac_ctypes(order, uplo, n, kd, nrhs, _ab, pdab, _b, pdb, fail)

f07hbc_ctypes = _library_cl.f07hbc
f07hbc_ctypes.restype = None
f07hbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07hbc(order, fact, uplo, n, kd, nrhs, ab, pdab, afb, pdafb, equed, s, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the Cholesky factorization to compute the solution, error-bound and
    condition estimate for a real symmetric positive definite banded system
    of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, s, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, afb, s, b, x, rcond, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07hbc', ab, 'ab')
    _afb = nag_double_type_check_and_cast('f07hbc', afb, 'afb')
    _s = nag_double_type_check_and_cast('f07hbc', s, 's')
    _b = nag_double_type_check_and_cast('f07hbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07hbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07hbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07hbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07hbc', berr, 'berr')
    f07hbc_ctypes(order, fact, uplo, n, kd, nrhs, _ab, pdab, _afb, pdafb, equed, _s, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07hdc_ctypes = _library_cl.f07hdc
f07hdc_ctypes.restype = None
f07hdc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07hdc(order, uplo, n, kd, ab, pdab, fail):
    """
    Cholesky factorization of real symmetric positive definite band matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07hdc', ab, 'ab')
    f07hdc_ctypes(order, uplo, n, kd, _ab, pdab, fail)

f07hec_ctypes = _library_cl.f07hec
f07hec_ctypes.restype = None
f07hec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07hec(order, uplo, n, kd, nrhs, ab, pdab, b, pdb, fail):
    """
    Solution of real symmetric positive definite band system of linear
    equations, multiple right-hand sides, matrix already factorized by
    f07hdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, b
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07hec', ab, 'ab')
    _b = nag_double_type_check_and_cast('f07hec', b, 'b')
    f07hec_ctypes(order, uplo, n, kd, nrhs, _ab, pdab, _b, pdb, fail)

f07hfc_ctypes = _library_cl.f07hfc
f07hfc_ctypes.restype = None
f07hfc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07hfc(order, uplo, n, kd, ab, pdab, s, scond, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite banded matrix and reduce its condition
    number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hfc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, s, scond, amax
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07hfc', ab, 'ab')
    _s = nag_double_type_check_and_cast('f07hfc', s, 's')
    _scond = nag_double_type_check_and_cast('f07hfc', scond, 'scond')
    _amax = nag_double_type_check_and_cast('f07hfc', amax, 'amax')
    f07hfc_ctypes(order, uplo, n, kd, _ab, pdab, _s, _scond, _amax, fail)

f07hgc_ctypes = _library_cl.f07hgc
f07hgc_ctypes.restype = None
f07hgc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, c_double, POINTER(c_double), POINTER(NagError)]


def f07hgc(order, uplo, n, kd, ab, pdab, anorm, rcond, fail):
    """
    Estimate condition number of real symmetric positive definite band
    matrix, matrix already factorized by f07hdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, rcond
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07hgc', ab, 'ab')
    _rcond = nag_double_type_check_and_cast('f07hgc', rcond, 'rcond')
    f07hgc_ctypes(order, uplo, n, kd, _ab, pdab, anorm, _rcond, fail)

f07hhc_ctypes = _library_cl.f07hhc
f07hhc_ctypes.restype = None
f07hhc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07hhc(order, uplo, n, kd, nrhs, ab, pdab, afb, pdafb, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real symmetric positive definite
    band system of linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, afb, b, x, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07hhc', ab, 'ab')
    _afb = nag_double_type_check_and_cast('f07hhc', afb, 'afb')
    _b = nag_double_type_check_and_cast('f07hhc', b, 'b')
    _x = nag_double_type_check_and_cast('f07hhc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07hhc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07hhc', berr, 'berr')
    f07hhc_ctypes(order, uplo, n, kd, nrhs, _ab, pdab, _afb, pdafb, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07hnc_ctypes = _library_cl.f07hnc
f07hnc_ctypes.restype = None
f07hnc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07hnc(order, uplo, n, kd, nrhs, ab, pdab, b, pdb, fail):
    """
    Computes the solution to a complex Hermitian positive definite banded
    system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, b
    """
    from .util import nag_complex_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07hnc', ab, 'ab')
    _b = nag_complex_type_check_and_cast('f07hnc', b, 'b')
    f07hnc_ctypes(order, uplo, n, kd, nrhs, _ab, pdab, _b, pdb, fail)

f07hpc_ctypes = _library_cl.f07hpc
f07hpc_ctypes.restype = None
f07hpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Nag_EquilibrationType), POINTER(c_double), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07hpc(order, fact, uplo, n, kd, nrhs, ab, pdab, afb, pdafb, equed, s, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the Cholesky factorization to compute the solution, error-bound and
    condition estimate for a complex Hermitian positive definite banded
    system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, s, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        s, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, afb, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07hpc', ab, 'ab')
    _afb = nag_complex_type_check_and_cast('f07hpc', afb, 'afb')
    _s = nag_double_type_check_and_cast('f07hpc', s, 's')
    _b = nag_complex_type_check_and_cast('f07hpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07hpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07hpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07hpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07hpc', berr, 'berr')
    f07hpc_ctypes(order, fact, uplo, n, kd, nrhs, _ab, pdab, _afb, pdafb, equed, _s, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07hrc_ctypes = _library_cl.f07hrc
f07hrc_ctypes.restype = None
f07hrc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07hrc(order, uplo, n, kd, ab, pdab, fail):
    """
    Cholesky factorization of complex Hermitian positive definite band
    matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hrc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab
    """
    from .util import nag_complex_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07hrc', ab, 'ab')
    f07hrc_ctypes(order, uplo, n, kd, _ab, pdab, fail)

f07hsc_ctypes = _library_cl.f07hsc
f07hsc_ctypes.restype = None
f07hsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07hsc(order, uplo, n, kd, nrhs, ab, pdab, b, pdb, fail):
    """
    Solution of complex Hermitian positive definite band system of linear
    equations, multiple right-hand sides, matrix already factorized by
    f07hrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, b
    """
    from .util import nag_complex_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07hsc', ab, 'ab')
    _b = nag_complex_type_check_and_cast('f07hsc', b, 'b')
    f07hsc_ctypes(order, uplo, n, kd, nrhs, _ab, pdab, _b, pdb, fail)

f07htc_ctypes = _library_cl.f07htc
f07htc_ctypes.restype = None
f07htc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07htc(order, uplo, n, kd, ab, pdab, s, scond, amax, fail):
    """
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite banded matrix and reduce its condition
    number.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07htc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        s, scond, amax
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07htc', ab, 'ab')
    _s = nag_double_type_check_and_cast('f07htc', s, 's')
    _scond = nag_double_type_check_and_cast('f07htc', scond, 'scond')
    _amax = nag_double_type_check_and_cast('f07htc', amax, 'amax')
    f07htc_ctypes(order, uplo, n, kd, _ab, pdab, _s, _scond, _amax, fail)

f07huc_ctypes = _library_cl.f07huc
f07huc_ctypes.restype = None
f07huc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, c_double, POINTER(c_double), POINTER(NagError)]


def f07huc(order, uplo, n, kd, ab, pdab, anorm, rcond, fail):
    """
    Estimate condition number of complex Hermitian positive definite band
    matrix, matrix already factorized by f07hrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07huc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07huc', ab, 'ab')
    _rcond = nag_double_type_check_and_cast('f07huc', rcond, 'rcond')
    f07huc_ctypes(order, uplo, n, kd, _ab, pdab, anorm, _rcond, fail)

f07hvc_ctypes = _library_cl.f07hvc
f07hvc_ctypes.restype = None
f07hvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07hvc(order, uplo, n, kd, nrhs, ab, pdab, afb, pdafb, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex Hermitian positive
    definite band system of linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07hvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, afb, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, afb, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07hvc', ab, 'ab')
    _afb = nag_complex_type_check_and_cast('f07hvc', afb, 'afb')
    _b = nag_complex_type_check_and_cast('f07hvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07hvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07hvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07hvc', berr, 'berr')
    f07hvc_ctypes(order, uplo, n, kd, nrhs, _ab, pdab, _afb, pdafb, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07jac_ctypes = _library_cl.f07jac
f07jac_ctypes.restype = None
f07jac_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(NagError)]


def f07jac(order, n, nrhs, d, e, b, pdb, fail):
    """
    Computes the solution to a real symmetric positive definite tridiagonal
    system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, e, b
    """
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jac', d, 'd')
    _e = nag_double_type_check_and_cast('f07jac', e, 'e')
    _b = nag_double_type_check_and_cast('f07jac', b, 'b')
    f07jac_ctypes(order, n, nrhs, _d, _e, _b, pdb, fail)

f07jbc_ctypes = _library_cl.f07jbc
f07jbc_ctypes.restype = None
f07jbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07jbc(order, fact, n, nrhs, d, e, df, ef, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the LDL^T factorization to compute the solution, error-bound and
    condition estimate for a real symmetric positive definite tridiagonal
    system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, df, ef, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, e, df, ef, b, x, rcond, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jbc', d, 'd')
    _e = nag_double_type_check_and_cast('f07jbc', e, 'e')
    _df = nag_double_type_check_and_cast('f07jbc', df, 'df')
    _ef = nag_double_type_check_and_cast('f07jbc', ef, 'ef')
    _b = nag_double_type_check_and_cast('f07jbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07jbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07jbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07jbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07jbc', berr, 'berr')
    f07jbc_ctypes(order, fact, n, nrhs, _d, _e, _df, _ef, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07jdc_ctypes = _library_cl.f07jdc
f07jdc_ctypes.restype = None
f07jdc_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07jdc(n, d, e, fail):
    """
    Computes the LDL^T factorization of a real symmetric positive definite
    tridiagonal matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, e
    """
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jdc', d, 'd')
    _e = nag_double_type_check_and_cast('f07jdc', e, 'e')
    f07jdc_ctypes(n, _d, _e, fail)

f07jec_ctypes = _library_cl.f07jec
f07jec_ctypes.restype = None
f07jec_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(NagError)]


def f07jec(order, n, nrhs, d, e, b, pdb, fail):
    """
    Solves a real symmetric positive definite tridiagonal system using the
    LDL^T factorization computed by f07jdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, e, b
    """
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jec', d, 'd')
    _e = nag_double_type_check_and_cast('f07jec', e, 'e')
    _b = nag_double_type_check_and_cast('f07jec', b, 'b')
    f07jec_ctypes(order, n, nrhs, _d, _e, _b, pdb, fail)

f07jgc_ctypes = _library_cl.f07jgc
f07jgc_ctypes.restype = None
f07jgc_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(NagError)]


def f07jgc(n, d, e, anorm, rcond, fail):
    """
    Computes the reciprocal of the condition number of a real symmetric
    positive definite tridiagonal system using the LDL^T factorization
    computed by f07jdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, e, rcond
    """
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jgc', d, 'd')
    _e = nag_double_type_check_and_cast('f07jgc', e, 'e')
    _rcond = nag_double_type_check_and_cast('f07jgc', rcond, 'rcond')
    f07jgc_ctypes(n, _d, _e, anorm, _rcond, fail)

f07jhc_ctypes = _library_cl.f07jhc
f07jhc_ctypes.restype = None
f07jhc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07jhc(order, n, nrhs, d, e, df, ef, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real symmetric positive definite
    tridiagonal system of linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, df, ef, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, e, df, ef, b, x, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jhc', d, 'd')
    _e = nag_double_type_check_and_cast('f07jhc', e, 'e')
    _df = nag_double_type_check_and_cast('f07jhc', df, 'df')
    _ef = nag_double_type_check_and_cast('f07jhc', ef, 'ef')
    _b = nag_double_type_check_and_cast('f07jhc', b, 'b')
    _x = nag_double_type_check_and_cast('f07jhc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07jhc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07jhc', berr, 'berr')
    f07jhc_ctypes(order, n, nrhs, _d, _e, _df, _ef, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07jnc_ctypes = _library_cl.f07jnc
f07jnc_ctypes.restype = None
f07jnc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(c_double), POINTER(Complex), POINTER(Complex), Integer, POINTER(NagError)]


def f07jnc(order, n, nrhs, d, e, b, pdb, fail):
    """
    Computes the solution to a complex Hermitian positive definite
    tridiagonal system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d
    TypeError will be raised when the following arguments are not of data type
    Complex:
        e, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jnc', d, 'd')
    _e = nag_complex_type_check_and_cast('f07jnc', e, 'e')
    _b = nag_complex_type_check_and_cast('f07jnc', b, 'b')
    f07jnc_ctypes(order, n, nrhs, _d, _e, _b, pdb, fail)

f07jpc_ctypes = _library_cl.f07jpc
f07jpc_ctypes.restype = None
f07jpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Integer, Integer, POINTER(c_double), POINTER(Complex), POINTER(c_double), POINTER(Complex), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07jpc(order, fact, n, nrhs, d, e, df, ef, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the LDL^T factorization to compute the solution, error-bound and
    condition estimate for a complex Hermitian positive definite tridiagonal
    system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, df, ef, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, df, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        e, ef, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jpc', d, 'd')
    _e = nag_complex_type_check_and_cast('f07jpc', e, 'e')
    _df = nag_double_type_check_and_cast('f07jpc', df, 'df')
    _ef = nag_complex_type_check_and_cast('f07jpc', ef, 'ef')
    _b = nag_complex_type_check_and_cast('f07jpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07jpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07jpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07jpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07jpc', berr, 'berr')
    f07jpc_ctypes(order, fact, n, nrhs, _d, _e, _df, _ef, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07jrc_ctypes = _library_cl.f07jrc
f07jrc_ctypes.restype = None
f07jrc_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(Complex), POINTER(NagError)]


def f07jrc(n, d, e, fail):
    """
    Computes the LDL^T factorization of a complex Hermitian positive
    definite tridiagonal matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jrc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d
    TypeError will be raised when the following arguments are not of data type
    Complex:
        e
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jrc', d, 'd')
    _e = nag_complex_type_check_and_cast('f07jrc', e, 'e')
    f07jrc_ctypes(n, _d, _e, fail)

f07jsc_ctypes = _library_cl.f07jsc
f07jsc_ctypes.restype = None
f07jsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(Complex), POINTER(Complex), Integer, POINTER(NagError)]


def f07jsc(order, uplo, n, nrhs, d, e, b, pdb, fail):
    """
    Solves a complex Hermitian positive definite tridiagonal system using
    the LDL^T factorization computed by f07jrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d
    TypeError will be raised when the following arguments are not of data type
    Complex:
        e, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jsc', d, 'd')
    _e = nag_complex_type_check_and_cast('f07jsc', e, 'e')
    _b = nag_complex_type_check_and_cast('f07jsc', b, 'b')
    f07jsc_ctypes(order, uplo, n, nrhs, _d, _e, _b, pdb, fail)

f07juc_ctypes = _library_cl.f07juc
f07juc_ctypes.restype = None
f07juc_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(Complex), c_double, POINTER(c_double), POINTER(NagError)]


def f07juc(n, d, e, anorm, rcond, fail):
    """
    Computes the reciprocal of the condition number of a complex Hermitian
    positive definite tridiagonal system using the LDL^T factorization
    computed by f07jrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07juc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        e
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07juc', d, 'd')
    _e = nag_complex_type_check_and_cast('f07juc', e, 'e')
    _rcond = nag_double_type_check_and_cast('f07juc', rcond, 'rcond')
    f07juc_ctypes(n, _d, _e, anorm, _rcond, fail)

f07jvc_ctypes = _library_cl.f07jvc
f07jvc_ctypes.restype = None
f07jvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(Complex), POINTER(c_double), POINTER(Complex), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07jvc(order, uplo, n, nrhs, d, e, df, ef, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex Hermitian positive
    definite tridiagonal system of linear equations, multiple right-hand
    sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07jvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        d, e, df, ef, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, df, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        e, ef, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _d = nag_double_type_check_and_cast('f07jvc', d, 'd')
    _e = nag_complex_type_check_and_cast('f07jvc', e, 'e')
    _df = nag_double_type_check_and_cast('f07jvc', df, 'df')
    _ef = nag_complex_type_check_and_cast('f07jvc', ef, 'ef')
    _b = nag_complex_type_check_and_cast('f07jvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07jvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07jvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07jvc', berr, 'berr')
    f07jvc_ctypes(order, uplo, n, nrhs, _d, _e, _df, _ef, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07kdc_ctypes = _library_cl.f07kdc
f07kdc_ctypes.restype = None
f07kdc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(Integer), c_double, POINTER(NagError)]


def f07kdc(order, uplo, n, a, pda, piv, rank, tol, fail):
    """
    Cholesky factorization, with complete pivoting, of a real, symmetric,
    positive semidefinite matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07kdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, piv, rank
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        piv, rank
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07kdc', a, 'a')
    _piv = nag_integer_type_check_and_cast('f07kdc', piv, 'piv')
    _rank = nag_integer_type_check_and_cast('f07kdc', rank, 'rank')
    f07kdc_ctypes(order, uplo, n, _a, pda, _piv, _rank, tol, fail)

nag_dpstrf = f07kdc

f07krc_ctypes = _library_cl.f07krc
f07krc_ctypes.restype = None
f07krc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Integer), c_double, POINTER(NagError)]


def f07krc(order, uplo, n, a, pda, piv, rank, tol, fail):
    """
    Cholesky factorization of complex Hermitian positive semidefinite
    matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07krc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, piv, rank
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        piv, rank
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_integer_type_check_and_cast
    from .util import nag_complex_type_check_and_cast
    _piv = nag_integer_type_check_and_cast('f07krc', piv, 'piv')
    _rank = nag_integer_type_check_and_cast('f07krc', rank, 'rank')
    _a = nag_complex_type_check_and_cast('f07krc', a, 'a')
    f07krc_ctypes(order, uplo, n, _a, pda, _piv, _rank, tol, fail)

nag_zpstrf = f07krc

f07mac_ctypes = _library_cl.f07mac
f07mac_ctypes.restype = None
f07mac_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07mac(order, uplo, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Computes the solution to a real symmetric system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07mac', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mac', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07mac', b, 'b')
    f07mac_ctypes(order, uplo, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07mbc_ctypes = _library_cl.f07mbc
f07mbc_ctypes.restype = None
f07mbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07mbc(order, fact, uplo, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the diagonal pivoting factorization to compute the solution to a
    real symmetric system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, af, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07mbc', a, 'a')
    _af = nag_double_type_check_and_cast('f07mbc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07mbc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07mbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07mbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07mbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07mbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07mbc', berr, 'berr')
    f07mbc_ctypes(order, fact, uplo, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07mdc_ctypes = _library_cl.f07mdc
f07mdc_ctypes.restype = None
f07mdc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(NagError)]


def f07mdc(order, uplo, n, a, pda, ipiv, fail):
    """
    Bunch--Kaufman factorization of real symmetric indefinite matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07mdc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mdc', ipiv, 'ipiv')
    f07mdc_ctypes(order, uplo, n, _a, pda, _ipiv, fail)

f07mec_ctypes = _library_cl.f07mec
f07mec_ctypes.restype = None
f07mec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07mec(order, uplo, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Solution of real symmetric indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by f07mdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07mec', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mec', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07mec', b, 'b')
    f07mec_ctypes(order, uplo, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07mgc_ctypes = _library_cl.f07mgc
f07mgc_ctypes.restype = None
f07mgc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), Integer, POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07mgc(order, uplo, n, a, pda, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of real symmetric indefinite matrix, matrix
    already factorized by f07mdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07mgc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mgc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07mgc', rcond, 'rcond')
    f07mgc_ctypes(order, uplo, n, _a, pda, _ipiv, anorm, _rcond, fail)

f07mhc_ctypes = _library_cl.f07mhc
f07mhc_ctypes.restype = None
f07mhc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07mhc(order, uplo, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real symmetric indefinite system
    of linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, af, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07mhc', a, 'a')
    _af = nag_double_type_check_and_cast('f07mhc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07mhc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07mhc', b, 'b')
    _x = nag_double_type_check_and_cast('f07mhc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07mhc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07mhc', berr, 'berr')
    f07mhc_ctypes(order, uplo, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07mjc_ctypes = _library_cl.f07mjc
f07mjc_ctypes.restype = None
f07mjc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(NagError)]


def f07mjc(order, uplo, n, a, pda, ipiv, fail):
    """
    Inverse of real symmetric indefinite matrix, matrix already factorized
    by f07mdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mjc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07mjc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mjc', ipiv, 'ipiv')
    f07mjc_ctypes(order, uplo, n, _a, pda, _ipiv, fail)

f07mnc_ctypes = _library_cl.f07mnc
f07mnc_ctypes.restype = None
f07mnc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07mnc(order, uplo, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Computes the solution to a complex Hermitian system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07mnc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mnc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07mnc', b, 'b')
    f07mnc_ctypes(order, uplo, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07mpc_ctypes = _library_cl.f07mpc
f07mpc_ctypes.restype = None
f07mpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07mpc(order, fact, uplo, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the diagonal pivoting factorization to compute the solution to a
    complex Hermitian system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07mpc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07mpc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07mpc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07mpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07mpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07mpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07mpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07mpc', berr, 'berr')
    f07mpc_ctypes(order, fact, uplo, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07mrc_ctypes = _library_cl.f07mrc
f07mrc_ctypes.restype = None
f07mrc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07mrc(order, uplo, n, a, pda, ipiv, fail):
    """
    Bunch--Kaufman factorization of complex Hermitian indefinite matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mrc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07mrc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mrc', ipiv, 'ipiv')
    f07mrc_ctypes(order, uplo, n, _a, pda, _ipiv, fail)

f07msc_ctypes = _library_cl.f07msc
f07msc_ctypes.restype = None
f07msc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07msc(order, uplo, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Solution of complex Hermitian indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by f07mrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07msc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07msc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07msc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07msc', b, 'b')
    f07msc_ctypes(order, uplo, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07muc_ctypes = _library_cl.f07muc
f07muc_ctypes.restype = None
f07muc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07muc(order, uplo, n, a, pda, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of complex Hermitian indefinite matrix, matrix
    already factorized by f07mrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07muc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07muc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07muc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07muc', rcond, 'rcond')
    f07muc_ctypes(order, uplo, n, _a, pda, _ipiv, anorm, _rcond, fail)

f07mvc_ctypes = _library_cl.f07mvc
f07mvc_ctypes.restype = None
f07mvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07mvc(order, uplo, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex Hermitian indefinite
    system of linear equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07mvc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07mvc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07mvc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07mvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07mvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07mvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07mvc', berr, 'berr')
    f07mvc_ctypes(order, uplo, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07mwc_ctypes = _library_cl.f07mwc
f07mwc_ctypes.restype = None
f07mwc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07mwc(order, uplo, n, a, pda, ipiv, fail):
    """
    Inverse of complex Hermitian indefinite matrix, matrix already
    factorized by f07mrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07mwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07mwc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07mwc', ipiv, 'ipiv')
    f07mwc_ctypes(order, uplo, n, _a, pda, _ipiv, fail)

f07nnc_ctypes = _library_cl.f07nnc
f07nnc_ctypes.restype = None
f07nnc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07nnc(order, uplo, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Computes the solution to a complex symmetric system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07nnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07nnc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07nnc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07nnc', b, 'b')
    f07nnc_ctypes(order, uplo, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07npc_ctypes = _library_cl.f07npc
f07npc_ctypes.restype = None
f07npc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07npc(order, fact, uplo, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the diagonal pivoting factorization to compute the solution to a
    complex symmetric system of linear equations.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07npc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07npc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07npc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07npc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07npc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07npc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07npc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07npc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07npc', berr, 'berr')
    f07npc_ctypes(order, fact, uplo, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07nrc_ctypes = _library_cl.f07nrc
f07nrc_ctypes.restype = None
f07nrc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07nrc(order, uplo, n, a, pda, ipiv, fail):
    """
    Bunch--Kaufman factorization of complex symmetric matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07nrc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07nrc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07nrc', ipiv, 'ipiv')
    f07nrc_ctypes(order, uplo, n, _a, pda, _ipiv, fail)

f07nsc_ctypes = _library_cl.f07nsc
f07nsc_ctypes.restype = None
f07nsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07nsc(order, uplo, n, nrhs, a, pda, ipiv, b, pdb, fail):
    """
    Solution of complex symmetric system of linear equations, multiple
    right-hand sides, matrix already factorized by f07nrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07nsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07nsc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07nsc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07nsc', b, 'b')
    f07nsc_ctypes(order, uplo, n, nrhs, _a, pda, _ipiv, _b, pdb, fail)

f07nuc_ctypes = _library_cl.f07nuc
f07nuc_ctypes.restype = None
f07nuc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07nuc(order, uplo, n, a, pda, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of complex symmetric matrix, matrix already
    factorized by f07nrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07nuc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07nuc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07nuc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07nuc', rcond, 'rcond')
    f07nuc_ctypes(order, uplo, n, _a, pda, _ipiv, anorm, _rcond, fail)

f07nvc_ctypes = _library_cl.f07nvc
f07nvc_ctypes.restype = None
f07nvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07nvc(order, uplo, n, nrhs, a, pda, af, pdaf, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex symmetric system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07nvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, af, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, af, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07nvc', a, 'a')
    _af = nag_complex_type_check_and_cast('f07nvc', af, 'af')
    _ipiv = nag_integer_type_check_and_cast('f07nvc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07nvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07nvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07nvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07nvc', berr, 'berr')
    f07nvc_ctypes(order, uplo, n, nrhs, _a, pda, _af, pdaf, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07nwc_ctypes = _library_cl.f07nwc
f07nwc_ctypes.restype = None
f07nwc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(NagError)]


def f07nwc(order, uplo, n, a, pda, ipiv, fail):
    """
    Inverse of complex symmetric matrix, matrix already factorized by
    f07nrc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07nwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07nwc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f07nwc', ipiv, 'ipiv')
    f07nwc_ctypes(order, uplo, n, _a, pda, _ipiv, fail)

f07pac_ctypes = _library_cl.f07pac
f07pac_ctypes.restype = None
f07pac_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07pac(order, uplo, n, nrhs, ap, ipiv, b, pdb, fail):
    """
    Computes the solution to a real symmetric system of linear equations,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07pac', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07pac', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07pac', b, 'b')
    f07pac_ctypes(order, uplo, n, nrhs, _ap, _ipiv, _b, pdb, fail)

f07pbc_ctypes = _library_cl.f07pbc
f07pbc_ctypes.restype = None
f07pbc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07pbc(order, fact, uplo, n, nrhs, ap, afp, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the diagonal pivoting factorization to compute the solution to a
    real symmetric system of linear equations, packed storage. Error bounds
    and a condition estimate are also computed.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, afp, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07pbc', ap, 'ap')
    _afp = nag_double_type_check_and_cast('f07pbc', afp, 'afp')
    _ipiv = nag_integer_type_check_and_cast('f07pbc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07pbc', b, 'b')
    _x = nag_double_type_check_and_cast('f07pbc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07pbc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07pbc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07pbc', berr, 'berr')
    f07pbc_ctypes(order, fact, uplo, n, nrhs, _ap, _afp, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07pdc_ctypes = _library_cl.f07pdc
f07pdc_ctypes.restype = None
f07pdc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f07pdc(order, uplo, n, ap, ipiv, fail):
    """
    Bunch--Kaufman factorization of real symmetric indefinite matrix, packed
    storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07pdc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07pdc', ipiv, 'ipiv')
    f07pdc_ctypes(order, uplo, n, _ap, _ipiv, fail)

f07pec_ctypes = _library_cl.f07pec
f07pec_ctypes.restype = None
f07pec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, POINTER(NagError)]


def f07pec(order, uplo, n, nrhs, ap, ipiv, b, pdb, fail):
    """
    Solution of real symmetric indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by f07pdc, packed
    storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07pec', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07pec', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07pec', b, 'b')
    f07pec_ctypes(order, uplo, n, nrhs, _ap, _ipiv, _b, pdb, fail)

f07pgc_ctypes = _library_cl.f07pgc
f07pgc_ctypes.restype = None
f07pgc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07pgc(order, uplo, n, ap, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of real symmetric indefinite matrix, matrix
    already factorized by f07pdc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07pgc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07pgc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07pgc', rcond, 'rcond')
    f07pgc_ctypes(order, uplo, n, _ap, _ipiv, anorm, _rcond, fail)

f07phc_ctypes = _library_cl.f07phc
f07phc_ctypes.restype = None
f07phc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07phc(order, uplo, n, nrhs, ap, afp, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of real symmetric indefinite system
    of linear equations, multiple right-hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07phc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, afp, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07phc', ap, 'ap')
    _afp = nag_double_type_check_and_cast('f07phc', afp, 'afp')
    _ipiv = nag_integer_type_check_and_cast('f07phc', ipiv, 'ipiv')
    _b = nag_double_type_check_and_cast('f07phc', b, 'b')
    _x = nag_double_type_check_and_cast('f07phc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07phc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07phc', berr, 'berr')
    f07phc_ctypes(order, uplo, n, nrhs, _ap, _afp, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07pjc_ctypes = _library_cl.f07pjc
f07pjc_ctypes.restype = None
f07pjc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f07pjc(order, uplo, n, ap, ipiv, fail):
    """
    Inverse of real symmetric indefinite matrix, matrix already factorized
    by f07pdc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pjc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07pjc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07pjc', ipiv, 'ipiv')
    f07pjc_ctypes(order, uplo, n, _ap, _ipiv, fail)

f07pnc_ctypes = _library_cl.f07pnc
f07pnc_ctypes.restype = None
f07pnc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07pnc(order, uplo, n, nrhs, ap, ipiv, b, pdb, fail):
    """
    Computes the solution to a complex Hermitian system of linear equations,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07pnc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07pnc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07pnc', b, 'b')
    f07pnc_ctypes(order, uplo, n, nrhs, _ap, _ipiv, _b, pdb, fail)

f07ppc_ctypes = _library_cl.f07ppc
f07ppc_ctypes.restype = None
f07ppc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07ppc(order, fact, uplo, n, nrhs, ap, afp, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the diagonal pivoting factorization to compute the solution to a
    complex, Hermitian, system of linear equations, error bounds and
    condition estimates. Packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ppc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, afp, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07ppc', ap, 'ap')
    _afp = nag_complex_type_check_and_cast('f07ppc', afp, 'afp')
    _ipiv = nag_integer_type_check_and_cast('f07ppc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07ppc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07ppc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07ppc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07ppc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07ppc', berr, 'berr')
    f07ppc_ctypes(order, fact, uplo, n, nrhs, _ap, _afp, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07prc_ctypes = _library_cl.f07prc
f07prc_ctypes.restype = None
f07prc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f07prc(order, uplo, n, ap, ipiv, fail):
    """
    Bunch--Kaufman factorization of complex Hermitian indefinite matrix,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07prc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07prc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07prc', ipiv, 'ipiv')
    f07prc_ctypes(order, uplo, n, _ap, _ipiv, fail)

f07psc_ctypes = _library_cl.f07psc
f07psc_ctypes.restype = None
f07psc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07psc(order, uplo, n, nrhs, ap, ipiv, b, pdb, fail):
    """
    Solution of complex Hermitian indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by f07prc, packed
    storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07psc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07psc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07psc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07psc', b, 'b')
    f07psc_ctypes(order, uplo, n, nrhs, _ap, _ipiv, _b, pdb, fail)

f07puc_ctypes = _library_cl.f07puc
f07puc_ctypes.restype = None
f07puc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07puc(order, uplo, n, ap, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of complex Hermitian indefinite matrix, matrix
    already factorized by f07prc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07puc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07puc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07puc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07puc', rcond, 'rcond')
    f07puc_ctypes(order, uplo, n, _ap, _ipiv, anorm, _rcond, fail)

f07pvc_ctypes = _library_cl.f07pvc
f07pvc_ctypes.restype = None
f07pvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07pvc(order, uplo, n, nrhs, ap, afp, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex Hermitian indefinite
    system of linear equations, multiple right-hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, afp, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07pvc', ap, 'ap')
    _afp = nag_complex_type_check_and_cast('f07pvc', afp, 'afp')
    _ipiv = nag_integer_type_check_and_cast('f07pvc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07pvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07pvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07pvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07pvc', berr, 'berr')
    f07pvc_ctypes(order, uplo, n, nrhs, _ap, _afp, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07pwc_ctypes = _library_cl.f07pwc
f07pwc_ctypes.restype = None
f07pwc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f07pwc(order, uplo, n, ap, ipiv, fail):
    """
    Inverse of complex Hermitian indefinite matrix, matrix already
    factorized by f07prc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07pwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07pwc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07pwc', ipiv, 'ipiv')
    f07pwc_ctypes(order, uplo, n, _ap, _ipiv, fail)

f07qnc_ctypes = _library_cl.f07qnc
f07qnc_ctypes.restype = None
f07qnc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07qnc(order, uplo, n, nrhs, ap, ipiv, b, pdb, fail):
    """
    Computes the solution to a complex symmetric system of linear equations,
    packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07qnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07qnc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07qnc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07qnc', b, 'b')
    f07qnc_ctypes(order, uplo, n, nrhs, _ap, _ipiv, _b, pdb, fail)

f07qpc_ctypes = _library_cl.f07qpc
f07qpc_ctypes.restype = None
f07qpc_ctypes.argtypes = [Nag_OrderType, Nag_FactoredFormType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07qpc(order, fact, uplo, n, nrhs, ap, afp, ipiv, b, pdb, x, pdx, rcond, ferr, berr, fail):
    """
    Uses the diagonal pivoting factorization to compute the solution to a
    complex, symmetric, system of linear equations, error bounds and
    condition estimates. Packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07qpc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, ipiv, b, x, rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, afp, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07qpc', ap, 'ap')
    _afp = nag_complex_type_check_and_cast('f07qpc', afp, 'afp')
    _ipiv = nag_integer_type_check_and_cast('f07qpc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07qpc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07qpc', x, 'x')
    _rcond = nag_double_type_check_and_cast('f07qpc', rcond, 'rcond')
    _ferr = nag_double_type_check_and_cast('f07qpc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07qpc', berr, 'berr')
    f07qpc_ctypes(order, fact, uplo, n, nrhs, _ap, _afp, _ipiv, _b, pdb, _x, pdx, _rcond, _ferr, _berr, fail)

f07qrc_ctypes = _library_cl.f07qrc
f07qrc_ctypes.restype = None
f07qrc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f07qrc(order, uplo, n, ap, ipiv, fail):
    """
    Bunch--Kaufman factorization of complex symmetric matrix, packed
    storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07qrc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07qrc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07qrc', ipiv, 'ipiv')
    f07qrc_ctypes(order, uplo, n, _ap, _ipiv, fail)

f07qsc_ctypes = _library_cl.f07qsc
f07qsc_ctypes.restype = None
f07qsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(NagError)]


def f07qsc(order, uplo, n, nrhs, ap, ipiv, b, pdb, fail):
    """
    Solution of complex symmetric system of linear equations, multiple
    right-hand sides, matrix already factorized by f07qrc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07qsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, b
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07qsc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07qsc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07qsc', b, 'b')
    f07qsc_ctypes(order, uplo, n, nrhs, _ap, _ipiv, _b, pdb, fail)

f07quc_ctypes = _library_cl.f07quc
f07quc_ctypes.restype = None
f07quc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(Integer), c_double, POINTER(c_double), POINTER(NagError)]


def f07quc(order, uplo, n, ap, ipiv, anorm, rcond, fail):
    """
    Estimate condition number of complex symmetric matrix, matrix already
    factorized by f07qrc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07quc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07quc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07quc', ipiv, 'ipiv')
    _rcond = nag_double_type_check_and_cast('f07quc', rcond, 'rcond')
    f07quc_ctypes(order, uplo, n, _ap, _ipiv, anorm, _rcond, fail)

f07qvc_ctypes = _library_cl.f07qvc
f07qvc_ctypes.restype = None
f07qvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07qvc(order, uplo, n, nrhs, ap, afp, ipiv, b, pdb, x, pdx, ferr, berr, fail):
    """
    Refined solution with error bounds of complex symmetric system of linear
    equations, multiple right-hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07qvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, afp, ipiv, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, afp, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07qvc', ap, 'ap')
    _afp = nag_complex_type_check_and_cast('f07qvc', afp, 'afp')
    _ipiv = nag_integer_type_check_and_cast('f07qvc', ipiv, 'ipiv')
    _b = nag_complex_type_check_and_cast('f07qvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07qvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07qvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07qvc', berr, 'berr')
    f07qvc_ctypes(order, uplo, n, nrhs, _ap, _afp, _ipiv, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07qwc_ctypes = _library_cl.f07qwc
f07qwc_ctypes.restype = None
f07qwc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f07qwc(order, uplo, n, ap, ipiv, fail):
    """
    Inverse of complex symmetric matrix, matrix already factorized by
    f07qrc, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07qwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, ipiv
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07qwc', ap, 'ap')
    _ipiv = nag_integer_type_check_and_cast('f07qwc', ipiv, 'ipiv')
    f07qwc_ctypes(order, uplo, n, _ap, _ipiv, fail)

f07tec_ctypes = _library_cl.f07tec
f07tec_ctypes.restype = None
f07tec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07tec(order, uplo, trans, diag, n, nrhs, a, pda, b, pdb, fail):
    """
    Solution of real triangular system of linear equations, multiple right-
    hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07tec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07tec', a, 'a')
    _b = nag_double_type_check_and_cast('f07tec', b, 'b')
    f07tec_ctypes(order, uplo, trans, diag, n, nrhs, _a, pda, _b, pdb, fail)

f07tgc_ctypes = _library_cl.f07tgc
f07tgc_ctypes.restype = None
f07tgc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Nag_UploType, Nag_DiagType, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(NagError)]


def f07tgc(order, norm, uplo, diag, n, a, pda, rcond, fail):
    """
    Estimate condition number of real triangular matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07tgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, rcond
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07tgc', a, 'a')
    _rcond = nag_double_type_check_and_cast('f07tgc', rcond, 'rcond')
    f07tgc_ctypes(order, norm, uplo, diag, n, _a, pda, _rcond, fail)

f07thc_ctypes = _library_cl.f07thc
f07thc_ctypes.restype = None
f07thc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07thc(order, uplo, trans, diag, n, nrhs, a, pda, b, pdb, x, pdx, ferr, berr, fail):
    """
    Error bounds for solution of real triangular system of linear equations,
    multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07thc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, b, x, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07thc', a, 'a')
    _b = nag_double_type_check_and_cast('f07thc', b, 'b')
    _x = nag_double_type_check_and_cast('f07thc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07thc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07thc', berr, 'berr')
    f07thc_ctypes(order, uplo, trans, diag, n, nrhs, _a, pda, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07tjc_ctypes = _library_cl.f07tjc
f07tjc_ctypes.restype = None
f07tjc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_DiagType, Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07tjc(order, uplo, diag, n, a, pda, fail):
    """
    Inverse of real triangular matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07tjc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a
    """
    from .util import nag_double_type_check_and_cast
    _a = nag_double_type_check_and_cast('f07tjc', a, 'a')
    f07tjc_ctypes(order, uplo, diag, n, _a, pda, fail)

f07tsc_ctypes = _library_cl.f07tsc
f07tsc_ctypes.restype = None
f07tsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07tsc(order, uplo, trans, diag, n, nrhs, a, pda, b, pdb, fail):
    """
    Solution of complex triangular system of linear equations, multiple
    right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07tsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b
    """
    from .util import nag_complex_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07tsc', a, 'a')
    _b = nag_complex_type_check_and_cast('f07tsc', b, 'b')
    f07tsc_ctypes(order, uplo, trans, diag, n, nrhs, _a, pda, _b, pdb, fail)

f07tuc_ctypes = _library_cl.f07tuc
f07tuc_ctypes.restype = None
f07tuc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Nag_UploType, Nag_DiagType, Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(NagError)]


def f07tuc(order, norm, uplo, diag, n, a, pda, rcond, fail):
    """
    Estimate condition number of complex triangular matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07tuc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07tuc', a, 'a')
    _rcond = nag_double_type_check_and_cast('f07tuc', rcond, 'rcond')
    f07tuc_ctypes(order, norm, uplo, diag, n, _a, pda, _rcond, fail)

f07tvc_ctypes = _library_cl.f07tvc
f07tvc_ctypes.restype = None
f07tvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07tvc(order, uplo, trans, diag, n, nrhs, a, pda, b, pdb, x, pdx, ferr, berr, fail):
    """
    Error bounds for solution of complex triangular system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07tvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07tvc', a, 'a')
    _b = nag_complex_type_check_and_cast('f07tvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07tvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07tvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07tvc', berr, 'berr')
    f07tvc_ctypes(order, uplo, trans, diag, n, nrhs, _a, pda, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07twc_ctypes = _library_cl.f07twc
f07twc_ctypes.restype = None
f07twc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_DiagType, Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07twc(order, uplo, diag, n, a, pda, fail):
    """
    Inverse of complex triangular matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07twc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a
    """
    from .util import nag_complex_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f07twc', a, 'a')
    f07twc_ctypes(order, uplo, diag, n, _a, pda, fail)

f07uec_ctypes = _library_cl.f07uec
f07uec_ctypes.restype = None
f07uec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(NagError)]


def f07uec(order, uplo, trans, diag, n, nrhs, ap, b, pdb, fail):
    """
    Solution of real triangular system of linear equations, multiple right-
    hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07uec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, b
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07uec', ap, 'ap')
    _b = nag_double_type_check_and_cast('f07uec', b, 'b')
    f07uec_ctypes(order, uplo, trans, diag, n, nrhs, _ap, _b, pdb, fail)

f07ugc_ctypes = _library_cl.f07ugc
f07ugc_ctypes.restype = None
f07ugc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Nag_UploType, Nag_DiagType, Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07ugc(order, norm, uplo, diag, n, ap, rcond, fail):
    """
    Estimate condition number of real triangular matrix, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ugc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, rcond
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07ugc', ap, 'ap')
    _rcond = nag_double_type_check_and_cast('f07ugc', rcond, 'rcond')
    f07ugc_ctypes(order, norm, uplo, diag, n, _ap, _rcond, fail)

f07uhc_ctypes = _library_cl.f07uhc
f07uhc_ctypes.restype = None
f07uhc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07uhc(order, uplo, trans, diag, n, nrhs, ap, b, pdb, x, pdx, ferr, berr, fail):
    """
    Error bounds for solution of real triangular system of linear equations,
    multiple right-hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07uhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap, b, x, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07uhc', ap, 'ap')
    _b = nag_double_type_check_and_cast('f07uhc', b, 'b')
    _x = nag_double_type_check_and_cast('f07uhc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07uhc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07uhc', berr, 'berr')
    f07uhc_ctypes(order, uplo, trans, diag, n, nrhs, _ap, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07ujc_ctypes = _library_cl.f07ujc
f07ujc_ctypes.restype = None
f07ujc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_DiagType, Integer, POINTER(c_double), POINTER(NagError)]


def f07ujc(order, uplo, diag, n, ap, fail):
    """
    Inverse of real triangular matrix, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07ujc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ap
    """
    from .util import nag_double_type_check_and_cast
    _ap = nag_double_type_check_and_cast('f07ujc', ap, 'ap')
    f07ujc_ctypes(order, uplo, diag, n, _ap, fail)

f07usc_ctypes = _library_cl.f07usc
f07usc_ctypes.restype = None
f07usc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(Complex), POINTER(Complex), Integer, POINTER(NagError)]


def f07usc(order, uplo, trans, diag, n, nrhs, ap, b, pdb, fail):
    """
    Solution of complex triangular system of linear equations, multiple
    right-hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07usc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b
    """
    from .util import nag_complex_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07usc', ap, 'ap')
    _b = nag_complex_type_check_and_cast('f07usc', b, 'b')
    f07usc_ctypes(order, uplo, trans, diag, n, nrhs, _ap, _b, pdb, fail)

f07uuc_ctypes = _library_cl.f07uuc
f07uuc_ctypes.restype = None
f07uuc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Nag_UploType, Nag_DiagType, Integer, POINTER(Complex), POINTER(c_double), POINTER(NagError)]


def f07uuc(order, norm, uplo, diag, n, ap, rcond, fail):
    """
    Estimate condition number of complex triangular matrix, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07uuc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07uuc', ap, 'ap')
    _rcond = nag_double_type_check_and_cast('f07uuc', rcond, 'rcond')
    f07uuc_ctypes(order, norm, uplo, diag, n, _ap, _rcond, fail)

f07uvc_ctypes = _library_cl.f07uvc
f07uvc_ctypes.restype = None
f07uvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, POINTER(Complex), POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07uvc(order, uplo, trans, diag, n, nrhs, ap, b, pdb, x, pdx, ferr, berr, fail):
    """
    Error bounds for solution of complex triangular system of linear
    equations, multiple right-hand sides, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07uvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07uvc', ap, 'ap')
    _b = nag_complex_type_check_and_cast('f07uvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07uvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07uvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07uvc', berr, 'berr')
    f07uvc_ctypes(order, uplo, trans, diag, n, nrhs, _ap, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07uwc_ctypes = _library_cl.f07uwc
f07uwc_ctypes.restype = None
f07uwc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_DiagType, Integer, POINTER(Complex), POINTER(NagError)]


def f07uwc(order, uplo, diag, n, ap, fail):
    """
    Inverse of complex triangular matrix, packed storage.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07uwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ap
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ap
    """
    from .util import nag_complex_type_check_and_cast
    _ap = nag_complex_type_check_and_cast('f07uwc', ap, 'ap')
    f07uwc_ctypes(order, uplo, diag, n, _ap, fail)

f07vec_ctypes = _library_cl.f07vec
f07vec_ctypes.restype = None
f07vec_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f07vec(order, uplo, trans, diag, n, kd, nrhs, ab, pdab, b, pdb, fail):
    """
    Solution of real band triangular system of linear equations, multiple
    right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07vec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, b
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07vec', ab, 'ab')
    _b = nag_double_type_check_and_cast('f07vec', b, 'b')
    f07vec_ctypes(order, uplo, trans, diag, n, kd, nrhs, _ab, pdab, _b, pdb, fail)

f07vgc_ctypes = _library_cl.f07vgc
f07vgc_ctypes.restype = None
f07vgc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Nag_UploType, Nag_DiagType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(NagError)]


def f07vgc(order, norm, uplo, diag, n, kd, ab, pdab, rcond, fail):
    """
    Estimate condition number of real band triangular matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07vgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, rcond
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07vgc', ab, 'ab')
    _rcond = nag_double_type_check_and_cast('f07vgc', rcond, 'rcond')
    f07vgc_ctypes(order, norm, uplo, diag, n, kd, _ab, pdab, _rcond, fail)

f07vhc_ctypes = _library_cl.f07vhc
f07vhc_ctypes.restype = None
f07vhc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07vhc(order, uplo, trans, diag, n, kd, nrhs, ab, pdab, b, pdb, x, pdx, ferr, berr, fail):
    """
    Error bounds for solution of real band triangular system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07vhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, b, x, ferr, berr
    """
    from .util import nag_double_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f07vhc', ab, 'ab')
    _b = nag_double_type_check_and_cast('f07vhc', b, 'b')
    _x = nag_double_type_check_and_cast('f07vhc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07vhc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07vhc', berr, 'berr')
    f07vhc_ctypes(order, uplo, trans, diag, n, kd, nrhs, _ab, pdab, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07vsc_ctypes = _library_cl.f07vsc
f07vsc_ctypes.restype = None
f07vsc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f07vsc(order, uplo, trans, diag, n, kd, nrhs, ab, pdab, b, pdb, fail):
    """
    Solution of complex band triangular system of linear equations, multiple
    right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07vsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, b
    """
    from .util import nag_complex_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07vsc', ab, 'ab')
    _b = nag_complex_type_check_and_cast('f07vsc', b, 'b')
    f07vsc_ctypes(order, uplo, trans, diag, n, kd, nrhs, _ab, pdab, _b, pdb, fail)

f07vuc_ctypes = _library_cl.f07vuc
f07vuc_ctypes.restype = None
f07vuc_ctypes.argtypes = [Nag_OrderType, Nag_NormType, Nag_UploType, Nag_DiagType, Integer, Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(NagError)]


def f07vuc(order, norm, uplo, diag, n, kd, ab, pdab, rcond, fail):
    """
    Estimate condition number of complex band triangular matrix.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07vuc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, rcond
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        rcond
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07vuc', ab, 'ab')
    _rcond = nag_double_type_check_and_cast('f07vuc', rcond, 'rcond')
    f07vuc_ctypes(order, norm, uplo, diag, n, kd, _ab, pdab, _rcond, fail)

f07vvc_ctypes = _library_cl.f07vvc
f07vvc_ctypes.restype = None
f07vvc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Nag_TransType, Nag_DiagType, Integer, Integer, Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(Complex), Integer, POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def f07vvc(order, uplo, trans, diag, n, kd, nrhs, ab, pdab, b, pdb, x, pdx, ferr, berr, fail):
    """
    Error bounds for solution of complex band triangular system of linear
    equations, multiple right-hand sides.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07vvc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, b, x, ferr, berr
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ferr, berr
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, b, x
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f07vvc', ab, 'ab')
    _b = nag_complex_type_check_and_cast('f07vvc', b, 'b')
    _x = nag_complex_type_check_and_cast('f07vvc', x, 'x')
    _ferr = nag_double_type_check_and_cast('f07vvc', ferr, 'ferr')
    _berr = nag_double_type_check_and_cast('f07vvc', berr, 'berr')
    f07vvc_ctypes(order, uplo, trans, diag, n, kd, nrhs, _ab, pdab, _b, pdb, _x, pdx, _ferr, _berr, fail)

f07wdc_ctypes = _library_cl.f07wdc
f07wdc_ctypes.restype = None
f07wdc_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Integer, POINTER(c_double), POINTER(NagError)]


def f07wdc(order, transr, uplo, n, ar, fail):
    """
    Cholesky factorization of real symmetric positive definite matrix,
    Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ar
    """
    from .util import nag_double_type_check_and_cast
    _ar = nag_double_type_check_and_cast('f07wdc', ar, 'ar')
    f07wdc_ctypes(order, transr, uplo, n, _ar, fail)

nag_dpftrf = f07wdc

f07wec_ctypes = _library_cl.f07wec
f07wec_ctypes.restype = None
f07wec_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(NagError)]


def f07wec(order, transr, uplo, n, nrhs, ar, b, pdb, fail):
    """
    Solution of real symmetric positive definite system of linear equations,
    multiple right-hand sides, coefficient matrix already factorized by
    f07wdc, Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar, b
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ar, b
    """
    from .util import nag_double_type_check_and_cast
    _ar = nag_double_type_check_and_cast('f07wec', ar, 'ar')
    _b = nag_double_type_check_and_cast('f07wec', b, 'b')
    f07wec_ctypes(order, transr, uplo, n, nrhs, _ar, _b, pdb, fail)

nag_dpftrs = f07wec

f07wjc_ctypes = _library_cl.f07wjc
f07wjc_ctypes.restype = None
f07wjc_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Integer, POINTER(c_double), POINTER(NagError)]


def f07wjc(order, transr, uplo, n, ar, fail):
    """
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by f07wdc, Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wjc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ar
    """
    from .util import nag_double_type_check_and_cast
    _ar = nag_double_type_check_and_cast('f07wjc', ar, 'ar')
    f07wjc_ctypes(order, transr, uplo, n, _ar, fail)

nag_dpftri = f07wjc

f07wkc_ctypes = _library_cl.f07wkc
f07wkc_ctypes.restype = None
f07wkc_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Nag_DiagType, Integer, POINTER(c_double), POINTER(NagError)]


def f07wkc(order, transr, uplo, diag, n, ar, fail):
    """
    Inverse of real triangular matrix, Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wkc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ar
    """
    from .util import nag_double_type_check_and_cast
    _ar = nag_double_type_check_and_cast('f07wkc', ar, 'ar')
    f07wkc_ctypes(order, transr, uplo, diag, n, _ar, fail)

nag_dtftri = f07wkc

f07wrc_ctypes = _library_cl.f07wrc
f07wrc_ctypes.restype = None
f07wrc_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Integer, POINTER(Complex), POINTER(NagError)]


def f07wrc(order, transr, uplo, n, ar, fail):
    """
    Cholesky factorization of complex Hermitian positive definite matrix,
    Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wrc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ar
    """
    from .util import nag_complex_type_check_and_cast
    _ar = nag_complex_type_check_and_cast('f07wrc', ar, 'ar')
    f07wrc_ctypes(order, transr, uplo, n, _ar, fail)

nag_zpftrf = f07wrc

f07wsc_ctypes = _library_cl.f07wsc
f07wsc_ctypes.restype = None
f07wsc_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Integer, Integer, POINTER(Complex), POINTER(Complex), Integer, POINTER(NagError)]


def f07wsc(order, transr, uplo, n, nrhs, ar, b, pdb, fail):
    """
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, coefficient matrix already
    factorized by f07wrc, Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wsc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar, b
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ar, b
    """
    from .util import nag_complex_type_check_and_cast
    _ar = nag_complex_type_check_and_cast('f07wsc', ar, 'ar')
    _b = nag_complex_type_check_and_cast('f07wsc', b, 'b')
    f07wsc_ctypes(order, transr, uplo, n, nrhs, _ar, _b, pdb, fail)

nag_zpftrs = f07wsc

f07wwc_ctypes = _library_cl.f07wwc
f07wwc_ctypes.restype = None
f07wwc_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Integer, POINTER(Complex), POINTER(NagError)]


def f07wwc(order, transr, uplo, n, ar, fail):
    """
    Inverse of complex Hermitian positive definite matrix, matrix already
    factorized by f07wrc, Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wwc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ar
    """
    from .util import nag_complex_type_check_and_cast
    _ar = nag_complex_type_check_and_cast('f07wwc', ar, 'ar')
    f07wwc_ctypes(order, transr, uplo, n, _ar, fail)

nag_zpftri = f07wwc

f07wxc_ctypes = _library_cl.f07wxc
f07wxc_ctypes.restype = None
f07wxc_ctypes.argtypes = [Nag_OrderType, Nag_RFP_Store, Nag_UploType, Nag_DiagType, Integer, POINTER(Complex), POINTER(NagError)]


def f07wxc(order, transr, uplo, diag, n, ar, fail):
    """
    Inverse of complex triangular matrix, Rectangular Full Packed format.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f07/f07wxc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ar
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ar
    """
    from .util import nag_complex_type_check_and_cast
    _ar = nag_complex_type_check_and_cast('f07wxc', ar, 'ar')
    f07wxc_ctypes(order, transr, uplo, diag, n, _ar, fail)

nag_ztftri = f07wxc

nag_dgesv = f07aac
nag_dgesvx = f07abc
nag_dsgesv = f07acc
nag_dgetrf = f07adc
nag_dgetrs = f07aec
nag_dgeequ = f07afc
nag_dgecon = f07agc
nag_dgerfs = f07ahc
nag_dgetri = f07ajc
nag_zgesv = f07anc
nag_zgesvx = f07apc
nag_zcgesv = f07aqc
nag_zgetrf = f07arc
nag_zgetrs = f07asc
nag_zgeequ = f07atc
nag_zgecon = f07auc
nag_zgerfs = f07avc
nag_zgetri = f07awc
nag_dgbsv = f07bac
nag_dgbsvx = f07bbc
nag_dgbtrf = f07bdc
nag_dgbtrs = f07bec
nag_dgbequ = f07bfc
nag_dgbcon = f07bgc
nag_dgbrfs = f07bhc
nag_zgbsv = f07bnc
nag_zgbsvx = f07bpc
nag_zgbtrf = f07brc
nag_zgbtrs = f07bsc
nag_zgbequ = f07btc
nag_zgbcon = f07buc
nag_zgbrfs = f07bvc
nag_dgtsv = f07cac
nag_dgtsvx = f07cbc
nag_dgttrf = f07cdc
nag_dgttrs = f07cec
nag_dgtcon = f07cgc
nag_dgtrfs = f07chc
nag_zgtsv = f07cnc
nag_zgtsvx = f07cpc
nag_zgttrf = f07crc
nag_zgttrs = f07csc
nag_zgtcon = f07cuc
nag_zgtrfs = f07cvc
nag_dposv = f07fac
nag_dposvx = f07fbc
nag_dpotrf = f07fdc
nag_dpotrs = f07fec
nag_dpoequ = f07ffc
nag_dpocon = f07fgc
nag_dporfs = f07fhc
nag_dpotri = f07fjc
nag_zposv = f07fnc
nag_zposvx = f07fpc
nag_zpotrf = f07frc
nag_zpotrs = f07fsc
nag_zpoequ = f07ftc
nag_zpocon = f07fuc
nag_zporfs = f07fvc
nag_zpotri = f07fwc
nag_dppsv = f07gac
nag_dppsvx = f07gbc
nag_dpptrf = f07gdc
nag_dpptrs = f07gec
nag_dppequ = f07gfc
nag_dppcon = f07ggc
nag_dpprfs = f07ghc
nag_dpptri = f07gjc
nag_zppsv = f07gnc
nag_zppsvx = f07gpc
nag_zpptrf = f07grc
nag_zpptrs = f07gsc
nag_zppequ = f07gtc
nag_zppcon = f07guc
nag_zpprfs = f07gvc
nag_zpptri = f07gwc
nag_dpbsv = f07hac
nag_dpbsvx = f07hbc
nag_dpbtrf = f07hdc
nag_dpbtrs = f07hec
nag_dpbequ = f07hfc
nag_dpbcon = f07hgc
nag_dpbrfs = f07hhc
nag_zpbsv = f07hnc
nag_zpbsvx = f07hpc
nag_zpbtrf = f07hrc
nag_zpbtrs = f07hsc
nag_zpbequ = f07htc
nag_zpbcon = f07huc
nag_zpbrfs = f07hvc
nag_dptsv = f07jac
nag_dptsvx = f07jbc
nag_dpttrf = f07jdc
nag_dpttrs = f07jec
nag_dptcon = f07jgc
nag_dptrfs = f07jhc
nag_zptsv = f07jnc
nag_zptsvx = f07jpc
nag_zpttrf = f07jrc
nag_zpttrs = f07jsc
nag_zptcon = f07juc
nag_zptrfs = f07jvc
nag_dsysv = f07mac
nag_dsysvx = f07mbc
nag_dsytrf = f07mdc
nag_dsytrs = f07mec
nag_dsycon = f07mgc
nag_dsyrfs = f07mhc
nag_dsytri = f07mjc
nag_zhesv = f07mnc
nag_zhesvx = f07mpc
nag_zhetrf = f07mrc
nag_zhetrs = f07msc
nag_zhecon = f07muc
nag_zherfs = f07mvc
nag_zhetri = f07mwc
nag_zsysv = f07nnc
nag_zsysvx = f07npc
nag_zsytrf = f07nrc
nag_zsytrs = f07nsc
nag_zsycon = f07nuc
nag_zsyrfs = f07nvc
nag_zsytri = f07nwc
nag_dspsv = f07pac
nag_dspsvx = f07pbc
nag_dsptrf = f07pdc
nag_dsptrs = f07pec
nag_dspcon = f07pgc
nag_dsprfs = f07phc
nag_dsptri = f07pjc
nag_zhpsv = f07pnc
nag_zhpsvx = f07ppc
nag_zhptrf = f07prc
nag_zhptrs = f07psc
nag_zhpcon = f07puc
nag_zhprfs = f07pvc
nag_zhptri = f07pwc
nag_zspsv = f07qnc
nag_zspsvx = f07qpc
nag_zsptrf = f07qrc
nag_zsptrs = f07qsc
nag_zspcon = f07quc
nag_zsprfs = f07qvc
nag_zsptri = f07qwc
nag_dtrtrs = f07tec
nag_dtrcon = f07tgc
nag_dtrrfs = f07thc
nag_dtrtri = f07tjc
nag_ztrtrs = f07tsc
nag_ztrcon = f07tuc
nag_ztrrfs = f07tvc
nag_ztrtri = f07twc
nag_dtptrs = f07uec
nag_dtpcon = f07ugc
nag_dtprfs = f07uhc
nag_dtptri = f07ujc
nag_ztptrs = f07usc
nag_ztpcon = f07uuc
nag_ztprfs = f07uvc
nag_ztptri = f07uwc
nag_dtbtrs = f07vec
nag_dtbcon = f07vgc
nag_dtbrfs = f07vhc
nag_ztbtrs = f07vsc
nag_ztbcon = f07vuc
nag_ztbrfs = f07vvc
__all__ = ['f07aac', 'f07abc', 'f07acc', 'f07adc', 'f07aec', 'f07afc',
           'f07agc', 'f07ahc', 'f07ajc', 'f07anc', 'f07apc', 'f07aqc',
           'f07arc', 'f07asc', 'f07atc', 'f07auc', 'f07avc', 'f07awc',
           'f07bac', 'f07bbc', 'f07bdc', 'f07bec', 'f07bfc', 'f07bgc',
           'f07bhc', 'f07bnc', 'f07bpc', 'f07brc', 'f07bsc', 'f07btc',
           'f07buc', 'f07bvc', 'f07cac', 'f07cbc', 'f07cdc', 'f07cec',
           'f07cgc', 'f07chc', 'f07cnc', 'f07cpc', 'f07crc', 'f07csc',
           'f07cuc', 'f07cvc', 'f07fac', 'f07fbc', 'f07fdc',
           'f07fec', 'f07ffc', 'f07fgc', 'f07fhc', 'f07fjc', 'f07fnc',
           'f07fpc', 'f07frc', 'f07fsc', 'f07ftc', 'f07fuc',
           'f07fvc', 'f07fwc', 'f07gac', 'f07gbc', 'f07gdc', 'f07gec',
           'f07gfc', 'f07ggc', 'f07ghc', 'f07gjc', 'f07gnc', 'f07gpc',
           'f07grc', 'f07gsc', 'f07gtc', 'f07guc', 'f07gvc', 'f07gwc',
           'f07hac', 'f07hbc', 'f07hdc', 'f07hec', 'f07hfc', 'f07hgc',
           'f07hhc', 'f07hnc', 'f07hpc', 'f07hrc', 'f07hsc', 'f07htc',
           'f07huc', 'f07hvc', 'f07jac', 'f07jbc', 'f07jdc', 'f07jec',
           'f07jgc', 'f07jhc', 'f07jnc', 'f07jpc', 'f07jrc', 'f07jsc',
           'f07juc', 'f07jvc', 'f07mac', 'f07mbc',
           'f07mdc', 'f07mec', 'f07mgc', 'f07mhc', 'f07mjc', 'f07mnc',
           'f07mpc', 'f07mrc', 'f07msc', 'f07muc', 'f07mvc', 'f07mwc',
           'f07nnc', 'f07npc', 'f07nrc', 'f07nsc', 'f07nuc', 'f07nvc',
           'f07nwc', 'f07pac', 'f07pbc', 'f07pdc', 'f07pec', 'f07pgc',
           'f07phc', 'f07pjc', 'f07pnc', 'f07ppc', 'f07prc', 'f07psc',
           'f07puc', 'f07pvc', 'f07pwc', 'f07qnc', 'f07qpc', 'f07qrc',
           'f07qsc', 'f07quc', 'f07qvc', 'f07qwc', 'f07tec', 'f07tgc',
           'f07thc', 'f07tjc', 'f07tsc', 'f07tuc', 'f07tvc', 'f07twc',
           'f07uec', 'f07ugc', 'f07uhc', 'f07ujc', 'f07usc', 'f07uuc',
           'f07uvc', 'f07uwc', 'f07vec', 'f07vgc', 'f07vhc', 'f07vsc',
           'f07vuc', 'f07vvc',
           'nag_dgbcon',
           'nag_dgbequ', 'nag_dgbrfs', 'nag_dgbsv', 'nag_dgbsvx',
           'nag_dgbtrf', 'nag_dgbtrs', 'nag_dgecon', 'nag_dgeequ',
           'nag_dgerfs', 'nag_dgesv', 'nag_dgesvx', 'nag_dgetrf',
           'nag_dgetri', 'nag_dgetrs', 'nag_dgtcon', 'nag_dgtrfs',
           'nag_dgtsv', 'nag_dgtsvx', 'nag_dgttrf', 'nag_dgttrs',
           'nag_dpbcon', 'nag_dpbequ', 'nag_dpbrfs', 'nag_dpbsv',
           'nag_dpbsvx', 'nag_dpbtrf', 'nag_dpbtrs',
           'nag_dpocon', 'nag_dpoequ',
           'nag_dporfs', 'nag_dposv', 'nag_dposvx', 'nag_dpotrf',
           'nag_dpotri', 'nag_dpotrs', 'nag_dppcon', 'nag_dppequ',
           'nag_dpprfs', 'nag_dppsv', 'nag_dppsvx', 'nag_dpptrf',
           'nag_dpptri', 'nag_dpptrs', 'nag_dptcon',
           'nag_dptrfs', 'nag_dptsv', 'nag_dptsvx', 'nag_dpttrf',
           'nag_dpttrs', 'nag_dsgesv', 'nag_dspcon',
           'nag_dsprfs', 'nag_dspsv', 'nag_dspsvx', 'nag_dsptrf',
           'nag_dsptri', 'nag_dsptrs', 'nag_dsycon', 'nag_dsyrfs',
           'nag_dsysv', 'nag_dsysvx', 'nag_dsytrf', 'nag_dsytri',
           'nag_dsytrs', 'nag_dtbcon', 'nag_dtbrfs', 'nag_dtbtrs',
           'nag_dtpcon', 'nag_dtprfs', 'nag_dtptri', 'nag_dtptrs',
           'nag_dtrcon', 'nag_dtrrfs', 'nag_dtrtri', 'nag_dtrtrs',
           'nag_zcgesv', 'nag_zgbcon', 'nag_zgbequ',
           'nag_zgbrfs', 'nag_zgbsv', 'nag_zgbsvx', 'nag_zgbtrf',
           'nag_zgbtrs', 'nag_zgecon', 'nag_zgeequ', 'nag_zgerfs',
           'nag_zgesv', 'nag_zgesvx', 'nag_zgetrf', 'nag_zgetri',
           'nag_zgetrs', 'nag_zgtcon', 'nag_zgtrfs', 'nag_zgtsv',
           'nag_zgtsvx', 'nag_zgttrf', 'nag_zgttrs', 'nag_zhecon',
           'nag_zherfs', 'nag_zhesv', 'nag_zhesvx', 'nag_zhetrf',
           'nag_zhetri', 'nag_zhetrs', 'nag_zhpcon', 'nag_zhprfs',
           'nag_zhpsv', 'nag_zhpsvx', 'nag_zhptrf', 'nag_zhptri',
           'nag_zhptrs', 'nag_zpbcon', 'nag_zpbequ', 'nag_zpbrfs',
           'nag_zpbsv', 'nag_zpbsvx', 'nag_zpbtrf', 'nag_zpbtrs',
           'nag_zpocon',
           'nag_zpoequ', 'nag_zporfs', 'nag_zposv', 'nag_zposvx',
           'nag_zpotrf', 'nag_zpotri', 'nag_zpotrs', 'nag_zppcon',
           'nag_zppequ', 'nag_zpprfs', 'nag_zppsv', 'nag_zppsvx',
           'nag_zpptrf', 'nag_zpptri', 'nag_zpptrs', 'nag_zptcon',
           'nag_zptrfs', 'nag_zptsv', 'nag_zptsvx', 'nag_zpttrf',
           'nag_zpttrs', 'nag_zspcon', 'nag_zsprfs', 'nag_zspsv',
           'nag_zspsvx', 'nag_zsptrf', 'nag_zsptri', 'nag_zsptrs',
           'nag_zsycon', 'nag_zsyrfs', 'nag_zsysv', 'nag_zsysvx',
           'nag_zsytrf', 'nag_zsytri', 'nag_zsytrs', 'nag_ztbcon',
           'nag_ztbrfs', 'nag_ztbtrs', 'nag_ztpcon', 'nag_ztprfs',
           'nag_ztptri', 'nag_ztptrs', 'nag_ztrcon', 'nag_ztrrfs',
           'nag_ztrtri', 'nag_ztrtrs']
