"""
Deprecated support objects for use with the deprecated Python wrappers for the
NAG C Library (Mark 26.0).

Raises RuntimeError if the integer width used in the associated
NAG C Library could not be determined.

NAG Copyright 2017.
"""
from ctypes import c_char
from ctypes import c_char_p
from ctypes import c_double
from ctypes import c_float
from ctypes import c_int
from ctypes import c_long
from ctypes import c_ulong
from ctypes import c_void_p
from ctypes import POINTER
from ctypes import Structure
from numpy import int32, int64

def _reset_units(func):
    """
    Decorate with resetting to Fortran-style advisory and error units.
    """
    import functools
    from ..base.utils import FileObjManager
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        FileObjManager._set_default_units()
        result = func(*args, **kwargs)
        FileObjManager._set_ni_units()
        return result
    return wrapper

def wrapNagErrorMessageLines(msg_lines):
    """
    Return a NAG formatting of the error message msg_lines.
    """
    from textwrap import fill
    indent = '** '
    return '\n'.join([fill(msg,
                           initial_indent=indent,
                           subsequent_indent=indent,
                           width=80)
                      for msg in msg_lines])

def nagBadArgInstanceMessage(rout_name, arg_name):
    """
    Return a NAG-formatted error message for argument arg_name in
    NAG Python function rout_name not being an instance of numpy.ndarray.
    """
    return nagFunctionErrorMessage(
        rout_name,
        ['Argument ' + arg_name + ' must be an instance of numpy.ndarray.']
    )

def nagBadArgTypeMessage(rout_name, arg_name, exp_dtype):
    """
    Return a NAG-formatted error message for argument arg_name in
    NAG Python function rout_name not being of type exp_dtype.
    """
    from numpy import dtype
    return nagFunctionErrorMessage(
        rout_name,
        ['Argument ' + arg_name + ' must be of type ' + str(dtype(exp_dtype)) +
         '.']
        )

def nagBadHandleInstanceMessage(rout_name, arg_name):
    """
    Return a NAG-formatted error message for argument arg_name in
    NAG Python function rout_name not being a nag_opt_handle_object.
    """
    return nagFunctionErrorMessage(
        rout_name,
        ['Argument ' + arg_name + ' must be a nag_opt_handle_object.']
    )

def nagBadPkgIntNotFoundMessage(naglib_maxint):
    """
    Return a NAG-formatted error message for not being able to map
    naglib_maxint to a numpy integer.
    """
    return nagPackageErrorMessage([
        'Could not determine int type for x02bbc result ' + str(naglib_maxint) +
        '.',
    ])

def nagBadPkgSystemMessage(system, bits):
    """
    Return a NAG-formatted error message for being on an unsupported
    (bits-bit) system.
    """
    return wrapNagErrorMessageLines([
        nagPackageErrorMessageHeader(),
        'Unexpected system ' + system + ' (' + str(bits) + '-bit).',
        nagPackageErrorMessageFooter()
    ])

def nagFunctionErrorMessageHeader(rout_name):
    """
    Return a header string for an error occurring in
    NAG Python function rout_name.
    """
    return 'Error detected by NAG Python function ' + rout_name + ':'

def nagFunctionErrorMessage(rout_name, msg_lines):
    """
    Return a NAG-formatted error message for message msg_lines occurring in
    NAG Python function rout_name.
    """
    these_msg_lines = [nagFunctionErrorMessageHeader(rout_name)]
    these_msg_lines.extend(msg_lines)
    return wrapNagErrorMessageLines(these_msg_lines)

def nagImportErrorMessage(rout_name, system):
    """
    Return a NAG-formatted error message for function rout_name not being
    available on system system.
    """
    return wrapNagErrorMessageLines([
        'Error importing NAG Python function ' + rout_name + ':',
        'Function not supported on ' + system + '.',
        nagPackageErrorMessageFooter()
    ])

def nagLibraryHandleErrorMessage():
    """
    Return a NAG-formatted error message for failure to retrieve a handle to
    the NAG C Library.
    """
    these_msg_lines = [nagPackageErrorMessageHeader()]
    these_msg_lines.append(
        'Could not retrieve a handle for the NAG C Library.'
    )
    return wrapNagErrorMessageLines(these_msg_lines)

def nagPackageErrorMessageFooter():
    """
    Return a footer string for an error occurring in the set up of the
    NAG Python package.
    """
    return 'Please contact NAG.'

def nagPackageErrorMessageHeader():
    """
    Return a header string for an error occurring in the set up of the
    NAG Python package.
    """
    return 'Error initializing NAG Python wrappers.'

def nagPackageErrorMessage(msg_lines):
    """
    Return a NAG-formatted error message for message msg_lines occurring in
    the set up of the NAG Python package.
    """
    these_msg_lines = [nagPackageErrorMessageHeader()]
    these_msg_lines.extend(msg_lines)
    these_msg_lines.append(nagPackageErrorMessageFooter())
    return wrapNagErrorMessageLines(these_msg_lines)

def get_library():
    """
    Return a handle to the NAG Library.
    """
    from ..base.utils import _EngineState
    return _EngineState._engine_handle

def get_functype():
    """
    Return the correct function type for the calling convention in use:
    that is, WINFUNCTYPE on Win32 and CFUNCTYPE otherwise.
    """
    from platform import system
    from struct import calcsize
    bits = 8 * calcsize("P")
    if system() == 'Windows' and bits == 32:
        from ctypes import WINFUNCTYPE
        return WINFUNCTYPE
    else:
        from ctypes import CFUNCTYPE
        return CFUNCTYPE
NAG_FUNCTYPE = get_functype()

_libraries = {}
_library_cl = get_library()

def get_int_type():
    """
    Return the integer type based on the value of x02bbc in the
    NAG C Library.
    """
    from numpy import iinfo
    x02bbc_ctypes = _library_cl.x02bbc
    x02bbc_ctypes.argtypes = []
    x02bbc_ctypes.restype = c_long
    naglib_maxint = x02bbc_ctypes()
    for int_try in [int32, int64]:
        if naglib_maxint == iinfo(int_try).max:
            return int_try
    raise RuntimeError(nagBadPkgIntNotFoundMessage(naglib_maxint))

nag_int_type = get_int_type()

NE_NOERROR = 0
NE_CODE_NOT_SET = 1
NE_ERR_MSG_STRING = 2
NE_TOO_MANY = 3
NE_REAL_ARG_EQ = 4
NE_REAL_ARG_LT = 5
NE_REAL_ARG_LE = 6
NE_REAL_ARG_GT = 7
NE_REAL_ARG_GE = 8
NE_REAL_ARG_CONS = 9
NE_INT_ARG_EQ = 10
NE_INT_ARG_LT = 11
NE_INT_ARG_LE = 12
NE_INT_ARG_GT = 13
NE_INT_ARG_GE = 14
NE_INT_ARG_CONS = 15
NE_2_INT_ARG_EQ = 16
NE_2_INT_ARG_LT = 17
NE_2_INT_ARG_LE = 18
NE_2_INT_ARG_GT = 19
NE_2_INT_ARG_GE = 20
NE_2_INT_ARG_ZERO = 21
NE_2_REAL_ARG_EQ = 22
NE_2_REAL_ARG_LT = 23
NE_2_REAL_ARG_LE = 24
NE_2_REAL_ARG_GT = 25
NE_2_REAL_ARG_GE = 26
NE_2_REAL_ARG_TOO_CLOSE = 27
NE_2_INT_ARG_CONS = 28
NE_3_INT_ARG_CONS = 29
NE_4_INT_ARG_CONS = 30
NE_2_REAL_ARG_CONS = 31
NE_3_REAL_ARG_CONS = 32
NE_4_REAL_ARG_CONS = 33
NE_2_REAL_ENUM_ARG_CONS = 34
NE_3_REAL_ENUM_ARG_CONS = 35
NE_REAL_INT_ARG_CONS = 36
NE_REAL_ENUM_ARG_CONS = 37
NE_ENUM_ENUM_ARG_CONS = 38
NE_INVALID_INT_RANGE_1 = 39
NE_INVALID_INT_RANGE_2 = 40
NE_INVALID_REAL_RANGE_E = 41
NE_INVALID_REAL_RANGE_F = 42
NE_INVALID_REAL_RANGE_EE = 43
NE_INVALID_REAL_RANGE_EF = 44
NE_INVALID_REAL_RANGE_FF = 45
NE_INVALID_REAL_RANGE_FE = 46
NE_INVALID_REAL_RANGE_CONS = 47
NE_INT_RANGE_CONS = 48
NE_INT_RANGE_CONS_2 = 49
NE_REAL_RANGE_CONS = 50
NE_2_INT_REAL_ARG_CONS = 51
NE_2_REAL_INT_ARG_CONS = 52
NE_INT_ARRAY_1 = 53
NE_INT_ARRAY_2 = 54
NE_REAL_ARRAY_INPUT = 55
NE_INT_ARRAY_VAL_1_OR_2 = 56
NE_REAL_ARRAY_CONS = 57
NE_ARRAY_CONS = 58
NE_ARRAY_INPUT = 59
NE_2_REAL_ARRAYS_CONS = 60
NE_REAL_ARRAY_ELEM_CONS = 61
NE_NOT_INCREASING = 62
NE_NOT_STRICTLY_INCREASING = 63
NE_NOT_DECREASING = 64
NE_NOT_STRICTLY_DECREASING = 65
NE_NOT_NON_DECREASING = 66
NE_NO_NORMALISATION = 67
NE_USER_STOP = 68
NE_BOOL_NOT_SET = 69
NE_BAD_PARAM = 70
NE_F06_BAD_PARAM = 71
NE_ARG_INVALID = 72
NE_ALLOC_FAIL = 73
NE_INTERNAL_ERROR = 74
NE_MESG_NOT_INIT = 75
NE_NOT_READ_FILE = 76
NE_NOT_WRITE_FILE = 77
NE_NOT_APPEND_FILE = 78
NE_NOT_CLOSE_FILE = 79
NE_WRITE_ERROR = 80
NE_STRUCT_CORRUPT = 81
NE_NULL_ARRAY = 82
NE_NON_MONOTONIC = 83
NE_INT_OPT_ARG_LT = 84
NE_INT_OPT_ARG_GT = 85
NE_2_INT_OPT_ARG_CONS = 86
NE_2_REAL_EE_OPT_ARG_CONS = 87
NE_INCOMPAT_PARAM = 88
NE_TEST_FAILED = 89
NE_INT = 90
NE_INT_2 = 91
NE_INT_3 = 92
NE_INT_4 = 93
NE_INT_5 = 94
NE_INT_6 = 95
NE_ENUM_INT = 96
NE_ENUM_INT_2 = 97
NE_ENUM_INT_3 = 98
NE_ENUM_INT_4 = 99
NE_REAL = 100
NE_REAL_2 = 101
NE_REAL_3 = 102
NE_REAL_4 = 103
NE_REAL_5 = 104
NE_REAL_6 = 105
NE_ENUM_REAL_1 = 106
NE_ENUM_REAL_2 = 107
NE_ENUM_REAL_3 = 108
NE_ENUM_REAL_4 = 109
NE_CONSTRAINT = 110
NE_COMPLEX = 111
NE_INT_ARRAY_CONS = 112
NE_2_INT_ARRAY_CONS = 113
NE_2D_INT_ARRAY_CONS = 114
NE_2D_REAL_ARRAY_CONS = 115
NE_2D_1D_INT_ARRAYS_CONS = 116
NE_INT_ARG_ENUM_CONS = 117
NE_2_INT_ARG_ENUM_CONS = 118
NE_2_INTARR_INT = 119
NE_INTARR = 120
NE_INTARR_INT = 121
NE_INTARR_2_INT = 122
NE_INTARR_3_INT = 123
NE_REALARR = 124
NE_INCOMP_ARRAYS = 125
NE_REAL_MACH_PREC = 126
NE_REAL_INT = 127
NE_REALARR_REAL = 128
NW_INT = 129
NW_2_INT = 130
NW_REAL_REALARR = 131
NE_INTARR_REALARR = 132
NE_COMPLEX_ZERO = 133
NE_POLY_NOT_CONV = 134
NE_POLY_UNFLOW = 135
NE_POLY_OVFLOW = 136
NE_C02_OVERFLOW = 137
NE_C02_NOT_CONV = 138
NE_FUNC_END_VAL = 139
NE_PROBABLE_POLE = 140
NE_XTOL_TOO_SMALL = 141
NE_TOO_MANY_FUNC_EVAL = 142
NE_NO_IMPROVEMENT = 143
NE_FACTOR_GT = 144
NE_TOO_MANY_FACTORS = 145
NE_C06_FACTOR_GT = 146
NE_C06_TOO_MANY_FACTORS = 147
NE_C06_NOT_TRIG = 148
NE_C06_FACTORS = 149
NE_QUAD_MAX_SUBDIV = 150
NE_QUAD_BAD_SUBDIV = 151
NE_QUAD_BAD_SUBDIV_INTS = 152
NE_QUAD_ROUNDOFF_TOL = 153
NE_QUAD_ROUNDOFF_ABS_TOL = 154
NE_QUAD_ROUNDOFF_EXTRAPL = 155
NE_QUAD_NO_CONV = 156
NE_QUAD_BRKPTS_INVAL = 157
NE_QUAD_BAD_SUBDIV_INT = 158
NE_QUAD_MAX_INT = 159
NE_QUAD_EXTRAPL_INT = 160
NE_QUAD_MAX_SUBDIV_SPEC_INT = 161
NE_QUAD_ROUNDOFF_TOL_SPEC_INT = 162
NE_QUAD_NO_CONV_SPEC_INT = 163
NE_QUAD_DIVERGENCE_SPEC_INT = 164
NE_QUAD_BAD_SPEC_INT = 165
NE_QUAD_GAUSS_NPTS_RULE = 166
NE_QUAD_GAUSS_CONS = 167
NE_QUAD_GAUSS_LAG_HER_WTS = 168
NE_QUAD_MAX_INTEGRAND_EVAL = 169
NE_QUAD_MAX_INTEGRAND_CONS = 170
NE_QUAD_FIRST_TWO_PTS_EQL = 171
NE_XSOL_NOT_RESET = 172
NE_XSOL_SET_WRONG = 173
NE_XSOL_INCONSIST = 174
NE_TOL_TOO_SMALL = 175
NE_TOL_PROGRESS = 176
NE_NO_SIGN_CHANGE = 177
NE_STATE = 178
NE_NEQF_CHANGED = 179
NE_NEQG_CHANGED = 180
NE_VECTOL_CHANGED = 181
NE_REAL_LT_COND = 182
NE_NO_INTEGRATE = 183
NE_NO_STEPS = 184
NW_EXTRAPOLATION = 185
NE_NWANT_GT = 186
NE_NO_SETUP = 187
NE_SETUP_ERROR = 188
NE_T_SAME_TOUT = 189
NE_T_CHANGED = 190
NE_DIRECTION_CHANGE = 191
NE_TOUT_TCRIT = 192
NE_MAX_STEP = 193
NE_ODE_TOL = 194
NE_WEIGHT_ZERO = 195
NE_STIFF_PROBLEM = 196
NE_SINGULAR_POINT = 197
NE_NO_G_FUN = 198
NE_NEQF = 199
NE_NEQ = 200
NE_LF_B_VAL = 201
NE_RT_B_VAL = 202
NE_LFRT_B_VAL = 203
NE_LF_B_MESH = 204
NE_RT_B_MESH = 205
NE_CONV_MESH = 206
NE_CONV_MESH_INIT = 207
NE_CONV_ROUNDOFF = 208
NE_CONV_CONT = 209
NE_CONV_JACOBG = 210
NE_CONV_CONT_DEP = 211
NE_CONV_CONT_DELEPS = 212
NE_BOUND_COND_ROW = 213
NE_BOUND_COND_COL = 214
NE_BOUND_COND_MAT = 215
NE_BOUND_COND_LC = 216
NE_BOUND_COND_NLC = 217
NE_INVALID_FUN_JAC = 218
NE_INVALID_FUN_JAC_CONT = 219
NE_INVALID_FUN_JAC_NO_CONT = 220
NE_INVALID_FUN_NO_JAC_CONT = 221
NE_RK_INVALID_CALL = 222
NE_PREV_CALL = 223
NE_PREV_CALL_INI = 224
NE_RK_TGOT_EQ_TEND = 225
NE_RK_TGOT_RANGE_TEND = 226
NE_RK_TGOT_RANGE_TEND_CLOSE = 227
NE_RK_TWANT_CLOSE_TGOT = 228
NE_RK_PCC_METHOD = 229
NE_RK_PDC_TEND = 230
NE_RK_PDC_STEP = 231
NE_RK_PDC_POINTS = 232
NE_RK_PDC_GLOBAL_ERROR_T = 233
NE_RK_PDC_GLOBAL_ERROR_S = 234
NE_MISSING_CALL = 235
NE_RK_DIRECTION_POS = 236
NE_RK_DIRECTION_NEG = 237
NE_RK_STEP = 238
NE_ERRASS_REQ = 239
NE_RK_STIFF = 240
NE_RK_PX_METHOD = 241
NW_RK_TOO_MANY = 242
NE_RK_NOSTEP = 243
NE_MEMORY_FREED = 244
NE_NOT_MONOTONIC = 245
NW_EXTRAPOLATE = 246
NE_ABSCI_OUTSIDE_KNOT_INTVL = 247
NE_KNOTS_OUTSIDE_DATA_INTVL = 248
NE_WEIGHTS_NOT_POSITIVE = 249
NE_SW_COND_FAIL = 250
NW_INTERVAL_EXTRAPOLATE = 251
NE_DATA_ILL_CONDITIONED = 252
NE_KNOTS_DISTINCT_ABSCI_CONS = 253
NE_ALL_DATA_COLLINEAR = 254
NE_DATA_NOT_UNIQUE = 255
NE_NUM_PARAM_INVALID = 256
NE_RAD_PARAM_INVALID = 257
NW_SPARSE_DATA_FIT = 258
NE_TRIANG_INVALID = 259
NW_VALUE_EXTRAPOLATED = 260
NE_BAD_INTERPOLANT = 261
NE_KNOTS_CONS = 262
NE_SPLINE_COEFF_CONV = 263
NE_END_KNOTS_CONS = 264
NE_KNOTS_COORD_CONS = 265
NE_POINT_OUTSIDE_RECT = 266
NE_KNOTS_COINCIDE = 267
NE_PANEL_ORDER = 268
NE_ALL_ELEMENTS_EQUAL = 269
NE_NUM_KNOTS_2D_GT_RECT = 270
NE_NUM_KNOTS_2D_GT_SCAT = 271
NE_NUM_COEFF_GT = 272
NE_NUM_KNOTS_1D_GT = 273
NE_NO_ADDITIONAL_KNOTS = 274
NE_ENUMTYPE_WARM = 275
NE_NUM_KNOTS_2D_LT = 276
NE_SF_D_K_CONS = 277
NE_SPLINE_RANGE_INVALID = 278
NE_NON_ZERO_WEIGHTS = 279
NE_INVALID_XCAP = 280
NE_OPT_NOT_INIT = 281
NE_INVALID_OPTION = 282
NE_FIELD_UNKNOWN = 283
NE_INVALID_OPTION_NAME = 284
NE_INVALID_VALUE = 285
NE_NO_VALUE = 286
NE_UNBALANCED_COMMENT = 287
NW_SMALL_DERIV = 288
NE_DERIV_ERRORS = 289
NE_GRAD_TOO_SMALL = 290
NE_SVD_FAIL = 291
NW_NO_IMPROVEMENT = 292
NW_COND_MIN = 293
NW_TOO_MANY_ITER = 294
NW_STEP_BOUND_TOO_SMALL = 295
NE_MISSING_DERIV = 296
NE_INVALID_TEXT_RANGE = 297
NE_INVALID_ENUM_RANGE = 298
NE_NOT_FUN_NAME = 299
NE_CHOLESKY_OVERFLOW = 300
NW_LOCAL_SEARCH = 301
NE_BOUND = 302
NE_BOUND_LCON = 303
NE_BOUND_NLCON = 304
NE_BOUND_EQ = 305
NE_BOUND_EQ_LCON = 306
NE_BOUND_EQ_NLCON = 307
NE_FD_INT = 308
NE_HESD = 309
NE_CANCEL_ERR = 310
NE_INIT_MEM = 311
NE_NO_MEM = 312
NE_SINGULAR_VALUES = 313
NW_LIN_DEPEND = 314
NE_SVD_ZERO = 315
NE_C_OVERFLOW = 316
NW_DEAD_POINT = 317
NW_SOLN_NOT_UNIQUE = 318
NE_UNBOUNDED = 319
NW_NOT_FEASIBLE = 320
NW_LIN_NOT_FEASIBLE = 321
NW_NONLIN_NOT_FEASIBLE = 322
NE_HESS_TOO_BIG = 323
NW_OVERFLOW_WARN = 324
NE_STATE_VAL = 325
NE_STR_UNKNOWN = 326
NE_STOP_LT_START = 327
NE_CHECK_LT_ONE = 328
NE_WARM_START = 329
NE_CVEC_NULL = 330
NE_H_NULL = 331
NE_H_NULL_QP = 332
NW_NOT_CONVERGED = 333
NW_KT_CONDITIONS = 334
NE_INVALID_BEGIN = 335
NW_LAGRANGE_MULT_ZERO = 336
NE_CYCLING = 337
NE_OBJ_BOUND = 338
NE_DUPLICATE_ELEMENT = 339
NE_HESS_INDEF = 340
NE_BASIS_ILL_COND = 341
NE_OUT_OF_WORKSPACE = 342
NE_BASIS_SINGULAR = 343
NE_NULL_QPHX = 344
NE_NAME_TOO_LONG = 345
NE_STATE_VAL_ARG = 346
NE_NAMES_NOT_NAG_MEM = 347
NE_NULL_ARGUMENT = 348
NE_B_NULL = 349
NW_MAX_FUN = 350
NE_H_FORWARD_NULL = 351
NW_DERIV_INFO = 352
NE_INFEASIBLE = 353
NE_LIN_NOT_FEASIBLE = 354
NE_NONLIN_NOT_FEASIBLE = 355
NE_MAYBE_UNBOUNDED = 356
NE_TOO_MANY_ITER = 357
NE_TOO_MANY_MAJOR_ITER = 358
NE_TOO_MANY_MINOR_ITER = 359
NE_SUPERBASICS_LIMIT = 360
NE_OBJ_DERIV_ERRORS = 361
NE_CON_DERIV_ERRORS = 362
NE_NO_IMPROVE = 363
NE_CANNOT_CALCULATE = 364
NE_NOT_REQUIRED_ACC = 365
NE_SINGULAR = 366
NE_NOT_POS_DEF = 367
NE_NOT_POS_DEF_FACT = 368
NE_ZERO_DIAG = 369
NE_NOT_UNIT_DIAG = 370
NE_TOO_MANY_ITERATIONS = 371
NE_TDQ_LT_M = 372
NE_TDQ_LT_N = 373
NE_TDP_LT_N = 374
NE_QR_NOT_CONV = 375
NE_RANK_LT_N = 376
NE_RANK_LT_NOT_UNIQ = 377
NE_ITER_FAIL_CONV = 378
NE_DIAG_IMAG_NON_ZERO = 379
NE_SINGULAR_DET_ZERO = 380
NE_DET_OVERFLOW = 381
NE_DET_UNDERFLOW = 382
NE_ITERATIONS_QZ = 383
NE_MATRIX_NOT_ZERO = 384
NE_QR_FAIL_CONV = 385
NE_MAT_SINGULAR = 386
NE_QR_FAIL = 387
NE_REQD_EIGVAL = 388
NE_EIGVEC = 389
NE_FORMAT_LEN = 390
NE_INVALID_FORMAT = 391
NE_COL_WIDTH = 392
NE_EXCHANGE = 393
NE_REORDER = 394
NE_PERTURBED = 395
NE_SYMM_MATRIX = 396
NE_ZERO_DIAGONAL_ELEM = 397
NE_PRECOND_NOT_POS_DEF = 398
NE_COEFF_NOT_POS_DEF = 399
NE_NOT_REQ_ACC = 400
NE_ACC_LIMIT = 401
NE_SYMM_MATRIX_DUP = 402
NE_INVALID_ROW_PIVOT = 403
NE_LA_TOO_SMALL = 404
NE_INVALID_SCS = 405
NE_INVALID_SCS_PRECOND = 406
NE_NON_ZERO_DUP = 407
NE_ZERO_COEFF = 408
NE_INVALID_CS = 409
NE_INVALID_CS_PRECOND = 410
NE_INVALID_ROWCOL_PIVOT = 411
NE_NONSYMM_MATRIX = 412
NE_NONSYMM_MATRIX_DUP = 413
NE_NEG_WEIGHT = 414
NE_CASES_ZERO = 415
NE_CASES_ONE = 416
NE_FUN_INF = 417
NE_GAM_NOT_CONV = 418
NE_RES_NOT_ACC = 419
NE_ALG_NOT_CONV = 420
NE_SOL_CLOSE_TO_ZERO = 421
NE_SOL_NOT_CONV = 422
NE_ACC_IN_DOUBT = 423
NE_PROBAB_CLOSE_TO_TAIL = 424
NE_ARG_TOO_LARGE = 425
NE_VARIANCE_TOO_LARGE = 426
NE_SAMPLES_NOT_IN_ORDER = 427
NE_SAMPLES_ARE_ALL_EQUAL = 428
NE_NORMAL_SCORE_EST = 429
NE_SERIES = 430
NE_PROBABILITY = 431
NE_POISSON_WEIGHT = 432
NE_TERM_LARGE = 433
NE_CHI_PROB = 434
NE_CONV = 435
NE_PROB_F = 436
NE_PROB_F_INIT = 437
NE_PROB_LIMIT = 438
NE_PROB_B_INIT = 439
NE_BAD_SX_OR_IP = 440
NE_ZERO_DOF_RESID = 441
NE_SVD_NOT_CONV = 442
NE_MAT_NOT_UPD = 443
NE_RSS_NOT_UPD = 444
NE_DOF_LE_ZERO = 445
NE_NVAR_NOT_IND = 446
NE_ZERO_DIAG_ELEM = 447
NE_RESID_LARG = 448
NE_SVD_RANK_NE_IP = 449
NE_SVD_RANK_GT_IP = 450
NE_ILLCOND = 451
NE_IP_INCOMP_WITH_SX = 452
NE_MAT_NOT_FULL_RANK = 453
NE_STDES_ZERO = 454
NE_RANK_EQ_IP = 455
NE_DIAG_ELEM_ZERO = 456
NE_SW_LT_ONE = 457
NE_VAR_EQ_ZERO = 458
NE_NEG_SX = 459
NE_POS_SX = 460
NE_WT_LOW = 461
NE_SW_LOW = 462
NW_RSS_EQ_ZERO = 463
NW_RMS_EQ_ZERO = 464
NE_SOBS_LOW = 465
NE_SVAR_LOW = 466
NE_X_OR_Y_IDEN = 467
NE_X_IDEN = 468
NE_PREC_NOT_INIT = 469
NE_STRUCT_NOT_DEFINED = 470
NE_EX_POWER_EQ_ZERO = 471
NE_IP_INCOMP_SX = 472
NE_IP_GT_OBSERV = 473
NE_VALUE_AT_BOUNDARY_A = 474
NE_VALUE_AT_BOUNDARY_B = 475
NE_VALUE_AT_BOUNDARY_C = 476
NE_VALUE_AT_BOUNDARY_D = 477
NE_LSQ_ITER_NOT_CONV = 478
NE_RANK_CHANGED = 479
NE_ZERO_DOF_ERROR = 480
NE_BAD_NY_SET = 481
NE_BAD_NX_SET = 482
NE_VARCOV_MAT_SINGULAR = 483
NE_SQRT_VARCOV_SINGULAR = 484
NE_DIAG_ELEM_INVALID = 485
NE_DIAG_EQ_ZERO = 486
NE_ELEM_GT_ONE = 487
NE_INVALID_HUBER_FUN = 488
NE_INVALID_DCHI_FUN = 489
NE_INVALID_MALLOWS_REG_C = 490
NE_INVALID_SCHWEPPE_REG_C = 491
NE_BAD_HAMPEL_PSI_FUN = 492
NE_WT_ITER_EXCEEDED = 493
NE_BETA1_ITER_EXCEEDED = 494
NE_THETA_ITER_EXCEEDED = 495
NE_LSQ_FAIL_CONV = 496
NE_WT_LSQ_NOT_FULL_RANK = 497
NE_REG_MAT_SINGULAR = 498
NE_COV_MAT_FACTOR_ZERO = 499
NE_VAR_THETA_LEQ_ZERO = 500
NE_ERR_DOF_LEQ_ZERO = 501
NE_ESTIM_SIGMA_ZERO = 502
NE_OBS_VAR_CONST = 503
NE_C_ITER_NOT_CONV = 504
NE_C_ITER_UNSTABLE = 505
NE_CPSI_NOT_SET = 506
NE_CUCV_NOT_SET = 507
NE_HPSI_NOT_SET = 508
NE_DCHI_NOT_SET = 509
NE_SUM_ZERO = 510
NE_UCV = 511
NE_CONST_COL = 512
NE_COR_MAT_RANK = 513
NE_COR_MAT_POSDEF = 514
NE_NEG_WEIGHT_ELEMENT = 515
NE_VAR_INCL_INDICATED = 516
NE_VAR_INCL_COND = 517
NE_OBSERV_LT_VAR = 518
NE_VAR_INCL_STANDARD = 519
NE_ZERO_EIGVALS = 520
NE_CANON_CORR_1 = 521
NE_GROUPS = 522
NE_RANK_ZERO = 523
NE_MAT_RANK_ZERO = 524
NE_WT_ARGS = 525
NE_ACC_ITER = 526
NE_NORM_ZERO_PTS = 527
NE_LSQ_SCAL_ZERO_PTS = 528
NE_MAT_RANK = 529
NE_GROUP_OBSERV = 530
NE_GROUP_VAR = 531
NE_DIAG_0_COND = 532
NE_DIAG_0_J_COND = 533
NE_VAR_RANK = 534
NE_GROUP_VAR_RANK = 535
NE_GROUP_SUM = 536
NE_PRIOR_SUM = 537
NE_IDEN_ELEM_COND = 538
NE_DENDROGRAM = 539
NE_WT_ZERO = 540
NE_CLUSTER_EMPTY = 541
NE_DENDROGRAM_ARRAY = 542
NE_CLUSTER = 543
NE_NONZERO_EIGVALS = 544
NE_EIGVAL = 545
NE_NEG_OR_ZERO_ARRAY = 546
NE_NEG_ELEMENT = 547
NE_IT_ARRAY = 548
NE_ARRAY_CONSTANT = 549
NE_G04BB_STDERR = 550
NE_G04BB_CONV = 551
NE_G04BB_DESIGN = 552
NE_G04BB_TREAT = 553
NE_G04BB_RES_DF = 554
NE_TOO_SMALL = 555
NE_PLOT_TREAT = 556
NE_G04CA_RES_DF = 557
NE_G04BC_ST_ERR = 558
NE_G04BC_REPS = 559
NE_G04BC_RESD = 560
NE_G04BC_DISCON = 561
NE_G04EA_LEVELS = 562
NE_G04EA_ORTHO_POLY = 563
NE_STUDENTIZED_STAT = 564
NE_STATE_CORRUPT = 565
NE_REAL_ARG_NON_INTEGRAL = 566
NE_SYSTEM_TIME = 567
NE_NOT_POS_SEM_DEF = 568
NE_START_P_Q = 569
NE_REF_VEC = 570
NE_PROB_NOT_ONE = 571
NE_ALL_PROB_ZERO = 572
NE_NEG_PROB = 573
NE_STATIONARITY = 574
NE_WINS_RES_ZERO = 575
NE_ACC = 576
NE_POS_DEF = 577
NE_ROUND_OFF = 578
NE_G08AA_NON_TIED = 579
NE_G08AG_SAMP_IDEN = 580
NE_G08AH_SAMP_IDEN = 581
NE_G08CB_PARAM = 582
NE_G08CB_DATA = 583
NE_G08CB_SAMPLE = 584
NE_G08CB_VARIANCE = 585
NE_G08CB_INCOMP_GAMMA = 586
NE_G08CD_CONV = 587
NE_G08CG_FREQ = 588
NE_G08CG_CLASS_VAL = 589
NE_G08CG_CONV = 590
NE_G08EA_TIE = 591
NE_G08EA_RUNS_LENGTH = 592
NE_G08EA_COVAR = 593
NE_G08EA_RUNS = 594
NE_G08EB_PAIRS = 595
NE_G08EB_CELL = 596
NE_G08EC_CELL = 597
NE_G08EC_TRIPLETS = 598
NE_G08ED_GAPS_ZERO = 599
NE_G08ED_FREQ_ZERO = 600
NE_G08ED_GAPS = 601
NE_G08ED_FREQ_LT_ONE = 602
NE_MANN_WHITNEY_STAT = 603
NE_SEQUENCE_SHORT = 604
NE_G10AC_DF_RHO = 605
NE_G10AC_DF_TOL = 606
NE_G10AC_ACC = 607
NE_G10AC_CG_RHO = 608
NE_G10BA_INIT = 609
NE_G10BA_INTERVAL = 610
NE_2D_INT_ARR_ELEM = 611
NE_2D_INT_ARR_ELEMS = 612
NE_TABLE_DEGENERATE = 613
NE_LOW_EXPECTED_FREQ = 614
NE_MAXT = 615
NE_VAR_DIV = 616
NE_G11BA_CHANGED = 617
NE_CELL_EMPTY = 618
NE_INVALID_CENSOR_CODE = 619
NE_INVALID_FREQ = 620
NE_CHAR_ARG_INVALID = 621
NE_G12BA_SZ_IP = 622
NE_G12BA_SZ_ISI = 623
NE_G12BA_NDMAX = 624
NE_G12BA_MAT_SING = 625
NE_G12BA_OVERFLOW = 626
NE_G12BA_CONV = 627
NE_G12BA_DEV = 628
NE_TIME_SERIES_IDEN = 629
NE_INVALID_AUTOCO_COEF = 630
NE_CORR_NOT_POS_DEF = 631
NE_INVALID_NSER = 632
NE_NPARA_MR_MT_INCONSIST = 633
NE_DELTA_TEST_FAILED = 634
NE_SOLUTION_FAIL_CONV = 635
NE_MAT_NOT_POS_DEF = 636
NE_ARIMA_TEST_FAILED = 637
NE_ITER_FAIL_NIT = 638
NE_G13_OPTIONS_NOT_INIT = 639
NE_G13_ORDERS_NOT_INIT = 640
NE_NSER_INCONSIST = 641
NE_DIFORDER_LEN_INCONSIST = 642
NE_SPECTRAL_ESTIM_NEG = 643
NE_CONFID_LIMIT_FACT = 644
NE_BIVAR_SPECTRAL_ESTIM_ZERO = 645
NE_UNIVAR_SPECTRAL_ESTIM_ZERO = 646
NE_UNIVAR_SPECTRAL_ESTIM_NEG = 647
NE_SQUARED_FREQ_GT_ONE = 648
NE_ARIMA_INPUT = 649
NE_INPUT_NARMA = 650
NE_G13AS_AR = 651
NE_G13AS_ZERO_VAR = 652
NE_G13AS_ITER = 653
NE_G13AS_FACT = 654
NE_G13AS_DIAG = 655
NE_REQ_AVAIL = 656
NE_MPS_USER_MEM = 657
NE_MPS_ILLEGAL_DATA_LINE = 658
NE_MPS_SPLIT_COL = 659
NE_MPS_REPEAT_ROW = 660
NE_MPS_ILLEGAL_NAME = 661
NE_MPS_ILLEGAL_SETNAME = 662
NE_MPS_ILLEGAL_NUMBER = 663
NE_MPS_INVALID_INF_BOUND = 664
NE_MPS_INVALID_BND_TYPE = 665
NE_MPS_INVALID_ROW_TYPE = 666
NE_MPS_INVALID_INTORG_INTEND = 667
NE_MPS_INVALID_RANGE = 668
NE_MPS_INVALID_BND_VAL = 669
NE_MPS_NO_NEWLINE = 670
NE_MPS_NO_COLS = 671
NE_MPS_NO_OBJ = 672
NE_MPS_NO_ROWS = 673
NE_MPS_TOO_BIG = 674
NE_MPS_UNKNOWN_COLNAME = 675
NE_MPS_INVALID_INDICATOR = 676
NE_MPS_UNKNOWN_ROWNAME = 677
NE_MPS_PROB_NOT_FOUND = 678
NE_MPS_RHS_RANGE_BND_NOT_FOUND = 679
NE_MPS_OPT_NOT_INIT = 680
NE_MPS_ENDATA_NOT_FOUND = 681
NE_ALLOC_AFTER_MPS = 682
NE_PRIORITY_NULL = 683
NE_MIP_PROB_NULL = 684
NE_MIP_PROB_NOT_INIT = 685
NE_MIP_ROOT_UNBOUNDED = 686
NE_MIP_ROOT_INFEAS = 687
NE_MIP_ROOT_MAX_ITER = 688
NW_MIP_FIRST_INT_SOL = 689
NW_MIP_NO_INT_SOL = 690
NW_MIP_MAX_ITER_INT_SOL = 691
NW_MIP_MAX_ITER_NO_INT_SOL = 692
NW_MIP_MAX_NODES_INT_SOL = 693
NW_MIP_MAX_NODES_NO_INT_SOL = 694
NW_MIP_MAX_DEPTH_INT_SOL = 695
NW_MIP_MAX_DEPTH_NO_INT_SOL = 696
NE_MIP_ROOT_HESS_TOO_BIG = 697
NE_MIP_HESS_TOO_BIG_INT_SOL = 698
NE_MIP_HESS_TOO_BIG_NO_INT_SOL = 699
NE_CH_LOOP = 700
NE_BAD_RANK = 701
NE_REZ_OVERFLOW = 702
NE_IMZ_OVERFLOW = 703
NE_Z_OVERFLOW = 704
NE_IMZ_HALF_PREC = 705
NE_IMZ_NO_PREC = 706
NE_REAL_ARG_TOO_SMALL = 707
NE_REAL_ARG_NEG_INT = 708
NE_UNDERFLOW_LIKELY = 709
NE_OVERFLOW_LIKELY = 710
NW_SOME_PRECISION_LOSS = 711
NE_TOTAL_PRECISION_LOSS = 712
NE_TERMINATION_FAILURE = 713
NE_INFINITE = 714
NE_S21_CONV = 715
NE_NO_LICENCE = 716
NE_INT_ARRAY = 717
NE_VACUUM = 718
NE_RESULT_NO_PRECISION = 719
NE_RESULT_HALF_PRECISION = 720
NE_RESULT_REAL_OVERFLOW = 721
NE_RESULT_IMAGINARY_OVERFLOW = 722
NE_RESULT_OVERFLOW = 723
NE_INTERNAL_WORKSPACE = 724
NE_TOO_CLOSE_INTEGER = 725
NE_NO_ARNOLDI_FAC = 726
NE_NO_SHIFTS_APPLIED = 727
NE_MAX_ITER = 728
NE_ZERO_INIT_RESID = 729
NE_OPT_INCOMPAT = 730
NE_INTERNAL_EIGVAL_FAIL = 731
NE_SCHUR_REORDER = 732
NE_SCHUR_EIG_FAIL = 733
NE_RITZ_COUNT = 734
NE_ZERO_EIGS_FOUND = 735
NE_TRANSFORM_OVFL = 736
NE_COMP_BAND_SOL = 737
NE_COMP_BAND_FAC = 738
NE_REAL_BAND_SOL = 739
NE_REAL_BAND_FAC = 740
NE_ZERO_SHIFT = 741
NE_NO_LANCZOS_FAC = 742
NE_BOTH_ENDS_1 = 743
NE_INTERNAL_EIGVEC_FAIL = 744
NE_FREQ_CONS = 2001
NE_FREQ_SUM = 2002
NE_ACCURACY = 2003
NE_CONVERGENCE = 2004
NE_INIT_ESTIMATE = 2005
NE_PROB_BOUNDARY = 2006
NE_REAL_ARRAY_ENUM = 2007
NE_REAL_ARRAY = 2008
NE_SOME_MOMENTS = 2009
NE_MOMENTS = 2010
NE_POS_SEMI_DEF = 2011
NE_EIGENVALUES = 2012
NE_SUM_WEIGHT = 2013
NE_ZERO_VARIANCE = 2014
NE_INDEP_VARS_OBS = 2015
NE_FREE_VARS = 2016
NE_FULL_RANK = 2017
NE_INT_ARRAY_ELEM_CONS = 2018
NE_MODEL_PARAMETERS = 2019
NE_ZERO_VARS = 2020
NE_ZERO_DF = 2022
NE_DENOM_ZERO = 2023
NE_RESID_IDEN = 2024
NE_RESID_MEAN = 2025
NE_FUN_RET_VAL = 2026
NE_ZERO_DIAGONAL = 2027
NE_CONVERGENCE_THETA = 2028
NE_CONVERGENCE_SOL = 2029
NE_ZERO_VALUE = 2030
NE_CHI = 2031
NE_CORRECTION_FACTOR = 2032
NE_ZERO_SUM = 2033
NE_ZERO_DIAGNOL = 2034
NE_CHARACTER = 2035
NE_ENUM_CHARACTER = 2036
NE_DIM_INFEASIBLE = 2037
NE_REAL_ARRAY_SUM = 2038
NE_STATIONARY_AR = 2039
NE_CLOSE_TO_STATIONARITY = 2040
NE_INVERTIBILITY = 2041
NE_OUTSIDE_STATIONARITY = 2042
NE_ORTHOGONAL_MATRIX = 2043
NE_DIAG_ELEMENTS = 2044
NE_EIGVAL_SUM = 2045
NE_NEGATIVE_EIGVAL = 2046
NE_REAL_ARRAYS_SUM = 2047
NE_TOO_MANY_CALLS = 2048
NE_INITIALIZATION = 2049
NE_STANDARD_ERRORS = 2050
NE_DIVERGENCE = 2051
NE_EM_PROCESS = 2052
NE_OBSERVATIONS = 2053
NE_OVERFLOW = 2054
NE_ZERO_RESID = 2055
NE_SIGMA_NEGATIVE = 2056
NE_SAMPLE_IDEN = 2057
NE_SAMPLE = 2058
NE_MAT_ILL_DEFINED = 2059
NE_RESPONSE_LEVEL = 2060
NE_MAT_INV = 2061
NE_CONT_FRAC = 2062
NE_INT_ARG = 2063
NE_NOT_COMPLEX = 2064
NE_INIT_FILTER = 2065
NE_ORDERS_ARIMA = 2066
NE_ORDERS_FILTER = 2067
NE_ARRAY_SIZE = 2068
NE_TIME_SERIES = 2069
NE_MODEL_PARAMS = 2070
MATRIX_ILL_CONDITIONED = 2071
NE_PRIME_FACTOR = 2072
NE_INT_REAL = 2073
NE_CONFID_LIMITS = 2074
NE_SPECTRAL_ESTIMATES = 2075
NE_TRANSFORMATION = 2076
NE_MODEL_INFEASIBLE = 2077
NE_OPTIMIZE = 2078
NE_ITER_FAIL = 2080
NE_NOT_ACC = 2081
NE_POLE_PRESENT = 2082
NE_DATA_COPLANAR = 2083
NE_DUPLICATE_NODE = 2084
NE_BAD_POINT = 2085
NE_NON_UNIQUE = 2086
NE_ILL_CONDITIONED = 2087
NE_DEGENERATE = 2088
NE_EXTRAPOLATION = 2089
NE_ERROR_ESTIMATE = 2090
NE_FAILED_START = 2091
NE_SING_JAC = 2092
NE_FAILED_DERIV = 2093
NE_TIME_DERIV_DEP = 2094
NE_FAILED_STEP = 2095
NE_ZERO_WTS = 2096
NE_REMESH_CHANGED = 2097
NE_BAD_MONIT = 2098
NE_MESH_ERROR = 2099
NE_FAIL_SPARSITY = 2100
NE_FAIL_TOL = 2101
NE_ALG_FAIL = 2102
NE_INVALID_PERM_COL = 2103
NE_INVALID_PERM_ROW = 2104
NE_NZLMX_TOO_SMALL = 2105
NE_NZLUMX_TOO_SMALL = 2106
NE_NZUMX_TOO_SMALL = 2107
NE_SINGULAR_MATRIX = 2108
NE_SPARSE_COL = 2109
NE_SPARSE_ROW = 2110
NE_RCOND = 2111
NE_E04NPC_NOT_INIT = 2112
NE_ALLOC_INSUFFICIENT = 2113
NE_BASIS_FAILURE = 2114
NE_OPTIONS_FILE_READ_FAILURE = 2115
NE_E04_OPTION_INVALID = 2116
NE_E04VGC_NOT_INIT = 2117
NE_NUM_DIFFICULTIES = 2118
NE_USRFUN_UNDEFINED = 2119
NE_ARRAY_TOO_SMALL = 2120
NE_JACOBIAN_STRUCTURE_FAIL = 2121
NE_E04WCC_NOT_INIT = 2122
NE_G13D_DIAG = 2123
NE_G13D_FACT = 2124
NE_G13D_ITER = 2125
NE_G13D_RES = 2126
NE_G13D_ZERO_VAR = 2127
NE_G13D_AR = 2128
NE_G13D_MA = 2129
NE_G13D_ARMA = 2130
NE_HESS_NOT_POS_DEF = 2131
NE_G13D_DERIV = 2132
NE_G13D_BOUND = 2133
NE_G13D_MAX_LOGLIK = 2134
NE_G13D_MAXCAL = 2135
NE_MAX_LOGLIK = 2136
NE_G13D_GRAD = 2137
NE_G13D_START = 2138
NE_NEARLY_POS_DEF = 2139
NE_OBSERV_LT_P = 2140
NE_INVALID_ARRAY = 2141
NE_READ_ERROR = 2139
NE_NOT_CLOSE_FILEID = 2140
NW_TOO_MUCH_ACC_REQUESTED = 2143
NE_ZERO_NOT_FOUND = 2144
NE_NOT_SIGN_CHANGE = 2145
NW_TOO_MANY_FEVALS = 2146
NE_E05JAC_NOT_INIT = 2147
NE_E05_OPTION_INVALID = 2148
NE_WRONG_NUM_TOKENS = 2149
NE_NOT_PARSED = 2150
NE_TYPE_DISCREP = 2151
NE_OUT_OF_RANGE = 2152
NE_NOT_INIT = 2153
NE_OPT_NOT_READ = 2154
NE_NOT_ONOFF_OPTION = 2155
NE_NOT_INT_OPTION = 2156
NE_NOT_REAL_OPTION = 2157
NE_MONIT_TERMIN = 2158
NE_OBJFUN_TERMIN = 2159
NE_TOO_MANY_FEVALS = 2160
NE_DIV_COMPLETE = 2161
NE_INF_INIT_LIST = 2162
NE_N_CHANGED = 2163
NE_LINESEARCH_ERROR = 2164
NE_FILE_NOT_READ = 2165
NW_INTEGRAL_INFINITE = 2166
NE_NOT_CLOSE_OR_DEL_FILEID = 2167
NE_CONTIN_PROB_NOT_SOLVED = 2168
NE_CONTIN_AWAY_NOT_POSS = 2169
NE_FINAL_PROB_NOT_SOLVED = 2170
NE_SIGNIF_DERIVS_NOT_COMPUT = 2171
NW_REAL = 2172
NE_SCHUR_REORDER_SELECT = 2173
NE_SINGULAR_WP = 2174
NE_INVALID_STATE = 2175
NE_INVALID_SEQUENCE = 2176
NE_INVALID_PRED = 2177
NW_LIMIT_REACHED = 2178
NE_ITERATION_QZ = 2179
NE_SCHUR = 2180
NE_INLIST_CLOSE = 2181
NE_MAT_ROW_ZERO = 2182
NE_MAT_COL_ZERO = 2183
NE_EIGENVECTORS = 2184
NE_STACK_OVERFLOW = 2185
NE_Q_NOT_ASCENDING = 2186
NE_Q_OUT_OF_RANGE = 2187
NE_IP_G_OBSERV = 2188
NE_CHAR_LEN_INVALID = 2189
NE_ARRAY_DIM_LEN = 2190
NE_EIGENPROBLEM = 2191
NE_LANCZOS_ITERATION = 2192
NE_UNDERFLOW = 2193
NW_HI = 2194
NW_NEG = 2195
NW_PARAM_DIST_ASYM = 2196
NW_PARAM_DIST_OBS = 2197
NW_PARAM_DIST = 2198
NE_RES_FLAG = 2199
NE_TR_STEP_FAILED = 2200
NE_RESCUE_FAILED = 2201
NE_SECANT_ITER_FAILED = 2202
NE_DATA_HYPERSURFACE = 2203
NE_OUT_OF_SEQUENCE = 2204
NE_ILLEGAL_COMM = 2205
NE_H_TOO_SMALL = 2206
NE_SPACING_INCORRECT = 2207
NW_POTENTIAL_PROBLEM = 2208
NW_SOLUTION_NOT_GUARANTEED = 2209
NW_FAST_SOLUTION = 2210
NE_ILLEGAL_CALLBACK = 2211
NE_NO_SMP_CALLBACK_THREADSAFE = 2212
NE_INTERPOLATION = 2213
NE_ITASK_FLAG = 2214
NE_INT_CHANGED = 2215
NE_REAL_CHANGED = 2216
NW_TRUNCATED = 2217
NW_OUT_OF_WORKSPACE = 2218
NE_NO_METHOD = 2219
NE_ILLEGAL_RWORK = 2220
NE_TOO_BIG = 2221
NE_INDETERMINATE = 2222
NW_IVALID = 2223
NE_ENUM_2_INT_3 = 2224
NE_ENUM_3_INT_3 = 2225
NE_NO_SOLUTION = 2226
NW_SOME_SOLUTIONS = 2227
NE_MISSING_BEGIN = 2228
NE_MISSING_END = 2229
NE_ENUM_3_INT = 2230
NE_INVALID_SPLINE = 2231
NE_BRACKET = 2232
NE_COMPLEX_2 = 2233
NE_COMPLEX_3 = 2234
NE_INT_REAL_2 = 2235
NE_REAL_INT_2 = 2236
NE_REAL_INT_3 = 2237
NE_RK_TEND = 2301
NE_RK_STEP_TOO_SMALL = 2302
NE_RK_POINTS = 2303
NE_RK_GLOBAL_ERROR_T = 2304
NE_RK_GLOBAL_ERROR_S = 2305
NE_RK_NO_INTERP = 2306
NW_MAX_SUBINT = 2307
NE_FILEID = 2310
NE_INT_MAX = 2311
NE_MPS_BOUNDS = 2312
NE_MPS_COLUMNS = 2313
NE_MPS_FORMAT = 2314
NE_MPS_INDICATOR = 2315
NE_MPS_MANDATORY = 2316
NE_MPS_OBJNAME = 2317
NE_MPS_PRINTABLE = 2318
NE_MPS_RANGES = 2319
NE_MPS_REPEAT_COLUMN = 2320
NE_MPS_RHS = 2321
NE_MPS_ROWS = 2322
NE_MPS_ROWS_OR_CONS = 2323
NE_HANDLE = 2324
NE_PHASE = 2325
NE_ALREADY_DEFINED = 2326
NE_REF_MATCH = 2327
NE_DIM_MATCH = 2328
NE_INDICES = 2329
NE_FILE_INCOMPLETE = 2330
NE_SUBPROBLEM = 2331
NE_MAYBE_INFEASIBLE = 2332
NW_UNDERFLOW_WARN = 2341
NE_UNDEFINED_ERROR = 9999
NE_UNKNOWNERROR = 9999
NE_UNKNOWN_ERROR = 9999
STRING = c_char_p

BackwardSequence = 1015
BottomPivot = 1010
ConjugateTranspose = 1003
FixedPivot = 1013
ForwardSequence = 1014
Nag_1stDeriv_Small = 1164
Nag_2ndDeriv_Large = 1163
Nag_2ndOrderDefaultParam = 2870
Nag_2ndOrderSuppliedParam = 2872
Nag_3RSSH = 1506
Nag_4253H = 1505
Nag_4thOrderDefaultParam = 2871
Nag_4thOrderSuppliedParam = 2873
Nag_ACORN = 2154
Nag_AboutMean = 1362
Nag_AboutZero = 1363
Nag_AboveForm = 1609
Nag_AbsDiff = 2862
Nag_AbsDiffScaled = 2864
Nag_AbsVal = 2861
Nag_AbsValScaled = 2863
Nag_Absolute = 1088
Nag_AccumulateQ = 1029
Nag_AccumulateZ = 1033
Nag_Adams = 2750
Nag_AddVar = 2140
Nag_AdditiveHoltWinters = 2583
Nag_Alg = 1072
Nag_Alg_loga = 1073
Nag_Alg_loga_logb = 1075
Nag_Alg_logb = 1074
Nag_AllEigVals = 1450
Nag_AllG_AllJ = 1169
Nag_AllG_SomeJ = 1167
Nag_AllLevels = 1461
Nag_AllU = 2060
Nag_AllVT = 2101
Nag_AllValues = 1054
Nag_AlternativeDAbs = 500029
Nag_AlternativeDNeg = 500031
Nag_AlternativeDPos = 500030
Nag_AmericanCall = 1114
Nag_AmericanPut = 1116
Nag_AnalyticalJacobian = 2180
Nag_AnalyticalStructJacobian = 2183
Nag_AndrewFun = 1382
Nag_ApplyP = 1060
Nag_ApplyQ = 1062
Nag_Ascending = 1597
Nag_AutoCorr = 1546
Nag_AutoCov = 1545
Nag_AvTwoNorm = 179
Nag_AverageTies = 1341
Nag_BB_Alloc_Fail = 1568
Nag_BB_Internal_Error = 1569
Nag_BB_OK = 1567
Nag_BDF = 2751
Nag_BIC = 2600
Nag_BackErrBoth = 3044
Nag_BackErrLeft = 3042
Nag_BackErrRight = 3043
Nag_BackTransform = 1038
Nag_BackwardTransform = 1071
Nag_BalanceBoth = 2023
Nag_BalancePermute = 2021
Nag_BalanceScale = 2022
Nag_Bartlett = 1535
Nag_Base = 151
Nag_Basic = 2150
Nag_BasisFactB = 1250
Nag_BasisFactBS = 1251
Nag_BasisFile = 50001
Nag_Basis_Ill_Cond = 1220
Nag_Basis_Singular = 1221
Nag_BeginBackward = 2141
Nag_BeginForward = 2144
Nag_BeginStepwise = 2146
Nag_Beta = 1488
Nag_Binomial = 1490
Nag_Biorthogonal1_1 = 2253
Nag_Biorthogonal1_3 = 2254
Nag_Biorthogonal1_5 = 2255
Nag_Biorthogonal2_2 = 2256
Nag_Biorthogonal2_4 = 2257
Nag_Biorthogonal2_6 = 2258
Nag_Biorthogonal2_8 = 2259
Nag_Biorthogonal3_1 = 2260
Nag_Biorthogonal3_3 = 2261
Nag_Biorthogonal3_5 = 2262
Nag_Biorthogonal3_7 = 2263
Nag_BlomScores = 1337
Nag_BonferroniInterval = 1463
Nag_Both = 2812
Nag_BothSides = 143
Nag_BoundNotSet = 1181
Nag_Bounds = 1182
Nag_BoundsEqual = 1184
Nag_BoundsZero = 1183
Nag_BrDir_NotSet = 1586
Nag_BracketForm = 1610
Nag_Branch_InitX = 1589
Nag_Branch_Left = 1587
Nag_Branch_Right = 1588
Nag_Broad_Search = 1578
Nag_BrownsExponential = 2581
Nag_ByBlock = 1052
Nag_CDF = 1470
Nag_CE_EM = 1475
Nag_CE_NR = 1476
Nag_Call = 2190
Nag_Censored = 1477
Nag_Central = 1334
Nag_CentroSymmEven = 2762
Nag_CentroSymmNeither = 2763
Nag_CentroSymmOdd = 2761
Nag_Centroid = 1443
Nag_Character = 2742
Nag_CharacterLabels = 1605
Nag_CheckCon = 1140
Nag_CheckObj = 1139
Nag_CheckObjCon = 1141
Nag_ChiSquare = 1489
Nag_ChkNotSet = 1136
Nag_ClassBoundaryComp = 1347
Nag_ClassBoundaryUser = 1346
Nag_ColMajor = 102
Nag_Cold = 1107
Nag_ColinearVar = 2142
Nag_ColpivRrank = 2780
Nag_ColpivRrankCond = 2781
Nag_ColpivSVrankAbs = 2784
Nag_ColpivSVrankRel = 2785
Nag_ColumnEquilibration = 2003
Nag_CompProbApprox = 1468
Nag_CompProbExact = 1467
Nag_Compl = 1373
Nag_CompleteLink = 1441
Nag_ComputeAll = 1037
Nag_ComputeMoments = 1352
Nag_ComputeQ = 50000
Nag_ComputeSelected = 1039
Nag_ComputeV = 2070
Nag_Compute_1 = 1065
Nag_Compute_2 = 1066
Nag_Compute_3 = 1067
Nag_ConCheck = 1153
Nag_CondBackErrBoth = 3047
Nag_CondBackErrLeft = 3045
Nag_CondBackErrRight = 3046
Nag_CondFanLinVanDooren = 500042
Nag_CondOnly = 3041
Nag_Conj = 191
Nag_ConjTrans = 113
Nag_Continue = 1112
Nag_ContinueAndUpdate = 2170
Nag_ContinueNoUpdate = 2173
Nag_Convolution = 1068
Nag_Coord_Cartesian = 2820
Nag_Coord_Spherical = 2821
Nag_Coord_Spherical_Special = 2822
Nag_Correlation = 1069
Nag_Cos = 2832
Nag_Cosh = 2834
Nag_Cosine = 1079
Nag_CovMatAve = 1389
Nag_CovMatObs = 1390
Nag_CovNotSet = 1388
Nag_CrashNotSet = 1252
Nag_CrashOnce = 1254
Nag_CrashThreeTimes = 1256
Nag_CrashTwice = 1255
Nag_CriteriaNotSet = 1530
Nag_CumulantsOnly = 1351
Nag_CustomMesh = 1118
Nag_Cycling = 1215
Nag_DGauss = 2241
Nag_DInfo_NotSet = 1159
Nag_DLTX = 1269
Nag_DWant_NotSet = 1155
Nag_D_Dbg = 1151
Nag_D_Full = 1150
Nag_D_NoPrint = 1147
Nag_D_NotSet = 1146
Nag_D_Print = 1148
Nag_D_Sum = 1149
Nag_DataByVar = 3070
Nag_DataByObs = 3071
Nag_DataCorr = 1415
Nag_DataCovar = 1416
Nag_Daubechies10 = 2252
Nag_Daubechies2 = 2244
Nag_Daubechies3 = 2245
Nag_Daubechies4 = 2246
Nag_Daubechies5 = 2247
Nag_Daubechies6 = 2248
Nag_Daubechies7 = 2249
Nag_Daubechies8 = 2250
Nag_Daubechies9 = 2251
Nag_DavidsAndMoodsTest = 500032
Nag_DavidsTest = 500033
Nag_Deadpoint = 1209
Nag_DecreasingOrder = 182
Nag_DeepBroad_Search = 1580
Nag_DeepMinObj_Search = 1581
Nag_Deep_Search = 1577
Nag_DefInitMesh = 1091
Nag_Default = 500028
Nag_DefaultJacobian = 2182
Nag_DegenEndCycle = 1241
Nag_DegenInit = 1239
Nag_DegenOptimal = 1240
Nag_DendEast = 1448
Nag_DendNorth = 1446
Nag_DendSouth = 1447
Nag_DendWest = 1449
Nag_Der = 1100
Nag_Deriv = 1170
Nag_DerivNotSet = 1165
Nag_Deriv_Error = 1218
Nag_Deriv_OK = 1160
Nag_Descending = 1598
Nag_DiagonalMatrix = 500023
Nag_DiffInt = 1154
Nag_DirectForm = 1611
Nag_DirectionNotSet = 1617
Nag_DiscrimEstimate = 1426
Nag_DiscrimPredict = 1427
Nag_DistAbs = 1433
Nag_DistEuclid = 1434
Nag_DistSquared = 1435
Nag_DoAll = 50004
Nag_DoBoth = 1044
Nag_DoNotForm = 1058
Nag_DoNothing = 1046
Nag_DoOverwrite = 50006
Nag_DoSquare = 50005
Nag_DownandIn = 2200
Nag_DownandOut = 2201
Nag_DropVar = 2143
Nag_DunnInterval = 1464
Nag_EUV = 2603
Nag_EigVals = 1045
Nag_EigVecs = 1051
Nag_ElementsIn = 1261
Nag_ElementsSeparate = 1262
Nag_Emax = 156
Nag_EmbedPadValues = 2951
Nag_EmbedPadZeros = 2950
Nag_EmbedScaleOne = 2962
Nag_EmbedScaleSqrtTraces = 2961
Nag_EmbedScaleTraces = 2960
Nag_Emin = 155
Nag_EndStateNotSet = 1206
Nag_Entire = 1053
Nag_Eps = 157
Nag_EqualCovar = 1422
Nag_EqualPrior = 1428
Nag_EquilibrateAndFactor = 2012
Nag_Equilibrated = 2001
Nag_ErrorAssess_off = 1097
Nag_ErrorAssess_on = 1098
Nag_EstimateInitialValues = 2171
Nag_EstimatesOrig = 2610
Nag_EstimatesStand = 2611
Nag_EuropeanCall = 1113
Nag_EuropeanPut = 1115
Nag_Even = 2970
Nag_Exact = 1532
Nag_Exp = 2830
Nag_Expo = 1366
Nag_Exponential = 1491
Nag_ExponentialLambda = 2994
Nag_ExtraScale = 1260
Nag_FALSE = 0
Nag_FFT_Initial = 2890
Nag_FFT_Restart = 2892
Nag_FFT_Subsequent = 2891
Nag_FP = 1188
Nag_FPE = 2601
Nag_FacNoRotate = 1421
Nag_FacRotate = 1420
Nag_FacScoreBart = 1419
Nag_FacScoreRegsn = 1418
Nag_Factored = 2010
Nag_FanLinVanDooren = 500043
Nag_FaureTezuka = 2592
Nag_Feasible = 1207
Nag_FeasiblePoint = 1620
Nag_FinishStepwise = 2148
Nag_First = 1599
Nag_FirstCall = 1364
Nag_FirstCallR = 1502
Nag_FirstLevel = 1459
Nag_FirstSegCall = 3000
Nag_First_Int = 1583
Nag_FisherInterval = 1465
Nag_FormBoth = 1057
Nag_FormP = 1059
Nag_FormQ = 1061
Nag_FormX = 1064
Nag_ForwardTransform = 1070
Nag_Freq = 1526
Nag_FrobeniusNorm = 174
Nag_FullpivRrank = 2782
Nag_FullpivRrankCond = 2783
Nag_Fun_Constant = 1161
Nag_Fun_LinearOdd = 1162
Nag_Function = 1172
Nag_Functional = 2881
Nag_GCV = 2604
Nag_Gamma = 1487
Nag_GammaScale = 2993
Nag_Garch_Est_Initial_False = 1562
Nag_Garch_Est_Initial_True = 1561
Nag_Garch_Fcall_False = 1558
Nag_Garch_Fcall_True = 1557
Nag_Garch_Stationary_False = 1560
Nag_Garch_Stationary_True = 1559
Nag_Garch_Useht_False = 1556
Nag_Garch_Useht_True = 1555
Nag_GenPareto = 500012
Nag_GeneralMatrix = 1606
Nag_GenerateFromReference = 2161
Nag_GenerateWithoutReference = 2163
Nag_Grad_HessDiag = 1156
Nag_Grad_HessFull = 1158
Nag_GroupAverage = 1442
Nag_GroupCovar = 2920
Nag_GroupMeans = 1425
Nag_GroupSizePrior = 1429
Nag_GroupVar = 2922
Nag_HSEQRSource = 1035
Nag_Haar = 2243
Nag_HalfPointSymmetric = 2220
Nag_HampelFun = 1381
Nag_HashOK = 1612
Nag_HashTableTooBig = 1613
Nag_Helmert = 1458
Nag_Hermite = 1084
Nag_HessFull = 1157
Nag_Hess_Indefinite = 1219
Nag_Hess_Too_Big = 1214
Nag_HessianFull = 1615
Nag_HessianLimited = 1616
Nag_HessianNotSet = 1614
Nag_HighestTies = 1343
Nag_Hot = 1109
Nag_HuberFun = 1380
Nag_HuberReg = 1375
Nag_Iden = 1367
Nag_Identity = 2860
Nag_Ieee = 154
Nag_IgnoreTies = 1345
Nag_IncSignZerosN = 1484
Nag_IncSignZerosY = 1483
Nag_IncreasingOrder = 181
Nag_Indices = 1056
Nag_InfNorm = 175
Nag_Infeasible = 1212
Nag_Infinite = 1078
Nag_InitNotSet = 1176
Nag_InitQ = 1027
Nag_InitU = 2064
Nag_InitV = 2072
Nag_InitWeights = 2770
Nag_InitZ = 1031
Nag_Init_All = 1179
Nag_Init_F_G_H = 1178
Nag_Init_H_S = 1180
Nag_Init_None = 1177
Nag_InitialValuesSupplied = 2172
Nag_InitializeAndGenerate = 2162
Nag_InitializeI = 1471
Nag_InitializeReference = 2160
Nag_InputA = 1472
Nag_Integer = 2741
Nag_IntegerLabels = 1604
Nag_Integer_Additional = 2743
Nag_Intercept = 2840
Nag_IntermediateCallR = 1503
Nag_Interval = 1055
Nag_Iter = 1128
Nag_Iter_Long = 1129
Nag_JrotInner = 201
Nag_JrotOuter = 202
Nag_JrotSorted = 203
Nag_LARS_LAR = 4000
Nag_LARS_ForwardStagewise = 4001
Nag_LARS_LASSO = 4002
Nag_LARS_PositiveLASSO = 4003
Nag_LARS_None = 4010
Nag_LARS_Centered = 4011
Nag_LARS_Normalized = 4012
Nag_LARS_CenteredNormalized = 4013
Nag_LARS_StepNumber = 4020
Nag_LARS_ScaledNorm = 4021
Nag_LARS_ProportionScaledNorm = 4022
Nag_LARS_UnscaledNorm = 4023
Nag_LARS_ProportionUnscaledNorm = 4024
Nag_LCCDefault = 1350
Nag_LCCImhof = 1349
Nag_LCCPan = 1348
Nag_LDLTX = 1267
Nag_LDX = 1268
Nag_LH_Controller = 1554
Nag_LH_Observer = 1552
Nag_LLTX = 1270
Nag_LOOCV = 2602
Nag_LP = 1189
Nag_LRRoundDown = 2980
Nag_LRRoundUp = 2981
Nag_LS1 = 1194
Nag_LS2 = 1195
Nag_LS3 = 1196
Nag_LS4 = 1197
Nag_LTX = 1272
Nag_LU_SolveB = 1248
Nag_LU_SolveBt = 1249
Nag_LU_SolveL = 1247
Nag_LU_TypeB = 1244
Nag_LU_TypeBS = 1245
Nag_LU_TypeBT = 1246
Nag_LX = 1271
Nag_Laguerre = 1083
Nag_LargeEigVals = 1451
Nag_Last = 1600
Nag_LastCallR = 1504
Nag_LastLevel = 1460
Nag_LeastSquares = 1531
Nag_LeftDerivs = 1104
Nag_LeftDerivs_1 = 500036
Nag_LeftDerivs_2 = 500038
Nag_LeftDerivs_3 = 500040
Nag_LeftSide = 141
Nag_LeftSingVecs = 50007
Nag_LeftSpan = 2082
Nag_LeftSubSeg = 3002
Nag_LeftVecs = 2080
Nag_LeftVecsCtol = 2084
Nag_Legendre = 1081
Nag_LinAlgBand = 1120
Nag_LinAlgFull = 1119
Nag_LinAlgSparse = 1121
Nag_LinFunNotSet = 1173
Nag_Lin_Deriv = 1174
Nag_Lin_NoDeriv = 1175
Nag_Linear = 2851
Nag_LinearHolt = 2582
Nag_LinearRatio = 1354
Nag_Linesearch = 2672
Nag_Log = 1368
Nag_LogNormal = 1494
Nag_LogTransform = 2131
Nag_Loga = 2835
Nag_Logged = 1542
Nag_Logistic = 1371
Nag_Lower = 122
Nag_LowerBound = 2810
Nag_LowerFactored = 500024
Nag_LowerMatrix = 1607
Nag_LowerMinus = 3050
Nag_LowerSemiInfinite = 1077
Nag_LowerTail = 1329
Nag_LowestTies = 1342
Nag_Lsq = 1379
Nag_LsqScale = 1413
Nag_MILP = 1571
Nag_MIP_Best_ISol = 1223
Nag_MIP_Big_Hess_ISol = 1232
Nag_MIP_Big_Hess_No_ISol = 1234
Nag_MIP_Max_Depth_ISol = 1233
Nag_MIP_Max_Depth_No_ISol = 1237
Nag_MIP_Max_Itn_ISol = 1230
Nag_MIP_Max_Itn_No_ISol = 1235
Nag_MIP_Max_Nodes_ISol = 1231
Nag_MIP_Max_Nodes_No_ISol = 1236
Nag_MIP_No_ISol = 1225
Nag_MIP_Root_Big_Hess = 1229
Nag_MIP_Root_Infeasible = 1227
Nag_MIP_Root_Max_Itn = 1228
Nag_MIP_Root_Unbounded = 1226
Nag_MIP_Stop_First_ISol = 1224
Nag_MIP_TypeNotSet = 1570
Nag_MIP_User_Stop = 1238
Nag_MIQP1 = 1572
Nag_MIQP2 = 1573
Nag_MIQP3 = 1574
Nag_MIQP4 = 1575
Nag_MODWTMulti = 2563
Nag_MODWTSingle = 2562
Nag_MOM = 2661
Nag_MOMMLE = 2662
Nag_MPS_List = 1566
Nag_MPS_Summary = 1565
Nag_MRG32k3a = 2155
Nag_MallowsReg = 1376
Nag_ManyIterations = 1086
Nag_Marginal = 1533
Nag_MatCorr_Covar = 1417
Nag_MatCorrelation = 1393
Nag_MatStandardised = 1394
Nag_MatSumSq = 1395
Nag_MatUp = 1431
Nag_MatVarCovar = 1396
Nag_MaxNorm = 177
Nag_Maximize = 1619
Nag_Mean = 1539
Nag_MeanInclude = 1358
Nag_MeanZero = 1359
Nag_Median = 1444
Nag_MersenneTwister = 2152
Nag_MexHat = 2242
Nag_MinObj_Search = 1579
Nag_MinVariance = 1445
Nag_Minimize = 1618
Nag_Minus = 1043
Nag_Mixed = 1089
Nag_ModifiedNewton = 2880
Nag_MoodsTest = 500034
Nag_Morlet = 2240
Nag_MultiLevel = 2561
Nag_MultiplicativeHoltWinters = 2584
Nag_NS_Bounded = 1594
Nag_NS_Infeasible = 1595
Nag_NS_Integer = 1593
Nag_NS_NotBranched = 1592
Nag_NS_NotSet = 1590
Nag_NS_NotSolved = 1591
Nag_NS_Terminated = 1596
Nag_Nearest_Half = 1584
Nag_NegBinomial = 500011
Nag_NewStart = 1110
Nag_NextPoint = 2852
Nag_NoBalancing = 2020
Nag_NoBlocks = 1454
Nag_NoBounds = 1185
Nag_NoBounds_One_Call = 1186
Nag_NoCensored = 1478
Nag_NoCentroSymm = 2760
Nag_NoCheck = 1137
Nag_NoCondBackErr = 3040
Nag_NoConj = 192
Nag_NoCorrection = 1538
Nag_NoCrash = 1253
Nag_NoDeriv = 1171
Nag_NoDerivs = 500035
Nag_NoEquilibration = 2000
Nag_NoFreq = 1527
Nag_NoIntercept = 2841
Nag_NoLOO = 2641
Nag_NoLabels = 1603
Nag_NoMatUp = 1432
Nag_NoOutput = 1564
Nag_NoPara = 2630
Nag_NoPrint = 1125
Nag_NoRemoveVar = 2145
Nag_NoScale = 1258
Nag_NoScaleProvided = 500015
Nag_NoScrambling = 2590
Nag_NoSortEigVals = 2110
Nag_NoTrans = 111
Nag_NoTransNorm = 1407
Nag_NoTransform = 2130
Nag_NoVIF = 2650
Nag_NoVarScale = 1439
Nag_NoVec = 1040
Nag_NoWeights = 1401
Nag_NodSel_NotSet = 1576
Nag_NonRefDiag = 130
Nag_NonUnitDiag = 131
Nag_Norm = 1410
Nag_Normal = 1486
Nag_NormalDistn = 1481
Nag_NormalMean = 2990
Nag_NormalMeanStd = 2992
Nag_NormalScores = 1336
Nag_NormalStd = 2991
Nag_NotEigVecs = 2040
Nag_NotEqualCovar = 1423
Nag_NotFactored = 2011
Nag_NotKnown = 1036
Nag_NotLeftVecs = 2081
Nag_NotLeftWork = 2083
Nag_NotLsqScale = 1414
Nag_NotQ = 1028
Nag_NotRCond = 2050
Nag_NotRegularized = 2911
Nag_NotRightVecs = 2091
Nag_NotRightWork = 2093
Nag_NotSingularVecs = 2030
Nag_NotSupplied = 1264
Nag_NotU = 2061
Nag_NotV = 2071
Nag_NotVT = 2100
Nag_NotZ = 1032
Nag_Not_Converged = 1216
Nag_Not_Kuhn_Tucker = 1217
Nag_NumericalJacobian = 2181
Nag_NumericalStructJacobian = 2184
Nag_ObjCheck = 1152
Nag_ObservAdd = 1360
Nag_ObservDel = 1361
Nag_Odd = 2971
Nag_OneIteration = 1085
Nag_OneNorm = 171
Nag_Optimal = 1208
Nag_Orig = 1408
Nag_OrigCentroid = 1409
Nag_OrigEigVecs = 2042
Nag_OrigNorm = 1411
Nag_OrigNormCentroid = 1412
Nag_OrigPara = 2631
Nag_OsherOriginal = 1122
Nag_OsherPhysical = 1123
Nag_Out_Of_Workspace = 1222
Nag_OutputNotSet = 1563
Nag_OverallVar = 2924
Nag_Overflow = 160
Nag_Overwrite = 2063
Nag_OverwriteVT = 2103
Nag_OwenFaureTezuka = 2593
Nag_OwenLike = 2591
Nag_PDF = 1469
Nag_PWM = 2660
Nag_PWMMLE = 2663
Nag_Pack = 500026
Nag_PackedSingularVecs = 2031
Nag_ParaEstimated = 1497
Nag_ParaSupplied = 1496
Nag_ParallelBlocks = 1456
Nag_Parzen = 1537
Nag_Periodic = 2222
Nag_Permute = 1047
Nag_PerturbOff = 2801
Nag_PerturbOn = 2800
Nag_Plus = 1042
Nag_Poisson = 1492
Nag_PoissonLambda = 2995
Nag_Poly = 1457
Nag_PooledCovar = 2921
Nag_PooledVar = 2923
Nag_PopVarEqual = 1473
Nag_PopVarNotEqual = 1474
Nag_Prec = 158
Nag_PredNoScale = 2620
Nag_PredStdScale = 2621
Nag_PredUserScale = 2622
Nag_PreviousPoint = 2850
Nag_PrintNotSet = 1124
Nag_ProbTypeNotSet = 1187
Nag_Probit = 1372
Nag_PsiNotSet = 1378
Nag_Put = 2191
Nag_QP1 = 1190
Nag_QP2 = 1191
Nag_QP3 = 1192
Nag_QP4 = 1193
Nag_QuadRatio = 1355
Nag_Quad_Gauss_Exponential = 2712
Nag_Quad_Gauss_Exponential_Adjusted = 2724
Nag_Quad_Gauss_Hermite = 2714
Nag_Quad_Gauss_Hermite_Adjusted = 2726
Nag_Quad_Gauss_Jacobi = 2711
Nag_Quad_Gauss_Jacobi_Adjusted = 2725
Nag_Quad_Gauss_Kronrod_15 = 2716
Nag_Quad_Gauss_Kronrod_21 = 2717
Nag_Quad_Gauss_Kronrod_31 = 2718
Nag_Quad_Gauss_Kronrod_41 = 2719
Nag_Quad_Gauss_Kronrod_51 = 2720
Nag_Quad_Gauss_Kronrod_61 = 2721
Nag_Quad_Gauss_Laguerre = 2713
Nag_Quad_Gauss_Laguerre_Adjusted = 2722
Nag_Quad_Gauss_Legendre = 2710
Nag_Quad_Gauss_Rational = 2715
Nag_Quad_Gauss_Rational_Adjusted = 2723
Nag_Quad_Sphere_Default = 2733
Nag_Quad_Sphere_HighAcc = 2728
Nag_Quad_Sphere_HighAcc_BadCentre = 2732
Nag_Quad_Sphere_HighAcc_BadSurface = 2730
Nag_Quad_Sphere_LowAcc = 2727
Nag_Quad_Sphere_LowAcc_BadCentre = 2731
Nag_Quad_Sphere_LowAcc_BadSurface = 2729
Nag_QuasiRandom_Cont = 1622
Nag_QuasiRandom_Faure = 1625
Nag_QuasiRandom_Finish = 1623
Nag_QuasiRandom_Init = 1621
Nag_QuasiRandom_Nied = 1626
Nag_QuasiRandom_Sobol = 1624
Nag_QuasiRandom_SobolA659 = 1627
Nag_RC = 1102
Nag_RCI_Approx = 1480
Nag_RCI_Exact = 1479
Nag_RCondBoth = 50003
Nag_RCondEigVals = 2051
Nag_RCondEigVecs = 50002
Nag_RCondLeft = 2052
Nag_RCondRight = 2053
Nag_RFP_ConjTrans = 2702
Nag_RFP_Normal = 2700
Nag_RFP_Trans = 2701
Nag_RK_2_3 = 1094
Nag_RK_4_5 = 1095
Nag_RK_7_8 = 1096
Nag_RK_onestep = 1093
Nag_RK_range = 1092
Nag_RLRoundDown = 2982
Nag_RLRoundUp = 2983
Nag_Random = 2674
Nag_RandomTies = 1344
Nag_RankScores = 1335
Nag_RatioMoments = 1353
Nag_Rational = 1082
Nag_ReGenerateFromReference = 2165
Nag_ReInitializeAndGenerate = 2164
Nag_ReStart = 1111
Nag_Real = 2740
Nag_RealInfNorm = 176
Nag_RealMaxNorm = 178
Nag_RealOneNorm = 172
Nag_Real_Additional = 2744
Nag_Reci = 1370
Nag_Rectangular = 1534
Nag_RegNotSet = 1374
Nag_Regularized = 2910
Nag_Relative = 1087
Nag_ReuseWeights = 2771
Nag_RightDerivs = 1105
Nag_RightDerivs_1 = 500037
Nag_RightDerivs_2 = 500039
Nag_RightDerivs_3 = 500041
Nag_RightSide = 142
Nag_RightSingVecs = 50008
Nag_RightSubSeg = 3003
Nag_RightVecs = 2090
Nag_RightVecsJRots = 2092
Nag_RightVecsMV = 2094
Nag_Rnd = 153
Nag_RoLoadNotStand = 1406
Nag_RoLoadStand = 1405
Nag_RowAndColumnEquilibration = 2004
Nag_RowColScale = 1259
Nag_RowEquilibration = 2002
Nag_RowMajor = 101
Nag_RuleOfThumb = 2901
Nag_SStress = 1453
Nag_SamplePoints = 1424
Nag_SavageScores = 1340
Nag_Scale = 1049
Nag_ScaleNotSet = 1257
Nag_ScaleProvided = 500014
Nag_ScaleRes = 1601
Nag_ScheffeInterval = 1466
Nag_Schur = 1048
Nag_SchweppeReg = 1377
Nag_ScoresEigenval = 1400
Nag_ScoresNotStand = 1398
Nag_ScoresStand = 1397
Nag_ScoresUnitVar = 1399
Nag_SecondSegCall = 3001
Nag_Select_Modulus = 1265
Nag_Select_RealPart = 1266
Nag_SerialBlocks = 1455
Nag_SeriesEven = 2691
Nag_SeriesGeneral = 2690
Nag_SeriesOdd = 2692
Nag_Sfmin = 161
Nag_Shep = 1103
Nag_SigmaBypas = 1392
Nag_SigmaChi = 1387
Nag_SigmaConst = 1386
Nag_SigmaNotSet = 1384
Nag_SigmaRes = 1385
Nag_SigmaSimul = 1391
Nag_SimpleBdry = 2670
Nag_SimpleCheck = 1138
Nag_SimpleOffBdry = 2671
Nag_Sin = 2831
Nag_Sine = 1080
Nag_SingleCallR = 1501
Nag_SingleExponential = 2580
Nag_SingleLevel = 2560
Nag_SingleLink = 1440
Nag_SingularVecs = 2032
Nag_SingularVecsU = 2062
Nag_SingularVecsVT = 2102
Nag_Sinh = 2833
Nag_SmoothFitFull = 1509
Nag_SmoothFitPartial = 1507
Nag_SmoothFitQuick = 1508
Nag_SmoothParamCV = 1510
Nag_SmoothParamDF = 1512
Nag_SmoothParamGCV = 1511
Nag_Sol = 1099
Nag_SolDer = 1101
Nag_Soln = 1127
Nag_Soln_Iter = 1131
Nag_Soln_Iter_Const = 1133
Nag_Soln_Iter_Diag = 1134
Nag_Soln_Iter_Full = 1135
Nag_Soln_Iter_Long = 1132
Nag_Soln_Root = 1126
Nag_Soln_Root_Iter = 1130
Nag_SomeG_AllJ = 1168
Nag_SomeG_SomeJ = 1166
Nag_SortEigVals = 2111
Nag_SparseFP = 1198
Nag_SparseFPE = 1201
Nag_SparseFPL = 1202
Nag_SparseFPS = 1203
Nag_SparseLP = 1199
Nag_SparseNsym_BiCGSTAB = 1275
Nag_SparseNsym_CG = 1276
Nag_SparseNsym_CGS = 1274
Nag_SparseNsym_Check = 1315
Nag_SparseNsym_CompletePiv = 1284
Nag_SparseNsym_FailDups = 1291
Nag_SparseNsym_FailZeros = 1288
Nag_SparseNsym_JacPrec = 1301
Nag_SparseNsym_KeepZeros = 1287
Nag_SparseNsym_MINRES = 500010
Nag_SparseNsym_MarkowitzPiv = 1285
Nag_SparseNsym_ModFact = 1279
Nag_SparseNsym_NoCheck = 1314
Nag_SparseNsym_NoPiv = 1281
Nag_SparseNsym_NoPrec = 1299
Nag_SparseNsym_PartialPiv = 1283
Nag_SparseNsym_Prec = 1302
Nag_SparseNsym_RGMRES = 1273
Nag_SparseNsym_RemoveDups = 1289
Nag_SparseNsym_RemoveZeros = 1286
Nag_SparseNsym_SSORPrec = 1300
Nag_SparseNsym_StoreCS = 1292
Nag_SparseNsym_StoreSCS = 1293
Nag_SparseNsym_SumDups = 1290
Nag_SparseNsym_TFQMR = 1277
Nag_SparseNsym_UnModFact = 1280
Nag_SparseNsym_UnWeighted = 1304
Nag_SparseNsym_UserPiv = 1282
Nag_SparseNsym_Weighted = 1303
Nag_SparseQP = 1200
Nag_SparseQPP = 1204
Nag_SparseQPS = 1205
Nag_SparseSym_Bisect = 1307
Nag_SparseSym_CG = 1294
Nag_SparseSym_Check = 1317
Nag_SparseSym_FailDups = 500016
Nag_SparseSym_FailZeros = 500017
Nag_SparseSym_InfNorm = 1311
Nag_SparseSym_JacPrec = 1324
Nag_SparseSym_KeepZeros = 1327
Nag_SparseSym_Lanczos = 1295
Nag_SparseSym_MINRES = 500019
Nag_SparseSym_MarkPiv = 1319
Nag_SparseSym_ModFact = 1297
Nag_SparseSym_NoBisect = 1308
Nag_SparseSym_NoCheck = 1316
Nag_SparseSym_NoPiv = 1318
Nag_SparseSym_NoPrec = 1322
Nag_SparseSym_OneNorm = 1309
Nag_SparseSym_Prec = 1321
Nag_SparseSym_RemoveDups = 1326
Nag_SparseSym_RemoveZeros = 1328
Nag_SparseSym_SSORPrec = 1323
Nag_SparseSym_SYMMLQ = 500018
Nag_SparseSym_SumDups = 1325
Nag_SparseSym_TFQMR = 1296
Nag_SparseSym_TermLanczos = 1312
Nag_SparseSym_TermLanczosCG = 1313
Nag_SparseSym_TwoNorm = 1310
Nag_SparseSym_UnModFact = 1298
Nag_SparseSym_UnWeighted = 1306
Nag_SparseSym_UserPiv = 1320
Nag_SparseSym_Weighted = 1305
Nag_Sparse_Colamd = 2122
Nag_Sparse_Identity = 2120
Nag_Sparse_User = 2121
Nag_SplineVectorSort_Sorted = 3010
Nag_SplineVectorSort_Sorted_Indexed = 3011
Nag_SplineVectorSort_Sorted_Indexed_Perm = 3012
Nag_SplineVectorSort_Unsorted = 3013
Nag_SplineVectorSort_Unsorted_Indexed = 3014
Nag_Sqrt = 1369
Nag_SqrtTransform = 2132
Nag_StandPara = 2632
Nag_StartNotSet = 1106
Nag_StatesExternal = 1243
Nag_StatesInternal = 1242
Nag_StoreAll = 2570
Nag_StoreFinal = 2571
Nag_Stress = 1452
Nag_StudentT = 1495
Nag_Subspace = 1050
Nag_Supplied = 1263
Nag_T = 152
Nag_TRUE = 1
Nag_TableStatsAv = 1517
Nag_TableStatsLarge = 1520
Nag_TableStatsMedian = 1518
Nag_TableStatsNObs = 1515
Nag_TableStatsSmall = 1521
Nag_TableStatsTotal = 1516
Nag_TableStatsVar = 1519
Nag_TableUpdateI = 1523
Nag_TableUpdateU = 1522
Nag_TabulateVarCont = 1525
Nag_TabulateVarDiscr = 1524
Nag_Tdistn = 1482
Nag_TestStatisticsDAbs = 1498
Nag_TestStatisticsDNeg = 1500
Nag_TestStatisticsDPos = 1499
Nag_Too_Many_Iter = 1213
Nag_Trans = 112
Nag_Trend = 1540
Nag_TridiagEigVecs = 2041
Nag_TropicalLargest = 500044
Nag_TropicalSmallest = 500045
Nag_Tukey = 1536
Nag_TukeyFun = 1383
Nag_TukeyInterval = 1462
Nag_TukeyScores = 1338
Nag_TwoNorm = 173
Nag_TwoTail = 1333
Nag_TwoTailConfid = 1332
Nag_TwoTailSignif = 1331
Nag_UEV = 2605
Nag_UH_Controller = 1553
Nag_UH_Observer = 1551
Nag_Unbounded = 1211
Nag_Underflow = 159
Nag_Uniform = 1485
Nag_UniformMesh = 1117
Nag_UnitDiag = 132
Nag_UnitRoot = 3020
Nag_UnitRootWithDrift = 3021
Nag_UnitRootWithDriftAndTrend = 3022
Nag_Unlogged = 1541
Nag_Unpack = 500027
Nag_UnscaleRes = 1602
Nag_Unsorted = 500013
Nag_UnweightedEstimate = 1357
Nag_UpandIn = 2202
Nag_UpandOut = 2203
Nag_Update = 1365
Nag_UpdateQ = 1063
Nag_UpdateSchur = 1030
Nag_UpdateZ = 1034
Nag_Upper = 121
Nag_UpperFactored = 500025
Nag_UpperMatrix = 1608
Nag_UpperMinus = 3051
Nag_UpperSemiInfinite = 1076
Nag_UpperTail = 1330
Nag_UseFFT_N = 1514
Nag_UseFFT_Y = 1513
Nag_UseOffset_N = 1529
Nag_UseOffset_Y = 1528
Nag_UseRange = 1543
Nag_UseSD = 1544
Nag_Use_Default = 2680
Nag_Use_Priority = 1585
Nag_Use_Supplied = 2681
Nag_UserInitMesh = 1090
Nag_UserPrior = 1430
Nag_UserProb = 1493
Nag_UserSet = 2673
Nag_UserVec = 1041
Nag_VarScaleRange = 1437
Nag_VarScaleStd = 1436
Nag_VarScaleUser = 1438
Nag_VarSel_NotSet = 1582
Nag_VarianceRatio = 2147
Nag_VgmBessel = 2937
Nag_VgmBrownian = 2943
Nag_VgmCauchy = 2931
Nag_VgmContParam = 2940
Nag_VgmCosine = 2942
Nag_VgmDifferential = 2932
Nag_VgmExponential = 2933
Nag_VgmGauss = 2934
Nag_VgmGenHyp = 2941
Nag_VgmHole = 2938
Nag_VgmNugget = 2935
Nag_VgmSpherical = 2936
Nag_VgmSymmStab = 2930
Nag_VgmWhittleMatern = 2939
Nag_ViaLookUp = 3030
Nag_ViaLookUpOriginal = 3031
Nag_ViaSimulation = 3032
Nag_WaerdenScores = 1339
Nag_WantLOO = 2640
Nag_WantVIF = 2651
Nag_Warm = 1108
Nag_Weakmin = 1210
Nag_WeightObs = 500020
Nag_WeightWindow = 500021
Nag_WeightWindowPos = 500022
Nag_WeightedEstimate = 1356
Nag_WeightedNorm = 2811
Nag_Weights = 1402
Nag_Weightsfreq = 1403
Nag_Weightsvar = 1404
Nag_WholePointSymmetric = 2221
Nag_WichmannHill_I = 2151
Nag_WichmannHill_II = 2153
Nag_WindowSupplied = 2900
Nag_XCheckCon = 1144
Nag_XCheckObj = 1143
Nag_XCheckObjCon = 1145
Nag_XSimpleCheck = 1142
Nag_ZeroColsNormal = 2790
Nag_ZeroColsRestrict = 2791
Nag_ZeroPadded = 2223
Nag_ab_prod = 1549
Nag_ab_sep = 1550
Nag_curr_state = 1548
Nag_next_state = 1547
NoTranspose = 1001
TopPivot = 1011
Transpose = 1002
VariablePivot = 1012
nag_errlist = (STRING * 0).in_dll(_library_cl, 'nag_errlist')

# values for enumeration 'Nag_Boolean'
Nag_Boolean = c_int # enum

if nag_int_type == int32:
    Integer = c_int
else:
    Integer = c_long
Nag_FileID = Integer
Pointer = c_void_p
class Complex(Structure):
    pass
Complex._fields_ = [
    ('re', c_double),
    ('im', c_double),
]

def get_pointer_int_type():
    """
    Return the correct ctype integer for holding a pointer:
    that is, c_longlong on Win64 and c_long otherwise.
    """
    from platform import system
    from struct import calcsize
    bits = 8 * calcsize("P")
    if system() == 'Windows' and bits == 64:
        from ctypes import c_longlong
        return c_longlong
    else:
        return c_long

def get_complex_type():
    """
    Return the correct Complex object for the system API:
    that is, POINTER(Complex) on Win64 and Complex otherwise.
    """
    from platform import system
    from struct import calcsize
    bits = 8 * calcsize("P")
    if system() == 'Windows' and bits == 64:
        return POINTER(Complex)
    else:
        return Complex
Nag_Complex_Type = get_complex_type()

class Complexf(Structure):
    pass
Complexf._fields_ = [
    ('re', c_float),
    ('im', c_float),
]

# values for enumeration 'MatrixTranspose'
MatrixTranspose = c_int # enum

# values for enumeration 'PivotType'
PivotType = c_int # enum

# values for enumeration 'SequenceDirection'
SequenceDirection = c_int # enum

# values for enumeration 'Nag_OrderType'
Nag_OrderType = c_int # enum

# values for enumeration 'Nag_TransType'
Nag_TransType = c_int # enum

# values for enumeration 'Nag_UploType'
Nag_UploType = c_int # enum

# values for enumeration 'Nag_DiagType'
Nag_DiagType = c_int # enum

# values for enumeration 'Nag_SideType'
Nag_SideType = c_int # enum

# values for enumeration 'Nag_CmachType'
Nag_CmachType = c_int # enum

# values for enumeration 'Nag_NormType'
Nag_NormType = c_int # enum

# values for enumeration 'Nag_SortType'
Nag_SortType = c_int # enum

# values for enumeration 'Nag_ConjType'
Nag_ConjType = c_int # enum

# values for enumeration 'Nag_JrotType'
Nag_JrotType = c_int # enum

# values for enumeration 'Nag_ComputeQType'
Nag_ComputeQType = c_int # enum

# values for enumeration 'Nag_ComputeZType'
Nag_ComputeZType = c_int # enum

# values for enumeration 'Nag_EigValsSourceType'
Nag_EigValsSourceType = c_int # enum

# values for enumeration 'Nag_HowManyType'
Nag_HowManyType = c_int # enum

# values for enumeration 'Nag_InitVeenumtype'
Nag_InitVeenumtype = c_int # enum

# values for enumeration 'Nag_SignType'
Nag_SignType = c_int # enum

# values for enumeration 'Nag_JobType'
Nag_JobType = c_int # enum

# values for enumeration 'Nag_EigValRankType'
Nag_EigValRankType = c_int # enum

# values for enumeration 'Nag_RangeType'
Nag_RangeType = c_int # enum

# values for enumeration 'Nag_VectType'
Nag_VectType = c_int # enum

# values for enumeration 'Nag_ComputeType'
Nag_ComputeType = c_int # enum

# values for enumeration 'Nag_EquilibrationType'
Nag_EquilibrationType = c_int # enum

# values for enumeration 'Nag_FactoredFormType'
Nag_FactoredFormType = c_int # enum

# values for enumeration 'Nag_BalanceType'
Nag_BalanceType = c_int # enum

# values for enumeration 'Nag_ComputeSingularVecsType'
Nag_ComputeSingularVecsType = c_int # enum

# values for enumeration 'Nag_ComputeEigVecsType'
Nag_ComputeEigVecsType = c_int # enum

# values for enumeration 'Nag_RCondType'
Nag_RCondType = c_int # enum

# values for enumeration 'Nag_ComputeUType'
Nag_ComputeUType = c_int # enum

# values for enumeration 'Nag_ComputeVType'
Nag_ComputeVType = c_int # enum

# values for enumeration 'Nag_LeftVecsType'
Nag_LeftVecsType = c_int # enum

# values for enumeration 'Nag_RightVecsType'
Nag_RightVecsType = c_int # enum

# values for enumeration 'Nag_ComputeVTType'
Nag_ComputeVTType = c_int # enum

# values for enumeration 'Nag_SortEigValsType'
Nag_SortEigValsType = c_int # enum
class Ac02af(Structure):
    pass
Ac02af._fields_ = [
    ('ovflow', Nag_Boolean),
    ('unflow', Nag_Boolean),
]
class Bc02af(Structure):
    pass
Bc02af._fields_ = [
    ('expdep', Integer),
    ('finity', c_double),
    ('lrgexp', Integer),
    ('sqrtfy', c_double),
    ('sqrtty', c_double),
    ('tiny', c_double),
    ('deps', c_double),
    ('emin', Integer),
    ('emax', Integer),
]
class Ac02ag(Structure):
    pass
Ac02ag._fields_ = [
    ('ovflow', Nag_Boolean),
    ('unflow', Nag_Boolean),
]
class Bc02ag(Structure):
    pass
Bc02ag._fields_ = [
    ('finity', c_double),
    ('sqrtfy', c_double),
    ('sqrtty', c_double),
    ('tiny', c_double),
    ('deps', c_double),
    ('eminm1', Integer),
    ('emaxm1', Integer),
    ('expdep', Integer),
    ('lrgexp', Integer),
]
class Cc02ag(Structure):
    pass
Cc02ag._fields_ = [
    ('dpnewl', c_double),
    ('dpnewu', c_double),
    ('deps', c_double),
    ('temp', c_double),
    ('fact', c_double),
    ('dbase', Integer),
    ('mnexp', Integer),
    ('mxexp', Integer),
    ('newl', Integer),
    ('newu', Integer),
]

# values for enumeration 'Nag_VectorOp'
Nag_VectorOp = c_int # enum

# values for enumeration 'Nag_TransformDirection'
Nag_TransformDirection = c_int # enum

# values for enumeration 'Nag_Series'
Nag_Series = c_int # enum

# values for enumeration 'Nag_FFT_InitFlag'
Nag_FFT_InitFlag = c_int # enum

# values for enumeration 'Nag_QuadWeight'
Nag_QuadWeight = c_int # enum

# values for enumeration 'Nag_BoundInterval'
Nag_BoundInterval = c_int # enum

# values for enumeration 'Nag_TrigTransform'
Nag_TrigTransform = c_int # enum

# values for enumeration 'Nag_GaussFormulae'
Nag_GaussFormulae = c_int # enum

# values for enumeration 'Nag_MCMethod'
Nag_MCMethod = c_int # enum

# values for enumeration 'Nag_ErrorControl'
Nag_ErrorControl = c_int # enum

# values for enumeration 'Nag_MeshSet'
Nag_MeshSet = c_int # enum

# values for enumeration 'Nag_RK_task'
Nag_RK_task = c_int # enum

# values for enumeration 'Nag_RK_method'
Nag_RK_method = c_int # enum

# values for enumeration 'Nag_ErrorAssess'
Nag_ErrorAssess = c_int # enum

# values for enumeration 'Nag_SolDeriv'
Nag_SolDeriv = c_int # enum

# values for enumeration 'Nag_EvaluateJacobian'
Nag_EvaluateJacobian = c_int # enum

# values for enumeration 'Nag_2d_Scat_Method'
Nag_2d_Scat_Method = c_int # enum

# values for enumeration 'Nag_DerivType'
Nag_DerivType = c_int # enum

# values for enumeration 'Nag_SplineVectorSort'
Nag_SplineVectorSort = c_int # enum

# values for enumeration 'Nag_Start'
Nag_Start = c_int # enum

# values for enumeration 'Nag_OptionType'
Nag_OptionType = c_int # enum

# values for enumeration 'Nag_MeshType'
Nag_MeshType = c_int # enum

# values for enumeration 'Nag_LinAlgOption'
Nag_LinAlgOption = c_int # enum

# values for enumeration 'Nag_OsherVersion'
Nag_OsherVersion = c_int # enum

# values for enumeration 'Nag_PrintType'
Nag_PrintType = c_int # enum

# values for enumeration 'Nag_GradChk'
Nag_GradChk = c_int # enum

# values for enumeration 'Nag_DPrintType'
Nag_DPrintType = c_int # enum

# values for enumeration 'Nag_CheckType'
Nag_CheckType = c_int # enum

# values for enumeration 'Nag_DWantType'
Nag_DWantType = c_int # enum

# values for enumeration 'Nag_DerivInfo'
Nag_DerivInfo = c_int # enum

# values for enumeration 'Nag_DerivSet'
Nag_DerivSet = c_int # enum

# values for enumeration 'Nag_FunType'
Nag_FunType = c_int # enum

# values for enumeration 'Nag_LinFun'
Nag_LinFun = c_int # enum

# values for enumeration 'Nag_InitType'
Nag_InitType = c_int # enum

# values for enumeration 'Nag_BoundType'
Nag_BoundType = c_int # enum

# values for enumeration 'Nag_ProblemType'
Nag_ProblemType = c_int # enum

# values for enumeration 'Nag_EndState'
Nag_EndState = c_int # enum

# values for enumeration 'Nag_DegenJob'
Nag_DegenJob = c_int # enum

# values for enumeration 'Nag_SetStateMode'
Nag_SetStateMode = c_int # enum

# values for enumeration 'Nag_SparseQP_LU_Type'
Nag_SparseQP_LU_Type = c_int # enum

# values for enumeration 'Nag_SparseQP_LU_SolveType'
Nag_SparseQP_LU_SolveType = c_int # enum

# values for enumeration 'Nag_SparseQP_BasisFactType'
Nag_SparseQP_BasisFactType = c_int # enum

# values for enumeration 'Nag_CrashType'
Nag_CrashType = c_int # enum

# values for enumeration 'Nag_DataByObsOrVar'
Nag_DataByObsOrVar = c_int # enum

# values for enumeration 'Nag_ScaleType'
Nag_ScaleType = c_int # enum

# values for enumeration 'Nag_WhereElements'
Nag_WhereElements = c_int # enum

# values for enumeration 'Nag_InitRotation'
Nag_InitRotation = c_int # enum

# values for enumeration 'Nag_Select_Eigenvalues'
Nag_Select_Eigenvalues = c_int # enum

# values for enumeration 'Nag_SolveSystem'
Nag_SolveSystem = c_int # enum

# values for enumeration 'Nag_SparseNsym_Method'
Nag_SparseNsym_Method = c_int # enum

# values for enumeration 'Nag_SparseNsym_Fact'
Nag_SparseNsym_Fact = c_int # enum

# values for enumeration 'Nag_SparseNsym_Piv'
Nag_SparseNsym_Piv = c_int # enum

# values for enumeration 'Nag_SparseNsym_Zeros'
Nag_SparseNsym_Zeros = c_int # enum

# values for enumeration 'Nag_SparseNsym_Dups'
Nag_SparseNsym_Dups = c_int # enum

# values for enumeration 'Nag_SparseNsym_Store'
Nag_SparseNsym_Store = c_int # enum

# values for enumeration 'Nag_SparseSym_Method'
Nag_SparseSym_Method = c_int # enum

# values for enumeration 'Nag_SparseSym_Fact'
Nag_SparseSym_Fact = c_int # enum

# values for enumeration 'Nag_SparseNsym_PrecType'
Nag_SparseNsym_PrecType = c_int # enum

# values for enumeration 'Nag_SparseNsym_Weight'
Nag_SparseNsym_Weight = c_int # enum

# values for enumeration 'Nag_SparseSym_Weight'
Nag_SparseSym_Weight = c_int # enum

# values for enumeration 'Nag_SparseSym_Bisection'
Nag_SparseSym_Bisection = c_int # enum

# values for enumeration 'Nag_SparseSym_Norm'
Nag_SparseSym_Norm = c_int # enum

# values for enumeration 'Nag_SparseSym_Term'
Nag_SparseSym_Term = c_int # enum

# values for enumeration 'Nag_SparseNsym_CheckData'
Nag_SparseNsym_CheckData = c_int # enum

# values for enumeration 'Nag_SparseSym_CheckData'
Nag_SparseSym_CheckData = c_int # enum

# values for enumeration 'Nag_SparseSym_Piv'
Nag_SparseSym_Piv = c_int # enum

# values for enumeration 'Nag_SparseSym_PrecType'
Nag_SparseSym_PrecType = c_int # enum

# values for enumeration 'Nag_SparseSym_Dups'
Nag_SparseSym_Dups = c_int # enum

# values for enumeration 'Nag_SparseSym_Zeros'
Nag_SparseSym_Zeros = c_int # enum

# values for enumeration 'Nag_LARSTargetType'
Nag_LARSTargetType = c_int # enum

# values for enumeration 'Nag_LARSModelType'
Nag_LARSModelType = c_int # enum

# values for enumeration 'Nag_LARSPreProcess'
Nag_LARSPreProcess = c_int # enum

# values for enumeration 'Nag_ColumnPermutationType'
Nag_ColumnPermutationType = c_int # enum

# values for enumeration 'Nag_Transform'
Nag_Transform = c_int # enum

# values for enumeration 'Nag_FullStepwise'
Nag_FullStepwise = c_int # enum

# values for enumeration 'Nag_CallPut'
Nag_CallPut = c_int # enum

# values for enumeration 'Nag_Barrier'
Nag_Barrier = c_int # enum

# values for enumeration 'Nag_OptimOpt'
Nag_OptimOpt = c_int # enum
class Nag_f04ycf_save(Structure):
    pass
Nag_f04ycf_save._fields_ = [
    ('iter', Integer),
    ('jump', Integer),
    ('j', Integer),
]

# values for enumeration 'Nag_Wavelet'
Nag_Wavelet = c_int # enum

# values for enumeration 'Nag_WaveletMode'
Nag_WaveletMode = c_int # enum

# values for enumeration 'Nag_WaveletTransform'
Nag_WaveletTransform = c_int # enum

# values for enumeration 'Nag_WaveletCoefficients'
Nag_WaveletCoefficients = c_int # enum

# values for enumeration 'Nag_ExpSmoothType'
Nag_ExpSmoothType = c_int # enum
class Nag_Sparse_Comm_tag(Structure):
    pass
Nag_Sparse_Comm_tag._fields_ = [
    ('idatax', Integer * 20),
    ('rdatax', c_double * 20),
    ('idata', Integer * 20),
    ('rdata', c_double * 20),
    ('work', POINTER(c_double)),
    ('iwork', POINTER(Integer)),
    ('lwork', Integer),
    ('liwork', Integer),
    ('f11gac_lwreq', Integer),
    ('f11bby_iter', Integer),
    ('f11bby_jump', Integer),
    ('f11bby_j', Integer),
    ('f11bby_kase', Integer),
    ('f11gbz_ib', Integer),
    ('f11gbz_id', Integer),
    ('f11gbz_ie2', Integer),
    ('f11gbz_infoch', Integer),
    ('f11gbz_ip', Integer),
    ('f11gbz_ir', Integer),
    ('f11gbz_iterm', Integer),
    ('f11gbz_itn', Integer),
    ('f11gbz_its', Integer),
    ('f11gbz_iw', Integer),
    ('f11gbz_iwgt', Integer),
    ('f11gbz_ix', Integer),
    ('f11gbz_maxitn', Integer),
    ('f11gbz_maxits', Integer),
    ('f11gbz_monit', Integer),
    ('f11gbz_n', Integer),
    ('f11gbz_norm', Integer),
    ('f11gbz_sigcmp', Integer),
    ('f11gbz_sigcmx', Integer),
    ('f11gbz_stage', Integer),
    ('f11gbz_anorm', c_double),
    ('f11gbz_bnorm', c_double),
    ('f11gbz_sigerr', c_double),
    ('f11gbz_sigerc', c_double),
    ('f11gbz_sigmax', c_double),
    ('f11gbz_sigmac', c_double),
    ('f11gbz_stplhs', c_double),
    ('f11gbz_stprhs', c_double),
    ('f11gbz_talpha', c_double),
    ('f11gbz_tbeta', c_double),
    ('f11gbz_tol', c_double),
    ('f11gbz_xnorm', c_double),
    ('f11gbz_floop', Nag_Boolean),
    ('f11gbz_next', Nag_Boolean),
    ('f11gbz_precon', Nag_SparseSym_PrecType),
    ('f11gby_alpha', c_double),
    ('f11gby_anorm', c_double),
    ('f11gby_beta1', c_double),
    ('f11gby_beta2', c_double),
    ('f11gby_betain', c_double),
    ('f11gby_bnorm', c_double),
    ('f11gby_bnorm2', c_double),
    ('f11gby_gammab', c_double),
    ('f11gby_pi', c_double),
    ('f11gby_rho1', c_double),
    ('f11gby_rho2', c_double),
    ('f11gby_rnorm2', c_double),
    ('f11gby_sigerr', c_double),
    ('f11gby_sigerc', c_double),
    ('f11gby_sigmax', c_double),
    ('f11gby_sigmac', c_double),
    ('f11gby_stplhs', c_double),
    ('f11gby_stprhs', c_double),
    ('f11gby_tol', c_double),
    ('f11gby_xnorm', c_double),
    ('f11gby_xnorm0', c_double),
    ('f11gby_zeta', c_double),
    ('f11gby_zetab', c_double),
    ('f11gby_ib', Integer),
    ('f11gby_id', Integer),
    ('f11gby_ie2', Integer),
    ('f11gby_infoch', Integer),
    ('f11gby_iterm', Integer),
    ('f11gby_itn', Integer),
    ('f11gby_its', Integer),
    ('f11gby_iv1', Integer),
    ('f11gby_iv2', Integer),
    ('f11gby_iw1', Integer),
    ('f11gby_iw2', Integer),
    ('f11gby_iwgt', Integer),
    ('f11gby_ix', Integer),
    ('f11gby_maxitn', Integer),
    ('f11gby_maxits', Integer),
    ('f11gby_monit', Integer),
    ('f11gby_n', Integer),
    ('f11gby_norm', Integer),
    ('f11gby_resid', Integer),
    ('f11gby_sigcmp', Integer),
    ('f11gby_sigcmx', Integer),
    ('f11gby_stage', Integer),
    ('f11gby_floop', Nag_Boolean),
    ('f11gby_next', Nag_Boolean),
    ('f11gby_usexc', Nag_Boolean),
    ('f11gby_precon', Nag_SparseSym_PrecType),
    ('f11gbx_fpass', Nag_Boolean),
    ('f11gbv_rho', c_double),
    ('f11gbv_xi', c_double),
    ('f11gbs_dov1', Nag_Boolean),
    ('f11gbr_deltab', c_double),
    ('f11gbr_epslon', c_double),
    ('f11baz_idatax', Integer * 20),
    ('f11baz_rdatax', c_double * 20),
    ('f11gbw_eps', c_double),
    ('f11gbw_epslon', c_double),
    ('f11gbw_mu1', c_double),
    ('f11gbw_mu1l', c_double),
    ('f11gbw_mun', c_double),
    ('f11gbw_munh', c_double),
    ('f11gbw_sigmx0', c_double),
    ('f11gbw_sigmx1', c_double),
    ('f11gbw_sigtol', c_double),
    ('f11gbw_ta2', c_double),
    ('f11gbw_tb2', c_double),
    ('f11gbw_tnorm', c_double),
    ('f11gbw_maxits', Integer),
    ('f11gbw_sigcmp', Nag_Boolean),
    ('f11gbc_ifaill', Integer),
    ('f11gbc_ifailm', Integer),
    ('f11gbc_info', Integer),
    ('f11gbc_irevcx', Integer),
    ('f11gbc_kill', Integer),
    ('f11gbc_idata', Integer * 20),
    ('f11gbc_lworkl', Integer),
    ('f11gbc_done', Nag_Boolean),
    ('f11gbc_first', Nag_Boolean),
    ('f11gbc_rdata', c_double * 20),
    ('f11gbq_factol', c_double),
    ('f11gbq_stprhx', c_double),
    ('f11gbq_tolc', c_double),
    ('f11gbq_toleps', c_double),
    ('f11gbq_tolf', c_double),
    ('f11gbq_toll', c_double),
    ('f11gbq_xcbnrm', c_double),
    ('f11gbq_xlbnrm', c_double),
    ('f11gbq_xnorm', c_double),
    ('f11gbq_loop', Integer),
    ('f11gbq_fterm', Nag_Boolean),
    ('f11gbu_epslon', c_double),
    ('f11gbu_stprhx', c_double),
    ('f11gbu_toleps', c_double),
    ('f11gbu_tolf', c_double),
    ('f11gbu_xnorm0', c_double),
    ('f11gbu_loop', Integer),
    ('f11gbu_nrestl', Integer),
    ('f11gbu_nrestc', Integer),
    ('f11gbu_fterm', Nag_Boolean),
    ('f11bbz_iwgt', Integer),
    ('f11bbz_norm', Integer),
    ('f11bbz_iterm', Integer),
    ('f11bbz_monit', Integer),
    ('f11bbz_m', Integer),
    ('f11bbz_n', Integer),
    ('f11bbz_stage', Integer),
    ('f11bbz_ib', Integer),
    ('f11bbz_ic', Integer),
    ('f11bbz_ih', Integer),
    ('f11bbz_iq', Integer),
    ('f11bbz_ir', Integer),
    ('f11bbz_is', Integer),
    ('f11bbz_it', Integer),
    ('f11bbz_ix', Integer),
    ('f11bbz_infoch', Integer),
    ('f11bbz_imonit', Integer),
    ('f11bbz_maxitn', Integer),
    ('f11bbz_itn', Integer),
    ('f11bbz_anorm', c_double),
    ('f11bbz_bnorm', c_double),
    ('f11bbz_rnrm20', c_double),
    ('f11bbz_xnorm', c_double),
    ('f11bbz_rnorm2', c_double),
    ('f11bbz_xnorm2', c_double),
    ('f11bbz_sigmax', c_double),
    ('f11bbz_stplhs', c_double),
    ('f11bbz_stprhs', c_double),
    ('f11bbz_tol', c_double),
    ('f11bbz_anrcmp', Nag_Boolean),
    ('f11bbz_precon', Nag_Boolean),
    ('f11bbz_next', Nag_Boolean),
    ('f11bbz_fcall', Nag_Boolean),
    ('f11bbu_first', Nag_Boolean),
    ('f11bbu_usedot', Nag_Boolean),
    ('f11bbv_tolf', c_double),
    ('f11bbr_alpha', c_double),
    ('f11bbr_rho', c_double),
    ('f11bbr_do1', Nag_Boolean),
    ('f11bbw_stpr', c_double),
    ('f11bbw_m1', Integer),
    ('f11bbw_tnorm', c_double),
    ('f11bbw_im', Integer),
    ('f11bbw_sigcmp', Nag_Boolean),
    ('f11bbw_im1', Integer),
    ('f11bbw_tinorm', c_double),
    ('f11bbw_restrt', Nag_Boolean),
    ('f11bbw_tol1', c_double),
    ('f11bbt_iwgt', Integer),
    ('f11bbt_norm', Integer),
    ('f11bbt_next', Nag_Boolean),
    ('f11bbt_n', Integer),
    ('f11bbt_fcall', Nag_Boolean),
    ('f11bbt_irbar', Integer),
    ('f11bbt_stage', Integer),
    ('f11bbt_anorm', c_double),
    ('f11bbt_bnorm', c_double),
    ('f11bbt_iterm', Integer),
    ('f11bbt_monit', Integer),
    ('f11bbt_xnorm', c_double),
    ('f11bbt_ib', Integer),
    ('f11bbt_ip', Integer),
    ('f11bbt_ir', Integer),
    ('f11bbt_it', Integer),
    ('f11bbt_ix', Integer),
    ('f11bbt_iy', Integer),
    ('f11bbt_infoch', Integer),
    ('f11bbt_anrcmp', Nag_Boolean),
    ('f11bbt_precon', Nag_Boolean),
    ('f11bbt_sigmax', c_double),
    ('f11bbt_maxitn', Integer),
    ('f11bbt_stplhs', c_double),
    ('f11bbt_stprhs', c_double),
    ('f11bbt_restrt', Nag_Boolean),
    ('f11bbt_itn', Integer),
    ('f11bbt_tol', c_double),
    ('f11bbq_tolf', c_double),
    ('f11bbq_loop', Integer),
    ('f11bbq_pass1', Nag_Boolean),
    ('f11bbq_frest', Nag_Boolean),
    ('f11bbq_xnmax', c_double),
    ('f11bbq_toleps', c_double),
    ('f11bbq_stplhx', c_double),
    ('f11bbq_stprhx', c_double),
    ('f11bbq_zz1', c_double),
    ('f11bbq_zz2', c_double),
    ('f11bbp_doxn', Nag_Boolean),
    ('f11bbp_next', Nag_Boolean),
    ('f11bbp_fcall', Nag_Boolean),
    ('f11bbp_anrcmp', Nag_Boolean),
    ('f11bbp_finish', Nag_Boolean),
    ('f11bbp_precon', Nag_Boolean),
    ('f11bbp_restrt', Nag_Boolean),
    ('f11bbp_imonit', Integer),
    ('f11bbp_maxitn', Integer),
    ('f11bbp_igam', Integer),
    ('f11bbp_itau', Integer),
    ('f11bbp_iwgt', Integer),
    ('f11bbp_norm', Integer),
    ('f11bbp_m', Integer),
    ('f11bbp_n', Integer),
    ('f11bbp_igamp', Integer),
    ('f11bbp_irbar', Integer),
    ('f11bbp_stage', Integer),
    ('f11bbp_xnorm', c_double),
    ('f11bbp_anorm', c_double),
    ('f11bbp_bnorm', c_double),
    ('f11bbp_iterm', Integer),
    ('f11bbp_monit', Integer),
    ('f11bbp_ib', Integer),
    ('f11bbp_iq', Integer),
    ('f11bbp_ir', Integer),
    ('f11bbp_ix', Integer),
    ('f11bbp_infoch', Integer),
    ('f11bbp_idx', Integer),
    ('f11bbp_itn', Integer),
    ('f11bbp_ipx', Integer),
    ('f11bbp_sigmax', c_double),
    ('f11bbp_stplhs', c_double),
    ('f11bbp_stprhs', c_double),
    ('f11bbp_tol', c_double),
    ('f11bbn_rhox', c_double),
    ('f11bbn_alpha', c_double),
    ('f11bbn_tol1', c_double),
    ('f11bbn_tol2', c_double),
    ('f11bbn_rn1', c_double),
    ('f11bbn_i1', Integer),
    ('f11bbn_j1', Integer),
    ('f11bbn_n1', Integer),
    ('f11bbn_im', Integer),
    ('f11bbn_jn1', Integer),
    ('f11bbn_imn', Integer),
    ('f11bbn_doq', Nag_Boolean),
    ('f11bbm_tolf', c_double),
    ('f11bbm_jitn', Integer),
    ('f11bbm_kitn', Integer),
    ('f11bbm_loop', Integer),
    ('f11bbm_pass1', Nag_Boolean),
    ('f11bbm_frest', Nag_Boolean),
    ('f11bbm_irest', Integer),
    ('f11bbm_xnmax', c_double),
    ('f11bbm_toleps', c_double),
    ('f11bbm_stplhx', c_double),
    ('f11bbm_stprhx', c_double),
    ('f11bbm_zz1', c_double),
    ('f11bbm_zz2', c_double),
    ('f11bbf_irevcx', Integer),
    ('f11bbf_done', Nag_Boolean),
    ('f11bbf_first', Nag_Boolean),
    ('f11bbf_info', Integer),
    ('f11bbf_kill', Integer),
    ('f11bbf_idata', Integer * 20),
    ('f11bbf_rdata', c_double * 20),
    ('f11bbf_ifaill', Integer),
    ('f11bbf_ifailm', Integer),
    ('f11bbf_lworkl', Integer),
]
Nag_Sparse_Comm = Nag_Sparse_Comm_tag

# values for enumeration 'Nag_TailProbability'
Nag_TailProbability = c_int # enum

# values for enumeration 'Nag_Scores'
Nag_Scores = c_int # enum

# values for enumeration 'Nag_Ties'
Nag_Ties = c_int # enum

# values for enumeration 'Nag_ClassBoundary'
Nag_ClassBoundary = c_int # enum

# values for enumeration 'Nag_LCCMethod'
Nag_LCCMethod = c_int # enum

# values for enumeration 'Nag_SelectMoments'
Nag_SelectMoments = c_int # enum

# values for enumeration 'Nag_MomentType'
Nag_MomentType = c_int # enum

# values for enumeration 'Nag_IncludeWeight'
Nag_IncludeWeight = c_int # enum

# values for enumeration 'Nag_IncludeMean'
Nag_IncludeMean = c_int # enum

# values for enumeration 'Nag_UpdateObserv'
Nag_UpdateObserv = c_int # enum

# values for enumeration 'Nag_SumSquare'
Nag_SumSquare = c_int # enum

# values for enumeration 'Nag_Initialize'
Nag_Initialize = c_int # enum

# values for enumeration 'Nag_Link'
Nag_Link = c_int # enum

# values for enumeration 'Nag_RegType'
Nag_RegType = c_int # enum

# values for enumeration 'Nag_PsiFun'
Nag_PsiFun = c_int # enum

# values for enumeration 'Nag_SigmaEst'
Nag_SigmaEst = c_int # enum

# values for enumeration 'Nag_CovMatrixEst'
Nag_CovMatrixEst = c_int # enum

# values for enumeration 'Nag_SigmaSimulEst'
Nag_SigmaSimulEst = c_int # enum

# values for enumeration 'Nag_PrinCompMat'
Nag_PrinCompMat = c_int # enum

# values for enumeration 'Nag_PrinCompScores'
Nag_PrinCompScores = c_int # enum

# values for enumeration 'Nag_Weightstype'
Nag_Weightstype = c_int # enum

# values for enumeration 'Nag_RotationLoading'
Nag_RotationLoading = c_int # enum

# values for enumeration 'Nag_TransNorm'
Nag_TransNorm = c_int # enum

# values for enumeration 'Nag_RotationScale'
Nag_RotationScale = c_int # enum

# values for enumeration 'Nag_FacMat'
Nag_FacMat = c_int # enum

# values for enumeration 'Nag_FacScoreMethod'
Nag_FacScoreMethod = c_int # enum

# values for enumeration 'Nag_FacRotation'
Nag_FacRotation = c_int # enum

# values for enumeration 'Nag_GroupCovars'
Nag_GroupCovars = c_int # enum

# values for enumeration 'Nag_MahalDist'
Nag_MahalDist = c_int # enum

# values for enumeration 'Nag_DiscrimMethod'
Nag_DiscrimMethod = c_int # enum

# values for enumeration 'Nag_PriorProbability'
Nag_PriorProbability = c_int # enum

# values for enumeration 'Nag_MatUpdate'
Nag_MatUpdate = c_int # enum

# values for enumeration 'Nag_DistanceType'
Nag_DistanceType = c_int # enum

# values for enumeration 'Nag_VarScaleType'
Nag_VarScaleType = c_int # enum

# values for enumeration 'Nag_ClusterMethod'
Nag_ClusterMethod = c_int # enum

# values for enumeration 'Nag_DendOrient'
Nag_DendOrient = c_int # enum

# values for enumeration 'Nag_Eigenvalues'
Nag_Eigenvalues = c_int # enum

# values for enumeration 'Nag_ScaleCriterion'
Nag_ScaleCriterion = c_int # enum

# values for enumeration 'Nag_VarCovar'
Nag_VarCovar = c_int # enum

# values for enumeration 'Nag_Blocks'
Nag_Blocks = c_int # enum

# values for enumeration 'Nag_DummyType'
Nag_DummyType = c_int # enum

# values for enumeration 'Nag_IntervalType'
Nag_IntervalType = c_int # enum

# values for enumeration 'Nag_CompProb'
Nag_CompProb = c_int # enum

# values for enumeration 'Nag_DiscreteDistrib'
Nag_DiscreteDistrib = c_int # enum

# values for enumeration 'Nag_InitializeA'
Nag_InitializeA = c_int # enum

# values for enumeration 'Nag_BaseRNG'
Nag_BaseRNG = c_int # enum

# values for enumeration 'Nag_ModeRNG'
Nag_ModeRNG = c_int # enum

# values for enumeration 'Nag_InitialValues'
Nag_InitialValues = c_int # enum

# values for enumeration 'Nag_Pop'
Nag_Pop = c_int # enum

# values for enumeration 'Nag_Variogram'
Nag_Variogram = c_int # enum

# values for enumeration 'Nag_EmbedPad'
Nag_EmbedPad = c_int # enum

# values for enumeration 'Nag_EmbedScale'
Nag_EmbedScale = c_int # enum

# values for enumeration 'Nag_Parity'
Nag_Parity = c_int # enum

# values for enumeration 'Nag_BridgeOrder'
Nag_BridgeOrder = c_int # enum

# values for enumeration 'Nag_PopVar'
Nag_PopVar = c_int # enum

# values for enumeration 'Nag_CEMethod'
Nag_CEMethod = c_int # enum

# values for enumeration 'Nag_CestMethod'
Nag_CestMethod = c_int # enum

# values for enumeration 'Nag_RCIMethod'
Nag_RCIMethod = c_int # enum

# values for enumeration 'Nag_ErrorDistn'
Nag_ErrorDistn = c_int # enum

# values for enumeration 'Nag_IncSignZeros'
Nag_IncSignZeros = c_int # enum

# values for enumeration 'Nag_Distributions'
Nag_Distributions = c_int # enum

# values for enumeration 'Nag_ParaEstimates'
Nag_ParaEstimates = c_int # enum

# values for enumeration 'Nag_TestStatistics'
Nag_TestStatistics = c_int # enum

# values for enumeration 'Nag_CallTypes'
Nag_CallTypes = c_int # enum

# values for enumeration 'Nag_Smooth_Type'
Nag_Smooth_Type = c_int # enum

# values for enumeration 'Nag_SmoothFitType'
Nag_SmoothFitType = c_int # enum

# values for enumeration 'Nag_SmoothParamMethods'
Nag_SmoothParamMethods = c_int # enum

# values for enumeration 'Nag_UseFFT'
Nag_UseFFT = c_int # enum

# values for enumeration 'Nag_TableStats'
Nag_TableStats = c_int # enum

# values for enumeration 'Nag_TableUpdate'
Nag_TableUpdate = c_int # enum

# values for enumeration 'Nag_TabulateVar'
Nag_TabulateVar = c_int # enum

# values for enumeration 'Nag_FreqTime'
Nag_FreqTime = c_int # enum

# values for enumeration 'Nag_UseOffset'
Nag_UseOffset = c_int # enum

# values for enumeration 'Nag_Likelihood'
Nag_Likelihood = c_int # enum

# values for enumeration 'Nag_LagWindow'
Nag_LagWindow = c_int # enum

# values for enumeration 'NagMeanOrTrend'
NagMeanOrTrend = c_int # enum

# values for enumeration 'Nag_LoggedSpectra'
Nag_LoggedSpectra = c_int # enum

# values for enumeration 'Nag_RangeStat'
Nag_RangeStat = c_int # enum

# values for enumeration 'Nag_CovOrCorr'
Nag_CovOrCorr = c_int # enum

# values for enumeration 'Nag_state'
Nag_state = c_int # enum

# values for enumeration 'Nag_ab_input'
Nag_ab_input = c_int # enum

# values for enumeration 'Nag_ObserverForm'
Nag_ObserverForm = c_int # enum

# values for enumeration 'Nag_ControllerForm'
Nag_ControllerForm = c_int # enum

# values for enumeration 'Nag_Garch_Useht_Type'
Nag_Garch_Useht_Type = c_int # enum

# values for enumeration 'Nag_Garch_Fcall_Type'
Nag_Garch_Fcall_Type = c_int # enum

# values for enumeration 'Nag_Garch_Stationary_Type'
Nag_Garch_Stationary_Type = c_int # enum

# values for enumeration 'Nag_Garch_Est_Initial_Type'
Nag_Garch_Est_Initial_Type = c_int # enum

# values for enumeration 'Nag_WindowType'
Nag_WindowType = c_int # enum

# values for enumeration 'Nag_RegularizedType'
Nag_RegularizedType = c_int # enum

# values for enumeration 'Nag_OutputType'
Nag_OutputType = c_int # enum

# values for enumeration 'Nag_BB_Fail'
Nag_BB_Fail = c_int # enum

# values for enumeration 'Nag_MIP_ProbType'
Nag_MIP_ProbType = c_int # enum

# values for enumeration 'Nag_Node_Selection'
Nag_Node_Selection = c_int # enum

# values for enumeration 'Nag_Var_Selection'
Nag_Var_Selection = c_int # enum

# values for enumeration 'Nag_Branch_Direction'
Nag_Branch_Direction = c_int # enum

# values for enumeration 'Nag_NodeStatus'
Nag_NodeStatus = c_int # enum

# values for enumeration 'Nag_SortOrder'
Nag_SortOrder = c_int # enum

# values for enumeration 'Nag_SearchMatch'
Nag_SearchMatch = c_int # enum

# values for enumeration 'Nag_ScaleResType'
Nag_ScaleResType = c_int # enum

# values for enumeration 'Nag_LabelType'
Nag_LabelType = c_int # enum

# values for enumeration 'Nag_MatrixType'
Nag_MatrixType = c_int # enum

# values for enumeration 'Nag_ComplexFormType'
Nag_ComplexFormType = c_int # enum

# values for enumeration 'Nag_HashError'
Nag_HashError = c_int # enum

# values for enumeration 'Nag_PredictError'
Nag_PredictError = c_int # enum

# values for enumeration 'Nag_EstimatesOption'
Nag_EstimatesOption = c_int # enum

# values for enumeration 'Nag_ScalePredictor'
Nag_ScalePredictor = c_int # enum

# values for enumeration 'Nag_ParaOption'
Nag_ParaOption = c_int # enum

# values for enumeration 'Nag_OptionLOO'
Nag_OptionLOO = c_int # enum

# values for enumeration 'Nag_VIFOption'
Nag_VIFOption = c_int # enum

# values for enumeration 'Nag_HessianType'
Nag_HessianType = c_int # enum

# values for enumeration 'Nag_DirectionType'
Nag_DirectionType = c_int # enum

# values for enumeration 'Nag_TS_URTestType'
Nag_TS_URTestType = c_int # enum

# values for enumeration 'Nag_TS_URProbMethod'
Nag_TS_URProbMethod = c_int # enum

# values for enumeration 'Nag_TS_ChangeType'
Nag_TS_ChangeType = c_int # enum

# values for enumeration 'Nag_TS_SegSide'
Nag_TS_SegSide = c_int # enum

# values for enumeration 'Nag_CondErrType'
Nag_CondErrType = c_int # enum

# values for enumeration 'Nag_SignsType'
Nag_SignsType = c_int # enum
class Nag_EnumTableType(Structure):
    pass
Nag_EnumTableType._fields_ = [
    ('enum_string', STRING),
    ('enum_value', c_int),
]
class Nag_ErrorNameTable(Structure):
    pass
Nag_ErrorNameTable._fields_ = [
    ('error_name', STRING),
    ('error_value', c_int),
]
NAG_ERRHAN = NAG_FUNCTYPE(None, STRING, c_int, STRING)
class NagError(Structure):
    pass
NagError._fields_ = [
    ('code', c_int),
    ('eprint', Nag_Boolean),
    ('message', c_char * 512),
    ('handler', NAG_ERRHAN),
    ('errnum', Integer),
    ('iflag', Integer),
    ('ival', Integer),
]
class Nag_User(Structure):
    pass
Nag_User._fields_ = [
    ('p', Pointer),
]
class comm_struct(Structure):
    pass
comm_struct._fields_ = [
    ('flag', Integer),
    ('first', Nag_Boolean),
    ('last', Nag_Boolean),
    ('nf', Integer),
    ('it_prt', Nag_Boolean),
    ('it_maj_prt', Nag_Boolean),
    ('sol_sqp_prt', Nag_Boolean),
    ('sol_prt', Nag_Boolean),
    ('rootnode_sol_prt', Nag_Boolean),
    ('node_prt', Nag_Boolean),
    ('rootnode_prt', Nag_Boolean),
    ('g_prt', Nag_Boolean),
    ('new_lm', Nag_Boolean),
    ('needf', Integer),
    ('p', Pointer),
    ('iuser', POINTER(Integer)),
    ('user', POINTER(c_double)),
    ('nag_w', POINTER(c_double)),
    ('nag_iw', POINTER(Integer)),
    ('nag_p', Pointer),
    ('nag_print_w', POINTER(c_double)),
    ('nag_print_iw', POINTER(Integer)),
    ('nrealloc', Integer),
]
Nag_Comm = comm_struct
class User_Work_t(Structure):
    pass
User_Work_t._fields_ = [
    ('nc', POINTER(Nag_Comm)),
]
class Nag_FileSt(Structure):
    pass
Nag_FileSt._fields_ = [
    ('set', Nag_Boolean),
    ('st_out', Nag_Boolean),
    ('st_err', Nag_Boolean),
    ('st_in', Nag_Boolean),
    ('name', c_char * 512),
    ('chapter', STRING),
    ('error', c_int),
]
class Nag_Mesg(Structure):
    pass
Nag_Mesg._fields_ = [
    ('init_mesg1', Integer),
    ('init_mesg2', Integer),
    ('res_file', STRING),
    ('eprint', Nag_Boolean),
    ('code', c_int),
    ('name', STRING),
    ('opt_name', STRING),
    ('print_mesg', NAG_FUNCTYPE(None, STRING, c_int, STRING)),
]
class Nag_QuadProgress(Structure):
    pass
Nag_QuadProgress._fields_ = [
    ('num_subint', Integer),
    ('fun_count', Integer),
    ('sub_int_beg_pts', POINTER(c_double)),
    ('sub_int_end_pts', POINTER(c_double)),
    ('sub_int_result', POINTER(c_double)),
    ('sub_int_error', POINTER(c_double)),
]
class Nag_QuadSubProgress(Structure):
    pass
Nag_QuadSubProgress._fields_ = [
    ('intervals', Integer),
    ('fun_count', Integer),
    ('subints_per_interval', Integer),
    ('interval_error', POINTER(c_double)),
    ('interval_result', POINTER(c_double)),
    ('interval_flag', POINTER(Integer)),
]
class Nag_ad02(Structure):
    pass
Nag_ad02._fields_ = [
    ('alpha', c_double * 12),
    ('beta', c_double * 12),
    ('psi', c_double * 12),
    ('v', c_double * 12),
    ('w', c_double * 12),
    ('sig', c_double * 13),
    ('g', c_double * 13),
    ('gi', c_double * 11),
    ('xold', c_double),
    ('hold', c_double),
    ('told', c_double),
    ('xsave', c_double),
    ('tstar', c_double),
    ('twou', c_double),
    ('init', Integer),
    ('ibegin', Integer),
    ('itol', Integer),
    ('iinteg', Integer),
    ('itstop', Integer),
    ('inflop', Integer),
    ('iquit', Integer),
    ('iv', Integer * 10),
    ('ns', Integer),
    ('kord', Integer),
    ('kold', Integer),
    ('ksteps', Integer),
    ('kle4', Integer),
    ('kprev', Integer),
    ('ivc', Integer),
    ('kgi', Integer),
    ('start', Nag_Boolean),
    ('phase1', Nag_Boolean),
    ('nornd', Nag_Boolean),
    ('stiff', Nag_Boolean),
    ('intout', Nag_Boolean),
]
class Nag_bd02(Structure):
    pass
Nag_bd02._fields_ = [
    ('zero', c_double),
    ('u', c_double),
    ('fouru', c_double),
    ('sru', c_double),
    ('u34', c_double),
    ('u78', c_double),
    ('srbig', c_double),
    ('delsgn', c_double),
    ('troots', c_double),
    ('tleft', c_double),
    ('svtnew', c_double),
    ('krootp', Integer),
    ('inrotp', Integer),
    ('gstop', Nag_Boolean),
    ('pgstop', Nag_Boolean),
    ('root', Nag_Boolean),
    ('roots', Nag_Boolean),
    ('needg', Nag_Boolean),
    ('discop', Nag_Boolean),
    ('newgeq', Nag_Boolean),
    ('search', Nag_Boolean),
    ('pserch', Nag_Boolean),
]
class Nag_ad02ra(Structure):
    pass
Nag_ad02ra._fields_ = [
    ('nfun', Integer),
    ('nimp', Integer),
    ('repeat', Nag_Boolean),
    ('l1', Integer),
    ('l2', Integer),
    ('ng', Integer),
    ('n1', Integer),
    ('nc', Integer),
]
class Nag_dd02(Structure):
    pass
Nag_dd02._fields_ = [
    ('ts', c_double),
    ('gts', c_double),
    ('fd', c_double),
    ('ftd', c_double),
    ('fm', c_double),
    ('fmx', c_double),
    ('prev', c_double),
    ('min', c_double),
    ('max', c_double),
    ('avg', c_double),
    ('amb', c_double),
    ('ambs', c_double),
    ('sbma', c_double),
    ('sf', c_double),
    ('dezu34', c_double),
    ('ic', Integer),
    ('kmc', Integer),
    ('ising', Integer),
    ('kount', Integer),
    ('ktest', Integer),
    ('locext', Nag_Boolean),
]
class Nag_ed02(Structure):
    pass
Nag_ed02._fields_ = [
    ('aez', c_double),
    ('agtrl', c_double),
    ('at', c_double),
    ('atm', c_double),
    ('atmtr', c_double),
    ('brd', c_double),
    ('brds', c_double),
    ('ddt', c_double),
    ('deltg', c_double),
    ('deltgm', c_double),
    ('deltr', c_double),
    ('dg', c_double),
    ('dtg', c_double),
    ('dtgk', c_double),
    ('dtgp', c_double),
    ('dtqr', c_double),
    ('fac', c_double),
    ('g21', c_double),
    ('g32', c_double),
    ('gat', c_double),
    ('gbr', c_double),
    ('gbrc', c_double),
    ('gbrmin', c_double),
    ('gbrmns', c_double),
    ('gct', c_double),
    ('glocmx', c_double),
    ('gmsign', c_double),
    ('goldsv', c_double),
    ('got', c_double),
    ('gres', c_double),
    ('gright', c_double),
    ('gt21', c_double),
    ('gt32', c_double),
    ('gtint', c_double),
    ('gtqr', c_double),
    ('gtrl', c_double),
    ('gtt', c_double),
    ('gttmr', c_double),
    ('gvalue', c_double),
    ('gvdif', c_double),
    ('gvmax', c_double),
    ('gvmin', c_double),
    ('omu78', c_double),
    ('opu78', c_double),
    ('pu', c_double),
    ('qc0', c_double),
    ('qc1', c_double),
    ('qc2', c_double),
    ('qd', c_double),
    ('qdq', c_double),
    ('qr1', c_double),
    ('qr1n', c_double),
    ('qr1p', c_double),
    ('quadmn', c_double),
    ('rd', c_double),
    ('rdg', c_double),
    ('rem', c_double),
    ('remd', c_double),
    ('rez', c_double),
    ('rootno', c_double),
    ('sdg', c_double),
    ('signg', c_double),
    ('slope1', c_double),
    ('slope2', c_double),
    ('slopep', c_double),
    ('sotgv', c_double),
    ('ss', c_double),
    ('sstu78', c_double),
    ('t21', c_double),
    ('t31', c_double),
    ('t32', c_double),
    ('taddon', c_double),
    ('tcb', c_double),
    ('tcf', c_double),
    ('tclose', c_double),
    ('tdif', c_double),
    ('tdmn', c_double),
    ('tdmx', c_double),
    ('tfar', c_double),
    ('tgbrmn', c_double),
    ('tgvdif', c_double),
    ('tgvmt', c_double),
    ('tint', c_double),
    ('tintp', c_double),
    ('tk', c_double),
    ('tkold', c_double),
    ('tlbk', c_double),
    ('tnew', c_double),
    ('tp', c_double),
    ('tpoint', c_double),
    ('tprev', c_double),
    ('tqr', c_double),
    ('tqr1', c_double),
    ('tqr2', c_double),
    ('tqrf', c_double),
    ('tqrmp', c_double),
    ('tqrp', c_double),
    ('tqs', c_double),
    ('tqt', c_double),
    ('trd', c_double),
    ('trj', c_double),
    ('trl', c_double),
    ('trn', c_double),
    ('trnow', c_double),
    ('troot', c_double),
    ('tsave', c_double),
    ('tslop', c_double),
    ('tsru', c_double),
    ('tstep', c_double),
    ('tt', c_double),
    ('tu78', c_double),
    ('tut', c_double),
    ('tz', c_double),
    ('tzero', c_double),
    ('tzerok', c_double),
    ('tzj', c_double),
    ('tzk', c_double),
    ('grno', c_double),
    ('gt', c_double),
    ('sttu78', c_double),
    ('tadd', c_double * 2),
    ('tr', c_double * 3),
]
class Nag_fd02(Structure):
    pass
Nag_fd02._fields_ = [
    ('backr', Nag_Boolean),
    ('mroot', Nag_Boolean),
    ('multr', Nag_Boolean),
    ('peak', Nag_Boolean),
    ('qrreal', Nag_Boolean),
    ('skpmmt', Nag_Boolean),
]
class Nag_gd02(Structure):
    pass
Nag_gd02._fields_ = [
    ('i', Integer),
    ('i1', Integer),
    ('i2', Integer),
    ('i21', Integer),
    ('i3', Integer),
    ('iadd', Integer),
    ('ic1', Integer),
    ('ic2', Integer),
    ('icase', Integer),
    ('icbr', Integer),
    ('icbrk', Integer),
    ('idtg', Integer),
    ('iga', Integer),
    ('ijk', Integer),
    ('ik', Integer),
    ('ikmr', Integer),
    ('iktqr', Integer),
    ('indx', Integer),
    ('isi', Integer),
    ('isin', Integer),
    ('isr', Integer),
    ('itgvt', Integer),
    ('itry', Integer),
    ('j', Integer),
    ('j1', Integer),
    ('j2', Integer),
    ('j23', Integer),
    ('j3', Integer),
    ('jp1', Integer),
    ('jr', Integer),
    ('jsr', Integer),
    ('k', Integer),
    ('kge', Integer),
    ('kp1', Integer),
    ('krootn', Integer),
    ('krooto', Integer),
    ('kroots', Integer),
    ('kskpmt', Integer),
    ('ktry', Integer),
    ('l', Integer),
    ('ltqbig', Integer),
    ('ltr', Integer),
    ('mmr', Integer),
    ('mmrin', Integer),
    ('mmrk', Integer),
    ('nsr', Integer),
    ('ntin', Integer),
    ('ntinm', Integer),
    ('ipath', c_int),
]
class Nag_jd02(Structure):
    pass
Nag_jd02._fields_ = [
    ('a', c_double),
    ('absdx', c_double),
    ('da', c_double),
    ('delx', c_double),
    ('dfdub', c_double),
    ('dfdxb', c_double),
    ('dx', c_double),
    ('fbnd', c_double),
    ('power', c_double),
    ('relper', c_double),
    ('ynorm', c_double),
    ('ypnorm', c_double),
    ('icase', Integer),
    ('k', Integer),
    ('lk', Integer),
]
class Nag_kd02(Structure):
    pass
Nag_kd02._fields_ = [
    ('two', c_double * 13),
    ('gstr', c_double * 13),
    ('absh', c_double),
    ('big', c_double),
    ('erk', c_double),
    ('erkm1', c_double),
    ('p5eps', c_double),
    ('round', c_double),
    ('u', c_double),
    ('ifail', Integer),
    ('km1', Integer),
    ('km2', Integer),
    ('knew', Integer),
    ('kp1', Integer),
    ('kp2', Integer),
]
class Nag_ODE_Adams(Structure):
    pass
Nag_ODE_Adams._fields_ = [
    ('init1', Integer),
    ('init2', Integer),
    ('neqf', Integer),
    ('neqg', Integer),
    ('state', Nag_Start),
    ('alter_g', Nag_Boolean),
    ('sophist', Nag_Boolean),
    ('vectol', Nag_Boolean),
    ('a_tol', POINTER(c_double)),
    ('rtol', POINTER(c_double)),
    ('crit', Nag_Boolean),
    ('tcrit', c_double),
    ('one_step', Nag_Boolean),
    ('hmax', c_double),
    ('max_step', Integer),
    ('root', Nag_Boolean),
    ('badcmp', Integer),
    ('hnext', c_double),
    ('hlast', c_double),
    ('tolfac', c_double),
    ('tcurr', c_double),
    ('nsuccess', Integer),
    ('nfail', Integer),
    ('ord_last', Integer),
    ('ord_next', Integer),
    ('events', POINTER(Integer)),
    ('resids', POINTER(c_double)),
    ('yp', POINTER(c_double)),
    ('yy', POINTER(c_double)),
    ('p', POINTER(c_double)),
    ('phi', POINTER(c_double)),
    ('qwc_state', Integer),
    ('qfc_state', Integer),
    ('giwork', POINTER(Integer)),
    ('siwork', POINTER(Integer)),
    ('w', POINTER(c_double)),
    ('gwork', POINTER(c_double)),
    ('swork', POINTER(c_double)),
    ('free', NAG_FUNCTYPE(None, POINTER(Nag_ODE_Adams))),
    ('index', Integer),
    ('type', Integer),
    ('ad02qf', Nag_ad02),
    ('bd02qf', Nag_bd02),
    ('crash', Nag_Boolean),
    ('izflag', Integer),
    ('dd02qf', Nag_dd02),
    ('ed02qf', Nag_ed02),
    ('fd02qf', Nag_fd02),
    ('gd02qf', Nag_gd02),
    ('kroo', Integer),
    ('jd02qf', Nag_jd02),
    ('kd02qf', Nag_kd02),
]
class Nag_ad02nm(Structure):
    pass
Nag_ad02nm._fields_ = [
    ('itrace', Integer),
    ('idev', Integer),
]
class Nag_bd02nm(Structure):
    pass
Nag_bd02nm._fields_ = [
    ('hold', c_double),
    ('nq', Integer),
    ('nqu', Integer),
    ('nst', Integer),
    ('nre', Integer),
    ('nje', Integer),
    ('niter', Integer),
    ('ninter', Integer),
    ('kcur', Integer),
]
class Nag_cd02nm(Structure):
    pass
Nag_cd02nm._fields_ = [
    ('hmxstt', c_double),
]
class Nag_dd02nm(Structure):
    pass
Nag_dd02nm._fields_ = [
    ('h', c_double),
    ('el0', c_double),
    ('tn', c_double),
    ('hu', c_double),
    ('hmin', c_double),
    ('hmxi', c_double),
    ('dsqu', c_double),
    ('els', c_double),
    ('wb', c_double),
    ('elc', c_double),
    ('ts', c_double),
    ('init', Integer),
    ('lewt', Integer),
    ('lacor', Integer),
    ('lsavr', Integer),
    ('ldae', Integer),
    ('mxstep', Integer),
    ('mxhnil', Integer),
    ('nhnil', Integer),
    ('nslast', Integer),
]
class Nag_ed02nm(Structure):
    pass
Nag_ed02nm._fields_ = [
    ('ifunc', Integer),
    ('istep', Integer),
    ('ier', Integer),
    ('iewset', Integer),
    ('iretur', Integer),
    ('idacnt', Integer),
    ('idaold', Integer),
    ('initl', Integer),
    ('iode', Integer),
    ('icrash', Integer),
    ('ifn', Integer),
    ('idachk', Integer),
]
class Nag_fd02nm(Structure):
    pass
Nag_fd02nm._fields_ = [
    ('dunflo', c_double),
    ('uround', c_double),
    ('iovflo', Integer),
]
class Nag_gd02nm(Structure):
    pass
Nag_gd02nm._fields_ = [
    ('damp', c_double),
    ('rjnorm', c_double),
    ('crate', c_double),
    ('maxit', Integer),
]
class Nag_hd02nm(Structure):
    pass
Nag_hd02nm._fields_ = [
    ('inorm', Integer),
]
class Nag_jd02nm(Structure):
    pass
Nag_jd02nm._fields_ = [
    ('big', c_double),
    ('qtymin', c_double),
    ('rootn', c_double),
]
class Nag_kd02nm(Structure):
    pass
Nag_kd02nm._fields_ = [
    ('ihit', Nag_Boolean),
]
class Nag_ld02nm(Structure):
    pass
Nag_ld02nm._fields_ = [
    ('a_toli', c_double),
    ('big', c_double),
    ('ewti', c_double),
    ('h0', c_double),
    ('rh', c_double),
    ('rtoli', c_double),
    ('size', c_double),
    ('tcrit', c_double),
    ('tem', c_double),
    ('hnext', c_double),
    ('tolsf', c_double),
    ('tp', c_double),
    ('hmax', c_double),
    ('i', Integer),
    ('ifj', Integer),
    ('iflag', Integer),
    ('ichmax', Integer),
    ('imxer', Integer),
    ('j', Integer),
    ('kgo', Integer),
    ('lenrw', Integer),
    ('mxhnl0', Integer),
    ('mxstp0', Integer),
    ('n', Integer),
    ('isave', Integer),
    ('nold', Integer),
    ('crit1', Nag_Boolean),
    ('crit2', Nag_Boolean),
    ('nt1stp', Nag_Boolean),
]
class Nag_md02nm(Structure):
    pass
Nag_md02nm._fields_ = [
    ('singlr', c_char * 7),
]
class Nag_nd02nm(Structure):
    pass
Nag_nd02nm._fields_ = [
    ('irow', Integer),
    ('icol', Integer),
]
class Nag_rd02nm(Structure):
    pass
Nag_rd02nm._fields_ = [
    ('con', c_double),
    ('di', c_double),
    ('fac', c_double),
    ('hl0', c_double),
    ('r', c_double),
    ('r0', c_double),
    ('srur', c_double),
    ('yi', c_double),
    ('yj', c_double),
    ('yjj', c_double),
    ('i', Integer),
    ('ier', Integer),
    ('ii', Integer),
    ('j', Integer),
    ('j1', Integer),
    ('jj', Integer),
    ('n', Integer),
]
class Nag_sd02nm(Structure):
    pass
Nag_sd02nm._fields_ = [
    ('iwp', Integer),
]
class Nag_yd02nm(Structure):
    pass
Nag_yd02nm._fields_ = [
    ('rdm10', c_double * 7),
    ('bdel', c_double),
    ('rdm11', c_double),
    ('bel1h', c_double),
    ('rdm12', c_double),
    ('bdcon', c_double),
    ('rdm13', c_double * 34),
    ('idm11', Integer * 17),
]
class Nag_zd02nm(Structure):
    pass
Nag_zd02nm._fields_ = [
    ('odcode', c_char * 7),
]
class Nag_nd02nn(Structure):
    pass
Nag_nd02nn._fields_ = [
    ('const1', c_double),
    ('const2', c_double),
    ('const3', c_double),
    ('const4', c_double),
    ('const5', c_double),
    ('const6', c_double),
    ('start', Nag_Boolean),
    ('speclh', Nag_Boolean),
]
class Nag_pd02nn(Structure):
    pass
Nag_pd02nn._fields_ = [
    ('nzeros', Integer),
    ('nonzer', Nag_Boolean),
    ('somzer', Nag_Boolean),
    ('allzer', Nag_Boolean),
]
class Nag_rd02nn(Structure):
    pass
Nag_rd02nn._fields_ = [
    ('hprop', c_double),
    ('change', Nag_Boolean),
]
class Nag_sd02nn(Structure):
    pass
Nag_sd02nn._fields_ = [
    ('hold', c_double),
]
class Nag_td02nn(Structure):
    pass
Nag_td02nn._fields_ = [
    ('errcwe', c_double),
]
class Nag_ud02nn(Structure):
    pass
Nag_ud02nn._fields_ = [
    ('ddn', c_double),
    ('dsm', c_double),
    ('dup', c_double),
    ('eljh', c_double),
    ('terk', c_double),
    ('terkm1', c_double),
    ('terkp1', c_double),
    ('exdn', c_double),
    ('exsm', c_double),
    ('exup', c_double),
    ('r', c_double),
    ('rh', c_double),
    ('rhdn', c_double),
    ('rhsm', c_double),
    ('rhup', c_double),
    ('told', c_double),
    ('terkm2', c_double),
    ('ddn1', c_double),
    ('i', Integer),
    ('jc', Integer),
    ('iredo', Integer),
    ('iret', Integer),
    ('j', Integer),
    ('jb', Integer),
    ('kgo', Integer),
    ('ncf', Integer),
    ('iz', Integer),
    ('jstart', Integer),
    ('kflag', Integer),
    ('raise', Nag_Boolean),
]
class Nag_yd02nn(Structure):
    pass
Nag_yd02nn._fields_ = [
    ('dcon', c_double),
    ('del', c_double),
    ('delp', c_double),
    ('d1', c_double),
    ('d2', c_double),
    ('el1h', c_double),
    ('maxcor', Integer),
    ('isave', Integer),
    ('m', Integer),
    ('iz', Integer),
    ('i', Integer),
]
class Nag_ad02xk(Structure):
    pass
Nag_ad02xk._fields_ = [
    ('rownd', c_double),
    ('conit', c_double),
    ('el', c_double * 13),
    ('elco', c_double * 156),
    ('rmax', c_double),
    ('tesco', c_double * 36),
    ('ccmax', c_double),
    ('rc', c_double),
    ('ialth', Integer),
    ('ipup', Integer),
    ('lmax', Integer),
    ('meo', Integer),
    ('nslp', Integer),
    ('icf', Integer),
    ('jcur', Integer),
    ('l', Integer),
    ('meth', Integer),
    ('miter', Integer),
    ('maxord', Integer),
    ('msbp', Integer),
    ('mxncf', Integer),
    ('n', Integer),
]
class Nag_bd02xk(Structure):
    pass
Nag_bd02xk._fields_ = [
    ('hhused', c_double),
    ('petzld', Nag_Boolean),
    ('fniter', Nag_Boolean),
]
class Nag_cd02ej(Structure):
    pass
Nag_cd02ej._fields_ = [
    ('jceval', c_char * 7),
    ('outopt', c_char * 7),
    ('gopt', c_char * 7),
]
class Nag_ODE_BDF(Structure):
    pass
Nag_ODE_BDF._fields_ = [
    ('ad02nm', Nag_ad02nm),
    ('bd02nm', Nag_bd02nm),
    ('cd02nm', Nag_cd02nm),
    ('dd02nm', Nag_dd02nm),
    ('ed02nm', Nag_ed02nm),
    ('fd02nm', Nag_fd02nm),
    ('gd02nm', Nag_gd02nm),
    ('hd02nm', Nag_hd02nm),
    ('jd02nm', Nag_jd02nm),
    ('kd02nm', Nag_kd02nm),
    ('ld02nm', Nag_ld02nm),
    ('md02nm', Nag_md02nm),
    ('nd02nm', Nag_nd02nm),
    ('rd02nm', Nag_rd02nm),
    ('sd02nm', Nag_sd02nm),
    ('yd02nm', Nag_yd02nm),
    ('zd02nm', Nag_zd02nm),
    ('nd02nn', Nag_nd02nn),
    ('pd02nn', Nag_pd02nn),
    ('rd02nn', Nag_rd02nn),
    ('sd02nn', Nag_sd02nn),
    ('td02nn', Nag_td02nn),
    ('ud02nn', Nag_ud02nn),
    ('yd02nn', Nag_yd02nn),
    ('ad02xk', Nag_ad02xk),
    ('bd02xk', Nag_bd02xk),
    ('cd02ej', Nag_cd02ej),
    ('icall', Integer),
]
class Nag_D03_Save(Structure):
    pass
Nag_D03_Save._fields_ = [
    ('cwsav', c_char * 810),
    ('rwsav', c_double * 1100),
    ('iwsav', Integer * 505),
    ('lwsav', Integer * 100),
]
class Nag_Scat_Struct(Structure):
    pass
Nag_Scat_Struct._fields_ = [
    ('init1', Integer),
    ('init2', Integer),
    ('exit_status', Integer),
    ('method', Nag_2d_Scat_Method),
    ('m', Integer),
    ('rnw', c_double),
    ('mem_set', Nag_Boolean),
    ('x', POINTER(c_double)),
    ('y', POINTER(c_double)),
    ('f', POINTER(c_double)),
    ('fnodes', POINTER(c_double)),
    ('triang', POINTER(Integer)),
    ('grads', POINTER(c_double)),
]
class Nag_E01_Opt(Structure):
    pass
Nag_E01_Opt._fields_ = [
    ('rnw', c_double),
    ('rnq', c_double),
    ('nw', Integer),
    ('nq', Integer),
    ('minnq', Integer),
]
class Nag_Spline(Structure):
    pass
Nag_Spline._fields_ = [
    ('n', Integer),
    ('lamda', POINTER(c_double)),
    ('c', POINTER(c_double)),
    ('init1', Integer),
    ('init2', Integer),
]
class Nag_2dSpline(Structure):
    pass
Nag_2dSpline._fields_ = [
    ('nx', Integer),
    ('lamda', POINTER(c_double)),
    ('ny', Integer),
    ('mu', POINTER(c_double)),
    ('c', POINTER(c_double)),
    ('init1', Integer),
    ('init2', Integer),
]
class Nag_ae04mf(Structure):
    pass
Nag_ae04mf._fields_ = [
    ('loclc', Integer * 20),
    ('litotl', Integer),
    ('lwtotl', Integer),
    ('lennam', Integer),
    ('ldt', Integer),
    ('ncolt', Integer),
    ('ldq', Integer),
]
class Nag_e04nfu(Structure):
    pass
Nag_e04nfu._fields_ = [
    ('prob', Nag_ProblemType),
    ('m', Integer),
]
class Nag_de04nb(Structure):
    pass
Nag_de04nb._fields_ = [
    ('asize', c_double),
    ('dtmax', c_double),
    ('dtmin', c_double),
]
class Nag_ce04mf(Structure):
    pass
Nag_ce04mf._fields_ = [
    ('tolx0', c_double),
    ('tolinc', c_double),
    ('idegen', Integer),
    ('kdegen', Integer),
    ('ndegen', Integer),
    ('itnfix', Integer),
    ('nfix', Integer * 2),
]
class Nag_de04mf(Structure):
    pass
Nag_de04mf._fields_ = [
    ('alfa', c_double),
    ('trulam', c_double),
    ('isdel', Integer),
    ('jdel', Integer),
    ('jadd', Integer),
    ('header', Nag_Boolean),
    ('prnt', Nag_Boolean),
]
class Nag_ee04mf(Structure):
    pass
Nag_ee04mf._fields_ = [
    ('idbg', Integer),
    ('lcdbg', Nag_Boolean),
    ('ilcdbg', Integer * 5),
    ('cmdbg', Nag_Boolean),
    ('icmdbg', Integer * 5),
]
class Nag_XazSt(Structure):
    pass
Nag_XazSt._fields_ = [
    ('cdsave', c_double),
    ('fdsave', c_double),
    ('hsave', c_double),
    ('oldh', c_double),
    ('rho', c_double),
    ('sdsave', c_double),
    ('ce1big', c_double),
    ('ce2big', c_double),
    ('te2big', c_double),
]
class Nag_UcjkSt(Structure):
    pass
Nag_UcjkSt._fields_ = [
    ('braktd', Nag_Boolean),
    ('crampd', Nag_Boolean),
    ('extrap', Nag_Boolean),
    ('moved', Nag_Boolean),
    ('wset', Nag_Boolean),
    ('vset', Nag_Boolean),
    ('nsamea', Integer),
    ('nsameb', Integer),
    ('a', c_double),
    ('b', c_double),
    ('xtry', c_double),
    ('factor', c_double),
    ('xw', c_double),
    ('fw', c_double),
    ('tolmax', c_double),
    ('gw', c_double),
    ('fa', c_double),
    ('xv', c_double),
    ('fv', c_double),
]
class Nag_Uc01St(Structure):
    pass
Nag_Uc01St._fields_ = [
    ('braktd', Nag_Boolean),
    ('crampd', Nag_Boolean),
    ('extrap', Nag_Boolean),
    ('moved', Nag_Boolean),
    ('wset', Nag_Boolean),
    ('vset', Nag_Boolean),
    ('nfsrch', Integer),
    ('nsamea', Integer),
    ('nsameb', Integer),
    ('a', c_double),
    ('b', c_double),
    ('alfuzz', c_double),
    ('factor', c_double),
    ('xw', c_double),
    ('fw', c_double),
    ('tolmax', c_double),
    ('rtmin', c_double),
    ('gw', c_double),
    ('fa', c_double),
    ('xv', c_double),
    ('fv', c_double),
]
class Nag_Deriv_Inf(Structure):
    pass
Nag_Deriv_Inf._fields_ = [
    ('lfdset', Integer),
    ('ncdiff', Integer),
    ('nfdiff', Integer),
    ('lvldif', Integer),
]
class Nag_DebugSt(Structure):
    pass
Nag_DebugSt._fields_ = [
    ('inpdbg', Integer * 5),
    ('eprint', Nag_Boolean),
]
class Nag_Grad_Chk_St(Structure):
    pass
Nag_Grad_Chk_St._fields_ = [
    ('type', Nag_GradChk),
    ('obj_start', Integer),
    ('obj_stop', Integer),
    ('con_start', Integer),
    ('con_stop', Integer),
    ('g_error', c_int),
]
class Nag_SimSt(Structure):
    pass
Nag_SimSt._fields_ = [
    ('n_elements', Integer),
    ('dir_deriv', c_double),
    ('fd_approx', c_double),
    ('correct', Nag_Boolean),
    ('max_error', c_double),
    ('max_subfunction', Integer),
    ('max_constraint', Integer),
]
class Nag_CompSt(Structure):
    pass
Nag_CompSt._fields_ = [
    ('assigned', Nag_Boolean),
    ('hopt', c_double),
    ('gdiff', c_double),
    ('iter', Integer),
    ('correct', Nag_Boolean),
    ('comment', STRING),
    ('zero', Nag_Boolean),
    ('error', c_double),
]
class Nag_FDIntSt(Structure):
    pass
Nag_FDIntSt._fields_ = [
    ('hfd', c_double),
    ('hcd', c_double),
    ('errmax', c_double),
]
class Nag_GPrintSt(Structure):
    pass
Nag_GPrintSt._fields_ = [
    ('g_chk', Nag_Grad_Chk_St),
    ('nfdiff', Integer),
    ('ncdiff', Integer),
    ('nf_const', Integer),
    ('nc_const', Integer),
    ('deriv_level', Nag_DerivSet),
    ('old_deriv_level', Nag_DerivSet),
    ('f_sim', Nag_SimSt),
    ('c_sim', Nag_SimSt),
    ('f_comp', POINTER(Nag_CompSt)),
    ('c_comp', POINTER(Nag_CompSt)),
    ('intervals', POINTER(Nag_FDIntSt)),
    ('print_deriv', Nag_DPrintType),
]
class Nag_QP_Print(Structure):
    pass
Nag_QP_Print._fields_ = [
    ('refactor', Nag_Boolean),
    ('jmax', Integer),
    ('errmax', c_double),
    ('moved', Nag_Boolean),
    ('nmoved', Integer),
    ('feasible', Nag_Boolean),
    ('rowerr', Nag_Boolean),
    ('first', Nag_Boolean),
    ('header', Nag_Boolean),
    ('prnt', Nag_Boolean),
]
class Nag_Search_State(Structure):
    pass
Nag_Search_State._fields_ = [
    ('m', Integer),
    ('n', Integer),
    ('numinf', Integer),
    ('x', POINTER(c_double)),
    ('f', c_double),
    ('objf', c_double),
    ('g', POINTER(c_double)),
    ('diagt', POINTER(c_double)),
    ('fvec', POINTER(c_double)),
    ('fjac', POINTER(c_double)),
    ('conjac', POINTER(c_double)),
    ('tdj', Integer),
    ('tdconjac', Integer),
    ('step', c_double),
    ('xk_norm', c_double),
    ('gpj_norm', c_double),
    ('cond', c_double),
    ('cond_h', c_double),
    ('cond_hz', c_double),
    ('cond_t', c_double),
    ('cond_r', c_double),
    ('iter_conv', Nag_Boolean),
    ('norm_gz_small', Nag_Boolean),
    ('violtn_small', Nag_Boolean),
    ('update_modified', Nag_Boolean),
    ('c_diff', Nag_Boolean),
    ('qp_not_feasible', Nag_Boolean),
    ('step_limit_exceeded', Nag_Boolean),
    ('refactor', Nag_Boolean),
    ('mgrmsg', STRING),
    ('norm_nlnviol', c_double),
    ('grade', Integer),
    ('s', POINTER(c_double)),
    ('state', POINTER(Integer)),
    ('iter', Integer),
    ('major_iter', Integer),
    ('minor_iter', Integer),
    ('nfun', Integer),
    ('merit', c_double),
    ('violtn', c_double),
    ('penalty', c_double),
    ('itn', Integer),
    ('nf', Integer),
    ('nctotal', Integer),
    ('lncons_val', POINTER(c_double)),
    ('major_iter_soln', Nag_Boolean),
    ('local_search', Nag_Boolean),
    ('nclin', Integer),
    ('ncnlin', Integer),
    ('nactiv', Integer),
    ('jdel', Integer),
    ('jadd', Integer),
    ('ninf', Integer),
    ('sum_infeasibilities', c_double),
    ('bnd', Integer),
    ('lin', Integer),
    ('nln', Integer),
    ('t', POINTER(c_double)),
    ('tdt', Integer),
    ('r', POINTER(c_double)),
    ('diagr', POINTER(c_double)),
    ('tdr', Integer),
    ('nart', Integer),
    ('nartm', Integer),
    ('nrz', Integer),
    ('nz', Integer),
    ('norm_gz', c_double),
    ('norm_gf', c_double),
    ('nopt', Integer),
    ('min_lm', c_double),
    ('condt', c_double),
    ('condr', c_double),
    ('rzz', c_double),
    ('rset', Nag_Boolean),
    ('ax', POINTER(c_double)),
    ('cx', POINTER(c_double)),
    ('nrank', Integer),
    ('nfixed', Integer),
    ('kx', POINTER(Integer)),
    ('kactive', POINTER(Integer)),
    ('lambda', POINTER(c_double)),
    ('bclambda', POINTER(c_double)),
    ('gq', POINTER(c_double)),
    ('jmax', Integer),
    ('errmax', c_double),
    ('moved', Nag_Boolean),
    ('nmoved', Integer),
    ('feasible', Nag_Boolean),
    ('rowerr', Nag_Boolean),
    ('first', Nag_Boolean),
    ('bl', POINTER(c_double)),
    ('bu', POINTER(c_double)),
    ('endstate', Nag_EndState),
    ('gprint', POINTER(Nag_GPrintSt)),
    ('fmin', c_double),
    ('fmax', c_double),
    ('simplex', POINTER(c_double)),
    ('prbtyp', STRING),
    ('tdfjac', Integer),
    ('col', Nag_Boolean),
    ('index', Integer),
    ('name', STRING),
    ('key', STRING),
    ('sstate', STRING),
    ('val', c_double),
    ('blo', c_double),
    ('bup', c_double),
    ('lmult', c_double),
    ('objg', c_double),
    ('just_feas', Nag_Boolean),
    ('qp', Nag_Boolean),
    ('pprice', Integer),
    ('rgval', c_double),
    ('sb_add', Integer),
    ('sb_leave', Integer),
    ('b_leave', Integer),
    ('bswap_leave', Integer),
    ('pivot', c_double),
    ('nnz_l', Integer),
    ('nnz_u', Integer),
    ('ncp', Integer),
    ('norm_rg', c_double),
    ('nsb', Integer),
    ('root', Nag_Boolean),
    ('nnodes', Integer),
    ('node_num', Integer),
    ('parent_node', Integer),
    ('crnames', POINTER(STRING)),
    ('node_status', Nag_NodeStatus),
    ('branch_index', Integer),
    ('x_before', c_double),
    ('x_after', c_double),
    ('x_lo', c_double),
    ('x_up', c_double),
    ('depth', Integer),
    ('posdef', Nag_Boolean),
]
class Nag_Opt_Found(Structure):
    pass
Nag_Opt_Found._fields_ = [
    ('options', Nag_Boolean),
    ('begin', Nag_Boolean),
    ('end', Nag_Boolean),
    ('eof', Nag_Boolean),
]
NAG_E04_PRINTFUN = NAG_FUNCTYPE(None, POINTER(Nag_Search_State), POINTER(Nag_Comm))
class Nag_E04_Opt(Structure):
    pass
Nag_E04_Opt._fields_ = [
    ('f_name', STRING),
    ('f_longname', STRING),
    ('init1', Integer),
    ('init2', Integer),
    ('minlin', Nag_LinFun),
    ('grade', Integer),
    ('nf', Integer),
    ('iter', Integer),
    ('major_iter', Integer),
    ('max_iter', Integer),
    ('minor_max_iter', Integer),
    ('min_infeas', Nag_Boolean),
    ('f_est', c_double),
    ('step_max', c_double),
    ('step_limit', c_double),
    ('max_line_step', c_double),
    ('f_prec', c_double),
    ('linesearch_tol', c_double),
    ('optim_tol', c_double),
    ('conf', POINTER(c_double)),
    ('nag_conf', c_long),
    ('conjac', POINTER(c_double)),
    ('nag_conjac', c_long),
    ('h', POINTER(c_double)),
    ('nag_h', c_long),
    ('h_unit_init', Nag_Boolean),
    ('h_reset_freq', Integer),
    ('s', POINTER(c_double)),
    ('nag_s', c_long),
    ('v', POINTER(c_double)),
    ('nag_v', c_long),
    ('tdv', Integer),
    ('delta', POINTER(c_double)),
    ('nag_delta', c_long),
    ('hesl', POINTER(c_double)),
    ('nag_hesl', c_long),
    ('hesd', POINTER(c_double)),
    ('nag_hesd', c_long),
    ('state', POINTER(Integer)),
    ('nag_state', c_long),
    ('init_state', Nag_InitType),
    ('obj_deriv', Nag_Boolean),
    ('con_deriv', Nag_Boolean),
    ('f_diff_int', c_double),
    ('c_diff_int', c_double),
    ('obj_check_start', Integer),
    ('obj_check_stop', Integer),
    ('con_check_start', Integer),
    ('con_check_stop', Integer),
    ('verify_grad', Nag_GradChk),
    ('hessian', Nag_Boolean),
    ('local_search', Nag_Boolean),
    ('deriv_check', Nag_Boolean),
    ('list', Nag_Boolean),
    ('print_gcheck', Nag_Boolean),
    ('print_deriv', Nag_DPrintType),
    ('print_level', Nag_PrintType),
    ('minor_print_level', Nag_PrintType),
    ('print_iter', Integer),
    ('outfile', c_char * 512),
    ('print_fun', NAG_E04_PRINTFUN),
    ('ax', POINTER(c_double)),
    ('nag_ax', c_long),
    ('lambda', POINTER(c_double)),
    ('nag_lambda', c_long),
    ('fcheck', Integer),
    ('crash_tol', c_double),
    ('reset_ftol', Integer),
    ('fmax_iter', Integer),
    ('ftol', c_double),
    ('lin_feas_tol', c_double),
    ('nonlin_feas_tol', c_double),
    ('hrows', Integer),
    ('inf_bound', c_double),
    ('inf_step', c_double),
    ('max_df', Integer),
    ('prob', Nag_ProblemType),
    ('rank_tol', c_double),
    ('form_hessian', Nag_Boolean),
    ('start', Nag_Start),
    ('debug_iter', Integer),
    ('minor_debug_iter', Integer),
    ('debug', Nag_Boolean),
    ('used', Nag_Boolean),
    ('unitq', Nag_Boolean),
    ('nfree', Integer),
    ('nactiv', Integer),
    ('nprob', Integer),
    ('n', Integer),
    ('nclin', Integer),
    ('ncnlin', Integer),
    ('bndlow', c_double),
    ('bndupp', c_double),
    ('minimize', Nag_Boolean),
    ('factor_freq', Integer),
    ('partial_price', Integer),
    ('max_sb', Integer),
    ('lu_factor_tol', c_double),
    ('lu_sing_tol', c_double),
    ('lu_update_tol', c_double),
    ('pivot_tol', c_double),
    ('scale_tol', c_double),
    ('crash', Nag_CrashType),
    ('scale', Nag_ScaleType),
    ('prob_name', c_char * 9),
    ('obj_name', c_char * 9),
    ('bnd_name', c_char * 9),
    ('rhs_name', c_char * 9),
    ('range_name', c_char * 9),
    ('crnames', POINTER(STRING)),
    ('nag_crnames', c_long),
    ('nsb', Integer),
    ('max_restart', Integer),
    ('max_basis_len', Integer),
    ('max_compress', Integer),
    ('max_basis_nfactor', Integer),
    ('col_lo_default', c_double),
    ('col_up_default', c_double),
    ('infinity', c_double),
    ('ncol_approx', Integer),
    ('nrow_approx', Integer),
    ('est_density', c_double),
    ('output_level', Nag_OutputType),
    ('deriv_want', Nag_DWantType),
    ('use_hfwd_init', Nag_Boolean),
    ('f_prec_used', c_double),
    ('print_80ch', Nag_Boolean),
    ('feas_exit', Nag_Boolean),
    ('hess_freq', Integer),
    ('elastic_wt', c_double),
    ('lu_den_tol', c_double),
    ('part_price', Integer),
    ('scale_opt', Integer),
    ('expand_freq', Integer),
    ('hess_update', Integer),
    ('iter_lim', Integer),
    ('major_opt_tol', c_double),
    ('minor_opt_tol', c_double),
    ('unbounded_obj', c_double),
    ('major_step_lim', c_double),
    ('major_iter_lim', Integer),
    ('major_feas_tol', c_double),
    ('minor_iter_lim', Integer),
    ('minor_feas_tol', c_double),
    ('violation_limit', c_double),
    ('nz_coef', c_double),
    ('deriv_linesearch', Nag_Boolean),
    ('hess_storage', Nag_HessianType),
    ('direction', Nag_DirectionType),
    ('ncon', Integer),
]
class Nag_SparseQP_Save_Vars(Structure):
    pass
Nag_SparseQP_Save_Vars._fields_ = [
    ('itnfix', Integer),
    ('nfix', Integer * 2),
    ('tolx0', c_double),
    ('nbfac', Integer),
    ('umin', c_double),
]
class Nag_ae04nc(Structure):
    pass
Nag_ae04nc._fields_ = [
    ('locls', Integer * 20),
]
class Nag_ae04uc(Structure):
    pass
Nag_ae04uc._fields_ = [
    ('locnp', Integer * 35),
]
class Nag_ae04up(Structure):
    pass
Nag_ae04up._fields_ = [
    ('locnl', Integer * 20),
]
class Nag_be04nb(Structure):
    pass
Nag_be04nb._fields_ = [
    ('tdt', Integer),
    ('ncolt', Integer),
    ('tdq', Integer),
]
class Nag_ce04nc(Structure):
    pass
Nag_ce04nc._fields_ = [
    ('ilsdbg', Integer * 5),
    ('lsdbg', Nag_Boolean),
]
class Nag_de04uc(Structure):
    pass
Nag_de04uc._fields_ = [
    ('rhomax', c_double),
    ('rhonrm', c_double),
    ('rhodmp', c_double),
    ('scale', c_double),
    ('incrun', Nag_Boolean),
]
class Nag_ee04nb(Structure):
    pass
Nag_ee04nb._fields_ = [
    ('rcndbd', c_double),
    ('rfrobn', c_double),
    ('drmax', c_double),
    ('drmin', c_double),
]
class Nag_fe04nb(Structure):
    pass
Nag_fe04nb._fields_ = [
    ('icmdbg', Integer * 5),
    ('cmdbg', Nag_Boolean),
]
class Nag_fe04nc(Structure):
    pass
Nag_fe04nc._fields_ = [
    ('nactiv', Integer),
    ('nfree', Integer),
    ('nz', Integer),
    ('unitq', Nag_Boolean),
]
class Nag_fe04uc(Structure):
    pass
Nag_fe04uc._fields_ = [
    ('inpdbg', Integer * 5),
    ('npdbg', Nag_Boolean),
]
class Nag_E04State(Structure):
    pass
Nag_E04State._fields_ = [
    ('cw', c_char * 4800),
    ('iw', Integer * 600),
    ('rw', c_double * 600),
]
class Nag_E04StateB(Structure):
    pass
Nag_E04StateB._fields_ = [
    ('CWSAV', c_char * 5600),
    ('IWSAV', Integer * 700),
    ('LWSAV', Integer * 700),
    ('RWSAV', c_double * 700),
]

# values for enumeration 'Nag_MCSInitMethod'
Nag_MCSInitMethod = c_int # enum
class Nag_E05State(Structure):
    pass
Nag_E05State._fields_ = [
    ('comm', c_double * 100),
]
class Nag_GlmAcc(Structure):
    pass
Nag_GlmAcc._fields_ = [
    ('max_iter', Integer),
    ('tol', c_double),
    ('eps', c_double),
    ('print_val', Integer),
    ('init1', Integer),
    ('init2', Integer),
]
class Nag_GlmRes(Structure):
    pass
Nag_GlmRes._fields_ = [
    ('rank', Integer),
    ('df', c_double),
    ('rss', c_double),
    ('dev', c_double),
    ('se_beta', POINTER(c_double)),
    ('cov', POINTER(c_double)),
]
class Nag_RegData(Structure):
    pass
Nag_RegData._fields_ = [
    ('max_iter', Integer),
    ('iter_mon', Integer),
    ('cpsi', c_double),
    ('cucv', c_double),
    ('dchi', c_double),
    ('hpsi', c_double * 3),
    ('tol', c_double),
    ('init1', Integer),
    ('init2', Integer),
]
class Nag_ResWrk(Structure):
    pass
Nag_ResWrk._fields_ = [
    ('rs', POINTER(c_double)),
    ('wgt', POINTER(c_double)),
    ('work', POINTER(c_double)),
]
class Nag_CorrData(Structure):
    pass
Nag_CorrData._fields_ = [
    ('max_iter', Integer),
    ('indm', Integer),
    ('iter_mon', Integer),
    ('bl', c_double),
    ('bd', c_double),
    ('tol', c_double),
    ('init1', Integer),
    ('init2', Integer),
]
class Nag_ag01hb(Structure):
    pass
Nag_ag01hb._fields_ = [
    ('c', c_double * 100),
    ('const_', c_double),
    ('l1', c_double),
    ('l2', c_double),
    ('u1', c_double),
    ('u2', c_double),
    ('sd1', c_double),
    ('sd2', c_double),
    ('rho', c_double),
    ('ind', Integer),
]
class Nag_ag01em(Structure):
    pass
Nag_ag01em._fields_ = [
    ('cq', c_double),
    ('cr', c_double),
    ('cv', c_double),
    ('cc', c_double),
    ('ceps', c_double),
    ('uflow', c_double),
    ('icr', Integer),
]
class Nag_QuasiRandom(Structure):
    pass
Nag_QuasiRandom._fields_ = [
    ('Faure', POINTER(c_long)),
    ('iref', c_long * 2000),
    ('GFdimension', c_int),
    ('GFbase', c_int),
    ('GFmbit', c_int),
    ('GFadd', POINTER(POINTER(c_int))),
    ('GFsub', POINTER(POINTER(c_int))),
    ('GFgray', POINTER(c_int)),
    ('GFb_ary', POINTER(c_int)),
    ('GFfaure', POINTER(POINTER(c_int))),
    ('GFplusone', POINTER(c_int)),
    ('GFpascal', POINTER(POINTER(POINTER(c_int)))),
    ('GFpowbase', POINTER(POINTER(c_long))),
    ('GFoutput', POINTER(c_long)),
]

# values for enumeration 'Nag_QuasiRandom_State'
Nag_QuasiRandom_State = c_int # enum

# values for enumeration 'Nag_QuasiRandom_Sequence'
Nag_QuasiRandom_Sequence = c_int # enum

# values for enumeration 'Nag_QuasiRandom_Scrambling'
Nag_QuasiRandom_Scrambling = c_int # enum

# values for enumeration 'Nag_RFP_Store'
Nag_RFP_Store = c_int # enum

# values for enumeration 'Nag_QuadType'
Nag_QuadType = c_int # enum

# values for enumeration 'Nag_VariableType'
Nag_VariableType = c_int # enum

# values for enumeration 'Nag_ODEMethod'
Nag_ODEMethod = c_int # enum

# values for enumeration 'Nag_KernelForm'
Nag_KernelForm = c_int # enum

# values for enumeration 'Nag_WeightMode'
Nag_WeightMode = c_int # enum

# values for enumeration 'Nag_Preprocess'
Nag_Preprocess = c_int # enum

# values for enumeration 'Nag_ZeroCols'
Nag_ZeroCols = c_int # enum

# values for enumeration 'Nag_Perturb'
Nag_Perturb = c_int # enum

# values for enumeration 'Nag_NearCorr_ProbType'
Nag_NearCorr_ProbType = c_int # enum

# values for enumeration 'Nag_CoordType'
Nag_CoordType = c_int # enum

# values for enumeration 'Nag_MatFunType'
Nag_MatFunType = c_int # enum

# values for enumeration 'Nag_IncludeIntercept'
Nag_IncludeIntercept = c_int # enum

# values for enumeration 'Nag_TS_Interpolation'
Nag_TS_Interpolation = c_int # enum

# values for enumeration 'Nag_TS_Transform'
Nag_TS_Transform = c_int # enum

# values for enumeration 'Nag_UnscentedTransform'
Nag_UnscentedTransform = c_int # enum

# values for enumeration 'Nag_NLSolver'
Nag_NLSolver = c_int # enum
class Nag_UserPrintFun(Structure):
    pass
Nag_UserPrintFun._fields_ = [
    ('itc', Integer),
    ('rss', c_double),
    ('objf', c_double),
    ('para', POINTER(c_double)),
    ('sd', POINTER(c_double)),
    ('df', c_double),
    ('npara', Integer),
    ('npe', Integer),
    ('mtyp', POINTER(Integer)),
    ('mser', POINTER(Integer)),
]
NAG_G13_PRINTFUN = NAG_FUNCTYPE(None, POINTER(Nag_UserPrintFun), POINTER(Nag_Comm))
class Nag_G13_Opt(Structure):
    pass
Nag_G13_Opt._fields_ = [
    ('used', Nag_Boolean),
    ('f_name', STRING),
    ('init1', Integer),
    ('init2', Integer),
    ('cfixed', Nag_Boolean),
    ('criteria', Nag_Likelihood),
    ('max_iter', Integer),
    ('alpha', c_double),
    ('beta', c_double),
    ('delta', c_double),
    ('gamma', c_double),
    ('iter', Integer),
    ('cm', POINTER(c_double)),
    ('tdcm', Integer),
    ('nag_cm', c_long),
    ('res', POINTER(c_double)),
    ('lenres', Integer),
    ('nag_res', c_long),
    ('zt', POINTER(c_double)),
    ('tdzt', Integer),
    ('nag_zt', c_long),
    ('noise', POINTER(c_double)),
    ('nag_noise', c_long),
    ('isttf', Integer),
    ('sttf', POINTER(c_double)),
    ('nag_sttf', c_long),
    ('nsttf', Integer),
    ('print_level', Nag_PrintType),
    ('list', Nag_Boolean),
    ('print_fun', NAG_G13_PRINTFUN),
    ('outfile', c_char * 512),
]
class Nag_ArimaOrder(Structure):
    pass
Nag_ArimaOrder._fields_ = [
    ('p', Integer),
    ('d', Integer),
    ('q', Integer),
    ('bigp', Integer),
    ('bigd', Integer),
    ('bigq', Integer),
    ('s', Integer),
]
class Nag_TransfOrder(Structure):
    pass
Nag_TransfOrder._fields_ = [
    ('init3', Integer),
    ('b', POINTER(Integer)),
    ('nag_b', c_long),
    ('q', POINTER(Integer)),
    ('nag_q', c_long),
    ('p', POINTER(Integer)),
    ('nag_p', c_long),
    ('r', POINTER(Integer)),
    ('nag_r', c_long),
]
NAG_H02_PRINTFUN = NAG_FUNCTYPE(None, POINTER(Nag_Search_State), POINTER(Nag_Comm))
NAG_H02_QPFUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
class Nag_MPS_Opt(Structure):
    pass
Nag_MPS_Opt._fields_ = [
    ('init1', Integer),
    ('init2', Integer),
    ('prob_name', c_char * 9),
    ('bnd_name', c_char * 9),
    ('obj_name', c_char * 9),
    ('range_name', c_char * 9),
    ('rhs_name', c_char * 9),
    ('col_lo_default', c_double),
    ('col_up_default', c_double),
    ('infinity', c_double),
    ('ncol_approx', Integer),
    ('nrow_approx', Integer),
    ('sparse', Nag_Boolean),
    ('est_density', c_double),
    ('requ_ivar', Nag_Boolean),
    ('list', Nag_Boolean),
    ('output_level', Nag_OutputType),
    ('outfile', c_char * 512),
]
class Nag_H02_Opt(Structure):
    pass
Nag_H02_Opt._fields_ = [
    ('f_name', STRING),
    ('f_longname', STRING),
    ('prob', Nag_MIP_ProbType),
    ('init1', Integer),
    ('init2', Integer),
    ('used', Nag_Boolean),
    ('list', Nag_Boolean),
    ('print_level', Nag_PrintType),
    ('print_fun', NAG_H02_PRINTFUN),
    ('outfile', c_char * 512),
    ('max_iter', Integer),
    ('inf_bound', c_double),
    ('feas_tol', c_double),
    ('hrows', Integer),
    ('max_df', Integer),
    ('rank_tol', c_double),
    ('first_soln', Nag_Boolean),
    ('max_depth', Integer),
    ('max_nodes', Integer),
    ('int_obj_bound', c_double),
    ('soln_tol', c_double),
    ('int_tol', c_double),
    ('nodsel', Nag_Node_Selection),
    ('varsel', Nag_Var_Selection),
    ('branch_dir', Nag_Branch_Direction),
    ('priority', POINTER(c_double)),
    ('lower', POINTER(c_double)),
    ('upper', POINTER(c_double)),
    ('lambda', POINTER(c_double)),
    ('state', POINTER(Integer)),
    ('nag_lower', get_pointer_int_type()),
    ('nag_upper', get_pointer_int_type()),
    ('nag_state', get_pointer_int_type()),
    ('nag_lambda', get_pointer_int_type()),
    ('prob_name', c_char * 9),
    ('obj_name', c_char * 9),
    ('bnd_name', c_char * 9),
    ('rhs_name', c_char * 9),
    ('range_name', c_char * 9),
    ('crnames', POINTER(STRING)),
    ('nag_crnames', c_long),
    ('col_lo_default', c_double),
    ('col_up_default', c_double),
    ('infinity', c_double),
    ('ncol_approx', Integer),
    ('nrow_approx', Integer),
    ('output_level', Nag_OutputType),
    ('n_ivar', Integer),
    ('dbg', Nag_Boolean),
    ('dbg_rt', Nag_Boolean),
    ('dbg_rqip', Nag_Boolean),
    ('dbg_es', Nag_EndState),
]
class Nag_MIP_Problem(Structure):
    pass
Nag_MIP_Problem._fields_ = [
    ('init1', Integer),
    ('init2', Integer),
    ('used', Nag_Boolean),
    ('type', Nag_MIP_ProbType),
    ('n_ivar', Integer),
    ('ivar', POINTER(Integer)),
    ('n', Integer),
    ('m', Integer),
    ('bl', POINTER(c_double)),
    ('bu', POINTER(c_double)),
    ('names', POINTER(STRING)),
    ('c', POINTER(c_double)),
    ('h', POINTER(c_double)),
    ('tdh', Integer),
    ('qphess', NAG_H02_QPFUN),
    ('a', POINTER(c_double)),
    ('tda', Integer),
    ('prob_name', c_char * 9),
    ('bnd_name', c_char * 9),
    ('obj_name', c_char * 9),
    ('range_name', c_char * 9),
    ('rhs_name', c_char * 9),
    ('nnz', Integer),
    ('iobj', Integer),
    ('ha', POINTER(Integer)),
    ('ka', POINTER(Integer)),
]
class Nag_MIP_Solution(Structure):
    pass
Nag_MIP_Solution._fields_ = [
    ('objval', c_double),
    ('x', POINTER(c_double)),
    ('lambda', POINTER(c_double)),
    ('state', POINTER(Integer)),
    ('endstate', Nag_EndState),
]
class Nag_e04nfg_Args(Structure):
    pass
Nag_e04nfg_Args._fields_ = [
    ('prbtyp', STRING),
    ('vertex', Nag_Boolean),
    ('cset', Nag_Boolean),
    ('iw', POINTER(Integer)),
    ('w', POINTER(c_double)),
    ('ae04mf', POINTER(Nag_ae04mf)),
    ('ee04mf', POINTER(Nag_ee04mf)),
    ('e04_opt', POINTER(Nag_E04_Opt)),
    ('e04_comm', POINTER(Nag_Comm)),
    ('e04_stream', POINTER(Nag_FileSt)),
]
class _s_Nag_HashWrap(Structure):
    pass
Nag_HashWrap = _s_Nag_HashWrap
_s_Nag_HashWrap._fields_ = [
    ('data', Pointer),
    ('name', STRING),
    ('next', POINTER(Nag_HashWrap)),
]
class Nag_HashPool(Structure):
    pass
Nag_HashPool._fields_ = [
    ('pool', POINTER(Nag_HashWrap)),
    ('next_free', POINTER(Nag_HashWrap)),
    ('pool_size', c_ulong),
    ('n_used', c_ulong),
    ('grow', POINTER(POINTER(Nag_HashWrap))),
    ('total_pool_size', c_ulong),
    ('ngrow', Integer),
    ('maxgrow', Integer),
]
class Nag_HashTable(Structure):
    pass
Nag_HashTable._fields_ = [
    ('hw_array', POINTER(Nag_HashWrap)),
    ('hpool', Nag_HashPool),
    ('table_size', c_ulong),
]
NAG_C05AGC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_C05AUC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_C05AWC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_C05AYC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_C05QBC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_C05QCC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_C05QSC_FUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_C05RBC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_C05RCC_FCN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_C05SDC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_C05UBC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Integer), POINTER(Nag_User))
NAG_D01BDC_F = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D01DAC_PHI1 = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D01DAC_PHI2 = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D01DAC_F = NAG_FUNCTYPE(c_double, c_double, c_double, POINTER(Nag_Comm))
NAG_D01ESC_F = NAG_FUNCTYPE(None, Integer, Integer, Integer, c_double, Integer, POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D01FBC_FUN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D01FCC_FUN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double))
NAG_D01FDC_F = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D01FDC_REGION = NAG_FUNCTYPE(None, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D01GBC_FUN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double))
NAG_D01GDC_VECFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D01GDC_VECREG = NAG_FUNCTYPE(None, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D01PAC_FUNCTN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D01RGC_F = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D01SJC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01SKC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01SLC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01SMC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01SNC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01SPC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01SQC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01SSC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01TAC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_User))
NAG_D01UAC_F = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D01UBC_F = NAG_FUNCTYPE(None, POINTER(c_double), POINTER(c_double), Integer,
                            POINTER(Nag_Comm), POINTER(Integer))
NAG_D01WCC_FUN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double), POINTER(Nag_User))
NAG_D01XBC_FUN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double), POINTER(Nag_User))
NAG_D02CJC_FUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02CJC_OUTFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02CJC_GFUN = NAG_FUNCTYPE(c_double, Integer, c_double, POINTER(c_double), POINTER(Nag_User))
NAG_D02EJC_FUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02EJC_PFUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02EJC_OUTFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02EJC_GFUN = NAG_FUNCTYPE(c_double, Integer, c_double, POINTER(c_double), POINTER(Nag_User))
NAG_D02GAC_FUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02GBC_FUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(Nag_User))
NAG_D02GBC_GFUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(Nag_User))
NAG_D02NEC_RES = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D02NEC_JAC = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, POINTER(Nag_Comm))
NAG_D02NJC_RESID = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D02NJC_JAC = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), c_double, c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02NJC_MONITR = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02QFC_FUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02QFC_GFUN = NAG_FUNCTYPE(c_double, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_User))
NAG_D02RAC_FUN = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAC_GFUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAC_JFUN = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAC_JGFUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAC_JEPSFUN = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAC_JGEPSFUN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D03PCC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PCC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PDC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PDC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PDC_UINIT = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03PEC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PEC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PFC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PFC_NUMFLX = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm), POINTER(Nag_D03_Save))
NAG_D03PFC_BNDARY = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PHC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PHC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PHC_ODEDEF = NAG_FUNCTYPE(None, Integer, c_double, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PJC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PJC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PJC_ODEDEF = NAG_FUNCTYPE(None, Integer, c_double, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PJC_UVINIT = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D03PKC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PKC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PKC_ODEDEF = NAG_FUNCTYPE(None, Integer, c_double, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PLC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PLC_NUMFLX = NAG_FUNCTYPE(None, Integer, c_double, c_double, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm), POINTER(Nag_D03_Save))
NAG_D03PLC_BNDARY = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PLC_ODEDEF = NAG_FUNCTYPE(None, Integer, c_double, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PPC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PPC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PPC_UVINIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D03PPC_ODEDEF = NAG_FUNCTYPE(None, Integer, c_double, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PPC_MONITF = NAG_FUNCTYPE(None, c_double, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03PRC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PRC_BNDARY = NAG_FUNCTYPE(None, Integer, c_double, Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PRC_UVINIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D03PRC_ODEDEF = NAG_FUNCTYPE(None, Integer, c_double, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PRC_MONITF = NAG_FUNCTYPE(None, c_double, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03PSC_PDEDEF = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PSC_NUMFLX = NAG_FUNCTYPE(None, Integer, c_double, c_double, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm), POINTER(Nag_D03_Save))
NAG_D03PSC_BNDARY = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PSC_UVINIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D03PSC_ODEDEF = NAG_FUNCTYPE(None, Integer, c_double, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03PSC_MONITF = NAG_FUNCTYPE(None, c_double, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D04AAC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05AAC_K1 = NAG_FUNCTYPE(c_double, c_double, c_double, POINTER(Nag_Comm))
NAG_D05AAC_K2 = NAG_FUNCTYPE(c_double, c_double, c_double, POINTER(Nag_Comm))
NAG_D05AAC_G = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05ABC_K = NAG_FUNCTYPE(c_double, c_double, c_double, POINTER(Nag_Comm))
NAG_D05ABC_G = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05BAC_CK = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05BAC_CG = NAG_FUNCTYPE(c_double, c_double, c_double, POINTER(Nag_Comm))
NAG_D05BAC_CF = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05BDC_CK = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05BDC_CF = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05BDC_CG = NAG_FUNCTYPE(c_double, c_double, c_double, POINTER(Nag_Comm))
NAG_D05BEC_CK = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05BEC_CF = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D05BEC_CG = NAG_FUNCTYPE(c_double, c_double, c_double, POINTER(Nag_Comm))
NAG_D06BAC_FBND = NAG_FUNCTYPE(c_double, Integer, c_double, c_double, POINTER(Nag_Comm))
NAG_E04ABC_FUN = NAG_FUNCTYPE(None, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_E04BBC_FUN = NAG_FUNCTYPE(None, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04CBC_FUNCT = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04CBC_MONIT = NAG_FUNCTYPE(None, c_double, c_double, POINTER(c_double), Integer, Integer, c_double, c_double, POINTER(Nag_Comm))
NAG_E04DGC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04FCC_FUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04GBC_FUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04JCC_OBJFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_E04JCC_MONFUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), c_double, c_double, POINTER(Nag_Comm), POINTER(Integer))
NAG_E04LBC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04LBC_HESS = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04NFC_FUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04NKC_HESSFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04NQC_QPHX = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04UCC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04UCC_CONFUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04UGC_CONFUN = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04UGC_OBJFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04UNC_OBJFUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04UNC_CONFUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04VHC_USRFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), Integer, Integer, POINTER(c_double), Integer, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_E04VJC_USRFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), Integer, Integer, POINTER(c_double), Integer, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_E04WDC_CONFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04WDC_OBJFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04XAC_OBJFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04YAC_FUN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05JBC_OBJFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm), POINTER(Integer))
NAG_E05JBC_MON = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(Integer), Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm), POINTER(Integer))
NAG_E05SAC_OBJFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05SAC_MONMOD = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm), POINTER(Integer))
NAG_E05SBC_OBJFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05SBC_CONFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05SBC_MONMOD = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm), POINTER(Integer))
NAG_E05UCC_CONFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05UCC_OBJFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05UCC_START = NAG_FUNCTYPE(None, Integer, POINTER(c_double), Integer, Nag_Boolean, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_F01EFC_F = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_F01EMC_F = NAG_FUNCTYPE(None, Integer, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F01FFC_F = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_F01FMC_F = NAG_FUNCTYPE(None, Integer, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F02FKC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_F02WGC_AV = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_F02FKC_OPTION = NAG_FUNCTYPE(None, POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_F08PAC_SELECT = NAG_FUNCTYPE(Nag_Boolean, c_double, c_double)
NAG_F08PBC_SELECT = NAG_FUNCTYPE(Nag_Boolean, c_double, c_double)
NAG_F08PNC_SELECT = NAG_FUNCTYPE(Nag_Boolean, Nag_Complex_Type)
NAG_F08PPC_SELECT = NAG_FUNCTYPE(Nag_Boolean, Nag_Complex_Type)
NAG_F08XAC_SELCTG = NAG_FUNCTYPE(Nag_Boolean, c_double, c_double, c_double)
NAG_F08XBC_SELCTG = NAG_FUNCTYPE(Nag_Boolean, c_double, c_double, c_double)
NAG_F08XCC_SELCTG = NAG_FUNCTYPE(Nag_Boolean, c_double, c_double, c_double)
NAG_F08XNC_SELCTG = NAG_FUNCTYPE(Nag_Boolean, Nag_Complex_Type, Nag_Complex_Type)
NAG_F08XPC_SELCTG = NAG_FUNCTYPE(Nag_Boolean, Nag_Complex_Type, Nag_Complex_Type)
NAG_F08XQC_SELCTG = NAG_FUNCTYPE(Nag_Boolean, Nag_Complex_Type, Nag_Complex_Type)
NAG_G02EFC_MONFUN = NAG_FUNCTYPE(None, Nag_FullStepwise, Integer, c_double, POINTER(Nag_Comm))
NAG_G02HBC_UCV = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_G02HDC_CHI = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_G02HDC_PSI = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_G02HFC_PSI = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_G02HFC_PSP = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_G02HLC_UCV = NAG_FUNCTYPE(None, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_G02HMC_UCV = NAG_FUNCTYPE(None, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_G07DCC_CHI = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_G07DCC_PSI = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_H02BBC_QPHESS = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_M01CSC_COMPARE = NAG_FUNCTYPE(Integer, c_void_p, c_void_p)
NAG_M01CTC_COMPARE = NAG_FUNCTYPE(Integer, c_void_p, c_void_p)
NAG_M01CUC_COMPARE = NAG_FUNCTYPE(Integer, c_void_p, c_void_p)
NAG_M01DSC_COMPARE = NAG_FUNCTYPE(Integer, c_void_p, c_void_p)
NAG_M01FSC_COMPARE = NAG_FUNCTYPE(Integer, c_void_p, c_void_p)
NAG_E04NKC_HESSFUN_WRAP = NAG_FUNCTYPE(None, NAG_E04NKC_HESSFUN, Integer, Integer, Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm))
NAG_C06_FUN = NAG_FUNCTYPE(Nag_Complex_Type, Nag_Complex_Type, POINTER(Nag_Comm))
NAG_C06LBC_FUN = NAG_C06_FUN
NAG_D01_TS_FUN = NAG_D01SJC_FUN
NAG_C06LAC_FUN = NAG_FUNCTYPE(None, Nag_Complex_Type, POINTER(Complex), POINTER(Nag_Comm))
NAG_D01ARC_FUN = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_D01EAC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D01GCC_FUN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D01GCC_REG = NAG_FUNCTYPE(None, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D01JAC_FUN = NAG_FUNCTYPE(c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D01RBC_MON = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02AGC_AUX = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02AGC_BECAUX = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02AGC_RAAUX = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02AGC_PRSOL = NAG_FUNCTYPE(None, POINTER(c_double), c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02BGC_FCN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02BHC_FCN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02BHC_G = NAG_FUNCTYPE(c_double, Integer, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02BJC_FCN = NAG_FUNCTYPE(None, c_double, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02BJC_OUTPUT = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02BJC_G = NAG_FUNCTYPE(c_double, c_double, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02HAC_FCN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02HBC_FCN = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02HBC_BC = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02HBC_RANGE = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02JAC_CF = NAG_FUNCTYPE(c_double, Integer, c_double, POINTER(Nag_Comm))
NAG_D02JAC_BC = NAG_FUNCTYPE(None, Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02JBC_CF = NAG_FUNCTYPE(c_double, Integer, Integer, c_double, POINTER(Nag_Comm))
NAG_D02JBC_BC = NAG_FUNCTYPE(None, Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02KAC_COEFFN = NAG_FUNCTYPE(None, POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, c_double, Integer, POINTER(Nag_Comm))
NAG_D02KAC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02KDC_COEFFN = NAG_FUNCTYPE(None, POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, c_double, Integer, POINTER(Nag_Comm))
NAG_D02KDC_BDYVAL = NAG_FUNCTYPE(None, c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02KDC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02KEC_COEFFN = NAG_FUNCTYPE(None, POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, c_double, Integer, POINTER(Nag_Comm))
NAG_D02KEC_BDYVAL = NAG_FUNCTYPE(None, c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02KEC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02KEC_REPORT = NAG_FUNCTYPE(None, c_double, POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02LAC_FCN = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02NGC_RESID = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D02NGC_JAC = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), c_double, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02NGC_MONITR = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02NHC_RESID = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D02NHC_JAC = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), c_double, c_double, Integer, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02NHC_MONITR = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02PEC_F = NAG_FUNCTYPE(None, c_double, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02PFC_F = NAG_FUNCTYPE(None, c_double, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02PSC_F = NAG_FUNCTYPE(None, c_double, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02SAC_RANGE = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02SAC_BC = NAG_FUNCTYPE(None, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, Integer, POINTER(Nag_Comm))
NAG_D02SAC_FCN = NAG_FUNCTYPE(None, c_double, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, Integer, POINTER(Nag_Comm))
NAG_D02SAC_EQN = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02SAC_CONSTR = NAG_FUNCTYPE(Nag_Boolean, POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02SAC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), c_double, c_double, c_double, POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer), POINTER(c_double))
NAG_D02SAC_PRSOL = NAG_FUNCTYPE(None, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D02TGC_COEFF = NAG_FUNCTYPE(None, c_double, Integer, POINTER(c_double), Integer, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02TGC_BDYC = NAG_FUNCTYPE(None, POINTER(c_double), Integer, Integer, POINTER(c_double), Integer, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_E02GBC_MONIT = NAG_FUNCTYPE(None, Integer, POINTER(c_double), Integer, Integer, c_double, POINTER(Nag_Comm))
NAG_E04FYC_LSFUN1 = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04GDC_LSQFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04GDC_LSQMON = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, Integer, Integer, POINTER(Nag_Comm))
NAG_E04GYC_LSFUN2 = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04GZC_LSFUN2 = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04HEC_LSQFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04HEC_LSQHES = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04HEC_LSQMON = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), Integer, Integer, Integer, POINTER(Nag_Comm))
NAG_E04HYC_LSFUN2 = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04HYC_LSHES2 = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04JYC_FUNCT1 = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04KDC_FUNCT = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04KDC_MONIT = NAG_FUNCTYPE(None, Integer, POINTER(c_double), c_double, POINTER(c_double), POINTER(Integer), c_double, c_double, Nag_Boolean, Integer, Integer, POINTER(Nag_Comm))
NAG_E04KYC_FUNCT2 = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04KZC_FUNCT2 = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04LYC_FUNCT2 = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04LYC_HESS2 = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_E04USC_CONFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04USC_OBJFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04YBC_LSQFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E04YBC_LSQHES = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05USC_CONFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05USC_OBJFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_E05USC_START = NAG_FUNCTYPE(None, Integer, POINTER(c_double), Integer, Nag_Boolean, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_F01ELC_F = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F01FLC_F = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F01JBC_F = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F01JCC_F = NAG_FUNCTYPE(None, Integer, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F01KBC_F = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F01KCC_F = NAG_FUNCTYPE(None, Integer, POINTER(Integer), Integer, POINTER(Complex), POINTER(Complex), POINTER(Nag_Comm))
NAG_F02EKC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(Complex), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_F02EKC_OPTION = NAG_FUNCTYPE(None, POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_F02FJC_DOT = NAG_FUNCTYPE(c_double, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm), Integer)
NAG_F02FJC_IMAGE = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm), Integer)
NAG_F02FJC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_F04QAC_APROD = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(c_double), POINTER(c_double), Integer, Integer, POINTER(Nag_Comm))
NAG_G05ZMC_COV1 = NAG_FUNCTYPE(None, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_G05ZQC_COV2 = NAG_FUNCTYPE(None, c_double, c_double, POINTER(c_double), POINTER(Nag_Comm))
NAG_G08CCC_CDF = NAG_FUNCTYPE(c_double, c_double, POINTER(Nag_Comm))
NAG_G13AEC_PIV = NAG_FUNCTYPE(None, POINTER(Integer), POINTER(c_double), Integer, c_double, Integer, POINTER(Integer), c_double, POINTER(c_double), POINTER(c_double), Integer, Integer, Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_G13EKC_F = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_G13EKC_H = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_G13NEC_CHGPFN = NAG_FUNCTYPE(None, Nag_TS_SegSide, Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_G13NBC_COSTFN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_G13NBC_CONSTFN = NAG_FUNCTYPE(None, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_G13NEC_CHGPFN = NAG_FUNCTYPE(None, Nag_TS_SegSide, Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_H02CBC_QPHESS = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_H02CBC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Boolean), POINTER(Integer), POINTER(Nag_Comm))
NAG_H02CEC_QPHX = NAG_FUNCTYPE(None, Integer, Integer, POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_H02CEC_MONIT = NAG_FUNCTYPE(None, Integer, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Boolean), POINTER(Integer), POINTER(Nag_Comm))
NAG_H02DAC_CONFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_H02DAC_OBJFUN = NAG_FUNCTYPE(None, POINTER(Integer), Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_H05ABC_F = NAG_FUNCTYPE(None, Integer, Integer, Integer, POINTER(Integer), Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm), POINTER(Integer))
NAG_D02TLC_FFUN = NAG_FUNCTYPE(None, c_double, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02TLC_FJAC = NAG_FUNCTYPE(None, c_double, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm))
NAG_D02TLC_GAFUN = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02TLC_GBFUN = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02TLC_GAJAC = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02TLC_GBJAC = NAG_FUNCTYPE(None, POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(c_double), POINTER(Nag_Comm))
NAG_D02TLC_GUESS = NAG_FUNCTYPE(None, c_double, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03EEC_PDEF = NAG_FUNCTYPE(None, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03EEC_BNDY = NAG_FUNCTYPE(None, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))
NAG_D03RAC_PDEDEF = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03RAC_BNDARY = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03RAC_PDEIV = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03RAC_MONITR = NAG_FUNCTYPE(None, Integer, c_double, c_double, c_double, Nag_Boolean, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D03RBC_INIDOM = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Nag_Comm), POINTER(Integer))
NAG_D03RBC_PDEDEF = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03RBC_BNDARY = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03RBC_PDEIV = NAG_FUNCTYPE(None, Integer, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_D03RBC_MONITR = NAG_FUNCTYPE(None, Integer, c_double, c_double, c_double, Nag_Boolean, Integer, c_double, c_double, c_double, c_double, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Nag_Comm))
NAG_D02RAZ_fcn = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_g = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_fcnep = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_fcna = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_fcnb = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_jacobe = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_jacobg = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_jaceps = NAG_FUNCTYPE(None, Integer, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_D02RAZ_jacgep = NAG_FUNCTYPE(None, Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_User))
NAG_E04HBC_FUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Nag_Comm))
NAG_E04HCC_FUN = NAG_E04HBC_FUN
NAG_E04KBC_FUN = NAG_E04HBC_FUN
class Nag_Fun(Structure):
    pass
Nag_Fun._fields_ = [
    ('type', Nag_FunType),
    ('call_lsf_noderiv', NAG_FUNCTYPE(None, Integer, Integer, NAG_E04FCC_FUN, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))),
    ('lsf_noderiv', NAG_E04FCC_FUN),
    ('call_lsf_deriv', NAG_FUNCTYPE(None, Integer, Integer, NAG_E04GBC_FUN, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Nag_Comm))),
    ('lsf_deriv', NAG_E04GBC_FUN),
]
__all__ = ['Ac02af', 'Ac02ag', 'BackwardSequence', 'Bc02af', 'Bc02ag',
           'BottomPivot', 'Cc02ag', 'Complex', 'Complexf',
           'ConjugateTranspose', 'FixedPivot', 'ForwardSequence',
           'NAG_C05AGC_FUN', 'NAG_C05AUC_FUN', 'NAG_C05AWC_FUN',
           'NAG_C05AYC_FUN', 'NAG_C05QBC_FUN', 'NAG_C05QCC_FUN',
           'NAG_C05QSC_FUN', 'NAG_C05RBC_FUN', 'NAG_C05RCC_FCN',
           'NAG_C05SDC_FUN', 'NAG_C05UBC_FUN', 'NAG_C06LAC_FUN',
           'NAG_C06LBC_FUN', 'NAG_C06_FUN', 'NAG_D01ARC_FUN',
           'NAG_D01BDC_F', 'NAG_D01DAC_F', 'NAG_D01DAC_PHI1',
           'NAG_D01DAC_PHI2', 'NAG_D01EAC_FUN', 'NAG_D01FBC_FUN',
           'NAG_D01FCC_FUN', 'NAG_D01FDC_F', 'NAG_D01FDC_REGION',
           'NAG_D01GBC_FUN', 'NAG_D01GCC_FUN', 'NAG_D01GCC_REG',
           'NAG_D01GDC_VECFUN', 'NAG_D01GDC_VECREG', 'NAG_D01JAC_FUN',
           'NAG_D01PAC_FUNCTN', 'NAG_D01RBC_MON', 'NAG_D01RGC_F',
           'NAG_D01SJC_FUN', 'NAG_D01SKC_FUN', 'NAG_D01SLC_FUN',
           'NAG_D01SMC_FUN', 'NAG_D01SNC_FUN', 'NAG_D01SPC_FUN',
           'NAG_D01SQC_FUN', 'NAG_D01SSC_FUN', 'NAG_D01TAC_FUN',
           'NAG_D01UAC_F', 'NAG_D01UBC_F',
           'NAG_D01WCC_FUN', 'NAG_D01XBC_FUN',
           'NAG_D01_TS_FUN', 'NAG_D02AGC_AUX', 'NAG_D02AGC_BECAUX',
           'NAG_D02AGC_PRSOL', 'NAG_D02AGC_RAAUX', 'NAG_D02BGC_FCN',
           'NAG_D02BHC_FCN', 'NAG_D02BHC_G', 'NAG_D02BJC_FCN',
           'NAG_D02BJC_G', 'NAG_D02BJC_OUTPUT', 'NAG_D02CJC_FUN',
           'NAG_D02CJC_GFUN', 'NAG_D02CJC_OUTFUN', 'NAG_D02EJC_FUN',
           'NAG_D02EJC_GFUN', 'NAG_D02EJC_OUTFUN', 'NAG_D02EJC_PFUN',
           'NAG_D02GAC_FUN', 'NAG_D02GBC_FUN', 'NAG_D02GBC_GFUN',
           'NAG_D02HAC_FCN', 'NAG_D02HBC_BC', 'NAG_D02HBC_FCN',
           'NAG_D02HBC_RANGE', 'NAG_D02JAC_BC', 'NAG_D02JAC_CF',
           'NAG_D02JBC_BC', 'NAG_D02JBC_CF', 'NAG_D02KAC_COEFFN',
           'NAG_D02KAC_MONIT', 'NAG_D02KDC_BDYVAL',
           'NAG_D02KDC_COEFFN', 'NAG_D02KDC_MONIT',
           'NAG_D02KEC_BDYVAL', 'NAG_D02KEC_COEFFN',
           'NAG_D02KEC_MONIT', 'NAG_D02KEC_REPORT', 'NAG_D02LAC_FCN',
           'NAG_D02NEC_JAC', 'NAG_D02NEC_RES', 'NAG_D02NGC_JAC',
           'NAG_D02NGC_MONITR', 'NAG_D02NGC_RESID', 'NAG_D02NHC_JAC',
           'NAG_D02NHC_MONITR', 'NAG_D02NHC_RESID', 'NAG_D02NJC_JAC',
           'NAG_D02NJC_MONITR', 'NAG_D02NJC_RESID',
           'NAG_D02PEC_F', 'NAG_D02PFC_F',
           'NAG_D02PSC_F', 'NAG_D02QFC_FUN',
           'NAG_D02QFC_GFUN', 'NAG_D02RAC_FUN', 'NAG_D02RAC_GFUN',
           'NAG_D02RAC_JEPSFUN', 'NAG_D02RAC_JFUN',
           'NAG_D02RAC_JGEPSFUN', 'NAG_D02RAC_JGFUN',
           'NAG_D02RAZ_fcn', 'NAG_D02RAZ_fcna', 'NAG_D02RAZ_fcnb',
           'NAG_D02RAZ_fcnep', 'NAG_D02RAZ_g', 'NAG_D02RAZ_jaceps',
           'NAG_D02RAZ_jacgep', 'NAG_D02RAZ_jacobe',
           'NAG_D02RAZ_jacobg', 'NAG_D02SAC_BC', 'NAG_D02SAC_CONSTR',
           'NAG_D02SAC_EQN', 'NAG_D02SAC_FCN', 'NAG_D02SAC_MONIT',
           'NAG_D02SAC_PRSOL', 'NAG_D02SAC_RANGE', 'NAG_D02TGC_BDYC',
           'NAG_D02TGC_COEFF', 'NAG_D02TLC_FFUN', 'NAG_D02TLC_FJAC',
           'NAG_D02TLC_GAFUN', 'NAG_D02TLC_GAJAC', 'NAG_D02TLC_GBFUN',
           'NAG_D02TLC_GBJAC', 'NAG_D02TLC_GUESS', 'NAG_D03EEC_BNDY',
           'NAG_D03EEC_PDEF', 'NAG_D03PCC_BNDARY',
           'NAG_D03PCC_PDEDEF', 'NAG_D03PDC_BNDARY',
           'NAG_D03PDC_PDEDEF', 'NAG_D03PDC_UINIT',
           'NAG_D03PEC_BNDARY', 'NAG_D03PEC_PDEDEF',
           'NAG_D03PFC_BNDARY', 'NAG_D03PFC_NUMFLX',
           'NAG_D03PFC_PDEDEF', 'NAG_D03PHC_BNDARY',
           'NAG_D03PHC_ODEDEF', 'NAG_D03PHC_PDEDEF',
           'NAG_D03PJC_BNDARY', 'NAG_D03PJC_ODEDEF',
           'NAG_D03PJC_PDEDEF', 'NAG_D03PJC_UVINIT',
           'NAG_D03PKC_BNDARY', 'NAG_D03PKC_ODEDEF',
           'NAG_D03PKC_PDEDEF', 'NAG_D03PLC_BNDARY',
           'NAG_D03PLC_NUMFLX', 'NAG_D03PLC_ODEDEF',
           'NAG_D03PLC_PDEDEF', 'NAG_D03PPC_BNDARY',
           'NAG_D03PPC_MONITF', 'NAG_D03PPC_ODEDEF',
           'NAG_D03PPC_PDEDEF', 'NAG_D03PPC_UVINIT',
           'NAG_D03PRC_BNDARY', 'NAG_D03PRC_MONITF',
           'NAG_D03PRC_ODEDEF', 'NAG_D03PRC_PDEDEF',
           'NAG_D03PRC_UVINIT', 'NAG_D03PSC_BNDARY',
           'NAG_D03PSC_MONITF', 'NAG_D03PSC_NUMFLX',
           'NAG_D03PSC_ODEDEF', 'NAG_D03PSC_PDEDEF',
           'NAG_D03PSC_UVINIT', 'NAG_D03RAC_BNDARY',
           'NAG_D03RAC_MONITR', 'NAG_D03RAC_PDEDEF',
           'NAG_D03RAC_PDEIV', 'NAG_D03RBC_BNDARY',
           'NAG_D03RBC_INIDOM', 'NAG_D03RBC_MONITR',
           'NAG_D03RBC_PDEDEF', 'NAG_D03RBC_PDEIV', 'NAG_D04AAC_FUN',
           'NAG_D05AAC_G', 'NAG_D05AAC_K1', 'NAG_D05AAC_K2',
           'NAG_D05ABC_G', 'NAG_D05ABC_K', 'NAG_D05BAC_CF',
           'NAG_D05BAC_CG', 'NAG_D05BAC_CK', 'NAG_D05BDC_CF',
           'NAG_D05BDC_CG', 'NAG_D05BDC_CK', 'NAG_D05BEC_CF',
           'NAG_D05BEC_CG', 'NAG_D05BEC_CK', 'NAG_D06BAC_FBND',
           'NAG_E02GBC_MONIT', 'NAG_E04ABC_FUN', 'NAG_E04BBC_FUN',
           'NAG_E04CBC_FUNCT', 'NAG_E04CBC_MONIT', 'NAG_E04DGC_FUN',
           'NAG_E04FCC_FUN', 'NAG_E04FYC_LSFUN1', 'NAG_E04GBC_FUN',
           'NAG_E04GDC_LSQFUN', 'NAG_E04GDC_LSQMON',
           'NAG_E04GYC_LSFUN2', 'NAG_E04GZC_LSFUN2', 'NAG_E04HBC_FUN',
           'NAG_E04HCC_FUN', 'NAG_E04HEC_LSQFUN', 'NAG_E04HEC_LSQHES',
           'NAG_E04HEC_LSQMON', 'NAG_E04HYC_LSFUN2',
           'NAG_E04HYC_LSHES2', 'NAG_E04JCC_MONFUN',
           'NAG_E04JCC_OBJFUN', 'NAG_E04JYC_FUNCT1', 'NAG_E04KBC_FUN',
           'NAG_E04KDC_FUNCT', 'NAG_E04KDC_MONIT',
           'NAG_E04KYC_FUNCT2', 'NAG_E04KZC_FUNCT2', 'NAG_E04LBC_FUN',
           'NAG_E04LBC_HESS', 'NAG_E04LYC_FUNCT2', 'NAG_E04LYC_HESS2',
           'NAG_E04NFC_FUN', 'NAG_E04NKC_HESSFUN',
           'NAG_E04NKC_HESSFUN_WRAP', 'NAG_E04NQC_QPHX',
           'NAG_E04UCC_CONFUN', 'NAG_E04UCC_FUN', 'NAG_E04UGC_CONFUN',
           'NAG_E04UGC_OBJFUN', 'NAG_E04UNC_CONFUN',
           'NAG_E04UNC_OBJFUN', 'NAG_E04USC_CONFUN',
           'NAG_E04USC_OBJFUN', 'NAG_E04VHC_USRFUN',
           'NAG_E04VJC_USRFUN', 'NAG_E04WDC_CONFUN',
           'NAG_E04WDC_OBJFUN', 'NAG_E04XAC_OBJFUN', 'NAG_E04YAC_FUN',
           'NAG_E04YBC_LSQFUN', 'NAG_E04YBC_LSQHES',
           'NAG_E04_PRINTFUN', 'NAG_E05JBC_MON', 'NAG_E05JBC_OBJFUN',
           'NAG_E05SAC_MONMOD', 'NAG_E05SAC_OBJFUN',
           'NAG_E05SBC_CONFUN', 'NAG_E05SBC_MONMOD',
           'NAG_E05SBC_OBJFUN', 'NAG_E05UCC_CONFUN',
           'NAG_E05UCC_OBJFUN', 'NAG_E05UCC_START',
           'NAG_E05USC_CONFUN', 'NAG_E05USC_OBJFUN',
           'NAG_E05USC_START', 'NAG_ERRHAN', 'NAG_F01EFC_F',
           'NAG_F01ELC_F', 'NAG_F01EMC_F', 'NAG_F01FFC_F',
           'NAG_F01FLC_F', 'NAG_F01FMC_F', 'NAG_F01JBC_F',
           'NAG_F01JCC_F', 'NAG_F01KBC_F', 'NAG_F01KCC_F',
           'NAG_F02EKC_MONIT', 'NAG_F02EKC_OPTION', 'NAG_F02FJC_DOT',
           'NAG_F02FJC_IMAGE', 'NAG_F02FJC_MONIT', 'NAG_F02WGC_AV',
           'NAG_F04QAC_APROD', 'NAG_F08PAC_SELECT',
           'NAG_F08PBC_SELECT', 'NAG_F08PNC_SELECT',
           'NAG_F08PPC_SELECT', 'NAG_F08XAC_SELCTG',
           'NAG_F08XBC_SELCTG', 'NAG_F08XCC_SELCTG', 'NAG_F08XNC_SELCTG',
           'NAG_F08XPC_SELCTG', 'NAG_F08XQC_SELCTG',
           'NAG_G02EFC_MONFUN', 'NAG_G02HBC_UCV',
           'NAG_G02HDC_CHI', 'NAG_G02HDC_PSI', 'NAG_G02HFC_PSI',
           'NAG_G02HFC_PSP', 'NAG_G02HLC_UCV', 'NAG_G02HMC_UCV',
           'NAG_G05ZMC_COV1', 'NAG_G05ZQC_COV2', 'NAG_G07DCC_CHI',
           'NAG_G07DCC_PSI', 'NAG_G08CCC_CDF', 'NAG_G13AEC_PIV',
           'NAG_G13NBC_COSTFN', 'NAG_G13NEC_CHGPFN',
           'NAG_G13_PRINTFUN', 'NAG_H02BBC_QPHESS',
           'NAG_H02CBC_MONIT', 'NAG_H02CBC_QPHESS',
           'NAG_H02CEC_MONIT', 'NAG_H02CEC_QPHX', 'NAG_H02_PRINTFUN',
           'NAG_H02_QPFUN', 'NAG_H05ABC_F', 'NAG_M01CSC_COMPARE',
           'NAG_M01CTC_COMPARE', 'NAG_M01CUC_COMPARE',
           'NAG_M01DSC_COMPARE', 'NAG_M01FSC_COMPARE', 'NagError',
           'Nag_1stDeriv_Small', 'Nag_2dSpline', 'Nag_2ndDeriv_Large',
           'Nag_2ndOrderDefaultParam', 'Nag_2ndOrderSuppliedParam',
           'Nag_3RSSH', 'Nag_4253H', 'Nag_4thOrderDefaultParam',
           'Nag_4thOrderSuppliedParam', 'Nag_ACORN', 'Nag_AboutMean',
           'Nag_AboutZero', 'Nag_AboveForm', 'Nag_AbsDiff',
           'Nag_AbsDiffScaled', 'Nag_AbsVal', 'Nag_AbsValScaled',
           'Nag_Absolute', 'Nag_AccumulateQ', 'Nag_AccumulateZ',
           'Nag_Adams', 'Nag_AddVar', 'Nag_AdditiveHoltWinters',
           'Nag_Alg', 'Nag_Alg_loga', 'Nag_Alg_loga_logb',
           'Nag_Alg_logb', 'Nag_AllEigVals', 'Nag_AllG_AllJ',
           'Nag_AllG_SomeJ', 'Nag_AllLevels', 'Nag_AllU', 'Nag_AllVT',
           'Nag_AllValues', 'Nag_AlternativeDAbs',
           'Nag_AlternativeDNeg', 'Nag_AlternativeDPos',
           'Nag_AmericanCall', 'Nag_AmericanPut',
           'Nag_AnalyticalJacobian', 'Nag_AnalyticalStructJacobian',
           'Nag_AndrewFun', 'Nag_ApplyP', 'Nag_ApplyQ',
           'Nag_ArimaOrder', 'Nag_Ascending', 'Nag_AutoCorr',
           'Nag_AutoCov', 'Nag_AvTwoNorm', 'Nag_AverageTies',
           'Nag_BB_Alloc_Fail', 'Nag_BB_Internal_Error', 'Nag_BB_OK',
           'Nag_BDF', 'Nag_BIC', 'Nag_BackErrBoth', 'Nag_BackErrLeft',
           'Nag_BackErrRight', 'Nag_BackTransform',
           'Nag_BackwardTransform', 'Nag_BalanceBoth',
           'Nag_BalancePermute', 'Nag_BalanceScale', 'Nag_Bartlett',
           'Nag_Base', 'Nag_Basic', 'Nag_BasisFactB',
           'Nag_BasisFactBS', 'Nag_BasisFile', 'Nag_Basis_Ill_Cond',
           'Nag_Basis_Singular', 'Nag_BeginBackward',
           'Nag_BeginForward', 'Nag_BeginStepwise', 'Nag_Beta',
           'Nag_Binomial', 'Nag_Biorthogonal1_1',
           'Nag_Biorthogonal1_3', 'Nag_Biorthogonal1_5',
           'Nag_Biorthogonal2_2', 'Nag_Biorthogonal2_4',
           'Nag_Biorthogonal2_6', 'Nag_Biorthogonal2_8',
           'Nag_Biorthogonal3_1', 'Nag_Biorthogonal3_3',
           'Nag_Biorthogonal3_5', 'Nag_Biorthogonal3_7',
           'Nag_BlomScores', 'Nag_BonferroniInterval', 'Nag_Both',
           'Nag_BothSides', 'Nag_BoundNotSet', 'Nag_Bounds',
           'Nag_BoundsEqual', 'Nag_BoundsZero', 'Nag_BrDir_NotSet',
           'Nag_BracketForm', 'Nag_Branch_InitX', 'Nag_Branch_Left',
           'Nag_Branch_Right', 'Nag_Broad_Search',
           'Nag_BrownsExponential', 'Nag_ByBlock', 'Nag_CDF',
           'Nag_CE_EM', 'Nag_CE_NR', 'Nag_Call', 'Nag_Censored',
           'Nag_Central', 'Nag_CentroSymmEven',
           'Nag_CentroSymmNeither', 'Nag_CentroSymmOdd',
           'Nag_Centroid', 'Nag_Character', 'Nag_CharacterLabels',
           'Nag_CheckCon', 'Nag_CheckObj', 'Nag_CheckObjCon',
           'Nag_ChiSquare', 'Nag_ChkNotSet', 'Nag_ClassBoundaryComp',
           'Nag_ClassBoundaryUser', 'Nag_ColMajor', 'Nag_Cold',
           'Nag_ColinearVar', 'Nag_ColpivRrank',
           'Nag_ColpivRrankCond', 'Nag_ColpivSVrankAbs',
           'Nag_ColpivSVrankRel', 'Nag_ColumnEquilibration',
           'Nag_Comm', 'Nag_CompProbApprox', 'Nag_CompProbExact',
           'Nag_CompSt', 'Nag_Compl', 'Nag_CompleteLink',
           'Nag_ComputeAll', 'Nag_ComputeMoments', 'Nag_ComputeQ',
           'Nag_ComputeSelected', 'Nag_ComputeV', 'Nag_Compute_1',
           'Nag_Compute_2', 'Nag_Compute_3', 'Nag_ConCheck',
           'Nag_CondBackErrBoth', 'Nag_CondBackErrLeft',
           'Nag_CondBackErrRight', 'Nag_CondFanLinVanDooren',
           'Nag_CondOnly', 'Nag_Conj', 'Nag_ConjTrans',
           'Nag_Continue', 'Nag_ContinueAndUpdate',
           'Nag_ContinueNoUpdate', 'Nag_Convolution',
           'Nag_Coord_Cartesian', 'Nag_Coord_Spherical',
           'Nag_Coord_Spherical_Special', 'Nag_CorrData',
           'Nag_Correlation', 'Nag_Cos', 'Nag_Cosh', 'Nag_Cosine',
           'Nag_CovMatAve', 'Nag_CovMatObs', 'Nag_CovNotSet',
           'Nag_CrashNotSet', 'Nag_CrashOnce', 'Nag_CrashThreeTimes',
           'Nag_CrashTwice', 'Nag_CriteriaNotSet',
           'Nag_CumulantsOnly', 'Nag_CustomMesh', 'Nag_Cycling',
           'Nag_D03_Save', 'Nag_DGauss', 'Nag_DInfo_NotSet',
           'Nag_DLTX', 'Nag_DWant_NotSet', 'Nag_D_Dbg', 'Nag_D_Full',
           'Nag_D_NoPrint', 'Nag_D_NotSet', 'Nag_D_Print',
           'Nag_D_Sum', 'Nag_DataByVar', 'Nag_DataByObs', 'Nag_DataCorr', 'Nag_DataCovar',
           'Nag_Daubechies10', 'Nag_Daubechies2', 'Nag_Daubechies3',
           'Nag_Daubechies4', 'Nag_Daubechies5', 'Nag_Daubechies6',
           'Nag_Daubechies7', 'Nag_Daubechies8', 'Nag_Daubechies9',
           'Nag_DavidsAndMoodsTest', 'Nag_DavidsTest',
           'Nag_Deadpoint', 'Nag_DebugSt', 'Nag_DecreasingOrder',
           'Nag_DeepBroad_Search', 'Nag_DeepMinObj_Search',
           'Nag_Deep_Search', 'Nag_DefInitMesh', 'Nag_Default',
           'Nag_DefaultJacobian', 'Nag_DegenEndCycle',
           'Nag_DegenInit', 'Nag_DegenOptimal', 'Nag_DendEast',
           'Nag_DendNorth', 'Nag_DendSouth', 'Nag_DendWest',
           'Nag_Der', 'Nag_Deriv', 'Nag_DerivNotSet',
           'Nag_Deriv_Error', 'Nag_Deriv_Inf', 'Nag_Deriv_OK',
           'Nag_Descending', 'Nag_DiagonalMatrix', 'Nag_DiffInt',
           'Nag_DirectForm', 'Nag_DirectionNotSet',
           'Nag_DiscrimEstimate', 'Nag_DiscrimPredict', 'Nag_DistAbs',
           'Nag_DistEuclid', 'Nag_DistSquared', 'Nag_DoAll',
           'Nag_DoBoth', 'Nag_DoNotForm', 'Nag_DoNothing',
           'Nag_DoOverwrite', 'Nag_DoSquare', 'Nag_DownandIn',
           'Nag_DownandOut', 'Nag_DropVar', 'Nag_DunnInterval',
           'Nag_E01_Opt', 'Nag_E04State', 'Nag_E04StateB',
           'Nag_E04_Opt', 'Nag_E05State', 'Nag_EUV', 'Nag_EigVals',
           'Nag_EigVecs', 'Nag_ElementsIn', 'Nag_ElementsSeparate',
           'Nag_Emax', 'Nag_EmbedPadValues', 'Nag_EmbedPadZeros',
           'Nag_EmbedScaleOne', 'Nag_EmbedScaleSqrtTraces',
           'Nag_EmbedScaleTraces', 'Nag_Emin', 'Nag_EndStateNotSet',
           'Nag_Entire', 'Nag_EnumTableType', 'Nag_Eps',
           'Nag_EqualCovar', 'Nag_EqualPrior',
           'Nag_EquilibrateAndFactor', 'Nag_Equilibrated',
           'Nag_ErrorAssess_off', 'Nag_ErrorAssess_on',
           'Nag_ErrorNameTable', 'Nag_EstimateInitialValues',
           'Nag_EstimatesOrig', 'Nag_EstimatesStand',
           'Nag_EuropeanCall', 'Nag_EuropeanPut', 'Nag_Even',
           'Nag_Exact', 'Nag_Exp', 'Nag_Expo', 'Nag_Exponential',
           'Nag_ExponentialLambda', 'Nag_ExtraScale', 'Nag_FALSE',
           'Nag_FDIntSt', 'Nag_FFT_Initial', 'Nag_FFT_Restart',
           'Nag_FFT_Subsequent', 'Nag_FP', 'Nag_FPE',
           'Nag_FacNoRotate', 'Nag_FacRotate', 'Nag_FacScoreBart',
           'Nag_FacScoreRegsn', 'Nag_Factored', 'Nag_FanLinVanDooren',
           'Nag_FaureTezuka', 'Nag_Feasible', 'Nag_FeasiblePoint',
           'Nag_FileID', 'Nag_FileSt', 'Nag_FinishStepwise',
           'Nag_First', 'Nag_FirstCall', 'Nag_FirstCallR',
           'Nag_FirstLevel', 'Nag_FirstSegCall', 'Nag_First_Int',
           'Nag_FisherInterval', 'Nag_FormBoth', 'Nag_FormP',
           'Nag_FormQ', 'Nag_FormX', 'Nag_ForwardTransform',
           'Nag_Freq', 'Nag_FrobeniusNorm', 'Nag_FullpivRrank',
           'Nag_FullpivRrankCond', 'Nag_Fun', 'Nag_Fun_Constant',
           'Nag_Fun_LinearOdd', 'Nag_Function', 'Nag_Functional',
           'Nag_G13_Opt', 'Nag_GCV', 'Nag_GPrintSt', 'Nag_Gamma',
           'Nag_GammaScale', 'Nag_Garch_Est_Initial_False',
           'Nag_Garch_Est_Initial_True', 'Nag_Garch_Fcall_False',
           'Nag_Garch_Fcall_True', 'Nag_Garch_Stationary_False',
           'Nag_Garch_Stationary_True', 'Nag_Garch_Useht_False',
           'Nag_Garch_Useht_True', 'Nag_GenPareto',
           'Nag_GeneralMatrix', 'Nag_GenerateFromReference',
           'Nag_GenerateWithoutReference', 'Nag_GlmAcc', 'Nag_GlmRes',
           'Nag_Grad_Chk_St', 'Nag_Grad_HessDiag',
           'Nag_Grad_HessFull', 'Nag_GroupAverage', 'Nag_GroupCovar',
           'Nag_GroupMeans', 'Nag_GroupSizePrior', 'Nag_GroupVar',
           'Nag_H02_Opt', 'Nag_HSEQRSource', 'Nag_Haar',
           'Nag_HalfPointSymmetric', 'Nag_HampelFun', 'Nag_HashOK',
           'Nag_HashPool', 'Nag_HashTable', 'Nag_HashTableTooBig',
           'Nag_HashWrap', 'Nag_Helmert', 'Nag_Hermite',
           'Nag_HessFull', 'Nag_Hess_Indefinite', 'Nag_Hess_Too_Big',
           'Nag_HessianFull', 'Nag_HessianLimited',
           'Nag_HessianNotSet', 'Nag_HighestTies', 'Nag_Hot',
           'Nag_HuberFun', 'Nag_HuberReg', 'Nag_Iden', 'Nag_Identity',
           'Nag_Ieee', 'Nag_IgnoreTies', 'Nag_IncSignZerosN',
           'Nag_IncSignZerosY', 'Nag_IncreasingOrder', 'Nag_Indices',
           'Nag_InfNorm', 'Nag_Infeasible', 'Nag_Infinite',
           'Nag_InitNotSet', 'Nag_InitQ', 'Nag_InitU', 'Nag_InitV',
           'Nag_InitWeights', 'Nag_InitZ', 'Nag_Init_All',
           'Nag_Init_F_G_H', 'Nag_Init_H_S', 'Nag_Init_None',
           'Nag_InitialValuesSupplied', 'Nag_InitializeAndGenerate',
           'Nag_InitializeI', 'Nag_InitializeReference', 'Nag_InputA',
           'Nag_Integer', 'Nag_IntegerLabels',
           'Nag_Integer_Additional', 'Nag_Intercept',
           'Nag_IntermediateCallR', 'Nag_Interval', 'Nag_Iter',
           'Nag_Iter_Long', 'Nag_JrotInner', 'Nag_JrotOuter',
           'Nag_JrotSorted',
           'Nag_LARS_LAR', 'Nag_LARS_ForwardStagewise', 'Nag_LARS_LASSO',
           'Nag_LARS_PositiveLASSO', 'Nag_LARS_None', 'Nag_LARS_Centered',
           'Nag_LARS_Normalized', 'Nag_LARS_CenteredNormalized',
           'Nag_LARS_StepNumber', 'Nag_LARS_ScaledNorm',
           'Nag_LARS_ProportionScaledNorm', 'Nag_LARS_UnscaledNorm',
           'Nag_LARS_ProportionUnscaledNorm',
           'Nag_LCCDefault', 'Nag_LCCImhof',
           'Nag_LCCPan', 'Nag_LDLTX', 'Nag_LDX', 'Nag_LH_Controller',
           'Nag_LH_Observer', 'Nag_LLTX', 'Nag_LOOCV', 'Nag_LP',
           'Nag_LRRoundDown', 'Nag_LRRoundUp', 'Nag_LS1', 'Nag_LS2',
           'Nag_LS3', 'Nag_LS4', 'Nag_LTX', 'Nag_LU_SolveB',
           'Nag_LU_SolveBt', 'Nag_LU_SolveL', 'Nag_LU_TypeB',
           'Nag_LU_TypeBS', 'Nag_LU_TypeBT', 'Nag_LX', 'Nag_Laguerre',
           'Nag_LargeEigVals', 'Nag_Last', 'Nag_LastCallR',
           'Nag_LastLevel', 'Nag_LeastSquares', 'Nag_LeftDerivs',
           'Nag_LeftDerivs_1', 'Nag_LeftDerivs_2', 'Nag_LeftDerivs_3',
           'Nag_LeftSide', 'Nag_LeftSingVecs', 'Nag_LeftSpan',
           'Nag_LeftSubSeg', 'Nag_LeftVecs', 'Nag_LeftVecsCtol',
           'Nag_Legendre', 'Nag_LinAlgBand', 'Nag_LinAlgFull',
           'Nag_LinAlgSparse', 'Nag_LinFunNotSet', 'Nag_Lin_Deriv',
           'Nag_Lin_NoDeriv', 'Nag_Linear', 'Nag_LinearHolt',
           'Nag_LinearRatio', 'Nag_Linesearch', 'Nag_Log',
           'Nag_LogNormal', 'Nag_LogTransform', 'Nag_Loga',
           'Nag_Logged', 'Nag_Logistic', 'Nag_Lower',
           'Nag_LowerBound', 'Nag_LowerFactored', 'Nag_LowerMatrix',
           'Nag_LowerMinus', 'Nag_LowerSemiInfinite', 'Nag_LowerTail',
           'Nag_LowestTies', 'Nag_Lsq', 'Nag_LsqScale', 'Nag_MILP',
           'Nag_MIP_Best_ISol', 'Nag_MIP_Big_Hess_ISol',
           'Nag_MIP_Big_Hess_No_ISol', 'Nag_MIP_Max_Depth_ISol',
           'Nag_MIP_Max_Depth_No_ISol', 'Nag_MIP_Max_Itn_ISol',
           'Nag_MIP_Max_Itn_No_ISol', 'Nag_MIP_Max_Nodes_ISol',
           'Nag_MIP_Max_Nodes_No_ISol', 'Nag_MIP_No_ISol',
           'Nag_MIP_Problem', 'Nag_MIP_Root_Big_Hess',
           'Nag_MIP_Root_Infeasible', 'Nag_MIP_Root_Max_Itn',
           'Nag_MIP_Root_Unbounded', 'Nag_MIP_Solution',
           'Nag_MIP_Stop_First_ISol', 'Nag_MIP_TypeNotSet',
           'Nag_MIP_User_Stop', 'Nag_MIQP1', 'Nag_MIQP2', 'Nag_MIQP3',
           'Nag_MIQP4', 'Nag_MODWTMulti', 'Nag_MODWTSingle',
           'Nag_MOM', 'Nag_MOMMLE', 'Nag_MPS_List', 'Nag_MPS_Opt',
           'Nag_MPS_Summary', 'Nag_MRG32k3a', 'Nag_MallowsReg',
           'Nag_ManyIterations', 'Nag_Marginal', 'Nag_MatCorr_Covar',
           'Nag_MatCorrelation', 'Nag_MatStandardised',
           'Nag_MatSumSq', 'Nag_MatUp', 'Nag_MatVarCovar',
           'Nag_MaxNorm', 'Nag_Maximize', 'Nag_Mean',
           'Nag_MeanInclude', 'Nag_MeanZero', 'Nag_Median',
           'Nag_MersenneTwister', 'Nag_Mesg', 'Nag_MexHat',
           'Nag_MinObj_Search', 'Nag_MinVariance', 'Nag_Minimize',
           'Nag_Minus', 'Nag_Mixed', 'Nag_ModifiedNewton',
           'Nag_MoodsTest', 'Nag_Morlet', 'Nag_MultiLevel',
           'Nag_MultiplicativeHoltWinters', 'Nag_NS_Bounded',
           'Nag_NS_Infeasible', 'Nag_NS_Integer',
           'Nag_NS_NotBranched', 'Nag_NS_NotSet', 'Nag_NS_NotSolved',
           'Nag_NS_Terminated', 'Nag_Nearest_Half', 'Nag_NegBinomial',
           'Nag_NewStart', 'Nag_NextPoint', 'Nag_NoBalancing',
           'Nag_NoBlocks', 'Nag_NoBounds', 'Nag_NoBounds_One_Call',
           'Nag_NoCensored', 'Nag_NoCentroSymm', 'Nag_NoCheck',
           'Nag_NoCondBackErr', 'Nag_NoConj', 'Nag_NoCorrection',
           'Nag_NoCrash', 'Nag_NoDeriv', 'Nag_NoDerivs',
           'Nag_NoEquilibration', 'Nag_NoFreq', 'Nag_NoIntercept',
           'Nag_NoLOO', 'Nag_NoLabels', 'Nag_NoMatUp', 'Nag_NoOutput',
           'Nag_NoPara', 'Nag_NoPrint', 'Nag_NoRemoveVar',
           'Nag_NoScale', 'Nag_NoScaleProvided', 'Nag_NoScrambling',
           'Nag_NoSortEigVals', 'Nag_NoTrans', 'Nag_NoTransNorm',
           'Nag_NoTransform', 'Nag_NoVIF', 'Nag_NoVarScale',
           'Nag_NoVec', 'Nag_NoWeights', 'Nag_NodSel_NotSet',
           'Nag_NonRefDiag', 'Nag_NonUnitDiag', 'Nag_Norm',
           'Nag_Normal', 'Nag_NormalDistn', 'Nag_NormalMean',
           'Nag_NormalMeanStd', 'Nag_NormalScores', 'Nag_NormalStd',
           'Nag_NotEigVecs', 'Nag_NotEqualCovar', 'Nag_NotFactored',
           'Nag_NotKnown', 'Nag_NotLeftVecs', 'Nag_NotLeftWork',
           'Nag_NotLsqScale', 'Nag_NotQ', 'Nag_NotRCond',
           'Nag_NotRegularized', 'Nag_NotRightVecs',
           'Nag_NotRightWork', 'Nag_NotSingularVecs',
           'Nag_NotSupplied', 'Nag_NotU', 'Nag_NotV', 'Nag_NotVT',
           'Nag_NotZ', 'Nag_Not_Converged', 'Nag_Not_Kuhn_Tucker',
           'Nag_NumericalJacobian', 'Nag_NumericalStructJacobian',
           'Nag_ODE_Adams', 'Nag_ODE_BDF',
           'Nag_ObjCheck', 'Nag_ObservAdd', 'Nag_ObservDel',
           'Nag_Odd', 'Nag_OneIteration', 'Nag_OneNorm',
           'Nag_Opt_Found', 'Nag_Optimal', 'Nag_Orig',
           'Nag_OrigCentroid', 'Nag_OrigEigVecs', 'Nag_OrigNorm',
           'Nag_OrigNormCentroid', 'Nag_OrigPara',
           'Nag_OsherOriginal', 'Nag_OsherPhysical',
           'Nag_Out_Of_Workspace', 'Nag_OutputNotSet',
           'Nag_OverallVar', 'Nag_Overflow', 'Nag_Overwrite',
           'Nag_OverwriteVT', 'Nag_OwenFaureTezuka', 'Nag_OwenLike',
           'Nag_PDF', 'Nag_PWM', 'Nag_PWMMLE', 'Nag_Pack',
           'Nag_PackedSingularVecs', 'Nag_ParaEstimated',
           'Nag_ParaSupplied', 'Nag_ParallelBlocks', 'Nag_Parzen',
           'Nag_Periodic', 'Nag_Permute', 'Nag_PerturbOff',
           'Nag_PerturbOn', 'Nag_Plus', 'Nag_Poisson',
           'Nag_PoissonLambda', 'Nag_Poly', 'Nag_PooledCovar',
           'Nag_PooledVar', 'Nag_PopVarEqual', 'Nag_PopVarNotEqual',
           'Nag_Prec', 'Nag_PredNoScale', 'Nag_PredStdScale',
           'Nag_PredUserScale', 'Nag_PreviousPoint',
           'Nag_PrintNotSet', 'Nag_ProbTypeNotSet', 'Nag_Probit',
           'Nag_PsiNotSet', 'Nag_Put', 'Nag_QP1', 'Nag_QP2',
           'Nag_QP3', 'Nag_QP4', 'Nag_QP_Print', 'Nag_QuadProgress',
           'Nag_QuadRatio', 'Nag_QuadSubProgress',
           'Nag_Quad_Gauss_Exponential',
           'Nag_Quad_Gauss_Exponential_Adjusted',
           'Nag_Quad_Gauss_Hermite',
           'Nag_Quad_Gauss_Hermite_Adjusted', 'Nag_Quad_Gauss_Jacobi',
           'Nag_Quad_Gauss_Jacobi_Adjusted',
           'Nag_Quad_Gauss_Kronrod_15', 'Nag_Quad_Gauss_Kronrod_21',
           'Nag_Quad_Gauss_Kronrod_31', 'Nag_Quad_Gauss_Kronrod_41',
           'Nag_Quad_Gauss_Kronrod_51', 'Nag_Quad_Gauss_Kronrod_61',
           'Nag_Quad_Gauss_Laguerre',
           'Nag_Quad_Gauss_Laguerre_Adjusted',
           'Nag_Quad_Gauss_Legendre', 'Nag_Quad_Gauss_Rational',
           'Nag_Quad_Gauss_Rational_Adjusted',
           'Nag_Quad_Sphere_Default', 'Nag_Quad_Sphere_HighAcc',
           'Nag_Quad_Sphere_HighAcc_BadCentre',
           'Nag_Quad_Sphere_HighAcc_BadSurface',
           'Nag_Quad_Sphere_LowAcc',
           'Nag_Quad_Sphere_LowAcc_BadCentre',
           'Nag_Quad_Sphere_LowAcc_BadSurface', 'Nag_QuasiRandom',
           'Nag_QuasiRandom_Cont', 'Nag_QuasiRandom_Faure',
           'Nag_QuasiRandom_Finish', 'Nag_QuasiRandom_Init',
           'Nag_QuasiRandom_Nied', 'Nag_QuasiRandom_Sobol',
           'Nag_QuasiRandom_SobolA659', 'Nag_RC', 'Nag_RCI_Approx',
           'Nag_RCI_Exact', 'Nag_RCondBoth', 'Nag_RCondEigVals',
           'Nag_RCondEigVecs', 'Nag_RCondLeft', 'Nag_RCondRight',
           'Nag_RFP_ConjTrans', 'Nag_RFP_Normal', 'Nag_RFP_Trans',
           'Nag_RK_2_3', 'Nag_RK_4_5', 'Nag_RK_7_8',
           'Nag_RK_onestep', 'Nag_RK_range', 'Nag_RLRoundDown',
           'Nag_RLRoundUp', 'Nag_Random', 'Nag_RandomTies',
           'Nag_RankScores', 'Nag_RatioMoments', 'Nag_Rational',
           'Nag_ReGenerateFromReference',
           'Nag_ReInitializeAndGenerate', 'Nag_ReStart', 'Nag_Real',
           'Nag_RealInfNorm', 'Nag_RealMaxNorm', 'Nag_RealOneNorm',
           'Nag_Real_Additional', 'Nag_Reci', 'Nag_Rectangular',
           'Nag_RegData', 'Nag_RegNotSet', 'Nag_Regularized',
           'Nag_Relative', 'Nag_ResWrk', 'Nag_ReuseWeights',
           'Nag_RightDerivs', 'Nag_RightDerivs_1',
           'Nag_RightDerivs_2', 'Nag_RightDerivs_3', 'Nag_RightSide',
           'Nag_RightSingVecs', 'Nag_RightSubSeg', 'Nag_RightVecs',
           'Nag_RightVecsJRots', 'Nag_RightVecsMV', 'Nag_Rnd',
           'Nag_RoLoadNotStand', 'Nag_RoLoadStand',
           'Nag_RowAndColumnEquilibration', 'Nag_RowColScale',
           'Nag_RowEquilibration', 'Nag_RowMajor', 'Nag_RuleOfThumb',
           'Nag_SStress', 'Nag_SamplePoints', 'Nag_SavageScores',
           'Nag_Scale', 'Nag_ScaleNotSet', 'Nag_ScaleProvided',
           'Nag_ScaleRes', 'Nag_Scat_Struct', 'Nag_ScheffeInterval',
           'Nag_Schur', 'Nag_SchweppeReg', 'Nag_ScoresEigenval',
           'Nag_ScoresNotStand', 'Nag_ScoresStand',
           'Nag_ScoresUnitVar', 'Nag_Search_State',
           'Nag_SecondSegCall', 'Nag_Select_Modulus',
           'Nag_Select_RealPart', 'Nag_SerialBlocks',
           'Nag_SeriesEven', 'Nag_SeriesGeneral', 'Nag_SeriesOdd',
           'Nag_Sfmin', 'Nag_Shep', 'Nag_SigmaBypas', 'Nag_SigmaChi',
           'Nag_SigmaConst', 'Nag_SigmaNotSet', 'Nag_SigmaRes',
           'Nag_SigmaSimul', 'Nag_SimSt', 'Nag_SimpleBdry',
           'Nag_SimpleCheck', 'Nag_SimpleOffBdry', 'Nag_Sin',
           'Nag_Sine', 'Nag_SingleCallR', 'Nag_SingleExponential',
           'Nag_SingleLevel', 'Nag_SingleLink', 'Nag_SingularVecs',
           'Nag_SingularVecsU', 'Nag_SingularVecsVT', 'Nag_Sinh',
           'Nag_SmoothFitFull', 'Nag_SmoothFitPartial',
           'Nag_SmoothFitQuick', 'Nag_SmoothParamCV',
           'Nag_SmoothParamDF', 'Nag_SmoothParamGCV', 'Nag_Sol',
           'Nag_SolDer', 'Nag_Soln', 'Nag_Soln_Iter',
           'Nag_Soln_Iter_Const', 'Nag_Soln_Iter_Diag',
           'Nag_Soln_Iter_Full', 'Nag_Soln_Iter_Long',
           'Nag_Soln_Root', 'Nag_Soln_Root_Iter', 'Nag_SomeG_AllJ',
           'Nag_SomeG_SomeJ', 'Nag_SortEigVals', 'Nag_SparseFP',
           'Nag_SparseFPE', 'Nag_SparseFPL', 'Nag_SparseFPS',
           'Nag_SparseLP', 'Nag_SparseNsym_BiCGSTAB',
           'Nag_SparseNsym_CG', 'Nag_SparseNsym_CGS',
           'Nag_SparseNsym_Check', 'Nag_SparseNsym_CompletePiv',
           'Nag_SparseNsym_FailDups', 'Nag_SparseNsym_FailZeros',
           'Nag_SparseNsym_JacPrec', 'Nag_SparseNsym_KeepZeros',
           'Nag_SparseNsym_MINRES', 'Nag_SparseNsym_MarkowitzPiv',
           'Nag_SparseNsym_ModFact', 'Nag_SparseNsym_NoCheck',
           'Nag_SparseNsym_NoPiv', 'Nag_SparseNsym_NoPrec',
           'Nag_SparseNsym_PartialPiv', 'Nag_SparseNsym_Prec',
           'Nag_SparseNsym_RGMRES', 'Nag_SparseNsym_RemoveDups',
           'Nag_SparseNsym_RemoveZeros', 'Nag_SparseNsym_SSORPrec',
           'Nag_SparseNsym_StoreCS', 'Nag_SparseNsym_StoreSCS',
           'Nag_SparseNsym_SumDups', 'Nag_SparseNsym_TFQMR',
           'Nag_SparseNsym_UnModFact', 'Nag_SparseNsym_UnWeighted',
           'Nag_SparseNsym_UserPiv', 'Nag_SparseNsym_Weighted',
           'Nag_SparseQP', 'Nag_SparseQPP', 'Nag_SparseQPS',
           'Nag_SparseQP_Save_Vars', 'Nag_SparseSym_Bisect',
           'Nag_SparseSym_CG', 'Nag_SparseSym_Check',
           'Nag_SparseSym_FailDups', 'Nag_SparseSym_FailZeros',
           'Nag_SparseSym_InfNorm', 'Nag_SparseSym_JacPrec',
           'Nag_SparseSym_KeepZeros', 'Nag_SparseSym_Lanczos',
           'Nag_SparseSym_MINRES', 'Nag_SparseSym_MarkPiv',
           'Nag_SparseSym_ModFact', 'Nag_SparseSym_NoBisect',
           'Nag_SparseSym_NoCheck', 'Nag_SparseSym_NoPiv',
           'Nag_SparseSym_NoPrec', 'Nag_SparseSym_OneNorm',
           'Nag_SparseSym_Prec', 'Nag_SparseSym_RemoveDups',
           'Nag_SparseSym_RemoveZeros', 'Nag_SparseSym_SSORPrec',
           'Nag_SparseSym_SYMMLQ', 'Nag_SparseSym_SumDups',
           'Nag_SparseSym_TFQMR', 'Nag_SparseSym_TermLanczos',
           'Nag_SparseSym_TermLanczosCG', 'Nag_SparseSym_TwoNorm',
           'Nag_SparseSym_UnModFact', 'Nag_SparseSym_UnWeighted',
           'Nag_SparseSym_UserPiv', 'Nag_SparseSym_Weighted',
           'Nag_Sparse_Colamd', 'Nag_Sparse_Comm',
           'Nag_Sparse_Comm_tag', 'Nag_Sparse_Identity',
           'Nag_Sparse_User', 'Nag_Spline',
           'Nag_SplineVectorSort_Sorted',
           'Nag_SplineVectorSort_Sorted_Indexed',
           'Nag_SplineVectorSort_Sorted_Indexed_Perm',
           'Nag_SplineVectorSort_Unsorted',
           'Nag_SplineVectorSort_Unsorted_Indexed', 'Nag_Sqrt',
           'Nag_SqrtTransform', 'Nag_StandPara', 'Nag_StartNotSet',
           'Nag_StatesExternal', 'Nag_StatesInternal', 'Nag_StoreAll',
           'Nag_StoreFinal', 'Nag_Stress', 'Nag_StudentT',
           'Nag_Subspace', 'Nag_Supplied', 'Nag_T', 'Nag_TRUE',
           'Nag_TableStatsAv', 'Nag_TableStatsLarge',
           'Nag_TableStatsMedian', 'Nag_TableStatsNObs',
           'Nag_TableStatsSmall', 'Nag_TableStatsTotal',
           'Nag_TableStatsVar', 'Nag_TableUpdateI',
           'Nag_TableUpdateU', 'Nag_TabulateVarCont',
           'Nag_TabulateVarDiscr', 'Nag_Tdistn',
           'Nag_TestStatisticsDAbs', 'Nag_TestStatisticsDNeg',
           'Nag_TestStatisticsDPos', 'Nag_Too_Many_Iter', 'Nag_Trans',
           'Nag_TransfOrder', 'Nag_Trend', 'Nag_TridiagEigVecs',
           'Nag_TropicalLargest', 'Nag_TropicalSmallest', 'Nag_Tukey',
           'Nag_TukeyFun', 'Nag_TukeyInterval', 'Nag_TukeyScores',
           'Nag_TwoNorm', 'Nag_TwoTail', 'Nag_TwoTailConfid',
           'Nag_TwoTailSignif', 'Nag_UEV', 'Nag_UH_Controller',
           'Nag_UH_Observer', 'Nag_Uc01St', 'Nag_UcjkSt',
           'Nag_Unbounded', 'Nag_Underflow', 'Nag_Uniform',
           'Nag_UniformMesh', 'Nag_UnitDiag', 'Nag_UnitRoot',
           'Nag_UnitRootWithDrift', 'Nag_UnitRootWithDriftAndTrend',
           'Nag_Unlogged', 'Nag_Unpack', 'Nag_UnscaleRes',
           'Nag_Unsorted', 'Nag_UnweightedEstimate', 'Nag_UpandIn',
           'Nag_UpandOut', 'Nag_Update', 'Nag_UpdateQ',
           'Nag_UpdateSchur', 'Nag_UpdateZ', 'Nag_Upper',
           'Nag_UpperFactored', 'Nag_UpperMatrix', 'Nag_UpperMinus',
           'Nag_UpperSemiInfinite', 'Nag_UpperTail', 'Nag_UseFFT_N',
           'Nag_UseFFT_Y', 'Nag_UseOffset_N', 'Nag_UseOffset_Y',
           'Nag_UseRange', 'Nag_UseSD', 'Nag_Use_Default',
           'Nag_Use_Priority', 'Nag_Use_Supplied', 'Nag_User',
           'Nag_UserInitMesh', 'Nag_UserPrintFun', 'Nag_UserPrior',
           'Nag_UserProb', 'Nag_UserSet', 'Nag_UserVec',
           'Nag_VarScaleRange', 'Nag_VarScaleStd', 'Nag_VarScaleUser',
           'Nag_VarSel_NotSet', 'Nag_VarianceRatio', 'Nag_VgmBessel',
           'Nag_VgmBrownian', 'Nag_VgmCauchy', 'Nag_VgmContParam',
           'Nag_VgmCosine', 'Nag_VgmDifferential',
           'Nag_VgmExponential', 'Nag_VgmGauss', 'Nag_VgmGenHyp',
           'Nag_VgmHole', 'Nag_VgmNugget', 'Nag_VgmSpherical',
           'Nag_VgmSymmStab', 'Nag_VgmWhittleMatern', 'Nag_ViaLookUp',
           'Nag_ViaLookUpOriginal', 'Nag_ViaSimulation',
           'Nag_WaerdenScores', 'Nag_WantLOO', 'Nag_WantVIF',
           'Nag_Warm', 'Nag_Weakmin', 'Nag_WeightObs',
           'Nag_WeightWindow', 'Nag_WeightWindowPos',
           'Nag_WeightedEstimate', 'Nag_WeightedNorm', 'Nag_Weights',
           'Nag_Weightsfreq', 'Nag_Weightsvar',
           'Nag_WholePointSymmetric', 'Nag_WichmannHill_I',
           'Nag_WichmannHill_II', 'Nag_WindowSupplied',
           'Nag_XCheckCon', 'Nag_XCheckObj', 'Nag_XCheckObjCon',
           'Nag_XSimpleCheck', 'Nag_XazSt', 'Nag_ZeroColsNormal',
           'Nag_ZeroColsRestrict', 'Nag_ZeroPadded', 'Nag_ab_prod',
           'Nag_ab_sep', 'Nag_ad02', 'Nag_ad02nm',
           'Nag_ad02ra', 'Nag_ad02xk', 'Nag_ae04mf', 'Nag_ae04nc',
           'Nag_ae04uc', 'Nag_ae04up', 'Nag_ag01em', 'Nag_ag01hb',
           'Nag_bd02', 'Nag_bd02nm', 'Nag_bd02xk',
           'Nag_be04nb', 'Nag_cd02ej', 'Nag_cd02nm',
           'Nag_ce04mf', 'Nag_ce04nc', 'Nag_curr_state', 'Nag_dd02',
           'Nag_dd02nm', 'Nag_de04mf', 'Nag_de04nb',
           'Nag_de04uc', 'Nag_e04nfg_Args', 'Nag_e04nfu', 'Nag_ed02',
           'Nag_ed02nm', 'Nag_ee04mf', 'Nag_ee04nb',
           'Nag_f04ycf_save', 'Nag_fd02', 'Nag_fd02nm',
           'Nag_fe04nb', 'Nag_fe04nc', 'Nag_fe04uc', 'Nag_gd02',
           'Nag_gd02nm', 'Nag_hd02nm',
           'Nag_jd02', 'Nag_jd02nm', 'Nag_kd02',
           'Nag_kd02nm', 'Nag_ld02nm', 'Nag_md02nm', 'Nag_nd02nm',
           'Nag_nd02nn', 'Nag_next_state', 'Nag_pd02nn', 'Nag_rd02nm',
           'Nag_rd02nn', 'Nag_sd02nm', 'Nag_sd02nn', 'Nag_td02nn',
           'Nag_ud02nn', 'Nag_yd02nm', 'Nag_yd02nn', 'Nag_zd02nm',
           'NoTranspose', 'TopPivot', 'Transpose', 'User_Work_t',
           'VariablePivot', '_s_Nag_HashWrap', 'comm_struct',
           'nag_errlist']

__all__.extend(['NAG_D01ESC_F', 'NAG_F02FKC_MONIT', 'NAG_F02FKC_OPTION',
                'NAG_G13EKC_F', 'NAG_G13EKC_H', 'NAG_G13NBC_CONSTFN',
                'NAG_G13NEC_CHGPFN', 'NAG_H02DAC_CONFUN', 'NAG_H02DAC_OBJFUN'])
NAG_E04STC_OBJFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double),
                                 POINTER(c_double), POINTER(Integer),
                                 POINTER(Nag_Comm))
NAG_E04STC_OBJGRD = NAG_FUNCTYPE(None, Integer, POINTER(c_double),
                                 Integer, POINTER(c_double),
                                 POINTER(Integer), POINTER(Nag_Comm))
NAG_E04STC_CONFUN = NAG_FUNCTYPE(None, Integer, POINTER(c_double),
                                 Integer, POINTER(c_double),
                                 POINTER(Integer), POINTER(Nag_Comm))
NAG_E04STC_CONGRD = NAG_FUNCTYPE(None, Integer, POINTER(c_double),
                                 Integer, POINTER(c_double),
                                 POINTER(Integer), POINTER(Nag_Comm))
NAG_E04STC_HESS = NAG_FUNCTYPE(None, Integer, POINTER(c_double),
                               Integer, Integer, c_double,
                               POINTER(c_double), Integer,
                               POINTER(c_double), POINTER(Integer),
                               POINTER(Nag_Comm))
NAG_E04STC_MON = NAG_FUNCTYPE(None, Integer, POINTER(c_double),
                              Integer, POINTER(c_double),
                              POINTER(Integer), POINTER(c_double),
                              POINTER(c_double),
                              POINTER(Nag_Comm))
__all__.append(["NAG_E04STC_OBJFUN",
                "NAG_E04STC_OBJGRD",
                "NAG_E04STC_CONFUN",
                "NAG_E04STC_CONGRD",
                "NAG_E04STC_HESS",
                "NAG_E04STC_MON"])

def nag_char_cast(arg, arg_name):
    """
    Return arg cast to a structure of c_char_ps.
    Raises TypeError if arg is not an array of strings.
    """
    return nag_char_type_check_and_cast('nag_char_cast', arg, arg_name)

__all__.append('nag_char_cast')

def nag_char_type_check(rout_name, arg, arg_name):
    """
    Raises TypeError if arg is not an array of strings.
    """
    if not hasattr(arg, '__len__'):
        raise TypeError(nagFunctionErrorMessage(
            rout_name,
            ['Argument ' + arg_name + ' is not an array (no __len__ attribute).']
        ))
    if len(arg) and not isinstance(arg[0], (str, bytes)):
        raise TypeError(nagFunctionErrorMessage(
            rout_name,
            ['Argument ' + arg_name + ' must be an array of strings.']
        ))

def nag_char_type_check_and_cast(rout_name, arg, arg_name):
    """
    Return arg cast to a structure of STRINGs, or as None.
    Raises TypeError if arg is not an array of strings.
    """
    if arg is None:
        return None
    nag_char_type_check(rout_name, arg, arg_name)
    return (STRING*len(arg))(*arg)

def nag_double_type_check_and_cast(rout_name, arg, arg_name):
    """
    Return arg cast as a pointer to a c_double, or as None.
    Raises TypeError if arg is not an instance of numpy.ndarray.
    Raises TypeError if arg is not a numpy float.
    """
    from numpy import ndarray
    from numpy import float as numpy_float
    if arg is None:
        return None
    if not isinstance(arg, ndarray):
        raise TypeError(nagBadArgInstanceMessage(rout_name, arg_name))
    if arg.dtype != numpy_float:
        raise TypeError(nagBadArgTypeMessage(rout_name, arg_name, numpy_float))
    return arg.ctypes.data_as(POINTER(c_double))

def nag_integer_type_check_and_cast(rout_name, arg, arg_name):
    """
    Return arg cast as a pointer to an integer of the correct size for the
    NAG C Library, or as None.
    Raises TypeError if arg is not an instance of numpy.ndarray.
    Raises TypeError if arg is not an integer of the correct size.
    """
    from numpy import ndarray
    if arg is None:
        return None
    if not isinstance(arg, ndarray):
        raise TypeError(nagBadArgInstanceMessage(rout_name, arg_name))
    if arg.dtype != nag_int_type:
        raise TypeError(nagBadArgTypeMessage(rout_name, arg_name, nag_int_type))
    return arg.ctypes.data_as(POINTER(Integer))

def nag_enum_type_check_and_cast(rout_name, arg, arg_name, enum):
    """
    Return arg cast as a pointer to enum.
    Raises TypeError if arg is not an instance of numpy.ndarray.
    Raises TypeError if arg is not an int32.
    """
    from numpy import ndarray
    if not isinstance(arg, ndarray):
        raise TypeError(nagBadArgInstanceMessage(rout_name, arg_name))
    if arg.dtype != int32:
        raise TypeError(nagBadArgTypeMessage(rout_name, arg_name, enum))
    return arg.ctypes.data_as(POINTER(enum))

def nag_complex_type_check_and_cast(rout_name, arg, arg_name):
    """
    Return arg cast as a pointer to a Complex.
    Raises TypeError if arg is not an instance of numpy.ndarray.
    Raises TypeError if arg is not a Complex.
    """
    from numpy import ndarray
    if not isinstance(arg, ndarray):
        raise TypeError(nagBadArgInstanceMessage(rout_name, arg_name))
    if arg.dtype != [('re', '<f8'), ('im', '<f8')]:
        raise TypeError(nagBadArgTypeMessage(rout_name, arg_name, Complex))
    return arg.ctypes.data_as(POINTER(Complex))

def nag_char2array_type_check_and_cast(rout_name, arg, arg_name, strlen, length):
    """
    Return arg reshaped to a single contiguous string consisting of blocks
    of length strlen, suitable for passing to a C function that expects a
    char[][strlen] argument.
    Raises TypeError if arg is not an array of strings.
    Raises TypeError if len(arg) is smaller than length.
    """
    if arg is None:
        return arg
    from numpy import ndarray
    if not isinstance(arg, ndarray):
        raise TypeError(nagBadArgInstanceMessage(rout_name, arg_name))
    nag_char_type_check(rout_name, arg, arg_name)
    if len(arg) < length:
        raise TypeError(nagFunctionErrorMessage(
            rout_name,
            ['Argument ' + arg_name + ' has length ' + str(len(arg)) +
             ', which is smaller than the minimum expected length of ' +
             str(length) + '.']))
    arg_type = 'S{}'.format(strlen)
    for n, string in enumerate(arg):
        _arg = b''
        if isinstance(string, str):
            string = string.encode("utf-8")
        _arg += string[:strlen].ljust(strlen, b'.')
        arg[n] = _arg
    return arg.astype(arg_type, copy=False).ctypes.data_as(STRING)

def nag_opt_handle_object():
    """
    Return a ctypes.c_void_p object to be used as the handle for the
    optimization modelling suite.
    """
    return c_void_p()

def nag_handle_check(rout_name, arg_name, handle):
    """
    Raises TypeError if handle is not a nag_opt_handle_object.
    No return value.
    """
    if not isinstance(handle, type(nag_opt_handle_object())):
        raise TypeError(nagBadHandleInstanceMessage(rout_name, arg_name))

def quiet_fail():
    "Return a NagError instance with printing disabled."
    fail = NagError()
    fail.eprint = Nag_FALSE
    return fail

def noisy_fail():
    "Return a NagError instance with printing enabled."
    fail = NagError()
    fail.eprint = Nag_TRUE
    return fail

__all__.append("quiet_fail")
__all__.append("noisy_fail")

def get_input_func():
    """
    Return the correct function for expecting input, depending on the
    Python version: that is, raw_input for Python < 3 and input otherwise.
    """
    from sys import version_info
    if version_info < (3, ):
        return raw_input
    else:
        return input

nag_input = get_input_func()
__all__.append("nag_input")
